/*
** Module   : DumpTree.java
** Abstract : print an Aast to a user-defined output stream or to stdout 
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 GES 20041117   @18671 First version based on antlr.DumpASTVisitor.
**                           The formatting was changed to meet GCD
**                           coding standards.  Most importantly, it
**                           now prints token type names instead of the
**                           integer token type itself. Moved the package
**                           to include it in util.  Added javadoc.
** 002 ECF 20050124   @19406 Added option to report position (line and
**                           column numbers) at which underlying token was
**                           found.
** 003 GES 20050126   @19408 Added support for user control over the
**                           output stream.  This allows the caller to
**                           redirect the output to a file or other
**                           destination.
** 004 GES 20050303   @20163 Moved from util to uast package and switched
**                           to using the Aast interface for all dump
**                           activities. Rewrote the code to fix a defect
**                           in the visitor which would walk right side
**                           siblings of the node passed in.  This worked
**                           in the past because we only used it on a
**                           root node which has no such siblings, so the
**                           bug was not seen. There is no code left from
**                           the original antlr.DumpASTVisitor
**                           implementation, it is all GCD now.
** 005 EVL 20070609   @34005 Renaming reserved in Java 6 enum identifier
**                           to be able to compile for Java 6.
** 006 GES 20090424   @41954 Import change.
** 007 GES 20090427   @42017 Moved this class to a new package.
** 008 GES 20090518   @42371 Import change.
** 009 CA  20131023          Added debug aid to dump annotation info.
** 010 OM  20150210          Added support for settable annotation output.
** 011 EVL 20160222          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 012 CA  20170825          Emit the node ID too.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast; 
 
import java.io.*;
import java.util.*;
import com.goldencode.util.*;

/**
 * Dumps the contents of an <code>AST</code> to <code>stdout</code> or to
 * a user-defined output stream. 
 * <p>
 * Each node prints the symbolic text name of the token type rather than the
 * integer representation. The token text, line and column number information
 * for the AST will also be rendered.
 */
public class DumpTree
{
   /** Default value for dump annotations information. */
   private static final boolean DEFAULT_DUMP_ANNOTATIONS = false;
   
   /** The size of indentation in spaces.*/
   private static final int TAB_SIZE = 3;
   
   /** The destination of all dumped output. */
   private PrintStream out = null;
   
   /** Debug aid to dump annotations information. */
   private boolean dumpAnnotations = DEFAULT_DUMP_ANNOTATIONS;

   /**
    * Default constructor which uses <code>stdout</code>.
    */
   public DumpTree()
   {
      this(null);
   }   
   
   /**
    * Constructs an instance and allows control over the output destination.
    * If the output stream is <code>null</code> it will be set to <code>System.out</code>.
    *
    * @param    out
    *           The new stream to use as the output destination.
    */
   public DumpTree(PrintStream out)
   {
      setOutputStream(out);
   }
   
   /**
    * Allows the current output stream to be switched for a user-defined
    * stream.  If set to <code>null</code>, <code>System.out</code> will
    * be used.  The caller must close this stream when done!
    *
    * @param    out
    *           The new stream to use as the output destination.
    */
   public void setOutputStream(PrintStream out)
   {
      this.out = (out != null) ? out : System.out;
   }
   
   /**
    * Prints a 3 space indent for every additional nested level in the tree.
    *
    * @param    tabs
    *           The numbers of TAB to print.
    */
   private void printIndent(int tabs)
   {
      if (tabs > 0)
      {
         out.print(StringHelper.repeatChar(' ', tabs * TAB_SIZE));
      }
   }
   
   /**
    * Prints out a text-based representation of an <code>Aast</code> and
    * all children.  This method relies upon the fact that all tree nodes 
    * represent a full tree themselves, even if they have no children. An
    * iterator is obtained that allows a simple depth-first, left-right
    * walk of the tree in a single loop.
    * 
    * @param   node
    *          The root of the tree or subtree which should be printed.
    */
   public void visit(Aast node)
   {
      Iterator iter = node.iterator();
      int      root = node.getDepth();
      
      while (iter.hasNext())
      {
         Aast next = (Aast) iter.next();
         printIndent(next.getDepth() - root);
         
         // print the token type as a name and token position
         out.println(next.getText() + " [" + next.getSymbolicTokenType() + "]:" + next.getId() + 
                     " @" + next.getLine() + ':' + next.getColumn());
         
         // output annotations (only for debugging)
         Iterator<String> itra = next.annotationKeys();
         if (dumpAnnotations && itra.hasNext())
         {
            printIndent(next.getDepth() - root);
            StringBuilder sb = new StringBuilder("(");
            while (itra.hasNext())
            {
               String key = itra.next();
               sb.append(key).append("=").append(next.getAnnotation(key));
               if (itra.hasNext())
               {
                  sb.append(", ");
               }
            }
            sb.append(")");
            
            out.println(sb.toString());
         }
      }
   }
   
   /**
    * Enable/disable the output of annotations for each visited node.
    * 
    * @param   dumpAnnotations
    *          if {@code true} then the annotations for each node will be printed in the output
    *          stream 
    */
   public void setDumpAnnotations(boolean dumpAnnotations)
   {
      this.dumpAnnotations = dumpAnnotations;
   }
}