/*
** Module   : InMemoryRegistryPlugin.java
** Abstract : implements an in-memory AST storage option for an AstManagerPlugin
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20131023 Created first version.
** 002 ECF 20140919 Fixed memory leaks, added usesMemoryStorage method, to help with memory
**                  management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import java.util.*;

/**
 * Defines an in-memory
 */
public class InMemoryRegistryPlugin
extends BasePlugin
{
   /** In-memory storage of ASTs, by their name. */
   private final Map<String, Aast> asts = new HashMap<String, Aast>();
   
   /**
    * Does this plug-in use memory as its primary storage medium? This implementation overrides
    * the superclass' implementation to return <code>true</code>.
    * 
    * @return  <code>true</code>.
    */
   public boolean usesMemoryStorage()
   {
      return true;
   }
   
   /**
    * Load a named AST from the in-memory storage, by returning a copy of the {@link #asts}
    * entry after using the specified <code>name</code> as the key.
    *
    * @param    name
    *           The name of the tree. Often this is a file name.
    *
    * @return   The loaded AST.
    *
    * @throws   AstException
    *           On any error during the loading process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource doesn't exist.
    */
   @Override
   public Aast loadTree(String name)
   throws AstException,
          IllegalArgumentException
   {
      Aast ast = asts.get(name);
      
      Aast clone = ast.duplicate();
      clone.removeAnnotation("terse");
      
      if (ast.isAnnotation("terse") && (Boolean) ast.getAnnotation("terse"))
      {
         terseTree(clone);
      }
      
      return clone;
   }

   /**
    * Save the given AST to in-memory storage under the given name. This will save a copy of the
    * specified AST in the {@link #asts} map, usin the specified name as the key.
    *
    * @param    ast
    *           The tree to save.
    * @param    name
    *           The name of the tree. Often this is a file name.
    * @param    terse
    *           Do not store line and column data when the node data is saved.
    * 
    * @throws   AstException
    *           On any error during the saving process.
    * @throws   IllegalArgumentException
    *           If the name is invalid or the specified resource cannot be
    *           used as a target to store the AST.
    */
   @Override
   public void saveTree(Aast ast, String name, boolean terse)
   throws AstException,
          IllegalArgumentException
   {
      addTree(name);
      
      Aast clone = ast.duplicate();
      
      clone.putAnnotation("terse", terse);
      if (terse)
      {
         terseTree(clone);
      }
      
      asts.put(name, clone);
   }

   /**
    * A no-op for this in-memory storage.
    */
   @Override
   public void save()
   throws AstException
   {
      // no-op
   }
   
   /**
    * Removes this tree from the in-memory storage.
    *
    * @param    id
    *           The ID the tree to remove.
    *
    * @return   <code>true</code> if the tree existed in the registry (and 
    *           was removed), <code>false</code> if the tree didn't exist.
    *           
    * @see      BasePlugin#removeTree(long)
    */
   @Override
   public boolean removeTree(long id)
   throws AstException
   {
      String name = getTreeName(id);
      asts.remove(name);

      return super.removeTree(id);
   }

   /**
    * Check if the AST specified by the given name exists already in the plugin's storage.
    * 
    * @param    name
    *           The name (including path) of the AST file to be checked.
    * 
    * @return   <code>true</code> if the specified AST is in the plugin's storage.
    */
   public boolean isExistingAst(String name)
   {
      return asts.containsKey(name);
   }
   
   /**
    * Set the line and column information to 0 for the entire tree.
    * 
    * @param    ast
    *           The AST which needs the column/line information to be removed.
    */
   private void terseTree(Aast ast)
   {
      ast.setLine(0);
      ast.setColumn(0);
      
      Iterator<Aast> iter = ast.iterator();
      while (iter.hasNext())
      {
         ast = iter.next();
         ast.setLine(0);
         ast.setColumn(0);
      }
   }
}