/*
** Module   : LexerDumpHelper.java
** Abstract : writes lexer output to a stream
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20070821   @34947 Based on prior work. Writes a text 
**                           representation of a lexer's token stream
**                           by driving the lexer and for each token:
**                           - decoding token types
**                           - writing output with line numbers, token
**                             types and text
** 002 GES 20090424   @41971 Import change.
** 003 GES 20090428   @42019 Package and class name change. Added an alternate 
**                           constructor.
** 004 EVL 20160225          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import java.io.*;
import java.text.*;
import antlr.*;
import com.goldencode.util.*;

/**
 * Provides a simple mechanism to run a given lexer on and to print the
 * resulting stream of tokens.  The output is a column formatted list with
 * one line printed for each token.
 * <p>
 * The first column is formatted as "[" line number ":" column number "]".
 * <p> 
 * The second column is the symbolic representation of the integer token
 * type of that token.  The third column is the actual text from the file
 * that corresponds to that token.
 */
public class LexerDumpHelper
{
   /** Formats a line number padded on left with zeros. */
   private NumberFormat lfmt = NumberFormat.getInstance();
   
   /** Formats a column number padded on left with zeros. */   
   private NumberFormat cfmt = NumberFormat.getInstance();
   
   /** Size in characters of the largest possible token type name. */ 
   private int maxSymSize = 0;
      
   /** An instance of the lexer object. */  
   private TokenStream lexer = null;
   
   /** Maps token types (as indexes) to token names. */
   private String[] names = null;
   
   /** Output destination. */
   private PrintWriter out = null;
   
   /**
    * Constructs an instance of the test driver using a specific instance
    * of the lexer.  This will initialize our formatting objects to a 
    * known state.
    *
    * @param    lexer
    *           Instance of the lexer to use for tokenizing the input file.
    * @param    names
    *           The array of token names indexed by token type.
    * @param    out
    *           Destination for output.
    */
   public LexerDumpHelper(TokenStream lexer, String[] names, PrintWriter out)   
   {
      this.lexer = lexer;
      this.names = names;
      this.out   = out;
      
      lfmt.setMinimumIntegerDigits(5);
      lfmt.setGroupingUsed(false);
      cfmt.setMinimumIntegerDigits(3);
      cfmt.setGroupingUsed(false);
      
      calcMaximumSymbolNameSize();
   }
      
   /**
    * Constructs an instance of the test driver using a specific instance
    * of the lexer.  This will initialize our formatting objects to a 
    * known state.
    *
    * @param    lexer
    *           Instance of the lexer to use for tokenizing the input file.
    * @param    names
    *           The array of token names indexed by token type.
    * @param    out
    *           Destination for output.
    */
   public LexerDumpHelper(TokenStream lexer, String[] names, PrintStream out)   
   {
      this(lexer, names, new PrintWriter(out, true));
   }
      
   /**
    * Scans the list of token type names (text symbols that represent each 
    * integer token type) and calculates the width in characters of the
    * largest one.  This is saved as the column width in the output report.
    */
   private void calcMaximumSymbolNameSize()
   {
      for (int i = 0; i < names.length; i++)
      {
         int current = names[i].length();
         
         if (current > maxSymSize)
            maxSymSize = current;
      }
      
      // make room for the angle brackets and an extra space for padding
      maxSymSize += 3;
   }
   
   /**   
    * Prints a report with one line per token in the token stream. 
    * 
    * @throws   antlr.TokenStreamException 
    *           If an respective error occurs. 
    * @throws   IOException 
    *           If an respective error occurs. 
    */
   public void dump()
   throws antlr.TokenStreamException,
          IOException
   {
      CommonToken next;
      
      do
      {
         next = (CommonToken) lexer.nextToken();
         printToken(next);
      } 
      while (next.getType() != Token.EOF_TYPE);
   }
   
   /**
    * Prints a one line record for the token object passed as a parameter.
    * The record is printed in a 3 column format as documented in the
    * overview section.
    *
    * @param    next
    *           The token to print.
    * 
    * @throws   IOException 
    *           If an respective error occurs. 
    */
   private void printToken(CommonToken next)
   throws IOException
   {
      String decode = names[next.getType()];
      
      out.println("["                                                  +
                  lfmt.format( next.getLine() )                        +
                  ":"                                                  +
                  cfmt.format( next.getColumn() )                      +
                  "] <"                                                +
                  StringHelper.leftAlignText(decode + ">", maxSymSize) +
                  " "                                                  +
                  next.getText());
   }
}