/*
** Module   : EmailDefinition.java
** Abstract : stores email configuration and data
**                       
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20171020 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.email;

import java.io.*;
import java.util.*;

/**
 * This is a container class that stores the configuration and data needed to send an email.
 */
public class EmailDefinition
implements Externalizable
{
   /** Type of connection security. */ 
   public ConnectionSecurity connectType = ConnectionSecurity.UNENCRYPTED;
   
   /** Host name or IP address of the SMTP server. Must not be <code>null</code>. */ 
   public String host = null;
   
   /** TCP port for the SMTP server. If <code>null</code>, the default port will be used. */ 
   public String port = null;
   
   /**
    * Userid for authentication. If both userid and password are <code>null</code>,
    * authentication will be disabled.
    */ 
   public String userid = null;
   
   /**
    * Password for authentication. If both userid and password are <code>null</code>,
    * authentication will be disabled.
    */ 
   public String password = null;
   
   /**
    * Flag to determine if the server's SSL certificate is checked for validity. This is only
    * used for SSL and STARTTLS connection types. If this is <code>false</code>, the hostname
    * of the SMTP server will be defined as trusted AND validity checking will be disabled.
    */
   public boolean validate = true;
   
   /** Flag to enable debug output to the console. */
   public boolean debug = false;
   
   /** FROM address. */
   public String from = null;
         
   /** REPLY-TO address. */
   public String replyTo = null;
         
   /** Subject line. */
   public String subject = null;
   
   /** Plain text body content. */
   public String text = null;
   
   /** HTML body content. */
   public String html = null;
   
   /** List of TO addresses. */
   private List<String> to = null;
   
   /** List of CC addresses. */
   private List<String> cc = null;
   
   /** List of BCC addresses. */
   private List<String> bcc = null;
   
   /** List of attachment descriptors. */
   private List<AttachmentDescriptor> attachments = null;
   
   /** List of embedded image descriptors. */
   private List<EmbedDescriptor> embeds = null;
   
   /**
    * Default constructor.
    */
   public EmailDefinition()
   {
   }
   
   /**
    * Add the given address to the list of TO addresses.
    *
    * @param    addr
    *           The TO address to add.
    */
   public synchronized void addTo(String addr)
   {
      if (addr == null)
      {
         // do nothing
         return;
      }
      
      if (to == null)
      {
         to = new ArrayList<String>();
      }
      
      to.add(addr);
   }
   
   /**
    * Add the given addresses to the list of TO addresses.
    *
    * @param    addrs
    *           The list of TO addresses to add.
    */
   public synchronized void addTo(String[] addrs)
   {
      if (addrs == null || addrs.length == 0)
      {
         // do nothing
         return;
      }
      
      for (String addr : addrs)
      {
         addTo(addr);
      }
   }
   
   /**
    * Obtain the current list of all TO addresses.
    *
    * @return   The list of TO addresses.
    */
   public synchronized String[] getToList()
   {
      if (to == null)
      {
         return new String[0];
      }
      
      return to.toArray(new String[to.size()]);
   }
   
   /**
    * Clear the current list of all TO addresses.
    */
   public synchronized void clearToList()
   {
      if (to != null)
      {
         to = null;
      }
   }
   
   /**
    * Add the given address to the list of CC addresses.
    *
    * @param    addr
    *           The CC address to add.
    */
   public synchronized void addCc(String addr)
   {
      if (addr == null)
      {
         // do nothing
         return;
      }
      
      if (cc == null)
      {
         cc = new ArrayList<String>();
      }
      
      cc.add(addr);
   }
   
   /**
    * Add the given addresses to the list of CC addresses.
    *
    * @param    addrs
    *           The list of CC addresses to add.
    */
   public synchronized void addCc(String[] addrs)
   {
      if (addrs == null || addrs.length == 0)
      {
         // do nothing
         return;
      }
      
      for (String addr : addrs)
      {
         addCc(addr);
      }
   }
   
   /**
    * Obtain the current list of all CC addresses.
    *
    * @return   The list of CC addresses.
    */
   public synchronized String[] getCcList()
   {
      if (cc == null)
      {
         return new String[0];
      }
      
      return cc.toArray(new String[cc.size()]);
   }
   
   /**
    * Clear the current list of all CC addresses.
    */
   public synchronized void clearCcList()
   {
      if (cc != null)
      {
         cc = null;
      }
   }
   
   /**
    * Add the given address to the list of BCC addresses.
    *
    * @param    addr
    *           The BCC address to add.
    */
   public synchronized void addBcc(String addr)
   {
      if (addr == null || addr.length() == 0)
      {
         // do nothing
         return;
      }
      
      if (bcc == null)
      {
         bcc = new ArrayList<String>();
      }
      
      bcc.add(addr);
   }
   
   /**
    * Add the given addresses to the list of BCC addresses.
    *
    * @param    addrs
    *           The list of BCC addresses to add.
    */
   public synchronized void addBcc(String[] addrs)
   {
      if (addrs == null || addrs.length == 0)
      {
         // do nothing
         return;
      }
      
      for (String addr : addrs)
      {
         addBcc(addr);
      }
   }
   
   /**
    * Obtain the current list of all BCC addresses.
    *
    * @return   The list of BCC addresses.
    */
   public synchronized String[] getBccList()
   {
      if (bcc == null)
      {
         return new String[0];
      }
      
      return bcc.toArray(new String[bcc.size()]);
   }
   
   /**
    * Clear the current list of all BCC addresses.
    */
   public synchronized void clearBccList()
   {
      if (bcc != null)
      {
         bcc = null;
      }
   }
   
   /**
    * Add the given filename to the list of attachments.
    *
    * @param    filename
    *           The filename of the resource to attach.
    */
   public synchronized void attachFile(String filename)
   {
      if (filename == null || filename.length() == 0)
      {
         // do nothing
         return;
      }
      
      if (attachments == null)
      {
         attachments = new ArrayList<AttachmentDescriptor>();
      }
      
      attachments.add(new AttachmentDescriptor(ResourceType.PATH, filename));
   }
   
   /**
    * Add the given URL to the list of attachments.
    *
    * @param    url
    *           The URL of the resource to attach.
    */
   public synchronized void attachURL(String url)
   {
      if (url == null || url.length() == 0)
      {
         // do nothing
         return;
      }
      
      if (attachments == null)
      {
         attachments = new ArrayList<AttachmentDescriptor>();
      }
      
      attachments.add(new AttachmentDescriptor(ResourceType.URL, url));
   }
   
   /**
    * Obtain the current list of all attachments.
    *
    * @return   The list of attachments.
    */
   public synchronized AttachmentDescriptor[] getAttachmentsList()
   {
      if (attachments == null)
      {
         return new AttachmentDescriptor[0];
      }
      
      return attachments.toArray(new AttachmentDescriptor[attachments.size()]);
   }
   
   /**
    * Clear the current list of all attachments.
    */
   public synchronized void clearAttachmentsList()
   {
      if (attachments != null)
      {
         attachments = null;
      }
   }
   
   /**
    * Add the given filename to the list of embedded images.
    *
    * @param    filename
    *           The filename of the resource to attach.
    *
    * @return   Content ID to use in the HTML IMG tag or <code>null</code> if there was an error.
    */
   public synchronized String embedFile(String filename)
   {
      if (filename == null || filename.length() == 0)
      {
         // do nothing
         return null;
      }
      
      if (embeds == null)
      {
         embeds = new ArrayList<EmbedDescriptor>();
      }
      
      EmbedDescriptor descr = new EmbedDescriptor(ResourceType.PATH, filename, null);
      
      embeds.add(descr);
      
      return descr.cid;
   }
   
   /**
    * Add the given URL to the list of embedded images.
    *
    * @param    url
    *           The URL of the resource to attach.
    * @param    name
    *           The name of the resource.  This may not be <code>null</code> and it cannot be
    *           an empty string.
    *
    * @return   Content ID to use in the HTML IMG tag or <code>null</code> if there was an error.
    */
   public synchronized String embedURL(String url, String name)
   {
      if (url == null || url.length() == 0 || name == null || name.length() == 0)
      {
         // do nothing
         return null;
      }
      
      if (embeds == null)
      {
         embeds = new ArrayList<EmbedDescriptor>();
      }
      
      EmbedDescriptor descr = new EmbedDescriptor(ResourceType.URL, url, name);
      
      embeds.add(descr);
      
      return descr.cid;
   }
   
   /**
    * Obtain the current list of all embedded images.
    *
    * @return   The list of embedded images.
    */
   public synchronized EmbedDescriptor[] getEmbedsList()
   {
      if (embeds == null)
      {
         return new EmbedDescriptor[0];
      }
      
      return embeds.toArray(new EmbedDescriptor[embeds.size()]);
   }
   
   /**
    * Clear the current list of all embedded images.
    */
   public synchronized void clearEmbedsList()
   {
      if (embeds != null)
      {
         embeds = null;
      }
   }
   
   /**
    * Replacement for the default object reading method. The latest state is read from the input
    * source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      connectType     = ConnectionSecurity.valueOf((String) in.readObject());
      host            = (String) in.readObject();
      port            = (String) in.readObject();
      userid          = (String) in.readObject();
      password        = (String) in.readObject();
      validate        = in.readBoolean();
      debug           = in.readBoolean();
      from            = (String) in.readObject();
      replyTo         = (String) in.readObject();
      subject         = (String) in.readObject();
      text            = (String) in.readObject();
      html            = (String) in.readObject();
      addTo((String[]) in.readObject());
      addCc((String[]) in.readObject());
      addBcc((String[]) in.readObject());
      
      AttachmentDescriptor[] atts = (AttachmentDescriptor[]) in.readObject();
      
      if (atts.length > 0)
      {
         attachments = new ArrayList<AttachmentDescriptor>(Arrays.asList(atts));
      }
      
      EmbedDescriptor[] ems = (EmbedDescriptor[]) in.readObject();
      
      if (ems.length > 0)
      {
         embeds = new ArrayList<EmbedDescriptor>(Arrays.asList(ems));
      }
   }

   /**
    * Replacement for the default object writing method. The latest state is written to the
    * output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(connectType.toString());
      out.writeObject(host);
      out.writeObject(port);
      out.writeObject(userid);
      out.writeObject(password);
      out.writeBoolean(validate);
      out.writeBoolean(debug);
      out.writeObject(from);
      out.writeObject(replyTo);
      out.writeObject(subject);
      out.writeObject(text);
      out.writeObject(html);
      out.writeObject(getToList());
      out.writeObject(getCcList());
      out.writeObject(getBccList());
      out.writeObject(getAttachmentsList());        
      out.writeObject(getEmbedsList());
   }   
}