/*
** Module   : BytecodeContainer.java
** Abstract : Bytecode support object which can nest instances of itself
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @19644 Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 002 ECF 20050207 CHG   @19645 Updated to conform with coding standards.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.util.*;

/**
 * This class represents any object which contains bytecode, such as a
 * bytecode instruction or a code unit containing multiple instructions.
 * Each instance of this class has a position (offset) within the overall
 * bytecode for a method, as well as a length in bytes.  When the bytecode
 * for a method is compiled, each container is asked to layout itself at
 * a certain offset.  This offset must be available to other containers
 * which may need to branch to the first instruction in this container.
 * Each container must also be aware of its impact on the maximum stack
 * depth of the method of which it is a part.
 */
abstract class BytecodeContainer
implements WritableData,
           CompilerConstants
{
   /** Byte position of the beginning of this container. */
   protected int offset = 0;
   
   /** Number of bytes which this container consumes. */
   protected int length = 0;
   
   /**
    * Return the offset of the first instruction in this container.
    *
    * @return  Beginning position of this container.
    */
   int getOffset()
   {
      return offset;
   }
   
   /**
    * Return the length, in bytes, of this code container.
    *
    * @return  Number of bytes contained in this object.
    */
   int getLength()
   {
      return length;
   }
   
   /**
    * This method must be implemented to update the stack depth of the
    * current bytecode array, based upon the impact of the instruction(s)
    * which compose that bytecode array.  The expression compiler maintains
    * and ultimately records the deepest depth which the stack may achieve
    * while running the resulting bytecode in the method currently being
    * compiled.
    *
    * @param   compiler
    *          Expression compiler which maintains stack depth information.
    */
   abstract void modifyStackDepth(Compiler compiler);
   
   /**
    * This method must be implemented to assign the specified offset to
    * this bytecode container, and to return the offset pointing to the
    * beginning of the next available position in the bytecode array.
    *
    * @param   offset
    *          Integer offset (position) of the beginning bytecode in this
    *          bytecode container.
    *
    * @return  The result of <tt>offset</tt> plus <tt>length</tt>
    *          of this bytecode container.
    */
   abstract int layout(int offset);
   
   /**
    * This method must be implemented to pop away any superfluous runtime
    * stack entries in the event the compiled code must branch control.
    *
    * @param   stack
    *          Stack object which represents the state the runtime stack
    *          will be in at the time the instructions in this bytecode
    *          container are executed.
    */
   abstract void fixupStack(Stack stack);
   
   /**
    * This method may be overridden if the subclass needs to perform any
    * special processing to fix up its branch instructions (if any) so
    * that their arguments will branch to the proper offset in the bytecode
    * array.  This method will be called on each bytecode container in the
    * final bytecode array for a method, in order, after the final offsets
    * have been assigned to each container.
    */
   public void fixupBranchOffset()
   {
   }
}