/*
** Module   : CompiledExpression.java
** Abstract : Base class of all custom, compiled expressions
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @21346 Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 002 ECF 20050527 CHG   @21347 Expression engine rewrite. Removed symbol
**                               resolver instance members; added returnType,
**                               libraries, and variables members and access
**                               methods. Added testRound method for floating
**                               point rounding ops.
** 003 ECF 20050604 CHG   @21513 Enhanced to allow for more detailed error
**                               reporting during runtime exception handling.
**                               This involves storing the expression's AST
**                               when compiling with error/debug diagnostics,
**                               and storing the ID of each active AST node as
**                               the expression executes.
** 004 ECF 20071121 CHG   @39254 Added clone(SymbolResolver) method. Allows an
**                               instance of this class to be copied.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

/**
 * This is the base class for all custom, compiled expression classes. It
 * provides an abstract {@link #execute} method. This method contains the
 * compiled bytecode instructions which define the essential operation of
 * the expression; it returns the overall result of execution.
 * <p>
 * Each custom subclass defines a specific return type which its
 * implementation of the <code>execute</code> method will return. This is
 * accessed through the {@link #getReturnType} method. Likewise, each custom
 * subclass may use zero or more libraries and zero or more variables. These
 * are stored at compile time via the {@link #setLibraries} and {@link
 * #setVariables} methods. Though all three of these attributes would better
 * be stored as static members of the custom subclasses, a design decision
 * was made to store them as instance variables in the superclass instead.
 * This is deemed acceptable as a trade-off for the much more complicated
 * implementation which would have been required to do this dynamically.
 * Also, because the {@link Expression} wrapper class manages a cache of
 * subclass instances, in practice no more than one instance of a custom
 * subclass exists at a time, such that these custom implementations are
 * essentially singletons.
 */
public abstract class CompiledExpression
implements Cloneable
{
   /** Object type returned by concrete implementation of execute method */
   private Class<?> returnType = null;
   
   /** Objects upon which user functions are invoked by the expression */
   private Object[] libraries = null;
   
   /** Runtime variables accessed by the expression */
   private Variable[] variables = null;
   
   /** AST associated with underlying expression */
   private ExtraAst ast = null;
   
   /** ID of AST corresponding with most recently active code unit */
   private int activeAstId = -1;
   
   /**
    * This method is invoked when rounding floating point calculation results
    * to a fixed number of decimal places. The rounding algorithm multiplies
    * the operation result by the rounding constant defined for the current
    * expression, rounds the resulting double to a long, then converts the
    * long to a double, and divides by the rounding constant. The rounding
    * constant is a power of 10, where the exponent is the number of decimal
    * places of precision defined for the expression. Thus, the result of
    * the first multiplication can become quite large. In cases where that
    * number is a larger positive than the maximum, 64 bit long value (or a
    * larger negative than the minimum long value), the result of the round
    * operation will be inaccurate.
    * <p>
    * The purpose of this method is to test for these conditions and to
    * throw an exception in the event one of them occurs. When this happens,
    * it will be necessary for the expression's precision to be reduced to
    * the point where the rounding operation will be meaningful.
    *
    * @param   num
    *          The intermediate numeric value to test against the boundary
    *          conditions described above.
    *
    * @return  <code>num</code>. This value is returned for convenience in
    *          assembling the bytecode for this test, in that the runtime
    *          stack is left in the same condition when this method returns
    *          successfully.
    *
    * @throws  IllegalArgumentException
    *          if <code>num</code> exceeds the range which can be represented
    *          by a 64 bit <code>long</code> value.
    */
   protected static double testRound(double num)
   {
      if ((num + 0.5) > Long.MAX_VALUE || (num - 0.5) < Long.MIN_VALUE)
      {
         throw new IllegalArgumentException(
            "Floating point rounding precision too high");
      }
      
      return num;
   }
   
   /**
    * Create a copy of this object which uses the same libraries and variables
    * as the original.
    * 
    * @param   resolver
    *          Symbol resolver.
    * 
    * @return  Copy of this object.
    */
   protected final CompiledExpression clone(SymbolResolver resolver)
   {
      CompiledExpression copy = null;
      
      try
      {
         copy = (CompiledExpression) super.clone();
      }
      catch (CloneNotSupportedException exc)
      {
         throw new ExpressionException(exc);
      }
         
      Scope scope = resolver.getCurrentScope();
      int len;
      
      len = libraries.length;
      for (int i = 0; i < len; i++)
      {
         Object lib = libraries[i];
         Class<?> cls = lib.getClass();
         lib = resolver.getLibraryTarget(scope, cls);
         assert (lib != null);
         copy.libraries[i] = lib;
      }
      
      len = variables.length;
      for (int i = 0; i < len; i++)
      {
         Variable var = variables[i];
         String name = var.getName();
         var = resolver.getVariable(scope, name);
         assert (var != null);
         copy.variables[i] = var;
      }
      
      return copy;
   }
   
   /**
    * Get the library object stored at the specified index. This method is
    * called from within the <code>execute</code> method of compiled
    * expressions; it is not intended to be called directly by client code.
    *
    * @param   index
    *          Index of the desired library object; this is known at compile
    *          time and is hard-coded directly into the bytecode instructions
    *          of a subclass' <code>execute</code> method.
    *
    * @return  A library object whose methods are callable by compiled
    *          expression implementations.
    *
    * @see     #setLibraries
    */
   protected Object getLib(int index)
   {
      return libraries[index];
   }
   
   /**
    * Get the <em>value of</em> the user variable stored at the specified
    * index. This method is called from within the <code>execute</code>
    * method of compiled expressions; it is not intended to be called
    * directly by client code.
    *
    * @param   index
    *          Index of the desired user variable; this is known at compile
    *          time and is hard-coded directly into the bytecode instructions
    *          of a subclass' <code>execute</code> method.
    *
    * @return  The value of a {@link Variable user variable} object, the
    *          methods of which can be accessed by compiled expression
    *          implementations.
    *
    * @see     #setVar
    * @see     #setVariables
    */
   protected Object getVar(int index)
   {
      return variables[index].getValue();
   }
   
   /**
    * Set the <em>value of</em> the user variable stored at the specified
    * index. This method is called from within the <code>execute</code>
    * method of compiled expressions; it is not intended to be called
    * directly by client code.
    * <p>
    * This method can only be compiled into an expression if the target
    * user variable is not {@link SymbolResolver#registerVariable registered}
    * as read-only.
    *
    * @param   index
    *          Index of the target user variable; this is known at compile
    *          time and is hard-coded directly into the bytecode instructions
    *          of a subclass' <code>execute</code> method.
    * @param   value
    *          The value to store within the {@link Variable user variable}
    *          object located at <code>index</code> in the array of user
    *          variables which this expression accesses. The type of this
    *          parameter must be compatible with the type of the variable.
    *
    * @see     #getVar
    * @see     #setVariables
    */
   protected void setVar(int index, Object value)
   {
      variables[index].setValue(value);
   }
   
   /**
    * Store the ID of the AST which corresponds with the code unit which is
    * about to be executed. This is called by certain compiled expression
    * instances at runtime at key execution points; namely, just before any
    * code unit's instructions are executed for code units which invoke
    * method calls which throw an exception (or cause an exception to be
    * thrown in the immediately following instructions). This effectively
    * associates a set of runtime instructions with a specific segment of
    * the expression being executed. The ID can be retrieved during
    * exception handling at runtime, allowing an error message to be
    * generated which pinpoints the area of failure within an expression.
    * <p>
    * This method is only invoked by expressions which have been compiled
    * with one (or more) of the following runtime flags set:
    * <ul>
    * <li>{@link Expression#RT_ERROR}
    * <li>{@link Expression#RT_DEBUG}
    * </ul>
    *
    * @param   activeAstId
    *          ID of an AST node from the postfix representation of the
    *          expression which was used to generate this object's class.
    *          This ID will be unique within the expression's syntax tree.
    *
    * @see     #getActiveAstId
    */
   protected void setActiveAstId(int activeAstId)
   {
      this.activeAstId = activeAstId;
   }
   
   /**
    * Return the AST associated with this compiled expression. This will
    * only be non-<code>null</code> for expressions compiled with runtime
    * flags of {@link Expression#RT_ERROR} or {@link Expression#RT_DEBUG}.
    *
    * @return  AST associated with this compiled expression.
    */
   ExtraAst getAst()
   {
      return ast;
   }
   
   /**
    * Set the AST associated with this compiled expression. This will
    * only be called for expressions compiled with runtime flags of {@link
    * Expression#RT_ERROR} or {@link Expression#RT_DEBUG}.
    *
    * @param   ast
    *          AST associated with this compiled expression.
    */
   void setAst(ExtraAst ast)
   {
      this.ast = ast;
   }
   
   /**
    * Reports the type of object which the {@link #execute} method of this
    * object returns.
    *
    * @return  Class of the object returned by the <code>execute</code>
    *          method.
    */
   Class<?> getReturnType()
   {
      return returnType;
   }
   
   /**
    * Set the type of object which the {@link #execute} method of this object
    * is expected to return. This method is called directly by the {@link
    * Compiler} during expression compilation and should not be called by
    * any other code.
    *
    * @param   returnType
    *          Class of the object returned by the <code>execute</code>
    *          method.
    */
   void setReturnType(Class<?> returnType)
   {
      this.returnType = returnType;
   }
   
   /**
    * Called by the {@link Compiler} to set the library objects upon which
    * this expression will invoke methods at expression execution time.
    * This method should not be called by any other code. Bytecode
    * instructions are hard-coded into the expression's <code>execute</code>
    * method to access libraries at fixed indices in this array.
    *
    * @param   libraries
    *          The array of libraries to be accessed by the expression at
    *          execution time. This array will have a length of zero if the
    *          expression references no library objects, but it will not be
    *          <code>null</code>.
    */
   void setLibraries(Object[] libraries)
   {
      this.libraries = libraries;
   }
   
   /**
    * Called by the {@link Compiler} to set the user variables upon which
    * this expression will invoke methods at expression execution time.
    * This method should not be called by any other code. Bytecode
    * instructions are hard-coded into the expression's <code>execute</code>
    * method to access user variables at fixed indices in this array.
    *
    * @param   variables
    *          The array of variables to be accessed by the expression at
    *          execution time. This array will have a length of zero if the
    *          expression references no user variables, but it will not be
    *          <code>null</code>.
    */
   void setVariables(Variable[] variables)
   {
      this.variables = variables;
   }
   
   /**
    * Get the ID of the AST node which corresponds with the code unit which
    * was last executing during a run of this expression's {@link #execute}
    * method. This ID is set at runtime by certain compiled expression
    * instances, to enable better exception handling and error reporting.
    *
    * @return  ID of an AST node from the postfix representation of the
    *          expression which was used to generate this object's class.
    *          This ID will be unique within the expression's syntax tree.
    *          <code>-1</code> is returned for expressions which were not
    *          compiled with error or debug diagnostics turned on.
    *
    * @see     #setActiveAstId
    */
   int getActiveAstId()
   {
      return activeAstId;
   }
   
   /**
    * Each custom, concrete implementation of this abstract class will
    * implement this method to perform the operations described by the
    * underlying expression. The concrete implementations of this method
    * are generated by the {@link Compiler} in bytecode form only (no
    * associated source code is generated). The basis for these instructions
    * is the infix notation expression string provided by the instance of
    * the {@link Expression} class which is associated with an instance of
    * this class.
    * <p>
    * Different concrete implementations of this method will return different
    * objects, depending upon the structure of the underlying expression.
    * The exact data type returned by this method is determined by the {@link
    * #getReturnType} method. Most expressions return an object reference
    * (or <code>null</code>). Assignment expressions and expressions whose
    * top level operation is a method call with a <code>void</code> return
    * type always return <code>null</code> from this method. If an expression
    * is written to return a primitive data type, that data will be wrapped
    * in the appropriate wrapper object before it is returned from this
    * method.
    *
    * @return  The result of executing the underlying expression: an object
    *          reference or <code>null</code>, as described above.
    */
   protected abstract Object execute();
}