/*
** Module   : Compiler.java
** Abstract : Assembles bytecode instructions and compiles and loads
**            in-memory class files
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 ECF 20050505   @21343 Created initial version. Parses, assembles, compiles, loads, and 
**                           instantiates compiled expression classes. Represents a complete
**                           rewrite of the expression compiler originally created for the Java 
**                           Trace Analyzer.
** 002 ECF 20050604   @21411 Added diagnostic feature to help isolate problematic expression 
**                           components. Maintains linkage between AST and runtime instructions
**                           to permit better error reporting at runtime.
** 003 ECF 20050606   @21428 Fix to convertToRequiredType. Allow narrowing reference conversions 
**                           to primitive wrappers as well as compound conversions from a wrapper 
**                           of one primitive type to a different primitive type (e.g., 
**                           java.lang.Long to int).
** 004 ECF 20050607   @21436 Yet another fix for auto type conversion. Allow conversions between 
**                           Long, Integer, and Double. Enforce explicit casts of method 
**                           parameters to types other than those expected by the method.
** 005 ECF 20050613   @21471 Fixed NPEs in convertToRequiredType. Added safety checks for null. 
**                           Assignments of null to settable properties and variables were
**                           triggering NPE.
** 006 ECF 20050615   @21494 Fixed conjunction operand casts to boolean. The cast logic now 
**                           properly branches based upon the result of casting an operand to a
**                           boolean (or Boolean) type. Previously the casted value was simply 
**                           pushed onto the stack, causing a VerifyError. Also removed references
**                           to the NEG_METH_BOOLEAN, which was never being set as a token type.
** 007 ECF 20050720   @21748 Fixed NPE in array store code. Passing a null value in a var-arg list
**                           was causing an NPE during parameter auto-conversion. Simple null
**                           check prevents this.
** 008 ECF 20051003   @22923 Modified assembleComparatorOp method and code which invokes it. We 
**                           now pass in the "least common denominator" operand type as reported
**                           by the Verifier. Previously, we used the value of the "extra" object 
**                           itself for this purpose, but this was breaking function signature 
**                           matching logic. ExtraAst has been upgraded to provide this new data 
**                           (see also @22922 and @22924).
** 009 ECF 20060226   @24833 Minor optimization. Iterate map entries for variable pool and library
**                           maps, instead of iterating keys and doing an extra lookup on each.
** 010 ECF 20070629   @34348 Minor enhancement. Replaced StringBuffer with StringBuilder.
** 011 ECF 20071103   @39257 Implement ExpressionFlags interface. Added support for SQL-like 
**                           substitution parameters and equivalence syntax. The latter requires
**                           no-assign mode ('=' is converted to '==').
** 012 CA  20140630          Always use ldc_w instead of ldc, otherwise the constant pool index 
**                           is limited to a max of 255.
** 013 EVL 20160222          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import antlr.*;
import antlr.collections.AST;

/**
 * Creates in-memory Java classes on the fly from expression strings written
 * in infix notation. A custom, compiled expression is generated with a call
 * to the static {@link #compile} method. This method is intended to be
 * called only from the {@link Expression} class. At a high level, it does
 * the following:
 * <ol>
 * <li>Creates a new <code>Compiler</code> instance and associates it with
 *     a particular <code>Expression</code> instance.
 * <li>Parses the infix expression string provided by the
 *     <code>Expression</code> object into an abstract syntax tree (AST);
 * <li>Preprocesses the AST to associate information with each AST node to
 *     assist with bytecode assembly for logical expressions (or logical
 *     subexpressions within a larger, non-logical expression);
 * <li>Recursively walks the tree in a depth-first iteration of all the
 *     nodes, assembling bytecode instructions for the operation represented
 *     by each node. The assembled bytecode unit for each AST node is
 *     attached to the associated node. Upon return from each recursion, the
 *     assembled bytecode units for the children nodes are combined to
 *     perform the higher level operation represented by the parent node.
 * <li>Prepares an in-memory Java class file with an <code>execute</code>
 *     method containing the bytecode assembled in the previous step. The
 *     class generated is a direct subclass of {@link CompiledExpression}.
 * <li>Loads the class file created in the previous step using a custom
 *     class loader.
 * <li>Instantiates the custom, <code>CompiledExpression</code> subclass
 *     and returns it.
 * </ol>
 */
final class Compiler
implements CompilerConstants,
           ExpressionFlags,
           ExpressionParserTokenTypes
{
   /** Master debug flag;  verbose output to stdout generated when true */
   static final boolean debug = false;
   
   /** Package prefix using slash separator; e.g., 'foo/bar' */
   static final String PKG_PREFIX;
   
   /** Primitive signature names by primitive class */
   private static final Map PRIMITIVES = new HashMap();
   
   /** Map of primitive object wrapper classes to primitive types */
   private static final Map WRAPPERS = new HashMap();
   
   // Dynamically retrieve package prefix.
   static
   {
      Package pkg = Compiler.class.getPackage();
      String name = pkg.getName();
      PKG_PREFIX = name.replace('.', '/') + '/';
      
      PRIMITIVES.put(Boolean.TYPE, new PrimitiveInfo(1, "Z",
                                                     Boolean.class,
                                                     "booleanValue"));
      PRIMITIVES.put(Integer.TYPE, new PrimitiveInfo(1, "I",
                                                     Integer.class,
                                                     "intValue"));
      PRIMITIVES.put(Long.TYPE   , new PrimitiveInfo(2, "J",
                                                     Long.class,
                                                     "longValue"));
      PRIMITIVES.put(Double.TYPE , new PrimitiveInfo(2, "D",
                                                     Double.class,
                                                     "doubleValue"));
      PRIMITIVES.put(Void.TYPE   , new PrimitiveInfo(0, "V",
                                                     Void.class,
                                                     null));
      
      WRAPPERS.put(Boolean.class , Boolean.TYPE);
      WRAPPERS.put(Integer.class , Integer.TYPE);
      WRAPPERS.put(Long.class    , Long.TYPE   );
      WRAPPERS.put(Double.class  , Double.TYPE );
   }
   
   /** Class name prefix */
   private static final String CLASS_NAME_PREFIX = "CE";
   
   /** Value for floating point epsilon testing */
   private static final double EPSILON = Math.pow(2, -52);
   
   /** Custom classloader for in-memory compiled expression class loading */
   private static ExpressionClassLoader loader = new ExpressionClassLoader();
   
   /** Counter appended to each new expression class' name */
   private static volatile int classCount = 0;
   
   /** Object which holds expression information */
   private Expression expression = null;
   
   /** Map of library name symbols to library runtime index */
   private Map libraries = new HashMap();
   
   /** Map of variable name symbols to variable runtime index */
   private Map variables = new HashMap();
   
   /** Parsed infix expression in AST form */
   private ExtraAst tree = null;
   
   /** Data type of result of evaluating expression (may be primitive) */
   private Class resultType = null;
   
   /** Data type which must be returned from execute() (must be object) */
   private Class returnType = null;
   
   /** Class file object which represents compiled class */
   private ClassFile classFile = null;
   
   /** Constant pool of class file */
   private ConstantPool constantPool = null;
   
   /** Expression class after classloading */
   private Class customClass = null;
   
   /** Stack depth at current instruction */
   private int stackDepth = 0;
   
   /** Maximum stack depth across all instructions */
   private int maxStackDepth = 0;
   
   /** Maximum local variables in use within execute() method */
   private int maxLocals = 1;
   
   /** Instance of compiled expression custom subclass */
   private CompiledExpression compiled = null;
   
   /** Constant for rounding floating point calculations to fixed precision */
   private double roundingConstant = 0.0;
   
   /** AST currently being assembled */
   private ExtraAst currentAst = null;
   
   /** Next AST ID to be assigned during assembly of bytecode instructions */
   private int nextAstId = 0;
   
   /**
    * Private constructor. Access is only available through the static
    * {@link #compile} method. This ensures that a separate instance of the
    * compiler is created for each expression to be compiled. This is
    * necessary because the compiler manages state specific to the current
    * expression, and it is not designed for asynchronous access.
    *
    * @param   expression
    *          Object containing details about the expression to be compiled,
    *          including the original infix text and runtime flags which
    *          affect the compilation.
    */
   private Compiler(Expression expression)
   {
      this.expression = expression;
      roundingConstant = Math.pow(10, expression.getPrecision());
   }
   
   /**
    * Creates an instance of this class which is associated with the
    * specified expression, compiles the expression into a custom subclass
    * of <code>CompiledExpression</code>, loads it, instantiates it, and
    * returns the new compiled expression instance.
    *
    * @param   expression
    *          Object which contains information specific to the expression
    *          to be compiled, including the expression text in infix
    *          notation, and runtime flags which impact the behavior of
    *          the compiled expression.
    *
    * @return  Instance of a custom subclass of
    *          <code>CompiledExpression</code>.
    *
    * @throws  CompilerException
    *          if any error occurs during compilation of the expression.
    */
   static synchronized CompiledExpression compile(Expression expression)
   throws CompilerException
   {
      Compiler compiler = new Compiler(expression);
      
      return compiler.process();
   }
   
   /**
    * Provides the next, fully qualified class name available for use as
    * a custom, {@link CompiledExpression} subclass. The form of this name
    * is <code>com.goldencode.expr.CE&lt;<i>n</i>&gt;</code>, where <i>n</i>
    * is a sequential number, beginning with 0.
    *
    * @return  Next available class name.
    */
   private static String nextClassName()
   {
      // Create unique, fully qualified class name.
      StringBuilder buf = new StringBuilder(PKG_PREFIX);
      buf.append(CLASS_NAME_PREFIX)
         .append(classCount);
      
      return buf.toString();
   }
   
   /**
    * The primary driver method for the compilation process. It parses the
    * infix expression into an AST, preprocesses the AST, assembles the
    * bytecode instructions, prepares, the custom class file in memory,
    * loads it with a custom class loader, then creates and returns an
    * instance of the new class.
    * <p>
    * If the {@link Expression} object associated with this compiler
    * instance has the <code>RT_DEBUG</code> runtime flag set, a report
    * detailing the compilation is generated to <code>stdout</code> at the
    * end of this process. If, additionally, this class' static {@link
    * #debug} flag is set to <code>true</code>, (and the generation of a
    * class file completed successfully), the custom class file generated
    * for the current expression is written to the working directory.
    *
    * @return  An instance of the custom, compiled expression subclass.
    *
    * @throws  CompilerException
    *          if any error occurs during the compilation process.
    */
   private CompiledExpression process()
   throws CompilerException
   {
      // Create a class file object into which to install execute method.
      classFile = new ClassFile(nextClassName(), CLASS_COMPILEDEXPRESSION);
      
      // Get constant pool, which is needed during assembly and
      // compilation.
      constantPool = classFile.getConstantPool();
      
      try
      {
         // Parse infix expression into AST.
         tree = parse();
         
         // Collect data about runtime libraries and variable pool.
         preprocess(tree);
         if (debug)
         {
            BaseAST.setVerboseStringConversion(true,
                                               ExpressionParser._tokenNames);
            ExpressionParser.visit(tree, 0);
         }
         
         // Assemble core bytecode for execute method.
         CodeUnit unit = assemble(tree);
         
         // Prepare class file and wrap execute bytecode with return logic.
         prepare(unit);
         
         // Load the class using the custom classloader.
         load();
         
         // Create an instance of the compiled class and install runtime
         // libraries and variables.
         instantiate();
      }
      catch (CompilerException exc)
      {
         throw exc;
      }
      catch (ANTLRException exc)
      {
         throw new CompilerException("Error parsing expression", exc);
      }
      catch (UnresolvedSymbolException exc)
      {
         throw new CompilerException("Error parsing expression", exc);
      }
      catch (Exception exc)
      {
         throw new CompilerException("Error compiling expression", exc);
      }
      catch (VerifyError err)
      {
         throw new CompilerException("Compiler defect", err);
      }
      finally
      {
         if (debug || (expression.getRuntimeFlags() & RT_DEBUG) != 0)
         {
            try
            {
               dumpState();
            }
            catch (Exception exc)
            {
               System.err.println("Error during reporting:");
               exc.printStackTrace();
               System.err.println("-----------------------");
            }
         }
      }
      
      // Processing was successful;  increment class count.
      classCount++;
      
      return compiled;
   }
   
   /**
    * This method tracks changes in the current and maximum stack depths. It
    * is called by the <code>Instruction</code> class for each instruction
    * in a method's code attribute.  The maximum stack depth is ultimately
    * used in the <code>Code</code> attribute of the method.
    *
    * @param   increment
    *          An incremental change to the stack depth for the bytecode in
    *          the method currently being assembled.  May be zero, positive,
    *          or negative.
    */
   void incrementStackDepth(int increment)
   {
      if (increment != 0)
      {
         stackDepth += increment;
         maxStackDepth = Math.max(stackDepth, maxStackDepth);
      }
      
//      if (debug)
//      {
//         System.out.println("   increment = " + increment);
//         System.out.println("   stackDepth = " + stackDepth);
//         System.out.println("   maxStackDepth = " + maxStackDepth);
//      }
   }
   
   /**
    * Return the maximum number of 32-bit slots which the JVM runtime stack
    * must be able to accommodate for the method whose bytecode currently is
    * being assembled.
    *
    * @return  Maximum stack depth in 32-bit slots.
    */
   int getMaxStackDepth()
   {
      return maxStackDepth;
   }
   
   /**
    * Reset to zero each of the state variables which track current and
    * maximum stack depth. These variables should be reset before each new
    * method is assembled.
    */
   private void resetStackVariables()
   {
      stackDepth = 0;
      maxStackDepth = 0;
   }
   
   /**
    * Creates and uses the {@link ExpressionLexer} to scan the infix notation
    * expression into tokens, and the {@link ExpressionParser} to generate an
    * abstract syntax tree (AST) from these tokens. The AST represents the
    * expression in postfix notation, which is a more appropriate form for
    * the following steps in the compilation process.
    *
    * @return  The expression, represented as an AST.
    *
    * @throws  ANTLRException
    *          if there were any errors lexing or parsing the expression.
    */
   private ExtraAst parse()
   throws ANTLRException
   {
      boolean noAssign = ((expression.getRuntimeFlags() & RT_NO_ASSIGN) != 0);
      StringReader reader = new StringReader(expression.getInfix());
      ExpressionLexer lexer = new ExpressionLexer(reader);
      SymbolResolver resolver = expression.getResolver();
      ExpressionParser parser = new ExpressionParser(lexer,
                                                     resolver,
                                                     noAssign);
      
      resultType = parser.expression();
      
      return (ExtraAst) parser.getAST();
   }
   
   /**
    * Preprocess the AST form of the expression to annotate operations which
    * must produce a boolean result. For each logical operator encountered
    * in a depth-first, recursive iteration of the tree, a {@link
    * Compiler.LogicalData} object is stored in the {@link ExtraAst} node's
    * "extra" field. The  <code>LogicalData</code> object contains data used
    * to determine how the flow of control should branch to other locations
    * in the bytecode instruction array at runtime, depending upon the
    * outcome of the logical operation.
    * <p>
    * When a logical operation takes place, the flow of control moves to the
    * next <em>relevant</em> instruction. Relevance of a set of instructions
    * is determined by the appication of some simple boolean algebra
    * principles:
    * <ul>
    * <li>The right operand of a logical <code>OR</code> operation is
    *     relevant only if the left operand evaluates to <code>false</code>.
    * <li>The right operand of a logical <code>AND</code> operation is
    *     relevant only if the left operand evaluates to <code>true</code>.
    * </ul>
    * <p>
    * For each logical operator encountered during this stage, a
    * <code>LogicalData</code> object is created. This object encapsulates
    * the notion of a branch <em>trigger</em>. The trigger activates a jump
    * based upon the rules declared above. Another <code>LogicalData</code>
    * object is attached to each operand of the operator which cascades the
    * application of these rules. The jump will always be to an instruction
    * <em>outside of the logical operation from which it originates</em> and
    * <em>after the code unit which represents that logical operation</em>.
    * All irrelevant operations are always skipped as soon as possible. The
    * number of operations skipped depends upon the structure of the
    * expression. Thus, jumps may take place through several layers of
    * nested, logical tests if the expression is so written.
    * <p>
    * At the top level of a logical expression (or subexpression), the flow
    * of control must move to an instruction which pushes either a
    * <code>true</code> value or a <code>false</code> value onto the runtime
    * stack, such that following instructions can use this value. If the top
    * level represents the highest level operation in the expression, this
    * value will be wrapped in a <code>Boolean</code> object and returned
    * from the compiled expression's {@link CompiledExpression#execute
    * execute} method. If the top level represents a subexpression within
    * the overall expression, this value will simply be an input into the
    * next higher level operation within the overall expression which does
    * not represent a logical operation (for example, the value may be used
    * as a boolean parameter to a method call).
    * <p>
    * As a result of this method, the AST will contain enough information
    * such that all necessary flow of control branches for logical operations
    * can be assembled into bytecode instructions in the assembly step.
    *
    * @param   ast
    *          An AST node in the expression tree. Upon the first call into
    *          this method, this will be the root of the entire tree.
    *          Thereafter, this method will be called recursively, such that
    *          it will be called for each node in the tree.
    */
   private void preprocess(ExtraAst ast)
   {
      int type = ast.getType();
      Object extra = ast.getExtra();
      LogicalData logData = null;
      if (extra instanceof LogicalData)
      {
         logData = (LogicalData) extra;
         extra = logData.getExtra();
      }
      
      switch (type)
      {
         case AND:
            preprocessConjunction(ast, false);
            break;
            
         case OR:
            preprocessConjunction(ast, true);
            break;
            
         case GT:
         case LT:
         case GTE:
         case LTE:
         case EQUALS:
         case NOT_EQ:
            preprocessBinaryLogical(ast);
            break;
            
         case IS_NULL:
         case NOT_NULL:
            preprocessLogicalTest(ast, true);
            break;
            
         case METH_BOOLEAN:
         case METH_INT:
         case METH_LONG:
         case METH_DOUBLE:
         case METH_OBJ:
            Function func = (Function) extra;
            String name = func.getSymbol();
            Integer size;
            switch (func.getType())
            {
               case Function.GET_LIB:
                  if (!libraries.containsKey(name))
                  {
                     size = new Integer(libraries.size());
                     libraries.put(name, size);
                  }
                  break;
                  
               case Function.GET_VAR:
               case Function.SET_VAR:
                  if (!variables.containsKey(name))
                  {
                     size = new Integer(variables.size());
                     variables.put(name, size);
                  }
                  break;
                  
               default:
                  break;
            }
            
            if (logData != null && ast.getFirstChild() != null)
            {
               preprocessMethodInvocation(ast);
            }
            break;
            
         default:
            break;
      }
      
      // Recursively descend tree, depth-first.
      AST child = ast.getFirstChild();
      while (child != null)
      {
         preprocess((ExtraAst) child);
         child = child.getNextSibling();
      }
   }
   
   /**
    * Preprocess an AST which represents a method invocation. This propagates
    * the proper <code>LogicalData</code> object to the first child of the
    * AST, the result of which always represents the target object of this
    * method invocation.
    *
    * @param   ast
    *          AST which represents a method invocation.
    */
   private void preprocessMethodInvocation(ExtraAst ast)
   {
      LogicalData data = preprocessLogicalTest(ast, true);
      
      ExtraAst operand = (ExtraAst) ast.getFirstChild();
      LogicalData operandData = new LogicalData(true, operand.getExtra());
      operandData.setTrueTarget(data.getTrueTarget());
      operandData.setFalseTarget(data.getFalseTarget());
      operand.setExtra(operandData);
   }
   
   /**
    * Preprocess an AST which represents a binary, logical (non-conjunctive)
    * operation, such as a value-based comparison of two operands. This
    * propagates the appropriate <code>LogicalData</code> object to each
    * child AST which represents a method invocation.
    *
    * @param   ast
    *          AST which represents a binary, logical comparison.
    */
   private void preprocessBinaryLogical(ExtraAst ast)
   {
      LogicalData data = preprocessLogicalTest(ast, true);
      
      ExtraAst left = (ExtraAst) ast.getFirstChild();
      ExtraAst right = (ExtraAst) left.getNextSibling();
      
      if (isMethodInvocation(left))
      {
         LogicalData leftData = new LogicalData(false, left.getExtra());
         leftData.setFalseTarget(data.getFalseTarget());
         left.setExtra(leftData);
      }
      
      if (isMethodInvocation(right))
      {
         LogicalData rightData = new LogicalData(false, right.getExtra());
         rightData.setFalseTarget(data.getFalseTarget());
         right.setExtra(rightData);
      }
   }
   
   /**
    * Preprocess an AST which represents a binary, logical, conjunction
    * operation (logical <code>AND</code> or <code>OR</code>). This
    * propagates the appropriate <code>LogicalData</code> object to each
    * child AST, following the rules of boolean algebra.
    * <p>
    * Worth noting is that the <code>LogicalData</code> object attached to
    * the second (right-hand) operand must enforce a jump trigger on a result
    * of <code>true</code>, if <code>ast</code> represents the top of a chain
    * of logical operations. This is because at the top level, the flow of
    * control must move to a manufactured set of instructions that always
    * has the following structure:
    * <pre>
    *    iconst_0       // push false onto stack
    *    goto 4         // continue processing after the next instruction
    *    iconst_1       // push true onto stack
    *    ...
    * </pre>
    * Thus, it is <em>always</em> necessary to jump over the first two
    * instructions in this code unit in order to push a <code>true</code>
    * value onto the stack. Flow of control naturally proceeds to the
    * <code>iconst_0</code> instruction upon evaluation of the top level
    * operation to <code>false</code>.
    *
    * @param   ast
    *          AST which represents a logical <code>AND</code> or a logical
    *          <code>OR</code> operator.
    * @param   trigger
    *          Trigger which causes a flow of control jump: <code>true</code>
    *          for logical <code>OR</code>, <code>false</code> for logical
    *          <code>AND</code>.
    */
   private void preprocessConjunction(ExtraAst ast, boolean trigger)
   {
      // trigger is true for logical OR, false for logical AND.
      LogicalData data = preprocessLogicalTest(ast, trigger);
      
      ExtraAst left = (ExtraAst) ast.getFirstChild();
      ExtraAst right = (ExtraAst) left.getNextSibling();
      
      // For a logical OR, we always trigger a jump on a true evaluation
      // of the left (first) operand.  False falls through naturally to
      // the right (second) operand.  For a logical AND, we always trigger
      // a jump on a false evaluation of the left (first) operand.  True
      // falls through naturally to the right (second) operand.
      LogicalData leftData = new LogicalData(trigger, left.getExtra());
      leftData.setTrueTarget(trigger ? data.getTrueTarget() : right);
      leftData.setFalseTarget(trigger ? right : data.getFalseTarget());
      left.setExtra(leftData);
      
      // For the right (second) operand, we always trigger a jump on
      // evaluation to true if our parent is the top of a logical branch.
      // Otherwise we trigger a jump on the same condition as our parent.
      trigger = (data.isTopOfBranch() ? true : data.getTrigger());
      LogicalData rightData = new LogicalData(trigger, right.getExtra());
      rightData.setTrueTarget(data.getTrueTarget());
      rightData.setFalseTarget(data.getFalseTarget());
      right.setExtra(rightData);
   }
   
   /**
    * Preprocess an AST which represents <em>any</em> logical test. This is
    * called by the other <code>preprocessXXXX</code> methods as a precursor
    * to their work. This method first retrieves the "extra" field from
    * <code>ast</code>. If it is an instance of {@link Compiler.LogicalData},
    * it is simply returned.
    * <p>
    * If not, a new <code>LogicalData</code> object is created and stored in
    * the "extra" field of <code>ast</code> (the original value is stored in
    * <code>LogicalData</code>'s extra field so it is not lost). New ASTs
    * are created and stored as the <code>true</code> and <code>false</code>
    * targets of the new <code>LogicalData</code> object. Each AST has a code
    * unit associated with it which pushes the corresponding boolean literal
    * onto the runtime stack. <code>ast</code> is marked as the top of a
    * chain of logical operations. The new <code>LogicalData</code> object
    * is returned.
    *
    * @param   ast
    *          AST which represents any sort of logical test.
    * @param   trigger
    *          <code>true</code> if the flow of control should jump upon
    *          <code>ast</code>'s logical test evaluation to <code>true</code>
    *          at runtime; <code>false</code> if it should jump upon
    *          evaluation to <code>false</code> at runtime.
    *
    * @return  The <code>LogicalData</code> object associated with
    *          <code>ast</code>.
    */
   private LogicalData preprocessLogicalTest(ExtraAst ast, boolean trigger)
   {
      Object extra = ast.getExtra();
      LogicalData data;
      if (extra instanceof LogicalData)
      {
         data = (LogicalData) extra;
      }
      else
      {
         data = new LogicalData(trigger, extra);
         data.setTopOfBranch(true);
         ExtraAst tmpAst;
         
         tmpAst = new ExtraAst();
         tmpAst.setType(BOOL_TRUE);
         tmpAst.setText("true*");
         tmpAst.setExtra(assembleBoolLiteral(BOOL_TRUE, null));
         data.setTrueTarget(tmpAst);
         
         tmpAst = new ExtraAst();
         tmpAst.setType(BOOL_FALSE);
         tmpAst.setText("false*");
         tmpAst.setExtra(assembleBoolLiteral(BOOL_FALSE, null));
         data.setFalseTarget(tmpAst);
         
         ast.setExtra(data);
      }
      
      return data;
   }
   
   /**
    * Primary driver method for the assembly of the compiled expression's
    * {@link CompiledExpression#execute execute} method's bytecode
    * instructions. This method executes recursively to walk the AST
    * in a depth-first iteration. At each node encountered, it assembles
    * bytecode instructions to implement the action represented by that
    * node.
    * <p>
    * The key implementation point to understand here is that recursive calls
    * to assemble child nodes take place <em>before</em> any actual assembly
    * is performed. Thus, we first iterate down to the leaf of each AST
    * branch, which represents the most fundamental set of instructions for
    * that branch's operation(s). We then assemble the instructions for that
    * leaf and replace that leaf's "extra" field with the assembled code
    * unit. Next, we visit that leaf's siblings (if any) and recursively
    * assemble them. Then, we return from the recursive call which visited
    * the first leaf, and assemble the instructions which represent that
    * leaf's parent operation, and so on. The last node to be assembled is
    * the root node of the entire tree.
    * <p>
    * At each return from a recursive call, it can thus be assumed that all
    * child nodes of the current level already have been assembled. This
    * simplifies the assembly of the current node in that assembly at this
    * level becomes largely a matter of combining the code units produced
    * for each child and applying the current operation to these units
    * appropriately.
    * <p>
    * The code unit returned from the top level invocation of this method
    * thus represents the core functionality of the compiled expression's
    * <code>execute</code> method. The only instructions missing will be
    * those to wrapper (if necessary) the resulting value into an object,
    * and those to return that object from the method. These instructions
    * are later added by the {@link #prepare} method.
    *
    * @param   ast
    *          AST for which bytecode instructions are to be assembled next.
    *
    * @return  A code unit which contains the proper bytecode instructions
    *          to implement at expression execution time the operation
    *          represented by <code>ast</code>.
    *
    * @throws  CompilerException
    *          if an unrecoverable error occurs while assembling bytecode
    *          instructions.
    */
   private CodeUnit assemble(ExtraAst ast)
   throws CompilerException
   {
      // Recursively descend tree, depth-first.  Collect resulting code
      // units in a list which is passed to higher level assembly routines.
      List units = new LinkedList();
      ExtraAst child = (ExtraAst) ast.getFirstChild();
      while (child != null)
      {
         CodeUnit result = assemble(child);
         units.add(result);
         child = (ExtraAst) child.getNextSibling();
      }
      
      // Assign sequential IDs to AST nodes for runtime diagnostic use.
      ast.setId(nextAstId++);
      currentAst = ast;
      
      CodeUnit unit = null;
      int type = ast.getType();
      int radix = 10;
      LogicalData logData = null;
      Object extra = ast.getExtra();
      if (extra instanceof LogicalData)
      {
         logData = (LogicalData) extra;
         extra = logData.getExtra();
      }
      
      switch (type)
      {
         case NOT:
            unit = assembleLogicalNegation(logData,
                                           (Class) extra,
                                           (CodeUnit) units.get(0));
            break;
            
         case AND:
         case OR:
            unit = assembleConjunctionOp(ast);
            break;
            
         case GT:
         case LT:
         case GTE:
         case LTE:
         case EQUALS:
         case NOT_EQ:
            unit = assembleComparatorOp(type,
                                        logData,
                                        ast.getOperandType(),
                                        units);
            break;
            
         case BIT_COMP:
            unit = assembleBitwiseNegation((Class) extra,
                                           (CodeUnit) units.get(0));
            break;
            
         case RSHIFT:
         case LSHIFT:
         case ZRSHIFT:
         case BIT_AND:
         case BIT_OR:
         case BIT_XOR:
            unit = assembleBinaryBitwiseOp(type, (Class) extra, units);
            break;
            
         case UN_MINUS:
            unit = assembleArithmeticNegation((Class) extra,
                                              (CodeUnit) units.get(0));
            break;
            
         case MULTIPLY:
         case DIVIDE:
         case MODULO:
         case PLUS:
         case MINUS:
            unit = assembleBinaryArithmeticOp(type, (Class) extra, units);
            break;
            
         case STRING:
            unit = assembleStringLiteral(ast.getText());
            break;
            
         case NULL:
            unit = new CodeUnit();
            unit.addInstruction(_aconst_null);
            break;
            
         case IS_NULL:
         case NOT_NULL:
            unit = assembleNullTest(type, logData, (CodeUnit) units.get(0));
            break;
            
         case BOOL_TRUE:
         case BOOL_FALSE:
            unit = assembleBoolLiteral(type, logData);
            break;
            
         case DEC_LITERAL:
            unit = assembleDecLiteral(parseDouble(ast));
            break;
            
         case HEX_LITERAL:
            radix = 16;
         case NUM_LITERAL:
            unit = assembleNumLiteral((Class) extra, parseLong(ast, radix));
            break;
            
         case CAST:
            unit = assembleUserCast(logData,
                                    (CodeUnit) units.get(0),
                                    (Class) extra);
            break;
            
         case METH_BOOLEAN:
         case METH_INT:
         case METH_LONG:
         case METH_DOUBLE:
         case METH_OBJ:
            Function func = (Function) extra;
            switch (func.getType())
            {
               case Function.GET_LIB:
               case Function.GET_VAR:
               case Function.SET_VAR:
                  unit = assembleBootstrapMethod(type, logData, func, units);
                  break;
                  
               default:
                  unit = assembleRegularMethod(type, logData, func, units);
                  break;
            }
            break;
            
         default:
            break;
      }
      
      // Manufacture instructions to push proper boolean value onto stack
      // if we encounter the top of a logical branch.
      if (logData != null && logData.isTopOfBranch())
      {
         ExtraAst falseAst = logData.getFalseTarget();
         CodeUnit falseUnit = (CodeUnit) falseAst.getExtra();
         
         ExtraAst trueAst = logData.getTrueTarget();
         CodeUnit trueUnit = (CodeUnit) trueAst.getExtra();
         
         // Add unit to push false onto stack.
         unit.addCodeUnit(falseUnit);
         
         // Add instruction to jump over this instruction and following
         // code unit.
         Instruction instruc = new Instruction(_goto);
         instruc.addArgument(new NumericArgument(3 + trueUnit.getLength(),
                                                 Argument.SHORT));
         unit.addInstruction(instruc);
         
         // Add unit to push true onto stack.
         unit.addCodeUnit(trueUnit);
         
         unit.setStackMarker(true);
      }
      
      // All token types should be covered by the above switch; if not, we
      // will throw NPE here...
      if (debug && unit == null)
      {
         BaseAST.setVerboseStringConversion(true,
                                            ExpressionParser._tokenNames);
         System.err.println("Error assembling:");
         System.err.println("AST  :  " + ast.toString());
         System.err.println("Extra:  " + extra);
         
         throw new NullPointerException();
      }
      
      // Store assembled code unit as AST's extra data.
      ast.setExtra(unit);
      
      return unit;
   }
   
   /**
    * Adds instructions to the specified code unit to update diagnostic
    * information at runtime. Specifically, the ID of the AST associated
    * with <code>unit</code>'s instructions is set in the compiled
    * expression runtime instance. These are executed just before
    * instructions which may cause an exception to be generated at runtime,
    * such that an exception handler around the executing expression can
    * query this information and provide useful diagnostic messages about
    * the error.
    * <p>
    * These special diagnostic instructions are only added if the expression
    * associated with this compilation have one or more of the following
    * runtime flags turned on:
    * <ul>
    * <li>{@link Expression#RT_ERROR}
    * <li>{@link Expression#RT_DEBUG}
    * </ul>
    * <p>
    * Otherwise, this method returns immediately, without adding any
    * instructions to the code unit.
    *
    * @param   unit
    *          Code unit to which diagnostic instructions are added.
    */
   private void addDebug(CodeUnit unit)
   {
      int flags = expression.getRuntimeFlags();
      if ((flags & (RT_ERROR | RT_DEBUG)) == 0)
      {
         return;
      }
      
      // Load the compiled expression's "this" reference.
      unit.addInstruction(_aload_0);
      
      // Load the index constant, which will be the sole parameter to the
      // setActiveAstIndex method call.
      unit.addCodeUnit(assembleNumLiteral(Integer.TYPE, currentAst.getId()));
      
      // Invoke the setActiveAstIndex method.
      int[] stackChange = { -1, -1 };
      String name = formatClassName(CompiledExpression.class, false);
      Instruction instruc = new Instruction(_invokevirtual, stackChange);
      instruc.addArgument(new ConstantArgument(
         MethodrefConstant.create(
            constantPool, name, "setActiveAstId", "(I)V")));
      unit.addInstruction(instruc);
   }
   
   /**
    * Prepare the compiled class file object for class loading. This
    * involves:
    * <ol>
    * <li>setting class file access flags
    * <li>adding a default constructor
    * <li>adding the <code>execute</code> method
    * <li>adding to the <code>execute</code> method the code attribute
    *     containing the assembled bytecode instructions for the expression
    * <li>wrapping the core bytecode instructions with additional
    *     instructions to ensure the proper object type is returned
    * <li>walking the bytecode array to determine the maximum stack depth
    *     required for the method
    * <li>fixing up the stack by adding instructions where necessary to pop
    *     unnecessary values from the runtime stack before branching flow of
    *     control
    * </ol>
    *
    * @param   core
    *          Code unit which contains the bytecode instructions which
    *          perform the core functionality of the <code>execute</code>
    *          method.
    */
   private void prepare(CodeUnit core)
   {
      Class coreType = core.getDataType();
      
      // Make class public.
      classFile.setAccessFlags(ClassFile.ACC_PUBLIC);
      
      // Add constructor.
      classFile.addDefaultConstructor(ClassFile.ACC_PUBLIC, this);
      resetStackVariables();
      
      // Add execute() method.
      MethodInfo method = new MethodInfo(constantPool,
                                         METHNAME_EXECUTE,
                                         METHDESC_EXECUTE,
                                         ClassFile.ACC_PUBLIC);
      classFile.addMethod(method);
      
      // Prepare execute() code attribute.
      CodeAttribute code = new CodeAttribute(constantPool);
      method.addAttribute(code);
      
      // Assemble instructions to optimistically push a new object reference
      // onto the stack for eventual return from the method.
      // The execute() method must return an object, not a primitive, so
      // ensure returnType is a wrapper type if resultType is a primitive.
      if (coreType.isPrimitive())
      {
         PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(coreType);
         returnType = info.wrapperClass;
         code.addCodeUnit(assembleNewObjectReference(returnType));
      }
      else
      {
         returnType = coreType;
      }
      
      // Add primary code.
      code.addCodeUnit(core);
      
      // Append return instructions.
      assembleReturn(core, code);
      
      // Set necessary maximums into code attribute.
      code.setMaxStack(this);
      code.setMaxLocals(maxLocals);
      
      // Ensure proper stack handling during branching.
      code.fixupStack();
   }
   
   /**
    * Load the in-memory byte array which represents the compiled expression
    * class file, using a {@link ExpressionClassLoader custom class loader}.
    *
    * @throws  ClassNotFoundException
    *          if any error occurs loading the class file.
    */
   private void load()
   throws ClassNotFoundException
   {
      String name = classFile.getFullyQualifiedName();
      loader.putClassFile(name, classFile);
      customClass = loader.loadClass(name, true);
   }
   
   /**
    * Create a new instance of the custom, compiled expression class and
    * initialize it with:
    * <ul>
    * <li>The class which represents the expected return type from the
    *     compiled expression's <code>execute</code> method.
    * <li>The array of library objects to which the expression refers.
    * <li>The array of user variables to which the expression refers.
    * </ul>
    * <p>
    * The new instance is stored in the {@link #compiled} instance variable.
    *
    * @throws  InstantiationException
    *          if instantiation fails due to a compiler defect.
    * @throws  IllegalAccessException
    *          if the class or its default constructor are not accessible.
    *          This should only happen in the case of a compiler defect.
    */
   private void instantiate()
   throws InstantiationException,
          IllegalAccessException
   {
      compiled = (CompiledExpression) customClass.newInstance();
      
      // Store return type.
      compiled.setReturnType(returnType);
      
      SymbolResolver resolver = expression.getResolver();
      
      // Populate array of runtime libraries.
      Object[] libs = new Object[libraries.size()];
      Iterator iter = libraries.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry next = (Map.Entry) iter.next();
         String key = (String) next.getKey();
         int index = ((Integer) next.getValue()).intValue();
         Scope scope = resolver.getCurrentScope();
         while (scope != null && libs[index] == null)
         {
            libs[index] = resolver.getLibraryTarget(scope, key);
            scope = scope.getEnclosingScope();
         }
      }
      compiled.setLibraries(libs);
      
      // Populate array of user variables.
      Variable[] vars = new Variable[variables.size()];
      iter = variables.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry next = (Map.Entry) iter.next();
         String key = (String) next.getKey();
         int index = ((Integer) next.getValue()).intValue();
         Scope scope = resolver.getCurrentScope();
         while (scope != null && vars[index] == null)
         {
            vars[index] = resolver.getVariable(scope, key);
            scope = scope.getEnclosingScope();
         }
      }
      compiled.setVariables(vars);
      
      // Store the expression's AST in the compiled object if runtime
      // diagnostics are required.
      int flags = expression.getRuntimeFlags();
      if ((flags & (RT_ERROR | RT_DEBUG)) != 0)
      {
         compiled.setAst(tree);
      }
   }
   
   /**
    * Assemble instructions to push a new object reference of the specified
    * type onto the runtime stack. If <code>type</code> represents a
    * <code>void</code> return, a <code>null</code> reference is pushed
    * onto the stack instead of an object reference. For non-<code>null</code>
    * values, the reference is duplicated, since it is assumed that the
    * duplicate reference will be consumed in a later call to the class'
    * constructor.
    *
    * @param   type
    *          Class representing the object reference type to create.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleNewObjectReference(Class type)
   {
      CodeUnit unit = new CodeUnit(type);
      
      if (type == Void.class)
      {
         // Special case:  void result type means we return null. So push
         // null onto the stack here instead.
         unit.addInstruction(_aconst_null);
      }
      else
      {
         String name = formatClassName(type, false);
         
         // Create a new object reference on the stack.
         Instruction instruc = new Instruction(_new);
         instruc.addArgument(new ConstantArgument(
            ClassConstant.create(constantPool, name)));
         unit.addInstruction(instruc);
         
         // Duplicate the object reference so it is at the top of the stack
         // after the constructor is called at the end of the method.
         unit.addInstruction(_dup);
      }
      
      return unit;
   }
   
   /**
    * Assemble instructions to wrapper (if necessary) the data value on top
    * of the runtime stack, and to return an object from the current method.
    *
    * @param   core
    *          Code unit which contains the bytecode instructions which
    *          perform the core functionality of the <code>execute</code>
    *          method.
    * @param   code
    *          Code attribute object to which the return instructions are
    *          appended.
    */
   private void assembleReturn(CodeUnit core, CodeAttribute code)
   {
      // If result of expression execution is a primitive, we need to
      // wrap it in a wrapper object first.
      Class coreType = core.getDataType();
      if (coreType.isPrimitive() && resultType != Void.TYPE)
      {
         code.addCodeUnit(assembleWrapperConstructor(coreType));
      }
      
      // Add an instruction to return the object reference from the top
      // of the stack.
      code.addInstruction(new Instruction(_areturn));
   }
   
   /**
    * Assemble instructions to perform a manual type cast of the runtime
    * result of executing <code>castTarget</code> to the required type. If
    * the code unit being cast is a logical code unit, ensure it branches to
    * the correct destination instruction.
    *
    * @param   logData
    *          Object which contains information on how flow of control
    *          should proceed upon the execution of <code>castTarget</code>'s
    *          instructions.
    * @param   castTarget
    *          Unit of code whose result should be typecast.
    * @param   required
    *          Class type to which <code>targetCast</code> is to be casted.
    *
    * @return  Code unit containing <code>castTarget</code>, followed by the
    *          instruction(s) necessary to typecast its result.
    */
   private CodeUnit assembleUserCast(LogicalData logData,
                                     CodeUnit castTarget,
                                     Class required)
   {
      CodeUnit unit = convertToRequiredType(logData,
                                            castTarget,
                                            castTarget.getDataType(),
                                            required);
      
      if (logData != null && logData.getTrueTarget() != null)
      {
         unit = assembleBranchOnBooleanResult(NOT_EQ, // i.e., result != 0
                                              logData,
                                              unit,
                                              required);
      }
      
      return unit;
   }
   
   /**
    * Assemble instructions necessary to branch the flow of control based
    * upon the result of executing the instructions within the given code
    * unit. Destination information is provided by <code>logData</code>.
    *
    * @param   type
    *          Type of branch logic;  indicates how branching test is
    *          performed and which bytecode instruction is used to jump.
    * @param   logData
    *          Object which contains information on how flow of control
    *          should proceed upon the execution of <code>core</code>'s
    *          instructions.
    * @param   core
    *          Underlying code unit whose bytecode instructions are
    *          executed and which must produce a result which determines
    *          whether or not we jump to a new location.
    * @param   required
    *          Type of result from the execution of <code>core</code> which
    *          is required to perform the branch.
    *
    * @return  A code unit which wraps <code>core</code>, appended with the
    *          instructions necessary to branch based upon the result of
    *          executing <code>core</code>'s instructions.
    */
   private CodeUnit assembleBranchOnBooleanResult(int type,
                                                  LogicalData logData,
                                                  CodeUnit core,
                                                  Class required)
   {
      if (required == Boolean.class)
      {
         core = operandToPrimitive(core, Boolean.TYPE);
      }
      
      CodeUnit unit = new CodeUnit(logData.getTrigger(),
                                   logData.getTarget(),
                                   type,
                                   false);
      unit.addCodeUnit(core);
      unit.addBranchInstruction();
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to perform a logical negation of
    * the specified operand code unit. The negation will take one of two
    * forms, depending upon whether a non-<code>null</code>
    * <code>logData</code> object is provided:
    * <p>
    * If <code>logData</code> is non-<code>null</code>, the negation amounts
    * to causing a jump to the instruction indicated by <code>logData</code>'s
    * appropriate target, if the result of executing <code>operand</code>'s
    * instructions is <code>false</code>.
    * <p>
    * If no <code>logData</code> object is provided, the negation amounts to
    * examining the boolean value at the top of the runtime stack and pushing
    * its opposite back onto the stack.
    *
    * @param   logData
    *          Object which contains information on how flow of control
    *          should proceed upon the execution of <code>operand</code>'s
    *          instructions.
    * @param   required
    *          Data type expected by the operation containing
    *          <code>operand</code>.
    * @param   operand
    *          Code unit containing instructions which should evaluate to a
    *          <code>true</code> or <code>false</code> result.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleLogicalNegation(LogicalData logData,
                                            Class required,
                                            CodeUnit operand)
   {
      // Determine required primitive type.
      Class primitive = (required.isPrimitive()
                         ? required
                         : (Class) WRAPPERS.get(required));
      
      operand = operandToPrimitive(operand, primitive);
      
      CodeUnit unit;
      
      if (logData != null)
      {
         unit = new CodeUnit(logData.getTrigger(),
                             logData.getTarget(),
                             EQUALS,   // i.e., result equals 0
                             false);
         unit.addCodeUnit(operand);
         unit.addBranchInstruction();
      }
      else
      {
         unit = new CodeUnit(primitive);
         unit.addCodeUnit(operand);
         
         // Negate boolean value at top of stack.
         Instruction instruc = new Instruction(_ifeq);
         instruc.addArgument(new NumericArgument(7, Argument.SHORT));
         unit.addInstruction(instruc);
         unit.addInstruction(_iconst_0);
         instruc = new Instruction(_goto);
         instruc.addArgument(new NumericArgument(4, Argument.SHORT));
         unit.addInstruction(instruc);
         unit.addInstruction(_iconst_1);
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to execute a logical conjunction
    * operation (<code>AND</code> or <code>OR</code>). This consists of
    * ensuring both operands produce a primitive, boolean result, then
    * creating a new code unit which simply contains each of the operands'
    * code units in the order encountered in the AST. This is possible
    * because each child code unit already contains instructions to proceed
    * (possibly branch) to the proper instructions, based upon their
    * respective results.
    *
    * @param   ast
    *          AST representing the logical conjunction operation. This AST
    *          will have two children: one each for the left and right sides
    *          of the conjunction operation.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleConjunctionOp(ExtraAst ast)
   {
      CodeUnit unit = new CodeUnit(Boolean.TYPE);
      
      int type;
      ExtraAst child;
      CodeUnit next;
      
      // Left child of conjunction.
      child = (ExtraAst) ast.getFirstChild();
      next = (CodeUnit) child.getExtra();
      type = child.getType();
      
      if (type == METH_BOOLEAN && next.getDataType() == Boolean.class)
      {
         next = operandToPrimitive(next, Boolean.TYPE);
      }
      unit.addCodeUnit(next);
         
      // Right child of conjunction.
      child = (ExtraAst) child.getNextSibling();
      next = (CodeUnit) child.getExtra();
      type = child.getType();
      
      if (type == METH_BOOLEAN && next.getDataType() == Boolean.class)
      {
         next = operandToPrimitive(next, Boolean.TYPE);
      }
      unit.addCodeUnit(next);
      
      return unit;
   }
   
   /**
    * Assemble instructions to perform a logical comparison of two values,
    * and to proceed (possibly branch) to the proper instruction, based upon
    * the outcome of that comparison. Comparator operations will
    * <em>always</em> contain a branching instruction. The destination of
    * the branch is determined during the {@link #preprocess} phase of
    * compilation, when the appropriate <code>LogicalData</code> object is
    * attached to each operand which may be required to branch within a
    * logical operation.
    * <p>
    * This method handles numeric, boolean, and <code>Comparable</code>
    * operands. Numeric and boolean operands may be primitive or object
    * wrappers. The latter are converted to primitive types for comparison
    * purposes.
    *
    * @param   type
    *          The token type associated with the operation; defines the
    *          nature of the comparison.
    * @param   logData
    *          Object which contains information on how flow of control
    *          should proceed based upon the result of the comparison.
    * @param   opClass
    *          Class which represents the least common denominator type among
    *          the operands. This is used to determine the type of comparison
    *          operation to be applied.
    * @param   units
    *          Code units which contain the assembled instructions of the
    *          two operands of the comparison operation.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleComparatorOp(int type,
                                         LogicalData logData,
                                         Class opClass,
                                         List units)
   {
      CodeUnit comparison = null;
      
      if (opClass == Long.TYPE)
      {
         comparison = new CodeUnit(Integer.TYPE);
         comparison.addInstruction(_lcmp);
      }
      else if (opClass == Double.TYPE)
      {
         comparison = new CodeUnit(Integer.TYPE);
         switch (type)
         {
            case EQUALS:
               comparison.addCodeUnit(assembleEpsilonTest());
               type = LT;
               break;
               
            case NOT_EQ:
               comparison.addCodeUnit(assembleEpsilonTest());
               type = GTE;
               break;
         }
         comparison.addInstruction(_dcmpg);
      }
      
      boolean binary = (opClass == Integer.TYPE || opClass == Boolean.TYPE);
      CodeUnit unit = new CodeUnit(logData.getTrigger(),
                                   logData.getTarget(),
                                   type,
                                   binary);
      
      // Get operands.
      CodeUnit left = (CodeUnit) units.get(0);
      CodeUnit right = (CodeUnit) units.get(1);
      
      if (opClass.isPrimitive())
      {
         // Convert left operand to primitive, if necessary.
         unit.addCodeUnit(operandToPrimitive(left, opClass));
         
         // Convert right operand to primitive, if necessary.
         unit.addCodeUnit(operandToPrimitive(right, opClass));
         
         if (comparison != null)
         {
            unit.addCodeUnit(comparison);
         }
      }
      else
      {
         Class leftType = left.getDataType();
         if (leftType != opClass)
         {
            left = convertToRequiredType(null, left, leftType, opClass);
         }
         
         Class rightType = right.getDataType();
         if (rightType != opClass)
         {
            right = convertToRequiredType(null, right, rightType, opClass);
         }
         
         unit.addCodeUnit(left);
         unit.addCodeUnit(right);
         
         // Compare two object references as Comparable implementations.
         int[] stackChange = { -1, -1, 1 };
         Instruction instruc = new Instruction(_invokeinterface, stackChange);
         instruc.addArgument(
            new ConstantArgument(
               InterfaceMethodrefConstant.create(
                  constantPool,
                  "java/lang/Comparable",
                  "compareTo",
                  "(Ljava/lang/Object;)I")));
         instruc.addArgument(new NumericArgument(2, Argument.BYTE));
         instruc.addArgument(new NumericArgument(0, Argument.BYTE));
         unit.addInstruction(instruc);
      }
      
      unit.addBranchInstruction();
      
      return unit;
   }
   
   /**
    * Assemble instructions necessary to perform an "epsilon test" to
    * determine the equality of two floating point values. The values are
    * considered equal if the absolute value of the result of subtracting
    * one from the other is less than the smallest difference which can be
    * represented between two floating point numbers. 
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleEpsilonTest()
   {
      CodeUnit unit = new CodeUnit(Double.TYPE);
      
      // Subtract top two operands.
      unit.addInstruction(new Instruction(_dsub));
      
      // Invoke Math.abs() method to get absolute value of result.
      int[] stackChange = { -2, 2 };
      Instruction instruc = new Instruction(_invokestatic, stackChange);
      instruc.addArgument(
         new ConstantArgument(
            MethodrefConstant.create(
               constantPool, "java/lang/Math", "abs", "(D)D")));
      unit.addInstruction(instruc);
      
      // Push EPSILON value onto stack.
      instruc = new Instruction(_ldc2_w);
      instruc.addArgument(
         new ConstantArgument(DoubleConstant.create(constantPool, EPSILON)));
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble instructions necessary to perform a bitwise negation of the
    * specified operand. Integer and long operand types are handled
    * (wrapped primitives are first unwrapped). There is no VM op code for
    * a bitwise negation, so the following arithmetic operation is assembled
    * to perform the negation: <code>-(operand + 1)</code>.
    *
    * @param   required
    *          Data type expected by the operation containing
    *          <code>operand</code>.
    * @param   operand
    *          Code unit containing instructions which generate the value
    *          to be negated.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleBitwiseNegation(Class required, CodeUnit operand)
   {
      // Determine required primitive type.
      Class primitive = (required.isPrimitive()
                         ? required
                         : (Class) WRAPPERS.get(required));
      CodeUnit unit = new CodeUnit(primitive);
      
      unit.addCodeUnit(operandToPrimitive(operand, primitive));
      
      // There is no opcode for bitwise negation, so we use a simple
      // arithmetic device:
      // ~x == (-x - 1) == -(x + 1)
      
      if (primitive == Integer.TYPE)
      {
         unit.addInstruction(_iconst_1);
         unit.addInstruction(_iadd);
         unit.addInstruction(_ineg);
      }
      else if (primitive == Long.TYPE)
      {
         unit.addInstruction(_lconst_1);
         unit.addInstruction(_ladd);
         unit.addInstruction(_lneg);
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to perform a binary, bitwise
    * operation. Supported operations are left and right shifts (including
    * signed and unsigned), as well as bitwise <code>AND</code>,
    * <code>OR</code>, and <code>XOR</code>. Handles both long and integer
    * operands (primitive or wrapped). Both operands must be of the same
    * data type. In the case of the shift operations, the right operand
    * (the shift magnitude) is always converted to an integer before the
    * operation is performed.
    *
    * @param   type
    *          The token type associated with the operation; defines the
    *          nature of the operation.
    * @param   required
    *          Data type of the result of this operation. If a wrapper class,
    *          the corresponding primitive is used instead.
    * @param   units
    *          Code units containing the instructions necessary to generate
    *          each of the operands, in the order encountered in the AST.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleBinaryBitwiseOp(int type,
                                            Class required,
                                            List units)
   {
      // Determine required primitive type.
      Class primitive = (required.isPrimitive()
                         ? required
                         : (Class) WRAPPERS.get(required));
      CodeUnit unit = new CodeUnit(primitive);
      
      // The type of the right operand must be an int always for shift
      // operations; otherwise, it must be the same type as the left
      // operand.
      Class rightType = primitive;
      
      // Determine required opCode.
      final int[] opCodes =
      {
         _ishr , _lshr ,
         _ishl , _lshl ,
         _iushr, _lushr,
         _iand , _land ,
         _ior  , _lor  ,
         _ixor , _lxor ,
      };
      
      int col = -1;
      if (primitive == Integer.TYPE)
      {
         col = 0;
      }
      else if (primitive == Long.TYPE)
      {
         col = 1;
      }
      
      int row = -1;
      switch (type)
      {
         case RSHIFT:
            row = 0;
            rightType = Integer.TYPE;
            break;
            
         case LSHIFT:
            row = 1;
            rightType = Integer.TYPE;
            break;
            
         case ZRSHIFT:
            row = 2;
            rightType = Integer.TYPE;
            break;
            
         case BIT_AND:
            row = 3;
            break;
            
         case BIT_OR:
            row = 4;
            break;
            
         case BIT_XOR:
            row = 5;
            break;
      }
      
      // Get operands.
      CodeUnit left = (CodeUnit) units.get(0);
      CodeUnit right = (CodeUnit) units.get(1);
      
      // Convert left operand to primitive, if necessary.
      unit.addCodeUnit(operandToPrimitive(left, primitive));
      
      // Convert right operand to primitive, if necessary.
      unit.addCodeUnit(operandToPrimitive(right, rightType));
      
      // Add operation instruction.
      unit.addInstruction(opCodes[row * 2 + col]);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to perform an arithmetic (sign)
    * negation of the specified operand. Handles integer, long, and double
    * operand types (primitive or wrapped). The resulting stack value will
    * always be a primitive of the same general type.
    *
    * @param   required
    *          Data type expected by the operation containing
    *          <code>operand</code>.
    * @param   operand
    *          Code unit containing instructions which generate the value
    *          to be negated.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleArithmeticNegation(Class required,
                                               CodeUnit operand)
   {
      // Determine required primitive type.
      Class primitive = (required.isPrimitive()
                         ? required
                         : (Class) WRAPPERS.get(required));
      CodeUnit unit = new CodeUnit(primitive);
      
      unit.addCodeUnit(operandToPrimitive(operand, primitive));
      
      int opCode = -1;
      if (primitive == Integer.TYPE)
      {
         opCode = _ineg;
      }
      else if (primitive == Long.TYPE)
      {
         opCode = _lneg;
      }
      else if (primitive == Double.TYPE)
      {
         opCode = _dneg;
      }
      
      unit.addInstruction(opCode);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to perform a binary, arithmetic
    * operation (multiplication, division, remainder, addition, subtraction).
    * Integer, long and double operand types (primitive or wrapped) are
    * handled, including mixed operand types (the narrower is converted to
    * the wider).
    * <p>
    * If the operation uses floating point math, and the associated {@link
    * Expression} object has the <code>RT_FP_ROUNDING</code> runtime flag
    * set, the result is rounded to the number of decimal places precision
    * specified in that object. Note that this may generate an exception if
    * the requested level of precision cannot be supported at runtime.
    * <p>
    * The resulting stack value will always be a primitive of the same data
    * type as the widest operand.
    *
    * @param   type
    *          The token type associated with the operation; defines the
    *          nature of the operation.
    * @param   required
    *          Data type of the result of this operation. If a wrapper class,
    *          the corresponding primitive is used instead.
    * @param   units
    *          Code units containing the instructions necessary to generate
    *          each of the operands, in the order encountered in the AST.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    *
    * @see     CompiledExpression#testRound
    */
   private CodeUnit assembleBinaryArithmeticOp(int type,
                                               Class required,
                                               List units)
   {
      // Get operands.
      CodeUnit left = (CodeUnit) units.get(0);
      CodeUnit right = (CodeUnit) units.get(1);
      
      // Determine required primitive type.
      Class primitive = (required.isPrimitive()
                         ? required
                         : (Class) WRAPPERS.get(required));
      CodeUnit unit = new CodeUnit(primitive);
      
      // Convert left operand to primitive, if necessary.
      unit.addCodeUnit(operandToPrimitive(left, primitive));
      
      // Convert right operand to primitive, if necessary.
      unit.addCodeUnit(operandToPrimitive(right, primitive));
      
      final int[] opCodes =
      {
         _imul, _lmul, _dmul,
         _idiv, _ldiv, _ddiv,
         _irem, _lrem, _drem,
         _iadd, _ladd, _dadd,
         _isub, _lsub, _dsub,
      };
      
      int col = -1;
      if (primitive == Integer.TYPE)
      {
         col = 0;
      }
      else if (primitive == Long.TYPE)
      {
         col = 1;
      }
      else if (primitive == Double.TYPE)
      {
         col = 2;
      }
      
      int row = -1;
      switch (type)
      {
         case MULTIPLY:
            row = 0;
            break;
            
         case DIVIDE:
            row = 1;
            break;
            
         case MODULO:
            row = 2;
            break;
            
         case PLUS:
            row = 3;
            break;
            
         case MINUS:
            row = 4;
            break;
      }
      
      unit.addInstruction(opCodes[row * 3 + col]);
      
      // Perform rounding to a fixed number of decimal places for double
      // calculations. The level of precision is set in the Expression
      // object. Perform only if floating point rounding flag is enabled.
      if (primitive == Double.TYPE &&
          (expression.getRuntimeFlags() & RT_FP_ROUNDING) != 0)
      {
         unit.addCodeUnit(assembleDecLiteral(roundingConstant));
         unit.addInstruction(_dmul);
         int[] stackChange = { -2, 2 };
         Instruction instruc = new Instruction(_invokestatic, stackChange);
         instruc.addArgument(
            new ConstantArgument(
               MethodrefConstant.create(constantPool,
                                        CLASS_COMPILEDEXPRESSION,
                                        "testRound",
                                        "(D)D")));
         unit.addInstruction(instruc);
         instruc = new Instruction(_invokestatic, stackChange);
         instruc.addArgument(
            new ConstantArgument(
               MethodrefConstant.create(constantPool,
                                        "java/lang/Math",
                                        "round",
                                        "(D)J")));
         unit.addInstruction(instruc);
         unit.addInstruction(_l2d);
         unit.addCodeUnit(assembleDecLiteral(roundingConstant));
         unit.addInstruction(_ddiv);
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to invoke a "bootstrap" method. A
    * bootstrap method is one which retrieves a registered library or which
    * retrieves or sets the value of a registered user variable. All the
    * library objects (if any) to which the current expression refers are
    * stored within the {@link CompiledExpression#libraries} array by the
    * compiler at expression instantiation time. Likewise, all the user
    * variables (if any) to which the expression refers are stored within
    * the {@link CompiledExpression#variables} array by the compiler at the
    * same time. The indices of these objects in their respective arrays are
    * known and will not change. They are thus hard-coded as operands to the
    * bootstrap method invocations generated by this method.
    * <p>
    * The following methods are considered bootstrap methods:
    * <ul>
    * <li>{@link CompiledExpression#getLib}
    * <li>{@link CompiledExpression#getVar}
    * <li>{@link CompiledExpression#setVar}
    * </ul>
    * <p>
    * The special handling required for these methods consists of loading
    * the compiled expression's "this" reference onto the runtime stack,
    * then loading the proper constant representing the index of the library
    * or user variable to be accessed onto the stack. After this, the
    * remaining method invocation assembly is delegated to the {@link
    * #assembleRegularMethod} worker method.
    *
    * @param   type
    *          Token type of the AST which represents the method invocation;
    *          passed through to the worker method.
    * @param   logData
    *          Information required for branching within logical operations;
    *          not used here, but passed through to the worker method.
    * @param   function
    *          Defines the type of bootstrap method being invoked; also
    *          passed through to the worker method.
    * @param   units
    *          List of code units which represent any other arguments (i.e.,
    *          in addition to the hard-coded index assembled here) to the
    *          bootstrap method. Currently, the only bootstrap method which
    *          contains an additional argument is the <code>setVar</code>
    *          method. Code units for loading the "this" reference and for
    *          loading the index constant are inserted at the front of this
    *          list before it is passed on to the worker method.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleBootstrapMethod(int type,
                                            LogicalData logData,
                                            Function function,
                                            List units)
   {
      CodeUnit unit = new CodeUnit(CompiledExpression.class);
      
      // Load target object, which is the compiled expression object itself
      // (i.e., "this").
      unit.addInstruction(_aload_0);
      units.add(0, unit);
      
      // Load the implicit index parameter not reflected in the AST. This
      // is used to select the proper library or user variable. We get this
      // from the mappings we made during the preprocessing phase.
      String name = function.getSymbol();
      Integer index = null;
      switch (function.getType())
      {
         case Function.GET_LIB:
            index = (Integer) libraries.get(name);
            break;
            
         case Function.GET_VAR:
         case Function.SET_VAR:
            index = (Integer) variables.get(name);
            break;
      }
      units.add(1, assembleNumLiteral(Integer.TYPE, index.intValue()));
      
      return assembleRegularMethod(type, logData, function, units);
   }
   
   /**
    * Assemble the instructions necessary to perform a method invocation.
    * It is assumed that all instructions necessary to push the proper
    * invocation target object and any required method parameters onto the
    * runtime stack already have been assembled. These are contained within
    * the <code>units</code> list, but they may require additional conversion
    * to match the type of the required target object and the target method's
    * parameters signature.
    * <p>
    * The <code>function</code> parameter contains information about the
    * invocation target class, the return type expected from this
    * invocation (which may be different from the return type of the
    * underlying method itself and require conversion), and a reference to
    * underlying method.
    * <p>
    * The <code>units</code> list contains one or more code units. The first
    * code unit loads the invocation target object onto the stack. The
    * remaining units, if any, load each required parameter onto the stack,
    * in the order required by the method. If <code>function</code> indicates
    * that the underlying method supports variable argument lists, all
    * arguments after the number of required minimum parameters are
    * marshalled into an <code>Object</code> array, which is used as the
    * method's final parameter.
    * <p>
    * In order to convert the invocation target object and parameters to
    * the required types, instructions are assembled which perform primitive
    * wrapping/unwrapping and casting of primitive or reference types. Next,
    * instructions to invoke the method are assembled. Finally, the return
    * type is converted to the data type defined by <code>function</code>
    * as the required return type. If the <code>logData</code> argument to
    * this method is non-<code>null</code>, this conversion includes a check
    * to determine whether the runtime value returned by the invoked method
    * is <code>null</code>. If so, instructions are assembled to branch to
    * the proper destination, as defined by <code>logData</code>.
    * <p>
    * This method handles invocation of static, virtual, and interface
    * methods. There is one limitation on the invocation of static methods
    * that should be noted: a static method must always be invoked on an
    * instance of that method's declaring class. Thus, invocations of the
    * form <code>Class.staticMethod(parameter list)</code> are not permitted
    * at this time. This limitation may be removed in a future version of
    * the compiler.
    *
    * @param   type
    *          Token type of the AST which represents the method invocation.
    * @param   logData
    *          Information required for branching within logical operations
    *          when a <code>null</code> result is encountered or when the
    *          return type of the invoked method is boolean.
    * @param   function
    *          Contains details about the method being invoked.
    * @param   units
    *          List of code units which represent the invocation target
    *          object and any arguments to the method.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleRegularMethod(int type,
                                          LogicalData logData,
                                          Function function,
                                          List units)
   {
      Method method = function.getMethod();
      Class required = function.getReturnType();
      Class provided = method.getReturnType();
      CodeUnit unit = new CodeUnit(required);
      Class[] sig = method.getParameterTypes();
      Class[] castSig = function.getParameterTypes();
      int minArgs = function.getMinArgs();
      
      ExtraAst ast = currentAst;
      ExtraAst child = (ExtraAst) ast.getFirstChild();
      boolean bootstrap = false;
      switch (function.getType())
      {
         case Function.GET_LIB:
         case Function.GET_VAR:
         case Function.SET_VAR:
            bootstrap = true;
            break;
      }
      
      int size = units.size();
      for (int i = 0; i < size; i++)
      {
         if (!bootstrap)
         {
            currentAst = child;
            child = (ExtraAst) child.getNextSibling();
         }
         
         CodeUnit next = (CodeUnit) units.get(i);
         Class requiredArgType = null;
         Class providedArgType = next.getDataType();
         if (minArgs >= 0 && i == (minArgs + 1))
         {
            // Special handling for variable argument lists.
            List varArgUnits = units.subList(i, size);
            int varArgSize = size - minArgs - 1;
            unit.addCodeUnit(assembleVarArgArray(minArgs,
                                                 varArgSize,
                                                 castSig,
                                                 varArgUnits));
            
            break;
         }
         else if (i == 0)
         {
            // The invocation target object.
            requiredArgType = function.getTarget();
         }
         else
         {
            // An argument.
            requiredArgType = sig[i - 1];
            
            // If we have an explicit conversion caused by a user cast,
            // which is different than the required argument type, we must
            // convert now, then convert again to the type the method
            // expects in the following step.
            Class castType = castSig[i - 1];
            if (castType != null && providedArgType != castType)
            {
               next = convertToRequiredType(null,
                                            next,
                                            providedArgType,
                                            castType);
               providedArgType = castType;
            }
         }
         
         if (providedArgType != null && providedArgType != requiredArgType)
         {
            next = convertToRequiredType(null,
                                         next,
                                         providedArgType,
                                         requiredArgType);
         }
         unit.addCodeUnit(next);
      }
      
      // If function supports var-args, but none were provided, we still
      // need to push an empty object array onto the stack.
      if (minArgs == size - 1)
      {
         unit.addCodeUnit(assembleVarArgArray(0, 0, null, null));
      }
      
      currentAst = ast;
      
      // Analyze method parameters to determine signature string and stack
      // depth changes.
      StringBuilder buf = new StringBuilder("(");
      int len = sig.length;
      int offset;
      int[] deltas;
      boolean isStatic = ((method.getModifiers() & Modifier.STATIC) != 0);
      if (isStatic)
      {
         offset = 0;
         deltas = new int[len + 1];
      }
      else
      {
         offset = 1;
         deltas = new int[len + 2];
         deltas[0] = -1;
      }
      for (int i = 0; i < len; i++)
      {
         Class next = sig[i];
         
         // Signature string.
         buf.append(formatClassName(next, true));
         
         // Stack depth changes.
         deltas[i + offset] = ((next == Long.TYPE || next == Double.TYPE)
                               ? -2 : -1);
      }
      buf.append(")")
         .append(formatClassName(method.getReturnType(), true));
      String methDesc = buf.toString();
      
      // Figure stack change for return type:  2 for wide types, 0 for
      // void, 1 for everything else.
      if (provided == Long.TYPE || provided == Double.TYPE)
      {
         deltas[len + offset] = 2;
      }
      else if (provided == Void.TYPE)
      {
         deltas[len + offset] = 0;
      }
      else
      {
         deltas[len + offset] = 1;
      }
      
      addDebug(unit);
      
      // Now that the target and all arguments have been fixed up, invoke
      // the method.
      Class target = method.getDeclaringClass();
      String className = formatClassName(target, false);
      Instruction instruc;
      if (target.isInterface())
      {
         // Count arguments, where category 2 types count for 2 slots each.
         int count = 0;
         for (int i = 1; i <= len; i++)
         {
            count += -deltas[i];
         }
         instruc = new Instruction(_invokeinterface, deltas);
         instruc.addArgument(
            new ConstantArgument(
               InterfaceMethodrefConstant.create(constantPool,
                                                 className,
                                                 method.getName(),
                                                 methDesc)));
         instruc.addArgument(new NumericArgument(++count, Argument.BYTE));
         instruc.addArgument(new NumericArgument(0, Argument.BYTE));
      }
      else
      {
         int opCode = (isStatic ? _invokestatic : _invokevirtual);
         instruc = new Instruction(opCode, deltas);
         instruc.addArgument(
            new ConstantArgument(
               MethodrefConstant.create(constantPool,
                                        className,
                                        method.getName(),
                                        methDesc)));
      }
      unit.addInstruction(instruc);
      
      // Convert to return type required by function.  It is safe to call
      // this method even if no conversion is required, and it is necessary
      // in order to do null reference processing.
      unit = convertToRequiredType(logData, unit, provided, required);
      
      // If method returns a boolean value and must branch somewhere based
      // on its result, add branching logic here.
      if (logData != null &&
          logData.getTrueTarget() != null &&
          type == METH_BOOLEAN)
      {
         unit = assembleBranchOnBooleanResult(type,
                                              logData,
                                              unit,
                                              required);
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to instantiate an array of
    * <code>Object</code>s which is used as a variable argument list at
    * runtime. The array is of the specified size and contains as its
    * members the data values pushed onto the stack by the instructions
    * contained in the specified list of code units. If any of these code
    * units produces a primitive value, it is first wrapped in the proper
    * object, then added to the array.
    *
    * @param   minArgs
    *          Minimum number of arguments required by method.
    * @param   varSize
    *          Number of elements in the object array. If zero, an array of
    *          length zero will be created.
    * @param   signature
    *          Array of parameter types required for the method (including
    *          required arguments and variable arguments).
    * @param   units
    *          List of code units which produce the data values to be stored
    *          in the object array. If <code>size</code> is 0, this argument
    *          may be <code>null</code>.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleVarArgArray(int minArgs,
                                        int varSize,
                                        Class[] signature,
                                        List units)
   {
      CodeUnit unit = new CodeUnit(Object[].class);
      
      // Push the size of the array to be created onto the runtime stack.
      unit.addCodeUnit(assembleNumLiteral(Integer.TYPE, varSize));
      
      // Create an array of Object references.
      Instruction instruc = new Instruction(_anewarray);
      instruc.addArgument(
         new ConstantArgument(
            ClassConstant.create(constantPool, "java/lang/Object")));
      unit.addInstruction(instruc);
      
      // Populate the array by looping backwards through the function's
      // parameters and generating code to load each into the array.
      // We start at the end of the list because the last item for the
      // array will be at the top of the stack at runtime.
      for (int i = varSize - 1; i >= 0; i--)
      {
         Class required = signature[minArgs + i];
         CodeUnit argUnit = (CodeUnit) units.get(i);
         unit.addCodeUnit(assembleArrayStore(i, required, argUnit));
         
         currentAst = (ExtraAst) currentAst.getNextSibling();
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to store a single data value in an
    * object array. A primitive value is first converted to an object of the
    * appropriate wrapper type, if necessary. An object reference to the
    * array is always left at the top of the runtime stack after the array
    * store completes.
    *
    * @param   index
    *          Index at which to store the data value in the array.
    * @param   required
    *          Required data type of the item to be stored in the array.
    * @param   argUnit
    *          Code unit which produces the data value to be stored.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleArrayStore(int index,
                                       Class required,
                                       CodeUnit argUnit)
   {
      CodeUnit unit = new CodeUnit();
      
      // The array reference is at the top of the stack. Duplicate the array
      // reference so that we can use the duplicate for the array store
      // instruction, but still have the array reference at the top of the
      // stack when we're done.
      unit.addInstruction(_dup);
      
      // Lay down the index at which the item is to be stored in the array.
      unit.addCodeUnit(assembleNumLiteral(Integer.TYPE, index));
      
      // Lay down the data value to be stored. Convert it if necessary.
      CodeUnit converted = convertToRequiredType(null,
                                                 argUnit,
                                                 argUnit.getDataType(),
                                                 required);
      
      // Wrapper primitive data if necessary.
      if (required != null && required.isPrimitive())
      {
         PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(required);
         converted = convertToRequiredType(null,
                                           converted,
                                           required,
                                           info.wrapperClass);
      }
      
      unit.addCodeUnit(converted);
      
      // Store the data in the array.
      unit.addInstruction(_aastore);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to wrapper a primitive data value
    * of the specified type in an object wrapper of the appropriate type.
    * Assumes the target primitive data value is at the top of the runtime
    * stack and that an object reference of the appropriate wrapper type
    * exists just beneath it on the stack. Both values are consumed in the
    * operation.
    *
    * @param   primitive
    *          Class which represents the primitive data type to be
    *          encapsulated in a wrapper object.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleWrapperConstructor(Class primitive)
   {
      PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(primitive);
      CodeUnit unit = new CodeUnit(info.wrapperClass);
      String name = formatClassName(info.wrapperClass, false);
      
      // Assemble constructor invocation.
      int[] stackChange = { -info.width, -1 };
      StringBuilder buf = new StringBuilder();
      buf.append("(")
         .append(info.signature)
         .append(")V");
      String sig = buf.toString();
      
      // Instruction which invokes constructor.
      Instruction instruc = new Instruction(_invokespecial, stackChange);
      instruc.addArgument(new ConstantArgument(
         MethodrefConstant.create(constantPool, name, METHOD_INIT, sig)));
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to unwrap a primitive data value
    * from a containing object wrapper. The wrapper object is assumed to be
    * at the top of the runtime stack. This reference will be consumed and
    * will be replaced at the top of the stack by the primitive value after
    * the operation completes.
    *
    * @param   primitive
    *          Class which represents the primitive data type encapsulated
    *          within the wrapper object.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assemblePrimitiveUnwrap(Class primitive)
   {
      CodeUnit unit = new CodeUnit(primitive);
      addDebug(unit);
      PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(primitive);
      
      int[] stackChange = { -1, info.width };
      StringBuilder buf = new StringBuilder();
      buf.append("()")
         .append(info.signature);
      String methDesc = buf.toString();
      String name = formatClassName(info.wrapperClass, false);
      
      Instruction instruc = new Instruction(_invokevirtual, stackChange);
      instruc.addArgument(new ConstantArgument(
         MethodrefConstant.create(
            constantPool, name, info.unwrapMethod, methDesc)));
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to load an integral constant onto
    * the runtime stack. The constant is added to the constant pool if the
    * load cannot be represented by a no-argument instruction.
    *
    * @param   type
    *          Type of the integral constant to load; must be a primitive
    *          integer or long.
    * @param   num
    *          The integral value to load as a constant.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleNumLiteral(Class type, long num)
   {
      CodeUnit unit = new CodeUnit(type);
      Instruction instruc = null;
      
      if (type == Integer.TYPE)
      {
         int inum = (int) num;
         switch (inum)
         {
            case 0:
               instruc = new Instruction(_iconst_0);
               break;
            case 1:
               instruc = new Instruction(_iconst_1);
               break;
            case 2:
               instruc = new Instruction(_iconst_2);
               break;
            case 3:
               instruc = new Instruction(_iconst_3);
               break;
            case 4:
               instruc = new Instruction(_iconst_4);
               break;
            case 5:
               instruc = new Instruction(_iconst_5);
               break;
            default:
               instruc = new Instruction(_ldc_w);
               instruc.addArgument(
                  new ConstantArgument(
                     IntegerConstant.create(constantPool, inum), 
                     Argument.SHORT));
               break;
         }
      }
      else if (type == Long.TYPE)
      {
         if (num == 0L)
            instruc = new Instruction(_lconst_0);
         else if (num == 1L)
            instruc = new Instruction(_lconst_1);
         else
         {
            instruc = new Instruction(_ldc2_w);
            instruc.addArgument(
               new ConstantArgument(
                  LongConstant.create(constantPool, num)));
         }
      }
      
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to load a floating point constant
    * onto the runtime stack. The constant is added to the constant pool if
    * the load cannot be represented by a no-argument instruction. Currently,
    * only the primitive double type is supported.
    *
    * @param   num
    *          The floating point value to load as a constant.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleDecLiteral(double num)
   {
      CodeUnit unit = new CodeUnit(Double.TYPE);
      Instruction instruc = null;
      
      if (num == 0.)
         instruc = new Instruction(_dconst_0);
      else if (num == 1.)
         instruc = new Instruction(_dconst_1);
      else
      {
         instruc = new Instruction(_ldc2_w);
         instruc.addArgument(
            new ConstantArgument(
            DoubleConstant.create(constantPool, num)));
      }
      
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to load either a boolean constant
    * onto the runtime stack, or to branch to another instruction, based
    * upon the value of a boolean literal and the provided
    * <code>logData</code> information. The former is generated if the
    * <code>logData</code> parameter is <code>null</code>; else the latter
    * is generated.
    *
    * @param   type
    *          Token type of a boolean literal: either <code>BOOL_TRUE</code>
    *          or <code>BOOL_FALSE</code>.
    * @param   logData
    *          Information which indicates to which instruction to branch
    *          based upon the value of <code>type</code>.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleBoolLiteral(int type, LogicalData logData)
   {
      CodeUnit unit;
      boolean bool = (type == BOOL_TRUE);
      int opCode = (bool ? _iconst_1 : _iconst_0);
      
      if (logData != null)
      {
         ExtraAst target = (bool
                            ? logData.getTrueTarget()
                            : logData.getFalseTarget());
         
         unit = new CodeUnit(bool, target, GOTO, false);
         unit.addBranchInstruction();
      }
      else
      {
         unit = new CodeUnit(Boolean.TYPE);
         unit.addInstruction(opCode);
      }
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to load a string constant onto the
    * runtime stack. The string constant is added to the constant pool.
    *
    * @param   text
    *          The string to be loaded at runtime as a constant.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleStringLiteral(String text)
   {
      CodeUnit unit = new CodeUnit(String.class);
      
      Instruction instruc = new Instruction(_ldc_w);
      instruc.addArgument(
         new ConstantArgument(
            StringConstant.create(constantPool, text), Argument.SHORT));
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to perform an explicit test for a
    * <code>null</code> object reference. It is assumed that the reference
    * to be tested is at the top of the runtime stack. The result of the
    * test is to branch to an instruction defined by the <code>logData</code>
    * parameter, or to proceed normally to the next instruction in the
    * bytecode array.
    *
    * @param   type
    *          Token type of the AST which describes the test to perform:
    *          <code>IS_NULL</code> or <code>NOT_NULL</code>. The former
    *          tests positive if the reference is <code>null</code>, the
    *          latter tests positive if the reference is not
    *          <code>null</code>.
    * @param   logData
    *          Information which indicates to which instruction to branch
    *          based upon the value of <code>type</code> and the outcome of
    *          the runtime check.
    * @param   testUnit
    *          Code unit which contains the instructions which load the
    *          object reference to be tested.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit assembleNullTest(int type,
                                     LogicalData logData,
                                     CodeUnit testUnit)
   {
      if (testUnit.getDataType().isPrimitive())
      {
         throw new IllegalArgumentException(
            "A primitive value cannot be compared to a null reference");
      }
      
      CodeUnit unit = new CodeUnit(logData.getTrigger(),
                                   logData.getTarget(),
                                   type,
                                   false);
      unit.addCodeUnit(testUnit);
      unit.addBranchInstruction();
      
      return unit;
   }
   
   /**
    * Assemble the instructions necessary to convert the specified operand
    * to a primitive and to cast the result if necessary to the specified
    * primitive type. It is assumed that the code unit <code>operand</code>
    * generates either a primitive value or a reference to a wrapper object
    * which contains a primitive value which is of type <code>primitive</code>
    * or which can be safely cast to this type.
    *
    * @param   operand
    *          Code unit containing instructions to generate the value to be
    *          converted.
    * @param   primitive
    *          Class which represents the primitive type to which
    *          <code>operand</code> must be converted.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit operandToPrimitive(CodeUnit operand, Class primitive)
   {
      // Convert operand code unit to primitive, if necessary.
      Class operandType = operand.getDataType();
      if (!operandType.isPrimitive())
      {
         Class reqOperand = (Class) WRAPPERS.get(operandType);
         operand.addCodeUnit(assemblePrimitiveUnwrap(reqOperand));
         operand.setDataType(reqOperand);
      }
      
      // Cast operand to required primitive type, if necessary.
      if ((operandType = operand.getDataType()) != primitive)
      {
         operand = convertToRequiredType(null,
                                         operand,
                                         operandType,
                                         primitive);
      }
      
      return operand;
   }
   
   /**
    * Assemble the instructions necessary to convert a data value of one
    * type into a value of another type, optionally branching to a safe
    * location if the data value represents a <code>null</code> object
    * reference.
    * <p>
    * This method will convert:
    * <ul>
    * <li>an object reference to a primitive value;
    * <li>a primitive value to an object reference;
    * <li>an object reference of one type into a reference of another type
    *     through an explicit cast;
    * <li>a primitive value of one numeric type to another numeric type
    *     through a widening or narrowing conversion.
    * </ul>
    * <p>
    * Where possible given the data provided, multiple conversions will be
    * attempted to achieve the desired, overall conversion.
    * <p>
    * In certain cases (when the <code>logData</code> parameter is not
    * <code>null</code>), an additional check will be performed during
    * conversion to test whether the target data value is a <code>null</code>
    * object reference. If so, instructions will be added to branch
    * execution to the <code>falseTarget</code> defined by the
    * <code>logData</code> object. This has the effect of "failing" the
    * innermost logical test containing the <code>null</code> reference.
    * The result is effectively the same as if the reference had not been
    * <code>null</code>, and the enclosing logical test had evaluated to
    * <code>false</code>. In the event of such a branch, instructions may
    * be inserted to pop obsolete data values off the runtime stack in
    * order to ensure integrity of the stack at the branch destination
    * instruction.
    * <p>
    * If no conversion is necessary, the above <code>null</code> check will
    * still take place if a <code>logData</code> parameter is provided.
    *
    * @param   logData
    *          Object which contains information on how flow of control
    *          should proceed based upon the result of the null object
    *          reference check.
    * @param   unit
    *          Code unit which contains the instructions which will generate
    *          the data value needing conversion.
    * @param   provided
    *          Class which represents the data type to be converted.
    * @param   required
    *          Class which represents the data type to which the data value
    *          should be converted.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation.
    */
   private CodeUnit convertToRequiredType(LogicalData logData,
                                          CodeUnit unit,
                                          Class provided,
                                          Class required)
   {
      CodeUnit outer = new CodeUnit(required);
      
      boolean providedPrim = (provided != null && provided.isPrimitive());
      boolean requiredPrim = (required != null && required.isPrimitive());
      boolean primToRef = (providedPrim && !requiredPrim);
      boolean refToPrim = (requiredPrim && !providedPrim);
      boolean wrapToWrap = (provided != null &&
                            required != null &&
                            !requiredPrim &&
                            !providedPrim &&
                            provided != required &&
                            Number.class.isAssignableFrom(provided) &&
                            Number.class.isAssignableFrom(required));
      
      // If we have a primitive and we need an object wrapper, push a
      // new instance onto the stack.
      if (primToRef || wrapToWrap)
      {
         Class type = required;
         if (!WRAPPERS.containsKey(required))
         {
            PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(provided);
            type = info.wrapperClass;
         }
         outer.addCodeUnit(assembleNewObjectReference(type));
      }
      
      outer.addCodeUnit(unit);
      
      CodeUnit finishUnit = new CodeUnit(required);
      
      // Perform a null check on the result if necessary.
      if (logData != null && !providedPrim)
      {
         // Create a bogus AST which represents the destination of a jump
         // in the case where the object reference under test is determined
         // to be non-null.
         ExtraAst notNullDest = new ExtraAst();
         notNullDest.setType(NOT_NULL);
         notNullDest.setText("null check dest");
         notNullDest.setExtra(finishUnit);
         
         // Set up a branching code unit to branch to the finishing code
         // unit in the event the object reference in non-null.
         CodeUnit notNullUnit = new CodeUnit(true,
                                             notNullDest,
                                             NOT_NULL,
                                             false);
         notNullUnit.addInstruction(_dup);
         notNullUnit.addBranchInstruction();
         
         // Now set up the code unit which will jump to the closest false
         // target if the above null check failed.
         CodeUnit gotoUnit = new CodeUnit(true,
                                          logData.getFalseTarget(),
                                          GOTO,
                                          false);
         
         // Before the actual goto instruction, insert a special code unit
         // which will insert as many pops as are necessary to clean up the
         // stack to the proper height before we jump.
         CodeUnit popUnit = new CodeUnit();
         popUnit.setStackPopper(true);
         gotoUnit.addCodeUnit(popUnit);
         
         // Add the actual goto instruction.
         gotoUnit.addBranchInstruction();
         
         // Finally, add both the null check and goto units to the
         // code unit for this method invocation.
         outer.addCodeUnit(notNullUnit);
         outer.addCodeUnit(gotoUnit);
      }
      
      // If we have a primitive and we need an object wrapper, invoke the
      // wrapper constructor.
      if (primToRef)
      {
         // Do we need to cast to another primitive type before wrapping?
         Class reqPrimType = (Class) WRAPPERS.get(required);
         CodeUnit cast = assemblePrimitiveConversion(reqPrimType, provided);
         if (cast != null)
         {
            finishUnit.addCodeUnit(cast);
         }
         
         // Wrap the primitive in an object.
         if (reqPrimType == null)
         {
            reqPrimType = provided;
         }
         finishUnit.addCodeUnit(assembleWrapperConstructor(reqPrimType));
      }
      // If we have an object wrapper and we need a primitive, unwrap the
      // primitive.
      else if (refToPrim)
      {
         // Do we need to cast to another reference type before unwrapping?
         PrimitiveInfo info = (PrimitiveInfo) PRIMITIVES.get(required);
         Class reqRefType = info.wrapperClass;
         if (provided != reqRefType && reqRefType.isAssignableFrom(provided))
         {
            finishUnit.addCodeUnit(assembleExplicitCast(reqRefType));
         }
         
         Class prvPrimType = (Class) WRAPPERS.get(provided);
         if (prvPrimType != null)
         {
            // Unwrap the primitive.
            finishUnit.addCodeUnit(assemblePrimitiveUnwrap(prvPrimType));
            
            // Do we need to cast to another primitive type to meet the
            // requirement?
            CodeUnit cast = assemblePrimitiveConversion(required, prvPrimType);
            if (cast != null)
            {
               finishUnit.addCodeUnit(cast);
            }
         }
         else
         {
            // Provided type was not a primitive wrapper. Try to cast it to
            // one, then unwrap.
            finishUnit.addCodeUnit(assembleExplicitCast(reqRefType));
            finishUnit.addCodeUnit(assemblePrimitiveUnwrap(required));
         }
      }
      // Convert to the required return type if necessary.
      else if (required != provided && providedPrim == requiredPrim)
      {
         if (requiredPrim)
         {
            // Handle primitive conversion.
            finishUnit.addCodeUnit(assemblePrimitiveConversion(required,
                                                               provided));
         }
         else if (provided != null &&
                  (required.isAssignableFrom(provided) ||
                   provided.isAssignableFrom(required)))
         {
            // Handle simple reference cast.
            finishUnit.addCodeUnit(assembleExplicitCast(required));
         }
         else if (wrapToWrap)
         {
            // provided and required both represent different concrete
            // subclasses of java.lang.Number. Try to unwrap provided type
            // and rewrap primitive in required type.
            Class prvPrimType = (Class) WRAPPERS.get(provided);
            Class reqPrimType = (Class) WRAPPERS.get(required);
            if (prvPrimType != null && reqPrimType != null)
            {
               // Unwrap the primitive.
               finishUnit.addCodeUnit(
                  assemblePrimitiveUnwrap(prvPrimType));
               
               // Do we need to cast to another primitive type to meet the
               // requirement?
               CodeUnit cast = assemblePrimitiveConversion(reqPrimType,
                                                           prvPrimType);
               if (cast != null)
               {
                  finishUnit.addCodeUnit(cast);
               }
               
               // Rewrap in the new reference type.
               finishUnit.addCodeUnit(
                  assembleWrapperConstructor(reqPrimType));
            }
         }
      }
      
      outer.addCodeUnit(finishUnit);
      
      return outer;
   }
   
   /**
    * Assemble the instructions necessary to convert the data value at the
    * top of the runtime stack from primitive type <code>provided</code> to
    * primitive type <code>required</code>. This conversion deals with
    * numeric values only, and performs both narrowing and widening
    * conversions. Currently, only conversions between integers, longs, and
    * doubles are supported.
    * <p>
    * At a future date, this method may add instructions to check for
    * precision loss if certain runtime flags are set in the {@link
    * Expression} object associated with this compilation.
    *
    * @param   required
    *          Required primitive type.
    * @param   provided
    *          Provided primitive type.
    *
    * @return  Code unit containing the appropriate instructions for this
    *          operation, or <code>null</code> if either argument is
    *          <code>null</code> or does not represent a primitive type, or
    *          if no conversion is necessary.
    */
   private CodeUnit assemblePrimitiveConversion(Class required,
                                                Class provided)
   {
      if (required == null || !required.isPrimitive() ||
          provided == null || !provided.isPrimitive())
      {
         return null;
      }
      
      int opCode = -1;
      
      if (provided == Integer.TYPE)
      {
         if (required == Long.TYPE)
         {
            opCode = _i2l;
         }
         else if (required == Double.TYPE)
         {
            opCode = _i2d;
         }
      }
      else if (provided == Long.TYPE)
      {
         if (required == Integer.TYPE)
         {
            opCode = _l2i;
         }
         else if (required == Double.TYPE)
         {
            opCode = _l2d;
         }
      }
      else if (provided == Double.TYPE)
      {
         if (required == Integer.TYPE)
         {
            opCode = _d2i;
         }
         else if (required == Long.TYPE)
         {
            opCode = _d2l;
         }
      }
      
      if (opCode < 0)
      {
         return null;
      }
      
      CodeUnit unit = new CodeUnit(required);
      
      unit.addInstruction(opCode);
      
      return unit;
   }
   
   /**
    * Assemble the instruction necessary to cast the object reference at the
    * top of the runtime stack to the required object type.
    *
    * @param   required
    *          Class to which the target reference will be cast.
    *
    * @return  Code unit containing the appropriate instruction for this
    *          operation.
    */
   private CodeUnit assembleExplicitCast(Class required)
   {
      CodeUnit unit = new CodeUnit(required);
      addDebug(unit);
      Instruction instruc = new Instruction(_checkcast);
      instruc.addArgument(
         new ConstantArgument(
            ClassConstant.create(constantPool,
                                 formatClassName(required, false))));
      unit.addInstruction(instruc);
      
      return unit;
   }
   
   /**
    * Given a class, format the class' name into a form which can be used
    * within the class file constant pool. For primitive types, this is the
    * <i>BaseType</i> character identifier specified in the table on page 101
    * of <u>The Java(TM) Virtual Machine Specification Second Edition</u>.
    * For classes, this is the fully qualified class name, with all instances
    * of the '.' character replaced with the '/' character.
    * <p>
    * Optionally, enclose class names within a leading 'L' character and a
    * trailing ';' character, for use within method signature specifications.
    *
    * @param   cls
    *          The class whose name is to be formatted.
    * @param   encloseRefs
    *          <code>true</code> to wrapper the class name for method
    *          signature use as described above; else <code>false</code>.
    *
    * @return  Formatted class name.
    */
   private static String formatClassName(Class cls, boolean encloseRefs)
   {
      String name = null;
      if (cls.isPrimitive())
      {
         name = ((PrimitiveInfo) PRIMITIVES.get(cls)).signature;
      }
      else
      {
         name = cls.getName().replace('.', '/');
         if (encloseRefs && !cls.isArray())
         {
            StringBuilder buf = new StringBuilder("L");
            buf.append(name)
               .append(";");
            name = buf.toString();
         }
      }
      
      return name;
   }
   
   /**
    * Parse a long value from the text of an AST, using the specified radix.
    *
    * @param   ast
    *          AST containing the text to be parsed.
    * @param   radix
    *          Radix value to use for the parse operation.
    *
    * @return  The numeric value parsed from the AST's text.
    *
    * @throws  CompilerException
    *          if the AST's text could not be parsed into an integral number.
    */
   private static long parseLong(ExtraAst ast, int radix)
   throws CompilerException
   {
      long num = 0L;
      try
      {
         num = Long.parseLong(ast.getText(), radix);
      }
      catch (NumberFormatException exc)
      {
         throw new CompilerException("Invalid numeric literal", exc);
      }
      
      return num;
   }
   
   /**
    * Parse a double value from the text of an AST.
    *
    * @param   ast
    *          AST containing the text to be parsed.
    *
    * @return  The numeric value parsed from the AST's text.
    *
    * @throws  CompilerException
    *          if the AST's text could not be parsed into a floating point
    *          number.
    */
   private static double parseDouble(ExtraAst ast)
   throws CompilerException
   {
      double num = 0.;
      try
      {
         num = Double.parseDouble(ast.getText());
      }
      catch (NumberFormatException exc)
      {
         throw new CompilerException("Invalid decimal literal", exc);
      }
      
      return num;
   }
   
   /**
    * Detect whether the given AST represents a method invocation or not.
    *
    * @param   ast
    *          AST to be tested.
    *
    * @return  <code>true</code> if AST represents a method invocation, else
    *          <code>false</code>.
    */
   private static boolean isMethodInvocation(ExtraAst ast)
   {
      if (ast == null)
      {
         return false;
      }
      
      switch (ast.getType())
      {
         case METH_BOOLEAN:
         case METH_INT:
         case METH_LONG:
         case METH_DOUBLE:
         case METH_OBJ:
            return true;
         case CAST:
            return isMethodInvocation((ExtraAst) ast.getFirstChild());
      }
      
      return false;
   }
   
   /**
    * Produce an expression compilation report which describes the internal
    * state of the compiler and the expression being compiled. This report
    * includes the name of the compiled class, the infix expression text,
    * a postfix representation of the expression in AST form, with each
    * tree level indented for readability, the result and return types of
    * the expression, and all referenced libraries and variables.
    * <p>
    * If the {@link #debug} flag is set to true (typically only when tracking
    * down compiler defects), the class file generated for the current
    * expression (if compilation proceeded that far without error) is written
    * to the current directory.
    * <p>
    * Note: this method is only called when the {@link Expression} object
    * associated with the compiler has the <code>RT_DEBUG</code> flag set.
    */
   private void dumpState()
   {
      System.out.println("-----------------------------");
      System.out.println("EXPRESSION COMPILATION REPORT");
      System.out.println("-----------------------------");
      System.out.println("Class:  " + nextClassName());
      System.out.println("Infix:  " + expression.getInfix());
      System.out.print("Postfix:");
      if (tree != null)
      {
         System.out.println();
         BaseAST.setVerboseStringConversion(true,
                                            ExpressionParser._tokenNames);
         ExpressionParser.visit(tree, 0);
         System.out.println("Result type         :  " + resultType);
         System.out.println("Return type         :  " + returnType);
      }
      else
      {
         System.out.println("  <unavailable>");
      }
      
      System.out.println("Referenced Libraries:  " + libraries);
      System.out.println("Referenced Variables:  " + variables);
      
      System.out.println("-------- END REPORT ---------");
      
      if (debug)
      {
         // Write class file to current directory.
         DataOutputStream out = null;
         try
         {
            File file = new File(nextClassName() + ".class");
            FileOutputStream fos = new FileOutputStream(file.getName());
            BufferedOutputStream bos = new BufferedOutputStream(fos);
            out = new DataOutputStream(bos);
            
            classFile.write(out);
            out.flush();
         }
         catch (IOException exc)
         {
            exc.printStackTrace();
         }
         finally
         {
            if (out != null)
            {
               try
               {
                  out.close();
               }
               catch (IOException exc)
               {
               }
            }
         }
      }
   }
   
   /**
    * Helper class which stores information to assist when assembling branch
    * instructions for logical tests. Instances of this class are attached to
    * nodes in the expression's AST during the compiler's preprocessing step.
    * When these nodes (which represent logical tests which must evaluate to
    * <code>true</code> or <code>false</code>) are assembled into bytecode
    * instructions, this object is consulted to determine whether the flow
    * of control should proceed normally to the next instruction, or whether
    * it should branch to another instruction, effectively bypassing sets of
    * instructions made irrelevant at runtime.
    * <p>
    * An instance of this class defines a <code>trueTarget</code>, which is
    * the AST which logically represents the destination instruction in the
    * event the underlying logical test evaluates to <code>true</code>. It
    * also defines a <code>falseTarget</code>, which is the AST which
    * logically represents the destination instruction in the event the
    * underlying test evaluates to <code>false</code>. Since branching
    * instructions which test a condition only allow a jump in one case or
    * the other (otherwise execution proceeds normally to the next
    * instruction in the bytecode array), only one target can be active.
    * The active target is defined by the <code>trigger</code> flag. The
    * preprocess step of compilation does the hard work of determining how
    * to apply instances of this class to AST nodes appropriately, based
    * upon some basic rules of boolean algebra.
    * <p>
    * An <code>extra</code> field exists to hold the data previously stored
    * in the <code>extra</code> field of the associated AST. This is
    * necessary because we usurp that field in the AST to store an instance
    * of this class, but we do not want to lose the original data, which
    * is still needed after preprocessing the expression AST.
    * <p>
    * Finally, this class defines a <code>topOfBranch</code> flag, which is
    * used to denote that the associated AST represents the highest level
    * logical operation in a chain. This may be the root of the entire
    * expression, or it may be the base of an AST branch which resolves to
    * a boolean value, but whose parent is not a logical operator (for
    * example, the parent could represent a method call which accepts a
    * boolean argument, but does not have a boolean return type). In either
    * case the <code>topOfBranch</code> designation is used to determine
    * that an additional pair of instructions is necessary as the targets
    * of the top level AST: one that pushes a <code>true</code> result onto
    * the runtime stack, and one that pushes a <code>false</code> result
    * onto the stack. These then become the targets of the top level, logical
    * operation, since no other natural jump target exists.
    *
    * @see  #preprocess
    * @see  #assemble
    * @see  CodeUnit.BranchData
    */
   private static class LogicalData
   {
      /** AST which defines the destination instruction for a true result */
      private ExtraAst trueTarget = null;
      
      /** AST which defines the destination instruction for a false result */
      private ExtraAst falseTarget = null;
      
      /** Data object originally stored in the extra field of an AST */
      private Object extra = null;
      
      /** Logical test result which will trigger a jump */
      private boolean trigger = false;
      
      /** Flag indicating AST represents top of chain of logical operations */
      private boolean topOfBranch = false;
      
      /**
       * Constructor which defines this object's trigger condition and stores
       * the "extra" data previously stored in this object's new host AST.
       *
       * @param   trigger
       *          <code>true</code> to jump to <code>trueTarget</code> on a
       *          <code>true</code> result; <code>false</code> to jump to
       *          <code>falseTarget</code> on a <code>false</code> result.
       * @param   extra
       *          "Extra" data previously stored in this object's new host
       *          AST.
       */
      LogicalData(boolean trigger, Object extra)
      {
         this.trigger = trigger;
         this.extra = extra;
      }
      
      /**
       * Get the trigger condition for this object.
       *
       * @return  <code>true</code> if we jump on <code>true</code>;
       *          <code>false</code> if we jump on <code>false</code>.
       */
      boolean getTrigger()
      {
         return trigger;
      }
      
      /**
       * Report whether the underlying AST represents the base of a branch
       * of logical operations.
       *
       * @return  <code>true</code> if AST is such a node, else
       *          <code>false</code>.
       */
      boolean isTopOfBranch()
      {
         return topOfBranch;
      }
      
      /**
       * Set whether the underlying AST represents the base of a branch
       * of logical operations.
       *
       * @param   topOfBranch
       *          <code>true</code> if AST is such a node, else
       *          <code>false</code>.
       */
      void setTopOfBranch(boolean topOfBranch)
      {
         this.topOfBranch = topOfBranch;
      }
      
      /**
       * Get the "extra" data previously associated with the underlying AST.
       *
       * @return  AST's original "extra" object.
       */
      Object getExtra()
      {
         return extra;
      }
      
      /**
       * Get the AST which represents the target to which the underlying
       * AST's compiled code will jump when the trigger result is produced.
       * A jump will only trigger on one result, <code>true</code> or
       * <code>false</code>. The opposite result will cause execution to
       * proceed normally to the next instruction in the bytecode array
       * instead of causing a jump.
       *
       * @return  The <code>trueTarget</code> defined for this object if
       *          its <code>trigger</code> is <code>true</code>; else its
       *          <code>falseTarget</code>.
       */
      ExtraAst getTarget()
      {
         return (trigger ? trueTarget : falseTarget);
      }
      
      /**
       * Get the AST which represents this object's logical destination if
       * a <code>true</code> result is achieved. This transition represents
       * a branch only if <code>trigger</code> is <code>true</code>;
       * otherwise it represents normal, contiguous execution.
       *
       * @return  Target AST for a <code>true</code> result.
       */
      ExtraAst getTrueTarget()
      {
         return trueTarget;
      }
      
      /**
       * Set the AST which represents this object's logical destination if
       * a <code>true</code> result is achieved.
       *
       * @param   trueTarget
       *          Target AST for a <code>true</code> result.
       */
      void setTrueTarget(ExtraAst trueTarget)
      {
         this.trueTarget = trueTarget;
      }
      
      /**
       * Get the AST which represents this object's logical destination if
       * a <code>false</code> result is achieved. This transition represents
       * a branch only if <code>trigger</code> is <code>false</code>;
       * otherwise it represents normal, contiguous execution.
       *
       * @return  Target AST for a <code>false</code> result.
       */
      ExtraAst getFalseTarget()
      {
         return falseTarget;
      }
      
      /**
       * Set the AST which represents this object's logical destination if
       * a <code>false</code> result is achieved.
       *
       * @param   falseTarget
       *          Target AST for a <code>false</code> result.
       */
      void setFalseTarget(ExtraAst falseTarget)
      {
         this.falseTarget = falseTarget;
      }
      
      /**
       * Generate a string representation of this object for debugging
       * purposes.
       *
       * @return  String representing this object's internal state.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder();
         
         if (topOfBranch)
         {
            buf.append("{TOP} ");
         }
         
         buf.append("T: ")
            .append(trueTarget != null ? trueTarget.getText() : "null")
            .append("; F: ")
            .append(falseTarget != null ? falseTarget.getText() : "null")
            .append("; trigger = ")
            .append(trigger)
            .append(" (")
            .append(extra)
            .append(")");
            
         return buf.toString();
      }
   }
   
   /**
    * A helper class which stores information about the various primitive
    * data types supported by the expression compiler. This includes:
    * <ul>
    * <li>the width (in 32-bit slots) of a data value of the type;
    * <li>the signature of the type (i.e., the <i>BaseType</i> character
    *     identifier specified in the table on page 101 of <u>The Java(TM)
    *     Virtual Machine Specification Second Edition</u>);
    * <li>the class used to wrapper the data type (e.g.,
    *     <code>java.lang.Integer</code> for type <code>int</code>);
    * <li>the name of the method used to "unwrap" the primitive data value
    *     from an instance of the associated wrapper class (e.g.,
    *     <code>intValue</code> for <code>java.lang.Integer</code>).
    * </ul>
    * <p>
    * This information is stored in the static {@link #PRIMITIVES} map, such
    * that it can be looked up when assembling bytecode instructions (in
    * place of <code>switch</code> statements).
    */
   private static class PrimitiveInfo
   {
      /** Width (in 32-bit slots) of a data value of this primitive type */
      int width = 0;
      
      /** BaseType identifier for this primitive type */
      String signature = null;
      
      /** Class used to wrapper data values of this primitive type */
      Class wrapperClass = null;
      
      /** Name of wrapper class method used to unwrap primitive value */
      String unwrapMethod = null;
      
      /**
       * Convenience constructor which stores all required data fields.
       *
       * @param   width
       *          Data value width.
       * @param   signature
       *          BaseType identifier.
       * @param   wrapperClass
       *          Wrapper class
       * @param   unwrapMethod
       *          Name of unwrap method.
       */
      PrimitiveInfo(int width,
                    String signature,
                    Class wrapperClass,
                    String unwrapMethod)
      {
         this.width = width;
         this.signature = signature;
         this.wrapperClass = wrapperClass;
         this.unwrapMethod = unwrapMethod;
      }
   }
}