/*
** Module   : ConstantPool.java
** Abstract : 
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 002 EVL 20160222              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;


import java.io.*;
import java.util.*;


/**
 * This class represents the constant pool found within the Java class
 * file format (see <u>The Java Virtual Machine Specification</u>, Second
 * Edition).  Each instance of this class contains a map of constants,
 * sorted according to the natural order in which the various constant types
 * should appear in the pool.
 * <p>
 * Classes which access the constant pool to add constants must do so via
 * the <code>get()</code> method.  This method will accept an instance of
 * a <code>Constant</code> subclass.  If an equivalent object alreay exists
 * in the pool, it is returned.  If not, the specified object is added to
 * the pool, then returned from the method.  This mechanism preserves the
 * integrity of the pool, such that two equivalent constants are never
 * present.
 * <p>
 * The constant pool is written to an output stream as part of the class
 * file compilation.  This process causes all constants to be indexed,
 * during which each is assigned a unique index associated with its
 * ordered position within the pool.  Note that <code>longs</code> and
 * <code>doubles</code> take up two slots each;  all other constants
 * take up one slot.  It is these indices which are used to refer to a
 * particular constant from within the bytecode of Java methods.
 */
final class ConstantPool implements WritableData
{

   /** Map of all <code>Constant</code> subclass instances. */
   private SortedMap constantMap = new TreeMap();

   /** Size (in 32-bit slots) of the constant pool. */
   private int size;


   /**
    * Default constructor.
    */
   ConstantPool()
   {
   }


   /** 
    * This method checks whether the specified constant is already in the
    * pool.  If so, the instance in the pool is returned and the specified
    * constant is ignored.  If not, the specified constant is added to the
    * pool and is returned.
    *
    * @param   constant
    *          Instance of the <code>Constant</code> subclass.  This
    *          parameter is both used as a key to check for an equivalent
    *          constant already in the pool, and as a constant instance
    *          to add to the pool if an equivalent one is not yet present.
    *
    * @return  The specified <code>constant</code>, if it was added to the
    *          pool, else the equivalent constant which was already in the
    *          pool.
    */
   Constant get(Constant constant)
   {

      Constant cached = (Constant) constantMap.get(constant);
      if (cached == null)
      {
/*         if (constant instanceof Utf8Constant) {
            System.out.println("------------------------------");
            System.out.println("Adding " + constant);
            System.out.println("Current UTF8 keys:");
            Iterator iter = constantMap.keySet().iterator();
            while (iter.hasNext()) {
               Comparable next = (Comparable) iter.next();
               int comp = constant.compareTo(next);
               if (next instanceof Utf8Constant) {
                  System.out.println("   " + comp + "   " + next);
               } // end if
            } // end while
            System.out.println("Current UTF8 values:");
            iter = constantMap.values().iterator();
            while (iter.hasNext()) {
               Object next = iter.next();
               if (next instanceof Utf8Constant) {
                  System.out.println("   " + next);
               } // end if
            } // end while
         } */
         cached = constant;
         constantMap.put(cached, cached);
      } // end if

      return cached;

   }


   /** 
    * This method is called just before writing the pool to an output
    * stream, presumably to create a Java class file.  It iterates through
    * each constant in the pool, assigning each its index in terms of
    * 32-bit slots.  Note that the indices assigned here will not
    * necessarily match each constant's position in the pool, as some
    * constants (i.e., doubles and longs) will take up two slots rather
    * than one.
    */
   private void indexPool()
   {

      Iterator iter = constantMap.values().iterator();
      int i = 1;
      while (iter.hasNext())
      {
         Constant next = (Constant) iter.next();
         next.setCpIndex(i);
         i += next.getWidth();
      } // end while

      size = i;

   }


   /** 
    * This method indexes the constant pool, which is already sorted in the
    * natural order of its constant objects, then writes the constant pool
    * data to the specified output stream, presumably as part of the
    * creation of a Java class file.
    *
    * @param   out
    *          Output stream to which to write constant pool data.
    *
    * @exception  IOException
    *             If any error is encountered while writing the data.
    */
   public void write(DataOutput out) throws IOException {

      indexPool();

      // Write out constant pool count (plus 1).
      out.writeShort(size);

      // Write each entry to the output stream.
      Iterator iter = constantMap.values().iterator();
      while (iter.hasNext())
      {
         Constant next = (Constant) iter.next();
         next.write(out);
      } // end while

   }


   /**
    * Test harness entry point.
    *
    * @param   args
    *          Main function argumens.
    */
   public static void main(String args[])
   {

      try
      {

         ConstantPool pool = new ConstantPool();
         FileOutputStream fos = new FileOutputStream(args[0]);
         BufferedOutputStream bos = new BufferedOutputStream(fos);
         DataOutputStream out = new DataOutputStream(bos);

         InterfaceMethodrefConstant.create(
            pool, "com/goldencode/util/ArithmeticExpression",
            "compute", "(Lcom/goldencode/util/VariableManager)D;");
         ClassConstant.create(pool, "com/goldencode/ProtAnalyzer/ConcreteAE000000");
         ClassConstant.create(pool, "com/goldencode/util/ArithmeticExpression");
         ClassConstant.create(pool, "com/goldencode/util/VariableManager");
         ClassConstant.create(pool, "java/lang/Object");
         pool.write(out);

         out.close();
         bos.close();
         fos.close();

      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      } // end try-catch

   }


}