/*
** Module   : NumberedException.java
** Abstract : Exception type which supports a queryable error number
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------------------------------
** 001 ECF 20051104   @23243 Created initial version. Exception type which supports a queryable
**                           error number in addition to regular Exception features.
** 002 SVL 20140709          Added prefix field.
** 003 VIG 20140709          Changed field number to numbers and added additional constructor and
**                           getter for it.
** 004 ECF 20141001          Fixed regressions introduced by 003. Removed use of Apache commons.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j;

/**
 * Exception type which supports a queryable error number. The number may
 * be set only upon construction and can be queried using the {@link
 * #getNumber} method.  The default error number is <code>0</code>, in case
 * a constructor variant is used which does not accept an error number.
 * These are provided primarily for compatibility with the superclass, so
 * that subclasses need not expressly set an error number.
 */
public class NumberedException
extends Exception
{
   /** Errors numbers in order of appearance */
   private int[] numbers;

   /** Determines if the prefix should be displayed for the error message. */
   private boolean prefix = true;
   
   /**
    * Constructor which sets the message text only.  The default error
    * number is used.
    *
    * @param   message
    *          Error message. Corresponds to the last error number in the list. 
    */
   public NumberedException(String message)
   {
      this(message, 0);
   }
   
   /**
    * Constructor which sets message text and error number.
    *
    * @param   message
    *          Error message.
    * @param   number
    *          Error number.
    */
   public NumberedException(String message, int number)
   {
      this(message, number, (Throwable) null);
   }
   
   /**
    * Constructor which sets message text and errors numbers.
    *
    * @param   message
    *          Error message.
    * @param   numbers
    *          Contains numbers of errors with previous during one statement execution.
    */
   public NumberedException(String message, int[] numbers)
   {
      super(message, null);
      
      int len = numbers.length;
      this.numbers = new int[len];
      System.arraycopy(numbers, 0, this.numbers, 0, len);
   }

   /**
    * Constructor which sets message text and error number.
    *
    * @param   message
    *          Error message.
    * @param   number
    *          Error number.
    * @param   prefix
    *          Determines if the prefix should be displayed for the error message.
    */
   public NumberedException(String message, int number, boolean prefix)
   {
      this(message, number, prefix, null);
   }
   
   /**
    * Constructor which sets message text and error number, and which chains
    * a root cause <code>Throwable</code>.
    *
    * @param   message
    *          Error message.
    * @param   number
    *          Error number.
    * @param   rootCause
    *          Root cause exception/error.
    */
   public NumberedException(String message, int number, Throwable rootCause)
   {
      super(message, rootCause);
      
      this.numbers = new int[] { number };
   }
   
   /**
    * Constructor which sets message text and error number, and which chains
    * a root cause <code>Throwable</code>.
    *
    * @param   message
    *          Error message.
    * @param   number
    *          Error number.
    * @param   prefix
    *          Determines if the prefix should be displayed for the error message.
    * @param   rootCause
    *          Root cause exception/error.
    */
   public NumberedException(String message, int number, boolean prefix, Throwable rootCause)
   {
      super(message, rootCause);
      
      this.numbers = new int[] { number };
      this.prefix = prefix;
   }
   
   /**
    * Constructor which sets message text and which chains a root cause
    * <code>Throwable</code>.  The default error number is used.
    *
    * @param   message
    *          Error message.
    * @param   rootCause
    *          Root cause exception/error.
    */
   public NumberedException(String message, Throwable rootCause)
   {
      super(message, rootCause);
   }
   
   /**
    * Constructor which chains a root cause <code>Throwable</code>, but
    * provides no message text.  The default error number is used.
    *
    * @param   rootCause
    *          Root cause exception/error.
    */
   public NumberedException(Throwable rootCause)
   {
      super(rootCause);
   }
   
   /**
    * Get the error number associated with this exception.
    *
    * @return  Error number (<code>0</code> if never explicitly set to a
    *          different value at construction).
    */
   public int getNumber()
   {
      return numbers == null ? 0 : numbers[numbers.length - 1];
   }
   
   /**
    * Determines if the prefix should be displayed for the error message.
    *
    * @return <code>true</code> if the prefix should be displayed.
    */
   public boolean isPrefix()
   {
      return prefix;
   }
   
   /**
    * Get the errors numbers containing - current error number and previous
    *
    * @return  array of error numbers (list with one element <code>0</code> if never explicitly 
    *          set to a different value at construction).
    */
   public int[] getNumbers()
   {
      return numbers;
   }
}