/*
** Module   : BulkAuthModeUpdateView.java
** Abstract : BulkAuthModeUpdateView represents a custom accounts update dialog to modify all
**            target accounts in a batch mode.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home;

import java.util.Map;
import java.util.logging.Level;

import org.gwtbootstrap3.client.ui.Button;
import org.gwtbootstrap3.client.ui.ListBox;
import org.gwtbootstrap3.client.ui.Modal;

import com.goldencode.p2j.admin.TaggedName;
import com.goldencode.p2j.admin.client.application.home.accounts.AccountType;
import com.goldencode.p2j.admin.client.application.home.accounts.AccountsConstants;
import com.goldencode.p2j.admin.client.application.home.accounts.AuthModes;
import com.goldencode.p2j.admin.client.widget.dialog.ModalViewWithUiHandlers;
import com.google.gwt.core.client.GWT;
import com.google.gwt.event.dom.client.ChangeEvent;
import com.google.gwt.event.dom.client.ChangeHandler;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.uibinder.client.UiBinder;
import com.google.gwt.uibinder.client.UiField;
import com.google.gwt.uibinder.client.UiHandler;
import com.google.gwt.user.client.ui.Widget;
import com.google.inject.Inject;
import com.gwtplatform.mvp.client.UiHandlers;


/**
 * BulkAuthModeUpdateView represents a custom accounts update dialog to modify all target accounts
 * in a batch mode.
 */
public class BulkAuthModeUpdateView
extends ModalViewWithUiHandlers<UiHandlers>
{
   /** The constants resources used by accounts package */ 
   private static final AccountsConstants CONSTANTS = GWT.create(AccountsConstants.class);
   
   /** The accounts targets map */
   private static final Map<String, String> ACCOUNTS_TARGETS = CONSTANTS.getAccountsTargets();

   /**
    * GWT UI creator
    */
   interface Binder
   extends UiBinder<Widget, BulkAuthModeUpdateView>
   {}
   
   /** The underlined modal dialog */
   @UiField
   Modal modal;
   
   /** The list box of account types */
   @UiField
   ListBox accountTypes;
   
   /** The list box of authentication modes */
   @UiField
   ListBox modes;
   
   /** The list box of custom authentication plugins */
   @UiField
   ListBox plugins;
   
   /** The Bulk Update button */
   @UiField
   Button updateAction;
   
   /** The Close button */
   @UiField
   Button closeAction;
   
   /** The plugins model */
   private TaggedName[] pluginValues;

   /**
    * Creates this view.
    * 
    * @param    binder
    *           The injected GWT UI creator
    */
   @Inject
   public BulkAuthModeUpdateView(BulkAuthModeUpdateView.Binder binder)
   {
      initWidget(binder.createAndBindUi(this));
      for(AccountType type : AccountType.values())
      {
         String key = type.name();
         accountTypes.addItem(ACCOUNTS_TARGETS.get(key), key);
      }
      
      accountTypes.addChangeHandler(new ChangeHandler()
      {
         @Override
         public void onChange(ChangeEvent event)
         {
            updateModeListBox();
         }
      });
      
      updateModeListBox();
      
      plugins.setEnabled(AuthModes.CUST_AUTH.name().equals(modes.getSelectedValue()));
      
      modes.addChangeHandler(new ChangeHandler()
      {
         @Override
         public void onChange(ChangeEvent event)
         {
            plugins.setEnabled(AuthModes.CUST_AUTH.name().equals(modes.getSelectedValue()));
         }
      });
   }

   /**
    * The Bulk Update button click handler.
    * 
    * @param    e
    *           The click event
    */
   @UiHandler("updateAction")
   void onUpdateAction(ClickEvent e)
   {
      logger.log(Level.INFO, "onUpdateAction");
   }
   
   /**
    * The Close button click handler.
    * 
    * @param    e
    *           The click event
    */
   @UiHandler("closeAction")
   void onCancelAction(ClickEvent e)
   {
      logger.log(Level.INFO, "onCloseAction");
   }
   
   /**
    * Sets the custom authentication plugins.
    * 
    * @param    pluginValues
    *           The array of custom authentication plugins
    */
   public void setCustomAuthPlugins(TaggedName[] pluginValues)
   {
      plugins.clear();
      if (pluginValues == null)
      {
         return;
      }
      // ListBox doesn't have model
      this.pluginValues = pluginValues;
      for (TaggedName value : this.pluginValues)
      {
         plugins.addItem(value.getTag(), value.getName());
      }
   }
   
   /**
    * Setups the authentication mode list box.
    */
   private void updateModeListBox()
   {
      modes.clear();
      for(AuthModes value : AuthModes.values())
      {
         if (value == AuthModes.SERVER_AUTH &&
                  AccountType.GROUP.name().equals(accountTypes.getSelectedValue()))
         {
            modes.addItem(AuthModes.getTranslation("GROUP_SERVER_AUTH"), value.name());
         }
         else
         {
            modes.addItem(value.toString(), value.name());
         }
      }
      plugins.setEnabled(AuthModes.CUST_AUTH.name().equals(modes.getSelectedValue()));
   }
   
   /**
    * Gets the account type list box.
    * 
    * @return   The account type list box
    */
   ListBox getAccountsType()
   {
      return accountTypes;
   }
   
   /**
    * Gets the authentication mode list box.
    * 
    * @return   The authentication mode list box
    */
   ListBox getModes()
   {
      return modes;
   }
   
   /**
    * Gets the list box of custom authentication plugins.
    * 
    * @return   The list box of custom authentication plugins
    */
   ListBox getPlugins()
   {
      return plugins;
   }
   
   /**
    * Gets the Bulk Update button.
    * 
    * @return   The Bulk Update button
    */
   Button getUpdateAction()
   {
      return updateAction;
   }

   /**
    * Gets the Close button.
    * 
    * @return   The Close button
    */
   Button getCloseAction()
   {
      return closeAction;
   }
   
   /**
    * Provides an access to the target modal dialog.
    * 
    * @return   The underlined modal dialog
    */
   @Override
   protected Modal getModal()
   {
      return modal;
   }
}
