/*
** Module   : ModalDialogs.java
** Abstract : A facade to the common modal dialogs implementation
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20170612 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.admin.client.widget.dialog;

import org.gwtbootstrap3.client.ui.constants.*;

import java.util.function.*;

/**
 * A facade to the common modal dialogs implementation.
 */
public interface ModalDialogs
{
   /**
    * Displays an info modal dialog.
    *
    * @param   message
    *          The message to show.
    */
   default void showInfo(String message)
   {
      showInfo(message, null);
   }

   /**
    * Displays an info modal dialog.
    *
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showInfo(String message, Runnable closeCallback)
   {
      showInfo("Information", message, closeCallback);
   }

   /**
    * Displays an info modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showInfo(String title, String message, Runnable closeCallback)
   {
      getMessageDialogPresenter().showMessage(
         title,
         message,
         MessageType.INFO,
         new String[] {"OK"},
         new ButtonType[] {ButtonType.PRIMARY},
         closeCallback != null ? integer -> closeCallback.run() : null);
   }

   /**
    * Displays a warning modal dialog.
    *
    * @param   message
    *          The message to show.
    */
   default void showWarning(String message)
   {
      showError(message, null);
   }

   /**
    * Displays a warning modal dialog.
    *
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showWarning(String message, Runnable closeCallback)
   {
      showWarning("Warning", message, closeCallback);
   }

   /**
    * Displays a warning modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showWarning(String title, String message, Runnable closeCallback)
   {
      getMessageDialogPresenter().showMessage(
      title,
      message,
      MessageType.WARNING,
      new String[] {"OK"},
      new ButtonType[] {ButtonType.PRIMARY},
      closeCallback != null ? integer -> closeCallback.run() : null);
   }

   /**
    * Displays an error modal dialog.
    *
    * @param   message
    *          The message to show.
    */
   default void showError(String message)
   {
      showError(message, null);
   }

   /**
    * Displays the error modal dialog.
    * 
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showError(String message, Runnable closeCallback)
   {
      showError("Error", message, closeCallback);
   }

   /**
    * Displays an error modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showError(String title, String message, Runnable closeCallback)
   {
      getMessageDialogPresenter().showMessage(
         title,
         message,
         MessageType.ERROR,
         new String[] {"OK"},
         new ButtonType[] {ButtonType.PRIMARY},
         closeCallback != null ? integer -> closeCallback.run() : null);
   }

   /**
    * Displays a yes-no modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   type
    *          The message type.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showYesNo(String title,
                          String message,
                          MessageType type,
                          Consumer<Integer> closeCallback)
   {

      getMessageDialogPresenter().showYesNo(title,
                                            message,
                                            type,
                                            closeCallback);
   }

   /**
    * Displays a yes-no-cancel modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   type
    *          The message type.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showYesNoCancel(String title,
                                String message,
                                MessageType type,
                                Consumer<Integer> closeCallback)
   {
      getMessageDialogPresenter().showYesNoCancel(title,
                                                  message,
                                                  type,
                                                  closeCallback);
   }

   /**
    * Displays a ok-cancel modal dialog.
    *
    * @param   title
    *          Dialog title.
    * @param   message
    *          The message to show.
    * @param   type
    *          The message type.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showOKCancel(String title,
                             String message,
                             MessageType type,
                             Consumer<Integer> closeCallback)
   {

      getMessageDialogPresenter().showYesNoCancel(title,
                                                  message,
                                                  type,
                                                  closeCallback);
   }


   /**
    * Shows a modal dialog according to the supplied arguments.
    *
    * @param   title
    *          Modal dialog title.
    * @param   message
    *          The dialog message.
    * @param   type
    *          The message type.
    * @param   buttonLabels
    *          Button labels.
    * @param   buttonTypes
    *          Button types.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showMessage(String title,
                            String message,
                            MessageType type,
                            String[] buttonLabels,
                            ButtonType[] buttonTypes,
                            Consumer<Integer> closeCallback)
   {

      getMessageDialogPresenter().showMessage(title,
                                              message,
                                              type,
                                              buttonLabels,
                                              buttonTypes,
                                              closeCallback);
   }

   /**
    * Shows a modal dialog according to the supplied arguments.
    *
    * @param   title
    *          Modal dialog title.
    * @param   message
    *          The dialog message.
    * @param   icon
    *          The message icon.
    * @param   buttonLabels
    *          Button labels.
    * @param   buttonTypes
    *          Button types.
    * @param   closeCallback
    *          Close callback called when the dialog is dismissed.
    */
   default void showMessage(String title,
                            String message,
                            IconType icon,
                            String[] buttonLabels,
                            ButtonType[] buttonTypes,
                            Consumer<Integer> closeCallback)
   {

      getMessageDialogPresenter().showMessage(title,
                                              message,
                                              icon,
                                              buttonLabels,
                                              buttonTypes,
                                              closeCallback);
   }

   /**
    * Returns the common message dialog presenter.
    *
    * @return  see above.
    */
   MessageDialogPresenter getMessageDialogPresenter();
}
