/*
** Module   : ReportRequest.java
** Abstract : ReportRequest is a request to build a target report.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.shared;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Map;


/**
 * ReportRequest is a request to build a target report.
 */
public class ReportRequest implements Serializable
{
   
   /** Defines report id. */
   private String reportId;
   
   /** Defines key data for which the requested report should be built. */
   private String[] selectedRows;
   
   /** Defines report parameters and their values. */
   private Map<String, String> reportParameters;

   /**
    * Defines a search string to select a data subset for which the requested report should be
    * built.
    */
   private String filter;
   
   /**
    * Creates the report request data transfer object.
    */
   public ReportRequest()
   {}

   /**
    * Constructs the report request data transfer object.
    *
    * @param    reportId
    *           The report id
    * @param    selectedRows
    *           The selected rows
    * @param    filter
    *           The filter
    * @param    reportParameters
    *           The report parameters
    */
   public ReportRequest(String reportId, String[] selectedRows, String filter,
            Map<String, String> reportParameters)
   {
      this.reportId = reportId;
      this.selectedRows = selectedRows;
      this.filter = filter;
      this.reportParameters = reportParameters;
   }
   
   /**
    * Gets the report id.
    *
    * @return   The report id
    */
   public String getReportId()
   {
      return reportId;
   }

   /**
    * Sets the report id.
    *
    * @param    reportId
    *           The new report id
    */
   public void setReportId(String reportId)
   {
      this.reportId = reportId;
   }

   /**
    * Gets the selected rows.
    *
    * @return   The selected rows
    */
   public String[] getSelectedRows()
   {
      return selectedRows;
   }

   /**
    * Sets the selected rows.
    *
    * @param    selectedRows
    *           The new selected rows
    */
   public void setSelectedRows(String[] selectedRows)
   {
      this.selectedRows = selectedRows;
   }

   /**
    * Gets the report parameters.
    *
    * @return   The report parameters
    */
   public Map<String, String> getReportParameters()
   {
      return reportParameters;
   }

   /**
    * Sets the report parameters.
    *
    * @param    reportParameters
    *           The report parameters
    */
   public void setReportParameters(Map<String, String> reportParameters)
   {
      this.reportParameters = reportParameters;
   }

   /**
    * Gets the filter.
    *
    * @return   The filter
    */
   public String getFilter()
   {
      return filter;
   }

   /**
    * Sets the filter.
    *
    * @param    filter
    *           The new filter
    */
   public void setFilter(String filter)
   {
      this.filter = filter;
   }

   /**
    * Tests if this is a request for the detailed report.
    * 
    * @return   True if the detailed report is requested.
    */
   public boolean isDetailedReportRequest()
   {
      if (this.reportParameters != null)
      {
         String value = this.reportParameters.get(ReportParameters.DETAILS.getParameter());
         if (value != null && !value.isEmpty())
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * {@inheritDoc}
    * @see java.lang.Object#hashCode()
    */
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      result = prime * result + ((filter == null) ? 0 : filter.hashCode());
      result = prime * result + ((reportId == null) ? 0 : reportId.hashCode());
      result = prime * result + ((reportParameters == null) ? 0 : reportParameters.hashCode());
      result = prime * result + Arrays.hashCode(selectedRows);
      return result;
   }
   
   /**
    * {@inheritDoc}
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (getClass() != obj.getClass())
         return false;
      ReportRequest other = (ReportRequest) obj;
      if (filter == null)
      {
         if (other.filter != null)
            return false;
      }
      else if (!filter.equals(other.filter))
         return false;
      if (reportId == null)
      {
         if (other.reportId != null)
            return false;
      }
      else if (!reportId.equals(other.reportId))
         return false;
      if (reportParameters == null)
      {
         if (other.reportParameters != null)
            return false;
      }
      else if (!reportParameters.equals(other.reportParameters))
         return false;
      if (!Arrays.equals(selectedRows, other.selectedRows))
         return false;
      return true;
   }
}
