/*
** Module   : ComObject.java
** Abstract : The base class for all COM objects, with minimal management.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit.
** 002 CA  20171026 Changes for native COM automation support.
** 003 CA  20171102 The comhandle must use the comId as its ID reported by the id() API.
**     EVL 20171110 Fix for array out of index exception issue when merging OUTPUT or INPUT-OUTPUT
**                  parameters back to Java side.
** 004 CA  20180517 Fixed comhandle FWD-specific resource management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.comauto;

import java.lang.reflect.*;
import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Base class for COM objects that will be stored using {@link comhandle}s. Does minimal
 * management of the COM Object.
 */
public abstract class ComObject
implements WrappedResource
{
   /**
    * The resource id. Used by {@code WrappedResource} management. 
    */
   private Long resourceId = null;
   
   /**
    * The parent {@code ControlFrameWidget} widget. This container is used during the dispatch of 
    * the events generated by this COM Object.
    */
   private ControlFrameWidget parentControlFrame = null;
   
   /**
    * Build a new COM Object.
    */
   public ComObject()
   {
   }
   
   /**
    * Converts a value to a BDT object. The method does not check whether the value is already a
    * {@code BaseDataType}, in fact is assume it is not. Note that, because this method uses the
    * best matching constructor, this method may do some implicit conversions if the passed in
    * value is compatible with the class required (like from a {@code String}).
    * 
    * @param   aVal
    *          The value to be converted.
    *
    * @return  A {@code BaseDataType} instance that represent the same value as {@code aVal}.
    */
   protected static BaseDataType marshal(Object aVal)
   {
      return marshal(aVal, null);
   }
   
   /**
    * Converts a value to a BDT object. The method does not check whether the value is already a
    * {@code BaseDataType}, in fact is assume it is not. Note that, because this method uses the
    * best matching constructor, this method may do some implicit conversions if the passed in
    * value is compatible with the class required (like from a {@code String}).
    * 
    * @param   aVal
    *          The value to be converted.
    * @param   bdtClass
    *          The type of the returned result (as a subclass of {@code BaseDataType}).
    *
    * @return  A {@code BaseDataType} instance that represent the same value as {@code aVal}.
    */
   protected static BaseDataType marshal(Object aVal, Class<?> bdtClass)
   {
      // TODO: extent arguments

      if (bdtClass != null)
      {
         try
         {
            Constructor<?> ctor = bdtClass.getConstructor(aVal.getClass());
            return (BaseDataType) ctor.newInstance(aVal);
         }
         catch (NoSuchMethodException  |
                IllegalAccessException |
                InstantiationException |
                InvocationTargetException e)
         {
            // TODO: log this
         }
      }
      else
      {
         if (aVal != null)
         {
            if (aVal instanceof BaseDataType)
            {
               return (BaseDataType) aVal;
            }
            else if (aVal instanceof String)
            {
               return new character((String) aVal);
            }
            else if (aVal instanceof Boolean)
            {
               return new logical((Boolean) aVal);
            }
            else if (aVal instanceof Integer)
            {
               return new integer((Integer) aVal);
            }
            else if (aVal instanceof Long)
            {
               return new int64((Long) aVal);
            }
            else if (aVal instanceof Double)
            {
               return new decimal((Double) aVal);
            }
            else if (aVal instanceof Float)
            {
               return new decimal((Float) aVal);
            }
            else if (aVal instanceof Date)
            {
               return new date((Date) aVal);
            }
         }
      }
      
      return new unknown();
   }

   /**
    * Converts a value from a {@link ComParameter}, {@link BaseDataType} or other Java type to
    * a {@link ComParameter} instance which can be passed as an argument to a COM method call.
    * 
    * @param   aVal
    *          The value to be converted.
    *
    * @return  A {@code ComParameter} instance.
    * 
    * @throws  IllegalArgumentException
    *          If the value can't be passed by FWD as an argument for a COM method call.
    */
   protected static ComParameter unmarshal(Object aVal)
   {
      if (aVal == comhandle.EMPTY)
      {
         return null;
      }
      
      boolean extent = false;
      
      if (aVal instanceof ComParameter)
      {
         ComParameter aparam = (ComParameter) aVal;
         extent = aparam.isExtent();
         
         if (aparam.isOutput())
         {
            // replace the BDT instance for OUTPUT parameters and set it to default value
            ComParameter oparam = new ComParameter(aparam);
            
            Object value;
            
            if (extent)
            {
               Class<?> cls = oparam.getValue().getClass().getComponentType();
               
               value = Array.newInstance(cls, oparam.getLength());
               for (int i = 0; i < oparam.getLength(); i++)
               {
                  Array.set(value, i, BaseDataType.generateDefault(cls));
               }
            }
            else
            {
               value = ((BaseDataType) aparam.getValue()).instantiateDefault();
            }
            oparam.setValue(value);
            
            aVal = oparam;
         }
      }
      
      String[] fwdDataType = new String[1];

      Function<Object, Object> process = (value) ->
      {
         if (value instanceof BaseDataType)
         {
            BaseDataType bdt = (BaseDataType) value;
            fwdDataType[0] = bdt.getTypeName();
            
            if (bdt.isUnknown() || bdt instanceof unknown)
            {
               value = null;
            }
            else if (bdt instanceof memptr)
            {
               value = ((memptr) bdt).getPointerValue().longValue();
            }
            else if (bdt instanceof raw)
            {
               value = ((raw) bdt).getByteArray();
            }
            else if (bdt instanceof comhandle)
            {
               ComObject com = ((comhandle) bdt).getResource();
               
               if (com instanceof NativeComObject)
               {
                  value = ((NativeComObject) com).getComId();
               }
               else
               {
                  throw new IllegalArgumentException(
                            "FWD COM implementations can't be invoked on native side!");
               }
            }
            else if (bdt instanceof date || 
                     bdt instanceof datetime ||
                     bdt instanceof datetimetz)
            {
               value = ((date) bdt).dateValue();
            }
            else if (bdt instanceof handle)
            {
               throw new IllegalArgumentException(
                        "HANDLE data-type is not supported by native COM arguments!");
            }
            else if (bdt instanceof logical)
            {
               value = Boolean.valueOf(((logical) bdt).booleanValue());
            }
            else if (bdt instanceof decimal)
            {
               value = new Double(((decimal) bdt).doubleValue());
            }
            else if (bdt instanceof integer)
            {
               value = new Integer(((integer) bdt).intValue());
            }
            else if (bdt instanceof int64)
            {
               value = new Long(((int64) bdt).longValue());
            }
            else if (bdt instanceof recid)
            {
               throw new IllegalArgumentException(
                        "RECID data-type is not supported by native COM arguments!");
            }
            else if (bdt instanceof rowid)
            {
               throw new IllegalArgumentException(
                        "ROWID data-type is not supported by native COM arguments!");
            }
            else if (bdt instanceof Text)
            {
               value = ((Text) bdt).getValue();
            }
            else
            {
               throw new IllegalArgumentException(
                     fwdDataType + " data-type is not supported by native COM arguments!");
            }
         }
         else
         {
            if (value == null)
            {
               fwdDataType[0] = "unknown";
            }
            if (value instanceof String)
            {
               fwdDataType[0] = "character";
            }
            else if (value instanceof Boolean)
            {
               fwdDataType[0] = "logical";
            }
            else if (value instanceof Integer)
            {
               fwdDataType[0] = "integer";
            }
            else if (value instanceof Long)
            {
               fwdDataType[0] = "int64";
            }
            else if (value instanceof Double)
            {
               fwdDataType[0] = "decimal";
            }
            else if (value instanceof Float)
            {
               fwdDataType[0] = "decimal";
            }
            else if (value instanceof Date)
            {
               fwdDataType[0] = "date";
            }
            else
            {
               throw new IllegalArgumentException(
                        value.getClass() + " data-type is not supported by native COM arguments!");
            }
         }
         
         return value;
      };

      ComParameter ret;
      Object value;
      
      if (aVal instanceof ComParameter)
      {
         ComParameter cparam = (ComParameter) aVal;
         
         ret = new ComParameter(cparam);
         ret.setByPointer(cparam.isByPointer());
         ret.setByVariantPointer(cparam.isByVariantPointer());
         ret.setDataType(cparam.getDataType());
         ret.setMode(cparam.getMode());

         value = cparam.getValue();
      }
      else
      {
         ret = new ComParameter();
         ret.setByPointer(false);
         ret.setByVariantPointer(false);
         ret.setDataType(null);
         ret.setMode('I');
         
         value = aVal;
      }
      
      if (value != null && value.getClass().isArray())
      {
         ret.setExtent(true);
         ret.setLength(Array.getLength(value));
      }
      
      if (ret.isExtent())
      {
         Object avalue = Array.newInstance(Object.class, ret.getLength());
         
         for (int i = 0; i < ret.getLength(); i++)
         {
            Array.set(avalue, i, process.apply(Array.get(value, i)));
         }
         
         value = avalue;
      }
      else
      {
         value = process.apply(value);
      }

      ret.setFwdDataType(fwdDataType[0].toUpperCase());
      ret.setValue(value);
      
      return ret;
   }

   /**
    * Converts an array of values from a {@link ComParameter}, {@link BaseDataType} or other Java 
    * type to a {@link ComParameter} instance which can be passed as an argument to a COM method 
    * call.
    * 
    * @param   aVals
    *          The values to be converted.
    *
    * @return  A {@code ComParameter} instance.
    * 
    * @throws  IllegalArgumentException
    *          If the value can't be passed by FWD as an argument for a COM method call.
    */
   protected static ComParameter[] unmarshal(Object[] aVals)
   {
      // the parameter list can be empty
      if (aVals == null)
      {
         return null;
      }
      
      ComParameter[] result = new ComParameter[aVals.length];
      
      for (int i = 0; i < aVals.length; i++)
      {
         result[i] = unmarshal(aVals[i]);
      }
      
      return result;
   }
   
   /**
    * Check the parameters received from a COM method call and copy the values into OUTPUT or
    * INPUT-OUTPUT arguments.
    * 
    * @param    vals
    *           The received values.  The arguments are 1-indexed (as on index 0 is the return 
    *           value), and each argument has an entry in this array; {@code null} values 
    *           represent INPUT arguments or {@link unknown} values.
    * @param    params
    *           The actual references received by FWD for this method call.  All OUTPUT or INPUT-
    *           OUTPUT arguments must be {@link ComParameter} instances.
    */
   protected static void checkParameters(BaseDataType[][] vals, Object[] params)
   {
      if (params == null)
      {
         return;
      }
      
      for (int i = 0, k = 0; i < params.length; i++)
      {
         if (!(params[i] instanceof ComParameter))
         {
            continue;
         }
         
         ComParameter comParam = (ComParameter) params[i];
         if (comParam.isOutput() || comParam.isInputOutput())
         {
            // we increase k counter before indexing the vals to account vals[0] is
            // not a parameter but result
            BaseDataType[] val = vals[++k];
            
            Object aval = comParam.getValue();
            
            if (comParam.isExtent())
            {
               if (comParam.getLength() == 0)
               {
                  aval = ArrayAssigner.resize((BaseDataType[]) aval, val.length);
                  // and also replace the reference...
                  comParam.getRefReplacement().accept((BaseDataType[]) aval);
               }
               
               if (comParam.getLength() != val.length)
               {
                  throw new IllegalStateException("Incorrect source and dest array dimension!");
               }
               
               // now we can just copy the values - a dynamic extent was resized and its reference
               // was properly switched
               for (int j = 0; j < comParam.getLength(); j++)
               {
                  ((BaseDataType) Array.get(aval, j)).assign(val[j]);
               }
            }
            else
            {
               if (!(aval instanceof BaseDataType))
               {
                  throw new IllegalStateException(
                     "COM call sent an OUTPUT/INPUT-OUTPUT argument which was not passed as a BDT " +
                     "from FWD side!");
               }
               
               BaseDataType bdt = (BaseDataType) aval;
               if (val[0] == null)
               {
                  bdt.setUnknown();
               }
               else
               {
                  bdt.assign(val[0]);
               }
            }
         }
      }
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return   <code>true</code> if object is unknown.
    */
   @Override
   public boolean unknown()
   {
      return false;
   }
   
   /**
    * Get this resource's ID, if is already set.
    *
    * @return   The resource's ID or <code>null</code> if not set.
    */
   @Override
   public Long id()
   {
      return resourceId;
   }
   
   /**
    * Set this resource's ID.
    *
    * @param    id
    *           The resource's ID.
    */
   @Override
   public void id(long id)
   {
      resourceId = id;
   }
   
   /**
    * Obtain the name of this COM object.
    * 
    * @return  the name of this COM object.
    */
   public abstract String getName();
   
   /**
    * Perform actual delete of all resources. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    */
   public abstract void release();
   
   /**
    * Destroy all resources kept by FWD related to this COM object.
    */
   public final void destroy()
   {
      comhandle.removeResource(this);
   }
   
   /**
    * Obtain the {@code ControlFrameWidget} container for this COM, if any.
    * 
    * @return  container for this COM.
    */
   public ControlFrameWidget getParentControlFrame()
   {
      return parentControlFrame;
   }
   
   /**
    * Sets the {@code ControlFrameWidget} container for this COM.
    *
    * @param   pcf
    *          The container for this COM.
    */
   public void setParentControlFrame(ControlFrameWidget pcf)
   {
      this.parentControlFrame = pcf;
   }
   
   /**
    * Sets a COM property for the COM object stored by the handle.
    * 
    * @param   prop
    *          The legacy property name. Case insensitive. If no property is found then a warning
    *          is displayed and method returns without altering the stored object.
    * @param   newVal
    *          The new value for the property. It must be of a compatible type with the property.
    * @param   indices
    *          A variable number of indices used to access this property's element.
    * 
    * @return  <code>true</code> if the property was set.
    * 
    * @throws  PropertyAccessException
    *          If the the property was found but there was an error during setting.
    */
   public boolean setProperty(String prop, Object newVal, Object... indices)
   {
      // TODO: the "AS data-type" clause
      
      Method setter = getComAccessor(prop, true);
      
      if (setter == null)
      {
         return false;
      }
      // prepare params
      BaseDataType[] params = new BaseDataType[1]; 
      if (newVal instanceof BaseDataType)
      {
         params[0] = (BaseDataType) newVal;
      }
      else 
      {
         // setter is guaranteed to have exactly one parameter because of getComAccessor()
         params[0] = marshal(newVal, setter.getParameterTypes()[0]);
      }
      
      try
      {
         setter.invoke((Object[]) params);
      }
      catch (IllegalAccessException | InvocationTargetException e)
      {
         // res remains unchanged 
      }
      
      return true;
   }
   
   /**
    * Obtain a property value from the COM object stored in this handle. 
    * 
    * @param   prop
    *          The legacy name of the property. Case insensitive. If no property is found then a
    *          warning is displayed and method returns {@code unknown} value.
    *
    * @return  The value of the requested property or {@code unknown} on exceptions.
    */
   public BaseDataType getProperty(String prop)
   {
      Method getter = getComAccessor(prop, false);
      if (getter != null)
      {
         Object ret = null;
         try
         {
            ret = getter.invoke(this, (Object[]) null);
         }
         catch (IllegalAccessException | InvocationTargetException e)
         {
            // res remains [null] and won't be cast to BDT and returned 
         }
         
         if (ret instanceof BaseDataType)
         {
            return (BaseDataType) ret;
         }
         else if (ret instanceof ComObject)
         {
            return new comhandle(ret);
         }
      }
      
      return null;
   }
   
   /**
    * Calls a COM-method without parameters on the COM-object stored in this com-handle.
    *
    * @param   methodName
    *          The method name. Case insensitive. If the COM-object does not declare such method
    *          an error message is displayed and this method returns {@code unknown} value.
    * @param   params
    *          The list of actual parameters. Their types must be compatible with the parameters
    *          of called method.
    *          TODO: marshal to BDT (?) When code is generated [param] is already BDT.
    *
    * @return  The result of the called method, if any.
    */
   public BaseDataType call(String methodName, Object... params)
   {
      // TODO: marshal parameters to BDT
      // TODO: OUTPUT/INPUT-OUTPUT parameters

      Class<?> clazz = this.getClass();
      while (clazz != Object.class)
      {
         List<Method> allMethods = new ArrayList<>(Arrays.asList(clazz.getDeclaredMethods()));
         for (Method method : allMethods)
         {
            ComMethod ann = method.getAnnotation(ComMethod.class);
            if (ann != null && ann.name().equalsIgnoreCase(methodName))
            {
               // check parameter compatibility / marshal params to BDT (?)
               try
               {
                  Object ret = method.invoke(this, params);
                  if (ret instanceof BaseDataType)
                  {
                     return (BaseDataType) ret;
                  }
                  else 
                  {
                     return marshal(ret, null);
                  }
               }
               catch (IllegalAccessException | InvocationTargetException e)
               {
                  e.printStackTrace();
               }
            }
         }
         // move to the upper class in the hierarchy in search for more methods
         clazz = clazz.getSuperclass();
      }
      
      return new comhandle();
   }

   /**
    * Calls a COM-method without parameters on the COM-object stored in this com-handle.
    *
    * @param   methodName
    *          The method name. Case insensitive. If the COM-object does not declare such method
    *          an error message is displayed and this method returns {@code unknown} value.
    * @param   params
    *          The list of actual parameters. Their types must be compatible with the parameters
    *          of called method.
    *          TODO: marshal to BDT (?) When code is generated [param] is already BDT.
    *
    * @return  The result of the called method, in a {@code comhandle}, so it can be used in a
    *          chained call or property access. 
    */
   public comhandle chainCall(String methodName, Object... params)
   {
      // TODO: marshal parameters to BDT
      // TODO: OUTPUT/INPUT-OUTPUT parameters
      
      Class<?> clazz = this.getClass();
      while (clazz != Object.class)
      {
         List<Method> allMethods = new ArrayList<>(Arrays.asList(clazz.getDeclaredMethods()));
         for (Method method : allMethods)
         {
            ComMethod ann = method.getAnnotation(ComMethod.class);
            if (ann != null && ann.name().equalsIgnoreCase(methodName))
            {
               // check parameter compatibility / marshal params to DBT (?)
               try
               {
                  Object ret = method.invoke(this, params);
                  // the result must be some king of ComObject so that the chained child
                  // call/property could be executed
                  if (ret instanceof ComObject)
                  {
                     return new comhandle(ret);
                  }
                  else 
                  {
                     // otherwise a NPE will occur because of invalid result 
                     return new comhandle();
                  }
               }
               catch (IllegalAccessException | InvocationTargetException e)
               {
                  e.printStackTrace();
               }
            }
         }
         // move to the upper class in the hierarchy in search for more methods
         clazz = clazz.getSuperclass();
      }
      
      return new comhandle();
   }

   /**
    * Internal utility method that selects an accessor for a property from a COM object.
    * 
    * @param   prop
    *          The property being searched. The accessor method (setter or getter) must be
    *          annotated with proper {@code ComProperty} annotation.
    * @param   setter
    *          {@code true} for setter methods and {@code false} for getters. In either case a
    *          quick check is performed to verify whether the eventually found method are the
    *          requested kind of accessors.
    * 
    * @return  A method for accessing the respective property or {@code null} if no such property
    *          was found or it does not meet the minimum syntactic conditions.
    */
   private Method getComAccessor(String prop, boolean setter)
   {
      Class<?> clazz = this.getClass();
      while (clazz != Object.class)
      {
         List<Method> allMethods = new ArrayList<>(Arrays.asList(clazz.getDeclaredMethods()));
         for (Method method : allMethods)
         {
            ComProperty ann = method.getAnnotation(ComProperty.class);
            if (ann != null                        && 
                ann.name().equalsIgnoreCase(prop)  &&
                ann.setter() == setter)
            {
               // check parameter compatibility for setter
               if (setter)
               {
                  Class<?>[] params = method.getParameterTypes();
                  if (params == null || params.length != 1)
                  {
                     return null;
                  }
               }
               else
               {
                  Class<?>[] params = method.getParameterTypes();
                  if (params != null && params.length != 0)
                  {
                     return null;
                  }
               }
               return method;
            }
         }
         // move to the upper class in the hierarchy in search for more methods
         clazz = clazz.getSuperclass();
      }
      
      return null;
   }
}
