/*
** Module   : CompatibilityHelper.java
** Abstract : hide differences in E4GL output generation
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20070822 NEW   @34964 Provides a standard interface to hide
**                               differences in E4GL output generation.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.e4gl;

/**
 * Provides a standard interface to hide differences in E4GL output
 * generation.
 */
public interface CompatibilityHelper
{
   /** Newline text on the system on which this tool runs. */
   public static final String NEWLINE = System.getProperty("line.separator");
   
   /**
    * Provides the known configuration to the helper and allows it to
    * prepare for use. This must be called before any other method can be
    * used.
    *
    * @param    opts
    *           The configuration.
    */
   public void initialize(Options opts);
   
   /**
    * Return the filename suffix (extension) to be used.
    *
    * @return   The filename extension to use based on the known options.
    */
   public String suffix();
   
   /**
    * Return the text for the header of the generated file. This is output
    * before any parsed content.
    *
    * @return   The text to use based on the known options.
    */
   public String header();
   
   /**
    * Return the text used to start a line of raw HTML output.
    *
    * @param    content
    *           <code>true</code> if there is HTML content that is about to
    *           be emitted on the line, <code>false</code> to handle the
    *           no content case (just start the line, without any
    *           content prefix text).
    *
    * @return   The text to use based on the known options.
    */
   public String openHTML(boolean content);
   
   /**
    * Return the text used to end a line of raw HTML output.
    *
    * @param    content
    *           <code>true</code> if there is content that needs to be
    *           finished on the line, <code>false</code> to handle the
    *           no content case (just terminate the line, without any
    *           content termination text).
    *
    * @return   The text to use based on the known options.
    */
   public String closeHTML(boolean content);
   
   /**
    * Convert any unsafe characters (in the given HTML text) into equivalents
    * that are safe to embed in a 4GL string literal and return that converted
    * text.
    *
    * @return   The replacement text.
    */
   public String escapeHTML(String html);
   
   /**
    * Convert any HTML entities (in the given E4GL text) into the equivalent
    * characters.
    *
    * @return   The replacement text with all conversion applied.
    */
   public String entityConvert(String e4gl);
   
   /**
    * Return the text used to start the output of an embedded 4GL expression.
    *
    * @return   The text to use based on the known options.
    */
   public String openExprEscape();
   
   /**
    * Return the text used to end the output of an embedded 4GL expression.
    *
    * @return   The text to use based on the known options.
    */
   public String closeExprEscape();
   
   /**
    * Return the text used to start a comment that removes the start or end
    * tag for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String openComment(boolean expr);
   
   /**
    * Return the text used to end a comment that removes the start or end
    * tag for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String closeComment(boolean expr);
   
   /**
    * Return the text used to start a comment that removes the META or
    * WSMETA element with an HTTP-EQUIV attribute that is set to
    * CONTENT-TYPE.
    *
    * @return   The text to use based on the known options. Return the empty
    *           string if the element should be left behind instead of
    *           commented.
    */
   public String openContentTypeComment();
   
   /**
    * Return the text used to end a comment that removes the META or
    * WSMETA element with an HTTP-EQUIV attribute that is set to
    * CONTENT-TYPE.
    *
    * @return   The text to use based on the known options. Return the empty
    *           string if the element should be left behind instead of
    *           commented.
    */
   public String closeContentTypeComment();
   
   /**
    * Return the text for the footer of the generated file. This is output
    * after all parsed content.
    *
    * @return   The text to use based on the known options.
    */
   public String footer();
}