/*
** Module   : AppServerEntry.java
** Abstract : Provides low-level APIs to connect to and communicate with an appserver.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130823 Allow identification of each request based on an unique ID. Added support to
**                  terminate a remote session. 
** 003 CA  20160404 Enhanced the appserver invocations, to allow implicit or explicit termination 
**                  of requests being executing on a certain connection.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import com.goldencode.p2j.util.*;

/**
 * Provide low-level APIs to connect and communicate with a remote appserver. This is exported
 * as a network server and statically implemented by the {@link AppServerManager} class.
 * <p>
 * These low-level APIs are meant to be used only by P2J runtime code. For outside access, use 
 * the APIs defined by the {@link AppServerHelper helper} class.
 */
public interface AppServerEntry
{
   /**
    * Establish a new appserver connection. This will invoke any connect procedure for this 
    * appserver, to which the given <code>user</code>, <code>pwd</code> and <code>serverInfo</code>
    * parameters will be passed.
    *  
    * @param    appServer
    *           The name of the target appserver.
    * @param    sessionFree
    *           Flag indicating if the connection request is for a session-free operating mode.
    * @param    user
    *           The first parameter to pass to the connect procedure, if any.
    * @param    pwd
    *           The second parameter to pass to the connect procedure, if any.
    * @param    serverInfo
    *           The third parameter to pass to the connect procedure, if any.
    *           
    * @return   An {@link AppServerInvocationResult} instance with any result or caught errors.
    */
   public AppServerInvocationResult connect(String  appServer,
                                            boolean sessionFree,
                                            String  user,
                                            String  pwd,
                                            String  serverInfo);

   /**
    * Disconnect the appserver connection, with the given ID.  It will invoke any configured 
    * disconnect procedure, but will not return any errors or values to the caller.
    * 
    * @param    id
    *           The connection ID to disconnect.
    */
   public void disconnect(String id);

   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side. Any errors, result or updated arguments will be
    * sent back to the caller.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *
    * @return   A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *           procedure, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    */
   public AppServerInvocationResult invokePersistent(String    id,
                                                     int       requestId,
                                                     long      timeout,
                                                     character name,
                                                     boolean   transactionDistinct,
                                                     String    modes,
                                                     Object... args);

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side. Any errors,
    * result or updated arguments will be sent back to the caller.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    resId
    *           The code of a remote procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invoke(String    id,
                                           int       requestId,
                                           long      timeout,
                                           character name,
                                           String    resId,
                                           boolean   function,
                                           boolean   dynamicFunction,
                                           boolean   superCall,
                                           boolean   transactionDistinct,
                                           String    modes,
                                           Object... args);

   /**
    * Delete the remote procedure, specified by the resource ID. This must be called at the same
    * time when the local proxy procedure is deleted.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    resId
    *           The ID of a remote procedure which needs to be deleted.
    *           
    * @return   <code>true</code> if the remote procedure was deleted.
    */
   public boolean deleteProcedure(String id, String resId);

   /**
    * Get the name of the remote procedure, specified by the resource ID.
    *  
    * @param    id
    *           The appserver connection id. 
    * @param    resId
    *           The resource ID of a remote procedure.
    *
    * @return   The procedure's name.
    */
   public character getProcedureName(String id, String resId);

   /**
    * Send the STOP condition to the request running with the specified ID.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    */
   public void sendStop(String id, int requestId);
   
   /**
    * Send a STOP condition to the request currently being executed on the specified connection.
    * 
    * @param    id
    *           The appserver connection id. 
    */
   public void sendStop(String id);
   
   /**
    * Obtain an unique ID from the remote side.
    * 
    * @return   An unique ID.
    */
   public int uniqueId();

   /**
    * Obtain a proxy for the given external procedure. This will also reserve an agent which will
    * be used for all subsequent invocation (including external procedure initialization).
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           A name of a remote external procedure.
    *           
    * @return   A proxy code for this procedure. May be invalid if the remote side can't resolve 
    *           the procedure.
    */
   public String obtainProxy(String id, String name);

   /**
    * Initialize the specified proxy procedure, by calling the <code>execute</code> method.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    code
    *           The code of a persistent external procedure to which the agent is associated. May
    *           be <code>null</code>.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *           
    * @return   The result of the external program initialization.
    */
   public AppServerInvocationResult initializeProxy(String   id,
                                                    int      requestId,
                                                    String   code, 
                                                    boolean  transactionDistinct,
                                                    String   modes,
                                                    Object[] args);
}