/*
** Module   : ClientParameters.java
** Abstract : Container of various client-side parameters which need to be exposed to the server.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 CA  20131004 Created initial version.
** 002 GES 20141228 Added some equivalents to 4GL command line options.
** 003 GES 20150108 Added OS userid, OS PID and terminal name.
** 004 IAS 20160328 Client log support.
** 005 OM  20160927 Added batchMode property.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;

/**
 * Container of client-side parameters which need to be exposed to the server-side application 
 * code.
 * <p>
 * This can be used for values that are effectively constants (set or queried at the time the
 * client starts and then unchangable for the remainer of the client's lifetime). It is suitable
 * for values that correspond to:
 * <p>
 * <ul>
 *    <li> Read-only environmental state that is exposed by the 4GL (e.g.
 *         <code>SESSION:WINDOW-SYSTEM</code>).
 *    <li> 4GL command line option equivalents (e.g. <code>-param</code>).
 *    <li> Operating system PID and userid. 
 * </ul>
 */
public class ClientParameters
implements Externalizable
{
   /** Constant for the "TTY" window-system (CHUI driver default). */
   public static final String WINDOW_SYSTEM_TTY = "TTY";
   
   /** Constant for the "MS-WINDOWS" window-system (GUI driver default). */
   public static final String WINDOW_SYSTEM_MS_WINDOWS = "MS-WINDOWS";
   
   /** Constant identifying the "MS-WIN95" window-system. */
   public static final String WINDOW_SYSTEM_MS_WIN95 = "MS-WIN95";
   
   /** Constant identifying the "MS-WINXP" window-system. */
   public static final String WINDOW_SYSTEM_MS_WINXP = "MS-WINXP";
   
   /** Identify the SESSION:WINDOW-SYSTEM attribute used by this client. */
   public String windowSystem = null;
   
   /** <code>-param</code> 4GL command line option (<code>SESSION:PARAMETER</code>). */
   public String param = null;
   
   /** <code>-p</code> 4GL command line option specifying the startup procedure to execute. */
   public String startupProc = null;
   
   /** User name as reported by the operating system. */
   public String osUserName = null;
   
   /** Terminal name as specified on the client command line. */
   public String terminalName = null;
   
   /** Process ID (PID) of the client process as reported by the operating system. */
   public long pid = -1;
   
   /** <code>true</code> if this client is ChUI, <code>false</code> for GUI. */
   public boolean chui = true;
   
   /** <code>-debugalert</code> 4GL command line option (<code>SESSION:DEBUG-ALERT</code>). */
   public boolean debugAlert = false;
   
   /** {@code -clientlog} 4GL command line option */ 
   public String clientlog = null;
   
   /** The batch mode. 0 = none, 1 = batch on, 2/3 = batch forced (client:mode:batch=true) */
   public int batchMode = 0;
   
   /**
    * Replacement for the default object writing method. The latest state is written to the output
    * destination.
    * 
    * @param    out
    *           The output destination to which parameters will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(windowSystem);
      out.writeObject(param);
      out.writeObject(startupProc);
      out.writeObject(osUserName);
      out.writeObject(terminalName);
      out.writeLong(pid);
      out.writeBoolean(chui);
      out.writeBoolean(debugAlert);
      out.writeObject(clientlog);
      out.writeInt(batchMode);
   }
   
   /**
    * Replacement for the default object reading method. The latest state is read from the input
    * source.
    * 
    * @param    in
    *           The input source from which parameters will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If a parameter can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      windowSystem = (String) in.readObject();
      param        = (String) in.readObject();
      startupProc  = (String) in.readObject();
      osUserName   = (String) in.readObject();
      terminalName = (String) in.readObject();
      pid          = in.readLong();
      chui         = in.readBoolean();
      debugAlert   = in.readBoolean();
      clientlog    = (String) in.readObject();
      batchMode    = in.readInt();
   }
}