/*
** Module   : EmbeddedWebAppHandler.java
** Abstract : Define handlers to run the embedded web app from within FWD.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20180724 Created initial version
** 002 CA  20180815 adm_windows.json will be served from the embedded/adm_windows.json path.
**                  The DefaultHandler must not be registered here, otherwise the /gui path will
**                  not be found.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.SessionManager;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;

import java.io.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.util.resource.*;

import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

/**
 * Define handlers to run the embedded web app from within fwd.  The handlers are run on the
 * <code>/embedded</code>, where <code>/embedded/fwd</code> will serve content from the 
 * <code>com/goldencode/p2j/ui/client/gui/driver/web/res/</code> package and the
 * <code>/embedded/custom</code> will server content from the <code>embedded/</code> sub-package
 * of the application's package root (as retrieved via {@link SourceNameMapper#getPackageRoot}. 
 */
public class EmbeddedWebAppHandler
{
   /** Anonymous log instance. */
   private static final Logger LOG = LogHelper.getLogger(EmbeddedWebAppHandler.class.getName());

   /**
    * The authentication worker - it authenticates an in-JVM FWD client, which will be used
    * to spawn FWD web clients. 
    */
   private static AuthWorker authWorker = null;

   /** The application's package root, in folder form. */
   private static String pkgroot;

   /**
    * Terminate the embedded web app.
    */
   public static void terminate()
   {
      synchronized (EmbeddedWebAppHandler.class)
      {    
         if (authWorker != null)
         {
            try
            {
               authWorker.stop();
            }            
            catch (Exception exc)
            {
               LOG.logp(Level.SEVERE,
                        "EmbeddedWebAppHandler.terminate()",
                        "",
                        LogHelper.generate("Could not shutdown the embedded web app auth worker!"),
                        exc);
            }
         }
      }
   }

   /**
    * Initialize all the handlers for the embedded web app.
    * 
    * @return   The handler for the embedded web app.
    */
   public static AbstractHandler initialize()
   {
      synchronized (EmbeddedWebAppHandler.class)
      {    
         if (authWorker == null)
         {
            authWorker = new AuthWorker();
            authWorker.start();

            String basePath = "/embedded";
            pkgroot = SourceNameMapper.getPackageRoot();
            pkgroot = pkgroot.replace('.', '/');
            if (!pkgroot.endsWith("/"))
            {
               pkgroot = pkgroot + "/";
            }
            if (!pkgroot.startsWith("/"))
            {
               pkgroot = "/" + pkgroot;
            }
            
            ContextHandler fwdCtx = createContextHandler(basePath + "/fwd",
                                                         "com/goldencode/p2j/ui/client/gui/driver/web/res/");
            ContextHandler customCtx = createContextHandler(basePath + "/custom",
                                                            pkgroot + "embedded/");
            ContextHandlerCollection contexts = new ContextHandlerCollection();
            contexts.addHandler(fwdCtx);
            contexts.addHandler(customCtx);

            HandlerCollection handlers = new HandlerCollection();
            handlers.addHandler(new EmbeddedPageHandler(basePath));
            handlers.addHandler(contexts);
            
            for (Handler h : WebResourceHandler.getWebResourceHandlers(basePath))
            {
               handlers.addHandler(h);
            }
            
            // the default handler must not be registered here - otherwise the /gui path will not
            // be served, as the request will be 'consumed' by the DefaultHandler.
            // handlers.addHandler(new DefaultHandler());
            
            return handlers;
         }
         else
         {
            return null;
         }
      }
   }
   
   /**
    * Create a new context handler serving class path resources from the given path.
    * 
    * @param    context
    *           The context used to server the resources.
    * @param    path
    *           The class path were the resources reside.
    *           
    * @return   A new handler.
    */
   private static ContextHandler createContextHandler(String context, String path)
   {
      ResourceHandler res = new ResourceHandler();
      res.setDirectoriesListed(false);
      ContextHandler ctx = new ContextHandler(context);
      ctx.setBaseResource(Resource.newClassPathResource(path));
      ctx.setHandler(res);
      
      return ctx;
   }
   
   /**
    * Implements the Jetty handler interface which calls the P2J server to spawn a web client and 
    * then it will return a page that loads that client's interface in an IFRAME.
    */
   private static class EmbeddedPageHandler
   extends AbstractHandler
   {
      /** The base path for this handler. */
      private final String basePath;
      
      /** The spawned client's configuration options. */
      private String[] options;

      /**
       * Create and initialize this handler with the given base path.
       * 
       * @param    basePath
       *           The base path to serve the requests.
       */
      public EmbeddedPageHandler(String basePath)
      {
         this.basePath = basePath;
         
         // read this from the server context - the handle() API has no context and can't access 
         // it.
         String cfgs =  Utils.getDirectoryNodeString(null,
                                                     "embeddedWebApp/cfgOverrides", 
                                                     null,
                                                     false);
         String[] cfgOpts = cfgs == null ? null : cfgs.split(" ");
         
         options = new String[2 + (cfgOpts == null ? 0 : cfgOpts.length)];
         if (cfgOpts != null)
         {
            System.arraycopy(cfgOpts, 0, options, 1, cfgOpts.length);
         }
         // let this first so that it can be overridden
         options[0] = "client:cmd-line-option:startup-procedure=emain.p";
         options[options.length - 1] = "client:web:embedded=true";
      }
      
      /**
       * Code to process web requests.
       */
      public void handle(String              target, 
                         Request             baseRequest,
                         HttpServletRequest  request,
                         HttpServletResponse response) 
      throws IOException,
             ServletException
      {
         if (!(target.equals(basePath) || 
               target.equals(basePath + "/launch") ||
               target.equals(basePath + "/adm_windows.json")))
         {
            baseRequest.setHandled(false);
            return;
         }
         
         response.setHeader("Access-Control-Allow-Origin", "*");
         response.setHeader("Access-Control-Allow-Methods", "POST, GET");
         response.setContentType("text/html;charset=utf-8");
         baseRequest.setHandled(true);

         if (target.equals(basePath + "/launch"))
         {
            response.setStatus(HttpServletResponse.SC_OK);

            CountDownLatch down = new CountDownLatch(1);
            String[] url = new String[1];
            authWorker.addWork(() -> 
            {
               url[0] = WebHandler.spawn(null,
                                         null,
                                         true,
                                         options,
                                         "https://www.google.com",
                                         WebHandler.getRequestParameters(request));
               down.countDown();
            });
            
            try
            {
               down.await();
            }
            catch (InterruptedException e)
            {
               // ignore
               url[0] = null;
            }
            
            if (url[0] == null || url[0].startsWith(WebHandler.ERROR_PREFIX))
            {
               url[0] = "null";
            }

            response.getWriter().println(url[0]);
            return;
         }

         String[] page = null;
         
         if (target.equals(basePath + "/adm_windows.json"))
         {
            page = loadTemplate(pkgroot + "embedded/adm_windows.json");
         }
         else
         {
            // load the template from ${pkgroot}/embedded/index.html
            // TODO: fallback to FWD
            page = loadTemplate(pkgroot + "embedded/index.html");
         }
         
         if (page == null)
         {
            response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            return;
         }
         
         response.setStatus(HttpServletResponse.SC_OK);

         for (String line : page)
         {
            response.getWriter().println(line);
         }
      }
      
      /**
       * Loads the target resource template by lines.
       * 
       * @param    path
       *           The target resource path within jar file.
       * 
       * @return   The resource content by lines.
       */
      private String[] loadTemplate(String path)
      {
         LinkedList<String> lines = new LinkedList<String>();
         try
         {
            InputStream is = getClass().getResourceAsStream(path);
            if (is == null)
            {
               return null;
            }

            try (LineNumberReader reader = new LineNumberReader(new InputStreamReader(is)))
            {
               String line;
               while ((line = reader.readLine()) != null)
               {
                  lines.add(line);
               }
            };
         }
         catch (IOException ex)
         {
            // ignore
            return null;
         }
         
         return lines.toArray(new String[lines.size()]);
      }
   }
   
   /**
    * The in-JVM authenticated client which will be used to spawn web clients.
    * <p>
    * This FWD client must be a process account and configured to authenticate via a certificate, 
    * with the certificate details configured in:
    * <ul>
    *    <li><code>embeddedWebApp/authAlias</code> for the process alias.</li>
    *    <li><code>embeddedWebApp/authKeyStore</code> for the store with the certificate.</li>
    *    <li><code>embeddedWebApp/authKeyStorePassword</code> for the store password.</li>
    * </ul>
    * <p>
    * The key store file path will be either a full path or a path relative to the folder where
    * the FWD server is started.
    */
   private static class AuthWorker
   implements Runnable
   {
      /** Flag indicating if the task worker is running. */
      private volatile boolean running = false;
      
      /** The tasks to be ran. */
      private LinkedList<Runnable> tasks = new LinkedList<>();

      /**
       * Start this worker - it will spawn a thread and create an in-JVM FWD context for it.
       */
      public void start()
      {
         // these must be read from the server's context
         String alias = Utils.getDirectoryNodeString(null, "embeddedWebApp/authAlias", 
                                                     null, false);
         String ksfile = Utils.getDirectoryNodeString(null, "embeddedWebApp/authKeyStore", 
                                                      null, false);
         String kspass = Utils.getDirectoryNodeString(null, "embeddedWebApp/authKeyStorePassword", 
                                                      null, false);
         
         if (alias == null || ksfile == null || kspass == null)
         {
            throw new IllegalStateException("The embedded web app is configured to start, but " +
                                            " the authentication details are not configured.");
         }

         // X509/certificate authentication is possible only for non-context threads - so start 
         // a simple Java thread.
         new Thread(() ->
         {
            try
            {
               // wait for the FWD server to start.
               SessionManager.get().waitUntilReady();
            }
            catch (InterruptedException e1)
            {
               e1.printStackTrace();
               
               return;
            }

            SecurityManager sm = SecurityManager.getInstance();
            try
            {
               BootstrapConfig cfg = new BootstrapConfig();
               cfg.setConfigItem("net", "connection", "secure", "true");
               cfg.setConfigItem("security", "keystore", "processalias", alias);
               cfg.setConfigItem("security", "keystore", "filename", ksfile);
               cfg.setConfigItem("access", "password", "keystore", kspass);

               Object context = sm.authenticateServer(cfg);
               
               if (context == null)
               {
                  LOG.logp(Level.SEVERE,
                           "Could not authenticate the worker for the embedded web app!",
                           "",
                           LogHelper.generate("!"));
                  return;
               }
               
               authWorker.run();
            }
            catch (ConfigurationException | 
                   RestrictedUseException e)
            {
               LOG.logp(Level.SEVERE,
                        "Could not start the embedded web app authentication worker!",
                        "",
                        LogHelper.generate("!"),
                        e);
            }
         }).start();
      }
      
      /**
       * Post a new work to the {@link #tasks queue}.
       * 
       * @param    work
       *           The work to execute.
       */
      public synchronized void addWork(Runnable work)
      {
         tasks.add(work);
         this.notifyAll();
      }
      
      /**
       * Stop the worker thread.
       */
      public synchronized void stop()
      {
         this.running = false;
         this.notifyAll();
      }
      
      /**
       * The worker loop - waits for tasks to be posted in {@link #tasks queue}.
       */
      @Override
      public void run()
      {
         running = true;

         while (running)
         {
            Runnable task;
            synchronized (this)
            {
               while (running && tasks.isEmpty())
               {
                  try
                  {
                     this.wait();
                  }
                  catch (InterruptedException e)
                  {
                     e.printStackTrace();
                     break;
                  }
               }
               
               task = tasks.isEmpty() ? null : tasks.removeFirst();
            }
            
            if (running && task != null)
            {
               try
               {
                  task.run();
               }
               catch (Throwable t)
               {
                  LOG.logp(Level.SEVERE,
                           "AuthWorker.run",
                           "",
                           LogHelper.generate("Problem executing the given task!"),
                           t);
               }
            }
         }
      }
   }
}