/*
** Module   : Spawner.java
** Abstract : define Spawner contract.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131118 First version.
** 002 CA  20140206 Refactored to make it unaware of the type of the P2J client beeing spawned.
** 003 SBI 20160628 Added new methods to allocate and to release the client's system resources.
** 004 SBI 20171019 Changed to allocate new system resources in order to run new spawned client on
**                  the target host.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

/**
 * This interface is exported using a RemoteObject protocol and is designed as a communication 
 * gateway between the server handler and spawned client. 
 * 
 * @author mag
 *
 */
public interface Spawner
{
   /**
    * An exception is thrown if the system can't allocate free resources for the spawned client.
    */
   class OutOfResources
   extends Exception
   {
   };

   /**
    * Notify when the client has started.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    * @param    data
    *           Custom data sent by the P2J client back to the server.
    */
   public void clientIsReady(String uuid, Object data);
   
   /**
    * Depending on the implementation, export custom data from the server back to the client.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    *
    * @return   Custom data sent by the server back to the client.
    */
   public Object exportData(String uuid);
   
   /**
    * Get a {@link TemporaryClient} worker which will do the work after authenticating on the P2J
    * server using the temporary credentials.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    * 
    * @return   See above.
    */
   public TemporaryClient getTemporaryClient(String uuid);
   
   /**
    * Allocates the system resources for the spawned client if it is necessarily.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    * @param    host
    *           The host address of the remote client
    * 
    * @return   The system resources for the spawned client.
    * 
    * @throws   OutOfResources
    *           If the required resources can't be allocated.
    */
   public Object allocateClient(String uuid, String host)
   throws OutOfResources;
   
   /**
    * Releases the system resources for this spawned client.
    * 
    * @param    uuid
    *           Remote client assigned identifier.
    */
   public void releaseClient(String uuid);
}