/*
** Module   : TwoURLsResource.java
** Abstract : Merges two Jetty resources into one.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 HC  20170925 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.main;

import org.eclipse.jetty.util.resource.*;

import java.io.*;
import java.net.*;
import java.nio.channels.*;

/**
 * Jetty resource implementation that is capable to merge two other resources.
 */
public class TwoURLsResource
extends Resource
{
   /** First resource */
   private Resource res1;

   /** Second resource */
   private Resource res2;

   /**
    * Ctor.
    *
    * @param   res1
    *          First resource.
    * @param   res2
    *          Second resource.
    */
   public TwoURLsResource(Resource res1, Resource res2)
   {
      this.res1 = res1;
      this.res2 = res2;
   }

   /**
    * Closes this stream and releases any system resources associated
    * with it. If the stream is already closed then invoking this
    * method has no effect.
    */
   @Override
   public synchronized void close()
   {
      res1.close();
      res2.close();
   }

   /**
    * Checks whether resource exists.
    *
    * @return  see above
    */
   @Override
   public boolean exists()
   {
      return res1.exists() || res2.exists();
   }

   /**
    * Returns is-directory flag.
    *
    * @return  see above
    */
   @Override
   public boolean isDirectory()
   {
      return res1.isDirectory() || res2.isDirectory();
   }

   /**
    * Returns the last-modified time stamp.
    *
    * @return  see above
    */
   @Override
   public long lastModified()
   {
      return Math.max(res1.lastModified(), res2.lastModified());
   }

   /**
    * Returns the stream byte length.
    *
    * @return  see above
    */
   @Override
   public long length()
   {
      return -1;
   }

   /**
    * Returns resource url.
    *
    * @return  see above
    */
   @Override
   @Deprecated
   public URL getURL()
   {
      return res1.getURL();
   }

   /**
    * Returns resource file.
    *
    * @return  see above
    * @throws  IOException
    */
   @Override
   public File getFile()
   throws IOException
   {
      return null;
   }

   /**
    * Returns rsource name.
    *
    * @return  see above
    */
   @Override
   public String getName()
   {
      return res1.getName();
   }

   /**
    * Returns resource as stream.
    *
    * @return  see above
    * @throws  java.io.IOException
    */
   @Override
   public synchronized InputStream getInputStream()
   throws java.io.IOException
   {
      return null;
   }

   /**
    * Returns resource as {@link ReadableByteChannel}.
    *
    * @return  see above
    * @throws  IOException
    */
   @Override
   public ReadableByteChannel getReadableByteChannel() throws IOException
   {
      return null;
   }

   /**
    * Deletes the resource.
    *
    * @return  <code>true</code> if deleted success
    * @throws  SecurityException
    */
   @Override
   public boolean delete()
   throws SecurityException
   {
      throw new SecurityException("Delete not supported");
   }

   /**
    * Renames resource to another resource.
    *
    * @param   dest
    *
    * @return  <code>true</code> if success
    * @throws  SecurityException
    */
   @Override
   public boolean renameTo(Resource dest)
   throws SecurityException
   {
      throw new SecurityException("RenameTo not supported");
   }

   /**
    * Lists the sub resources if any.
    *
    * @return  array of sub resources, if none returns <code>null</code>
    */
   @Override
   public String[] list()
   {
      return null;
   }

   /**
    * Adds a path to the resource and returns a new resource representing the path addition.
    * The method is immutable.
    *
    * @param   path
    *          The path to add.
    * @return  the new resource representing the result of the operation
    * @throws  IOException
    */
   @Override
   public Resource addPath(String path)
   throws IOException
   {
      Resource res = res1.addPath(path);
      if (res == null || !res.exists())
      {
         res = res2.addPath(path);
      }

      return res;
   }

   /**
    * Returns the string representation.
    *
    * @return  see above
    */
   @Override
   public String toString()
   {
      return res1.toString() + "; " + res2.toString();
   }

   /**
    * Checks whether this resource is contained in the supplied resource.
    *
    * @return <code>true</code> if contained, <code>false</code> otherwise
    */
   @Override
   public boolean isContainedIn (Resource containingResource)
   throws MalformedURLException
   {
      return false;
   }
}
