/*
** Module   : WebClientBuilderParameters.java
** Abstract : store the web client process parameters
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 CA  20140206 Created initial version, extracted from ClientBuilderParameters.
** 002 GES 20150312 Store the web client type.
** 003 GES 20160203 Added trusted mode support and removed some dead code.
** 004 CA  20160428 Added command-line configuration options for web-clients.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.util.*;

/**
 * This structure holds default parameters as well specific parameters used to create and spawn a 
 * process for a web client (either GUI or ChUI). 
 */
public class WebClientBuilderParameters
extends ClientBuilderParameters
{   
   /** Referrer. */
   private String referrer;
   
   /** Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for ChUI). */
   private boolean gui;

   /** Enables trusted launch mode (no password is used and the OS account is forced). */
   private boolean trusted = false;
   
   /** The command line-configuration options. */
   private final String[] cfgOptions;

   /**
    * Constructor. Assign a unique UUID to this process.
    *
    * @param    gui
    *           Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for
    *           ChUI).
    * @param    trusted
    *           If this instance will be used to launch clients in trusted mode.
    * @param    referrer
    *           The referrer string to provide the web client. If the client gets disconnected
    *           it will be redirected back to this URL.
    * @param    cfgOptions
    *           The command-line configuration options.  May be <code>null</code>.
    */
   public WebClientBuilderParameters(boolean  gui, 
                                     boolean  trusted, 
                                     String   referrer, 
                                     String[] cfgOptions)
   {
      super(new WebClientBuilderOptions(trusted));
      this.gui      = gui;
      this.trusted  = trusted;
      this.referrer = referrer;
      this.cfgOptions = cfgOptions;
   }
   
   /**
    * Get Terminal Options.
    * 
    * @return  The terminal options.
    */
   public Map<String, String> getClientOptions()
   {
      return ((WebClientBuilderOptions) options).getClientOptions();
   }

   /**
    * Get referrer address.
    * 
    * @return  Back link URL
    */
   public String getReferrer()
   {
      return referrer;
   }

   /**
    * Get the flag that reports which web client type is in use.
    * 
    * @return   Client type (<code>true</code> for GUI, <code>false</code> for ChUI).
    */
   public boolean isGui()
   {
      return gui;                  
   }                                 

   /**
    * Get the flag that reports is spawning should use trusted mode.
    *
    * @return   Trusted mode flag.
    */
   public boolean isTrusted()
   {
      return trusted;
   }

   /**
    * Get the command-line configuration options.
    * 
    * @return   The {@link #cfgOptions}.
    */
   public String[] getConfigOptions()
   {
      return cfgOptions;
   }
}