/*
** Module   : WebClientsManager.java
** Abstract : WebClientsManager implements business methods to allocate and release web clients.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170628 Created initial version.
** 002 SBI 20171020 Changed to allocate new system resources in order to run new web client on
**                  the target host, changed to map a host and its port to its web client
**                  identifier.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.main;

import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;
import java.util.logging.*;

import javax.net.ssl.SSLContext;

import org.eclipse.jetty.client.HttpClient;
import org.eclipse.jetty.client.api.Response.CompleteListener;
import org.eclipse.jetty.http.HttpScheme;
import org.eclipse.jetty.util.ssl.SslContextFactory;

import com.goldencode.p2j.net.RemoteObject;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.net.SessionListener;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.Utils;


/**
 * Manages the system resources allocated for web clients.
 */
public class WebClientsManager
implements WebClientAllocator,
           WebClientRegistrar,
           SessionListener
{
   /** Logger */
   private static final Logger LOG = 
            LogHelper.getLogger(WebClientsManager.class.getName());
   
   /** connect timeout */
   private static final int TIMEOUT = 60000;
   
   /** Maximal tries to release resources */
   private static final int MAX_RELEASE_TRIES = 10;
   
   /** interval between connection tests */
   private static final int TESTS_PERIOD = 10000;
   /**
    * one instance per standard server
    */
   private static WebClientsManager instance;
   
   /**
    * True value indicates that ports range is given
    */
   private final boolean portsRestricted;
   
   /** ports range started from this port number */
   private final int from;
   
   /** ports range bounded by this port number */
   private final int to;
   
   /**
    * Prefix is used to build mapping from ports to strings
    */
   private final String prefix;
   
   /**
    * Maps the web client's remote address pair to its configuration
    */
   private final Map<Integer, WebClientConfig> webClientSessions;
   
   /**
    * Maps client's uuid to its configuration, the client uuid are unique within the given server
    */
   private final Map<String, WebClientConfig> webClientConfigs;
   
   /**
    * Maps the web client host to the web client's port to its configuration 
    */
   private final Map<String, Map<Integer, WebClientConfig>> usedLocalPorts;
   
   /**
    * The forwarded host value provided by the directory settings, this value overrides
    * the corresponding value given by the "X-Forwarded-Host" HTTP header.
    */
   private final String forwardedHostByDirectory;
   
   /**
    * The forwarded protocol provided by the directory settings, this value overrides
    * the corresponding value given by the "X-Forwarded-Proto" HTTP header.
    */
   private final String forwardedProtoByDirectory;
   
   /**
    * Used to test if this web client is still alive and free its resources if it is not. Although
    * http client can allocate any free local ports it is supposed that the web client ports range
    * and the http client ports range cover different ports numbers.
    * TODO substitute it by asynchronous socket test 
    */
   private HttpClient httpClient;
   
   /**
    * Creates a unique instance per a standard P2J server.
    * 
    * @throws    Exception if it is not setup properly. 
    */
   private WebClientsManager()
   throws Exception
   {
      from = Utils.getDirectoryNodeInt(null, "webClient/portsRange/from", -1, false);
      
      to = Utils.getDirectoryNodeInt(null, "webClient/portsRange/to", -1, false);
      
      prefix = Utils.getDirectoryNodeString(null, "webClient/portsRange/namePrefix", "client",
               false);
      
      portsRestricted =  (to > from) && (from > 0);
      
      forwardedHostByDirectory = Utils.getDirectoryNodeString(
               null, "webClient/portsRange/forwardedHost", null, false);
      forwardedProtoByDirectory = Utils.getDirectoryNodeString(
               null, "webClient/portsRange/forwardedProto", HttpScheme.HTTPS.asString(), false);
      
      webClientSessions = new LinkedHashMap<Integer, WebClientConfig>();
      
      webClientConfigs = new LinkedHashMap<String, WebClientConfig>();
      
      usedLocalPorts = new LinkedHashMap<String, Map<Integer, WebClientConfig>>();
      
      RemoteObject.registerNetworkServer(WebClientRegistrar.class, this);
      
      if (portsRestricted)
      {
         // setup ssl engine to test
         SslContextFactory sslContextFactory = new SslContextFactory();
         SSLContext sslContext = SecurityManager.getInstance().getSecureSocketContext();
         sslContextFactory.setSslContext(sslContext);
         
         httpClient = new HttpClient(sslContextFactory);
         // Start HttpClient
         httpClient.start();
      }
   }

   /**
    * Returns the web clients manager or null if is is not setup properly.
    * <p>
    * At this time the code doesn't throw exceptions, it just logs them in the standard error
    * stream.
    * 
    * @return   The web clients manager
    */
   public static synchronized WebClientsManager getInstance()
   {
      if (instance == null)
      {
         try
         {
            instance = new WebClientsManager();
         }
         catch (Exception e)
         {
            LOG.log(Level.SEVERE, "Failed to create " + WebClientsManager.class.getName(), e);
         }
      }
      
      return instance;
   }
   
   /**
    * Allocates the dedicated port number from the restricted ports range for the new spawned
    * web client.
    * 
    * @param    requestParameters
    *           Holds additional parameters if they are provided. It is expected that the first
    *           requestParameters[0] holds the forwarded host of "X-Forwarded-Host" HTTP header
    *           and requestParameters[1] holds the corresponding forwarded protocol that is 
    *           provided by "X-Forwarded-Proto" value.
    * @param    host
    *           The remote host for the spawned web client
    * 
    * @return   The web client configuration. The allocated resources for the web client that
    *           will be spawned.
    */
   public WebClientConfig allocateClient(String[] requestParameters, String host)
   {
      
      String forwardedHost;
      String forwardedProto;
      if (requestParameters != null && requestParameters.length >= 2)
      {
         forwardedHost = requestParameters[0];
         forwardedProto = requestParameters[1];
      }
      else
      {
         forwardedHost = null;
         forwardedProto = null;
      }
      
      forwardedHost = forwardedHostByDirectory != null ? forwardedHostByDirectory : forwardedHost;
      forwardedProto = forwardedProtoByDirectory != null ? forwardedProtoByDirectory : forwardedProto;
      
      boolean viaProxyServer = forwardedHost != null && !forwardedHost.isEmpty();
      
      if (viaProxyServer && !portsRestricted)
      {
         throw new IllegalStateException("Correct range: "
                  + "'webClient/portsRange/from' < 'webClient/portsRange/to' must be provided");
      }
      
      StringBuilder webRootBuilder = new StringBuilder();
      webRootBuilder.append("/");
      String webClientHost = host;
      
      int webClientPort = getFreeLocalPort(host);

      if (webClientPort == -1)
      {
         // all available ports are in use
         //try to find free port via connection test
         webClientPort = tryAcquirePort(host);
         
         if (webClientPort == -1)
         {
            return null;
         }
      }
      
      if (viaProxyServer)
      {
         webRootBuilder.append("server");
         webRootBuilder.append("/").append(getPortName(webClientHost, webClientPort));
      }
      
      WebClientConfig webClient = new WebClientConfig();
      webClient.setHost(webClientHost);
      webClient.setPort(webClientPort);
      webClient.setWebRoot(webRootBuilder.toString());
      webClient.setForwardedHost(forwardedHost);
      webClient.setForwardedProto(forwardedProto);
      webClient.setSessionsCounter(new AtomicLong(0));
      
      if (portsRestricted)
      {
         Map<Integer, WebClientConfig> localPorts = usedLocalPorts.get(host);
         
         localPorts.put(webClientPort, webClient);
      }
      
      return webClient;
   }

   /**
    * Registers the P2J session between the server and the web client to release its system
    * resources on the server side in the case if the web client is failed unexpectedly.
    * 
    * @param    webClientId
    *           The web client unique identifier
    * @param    peerNode
    *           The remote node of the P2J session between the server and the web client
    */
   @Override
   public void registerWebClientSession(String webClientId, int peerNode)
   {
      if (!portsRestricted)
      {
         return;
      }
      synchronized (this)
      {
         WebClientConfig config = webClientConfigs.get(webClientId);
         if (config != null)
         {
            config.getSessionsCounter().incrementAndGet();
            
            webClientSessions.put(peerNode, config);
         }
      }
   }
   
   /**
    * Sets the unique identifier for the target web client.
    * 
    * @param    webClientId
    *           The provided unique identifier
    * @param    config
    *           The web client configuration
    */
   public void registerWebClientUuid(String webClientId, WebClientConfig config)
   {
      if (!portsRestricted)
      {
         return;
      }
      synchronized (this)
      {
         config.setUuid(webClientId);
         webClientConfigs.put(webClientId, config);
      }
   }

   /**
    * This method is not implemented by WebClientsManager.
    *
    * @param    session
    *           The session that is starting.
    */
   @Override
   public void initialize(Session session)
   {
   }

   /**
    * Releases the system resources for the web client.
    * 
    * @param    config
    *           The web client configuration
    */
   @Override
   public void freeClient(WebClientConfig config)
   {
      if (!portsRestricted)
      {
         return;
      }
      synchronized (this)
      {
         if (config == webClientConfigs.remove(config.getUuid()))
         {
            Map<Integer, WebClientConfig> localPorts = usedLocalPorts.get(config.getHost());
            localPorts.remove(config.getPort());
         }
      }
      
   }

   /**
    * Releases the web client resources managed by the server.
    * 
    * @param    webClientId
    *           The web client unique identifier
    */
   @Override
   public void freeWebClientResources(String webClientId)
   {
      if (!portsRestricted)
      {
         return;
      }
      synchronized (this)
      {
         WebClientConfig config = webClientConfigs.get(webClientId);
         if (config != null)
         {
            freeClient(config);
         }
      }
   }

   /**
    * This method is called when the session is ending. Removes this session from the active
    * sessions and to start asynchronous connect test in order to free the unused port.
    *
    * @param    session
    *           The session that is ending.
    */
   @Override
   public void terminate(Session session)
   {
      if (!portsRestricted)
      {
         return;
      }
      
      int peerNode = session.getRemoteAddress();
      
      synchronized (this)
      {
         WebClientConfig config = webClientSessions.remove(peerNode);
         if (config != null)
         {
            isStillAlive(config, result ->
            {
               long expectedCount = config.getSessionsCounter().get();
               if (result.isFailed())
               {
                  freeClient(config);
               }
               else
               {
                  tryReleasePort(config, expectedCount);
               }
            });
         }
      }
   }
   
   /**
    * Causes the current thread to sleep for a given period of milliseconds
    * 
    * @param    timeout
    *           A given period of milliseconds
    */
   private void sleep(long timeout)
   {
      long elapsed = 0;
      long start = System.currentTimeMillis();
      while (timeout > 0)
      {
         try
         {
            Thread.sleep(timeout);
         }
         catch (InterruptedException e)
         {
         }
         finally
         {
            elapsed = System.currentTimeMillis() - start;
            timeout -= elapsed;
         }
      }
   }
   
   /**
    * Try to release port acquired by the target web client on triggered terminated session events,
    * but it might not free it if is still in use (client was restarted) or if the client is
    * 'runaway' and the port is open at the OS level, without re-using it.
    * 
    * @param    config
    *           The web client configuration
    * @param    expectedCount
    *           The expected sessions count
    */
   private void tryReleasePort(WebClientConfig config, long expectedCount)
   {
      Executors.newSingleThreadExecutor().submit(() ->
      {
         int tries = 0;
         try
         {
            while (config.getSessionsCounter().compareAndSet(expectedCount, expectedCount) &&
                   (tries < MAX_RELEASE_TRIES))
            {
               tries++;
               sleep(TESTS_PERIOD);
               
               CountDownLatch synchronizer = new CountDownLatch(1);
               
               isStillAlive(config, result ->
               {
                  if (result.isFailed())
                  {
                     freeClient(config);
                     synchronizer.countDown();
                  }
               });
               
               if (synchronizer.await(TIMEOUT, TimeUnit.MILLISECONDS))
               {
                  return;
               }
            }
         }
         catch(InterruptedException e)
         {
            Thread.currentThread().interrupt();
            throw new RuntimeException(e);
         }
         if (config.getSessionsCounter().compareAndSet(expectedCount, expectedCount))
         {
            LOG.log(Level.SEVERE, "The port '" + config.getPort() + "' couldn't be freed because" +
                     "is still open at OS level and the client uuid='" + config.getUuid() +
                     "' is not re-using it.");
         }
      });
   }
   
   /**
    * Runs over the available ports range and gets a first free port number to use for the target
    * web client host.
    * 
    * @param    host
    *           The target web client host
    * 
    * @return   The free port or -1 if all available ports are used
    */
   private synchronized int getFreeLocalPort(String host)
   {
      if (!portsRestricted)
      {
         return 0;
      }

      int freePort = from;
      
      Map<Integer, WebClientConfig> localPorts = usedLocalPorts.get(host);
      
      if (localPorts == null)
      {
         localPorts = new LinkedHashMap<Integer, WebClientConfig>();
         usedLocalPorts.put(host, localPorts);
      }
      
      while (freePort <= to)
      {
         if (localPorts.get(freePort) == null)
         {
            return freePort;
         }
         
         freePort++;
      }
      
      return -1;
   }
   
   /**
    * Try to acquire a first available port by doing a connection test for the target web client
    * host.
    * 
    * @param    host
    *           The target web client host
    * 
    * @return   The free port or -1 if all available ports are used
    */
   private int tryAcquirePort(String host)
   {
      
      CountDownLatch synchronizer = new CountDownLatch(1);
      
      AtomicInteger availablePort = new AtomicInteger(-1);
      
      int port = from;
      
      while (port <= to)
      {
         synchronized (this)
         {
            Map<Integer, WebClientConfig> localPorts = usedLocalPorts.get(host);
            
            WebClientConfig config = localPorts.get(port);
            
            if (config == null)
            {
               return port;
            }
            else
            {
               if (availablePort.compareAndSet(-1, -1))
               {
                  isStillAlive(config, result ->
                  {
                     if (result.isFailed())
                     {
                        freeClient(config);
                        if (availablePort.compareAndSet(-1, config.getPort()))
                        {
                           synchronizer.countDown();
                        }
                     }
                  });
               }
               else
               {
                  return availablePort.get();
               }
            }
         }
         
         port++;
      }
      
      try
      {
         synchronizer.await(TIMEOUT, TimeUnit.MILLISECONDS);
      }
      catch (InterruptedException e)
      {
         Thread.currentThread().interrupt();
         throw new RuntimeException(e);
      }
      
      return availablePort.get();
   }
   
   /**
    * Represents a well-defined mapping from host ports numbers to their names. 
    * 
    * @param    host
    *           The host
    * @param    port
    *           The port number
    * 
    * @return   The port name
    */
   private String getPortName(String host, int port)
   {
      Integer hostIndex;
      try
      {
         hostIndex = BrokerManager.getOrderedNumber(host);
      }
      catch (UnknownHostException e)
      {
         hostIndex = null;
      }
      // hostIndex must be valid
      if (hostIndex == null)
      {
         // this case can't happen during runtime due to the logic implemented by the BrokerManager
         StringBuilder message = new StringBuilder("Unknown host: ");
         message.append(host);
         
         throw new IllegalStateException(message.toString());
      }
      
      StringBuilder portName = new StringBuilder();
      
      portName.append(prefix).append(hostIndex).append(port - from + 1);
      
      return portName.toString();
   }
   
   /**
    * Starts asynchronous connect test in order to free the unused port.
    * 
    * @param    config
    *           The tested web client configuration
    * @param    listener
    *           Listener for the request and response completed event.
    */
   private void isStillAlive(WebClientConfig config, CompleteListener listener)
   {
      if (config != null)
      {
         StringBuilder connectTest = new StringBuilder();
            connectTest.append(HttpScheme.HTTPS.asString()).append("://").append(config.getHost()).append(":")
               .append(config.getPort()).append("/");
         httpClient.newRequest(connectTest.toString()).timeout(TIMEOUT, TimeUnit.MILLISECONDS)
            .send(listener);
      }
   }   
}
