/*
** Module   : LocalStateSync.java
** Abstract : used by a proxy to implement state synchronization where both
**            "sides" are in the same JVM
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20111017 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
                              
package com.goldencode.p2j.net;


import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import com.goldencode.p2j.util.*;


/**
 * Used by a proxy to implement state synchronization where both "sides" are in
 * the same JVM.  Use {@link #sync} to drive the process. This class is needed
 * by proxy implementations which are run without a network environment. This
 * maintains the semantics of state synchronization which is necessary since
 * any local redirector environments have no {@link Session} object which would
 * otherwise service that requirement. 
 */
public class LocalStateSync
{
   /** Source of state changes. */
   private StateSynchronizer source = null;
   
   /** Sink for state changes. */
   private StateSynchronizer sink = null;
   
   /**
    * Constructor.
    *
    * @param    source
    *           The state synchronizer from which to obtain changes. Must not
    *           be <code>null</code>.
    * @param    sink
    *           The state synchronizer in which to apply changes. Must not
    *           be <code>null</code>.
    */      
   public LocalStateSync(StateSynchronizer source, StateSynchronizer sink)
   {
      this.source = source;
      this.sink   = sink;
   }
   
   /**
    * Drive the state synchronization process by reading any pending changes
    * and applying them if any changes are available. This method is normally
    * called before any method invocation can occur in a proxy object.
    * <p>
    * Changes are obtained via {@link StateSynchronizer#getChanges} and they
    * are applied via {@link StateSynchronizer#applyChanges}. This method is
    * synchronized to make all processing thread-safe.
    */
   public synchronized void sync()
   {
      Serializable state = source.getChanges();
      
      if (state != null)
      {
         sink.applyChanges(state);
      }
   }
}