/*
** Module   : CopySpec.java
** Abstract : defines a file copy operation
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20070716   @34527 Created initial version. Defines a file copy
                             operation (source, target and type of copy).
** 002 GES 20130206          Moved class to the pattern package.
*/                                                            
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.pattern;

/**
 * Defines a file copy operation (source, target and type of copy).
 * <p>
 * The following copy operation types are defined:
 * <p>
 * <pre>
 * Type                  Description
 * -------------------   ----------------------------------------
 * Type.BINARY           A byte for byte copy.
 * Type.TXT_TO_HTML      Each line of the input text file will be
 *                       copied to the output file as a line numbered
 *                       named anchor inside HTML preformatted tags.
 * </pre>
 */
public class CopySpec
{
   /** Operation type constants. */
   public enum Type { BINARY, TXT_TO_HTML };
   
   /** Source filename. */
   private String source = null;
   
   /** Target filename. */
   private String target = null;
   
   /** Operation. */
   private Type op = Type.BINARY;
   
   /**
    * Create an instance with the given filenames. The copy operation type
    * will default to <code>Type.BINARY</code>.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    */
   public CopySpec(String source, String target)
   {
      this.source = source;
      this.target = target;
   }
   
   /**
    * Create an instance with the given filenames.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    * @param    op
    *           Copy operation to use.
    */
   public CopySpec(String source, String target, Type op)
   {
      this(source, target);
      this.op = op;
   }
   
   /**
    * Access the source filename.
    *
    * @return   Source.
    */
   public String getSource()
   {
      return source;
   }
   
   /**
    * Set the source filename.
    *
    * @param    source
    *           New source filename.
    */
   public void setSource(String source)
   {
      this.source = source;
   }
   
   /**
    * Access the target filename.
    *
    * @return   Target.
    */
   public String getTarget()
   {
      return target;
   }
   
   /**
    * Set the target filename.
    *
    * @param    target
    *           New target filename.
    */
   public void setTarget(String target)
   {
      this.target = target;
   }
   
   /**
    * Access the copy operation.
    *
    * @return   Copy operation.
    */
   public Type getOperation()
   {
      return op;
   }
   
   /**
    * Set the copy operation type.
    *
    * @param    op
    *           New copy operation type.
    */
   public void setOperation(Type op)
   {
      this.op = op;
   }
}