/*
** Module   : BufferManager.java
** Abstract : Coordinates life cycle of record buffers
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050926   @23246 Created initial version. Manages multiple
**                           record buffers within each context. Maps
**                           legacy transactions to database-level
**                           transactions and registers buffers for commit
**                           and rollback notifications.
** 002 ECF 20051108   @23376 Adjust to new architecture of one Persistence
**                           instance per database. Most changes were in
**                           the inner class DBTxWrapper.
** 003 ECF 20051205   @23670 Renamed method obtain() to get() for
**                           consistency within package.
** 004 ECF 20060116   @23950 Fixed regression in initializeTransactions.
**                           Was impossible to begin a new database
**                           transaction.
** 005 ECF 20060120   @24037 Integration with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 006 ECF 20060130   @24285 Changed openScope signature. Now returns
**                           current scope depth.
** 007 ECF 20060205   @24286 Implement BatchListener and register with the
**                           TransactionManager for batchNotify events.
**                           Triggers start/end of record buffer batch
**                           assign mode.
** 008 ECF 20060428   @25869 Fixed dangling transaction defect. A database
**                           level which is not explicitly committed or
**                           rolled back by the TransactionManager is now
**                           closed by auto-rollback processing.
** 009 GES 20060517   @26197 Match new Finalizable interface requirement.
** 010 ECF 20060610   @27090 Fixed retry defect. A new database level
**                           transaction must be opened upon master
**                           transaction retries. This prevents a later
**                           commit or rollback from failing.
** 011 ECF 20060621   @27503 Added activeRecords method. Returns iterator
**                           on the records currently stored in all record
**                           buffers in the current client context for a
**                           particular database.
** 012 ECF 20060627   @27587 Modified activeRecords method. Only include
**                           non-transient records to avoid saving records
**                           which were not meant to be saved.
** 013 ECF 20060628   @27668 Added getCurrentScope method. Reports current
**                           buffer scope level.
** 014 ECF 20060719   @28110 Simplified activeRecords() implementation.
** 015 ECF 20060728   @28263 Moved DMO reference counting mechanism from
**                           RecordBuffer. incrementDMOUseCount() and
**                           decrementDMOUseCount() are now implemented
**                           here.
** 016 GES 20060804   @28431 Some javadoc improvements.
** 017 ECF 20060916   @29664 Added DMO property-level undo support. A
**                           scoped dictionary tracks all changes by DMO
**                           type, necessary to undo changes to any block
**                           scope. Also moved storage of record-level
**                           reversible actions from RecordBuffer class.
** 018 ECF 20060919   @29695 Fixed DMO eviction. The DMO implementation
**                           class must be passed to decrementDMOUseCount,
**                           because there previously was not enough
**                           information available to detect the proper
**                           DMO type when evicting objects from the
**                           persistence session.
** 019 ECF 20061010   @30307 Added support for auto-commit temp tables.
**                           Buffers which auto-commit are not managed by
**                           this class' transaction mechanism.
** 020 ECF 20061011   @30355 Rewrote logic to begin database transactions.
**                           Eliminated initializeTransactions() method
**                           and simplified scopeStart() and openScope()
**                           logic. This change closes some loopholes
**                           which would not have begun database level
**                           transactions at the correct time.
** 021 ECF 20061018   @30486 Moved processing to clean up batch assign
**                           mode from RecordBuffer. This was interfering
**                           with end-of-scope processing in some cases.
** 022 ECF 20061114   @31159 Moved batch assign mode state management from
**                           RecordBuffer to this class. Added package
**                           private methods to allow RecordBuffer access
**                           to batch mode state.
** 023 ECF 20061207   @31618 Fixed DMO validation problem. ChangeBroker
**                           needs to be marked for a pending session
**                           flush whenever a record buffer is added to
**                           the set of dirty buffers in batch assign
**                           mode.
** 024 ECF 20070130   @32037 Fixed handling of no-undo temp-tables. Change
**                           undone with a database-level transaction
**                           rollback is re-rolled forward and committed,
**                           as if the rollback had not occurred.
** 025 ECF 20070412   @32976 Fixed end of transaction lock processing.
**                           Record locks for a session are now managed by
**                           RecordLockContext.
** 026 ECF 20070414   @32993 Fixed memory leak. DBTxWrapper was holding on
**                           to record buffers after they had gone out of
**                           scope. The new locking architecture has
**                           eliminated the need for this object to hold
**                           buffer references at all.
** 027 ECF 20070502   @33368 Added lookupBuffer(). Required changes to how
**                           scoped dictionary for open buffers is managed
**                           internally.
** 028 ECF 20070503   @33390 Added code to track all defined buffers in a
**                           scoped dictionary. Added registerBuffer()
**                           method to allow new buffers to register
**                           themselves. Changed lookupBuffer() to use the
**                           new dictionary of all buffers rather than
**                           just open buffers.
** 029 ECF 20070515   @33668 Optimized startScope() slightly. Only process
**                           DBTxWrapper if beginning a full transaction.
** 030 ECF 20070601   @33916 Fixed potential memory leak. Release all
**                           reversible rollbacks at the end of a database
**                           transaction, rather than only at rollback.
** 031 ECF 20070821   @34901 Added setSilentError() method. Enables error
**                           suppression during query construction and/or
**                           initialization to mimic Progress' lenient
**                           error handling during its processing of a
**                           where clause.
** 032 ECF 20070911   @35283 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections.
** 033 ECF 20071024   @35594 Added call to allow ConnectionManager to
**                           clean up after a transaction ends. This was
**                           necessary to avoid this cleanup too early in
**                           the end-of-transaction logic flow. This
**                           problem was possible with the previous
**                           implementation, in which ConnectionManager
**                           registered itself with the TM independently.
** 034 ECF 20071029   @35632 Changed in[de]crementDMOUseCount() signature.
**                           Return use counts after change is made. Also
**                           integrated some use of generics.
** 035 ECF 20080516   @38600 Added support for dirty index update sharing.
**                           Integrated generics.
** 036 ECF 20080606   @38637 Fixed reversible rollback processing. Ensure
**                           that it only takes place when rolling back
**                           temp table transactions.
** 037 SVL 20080607   @38638 Fixed DBTxWrapper.iterate(). This method
**                           needs to perform cleanup processing for the
**                           master transaction by calling end() before
**                           begin(). [ECF: same change made for retry()].
** 038 CA  20080808   @39357 Implemented no-undo temp-table rollback - on
**                           db tx rollback, the re-roll actions are 
**                           processed for the no-undo buffers.
** 039 CA  20080826   @39568 Added rollbackPending which is called by TM
**                           when rollback targets a parent block.
** 040 ECF 20080827   @39601 Fixed lookupBuffer(). Addition of
**                           pendingBuffers logic regressed this method.
** 041 CA  20080820   @39451 Implemented DatabaseConnectionHandler - it 
**                           checks if an exception related to database
**                           connectivity errors should be re-thrown or 
**                           not. Fixed leak in allBuffers scoping.
** 042 ECF 20080925   @39945 Modified for ReversibleChain API change. That
**                           class' c'tor requires an additional argument.
** 043 ECF 20081020   @40207 Modified for Reversible API change. Method
**                           Reversible.rollback() now requires arguments.
** 044 ECF 20081118   @40548 Fixed reversible rollback processing in
**                           DBTxWrapper.rollback(). Needed a loop before
**                           actual rollback, to call Reversibles'
**                           prepareRollback() method.
** 045 ECF 20081127   @40711 Implement reclamation of temp table primary
**                           keys. On full transaction commit/rollback,
**                           keys which have been collected for recycling
**                           are made available for re-use, or discarded,
**                           respectively.
** 046 SVL 20081204   @40795 Added nestedOpen parameter to openScope()
**                           method.
** 047 ECF 20090114   @41161 Minor change to DBTxWrapper. begin() is now
**                           called from scopeStart() after all buffers
**                           have been processed. It is called from
**                           openScope() for a newly created DBTxWrapper.
** 048 ECF 20090121   @41193 Fixed defect in scopeStart()/scopeFinished().
**                           Buffers were having their enterBlock() and
**                           exitBlock() methods (respectively) invoked
**                           too often in the case of nested, open buffer
**                           scopes.
** 049 ECF 20090128   @41238 Changed activeRecords() to activeBuffers().
**                           Now returns an iterator on all active record
**                           buffers, rather than just the records they
**                           contain. The caller required additional
**                           information stored by the buffers themselves.
** 050 ECF 20090206   @41265 Added support for evicting unused DMOs in the
**                           P2J session interceptor. Exposed a view of
**                           DMOs currently in use so that the session
**                           interceptor can evict unused DMOs after a
**                           session flush. Also implemented a minor
**                           performance improvement for batch mode. We
**                           now pass the BufferManager instance to
**                           RecordBuffer.cleanupBatchMode() method to
**                           avoid an unnecessary context-local lookup.
** 051 ECF 20090218   @41330 Changed idiom for scoped maps of Reversibles.
**                           The get*() methods which return these maps
**                           will now return null if the map did not
**                           previously exist, and the caller did not set
**                           the create option to true. This avoids the
**                           unnecessary creation of maps in the common
**                           case of looking up an entry which will never
**                           be found.
** 052 ECF 20090221   @41444 Added removeAllReversibles(). Also removed
**                           unused parameter from decrementDMOUseCount().
** 053 ECF 20090315   @41608 Fixed NO-UNDO temp-table reversible rollback
**                           processing. It is necessary to prepare the
**                           reversibles for rollback before the old
**                           session is rolled back. The reversals are
**                           then executed in the new session. This allows
**                           us to initialize lazy persistent collections
**                           for extent fields before the old session is
**                           gone. This is a corner case for reversible
**                           deletes when the target DMO is also the
**                           current record in the buffer at the time of
**                           rollback.
** 054 ECF 20090610   @42632 Added evictDMOIfUnused(). Evicts DMO from
**                           Hibernate session if no resource references it.
** 055 ECF 20090623   @42962 Simplified internals and reduced memory
**                           footprint. Refactored four ScopedDictionary
**                           instances for reversible/undoable processing into
**                           one ScopedDictionary of UndoableData objects.
**                           Simplified API to access reversible/undoable
**                           action maps.
** 056 ECF 20090701   @43010 Fixed regressions introduced with #055 (@42962).
**                           getReversibles() and getUndoables() were not
**                           properly looking up or creating entries in the
**                           backing ScopedDictionary.
** 057 ECF 20090702   @43027 Fixed additional regressions from #055 (@42962).
**                           Reversible rollbacks were not being processed
**                           correctly.
** 058 ECF 20090716   @43213 Enabled record buffers to be lazily initialized.
**                           This required some refactoring with respect to
**                           scope open/close logic. Also ensure change is
**                           flushed to the database at the close of each
**                           block.
** 059 ECF 20090724   @43415 Fixed reversible/undoable mapping. Use master
**                           buffer instead of current buffer as map key. This
**                           avoids losing undo information for shared temp
**                           table buffers which go out of scope before their
**                           master temp tables. Also changed session flush
**                           strategy. We now try to flush before rollback at
**                           any level, rather than on scope finish events.
** 060 ECF 20090729   @43447 Adapted Commitable API change. Added validate()
**                           method to Commitable implementations.
** 061 CA  20090807   @43556 Reversibles must be kept by database and table
**                           name, not by RecordBuffer instance - to ensure
**                           proper reversible order when rolling back records
**                           in a table with multiple buffers defined for it.
**                           This fixes @43542.
** 062 ECF 20090810   @43582 Changed getBufferForEntity() to prevent NPE. This
**                           is needed, as RecordBuffer.getDatabaseAndTable()
**                           will now return null instead of throwing an
**                           IllegalStateException if the buffer is not yet
**                           initialized. Throw RuntimeException in various
**                           places (rather than a ConditionException), if an
**                           error is unrecoverable.
** 063 ECF 20090811   @43586 Changed decrementDMOUseCount() to prevent NPE.
**                           In a case where the thread is interrupted due to
**                           a session ending, it was possible for the
**                           increment and decrement calls to become out of
**                           balance.
** 064 ECF 20090811   @43598 Fixed regression in #061 (@43556). Reversibles
**                           were being stored by database and table name, but
**                           retrieved by RecordBuffer instance.
** 065 ECF 20090812   @43602 Changed Reversible key from String to new class
**                           BufferType. Removed getBufferForEntity(), which
**                           would fail to find a matching buffer in cases
**                           where buffers had their scopes closed before a
**                           transaction ended. A buffer lookup is no longer
**                           needed, since BufferType contains all information
**                           needed by former callers of that method.
** 066 ECF 20090816   @43665 Integrated with FieldAssigner. We get a context
**                           local Scopeable from that class, which we notify
**                           at each scope open and close event.
** 067 SVL 20091117   @44401 Added facility for releasing buffers into which a
**                           new record was loaded (or the old record was
**                           reloaded) at specific points of block processing.   
** 068 CA  20091202   @44466 Reworked the rollback mechanism, to maintain 
**                           proper reversible order and thus to solve some
**                           unique-constraint violations during rollback of
**                           some bulk delete cases and mixed record operations.
** 069 SIY 20100429   @44838 Fixed double rollback in some conditions.
** 070 CS  20121031          Removed client specific references.
** 071 OM  20130902          Added getDMOUseCount() for detecting if a dmo is referred by more
**                           than one record buffer.
** 072 HC  20130902          For #2164 added BufferManager initialization check.
**                           BufferManager is initialized when its initialize() 
**                           method is called. If not initialized get() throws.
** 073 SVL 20130822          Added openFreeScope.
** 074 ECF 20131028          Made class and getCurrentScope() public and added import statement to
**                           accommodate new lock sub-package.
** 075 SVL 20131028          Added openScopeAt and tracking functions for static temp tables.
** 076 CA  20131209          Added no-op deleted() and scopeDeleted() methods, requried by the
**                           changes in Scopeable and Finalizable interfaces.
** 077 SVL 20140124          Added includeTransient parameter for activeBuffers() function.
** 078 SVL 20140221          [INPUT-]OUTPUT TABLE-HANDLE parameters are tracked.
** 079 CA  20140301          Allow buffers to survive the persistent procedure's external
**                           procedure.
** 080 OM  20140417          Fixed typo in method name. Simplified string concatenation
**                           operations.
** 081 CA  20140407          Fixed scopeDeleted(), when it is called on deletion of an appserver
**                           persistent procedure.
** 082 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 083 ECF 20140813          Manage batch assign brackets in a scoped manner, since now these
**                           brackets can be nested (requirement of database triggers). Replaced
**                           Apache commons logging with J2SE logging.
** 084 ECF 20150124          Lazily initialize ConnectionManager to avoid premature use of
**                           persistence services during server startup.
** 085 ECF 20150220          Rewrote unique buffer name algorithm to ensure names are unique among
**                           all buffers in the current context, but are consistently reused for a
**                           given base, legacy name. This allows the dynamic query cache to
**                           operate more efficiently.
** 086 CA  20150312          Fixed cleanup of tracked buffers; scope notifications are now 
**                           properly raised when the external procedure gets terminated.
** 087 CA  20150320          Backed out H086 - it caused a regression and the change is not part 
**                           of the root mem leak cause.
** 086 ECF 20150322          Partially rewrote unique buffer algorithm introduced with H086 to not
**                           use String.format(), which is very inefficient. Added deregistration
**                           of dynamic buffers to prevent leakage of tracked, dynamic buffers.
**                           Made other performance improvements.
** 087 ECF 20150327          Enhanced logging to determine source of a conflicting transaction
**                           when attempting to begin a new one.
** 088 ECF 20150906          Reimplemented implicit silent error mode for query substitution
**                           parameter processing. Simplified and optimized in-use DMO tracking.
** 089 ECF 20150929          Optimized unused DMO eviction.
** 090 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 091 OM  20160212          Added cleanupPending() method. Finalizable API has changed. Upgrade
**                           code to Java8.
** 092 ECF 20160321          Replaced ScopedDictionary with more performant ScopedList for
**                           openBuffers data structure. Fixed memory leak in scopeDeleted. Fixed
**                           removeAllReversibles to remove reversibles from parent scope as well
**                           as current scope.
** 093 ECF 20160429          Optimized removeAllReversibles. Added toString to RefCount.
** 094 ECF 20160605          Replace TM.registerCommit with TM.registerCommitAt, to avoid extra
**                           context-local dereferencing in TM.
** 095 ECF 20160823          Fixed primary key reclamation algorithm. Unfortunately, it was later
**                           deemed that the reclamation concept itself seems inherently unsafe,
**                           because of the potential for deadlocking due to out-of-order keys.
**                           However, I want to keep this code around in case we figure out a way
**                           to use this safely in the future.
** 096 ECF 20160829          Performance enhancement: track transaction status so other objects
**                           in this context can use this local resource to check transaction
**                           status instead of TransactionManager.
** 097 GES 20171207          Removed silent error mode processing.
** 098 CA  20180619          DBTxWrapper notifications are DB-specific, so 
**                           ConnectionManager.transactionEnd must be informed of the database 
**                           being processed. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.id.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

/**
 * Manages all record buffers within the current user context.  This includes
 * transaction processing, buffer and block scope management.  A separate
 * instance of this class is created within each user context.  It acts as a
 * broker between the context local <code>TransactionManager</code>, the
 * persistence service layer which manages database resources, and multiple
 * record buffers opened by client code.
 * <p>
 * A record buffer registers itself with the buffer manager when client code
 * opens a new buffer scope.  If a transaction is active, that buffer is
 * immediately registered with the transaction manager for commit processing.
 * This allows the buffer to synchronize its state with the persistence
 * service when a transaction or subtransaction commit occurs.  If a
 * transaction is not active when a new buffer scope is opened, the buffer
 * manager keeps track of the open buffer, and the block scope at which it
 * was opened.  When a full transaction subsequently is opened, all buffers
 * with open scopes are batch registered for commit processing.
 * <p>
 * Also, when a full transaction begins, the buffer manager takes stock of
 * all databases referred to by open buffer scopes, and begins a new,
 * database-level transaction for each one.  The database transaction's life
 * cycle is managed by an inner class, {@link BufferManager.DBTxWrapper}.
 * This inner class registers itself for master commit and finish processing
 * with the transaction manager, such that it can reflect an application
 * level commit/rollback back to the underlying database transaction.
 * <p>
 * Updates made to existing records are tracked for undo purposes by this
 * class.  A scoped dictionary of reversible updates by DMO type is managed
 * by the context local instance of this class, for all open buffers in the
 * current client context.
 */
public final class BufferManager
implements Scopeable,
           BatchListener,
           StopConditionVetoHandler,
           Resettable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(BufferManager.class.getName());
   
   /** Stores context local instances of this class */
   private static final ContextLocal<BufferManager> context =
   new ContextLocal<BufferManager>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_2; }
      protected BufferManager initialValue() { return (new BufferManager()); }
   };
   
   /** Flag indicating whether class was initialized, i.e. {@link #initialize} was called */
   private static boolean initialized = false;
   
   /** Transaction finalizable to manage transaction state locally (avoid context-local calls) */
   private final Finalizable xactFin = new Finalizable()
   {
      @Override
      public void finished()
      {
         transactionDepth = -1;
      }
      @Override
      public void deleted()
      {
         // no-op
      }
      @Override
      public void iterate()
      {
         // no-op
      }
      @Override
      public void retry()
      {
         // no-op
      }
   };
   
   /** Finalizable which clears batch mode at the appropriate times */
   private final Finalizable batchCleaner = new Finalizable()
   {
      @Override
      public void finished()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
      @Override
      public void deleted()
      {
         // no-op
      }
      @Override
      public void iterate()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
      @Override
      public void retry()
      {
         RecordBuffer.cleanupBatchMode(BufferManager.this);
      }
   };
   
   /** Change broker */
   private final ChangeBroker changeBroker = ChangeBroker.get();
   
   /** Field assigner for function/procedure parameter assign-backs */
   private final Scopeable fieldScopes = FieldAssigner.getScopeable();
   
   /** The buffers which will be registered on the next scope entry. */
   private final Map<String, RecordBuffer> pendingBuffers = new HashMap<>();
   
   /** Scoped dictionary in which all defined record buffers are stored */
   private final ScopedDictionary<String, RecordBuffer> allBuffers = new ScopedDictionary<>();
   
   /**
    * Scoped dictionary which key set contains all buffers into which a new record was loaded (or
    * the old record was reloaded) into corresponding scopes. Value set doesn't contain any useful
    * information and is used as a temporary data storage during copying from one scope level to
    * another.
    */
   private final ScopedDictionary<RecordBuffer, Boolean> loadedBuffers = new ScopedDictionary<>();
   
   /** Scoped dictionary in which open record buffers are stored */
   private final ScopedList<RecordBuffer> openBuffers = new ScopedList<>();
   
   /**
    * Scoped dictionary which contains currently active static temp tables (i.e. table scope was
    * opened and not yet closed). Keyed by legacy table names.
    */
   private final ScopedDictionary<String, StaticTempTable> openStaticTempTables =
      new ScopedDictionary<>();
   
   /**
    * Scoped dictionary which contains the maps of output copiers for a procedure/function to
    * their 0-based indexes in the set of all TABLE-HANDLE parameters for the procedure/function
    * (excluding other types of parameters).
    */
   private final ScopedDictionary<OutputTableHandleCopier, Integer> thOutputParameters =
      new ScopedDictionary<>();
   
   /** Map of records in use (in a persistence session) by database/buffer */
   private final Map<Persistable, RefCount> inUseRecords = new IdentityHashMap<>(512);
   
   /** <code>Persistence</code> to <code>DBTxWrapper</code> instance map */
   private final Map<Persistence, DBTxWrapper> persistToDbTx = new IdentityHashMap<>();
   
   /** All undoable actions scoped by block and keyed by buffer type */
   private final ScopedDictionary<BufferType, UndoData> undoData = new ScopedDictionary<>();
   
   /** Stack of batch mode data */
   private final ArrayDeque<BatchModeData> batchModeStack = new ArrayDeque<>();
   
   /** Map with buffer-related data surviving a persistent proc's external procedure.*/
   private final Map<Object, PersistentProcScope> persistProcScopes = new HashMap<>();
   
   /** Primary keys pending reclamation */
   private final Map<Persistence, Set<Serializable>> pendingReclaimedKeys = new HashMap<>();
   
   /** Block depth at which full application transaction began, or -1 if not in a transaction */
   private int transactionDepth = -1;
   
   /** Connection manager */
   private ConnectionManager connMgr = null;
   
   /** Number of nested queries being processed */
   private int queryDepth = 0;
   
   /**
    * Constructor; instances of this class are context-local and must be accessed using the static
    * {@link #get} method.
    */
   private BufferManager()
   {
      TransactionManager.registerBatchListener(this);
      TransactionManager.registerStopVetoHandler(this);
      TransactionManager.registerResettable(this);
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "BufferManager created in context " + Utils.describeContext());
      }

      connMgr = ConnectionManager.get();
   }
   
   /**
    * Retrieve the context-local instance of this class, instantiating it
    * first if necessary.
    *
    * @return  Context-local instance of this class.
    * @throws  IllegalStateException
    *            when {@link BufferManager} is not activated, i.e. {@link
    *            #initialize()} is not called.
    */
   public static BufferManager get()
   {
      if (!BufferManager.initialized)
      {
         throw new IllegalStateException(
            "An attempt to use uninitialized BufferManager. Is persistence activated?");
      }
      
      return context.get();
   }
   
   /**
    * Cleanup of pending resources. This is called when the external procedure constructor failed 
    * to build a new object because of an exception thrown in a member initialization (shared 
    * frame/table).
    */
   public static void cleanupPending()
   {
      BufferManager bm = BufferManager.get();
      bm.pendingBuffers.clear();
   }
   
   /**
    * Register with the {@link com.goldencode.p2j.util.TransactionManager
    * TransactionManager} a factory object which creates instances of this
    * class, so that they can be registered to receive notifications of
    * runtime scope start and finish events.
    * <p>
    * This method should be invoked once during the server bootstrap phase.
    */
   static void initialize()
   {
      TransactionManager.registerScopeableFactory(BufferManager::get);
      BufferManager.initialized = true;
   }
   
   /**
    * Query the scope level which currently is open.
    *
    * @return  Current scope level.
    */
   public int getCurrentScope()
   {
      return openBuffers.scopes();
   }
   
   /**
    * Check the specified exception and determine whether we want to allow it
    * to be honored by the {@link TransactionManager}, or whether we want to
    * veto the STOP at this scope.
    * <p>
    * This implementation is interested only in STOP conditions which have
    * been generated by database connection errors.  STOP processing at the
    * current scope is vetoed iff the exception is an instance of {@link
    * DatabaseConnectionException} AND either of the following criteria are
    * met:
    * <ul>
    *   <li>the failing database is one that should be permanently connected
    *       to this P2J server instance;  OR
    *   <li>there is any record buffer still in scope, which is associated
    *       with the failing database.
    * </ul>
    * <p>
    * Otherwise, normal STOP processing is allowed to continue.
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if STOP should be vetoed;
    *          <code>false</code> if the STOP should be allowed. 
    */
   public boolean vetoStop(StopConditionException exc)
   {
      // If the exception was not caused by a database connection problem, it
      // it is not in our domain to veto;  allow STOP processing to continue.
      if (!(exc instanceof DatabaseConnectionException))
      {
         return false;
      }
      
      DatabaseConnectionException dce = (DatabaseConnectionException) exc;
      Database database = dce.getDatabase();
      
      // on database connection errors, the exception should be re-thrown if:
      
      // 1. this is a permanent database;  there is no recovery from this
      // failure
      Iterator<Database> iter = DatabaseManager.permanentConnections();
      while (iter.hasNext())
      {
         Database db = iter.next();
         if (db.equals(database))
         {
            return true;
         }
      }
      
      // 2. this is a virtual database and this block or a parent block uses
      // a table from this database
      for (RecordBuffer buffer : allBuffers.values())
      {
         // a buffer is in use, the STOP must be vetoed
         if (buffer.isActive() && database.equals(buffer.getDatabase()))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Process a notification that a new block scope has been entered.  Notify
    * every buffer with an open scope that a new block scope has started.  If
    * the new block scope is within the context of a transaction, each buffer
    * with an open scope is registered for commit processing with the
    * transaction manager.  If the block scope represents the opening of a
    * new, full transaction, we also initialize database transactions for
    * each database referenced by a buffer with an open scope.
    * 
    * @throws  ErrorConditionException
    *          if there is an error flushing pending changes to the database
    *          as we open a new database transaction.
    */
   public void scopeStart()
   {
      allBuffers.addScope(null);
      loadedBuffers.addScope(null);
      openBuffers.pushScope();
      openStaticTempTables.addScope(null);
      thOutputParameters.addScope(null);
      undoData.addScope(null);
      fieldScopes.scopeStart();
      batchModeStack.push(new BatchModeData());
      
      if (!pendingBuffers.isEmpty())
      {
         Iterator<RecordBuffer> iter = pendingBuffers.values().iterator();
         while (iter.hasNext())
         {
            RecordBuffer buffer = iter.next();
            registerBuffer(buffer);
            iter.remove();
         }
      }
      
      // Ensure that batch assign mode is turned off properly, regardless of
      // whether business logic invokes RecordBuffer.endBatch().  This
      // prevents a nesting error raised during RecordBuffer.startBatch(), if
      // a batch is interrupted and then a new one begins.
      TransactionManager.registerFinalizable(batchCleaner, false);
      
      int blockDepth = openBuffers.scopes();
      
      if (LOG.isLoggable(Level.FINER))
      {
         LOG.log(Level.FINER,
                 Utils.describeContext() + ":  block scope " + blockDepth + " started");
      }
      
      boolean inTx = TransactionManager.isTransaction();
      boolean fullTx = inTx && TransactionManager.isFullTransaction();
      
      if (fullTx)
      {
         transactionDepth = blockDepth - 1;  // match TM transaction level
         
         // register for master transaction finish/iterate to clear transaction depth
         TransactionManager.registerTransactionFinish(xactFin, true);
      }
      
      for (RecordBuffer buffer : openBuffers)
      {
         if (buffer.getBlockDepth() == blockDepth)
         {
            // Already processed this one (buffers can be open in multiple, nested scopes).
            continue;
         }
         
         try
         {
            if (inTx)
            {
               if (fullTx && buffer.isActive())
               {
                  bufferActiveInTransaction(buffer, false);
               }
               
               // Register the open buffer for commits at this block scope.
               TransactionManager.registerCommitAt(blockDepth - 1, buffer);
            }
         }
         finally
         {
            // Notify each open buffer that a new block scope has opened.
            buffer.enterBlock(blockDepth, inTx);
         }
      }
      
      if (fullTx)
      {
         for (DBTxWrapper dbTxWrapper : persistToDbTx.values())
         {
            dbTxWrapper.begin();
         }
      }
   }
   
   /**
    * Process a notification that a scope is about to be exited.  Notify each
    * record buffer with an open scope that the current block scope is ending.
    */
   public void scopeFinished()
   {
      PersistentProcScope ppScope = null;
      
      if (!TransactionManager.isGlobalBlock())
      {
         Object referent = ProcedureManager.thisProcedure().get();
         BlockType bt = TransactionManager.getBlockType();
         // if we have a persistent procedure, save the data
         if (bt == BlockType.EXTERNAL_PROC && ProcedureManager._isPersistent(referent))
         {
            ppScope = new PersistentProcScope();
            persistProcScopes.put(referent, ppScope);
         }
      }
      
      int blockDepth = openBuffers.scopes();
      
      // here we will be out of sync with the TM on a transaction boundary, because our
      // transaction depth is cleared with a transaction finalizable, which is called before this
      // scope finished notification, so we rely on the TM.isTransaction() instead of our own
      boolean inTx = TransactionManager.isTransaction();
      
      // Notify all buffers that a block scope is closing.
      for (RecordBuffer buffer : openBuffers)
      {
         if (buffer.getBlockDepth() < blockDepth)
         {
            // Already processed this one (buffers can be open in multiple, nested scopes).
            continue;
         }
         
         buffer.exitBlock(blockDepth, inTx);
      }
      
      undoData.deleteScope();
      
      // Copy the set of buffers contained in the current scope of the
      // loadedBuffers (which is about to be removed) to the enclosing scope.
      // Only buffers which are open at the enclosing or a higher scope
      // should be copied.
      if (loadedBuffers.size() > 1)
      {
         Set<RecordBuffer> currentLoadedBuffers =
            loadedBuffers.getDictionaryAtScope(0, false).keySet();
         
         Collection<RecordBuffer> currentOpenBuffers = openBuffers.sublistAtScope(0);
         
         for (RecordBuffer buffer : currentLoadedBuffers)
         {
            if (buffer.getOpenScopeCount() > 1 || !currentOpenBuffers.contains(buffer))
            {
               // if this buffer is missing at the current scope of open
               // buffers or it presents there but has open scope count > 1,
               // that means that it presents at a higher level and it should
               // be copied to the enclosing scope
               loadedBuffers.setValueAtScope(buffer, 1, false);
            }
         }
      }
      if (ppScope != null)
      {
         // save the data for this external procedure, as these will need to be removed from the
         // global scope, when the procedure gets deleted.
         ppScope.loadedBuffers.putAll(loadedBuffers.getDictionaryAtScope(0, false));
         ppScope.openBuffers.addAll(openBuffers.sublistAtScope(0));
         ppScope.allBuffers.putAll(allBuffers.getDictionaryAtScope(0, false));
         ppScope.openStaticTempTables.putAll(openStaticTempTables.getDictionaryAtScope(0, false));
         
         // add all these to the global scope, too
         
         if (!ppScope.loadedBuffers.isEmpty())
         {
            loadedBuffers.getDictionaryAtScope(loadedBuffers.size() - 1, true)
                         .putAll(ppScope.loadedBuffers);
         }
         
         if (!ppScope.openBuffers.isEmpty())
         {
            openBuffers.addAll(true, ppScope.openBuffers);
         }
         
         if (!ppScope.allBuffers.isEmpty())
         {
            allBuffers.getDictionaryAtScope(allBuffers.size() - 1, true)
                      .putAll(ppScope.allBuffers);
         }
         
         if (!ppScope.openStaticTempTables.isEmpty())
         {
            openStaticTempTables.getDictionaryAtScope(openStaticTempTables.size() - 1, true)
                                .putAll(ppScope.openStaticTempTables);
         }
      }
      
      loadedBuffers.deleteScope();
      openBuffers.popScope();
      allBuffers.deleteScope();
      fieldScopes.scopeFinished();
      openStaticTempTables.deleteScope();
      thOutputParameters.deleteScope();
      batchModeStack.pop();
      
      if (LOG.isLoggable(Level.FINER))
      {
         LOG.log(Level.FINER,
                 Utils.describeContext() + ":  block scope " + blockDepth + " finished");
      }
      
      // if any DBs were used in this tx block and remained connected, disconnect now
      if (TransactionManager.isFullTransaction())
      {
         connMgr.transactionEnded();
      }
   }
   
   /**
    * Provides notification that the external procedure scope has been deleted.
    * <p>
    * If we are deleting a persistent procedure, clean all its data from the global scope for the
    * following dictionaries: {@link #loadedBuffers}, {@link #openBuffers}, {@link #allBuffers}, 
    * {@link #openStaticTempTables}.
    */
   @Override
   public void scopeDeleted()
   {
      if (loadedBuffers.size() == 0)
      {
         // nothing to do
         return;
      }
      
      Object referent = ProcedureManager.getProcessedProcedure();
      if (referent == null || ProcedureManager._isPersistent(referent))
      {
         // if no referent or procedure is still persistent, do not clean up yet
         return;
      }
      
      PersistentProcScope scope = persistProcScopes.remove(referent);
      
      // this needs to be executed only when the procedure gets deleted.
      if (scope != null)
      {
         // remove all these from the global scope
         
         if (!scope.loadedBuffers.isEmpty())
         {
            loadedBuffers.getDictionaryAtScope(loadedBuffers.size() - 1, true)
                         .keySet()
                         .removeAll(scope.loadedBuffers.keySet());
         }
         
         if (!scope.openBuffers.isEmpty())
         {
            openBuffers.removeAllFrom(true, scope.openBuffers);
         }
         
         if (!scope.allBuffers.isEmpty())
         {
            allBuffers.getDictionaryAtScope(allBuffers.size() - 1, true)
                      .keySet()
                      .removeAll(scope.allBuffers.keySet());
         }
         
         if (!scope.openStaticTempTables.isEmpty())
         {
            openStaticTempTables.getDictionaryAtScope(openStaticTempTables.size() - 1, true)
                                .keySet()
                                .removeAll(scope.openStaticTempTables.keySet());
         }
      }
   }
   
   /**
    * Provides a notification that a batch is starting or ending.  Enters or
    * exits batch assignment mode for all record buffers in this context,
    * depending on the value of <code>start</code>.
    *
    * @param   start
    *          <code>true</code> if this is the start of a batch, otherwise
    *          <code>false</code>.
    *
    * @see     RecordBuffer#startBatch
    * @see     RecordBuffer#endBatch
    */
   public void batchNotify(boolean start)
   {
      if (start)
      {
         RecordBuffer.startBatch();
      }
      else
      {
         RecordBuffer.endBatch();
      }
   }
   
   /**
    * Releases the buffers into which a new record was loaded (or the old
    * record was reloaded) into the current or deeper scopes, or only clears
    * the list of buffers pending to be released.
    *
    * @param clearOnly
    *        If <code>true</code> then clear the list of buffers pending to be
    *        released, otherwise perform release of these buffers and clear
    *        the list after that.
    */
   public void resetState(boolean clearOnly)
   {
      Map<RecordBuffer, Boolean> buffers = loadedBuffers.getDictionaryAtScope(0, false);
      if (!clearOnly)
      {
         for (RecordBuffer buffer : buffers.keySet())
         {
            if (!buffer.isPendingRollbackProcessed())
            {
               buffer.release();
            }
            else
            {
               buffer.resetPendingRollbackProcessed();
            }
         }
      }
      
      buffers.clear();
   }
   
   /**
    * Marks the table for postponed delete, if it is possible. Postponed delete can take place if
    * the table is used as [INPUT-]OUTPUT TABLE-HANDLE parameter.
    *
    * @param  tempTable
    *         Temporary table to be marked for postponed delete.
    *
    * @return <code>true</code> if the table was postponed for deletion. <code>false</code> if
    *         this can't be done.
    */
   public boolean postponeTableHandleDelete(TempTable tempTable)
   {
      int topLevel =  TransactionManager.findNearestTopLevel();
      if (topLevel != -1)
      {
         int idx = thOutputParameters.size() - topLevel - 1;
         Map<OutputTableHandleCopier, Integer> map =
            thOutputParameters.getDictionaryAtScope(idx, false);
         if (map != null)
         {
            Set<OutputTableHandleCopier> set = map.keySet();
            for (OutputTableHandleCopier copier : set)
            {
               handle outputHandle = copier.getCalled();
               if (outputHandle._isValid() && tempTable.equals(outputHandle.getResource()))
               {
                  copier.setTableToDelete(tempTable);
                  return true;
               }
            }
            
            // It's not a TABLE-HANDLE parameter, just a dynamic table. Clear all pending
            // TABLE-HANDLE parameter deletes which current handle value is not an unknown handle.
            for (OutputTableHandleCopier copier : set)
            {
               if (!copier.getCalled()._isValid())
               {
                  copier.setTableToDelete(null);
               }
            }
         }
      }
      
      return false;
   }
   
   /**
    * Check whether the current block is within an application-level transaction.
    * 
    * @return  {@code true} if we are in a transaction, else {@code false}.
    */
   public boolean isTransaction()
   {
      return transactionDepth >= 0;
   }
   
   /**
    * Check whether the block at the specified block depth is within an application-level
    * transaction.
    * 
    * @param   level
    *          Zero-based block depth (counting from the outermost scope).
    * 
    * @return  {@code true} if we are in a transaction, else {@code false}.
    */
   public boolean isTransactionAt(int level)
   {
      return transactionDepth >= 0 && transactionDepth <= level;
   }
   
   /**
    * Check whether the current block marks (and is within) the boundary of an application-level
    * full transaction.
    * 
    * @return  {@code true} if we are at (and within) a full transaction boundary, else {@code
    *          false}.
    */
   public boolean isFullTransaction()
   {
      return transactionDepth >= 0 && getCurrentScope() - 1 == transactionDepth;
   }
   
   /**
    * Get the zero-based block depth (counting from the outermost scope) at which the current,
    * full, application-level transaction (if any) began.
    * 
    * @return  Zero-based block depth of current, full transaction, or -1 if we are not currently
    *          within a transaction.
    */
   public int getFullTransactionBlock()
   {
      return transactionDepth;
   }
   
   /**
    * Register an undoable which is responsible for setting the buffer's 
    * proper current record after all reversibles are processed.
    * 
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   undoTarget
    *          The undoable responsible with setting the buffer back to its
    *          proper record.
    */
   void registerBufferUndoable(RecordBuffer buffer, Undoable undoTarget)
   {
      Map<BufferType, UndoData> dict = undoData.getDictionaryAtScope(0, true);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         ud.bufferUndoable = undoTarget.deepCopy();
      }
   }
   
   /**
    * Register a newly defined record buffer in the current scope.  Buffers
    * are deregistered implicitly when the scope in which they were
    * registered ends.
    *
    * @param   buffer
    *          Buffer to be registered.
    */
   void registerBuffer(RecordBuffer buffer)
   {
      allBuffers.addEntry(false, buffer.getDMOAlias(), buffer);
   }
   
   /**
    * Register as pending the newly defined record buffer. When the next scope
    * is entered, all the <code>pendingBuffers</code> will be registered with
    * that scope. Buffers are deregistered implicitly when the scope in which 
    * they were registered ends.
    *
    * @param   buffer
    *          Buffer pending to be registered.
    */
   void registerPending(RecordBuffer buffer)
   {
      pendingBuffers.put(buffer.getDMOAlias(), buffer);
   }
   
   /**
    * Deregister the given, dynamic record buffer by removing it from the buffer manager's
    * tracking.
    * 
    * @param   buffer
    *          Dynamic record buffer to be removed.
    */
   void deregisterDynamicBuffer(RecordBuffer buffer)
   {
      String key = buffer.getDMOAlias();
      int blockDepth = getCurrentScope();
      
      for (int i = blockDepth - 1; i >= 0; i--)
      {
         allBuffers.removeEntryAtScope(key, i);
      }
      
      for (int i = blockDepth - 1; i >= 0; i--)
      {
         openBuffers.removeFrom(i, buffer);
      }
   }
   
   /**
    * Begin batch assignment mode.  While in this mode, validation of record buffers is deferred.
    * We remember the scope at which this occurs in order to test for errors now and allow proper
    * cleanup later.
    *
    * @see     RecordBuffer#startBatch
    * @see     RecordBuffer#endBatch
    * @see     RecordBuffer#cleanupBatchMode
    * 
    * @throws  IllegalStateException
    *          if batch mode is already active in the current block or was started but never
    *          ended in a previous block.
    */
   void startBatchAssignMode()
   {
      int currentScope = getCurrentScope();
      BatchModeData bmd = batchModeStack.peek();
      
      if (bmd.batchDepth == currentScope)
      {
         throw new IllegalStateException("Cannot nest batch assign modes within the same scope");
      }
      
      if (bmd.batchDepth != -1)
      {
         // this should only happen with hand-written or modified code; converted code generally
         // will have balanced pairs of start and end batch mode calls
         throw new IllegalStateException(
            "Batch assign mode was opened but not closed in the same scope (" +
            bmd.batchDepth + "; current scope is " + currentScope + ")");
      }
      
      bmd.activate(currentScope);
   }
   
   /**
    * End batch assignment mode if we are currently in batch assignment mode
    * and we are in the same scope at which batch assignment mode was
    * started.  Otherwise do nothing.
    *
    * @return  If the above conditions are met for batch mode to end, the
    *          set of record buffers which were modified during batch mode is
    *          returned.  Otherwise, return <code>null</code>, indicating
    *          either we are not in batch mode, or we are not in the proper
    *          scope to end it, or the buffers already have been processed, or
    *          the program is ending and we are in cleanup mode.
    *
    * @see     RecordBuffer#startBatch
    * @see     RecordBuffer#endBatch
    * @see     RecordBuffer#cleanupBatchMode
    */
   Set<RecordBuffer> endBatchAssignMode()
   {
      BatchModeData bmd = batchModeStack.peek();
      
      if (bmd == null)
      {
         // exiting program
         return null;
      }
      
      if (bmd.batchDepth < 0 || getCurrentScope() != bmd.batchDepth)
      {
         return null;
      }
      
      bmd.batchDepth = -1;
      
      return bmd.dirtyBatchBuffers;
   }
   
   /**
    * Indicate whether we currently are in batch assignment mode.
    *
    * @return  <code>true</code> if in batch assignment mode, else
    *          <code>false</code>.
    *
    * @see     #startBatchAssignMode
    * @see     #endBatchAssignMode
    */
   boolean isBatchAssignMode()
   {
      return (batchModeStack.peek().batchDepth > -1);
   }
   
   /**
    * If we currently are in batch assignment mode, add the given record
    * buffer to the set of buffers requiring validation at the end of the
    * batch.  Otherwise, do nothing.
    *
    * @param   buffer
    *          Buffer to be added to the set of dirty batch mode buffers.
    *
    * @see     #startBatchAssignMode
    * @see     #endBatchAssignMode
    */
   void addDirtyBatchBuffer(RecordBuffer buffer)
   {
      BatchModeData bmd = batchModeStack.peek();
      if (bmd.batchDepth > -1)
      {
         bmd.dirtyBatchBuffers.add(buffer);
         
         // Add buffer to those requiring session flush;  necessary for DMO
         // validation if query has to go to the database.
         changeBroker.forcePendingFlush(buffer);
      }
   }
   
   /**
    * Increment the query depth.
    * <p>
    * This method MUST be invoked in balanced pairs that involve a subsequent call to 
    * {@link #decrementDepth}.  Calls may be nested, but they must be balanced, such that
    * at the end of all nested calls, <code>queryDepth</code> must be <code>0</code>.
    *
    * @return   <code>true</code> if the starting depth is 0.
    */
   boolean incrementDepth()
   {
      return (queryDepth++ == 0);
   }
   
   /**
    * Decrement the query depth.
    * <p>
    * This method MUST be invoked in balanced pairs that involve a subsequent call to 
    * {@link #incrementDepth}.  Calls may be nested, but they must be balanced, such that
    * at the end of all nested calls, <code>queryDepth</code> must be <code>0</code>.
    *
    * @return   <code>true</code> if the resulting depth is 0.
    */
   boolean decrementDepth()
   {
      return (--queryDepth == 0);
   }
   
   /**
    * Check whether the current query depth is non-zero, indicating that we are processing a
    * query within the context of another query. If so, raise a 7254 error condition.
    */
   void checkNestedQuery()
   {
      if (queryDepth == 0)
      {
         return;
      }
      
      handle thisProc = ProcedureManager.thisProcedure();
      Object referent = thisProc.get();
      String iename = ProcedureManager.resolveClosestMethod(referent, true, false);
      String msg =
         "A FIND, FOR EACH or OPEN QUERY statement has been encountered in a user-defined " +
         "function, method or constructor '" +
         iename +
         "' which was called from a WHERE clause.  This statement cannot be executed in " +
         "this context";
      ErrorManager.recordOrThrowError(7254, msg);
   }
   
   /**
    * Add potentially unused keys to be reclaimed after the end of the current transaction or
    * as buffer scopes are closed after a transaction. Some of these keys may not be collected,
    * if the locks on them are downgraded instead of released.
    * 
    * @param   persistence
    *          Persistence helper object for a particular database.
    * @param   keys
    *          Keys to be reclaimed.
    */
   void addPendingReclaimedKeys(Persistence persistence, List<Serializable> keys)
   {
      if (keys.isEmpty())
      {
         return;
      }
      
      Set<Serializable> pending = pendingReclaimedKeys.get(persistence);
      if (pending == null)
      {
         pending = new HashSet<>();
         pendingReclaimedKeys.put(persistence, pending);
      }
      
      pending.addAll(keys);
   }
   
   /**
    * Reclaim unused keys collected during the transaction which just ended or from the buffer
    * scope which just closed.
    * 
    * @param   persistence
    *          Persistence helper object for a particular database.
    * @param   released
    *          Keys actually released, due to the corresponding locks on them being released
    *          (as opposed to downgraded).
    */
   void reclaimPendingKeys(Persistence persistence, Set<Serializable> released)
   {
      if (released == null)
      {
         return;
      }
      
      Set<Serializable> pending = pendingReclaimedKeys.get(persistence);
      if (pending == null)
      {
         return;
      }
      
      IdentityManager identityManager = persistence.getIdentityManager();
      if (identityManager == null)
      {
         return;
      }
      
      // only reclaim keys actually released (not downgraded); downgraded keys are still
      // pinned by open buffers and cannot be reclaimed in their locked state
      released.retainAll(pending);
      
      int size = released.size();
      if (size == 0)
      {
         return;
      }
      
      try
      {
         identityManager.reclaimKeys(released.toArray(new Serializable[size]));
      }
      finally
      {
         if (pending.removeAll(released) && pending.isEmpty())
         {
            pendingReclaimedKeys.remove(persistence);
         }
      }
   }
   
   /**
    * Retrieve the set of reversible action objects for a given block scope.
    * <p>
    * If <code>create</code> is <code>true</code> and no set is found for the
    * given alias at the given scope, a set is created for the requested
    * scope.  Otherwise, the existing set or <code>null</code> will be
    * returned.  If newly created, a set is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no set exists in the requested scope,
    *          a new set is created and is stored for future lookups;  if
    *          <code>false</code>, no new set is created.
    *
    * @return  A set with the reversible action for the current scope, 
    *          or <code>null</code>.
    *
    * @see     Reversible
    * @see     RecordBuffer ReversibleCreate and ReversibleDelete inner classes 
    */
   Set<Reversible> getReversibles(RecordBuffer buffer, int depth, boolean create)
   {
      Set<Reversible> set = null;
      Map<BufferType, UndoData> dict = undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            set = ud.reversibles;
            if (set == null && create)
            {
               set = new LinkedHashSet<>();
               ud.reversibles = set;
               ud.reversiblesMap = new HashMap<>();
            }
         }
      }
      
      return set;
   }
   
   
   /**
    * Retrieve a map of reversible action objects for a given block scope. The
    * reversibles are kept in a list, by DMO ID.
    * <p>
    * If <code>create</code> is <code>true</code> and no map is found for the
    * given alias at the given scope, a map is created for the requested
    * scope.  Otherwise, the existing map or <code>null</code> will be
    * returned.  If newly created, a map is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no map exists in the requested scope,
    *          a new map is created and is stored for future lookups;  if
    *          <code>false</code>, no new map is created.
    *
    * @return  A map with the reversible action for the current scope, 
    *          or <code>null</code>.
    *
    * @see     Reversible
    * @see     RecordBuffer ReversibleCreate and ReversibleDelete inner classes  
    */
   Map<Serializable, List<Reversible>> getReversiblesMap(RecordBuffer buffer,
                                                         int depth,
                                                         boolean create)
   {
      Map<Serializable, List<Reversible>> map = null;
      Map<BufferType, UndoData> dict = undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            map = ud.reversiblesMap;
            if (map == null && create)
            {
               map = new HashMap<>();
               ud.reversiblesMap = map;
            }
         }
      }
      
      return map;
   }
   
   /**
    * Add a {@link Reversible} object to the reversibles set. This method also
    * ensures that a subsequent chain of reversibles will be kept, in a 
    * separate map, by DMO ID. As reversible actions for 2 DMO's can mix, this
    * map will allow access to the reversibles for a certain DMO, in the
    * proper order.
    *
    * @param   buffer
    *          Record buffer with which changes are associated.
    * @param   key
    *          The DMO ID used by this reversible.
    * @param   item
    *          Object which will perform the re-application of changes.
    */
   void addReversible(RecordBuffer buffer, Serializable key, Reversible item)
   {
      addReversible(buffer, key, item, true);
   }
   
   /**
    * Add a {@link Reversible} object to the reversibles set. When the 
    * <code>chain</code> parameter is true, this method  ensures that a 
    * subsequent chain of reversibles will be kept, in a separate map, by DMO 
    * ID. As reversible actions for 2 DMO's can mix, this map will allow 
    * access to the reversibles for a certain DMO, in proper order.
    *
    * @param   buffer
    *          Record buffer with which changes are associated.
    * @param   key
    *          The DMO ID used by this reversible.
    * @param   item
    *          Object which will perform the re-application of changes.
    * @param   chain
    *          When true, it will add the reversible to the per-DMO map,
    *          to enable access to all reversibles registered for a DMO.
    */
   void addReversible(RecordBuffer buffer, Serializable key, Reversible item, boolean chain)
   {
      if (chain)
      {
         // the key must be non-null only if we chain the reversibles in the
         // fast-access map
         assert (key != null);
      }
      
      Set<Reversible> set = getReversibles(buffer, 0, true);
      
      set.add(item);
      
      if (chain)
      {
         // we must add the item to the fast-access map too
         Map<BufferType, UndoData> dict =  undoData.getDictionaryAtScope(0, false);
         RecordBuffer master = buffer.getMasterBuffer();
         BufferType bufKey = master.getBufferType();
         
         if (dict != null)
         {
            UndoData ud = dict.get(bufKey);
            
            if (ud != null)
            {
               List<Reversible> list = ud.reversiblesMap.get(key);
               if (list == null)
               {
                  list = new ArrayList<>(3);
                  ud.reversiblesMap.put(key, list);
               }
               
               list.add(item);
            }
         }
      }
   }
   
   /**
    * Retrieve the map of DMO IDs to reversible update objects for a given
    * block scope.
    * <p>
    * If <code>create</code> is <code>true</code> and no map is found for the
    * given alias at the given scope, a map is created for the requested
    * scope.  Otherwise, the existing map or <code>null</code> will be
    * returned.  If newly created, a map is stored in the scoped dictionary
    * for future lookups.
    *
    * @param   buffer
    *          Record buffer for which a change is being tracked for undo
    *          purposes.  It is assumed the buffer has been initialized.
    * @param   depth
    *          Depth from the top of the block scope stack at which to
    *          retrieve the map.  0 indicates the current scope, 1 the
    *          closest enclosing scope, and so on.  -1 indicates the global
    *          scope.
    * @param   create
    *          If <code>true</code> and no map exists in the requested scope,
    *          a new map is created and is stored for future lookups;  if
    *          <code>false</code>, no new map is created.
    *
    * @return  A map which associates DMO IDs with the reversible update for
    *          the current scope, or <code>null</code>.
    *
    * @see     RecordBuffer ReversibleUpdate inner class
    */
   Map<Serializable, Reversible> getUndoables(RecordBuffer buffer, int depth, boolean create)
   {
      Map<Serializable, Reversible> map = null;
      Map<BufferType, UndoData> dict = undoData.getDictionaryAtScope(depth, create);
      
      RecordBuffer master = buffer.getMasterBuffer();
      BufferType key = master.getBufferType();
      
      if (dict != null)
      {
         UndoData ud = dict.get(key);
         if (ud == null && create)
         {
            ud = new UndoData();
            dict.put(key, ud);
         }
         
         if (ud != null)
         {
            map = ud.undoables;
            if (map == null && create)
            {
               map = new LinkedHashMap<>(4);
               ud.undoables = map;
            }
         }
      }
      
      return map;
   }
   
   /**
    * Remove all <code>Reversible</code> objects associated with the given buffer from their
    * respective scoped dictionaries.
    * 
    * @param   buffer
    *          Buffer associated with the <code>Reversible</code>s to be removed.
    */
   void removeAllReversibles(TemporaryBuffer buffer)
   {
      // delete entries from current and parent scopes
      BufferType bufType = buffer.getMasterBuffer().getBufferType();
      undoData.removeEntryThroughScope(bufType, 1);
   }
   
   /**
    * Called by {@link RecordBuffer} when a new buffer scope is opened to
    * register the open scope with the buffer manager, so that the buffer
    * can receive block scope start and stop notifications, and be registered
    * for commit processing when a full transaction is begun.
    *
    * @param   buffer
    *          Buffer for which a new scope is being opened.
    * @param   nestedOpen
    *          <code>true</code> if this buffer is already open.
    *
    * @return  Current scope depth.
    */
   int openScope(RecordBuffer buffer, boolean nestedOpen)
   {
      int blockDepth = openBuffers.scopes();
      boolean inTx = isTransaction();
      
      if (!nestedOpen)
      {
         // This buffer was not open when the block scope was started, so
         // its enterBlock method was not called;  call it now.
         buffer.enterBlock(blockDepth, inTx);
      }
      
      openBuffers.add(false, buffer);
      
      if (inTx)
      {
         if (buffer.isActive())
         {
            bufferActiveInTransaction(buffer, true);
         }
         
         // Register buffer for commit notifications if it has not already
         // been opened at a higher scope (in that case, the scopeStart()
         // invocation for the current scope will already have registered it).
         if (!nestedOpen)
         {
            TransactionManager.registerCommitAt(blockDepth - 1, buffer);
         }
      }
      
      return blockDepth;
   }
   
   /**
    * Register static temp table when its scope is opened.
    *
    * @param  defaultBuffer
    *         Default buffer for the target temp table.
    * @param  global
    *         Indicates if it is a global temp table (declared as SHARED GLOBAL).
    *
    * @return temp table definition.
    */
   StaticTempTable registerStaticTempTable(Temporary defaultBuffer, boolean global)
   {
      StaticTempTable table = new StaticTempTable(defaultBuffer);
      openStaticTempTables.addEntry(global, table.name().toStringMessage(), table);
      
      return table;
   }
   
   /**
    * Register the output table copier.
    *
    * @param copier
    *        Copier to register.
    */
   void registerOutputTableHandleCopier(OutputTableHandleCopier copier)
   {
      Map<OutputTableHandleCopier, Integer> map =
         thOutputParameters.getDictionaryAtScope(0, true);
      thOutputParameters.addEntry(false, copier, map.size());
      copier.setSiblingsMap(map);
   }
   
   /**
    * Unregister the output table copier <b>in the current scope</b>.
    *
    * @param copier
    *        Copier to unregister.
    */
   void unregisterOutputTableHandleCopier(OutputTableHandleCopier copier)
   {
      thOutputParameters.removeEntryAtScope(copier, 0);
   }
   
   /**
    * Get definition for the active static temp table with the given name.
    *
    * @param  name
    *         Legacy name of the target static temp table.
    *
    * @return definition of the static temp table with the given name or <code>null</code> if
    *         there is no active table with the given name.
    */
   StaticTempTable getStaticTempTable(String name)
   {
      return openStaticTempTables.lookup(name);
   }
   
   /**
    * Get the depth of the scope at which the static temp table was opened.
    *
    * @param  name
    *         Legacy name of the target static temp table.
    *
    * @return zero-based index of the target scope, starting from the outermost scope. If there is
    *         no table with the given name, <code>-1</code> is returned.
    */
   int getStaticTempTableDepth(String name)
   {
      int topDepth = openStaticTempTables.locate(name);
      
      return topDepth == -1 ? -1 : openStaticTempTables.size() - topDepth - 1;
   }
   
   /**
    * Add the target buffer to the set of all buffers at the specified block depth.
    *
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost block) at which the
    *          buffer was opened.
    * @param   buffer
    *          Buffer to add.
    */
   void addToAllBuffersArray(int blockDepth, RecordBuffer buffer)
   {
      allBuffers.addEntryAt(blockDepth, buffer.getDMOAlias(), buffer);
   }
   
   /**
    * Called by {@link TemporaryBuffer} when a new dynamic buffer scope is opened to register the
    * open scope with the buffer manager, so that the buffer can receive block scope start and
    * stop notifications, and be registered for commit processing when a full transaction is
    * begun.
    * 
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost block) at which the
    *          buffer scope is opened.
    * @param   buffer
    *          Buffer for which a new scope is being opened.
    */
   void openScopeAt(int blockDepth, RecordBuffer buffer)
   {
      if (blockDepth == 0)
      {
         openBuffers.add(true, buffer);
      }
      else
      {
         openBuffers.addAt(blockDepth, buffer);
      }
      
      if (isTransaction())
      {
         if (buffer.isActive())
         {
            bufferActiveInTransaction(buffer, true);
         }
         
         // commitables processing is delegated to RecordBuffer.processMiddleBlocks
      }
   }
   
   /**
    * Given a record buffer with an open scope operating within an application
    * level transaction, possibly create a new database transaction wrapper
    * object and optionally begin a database level transaction in the buffer's
    * backing database.  If a transaction wrapper already exists for the
    * buffer's database, do nothing.
    * 
    * @param   buffer
    *          Record buffer with an open buffer scope.
    * @param   beginTx
    *          Indicates whether to begin a new database transaction, in
    *          addition to creating a transaction wrapper.  This parameter is
    *          ignored if a transaction wrapper already exists.
    */
   void bufferActiveInTransaction(RecordBuffer buffer, boolean beginTx)
   {
      Persistence persistence = buffer.getPersistence();
      DBTxWrapper dbTx = persistToDbTx.get(persistence);
      
      // If dbTx is null, we have to create one which is associated with
      // this buffer's Persistence object.
      if (dbTx == null)
      {
         // If we have any pending temp-table updates not yet flushed to the
         // database, we need to make sure they are flushed now, so that they
         // cannot be rolled back if the transaction we are about to open is
         // rolled back. This can happen, for instance, if an update is made
         // to a newly created record which was flushed to the session soon
         // after creation.
         if (buffer.isTemporary())
         {
            try
            {
               persistence.maybeFlush();
            }
            catch (PersistenceException exc)
            {
               // Database problem.
               throw new RuntimeException(exc);
            }
         }
         
         dbTx = new DBTxWrapper(persistence, buffer.getDirtyContext());
         
         if (beginTx)
         {
            dbTx.begin();
         }
      }
   }
   
   /**
    * Find the record buffer associated with the given alias, which is the
    * variable name with which the buffer was defined.
    *
    * @param   alias
    *          Buffer variable name in business logic.
    *
    * @return  Associated record buffer or <code>null</code> if not found.
    */
   RecordBuffer lookupBuffer(String alias)
   {
      RecordBuffer buffer = pendingBuffers.get(alias);
      if (buffer == null)
      {
         buffer = allBuffers.lookup(alias);
      }
      
      return buffer;
   }
   
   /**
    * Get an iterator on all open buffers which contain non-<code>null</code>
    * current records for the given persistence object, within the current
    * client context.  Auto-commit record buffers associated with a pending,
    * reversible rollback are not included in the iteration.  Records that
    * represent copies of DMOs from other sessions, obtained from a {@link
    * DirtyShareContext}, are likewise excluded.
    *
    * @param   persistence
    *          Persistence object which manages sessions.  Only record
    *          buffers associated with this persistence object are returned.
    *
    * @return  Iterator on all active buffers.
    */
   Iterator<RecordBuffer> activeBuffers(Persistence persistence)
   {
      return activeBuffers(persistence, false);
   }
   
   /**
    * Get an iterator on all open buffers which contain non-<code>null</code>
    * current records for the given persistence object, within the current
    * client context.  Auto-commit record buffers associated with a pending,
    * reversible rollback are not included in the iteration.  Records that
    * represent copies of DMOs from other sessions, obtained from a {@link
    * DirtyShareContext}, are likewise excluded.
    *
    * @param   persistence
    *          Persistence object which manages sessions.  Only record
    *          buffers associated with this persistence object are returned.
    * @param   includeTransient
    *          <code>true</code> to include transient records in the result set.
    *
    * @return  Iterator on all active buffers.
    */
   Iterator<RecordBuffer> activeBuffers(Persistence persistence, boolean includeTransient)
   {
      Set<RecordBuffer> set = new HashSet<>();
      for (RecordBuffer next : openBuffers)
      {
         if (next.isActive() &&
             (includeTransient || !next.isTransient()) &&
             !next.isDirtyCopy() &&
             persistence.equals(next.getPersistence()))
         {
            Persistable dmo = next.getCurrentRecord();
            if (dmo != null)
            {
               set.add(next);
            }
         }
      }
      
      return set.iterator();
   }
   
   /**
    * Increment the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #decrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    * 
    * @param   dmo
    *          DMO instance whose reference count is to be incremented.
    *
    * @return  Use count for <code>dmo</code> after incrementing.
    */
   int incrementDMOUseCount(Persistable dmo)
   {
      RefCount count = inUseRecords.get(dmo);
      if (count == null)
      {
         count = new RefCount();
         inUseRecords.put(dmo, count);
      }
      else
      {
         count.value++;
      }
      
      return count.value;
   }
   
   /**
    * Return the context local use count for the specific instance of the given record. Used
    * in conjunction with {@link #decrementDMOUseCount} and {@link #incrementDMOUseCount} to
    * maintain an evidence of the buffers referencing the dmo. Records are evicted from the
    * current persistence session when their use counts reach zero.
    * <p>
    * The actual count of buffers referencing the dmo is used to emulate the P4GL behavior when
    * <code>rereadnolock</code> is specified. When this parameter is active, the dmo is refreshed
    * from database no matter how many buffers refer the dmo, otherwise only refresh the dmo if
    * it is referred by only one buffer.
    *
    * @param   dmo
    *          DMO instance whose reference count is to be returned.
    *
    * @return  Use count for <code>dmo</code>.
    */
   int getDMOUseCount(Persistable dmo)
   {
      RefCount count = inUseRecords.get(dmo);
      
      return (count == null ? 0 : count.value);
   }
   
   /**
    * Decrement the context local use count for the specific instance of the
    * given record.  Used in conjunction with {@link #incrementDMOUseCount} to
    * implement an aggressive DMO eviction policy.  Records are evicted from
    * the current persistence session when their use counts reach zero.
    * <p>
    * Use counts must be maintained because multiple buffers may hold the
    * same instance of a record.  A record must not be evicted while any
    * buffer still needs it.
    *
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  Use count for <code>dmo</code> after decrementing, or -1 if use
    *          count could not be found (possible in rare cases of interrupted
    *          thread on session exit).
    *
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    */
   int decrementDMOUseCount(Persistable dmo)
   throws PersistenceException
   {
      RefCount count = inUseRecords.get(dmo);
      if (count == null)
      {
         return -1;
      }
      
      int i = --count.value;
      
      if (i == 0)
      {
         inUseRecords.remove(dmo);
      }
      
      return i;
   }
   
   /**
    * Evict the given DMO from the current Hibernate session if it is not
    * referenced by any buffer or other resource.  This method does not
    * modify the DMO's use count, if any.  It is intended to quickly evict
    * DMOs which have been loaded into the session by side effect.
    * 
    * @param   persistence
    *          Persistence object which manages sessions.
    * @param   local
    *          Current user persistence context.
    * @param   dmo
    *          DMO instance whose reference count is to be decremented.
    * 
    * @return  <code>true</code> if the DMO was evicted;  <code>false</code>
    *          if it was referenced by other resources.
    * 
    * @throws  PersistenceException
    *          if an error occurs detaching <code>dmo</code> from the
    *          persistence session.
    */
   boolean evictDMOIfUnused(Persistence persistence, Persistence.Context local, Persistable dmo)
   throws PersistenceException
   {
      RefCount count = inUseRecords.get(dmo);
      if (count == null)
      {
         if (local == null)
         {
            local = persistence.getContext();
         }
         persistence.evict(local, dmo);
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Notify the buffer manager that a new record was loaded (or the old
    * record was reloaded) into the specified record buffer. 
    *
    * @param recordBuffer
    *        The record buffer in which a record was loaded (or the old record
    *        was reloaded). 
    */
   void notifyRecordWasLoaded(RecordBuffer recordBuffer)
   {
      if (recordBuffer.getCurrentRecord() == null || recordBuffer.isTemporary())
      {
         return;
      }
      
      loadedBuffers.addEntry(false, recordBuffer, false);
   }
   
   /**
    * Generate unique (across the session) name (DMO alias) for a dynamic buffer. The name is
    * generated by using standard P2J buffer name conversion rules and if necessary, appending
    * <code>__{unique number}</code> postfix. This guarantees that the name will not contain
    * invalid characters and will be unique.
    *
    * @param   legacyBufferName
    *          Legacy name of the buffer.
    *
    * @return  unique name (alias) for a dynamic buffer.
    */
   String generateUniqueBufferName(String legacyBufferName)
   {
      NameConverter nc = new NameConverter();
      String javaName = nc.convert(legacyBufferName, MatchPhraseConstants.TYPE_VARIABLE);
      
      // quick return if base java name is not in use
      if (lookupBuffer(javaName) == null)
      {
         return javaName;
      }
      
      // optimized name generator, in case we are called many times; unfortunately, lookupBuffer
      // is still expensive
      
      String next;
      int capacity = javaName.length() + 5;
      StringBuilder buf = new StringBuilder(capacity);
      buf.append(javaName);
      buf.append("__");
      int baseLen = buf.length();
      int i = 0;
      
      do
      {
         buf.setLength(baseLen);
         buf.append(Integer.toString(++i));
         next = buf.toString();
      } while (lookupBuffer(next) != null);
      
      return next;
   }
   
   /**
    * Compose a message suitable for logging.  Reports buffer scope and
    * context information in addition to a custom message.
    *
    * @param   persistence
    *          Persistence object which will provide context info.
    * @param   text
    *          Custom message text.
    *
    * @return  Composed message text.
    */
   private String message(Persistence persistence, String text)
   {
      return persistence.message("scope " + openBuffers.scopes() + ": " + text);
   }
   
   /**
    * Storage mechanism for maps of reversible/undoable actions, keyed by
    * primary key.  One instance of this class holds all such objects for a
    * single buffer in a single scope.
    */
   private static class UndoData
   {
      /** Reversible record actions linked in natural order */
      Set<Reversible> reversibles;
      
      /** List of reversible record actions mapped by ID, in natural order */
      Map<Serializable, List<Reversible>> reversiblesMap;
      
      /** Undoable record actions (or rollbacks) keyed by primary key */
      Map<Serializable, Reversible> undoables;
      
      /** 
       * Undoable which will set the no-undo buffer to its proper record 
       * after the reversibles are processed.
       */
      Undoable bufferUndoable;
      
      /**
       * Indicate whether this data structure holds any data.
       * 
       * @return  <code>true</code> if this object holds any reversible or
       *          undoable actions, else <code>false</code>.
       */
      boolean isEmpty()
      {
         return (reversibles == null && undoables == null);
      }
   }
   
   /**
    * A helper class which is created whenever a new database-level
    * transaction is needed.  Upon creation, this object opens a new
    * database-level transaction, and registers itself with the
    * <code>TransactionManager</code> to receive commit/rollback and
    * finished/iterate/retry notifications.
    * <p>
    * Note that our architecture differs somewhat from Progress here in that
    * if we have an error at commit time, it is too late to undo in the
    * current block scope.  All attempts are made before commit is executed
    * to vet and prevent any errors, but it is still possible to have an
    * error at this level if say, two transactions in different contexts
    * attempt to commit conflicting data.  In this case, the error would be
    * handled in the next block scope up from the current one.
    * <p>
    * Auto-commit buffers require special handling during a database-level
    * rollback.  Changes to such buffers made during a database transaction
    * must be preserved, but a database-level transaction rollback will not
    * discriminate these changes from those made to other buffers.  Therefor,
    * we must reverse the rollback (re-roll forward?) for any auto-commit
    * buffers whose changes are rolled back with the transaction.  This is
    * done by beginning a new transaction just for these changes, immediately
    * after a database-level rollback.  Within that transaction, changes made
    * to an auto-commit buffer are rolled forward once again and the
    * transaction is then committed.
    */
   private class DBTxWrapper
   implements Finalizable,
              Commitable
   {
      /** Persistence service object */
      private final Persistence persistence;
      
      /** Dirty share context */
      private final DirtyShareContext dirtyContext;
      
      /** Do we need to rollback database transaction in finished()? */
      private boolean autoRollback = false;
      
      /**
       * Constructor which accepts a {@link Persistence} instance and the set
       * of currently open buffers for that database, registers with the
       * <code>TransactionManager</code> as a <code>Finalizable</code> and a
       * <code>Commitable</code>, and begins a new database-level
       * transaction.
       *
       * @param   persistence
       *          Persistence object which will be used for transaction
       *          support.
       * @param   dirtyContext
       *          Dirty share context which must be cleaned up at the end of
       *          this transaction.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      DBTxWrapper(Persistence persistence, DirtyShareContext dirtyContext)
      {
         if (persistence == null)
         {
            throw new NullPointerException();
         }
         
         persistToDbTx.put(persistence, this);
         
         this.persistence = persistence;
         this.dirtyContext = dirtyContext;
         
         // The SessionFlusher must be the first (regular) Commitable
         // registered, because it must do its work before any other
         // Commitable is called for the same scope.
         SessionFlusher flusher = new SessionFlusher(persistence);
         int blockDepth = openBuffers.scopes();
         TransactionManager.registerCommitAt(blockDepth - 1, flusher);
         
         // Register for master transaction commit/rollback.
         TransactionManager.registerTransactionCommit(this, true);
         
         // Register for master transaction finish/iterate.
         TransactionManager.registerTransactionFinish(this, true);
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE,
                    message(persistence, "registered for master transaction commit/finished"));
            
            persistence.enableTransactionTrace(true);
         }
      }
      
      /**
       * Called by the <code>TransactionManager</code> after commit or
       * rollback to clean up.  This closes the current Hibernate session
       * and removes this database name from the buffer manager's set of
       * databases currently involved in transactions.
       * <p>
       * In the event the transaction manager did not explicitly commit or
       * rollback (e.g., <code>blockSetup()</code> was never called, perhaps
       * due to conditional processing which evaluated to false), we
       * automatically rollback at this point.
       */
      public void finished()
      {
         try
         {
            if (autoRollback)
            {
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.log(Level.FINE, message(persistence, "triggering auto-rollback"));
               }
               
               rollback(true);
            }
         }
         finally
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, message(persistence, "master transaction finished"));
            }
            
            persistToDbTx.remove(persistence);
            end();
         }
      }
      
      /**
       * This implementation does nothing.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
      
      /**
       * Called by the <code>TransactionManager</code> from the top of the
       * block (just after the loop test) on the second and subsequent
       * iterations of an iterating loop.
       * <p>
       * We create a new, database-level transaction here, after the loop
       * test has been passed.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      public void iterate()
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, message(persistence, "master transaction iterate"));
         }
         
         end();
         begin();
      }
      
      /**
       * Processes the retry notification provided by the
       * <code>TransactionManager</code>.  This occurs just after we've
       * rolled back and we are about to jump to the top of the loop.  We
       * will not get a call to {@link #iterate}, because this is not a
       * normal iteration;  instead, we are retrying the same pass through
       * the block and need to set up a transaction as if we were coming in
       * fresh.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      public void retry()
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, message(persistence, "master transaction retry"));
         }
         
         end();
         begin();
      }
      
      /**
       * Called by the <code>TransactionManager</code> when the full
       * transaction must be committed.  This is our hook to commit the
       * database-level transaction.
       *
       * @param   transaction
       *          <code>true</code> if this is a full transaction.  Not used,
       *          since we only register for full transaction commits.
       */
      public void commit(boolean transaction)
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, message(persistence, "master transaction commit"));
         }
         
         try
         {
            autoRollback = false;
            persistence.commit();
         }
         catch (PersistenceException exc)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, message(persistence, "commit error"), exc);
            }
            
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
         finally
         {
            // Reclaim any primary keys which had been waiting to be recycled,
            // pending a commit.
            Database db = persistence.getDatabase();
            if (DatabaseManager.TEMP_TABLE_DB.equals(db))
            {
               Commitable ctx = TemporaryBuffer.getContextCommitable();
               ctx.commit(transaction);
            }
         }
      }
      
      /**
       * No-op
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       * 
       * @param   transaction
       *          Not used.
       */
      public void validate(boolean transaction)
      {
      }
      
      /**
       * Called by the <code>TransactionManager</code> when the full
       * transaction must be rolled back.  This is our hook to roll back the
       * database-level transaction.
       *
       * @param   transaction
       *          <code>true</code> if this is a full transaction.  Not used,
       *          since we only register for full transaction commits.
       */
      public void rollback(boolean transaction)
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, message(persistence, "master transaction rollback"));
         }
         
         Database database = persistence.getDatabase();
         boolean isTempDB = DatabaseManager.TEMP_TABLE_DB.equals(database);
         
         try
         {
            autoRollback = false;
            persistence.rollback();
         }
         catch (PersistenceException exc)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, message(persistence, "rollback error"), exc);
            }
            
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
         
         if (isTempDB)
         {
            // do this after the session rolledback to load the reversibles 
            // in the same state as before the transaction was begun
            prepareReversibleRollbacks();
            
            // Clear any primary keys which had been waiting to be recycled,
            // pending a commit.
            Commitable ctx = TemporaryBuffer.getContextCommitable();
            ctx.rollback(transaction);
            
            // If any auto-commit buffer changes were rolled back at the
            // database, we must reverse these now.
            executeReversibleRollbacks();
         }
      }
      
      /**
       * Begin a new database-level transaction.
       *
       * @throws  IllegalStateException
       *          if a transaction is already active in the current session;
       *          this represents a programming error.
       */
      private void begin()
      {
         try
         {
            if (!persistence.beginTransaction())
            {
               if (LOG.isLoggable(Level.FINE))
               {
                  Throwable beginTxTrace = persistence.getBeginTxTrace();
                  if (beginTxTrace != null)
                  {
                     String msg = "Current transaction previously started at location...";
                     LOG.log(Level.FINE, persistence.message(msg), beginTxTrace);
                  }
               }
               
               throw new IllegalStateException(
                  message(persistence, "database transaction already active"));
            }
            
            autoRollback = true;
         }
         catch (PersistenceException exc)
         {
            // No good way to recover from this.  Let it abend.
            throw new RuntimeException(exc);
         }
      }
      
      /**
       * Perform transaction post-processing to release or downgrade record
       * locks which may have been acquired during the transaction, to clean
       * up dirty index update sharing information contributed to the dirty
       * share manager by the current context for the transaction which is now
       * ending, and to clean up any other resources associated with the
       * transaction.
       */
      private void end()
      {
         // Clean up dirty shares for this transaction.
         if (dirtyContext != null)
         {
            dirtyContext.cleanup();
         }
         
         // remove all pending reversible rollbacks associated with this database
         Database database = persistence.getDatabase();
         Iterator<Map.Entry<BufferType, UndoData>> iter = undoData.entrySet().iterator();
         while (iter.hasNext())
         {
            Map.Entry<BufferType, UndoData> next = iter.next();
            BufferType key = next.getKey();
            
            if (database.equals(key.getDatabase()))
            {
               UndoData ud = next.getValue();
               ud.reversibles = null;
               ud.reversiblesMap = null;
               if (ud.isEmpty())
               {
                  iter.remove();
               }
            }
         }
         
         // release or downgrade record locks which may have been acquired during the transaction
         RecordLockContext lockCtx = persistence.getRecordLockContext();
         Set<Serializable> released = lockCtx.transactionEnded();
         
         // reclaim any primary keys for records deleted and committed or created and rolled back
         // during the transaction, provided those keys have been fully released (not just
         // downgraded) by the end of the transaction; note that this may "leak" reclaiming some
         // primary keys downgraded and still held as share locks by open buffers, for buffers
         // whose scope extends beyond the end of the transaction
         reclaimPendingKeys(persistence, released);
         
         // Allow the connection manager to perform its end of transaction cleanup.  This must
         // occur after all transaction processing, because it may result in tearing down the
         // database's session factory.
         connMgr.transactionEnded(database);
      }
      
      /**
       * Prepare temp table reversible rollbacks for execution.
       */
      private void prepareReversibleRollbacks()
      {
         Iterator<UndoData> iter = new ReversibleIterator();
         while (iter.hasNext())
         {
            UndoData data = iter.next();
            
            if (data.reversibles == null)
            {
               continue;
            }
            
            for (Reversible rev : data.reversibles)
            {
               rev.prepareRollback(null);
            }
         }
      }
      
      /**
       * Reverse temp table changes which were rolled back at the database, in
       * order to support the Progress temp-table NO-UNDO semantic.
       */
      private void executeReversibleRollbacks()
      {
         ReversibleIterator iter = new ReversibleIterator();
         while (iter.hasNext())
         {
            UndoData data = iter.next();
            
            try
            {
               Set<Reversible> set = data.reversibles;
               
               if (set != null && persistence.beginTransaction())
               {
                  Iterator<Reversible> revIter = set.iterator();
                  while (revIter.hasNext())
                  {
                     Reversible rev = revIter.next();
                     rev.rollback(true, true);
                     revIter.remove();
                     
                     // after its done processing, remove it from the fast 
                     // access map and change the DMO for all its right-side
                     // siblings. this is needed to ensure that the subsequent
                     // reversibles will use the same DMO as this one.
                     try
                     {
                        List<Reversible> list = data.reversiblesMap.get(rev.getId());
                        list.remove(rev);
                        
                        for (Reversible sibling : list)
                        {
                           Persistable dmo = rev.getDMO();
                           sibling.setDMO(dmo);
                        }
                     }
                     catch (UnsupportedOperationException exc) 
                     {
                        // ignore, is thrown only by bulk delete reversibles
                        // when getId() is called
                     }
                  }
                  
                  persistence.commit();
               }
            }
            catch (PersistenceException exc)
            {
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.log(Level.SEVERE, message(persistence, "reversible rollback error"), exc);
               }
               
               // No good way to recover from this.  Let it abend.
               throw new RuntimeException(exc);
            }
            finally
            {
               if (data.bufferUndoable != null)
               {
                  Undoable u = data.bufferUndoable;
                  // this ensures the buffer remains set on its correct record
                  u.assign(u);
               }
            }
         }
      }
      
      /**
       * Class used to iterate through all the reversible data, by buffer.
       */
      private class ReversibleIterator
      implements Iterator<UndoData>
      {
         private Iterator<Map.Entry<BufferType, UndoData>> outerIter =
            undoData.entrySet(0).iterator();
         
         private UndoData nextUndoData = null;
         
         public boolean hasNext()
         {
            fetchNextData();
            
            return (nextUndoData != null);
         }
         
         public UndoData next()
         {
            if (nextUndoData == null)
            {
               throw new NoSuchElementException();
            }
            
            UndoData data = nextUndoData;
            nextUndoData = null;
            
            return data;
         }
         
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
         
         private void fetchNextData()
         {
            while (nextUndoData == null && outerIter.hasNext())
            {
               Map.Entry<BufferType, UndoData> next = outerIter.next();
               
               BufferType key = next.getKey();
               
               if (key.isUndoable())
               {
                  continue;
               }
               
               nextUndoData = next.getValue();
            }
         }
      }
   }
   
   /**
    * Commitable implementation which tries to flush Hibernate session before
    * any rollback.  Must be the first <code>Commitable</code> registered at a
    * transaction scope (i.e., before any record buffers).
    */
   private static class SessionFlusher
   implements Commitable
   {
      /** Persistence service object */
      private final Persistence persistence;
      
      /**
       * Constructor.
       * 
       * @param   persistence
       *          Persistence service object.
       */
      SessionFlusher(Persistence persistence)
      {
         this.persistence = persistence;
      }
      
      /**
       * Called when <code>TransactionManager</code> processes a commit event.
       * <p>
       * This implementation does nothing.
       *
       * @param    transaction
       *           <code>true</code> if this is a full transaction and
       *           <code>false</code> if this is only a sub-transaction (a
       *           nested scope with transaction support).
       */
      public void commit(boolean transaction)
      {
      }
      
      /**
       * Called when <code>TransactionManager</code> processes a rollback
       * event.
       * <p>
       * Attempt to flush the Hibernate session if necessary.
       *
       * @param    transaction
       *           <code>true</code> if this is a full transaction and
       *           <code>false</code> if this is only a sub-transaction (a
       *           nested scope with transaction support).
       */
      public void rollback(boolean transaction)
      {
         try
         {
            persistence.maybeFlush();
         }
         catch (PersistenceException exc)
         {
            throw new RuntimeException(exc);
         }
      }
      
      /**
       * No-op
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       */
      public void validate(boolean transaction)
      {
      }
   }
   
   /**
    * Container for batch assign mode information.
    */
   private static class BatchModeData
   {
      /** Scope depth at which batch assign mode began; -1 if not in batch assign mode */
      private int batchDepth = -1;
      
      /** Set of dirty batch assign mode buffers requiring validation */
      private Set<RecordBuffer> dirtyBatchBuffers = null;
      
      /**
       * Default constructor.
       */
      BatchModeData()
      {
      }
      
      /**
       * Activate batch assign mode.
       * 
       * @param   batchDepth
       *          Scope depth at which batch assign mode was activated.
       */
      void activate(int batchDepth)
      {
         this.batchDepth = batchDepth;
         dirtyBatchBuffers = new LinkedHashSet<>();
      }
   }
   
   /**
    * Reference count object with a mutable counter value, for tracking buffer references to DMOs
    * currently in use within a Hibernate session.
    */
   private static class RefCount
   {
      /** Reference counter */
      int value = 1;
      
      /**
       * Return a string representation of the reference count.
       *
       * @return  String representation of the reference count.
       */
      @Override
      public String toString()
      {
         return String.valueOf(value);
      }
   }
   
   /**
    * Buffer-related data for a certain persistent procedure, surviving after its external 
    * procedure is finished.
    */
   private class PersistentProcScope
   {
      /** All defined record buffers for a persistent procedure. */
      private final Map<String, RecordBuffer> allBuffers = new HashMap<>();
      
      /**
       * All buffers into which a new record was loaded (or the old record was reloaded), for a
       * persistent procedure.
       */
      private final Map<RecordBuffer, Boolean> loadedBuffers = new HashMap<>();
      
      /** Open record buffers for a persistent procedure. */
      private final List<RecordBuffer> openBuffers = new LinkedList<>();
      
      /**
       * Active static temp tables (i.e. table scope was opened and not yet closed) for a 
       * persistent procedure.
       */
      private final Map<String, StaticTempTable> openStaticTempTables = new HashMap<>();
   }
}