/*
** Module   : BufferType.java
** Abstract : Object which describes a record buffer type
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20090812   @43605 Created initial version. Object which describes a RecordBuffer type.
** 002 ECF 20160605          Cache hash code.
** 002 ECF 20160827          Do not intern table name; instances of this class are not stored in
**                           long-term memory, so duplicates are OK to avoid the performance hit.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

/**
 * Object which describes a {@link RecordBuffer} type.  It does not uniquely
 * identify an individual buffer, only the main characteristics of a buffer;
 * namely, its physical database, backing table name, and whether changes made
 * to records it holds can be undone.
 */
final class BufferType
{
   /** Buffer's database */
   private final Database database;
   
   /** Buffer's backing table name (uppercase) */
   private final String table;
   
   /** Is buffer undoable? */
   private final boolean undoable;
   
   /** Cached hash code */
   private final int hashCode;
   
   /**
    * Constructor.
    * 
    * @param   database
    *          Database.
    * @param   table
    *          Backing table name.
    * @param   undoable
    *          Undoable flag.
    */
   BufferType(Database database, String table, boolean undoable)
   {
      this.database = database;
      this.table = table.toUpperCase();
      this.undoable = undoable;
      
      int result = 17;
      result = 37 * result + database.hashCode();
      result = 37 * result + table.hashCode();
      result = 37 * result + (undoable ? Boolean.TRUE.hashCode() : Boolean.FALSE.hashCode());
      this.hashCode = result;
   }
   
   /**
    * Calculate a hash code for this object.
    * 
    * @return  Hash code.
    */
   @Override
   public int hashCode()
   {
      return hashCode;
   }

   /**
    * Test this object for equality (equivalence) with the given object.
    * 
    * @param   o
    *          Object to test.
    * 
    * @return  <code>true</code> if this object is the same instance as the
    *          given object, or it is considered equivalent to the given
    *          object.
    */
   @Override
   public boolean equals(Object o)
   {
      if (this == o)
      {
         // Same instance.
         return true;
      }
      
      if (!(o instanceof BufferType))
      {
         // Impossible to be equivalent.
         return false;
      }
      
      BufferType that = (BufferType) o;
      
      return (this.database.equals(that.database) &&
              this.table.equals(that.table) &&
              this.undoable == that.undoable);
   }
   
   /**
    * Get the database.
    * 
    * @return  Database.
    */
   Database getDatabase()
   {
      return database;
   }
   
   /**
    * Get the undoable flag.
    * 
    * @return  Undoable flag.
    */
   boolean isUndoable()
   {
      return undoable;
   }
}