/*
** Module   : DatabaseStatistics.java
** Abstract : Collects query and second level cache database statistics
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 ECF 20150726 Created initial version.
** 002 ECF 20150827 Added execution row count to global query statistics.
** 003 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 004 ECF 20171127 Prepended "./" to database URL.
** 005 SBI 20180518 Renamed file-system/pkgroot into legacy-system/pkgroot.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.stat.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.util.*;

/**
 * A singleton instance of this class collects database statistics throughout a P2J server's run,
 * then collates this information into an H2 database as the server is shutting down. This adds
 * performance and memory overhead, so it is not recommended for production use, only for testing
 * and development.
 * <p>
 * The database is created in the server's current working directory, in a subdirectory named
 * <code>perfdb</code>. This subdirectory will be created if it does not exist (and the server
 * process has the file system permissions to do so).
 * <p>
 * Statistics collection is disabled by default and is enabled via the directory, by setting path
 * <code>persistence/collect-statistics</code> to <code>true</code>. This will enable collection
 * for ALL auto-connect databases. Currently, there is no more granular setting than this.
 * Statistics currently are not collected for remotely connected databases (i.e., those connected
 * via the converted CONNECT Progress statement).
 * <p>
 * The statistics collected are for every HQL statement executed, and the second-level cache
 * statistics which Hibernate provides. For queries, we use both Hibernate's query statistics
 * (recorded by each <code>SessionFactory</code>), and some simple use count statistics for each
 * query launched from a particular location in a converted application. This second bit allows
 * us to identify all the locations from which queries are executed in an application (the same
 * HQL can be executed from many places), whereas Hibernate's HQL statistics are global across an
 * application.
 * <p>
 * Note: Hibernate's timing statistics for query execution are based on millisecond precision
 * (or as close to that as the JVM has implemented), so they can be somewhat misleading. That
 * being said, the purpose of this class is to help identify query "hot spots" and evaluate
 * second level cache performance, and this level of precision will have to do.
 * <p>
 * This <code>perfdb</code> database can be accessed using the H2 console, which is available by
 * running:
 * <pre>
 *    java -jar &lt;path_to_h2_jar_file&gt;/h2.jar
 * </pre>
 * This will open the default browser to a database connect page. The simplest way to connect
 * to the <code>perfdb</code> database from this page is to use the embedded server option and
 * provide a URL as follows:
 * <pre>
 *    jdbc:h2:perfdb/perfdb;IFEXISTS=TRUE;ACCESS_MODE_DATA=r
 * </pre>
 * This relative URL assumes you ran the <code>java</code> command above from the parent
 * directory of the <code>perfdb</code> subdirectory. If not, you also can use an absolute URL:
 * <pre>
 *    jdbc:h2:/&lt;path_to_perfdb&gt;/perfdb;IFEXISTS=TRUE;ACCESS_MODE_DATA=r
 * </pre>
 * Provide the database user name and password configured in the application directory.
 * <p>
 * Refer to the H2 documentation (http://www.h2database.com) for additional information on how
 * to use H2.
 * <p>
 * To delete previous performance statistics, simple delete the <code>perfdb</code> subdirectory
 * (while the database is not open).
 * <p>
 * TODO: document <code>perfdb</code> schema.
 */
public final class DatabaseStatistics
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(DatabaseStatistics.class.getName());
   
   /** Relative directory path for statistics collection flag */
   private static final String COLLECT_STATS = "persistence/collect-statistics/enabled";
   
   /** Relative directory path for performance statistics database user name */
   private static final String DB_USER = "persistence/collect-statistics/db-user";
   
   /** Relative directory path for performance statistics database user password */
   private static final String DB_PASS = "persistence/collect-statistics/db-pass";
   
   /** Relative directory path for application package root */
   private static final String PKGROOT = "legacy-system/pkgroot";
   
   /** Singleton instance of this class */
   private static final DatabaseStatistics singleton;
   
   /** Package root prefix stripped from application class names collected with statistics */
   private static final String appPackagePrefix;
   
   /** Length of package root prefix, including trailing dot */
   private static final int appPackagePrefixLength;
   
   static
   {
      // determine whether statistics collection is enabled and set up state accordingly
      
      String prefix = null;
      
      DirectoryService dir = DirectoryService.getInstance();
      if (!dir.bind())
      {
         throw new RuntimeException("Directory bind failed");
      }
      
      if (Utils.getDirectoryNodeBoolean(dir, COLLECT_STATS, false, false))
      {
         String user = Utils.getDirectoryNodeString(dir, DB_USER, null, false);
         String pass = Utils.getDirectoryNodeString(dir, DB_PASS, null, false);
         prefix = Utils.getDirectoryNodeString(dir, PKGROOT, null, true);
         
         if (user != null)
         {
            DatabaseHelper.dbUser = user;
         }
         if (pass != null)
         {
            DatabaseHelper.dbPass = pass;
         }
      }
      
      dir.unbind();
      
      if (prefix != null)
      {
         if (!prefix.endsWith("."))
         {
            prefix += ".";
         }
         appPackagePrefixLength = prefix.length();
         singleton = new DatabaseStatistics();
         
         if (log.isLoggable(Level.WARNING))
         {
            String msg =
               "Database statistics collection is enabled for all databases; NOT " +
               "recommended for production use, due to performance and memory overhead!";
            log.warning(msg);
         }
      }
      else
      {
         appPackagePrefixLength = -1;
         singleton = null;
      }
      
      appPackagePrefix = prefix;
   }
   
   /** Map of databases to statistics collectors */
   private final Map<Database, Collector> collectors = new LinkedHashMap<>();
   
   /** Transient map of query instances to information about where they were constructed */
   private final Map<P2JQuery, Location> queryCtorLocs = new WeakHashMap<>();
   
   /**
    * Get the singleton instance of this class. This method is not synchronized because the
    * code which sets the singleton executes very early during server startup, and we should not
    * have any contention for "regular" calls during normal server operation.
    * 
    * @return  Singleton instance or <code>null</code> if statistics collection is disabled.
    */
   public static DatabaseStatistics get()
   {
      return singleton;
   }
   
   /**
    * Indicate whether statistics collection is enabled for this server instance.
    * 
    * @return  <code>true</code> if enabled, else <code>false</code>.
    */
   public static boolean isEnabled()
   {
      return (singleton != null);
   }
   
   /**
    * Private default constructor.
    */
   private DatabaseStatistics()
   {
   }
   
   /**
    * Collect statistics for all registered databases. Should only be called during server
    * termination processing.
    * <p>
    * Note: this method is invoked during server termination. It appears logging and writing to
    * stderr are not working properly at this point.
    */
   public void collectStatistics()
   {
      synchronized (collectors)
      {
         if (!collectors.isEmpty())
         {
            DatabaseHelper dbh = new DatabaseHelper();
            
            Iterator<Collector> iter = collectors.values().iterator();
            while (iter.hasNext())
            {
               Collector next = iter.next();
               next.collateStatistics(dbh);
               
               // prevent this from happening more than once
               iter.remove();
            }
            
            dbh.finalizeChanges();
         }
      }
   }
   
   /**
    * Register the given database for statistics collection.
    * 
    * @param   database
    *          Database.
    * @param   sessionFactory
    *          Session factory which gathers Hibernate's statistics.
    */
   void registerDatabase(Database database, SessionFactory sessionFactory)
   {
      Statistics statistics = sessionFactory.getStatistics();
      statistics.setStatisticsEnabled(true);
      
      synchronized (collectors)
      {
         collectors.put(database, new Collector(database, statistics));
      }
   }
   
   /**
    * Register the construction of a P2J query object by gathering application source code
    * location information for the query. This information is recorded transiently and is used
    * to collect additional information if the query is executed.
    * <p>
    * At this point, we are very early in the query's construction and do not yet have information
    * available as to the HQL which will be executed for this query, but we want to collect the
    * exact location in the application source code at which the construction happened.
    * <p>
    * Note: this method is expensive, as it constructs a <code>Throwable</code> on every call.
    * 
    * @param   query
    *          P2J query object which is being constructed.
    */
   void registerQuery(P2JQuery query)
   {
      Throwable t = new Throwable();
      String fileName = null;
      String methodName = null;
      int lineNumber = -1;
      for (StackTraceElement elem : t.getStackTrace())
      {
         String className = elem.getClassName();
         if (className.startsWith(appPackagePrefix))
         {
            if (lineNumber < 0)
            {
               fileName = elem.getFileName();
               lineNumber = elem.getLineNumber();
            }
            methodName = elem.getMethodName();
            
            if (className.lastIndexOf('$') < 0)
            {
               String shortClassName = className.substring(appPackagePrefixLength);
               Location loc = new Location(fileName, lineNumber, shortClassName, methodName);
               synchronized (queryCtorLocs)
               {
                  queryCtorLocs.put(query, loc);
               }
               
               break;
            }
         }
      }
   }
   
   /**
    * This method should be invoked just after P2J query object has executed its query. It
    * collects the HQL string used for the execution and increments the use count for all
    * executions of this specific HQL query from the location in the application source code at
    * which the given query was constructed.
    * 
    * @param   query
    *          P2J query object which executed its query.
    * @param   database
    *          Database upon which the query was executed.
    * @param   hql
    *          Dialect-specific HQL query statement which was executed.
    */
   void queryExecuted(P2JQuery query, Database database, String hql)
   {
      Location location = null;
      synchronized (queryCtorLocs)
      {
         location = queryCtorLocs.get(query);
      }
      if (location == null)
      {
         return;
      }
      
      Collector collector = null;
      synchronized (collectors)
      {
         collector = collectors.get(database);
      }
      
      if (collector != null)
      {
         collector.queryExecuted(location, query, hql);
      }
   }
   
   /**
    * Helper class which stores location information regarding the construction of a P2J query.
    */
   private static class Location
   {
      /** Unqualified application source code file name containing query */
      final String sourceFile;
      
      /** Line number at which query was constructed */
      final int lineNumber;
      
      /** Name of application class (sans package root name) containing query */
      final String className;
      
      /** Name of application-level method in which query was executed */
      final String methodName;
      
      /**
       * Constructor.
       * 
       * @param   sourceFile
       *          Unqualified application source code file name containing query.
       * @param   lineNumber
       *          Line number at which query was constructed.
       * @param   className
       *          Name of application class (sans package root name) containing query.
       * @param   methodName
       *          Name of application-level method in which query was executed
       */
      Location(String sourceFile, int lineNumber, String className, String methodName)
      {
         this.sourceFile = sourceFile.intern();
         this.lineNumber = lineNumber;
         this.className = className.intern();
         this.methodName = methodName.intern();
      }
      
      /**
       * Return the hash code computed for this object instance.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + lineNumber;
         result = 37 * result + className.hashCode();
         result = 37 * result + methodName.hashCode();
         result = 37 * result + sourceFile.hashCode();
         
         return result;
      }
      
      /**
       * Indicate whether this object instance is equivalent to the given instance, based on the
       * content of both.
       * 
       * @return  <code>true</code> if equivalent, else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof Location))
         {
            return false;
         }
         
         Location that = (Location) o;
         
         return (this.lineNumber == that.lineNumber &&
                 this.className == that.className &&
                 this.methodName == that.methodName &&
                 this.sourceFile == that.sourceFile);
      }
   }
   
   /**
    * Helper class which stores information about the localized execution of HQL queries. The
    * information is specific to particular locations in the application's source code. It
    * includes application source file name and line number where the query was constructed;
    * abbreviated application class name; application method name; and the number of times a
    * query at that location was executed. The execution count is cumulative across instances of
    * the same query (where "same" indicates the same HQL executed from a query constructed at
    * the same location in the application's source code).
    */
   private static class QueryInfo
   {
      /** Location information for query ctor */
      private final Location location;
      
      /** Dialect-specific HQL query string */
      private final String hql;
      
      /** Number of times query was executed at this location (across query instances) */
      private long useCount = 0;
      
      /**
       * Constructor.
       * 
       * @param   location
       *          Location information for query ctor.
       * @param   hql
       *          Dialect-specific HQL query string.
       */
      QueryInfo(Location location, String hql)
      {
         this.location = location;
         this.hql = hql;
      }
      
      /**
       * Get the query ctor location information.
       * 
       * @return  Location information.
       */
      Location getLocation()
      {
         return location;
      }
      
      /**
       * Get the dialect-specific HQL query string.
       * 
       * @return  HQL string.
       */
      String getHql()
      {
         return hql;
      }
      
      /**
       * Get the number of times the HQL query was executed at this location, across all query
       * instances constructed at the same location.
       * 
       * @return  Query execution count.
       */
      long getUseCount()
      {
         return useCount;
      }
      
      /**
       * Increment by one the number of times the HQL query was executed at this location, across
       * all query instances constructed at the same location.
       */
      void incrementUseCount()
      {
         useCount++;
      }
   }
   
   /**
    * Helper class which gathers local query statistics as the server is running, and which
    * collects global, Hibernate query and second level cache statistics as the server is
    * terminating. The statistics are collated at server termination and stored in an H2
    * database.
    */
   private static class Collector
   {
      /** Database for which statistics are collected */
      private final Database database;
      
      /** Statistics from the Hibernate <code>SessionFactory</code> for the database */
      private final Statistics statistics;
      
      /** Map of query ctor locations and HQL statements to local query statistics */
      private final Map<Key, QueryInfo> queryInfoMap = new HashMap<>();
      
      Collector(Database database, Statistics statistics)
      {
         this.database = database;
         this.statistics = statistics;
      }
      
      /**
       * This method is invoked just after P2J query object has executed its query. It collects
       * the HQL string used for the execution and increments the use count for all executions
       * of this specific HQL query from the location in the application source code at which
       * the given query was constructed.
       * 
       * @param   location
       *          Location at which the given query was constructed.
       * @param   query
       *          P2J query object which executed its query.
       * @param   hql
       *          Dialect-specific HQL query statement which was executed.
       */
      void queryExecuted(Location location, P2JQuery query, String hql)
      {
         synchronized (queryInfoMap)
         {
            Key key = new Key(location, hql);
            QueryInfo info = queryInfoMap.get(key);
            if (info == null)
            {
               info = new QueryInfo(location, key.getHql()); // hql from key has been intern'd
               queryInfoMap.put(key, info);
            }
            info.incrementUseCount();
         }
      }
      
      /**
       * Collate all statistics gathered by P2J queries and by Hibernate and store them in a
       * database for offline analysis.
       * 
       * @param   dbh
       *          Database helper object to manage storage of statistics in the stats database.
       */
      void collateStatistics(DatabaseHelper dbh)
      {
         // insert database record for this server run
         long did = dbh.insertDatabaseRecord(database.getName(), database.getType().getId());
         
         synchronized (queryInfoMap)
         {
            Map<String, List<QueryInfo>> hqlLocs = new HashMap<>();
            for (QueryInfo info : queryInfoMap.values())
            {
               String hql = info.getHql();
               List<QueryInfo> list = hqlLocs.get(hql);
               if (list == null)
               {
                  list = new ArrayList<>();
                  hqlLocs.put(hql, list);
               }
               list.add(info);
            }
            
            String[] queries = statistics.getQueries();
            for (String q : queries)
            {
               QueryStatistics stats = statistics.getQueryStatistics(q);
               
               long max = stats.getExecutionMaxTime();
               long min = stats.getExecutionMinTime();
               long avg = stats.getExecutionAvgTime();
               long execnt = stats.getExecutionCount();
               long rowcnt = stats.getExecutionRowCount();
               long hit = stats.getCacheHitCount();
               long miss = stats.getCacheMissCount();
               long put = stats.getCachePutCount();
               
               // insert global query statistics collected by Hibernate
               long gid = dbh.insertGlobalQueryStatistic(did,
                                                         min,
                                                         max,
                                                         avg,
                                                         hit,
                                                         miss,
                                                         put,
                                                         execnt,
                                                         rowcnt,
                                                         q);
               
               List<QueryInfo> list = hqlLocs.get(q);
               if (list != null)
               {
                  for (QueryInfo info : list)
                  {
                     // insert local query statistics, each associated with a global query
                     // record by foreign key
                     Location loc = info.getLocation();
                     dbh.insertLocalQueryStatistic(gid,
                                                   loc.sourceFile,
                                                   loc.className,
                                                   loc.methodName,
                                                   loc.lineNumber,
                                                   info.getUseCount());
                  }
               }
            }
            
            String[] regions = statistics.getSecondLevelCacheRegionNames();
            for (String r : regions)
            {
               SecondLevelCacheStatistics stats = statistics.getSecondLevelCacheStatistics(r);
               long cnt = stats.getElementCountInMemory();
               long mem = stats.getSizeInMemory();
               long hit = stats.getHitCount();
               long miss = stats.getMissCount();
               long put = stats.getPutCount();
               
               // insert second level cache statistics gathered by Hibernate
               dbh.insertCacheStatistic(did, cnt, mem, hit, miss, put, r);
            }
         }
      }
      
      /**
       * A hashable key comprised of both location information and an HQL string. Location info
       * alone is not enough to differentiate queries, because different HQL can be executed
       * from a query constructed at the same code location.
       */
      private static class Key
      {
         /** Query construction location information */
         private final Location location;
         
         /** Dialect-specific HQL query string */
         private final String hql;
         
         /**
          * Constructor.
          * 
          * @param   location
          *          Query construction location information.
          * @param   hql
          *          Dialect-specific HQL query string.
          */
         Key(Location location, String hql)
         {
            this.location = location;
            this.hql = hql.intern();
         }
         
         /**
          * Return the hash code computed for this object instance.
          * 
          * @return  Hash code.
          */
         public int hashCode()
         {
            int result = 17;
            result = 37 * result + location.hashCode();
            result = 37 * result + hql.hashCode();
            
            return result;
         }
         
         /**
          * Indicate whether this object instance is equivalent to the given instance, based on
          * the content of both.
          * 
          * @return  <code>true</code> if equivalent, else <code>false</code>.
          */
         public boolean equals(Object o)
         {
            if (!(o instanceof Key))
            {
               return false;
            }
            
            Key that = (Key) o;
            
            return (this.hql == that.hql && this.location.equals(that.location));
         }
         
         /**
          * Get the dialect-specific HQL query string.
          * 
          * @return  HQL query string.
          */
         String getHql()
         {
            return hql;
         }
      }
   }
   
   /**
    * Helper class to manage database operations while storing data to performance statistics
    * database. This class assumes an H2 back end and used H2-compliant syntax accordingly. It
    * uses JDBC directly, rather going through an ORM layer.
    */
   private static class DatabaseHelper
   {
      /** Connection user name; defaults to empty string, but may be overridden in directory */
      static String dbUser = "";
      
      /** Connection password; defaults to empty string, but may be overridden in directory */
      static String dbPass = "";
      
      /** Unique identifier for the current run of the application */
      private static final long runId = System.currentTimeMillis();
      
      /** JDBC URL to open/create/connect database */
      private static final String dbURL = "jdbc:h2:./perfdb/perfdb;";
      
      /** Database record insert statement */
      private final PreparedStatement stmtDatabaseInsert;
      
      /** Global query statistic record insert statement */
      private final PreparedStatement stmtGlobalQueryInsert;
      
      /** Local query statistic record insert statement */
      private final PreparedStatement stmtLocalQueryInsert;
      
      /** Cache statistic record insert statement */
      private final PreparedStatement stmtCacheInsert;
      
      /** Database connection */
      private Connection connection = null;
      
      /**
       * Default constructor.
       */
      DatabaseHelper()
      {
         Statement stmt = null;
         
         try
         {
            // load driver
            Class.forName("org.h2.Driver");
            
            openConnection();
            stmt = connection.createStatement();
            
            // create database table
            String create =
               "create table if not exists database (" +
               "id identity primary key, " +
               "run_id bigint not null, name varchar not null, type int not null)";
            stmt.execute(create);
            
            // prepare insert statement for global query statistics records
            String insert = "insert into database (run_id, name, type) values(?, ?, ?)";
            stmtDatabaseInsert = connection.prepareStatement(insert,
                                                             Statement.RETURN_GENERATED_KEYS);
            
            // create global query statistics table
            create =
               "create table if not exists global_query (" +
               "id identity primary key, " +
               "did bigint, " +
               "min bigint not null, max bigint not null, avg bigint not null, " +
               "hit bigint not null, miss bigint not null, put bigint not null, " +
               "execnt bigint not null, rowcnt bigint not null, " +
               "hql varchar not null)";
            stmt.execute(create);
            
            String alter =
               "alter table global_query add constraint if not exists fk_global_query_database " +
               "foreign key (did) references database";
            stmt.execute(alter);
            
            // prepare insert statement for global query statistics records
            insert =
               "insert into global_query (" +
               "did, min, max, avg, hit, miss, put, execnt, rowcnt, hql) " +
               "values(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
            stmtGlobalQueryInsert = connection.prepareStatement(insert,
                                                                Statement.RETURN_GENERATED_KEYS);
            
            // create local query table
            create =
               "create table if not exists local_query (" +
               "id identity, gid bigint, " +
               "file varchar not null, class varchar not null, method varchar not null, " +
               "line int not null, cnt bigint not null)";
            stmt.execute(create);
            
            alter =
               "alter table local_query add constraint if not exists " +
               "fk_local_query_global_query " + 
               "foreign key (gid) references global_query";
            stmt.execute(alter);
            
            // prepare insert statement for local query statistics records
            insert =
               "insert into local_query (gid, file, class, method, line, cnt) " +
               "values(?, ?, ?, ?, ?, ?)";
            stmtLocalQueryInsert = connection.prepareStatement(insert);
            
            // create cache statistics table
            create =
               "create table if not exists cache (" +
               "id identity, " +
               "did bigint not null, " +
               "cnt bigint not null, mem bigint not null, " +
               "hit bigint not null, miss bigint not null, put bigint not null, " +
               "region varchar not null)";
            stmt.execute(create);
            
            alter =
               "alter table cache add constraint if not exists fk_cache_database " + 
               "foreign key (did) references database";
            stmt.execute(alter);
                  
            // prepare insert statement for cache statistics records
            insert =
               "insert into cache (did, cnt, mem, hit, miss, put, region) " +
               "values(?, ?, ?, ?, ?, ?, ?)";
            stmtCacheInsert = connection.prepareStatement(insert);
            
            connection.commit();
         }
         catch (Exception exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error preparing database", exc);
         }
         finally
         {
            if (stmt != null)
            {
               try
               {
                  stmt.close();
               }
               catch (SQLException exc)
               {
                  // ignore
               }
            }
         }
      }
      
      /**
       * Insert a database record, which stores the server's run id in addition to the given
       * parameters.
       * 
       * @param   name
       *          Physical name of Progress database matching the database for which we are
       *          collecting results.
       * @param   type
       *          Database type:
       *          <ul>
       *          <li>0: primary</li>
       *          <li>1: dirty</li>
       *          <li>2: meta</li>
       *          </ul>
       * 
       * @return  Primary key of the created record.
       */
      long insertDatabaseRecord(String name, int type)
      {
         try
         {
            stmtDatabaseInsert.setLong(1, runId);
            stmtDatabaseInsert.setString(2, name);
            stmtDatabaseInsert.setInt(3, type);
            stmtDatabaseInsert.execute();
            stmtDatabaseInsert.clearParameters();
            
            // read back the primary key that was autogenerated
            ResultSet rs = stmtGlobalQueryInsert.getGeneratedKeys();
            
            if (rs.next())
            {
               return rs.getLong(1);
            }
            
            throw new RuntimeException("Error collecting database statistics");
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Insert a global query statistics record, which stores the server's run id in addition to
       * the given parameters.
       * 
       * @param   did
       *          Foreign key to related database.
       * @param   min
       *          Minimum query execution time, in ms.
       * @param   max
       *          Maximum query execution time, in ms.
       * @param   avg
       *          Average query execution time, in ms.
       * @param   hit
       *          Number of query cache hits for this query.
       * @param   miss
       *          Number of query cache misses for this query.
       * @param   put
       *          Number of query cache puts for this query.
       * @param   execnt
       *          Total number of executions of this query.
       * @param   rowcnt
       *          Total number of rows found by this query.
       * @param   hql
       *          Dialect-specific HQL which was executed.
       * 
       * @return  Primary key of the created record.
       */
      long insertGlobalQueryStatistic(long did,
                                      long min,
                                      long max,
                                      long avg,
                                      long hit,
                                      long miss,
                                      long put,
                                      long execnt,
                                      long rowcnt,
                                      String hql)
      {
         try
         {
            stmtGlobalQueryInsert.setLong(1, did);
            stmtGlobalQueryInsert.setLong(2, min);
            stmtGlobalQueryInsert.setLong(3, max);
            stmtGlobalQueryInsert.setLong(4, avg);
            stmtGlobalQueryInsert.setLong(5, hit);
            stmtGlobalQueryInsert.setLong(6, miss);
            stmtGlobalQueryInsert.setLong(7, put);
            stmtGlobalQueryInsert.setLong(8, execnt);
            stmtGlobalQueryInsert.setLong(9, rowcnt);
            stmtGlobalQueryInsert.setString(10, hql);
            stmtGlobalQueryInsert.execute();
            stmtGlobalQueryInsert.clearParameters();
            
            // read back the primary key that was autogenerated
            ResultSet rs = stmtGlobalQueryInsert.getGeneratedKeys();
            
            if (rs.next())
            {
               return rs.getLong(1);
            }
            
            throw new RuntimeException("Error collecting database statistics");
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Insert a local query statistics record.
       * 
       * @param   gid
       *          Foreign key to related global query statistics record.
       * @param   fileName
       *          Unqualified source file name.
       * @param   className
       *          Partially qualified class file name (package root prefix has been trimmed).
       * @param   methodName
       *          Application-level method name in which query was constructed.
       * @param   lineNumber
       *          Application-level line number at which query was constructed.
       * @param   cnt
       *          Number of times query was executed at this location.
       */
      void insertLocalQueryStatistic(long gid,
                                     String fileName,
                                     String className,
                                     String methodName,
                                     int lineNumber,
                                     long cnt)
      {
         try
         {
            stmtLocalQueryInsert.setLong(1, gid);
            stmtLocalQueryInsert.setString(2, fileName);
            stmtLocalQueryInsert.setString(3, className);
            stmtLocalQueryInsert.setString(4, methodName);
            stmtLocalQueryInsert.setInt(5, lineNumber);
            stmtLocalQueryInsert.setLong(6, cnt);
            stmtLocalQueryInsert.execute();
            stmtLocalQueryInsert.clearParameters();
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Insert a second level cache statistics record.
       * 
       * @param   did
       *          Foreign key to related database.
       * @param   cnt
       *          Number of elements in memory in this cache region.
       * @param   mem
       *          Amount of memory used, in bytes, for this cache region.
       * @param   hit
       *          Number of cache hits for this region.
       * @param   miss
       *          Number of cache misses for this region.
       * @param   put
       *          Number of cache puts for this region.
       * @param   region
       *          Name of the cache region represented by this record.
       */
      void insertCacheStatistic(long did,
                                long cnt,
                                long mem,
                                long hit,
                                long miss,
                                long put,
                                String region)
      {
         try
         {
            stmtCacheInsert.setLong(1, did);
            stmtCacheInsert.setLong(2, cnt);
            stmtCacheInsert.setLong(3, mem);
            stmtCacheInsert.setLong(4, hit);
            stmtCacheInsert.setLong(5, miss);
            stmtCacheInsert.setLong(6, put);
            stmtCacheInsert.setString(7, region);
            stmtCacheInsert.execute();
            stmtCacheInsert.clearParameters();
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Open a new connection to the database (if one already opened, it will close it first).
       * 
       * @throws  RuntimeException
       *          if new connection could not be established.
       */
      private void openConnection()
      {
         if (connection != null)
         {
            closeConnection();
         }
         
         try
         {
            connection = DriverManager.getConnection(dbURL, dbUser, dbPass);
         }
         catch (Exception exc)
         {
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Close the current connection.
       *  
       * @throws  RuntimeException
       *          if connection could not be closed.
       */
      private void closeConnection()
      {
         if (connection == null)
         {
            return;
         }
         
         try
         {
            connection.close();
         }
         catch (Exception exc)
         {
            throw new RuntimeException("Error collecting database statistics", exc);
         }
      }
      
      /**
       * Commit the outstanding transaction and run analyze on the database.
       * 
       * @throws  RuntimeException
       *          if there were errors during commit or analyze.
       */
      void finalizeChanges()
      {
         Statement stmt = null;
         
         try
         {
            connection.commit();
            
            // analyze (also commits)
            stmt = connection.createStatement();
            stmt.execute("analyze");
         }
         catch (Exception exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error collecting database statistics", exc);
         }
         finally
         {
            closeConnection();
         }
      }
   }
}