/*
** Module   : DmoPropertyAccessor.java
** Abstract : PropertyAccessor using ReflectASM instead of Java reflection.
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------- -------------------------Description---------------------------------
** 001 ECF 20180106 Created Initial version intended to improve performance.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.*;
import org.hibernate.*;
import org.hibernate.engine.spi.*;
import org.hibernate.property.*;
import com.esotericsoftware.reflectasm.*;

/**
 * A Hibernate property accessor for data model objects which at the time of writing is slightly
 * faster than the basic property accessor which uses reflection to invoke DMO getter and setter
 * methods.
 * <p>
 * This implementation uses the ReflectASM library to access DMO getter and setter methods
 * through custom bytecode assembly.
 * <p>
 * Getter and setter methods invoked by this accessor must not throw exceptions.
 */
public class DmoPropertyAccessor
implements PropertyAccessor
{
   /** Cached information needed to create accessors using shared MethodAccess instances */
   private static final Map<Class<?>, Access> accessCache = new ConcurrentHashMap<>();
   
   /** Cached getter property accessors */
   private static final Map<Key, DmoGetter> getters = new ConcurrentHashMap<>();
   
   /** Cached setter property accessors */
   private static final Map<Key, DmoSetter> setters = new ConcurrentHashMap<>();
   
   /**
    * Create a method access object for the given class, along with the information needed to
    * create getter and setter accessors using shared {@code MethodAccess} instances.
    * 
    * @param   dmoClass
    *          DMO class for which accessors are needed.
    * 
    * @return  Object containing the method access object and related information.
    */
   private static Access createAccess(Class<?> dmoClass)
   {
      // get DMO interface for property/method lookup
      
      // first check if we're an inner class; if so, we need our enclosing class
      Class<?> topClass = dmoClass.getEnclosingClass();
      if (topClass == null)
      {
         topClass = dmoClass;
      }
      
      String name = topClass.getSimpleName();
      Class<?> dmoIface = null;
      for (Class<?> iface : topClass.getInterfaces())
      {
         if (name.startsWith(iface.getSimpleName()))
         {
            dmoIface = iface;
            break;
         }
      }
      
      if (dmoIface == null)
      {
         return null;
      }
      
      Map<Method, String> g2p = PropertyHelper.pojoPropertiesByGetter(dmoIface);
      Map<Method, String> s2p = PropertyHelper.pojoPropertiesBySetter(dmoIface);
      Access access = null;
      
      // we synchronize here not for the cache integrity (which already is thread-safe), but
      // to ensure we don't compile two ASM access classes for the same DMO class
      synchronized (accessCache)
      {
         access = accessCache.get(dmoClass);
         
         if (access == null)
         {
            MethodAccess methodAccess = MethodAccess.get(dmoClass);
            
            Map<String, Integer> getterIndices = new HashMap<>();
            Map<String, Class<?>> types = new HashMap<>();
            for (Map.Entry<Method, String> e : g2p.entrySet())
            {
               Method method = e.getKey();
               String property = e.getValue();
               try
               {
                  Integer index = methodAccess.getIndex(method.getName(), 0);
                  getterIndices.put(property, index);
                  types.put(property, method.getReturnType());
               }
               catch (IllegalArgumentException exc)
               {
                  // this happens when interrogating the MethodAccess of a composite inner
                  // class using a getter method of the enclosing class; do not map these
               }
            }
            
            Map<String, Integer> setterIndices = new HashMap<>();
            for (Map.Entry<Method, String> e : s2p.entrySet())
            {
               Method method = e.getKey();
               String property = e.getValue();
               try
               {
                  Integer index = methodAccess.getIndex(method.getName(), 1);
                  setterIndices.put(property, index);
               }
               catch (IllegalArgumentException exc)
               {
                  // this happens when interrogating the MethodAccess of a composite inner
                  // class using a setter method of the enclosing class; do not map these
               }
            }
            
            access = new Access(methodAccess, getterIndices, setterIndices, types);
            
            accessCache.put(dmoClass, access);
         }
      }
      
      return access;
   }
   
   /**
    * Get an {@code Access} object, creating and caching it first if necessary.
    * 
    * @param   dmoClass
    *          DMO class for which accessors are needed.
    * 
    * @return  Object containing the method access object and related information.
    */
   private static Access getAccess(Class<?> dmoClass)
   {
      Access access = accessCache.get(dmoClass);
      if (access == null)
      {
         access = createAccess(dmoClass);
      }
      
      return access;
   }
   
   /**
    * Return a {@code Getter} Hibernate can use to read a DMO property.
    * 
    * @param   dmoClass
    *          DMO class for which access is needed.
    * @param   property
    *          Name of the DMO property to access.
    * 
    * @return  A {@code Getter} Hibernate can use to read a DMO property.
    * 
    * @throws  PropertyNotFoundException
    *          if the class is not a DMO or the property does not exist.
    */
   public Getter getGetter(Class dmoClass, String property)
   {
      Key key = new Key(dmoClass, property);
      
      DmoGetter getter = getters.get(key);
      if (getter == null)
      {
         Access access = getAccess(dmoClass);
         
         Integer methodIndex = null;
         if (access == null || (methodIndex = access.getterIndices.get(property)) == null)
         {
            throw new PropertyNotFoundException(
               "DMO property '" + dmoClass.getName() + '.' + property + "' not found");
         }
         Class<?> returnType = access.types.get(property);
         
         getter = new DmoGetter(access.methodAccess, methodIndex.intValue(), returnType);
         
         getters.putIfAbsent(key, getter);
      }
      
      return getter;
   }
   
   /**
    * Return a {@code Setter} Hibernate can use to write to a DMO property.
    * 
    * @param   dmoClass
    *          DMO class for which access is needed.
    * @param   property
    *          Name of the DMO property to access.
    * 
    * @return  A {@code Setter} Hibernate can use to write to a DMO property.
    * 
    * @throws  PropertyNotFoundException
    *          if the class is not a DMO or the property does not exist.
    */
   public Setter getSetter(Class dmoClass, String property)
   {
      Key key = new Key(dmoClass, property);
      
      DmoSetter setter = setters.get(key);
      if (setter == null)
      {
         Access access = getAccess(dmoClass);
         
         Integer methodIndex = null;
         if (access == null || (methodIndex = access.setterIndices.get(property)) == null)
         {
            throw new PropertyNotFoundException(
               "DMO property '" + dmoClass.getName() + '.' + property + "' not found");
         }
         
         setter = new DmoSetter(access.methodAccess, methodIndex.intValue());
         
         setters.putIfAbsent(key, setter);
      }
      
      return setter;
   }
   
   /**
    * Implementation of a {@code Getter} which uses assembled bytecode to call a getter method
    * instead of Java reflection.
    */
   public static final class DmoGetter
   implements Getter
   {
      /** Object used to invoke getter method */
      private final MethodAccess methodAccess;
      
      /** Index of the method to be invoked in the {@code MethodAccess}'s method array */
      private final int methodIndex;
      
      /** Type returned by the invoked method */
      private final Class<?> returnType;
      
      /**
       * Constructor.
       * 
       * @param   methodAccess
       *          Object used to invoke getter method.
       * @param   methodIndex
       *          Index of the method to be invoked in the {@code MethodAccess}'s method array.
       * @param   returnType
       *          Type returned by the invoked method.
       */
      DmoGetter(MethodAccess methodAccess, int methodIndex, Class<?> returnType)
      {
         this.methodAccess = methodAccess;
         this.methodIndex = methodIndex;
         this.returnType = returnType;
      }
      
      /**
       * Get the current value of the given DMO's property managed by this {@code Getter}.
       * 
       * @param   target
       *          DMO instance upon which to invoke method.
       * 
       * @return  Property's current value.
       */
      public Object get(Object target)
      {
         return methodAccess.invoke(target, methodIndex);
      }
      
      /**
       * Get the current value of the given DMO's property managed by this {@code Getter}.
       * 
       * @param   target
       *          DMO instance upon which to invoke method.
       * @param   mergeMap
       *          Not used.
       * @param   session
       *          Not used.
       * 
       * @return  Property's current value.
       */
      public Object getForInsert(Object target, Map mergeMap, SessionImplementor session)
      {
         return get(target);
      }
      
      /**
       * No-op.
       * 
       * @return  {@code null}.
       */
      public Member getMember()
      {
         return null;
      }
      
      /**
       * No-op.
       * 
       * @return  {@code null}.
       */
      public Method getMethod()
      {
         return null;
      }
      
      /**
       * Return the name of the method invoked by this accessor.
       * 
       * @return  Method name.
       */
      public String getMethodName()
      {
         return methodAccess.getMethodNames()[methodIndex];
      }
      
      /**
       * Get the return type of the method invoked by this accessor.
       * 
       * @return  Return type.
       */
      public Class getReturnType()
      {
         return returnType;
      }
      
      /**
       * Return a string representation of this object.
       * 
       * @return  String representation of this object.
       */
      public String toString()
      {
         return "DmoGetter(" + getMethodName() + ')';
      }
   }
   
   /**
    * Implementation of a {@code Setter} which uses assembled bytecode to call a setter method
    * instead of Java reflection.
    */
   public static final class DmoSetter
   implements Setter
   {
      /** Object used to invoke setter method */
      private final MethodAccess methodAccess;
      
      /** Index of the method to be invoked in the {@code MethodAccess}'s method array */
      private final int methodIndex;
      
      /**
       * Constructor.
       * 
       * @param   methodAccess
       *          Object used to invoke setter method.
       * @param   methodIndex
       *          Index of the method to be invoked in the {@code MethodAccess}'s method array.
       */
      DmoSetter(MethodAccess methodAccess, int methodIndex)
      {
         this.methodAccess = methodAccess;
         this.methodIndex = methodIndex;
      }
      
      /**
       * Invoke the property's setter method.
       * 
       * @param   target
       *          DMO instance upon which to invoke method.
       * @param   value
       *          Parameter to pass to setter method.
       * @param   factory
       *          Not used.
       */
      public void set(Object target, Object value, SessionFactoryImplementor factory)
      {
         methodAccess.invoke(target, methodIndex, value);
      }
      
      /**
       * No-op.
       * 
       * @return  {@code null}.
       */
      public Method getMethod()
      {
         return null;
      }
      
      /**
       * Return the name of the method invoked by this accessor.
       * 
       * @return  Method name.
       */
      public String getMethodName()
      {
         return methodAccess.getMethodNames()[methodIndex];
      }
      
      /**
       * Return a string representation of this object.
       * 
       * @return  String representation of this object.
       */
      public String toString()
      {
         return "DmoSetter(" + getMethodName() + ')';
      }
   }
   
   /**
    * A class containing the ReflectASM {@code MethodAccess} object and information about
    * method indices and return types used to create accessor objects using shared instances of
    * {@code MethodAccess}.
    *
    */
   private static class Access
   {
      /** Object used to invoke methods */
      final MethodAccess methodAccess;
      
      /** Map of property names to getter method indices */
      final Map<String, Integer> getterIndices;
      
      /** Map of property names to setter method indices */
      final Map<String, Integer> setterIndices;
      
      /** Map of property names to property types */
      final Map<String, Class<?>> types;
      
      /**
       * Constructor.
       * 
       * @param   methodAccess
       *          Object used to invoke methods.
       * @param   getterIndices
       *          Map of property names to getter method indices.
       * @param   setterIndices
       *          Map of property names to setter method indices.
       * @param   types
       *          Map of property names to property types.
       */
      Access(MethodAccess methodAccess,
             Map<String, Integer> getterIndices,
             Map<String, Integer> setterIndices,
             Map<String, Class<?>> types)
      {
         this.methodAccess = methodAccess;
         this.getterIndices = getterIndices;
         this.setterIndices = setterIndices;
         this.types = types;
      }
   }
   
   /**
    * Key for {@code Getter} and {@code Setter} cache maps.
    */
   private static class Key
   {
      /** DMO class */
      private final Class<?> dmoClass;
      
      /** Property name */
      private final String property;
      
      /**
       * Constructor.
       * 
       * @param   dmoClass
       *          DMO class.
       * @param   property
       *          Property name.
       */
      Key(Class<?> dmoClass, String property)
      {
         this.dmoClass = dmoClass;
         this.property = property;
      }
      
      /**
       * Override standard {@code hashCode} implementation to be consistent with {@code equals}.
       * 
       * @return  Hash code.
       */
      @Override
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + dmoClass.hashCode();
         result = 37 * result + property.hashCode();
         
         return result;
      }
      
      /**
       * Override standard {@code equals} implementation to be consistent with {@code hashCode}.
       * 
       * @return  {@code true} if this object is equivalent to the given object, else {@code
       *          false}.
       */
      @Override
      public boolean equals(Object o)
      {
         if (!(o instanceof Key))
         {
            return false;
         }
         
         Key that = (Key) o;
         
         return this.dmoClass.equals(that.dmoClass) && this.property.equals(that.property);
      }
   }
}