/*
** Module   : LegacyField.java
** Abstract : Annotation used to set the legacy field name and attributes for a DMO property.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20131015 Created initial version.
** 002 OM  20140117 Added fieldId (support for runtime indexed access for fields).
** 003 VMN 20140122 Added format(), initial(), columnLabel(), label(), mandatory(),
**                  validateMessage(), validateExpression().
** 004 VMN 20140226 Added index() and original().
** 005 OM  20140411 Added caseSensitive attribute.
** 006 ECF 20160219 Added extent attribute, changed index default to 0.
** 007 ECF 20170826 Added attributes for serialized/XML output.
** 008 GES 20180415 Added order and position attributes.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.annotation.*;

/**
 * Annotation definition to mark a DMO property with its legacy field name and attributes.
 */
@Target(ElementType.FIELD)
@Retention(RetentionPolicy.RUNTIME)
public @interface LegacyField
{
   /** COLUMN-LABEL attribute of this field. */
   String columnLabel() default "";
   
   /**
    * The id of this field in the sequence of the fields of the buffer. It is needed by the
    * dynamic <code>bufferField(int)</code> attribute of the <code>BufferImpl</code> for
    * emulating the <code>BUFFER-FIELD(integer)</code> buffer handle attribute.
    */
   int fieldId();
   
   /** 
    * Extent for array (extent) fields (defaults to 0 for scalar fields). Will be set to the
    * original extent for both normalized and denormalized fields.
    */
   int extent() default 0;
   
   /**
    * Index in the legacy extent field associated with a DMO property. It will be filled only
    * for field with custom extents.
    */
   int index() default 0;
   
   /** FORMAT attribute of this field. */
   String format() default "";
   
   /** INITIAL attribute of this field. */
   String initial() default "";
   
   /** LABEL attribute of this field. */
   String label() default "";
   
   /** MANDATORY attribute of this field. */
   boolean mandatory() default false;
   
   /** The name of the legacy field associated with a DMO property. */
   String name();
   
   /** ORDER attribute from the original schema. */
   int order() default 0;
   
   /** POSITION attribute from the original schema. */
   int position() default 0;
   
   /** VALIDATE-MESSAGE attribute of this field. */
   String validateMessage() default "";
   
   /** VALIDATE-EXPRESSION attribute of this field. */
   String validateExpression() default "";
   
   /** Original property name (not exposed in DMO) for denormalized field with extent. */
   String original() default "";
   
   /** Flag the case sensitive character fields. Other datatypes are not affected. */
   boolean caseSensitive() default false;
   
   /** Flag indicating the field should not be included in serialized output. */
   boolean serializeHidden() default false;
   
   /** Name of field in serialized output. */
   String serializeName() default "";
   
   /** XML schema data type of field. */
   String xmlDataType() default "";
   
   /** Name of element or attribute representing field in XML output. */
   String xmlNodeName() default "";
   
   /** Node type of field in XML output. */
   String xmlNodeType() default "ELEMENT";
}