/*
** Module   : ORMHandler.java
** Abstract : Manages loading of ORM documents into a Hibernate Configuration
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20080305   @37508 Created initial version. Manages the loading of Hibernate ORM 
**                           documents into a Hibernate Configuration.
** 002 GES 20080318   @37641 Switched from writer to stream to avoid unwanted character
**                           conversion when the default character set is not 8-bit ASCII.
** 003 CA  20080425   @38113 Added possibility to remove the foreign-key mappings from the hbm
**                           files, for the dirty databases.
** 004 ECF 20080509   @38497 Minor change to logging. Added database info to HBM dump.
** 005 GES 20090515   @42216 Import changes.
** 006 CA  20101026          Made the class and c'tor public. Built a new mapClass method which 
**                           takes a list of indexed, computed columns.
** 007 SVL 20120331          Upgraded to Hibernate 4. Now this class stores mappings instead of
**                           immediately applying it to Hibernate configuration.
** 008 VMN 20131012          Added injectComputedColumns.
** 009 CA  20131015          Added registration of legacy names for permanent tables.
** 010 ECF 20131111          Pass entityResolver when asking DatabaseManager for backing table
**                           name, to prevent error finding DTD when launching server offline.
** 011 OM  20140410          Fixed support case-sensitive fields in indexes.
**                           Reorder methods by visibility. Simplified string concatenation.
**                           Replaced Apache logging.
** 012 VMN 20140427          Added enhance schema name conversion support for hint "escape"
**                           attribute.
** 013 IAS 20160621          Added support for read-only tables.
** 014 ECF 20160830          Fixed support for read-only tables, which was only enabled for
**                           dialects needing computed columns.
** 015 ECF 20180106          Override default property access from BasicPropertyAccessor to
**                           DmoPropertyAccessor
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import javax.xml.namespace.*;
import javax.xml.stream.*;
import javax.xml.stream.events.*;
import org.hibernate.cfg.*;
import org.hibernate.internal.util.xml.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.util.*;

/**
 * Class which manages the loading of a Hibernate <code>Configuration</code>.
 * Each ORM is loaded via the {@link #mapClass} method.  Depending upon the
 * database dialect, this method may inject specialized nodes into the ORM XML
 * documents before handing them off to Hibernate.  This is how we manage the
 * use of computed columns for databases which do not allow expressions within
 * indexes.
 */
public class ORMHandler
{
   /** Foreign-key column name suffix */
   static final String FK_SUFFIX = "__fk";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ORMHandler.class);
   
   /** XML property element */
   private static final String ELEM_PROPERTY = "property";
   
   /** XML many-to-one element */
   private static final String ELEM_MANY_TO_ONE = "many-to-one";
   
   /** XML one-to-one element */
   private static final String ELEM_ONE_TO_ONE = "one-to-one";
   
   /** XML name attribute */
   private static final String ATTR_NAME = "name";
   
   /** XML column attribute */
   private static final String ATTR_COLUMN = "column";
   
   /** XML access attribute */
   private static final String ATTR_ACCESS = "access";
   
   /** XML insert attribute */
   private static final String ATTR_INSERT = "insert";
   
   /** XML update attribute */
   private static final String ATTR_UPDATE = "update";
   
   /** NopPropertyAccessor fully qualified class name */
   private static final String NOP_ACCESSOR = "com.goldencode.p2j.persist.NopPropertyAccessor";
   
   /** DmoPropertyAccessor fully qualified class name */
   private static final String DMO_ACCESSOR = "com.goldencode.p2j.persist.DmoPropertyAccessor";
   
   /** QName of the 'name' attribute */
   private static final QName CNAME = new QName("name");
   
   /** QName of the 'table' attribute */
   private static final QName TNAME = new QName("table");
   
   /** QName of the 'name' usage */
   private static final QName UNAME = new QName("usage");
   
  /** P2J database */
   private final Database database;
   
   /** Entity resolver used when parsing XML */
   private final EntityResolver entityResolver;
   
   /** This flag determines if the foreign keys should be removed or not */
   private boolean removeForeignKeys = false;
   
   /** Instance of XML helper. */
   private final XMLHelper xmlHelper;
   
   /** Stored Hibernate mappings. */
   private final Set<XmlDocument> hibernateMappings = new HashSet<>();
   
   /** Instance of entity resolver. */
   private EntityResolver streamEntityResolver = null;
   
   /** The dialect used by this handler. */
   private final P2JDialect dialect;
   
   /**
    * Constructor.
    * 
    * @param   dialect
    *          P2J database dialect.
    * @param   database
    *          Database whose ORM mappings require loading.
    */
   public ORMHandler(P2JDialect dialect, Database database)
   {
      this.removeForeignKeys = database.isDirty();
      this.database = database;
      this.dialect = dialect;
      this.xmlHelper = new XMLHelper();
      this.entityResolver = new DTDEntityResolver();
      
      if (!(dialect.needsComputedColumns() && dialect.injectComputedColumns()) &&
          !removeForeignKeys)
      {
         streamEntityResolver = new EJB3DTDEntityResolver();
      }
   }
   
   /**
    * Apply specified Hibernate mappings to the given Hibernate configuration.
    *
    * @param mappings
    *        Hibernate mappings to apply (a set of loaded XML documents).
    *
    * @param configuration
    *        Hibernate configuration to which the mappings are applied.
    */
   public static void applyHibernateMappings(Set<XmlDocument> mappings,
                                             Configuration configuration)
   {
      for (XmlDocument document : mappings)
      {
         configuration.add(document);
      }
   }
   
   /**
    * Retrieve the database table info for a given DMO implementation class, by reading it
    * directly from the .hbm.xml document.
    * 
    * @param   dmoClass
    *          DMO implementation class to which the table is mapped via ORM.
    * @param   resolver
    *          Entity resolver.
    * 
    * @return  Info of the database table as determined from the .hbm.xml document.
    */
   static TableInfo getBackingTableInfo(Class<?> dmoClass, EntityResolver resolver)
   throws PersistenceException
   {
      String clazzName = dmoClass.getSimpleName();
      String fileName = clazzName + ".hbm.xml";
      String table = null;
      boolean isReadOnly = false;
      InputStream in = null;
      try
      {
         in = dmoClass.getResourceAsStream(fileName);
         XMLInputFactory xif = XMLInputFactory.newInstance();
         xif.setXMLResolver(new XMLResolver()
         {
            @Override
            public Object
            resolveEntity(String publicID, String systemID, String baseURI, String namespace)
            throws XMLStreamException
            {
               try
               {
                  InputSource is = resolver.resolveEntity(publicID, systemID);
                  return is != null ? is.getByteStream() : null;
               }
               catch (SAXException | IOException e)
               {
                  throw new XMLStreamException("Failed to created XML resolver", e);
               }
            }
         });
         XMLEventReader eventReader = xif.createXMLEventReader(in);
         while(eventReader.hasNext())
         {
            XMLEvent event = eventReader.nextEvent();
            if (event.getEventType() != XMLStreamConstants.START_ELEMENT) 
            {
               continue;
            }
            StartElement startElement = event.asStartElement();
            String qName = startElement.getName().getLocalPart();
            if (qName.equals("class")) 
            {
               Attribute att = startElement.getAttributeByName(CNAME);
               if (att != null && clazzName.equals(att.getValue()))
               {
                  att = startElement.getAttributeByName(TNAME);
                  if (att != null) 
                  {
                     table = att.getValue();
                  }
                  break;
               }
            }
         }
         
         boolean done = false;
         while (!done && eventReader.hasNext())
         {
            XMLEvent event = eventReader.nextEvent();
            switch(event.getEventType())
            {
               case XMLStreamConstants.START_ELEMENT:
                  StartElement startElement = event.asStartElement();
                  String qName = startElement.getName().getLocalPart();
                  if (qName.equals("cache")) 
                  {
                     Attribute att = startElement.getAttributeByName(UNAME);
                     if (att != null ) 
                     {
                        isReadOnly = "read-only".equalsIgnoreCase(att.getValue());
                     }
                     done = true;
                  }
                  break;
               case XMLStreamConstants.END_ELEMENT:
               case XMLStreamConstants.END_DOCUMENT:
                  done = true;
                  break;
               default:
                  break;
            }
         }
      }
      catch (Exception e)
      {
         // couldn't read the .hbm.xml file, stop server initialization
         throw new PersistenceException("Couldn't read " + fileName + " document ", e);
      }
      finally
      {
         if (in != null) 
         {
            try
            {
               in.close();
            }
            catch (IOException e)
            {
               throw new PersistenceException("Couldn't close " + fileName + " document ", e);
            }
         }
      }
      
      return new TableInfo(table, isReadOnly);
   }
   /**
    * Visit an XML node and possibly inject a computed column property element directly after it.
    * The node must meet the following conditions for the element to be injected behind it:
    * <ul>
    *   <li>it must be an element node;
    *   <li>it must have a 'property' tag name;
    *   <li>it must reference a database text column which participates in an index.
    * </ul>
    *
    * @param   indexedColumns
    *          Set of those character column names which correspond with columns participating
    *          in database indexes.
    * @param   caseSensitiveCols
    *          The set of case-sensitive character columns that participate in the index.
    * @param   parent
    *          Visited node's parent node.
    * @param   child
    *          XML node being visited.
    *
    * @return  Injected XML node, or <code>null</code> if no node was injected.
    */
   static private Node visitChild(Set<String> indexedColumns,
                                  Set<String> caseSensitiveCols,
                                  Node parent,
                                  Node child,
                                  P2JDialect dialect)
   {
      short nodeType = child.getNodeType();
      if (nodeType != Node.ELEMENT_NODE)
      {
         return null;
      }
      
      Element elem = (Element) child;
      String tagName = elem.getTagName();
      if (!ELEM_PROPERTY.equals(tagName))
      {
         return null;
      }
      
      // override default property access from BasicPropertyAccessor to DmoPropertyAccessor
      if (!elem.hasAttribute(ATTR_ACCESS))
      {
         elem.setAttribute(ATTR_ACCESS, DMO_ACCESSOR);
      }
      
      String colName = elem.getAttribute(ATTR_COLUMN);
      if (!indexedColumns.contains(colName))
      {
         return null;
      }
      String propName = elem.getAttribute(ATTR_NAME);
      boolean cs = caseSensitiveCols.contains(colName);
      propName = dialect.getComputedColumnPrefix(cs) + propName;
      colName = DBUtils.getPrefixedParameter(colName, dialect.getComputedColumnPrefix(cs));
      
      Element copy = (Element) elem.cloneNode(false);
      copy.setAttribute(ATTR_NAME, propName);
      copy.setAttribute(ATTR_COLUMN, colName);
      copy.setAttribute(ATTR_INSERT, "false");
      copy.setAttribute(ATTR_UPDATE, "false");
      copy.setAttribute(ATTR_ACCESS, NOP_ACCESSOR);
      
      // Insert before current node.
      parent.insertBefore(copy, child);
      
      // Invert the order of these two nodes, so computed column comes last.
      parent.insertBefore(child, copy);
      
      return copy;
   }
   
   /**
    * Get stored Hibernate mappings.
    *
    * @return  Stored Hibernate mappings.
    */
   public Set<XmlDocument> getHibernateMappings()
   {
      return hibernateMappings;
   }
   
   /**
    * Configure the DMO class by reading its definition from the specified ORM file, updating the
    * object to relational mapping to inject computed column property nodes if the database
    * dialect requires it.
    *
    * @param   schema
    *          Database schema.
    * @param   ormLoc
    *          The ORM file location.
    * @param   indexedCols
    *          A set containing the names of the indexed character columns.
    * @param   caseSensitiveCols
    *          A set containing the names of the indexed, case-sensitive character columns.
    *
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes into the ORM XML.
    * @throws  FileNotFoundException
    *          if the ORM file was not found. 
    */
   public void mapClass(String schema,
                        String ormLoc,
                        Set<String> indexedCols,
                        Set<String> caseSensitiveCols)
   throws PersistenceException,
          FileNotFoundException
   {
      InputStream in = new FileInputStream(ormLoc);
      mapClass(schema, in, ormLoc, indexedCols, caseSensitiveCols);
   }
   
   /**
    * Walk the children of an XML element node, testing each to determine whether it represents a
    * property which corresponds with a text column which participates in a database index.  For
    * those that test positive, inject a computed a computed column property element after them.
    * <p>
    * This method calls itself recursively.
    *
    * @param   indexedColumns
    *          Set of those character column names which correspond with
    *          columns participating in database indexes.
    * @param   caseSensitiveCols
    *          The set of case-sensitive character columns that participate in the index.
    * @param   node
    *          XML parent node whose children are to be visited.
    */
   static void walkChildren(Set<String> indexedColumns,
                            Set<String> caseSensitiveCols,
                            Node node,
                            P2JDialect dialect)
   {
      Node next = node.getFirstChild();
      Node injected = null;
      while (next != null)
      {
         // Analyze child node and possibly inject a computed column entry after it.
         injected = visitChild(indexedColumns, caseSensitiveCols, node, next, dialect);
         
         // Recursively walk children, depth-first.
         walkChildren(indexedColumns, caseSensitiveCols, next, dialect);
         
         // Get next node.
         next = next.getNextSibling();
         
         // Skip over an injected node.
         if (injected != null && next == injected)
         {
            next = next.getNextSibling();
         }
      }
   }
   
   /**
    * Configure the specified DMO class, updating the object to relational
    * mapping to inject computed column property nodes if the database dialect
    * requires it.
    * 
    * @param   schema
    *          Database schema.
    * @param   dmoClass
    *          DMO implementation class.
    *          
    * @return  <code>true</code> if the corresponding table is read-only
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    */
   boolean mapClass(String schema, Class<?> dmoClass)
   throws PersistenceException
   {
      Set<String> indexedCols = null;
      Set<String> caseSensCols = new HashSet<>();
      boolean permanentDB = DatabaseManager.isPermamentDB(database);
      boolean isReadOnly = false;
      TableInfo ti = null;
      String table = null;
      
      if (permanentDB)
      {
         ti = getBackingTableInfo(dmoClass, entityResolver);
         table = ti.name;
         isReadOnly = ti.isReadOnly;
      }
      
      if (dialect.needsComputedColumns() && dialect.injectComputedColumns())
      {
         Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
         Map<String, P2JIndexComponent> indexComponentMap;
         
         if (!permanentDB)
         {
            indexComponentMap = TempTableHelper.getComputedCols(database, dmoIface);
         }
         else
         {
            // for permanent tables, we must read the indexes from the underlying database
            // using JDBC.
            indexComponentMap = DMOIndex.getIndexedCharacterColumns(schema, database, table);
         }
         
         // invert ignore-case set of character fields obtained to case-sensitive set
         indexedCols = indexComponentMap.keySet();
         for (P2JIndexComponent idxComp : indexComponentMap.values())
         {
            if (!idxComp.isIgnoreCase())
            {
               caseSensCols.add(idxComp.getName());
            }
         }
      }
      
      String ormLoc = getORMLoc(dmoClass);
      ClassLoader loader = dmoClass.getClassLoader();
      InputStream in = loader.getResourceAsStream(ormLoc);
      
      mapClass(schema, in, ormLoc, indexedCols, caseSensCols);
      
      if (permanentDB)
      {
         // only the tables in the permanent DBs are registered at server startup
         TableMapper.mapPermanentDMO(schema, dmoClass);
      }
      
      return isReadOnly;
   }
   
   /**
    * Configure the DMO class by reading its definition from the specified
    * stream, updating the object to relational mapping to inject computed 
    * column property nodes if the database dialect requires it.
    * 
    * @param   schema
    *          Database schema.
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    * @param   indexedCols
    *          A set containing the names of the indexed character columns.
    * @param   caseSensitiveCols
    *          The set of case-sensitive character columns that participate in the index.
    * 
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes into the ORM XML.
    */
   void mapClass(String schema,
                 InputStream in,
                 String ormLoc,
                 Set<String> indexedCols,
                 Set<String> caseSensitiveCols)
   throws PersistenceException
   {
      if (dialect.injectComputedColumns())
      {
         // after injected the computed property nodes, the foreign keys will be removed if needed
         injectComputedColumnsAndProcessFK(indexedCols, caseSensitiveCols, in, ormLoc);
      }
      else
      {
         if (removeForeignKeys)
         {
            Document doc = removeForeignKeys(in, ormLoc);
            storeMappingDoc(doc);
         }
         else
         {
            storeMappingStream(in);
         }
      }
   }
   
   /**
    * For each DMO character property which corresponds with a column
    * participating in a database index, inject a special property node into
    * the ORM XML document for the given class.  The injected property node
    * represents a read-only (no insert, no update) property, which does not
    * actually exist in the DMO.  Since there is no actual, corresponding
    * property in the DMO, a {@link NopPropertyAccessor} is used to intercept
    * calls to get and set the property's value in the DMO.  The element is
    * injected directly after the real property node.  The injected element
    * uses the same name as the real property, only with a leading prefix
    * prescribed by the corresponding database dialect.
    * <p>
    * Also, if the {@link #removeForeignKeys} flag is set to <code>true</code>,
    * all foreign key mappings are removed from the associated hbm.xml file.
    * 
    * @param   indexedColumns
    *          Set of those character column names which correspond with
    *          columns participating in database indexes. 
    * @param   caseSensitiveCols
    *          The set of case-sensitive character columns that participate in the index.
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    *
    * @throws  PersistenceException
    *          if there is an error injecting computed column property nodes
    *          into the ORM XML.
    */
   private void injectComputedColumnsAndProcessFK(Set<String> indexedColumns,
                                                  Set<String> caseSensitiveCols,
                                                  InputStream in,
                                                  String ormLoc)
   throws PersistenceException
   {
      try
      {
         Document doc = loadHBMFile(in);
         Element root = doc.getDocumentElement();
         
         walkChildren(indexedColumns, caseSensitiveCols, root, dialect);
         
         if (removeForeignKeys)
         {
            removeForeignKeys(root);
         }
         
         if (LOG.isLoggable(Level.FINE))
         {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            XmlHelper.write(doc, baos, true);
            String xml = new String(baos.toByteArray(), "UTF-8");
            
            LOG.log(Level.FINE, "Preprocessed ORM [" + database + "] " + ormLoc + "):\n" + xml);
         }
         
         storeMappingDoc(doc);
      }
      catch (Exception exc)
      {
         throw new PersistenceException(exc);
      }
   }
   
   /**
    * This method gets the ORM file location for the given DMO class.
    * 
    * @param   dmoClass
    *          The DMO class for which to load and parse the hbm.xml file.
    * 
    * @return  the ORM file location.
    */
   private String getORMLoc(Class<?> dmoClass)
   {
      // Build qualified mapping document filename.
      String dmoClsName = dmoClass.getName();
      String ormLoc = dmoClsName.replace('.', '/');
      return ormLoc + ".hbm.xml";
   }
   
   /**
    * This method loads and parses the specified hbm.xml.
    * 
    * @param   in
    *          The ORM stream.
    *          
    * @return  XML document created by parsing the associated hbm.xml file.
    * 
    * @throws  Exception
    *          If could not load or parse the hbm.xml file.
    */
   private Document loadHBMFile(InputStream in)
   throws Exception
   {
      return XmlHelper.parse(in, true, entityResolver, null);
   }
   
   /**
    * Remove the foreign keys from the associated hbm.xml ORM file associated
    * with the given DMO, and return the document.
    * 
    * @param   in
    *          The ORM input stream.
    * @param   ormLoc
    *          The ORM file location.
    * 
    * @return  the associated XML document with all the foreign keys removed.
    * 
    * @throws  PersistenceException
    *          if there is an error removing the foreign key nodes from the
    *          ORM XML.
    */
   private Document removeForeignKeys(InputStream in, String ormLoc)
   throws PersistenceException
   {
      try
      {
         Document doc = loadHBMFile(in);
         Element root = doc.getDocumentElement();
         
         removeForeignKeys(root);
         
         if (LOG.isLoggable(Level.FINE))
         {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            XmlHelper.write(doc, baos, true);
            String xml = new String(baos.toByteArray(), "UTF-8");
            
            LOG.log(Level.FINE, "Preprocessed ORM [" + database + "] " + ormLoc + "):\n" + xml);
         }
         
         return doc;
      }
      catch (Exception exc)
      {
         throw new PersistenceException(exc);
      }
   }
   
   /**
    * This method walks the DOM tree and removes the following nodes, when
    * a node with a <code>column</code> attribute containing a value which 
    * ends with the <code>__fk</code> suffix is encountered:
    * <ul>
    *    <li><code>many-to-one</code> nodes
    *    <li><code>one-to-one</code> nodes
    *    <li>if there is a <code>column</code> attribute with its value ending
    *        with the <code>__fk</code> suffix, remove the parent node if 
    *        this node is not a <code>many-to-one</code> or <code>one-to-one
    *        </code> node; this ensures that all collections - set, map, etc -
    *        are removed.
    * </ul>
    * 
    * @param   parent
    *          The parent node from which to start removing the foreign-key nodes.
    * 
    * @return  <code>true</code> if the <code>parent</code> node should be removed.
    */
   private boolean removeForeignKeys(Node parent)
   {
      Node child = parent.getFirstChild();
      Node nextChild = null;
      
      boolean removeParent = false;
      
      while (child != null)
      {
         boolean remove = removeForeignKeys(child);
         
         // if this node should be removed (it represents a collection - map,
         // set, etc) then remove it and continue processing
         if (remove)
         {
            nextChild = child.getNextSibling();
            parent.removeChild(child);
            child = nextChild;
            continue;
         }
         
         remove = false;
         
         // skip non-nodes elements
         short nodeType = child.getNodeType();
         if (nodeType == Node.ELEMENT_NODE)
         {
            Element elem = (Element) child;
            
            String column = elem.getAttribute(ATTR_COLUMN);
            String tag = elem.getTagName();
            
            // if this is a one-to-one relation, remove the node
            if (tag.equals(ELEM_ONE_TO_ONE))
            {
               remove = true;
            }
            else
            {
               if (column != null && column.endsWith(FK_SUFFIX))
               {
                  // if this is a many-to-one relation, remove the node
                  if (tag.equals(ELEM_MANY_TO_ONE))
                  {
                     remove = true;
                  }
                  else
                  {
                     // else, if this is not a many-to-one relation, the only
                     // possible conclusion is that this is a collection (map,
                     // set, etc) - so remove the parent node
                     removeParent = true;
                  }
               }
            }
         }
         // else do nothing
         
         if (remove)
         {
            nextChild = child.getNextSibling();
            parent.removeChild(child);
            child = nextChild;
         }
         else
         {
            child = child.getNextSibling();
         }
      }
      
      return removeParent;
   }
   
   /**
    * Read the Hibernate mappings from the given input stream and store them.
    *
    * @param inputStream
    *        Input stream containing Hibernate mappings.
    */
   private void storeMappingStream(InputStream inputStream)
   {
      InputSource inputSource = new InputSource(inputStream);
      Origin origin = new OriginImpl("input stream", null);
      try
      {
         XmlDocument xml = MappingReader.INSTANCE.readMappingDocument(streamEntityResolver,
                                                                      inputSource,
                                                                      origin);
         hibernateMappings.add(xml);
      }
      finally
      {
         try
         {
            inputStream.close();
         }
         catch (IOException ignore)
         {
            LOG.log(Level.WARNING, "Was unable to close input stream");
         }
      }
   }
   
   /**
    * Read the Hibernate mappings from the given document and store them.
    *
    * @param doc
    *        Document containing Hibernate mappings.
    */
   private void storeMappingDoc(Document doc)
   {
      org.dom4j.Document document = xmlHelper.createDOMReader().read(doc);
      hibernateMappings.add(new XmlDocumentImpl(document, "unknown", null));
   }
   
   /**
    *  Holder for the table info
    */
   public static class TableInfo
   {
      /** table name */
      public final String name;
      
      /** read-only flag */
      public final boolean isReadOnly;
      
      /**
       * Constructor
       * 
       * @param name
       *        table name
       * @param isReadOnly
       *        read-only flag
       */
      public TableInfo(String name, boolean isReadOnly) 
      {
         super();
         this.name = name;
         this.isReadOnly = isReadOnly;
      }
      
   }
}