/*
** Module   : PersistenceFactory.java
** Abstract : Factory for Persistence objects
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------------Description-------------------------
** 001 ECF 20071018   @35570 Created initial version. Factory for
**                           Persistence objects. Can create local and
**                           remote versions.
** 002 ECF 20080611   @38698 Disallow creation of a Persistence instance
**                           for a dirty database. In order to prevent
**                           programming errors, IllegalArgumentException
**                           is thrown if this is attempted.
** 003 ECF 20131010          Minor documentation update.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.hibernate.dialect.*;
import com.goldencode.p2j.persist.remote.*;

/**
 * Factory for {@link Persistence} objects.  Manages a cache of
 * <code>Persistence</code> objects, keyed by {@link Database}.  When an
 * instance is requested via one of the <code>getInstance()</code> method
 * variants, the cache is checked for an existing instance.  If none exists,
 * the factory determines what type of <code>Persistence</code> implementation
 * is necessary for a particular database, instantiates it, initializes it,
 * caches it, and returns it.
 * <p>
 * Currently, this factory knows how to generate instances of:
 * <ul>
 *   <li>{@link Persistence} - used for locally managed databases.
 *   <li>{@link com.goldencode.p2j.persist.remote.RemotePersistence
 *       RemotePersistence} - used for remotely managed databases.
 * </ul>
 * <p>
 * Persistence objects can be removed from the cache by other persistence
 * framework classes via the {@link #remove(Database)} method.
 * 
 * @author ECF
 */
public final class PersistenceFactory
{
   /** Cache of <code>Persistence</code> instances by database */
   private static final Map<Database, Persistence> cache =
      new HashMap<Database, Persistence>();
   
   /**
    * Get the {@link Persistence} instance associated with the given, physical database name,
    * creating it first if necessary. Assumes the name specifies a primary database. Newly
    * created instances are stored in a static cache.
    *
    * @param   name
    *          Name of physical database with which the instance is
    *          permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    */
   public static Persistence getInstance(String name)
   {
      return getInstance(new Database(name));
   }
   
   /**
    * Get the instance of this class associated with the given, physical database, creating it
    * first if necessary. Newly created instances are stored in a static cache.
    *
    * @param   database
    *          Database with which the instance is permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    * 
    * @throws  IllegalArgumentException
    *          if <code>database</code> represents a dirty database.
    */
   public static Persistence getInstance(Database database)
   {
      if (database.isDirty())
      {
         throw new IllegalArgumentException(
            "Cannot create a Persistence instance for dirty database '" +
            database +
            "'");
      }
      
      synchronized (cache)
      {
         Persistence instance = cache.get(database);
         if (instance == null)
         {
            Dialect dialect = DatabaseManager.getDialect(database);
            if (dialect == null)
            {
               throw new IllegalArgumentException(
                  "Unable to determine database dialect for database " +
                  database);
            }
            
            instance = (database.isLocal()
                        ? new Persistence(database, dialect)
                        : new RemotePersistence(database, dialect));
            instance.initializeInstance();
            
            cache.put(database, instance);
         }
         
         return instance;
      }
   }
   
   /**
    * Remove the <code>Persistence</code> instance associated with the given
    * database from the global cache.
    * <p>
    * This method should be called with caution.  It is provided primarily to
    * support transient databases, which are connected explicitly, used for a
    * time by one or more users, then disconnected.
    * 
    * @param   database
    *          Database with which the instance is associated.
    * 
    * @return  Instance of <code>Persistence</code> which was removed
    */
   static Persistence remove(Database database)
   {
      synchronized (cache)
      {
         return cache.remove(database);
      }
   }
}