/*
** Module   : TableSignature.java
** Abstract : Records the structure of a table in a manner that can be compared, serialized and
**            deserialized.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20180418 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Records the structure of a table in a manner that can be compared, serialized and
 * deserialized. This class is used for {@code RAW-TRANSFER}. This is a basic implementation.
 * The serialized form is not binary compatible with the 4GL. 
 */
class TableSignature
implements Externalizable
{
   /** Class to byte mapping table. **/
   private static final Map<Class<? extends BaseDataType>, Byte> c2b;
   
   /** Number of fields in the table. */
   private int num;
   
   /** Each field's data type in ascending order by the POSITION attribute in the schema. */
   private byte[] types;
   
   /** Each field's extent value in ascending order by the POSITION attribute in the schema. */
   private short[] extents;
   
   static
   {
      c2b = new HashMap<Class<? extends BaseDataType>, Byte>();
      
      c2b.put(character.class  , (byte) 0x0001);
      c2b.put(date.class       , (byte) 0x0002);
      c2b.put(logical.class    , (byte) 0x0003);
      c2b.put(integer.class    , (byte) 0x0004);
      c2b.put(decimal.class    , (byte) 0x0005);
      c2b.put(recid.class      , (byte) 0x0007);
      c2b.put(raw.class        , (byte) 0x0008);
      c2b.put(handle.class     , (byte) 0x000A);
      c2b.put(rowid.class      , (byte) 0x000D);
      c2b.put(datetime.class   , (byte) 0x0022);
      c2b.put(datetimetz.class , (byte) 0x0028);
      c2b.put(int64.class      , (byte) 0x0029);
   }
   
   /**
    * Default constructor.
    */
   public TableSignature()
   {
   }
   
   /**
    * Constructor.
    *
    * @param   fields
    *          List of field descriptors.
    */
   public TableSignature(List<TableMapper.LegacyFieldInfo> fields)
   {
      if (fields != null)
      {
         num     = fields.size();
         types   = new byte[num];
         extents = new short[num];
         
         for (int i = 0; i < num; i++)
         {
            TableMapper.LegacyFieldInfo fld = fields.get(i);
            
            Byte typ = c2b.get(fld.getDataType());
            
            types[i]   = (typ == null) ? (byte) 0x00FF : typ;
            extents[i] = (short) fld.getExtent();
         }
      }
   }
   
   /**
    * Test the equivalence of the current instance to the given object.
    *
    * @param    that
    *           The instance for comparison.
    */
   public boolean equals(Object that)
   {
      if (this == that)
         return true;
      
      if (that == null || !(that instanceof TableSignature))
         return false;
      
      TableSignature other = (TableSignature) that;
      
      return (num == other.num                        &&
              Arrays.equals(types, other.types)       && 
              Arrays.equals(extents, other.extents));
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      num     = in.readInt();
      types   = (byte[]) in.readObject();
      extents = (short[]) in.readObject();
   }
   
   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(num);
      out.writeObject(types);
      out.writeObject(extents);
   }
}