/*
** Module   : TableWrapper.java
** Abstract : A wrapper for table result sets, used to serialize a table data and send it to a 
*             remote side. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130620 Created initial version.
** 002 SVL 20140611 Fixed NPE in writeExternal when no result set is transferred. Added tableName.
** 003 SVL 20140797 Added tableHandle field.
** 004 IAS 20160919 Serialization optimization.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;

/**
 * Any {@link TableResultSet} instance will be wrapped in a {@link TableWrapper} instance before
 * sending it to a remote side.  This class is intended to be used only by P2J runtime.
 */
public class TableWrapper
implements Externalizable
{
   /** The result set to be sent to the remote side. */
   private TableResultSet resultSet = null;

   /** The list of read properties. */
   private List<PropertyDefinition> properties;
   
   /** The read data. */
   private List<Object[]> rows;

   /** Flag indicating this table is sent in INPUT or INPUT-OUTPUT mode. */
   private boolean input;

   /** Flag indicating this table is sent in OUTPUT or INPUT-OUTPUT mode. */
   private boolean output;
   
   /** Flag indicating this table is sent in APPEND mode. */
   private boolean append;

   /** Table name. */
   private String tableName;

   /**
    * Determines if this wrapper wraps data from a TABLE-HANDLE parameter. Necessary because of
    * quirks in error handling: calling side needs to know if it is a TABLE-HANDLE parameter (not
    * a TABLE parameter) on the called side.
    */
   private boolean tableHandle = false;
   
   /**
    * Default c'tor, explicitly added to allow instances of this class to be created on 
    * deserialization. Not for public use.
    */
   public TableWrapper()
   {
   }
   
   /**
    * Wrap a custom result set in a class known to P2J, so the remote side can deserialize it.
    * 
    * @param    resultSet
    *           The result set, which is a subclass of {@link TableResultSet}.
    */
   public TableWrapper(TableResultSet resultSet)
   {
      this.resultSet = resultSet;
   }
   
   /**
    * Check if this table is in INPUT or INPUT-OUTPUT mode.  This is used to access the data sent 
    * by a remote side, via {@link #readExternal}. Must not be called when {@link #resultSet} is 
    * set.
    * 
    * @return   The {@link #input} flag.
    */
   public boolean isInput()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return input;
   }
   
   /**
    * Check if this table is in OUTPUT or INPUT-OUTPUT mode.  This is used to access the data sent 
    * by a remote side, via {@link #readExternal}. Must not be called when {@link #resultSet} is 
    * set.
    * 
    * @return   The {@link #output} flag.
    */
   public boolean isOutput()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return output;
   }
   
   /**
    * Check if this table is in APPEND mode.  This is used to access the data sent 
    * by a remote side, via {@link #readExternal}. Must not be called when {@link #resultSet} is 
    * set.
    * 
    * @return   The {@link #append} flag.
    */
   public boolean isAppend()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return append;
   }

   /**
    * Returns table name.
    *
    * @return table name.
    */
   public String getTableName()
   {
      return tableName;
   }

   /**
    * Set table name.
    *
    * @param tableName
    *        table name.
    */
   public void setTableName(String tableName)
   {
      this.tableName = tableName;
   }

   /**
    * Provides an iterator over the read {@link #properties}.  This is used to access the data
    * sent by a remote side, via {@link #readExternal}. Must not be called when {@link #resultSet}
    * is set.
    * 
    * @return   See above.
    */
   public Iterator<PropertyDefinition> propertyIterator()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return (properties == null ? Collections.EMPTY_LIST : properties).iterator();
   }

   /**
    * Provides an iterator over the read {@link #rows}.  This is used to access the data sent by a
    * remote side, via {@link #readExternal}. Must not be called when {@link #resultSet} is set.
    * 
    * @return   See above.
    */
   public Iterator<Object[]> rowIterator()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return (rows == null ? Collections.EMPTY_LIST : rows).iterator();
   }

   /**
    * Get the list of read {@link #properties}.  This is used to access the data sent by a remote 
    * side, via {@link #readExternal}. Must not be called when {@link #resultSet} is set.
    * 
    * @return   See above.
    */
   public List<PropertyDefinition> getProperties()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return new ArrayList<PropertyDefinition>(properties);
   }

   /**
    * Get the list of read {@link #rows}.  This is used to access the data sent by a remote side,
    * via {@link #readExternal}. Must not be called when {@link #resultSet} is set.
    * 
    * @return   See above.
    */
   public List<Object[]> getRows()
   {
      if (resultSet != null)
      {
         throw new IllegalStateException(
           "A custom result set must not be specified when calling this method!");
      }

      return new ArrayList<Object[]>(rows);
   }

   /**
    * Send the result set (metadata and rows) to the specified output destination.
    * 
    * @param    out
    *           The output destination to which the result set will be sent.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public final void writeExternal(ObjectOutput out)
   throws IOException
   {
      // write modes
      if (resultSet != null)
      {
         out.writeByte(resultSet.isInput() ? 1 : 0);
         out.writeByte(resultSet.isOutput() ? 1 : 0);
         out.writeByte(resultSet.isAppend() ? 1 : 0);
      }
      else
      {
         out.writeByte(input ? 1 : 0);
         out.writeByte(output ? 1 : 0);
         out.writeByte(append ? 1 : 0);
      }

      out.writeObject(tableName);
      out.writeByte(tableHandle ? 1 : 0);
      
      // write properties
      if (resultSet != null)
      {
         while (resultSet.hasMoreProperties())
         {
            PropertyDefinition property = resultSet.nextProperty();
            out.writeByte(1);
            property.writeExternal(out);
         }
      }
      out.writeByte(0);
   
      // write data 
      if (resultSet != null)
      {
         while (resultSet.hasMoreRows())
         {
            Object[] row = resultSet.nextRow();
            out.writeObject(row);
         }
      }
      out.writeObject(null);
   }

   /**
    * Read the result set (metadata and rows) from the specified input source.
    * 
    * @param    in
    *           The input source from which the result set will be read.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public final void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      // read modes
      input = in.readByte() != 0;
      output = in.readByte() != 0;
      append = in.readByte() != 0;

      tableName = (String) in.readObject();
      tableHandle = in.readByte() != 0;
   
      properties = new ArrayList<PropertyDefinition>();
      do
      {
         if (in.readByte() == 0)
         {
            break;
         }
         PropertyDefinition pd = new PropertyDefinition();
         pd.readExternal(in);
         properties.add(pd);
      }
      while (true);
   
      // read data
      rows = new ArrayList<Object[]>();
      do
      {
         Object next = in.readObject();
         if (next == null)
         {
            break;
         }
         
         rows.add((Object[]) next);
      }
      while (true);
   }

   /**
    * Determines if this wrapper wraps data from a TABLE-HANDLE parameter.
    *
    * @return <code>true</code> if this wrapper wraps data from a TABLE-HANDLE parameter.
    */
   public boolean isTableHandle()
   {
      return tableHandle;
   }

   /**
    * Set if this wrapper wraps data from a TABLE-HANDLE parameter.
    *
    * @param tableHandle
    *        <code>true</code> if this wrapper wraps data from a TABLE-HANDLE parameter.
    */
   public void setTableHandle(boolean tableHandle)
   {
      this.tableHandle = tableHandle;
   }
}