/*
** Module   : TempTableConnectionProvider.java
** Abstract : specialized connection provider used for temp tables
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060619 @27502   Created initial version. A JDBC connection provider implementation 
**                           used by Hibernate to hold on to a particular JDBC connection for as
**                           long as one or more temp tables are using it.
** 002 SVL 20120331          Upgraded to Hibernate 4.
** 003 CA  20140513          Added a weight for the context-local var, to ensure predetermined
**                           order during context reset.
** 004 ECF 20150715          Replace StringBuffer with StringBuilder. Replace Apache commons
**                           logging with J2SE logging. Implement generics.
** 005 IAS 20160331          Fixed delegate declaration and initialization.
** 006 ECF 20160912          Exposed a private constant and slightly modified logging.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.InstantiationException;
import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;
import org.hibernate.service.*;
import org.hibernate.service.jdbc.connections.spi.*;
import org.hibernate.service.spi.*;

/**
 * A specialized connection provider which ensures that a database connection
 * used to create temp tables is preserved until the last temp table in use
 * is dropped.  A delegate connection provider is used to do the actual work
 * of obtaining and releasing connections, presumably from a connection pool.
 * This implementation will not release a connection back to the underlying
 * delegate until there are no more temp tables using the connection.  This
 * prevents the inadvertent loss of temp table data which would otherwise
 * occur if a Hibernate session is closed (as from a commit or rollback),
 * before all temp table processing begun on that connection was finished.
 * <p>
 * Another approach to solving this problem would have been to make the
 * persistence framework responsible for obtaining database connections for
 * temp tables and to supply these manually when opening new sessions via the
 * Hibernate <code>SessionFactory.openSession(Connection connection)</code>
 * method.  However, using that approach we would have lost the ability to
 * use Hibernate's second level cache for temp table records, AND we would
 * have had to code specialized logic in {@link Persistence} to differentially
 * handle temp tables and permanent tables.  Thus, an implementation of
 * Hibernate's <code>ConnectionProvider</code> interface was the preferred
 * approach.
 * <p>
 * This implementation wrappers a delegate connection provider whose
 * properties are specified in the directory.  To enable this implementation,
 * the fully qualified name of this class (i.e.,
 * <code>com.goldencode.p2j.persist.TempTableConnectionProvider</code>) is
 * specified for the <code>hibernate.connection.provider_class</code> property
 * in the directory.  The backing provider may be specified as well, using one
 * of the following mechanisms (see also the {@link #configure} method):
 * <ul>
 * <li>Specify an explicit, fully qualified provider class name using the
 *     property <code>hibernate.connection.delegate_provider_class</code>
 *     (note this is a P2J extension, not a standard Hibernate property name).
 * <li>Do not specify an explicit delegate provider class name, but instead
 *     let Hibernate's internal logic determine the provider to initialize,
 *     based upon the remaining Hibernate configuration properties provided
 *     in the directory (see Hibernate's configuration documentation for
 *     details on how it makes this determination).
 * </ul>
 * <p>
 * This class effectively operates as a singleton, even though it is possible
 * to create multiple instances, since Hibernate requires a public, default
 * constructor.  However, the implementation consists of only static
 * variables.  One of these is a <code>ContextLocal</code> variable
 * (<code>context</code>), which maintains connection state specific to each
 * client context.  This mechanism is used to ensure that even if multiple
 * instances of this class are created, requests to get and close JDBC
 * connections are managed at a client context level.
 */
public final class TempTableConnectionProvider
implements ConnectionProvider,
           Configurable
{
   /** Logger */
   private static final Logger LOG =
      LogHelper.getLogger(TempTableConnectionProvider.class.getName());
   
   /** Property name in configuration for delegate provider class name */
   static final String DELEGATE_CLASS_KEY = "hibernate.connection.delegate_provider_class";
   
   /** Property name in configuration for connection provider class name */
   private static final String PROVIDER_CLASS_KEY = "hibernate.connection.provider_class";
   
   /** Object which manages context local connection data */
   private static final ContextLocal<Context> context = new ContextLocal<Context>()
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LEVEL_2}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_2;
      };
      
      /**
       * Instantiate a new <code>Context</code> instance for each context.
       *
       * @return  Context object.
       */
      protected Context initialValue()
      {
         return new Context();
      }
      
      /**
       * Clean up the context before it is removed.
       *
       * @param   context
       *          Context object stored in this variable.
       */
      protected void cleanup(Context context)
      {
         if (context != null)
         {
            context.contextCleanup();
         }
      }
   };
   
   /** Connection provider to which actual work is delegated */
   private static volatile ConnectionProvider delegate = null;
   
   /**
    * Default constructor used by Hibernate's <code>ConnectionProviderFactory</code>
    * to instantiate this connection provider.
    */
   public TempTableConnectionProvider()
   {
   }
   
   /**
    * Configure this connection provider using the specified properties.
    * It is here where we construct and configure the delegate connection
    * provider.  The properties are for the most part passed through to the
    * connection provider factory to produce the delegate object.  However,
    * one property in particular is modified or removed.  The
    * <code>hibernate.connection.provider_class</code> property either:
    * <ol>
    * <li>has its value replaced with the value of the
    *     <code>hibernate.connection.delegate_provider_class</code> (if
    *     provided);  or
    * <li>is removed altogether if the
    *     <code>hibernate.connection.delegate_provider_class</code> property
    *     is not specified in the configuration.
    * </ol>
    * <p>
    * In the latter case, Hibernate's internal logic determines which
    * connection provider to create, based upon the remaining properties.
    *
    * @param   props
    *          Hibernate configuration properties.
    */
   public synchronized void configure(Map props)
   {
      if (delegate != null)
      {
         return;
      }
      
      // Copy all properties to pass to the connection provider factory when
      // requesting a delegate provider.
      Properties propCopy = new Properties();
      propCopy.putAll(props);
      
      // If a delegate connection provider class is specified, use it to
      // initialize the delegate.  Otherwise, remove the provider_class
      // property from the properties and try to create a new connection
      // provider instance as the delegate.
      String delegateClass = (String) props.get(DELEGATE_CLASS_KEY);
      if (delegateClass != null)
      {
         // Provide an explicit provider class for the delegate.
         propCopy.setProperty(PROVIDER_CLASS_KEY, delegateClass);
      }
      else
      {
         // Let Hibernate figure out which provider to create, based upon the
         // remaining properties.
         propCopy.remove(PROVIDER_CLASS_KEY);
      }
      
      // Instantiate and configure the proper backing provider.
      String cls = propCopy.getProperty(PROVIDER_CLASS_KEY);
      try
      {
         delegate = (ConnectionProvider) Class.forName(cls).newInstance();
         ((Configurable) delegate).configure(propCopy);
      }
      catch (InstantiationException e)
      {
         throw new HibernateException("Cannot instantiate connection provider " + cls);
      }
      catch (IllegalAccessException e)
      {
         throw new HibernateException("Cannot access connection provider " + cls);
      }
      catch (ClassNotFoundException e)
      {
         throw new HibernateException("Cannot find connection provider " + cls);
      }
      
      if (LOG.isLoggable(Level.CONFIG))
      {
         LOG.config("Temp table connection provider configured;  backed by " +
                    delegate.getClass().getName());
      }
   }
   
   /**
    * Obtain a JDBC connection on demand.  For a particular client context,
    * this method will fetch a new connection from the underlying delegate
    * provider only if no connection has previously been fetched, or if a
    * previously fetched connection was closed after all temp tables using it
    * were dropped.
    *
    * @return  Active temp table JDBC connection for the current context.
    *
    * @throws  SQLException
    *          if there is an error establishing a connection with the
    *          backing datastore.
    */
   public Connection getConnection()
   throws SQLException
   {
      return context.get().getConnection();
   }
   
   /**
    * Return a JDBC connection which is no longer needed in the current
    * context, releasing it to the underlying delegate provider if it is no
    * longer needed by any active temp tables.  If any temp table still
    * references it, it is not released to the delegate.
    *
    * @param   conn
    *          The connection to be released.
    *
    * @throws  SQLException
    *          if there is an error closing the connection with the backing
    *          datastore.
    */
   public void closeConnection(Connection conn)
   throws SQLException
   {
      context.get().closeConnection(conn);
   }
   
   /**
    * Returns <code>false</code> to indicate JDBC connections will not be
    * released aggressively, regardless of how the delegate would handle such
    * requests.
    *
    * @return  <code>false</code>.
    */
   public boolean supportsAggressiveRelease()
   {
      return false;
   }
   
   /**
    * Can this wrapped service be unwrapped as the indicated type?
    *
    * @param  unwrapType
    *         The type to check.
    *
    * @return <code>true</code>/<code>false</code>
    */
   public boolean isUnwrappableAs(Class unwrapType)
   {
      return ConnectionProvider.class.equals(unwrapType) ||
             TempTableConnectionProvider.class.isAssignableFrom(unwrapType);
   }
   
   /**
    * Unproxy the service proxy.
    *
    * @param  unwrapType
    *         The java type as which to unwrap this instance.
    *
    * @return The unwrapped reference.
    *
    * @throws UnknownUnwrapTypeException if the service cannot be unwrapped as the indicated type.
    */
   public <T> T unwrap(Class<T> unwrapType)
   {
      if (isUnwrappableAs(unwrapType))
      {
         return (T) this;
      }
      else
      {
         throw new UnknownUnwrapTypeException(unwrapType);
      }
   }
   
   /**
    * An instance of this class is created for each client context, in order
    * to manage connection data specific to that context.  This class
    * determines when the delegate will be asked to obtain and release the
    * current JDBC connection, using a reference count maintained by the
    * {@link TemporaryBuffer} class.
    */
   private static class Context
   {
      /** Active JDBC connection for temp tables in this client context */
      private Connection connection = null;
      
      /**
       * Fetch a JDBC connection, which will be either the cached connection
       * (if previously opened and not subsequently closed), or which will be
       * newly obtained from the delegate connection provider (if it is not
       * currently cached).
       *
       * @return  JDBC connection.
       *
       * @throws  SQLException
       *          if an error occurs obtaining the connection from the
       *          backing datastore.
       */
      private Connection getConnection()
      throws SQLException
      {
         boolean fine = LOG.isLoggable(Level.FINE);
         
         if (connection == null)
         {
            connection = delegate.getConnection();
            
            if (fine)
            {
               LOG.fine(status("Got backing connection"));
            }
         }
         
         if (fine)
         {
            LOG.fine(status("getConnection"));
         }
         
         return connection;
      }
      
      /**
       * Optionally release a JDBC connection to the delegate connection
       * provider.  The connection will be released to the delegate only if
       * no temp tables currently are using it.
       *
       * @param   conn
       *          JDBC connection to be returned to the provider.
       *
       * @throws  IllegalStateException
       *          if the connection returned (<code>conn</code>) is not the
       *          same connection currently cached by this object.
       * @throws  SQLException
       *          if an error occurs closing the connection.
       */
      private void closeConnection(Connection conn)
      throws SQLException
      {
         if (conn != connection && connection != null)
         {
            throw new IllegalStateException(
               "Attempted to close unknown temp table database connection");
         }
         
         boolean fine = LOG.isLoggable(Level.FINE);
         
         if (fine)
         {
            LOG.fine(status("closeConnection"));
         }
         
         if (connection != null && TemporaryBuffer.getTableCount() == 0)
         {
            delegate.closeConnection(connection);
            connection = null;
            
            if (fine)
            {
               LOG.fine(status("Released backing connection"));
            }
         }
      }
      
      /**
       * Perform clean up when the current client context is being closed.
       * If a connection currently is cached, is returned to the delegate
       * provider, so that it may be recycled or released.
       */
      private void contextCleanup()
      {
         if (connection == null)
         {
            return;
         }
         
         try
         {
            delegate.closeConnection(connection);
         }
         catch (SQLException exc)
         {
            LOG.log(Level.SEVERE, "Error closing JDBC connection", exc);
         }
      }
      
      /**
       * Generate a debug string which reports the current status of this
       * object along with a custom message.  Status includes the number of
       * temp tables currently accessing the cached connection, as well as
       * an identifier of the cached connection itself (if any).
       *
       * @param   message
       *          Message to be prepended to the status string.
       *
       * @return  Text which reports the given message plus the current
       *          status of this object.
       */
      private String status(String message)
      {
         StringBuilder buf = new StringBuilder(Utils.describeContext());
         buf.append(":  ");
         buf.append(message);
         buf.append(";  tableCount: ");
         buf.append(TemporaryBuffer.getTableCount());
         buf.append(";  connection: ");
         buf.append(connection);
         
         return buf.toString();
      }
   }
}