/*
** Module   : FunctionKey.java
** Abstract : Function descriptor which can be used as a hash table key
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20080312  @37493  Created initial version. Function descriptor which can be used as a
**                           hash table key.
** 002 ECF 20080508  @38486  Use Database object rather than database name String. This allows
**                           differentiation of the same function across multiple databases with
**                           the same name.
** 003 ECF 20150223          Lower case function name on construction to ensure case-insensitive
**                           look-ups.
** 004 OM  20151130          Removed NPE in toString() when database was not set.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.hql;

import com.goldencode.p2j.persist.*;

/**
 * Descriptor of a user-defined function, suitable for use as a key in a hash table.
 */
public final class FunctionKey
{
   /** Database name */
   private final Database database;
   
   /** Function name */
   private final String function;
   
   /** Function parameter signature as HQL types */
   private final HQLTypes[] signature;
   
   /**
    * Constructor.
    * 
    * @param   function
    *          Function name.
    * @param   signature
    *          Function parameter signature as HQL types.
    */
   public FunctionKey(String function, HQLTypes[] signature)
   {
      this(null, function, signature);
   }
   
   /**
    * Constructor.
    * 
    * @param   database
    *          Database.
    * @param   function
    *          Function name.
    * @param   signature
    *          Function parameter signature as HQL types.
    */
   public FunctionKey(Database database, String function, HQLTypes[] signature)
   {
      this.database = database;
      this.function = function.toLowerCase().intern();
      this.signature = signature;
   }
   
   /**
    * Return hash code for this object.
    * 
    * @return  Hash code.
    * 
    * @see java.lang.Object#hashCode()
    */
   public int hashCode()
   {
      int result = 17;
      
      result = 37 * result + (database != null ? database.hashCode() : 0);
      result = 37 * result + function.hashCode();
      for (HQLTypes type : signature)
      {
         result = 37 * result + type.hashCode();
      }
      
      return result;
   }
   
   /**
    * Test this object for equality with the given object, which is assumed to
    * be an instance of this class.
    * 
    * @return  <code>true</code> if objects are equal, else <code>false</code>.
    * 
    * @see java.lang.Object#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof FunctionKey))
      {
         return false;
      }
      
      FunctionKey that = (FunctionKey) o;
      
      boolean bothNull = false;
      
      if (database == null || that.database == null)
      {
         bothNull = (database == that.database);
         if (!bothNull)
         {
            return false;
         }
      }
      
      if (!bothNull && !database.equals(that.database))
      {
         return false;
      }
      
      if (!function.equals(that.function))
      {
         return false;
      }
      
      int len1 = signature.length;
      int len2 = that.signature.length;
      
      if (len1 != len2)
      {
         return false;
      }
      
      for (int i = 0; i < len1; i++)
      {
         if (!signature[i].equals(that.signature[i]))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Get a string representation of this object's state.
    * 
    * @return  String representation of this object.
    * 
    * @see java.lang.Object#toString()
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder(database == null ? "" : database.toString());
      buf.append("::");
      buf.append(function);
      buf.append("(");
      for (int i = 0; i < signature.length; i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(signature[i]);
      }
      buf.append(")");
      
      return buf.toString();
   }
}