/*
** Module   : JdbcKeyRetriever.java
** Abstract : KeyRetriever implementation which retrieves keys using SQL
**            queries.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 SVL 20080827  @39779 Created initial version. Retrieves free keys using two subsequent 
**                          queries.
** 002 SVL 20091126  @44444 Use external function for retrieving the list of database tables.
** 003 SVL 20120331         Upgraded to Hibernate 4.
** 004 OM  20140603         Use multiple queries in order to workaround the SQL Server limitation
**                          of maximum 2100 parameters per query.
** 005 OM  20151130         Adjusted the theoretical number of parameters per query to practical
**                          value.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.id;

import com.goldencode.p2j.persist.*;
import java.util.*;
import com.goldencode.p2j.persist.dialect.P2JDialect;
import org.hibernate.*;
import org.hibernate.type.*;

/**
 * Provides a service for retrieving available primary keys from database using SQL queries.
 */
public class JdbcKeyRetriever
implements KeyRetriever
{
   /** Query which returns the number of busy keys in the given range. */
   private String[] predictionQueries;
   
   /** Query which retrieves all busy keys in the given range. */
   private String[] retrievalQueries;
   
   /** The number of tables queried at once (in a single SQL statement). */
   private int bucketSize; 
   
   /**
    * Initialize key retriever for the given database.
    * Because the number of tables from a database may be too big (at least SQL Server has a
    * limitation of 2100 parameters in a query) the keys for the tables are retrieved in buckets.
    *
    * @param   database
    *          Database for which will be scanned for available keys.
    */
   @Override
   public void initialize(Database database)
   {
      // Get all table names.
      List<String> tableNames = DatabaseManager.getDatabaseTables(database);
      final int tableCount = tableNames.size();
      
      // Obtain the is the maximum number of tables that are scanned at in a single SQL query.
      P2JDialect dialect = (P2JDialect) DatabaseManager.getDialect(database);
      // There are 2 parameter for each table
      final int bucketSize = dialect.getMaxParameterCount() / 3; 
      
      // compute the necessary number of buckets to hold all tables
      int bucketCount;
      if (bucketSize <= 0)
      {
         // disable bucket-ing, all at once
         bucketCount = 1;
         this.bucketSize = tableCount;
      }
      else
      {
         if ((tableCount % bucketSize) == 0)
         {
            // no remainder / leftovers 
            bucketCount = tableCount / bucketSize;
         }
         else
         {
            // the remainder of tables go to extra bucket
            bucketCount = tableCount / bucketSize + 1;
         }
         this.bucketSize = bucketSize;
      }
      // debug only:
      // System.out.println("JdbcKeyRetriever is using bucketSize=" + this.bucketSize + 
      //   ", the queries (" + tableCount + ") will be split in " + bucketCount + " queries");
      
      predictionQueries = new String[bucketCount];
      retrievalQueries = new String[bucketCount];
      
      // Compose prediction queries which look like
      // 'select(
      //  (select count(id) from table1 where id >= :minId and id <= :maxId) +
      //  (select count(id) from table2 where id >= :minId and id <= :maxId) +
      //  ...) as cnt'
      // in the final step the subqueries intermediary result will be summed to a grand total
      int bucket = 0;
      StringBuilder buf = new StringBuilder("select (");
      for (int i = 0; i < tableCount; i++)
      {
         assert buf != null;
         buf.append("(select count(id) from ");
         buf.append(tableNames.get(i));
         buf.append(" where id >= :minId and id <= :maxId)");
         
         if (isLastInBucket(i, tableCount))
         {
            // bucket full, next bucket
            buf.append(") as cnt");
            predictionQueries[bucket] = buf.toString();
            ++bucket;
            
            // prepare for next bucket
            buf = (bucket != bucketCount) ? new StringBuilder("select (") : null;
         }
         else
         {
            // there are more items in this bucket
            buf.append(" + ");
         }
      }
      
      // Compose id retrieval query which looks like
      // 'select id from table1 where id >= :minId and id <= :maxId UNION
      //  select id from table2 where id >= :minId and id <= :maxId UNION
      //  ....  ORDER BY id asc'
      bucket = 0;
      buf = new StringBuilder();
      for (int i = 0; i < tableCount; i++)
      {
         assert buf != null;
         buf.append("select id from ");
         buf.append(tableNames.get(i));
         buf.append(" where id >= :minId and id <= :maxId");

         if (isLastInBucket(i, tableCount))
         {
            buf.append(" ORDER BY id asc");
            retrievalQueries[bucket] = buf.toString();
            ++bucket;
            buf = (bucket != bucketCount) ? new StringBuilder() : null;
         }
         else
         {
           // combine with other items from the current bucket
            buf.append(" UNION ");
         }
      }
   }
   
   /**
    * Get available keys.
    * Each bucket is queried independently and the  the results are put back together.
    *
    * @param   minKey
    *          Lower bound of a range of identifiers on which search will be made.
    * @param   maxKey
    *          Upper bound of a range of identifiers on which search will be made.
    * @param   session
    *          Hibernate session which will be used to perform queries.
    *          
    * @return  a list of ranges of available keys
    */
   @Override
   public List<Range> retrieveFreeKeys(Long minKey, Long maxKey, Session session)
   {
      long rangeSize = maxKey - minKey + 1;
      
      long count = 0;
      for (String pQuery : predictionQueries)
      {
         // Execute prediction queries.
         SQLQuery query = session.createSQLQuery(pQuery);
         query.addScalar("cnt", StandardBasicTypes.LONG);
         query.setLong("minId", minKey);
         query.setLong("maxId", maxKey);

         // sum up the bucket subtotals
         count += (Long) query.uniqueResult();
      }

      if (count == 0L)
      {
         // All ids in the given range are free.
         List<Range> res = new ArrayList<>();
         res.add(new Range(minKey, maxKey));
         return res;
      }
      else if (count == rangeSize)
      {
         // No ids available.
         return null;
      }
      
      SortedSet<Long> busyKeys = new TreeSet<>();
      for (String rQuery : retrievalQueries)
      {
         // Execute the query which will return all busy ids in the given range.
         SQLQuery query = session.createSQLQuery(rQuery);
         query.addScalar("id", StandardBasicTypes.LONG);
         query.setLong("minId", minKey);
         query.setLong("maxId", maxKey);
         
         busyKeys.addAll((List<Long>) query.list());
      }
      
      return new Range(minKey, maxKey).splitRange(busyKeys);
   }
   
   /**
    * Checks if <code>i</code> is the last item in the bucket. Useful for combining items in a
    * bucket; the last item is usually handled differently.
    * 
    * @param   i
    *          The current item, related to global item array.
    * @param   size
    *          The global item array size.
    *          
    * @return  <code>true</code> if i is the last item in his bucket
    */
   private boolean isLastInBucket(int i, int size)
   {
      if (i == size - 1) 
      {
         // since this is the last item globally, it is also the last item in the last bucket
         return true;
      }
      
      // compute the position in the bucket
      int bucketPosition = i % this.bucketSize;
      return bucketPosition == this.bucketSize - 1;
   }
}