/*
** Module   : MetadataManager.java
** Abstract : Manages databases which contain metadata information
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20131001 Created initial version.
** 002 ECF 20131028 Moved to meta sub-package.
** 003 ECF 20140315 Fixed path from which remote port is read in directory.
** 004 ECF 20141204 Use schema name rather than physical database name to load metaschema data.
**                  Do not initialize the metadata manager (and server) if no metadata is in use.
** 005 OM  20151130 Adjusted import for H2Helper.
** 006 ECF 20160225 Changes required by new PropertyHelper implementation.
** 007 OM  20160722 Added support for _file._template. Use caching for fast access.
** 008 OM  20160712 Added setupIdentityManager() method for creating p2j_id_generator_sequence.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.meta;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.sql.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.dialect.*;
import org.w3c.dom.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * This class creates, prepares, and populates the embedded database(s) used to serve metadata
 * information to converted business logic and to the persistence framework. All methods are
 * static; the class is never instantiated. Most methods are only called during server startup.
 * The {@link #isMetaDMO(Class)} method, however, is meant to be used by callers who need to
 * determine whether a particular DMO interface is metadata-related.
 */
public final class MetadataManager
{
   /** Well-known name of metadata schema */
   public static final String META_SCHEMA = "_meta";
   
   /** Well-known prefix of metadata DMOs */
   private static final String META_PREFIX = "Meta";
   
   /** Well-known name of DDL script to create tables */
   private static final String DDL_TABLE = "meta_table_ddl.sql";
   
   /** Well-known name of DDL script to create indexes */
   private static final String DDL_INDEX = "meta_index_ddl.sql";
   
   /** Logger */
   private static final Logger log = LogHelper.getLogger(MetadataManager.class);
   
   /** Map of short data type names to associated wrapper classes */
   private static final Map<String, Class<?>> fieldClassMap = new HashMap<>();
   
   /** Map of short data type names to constructors which instantiate those types */
   private static final Map<String, Constructor<?>> ctorMap = new HashMap<>();
   
   /** Map of short meta DMO names to helper objects */
   private static final Map<String, Helper> helpers = new HashMap<>();
   
   /** Set of DMO interfaces which are metadata DMOs */
   private static Set<Class<?>> dmoMetaTypes = new HashSet<>();
   
   /** Base directory for DMO resources in jar file */
   private static String baseDir;
   
   /** Stores the status of the metadata server. */
   private static boolean inUse = false;
   
   /**
    * Cached the template records ({@code _file._template} values) for fast access. The keys are
    * the DMO implementation classes of the tables from permanent database. The values are the
    * {@link rowid} values as {@link Long} data.
    */
   private static final Map<Class<?>, Long> templateRecords = new HashMap<>();
   
   static
   {
      // all types have constructors that accept a single String argument; abbreviations are
      // cryptic, but intended to save space in the XML format
      fieldClassMap.put("j", Long.class      );
      fieldClassMap.put("i", integer.class   );
      fieldClassMap.put("r", recid.class     );
      fieldClassMap.put("d", decimal.class   );
      fieldClassMap.put("c", character.class );
      fieldClassMap.put("l", logical.class   );
      fieldClassMap.put("t", date.class      );
      fieldClassMap.put("w", raw.class       );
      fieldClassMap.put("o", rowid.class     );
      fieldClassMap.put("6", int64.class     );
      fieldClassMap.put("m", datetime.class  );
      fieldClassMap.put("z", datetimetz.class);
   }
   
   /**
    * Private constructor; this class is not instantiated.
    */
   private MetadataManager()
   {
   }
   
   /**
    * Indicate whether the given DMO interface represents a metaschema table.
    * 
    * @param   dmoIface
    *          DMO interface to test.
    * 
    * @return  {@code true} if {@code dmoIface} represents a metaschema table, else {@code false}.
    */
   public static boolean isMetaDMO(Class<?> dmoIface)
   {
      return dmoMetaTypes.contains(dmoIface);
   }
   
   /**
    * Initialize this class and start up a secure database server to allow remote P2J servers to
    * access metadata.
    * 
    * @return  <code>true</code> if the metadata manager was initialized; <code>false</code> if
    *          there are no metadata classes found.
    * 
    * @throws  PersistenceException
    *          if there is any error initializing the class or starting the server.
    */
   public static boolean initialize()
   throws PersistenceException
   {
      Iterator<Class<?>> iter = DMOIndex.classes(META_SCHEMA);
      
      if (iter == null)
      {
         if (log.isLoggable(Level.INFO))
         {
            log.log(Level.INFO, "Metadata server was not started; no metadata in use");
         }
         
         return false;
      }
      
      int pfxLen = META_PREFIX.length();
      int sfxLen = "Impl".length();
      String basePkg = DMOIndex.getBasePackage();
      baseDir = basePkg.replaceAll("\\.", "/");
      
      while (iter.hasNext())
      {
         Class<?> dmoImpl = iter.next();
         Class<?> dmoIface = null;
         String longName = dmoImpl.getName();
         String shortName = longName;
         int pos = longName.lastIndexOf('.');
         if (pos > 0)
         {
            shortName = longName.substring(pos + 1);
         }
         shortName = shortName.substring(0, shortName.length() - sfxLen);
         String ifaceName = String.format("%s.%s.%s", basePkg, META_SCHEMA, shortName);
         try
         {
            dmoIface = Class.forName(ifaceName);
         }
         catch(ClassNotFoundException exc)
         {
            throw new PersistenceException("Missing DMO interface: " + ifaceName);
         }
         
         Helper h = new Helper(dmoIface, dmoImpl);
         shortName = shortName.substring(pfxLen);
         helpers.put(shortName, h);
         dmoMetaTypes.add(dmoIface);
      }
      
      // start the server for mixed mode connections
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         throw new PersistenceException("Directory bind failed");
      }
      try
      {
         if (Utils.getDirectoryNodeBoolean(ds, "persistence/remote-meta/enabled", false, false))
         {
            int port = Utils.getDirectoryNodeInt(ds,
                                                 "persistence/remote-meta/port",
                                                 DatabaseManager.MIXED_MODE_SSL_PORT,
                                                 false);
            String portString = String.valueOf(port);
            if (log.isLoggable(Level.INFO))
            {
               log.log(Level.INFO,
                       "Starting secure, mixed mode, H2 metaschema server on port " +
                       portString + "...");
            }
            H2Helper.startMixedModeServer("-tcpSSL", "-tcpPort", portString, "-tcpAllowOthers");
            if (log.isLoggable(Level.INFO))
            {
               log.log(Level.INFO, "Successfully started secure mixed mode H2 server");
            }
         }
      }
      catch (PersistenceException exc)
      {
         log.log(Level.SEVERE, "Error starting mixed mode, remote metaschema server", exc);
         
         throw exc;
      }
      finally
      {
         ds.unbind();
      }
      
      inUse = true;
      return true;
   }
   
   /**
    * Checks the status of the metadata subsystem. Note that this method returns {@code true} from
    * the moment the {@link #initialize()} has finished with success. At that moment the whole
    * metadata system is not yet ready as the metadata tables are not yet populated. 
    * <p>
    * Use this method to check if the project was not configured with this feature or if metadata
    * initialization failed, and quickly leave the calling method in these cases.
    *  
    * @return  {@code true} if the {@code MetadataManager} was successfully initialized and
    *          queries on internal data are possible and {@code false} if the initialization
    *          failed and the metadata is not "in use".
    */
   public static boolean inUse()
   {
      return MetadataManager.inUse;
   }
   
   /**
    * Create and load the embedded metadata database for the associated primary database.
    * start up a secure database server to allow remote P2J servers to access this metadata.
    * 
    * @param   database
    *          Database information object.
    * 
    * @throws  PersistenceException
    *          if there is any error creating or loading the database.
    */
   public static void prepareDatabase(Database database)
   throws PersistenceException
   {
      Connection conn = DatabaseManager.checkOutConnection(database);
      
      try
      {
         // define schema
         executeDDLScript(database, conn, DDL_TABLE);
         
         // apply indexes
         executeDDLScript(database, conn, DDL_INDEX);
      }
      finally
      {
         DatabaseManager.checkInConnection(conn, database);
      }
      
      // define user functions
//      H2Helper.preparePermanentDatabase(database);
   }
   
   /**
    * Set up sequence identity mManager for this database by creating the fresh instance of 
    * {@code p2j_id_generator_sequence}. Because it is repopulated at each startup the start value
    * of the sequence is not crucial; the value is hardcoded to 10000. 
    * 
    * @param   database
    *          The database to set up the identity manager for.
    * @param   dialect
    *          The dialect for the database. It is used to build the DDL for the sequence create
    *          statement.
    *          
    * @throws  PersistenceException
    *          if something goes wrong with database connection. 
    */
   public static void setupIdentityManager(Database database, Dialect dialect)
   throws PersistenceException
   {
      Connection conn = DatabaseManager.checkOutConnection(database);
      
      try
      {
         P2JDialect p2jDialect = (P2JDialect) dialect;
         String seq_ddl = p2jDialect.getCreateSequenceString(Persistence.ID_GEN_SEQUENCE, 10000);
         try
         {
            H2Helper.runScript(conn, new StringReader(seq_ddl));
         }
         catch (Exception exc)
         {
            throw new PersistenceException(
                  "Error creating " + Persistence.ID_GEN_SEQUENCE + " for " + database, exc);
         }
      }
      finally
      {
         DatabaseManager.checkInConnection(conn, database);
      }
   }
   
   /**
    * Populate the specified, embedded database with metaschema information read from XML files
    * stored in the application jar file with a well-known naming/pathing convention:
    * <pre>
    * {base DMO package}.{database name}.meta.xml
    * </pre>
    * 
    * @param   database
    *          Metaschema database to populate.
    * @param   schema
    *          Name of the associated, primary database's schema.
    * 
    * @throws  PersistenceException
    *          if there is any problem reading the XML file, instantiating DMO instances, or
    *          inserting records into the database.
    */
   public static void populateDatabase(Database database, String schema)
   throws PersistenceException
   {
      String path = String.format("%s/%s/%s.meta.xml", baseDir, schema, schema);
      Document dom = null;
      
      try
      {
         // find the XML data file
         InputStream in = ClassLoader.getSystemResourceAsStream(path);
         if (in == null)
         {
            throw new MissingDataException("Could not load DDL script " + path);
         }
         
         // generate a system ID to allow us to find the DTD, assuming it lives in the same jar
         // file as this class
         String clsLoc = MetadataManager.class.getName().replace('.', '/') + ".class";
         URL clsURL = ClassLoader.getSystemResource(clsLoc);
         path = clsURL.toString();
         String systemID = path.substring(0, path.indexOf(clsLoc));
         
         // parse it
         dom = XmlHelper.parse(in, true, systemID);
         
         Persistence p = PersistenceFactory.getInstance(database);
         
         p.beginTransaction();
         
         // walk the DOM and prepare metadata records
         Element root = dom.getDocumentElement();
         NodeList recNodes = root.getElementsByTagName("record");
         int numRecs = recNodes.getLength();
         for (int i = 0; i < numRecs; i++)
         {
            Element record = (Element) recNodes.item(i);
            Persistable dmo = prepareRecord(record, schema);
            p.save(dmo, dmo.getId());
         }
         
         p.commit();
         
         // update database statistics
         p.executeSQL("analyze");
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Error populating metadata database " + database, exc);
      }
      
      if (log.isLoggable(Level.INFO))
      {
         log.log(Level.INFO, "Metaschema database " + database + " ready");
         log.log(Level.INFO, "Caching " + templateRecords.size() + " template record ids");
      }
   }
   
   /**
    * Return the rowid of the template record for a permanent table.
    * 
    * @param   table
    *          The permanent table to lookup.
    * 
    * @return  The value of the {@link rowid} of the template record. If metadata is not enabled
    *          for this project or {@code _file} metatable is not configured or {@code table} is
    *          an invalid name {@code null} is returned.
    */
   public static Long getTemplateRowid(Class<?> table)
   {
      return MetadataManager.inUse() ? templateRecords.get(table) : null;
   }
   
   /**
    * Given an XML element describing an individual metaschema table record, instantiate a DMO
    * and initialize its fields for the purpose of inserting the record into the metaschema
    * database.
    * 
    * @param   record
    *          XML element describing the record to be inserted.
    * @param   schema
    *          Name of the associated, primary database's schema.
    * 
    * @return  DMO initialized with the information read from the XML record.
    * 
    * @throws  PersistenceException
    *          if a data value to be set into the DMO is invalid/illegal for the associated setter
    *          method.
    * @throws  NoSuchMethodException
    *          if a DMO setter method cannot be found.
    * @throws  InstantiationException
    *          if a DMO or data wrapper class cannot be instantiated.
    * @throws  InvocationTargetException
    *          if the underlying methods called with reflection throw an exception.
    * @throws  IllegalAccessException
    *          if a method called with reflection cannot be accessed.
    */
   private static Persistable prepareRecord(Element record, String schema)
   throws PersistenceException,
          NoSuchMethodException,
          InstantiationException,
          InvocationTargetException,
          IllegalAccessException
   {
      String recName = record.getAttribute("name");
      Helper h = helpers.get(recName);
      if (h == null)
      {
         throw new TypeNotPresentException(recName, null);
      }
      
      // prepare intercept _file._file-name _file._template
      boolean isFile = "File".equals(recName);
      String legacyName = null;
      String className = null;
      Long templateId = null;
      
      // instantiate DMO
      Persistable dmo = (Persistable) h.dmoImpl.newInstance();
      
      NodeList fldNodes = record.getElementsByTagName("field");
      int numFlds = fldNodes.getLength();
      for (int j = 0; j < numFlds; j++)
      {
         Element field = (Element) fldNodes.item(j);
         String fldName = field.getAttribute("name");
         String type = field.getAttribute("type");
         String value = field.getAttribute("value");
         String index = field.getAttribute("index");
         
         Constructor<?> ctor = ctorMap.get(type);
         if (ctor == null)
         {
            Class<?> argClass = fieldClassMap.get(type);
            if (argClass == null)
            {
               throw new TypeNotPresentException(type, null);
            }
            ctor = argClass.getConstructor(String.class);
            ctorMap.put(type, ctor);
         }
         Object arg = ctor.newInstance(value);
         
         try
         {
            // set DMO property
            h.setValue(dmo, fldName, index, arg);
         }
         catch (IllegalArgumentException exc)
         {
            String msg = String.format("Error setting metadata dmo property %s#%s(%s, %s)",
                                       h.dmoImpl.getName(),
                                       fldName,
                                       index,
                                       arg.toString());
            throw new PersistenceException(msg, exc);
         }
         
         // intercept _file._file-name _file._template values
         if (isFile)
         {
            if ("newName".equals(fldName))
            {
               className = value;
            }
            else if ("fileName".equals(fldName))
            {
               legacyName = value;
            }
            else if ("template".equals(fldName))
            {
               templateId = Long.parseLong(value); // a negative long value is expected in P2J
            }
         }
      }
      
      // intercept _file._file-name _file._template processing and add to template cache:
      if (isFile && className != null && legacyName != null && templateId != null)
      {
         String dmoClassName = String.format("%s/%s/impl/%sImpl", baseDir, schema, className);
         Class<?> dmoClass = null;
         try
         {
            dmoClass = Class.forName(dmoClassName.replace('/', '.'));
            templateRecords.put(dmoClass, templateId);
         }
         catch (ClassNotFoundException e)
         {
            // table not found ?
         }
      }
      
      return dmo;
   }
   
   /**
    * Execute a script containing DDL SQL statements.
    * 
    * @param   database
    *          Target database.
    * @param   connection
    *          Connection to target database.
    * @param   script
    *          Short name of script file, which is expected to be found in the application jar
    *          file. The following path is implied:
    *          <pre>
    *          {base DMO package}._meta.{script name}
    *          </pre>
    * 
    * @throws  PersistenceException
    *          if there is any error connecting to the target database or executing the DDL
    *          statements.
    */
   private static void executeDDLScript(Database database, Connection connection, String script)
   throws PersistenceException
   {
      String path = String.format("%s/%s/%s", baseDir, META_SCHEMA, script);
      BufferedReader br = null;
      
      try
      {
         InputStream in = ClassLoader.getSystemResourceAsStream(path);
         if (in == null)
         {
            throw new MissingDataException("Could not load DDL script " + path);
         }
         
         br = new BufferedReader(new InputStreamReader(in));
         H2Helper.runScript(connection, br);
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Error executing DDL for " + database, exc);
      }
      finally
      {
         if (br != null)
         {
            try
            {
               br.close();
            }
            catch (IOException exc)
            {
            }
         }
      }
   }
   
   /**
    * Helper class for calling DMO setter methods when populating a metaschema record.
    */
   private static class Helper
   {
      /** Well-known name of primary key setter method */
      private static final String SET_ID = "setId";
      
      /** Primary key setter method */
      private static final Method idSetter;
      
      /** DMO implementation class */
      private final Class<?> dmoImpl;
      
      /** Map of DMO property setters (excluding primary key setter) */
      private final Map<String, Method> setters;
      
      static
      {
         try
         {
            idSetter = Persistable.class.getDeclaredMethod(SET_ID, Long.class);
         }
         catch (NoSuchMethodException | SecurityException e)
         {
            throw new RuntimeException(e);
         }
      }
      
      /**
       * Constructor.
       * 
       * @param   dmoIface
       *          DMO interface.
       * @param   dmoImpl
       *          DMO implementation class.
       */
      Helper(Class<?> dmoIface, Class<?> dmoImpl)
      {
         this.dmoImpl = dmoImpl;
         this.setters = PropertyHelper.pojoSettersByProperty(dmoIface);
      }
      
      /**
       * Set a property value into a DMO instance.
       * 
       * @param   dmo
       *          Target DMO instance.
       * @param   property
       *          Name of property to be set.
       * @param   index
       *          Optional string representation of index (for extent fields only). Must be the
       *          empty string if property does not represent an extent field. May not be
       *          <code>null</code>.
       * @param   value
       *          Value of the appropriate data type (i.e., data wrapper instance) for the target
       *          property.
       * 
       * @throws  PersistenceException
       *          if <code>index</code> cannot be parsed into an <code>Integer</code>.
       * @throws  IllegalAccessException
       *          if the target setter method cannot be accessed.
       * @throws  IllegalArgumentException
       *          if <code>value</code> is the wrong type.
       * @throws  InvocationTargetException
       *          if the underlying setter method throws an exception.
       */
      void setValue(Object dmo, String property, String index, Object value)
      throws PersistenceException,
             IllegalAccessException,
             IllegalArgumentException,
             InvocationTargetException
      {
         Method m = getSetter(property);
         
         if ("".equals(index))
         {
            m.invoke(dmo, value);
         }
         else
         {
            try
            {
               Integer i = new Integer(index);
               m.invoke(dmo, i, value);
            }
            catch (NumberFormatException exc)
            {
               throw new IllegalArgumentException(
                  "Invalid DMO property setter index: " + index, exc);
            }
         }
      }
      
      /**
       * Retrieve the DMO property setter method for the specified property name.
       * 
       * @param   property
       *          Name of a DMO property (case-sensitive).
       * 
       * @return  Matching setter method.
       * 
       * @throws  PersistenceException
       *          if a setter method cannot be found for the specified property.
       */
      private Method getSetter(String property)
      throws PersistenceException
      {
         Method m = setters.get(property);
         if (m == null && DatabaseManager.PRIMARY_KEY.equals(property))
         {
            m = idSetter;
         }
         
         if (m == null)
         {
            String msg = String.format("No setter method found for %s#%s",
                                       dmoImpl.getName(),
                                       property);
            throw new PersistenceException(msg);
         }
         
         return m;
      }
   }
}