/*
** Module   : DecimalUserType.java
** Abstract : Custom user type to wrap P2J decimal for Hibernate access
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050630   @21591 Created initial version. Maps P2J decimal to
**                           SQL numeric.
** 002 ECF 20050729   @21954 Refactored to move nullSafeSet method
**                           implementation up to superclass. This was
**                           necessary to add centralized logging for
**                           prepared statement parameter binding.
** 003 ECF 20050809   @22074 Fixed nullSafeGet and added logging. Defect
**                           was that method was checking whether the
**                           result set column was null *before* reading
**                           data at that column. Also made class final.
** 004 ECF 20060907   @29317 Fixed ClassCastException in bindNonNull() and
**                           deepCopy() methods. Must cast to NumberType
**                           instead of decimal, and use doubleValue()
**                           instead of getValue().
** 005 ECF 20060907   @29331 Rewrote deepCopy(). Needed to handle unknown
**                           value safely and account for precision.
** 006 ECF 20070608   @34013 Replaced double with BigDecimal. Internal
**                           implementation of decimal now uses the latter
**                           instead of the former.
** 007 ECF 20080806   @39297 Moved to type sub-package of persist.
** 008 SVL 20120331          Upgraded to Hibernate 4.
** 009 ECF 20140220          Replaced Apache commons logging with J2SE standard logging.
** 010 ECF 20140624          Refactored SQL_TYPES to use TypeHelper.
** 011 ECF 20160617          Javadoc and format cleanup.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.math.*;
import java.sql.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;
import org.hibernate.engine.spi.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.decimal P2J decimal}. This permits Hibernate to
 * convert this data type to and from the SQL NUMERIC data type. The unknown
 * value is mapped to SQL NULL.
 */
public final class DecimalUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { TypeHelper.getSqlType(decimal.class) };
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a decimal to/from a database
    * record.
    *
    * @return  Array containing SQL type constant.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.decimal P2J decimal} class.
    */
   public Class returnedClass()
   {
      return decimal.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>decimal</code> instance.
    *
    * @param   value
    *          The <code>decimal</code> instance to be copied.
    *
    * @return  A new instance of <code>decimal</code> which contains the
    *          same internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      if (value instanceof decimal)
      {
         return (new decimal((decimal) value));
      }
      
      NumberType num = (NumberType) value;
      if (num.isUnknown())
      {
         return (new decimal());
      }
      
      return (new decimal(num.longValue()));
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return a <code>decimal</code>
    * instance which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>BigDecimal</code>.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   session
    *          Not used.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>decimal</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs,
                             String[] names,
                             SessionImplementor session,
                             Object owner)
   throws SQLException
   {
      BigDecimal value = rs.getBigDecimal(names[0]);
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "returning unknown value as column: " + names[0]);
         }
         return new decimal();
      }
      
      decimal wrapper = new decimal(value);
      if (LOG_TRACE)
      {
         LOG.log(Level.FINEST, "returning '" + toString(wrapper) + "' as column: " + names[0]);
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>BigDecimal</code> value into the given prepared statement,
    * based upon the <code>decimal</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>decimal</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps, BaseDataType value, int index)
   throws SQLException
   {
      decimal d = (value instanceof decimal
                   ? (decimal) value
                   : new decimal(((NumberType) value).longValue()));
      
      ps.setBigDecimal(index, d.toJavaType());
   }
}