/*
** Module   : AutoExport.java
** Abstract : Thread meant to run as a shutdown hook to persist user and configuration data.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 ECF 20170622 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.report.server;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import com.goldencode.p2j.report.*;

/**
 * Automatic export of user-defined report definitions and analytics configuration to XML.
 * This class is meant to run as a JVM shutdown hook. It shuts down the database as its last
 * action.
 */
public final class AutoExport
extends Thread
implements ReportConstants,
           XmlConstants
{
   /**
    * Export user-defined report definitions and analytics configuration to XML files.
    * 
    * @throws  RuntimeException
    *          if there is an error shutting down the database.
    */
   public void run()
   {
      ConfigurationPersistence cfgPers = new ConfigurationPersistence();
      
      try
      {
         // export user credentials
         List<UserRecord> users = ReportApi.getUserRecords();
         if (!users.isEmpty())
         {
            String path = DatabaseService.p2jHomePath + AUTO_EXPORT_USER_PATH;
            File file = new File(path);
            cfgPers.writeXmlUsers(file, users);
         }
         
         // export user-defined reports
         List<ReportDefinition> reports;
         
         // export user-defined schema report definitions, if any
         reports = ReportApi.getUserReports(SRC_SCHEMA);
         if (!reports.isEmpty())
         {
            String path = DatabaseService.p2jHomePath +
                          String.format(AUTO_EXPORT_RPT_SPEC, "schema");
            File file = new File(path);
            cfgPers.writeXmlReports(file, reports, true);
         }
         
         // export user-defined code report definitions, if any
         reports = ReportApi.getUserReports(SRC_CODE_CACHE);
         if (!reports.isEmpty())
         {
            String path = DatabaseService.p2jHomePath +
                          String.format(AUTO_EXPORT_RPT_SPEC, "code");
            File file = new File(path);
            cfgPers.writeXmlReports(file, reports, false);
         }
         
         // export non-default filter profiles
         List<FilterProfile> filters = ReportApi.getNonDefaultFilters();
         if (!filters.isEmpty())
         {
            String path = DatabaseService.p2jHomePath + AUTO_EXPORT_FILTER_PATH;
            File file = new File(path);
            cfgPers.writeXmlFilters(file, filters);
         }
         
         // export users' search history
         List<SearchHistory> searches = ReportApi.getUserSearchHistory();
         if (!searches.isEmpty())
         {
            String path = DatabaseService.p2jHomePath + AUTO_EXPORT_SRCH_PATH;
            File file = new File(path);
            cfgPers.writeXmlSearches(file, searches);
         }
      }
      catch (ParserConfigurationException | IOException exc)
      {
         exc.printStackTrace();
      }
      finally
      {
         // explicitly shut down the database, since we have disabled automatic shutdown on exit
         ReportApi.shutdownDatabase();
      }
   }
}
