/*
** Module   : event.js
** Abstract : Event processing functionality.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170531 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Very simple event processing service which routes events to listener callbacks immediately
 * (no queue). Events are dispatched to listeners in the order the listeners were added.
 * 
 */
define(function()
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Map of event listeners currently registered */
   var listeners = new Map();
   
   /**
    * Register a callback function to be invoked upon the occurrence of a named event. The
    * callback will be invoked synchronously when the event occurs, so it must execute quickly
    * to avoid suspending user interface processing.
    * 
    * @param   {String} eventName
    *          Name of the event to which the callback is registered.
    * @param   {Function} callback
    *          Callback function to invoke when the named event is fired.
    */
   me.addListener = function(eventName, callback)
   {
      var callbacks = listeners.get(eventName);
      if (!callbacks)
      {
         callbacks = [];
         listeners.set(eventName, callbacks);
      }
      
      if (callbacks.indexOf(callback) == -1)
      {
         callbacks.push(callback);
      }
   }
   
   /**
    * Deregister a callback function which previously was registered to a named event.
    * 
    * @param   {String} eventName
    *          Name of the event to which the callback is registered.
    * @param   {Function} callback
    *          Callback function to be deregistered.
    */
   me.removeListener = function(eventName, callback)
   {
      var callbacks = listeners.get(eventName);
      if (!callbacks)
      {
         return false;
      }
      
      var i = callbacks.indexOf(callback);
      if (i >= 0)
      {
         callbacks.splice(i, 1);
         if (callbacks.length == 0)
         {
            listeners.delete(eventName);
         }
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Fire a named event, invoking all registered callbacks synchronously.
    * 
    * @param   {String} name
    *          Name of the event being fired.
    * @param   {Object} source
    *          The object which represents the source of the event.
    * @param   {Object} data
    *          Optional data to be passed to registered callbacks.
    */
   me.fireEvent = function(name, source, data)
   {
      var callbacks = listeners.get(name);
      if (callbacks)
      {
         for (var func in callbacks)
         {
            try
            {
               callbacks[func].call(source, data);
            }
            catch (err)
            {
               console.log(err);
            }
         }
      }
   }
   
   /* exports */
   return me;
   
});