/*
** Module   : wsbase.js
** Abstract : Low level WebSocket functionality.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170428 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Module to establish a WebSocket connection with the server and process incoming and outgoing
 * messages via an RPC protocol. Different message types mirror the types of APIs on the server.
 * Each message is assigned a unique ID. At the time a message is sent to the server, a callback
 * is registered under that ID. The callback is invoked when the user sends a response message
 * with that ID in its header. The first parameter to each callback function is an error object,
 * which represents a serialized Java exception, if there was an error processing the API request
 * on the server side. If there was no error, this first parameter will be null. Message payloads
 * are serialized to JSON in both directions. The primary client API for use by other JavaScript
 * modules is this module's sendMessage function.
 */
define(["dojo/keys", "dijit/form/TextBox", "dojo/domReady!"], function(keys, TextBox)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Web socket used to communicate with server */
   var socket = null;
   
   /** Base value for message types */
   var MSG_BASE = 0;
   
   // base RPC message type for main report API
   me.MSG_RPT_BASE = MSG_BASE + 1000;
   
   // base RPC message type for callgraph API
   me.MSG_CG_BASE = MSG_BASE + 2000;
   
   // base RPC message type for AST API
   me.MSG_AST_BASE = MSG_BASE + 3000;
   
   // base RPC message type for file filter API
   me.MSG_FILT_BASE = MSG_BASE + 4000;
   
   // base RPC message type for source listing API
   me.MSG_LSTG_BASE = MSG_BASE + 5000;
   
   // message types used within this module
   var msgTypes =
   {
      AUTH              :  MSG_BASE + 1,
      CLOSE_NORMAL      :  MSG_BASE + 2,
      CHANGE_PASS       :  MSG_BASE + 3,
   }
   
   // next unique message ID
   var nextMessageId = 0;
   
   // map of active callbacks; elements are removed as responses are returned from the server
   var callbacks = new Map();
   
   /**
    * Initialize this module by creating and connecting a WebSocket to the server.
    * 
    * @param   {Function} onSuccess
    *          Callback to invoke if web socket connects to server successfully.
    */
   me.init = function(onSuccess)
   {
      var url = "wss://" + window.location.host + "/api";
      
      socket = new WebSocket(url);
      
      socket.onopen = function()
      {
         console.log("socket opened");
         onSuccess();
      }
      
      socket.onerror = function()
      {
         alert("The current session closed with a websocket error!");
         socket = null;
      }
      
      socket.onclose = function(evt)
      {
         var msg = "socket closed (" + evt.code + ")";
         if (evt.reason != null && evt.reason.length > 0)
         {
            msg += " - " + evt.reason;
         }
         console.log(msg);
         socket = null;
      }
      
      socket.onmessage = function(evt)
      {
         var msg = JSON.parse(evt.data);
         
         processMessage(msg);
      }
   }
   
   /**
    * Process an incoming message.
    */
   function processMessage(msg)
   {
      var callback = callbacks.get(msg.id);
      if (callback != null)
      {
         callbacks.delete(msg.id);
         
         if (msg.success)
         {
            callback(null, msg.pay);
         }
         else
         {
            callback(msg.pay);
         }
      }
      else
      {
         // TODO: implement server-originated requests
         alert("no callback registered");
      }
   }
   
   /**
    * Send a message across the web socket to the server.
    * 
    * @param   {Number} type
    *          Message type.
    * @param   {Object} payload
    *          Message payload, which is dependent upon the message type.
    * @param   {Function} callback
    *          Callback function to invoke when the server sends a response to the message.
    */
   me.sendMessage = function(type, payload, callback)
   {
      var msg = new Object();
      
      msg.id   = ++nextMessageId;
      msg.type = type;
      msg.pay  = payload;
      
      callbacks.set(msg.id, function(err, data) { callback(err, data); });
      
      socket.send(JSON.stringify(msg));
   }
   
   /* Server API */
   
   /**
    * Authenticate a user with the server.
    * 
    * @param   {String} user
    *          User name.
    * @param   {String} pass
    *          User password.
    * @param   {Function} callback
    *          Callback function to invoke when the server sends a response to the authentication
    *          request.
    */
   me.authenticate = function(user, pass, callback)
   {
      me.sendMessage(msgTypes.AUTH, [ user, pass ], callback);
   }
   
   /**
    * Prompt the user to change their password. Execute a callback in the event the user either
    * succeeds in doing so, or cancels the prompt. The prompt is recursive if the password change
    * fails. If the user cancels the prompt, the password change does not occur, and the
    * condition which led to the prompt (an expired password) will re-occur upon the next login.
    * 
    * @param   {String} user
    *          User name.
    * @param   {Function} callback
    *          Callback to be executed in the event the password change succeeds or is cancelled.
    */
   me.changePassword = function(user, callback)
   {
      var dialogId = "dialog_Change_Password";
      
      var dialog;
      
      // form content for the prompt
      var content =
         "<table>" +
         "  <tr>" +
         "    <td><label for='oldpw'>Old Password:</label></td>" +
         "    <td><input data-dojo-type='dijit/form/TextBox' " +
         "               type='password' name='oldpw' id='oldpw'></td>" +
         "  </tr>" +
         "  <tr>" +
         "    <td><label for='newpw1'>New Password:</label></td>" +
         "    <td><input data-dojo-type='dijit/form/TextBox' " +
         "               type='password' name='newpw1' id='newpw1'></td>" +
         "  </tr>" +
         "  <tr>" +
         "    <td><label for='newpw2'>Re-enter New Password:</label></td>" +
         "    <td><input data-dojo-type='dijit/form/TextBox' " +
         "               type='password' name='newpw2' id='newpw2'></td>" +
         "  </tr>" +
         "</table>";
      
      // execute callback - function which invokes the server's password change API
      var rpc = function()
      {
         var oldPass = dijit.byId("oldpw").get("value");
         var newPass1 = dijit.byId("newpw1").get("value");
         var newPass2 = dijit.byId("newpw2").get("value");
         
         me.sendMessage(msgTypes.CHANGE_PASS, [ user, oldPass, newPass1, newPass2 ], function(err)
         {
            if (err)
            {
               prompt(fwd.errorMessages(err)[0]);
            }
            else
            {
               callback();
            }
         });
      }
      
      // function to prompt the user; this is called recursively if the server password change
      // API returns an error
      var prompt = function(msg)
      {
         var cont = content;
         
         if (msg)
         {
            cont = "<div style='font-weight: bold; padding-bottom: 0.5em; color: red'>" +
                   msg + "</div>" + content;
         }
         
         var title = "Please Change Your Password";
         
         var listeners = new Map();
         
         listeners.set("execute", rpc);
         
         listeners.set("cancel", callback);
         
         listeners.set("show", function()
         {
            var widgets = [];
            
            var oldPassWidget  = dijit.byId("oldpw");
            
            widgets.push(oldPassWidget);
            
            var newPass1Widget = dijit.byId("newpw1");
            
            widgets.push(newPass1Widget);
            
            var newPass2Widget = dijit.byId("newpw2");
            
            widgets.push(newPass2Widget);
            
            var keyDownHandler = function(widget)
            {
               var index = widgets.indexOf(widget);
               
               return function(event)
               {
                  switch(event.keyCode)
                  {
                     case keys.UP_ARROW:
                        if (index > 0)
                        {
                           widgets[index - 1].focus();
                        }
                        break;
                     case keys.DOWN_ARROW:
                        if (index < widgets.length - 1)
                        {
                           widgets[index + 1].focus();
                        }
                        break;
                     case keys.HOME:
                        widgets[0].focus();
                        break;
                     case keys.END:
                        widgets[widgets.length - 1].focus();
                        break;
                     case keys.ENTER:
                        if (index < widgets.length - 1)
                        {
                           widgets[index + 1].focus();
                        }
                        else
                        {
                           var handle = dialog.on("hide", function()
                           {
                              handle.remove();
                              rpc();
                           })
                           dialog.hide();
                        }
                        break;
                     default:
                  }
               };
            };
            
            widgets.forEach(function(widget)
            {
               widget.on("keydown", keyDownHandler(widget)); 
            });
            
         });
         
         // create dialog
         dialog = fwd.createConfirmDialog(dialogId, title, cont, listeners);
         
         // show the dialog, executing the password change API on 'OK', the callback on 'Cancel'
         dialog.show();
      };
      
      // start the prompt
      prompt();
   }
   
   /**
    * Close the web socket to end communication with the server.
    * 
    * @param   {Function} callback
    *          Callback function to invoke when the server sends a response to the message.
    */
   me.close = function(callback)
   {
      me.sendMessage(msgTypes.CLOSE_NORMAL, null, function(err)
      {
         callback(err);
         socket.close();
         socket = null;
      });
   }
   
   /* exports */
   return me;
});
