/*
** Module   : BrowseJasperReportTemplate.java
** Abstract : Template for a browse Jasper report design, used by a browse.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180207 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

import com.goldencode.ast.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.Browse;
import com.goldencode.p2j.xml.*;
import com.goldencode.util.*;
import net.sf.jasperreports.engine.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import java.io.*;
import java.util.*;

import static com.goldencode.p2j.xml.XmlAst.domToAst;
import static com.goldencode.p2j.xml.XmlTokenTypes.*;

/**
 * Template for a Jasper report design, used by a browse. First, user creates a Jasper report
 * design. This design is called "template" because basing on this design a design applicable to
 * the given browse is created (design should have the matching number of columns, same width,
 * alignment and data type of columns). The following transformations is used to transform a
 * design template into the design for the given browse:
 * <ul>
 *    <li>Only the first column in the design template is taken into consideration. It is
 *    duplicated to form other columns.</li>
 *    <li>Column headers and details bands are widened to the page width (excluding margins)</li>
 *    <li>Widths of columns are set in the design in the same proportion as in the browse.</li>
 *    <li>Text alignment of a column (including column label) is set as in the browse.</li>
 *    <li>Label of a column is set as in the browse.</li>
 * </ul>
 * Transformations are performed in two steps: {@link #prepare(String)} function which loads
 * template into XML tree and performs preparation steps, and
 * {@link #generateReport(BrowseWidget)} function which transforms the XML tree into report
 * applicable to the specific browse and compiles the report.
 */
class BrowseJasperReportTemplate
{
   /** Root node of the prepared template. */
   private XmlAst templateRoot;

   /** Sub-tree representing template for a column header. */
   private XmlAst columnHeaderTemplate;

   /** Sub-tree representing template for a column data area. */
   private XmlAst columnDataTemplate;

   /** Name of the design template file. */
   private String designTemplateFile;

   /** Page width specified in the report. */
   private Integer pageWidth;

   /**
    * Load given design template into XML tree and perform preparation steps, which in the future
    * allow to quickly transform this tree into a report suitable for the specific browse.
    *
    * @param  designTemplateFile
    *         Name of the design template file.
    *
    * @throws JasperReportException
    *         if design template file could not be read or parsed or has invalid structure.
    */
   void prepare(String designTemplateFile)
   throws JasperReportException
   {
      InputStream stream =
            FwdJasperExtensionRegistry.getRepositoryService().getInputStream(designTemplateFile);

      Document dom;
      try
      {
         dom = XmlHelper.parse(stream);
      }
      catch (Exception e)
      {
         throw new JasperReportException("Could not parse browse report template file",
                                         designTemplateFile, e);
      }
      templateRoot = domToAst(dom);
      this.designTemplateFile = designTemplateFile;

      XmlAst jasperReportsRoot = findChild(templateRoot, "jasperReport");
      cleanUUID(jasperReportsRoot);

      // get page width
      pageWidth = getAttributeInt(jasperReportsRoot, "pageWidth", 0);
      pageWidth -= getAttributeInt(jasperReportsRoot, "leftMargin", 0);
      pageWidth -= getAttributeInt(jasperReportsRoot, "rightMargin", 0);

      if (pageWidth <= 0)
      {
         throw new JasperReportException("Cannot calculate page width", designTemplateFile);
      }

      // Remove FIELD nodes
      List<XmlAst> fields = findChildren(jasperReportsRoot, "field");
      fields.forEach(AnnotatedAst::remove);

      // Handle column headers (columnHeader/band nodes)
      XmlAst columnHeadersRoot = findChildPath(jasperReportsRoot, "columnHeader/band");
      List<XmlAst> headers = findChildren(columnHeadersRoot, "staticText");

      // Create template for column headers
      columnHeaderTemplate = (XmlAst) headers.get(0).duplicateFresh();

      // Remove original header nodes
      headers.forEach(AnnotatedAst::remove);

      // Handle column data (detail/band nodes)
      XmlAst columnsDataRoot = findChildPath(jasperReportsRoot, "detail/band");
      List<XmlAst> columns = findChildren(columnsDataRoot, "textField");

      // Create template for column data
      columnDataTemplate = (XmlAst) columns.get(0).duplicateFresh();

      // Remove original column data nodes
      columns.forEach(AnnotatedAst::remove);
   }

   /**
    * Transform the XML tree representing prepared template ({@link #templateRoot} field) into the
    * report applicable to the specific browse and compile it.
    *
    * @param  browse
    *         Browse for which the report is generated.
    *
    * @return compiled report applicable to the specific browse.
    *
    * @throws JasperReportException
    *         if there was an error during creation of the browse report or the report could not
    *         be compiled.
    */
   JasperReport generateReport(BrowseWidget browse)
   throws JasperReportException
   {
      XmlAst resultRoot = (XmlAst) templateRoot.duplicateFresh();
      XmlAst fieldsRoot = findChild(resultRoot, "jasperReport");
      XmlAst fieldsAnchor = findChild(fieldsRoot, "queryString");

      XmlAst columnHeadersRoot =
            findChildPath(resultRoot, "jasperReport/columnHeader/band");
      XmlAst columnDataRoot =
            findChildPath(resultRoot, "jasperReport/detail/band");

      // calc total columns width
      int colNo = browse.getNumColumns().toJavaIntegerType();
      if (colNo == 0)
      {
         throw new JasperReportException("Browse has no columns", designTemplateFile);
      }

      // calculate report columns' widths: widths should be in the same proportion as the browse
      // columns' widths
      int totalBrowseColumnsWidth = 0;
      int[] columnWidths = new int[colNo];
      for (int i = 1; i <= colNo; i++)
      {
         BrowseColumnWidget column =
               (BrowseColumnWidget) browse.getBrowseColumn(i).unwrapBrowseColumn();

         int columnWidth = column.getWidthPixels().toJavaIntegerType();
         columnWidths[i - 1] = columnWidth;
         totalBrowseColumnsWidth += columnWidth;
      }

      int totalReportColumnsWidth = 0;
      for (int i = 0; i < colNo - 1; i++)
      {
         columnWidths[i] = pageWidth * columnWidths[i] / totalBrowseColumnsWidth;
         totalReportColumnsWidth += columnWidths[i];
      }
      // calculate width of the last column separately
      columnWidths[colNo - 1] = pageWidth - totalReportColumnsWidth;

      int xOffset = 0;
      for (int i = 1; i <= colNo; i++)
      {
         BrowseColumnWidget column =
               (BrowseColumnWidget) browse.getBrowseColumn(i).unwrapBrowseColumn();
         int columnWidth = columnWidths[i - 1];

         // Create field node
         XmlAst fieldNode = new XmlAst(ELEMENT_NODE, "field");
         String columnVarName = "column" + i;
         setAttribute(fieldNode, "name", columnVarName);
         setAttribute(fieldNode, "class", "java.lang.String");
         fieldsRoot.graftAt(fieldNode, fieldsAnchor.getIndexPos() + 1);

         boolean rightAligned = Browse.isRightAligned(column.config());

         // Create column header
         XmlAst columnHeaderNode = (XmlAst) columnHeaderTemplate.duplicateFresh();
         columnHeadersRoot.graft(columnHeaderNode);

         // Set label text
         String[] labelLines = Browse.getLabelVar(column.config());
         StringBuilder sb = new StringBuilder();
         for (int j = 0; j < labelLines.length; j++)
         {
            sb.append(labelLines[j]);
            if (j > 0)
            {
               sb.append("\n");
            }
         }

         XmlAst node = findChild(columnHeaderNode, "text");
         node = (XmlAst) node.getImmediateChild(CDATA_SECTION_NODE, null);
         node = (XmlAst) node.getImmediateChild(CONTENT, null);
         node.setText(sb.toString());

         setPositionAndWidth(columnHeaderNode, xOffset, columnWidth);
         setTextAlignment(columnHeaderNode, rightAligned);

         // Create column data
         XmlAst columnDataNode = (XmlAst) columnDataTemplate.duplicateFresh();
         columnDataRoot.graft(columnDataNode);

         String content = String.format("$F{%s}", columnVarName);
         node = findChild(columnDataNode, "textFieldExpression");
         node = (XmlAst) node.getImmediateChild(CDATA_SECTION_NODE, null);
         node = (XmlAst) node.getImmediateChild(CONTENT, null);
         node.setText(content);

         setPositionAndWidth(columnDataNode, xOffset, columnWidth);
         setTextAlignment(columnDataNode, rightAligned);

         xOffset += columnWidth;
      }

      ByteArrayInputStream inputStream;
      try
      {
         String namespaceURI = getAttribute(fieldsRoot, "xmlns");
         Document doc = XmlAst.astToDom(resultRoot, namespaceURI);
         ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
         XmlHelper.write(doc, outputStream);
         //DEBUG: XmlHelper.write(doc, "generated-report.jrxml");
         inputStream = new ByteArrayInputStream(outputStream.toByteArray());
      }
      catch (ParserConfigurationException | IOException e)
      {
         throw new JasperReportException("Cannot create DOM from XML", designTemplateFile, e);
      }

      try
      {
         return JasperCompileManager.compileReport(inputStream);
      }
      catch (JRException e)
      {
         throw new JasperReportException("Cannot compile generated browse report",
                                         designTemplateFile, e);
      }
   }

   /**
    * Set position and width of the single <code>reportElement</code> element under the given
    * parent node.
    *
    * @param parent
    *        Parent node.
    * @param xPos
    *        Element X position.
    * @param columnWidth
 *           Element width.
    */
   private void setPositionAndWidth(XmlAst parent, int xPos, int columnWidth)
   {
      XmlAst element = findChild(parent, "reportElement");
      setAttribute(element, "width", Integer.toString(columnWidth));
      setAttribute(element, "x",     Integer.toString(xPos));
   }

   /**
    * Set text alignment of the single <code>textElement</code> element under the given
    * parent node.
    *
    * @param parent
    *        Parent node.
    * @param right
    *        <code>true</code> for right alignment, <code>false</code> for left alignment.
    */
   private void setTextAlignment(XmlAst parent, boolean right)
   {
      XmlAst element = findChild(parent, "textElement");
      setAttribute(element, "textAlignment", right ? "Right" : "Left");
   }

   /**
    * Find the single available path under the given parent node, and return the last node in this
    * path.
    *
    * @param  parent
    *         Parent node.
    * @param  path
    *         "/"-separated down path, e.g. "jasperReport/columnHeader/band". Elements of the path
    *         are XML tags.
    *
    * @return the last node in the path, if such path exists, or <code>null</code> if there is no
    *         path or there are multiple matching paths.
    */
   private XmlAst findChildPath(XmlAst parent, String path)
   {
      String[] tags = path.split("/");

      XmlAst res = parent;
      for (String tag : tags)
      {
         res = findChild(res, tag);
         if (res == null)
         {
            break;
         }
      }

      return res;
   }

   /**
    * Find the single available child with the specified XML tag under the given parent node.
    *
    * @param  parent
    *         Parent node.
    * @param  tag
    *         XML tag.
    *
    * @return the single available child with the specified XML tag.
    *
    * @throws JasperReportXmlSearchException
    *         if there is no available node, or there are multiple matching nodes.
    */
   private XmlAst findChild(XmlAst parent, String tag)
   {
      return findChild(parent, tag, false);
   }

   /**
    * Find the single available child with the specified XML tag under the given parent node.
    *
    * @param  parent
    *         Parent node.
    * @param  tag
    *         XML tag.
    *
    * @return the single available child with the specified XML tag. If <code>noError</code> is
    *         <code>true</code>: returns <code>null</code> if there is no available node, or
    *         there are multiple matching nodes.
    *
    * @throws JasperReportXmlSearchException
    *         if <code>noError</code> is <code>false</code>: there is no available node, or
    *         there are multiple matching nodes.
    */
   private XmlAst findChild(XmlAst parent, String tag, boolean noError)
   {
      return findChild(parent, ELEMENT_NODE, tag, noError);
   }

   /**
    * Find the single available child with the specified node type and text under the given parent
    * node.
    *
    * @param  parent
    *         Parent node.
    * @param  nodeType
    *         Node type (see constants in {@link XmlTokenTypes}).
    * @param  nodeText
    *         Node text (e.g. XML tag or attribute name).
    *
    * @return the single available child with the specified node type and text. If
    *         <code>noError</code> is <code>true</code>: returns <code>null</code> if there are no
    *         available node, or there are multiple matching nodes.
    *
    * @throws JasperReportXmlSearchException
    *         if <code>noError</code> is <code>false</code>: there are no available node, or
    *         there are multiple matching nodes.
    */
   private XmlAst findChild(XmlAst parent, int nodeType, String nodeText, boolean noError)
   {
      List<XmlAst> children = findChildren(parent, nodeType, nodeText, true);
      if (children == null || children.size() != 1)
      {
         if (noError)
         {
            return null;
         }
         else
         {
            throw new JasperReportXmlSearchException(parent, nodeText, designTemplateFile);
         }
      }

      return children.get(0);
   }

   /**
    * Find the child nodes with the specified XML tag under the given parent node.
    *
    * @param  parent
    *         Parent node.
    * @param  tag
    *         XML tag.
    *
    * @return available child nodes with the specified XML tag.
    *
    * @throws JasperReportXmlSearchException
    *         there are no available node.
    */
   private List<XmlAst> findChildren(XmlAst parent, String tag)
   {
      return findChildren(parent, ELEMENT_NODE, tag);
   }

   /**
    * Find the child nodes with the specified node type and text under the given parent node.
    *
    * @param  parent
    *         Parent node.
    * @param  nodeType
    *         Type of the target nodes (see constants in {@link XmlTokenTypes}).
    * @param  nodeText
    *         Node text (e.g. XML tag or attribute name).
    *
    * @return available child nodes with the specified node type and text.
    *
    * @throws JasperReportXmlSearchException
    *         there are no available node.
    */
   private List<XmlAst> findChildren(XmlAst parent, int nodeType, String nodeText)
   {
      return findChildren(parent, nodeType, nodeText, false);
   }

   /**
    * Find the child nodes with the specified node type and text under the given parent node.
    *
    * @param  parent
    *         Parent node.
    * @param  nodeType
    *         Type of the target nodes (see constants in {@link XmlTokenTypes}).
    * @param  nodeText
    *         Node text (e.g. XML tag or attribute name).
    *
    * @return available child nodes with the specified node type and text. If
    *         <code>noError</code> is <code>true</code>: returns <code>null</code> if there are no
    *         available nodes.
    *
    * @throws JasperReportXmlSearchException
    *         if <code>noError</code> is <code>false</code>: there are no available nodes.
    */
   private List<XmlAst> findChildren(XmlAst parent,
                                     int nodeType,
                                     String nodeText,
                                     boolean noError)
   {
      List<XmlAst> children = null;
      XmlAst child = null;
      while (true)
      {
         child = (XmlAst) parent.getImmediateChild(nodeType, child);
         if (child == null)
         {
            break;
         }

         if (Objects.equals(nodeText, child.getText()))
         {
            if (children == null)
            {
               children = new ArrayList<>();
            }
            children.add(child);
         }
      }

      if (children == null && !noError)
      {
         throw new JasperReportXmlSearchException(parent, nodeText, designTemplateFile);
      }

      return children;
   }

   /**
    * Remove all "uuid" attributes in the given tree.
    *
    * @param root
    *        Root of the tree.
    */
   private void cleanUUID(XmlAst root)
   {
      removeAttribute(root, "uuid");

      XmlAst child = null;
      while (true)
      {
         child = (XmlAst) root.getImmediateChild(ELEMENT_NODE, child);
         if (child == null)
         {
            return;
         }

         cleanUUID(child);
      }
   }

   /**
    * Get value of the specified integer attribute of the specified node.
    *
    * @param  node
    *         Parent node.
    * @param  name
    *         Attribute name.
    * @param  defaultValue
    *         Default value of the attribute.
    *
    * @return value of the attribute or the default value if there is no such attribute.
    */
   private int getAttributeInt(XmlAst node, String name, int defaultValue)
   {
      String val = getAttribute(node, name);
      if (val != null)
      {
         try
         {
            return Integer.valueOf(val);
         }
         catch (NumberFormatException e)
         {
            return defaultValue;
         }
      }
      else
      {
         return defaultValue;
      }
   }

   /**
    * Get value of the specified attribute of the specified node.
    *
    * @param  node
    *         Parent node.
    * @param  name
    *         Attribute name.
    *
    * @return value of the attribute or <code>null</code> if there is no such attribute.
    */
   private String getAttribute(XmlAst node, String name)
   {
      // Structure of an attributed node:
      //
      // <node> [ELEMENT_NODE]
      //    attributes [ATTR_SET]
      //       <name> [ATTRIBUTE_NODE]
      //          #text [TEXT_NODE]
      //             <value> [CONTENT]


      XmlAst attrSet = (XmlAst) node.getImmediateChild(ATTR_SET, null);
      if (attrSet == null)
      {
         return null;
      }

      XmlAst attrNode = findChild(attrSet, ATTRIBUTE_NODE, name, true);
      if (attrNode == null)
      {
         return null;
      }

      XmlAst textNode = (XmlAst) attrNode.getImmediateChild(TEXT_NODE, null);
      if (textNode == null)
      {
         return null;
      }

      XmlAst contentNode = (XmlAst) textNode.getImmediateChild(CONTENT, null);
      if (contentNode == null)
      {
         return null;
      }
      else
      {
         return contentNode.getText();
      }
   }

   /**
    * Delete the specified attribute of the specified node.
    *
    * @param  node
    *         Parent node.
    * @param  name
    *         Attribute name.
    */
   private void removeAttribute(XmlAst node, String name)
   {
      // Structure of an attributed node:
      //
      // <node> [ELEMENT_NODE]
      //    attributes [ATTR_SET]
      //       <name> [ATTRIBUTE_NODE]
      //          #text [TEXT_NODE]
      //             <value> [CONTENT]


      XmlAst attrSet = (XmlAst) node.getImmediateChild(ATTR_SET, null);
      if (attrSet == null)
      {
         return;
      }

      XmlAst attrNode = findChild(attrSet, ATTRIBUTE_NODE, name, true);
      if (attrNode != null)
      {
         attrNode.remove();
      }
   }

   /**
    * Set the value of the specified attribute.
    *
    * @param node
    *        Parent node.
    * @param name
    *        Attribute name.
    * @param value
    *        Attribute value to set.
    */
   private void setAttribute(XmlAst node, String name, String value)
   {
      // Structure of an attributed node:
      //
      // <node> [ELEMENT_NODE]
      //    attributes [ATTR_SET]
      //       <name> [ATTRIBUTE_NODE]
      //          #text [TEXT_NODE]
      //             <value> [CONTENT]


      XmlAst attrSet = (XmlAst) node.getImmediateChild(ATTR_SET, null);
      if (attrSet == null)
      {
         attrSet = new XmlAst(ATTR_SET, "attributes");
         node.graft(attrSet);
      }

      XmlAst attrNode = findChild(attrSet, ATTRIBUTE_NODE, name, true);
      if (attrNode == null)
      {
         attrNode = new XmlAst(ATTRIBUTE_NODE, name);
         attrSet.graft(attrNode);
      }

      XmlAst textNode = (XmlAst) attrNode.getImmediateChild(TEXT_NODE, null);
      if (textNode == null)
      {
         textNode = new XmlAst(TEXT_NODE, "#text");
         attrNode.graft(textNode);
      }

      XmlAst contentNode = (XmlAst) textNode.getImmediateChild(CONTENT, null);
      if (contentNode == null)
      {
         contentNode = new XmlAst(CONTENT, value);
         textNode.graft(contentNode);
      }
      else
      {
         contentNode.setText(value);
      }
   }
}
