/*
** Module   : JasperReportXmlSearchException.java
** Abstract : Exception thrown to indicate an error while searching in XML representing a Jasper
**            report design.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180207 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.reporting;

import com.goldencode.p2j.xml.XmlAst;

/**
 * This exception should be thrown to indicate an error while searching in XML representing a
 * Jasper report design.
 */
class JasperReportXmlSearchException
extends JasperReportException
{
   /**
    * Constructor which accepts searched node and report design file name.
    *
    * @param  parent
    *         Node which was supposed to be the parent of the node we are searching for.
    * @param  searchedNodeText
    *         Text which the node we were searching for was supposed to have.
    * @param  designFileName
    *         Name of the report design file processing of which caused the error.
    */
   JasperReportXmlSearchException(XmlAst parent, String searchedNodeText, String designFileName)
   {
      this(parent, searchedNodeText, designFileName, true);
   }

   /**
    * Constructor which accepts searched node and report design file name.
    *
    * @param  parent
    *         Node which was supposed to be the parent of the node we are searching for.
    * @param  searchedNodeText
    *         Text which the node we were searching for was supposed to have.
    * @param  designFileName
    *         Name of the report design file processing of which caused the error.
    * @param  node
    *         <code>true</code> if we were searching for a node, <code>false</code> if we were
    *         searching for an attribute.
    */
   JasperReportXmlSearchException(XmlAst parent,
                                  String searchedNodeText,
                                  String designFileName,
                                  boolean node)
   {
      super(String.format("Could not find '%s' %s",
                          getNodePath(parent, searchedNodeText),
                          node ? "node" : "attribute"),
            designFileName);
   }

   /**
    * Used for node search error text generation. Gets path from the root node to the requested
    * node as "/"-separated string. I.e.
    * <code>rootNode/node1/.../nodeN/parentNode/searchedNodeText</code>.
    *
    * @param  parent
    *         Node which was supposed to be the parent of the node we are searching for.
    * @param  searchedNodeText
    *         Text which the node we were searching for was supposed to have.
    *
    * @return path from the root node to the searched node as "/"-separated string
    */
   private static String getNodePath(XmlAst parent, String searchedNodeText)
   {
      String nodePath = searchedNodeText;
      XmlAst ast = parent;
      while (true)
      {
         String text = ast.getText();
         if (text == null)
         {
            text = "";
         }

         nodePath = text + "/" + nodePath;

         ast = (XmlAst) ast.getParent();
         if (ast == null)
         {
            break;
         }
      }

      return nodePath;
   }
}
