/*
** Module   : JobProcessor.java
** Abstract : Concrete job launcher, which listens for jobs posted via the timer thread and
**            executes them.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
** 002 CA  20140220 Fixed a bug in the job waiting loop - wait only if there are no pending jobs.
** 003 CA  20170228 Moved Scheduler.initialize to Scheduler.scheduleJobs.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.scheduler;

import java.util.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Specialized class which listens for incoming jobs (sent by the {@link Scheduler#timer}.  
 * This class runs in a separate thread, under the server's context.
 */
class JobProcessor
implements Runnable
{
   /** The list with the pending jobs. */
   private final LinkedList<Runnable> pendingJobs = new LinkedList<>();
   
   /**
    * Initialize the scheduler and schedule the initial jobs.  Then, start listening for new
    * jobs being posted to the {@link #pendingJobs queue} by the {@link Scheduler#timer}.
    */
   @Override
   public void run()
   {
      // before starting, wait for the server to start listening
      try
      {
         SessionManager.get().waitUntilReady();
      }
      catch (InterruptedException e)
      {
         return;
      }
      
      // add initial jobs
      synchronized (Scheduler.jobs)
      {
         for (Job job : Scheduler.jobs.values())
         {
            if (!job.isEnabled())
            {
               // skip disabled jobs
               continue;
            }
            
            Scheduler.scheduleJob(job.getName(), null);
         }
      }
      
      Runnable core = null;

      // listen for incoming jobs
      while (true)
      {
         synchronized (pendingJobs)
         {
            if (pendingJobs.isEmpty())
            {
               try
               {
                  pendingJobs.wait();
               }
               catch (InterruptedException e)
               {
                  // someone interrupted me, exit
                  break;
               }
            }
            
            core = (pendingJobs.isEmpty()) ? null : pendingJobs.removeFirst();
         }
         
         if (core != null)
         {
            // execute this in a different thread, but with the server's context
            Thread t = new AssociatedThread(core);
            String tname = LogHelper.createThreadName(String.format(core.toString()));
            t.setDaemon(true);
            t.setName(tname);
            t.start();
         }
      }
   }
   
   /**
    * Execute the given job.
    * 
    * @param    core
    *           The {@link Runnable} instance to execute the job.
    */
   public void executeJob(Runnable core)
   {
      synchronized (pendingJobs)
      {
         pendingJobs.addLast(core);
         pendingJobs.notify();
      }
   }
}