/*
** Module   : FieldInfo.java
** Abstract : Schema information about a database field.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20160328 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import com.goldencode.ast.*;

/**
 * A read-only bean which exposes schema-related information associated with a database field
 * which is referenced in source code.
 */
public class FieldInfo
{
   /** AST associated with the field */
   private final Aast ast;
   
   /** Fully qualified schema name of the field */
   private final String qualified;
   
   /** Field's buffer name as indicated in source */
   private final String buffer;
   
   /** Field's database logical name or alias */
   private final String database;
   
   /** Field's enclosing table's token type */
   private final int recordType;
   
   /**
    * Constructor.
    * 
    * @param   ast
    *          AST for the field.
    * @param   qualified
    *          Fully qualified schema name of the field
    * @param   buffer
    *          Name of the buffer indicated by the field reference in source code.
    * @param   database
    *          Name of the database which contains the field, as specified by the field reference
    *          in source code, or looked up if the source reference was unqualified.
    * @param   recordType
    *          Token type of the table backing the buffer indicated by the field reference in
    *          source code.
    */
   FieldInfo(Aast ast, String qualified, String buffer, String database, int recordType)
   {
      this.ast = ast;
      this.qualified = qualified;
      this.buffer = buffer;
      this.database = database;
      this.recordType = recordType;
   }
   
   /**
    * Get the AST which represents the field in the schema.
    * 
    * @return  Field AST.
    */
   public Aast getAst()
   {
      return ast;
   }
   
   /**
    * Get the fully qualified schema name of the field.
    * 
    * @return  Qualified field name.
    */
   public String getQualified()
   {
      return qualified;
   }
   
   /**
    * Get the buffer name associated with the field.
    * 
    * @return  Buffer name.
    */
   public String getBuffer()
   {
      return buffer;
   }
   
   /**
    * Get the logical database name or alias, if any, associated with the field.
    * 
    * @return  The logical name of the database associated with the buffer associated with the
    *          field. An empty string is returned if the containing table is a temp- or
    *          work-table.
    */
   public String getDatabase()
   {
      return database;
   }
   
   /**
    * Get the Progress parser token type of a table, temp-table, work-table, or buffer which
    * contains the field. If the record containing the field is a buffer, the type of the backing
    * record is returned. That is, a type of {@code BUFFER} is never returned, only the type of
    * the record which backs the buffer.
    * 
    * @return  Record token type.
    */
   public int getRecordType()
   {
      return recordType;
   }
}