/*
** Module   : schema.g
** Abstract : Progress database schema dump file (.df) parser
**
** Copyright (c) 2004-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041130   @18769 First version implementing a complete lexer,
**                           parser and tree walker that take a Progress
**                           schema dump (*.df file) on input and produce
**                           a hierarchy of Java objects which describe
**                           the schema.
** 002 ECF 20050101   @19158 Added ROWID and HANDLE field data types.
** 003 GES 20050204   @19585 Added support for standard meta-data in
**                           the lexer, parser and tree walker. New 
**                           keywords and data types were added.
** 004 GES 20050407   @20660 Total rewrite. Removal of lexer and tree
**                           walker. ProgressLexer was modified to handle
**                           (conditionally) schema-specific keywords and
**                           the parser was changed to use the same vocab
**                           as the ProgressParser. The parser was
**                           totally rewritten to be simpler and smaller
**                           and to create a tree that is compatible
**                           with the ProgressParser in the places that
**                           are overlapping.  Name change from data_df.g
**                           to schema.g
** 005 ECF 20050412   @20715 Added KW_TIME alternative to as_clause rule.
**                           Necessary for metadata. Cleaned up javadoc.
** 006 ECF 20050427   @20936 Fixed sqlWidth rule. No AST was being
**                           generated.
** 007 ECF 20050603   @21460 Strip extra quotes from schema area text and
**                           field label text.
** 008 ECF 20051017   @23049 Normalize case on all table, field, and index
**                           names. These are all set to lowercase.
** 009 ECF 20060115   @23941 Removed stripping of quotes from LABEL token.
**                           This was getting us into trouble with the UI
**                           conversion downstream.
** 010 ECF 20070219   @32178 Added placeholder rules for table triggers.
**                           These are dropped from the AST at this time,
**                           but the parser can handle table-trigger
**                           statements. Also added rules to parse (but
**                           ignore) some data server options.
** 011 GES 20070308   @32291 Added support for MAX-VAL in sequences. Added
**                           valExp support to tables. Added KW_NO_OVRRD
**                           to table trigger processing.
** 012 GES 20070418   @33113 Added support for COM-HANDLE type.
** 013 GES 20070629   @34330 Added line and column info to all artificial
**                           nodes. Added an option (CODEPAGE instead of
**                           CPSTREAM) which was found in client source.
** 014 GES 20070720   @34641 Add support for table permissions clauses.
**                           Improved the tree structure for table 
**                           triggers and now they are not dropped from
**                           the tree.
** 015 ECF 20070806   @34790 Fixed incorrect default value for logical in
**                           certain cases. When a logical field's format
**                           string was overridden to non-default string
**                           representations of true/false, we were always
**                           calculating a default of false, even when the
**                           true option was specified as the field's
**                           initial value.
** 016 GES 20070907   @35016 Support for 10.1B data types.
** 017 ECF 20070911   @35048 Added support for field triggers and more
**                           permission types. Added permissions as a
**                           valid field property.
** 018 GES 20071029   @35893 Added MAX-WIDTH support for fields. Fixed
**                           relexing to eliminate the use of the dot
**                           kludge.
** 019 GES 20090428   @42025 Moved command line driver code into main().
** 020 GES 20100220   @44693 JavaDoc update.
** 021 GES 20100809          Added rules to drop nonsense options that were
**                           found in real schemas. Added helpers to improve
**                           logging of warnings.
** 022 GES 20110623          Added support for parsing the INACTIVE keyword
**                           (whose meaning is unknown).
** 023 GES 20110831          Added support for field options CLOB-CODEPAGE,
**                           CLOB-COLLATION and CLOB-TYPE (meanings somewhat
**                           unknown).
** 024 CA  20140114          Save the legacy table name in an annotation, for later use.
** 025 VMN 20140120          Added annotation of original text of VALEXP.
** 026 ECF 20140125          Convert VALEXP original text to safe Java string before storing as
**                           original_text annotation.
** 027 GES 20140306          Match keyword processing changes for ascending which is a complete
**                           synonym for KW_ASC (ascii).
** 028 GES 20141202          Added UPDATE TABLE and UPDATE INDEX constructs (they are matched
**                           and dropped from the resulting tree). Added some optional trailer
**                           support for bufpool.
** 029 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 030 CA  20160721          Empty string initializer for numeric fields is emitted as 0.
** 031 GES 20170704          Added mergeUntilWS() and the any_ridiculous_symbol rule for matching
**                           symbols that start with numeric digits or other unexpected text.
** 032 GES 20180307          Add support for silently dealing with garbage initial date values.
**     HC  20180309          Added support for silently dealing with blank initial values for numeric and
**                           logical data types.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

header
{
/*
** Module   : SchemaParser.java
**            SchemaParserTokenTypes.java
** Abstract : Progress database schema dump file (.df) parser
**
** Copyright (c) 2004-2016, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041130   @18769 WARNING, THIS IS A GENERATED FILE!!! DO NOT EDIT THIS FILE. The
**                           original source file is schema.g!
*/ 

package com.goldencode.p2j.schema;

import java.io.*;
import java.util.*;
import antlr.*;
import antlr.debug.misc.ASTFrame;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
}

/**
 * Parser for a token stream generated by the <code>ProgressLexer</code>,
 * as part of the process of interpreting Progress schema dump (<code>*.df
 * </code>) files. Builds a normalized AST meant to be compatible in areas
 * which have similar options and structures as the AST created by the
 * <code>ProgressParser</code> from Progress 4GL source code.
 * <p>
 * <strong>NOTE:</strong>
 * DO NOT MAKE MODIFICATIONS TO THE SOURCE FILE FOR THIS CLASS DIRECTLY.
 * This class is generated by ANTLR, using the <code>schema.g</code> grammar.
 * All changes must be made at the grammar level.
 */
class SchemaParser extends Parser;

options
{
   buildAST = true;
   k = 2;
   ASTLabelType = "Aast";
   importVocab=ProgressParser;
}

{
   /** Name of the specialized AST class to use when parsing */
   public static final String AST_CLASS = ProgressAst.class.getName();
   
   /** Remember our platform so that we can properly parse strings containing escapes. */
   private static final boolean unixEsc = Configuration.getParameter("unix-escapes", true);
   
   /** Track the name of the table currently being parsed */
   private String currentTable = null;
                                                    
   /** Filter implementation to allow access to hidden tokens */
   private TokenStreamHiddenTokenFilter hidden = null;
   
   /**
    * Constructs a parser using a lexer instance.
    *
    * @param    lexer
    *           An instance of a <code>TokenStream</code> object.
    * @param    bogus
    *           Useless parameter which is only here to differentiate between
    *           this constructor and the standard one that ANTLR generates
    *           which doesn't have this parameter.
    */
   public SchemaParser(TokenStream lexer, boolean bogus)
   {
      // init our parent class
      this(lexer, 2);
      
      // shift into filter mode
      hidden = new TokenStreamHiddenTokenFilter(lexer);
      setTokenBuffer(new TokenBuffer(hidden));
      
      // setup the lexer
      ((ProgressLexer)lexer).setUnixEscapes(unixEsc);
      ((ProgressLexer)lexer).activateHiddenProcessing();
      setupHiddenFilter();
   }
   
   /**
    * Configure the hidden filter to ensure that token types that are not
    * normally recognized by the parser, are hidden (but still accessible)
    * using the filter.
    */
   private void setupHiddenFilter()
   {
      hidden.hide(WS);
      hidden.hide(TILDE);
      hidden.hide(COMMENT);
      hidden.hide(BACKSLASH);
   }
   
   /**
    * Helper to output a warning message with AST-specific type, line and
    * column data to STDERR.
    *
    * @param    node
    *           The AST to inspect for log data.
    * @param    front
    *           Log prefix.
    * @param    back
    *           Log suffix.
    */
   private void outputWarning(Aast node, String front, String back)
   {
      System.err.println(front +
                         node.getText() +
                         " [" +
                         node.getLine() +
                         ":" +
                         node.getColumn() +
                         "]" +
                         back);
   }
   
   /**
    * Helper to remove surrounding unnecessary double quotes. This method
    * assumes that there *are* double quotes as the first and last chars.
    */
   private String removeBogusQuotes(String quoted)
   {
      if (quoted == null)
         return "";
         
      return quoted.substring(1, quoted.length() - 1);
   }
   
   /**
    * Set the line and column numbers of the target AST to be the same as
    * the source AST.
    *
    * @param    source
    *           AST with the line and column numbers to copy.
    * @param    target
    *           AST to modify.
    */
   private void forceLineColumnNums(Aast source, Aast target)
   {
      target.setLine(source.getLine());
      target.setColumn(source.getColumn());
   }
   
   /**
    * Helper to instantiate a new (limited use) lexer and re-lex the input.
    * The problem is that the schema dump file puts everything inside quotes
    * and the lexer doesn't know when to remove the quotes and when to leave
    * them.  This results in the <code>INITIAL</code> having children that
    * are of the <code>STRING</code> type even when they should be some other
    * kind of literal.  <b>This method removes the unnecessary enclosing
    * double quotes and then calls the lexer to obtain the resulting token. 
    * It is assumed at this time that only one token will be present in the
    * passed input string.  If this is not the case, we will have to make
    * more serious changes!</b>
    *
    * @param    input
    *           The double quoted enclosed string containing a token to lex.
    *
    * @return   The token of the proper token type and text OR 
    *           <code>null</code> if the input text is an empty string.
    */
   private Token relex(String input)
   throws antlr.TokenStreamException
   {
      String cleaner = removeBogusQuotes(input);
      
      if ("".equals(cleaner))
         return null;
         
      SymbolResolver sr = new SymbolResolver(false);
      
      // disable the dot kludge as the following DOT will sometimes break
      // lexing, this needs to be a single token only
      ProgressLexer relex = new ProgressLexer(new StringReader(cleaner),
                                              sr,
                                              false);
                                              
      return relex.nextToken();
   }
   
   /**
    * Check if the given token is prefaced by whitespace in the original
    * source code (by reading from the hidden token stream).
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine any prefacing hidden tokens.
    *
    * @return   <code>true</code> if the given token is prefaced by whitespace
    *           in the hidden token stream.
    */
   private boolean prefacedByWhitespace(Token tok)
   {
      CommonHiddenStreamToken space = null;
      space = hidden.getHiddenBefore((CommonHiddenStreamToken) tok);      
      
      return (space != null && space.getType() == WS);
   }
   
   /**                          
    * Check if the given token is followed by whitespace in the original
    * source code (by reading from the hidden token stream).
    *
    * @param    tok
    *           The token from the token stream (the lexer) which is needed
    *           in order to examine if there are following hidden tokens.
    *
    * @return   <code>true</code> if the given token is followed by whitespace
    *           in the hidden token stream.
    */
   private boolean followedByWhitespace(Token tok)
   {
      CommonHiddenStreamToken space = null;
      space = hidden.getHiddenAfter((CommonHiddenStreamToken) tok);      
      
      return (space != null && space.getType() == WS);
   }
   
   /**
    * Merge the following non-hidden token text until (and including) a token
    * followed by whitespace or <code>EOF</code>.
    * <p>
    * All the merged tokens will be consumed and the merged text will be stored
    * in the last token before the token that triggered the end of merging. The
    * line and column numbers of the merged token will be forced to the same
    * values as the first token on entry.
    *
    * @return   <code>true</code> if merging was done, <code>false</code> if
    *           no changes were made.
    */
   private boolean mergeUntilWS()
   throws TokenStreamException,
          MismatchedTokenException
   {
      boolean merged = false;
   
      CommonToken current = (CommonToken) LT(1);
      CommonToken next    = (CommonToken) LT(2);
      
      int line = current.getLine();
      int col  = current.getColumn();
      
      StringBuilder sb = new StringBuilder();
      
      while (!followedByWhitespace(current)  && !(next != null && next.getType() == EOF))
      {
         merged = true;
      
         // save the text of the current token
         sb.append(current.getText());
         
         matchNot(EOF);
         
         // update the state for next iteration
         current = next;
         next    = (CommonToken) LT(2);
      }
      
      if (merged)
      {
         // put the last text into the buffer
         sb.append(current.getText());
      
         // override the last token's contents and positioning with the
         // merged results
         current.setText(sb.toString());
         current.setLine(line);
         current.setColumn(col);
         
         // we don't match here, so that the calling code can match on the
         // resulting token
      }
      
      return merged;
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * schema file lexing.
    * <p>
    * Syntax:
    * <pre>
    *    java SchemaParser [-l | -v] &lt;schema_dump_file&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> -l specifies to only run the lexer rather than the full parser
    *   <li> -v specifies to display the parsed tree in a Swing GUI panel
    *   <li> &lt;schema_dump_file&gt; is the filename of the schema dump
    *        (enclose the name in double quotes if spaces or other reserved
    *        command line characters are part of the filename)
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main( String[] args )  
   {
      if (args.length == 0)
      {
         System.err.println("Syntax: java.exe SchemaParser [-l | -v] " +
                            "<schema_dump_file>");
         System.exit( -1 );
      }
         
      try
      {
         // we know there is at least 1 arg
         boolean lexonly = "-l".equals(args[0]);
         boolean visual  = "-v".equals(args[0]);
         
         String fname = args[0];
         
         if (lexonly || visual)
         {
            fname = args[1];
         }
      
         // setup
         SymbolResolver sym = new SymbolResolver(false);
         FileReader     fr  = new FileReader(fname);
         BufferedReader br  = new BufferedReader(fr);
         
         if (lexonly)
         {
            // lex only
            LexerDumpFilter test = new LexerDumpFilter(br, sym);
            test.activateSchemaProcessing();
            test.setOutput(new PrintWriter(System.out));
            
            // start the lexing and print a record for each token                  
            test.dump();         
         }
         else
         {
            ProgressLexer lexer = new ProgressLexer(br, sym);
            lexer.activateSchemaProcessing();
            
            SchemaParser parser = new SchemaParser(lexer, true);
            
            // parse it and create a tree
            parser.schema();
            
            AnnotatedAst result = (AnnotatedAst) parser.getAST();
            
            // this is needed to ensure the parents are set right, or DumpTree
            // will be unable to calculate indents AND features like the
            // Aast.iterator() will also fail to operate properly
            result.fixups((Aast) null, null);
                     
            // print a report to stdout
            DumpTree visitor = new DumpTree();
            visitor.visit(result);         
            
            if (visual)
            {
               // display the tree in a frame window if debug is on
               ASTFrame frame = new ASTFrame("Database Schema", result);                                                
               frame.setVisible(true);
            }
         }
      }
       
      catch( Exception excpt )
      {
         excpt.printStackTrace();
      }
   }   
}

/**
 * Top-level parser method, from which the entire recursive descent parsing
 * of tokens for this file begins. This rule expects to match, in order:
 * <ol>
 * <li> zero or one {@link #database} construct;
 * <li> zero or more {@link #sequence} constructs;
 * <li> zero or more {@link #table} constructs;
 * <li> one {@link #trailer} construct;
 * </ol>
 * <p>
 * The root of the schema AST will be an artificial <code>DATABASE</code> node.
 * All sub-trees created by the rule references will be attached as child
 * nodes, in the order shown above.
 */
schema
   :
      {
         setASTNodeClass(AST_CLASS);
         astFactory.makeASTRoot(currentAST, #[DATABASE]);
         String name = null;
      }
      (
         (
            name=database!
            {
               if (name != null) ##.setText(name);
               ##.setLine(1);
               ##.setColumn(1);
            }
         )?
         (sequence)*
         (table)*
         (
            u:update_stuff!
            {
               outputWarning(#u, "WARNING: ignoring cruft ", "");
            }
         )*
         trailer!                            
      )
   ;
   
/**
 * Match a database construct, which specifies the database name. Curiously,
 * the database name seems to always show up as unknown (&quot;?&quot;).
 * <p>
 * The sub-tree created by this rule is rooted at the <code>KW_DATABASE</code>
 * node and has a single child which is the <code>STRING</code> that is the
 * database name. Any <code>KW_ADD</code> or <code>KW_UPDATE</code> token
 * is dropped.
 *
 * @return   The text name of the database.
 */
database returns [String name = null] 
   :
      (KW_ADD! | KW_UPDATE!) KW_DATABASE^ s:STRING
      { name = removeBogusQuotes(#s.getText()); }
   ;
   
/**
 * These UPDATE TABLE and UPDATE INDEX constructs were seen at the bottom of a customer
 * schema, but they don't seem to have any purpose in a converted environment. If these
 * need to be used later, the tree building will have to be improved.
 */
update_stuff
   :
      KW_UPDATE^
      (
           KW_TABLE STRING
         | KW_INDEX STRING KW_OF STRING
      )
      KW_BUFRPOOL STRING
   ;
   
/**
 * Match the trailer section of the file. The data contained in this section
 * is not used for P2J conversion purposes, but provides a handy mechanism
 * to detect that we have successfully reached the end of the database schema
 * definitions section.
 * <p>
 * This matches some <code>DOT</code> tokens, some code-page related keywords
 * and then matches the bytecount numeric token which completes the file.
 * <p>
 * The sub-tree created by this rule will be rooted at an artificial
 * <code>TRAILER</code> node. The first child will be the sub-tree from the
 * code_page rule and the second child will be the <code>NUM_LITERAL</code>
 * which represents the byte or character count for the file.
 */ 
trailer   
   :
      { astFactory.makeASTRoot(currentAST, #[TRAILER]); }
      DOT! KW_PSC!
      (bufpool)?
      c:code_page
      DOT!
      NUM_LITERAL  // bytecount
      {
         forceLineColumnNums(#c, ##);
      }
   ;
   
/**                                                      
 * Code-page related keywords and a code-page name as a <code>SYMBOL</code>.
 * <p>
 * The sub-tree created by this rule will be rooted at the <code>KW_PSC</code>
 * node and all following nodes will be it's direct children.
 */
code_page   
   :
      (KW_CPSTREAM^ | KW_CP^) EQUALS! any_ridiculous_symbol
   ;
   
/**                                                      
 * Optional <code>BUFPOOL</code> keyword followed by an <code>EQUALS</code> and
 * a boolean literal.  This can appear in the {@link #trailer}. 
 */
bufpool   
   :
      KW_BUFPOOL^ EQUALS! (BOOL_TRUE | BOOL_FALSE)
   ;
   
/**
 * Matches a sequence definition, which consists of the
 * <code>ADD SEQUENCE</code> phrase followed by sequence property
 * definitions.
 * <p>
 * The AST is produced rooted at the artificial <code>SEQUENCE</code> node.
 * The <code>KW_ADD</code>, <code>KW_SEQUENCE</code> and the <code>STRING</code>
 * tokens are all dropped from the tree, and sequence name (from the 
 * <code>STRING</code> node) will be set as the text for the
 * <code>SEQUENCE</code> node.  The children of this node will be the sub-trees
 * created by the rule references for the sequence properties.
 */
sequence
   :
      { astFactory.makeASTRoot(currentAST, #[SEQUENCE]); }
      (
         KW_ADD! seq:KW_SEQUENCE! s:STRING!
         { ##.setText(removeBogusQuotes(#s.getText())); } 
         
         (
              initial[FIELD_INT, null]
            | increment
            | cycleOnLimit
            | minVal
            | maxVal
         )*
      )
      {
         forceLineColumnNums(#seq, ##);
      }
   ;
   
/**
 * Matches a table definition, which consists of the <code>ADD TABLE</code>
 * phrase followed by table property definitions, some of which are optional,
 * then some number of {@link #field} and/or {@link #index} definitions.
 * <p>
 * The AST is produced rooted at an artificial <code>TABLE</code> node (not to
 * be confused with the <code>KW_TABLE</code>).  The <code>KW_ADD</code>,
 * <code>KW_TABLE</code>, <code>STRING</code>, <code>KW_TYPE</code> and
 * <code>SYMBOL</code> tokens are all dropped from the tree, and table name
 * (from the <code>STRING</code> node) will be set as the text for the
 * <code>TABLE</code> node.  The children of this node will be the sub-trees
 * created by the rule references for the table properties, fields and indexes.
 */
table
   :
      { astFactory.makeASTRoot(currentAST, #[TABLE]); }
      (
         KW_ADD! tab:KW_TABLE! s:STRING! (y:KW_TYPE! d:SYMBOL!)?
         {
            currentTable = removeBogusQuotes(#s.getText());
            ##.putAnnotation("legacy_name", currentTable);
            currentTable = currentTable.toLowerCase();
            ##.setText(currentTable);
            
            if (#y != null)
            {
               System.out.println("WARNING:  ignoring " +
                                  #y.getText() +
                                  " " +
                                  #d.getText() +
                                  " Data Server option to ADD TABLE statement");
            }
         }
         table_props
         (
            options { generateAmbigWarnings = false; }
            :
              field
            | index
         )*
      )
      {
         currentTable = null;
         forceLineColumnNums(#tab, ##);
      }
   ;
   
/**
 * Matches all possible table property definitions.
 * <p>                                       
 * The sub-tree created is rooted at an artificial <code>PROPERTIES</code>
 * node.  Each property is parsed in its own rule.  All of the sub-trees
 * for those properties are attached as children of the root node, in the same
 * order they are encountered in the file.  If no properties are found for
 * a table, this rule doesn't create a sub-tree. 
 */
table_props   
   :
      { astFactory.makeASTRoot(currentAST, #[PROPERTIES]); }
      (
           area
         | label
         | description
         | valExp
         | valMsg
         | frozen
         | hidden
         | dumpName
         | permissions
         | table_trigger
         | unknownOption!
      )*
      {
         if (##.getNumberOfChildren() == 0)
         {
            ## = null;
         }
         else
         {
            forceLineColumnNums(##.getChildAt(0), ##);
         }
      }
   ;
   
/**
 * This rule matches a field definition, which consists of the
 * <code>ADD FIELD</code> phrase followed by field property definitions, some
 * of which are optional.
 * <p>
 * An AST is produced as a result, consisting of an artificial
 * <code>FIELD_*</code> node, with its type assigned based on the data type of
 * the field as returned by the <code>as_clause</code> rule (see that rule for
 * the specific types that can be returned).  The text of that root node is the
 * field name as parsed from the text of the first <code>STRING</code> token.
 * The <code>KW_ADD</code>, <code>KW_FIELD</code>, <code>STRING</code>,
 * <code>KW_OF</code> and <code>STRING</code> tokens are all dropped from the
 * tree.  Likewise, the sub-tree for <code>as_clause</code> is dropped, but as
 * noted above, the data type is used as the root node type.  The children of
 * this node will be the sub-trees created by the rule references for the field
 * properties.  In the case of the <code>KW_MAND</code> and
 * <code>KW_CASE_SEN</code> tokens, the node is directly added as a child of
 * the root (but in the case of <code>KW_MAND</code> it is converted to
 * artificial type <code>MANDATORY</code>).
 * <p>
 * Implementation Note:
 * There is an assertion within this method to ensure that the table name
 * associated with the table field being parsed is the same as the name of
 * the table definition most recently parsed. This ensures that this field
 * definition, which is assumed to be in the context of the most recent table
 * definition, is properly "nested" as expected within that table. This is
 * not a critical assertion at runtime, but was a reality check during
 * development of this grammar, to confirm the assumption that a field
 * definition would always be a child of the most recently parsed table
 * definition.
 *
 * @throws  SemanticException
 *          if this field definition reports a different parent table name
 *          from the table definition most recently parsed.
 */
field
   :
      {
         int    ftype;
         String fmt = null;
      }
      (
         KW_ADD! f:KW_FIELD^ s:STRING!
         { #f.setText(removeBogusQuotes(#s.getText().toLowerCase())); } 
          
         // table name (dropped from the tree but used for a sanity check)
         KW_OF! t:STRING!
         { removeBogusQuotes(#t.getText()).equalsIgnoreCase(currentTable) }?
          
         // data type (dropped from the tree but used to set the data type)
         ftype=as_clause!
         { #f.setType(ftype); }
          
         // properties
         (
              description
            | ff:format  { fmt = #ff.getChildAt(0).getChildAt(0).getText(); }
            | initial[ftype, fmt]
            | label
            | position
            | sqlWidth
            | valExp
            | valMsg
            | columnLabel
            | permissions
            | field_trigger
            | help
            | extent
            | decimals
            | length
            | viewAs
            | order
            | m:KW_MAND  { #m.setType(MANDATORY); }
            | KW_CASE_SEN
            | n:KW_NOT_CS!  // no meaning, seen on non-character data types
              {
                 outputWarning(#n, "WARNING: ignoring cruft ", "");
              }
            | na:KW_NULLALWD! // unknown meaning
              {
                 outputWarning(#na, "WARNING: ignoring cruft ", "");
              }
            | maxWidth
            | lobArea
            | lobBytes
            | lobSize
            | clobCodepage
            | clobCollation
            | clobType
            | cruft!
            | unknownOption!
         )*
      )
   ;
   
/**
 * Matches an index definition, which consists of the <code>ADD INDEX</code>
 * phrase followed by index property definitions, some of which are optional,
 * followed by some number of child entity {@link #indexField definitions}
 * (field references).  The index name is found in the first <code>STRING</code>
 * token and the table name on which the index is based is found in the second
 * <code>STRING</code> token (after the <code>KW_ON</code>).
 * <p>
 * A sub-tree is produced as a result, consisting of an artificial
 * <code>INDEX</code> token node, with an artificial <code>PROPERTIES</code>
 * child node containing the index properties, and an artificial
 * <code>INDEX_FIELD</code> child node for each field of this index. The text
 * of the <code>INDEX</code> node will be set as the index name read from the
 * first <code>STRING</code> token.  The <code>KW_ADD</code>,
 * <code>KW_INDEX</code>, <code>STRING</code>, <code>KW_ON</code> and
 * <code>STRING</code> tokens are all dropped from the tree.  
 * <p>
 * Implementation Note:
 * There is an assertion within this method to ensure that the table name
 * associated with the table index being parsed is the same as the name of
 * the table definition most recently parsed. This ensures that this index
 * definition, which is assumed to be in the context of the most recent table
 * definition, is properly "nested" as expected within that table. This is
 * not a critical assertion at runtime, but was a reality check during
 * development of this grammar, to confirm the assumption that an index
 * definition would always be a child of the most recently parsed table
 * definition.
 *
 * @throws  SemanticException
 *          if this index definition reports a different parent table name
 *          from the table definition most recently parsed.
 */
index
   :
      (
         KW_ADD! i:KW_INDEX^ s:STRING!
         {
            #i.setType(INDEX);
            #i.setText(removeBogusQuotes(#s.getText().toLowerCase()));
         } 
         
         // table name (dropped from the tree but used for a sanity check)
         KW_ON! n:STRING!
         { removeBogusQuotes(#n.getText()).equalsIgnoreCase(currentTable) }?
            
         index_props (indexField)+
      )
   ;
   
/**
 * Index properties, aggregated as children of an artificial
 * <code>PROPERTIES</code> node, which is itself a direct child of the
 * <code>INDEX</code> node. Each matched keyword is a direct child
 * as are the sub-trees created in the referenced rules.
 */
index_props   
   :
      { astFactory.makeASTRoot(currentAST, #[PROPERTIES]); }   
      (
           area
         | description
         | KW_WORD_IDX
         | KW_UNIQUE
         | KW_PRIMARY
         | KW_INACTIVE
         | unknownOption!
      )*
      {
         if (##.getNumberOfChildren() == 0)
         {
            ## = null;
         }
         else
         {
            forceLineColumnNums(##.getChildAt(0), ##);
         }
      }
   ;
   
/**
 * Matches an index field definition, which consists of the
 * <code>INDEX-FIELD</code> keyword followed by index field property
 * definitions, some of which are optional.
 * <p>
 * An AST is produced as a result, consisting of an imaginary
 * <code>INDEX_FIELD</code> token node, with child nodes for each property
 * encountered.
 */
indexField 
   :
      k:KW_IDX_FLD^ { #k.setType(INDEX_FIELD); }
      s:STRING!     { #k.setText(removeBogusQuotes(#s.getText())); }
      // we don't have separate keyword processing for KW_ASCEND, but we differentiate it here
      // to make the downstream processing more meaningful
      (a:KW_ASC { #a.setType(KW_ASCEND); } | KW_DESCEND)?
      (KW_ABBV)?
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>INCREMENT</code> property of a sequence.
 * <p>
 * Matches:
 * <ol>
 * <li>the <code>KW_INCR</code> token
 * <li>the value of the property as a <code>NUM_LITERAL</code>
 * </ol>
 * <p>
 * The sub-tree produced as a result consists of a <code>INCREMENT</code>
 * node representing the property key (the <code>KW_INCR</code> type is
 * replaced) and a single child node with a type of <code>NUM_LITERAL</code>.
 */
increment
   :
      k:KW_INCR^ NUM_LITERAL { #k.setType(INCREMENT); }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>CYCLE-ON-LIMIT</code> property of a sequence.
 * <p>
 * The sub-tree produced as a result consists of a <code>CYCLE_ON_LIMIT</code>
 * node representing the property key (this replaces the <code>KW_CYCLE</code>
 * token) and a single child node with the value as a literal.
 */
cycleOnLimit
   :
      k:KW_CYCLE^ literal   { #k.setType(CYCLE_ON_LIMIT); }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>MIN-VAL</code> property of a sequence.
 * <p>
 * Matches:
 * <ol>
 *    <li> the <code>KW_MIN_VAL</code> keyword token
 *    <li> a <code>NUM_LITERAL</code>
 * </ol>
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_MIN_VAL</code>
 * node representing the property key and a single child node with a type of
 * <code>NUM_LITERAL</code>.
 */
minVal
   :
      KW_MIN_VAL^ NUM_LITERAL
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>MAX-VAL</code> property of a sequence.
 * <p>
 * Matches:
 * <ol>
 *    <li> the <code>KW_MAX_VAL</code> keyword token
 *    <li> a <code>NUM_LITERAL</code>
 * </ol>
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_MAX_VAL</code>
 * node representing the property key and a single child node with a type of
 * <code>NUM_LITERAL</code>.
 */
maxVal
   :
      KW_MAX_VAL^ NUM_LITERAL
   ;
   
/**                          
 * Match <code>MAX-WIDTH</code> keyword and a following 
 * <code>NUM_LITERAL</code> for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_MAX_WID</code>
 * node with a single child node with a type of <code>NUM_LITERAL</code>.
 */
maxWidth
   :
      KW_MAX_WID^ NUM_LITERAL
   ;
   
/**                          
 * Match <code>LOB-AREA</code> keyword and a following <code>STRING</code>
 * for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_LOB_AREA</code>
 * node with a single child node with a type of <code>STRING</code>.
 */
lobArea
   :
      KW_LOB_AREA^ STRING
   ;
   
/**                          
 * Match <code>LOB-BYTES</code> keyword and a following
 * <code>NUM_LITERAL</code> for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_LOB_BYTE</code>
 * node with a single child node with a type of <code>NUM_LITERAL</code>.
 */
lobBytes
   :
      KW_LOB_BYTE^ NUM_LITERAL
   ;
   
/**                          
 * Match <code>LOB-SIZE</code> keyword and a following size specification
 * for a field.  The size specification is a integral literal with an optional
 * single letter at the end to specify the units (e.g. 100M).  This will
 * appear as a <code>SYMBOL</code> in the resulting tree, although it
 * doesn't naturally lex that way.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_LOB_SIZE</code>
 * node with a single child node with a type of <code>SYMBOL</code>.  The
 * matched <code>NUM_LITERAL</code> is dropped but the text of that node is
 * prefixed to the text of the <code>SYMBOL</code> node. 
 */
lobSize
   :
      KW_LOB_SIZE^ n:NUM_LITERAL
      (
         { !prefacedByWhitespace(LT(1)) }?
         s:SYMBOL!
      )?
      {
         // put the single size spec back together, if needed
         if (#s != null)
         {
            StringBuilder sb = new StringBuilder();
            sb.append(#n.getText()).append(#s.getText());
            #n.setText(sb.toString());
         }
         
         #n.setType(SYMBOL);
      }
   ;
   
/**
 * Match <code>CLOB-CODEPAGE</code> keyword and a following <code>STRING</code>
 * for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_CLOB_CP</code>
 * node with a single child node with a type of <code>STRING</code>.
 */
clobCodepage
   :
      KW_CLOB_CP^ STRING
   ;
   
/**
 * Match <code>CLOB-COLLATION</code> keyword and a following <code>STRING</code>
 * for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_CLOB_COL</code>
 * node with a single child node with a type of <code>STRING</code>.
 */
clobCollation
   :
      KW_CLOB_COL^ STRING
   ;
   
/**
 * Match <code>CLOB-TYPE</code> keyword and a following <code>NUM_LITERAL</code>
 * for a field.
 * <p>
 * The sub-tree produced as a result consists of a <code>KW_CLOB_TYP</code>
 * node with a single child node with a type of <code>NUM_LITERAL</code>.
 */
clobType
   :
      KW_CLOB_TYP^ NUM_LITERAL
   ;
   
/**
 * This rule matches a set of tokens assumed from the caller's context to be
 * the value of the <code>AREA</code> property of a table or index.
 * <p>
 * The resulting sub-tree will be a single artficial node of type
 * <code>AREA</code> with it's text set as the text contents of the
 * <code>STRING</code> node. The actual tokens read by the rule will be dropped.
 */
area
   :
      k:KW_AREA^ v:STRING!
      {
         #k.setType(AREA);
         #k.setText(removeBogusQuotes(#v.getText()));
      }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>LABEL</code> property of a table or field.
 * <p>
 * The resulting sub-tree will be rooted at the <code>KW_LABEL</code> node and
 * the <code>STRING</code> will be its only child node.
 */
label
   :
      KW_LABEL^ s:STRING
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>DESCRIPTION</code> property of a table, field,
 * or index.
 * <p>
 * The resulting sub-tree will be rooted at an artificial 
 * <code>DESCRIPTION</code> node and the text of that node will be set as
 * the text contents of the <code>STRING</code> node.  Both tokens being
 * matched will be discarded.
 */
description
   :
      k:KW_DESCR^ v:STRING!
      {
         #k.setType(DESCRIPTION);
         #k.setText(removeBogusQuotes(#v.getText()));
      }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>VALMSG</code> property of a table or field.
 * <p>
 * Matches
 * <ol>
 *    <li> the <code>KW_VALMSG</code> token
 *    <li> a <code>STRING</code>
 * </ol>
 * <p>
 * <p>
 * The resulting AST will be rooted at an artificial <code>VALMSG</code>
 * node, with the <code>STRING</code> as a child of an artificial
 * <code>EXPRESSION</code> node (which makes the <code>STRING</code> the
 * grand-child node to the <code>VALMSG</code>.
 * <p>
 * <b>The child node will need to be post-processed if it contains a real
 * Progress 4GL expression rather than only a string literal. All samples
 * to date are only a simple string so this code will work properly.</b>
 */
valMsg!
   :
      m:KW_VALMSG v:STRING
      {
         ## = #([VALMSG, "valmsg"], ([EXPRESSION, "expression"], v));
         forceLineColumnNums(#m, ##);
      }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the
 * value of the <code>DUMP-NAME</code> property of a table.
 * <p>
 * The resulting sub-tree is rooted at an artificial <code>DUMP_NAME</code>
 * node and the text contents of the <code>STRING</code> are set as the text
 * of the root node.  The <code>KW_DMP_NAME</code> and <code>STRING</code>
 * tokens are discarded. 
 */
dumpName
   :
      k:KW_DMP_NAME^ v:STRING!
      {
         #k.setType(DUMP_NAME);
         #k.setText(removeBogusQuotes(#v.getText()));
      }
   ;
   
/**
 * Match a token assumed from the caller's context to be the
 * value of the <code>FROZEN</code> property of a table.
 * <p>
 * The resulting AST will be a single node of artificial type
 * <code>FROZEN</code>.
 */
frozen
   :
      k:KW_FROZEN   { #k.setType(FROZEN); }
   ;
   
/**
 * Match a token assumed from the caller's context to be the
 * value of the <code>HIDDEN</code> property of a table.
 * <p>
 * The resulting AST will be a single node of artificial type
 * <code>HIDDEN</code>.
 */
hidden
   :
      k:KW_HIDDEN   { #k.setType(HIDDEN); }
   ;
   
/**
 * Match a set of tokens for the <code>AS</code> clause, which defines the data
 * type of a field.
 * <p>
 * The resulting sub-tree is rooted at the <code>KW_AS</code> node and it will
 * have a single child node which is the data type keyword matched.  More
 * importantly, the data type keyword will set the type for the field as
 * follows:
 * <p>
 * <pre>
 * Type Keyword                Field Data Type
 * --------------------        ---------------------
 * KW_BLOB                     FIELD_BLOB
 * KW_CHAR                     FIELD_CHAR
 * KW_CLOB                     FIELD_CLOB
 * KW_COM_HNDL                 FIELD_COM_HANDLE
 * KW_DATE                     FIELD_DATE
 * KW_DATETIME                 FIELD_DATETIME
 * KW_DATE_TZ                  FIELD_DATETIME_TZ
 * KW_DEC                      FIELD_DEC
 * KW_INT                      FIELD_INT
 * KW_INT64                    FIELD_INT64
 * KW_LOGICAL                  FIELD_LOGICAL
 * KW_RECID                    FIELD_RECID
 * KW_ROWID                    FIELD_ROWID
 * KW_RAW                      FIELD_RAW
 * KW_HANDLE                   FIELD_HANDLE
 * </pre>
 * <p>
 * This parser is used for <code>.df</code> files which are only used with
 * permanent databases.  Since <code>CLASS</code> fields are for temp-tables
 * only, the <code>KW_CLASS</code> cannot be used in a <code>.df</code> file
 * and <code>FIELD_CLASS</code> will not be ever used as a field type in a
 * permanent database.
 * <p>
 * The following can also be matched, but they are only used in metadata
 * schemas, normal Progress databases don't provide these field types.
 * <p>
 * <pre>
 * Type Keyword                Metadata Field Data Type
 * --------------------        ------------------------
 * KW_BIGINT                   FIELD_BIGINT
 * KW_BYTE                     FIELD_BYTE
 * KW_DOUBLE                   FIELD_DOUBLE
 * KW_FIXCHAR                  FIELD_FIXCHAR
 * KW_FLOAT                    FIELD_FLOAT
 * KW_SHORT                    FIELD_SHORT
 * KW_TIMESTMP                 FIELD_TIMESTAMP
 * KW_TIME                     FIELD_TIME
 * </pre>
 *
 * @return   The data type of the field.
 */
as_clause returns [int ftype = -1]
   :
      // CLASS fields are for temp-tables only, this parser is used for
      // DF files which are only used with permanent databases... so there
      // is no need for support here
      KW_AS^
      (
           KW_BLOB                        { ftype = FIELD_BLOB; }
         | KW_CHAR                        { ftype = FIELD_CHAR; }
         | KW_CLOB                        { ftype = FIELD_CLOB; }
         | KW_COM_HNDL                    { ftype = FIELD_COM_HANDLE; }
         | KW_DATE                        { ftype = FIELD_DATE; }
         | KW_DATETIME                    { ftype = FIELD_DATETIME; }
         | KW_DATE_TZ                     { ftype = FIELD_DATETIME_TZ; }
         | KW_DEC                         { ftype = FIELD_DEC; }
         | KW_INT                         { ftype = FIELD_INT; }
         | KW_INT64                       { ftype = FIELD_INT64; }
         | KW_LOGICAL                     { ftype = FIELD_LOGICAL; }
         | KW_RECID                       { ftype = FIELD_RECID; }
         | KW_ROWID                       { ftype = FIELD_ROWID; }
         | KW_RAW                         { ftype = FIELD_RAW; }
         | KW_HANDLE                      { ftype = FIELD_HANDLE; }
         
         // "extended" meta types
         | KW_BIGINT                      { ftype = FIELD_BIGINT; }
         | KW_BYTE                        { ftype = FIELD_BYTE; }
         | KW_DOUBLE                      { ftype = FIELD_DOUBLE; }
         | KW_FIXCHAR                     { ftype = FIELD_FIXCHAR; }
         | KW_FLOAT                       { ftype = FIELD_FLOAT; }
         | KW_SHORT                       { ftype = FIELD_SHORT; }
         | KW_TIMESTMP                    { ftype = FIELD_TIMESTAMP; }
         | KW_TIME                        { ftype = FIELD_TIME; }
      )
   ;
   
/**
 * Match a set of tokens which is the value of the <code>FORMAT</code> property
 * of a field.
 * <p>
 * The resulting sub-tree is rooted at a <code>KW_FORMAT</code> node with an
 * artificial <code>EXPRESSION</code> child node and the <code>STRING</code>
 * node is a child of the <code>EXPRESSION</code>.
 */
format!
   :
      k:KW_FORMAT v:STRING
      { ## = #(k, ([EXPRESSION, "expression"], v)); }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>INITIAL</code> property of a sequence or field.
 * <p>
 * Matches:
 * <ol>
 *    <li> the <code>KW_INIT</code> keyword token
 *    <li> a {@link #literal}
 * </ol>
 * <p>
 * The schema dump file encloses all data values inside double quotes
 * and the lexer doesn't know when to remove the quotes and when to leave
 * them, since the necessary context is only known to the parser.  Only the
 * parser can know this based on the data type of the field or the fact that
 * this is a sequence.  This results in the <code>INITIAL</code> construct
 * having children that are of the <code>STRING</code> type even when they
 * should be some other kind of literal.  <b>To resolve this issue, the
 * <code>relex</code> helper method is used and the offending token is
 * overwritten in the init method. Then the matching continues as it should
 * with no problems and having the correct types.  WARNING: this rewriting does
 * not handle the situation where multiple tokens exist *inside* a single
 * <code>STRING</code> token!</b>
 * <p>
 * <b>WARNING: in normal Progress 4GL source, when extent &gt; 0, this same
 * construct can have a list of comma-separated values. No examples have been
 * encountered yet and the code in this rule won't match such conditions.</b>
 * <p>
 * The sub-tree created by this rule is rooted at a <code>KW_INIT</code> and
 * that node will have a single child which is the literal initial value. 
 * <p>
 * If this construct is encountered while parsing the definition for a field
 * of type <code>RAW</code>, the tokens will be discarded and no sub-tree will
 * be created. This is done since it is invalid to initialize a field of that
 * type.
 *
 * @param    ftype
 *           The data type of the field or sequence being processed for the
 *           initial clause.
 * @param    fmt
 *           The format string to be used in manhandling user-defined logical
 *           'constants' or <code>null</code> if no format string exists.
 */
initial [int ftype, String fmt]
   :
      {
         Token   content = LT(2);
         boolean eat     = false;
         int     oldtype = -1;
      
         // TODO: what does a BLOB or CLOB initializer look like?
         
         // it may be time to rewrite our token!
         if (content.getType() == STRING        &&
             ftype             != FIELD_CHAR    &&
             ftype             != FIELD_FIXCHAR)
         {
            // OK, we need to re-lex!
            Token fixed = relex(content.getText());
            
            if (fixed != null)
            {
               int newType = fixed.getType();
               
               // user-defined logical manhandling! 
               if (ftype    == FIELD_LOGICAL &&
                   newType  != BOOL_TRUE     &&
                   newType  != BOOL_FALSE    &&
                   newType  != UNKNOWN_VAL)
               {
                  // fixups for ridiculous Progress schema feature of allowing
                  // arbitrary user-defined strings (that are an exact match
                  // or an unambiguous abbreviated) match to a format string
                  // of the form "XXX/YYY" --> x, X, xx, XX, xxx, XXX all
                  // resolve to BOOL_TRUE and y, Y, yy, YY, yyy, YYY resolve
                  // to BOOL_FALSE

                  int match = -1;
                  if (fmt != null)
                  {
                     String lower = fmt.toLowerCase();
                     match = lower.indexOf(fixed.getText().toLowerCase());
                  }

                  if (match == 1)   // ignore opening double quote character
                  {
                     fixed.setType(BOOL_TRUE);
                  }
                  else
                  {
                     fixed.setType(BOOL_FALSE);
                  }
               }
            
               // date-specific fixups
               if (ftype == FIELD_DATE)
               {
                  if (newType == KW_TODAY)
                  {
                     // use of 'today' as a date literal manhandling!
                     oldtype = KW_TODAY;
                     fixed.setType(DATE_LITERAL);
                  }
                  else if (newType != DATE_LITERAL)
                  {
                     // deal with garbage input;  it is possible to code something like:
                     // INITIAL "this is total crap for a date initial value"
                     // we treat this as unknown value
                     fixed.setText("?");
                     fixed.setType(UNKNOWN_VAL);
                  }
               }
            
               // use of 'now' as a datetime literal manhandling! 
               if ((ftype == FIELD_DATETIME || ftype == FIELD_DATETIME_TZ) &&
                   newType == KW_NOW)
               {
                  oldtype = KW_NOW;
                  fixed.setType(DATETIME_TZ_LITERAL);
               }

               // blank string for numeric types is interpreted as 0 by Progress
               if ((ftype == FIELD_INT ||
                    ftype == FIELD_INT64 ||
                    ftype == FIELD_DEC ||
                    ftype == FIELD_RECID) && newType == EOF)
               {
                  fixed.setText("0");
                  fixed.setType(NUM_LITERAL);
               }

               content.setText(fixed.getText());
               content.setType(fixed.getType());
            }
            else
            {
               if (ftype == FIELD_RAW)
               {
                  // initializers don't make sense for raw fields! 
                  eat = true;
               }
               else if (ftype == FIELD_INT   || 
                        ftype == FIELD_INT64 || 
                        ftype == FIELD_DEC   || 
                        ftype == FIELD_RECID)
               {
                  content.setType(NUM_LITERAL);
                  content.setText("0");
               }
               else if (ftype == FIELD_LOGICAL)
               {
                  content.setType(BOOL_FALSE);
               }
            }
         }
      }
      KW_INIT^ l:literal
      {
         // consume the result and null out the subtree if the result makes
         // no sense (e.g. no initializers in raw fields)
         if (eat)
         {
            ## = null;
         }
         else
         {
            if (oldtype != -1)
            {
               // save off the old type as an annotation
               #l.putAnnotation("oldtype", new Long(oldtype));
            }
         }
      }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>POSITION</code> property of a Progress field.
 * <p>
 * The resulting sub-tree has an artificial <code>POSITION</code> node with a
 * child node of <code>NUM_LITERAL</code>. The <code>KW_POS</code> is dropped.
 */
position
   :
      k:KW_POS^ NUM_LITERAL
      { #k.setType(POSITION); }
   ;
   
/**
 * Match a set of tokens as the value of the <code>SQL-WIDTH</code> property of
 * a field.
 * <p>
 * The resulting sub-tree has an artificial <code>SQL_WIDTH</code> node with a
 * child node of <code>NUM_LITERAL</code>. The <code>KW_SQL_WID</code> is
 * dropped.
 */
sqlWidth
   :
      k:KW_SQL_WID^ NUM_LITERAL
      { #k.setType(SQL_WIDTH); }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>VALEXP</code> property of a Progress field. This results in
 * a <code>VALEXP</code> AST with a single child of type <code>STRING</code>.
 * The string as read from the DF file is enclosed within double quotes;
 * these are removed. The DF file escapes embedded double quotes in this
 * string by doubling them up. All doubled-up double quotes are replaced
 * with single, double quotes (e.g., ...<code>""some text""</code>... becomes
 * ...<code>"some text"</code>...).
 * <p>
 * <b>The child node will need to be post-processed if it contains a real
 * Progress 4GL expression.</b>
 */
valExp
   :
      k:KW_VALEXP^ s:STRING
      {
         String rawText = #s.getText();
         String text = rawText;
         
         // store the validation expression in a form suitable for compilation as a string
         // literal in a Java class; this ultimately will end up as a DMO annotation; here we
         // have to replace unsafe 4GL characters into a form javac can deal with
         text = character.progressToJavaString(text);
         #k.putAnnotation("original_text", text);
         
         // set the raw text with minimal preprocessing back into the node for post-processing
         // later
         text = removeBogusQuotes(rawText);
         text = text.replaceAll("\"\"", "\"");
         #s.setText(text);
         #k.setType(VALEXP);
      }
   ;
   
/**
 * Match a set of tokens assumed from the caller's context to be the value
 * of the <code>COLUMN-LABEL</code> property of a Progress field.
 * <p>
 * The resulting sub-tree is <code>KW_COL_LAB</code> node with a child
 * node of <code>STRING</code>.
 */
columnLabel
   :
      KW_COL_LAB^ STRING
   ;
   
/**
 * Match a set of tokens as the value of the <code>HELP</code> property of a
 * field.
 * <p>
 * The resulting sub-tree is the <code>KW_HELP</code> with a child of
 * <code>STRING</code>.
 */
help
   :
      KW_HELP^ STRING
   ;
   
/**
 * Match a set of tokens as the value of the <code>EXTENT</code> property of a
 * field.
 * <p>
 * The resulting sub-tree is the <code>KW_EXTENT</code> with a child of
 * <code>NUM_LITERAL</code>.
 */
extent
   :
      KW_EXTENT^ NUM_LITERAL
   ;
   
/**
 * Match a set of tokens as the value of the <code>DECIMALS</code> property of
 * a field.
 * <p>
 * The resulting sub-tree is the <code>KW_DECIMALS</code> with a child of
 * <code>NUM_LITERAL</code>.
 */
decimals
   :
      KW_DECIMALS^ NUM_LITERAL
   ;
   
/**
 * Match a set of tokens as the value of the <code>LENGTH</code> property of a
 * field.
 * <p>
 * The resulting sub-tree is the <code>KW_LENGTH</code> with a child of
 * <code>NUM_LITERAL</code>.
 */
length
   :
      KW_LENGTH^ NUM_LITERAL
   ;
   
/**
 * Match a set of tokens as the value of the <code>VIEW-AS</code> property of
 * a field. Strangely enough, the <code>KW_VIEW_AS</code> keyword is repeated
 * inside the <code>STRING</code> text.  That is just a Progress 4GL quirk.
 * <p>
 * The resulting sub-tree is the <code>KW_VIEW_AS</code> with a child of
 * <code>STRING</code>.
 * <p>
 * <b>The string literal child will need post-processing to convert it into
 * the full view-as phrase.</b>
 */
viewAs
   :
      KW_VIEW_AS^ s:STRING
      {
         // DO NOT use remoteBogusQuotes() here, parseProgressCharLiteral() is required to allow
         // more complex uses of escaped single and double quotes inside the string; the bottom
         // line is that we need the unescaped runtime version for downstream re-parsing
         #s.setText(character.parseProgressCharLiteral(#s.getText(), !unixEsc));
      }
   ;
   
/**
 * Match a set of tokens as the value of the <code>ORDER</code> property of a
 * field.
 * <p>
 * The resulting sub-tree is an artificial <code>ORDER</code> with a child of
 * <code>NUM_LITERAL</code>. The <code>KW_ORDER</code> is dropped.
 */
order
   :
      k:KW_ORDER^ NUM_LITERAL
      { #k.setType(ORDER); }
   ;
   
/**
 * Match permissions clause as a property of a table or field.
 * <p>
 * The resulting sub-tree will have the permission keyword (e.g. 
 * <code>KW_CAN_READ</code> as the root node and the <code>STRING</code> as
 * the child node.
 */
permissions
   :
      (
           KW_CAN_CREA^
         | KW_CAN_DEL^
         | KW_CAN_READ^
         | KW_CAN_WRT^
         | KW_CAN_DUMP^
         | KW_CAN_LOAD^
      )
      STRING
   ;
   
/**
 * Match a <code>TABLE-TRIGGER</code> clause in the {@link #table_props}.
 * Uses {@link #trigger_parms}.
 * <p>
 * The sub-tree will have the <code>KW_TAB_TRG</code> as the root and the
 * sub-tree created by <code>trigger_parms</code> will be the child.
 */
table_trigger
   :
      KW_TAB_TRG^ trigger_parms
   ;
   
/**
 * Match a <code>FIELD-TRIGGER</code> clause in the {@link #field}.
 * Uses {@link #trigger_parms}.
 * <p>
 * The sub-tree will have the <code>KW_FLD_TRG</code> as the root and the
 * sub-tree created by <code>trigger_parms</code> will be the child.
 */
field_trigger
   :
      KW_FLD_TRG^ trigger_parms
   ;
   
/**
 * Match parameters for a table or field trigger in {@link #table_trigger} and
 * {@link #field_trigger}, respectively.  Uses {@link #proc_clause} and {@link
 * #crc_clause}.
 * <p>
 * Each of the matched tokens will be attached (in order of occurrence) to the
 * tree of the calling rule.
 */
trigger_parms
   :
      STRING (KW_OVERRIDE | KW_NO_OVRRD)
      proc_clause
      crc_clause
   ;
   
/**
 * Matches the <code>PROCEDURE</code> keyword followed by a string literal.
 * Used by {@link #table_trigger} and separated to get the tree structure
 * correct.
 * <p>
 * The sub-tree will have the <code>KW_PROC</code> as the root and the
 * <code>STRING</code> will be the child.
 */
proc_clause
   :
      KW_PROC^ STRING
   ;
   
/**
 * Matches the <code>CRC</code> keyword followed by a string literal.
 * Used by {@link #table_trigger} and separated to get the tree structure
 * correct.
 * <p>
 * The sub-tree will have the <code>KW_CRC</code> as the root and the
 * <code>STRING</code> will be the child.
 */
crc_clause
   :
      KW_CRC^ STRING
   ;
   
/**
 * Match options that make no sense.  This may be related to newer features of
 * v10 which may be forcing all possible options to be output, even those that
 * make no sense for the given type and/or are undefined. The calling code will
 * actually drop the matched subtree.
 */
cruft
   :
      (
           d:KW_DECIMALS UNKNOWN_VAL
         | v1:KW_VALEXP UNKNOWN_VAL
         | v2:KW_VIEW_AS UNKNOWN_VAL
      )
      {
         Aast n = (#d != null) ? #d : ((#v1 != null) ? #v1 : #v2);
         outputWarning(#n, "WARNING: ignoring cruft ", " with value ?");
      }
   ;
   
/**
 * Match any other unrecognized option and warn that it is being dropped. The
 * calling code will actually drop the matched subtree.
 */
unknownOption
   :
      s:SYMBOL^ l:literal
      {
         outputWarning(#s,
                       "WARNING: ignoring unknown option ",
                       " with value " + #l.getText());
      }
   ;
   
/**
 * Matches a single instance of any and all legal Progress 4GL literals
 * (hard-coded instances of data types, otherwise called constants).  Such
 * constants are properly tokenized by the lexer and this rule simply allows
 * any of the following tokens to be matched:
 * <pre>
 *    NUM_LITERAL
 *    DEC_LITERAL
 *    STRING
 *    BOOL_TRUE
 *    BOOL_FALSE 
 *    DATE_LITERAL
 *    DATETIME_TZ_LITERAL
 *    UNKNOWN_VAL
 * </pre>
 * <p>
 * Note that the documentation states that there is no such thing as a date
 * constant in Progress 4GL, however experience shows that this is not true.
 * <p>
 * There are no literal representations for RECID, ROWID or HANDLE data
 * types since these are internal data structures of the Progress language.
 * One can only obtain an instance of such a value from Progress itself so
 * there is no way to hard-code such a reference as a constant.
 * <p>
 * There is no literal representation for RAW data type.
 * <p>
 * <b>WARNING: this rule is simply copied from the grammar for the
 * <code>ProgressParser</code>!</b>
 * <p>
 * The sub-tree created is a single node of the type and text as created by the
 * lexer.
 */
literal
   :
      (
           NUM_LITERAL
         | DEC_LITERAL
         | STRING
         | BOOL_TRUE
         | BOOL_FALSE 
         | DATE_LITERAL
         | DATETIME_TZ_LITERAL
         | UNKNOWN_VAL
      )
   ;
   
/**
 * Match any kind of text as a symbol, even if the text would not normally be allowed to start
 * a symbol (e.g. like a number or some punctuation). The resulting token will be of type
 * SYMBOL.
 */
any_ridiculous_symbol
   :
      {
         mergeUntilWS();
         LT(1).setType(SYMBOL);
      }
      (
           SYMBOL^
         
           // this reference is for "pull" only, its corresponding type
           // is rewritten unconditionally at entry to this rule so it
           // will never get executed
         | ~EOF
      )
   ;
