/*
** Module   : P2JCollatorProvider.java
** Abstract : Collator SPI implementation
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20080321 ADD  @37637  Created initial version. A service provider
**                               for specialized collators which are governed
**                               by Progress-like collation rules.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.spi;

import java.text.*;
import java.text.spi.*;
import java.util.*;

/**
 * A service provider for specialized collators which are governed by
 * Progress-like collation rules.
 */
public class P2JCollatorProvider
extends CollatorProvider
{
   /** Mapping of locales to fully qualified collator class names */
   private static final Map<Locale, String> collatorClasses =
      new HashMap<Locale, String>(3);
   
   /** All collator implementations managed by this provider */
   private static final Map<Locale, Collator> collators =
      new HashMap<Locale, Collator>(3);
   
   static
   {
      // Register all supported collator implementation class names.
      try
      {
         collatorClasses.put(new Locale("en", "US", "P2J"),
                             "com.goldencode.p2j.spi.ISO_8859_1Collator");
      }
      catch (Exception exc)
      {
         throw new RuntimeException(
            "Error initializing P2J collator(s)", exc);
      }
   }
   
   /**
    * Default constructor.
    */
   public P2JCollatorProvider()
   {
   }
   
   /**
    * Get an instance of a collator, if this provider is responsible for the
    * specified locale.  The collator is lazily instantiated and cached for
    * future requests.
    * 
    * @param   locale
    *          A locale possibly managed by this provider.
    * 
    * @return  Appropriate collator, or <code>null</code> if the specified
    *          locale is unknown to this provider.
    */
   public Collator getInstance(Locale locale)
   {
      synchronized (collators)
      {
         Collator collator = collators.get(locale);
         
         if (collator == null)
         {
            String className = collatorClasses.get(locale);
            if (className == null)
            {
               return null;
            }
            
            try
            {
               collator = (Collator) Class.forName(className).newInstance();
               collators.put(locale, collator);
            }
            catch (Exception exc)
            {
               // TODO:  what is the appropriate way to log using the Java
               // Extension Mechanism?
               exc.printStackTrace();
               
               return null;
            }
         }
         
         return collator;
      }
   }

   /**
    * Report all locales managed by this provider.
    * 
    * @return  Array of locales.
    */
   public Locale[] getAvailableLocales()
   {
      Set<Locale> set = collatorClasses.keySet();
      
      return set.toArray(new Locale[set.size()]);
   }

}