/*
** Module   : GenerateClassMap.java
** Abstract : command line utility to generate the class map file
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description-----------------
** 001 GES 20170706 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

import com.goldencode.util.*;

/**
 * Generates the Class Map file by searching for OO 4GL class files in a
 * given set of paths. 
 */
public class GenerateClassMap
{
   /**
    * Store the given class name map into the class name map XML file.
    * <p>
    * Each map element "entry" stores a "key" and "value" that maps a class
    * name in package form to a project relative filename of the associated
    * class file.
    * 
    * @param   map
    *          Map which describes the class name mapping.
    * 
    * @return  <code>true</code> if the operation was successful.
    */
   private static boolean writeClassMap(Map<String, String> map)
   {
      Document dom = null;
      
      try
      {
         dom = XmlHelper.newDocument();
      }
      catch (ParserConfigurationException e)
      {
         e.printStackTrace();
         return false;
      }

      Comment comment;

      comment = dom.createComment("OO 4GL Class Name Mapping File");
      dom.appendChild(comment);                                      

      comment = dom.createComment("Created at " + (new Date()));
      dom.appendChild(comment);

      Element root = dom.createElement(XmlHelper.MAP_ELEM_ROOT);
      dom.appendChild(root);

      Iterator<String> iter = new TreeSet(map.keySet()).iterator();
      
      while (iter.hasNext())
      {
         Element elem  = dom.createElement(XmlHelper.MAP_ELEM_ENTRY);
         String  key   = iter.next();
         String  value = map.get(key);
         
         elem.setAttribute(XmlHelper.MAP_ATTR_KEY, key);
         elem.setAttribute(XmlHelper.MAP_ATTR_VALUE, value);
         
         root.appendChild(elem);
      }

      try
      {
         XmlHelper.write(dom, SymbolResolver.CLASS_MAP_FILE);
      }
      catch (IOException e)
      {
         e.printStackTrace();
         return false;
      }

      return true;
   }
   
   /**
    * Command line runner. The result of execution will overwrite the existing
    * class map file.
    *
    * @param    args
    *           A list of one or more project relative "root package" paths. This is
    *           required even if it is just set to ".". The idea is that from this
    *           path any files with a .cls extension will be found and their relative
    *           path (after the root package path) will be the entry value stored in
    *           the map.
    */
   public static void main(String[] args)
   {
      Map<String, String> map = new HashMap();
      
      boolean curDir = false;
      
      for (int i = 0; i < args.length; i++)
      {
         if (args[i] != null)
         {
            String srch = args[i];
            
            if (srch.length() == 0 || srch.equals("."))
            {
               // only search the current directory once
               if (curDir)
                  continue;
               
               srch   = ".";
               curDir = true;
            }
            
            if (srch.charAt(srch.length() - 1) != File.separatorChar)
            {
               srch = srch + File.separator;
            }
            
            String[] list = SymbolResolver.getClassFileList(srch, false);
            
            SymbolResolver.createClassMappings(map, list, srch, true);
         }
      }
      
      if (!map.isEmpty())
      {
         File mapfile = new File(SymbolResolver.CLASS_MAP_FILE);
         
         if (mapfile.exists())
         {
            mapfile.delete();
         }
         
         writeClassMap(map);
      }
      else
      {
         System.out.println("Nothing to do!");
      }
   }
}