/*
** Module   : ChooseData.java
** Abstract : choose data for the client side
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20051207   @23700 Created initial version.
** 002 SIY 20060220   @24628 Changes necessary for choose()
**                           implementation.
** 003 GES 20060623   @27554 Added keyCode support to enable LASTKEY.
** 004 GES 20060630   @27733 Changed use of keyCode to be a screen buffer
**                           which will handle the LASTKEY requirement
**                           AND also meet the more general purpose that
**                           the data for the current down frame iteration
**                           (ROW mode) or for the frame in general (FIELD
**                           mode) needs to be updated in both directions
**                           on a CHOOSE.
** 005 SIY 20060622   @28140 Renamed isRowField() to isFieldMode().
** 006 GES 20060807   @28460 Moved to externalizable interface to optimize
**                           network performance.
** 007 SIY 20110112          Changes related to new color handling scheme.
** 008 SIY 20111101          Updated imports.
** 009 EVL 20160726          Changed usage of the screen buffers
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;


/**
 * This class is used as a container in which to transfer state between 
 * the client and server for the <b>CHOOSE</b> language statement.
 */
public class ChooseData
implements Externalizable
{
   /** <code>true</code> if in FIELD mode, <code>false</code> for ROW mode. */
   private boolean fieldMode = false;

   /** Holds the field widget in the ROW mode */
   private int rowId = -1;

   /** Holds the HELP text for the ROW item */
   private String rowHelp = null;

   /** AUTO-RETURN option  */
   private boolean autoReturn = false;

   /** Custom color */
   private Color color = Color.MESSAGES;

   /** highlight and navigation variable */
   private String keysVar = null;

   /** NO-ERROR option  */
   private boolean noError = false;

   /** CHOOSE timeout */
   private int pause = -1;
   
   /** Holds the IDs of field widgets in the FIELD mode */
   private int[] fields = null;

   /** Holds the HELP text for the FIELD items */
   private String[] fieldHelps = null;

   /** Holds the GO-ON keys */
   private String[] goOn = null;
   
   /** The current record's data for all editable frames. */
   private ScreenBuffer[] sb = null;

   /**
    * Default constructor (only used for de-serialization).
    */
   public ChooseData()
   {
   }
   
   /**
    * Construct instance with specified size of the internal arrays.
    * 
    * @param   size
    *          Size of the internal arrays.
    * @param   gosize
    *          Size of the GO-ON keys list.
    */
   public ChooseData(int size, int gosize)
   {
      if (size >= 0)
      {
         fields = new int[size];
         fieldHelps = new String[size];
      }
      
      if (gosize >= 0)
         goOn = new String[gosize];
   }
   
   /**
    * Get AUTO-RETURN state.
    * 
    * @return   Returns the state of auto-return processing.
    */
   public boolean isAutoReturn()
   {
      return autoReturn;
   }
   
   /**
    * Return color assigned to this <b>CHOOSE </b> statement.
    * 
    * @return   Returns the color.
    */
   public Color getColor()
   {
      return color;
   }
   
   /**
    * Returns array of help strings assigned for each field.
    * 
    * @return   Returns the fieldHelps.
    */
   public String[] getFieldHelps()
   {
      return fieldHelps;
   }

   /**
    * Returns help string assigned for particular field.
    * 
    * @param   index
    *          Index of the field for which help string is requested.
    * 
    * @return  Help text assigned to field.
    */
   public String getFieldHelp(int index)
   {
      return fieldHelps[index];
   }
   
   /**
    * Return array of widget IDs of the fields.
    * 
    * @return  Returns the fields.
    */
   public int[] getFields()
   {
      return fields;
   }
   
   /**
    * Returns ID of the particular field.
    * 
    * @param   index
    *          Index of the field for which ID is requested.
    * 
    * @return  Field ID.
    */
   public int getField(int index)
   {
      return fields[index];
   }
   
   /**
    * Returns array of <b>GO-ON </b> key labels.
    * 
    * @return  Returns the array of labels.
    */
   public String[] getGoOn()
   {
      return goOn;
   }
   
   /**
    * Returns highlight and navigation variable.
    * 
    * @return  Returns the variable.
    */
   public String getKeys()
   {
      return keysVar;
   }
   
   /**
    * Returns <b>NO-ERROR </b> flag state.
    * 
    * @return  Returns the flag state.
    */
   public boolean isNoError()
   {
      return noError;
   }
   
   /**
    * Returns amount of time in seconds provided to perform the processing.
    * Negative value means that processing will be performed as long as
    * necessary.
    * 
    * @return  Returns the time in seconds.
    */
   public int getPause()
   {
      return pause;
   }
   
   /**
    * Gets the mode of the <code>CHOOSE</code>. <code>true</code> means
    * <code>FIELD</code> mode and <code>false</code> means <code>ROW</code>
    * mode.
    * 
    * @return   The mode of the <code>CHOOSE</code>.
    */
   public boolean isFieldMode()
   {
      return fieldMode;
   }
   
   /**
    * Returns help string assigned for the <b>CHOOSE </b> in <b>ROW </b> mode.
    * 
    * @return  Returns the help text.
    */
   public String getRowHelp()
   {
      return rowHelp;
   }
   
   /**
    * Returns ID of the widget in <b>ROW </b> mode.
    * 
    * @return  Returns the row widget ID. <b>-1 </b> means no ID is assigned.
    */
   public int getRowId()
   {
      return rowId;
   }
   
   /**
    * Gets the screen buffer for the associated frame and all editable ones.
    * 
    * @return   The screen buffer.
    */
   public ScreenBuffer[] getScreenBuffers()
   {
      return sb;
   }
   
   /**
    * Sets the screen buffer array for the associated frame and all editable ones.
    * 
    * @param    sb
    *           The screen buffer array.
    */
   public void setScreenBuffers(ScreenBuffer[] sb)
   {
      this.sb = sb;
   }
   
   /**
    * Set <b>AUTO-RETURN</b> value.
    * 
    * @param    autoReturn 
    *           New value for the <b>AUTO-RETURN</b> attribute.
    */
   void setAutoReturn(boolean autoReturn)
   {
      this.autoReturn = autoReturn;
   }
   
   /**
    * Set color specification.
    * 
    * @param   color
    *          New color specification.
    */
   void setColor(ColorSpec color)
   {
      if (color != null)
         this.color = color.convert();
   }
   
   /**
    * Set help string for specified field.
    * 
    * @param   index
    *          Field index.
    * @param   text
    *          Help text.
    */
   void setFieldHelp(int index, String text)
   {
      this.fieldHelps[index] = text;
   }
   
   /**
    * Set widget ID for specified field.
    * 
    * @param   index
    *          Field index.
    * @param   id
    *          Field widget ID.
    */
   void setFields(int index, int id)
   {
      this.fields[index] = id;
   }
   
   /**
    * Set <b>GO-ON </b> key for specified index.
    * 
    * @param   index
    *          Index of the key.
    * @param   key
    *          New key value.
    */
   void setGoOn(int index, String key)
   {
      this.goOn[index] = new String(key);
   }
   
   /**
    * Set highlight and navigation variable value.
    * 
    * @param   keysVar
    *          New value for the variable.
    */
   public void setKeys(String keysVar)
   {
      if (keysVar == null)
         return;
      
      this.keysVar = keysVar;
   }
   
   /**
    * Set <b>NO-ERROR </b> flag value.
    * 
    * @param   noError
    *          New value of the flag.
    */
   void setNoError(boolean noError)
   {
      this.noError = noError;
   }
   
   /**
    * Set pause time in seconds.
    * 
    * @param   pause
    *          New value of the variable.
    */
   void setPause(int pause)
   {
      this.pause = pause;
   }
   
   /**
    * Set the mode of the <code>CHOOSE</code>. <code>true</code> selects
    * <code>FIELD</code> mode and <code>false</code> selects <code>ROW</code>
    * mode.
    * 
    * @param    fieldMode
    *           New mode flag.
    */
   void setFieldMode(boolean fieldMode)
   {
      this.fieldMode = fieldMode;
   }
   
   /**
    * Set help text for the <b>ROW </b> mode.
    * 
    * @param   rowHelp
    *          New help text.
    */
   void setRowHelp(String rowHelp)
   {
      this.rowHelp = rowHelp;
   }
   
   /**
    * Set field widget ID for the <b>ROW </b> mode.
    * 
    * @param   rowId
    *          New field widget ID.
    */
   void setRowId(int rowId)
   {
      this.rowId = rowId;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      rowId      = in.readInt();
      color      = (Color) in.readObject();
      pause      = in.readInt();
      fieldMode  = in.readBoolean();
      autoReturn = in.readBoolean();
      noError    = in.readBoolean();
      
      // the rest of the data are objects which can be null so we
      // read an indicator var first to determine if there is following
      // data to be read
      
      if (in.readBoolean())
         rowHelp = (String) in.readObject();
      
      if (in.readBoolean())
         keysVar = (String) in.readObject();
      
      if (in.readBoolean())
         fields = (int[]) in.readObject();
      
      if (in.readBoolean())
         fieldHelps = (String[]) in.readObject();
      
      if (in.readBoolean())
         goOn = (String[]) in.readObject();
      
      if (in.readBoolean())
         sb = (ScreenBuffer[]) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(rowId);
      out.writeObject(color);
      out.writeInt(pause);
      out.writeBoolean(fieldMode);
      out.writeBoolean(autoReturn);
      out.writeBoolean(noError);
      
      // the rest of the data are objects which can be null so we
      // write an indicator var first to specify if there is following
      // data to be read
      
      boolean there = (rowHelp != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(rowHelp);
      
      there = (keysVar != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(keysVar);
      
      there = (fields != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(fields);
      
      there = (fieldHelps != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(fieldHelps);
      
      there = (goOn != null);
      out.writeBoolean(there);
      if (there)
         out.writeObject(goOn);
      
      there = (sb != null);
      out.writeBoolean(there);
      if (there)
         out.writeObject(sb);
   }
}