/*
** Module   : ColorHSL.java
** Abstract : a simple color structure containing HSL values
**
** Copyright (c) 2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20171028 Created initial version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

/**
 * A simple color structure in HSL color-space. 
 */
public class ColorHSL
{
   /** The Hue HSL component (if computed, -1 otherwise). */
   private int hue = -1;
   
   /** The Saturation HSL component (if computed, -1 otherwise). */
   private int saturation = -1;
   
   /** The Luminosity HSL component (if computed, -1 otherwise). */
   private int luminosity = -1;

   /**
    * Default constructor. Creates a black color. 
    */
   public ColorHSL()
   {
      this(0, 0, 0);
   }
   
   /**
    * Constructor. Creates a custom color in HSL color space. 
    * 
    * @param   hue
    *          The hue.
    * @param   saturation
    *          The saturation.
    * @param   luminosity
    *          The luminosity.
    */
   public ColorHSL(int hue, int saturation, int luminosity)
   {
      this.hue = hue;
      this.saturation = saturation;
      this.luminosity = luminosity;
   }
   
   /**
    * Reports if the given object is the same color as this instance. This does not just
    * compare references, but instead it compares the contained HSL values.
    *
    * @param    obj
    *           The object to compare.
    *
    * @return   {@code true} if the objects are equivalent, {@code false} if the given instance is
    *           {@code null}, is not an instance of this class or if its HSL value is not
    *           equivalent.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
      {
         return true;
      }
      
      if (obj == null || !getClass().isAssignableFrom(obj.getClass()))
      {
         return false;
      }
      
      ColorHSL other = (ColorHSL) obj;
   
      return hue == other.hue && 
             saturation == other.saturation && 
             luminosity == other.luminosity;
   }
   
   /**
    * Calculate hash code compatible with equals()
    * 
    * @return hash code
    */
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      result = prime * result + hue;
      result = prime * result + saturation;
      result = prime * result + luminosity;
      
      return result;
   }
   
   /**
    * Obtains hue value from HSL color space.
    *
    * @return  the hue value in 0-240 interval. 
    */
   public int getHue()
   {
      return hue;
   }
   
   /**
    * Obtains saturation value from HSL color space.
    *
    * @return  the saturation value in 0-240 interval. 
    */
   public int getSaturation()
   {
      return saturation;
   }
   
   /**
    * Obtains luminosity value from HSL color space.
    *
    * @return  the luminosity value in 0-240 interval. 
    */
   public int getLuminosity()
   {
      return luminosity;
   }
   
   /**
    * Sets hue value from HSL color space.
    *
    * @param   hue
    *          the hue value in 0-240 interval. 
    */
   public void setHue(int hue)
   {
      this.hue = hue;
   }
   
   /**
    * Sets hue value from HSL color space.
    *
    * @param   saturation
    *          the saturation value in 0-240 interval. 
    */
   public void setSaturation(int saturation)
   {
      this.saturation = saturation;
   }
   
   /**
    * Sets luminosity value from HSL color space.
    *
    * @param   luminosity
    *          the luminosity value in 0-240 interval. 
    */
   public void setLuminosity(int luminosity)
   {
      this.luminosity = luminosity;
   }
   
   /**
    * Computes the HSL values from the objects RGB. Saves the result in volatile member fields. 
    */
   public static ColorHSL fromRGB(ColorRgb rgbColor)
   {
      float r = rgbColor.getRed() / 255F;
      float g = rgbColor.getGreen() / 255F;
      float b = rgbColor.getBlue() / 255F;
      float max = Math.max(r, Math.max(g, b));
      float min = Math.min(Math.min(r, g), b);
      float l = (max + min) / 2;
      float h, s; 
      
      if (max == min)
      {
         h = s = 0; // achromatic
      }
      else
      {
         float d = max - min;
         s = (l > 0.5) ? d / (2 - max - min) : d / (max + min);
         
         if (max == r)
         {
            h = (g - b) / d + (g < b ? 6 : 0);
         }
         else if (max == g)
         {
            h = (b - r) / d + 2;
         }
         else // max == b
         {
            h = (r - g) / d + 4;
         }
         h /= 6;
      }
      
      int hue = Math.round(h * 240);
      int saturation = Math.round(s * 240);
      int luminosity = Math.round(l * 240);
      return new ColorHSL(hue, saturation, luminosity);
   }
   
   /**
    * Convert the current color from HSL color space to RGB.
    * 
    * @return  A {@link ColorRgb} object representing the same color but in RGB. 
    */
   public ColorRgb toRGB()
   {
      float h = hue / 240F;
      float s = saturation / 240F;
      float l = luminosity / 240F;
      float r, g, b;
      
      if (s == 0)
      {
         r = g = b = l; // achromatic
      }
      else
      {
         float q = l < 1 / 2F ? l * (1 + s) : l + s - l * s;
         float p = 2 * l - q;
         r = hue2rgb(p, q, h + 1 / 3F);
         g = hue2rgb(p, q, h);
         b = hue2rgb(p, q, h - 1 / 3F);
      }
      
      int red = Math.round(r * 255);
      int green = Math.round(g * 255);
      int blue = Math.round(b * 255);
      
      return new ColorRgb(red, green, blue);
   }
   
   /**
    * Internal worker method used by {@code toRGB()} method.
    * 
    * @param   p
    *          Intermediary result in HSL to RGB conversion. 
    * @param   q
    *          Intermediary result in HSL to RGB conversion. 
    * @param   t
    *          Intermediary result in HSL to RGB conversion. 
    * 
    * @return  a color component.
    */
   private float hue2rgb(float p, float q, float t)
   {
      if (t < 0)
      {
         t += 1;
      }
      if (t > 1)
      {
         t -= 1;
      }
      if (t < 1 / 6F)
      {
         return p + (q - p) * 6 * t;
      }
      if (t < 1 / 2F)
      {
         return q;
      }
      if (t < 2 / 3D)
      {
         return p + (q - p) * (2 / 3F - t) * 6;
      }
      return p;
   }
}
