/*
** Module   : EditorConfig.java
** Abstract : stores Editor component state
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 SIY 20060428   @26085 Created initial version.
** 002 GES 20060803   @28405 Made externalizable for better performance. Coding std cleanups.
** 003 SIY 20060810   @28645 Moved code which validates and adjusts configuration variables 
**                           from Editor.java.
** 004 SIY 20100619          Switching to own implementation of character UI.
** 005 SIY 20100804          Changes in imports, some refactorings.
** 006 SIY 20111101          Moved from CHUI package. Minor cleanups/refactorings in preparation 
**                           for GUI client.
** 007 HC  20140911          GUI coordinate support - data types refactoring.
** 008 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 009 HC  20141021          Implemented window move/position support.
** 010 GES 20141124          Javadoc change.
** 011 GES 20150103          Moved state here from the client class to enable cursor-line and
**                           cursor-char attributes to be read/written on the server side. These
**                           are now calculated values since they don't directly correspond to
**                           the client-side state.
** 012 IAS 20150325          Added returnInserted attribute and selectionActive flag
** 013 HC  20150323          Removed business logic from config classes.
** 014 IAS 20150604          Moved selection-related field to the ControlConfig
** 015 CA  20150615          MAX-CHARS can be unknown, so its type is know set to Integer instead
**                           of native int.
** 016 IAS 20151214          MOUSE-POINTER attribute support.
** 017 IAS 20160322          Changed the constructor visibility.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * The configuration container class which holds attributes specific for the Editor widget.
 */
public class EditorConfig
extends ControlTextConfig
{
   /** BUFFER-CHARS Attribute */
   public int bufferChars = -1;
   
   /** BUFFER-LINES Attribute */
   public int bufferLines = -1;
   
   /** CURSOR-OFFSET Attribute */
   public int cursorOffset = -1;
   
   /** CURSOR-CHAR Attribute */
   public int cursorChar = -1;

   /** Cursor X position relative to left edge of the inner rectangle of widget (0-based). */
   public int cursorX = 0;
   
   /** Cursor Y position relative to to edge of the inner rectangle of widget (0-based). */
   public int cursorY = 0;
   
   /** Horizontal scrolling index of the text (0-based). */
   public int scrollX = 0;
   
   /** Vertical scrolling index of the text (0-based). */
   public int scrollY = 0;

   /** INNER-CHARS Attribute */
   public int innerChars = -1;
   
   /** The number of text lines in the widget. */
   public int innerLines = -1;
   
   /** LENGTH Attribute */
   public int length = 0;
   
   /** MAX-CHARS Attribute */
   public Integer maxChars = null;
   
   /** NUM-LINES Attribute */
   public int numLines = 0;
   
   /** NUM-REPLACED Attribute */
   public int numReplaced = 0;
  
   /** AUTO-INDENT Attribute */
   public boolean autoIndent = false;
   
   /** EMPTY Attribute */
   public boolean empty = false;
   
   /** SCROLLBAR-HORIZONTAL Attribute */
   public boolean scrollbarHorizontal = false;
   
   /** SCROLLBAR-VERTICAL Attribute */
   public boolean scrollbarVertical = false;
   
   /** WORD-WRAP Attribute */
   public boolean wordWrap = true;
   
   /** BOX attribute */
   public boolean box = true;

   /** SELECTION-TEXT Attribute */
   public String selectionText = null;
   
   /** RETURN-INSERTED attribute */
   public boolean returnInserted = false;
   
   /** TABLE Attribute */
   public String table = null;
   
   /** MOUSE-POINTER attribute */
   public String mousePointer = null;

   /**
    * Default constructor (only used in deserialization).
    */
   public EditorConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public EditorConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected EditorConfig(int id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      EditorConfig cfg = (EditorConfig) config;
      
      bufferChars = cfg.bufferChars;
      bufferLines = cfg.bufferLines;
      cursorX = cfg.cursorX;
      cursorY = cfg.cursorY;
      scrollX = cfg.scrollX;
      scrollY = cfg.scrollY;
      cursorChar = cfg.cursorChar;
      cursorOffset = cfg.cursorOffset;
      innerChars = cfg.innerChars;
      innerLines = cfg.innerLines;
      length = cfg.length;
      maxChars = cfg.maxChars;
      numLines = cfg.numLines;
      numReplaced = cfg.numReplaced;
      autoIndent = cfg.autoIndent;
      empty = cfg.empty;
      scrollbarHorizontal = cfg.scrollbarHorizontal;
      scrollbarVertical = cfg.scrollbarVertical;
      wordWrap = cfg.wordWrap;
      box = cfg.box;
      selectionText = cfg.selectionText;
      table = cfg.table;
      returnInserted = cfg.returnInserted;
      mousePointer = cfg.mousePointer;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      bufferChars         = in.readInt();
      bufferLines         = in.readInt();
      cursorX             = in.readInt();
      cursorY             = in.readInt();
      scrollX             = in.readInt();
      scrollY             = in.readInt();
      cursorChar          = in.readInt();
      cursorOffset        = in.readInt();
      innerChars          = in.readInt();
      innerLines          = in.readInt();
      length              = in.readInt();
      maxChars            = (Integer) in.readObject();
      numLines            = in.readInt();
      numReplaced         = in.readInt();
      autoIndent          = in.readBoolean();
      empty               = in.readBoolean();
      scrollbarHorizontal = in.readBoolean();
      scrollbarVertical   = in.readBoolean();
      wordWrap            = in.readBoolean();
      box                 = in.readBoolean();
      selectionText       = (String) in.readObject();
      table               = (String) in.readObject();
      returnInserted      = in.readBoolean();
      mousePointer        = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(bufferChars);
      out.writeInt(bufferLines);
      out.writeInt(cursorX);
      out.writeInt(cursorY);
      out.writeInt(scrollX);
      out.writeInt(scrollY);
      out.writeInt(cursorChar);
      out.writeInt(cursorOffset);
      out.writeInt(innerChars);
      out.writeInt(innerLines);
      out.writeInt(length);
      out.writeObject(maxChars);
      out.writeInt(numLines);
      out.writeInt(numReplaced);
      out.writeBoolean(autoIndent);
      out.writeBoolean(empty);
      out.writeBoolean(scrollbarHorizontal);
      out.writeBoolean(scrollbarVertical);
      out.writeBoolean(wordWrap);
      out.writeBoolean(box);
      out.writeObject(selectionText);
      out.writeObject(table);
      out.writeBoolean(returnInserted);
      out.writeObject(mousePointer);
  }
}