/*
** Module   : EditorInterface.java
** Abstract : Define EDITOR-specific widget attributes and methods available via a handle 
**            instance.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20141120 Created initial version.
** 002 EVK 20150209 Added conversion support for SAVE-FILE method.
** 003 IAS 20150318 Added runtime support for new attrs/methods (
**                  BOX, RETURN-INSERTED. CURSOR-XXX, REPLACE...)
** 004 IAS 20150604 Moved REPLACE and text selection methods to a separate interface
** 005 CA  20150615 Added support for WORD-WRAP.
** 006 OM  20170504 Added support for MOVE-TO-EOF, SEARCH and INSERT-FILE.
** 007 EVL 20170728 Adding javadoc fixes.
** 008 CA  20180502 Added LARGE attribute.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Define all EDITOR-specific widget attributes and methods which can be accessed via a handle
 * instance.  Attributes and methods shared by multiple widget types will be defined elsewhere.
 */
public interface EditorInterface
{
   /**
    * Implements the DELETE-LINE() method which deletes the line of text in which the cursor is
    * positioned.
    *
    * @return   <code>true</code> if the line was deleted.
    */
   @LegacyMethod(name = "DELETE-LINE")
   public logical deleteLine();
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   @LegacyMethod(name = "INSERT-STRING")
   public logical insertString(character txt);
   
   /**
    * Implements the INSERT-STRING() method which inserts the given text into the editor at the
    * location at which the cursor is positioned.
    *
    * @param    txt
    *           The text to be inserted.
    *
    * @return   <code>true</code> if the text was inserted.
    */
   @LegacyMethod(name = "INSERT-STRING")
   public logical insertString(String txt);
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   @LegacyMethod(name = "READ-FILE")
   public logical readFile(character name);
   
   /**
    * Implements the READ-FILE() method which replaces the editor's contents with those read
    * from the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully read.
    */
   @LegacyMethod(name = "READ-FILE")
   public logical readFile(String name);
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @LegacyMethod(name = "REPLACE-SELECTION-TEXT")
   public logical replaceSelectionText(character txt);
   
   /**
    * Implements the REPLACE-SELECTION-TEXT() method which replaces the editor's currently
    * selected text with the given text.
    *
    * @param    txt
    *           The replacement text.
    *
    * @return   <code>true</code> if the text was successfully replaced.
    */
   @LegacyMethod(name = "REPLACE-SELECTION-TEXT")
   public logical replaceSelectionText(String txt);

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @LegacyMethod(name = "SAVE-FILE")
   public logical saveFile(String name);

   /**
    * Implements the SAVE-FILE() method which write the editor's contents to the specified file.
    *
    * @param    name
    *           The relative or absolute file name.
    *
    * @return   <code>true</code> if the file was successfully written.
    */
   @LegacyMethod(name = "SAVE-FILE")
   public logical saveFile(character name);   
   
   /**
    * Get the value of the CURSOR-CHAR attribute which is the current cursor position as a
    * 1-based index on the current line.
    *
    * @return   The current position of the cursor on the current line.
    */
   @LegacyAttribute(name = "CURSOR-CHAR")
   public integer getCursorChar();
   
   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The position to which to move.
    */
   @LegacyAttribute(name = "CURSOR-CHAR", setter = true)
   public void setCursorChar(double idx);

   /**
    * Set the value of the CURSOR-CHAR attribute which moves the cursor position to the
    * specified 1-based index on the current line.
    *
    * @param    idx
    *           The position to which to move.
    */
   @LegacyAttribute(name = "CURSOR-CHAR", setter = true)
   public void setCursorChar(NumberType idx);

   /**
    * Get the value of the CURSOR-LINE attribute which is the current cursor position as a
    * 1-based line number.
    *
    * @return   The current position of the cursor on the current line.
    */
   @LegacyAttribute(name = "CURSOR-LINE")
   public integer getCursorLine();
   
   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   @LegacyAttribute(name = "CURSOR-LINE", setter = true)
   public void setCursorLine(double line);

   /**
    * Set the value of the CURSOR-LINE attribute which moves the cursor position to the
    * specified 1-based line number.
    *
    * @param    line
    *           The line to which to move.
    */
   @LegacyAttribute(name = "CURSOR-LINE", setter = true)
   public void setCursorLine(NumberType line);
   
   /**
    * Get the value of the NUM-LINES attribute which is the number of lines of text in the
    * editor.
    *
    * @return   The number of lines in the editor.
    */
   @LegacyAttribute(name = "NUM-LINES")
   public integer getNumLines();

   /**
    * Get the value of the RETURN-INSERTED attribute
    *
    * @return   the value of the attribute
    */
   @LegacyAttribute(name = "RETURN-INSERTED")
   public logical isReturnInserted();

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @LegacyAttribute(name = "RETURN-INSERTED", setter = true)
   public void setReturnInserted(logical value);

   /**
    * Set the value of the RETURN-INSERTED attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @LegacyAttribute(name = "RETURN-INSERTED", setter = true)
   public void setReturnInserted(boolean value);

   /**
    * Get the value of the WORD-WRAP attribute
    *
    * @return   the value of the attribute
    */
   @LegacyAttribute(name = "WORD-WRAP")
   public logical isWordWrap();

   /**
    * Set the value of the WORD-WRAP attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @LegacyAttribute(name = "WORD-WRAP", setter = true)
   public void setWordWrap(logical value);

   /**
    * Set the value of the WORD-WRAP attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @LegacyAttribute(name = "WORD-WRAP", setter = true)
   public void setWordWrap(boolean value);

   /**
    * Get the value of the LENGTH attribute
    *
    * @return   the value of the attribute
    */
   @LegacyAttribute(name = "LENGTH")
   public integer getLength();
   
   /**
    * Attempts to insert the content of a text file at the current position of the caret.
    *
    * @param   file
    *          The file whose content should be inserted. Full or relative path may be used. In
    *          latter case, the {@code PROPATH} is used to find the file.
    *
    * @return  if the text insertion succeeded, the method returns {@code yes}.
    */
   @LegacyMethod(name = "INSERT-FILE")
   public logical insertFile(String file);
   
   /**
    * Attempts to insert the content of a text file at the current position of the caret.
    *
    * @param   file
    *          The file whose content should be inserted. Full or relative path may be used. In
    *          latter case, the {@code PROPATH} is used to find the file.
    *
    * @return  if the text insertion succeeded, the method returns {@code yes}.
    */
   @LegacyMethod(name = "INSERT-FILE")
   public logical insertFile(character file);
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *  
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    * 
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @LegacyMethod(name = "SEARCH")
   public logical search(String pattern, long mode);
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @LegacyMethod(name = "SEARCH")
   public logical search(character pattern, long mode);
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @LegacyMethod(name = "SEARCH")
   public logical search(String pattern, int64 mode);
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>FIND-NEXT-OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>FIND-PREV-OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>FIND-CASE-SENSITIVE</td><td>4</td></tr>
    *              <tr><td>FIND-WRAP-AROUND</td><td>16</td></tr>
    *              <tr><td>FIND-SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   @LegacyMethod(name = "SEARCH")
   public logical search(character pattern, int64 mode);
   
   /**
    * Moves the caret position at the end of the current text
    * 
    * @return  if the operation is successful, the method returns {@code yes}
    */
   @LegacyMethod(name = "MOVE-TO-EOF")
   public logical moveCaretToEof();
   
   /**
    * Mark this EDITOR widget as EDITOR LARGE, which can be used to display a value of type 
    * <code>LONGCHAR</code>.
    * 
    * @param    large
    *           When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @LegacyAttribute(name = "LARGE", setter = true)
   public void setLarge(boolean large);
   
   /**
    * Mark this EDITOR widget as EDITOR LARGE, which can be used to display a value of type 
    * <code>LONGCHAR</code>.
    * 
    * @param    large
    *           When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @LegacyAttribute(name = "LARGE", setter = true)
   public void setLarge(logical large);
   
   /**
    * Get the EDITOR's LARGE attribute.
    * 
    * @return   When <code>true</code>, this is a EDITOR LARGE widget.
    */
   @LegacyAttribute(name = "LARGE")
   public logical isLarge();
}
