/*
** Module   : Element.java
** Abstract : represents the relationship between frame data and a widget
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060204 NEW   @24268 Created initial version to provide a simple
**                               relationship between frame data and a widget.
** 002 GES 20060915 CHG   @29626 Added support for an optional format string.
** 003 HC  20160714              Fixed runtime support of BROWSE-COLUMN:NAME.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * This class provides a simple association between a widget and the source 
 * and/or sink for the data to be displayed/edited in that widget. 
 */
public class Element
extends FrameElement
{
   /** The source and/or sink of the widget's data. */
   private Accessor data = null;

   /** The widget associated with this element. */
   private GenericWidget widget = null;
   
   /** The value's associated format string or <code>null</code>. */
   private String fmt = null;
   
   /**
    * Create an instance.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(Accessor data, GenericWidget widget)
   {
      this(data, null, widget);
   }
   
   /**
    * Create an instance.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(BaseDataType data, GenericWidget widget)
   {
      this(new AccessorWrapper(data), widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(String data, GenericWidget widget)
   {
      this(new character(data), widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(int data, GenericWidget widget)
   {
      this(new integer(data), widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(double data, GenericWidget widget)
   {
      this(new decimal(data), widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(boolean data, GenericWidget widget)
   {
      this(new logical(data), widget);
   }
   
   /**
    * Create an instance.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(Accessor data, String fmt, GenericWidget widget)
   {
      this.data   = data;
      this.widget = widget;
      this.fmt    = fmt;
   }
   
   /**
    * Create an instance.
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(BaseDataType data, String fmt, GenericWidget widget)
   {
      this(new AccessorWrapper(data), fmt, widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.  Must NOT be
    *           <code>null</code>.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(String data, String fmt, GenericWidget widget)
   {
      this(new character(data), fmt, widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(int data, String fmt, GenericWidget widget)
   {
      this(new integer(data), fmt, widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(double data, String fmt, GenericWidget widget)
   {
      this(new decimal(data), fmt, widget);
   }
   
   /**
    * Create an instance. 
    * <p>
    * <b>WARNING: this form of the constructor allows primative (non-wrapper)
    * types to be copied to the screen buffer BUT any assignment back to such 
    * widgets will silently fail since there is no backing wrapper instance
    * in the calling code to which to assign.</b>
    *
    * @param    data
    *           The data to display/edit in the given widget.
    * @param    fmt
    *           The format string to associate with the data value. May be
    *           <code>null</code>.
    * @param    widget
    *           The widget instance representing the exact frame field and
    *           state of that field.  Must NOT be <code>null</code>.
    */
   public Element(boolean data, String fmt, GenericWidget widget)
   {
      this(new logical(data), fmt, widget);
   }
   
   /**
    * Determines if the element stores a format string in addition to a data
    * value.
    *
    * @return   <code>true</code> if there is a format string encoded in this
    *           instance, <code>false</code> if the format string is
    *           <code>null</code>.
    */
   public boolean hasFormat()
   {
      return (fmt != null);
   }
   
   /**
    * Accesses the formatted value associated with this frame element.
    *
    * @return   The formatted value.
    */
   public FormattedValue getFormattedValue()
   {
      return new FormattedValue(data.get(), fmt); 
   }
   
   /**
    * Accesses the format string associated with this frame element.
    *
    * @return   The format string.
    */
   public String getFormat()
   {
      return fmt; 
   }
   
   /**
    * Accesses the widget associated with this frame element.
    *
    * @return   The widget.
    */
   public GenericWidget getWidget() 
   {
      return widget;
   }
   
   /**      
    * Get a copy of the referenced object's current runtime value (this does
    * not return the original instance itself).
    *
    * @return   The copy of the current value.
    */
   public BaseDataType get()
   {
      return data.get();
   }
   
   /**
    * Set the referenced object's current runtime value to the given value.
    *
    * @param    val
    *           The current value to be assigned into the referenced object.
    */
   public void set(BaseDataType val)
   {
      data.set(val);
   }
   
   /** 
    * Returns the source and/or sink of the widget's data.
    * 
    * @return   See above.
    */
   public Accessor getDataAccessor()
   {
      return data;
   }
}