/*
** Module   : EnhancedBrowseConfigManager.java
** Abstract : Manager for enhanced browse configurations.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SVL 20180529 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import java.util.*;
import java.util.concurrent.locks.*;
import java.util.logging.*;

/**
 * Manager for enhanced browse configurations. Loads configurations from the directory, caches
 * them and saves new and updated configurations back to the directory.
 */
public class EnhancedBrowseConfigManager
{
   /** Logger */
   private static final Logger LOG =
         LogHelper.getLogger(EnhancedBrowseConfigManager.class.getName());

   /** Root node for browse configurations. */
   private static final String ENHANCED_CONFIG_NODE = "enhanced-browse-configs";

   /** In-directory browse configuration class. */
   private static final String ENHANCED_CONFIG_CLASS = "enhancedBrowseConfig";

   /** In-directory column configuration class. */
   private static final String ENHANCED_COL_CONFIG_CLASS = "enhancedColumnConfig";

   /** Root node for column configurations. */
   private static final String COLUMNS_NODE = "columns";

   /** Default server or account node. */
   private static final String DEFAULT_NODE = "default";

   /** Key referencing ALL browses. */
   private static final EnhancedBrowseConfigKey ALL_BROWSES_KEY = new EnhancedBrowseConfigKey();

   /** Manager instance. */
   private static final EnhancedBrowseConfigManager instance;

   /** <code>true</code> if enhanced configurations are enabled. */
   private boolean enhanced;

   /** <code>true</code> if hyperlinks in browse columns are enabled. */
   private boolean hyperlinking;

   /** Array consisting of current server ID and "default" server. */
   private String[] serverIds;

   /**
    * Enhanced configurations keyed by Server ID -> Group/Account ID -> Browse key ->
    * Configuration.
    */
   private final Map<String, Map<String, Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig>>>
                                                                     configs = new HashMap<>();

   /** Lock for reading and modifying configurations. */
   private final ReadWriteLock configsLock = new ReentrantReadWriteLock();

   /** Maximum existing directory node ID inside each Server ID -> Group/Account ID section. */
   private final Map<String, Map<String, Integer>> maxNodeIds = new HashMap<>();

   /** Context-local save targets available for the user. */
   private final ContextLocal<EnhancedBrowseSaveTarget[]> enhancedSaveTargets =
                                                  new ContextLocal<EnhancedBrowseSaveTarget[]>()
   {
      /**
       * Check access right to specific directory nodes in order to determine save targets
       * available for the user.
       *
       * @return save targets available for the user.
       */
      protected EnhancedBrowseSaveTarget[] initialValue()
      {
         DirectoryService ds = DirectoryService.getInstance();
         List<EnhancedBrowseSaveTarget> res = new ArrayList<>();
         for (String serverId : serverIds)
         {
            for (String accountId : accountIds.get())
            {
               // test arbitrary sub-node under .../enhanced-browse-configs
               String testNode = getDirectoryNode(serverId, accountId) + "/some";
               if (ds.canAccessNode(testNode,
                                    DirectoryResource.DIR_WRITE_ACCESS,
                                    DirectoryResource.DIR_CREATE_ACCESS))
               {
                  for (boolean all : allBrowses)
                  {
                     res.add(new EnhancedBrowseSaveTarget(serverId, accountId, all));
                  }
               }
            }
         }

         return res.toArray(new EnhancedBrowseSaveTarget[res.size()]);
      }
   };

   /**
    * All subject IDs (user ID and group IDs) associated with the current context, including
    * "default".
    */
   private ContextLocal<String[]> accountIds = new ContextLocal<String[]>()
   {
      protected String[] initialValue()
      {
         String[] ids = SecurityManager.getInstance().getAccountIds();
         String[] res = new String[ids.length + 1];
         System.arraycopy(ids, 0, res, 0, ids.length);
         res[ids.length] = DEFAULT_NODE;

         return res;
      }
   };

   /**
    * Convenience array for looping, reflects if a configuration should be applied to a
    * specific browse or to all browses.
    */
   private boolean[] allBrowses = new boolean[] {false, true};

   static
   {
      instance = new EnhancedBrowseConfigManager();
      instance.init();
   }

   /**
    * Get instance of this manager.
    *
    * @return instance of this manager.
    */
   public static EnhancedBrowseConfigManager getInstance()
   {
      return instance;
   }

   /**
    * Initialize manager: read enhanced configurations from the directory.
    */
   private void init()
   {
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         return;
      }
      try
      {
         enhanced = Utils.getDirectoryNodeBoolean(ds, "enhancedBrowseEnabled", false, false);
         hyperlinking = Utils.getDirectoryNodeBoolean(ds,
                                                      "browseHyperlinkingEnabled",
                                                      false,
                                                      false);

         if (isEnhanced())
         {
            serverIds = new String[]{DirectoryService.getInstance().getServerId(), DEFAULT_NODE};
            for (String serverId : serverIds)
            {
               Map<String, Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig>> serverMap =
                                                                                 new HashMap<>();
               Map<String, Integer> serverMaxIdsMap = new HashMap<>();
               configs.put(serverId, serverMap);
               maxNodeIds.put(serverId, serverMaxIdsMap);

               String[] allSubjects = SecurityManager.getInstance().getAllSubjects();
               String[] allAccounts = new String[allSubjects.length + 1];
               System.arraycopy(allSubjects, 0, allAccounts, 0, allSubjects.length);
               allAccounts[allAccounts.length - 1] = DEFAULT_NODE;
               for (String accountId : allAccounts)
               {
                  List<EnhancedBrowseConfig> userConfigs = readConfigs(ds, serverId, accountId);
                  if (userConfigs != null && !userConfigs.isEmpty())
                  {
                     Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig> userMap = new HashMap<>();
                     serverMap.put(accountId, userMap);

                     int maxId = 1;
                     for (EnhancedBrowseConfig config : userConfigs)
                     {
                        userMap.put(config.getKey(), config);
                        maxId = Math.max(maxId, config.directoryNodeId);
                     }

                     serverMap.put(accountId, userMap);
                     serverMaxIdsMap.put(accountId, maxId);
                  }
               }
            }
         }
      }
      finally
      {
         ds.unbind();
      }
   }

   /**
    * Returns <code>true</code> if enhanced browses are enabled.
    *
    * @return see above.
    */
   public boolean isEnhanced()
   {
      return enhanced;
   }

   /**
    * Returns <code>true</code> if hyperlinks in browse columns are enabled.
    *
    * @return see above.
    */
   public boolean isHyperlinking()
   {
      return hyperlinking;
   }

   /**
    * Get save targets for an enhanced browse configuration, which are available for the current
    * user.
    *
    * @return see above.
    */
   public EnhancedBrowseSaveTarget[] getEnhancedBrowseSaveTargets()
   {
      return enhancedSaveTargets.get();
   }

   /**
    * Get an enhanced configuration from the cache. Cache is searched using the provided the key
    * in the following order:
    * <ul>
    *    <li>current server, current account, specific browse;</li>
    *    <li>current server, current account, all browses;</li>
    *    <li>current server, group account, specific browse;</li>
    *    <li>current server, group account, all browses;</li>
    *    <li>current server, default account, specific browse;</li>
    *    <li>current server, default account, all browses;</li>    
    *    <li>default server, current account, specific browse;</li>
    *    <li>default server, current account, all browses;</li>
    *    <li>default server, group account, specific browse;</li>
    *    <li>default server, group account, all browses;</li>
    *    <li>default server, default account, specific browse;</li>
    *    <li>default server, default account, all browses;</li>    
    * </ul>
    *
    * @param   key
    *          Enhanced configuration key.
    *
    * @return  first found enhanced configuration from the cache or <code>null</code> if a
    *          configuration wasn't found.
    */
   public EnhancedBrowseConfig getEnhancedConfig(EnhancedBrowseConfigKey key)
   {
      for (String server : serverIds)
      {
         for (String account : accountIds.get())
         {
            for (boolean all : allBrowses)
            {
               EnhancedBrowseConfig res =
                     getEnhancedConfig(server, account, all ? ALL_BROWSES_KEY : key);
               if (res != null)
               {
                  return res;
               }
            }
         }
      }

      return null;
   }

   /**
    * Save the specified enhanced browse configuration using the specified save target.
    *
    * @param  config
    *         Enhanced browse configuration to save.
    * @param  saveTarget
    *         Save target (target server, target account, applied to all browses or not).
    *
    * @return <code>true</code> on success.
    */
   public boolean saveEnhancedConfig(EnhancedBrowseConfig config,
                                     EnhancedBrowseSaveTarget saveTarget)
   {
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Cannot bind directory service.");
         }
         return false;
      }

      if (saveTarget.allBrowses)
      {
         config.ehKeyProcName = null;
         config.ehKeyBrowseName = null;
         config.columnConfigs = null;
      }

      configsLock.writeLock().lock();

      try
      {
         Map<String, Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig>> serverMap =
               configs.get(saveTarget.serverId);
         Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig> userMap =
               serverMap.get(saveTarget.accountId);

         // assign directory node id
         config.directoryNodeId = null;
         if (userMap != null)
         {
            EnhancedBrowseConfig saved = userMap.get(config.getKey());
            if (saved != null)
            {
               config.directoryNodeId = saved.directoryNodeId;
            }
         }

         if (config.directoryNodeId == null)
         {
            config.directoryNodeId =
                  getNextDirectoryNodeId(saveTarget.serverId, saveTarget.accountId);
         }

         String[] subNodes =
         {
            "server",
            saveTarget.serverId,
            "runtime",
            saveTarget.accountId,
            ENHANCED_CONFIG_NODE,
            String.valueOf(config.directoryNodeId)
         };

         String parentNode = "/";
         for (int i = 0; i < subNodes.length; i++)
         {
            String subNode = subNodes[i];
            String node = parentNode + subNode + "/";
            boolean lastNode = i == subNodes.length - 1;

            // add missing elements in target node path or the node itself
            if (ds.getNodeClass(node) == null)
            {
               if (!ds.openBatch(parentNode))
               {
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, "Cannot open batch for " + parentNode + ".");
                  }
                  return false;
               }

               if (!lastNode)
               {
                  // add path node
                  if (!ds.addNode(node, "container", null))
                  {
                     ds.closeBatch(false);
                     if (LOG.isLoggable(Level.SEVERE))
                     {
                        LOG.log(Level.SEVERE, "Cannot create path node " + node + ".");
                     }
                     return false;
                  }
               }
               else
               {
                  // add config node
                  Attribute[] attrs = getBrowseConfigNodeAttrs(ds, config);
                  if (!ds.addNode(node, ENHANCED_CONFIG_CLASS, attrs))
                  {
                     ds.closeBatch(false);
                     if (LOG.isLoggable(Level.SEVERE))
                     {
                        LOG.log(Level.SEVERE, "Cannot create enhanced config node " + node + ".");
                     }
                     return false;
                  }

                  if (!saveTarget.allBrowses && config.columnConfigs != null)
                  {
                     String columnsRoot = node + COLUMNS_NODE + "/";
                     if (!ds.addNode(columnsRoot, "container", null))
                     {
                        ds.closeBatch(false);
                        if (LOG.isLoggable(Level.SEVERE))
                        {
                           LOG.log(Level.SEVERE, "Cannot create enhanced columns' root node " +
                                                 columnsRoot + ".");
                        }
                        return false;
                     }

                     int count = 1;
                     for (Map.Entry<String, EnhancedColumnConfig> entry :
                                                                  config.columnConfigs.entrySet())
                     {
                        String columnNode = columnsRoot + String.valueOf(count++) + "/";
                        attrs = getColumnConfigNodeAttrs(ds,
                                                         entry.getKey(),
                                                         entry.getValue());
                        if (!ds.addNode(columnNode, ENHANCED_COL_CONFIG_CLASS, attrs))
                        {
                           ds.closeBatch(false);
                           if (LOG.isLoggable(Level.SEVERE))
                           {
                              LOG.log(Level.SEVERE, "Cannot create enhanced column node " +
                                                    columnNode + ".");
                           }
                           return false;
                        }
                     }
                  }
               }

               if (!ds.closeBatch(true))
               {
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, "Cannot close batch for node " + node + ".");
                  }
                  return false;
               }
            }
            else if (lastNode)
            {
               // config node already exists
               if (!ds.openBatch(node))
               {
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, "Cannot open batch for " + parentNode + ".");
                  }
                  return false;
               }

               if (!ds.setNodeAttributes(node, getBrowseConfigNodeAttrs(ds, config)))
               {
                  ds.closeBatch(false);
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, "Cannot set node attributes for node " + node + ".");
                  }
                  return false;
               }

               String columnsRoot = node + "columns/";
               if (ds.getNodeClass(columnsRoot) == null)
               {
                  if (!ds.addNode(columnsRoot, "container", null))
                  {
                     ds.closeBatch(false);
                     if (LOG.isLoggable(Level.SEVERE))
                     {
                        LOG.log(Level.SEVERE, "Cannot add columns node " + columnsRoot + ".");
                     }
                     return false;
                  }
               }

               if (config.columnConfigs != null && config.columnConfigs.size() > 0)
               {
                  int count = 1;
                  for (Map.Entry<String, EnhancedColumnConfig> entry :
                                                               config.columnConfigs.entrySet())
                  {
                     String columnNode = columnsRoot + String.valueOf(count++) + "/";
                     Attribute[] attrs = getColumnConfigNodeAttrs(ds,
                                                                  entry.getKey(),
                                                                  entry.getValue());
                     if (!ds.setNodeAttributes(columnNode, attrs))
                     {
                        ds.closeBatch(false);
                        if (LOG.isLoggable(Level.SEVERE))
                        {
                           LOG.log(Level.SEVERE, "Cannot node attributes for the column node " +
                                                 columnNode + ".");
                        }
                        return false;
                     }
                  }
               }

               // delete extra nodes under columns/
               int maxNode = config.columnConfigs != null ? config.columnConfigs.size() : 0;
               String[] columnNodes = ds.enumerateNodes(columnsRoot);
               if (columnNodes != null)
               {
                  for (String columnNode : columnNodes)
                  {
                     if (Integer.valueOf(columnNode) > maxNode)
                     {
                        if (!ds.deleteNode(columnsRoot + "/" + columnNode))
                        {
                           ds.closeBatch(false);
                           if (LOG.isLoggable(Level.SEVERE))
                           {
                              LOG.log(Level.SEVERE, "Cannot delete node " +
                                    columnNode + ".");
                           }
                           return false;
                        }
                     }
                  }
               }

               if (!ds.closeBatch(true))
               {
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     LOG.log(Level.SEVERE, "Cannot close batch for node " + node + ".");
                  }
                  return false;
               }
            }

            parentNode = node;
         }

         // update in-memory config
         if (userMap == null)
         {
            userMap = new HashMap<>();
            serverMap.put(saveTarget.accountId, userMap);
         }
         userMap.put(config.getKey(), config);
      }
      finally
      {
         configsLock.writeLock().unlock();
         ds.unbind();
      }

      return true;
   }

   /**
    * Get next directory node id for saving a new configuration.
    *
    * @param  serverId
    *         Server ID to which the configuration belongs.
    * @param  accountId
    *         Account ID to which the configuration belongs.
    *
    * @return next directory node id for saving a new configuration.
    */
   private int getNextDirectoryNodeId(String serverId, String accountId)
   {
      synchronized (maxNodeIds)
      {
         Map<String, Integer> accountMaxIds = maxNodeIds.get(serverId);
         Integer val = accountMaxIds.get(accountId);
         val = val == null ? 1 : val + 1;

         accountMaxIds.put(accountId, val);
         return val;
      }
   }

   /**
    * Get the enhanced browse configuration as an array of directory attributes.
    *
    * @param  ds
    *         Bound directory service.
    * @param  config
    *         Browse configuration to be processed.
    *
    * @return array of directory attributes created from the specified enhanced browse
    *         configuration.
    */
   private Attribute[] getBrowseConfigNodeAttrs(DirectoryService ds, EnhancedBrowseConfig config)
   {
      return new Attribute[]
      {
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "proc_name"),
               new String[] {config.ehKeyProcName}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "browse_name"),
               new String[] {config.ehKeyBrowseName}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "fgcolor"),
               new String[] {encodeColor(config.ehFgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "bgcolor"),
               new String[] {encodeColor(config.ehBgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "label_fgcolor"),
               new String[] {encodeColor(config.ehLabelFgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "label_bgcolor"),
               new String[] {encodeColor(config.ehLabelBgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "font"),
               new String[] {encodeFont(config.ehFont)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "label_font"),
               new String[] {encodeFont(config.ehLabelFont)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "separators"),
               new Boolean[] {config.separators}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "separators_color"),
               new String[] {encodeColor(config.ehSeparatorsColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "row_height_px"),
               new Integer[] {config.rowHeightPixels}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_CONFIG_CLASS, "labels_height_px"),
               new Integer[] {config.labelsHeightPixels})
      };
   }

   /**
    * Get the enhanced column configuration as an array of directory attributes.
    *
    * @param  ds
    *         Bound directory service.
    * @param  columnKey
    *         Column key (added to the resulting attributes).
    * @param  columnConfig
    *         Column configuration to be processed.
    *
    * @return array of directory attributes created from the specified enhanced column
    *         configuration.
    */
   private Attribute[] getColumnConfigNodeAttrs(DirectoryService ds,
                                                String columnKey,
                                                EnhancedColumnConfig columnConfig)
   {
      return new Attribute[]
      {
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "column_key"),
               new String[] {columnKey}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "ordinal"),
               new Integer[] {columnConfig.ordinal}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "visible"),
               new Boolean[] {columnConfig.visible}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "width_px"),
               new Integer[] {columnConfig.widthPixels}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "fgcolor"),
               new String[] {encodeColor(columnConfig.ehFgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "bgcolor"),
               new String[] {encodeColor(columnConfig.ehBgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "label_fgcolor"),
               new String[] {encodeColor(columnConfig.ehLabelFgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "label_bgcolor"),
               new String[] {encodeColor(columnConfig.ehLabelBgColor)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "font"),
               new String[] {encodeFont(columnConfig.ehFont)}),
         new Attribute(ds.getClassNodeAttribute(ENHANCED_COL_CONFIG_CLASS, "label_font"),
               new String[] {encodeFont(columnConfig.ehLabelFont)}),
      };
   }

   /**
    * Get the directory node path which corresponds the location of the "enhanced-browse-configs"
    * node for the given server and account IDs.
    *
    * @param  serverId
    *         Server ID.
    * @param  accountId
    *         Account ID.
    *
    * @return see above.
    */
   private String getDirectoryNode(String serverId, String accountId)
   {
      return String.format("/server/%s/runtime/%s/%s",
                           serverId,
                           accountId,
                           ENHANCED_CONFIG_NODE);
   }

   /**
    * Read the enhanced configurations from the directory.
    *
    * @param  ds
    *         Bound directory service.
    * @param  serverId
    *         Server ID of the target configurations.
    * @param  accountId
    *         Account ID of the target configurations.
    *
    * @return list of configurations belonging to the given server and account IDs read from the
    *         directory.
    */
   private List<EnhancedBrowseConfig> readConfigs(DirectoryService ds,
                                                  String serverId,
                                                  String accountId)
   {
      String nodeId = getDirectoryNode(serverId, accountId);

      String cls = ds.getNodeClass(nodeId);
      if (cls == null || !cls.endsWith("/container"))
      {
         return null;
      }

      String[] configNodes = ds.enumerateNodes(nodeId);
      if (configNodes == null || configNodes.length == 0)
      {
         return null;
      }

      List<EnhancedBrowseConfig> res = new ArrayList<>();
      for (String configNode : configNodes)
      {
         String browseNodeId = nodeId + "/" + configNode;
         EnhancedBrowseConfig config = new EnhancedBrowseConfig();

         config.directoryNodeId    = Integer.parseInt(configNode);
         config.ehKeyProcName      = nullify(ds.getNodeString(browseNodeId, "proc_name"));
         config.ehKeyBrowseName    = nullify(ds.getNodeString(browseNodeId, "browse_name"));
         config.ehFgColor          = decodeColor(ds.getNodeString(browseNodeId, "fgcolor"));
         config.ehBgColor          = decodeColor(ds.getNodeString(browseNodeId, "bgcolor"));
         config.ehLabelFgColor     = decodeColor(ds.getNodeString(browseNodeId, "label_fgcolor"));
         config.ehLabelBgColor     = decodeColor(ds.getNodeString(browseNodeId, "label_bgcolor"));
         config.ehFont             = decodeFont(ds.getNodeString(browseNodeId, "font"));
         config.ehLabelFont        = decodeFont(ds.getNodeString(browseNodeId, "label_font"));
         config.separators         = ds.getNodeBoolean(browseNodeId, "separators");
         config.ehSeparatorsColor  =
                                  decodeColor(ds.getNodeString(browseNodeId, "separators_color"));
         config.rowHeightPixels     = ds.getNodeInteger(browseNodeId, "row_height_px");
         config.labelsHeightPixels  = ds.getNodeInteger(browseNodeId, "labels_height_px");

         String columnsRoot = browseNodeId + "/" + COLUMNS_NODE;
         String[] columnNodes = ds.enumerateNodes(columnsRoot);
         if (columnNodes != null && columnNodes.length > 0)
         {
            config.columnConfigs = new HashMap<>(columnNodes.length);
            for (String columnNode : columnNodes)
            {
               String columnNodeId = columnsRoot + "/" + columnNode;
               EnhancedColumnConfig columnConfig = new EnhancedColumnConfig();
               String key                = nullify(ds.getNodeString(columnNodeId, "column_key"));
               columnConfig.ordinal      = ds.getNodeInteger(columnNodeId, "ordinal");
               columnConfig.ehFgColor    = decodeColor(ds.getNodeString(columnNodeId, "fgcolor"));
               columnConfig.ehBgColor    = decodeColor(ds.getNodeString(columnNodeId, "bgcolor"));
               columnConfig.ehLabelFgColor =
                                     decodeColor(ds.getNodeString(columnNodeId, "label_fgcolor"));
               columnConfig.ehLabelBgColor =
                                     decodeColor(ds.getNodeString(columnNodeId, "label_bgcolor"));
               columnConfig.visible        = ds.getNodeBoolean(columnNodeId, "visible");
               columnConfig.widthPixels    = ds.getNodeInteger(columnNodeId, "width_px");
               columnConfig.ehFont         = decodeFont(ds.getNodeString(columnNodeId, "font"));
               columnConfig.ehLabelFont    =
                                         decodeFont(ds.getNodeString(columnNodeId, "label_font"));
               config.columnConfigs.put(key, columnConfig);
            }
         }

         res.add(config);
      }

      return res;
   }

   /**
    * Convert string representation of the color into the {@link ColorRgb} object.
    *
    * @param  colorString
    *         String representation of the color in the format "#rrggbb".
    *
    * @return matching {@link ColorRgb} object.
    */
   private ColorRgb decodeColor(String colorString)
   {
      if (colorString == null || colorString.trim().isEmpty())
      {
         return null;
      }

      try
      {
         java.awt.Color color = java.awt.Color.decode(colorString);
         return new ColorRgb(color);
      }
      catch (NumberFormatException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Cannot decode color string \"" + colorString +
                       "\" in enhanced browse config");
         }
         return null;
      }
   }

   /**
    * Convert {@link ColorRgb} object into its string representation.
    *
    * @param  color
    *         Color to convert.
    *
    * @return string representation of the color in the format "#rrggbb".
    */
   private String encodeColor(ColorRgb color)
   {
      if (color == null)
      {
         return null;
      }

      return String.format("#%02x%02x%02x", color.getRed(), color.getGreen(), color.getBlue());
   }

   /**
    * Convert font into its string representation.
    *
    * @param  font
    *         Font to convert.
    *
    * @return string representation of the font in the format
    *         "font name,size,bold,italic,underline" (e.g. "dejavu sans,12,false,true,false").
    */
   private String encodeFont(FontDetails font)
   {
      if (font == null)
      {
         return null;
      }

      return font.key;
   }

   /**
    * Convert string representation of the font into the {@link FontDetails} object.
    *
    * @param  fontString
    *         String representation of the font in the format
    *         "font name,size,bold,italic,underline" (e.g. "dejavu sans,12,false,true,false").
    *
    * @return matching {@link FontDetails} object.
    */
   private FontDetails decodeFont(String fontString)
   {
      if (fontString == null || fontString.trim().isEmpty())
      {
         return null;
      }

      try
      {
         String[] parts = fontString.split(",");
         if (parts.length != 5)
         {
            throw new Exception("Invalid string format");
         }

         String fontName   = parts[0].trim();
         int size          = Integer.parseInt(parts[1].trim());
         boolean bold      = Boolean.parseBoolean(parts[2].trim());
         boolean italic    = Boolean.parseBoolean(parts[3].trim());
         boolean underline = Boolean.parseBoolean(parts[4].trim());

         FontStyle style = FontStyle.createStyle(bold, italic, underline);
         // rebuild trimmed key
         String key = FontManager.getFontKey(fontName, size, bold, italic, underline);
         return new FontDetails(key, fontName, style, size, -1, -1, -1, null, false);
      }
      catch (Exception e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Cannot decode font string \"" + fontString +
                       "\" in enhanced browse config");
         }
         return null;
      }
   }

   /**
    * If the source string is <code>null</code> or empty, return <code>null</code>.
    * Otherwise, the source string is returned.
    *
    * @param   str
    *          Source string.
    *
    * @return  see above.
    */
   private String nullify(String str)
   {
      if (str != null && str.trim().isEmpty())
      {
         return null;
      }

      return str;
   }

   /**
    * Get an enhanced configuration from the cache section under the given server and account IDs.
    *
    * @param   serverId
    *          Server ID.
    * @param   accountId
    *          Account ID.
    * @param   configKey
    *          Enhanced configuration key.
    *
    * @return  found enhanced configuration from the cache or <code>null</code> if a
    *          configuration wasn't found in the given section.
    */
   private EnhancedBrowseConfig getEnhancedConfig(String serverId,
                                                  String accountId,
                                                  EnhancedBrowseConfigKey configKey)
   {
      configsLock.readLock().lock();

      try
      {
         Map<String, Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig>> serverMap =
                                                                           configs.get(serverId);
         Map<EnhancedBrowseConfigKey, EnhancedBrowseConfig> userMap = serverMap.get(accountId);
         if (userMap != null)
         {
            EnhancedBrowseConfig config = userMap.get(configKey);
            if (config != null)
            {
               return new EnhancedBrowseConfig(config);
            }
         }
      }
      finally
      {
         configsLock.readLock().unlock();
      }

      return null;
   }
}
