/*
** Module   : ErrorWriterInteractive.java
** Abstract : Implementation for error writer for interactive mode.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131119 Created initial version.
** 002 VIG 20140810 Modification of pause method to use window Id as parameter.
** 003 VIG 20140922 Changed message method call. 
** 004 IAS 20160314 DEBUG-ALERT support.
** 005 HC  20160901 Error messages in GUI are always displayed in alert boxes.
** 006 HC  20171003 Implemented SESSION:SUPPRESS-WARNINGS runtime support and the ability to
**                  handle (display or output depending on the runtime state) legacy warnings.
** 007 SVL 20171024 Added displayError(errmsg, noMessageBox).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Implements display outputs for interactive mode.
 */
public class ErrorWriterInteractive
implements ErrorWriter
{
   /** Client exports access. */
   private ClientExports client = null; 
   
   /**
    * The only constructor used to create error writer.
    *
    * @param    client
    *           The client instance associated with writer.
    */
   public ErrorWriterInteractive(ClientExports client)
   {
      this.client = client;
   }
   
   /**
    * Displays error message for interactive mode.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayError(String errmsg)
   {
      displayError(errmsg, false);
   }

   /**
    * Displays error message for interactive mode.
    *
    * @param    errmsg
    *           Text of the message to display.
    * @param    noMessageBox
    *           <code>true</code> if error is always displayed in message area or logged rather
    *           than displayed as a message box.
    */
   public void displayError(String errmsg, boolean noMessageBox)
   {
      if (client != null)
      {
         client.displayErrorMessage(errmsg, noMessageBox);
      }
   }

   /**
    * Displays warning message using currently installed writer implementation.
    *
    * @param    wrnmsg
    *           Text of the message to display.
    */
   public void displayWarning(String wrnmsg)
   {
      if (client != null)
      {
         client.displayWarningMessage(wrnmsg);
      }
   }

   /**
    * Displays error message using currently installed writer implementation.  This is the
    * redirected version of the displayError.  If the output is redirected to file the message
    * is writing to the file.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void displayErrorRedirected(String errmsg)
   {
      displayError(errmsg);
   }
    
   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public boolean isInBatchMode()
   {
      return (client != null) ? client.isInBatchMode() : false;
   }

   /**
    * Displays error message for interactive mode.
    *
    * @param    errmsg
    *           Text of the message to display.
    */
   public void message(String errmsg)
   {
      if (client != null)
      {
         client.message(errmsg, false, null, false, null, null, WindowConfig.RESOLVE_WINDOW);
      }
   }
   
   /**
    * Displays error message box with title for interactive mode.
    *
    * @param    masterMsg
    *           Text of the message to display.
    * @param    title
    *           The title for the message box to display.
    */
   public void messageBox(String masterMsg, String title)
   {
      if (title == null)
      {
         title = new String("Error");
      }
      
      if (client != null)
      {
         client.messageBox(new Object[] {masterMsg},
                           null,
                           LogicalTerminal.ALERT_ERROR,
                           LogicalTerminal.BTN_OK,
                           title,
                           (new ColorSpec(ColorSpec.COLOR_NORMAL)).convert(),
                           WindowConfig.RESOLVE_WINDOW,
                           null /* TODO: is that always correct */);
      }
   }
   
   /**
    * Implements pause processing for interactive mode.
    */
   public void pause()
   {
      if (client != null)
      {
         if (LogicalTerminal.targetWindowWorker(null, true) == WindowConfig.RESOLVE_WINDOW) 
         {
            // TODO: This pause could be called in some other scenario than just error displaying
            // because of here is no error message.
            client.messageBox(new Object[] { /* TODO how to get an error message here? */},
                  null,
                  LogicalTerminal.ALERT_ERROR,
                  LogicalTerminal.BTN_OK,
                  "Error",
                  (new ColorSpec(ColorSpec.COLOR_NORMAL)).convert(),
                  WindowConfig.RESOLVE_WINDOW,
                  null /* TODO: is it always correct& */);
         }
         else {
            client.pause(-1, null, false, WindowConfig.RESOLVE_WINDOW);   
         }
      }
      
      return;
   }

   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public void stopBatchSession()
   {
      if (client != null)
      {
         client.stopBatchSession();
      }
   }

   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    */
   @Override
   public void logStackTrace(String message)
   {
      client.logStackTrace(message);
   }
}