/*
** Module   : FieldGroup.java
** Abstract : groups widgets in the frame on the server side
**
** Copyright (c) 2008-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20080318   @37589 Created initial version to group widgets in
**                           a frame on the server.
** 002 NVS 20081211   @40873 getParent() method is overriden here to get
**                           the frame or another field group as the
**                           parent.
** 003 GES 20081212   @40883 Only BaseEntity can accept a field group (at
**                           this time), so safety code was added.
** 004 GES 20081230   @41002 Return a NullWidget instance instead of null
**                           for safety.
** 005 SIY 20101122          Cleaned up imports.
** 006 SIY 20111101          Updated imports.
** 007 CA  20130221          Refactored usage of TYPE attribute.
** 008 CA  20130323          The PARENT attribute must be a CommonWidget, not Object.
** 019 CA  20130927          Added LegacyResource annotation, to determine resource type.
**                           Removed NullWidget usage and fixed getFirst/LastChild to return a 
**                           handle instance.
** 020 CA  20131115          Changes related to properly instantiate dynamic resources.
** 021 VIG 20140314          Added new parameter before to the constructor to support
**                           insertion semantic of new field-group.    
** 022 VIG 20140523          Added getFirstTabItem and setFirstTabItem methods of 
**                           the tab order management.
** 023 VIG 20140612          Removing TODO in getFirstTabItem.
** 024 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 026 IAS 20151020          addDynamicWidget method.
** 027 CA  20151026          Added add/remove widget support (to allow nested frames).
** 028 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 029 SBI 20151213          Removed apache dependency.
** 030 EVL 20160621          Fix for FIRST-TAB-ITEM wrong message for invisible frame.  If client
**                           is unable to provide valid widget P2J use first available field-level
**                           child as last resort.
** 031 HC  20160714          FIELD-GROUP:FIRST-TAB-ITEM must return an unknown handle when no
**                           items in the field group.
** 032 HC  20170131          Added runtime support for dynamic MENU* widgets.
** 033 SBI 20161011          Added moveWidgetAfter[Before] methods and first[last]TabItem fields,
**                           changed getFirstTabItem() accordingly.
** 034 SBI 20171016          Changed addNewTabItem to notify the client about tab items changes.
** 035 SBI 20171120          Exposes toTop/toBottom methods to move an owned widget to the top or
**                           to the bottom of its widget list.
** 036 CA  20180526          Added getters for width/height attributes, computed for the frame.
**                           This will not work with down frames.
** 037 HC  20180628          Added side-labels to the widget chain.
** 038 SBI 20180716          Changed to take into account z-order widget class to preserve widgets
**                           order according to their z-order classes hierarchy.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;


import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.util.*;
 
/**
 * Groups widgets together into a list of siblings that can be accessed
 * via the first or last entry in the list.
 */
@LegacyResource(resource = LegacyResource.FIELD_GROUP)
public class FieldGroup
extends BaseEntity<FieldGroupConfig>
implements Group,
           WidgetContainer,
           FieldGroupInterface
{
   /** The list of contained widgets. */
   private List<GenericWidget<?>> widgets = null;
   
   /** Function for adding widgets. 
    * 
    *  The FieldGroup is a server-side only object.
    *  However at least in the case of the BACKGROUND attribute they can be changed
    *  (widgets can be added)
    *  To make these changes visible the widgets should be added to the enclosing frame
    *  as well. This Consumer is used for this purpose instead of direct reference to a
    *  GenericFrame object to make this link less tight.
    */ 
   private Consumer<GenericWidget<?>> widgetAdder = null; 
   
   /** The first tab item */
   private GenericWidget<?> firstTabItem;
   
   /** The last tab item */
   private GenericWidget<?> lastTabItem;
   
   /** Tab item list */ 
   private List<GenericWidget<?>> tabItems;
   
   /**
    * Creates an instance to groups related widgets in a frame.
    *
    * @param    widgets
    *           The widgets contained in this instance. This is used as a
    *           read-only resource.
    * @param    beforeFg
    *           The right neighbor of the field-group been created or null if new field-group
    *           will be the last in the chain. If <b>beforeFg</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    */
   public FieldGroup(List<GenericWidget<?>> widgets, FieldGroup beforeFg)
   {
      super(false, new FieldGroupConfig(), beforeFg, true);
      this.widgets  = widgets;
      this.tabItems = new LinkedList<GenericWidget<?>>();
      if (widgets != null)
      {
         for (GenericWidget<?> next : widgets)
         {
            this.tabItems.add(next);
            if (next instanceof BaseEntity)
            {
               BaseEntity<?> be = (BaseEntity<?>) next;
               be.setGroup(this);
            }
         }
      }
      
      Utils.normalizeZOrder(this.widgets, w -> { return w.getZOrderClass(); });
   }
   
   /**
    * Get rectangle FIELD-GROUP:HEIGHT-PIXELS value.
    * 
    * @return  The height of the widget in pixels.
    */
   @Override
   public integer getHeightPixels()
   {
      // TODO: DOWN frames or top-label frames will report this incorrectly
      FrameWidget fw = frame.getFrameWidget();
      return new integer(fw.getHeightPixels().intValue() - 
                         fw.borderTopPixels().intValue() - 
                         fw.borderBottomPixels().intValue());
   }
   
   /**
    * Get rectangle FIELD-GROUP:HEIGHT-CHARS value.
    * 
    * @return  The height of the widget in character units.
    */
   @Override
   public decimal getHeightChars()
   {
      // TODO: DOWN frames or top-label frames will report this incorrectly
      FrameWidget fw = frame.getFrameWidget();
      return new decimal(fw.getHeightChars().doubleValue() - 
                         fw.borderTopChars().doubleValue() - 
                         fw.borderBottomChars().doubleValue());
   }
   
   /**
    * Get rectangle FIELD-GROUP:WIDTH-CHARS value.
    * 
    * @return  The width of the widget in character units.
    */
   @Override
   public decimal getWidthChars()
   {
      FrameWidget fw = frame.getFrameWidget();
      return new decimal(fw.getWidthChars().doubleValue() - 
                         fw.borderLeftChars().doubleValue() - 
                         fw.borderRightChars().doubleValue());
   }
   
   /**
    * Get rectangle FIELD-GROUP:WIDTH-PIXELS value.
    * 
    * @return  The width of the widget in pixels.
    */
   @Override
   public integer getWidthPixels()
   {
      FrameWidget fw = frame.getFrameWidget();
      return new integer(fw.getWidthPixels().intValue() - 
                         fw.borderLeftPixels().intValue() - 
                         fw.borderRightPixels().intValue());
   }
   
   /**
    * Remove the specified widget from this field-group.
    * 
    * @param    widget
    *           The widget to be removed.
    */
   public void removeWidget(GenericWidget<?> widget)
   {
      this.widgets.remove(widget);
      if (widget instanceof BaseEntity)
      {
         BaseEntity<?> be = (BaseEntity<?>) widget;
         be.setGroup(null);
      }
      this.tabItems.remove(widget);
      this.firstTabItem = null;
      this.lastTabItem  = null;
   }
   
   /**
    * Add the specified widget to this field-group.
    * 
    * @param    widget
    *           The widget to be added to this field group.
    */
   public void addWidget(GenericWidget<?> widget)
   {
      this.widgets.remove(widget);
      this.widgets.add(widget);
      
      Utils.normalizeZOrder(widgets,  w -> { return w.getZOrderClass(); });
      
      // add widget to the tab item list
      if (widget.isTabWidget())
      {
         moveTabItemToEnd(widget);
      }
      
      if (widget instanceof BaseEntity)
      {
         BaseEntity<?> be = (BaseEntity<?>) widget;
         be.setGroup(this);
      }
   }
   
   /**
    * Adds a new tab item to the end of the tab item list if the target widget is not in this list
    * and notifies the client about the changes. The widget can be added during addWidget calls.
    * 
    * @param    widget
    *           The widget to be added to the tab item list.
    */
   public void addNewTabItem(GenericWidget<?> widget)
   {
      if (!widget.isTabWidget())
      {
         return;
      }

      // add widget to the tab item list
      if (!this.tabItems.contains(widget))
      {
         this.tabItems.add(widget);
      }
      
      LogicalTerminal.updateTabItemList(this);
   }
   
   /**
    * Removes the target widget from the tab item list and add it to the end of the tab item list.
    * 
    * @param    widget
    *           The target widget
    */
   private void moveTabItemToEnd(GenericWidget<?> widget)
   {
      this.tabItems.remove(widget);
      this.tabItems.add(widget);
   }
   
   /**
    * Removes the target widget from the tab item list, add it at the top of the tab item
    * list at the given position and shifts the other tab items below this target widget. 
    * 
    * @param    widget
    *           The target widget
    * @param    pos
    *           The position to place new widget.
    */
   private void moveTabItemAtTop(GenericWidget<?> widget, int pos)
   {
      this.tabItems.remove(widget);
      this.tabItems.add(pos, widget);
   }
   
   /**
    * Places the given widget in the current tab item list.
    * 
    * @param    widget
    *           The widget
    * @param    op
    *           The current tab item list rearranging operation
    * @param    pos
    *           The widget position in the corresponding rearranging statement
    */
   private void placeWidgetInTabList(GenericWidget<?> widget, UpdatesOp op, int pos)
   {
      if (widget != null)
      {
               switch(op)
               {
                  case UPDATE:
                  case PROMPT_FOR:
                  case SET:
                     moveTabItemAtTop(widget, pos);
                     break;
                  case ENABLE:
                  default:
                     moveTabItemToEnd(widget);
               }
      }
   }

   /**
    * Updates tab items on tab stop.
    * 
    * @param    widget
    *           The given widget
    * @param    addToEnd
    *           The true value indicates that the given widget is placed at the end of the tab
    *           item list, otherwise its tab position is not changed.
    */
   public void updateTabItemsOnTabStop(GenericWidget<?> widget, boolean addToEnd)
   {
      if (addToEnd)
      {
         moveTabItemToEnd(widget);
      }
      updateExtremeTabItems();
   }
   
   /**
    * Move the target widget after the provided widget in this field-group. 
    * 
    * @param    widget
    *           The target widget to be moved.
    * @param    after
    *           The provided widget.
    */
   public void moveWidgetAfter(GenericWidget<?> widget, GenericWidget<?> after)
   {
      if (!tabItems.contains(widget) || !tabItems.contains(after))
      {
         throw new RuntimeException("The widgets must be in specified field-group");
      }
      
      setTabItemAfter(widget, after);
   }
   
   /**
    * Move the target widget before the provided widget in this field-group. 
    * 
    * @param    widget
    *           The target widget to be moved.
    * @param    before
    *           The provided widget.
    */
   public void moveWidgetBefore(GenericWidget<?> widget, GenericWidget<?> before)
   {
      if (!tabItems.contains(widget) || !tabItems.contains(before))
      {
         throw new RuntimeException("The widgets must be in specified field-group");
      }
      
      setTabItemBefore(widget, before);
   }
   
   /**
    * Sets the containing frame instance.
    * <p>
    * This updates the field-group's parentId, too.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(GenericFrame frame)
   {
      super.setFrame(frame);
      
      config.parentId = frame.getFrameId();
   }
   
   /**
    * Sets the name of this handle.This method is outside interfaces and only 
    * meant for initializing this within the implementation.
    * 
    * @param name
    *        The handle type to be set.
    */
   public void name(character name)
   {
      if (name == null || name.isUnknown())
      {
         String text = "Unable to assign UNKNOWN value to attribute NAME of FIELD-GROUP widget";
         ErrorManager.recordOrShowError(4083, text, false);
         return;
      }

      super.name(name); 
   }

   /**
    * Gets the PARENT writable attribute.
    * This is an override for the BaseEntity's implementation.
    * If this field group is a member of another field group,
    * that field group is returned as parent.
    * Otherwise, the owning frame is returned as the parent.
    *
    * @return   the parent widget for this widget
    */
   public CommonWidget getParent()
   {
      // explicitly set field group
      if (group != null)
      {
         return group;
      }

      // normal case
      return frame.asWidget();
   }

   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or an unknown handle if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      if (widgets == null || widgets.size() == 0)
      {
         return new handle();
      }

      GenericWidget w = widgets.get(0);
      return new handle(w instanceof WidgetTree ? ((WidgetTree) w).firstChild() : w);
   }

   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or an unknown handle if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      if (widgets == null || widgets.size() == 0)
      {
         return new handle();
      }

      GenericWidget w = widgets.get(widgets.size() - 1);
      return new handle(w instanceof WidgetTree ? ((WidgetTree) w).lastChild() : w);
   }
   
   /**
    * Accesses the next widget in the list of contained widgets.
    *
    * @param    child
    *           The widget in the list for which to find the following child.
    *
    * @return   The next widget or a <code>null</code> if there is no subsequent widget.
    */
   public GenericWidget<?> getNextChild(GenericWidget<?> child)
   {
      if (widgets == null || child == null)
         return null;
      
      int oldIdx = widgets.indexOf(child);
      int newIdx = oldIdx + 1;
      GenericWidget newWidget = (oldIdx < 0 || newIdx >= widgets.size()) ? null : widgets.get(newIdx);
      
      return newWidget instanceof WidgetTree ? ((WidgetTree) newWidget).firstChild() : newWidget;
   }

   /**
    * Accesses the previous widget in the list of contained widgets.
    *
    * @param    child
    *           The widget in the list for which to find the preceding child.
    *
    * @return   The previous widget or <code>null</code> if there is no preceding widget.
    */
   public GenericWidget<?> getPrevChild(GenericWidget<?> child)
   {
      if (widgets == null || child == null)
         return null;
      
      int oldIdx = widgets.indexOf(child);
      int newIdx = oldIdx - 1;
      GenericWidget newWidget = (oldIdx < 0 || newIdx < 0) ? null : widgets.get(newIdx);
      
      return newWidget instanceof WidgetTree ? ((WidgetTree) newWidget).lastChild() : newWidget;
   }
   
   /**
    * Getting the FIRST-TAB-ITEM attribute.
    * 
    * @return  The widget handle that is the first in the frame tab order
    */
   @Override
   public handle getFirstTabItem()
   {
      if (firstTabItem == null)
      {
         firstTabItem = getExtremeTabItem(true);
      }
      
      if (firstTabItem == null)
      {
         // when no item is found, unknown handle must be returned
         return new handle();
      }
      
      return firstTabItem.asWidgetHandle();
   }
   
   /**
    * Getting the LAST-TAB-ITEM attribute.
    * 
    * @return  The widget handle that is the last in the frame tab order
    */
   @Override
   public handle getLastTabItem()
   {
      if (lastTabItem == null)
      {
         lastTabItem = getExtremeTabItem(false);
      }
      
      if (lastTabItem == null)
      {
         // when no item is found, unknown handle must be returned
         return new handle();
      }
      
      return lastTabItem.asWidgetHandle();
   }

   /**
    * Setting the FIRST-TAB-ITEM attribute.
    * 
    * @param   wHandle
    *          The handle of widget in this field-group to assign as first in the tab order.
    */
   @Override
   public void setFirstTabItem(handle wHandle)
   {
      if (wHandle == null || !wHandle._isValid() || wHandle.isUnknown())
      {
         String ms = 
            "**Unable to assign UNKNOWN value to attribute FIRST-TAB-ITEM on FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4083, ms, false);
         return;
      }
      
      CommonWidget widget = wHandle.unwrapWidget();
      
      if (!tabItems.contains(widget))
      {
         String ms = "**Unable to set attribute FIRST-TAB-ITEM for FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4078, ms, false);
         return;
      }
      GenericWidget<?> anchor;
      if (firstTabItem != null)
      {
         anchor = firstTabItem;
      }
      else
      {
         anchor = tabItems.get(0);
      }
      firstTabItem = (GenericWidget<?>) widget;
      
      setTabItemBefore((GenericWidget<?>) widget, anchor);
      
      LogicalTerminal.updateTabItemList(this);
   }
   
   /**
    * Setting the LAST-TAB-ITEM attribute.
    * 
    * @param   wHandle
    *          The handle of widget in this field-group to assign as last in the tab order.
    */
   @Override
   public void setLastTabItem(handle wHandle)
   {
      if (wHandle == null || !wHandle._isValid() || wHandle.isUnknown())
      {
         String ms = 
            "**Unable to assign UNKNOWN value to attribute LAST-TAB-ITEM on FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4083, ms, false);
         return;
      }
      
      CommonWidget widget = wHandle.unwrapWidget();
      
      if (!tabItems.contains(widget))
      {
         String ms = "**Unable to set attribute LAST-TAB-ITEM for FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4078, ms, false);
         return;
      }
      GenericWidget<?> anchor;
      if (lastTabItem != null)
      {
         anchor = lastTabItem;
      }
      else
      {
         anchor = tabItems.get(tabItems.size() - 1);
      }
      
      lastTabItem = (GenericWidget<?>) widget;
      
      setTabItemAfter((GenericWidget<?>) widget, anchor);
      
      LogicalTerminal.updateTabItemList(this);
   }
   
   /**
    * Extracts field-group widgets.
    * 
    * @return   widgets collection.
    */
   public Collection<GenericWidget<?>> getWidgets()
   {
      return Collections.unmodifiableList(widgets);
   }
   
   /**
    * Adding new dynamic widget to the current group.
    *
    * @param    dynWidget
    *           The widget to be added to this group.
    */
   public void addDynamicWidget(GenericWidget<?> dynWidget)
   {
      addDynamicWidget(dynWidget, true);
   }
   
   /**
    * Adding new dynamic widget to this group.
    *
    * @param    dynWidget
    *           The widget to be added to this group.
    * @param    notifyOwnerFrame
    *           The true value indicates that the owner frame must be notified about new widget.
    */
   public void addDynamicWidget(GenericWidget<?> dynWidget, boolean notifyOwnerFrame)
   {
      // not expandable
      if (widgetAdder == null)
      {
         throw new UnsupportedOperationException("The FieldGroup is not expandable");
      }
      else
      {
         if (widgets.contains(dynWidget))
         {
            return;
         }
         
         addWidget(dynWidget);
         
         if (notifyOwnerFrame)
         {
            widgetAdder.accept(dynWidget);
         }
      }
   }
   
   /**
    * Set the value of the widgetAdder Consumer.
    *
    * @param    widgetAdder
    *           the new value of the widgetAdder.
    */
   public void setWidgetAdder(Consumer<GenericWidget<?>> widgetAdder)
   {
      this.widgetAdder = widgetAdder;
   }
   
   /**
    * Return the current tab item list.
    * 
    * @return   The tab item list
    */
   public int[] getTabItemList()
   {
      if (tabItems != null && !tabItems.isEmpty())
      {
         return tabItems.stream().filter(x -> (x.isTabStop().booleanValue()))
                  .mapToInt(x -> x.getId()).toArray();
      }
      
      return new int[0];
   }
   
   /**
    * Update the current tab item list.
    * 
    * @param    widgetIds
    *           The array of widget ids
    * @param    op
    *           The current tab item list rearranging operation
    * 
    * @return   The tab item list
    */
   public int[] updateTabItemList(int[] widgetIds, UpdatesOp op)
   {
      if (widgetIds != null)
      {
         int num = widgetIds.length;
         for(int i = 0; i < num; i++)
         {
            int id = widgetIds[i];
            GenericWidget<?> widget = LogicalTerminal.getWidgetForId(id);
            placeWidgetInTabList(widget, op, i);
         }
      }
      
      updateExtremeTabItems();
      
      return getTabItemList();
   }
   
   /**
    * Update the current tab item list.
    * 
    * @param    widgets
    *           The array of widgets
    * @param    op
    *           The current tab item list rearranging operation
    * 
    * @return   The tab item list
    */
   public int[] updateTabItemList(List<GenericWidget<?>> widgets, UpdatesOp op)
   {
      if (widgets != null)
      {
         int pos = 0;
         for(GenericWidget<?> widget : widgets)
         {
            placeWidgetInTabList(widget, op, pos);
            pos++;
         }
      }
      
      updateExtremeTabItems();
      
      return getTabItemList();
   }
   
   /**
    * Obtains the next tab item in this field group after the provided widget.
    * 
    * @param    widget
    *           The provided widget
    * 
    * @return   An instance of <code>GenericWidget</code> or <code>null</code>.
    */
   public handle nextTabItem(GenericWidget<?> widget)
   {
      return nextTabItem(widget, false);
   }
   
   /**
    * Obtains the previous tab item in this field group after the provided widget.
    * 
    * @param    widget
    *           The provided widget
    * 
    * @return   An instance of <code>GenericWidget</code> or <code>null</code>.
    */
   public handle previousTabItem(GenericWidget<?> widget)
   {
      return nextTabItem(widget, true);
   }
   
   /**
    * Moves the target widget to the top of this list.
    * 
    * @param    widget
    *           The target widget
    */
   protected void toTop(GenericWidget<?> widget)
   {
      Utils.moveToTopInClass(widgets, widget, w -> { return w.getZOrderClass(); });
   }

   /**
    * Moves the target widget to the bottom of this list.
    * 
    * @param    widget
    *           The target widget
    */
   protected void toBottom(GenericWidget<?> widget)
   {
      Utils.moveToBottomInClass(widgets, widget, w -> { return w.getZOrderClass(); });
   }
   
   /**
    * Obtains the next tab item in this field group after the provided widget following the given
    * direction.
    * 
    * @param    widget
    *           The provided widget
    * @param    reverse
    *           The boolean value that defines the iteration order direction
    * 
    * @return   An instance of <code>GenericWidget</code> or <code>null</code>.
    */
   private handle nextTabItem(GenericWidget<?> widget, boolean reverse)
   {
      Iterator<GenericWidget<?>> widgetsIter;
      
      if (reverse)
      {
         widgetsIter = Iterables.reverse(tabItems).iterator();
      }
      else
      {
         widgetsIter = tabItems.iterator();
      }
      
      boolean notFound = true;
      while(notFound && widgetsIter.hasNext())
      {
         GenericWidget<?> testableWidget = widgetsIter.next();
         if (widget.equals(testableWidget))
         {
            notFound = false;
         }
      }
      
      if (notFound || !widget.isTabStop().booleanValue())
      {
         return new handle();
      }
      
      notFound = true;
      GenericWidget<?> nextWidget = null;
      while(notFound && widgetsIter.hasNext())
      {
         nextWidget = widgetsIter.next();
         if (nextWidget._isRealized() && nextWidget.isTabStop().booleanValue())
         {
            notFound = false;
         }
      }
      
      if (notFound)
      {
         return new handle();
      }
      
      return nextWidget.asWidgetHandle();
   }
   
   /**
    * Indicates, that HandleChain implementation of siblings methods should be used.
    * 
    * @return   TRUE  - if the logic of parent class {@link HandleChain} is used,
    *           FALSE - if own siblings implementation logic is used. 
    */
   protected boolean useHandleChainSiblings()
   {
      return true;
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a frame
      return new String[] {LegacyResource.FRAME};
   }
   
   /**
    * Delete the resource.
    * <p>
    * Field groups need to be manually removed from the registry, when they are deleted.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      boolean delete = super.resourceDelete();
      
      if (delete)
      {
         LogicalTerminal.deregisterWidget(config.id.asInt());
      }
      
      return delete;
   }
   
   /**
    * Accesses the the first and the last field-level widget in the given field group.
    *
    * @param    first
    *           The true value indicates that the first will be accessed, otherwise
    *           the last widget.
    * 
    * @return   The extreme possible field-level widget in this group or
    *           <code>null</code> if there are no field level widgets.
    */
   private GenericWidget<?> getExtremeTabItem(boolean first)
   {
      if (tabItems == null)
      {
         return null;
      }
      
      Iterator<GenericWidget<?>> iter;
      
      if (first)
      {
         iter = tabItems.iterator();
      }
      else
      {
         iter = Iterables.reverse(tabItems).iterator();
      }
      
      // scan first available field-level widget
      
      while (iter.hasNext())
      {
         GenericWidget<?> widget = iter.next();
         if (widget._isRealized() && widget.isTabStop().booleanValue())
         {
            return widget;
         }
      }
      
      return null;
   }
   
   /**
    * Set the tab item after the provided widget in this field-group. 
    * 
    * @param    widget
    *           The tab item.
    * @param    after
    *           The provided widget.
    */
   private void setTabItemAfter(GenericWidget<?> widget, GenericWidget<?> after)
   {
      this.tabItems.remove(widget);
      
      int anchor = this.tabItems.indexOf(after) + 1;
      
      this.tabItems.add(anchor, widget);
      
      updateExtremeTabItems();
   }
   
   /**
    * Set the tab item before the provided widget in this field-group. 
    * 
    * @param    widget
    *           The tab item.
    * @param    before
    *           The provided widget.
    */
   private void setTabItemBefore(GenericWidget<?> widget, GenericWidget<?> before)
   {
      this.tabItems.remove(widget);
      
      int anchor = this.tabItems.indexOf(before);
      
      this.tabItems.add(anchor, widget);
      
      updateExtremeTabItems();
   }
   
   /**
    * Updates the first and last tab items if there are initialized.
    */
   private void updateExtremeTabItems()
   {
      if (firstTabItem != null)
      {
         firstTabItem = getExtremeTabItem(true);
      }
      
      if (lastTabItem != null)
      {
         lastTabItem = getExtremeTabItem(false);
      }
   }

   /**
    * Get the NUM-TABS attribute, which indicates the number of tab widgets inside this
    * field-group container.
    *
    * @return   The current value of the NUM-TABS attribute.
    */
   @Override
   public integer getNumTabs()
   {
      Iterator<GenericWidget<?>> widgetsIter = tabItems.iterator();
      
      int count = 0;
      
      while(widgetsIter.hasNext())
      {
         GenericWidget<?> nextWidget = widgetsIter.next();
         if (nextWidget._isRealized() && nextWidget.isTabStop().booleanValue())
         {
            count++;
         }
      }
      
      return new integer(count);
   }

   /**
    * Get the handle of the tab item by the given tab position.
    * 
    * @param    tabPosition
    *           The tab position
    * 
    * @return   The handle of the tab item by the given tab position.
    */
   @Override
   public handle getTabItem(integer tabPosition)
   {
      if (tabPosition.isUnknown())
      {
         return new handle();
      }
      
      return getTabItem(tabPosition.intValue());
   }
   
   /**
    * Get the handle of the tab item by the given tab position.
    * 
    * @param    tabPosition
    *           The tab position
    * 
    * @return   The handle of the tab item by the given tab position.
    */
   @Override
   public handle getTabItem(int tabPosition)
   {
      if (tabPosition <= 0)
      {
         return new handle();
      }
      
      Iterator<GenericWidget<?>> widgetsIter = tabItems.iterator();
      
      int count = 0;
      
      while(widgetsIter.hasNext())
      {
         GenericWidget<?> nextWidget = widgetsIter.next();
         if (nextWidget._isRealized() && nextWidget.isTabStop().booleanValue())
         {
            count++;
            if (count == tabPosition)
            {
               return nextWidget.asWidgetHandle();
            }
         }
      }
      
      return new handle();
   }
   
   /**
    * Returns the tab position of the given tab item. If the given widget doesn't belong to
    * this field-group, then it returns unknown value.
    * 
    * @param    widget
    *           The tab item.
    * 
    * @return   The tab order of a widget within its field group.
    */
   public integer getTabPosition(GenericWidget<?> widget)
   {
      if (!widget.isTabStop().booleanValue())
      {
         return new integer();
      }
      
      Iterator<GenericWidget<?>> widgetsIter = tabItems.iterator();
      
      int tabPosition = 0;
      
      while(widgetsIter.hasNext())
      {
         GenericWidget<?> nextWidget = widgetsIter.next();
         
         if (nextWidget._isRealized() && nextWidget.isTabStop().booleanValue())
         {
            tabPosition++;
            
            if (widget.equals(nextWidget))
            {
               return new integer(tabPosition);
            }
         }
      }
      
      return new integer();
   }
   
   /**
    * Defines the tab item list rearrangements.
    */
   public enum UpdatesOp
   {
      /** Defines the tab item list changes due to ENABLE statement */
      ENABLE,
      
      /** Defines the tab item list changes due to PROMPT-FOR statement */
      PROMPT_FOR,
      
      /** Defines the tab item list changes due to SET statement */
      SET,
      
      /** Defines the tab item list changes due to UPDATE statement */
      UPDATE
   }
}