/*
** Module   : FrameInterface.java
** Abstract : Define all FRAME-specific widget attributes and methods available via a handle
**            instance.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20141120 Created initial version.
** 002 IAS 20150922 Added support for the OVERLAY attribute.
** 003 OM  20150915 Added DEFAULT-BUTTON, CANCEL-BUTTON, CURRENT-ITERATION, BORDER-???-CHARS, 
**                  BORDER-???-PIXELS and BACKGROUND attributes.
** 004 EVL 20160621 Adding inheritance from IterableResource interface.  Otherwise current
**                  iteration can not be query and set by frame or dialog handle.  The iteration
**                  related methods are now defined in base interface.
** 005 CA  20180501 Added SIDE-LABELS attribute.
**     CA  20180514 Added CAREFUL-PAINT.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Define all FRAME-specific widget attributes and methods which can be accessed via a handle
 * instance.  Attributes and methods shared by multiple widget types will be defined elsewhere.
 */
public interface FrameInterface
extends IterableResource
{
   /**
    * Get the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @return   <code>true</code> if the frame is centered.
    */
   @LegacyAttribute(name = "CENTERED")
   public logical isCentered();

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   @LegacyAttribute(name = "CENTERED", setter = true)
   public void setCentered(boolean center);

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   @LegacyAttribute(name = "CENTERED", setter = true)
   public void setCentered(logical center);
   
   /**
    * Get the default button of this frame.
    *  
    * @return  handle to current default button.
    */
   @LegacyAttribute(name = "DEFAULT-BUTTON")
   public handle getDefaultButton();
   
   /**
    * Set the default button of this frame.
    *
    * @param   defButton
    *          handle to current default button.
    */
   @LegacyAttribute(name = "DEFAULT-BUTTON", setter = true)
   public void setDefaultButton(handle defButton);
   
   /**
    * Gets the cancel button for this frame.
    *
    * @return  button widget that receives the CHOOSE event when a user cancels the current 
    *          frame or dialog box by pressing the ESC key.
    */
   @LegacyAttribute(name = "CANCEL-BUTTON")
   public handle getCancelButton();
   
   /**
    * Sets the cancel button for this frame.
    *
    * @param    cancelButton
    *           button widget that receives the CHOOSE event when a user cancels the current frame
    *           or dialog box by pressing the ESC key.
    */
   @LegacyAttribute(name = "CANCEL-BUTTON", setter = true)
   public void setCancelButton(handle cancelButton);
   
   /**
    * Set the value of the OVERLAY attribute which means that the frame can overlay other frames
    * on the display.
    *
    * @param    overlay
    *           {@code true} if the frame can overlay other frames.
    */
   @LegacyAttribute(name = "OVERLAY", setter = true)
   public void setOverlay(boolean overlay);
   
   /**
    * Set the value of the CENTERED attribute which means that the frame can overlay other frames
    * on the display.
    *
    * @param    overlay
    *           {@code true} if the frame can overlay other frames.
    */
   @LegacyAttribute(name = "OVERLAY", setter = true)
   public void setOverlay(logical overlay);
   
   /**
    * Get the value of the OVERLAY attribute which means that the frame can overlay other frames
    * on the display.
    *
    * @return   {@code true} if the frame can overlay other frames.
    */
   @LegacyAttribute(name = "OVERLAY")
   public logical isOverlay();
   
   /**
    * Get the thickness (in character units) of the border at the top of the frame or dialog box.
    * If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-TOP-CHARS")
   public decimal borderTopChars();
   
   /**
    * Get the thickness (in character units) of the border at the left of the frame or dialog box.
    * If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-LEFT-CHARS")
   public decimal borderLeftChars();
   
   /**
    * Get the thickness (in character units) of the border at the right of the frame or dialog
    * box. If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-RIGHT-CHARS")
   public decimal borderRightChars();
   
   /**
    * Get the thickness (in character units) of the border at the bottom of the frame or dialog
    * box. If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-BOTTOM-CHARS")
   public decimal borderBottomChars();
   
   /**
    * Get the thickness (in pixels) of the border at the top of the frame or dialog box.
    * If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-TOP-PIXELS")
   public integer borderTopPixels();
   
   /**
    * Get the thickness (in pixels) of the border at the left of the frame or dialog box.
    * If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-LEFT-PIXELS")
   public integer borderLeftPixels();
   
   /**
    * Get the thickness (in pixels) of the border at the right of the frame or dialog
    * box. If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-RIGHT-PIXELS")
   public integer borderRightPixels();
   
   /**
    * Get the thickness (in pixels) of the border at the bottom of the frame or dialog
    * box. If the BOX attribute for the widget is {@code yes}, this attribute returns 0.
    *
    * @return  the thickness of the border at the top of the frame.
    */
   @LegacyAttribute(name = "BORDER-BOTTOM-PIXELS")
   public integer borderBottomPixels();
   
   /**
    * Get the handle for the background iteration of the frame.
    *
    * @return  the handle for the background iteration of the frame.
    */
   @LegacyAttribute(name = "BACKGROUND")
   public handle backgroundIteration();
   
   /**
    * Sets the SIDE-LABELS attribute.
    * 
    * @param   sideLabels
    *          <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "SIDE-LABELS", setter = true)
   public void setSideLabels(boolean sideLabels);
   
   /**
    * Sets the SIDE-LABELS attribute.
    * 
    * @param   sideLabels
    *          <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "SIDE-LABELS", setter = true)
   public void setSideLabels(logical sideLabels);
   
   /**
    * Get the SIDE-LABELS attribute.
    * 
    * @return   <code>true</code> if the attribute is ON.
    */
   @LegacyAttribute(name = "SIDE-LABELS")
   public logical isSideLabels();
   
   /**
    * Get the CAREFUL-PAINT attribute.
    * 
    * @return   <code>true</code> if the attribute is ON.
    */
   @LegacyAttribute(name = "CAREFUL-PAINT")
   public logical isCarefulPaint();
   
   /**
    * Sets the CAREFUL-PAINTattribute.
    * 
    * @param   careful
    *          <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "CAREFUL-PAINT", setter = true)
   public void setCarefulPaint(logical careful);
   
   /**
    * Sets the CAREFUL-PAINTattribute.
    * 
    * @param   careful
    *          <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "CAREFUL-PAINT", setter = true)
   public void setCarefulPaint(boolean careful);
}