/*
** Module   : LiteralWidget.java
** Abstract : server-side literal widget representation.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130927 Created initial version.
** 002 CA  20131115 Changes related to properly instantiate dynamic resources.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 CA  20141220 Track if an explicit VIEW-AS TEXT phrase is in use.  GUI has code dependent on
**                  this fact.
** 005 CA  20150222 Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 006 HC  20150405 Refactoring of config coordinate fields.
** 007 CA  20150725 Removed canSetPixelPositionInChui - all LITERAL widgets can be moved.
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 HC  20180628 Added side-labels to the widget chain.
** 010 CA  20180723 Fixed NAME, LABEL and SCREEN-VALUE attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Represents a LITERAL widget on server-side. This widget is used in following cases:
 * <ul>
 *   <li>FORM "string-literal"</li>
 * </ul>
 */
@LegacyResource(resource = LegacyResource.LITERAL)
public class LiteralWidget
extends ControlTextWidget<TextConfig>
{
   /**
    * Side label discriminator, when true this instance represents a side label.
    */
   private boolean sideLabel = false;
   
   /** The associated side widget for this label (when {@link #sideLabel} is true). */
   private GenericWidget<?> sideWidget = null;
   
   /**
    * Default constructor. To create a side label, instead use
    * {@link #createSideLabel(GenericWidget)}.
    */
   public LiteralWidget()
   {
      super(false, new TextConfig());
      setStatic(true);
   }
   
   /**
    * Creates new instance representing a side label.
    * 
    * @param    sideWidget
    *           The associated widget for this side-label.
    * 
    * @return   A reference to {@link LiteralWidget}. 
    */
   public static LiteralWidget createSideLabel(GenericWidget<?> sideWidget)
   {
      LiteralWidget result = new LiteralWidget();
      result.sideLabel = true;
      result.sideWidget = sideWidget;
      
      return result;
   }

   /**
    * Mark this widget as using an explicit VIEW-AS TEXT phrase.
    * 
    * @param    explicit
    *           <code>true</code> when there is a VIEW-AS TEXT phrase.
    */
   public void explicitViewAs(boolean explicit)
   {
      config.explicitViewAs = explicit;
   }

   /**
    * This widget can't be a tab item.
    * 
    * @return   False
    */
   @Override
   public boolean isTabWidget()
   {
      return false;
   }

   /**
    * Obtains the next sibling widget in the current field group.
    *
    * @return   The next sibling or an unknown handle if this is the only widget or if this is the
    *           last widget.
    */
   @Override
   public handle getNextSibling()
   {
      if (sideLabel && group instanceof FieldGroup)
      {
         // find the related widget
         for (GenericWidget w : ((FieldGroup) group).getWidgets())
         {
            if (!(w.config instanceof ControlConfig))
            {
               continue;
            }

            if (config.id.asInt() == ((ControlConfig) w.config).sideLabelId)
            {
               return new handle(w);
            }
         }
      }

      return super.getNextSibling();
   }

   /**
    * Obtains the previous sibling widget in the current field group.
    *
    * @return   The previous sibling or an unknown handle if this only widget or if this is the
    *           first widget.
    */
   @Override
   public handle getPrevSibling()
   {
      if (sideLabel && group instanceof FieldGroup)
      {
         // find the related widget
         for (GenericWidget w : ((FieldGroup) group).getWidgets())
         {
            if (!(w.config instanceof ControlConfig))
            {
               continue;
            }

            if (config.id.asInt() == ((ControlConfig) w.config).sideLabelId)
            {
               return new handle(((FieldGroup) group).getPrevChild(w));
            }
         }
      }

      return super.getPrevSibling();
   }

   /**
    * Get the label text for this widget.
    * <p>
    * Unlike other widget attribute getters, this API will raise an ERROR condition.
    * 
    * @return   The empty string.
    */
   @Override
   public character getLabel()
   {
      notQueryable("LABEL", false);
      return new character("");
   }

   /**
    * This API is a no-op - it just raises an ERROR condition.
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      if (frame == null || !frame.isInsideSetup())
      {
         notSettable("LABEL", false);
      }
      else
      {
         super.setLabel(label);
      }
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue()
   {
      if (!sideLabel)
      {
         return super.getScreenValue();
      }
      
      return sideWidget.getLabel();
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @Override
   public void setScreenValue(character value)
   {
      if (!sideLabel)
      {
         super.setScreenValue(value);
      }
      else
      {
         sideWidget.setLabel(value);
      }
   }
   
   /**
    * Get the name of this LITERAL widget - always unknown value.
    * 
    * @return    Always unknown value.
    */
   @Override
   public character name()
   {
      return new character();
   }
   
   /**
    * Sets the row or column position for this widget.
    * <p>
    * Sub-classes can override this method; at this point, the parameters are valid.
    * 
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    */
   @Override
   protected void setColumnOrRowWorker(boolean isUnknown, double value, boolean row)
   {
      int align = config.align;
      
      super.setColumnOrRowWorker(isUnknown, value, row);
      
      // preserve alignment when changing side label position
      if (sideLabel)
      {
         config.align = align;
      }
      
      if (row)
      {
         // nothing to do for rows
         return;
      }
      
      updateSideLabelColumn(value);
   }
   
   /**
    * Sets the x or y pixel position for this widget.
    * <p>
    * Sub-classes can override this method; at this point, the parameters are
    * valid.
    * 
    * @param    isUnknown
    *           If
    *           <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The x or y pixel value.
    * @param    x
    *           If <code>true</code> <code>value</code> represents an x
    *           position, otherwise y.
    */
   @Override
   protected void setXOrYWorker(boolean isUnknown, int value, boolean x)
   {
      int align = config.align;
      
      super.setXOrYWorker(isUnknown, value, x);
      
      // preserve alignment when changing side label position
      if (sideLabel)
      {
         config.align = align;
      }
      
      if (!x)
      {
         // nothing to do for y pixel position change
         return;
      }
      
      // use the converted column value set in the super implementation
      updateSideLabelColumn(config.column);
   }

   /**
    * The method updates side label metrics after a change of column
    * position. If this widget doesn't represent a side label, this
    * method is a no-op.
    * 
    * @param    col
    *           New widget column position.
    */
   private void updateSideLabelColumn(double col)
   {
      if (!sideLabel)
      {
         return;
      }
      
      // if the literal is not yet realized, compute its width
      if (!config.realized && config.name != null && config.align == ControlEntity.ALIGN_RIGHT)
      {
         if (config.name.length() > col - 1)
         {
            config.column = 0;
            config.widthChars = col - 1;
         }
         else
         {
            config.widthChars = config.name.length() + 1;
         }
      }
   }
}