/*
** Module   : SliderConfig.java
** Abstract : contains data for slider widget
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150715 Created initial version.
** 002 EVL 20160127 Changed superclass from WidgetConfig to ControlConfig.  The default frequency
**                  value is 0.  The default value for NO-CURRENT-VALUE is true.
** 003 IAS 20160217 MOUSE-POINTER attribute support.
** 004 OM  20180713 Fixed TIC-MARKS attribute.
** 005 SBI 20180716 Changed to set z-order widget class.
** 006 SBI 20180723 Fixed the slider z-order class to be the default WIDGET.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * Contains data for slider widget.
 */
public class SliderConfig
extends ControlConfig
{
   /** 
    * The tic marks disabled mode. Note that this value makes the Slider widget to be considered
    * old, migrated code and will be painted differently. For all other cases, the trackbar on the
    * slider widget will be relatively large. 
    */
   public static final String TM_NOT_SET = "NOT-SET";
   
   /** The tic marks are not drawn. */
   public static final String TM_NONE = "NONE";
   
   /** The tic marks left side mode. */
   public static final String TM_LEFT = "LEFT";
   
   /** The tic marks right side mode. */
   public static final String TM_RIGHT = "RIGHT";
   
   /** The tic marks top side mode. */
   public static final String TM_TOP = "TOP";
   
   /** The tic marks bottom side mode. */
   public static final String TM_BOTTOM = "BOTTOM";
   
   /** The tic marks both side mode. */
   public static final String TM_BOTH = "BOTH";
   
   /** Minimum value of the slider. */
   public int minValue = 0;
   
   /** Maximum value of the slider. */
   public int maxValue = 0;
   
   /** Current value of the slider. */
   public int currentValue = 0;
   
   /** Widget orientation, horizontal or vertival. */
   public boolean horizontal = true;
   
   /** The numeric range is large to small, override the default value is small to large. */
   public boolean largeToSmall = false;
   
   /** Flag to suppress display the current value for given position on slider control. */
   public boolean noCurrentValue = true;
   
   /** Tick marks mode. */
   public String ticMarks = TM_NOT_SET;
   
   /** How often to display the tic marks. */
   public int frequency = 0;

   /** MOUSE-POINTER attribute */
   public String mousePointer = null;

   /**
    * Default constructor (only used in deserialization).
    */
   public SliderConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected SliderConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected SliderConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      SliderConfig cfg = (SliderConfig) config;
      
      minValue = cfg.minValue;
      maxValue = cfg.maxValue;
      currentValue = cfg.currentValue;
      horizontal = cfg.horizontal;
      largeToSmall = cfg.largeToSmall;
      noCurrentValue = cfg.noCurrentValue;
      ticMarks = cfg.ticMarks;
      frequency = cfg.frequency;
      mousePointer = cfg.mousePointer;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      minValue = in.readInt();
      maxValue = in.readInt();
      currentValue = in.readInt();
      horizontal = in.readBoolean();
      largeToSmall = in.readBoolean();
      noCurrentValue = in.readBoolean();
      ticMarks = (String) in.readObject();
      frequency = in.readInt();
      mousePointer = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(minValue);
      out.writeInt(maxValue);
      out.writeInt(currentValue);
      out.writeBoolean(horizontal);
      out.writeBoolean(largeToSmall);
      out.writeBoolean(noCurrentValue);
      out.writeObject(ticMarks);
      out.writeInt(frequency);
      out.writeObject(mousePointer);
  }
}
