/*
** Module   : SubMenuWidget.java
** Abstract : server side sub-menu widget implementation
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20141124 Created initial version.
** 002 VIG 20150115 Changed base class, added initialization method.
** 003 VIG 20150302 Added name attribute setting, firstChild, lastChild methods.
**                  Added findSubMenuStatic implementation.
** 004 VIG 20150724 Removed methods which are already defined in BaseEntity.
** 005 VIG 20150823 Fixed firstChild and lastChild methods, removed mnemonic field.
** 006 GES 20150923 Javadoc fix.
** 007 HC  20170131 Added runtime support for dynamic MENU* widgets.
** 008 CA  20170929 Fixed LABEL attribute setter must call pushScreenDefinition.
**                  Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
**                  Improved menu registration with client-side.
** 009 CA  20180130 Replaced pushScreenDefinition() with pushWidgetAttr() for cases when business 
**                  logic is assigning a widget attribute.
** 010 CA  20180710 Fixed dynamic menu registration.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * This class represents sub-menu widget.
 */
@LegacyResource(resource = LegacyResource.SUB_MENU)
public class SubMenuWidget
extends MenuContainerWidget<SubMenuConfig>
implements MnemonicInterface
{
   /**
    * Default constructor.  
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public SubMenuWidget(boolean dynamic)
   {
      super(dynamic, new SubMenuConfig());
   }

   /**
    * Gets the mnemonic for this menu-item or sub-menu.
    * This is a FWD extension, not a real 4GL attribute.
    * 
    * @return  See above.
    */
   public integer getMnemonic()
   {
      return config.mnemonic == null ? new integer() : new integer(config.mnemonic);
   }

   /**
    * Gets the preprocessed label text, as it appears in the UI.
    * This is a FWD extension, not a real 4GL attribute.
    * 
    * @return  See above.
    */
   public character getPreprocessedLabel()
   {
      return config.preprocessedLabel == null 
               ? new character() 
               : new character(config.preprocessedLabel);
   }

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   public character getLabel()
   {
      return new character(config.title);
   }
   
   /**
    * Sets the label text, parsing mnemonic.
    * 
    * TODO: find out what is going on with undocumented feature of very long label
    * See parent's javadoc.
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      config.title = label;
      pushWidgetAttr("title", config.title);
   }
   
   /**
    * Sets new value of SENSITIVE attribute. Overrides {@link GenericWidget} logic 
    * (that enables widget in its frame).
    *
    * @param    s
    *           new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      // changing the SENSITIVE attribute does not affect the widget's visibility
      config.enabled = s;
      pushWidgetAttr("enabled", config.enabled);
   }
   
   /**
    * Finds menu-item in this menu tree. It can appear on different levels.
    * Menu-item is searched using some sort of depth-first algorithm.
    * See {@link MenuWidget#findMenuItemInt(String)}
    * 
    * @param   name
    *          MENU-ITEM name.
    *          
    * @return  concrete {@link MenuItemWidget} or null if not found.
    */
   public MenuItemWidget findMenuItem(String name)
   {
      return findMenuItemInt(name);
   }
   
   /**
    * Finds sub-menu in this sub-menu tree. It can appear only once. 
    * Simple depth-first search is used.
    * 
    * @param   name
    *          Sub-menu name.
    * 
    * @return  concrete {@link SubMenuWidget} or null if not found.
    */
   public SubMenuWidget findSubMenu(String name)
   {
      return findSubMenuInt(name);
   }
   
   /**
    * Accesses the first window in the list of contained windows, having this window as parent.
    *
    * @return   The first widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      Iterator<GenericWidget<?>> it = widgets().iterator();
      return it.hasNext() ? new handle(it.next()) : new handle();
   }
   
   /**
    * Accesses the last window in the list of contained windows, having this window as parent.
    *
    * @return   The last widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      Iterator<GenericWidget<?>> it = widgets().iterator();
      GenericWidget<?> last = null;
      while (it.hasNext())
      {
         last = it.next();
      }
      return last != null ? new handle(last) : new handle();
   }
   
   /**
    * Creates new sub-menu as a proxy object (as a local resource).  This loads
    * the static sub-menu definition.
    *
    * @param    <T> 
    *           Sub-menu definition class type.
    * @param    defCls
    *           The class of sub-menu definition.
    * @param    name
    *           The name of sub-menu.          
    *
    * @return   The created sub-menu widget.
    */
   public static <T extends SubMenuDefinition> T createStaticSubMenu(Class<T> defCls, String name)
   {
      SubMenuWidget subMenu = new SubMenuWidget(false);
      subMenu.name(name);
      
      final T menuDef;
      try
      {
         menuDef = defCls.getConstructor(new Class<?>[]{}).newInstance();
         subMenu.config.id = new WidgetId(LogicalTerminal.registerMenu(subMenu, true));
         subMenu.config.name = name;
         menuDef.setSelf(subMenu);
         menuDef.setup(subMenu);
         subMenu.coreInitialize(defCls, menuDef, false);
      }
      catch (NoSuchMethodException    | 
             SecurityException        | 
             InstantiationException   | 
             IllegalAccessException   | 
             IllegalArgumentException | 
             InvocationTargetException e)
      {
         throw new RuntimeException("Failed to instantiate sub-menu definition class", e);
      }
      
      return menuDef;
   }
   
   /**
    * Finds the first SUB-MENU with specified name in all menus.
    * 
    * @param   name
    *          The name of searched SUB-MENU
    *          
    * @return  The first found SubMenuWidget or null.
    */
   public static SubMenuWidget findSubMenuStatic(String name)
   {
      for (MenuWidget menu : LogicalTerminal.getAllMenus())
      {
         SubMenuWidget found = menu.findSubMenu(name);
         if (found != null)
         {
            return found;
         }
      }
      
      return null;
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a menu or sub-menu
      return new String[] {LegacyResource.MENU, LegacyResource.SUB_MENU};
   }
}