/*
** Module   : RedirectedTerminal.java
** Abstract : an API for rendering formatted output to a stream
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20060222   @24663 A plugin which provides a compatible back-
**                           end for formatted output which acts like a
**                           terminal but for which the transformed data
**                           is written to a stream.
** 002 GES 20060225   @24738 Minor change in stream interface usage for
**                           writing.
** 003 GES 20060228   @24782 Removed unnecessary try/catch.
** 004 GES 20060814   @28592 Honor clipping in sync() since setCursor()
**                           can modify data structures even outside of
**                           the clipping region. At this time, we only
**                           honor vertical clipping, since we assume
**                           that any horizontal clipping is not
**                           meaningful during sync() of redirected
**                           output.
** 005 NVS 20070716   @34540 Added syncable state and setSyncable() method
**                           that controls it. Required to support partial
**                           DISPLAY where multiple statements can be used
**                           to form single line of output.
** 006 NVS 20070724   @34673 Redone #005. The buffered output is tracked
**                           by the frame ID and gets flushed when the ID
**                           changes. For the same frame ID, the output
**                           remains buffered as long as flush() method is
**                           not used.
** 007 NVS 20070730   @34736 targetFrame() now returns the ID of the going
**                           frame to simplify its hiding.
** 008 SIY 20070806   @34806 Dynamical adjusting of terminal width.
** 009 NVS 20070925   @35213 Fixed bug in clipping rectangle calculation.
**                           Added setSyncable(boolean) method to allow
**                           direct control over the syncable state.
**                           Added debugging aid method view(). Useful in
**                           the debugging session to print the contents
**                           of the terminal buffers with "invoke view()".
** 010 NVS 20071011   @35430 More debug info is included in the view().
** 011 SIY 20071116   @35821 Added method to clear frame ID reference.
**                           Ref: #35800.
** 012 NVS 20071203   @36139 Since the update #118 to Frame.java, "in
**                           redirected mode draw even frames with zero
**                           width, because it may contain SKIP(n)",
**                           the call to blankBox() may happen for boxes
**                           of zero width. This update handles those
**                           properly.
** 014 NVS 20071229   @36612 Added isFresh() method which return true if
**                           the buffer is empty.
** 015 GES 20080103   @36662 On first output, notify the stream to force
**                           page headers to display if needed. This
**                           ensures that the page headers are processed
**                           immediately rather than waiting until this
**                           content is subsequently flushed (which could
**                           be an arbitrary amount of time later,
**                           possibly after those page headers are
**                           deregistered). This solves a problem where
**                           there was a missing page header because of
**                           ordering of language statements in the 4GL.
** 016 NVS 20080109   @36756 Change 015 caused a regression due to a
**                           recursive nature of the view() call now.
**                           Invalidation rectangle is preserved across
**                           the notification call.
** 017 NVS 20080110   @36769 Fixed a bug in blankBox() method. One line
**                           tall boxes were not blanked.
** 018 GES 20080116   @36827 Efficiency improvement in firstOutput() to
**                           avoid header processing when it isn't needed.
** 019 SIY 20080207   @36969 Minor cleanups in comments.
** 020 CA  20080229   @37212 Not all frames might initialize while in an
**                           interactive environment - so, for page
**                           element frames which initialize while in a
**                           redirected env, the screen width must be
**                           set when computing the frame layout.
**                           Ref: #37211
** 021 SIY 20080306   @37315 Extracted header processing in separate method
**                           processHeaders() and needsHeadersOutput()
**                           method.
**                           Ref: #37313
** 022 GES 20080421   @38062 I18N charset support.
** 023 NVS 20080812   @39371 Added support for redirected editors output.
**                           Two new methods were added, markHeader() and
**                           markBreakLines() and existing sync() method
**                           was extended. See methods for details.
** 024 NVS 20080826   @39596 Replaced invalidation rectangle technique
**                           with invalidation region.
** 025 NVS 20080828   @39626 Fixed clipping bug in addChar() where the
**                           cursor should move regardless, one position
**                           at a time.
** 026 CA  20080930   @39982 Added the possibility to move the entire output
**                           to a different column.
** 027 NVS 20081010   @40075 Fixed clipping bug in drawBox() where the
**                           cases of non-intersecting rectangles were not
**                           checked.
** 028 CA  20081114   @40446 All output primitives implementations will
**                           always draw data. Clipping responsibility
**                           belongs to the Toolkit class.
** 029 NVS 20090209   @41279 Due to changes in OutputDirectives interface
**                           added setCursorStay() and refresh() methods.
** 030 SIY 20090309   @41542 Added two stubs in order to satisfy interface
**                           definition.
** 031 SIY 20090319   @41642 Changed addArray() signature.
** 032 SVL 20090612   @42680 Notify BufferSizeManager about DISPLAY
**                           operations.
** 033 SIY 20090911   @43892 Now class is inherited from BasePrimitives.
** 034 SIY 20090925   @44035 Inform stream about number of pending (prepared 
**                           but not flushed yet) rows in the terminal.
** 035 SIY 20100619          Switching to own implementation of character UI.
** 036 SIY 20100804          Changes in imports, some refactorings.
** 037 EVL 20101220          Changining the approach of calculation the value
**                           for pending row counter. The correct value is very
**                           important for valid getNextLineNum() returns.
**                           Previously this value was computing in sync() call.
**                           This is not acceptable for the cases when we have
**                           several different frame outputting to the same
**                           stream. The result is the 4GL line-counter()
**                           function returns the value - 1 comparing to the
**                           valid one. The new approach changes:
**                           - Eliminating all pending row counter calculation
**                             in sync() method
**                           - The pending rows are now computed in cursorAt()
**                             method. We keep track on last - first changes
**                           - Initially the pending row == 1 on the first pass
**                           - In a next calls we compare new last - first with
**                             the previous one and if new one is more we
**                             increase the pending row value accordingly.
** 038 SIY 20110112          Changes related to new color handling scheme.
** 039 CA  20140805          Changes related to OutputManager refactoring.
** 040 HC  20140911          GUI coordinate support - data types refactoring.
** 041 HC  20141021          Implemented window move/position support.
** 042 HC  20151013          Changes to extend GUI multi-window focus management and ACTIVE-
**                           WINDOW system handle processing to modal windows.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Provides an implementation which with a compatible back-end for formatted
 * output acts like a terminal but for which the transformed data is written to
 * a stream.  This allows the output of the thin client's standard user
 * interface functions to be redirected to a stream with minimal changes.
 * <p>
 * The virtual terminal is the internal buffers/state of this class and the
 * physical terminal is the output stream.
 * <p>
 * Progress' semantics of streams require an instance of this class for every
 * stream.
 * <p>
 * The output to the redirected terminal remains buffered for as long as the
 * target frame remains unchanged. For this reason, the terminal maintains the
 * ID of the last used frame and flushes the buffered output, if any, the same
 * moment the ID change is detected.
 */
public class RedirectedTerminal
extends BasePrimitives
{
   /** Current logical width of terminal. */
   private int width = 0;
   
   /** Number of columns on the screen. */
   private int cols = 0;
   
   /** Number of lines on the screen. */
   private int lines = 0;
   
   /** The output destination. */
   private Stream out = null;
   
   /** The current X coordinate of the cursor. */
   private int cx = 0;
   
   /** The current Y coordinate of the cursor. */
   private int cy = 0;
   
   /** <code>true</code> if the last operation was flushing to the stream. */
   private boolean reset = true;
   
   /** The first line of valid output. */
   private int first = 0;
   
   /** The last line of valid output. */
   private int last = 0;
   
   /** The number of valid characters in each corresponding line. */
   private int[] num = null;
   
   /** The whether a box was requested for each corresponding line. */
   private boolean[] box = null;
   
   /** The 2-dimensional array of characters which is the virtual terminal. */
   private char[][] buffer = null;
   
   /** Terminal is syncable flag */
   private boolean syncable = false;
   
   /** The frame ID of the frame currently owning the buffer */
   private int lastId = -1;
   
   /** Set of line numbers where page breaks may occur */
   private Set<Integer> breakable = new HashSet<Integer>();

   /** Number of header lines in this frame */
   private int headerLines = 0;

   /** Accumulated headers */
   private String headers = null;

   /** breakable output is in headers processing */
   private boolean inHeaders = false;

   /** The screen state, which keeps track of all the clipped regions. */
   private ScreenBitmap bitmap = null;
   
   /** The coordinates conversion implementation. */ 
   private CoordinatesConversion conversion = CoordinatesConversion.chuiInstance();

   /**
    * Initializes an instance for a specific output destination.
    *
    * @param    out
    *           The output destination.  Must not be <code>null</code>.
    * @param    cols
    *           The width of each line in characters.
    * @param    lines
    *           The number of lines in the "screen".
    */
   public RedirectedTerminal(Stream out, int cols, int lines)
   {
      this.out   = out;
      this.cols  = cols;
      this.lines = lines;
      
      num    = new int[lines];
      box    = new boolean[lines];
      buffer = new char[lines][cols];
      
      clear();
         
      // when creating a new redirection, set the redirected terminal width to  
      // the screen width. this is because there can be header/footer frames  
      // which will initialize and register as a page element after the  
      // redirection is done (i.e. when first redirecting to an unnamed stream 
      // and after that running an external procedure which will output to the 
      // unnamed stream). if not setting the width of the redirected terminal
      // here, then ZeroColumnLayout.maxWidth will compute an incorrect width
      // for the frame - because the reported screen width will be zero.
      Window<?> wnd = WindowManager.resolveWindow();
      this.width = conversion.widthToNative(wnd.getScreenDimension().width);
   }

   /**
    * Get the screen bitmap for this output primitives implementation. 
    * 
    * @return   The screen bitmap.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      if (bitmap == null)
      {
         bitmap = new ScreenBitmap(lines, cols);
      }
      return bitmap;
   }

   /**
    * Replace the current bitmap instance with a previous one.
    * 
    * @param    bitmap
    *           The new bitmap instance.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      this.bitmap = bitmap;
   }

   /**
    * Add a character to the virtual terminal at the current cursor position.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           The color to use.
    */
   @Override
   public void append(int chr, Color color)
   {
      addChar(Utils.toChar(chr), color);
   }
   
   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The text to output.
    * @param    color
    *           The color to use.
    */
   @Override
   public void append(String str, Color color)
   {
      int len = str.length();
      
      for (int i = 0; i < len; i++)
      {
         addChar(str.charAt(i), color);
      }
   }
   
   /**
    * Draw a horizontal line of the specified length starting at the current
    * cursor position.  Always uses a '-' to draw.
    *
    * @param    length
    *           The length of the line to draw in columns.
    * @param    color
    *           The color to use.
    */
   @Override
   public void hLine(int length, Color color)
   {
      drawHorizontalLine(length, '-', color);
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.  Always uses a ' ' (space character) to draw.
    *
    * @param    length
    *           The length of the line to draw in rows.
    * @param    color
    *           The color to use.
    */
   @Override
   public void vLine(int length, Color color)
   {
      for (int i = 0; i < length; i++)
      {
         addChar(' ', color);
         cursorAt(cx - 1, cy + 1);
      }
   }
   
   /**
    * Blank out a box using the specified attributes and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The color to use.
    */
   @Override
   public void clearArea(int left, int top, int right, int bottom, Color color)
   {
      // this only erases output that may be there, it does not cause any
      // new output to be generated (so there is no need to call the
      // firstOutput() method)
      if (!reset)
      {
         // iterate through each row from top to bottom
         int vstart = top;
         int vend   = bottom;
         
         // blank the data but DON'T change the valid region or cursor position
         int lstart = left;
         int lend   = right;

         // protect against empty boxes
         if (lstart > lend)
         {
            return;
         }

         for (int i = vstart; i <= vend; i++)
         {
            Arrays.fill(buffer[i], lstart, lend, ' ');
         }
      }
   }
   
   /**
    * Draw the upper-left border corner. This is a no-op.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawULCorner(int col, int row, Color color)
   {
      // no-op
   }
   
   /**
    * Draw the upper-right border corner. This is a no-op.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawURCorner(int col, int row, Color color)
   {
      // no-op
   }

   /**
    * Draw the lower-left border corner. This is a no-op.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawLLCorner(int col, int row, Color color)
   {
      // no-op
   }
   
   /**
    * Draw the lower-right border corner. This is a no-op.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawLRCorner(int col, int row, Color color)
   {
      // no-op
   }

   /**
    * Draw the upper or lower border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    width
    *           The border width.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawHBorder(int col, int row, int width, Color color)
   {
      // mark proper top of region as valid and move to start
      cursorAt(col, row);

      drawHorizontalLine(width, ' ', color);
   }
   
   /**
    * Draw the left or right border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    height
    *           The border height.
    * @param    color
    *           The color to use.
    */
   @Override
   public void drawVBorder(int col, int row, int height, Color color)
   {
      // mark the lines which have had vertical line drawing requested
      for (int j = row; j <= row + height - 1; j++)
      {
         box[j] = true;
      }
   }      

   /**
    * Check if this output primitives implementation requires border corners
    * when drawing a box. This implementation does not require box border
    * corners, so this method will return <code>false</code>.
    * 
    * @return   <code>false</code>, as the box should not be drawn with border 
    *           corners.
    */
   @Override
   public boolean hasBoxBorderCorners()
   {
      return false;
   }

   /**
    * Draw a box using the specified dimensions and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The color to use.
    */
   @Override
   public void box(int left,
                       int top,
                       int right,
                       int bottom,
                       Color color)
   {
      int hlen = right - left + 1;
      int vlen = bottom - top + 1;
      
      // upper border
      drawHBorder(left, top, hlen, color);
      // lower border
      drawHBorder(left, bottom, hlen, color);
      // vertical border
      drawVBorder(top, bottom, vlen, color);
   }
   
   /**
    * Get absolute X position of the cursor.
    *
    * @return   The X position of the cursor.
    */
   @Override
   public int getX()
   {
      return cx;
   }
   
   /**
    * Get absolute Y position of the cursor.
    *
    * @return   The Y position of the cursor.
    */
   @Override
   public int getY()
   {
      return cy;
   }
   
   /**
    * Position the virtual cursor during the drawing.
    * 
    * @param     x
    *            Cursor X coordinate. 
    * @param     y 
    *            Cursor Y coordinate.
    *             
    * @return    <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      int pendingRowDiff = last - first;
      // manage the valid region (the region of the buffer that has content
      // that must be output to the stream)
      if (reset)
      {
         // first cursor move initializes our starting point
         firstOutput();
         
         if (y > 0 && y < lines)
         {
            first = y;
            last  = y;             
         }
         
         if (x > 0 && x < cols)
            num[y] = x;

         if (out != null)
            out.setPendingRowCount(1);
      }
      else
      {
         // relative cursor movements can expand the currently valid region
         
         if (y < first)
         {
            // move backward
            first = (y < 0) ? 0 : y;
         }
         else if (y > last)
         {
            // move forward
            last = (y >= lines) ? (lines - 1) : y;
         }
         
         if (x > num[y])
         {
            num[y] = (x >= cols) ? (cols - 1) : x;
         }
      }
      
      // set the active cursor location (may be outside the valid region)
      cx = x;
      cy = y;
      
      if (out != null && pendingRowDiff < last - first)
      {
         out.setPendingRowCount(out.getPendingRowCount() +
                                last - first - pendingRowDiff);
      }

      return true;
   }
   
   /**
    * Position the real cursor.
    * 
    * @param     x
    *            Cursor X coordinate. 
    * @param     y 
    *            Cursor Y coordinate. 
    * 
    * @return    <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      return cursorAt(x, y);
   }

   /**
    * Mark a rectangle as needing repainting.
    * Makes sense with the double buffered output only; ignored here.
    * 
    * @param    rect
    *           rectangle to mark for repainting
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
   }

   /**
    * Checks the state of the terminal's buffer.
    *
    * @return   <code>true</code> if the buffer is empty.
    */
   public boolean isFresh()
   {
      return reset;
   }

   /**
    * Notifies the terminal about the target frame for the subsequent output.
    * If the terminal detects the change of the target, it flushes the buffer
    * to the stream.
    *
    * @param  id
    *         frame ID for the subsequent output
    * @return frame ID of the going frame
    */
   public int targetFrame(int id)
   {
      if (lastId == id)
         return id;

      int oldId = lastId;

      flush();

      lastId = id;

      return oldId;
   }

   /**
    * Reset last frame ID. Note that ID is reset only if it matched passed
    * frame ID. This does mean that frame is not included into processing
    * anymore and should not be referenced.
    * 
    * @param   id
    *          The frame ID.
    */
   public void resetLastId(int id)
   {
      if (lastId == id)
         lastId = -1;
   }
   
   /**
    * Attempts flushing of the buffered output.
    *
    * @return <code>true</code> if the buffer wasn't empty and has been
    *         flushed.
    */
   public boolean flush()
   {
      if (reset)
         return false;

      boolean saveSync = syncable;
      syncable = true;
      sync();
      syncable = saveSync;

      return true;
   }

   /**
    * Overrides the syncable flag.
    *
    * @param  syncable
    *         new state of the flag
    */
   public void setSyncable(boolean syncable)
   {
      this.syncable = syncable;
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return width;
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return lines;
   }
   
   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    * <p>
    * During this synchronization the contents of the redirected terminal are
    * flushed to the target stream and the terminal is cleared.
    * <p>
    * If the contents belong to a frame without editors, flushing to the
    * stream is done in a single block that fits the current page.
    * Otherwise, the syncing procedure watches the current page growing and
    * inserts page break and, possibly, frame header if the output is at one
    * of the breakable lines.
    */
   @Override
   public void sync()   
   {
      if (!syncable)
         return;

      if (!reset)
      {
         int bytesWritten = 0;

         // only process content that is inside our vertical clipping region

         // set cursor processing can modify some of our data structures
         // even outside of the clipping region, but we only flush the
         // lines of the buffer that reside inside the clipping region
         int lstart = first; // (first >= top)   ? first : top;  
         int lend   = last;  // (last <= bottom) ? last  : bottom;

         // get the current page size and line counter
         // counter doesn't update after out.writeControlled()
         int ps = out.getPageSize().intValue();
         int pl = out.rawNextLineNum();

         // if breakable output (editors are present), don't allow the stream
         // to break the page
         if (!breakable.isEmpty())
         {
            out.setContiguous(true);
         }

         boolean gatheringHeaders = false;
         boolean writtenPageBreak = false;

         // write each line to the stream!
         for (int i = lstart; i <= lend; i++, pl++)
         {
            // find the last non-space char (remove trailing spaces)
            for (int j = num[i] - 1; j >= 0; j--)
            {
               if (buffer[i][j] != ' ')
                  break;
               
               num[i]--;
            }

            int idx = (box[i]) ? 1 : 0;
            int len = num[i] - idx;

            // output a new line for every touched line except for a
            // touched last line that has no data and had box drawing
            boolean newline = !(i == last && num[i] == 0 && box[i]);                

            if (newline)
            {
               String txt = (len > 0) ? new String(buffer[i], idx, len)  
                                      : "";
               out.writeControlled(txt);
               bytesWritten += txt.length();

               // if this buffer contains editors, page breaks should be
               // added whenever necessary
               if (!breakable.isEmpty())
               {
                  // save copy of frame headers for possible reuse
                  // headers are lines from lstart through lstart+headerLines
                  if (i == lstart && headerLines > 0 && headers == null)
                  {
                     headers = txt;
                     gatheringHeaders = true;
                  } 
                  else if (i < lstart + headerLines && gatheringHeaders)
                  {
                     headers = new String(headers + "\n" + txt);
                  }
                  else if (i == lstart + headerLines && gatheringHeaders)
                  {
                     gatheringHeaders = false;
                  } 

                  // check if the stream is paged
                  if (ps > 0)
                  {
                     // check if the page is full
                     if (pl >= ps)
                     {
                        // check if at a breakable line - line number is i
                        if (breakable.contains(new Integer(i)))
                        {
                           // insert page break and frame headers right here
                           String pbreak = "\f";
                           if (headers != null)
                           {
                              pbreak = new String(pbreak + headers + "\n");
                           }
                           out.writeControlledSpecial(pbreak, headerLines);
                           writtenPageBreak = true;
                           bytesWritten += pbreak.length();

                           // reset my copy of the line counter
                           pl = headerLines;
                        }
                     }
                  }
               }
            }               
         }

         // if page breaks were written during this run, produce page headers
         if (writtenPageBreak)
         {
            // reset before rendering headers
            clear();

            inHeaders = true;
            out.processHeadersUncond();
            inHeaders = false;
         }

         BufferSizeManager.getInstance().notifyOutputOperation(false,
                                                               out,
                                                               bytesWritten);
      }
      
      // reset
      clear();
   }

   /**
    * Debugging Aid.
    * Visualizes the redirected terminal's contents.
    *
    * @return array of strings that represent the contents of the terminal
    */
   public String[] view()   
   {
      String[] s = null;

      if (reset)
      {
         s = new String[1];
         s[0] = new String("empty; lines:" + first + "-" + last);
         return s;
      }

      if (!reset)
      {
         // set cursor processing can modify some of our data structures
         // even outside of the clipping region, but we only flush the
         // lines of the buffer that reside inside the clipping region
         int lstart = first; // (first >= top)   ? first : top;  
         int lend   = last;  // (last <= bottom) ? last  : bottom;
         
         s = new String[1 + lend - lstart + 1];
         s[0] = new String("lines:" + first + "-" + last);

         // write each line to the stream!
         for (int i = lstart; i <= lend; i++)
         {
            // find the last non-space char (remove trailing spaces)
            for (int j = num[i] - 1; j >= 0; j--)
            {
               if (buffer[i][j] != ' ')
                  break;

               num[i]--;
            }
            
            int idx = (box[i]) ? 1 : 0;
            int len = num[i] - idx;
            
            // output a new line for every touched line except for a
            // touched last line that has no data and had box drawing
            boolean newline = !(i == last && num[i] == 0 && box[i]);                
            
            if (newline)
            {
               String txt = (len > 0) ? new String(buffer[i], idx, len)  
                                      : "";
               s[1 + i - lstart] = txt;
            }               
         }
      }
      return s;
   }

   /**
    * Set logical terminal width.
    * 
    * @param   width
    *          New terminal width.
    */
   public void setLogicalWidth(int width)
   {
      this.width = width;
   }
   
   /**
    * Process headers if necessary.
    */
   public void processHeaders()
   {
      if (out.needsHeadersOutput())
      {
         // since the processHeaders() calls view() indirectly, the recursion
         // destroys the current invalidation rectangle so it has to be
         // preserved across the call
         ChuiOutputManager tk = (ChuiOutputManager) OutputManager.instance();
         ScreenBitmap bitmap = tk.getBitmapCopy();

         // output our headers
         out.processHeaders();

         // restore the screen clipping info
         tk.setBitmap(bitmap);
      }
   }

   /**
    * Notifies that the buffer image contains frame header and passes the
    * number of lines in the header. Frame header, when present, always starts
    * from the top of the buffer.
    * 
    * @param    height
    */
   public void markHeader(int height)
   {
      if (inHeaders)
      {
         return;
      }

      headerLines = height;
      headers = null;
   }

   /**
    * Notifies that the lines in the buffer belong to an editor and as such
    * can break the current page.
    * Page break may be inserted when flushing the buffer to the stream if
    * the current page exceeds the page size due to some previous contiguous
    * output and the line was marked as breakable with one of these calls.
    *
    * @param   first
    *          line number within the redirected terminal's buffer for the
    *          first line of the range
    * @param   height
    *          number of lines in this range
    */
   public void markBreakLines(int first, int height)
   {
      for (int i = 0; i < height; i++)
      {
         breakable.add(new Integer(first + i));
      }
   }

   /**
    * Detects if headers need to be output.
    *
    * @return   <code>true</code> if header processing is not in process and
    *           the stream has registered page headers and the next line
    *           number is 1 (we have no output on the current page).
    */
   public boolean needsHeadersOutput()
   {
      return out.needsHeadersOutput();
   }

   /**
    * Move the current output data to the new column.
    * 
    * @param   fromCol
    *          The original column where current output starts.
    * @param   toCol
    *          The column where to move the current output.
    */
   public void moveOutputToColumn(int fromCol, int toCol)
   {
      int firstRow = this.cy;
      int lastRow  = this.last;
      for (int i = firstRow; i <= lastRow; i++)
      {
         if (num[i] > 0)
         {
            char[] row = Arrays.copyOf(buffer[i], buffer[i].length);
            
            Arrays.fill(buffer[i], ' ');
            System.arraycopy(row, fromCol, buffer[i], toCol, num[i]);
            num[i] = num[i] - (fromCol - toCol);
         }
      }
   }
   
   /**
    * Add a character to the virtual terminal at the current cursor position.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           The color to use.
    */
   private void addChar(char chr, Color color)
   {
      if (reset)
      {
         firstOutput();
      }
      
      buffer[cy][cx] = chr;
      
      // increment the cursor position
      cx++;
      
      // if we wrote past the end, increment the length of this line
      if (cx > num[cy])
         num[cy] = cx;
   }
   
   /**
    * Any location in this class where content can be written to the buffer
    * must call this method to notify that this is the first output for a
    * fresh buffer.  This will invoke the method
    * {@link com.goldencode.p2j.util.Stream#processHeaders} to test if this
    * is the first output on a new page.  If so, any page headers will be
    * output. Note that erasing output is not considered causing output on
    * the page.  After this notification, the <code>reset</code> flag will
    * be forced to <code>false</code>.
    */
   private void firstOutput()
   {
      // clear the reset flag
      reset = false;
      
      processHeaders();
   }
   
   /**
    * Wipe the slate clean (initialize our state to its pristine condition).
    */
   private void clear()
   {
      Arrays.fill(num, 0);
      Arrays.fill(box, false);
      
      for (int i = 0; i < lines; i++)
      {
         Arrays.fill(buffer[i], ' ');
      }
      
      reset = true;
      cx    = 0;
      cy    = 0;
      first = 0;
      last  = 0;
      
      if (out != null)
         out.setPendingRowCount(0);
   }
   
   /**
    * Draw a horizontal line of the specified length starting at the current
    * cursor position.
    *
    * @param    length
    *           The length of the line to draw in columns.
    * @param    ch
    *           The drawing character to use.
    * @param    color
    *           The color to use.
    */
   private void drawHorizontalLine(int length, char ch, Color color)
   {
      for (int i = 0; i < length; i++)
      {
         addChar(ch, color);
      }
   }
}