/*
** Module   : ScrollPaneImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111202 CHUI-specific version extracted from ScrollPane.
** 002 SIY 20111221 Fixed constructor visibility.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 HC  20150323 Removed business logic from config classes.
** 008 HC  20150831 Scrolling support overhaul.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * CHUI-specific implementation of {@link ScrollPane}.
 */
public class ScrollPaneImpl
extends ScrollPane<ChuiOutputManager>
{
   /**
    * Create a ScrollPane that displays the contents of the specified
    * widget.
    * 
    * @param   widget
    *          The widget to be displayed. This widget must implement
    *          the {@link ScrollPublisher} interface.
    */
   ScrollPaneImpl(ScrollableWidget<ChuiOutputManager> widget)
   {
      super(widget);
   }
   
   /**
    * The main renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary.
    */
   @Override
   public void draw()
   {
      rootPanel.draw();
      
      Insets borderInsets = rootPanel.insets();
      
      // the actual useful parent, we are not interested in bordered panels
      Widget<?> parent = parent();
      while(parent instanceof BorderedPanel)
      {
         parent = parent.parent();
      }
               
      // don't draw scroll-bars if there are no borders
      if (borderInsets.widthDelta() == 0 && borderInsets.heightDelta() == 0 &&
          !(parent instanceof DropDown))
      {
         return;
      }
      
      Color color = Color.DEFAULT;

      BaseConfig parentCfg = UiUtils.getConfig(parent, BaseConfig.class);
      if (parentCfg != null)
      {
         color = UiUtils.getDColor(parentCfg.dcolor, this).withoutReverse();
      }
      
      Point origin = screenLocation();

      /*
       * If the child component is larger than the viewport, draw scroll-bars.
       */
   
      // The size of the component displayed within the viewport.
      Dimension childSize = viewport().getScrollWidget().dimension();
   
      // The size of the viewport
      Dimension extentSize = viewport().dimension();
      Point viewPosition = viewport().getScrollWidget().location();
      
      ScrollableWidget<ChuiOutputManager> scrollWidget = viewport().getScrollWidget();
   
      boolean hsb = true;
      boolean vsb = !(scrollWidget instanceof DefaultList<?, ?>);
      boolean force = false;
      
      if (parent instanceof DropDown)
      {
         vsb = true;
         // in chui char coordinates are always whole numbers
         force = ((int) extentSize.height > 2);
      }
      
      if (scrollWidget instanceof SelectionListBody)
      {
         hsb = ((SelectionListBody<ChuiOutputManager>) scrollWidget).isHorizontalScrollBar();
         force = vsb = ((SelectionListBody<ChuiOutputManager>) scrollWidget).isVerticalScrollBar();
      }
      
      if (parent instanceof Frame)
      {
         // in chui char coordinates are always whole numbers
         vsb = ((int) extentSize.height > 2);
         hsb = ((int) extentSize.width  > 2);
      }
   
      ChuiOutputManager screen = screen();

      Color rclr = color.withReverse();

      // in chui char coordinates are always whole numbers
      if ((force || (int) childSize.height > (int) extentSize.height) && vsb)
      {
         int offset = (int) calcVscroll(childSize, extentSize, viewPosition);             
   
         for (int i = 0; i < (int) extentSize.height; i++)
         {
            screen.at(origin.offset((int) (borderInsets.left + extentSize.width), 
                                    borderInsets.top + i))
                  .append(' ', (i == offset) ? color : rclr);
         }
      }
      else
      {
         if (parent.insets().right > 0)
         {
            screen.at(origin.offset((int) (borderInsets.left + extentSize.width),
                                    borderInsets.top))
                  .vLine(extentSize.height, color);
         }
      }
   
      // in chui char coordinates are always whole numbers
      if ((int) childSize.width > (int) extentSize.width && hsb)
      {
         int offset = (int) calcHscroll(childSize, extentSize, viewPosition);
         
         for (int i = 0; i < (int) extentSize.width; i++)
         {
            screen.at(origin.offset(borderInsets.left + i,
                                    (int) (borderInsets.top + extentSize.height)))
                  .append(' ', (i == offset) ? color : rclr);
         }
      }
      else
      {         
         if ((int) parent.insets().bottom > 0)
         {
            screen.at(origin.offset(borderInsets.left,
                                    (int) (borderInsets.top + extentSize.height)))
                  .hLine(extentSize.width, color);
         }
      }
   }

   /**
    * The secondary renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary. Note that this routine
    * assumes that it is invoked from Frame class.
    * 
    * @param   list
    *          Array of components to draw.
    */
   public void draw(Widget<ChuiOutputManager>[] list)
   {
      boolean redir = ThinClient.getInstance().isRedirected();
      
      if (list != null)
      {
         if (!redir)
         {
            ((BorderedPanelImpl) rootPanel).drawPanel();
         }
         
         Rectangle paneRect = bounds();
         
         for (int i = 0; i < list.length; i++)
         {
            if (!list[i].isVisible())
               continue;
   
            // check if component fits the visible area
            if (redir || paneRect.contains(list[i].bounds()))
               list[i].draw();
         }
      }
      
      // no need to draw the scrollbars or border if we are redirected
      if (redir)
      {
         return;
      }
      
      // the actual useful parent, we are not interested in bordered panels
      Widget<?> parent = parent();
      while(parent instanceof BorderedPanel)
      {
         parent = parent.parent();
      }

      Color color = Color.DEFAULT;

      BaseConfig parentConfig = UiUtils.getConfig(parent, BaseConfig.class);
      if (parentConfig != null)
      {
         color = UiUtils.getDColor(parentConfig.dcolor, this).withoutReverse();
      }
      
      // get the absolute origin of this component
      Point origin = screenLocation();
      Insets borderInsets;
      
      borderInsets = rootPanel.insets();
   
      // don't draw scroll-bars if there are no borders
      if (borderInsets.widthDelta() == 0 && borderInsets.heightDelta() == 0)
         return;
   
      // if the child component is larger than the viewport, draw scroll-bars
   
      // get the size of the component displayed within the viewport
      Dimension childSize = viewport().getScrollWidget().dimension();
   
      // get the size of the viewport
      Dimension extentSize = viewport().dimension();
      Point viewPosition = viewport().getScrollWidget().location();
      
      ChuiOutputManager screen = screen();

      Color rclr = color.withReverse();
      
      // in chui char coordinates are always whole numbers
      if ((int) childSize.height > (int) extentSize.height)
      {
         int offset = (int) calcVscroll(childSize, extentSize, viewPosition);

         for (int i = 0; i < (int) extentSize.height; i++)
         {
            screen.at(origin.offset((int) (borderInsets.left + extentSize.width),
                                    borderInsets.top + i))
                  .append(' ', (i == offset) ? color : rclr);
         }
      }
      else
      {
         if ((int) parent.insets().right > 0)
         {
            screen.at(origin.offset((int) (borderInsets.left + extentSize.width),
                                    borderInsets.top))
                  .vLine(extentSize.height, color);
         }
      }
   
      // in chui char coordinates are always whole numbers
      if ((int) childSize.width > (int) extentSize.width)
      {
         int offset = (int) calcHscroll(childSize, extentSize, viewPosition);
   
         for (int i = 0; i < (int) extentSize.width; i++)
         {
            screen.at(origin.offset(borderInsets.left + i,
                                    (int) (borderInsets.top + extentSize.height)))
                  .append(' ', (i == offset) ? color : rclr);
         }
      }
      else
      {         
         if ((int) parent.insets().bottom > 0)
         {
            screen.at(origin.offset(borderInsets.left,
                                    (int) (borderInsets.top + extentSize.height)))
                  .hLine(extentSize.width, color);
         }
      }
   }
}