/*
** Module   : TextImpl.java
** Abstract : static text implementation for ChUI
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111202 Extracted from Text.
** 002 CA  20130129 Import change.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 CA  20141220 Refactoring to integrate the GUI TEXT implementation with the ChUI logic.
** 008 CA  20150323 More refactoring related to GUI TEXT drawing.
** 009 HC  20150323 Removed business logic from config classes.
** 010 IAS 20150323 Added BLANK attribute support.
** 011 CA  20161101 Added support for TEXT widget which contains the ampersand char in its content.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * CHUI-specific static text widget ({@link Text}) implementation.
 */
public class TextImpl
extends Text<ChuiOutputManager>
{
   /** The content resolver for TEXT, LITERAL and FILL-IN widgets, in ChUI drivers. */
   private final ContentResolver resolver;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public TextImpl(WidgetId id, TextConfig cfg)
   {
      super(id, cfg, new ContentMnemonic<ChuiOutputManager>());
      
      this.resolver = new ContentResolver(this);
   }

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   @Override
   public void draw(Color color)
   {
      if (assigned && getState() == ScreenBuffer.UNINITIALIZED)
         setState(ScreenBuffer.UNCHANGED);
      
      Color dc = UiUtils.getDColor(this);

      selected = !Utils.equals(color, dc);

      String display = getText(true);
      
      if (config.blank)
      {
         display = StringHelper.repeatChar(' ', display.length());
      }

      // obtain the initialization status
      boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED);
      
      // If uninitialized and not for input and not date fill-in (in this case
      // separators should be displayed), display blanks instead.
      // TODO: probably it should be replaced by variable initialization
      //       in FrameHelper.clearComponents().
      if (uninit &&
          !(config.appFormat != null &&
            config.appFormat instanceof DateFormat &&
            isEnabled()))
      {
         int width = display.length();
         display = StringHelper.repeatChar(' ', width);
      }
      
      Point p = new Point(screenLocation());
      
      // in chui char coordinates are always whole numbers
      if ((int) p.x == -1 || (int) p.y == -1)
         return;

      ChuiOutputManager screen = screen();
      
      screen.at(p).append(display, color);

      // in chui char coordinates are always whole numbers
      if ((int) height() > 1)
      {
         String spaces = DisplayFormat.generateSpaces((int) width());
         
         for (int i = 1; i < (int) height(); ++i)
         {
            p.y++;
            screen.at(p);
            
            if (isHeader() && drawLine && i == (int) (height() - 1))
               screen.hLine(width(), color);
            else
               screen.append(spaces, Color.DEFAULT);
         }
      }

      //if (widgetFocused)
      if (hasFocus() && isEnabled())
         drawCaret();
   }

   /**
    * Allow the widget to adjust the cursor position, after a key was pressed.
    * <p>
    * For TEXT widgets, in ChUI the cursor can not go past the maximum cursor position.
    *  
    * @param    key
    *           The pressed key.
    * @param    pos
    *           The cursor position computed by the display format.
    * @param    max
    *           The maximum cursor position.
    *           
    * @return   The adjusted position.
    */
   @Override
   public int adjustCursor(int key, int pos, int max)
   {
      return pos == max ? pos - 1 : pos;
   }

   /**
    * Get screen representation of the data.
    * <p>
    * The work is delegated to the {@link ContentResolver#getText(boolean)} API.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative variable value is used 
    *          when available.
    * 
    * @return  Text representation of the variable value according to all current settings, 
    *          flags, format, etc.
    */
   public String getText(boolean isInternal)
   {
      return resolver.getText(isInternal);
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      return isAtField() ? config.atFormatLength : config.appFormat.getScreenWidth();
   }

   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      return 1;
   }
}