/*
** Module   : ToggleBoxImpl.java
** Abstract :
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20140610 Initial implementation.
** 002 GES 20140627 Changed color attribute method names to match naming standards.
** 003 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 HC  20150323 Removed business logic from config classes.
** 007 IAS 20150924 Added AUTO-RESIZE attribute support
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * Specific implementation of a "toggle box" with a text label for CHUI.
 */
public class ToggleBoxImpl
extends ToggleBox<ChuiOutputManager>
implements AutoResizable<ToggleBoxConfig>
{
   /**
    * This constructor creates a toggle-box with a given text label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle box text label.
    */
   ToggleBoxImpl(WidgetId id, Container<ChuiOutputManager> owner, String text) 
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ToggleBoxImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ToggleBoxConfig) cfg);
   }
   
   /**
    * Draw the button. Called by this ToggleBox's parent container.
    */
   @Override
   public void draw()
   {
      // get the absolute origin of this component
      Point origin = screenLocation();

      screen().at(origin);
      Color dc = UiUtils.getDColor(this);
      Color pf = config().pfcolor;

      // obtain the full text of the toggle-box's label
      String label = textLabel();
      
      // calculate the total size of the toggle-box
      // in chui char coordinates are always whole numbers
      int tbTotalSize = (int) width();
      
      // calculate the max size of the text part, 3 cells should be reserved for check-box
      int textMaxSize = tbTotalSize - 3;
      
      // calculate the actual text length (may be shorter or longer than the text size)
      int textActualSize = label.length(); 
      
      StringBuffer sb = new StringBuffer();
      
      // there is at least always the check-box with or without selection
      drawCheckBox(isEnabled() && widgetFocused, pf, dc);

      if (textMaxSize > 0)
      {
         // the full text labes size is bigger than space reserved for complete
         // toggle-box rendering
         if (textMaxSize < textActualSize)
         {
            // right truncate the text
            sb.append(label.substring(0, textMaxSize));
         }
         else
         {
            // we are just big enough to hold the full text
            sb.append(label);
         }
      }
      
      // actually draw the text label for check-box
      screen().append(sb.toString(), dc);
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(ToggleBoxConfig beforeUpdate)
   {
      setTextLabel(config().label);
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ToggleBoxConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.label, beforeUpdate.label))
      {
         config.widthChars = textLabel().length() + 3; 
      }
   }

   /**
    * Draws check box itself ("[X]" or "[ ]").
    *
    * @param highlighted
    *        Indicates whether the button is highlighted.
    * @param pf
    *        The color used inside this check box under focus.
    * @param dc
    *        The color which corresponds DCOLOR attribute for this radio
    *        button.
    */
   private void drawCheckBox(boolean highlighted, Color pf, Color dc)
   {
      screen().append("[", highlighted ? pf : dc);
      screen().append(config().checked ? "X" : " ", highlighted ? pf : dc);
      screen().append("]", highlighted ? pf : dc);
   }
}