/*
** Module   : Browse.java
** Abstract : core browse widget implementation
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22486 Refactoring (see @@22394-22409). 
** 003 SIY 20050907   @22623 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20050909   @22819 Added attribute container, added and updated 
**                           methods for interaction with BrowseColumn, 
**                           added support for SENSITIVE attribute, 
**                           implemented Validateable interface 
**                           (temporarily commented out).
** 005 SIY 20051205   @24198 Added necessary methods to support Progress 
**                           methods and attributes.
** 006 SIY 20060131   @24199 Use ThinClient special method instead of
**                           direct calls to Window.message().
** 007 SIY 20060209   @24487 Updated to accommodate changes in xxxConfig
**                           classes.
** 008 SIY 20060315   @25119 Major changes to complete BROWSE 
**                           implementation.
** 009 SIY 20060324   @25274 Fixes to make behavior match Progress 
**                           counterpart.
** 010 NVS 20060703   @27766 Reviewed the uses of requestFocus() and
**                           setFocus(). They seem to be safe as they are
**                           internal focus changes.
** 011 NVS 20060703   @27799 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 012 GES 20060713   @28001 Added support to honor the help attribute.
** 013 GES 20060808   @28493 Use int instead of long for row IDs. While
**                           this theoretically limits the maximum number 
**                           of rows per query, in practice the server
**                           side already has the same limitation AND it 
**                           is not practical to ever reach such limits
**                           in Progress code anyway. So we will take the
**                           small benefit of faster network performance
**                           over the capacity since the browse has a
**                           great potential to be a heavy consumer of
**                           network bandwidth.
** 014 SIY 20060824   @28844 Removed getP2JType() method.
** 015 SIY 20061025   @30637 Fixed size calculation and added support for
**                           column type override (@ fields).
** 016 SIY 20061027   @30757 Added refreshing of the browse on 
**                           VALUE-CHANGED event. Fixed drawing of empty
**                           browse.
** 017 SIY 20061102   @30939 Small change to encapsulate field.
** 018 SIY 20061105   @30982 Reposition browse to current row in the model.
** 019 SIY 20061114   @31361 Rewritten cache/navigation. Fixed insert mode
**                           handling and refreshing of row changes.
** 020 NVS 20061204   @31552 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 021 ECF 20061204   @31579 Changed code to interpret a current row of -1
**                           to indicate a query off-end condition. A row
**                           should not be selected in this case.
** 022 ECF 20061206   @31588 Modified refreshStartRow(). This method now
**                           invalidates the browse instead of drawing it
**                           immediately.
** 023 ECF 20061207   @31621 Backed out #022 (@31588). This was causing a
**                           regression in that a reposition against the
**                           query was not correctly updating the browse
**                           view to reflect a new highlighted row.
** 024 SIY 20061208   @31628 refreshStartRow() now explicitly calls
**                           refresh().
** 025 SIY 20061208   @31654 Fix for initial viewport position and drawing
**                           of empty browse.
** 026 SIY 20061219   @31715 Adjusted refresh() behavior to make it similar
**                           to Progress.
** 027 SIY 20070315   @32418 Fixed page width calculation.
** 028 SIY 20070319   @32461 Set model current row in selectFocusedRow().
** 029 SIY 20070321   @32587 Fixed initialization of empty browse, fixes
**                           for layout calculations.
** 030 NVS 20070426   @33247 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 031 SIY 20070504   @33420 New approach in model size handling.
** 032 SIY 20070515   @33747 Fixed positioning.
** 033 SIY 20070730   @34723 More positioning fixes.
** 034 SIY 20070731   @34743 Fixed handling of browse/column help messages.
** 035 SIY 20070926   @35251 Fixed numeric fields alignment. 
**                           Ref: #35250
** 036 ECF 20070928   @35272 Fixed row updates. Removed nullable check in
**                           stopEdit(). Added internal, END-ERROR driven,
**                           row level undo processing to match Progress'
**                           behavior on failed field constraint checks.
** 037 ECF 20071011   @35447 Key handling fixes. Added handler for END
**                           function (move to last row). Fixed PAGE UP
**                           and PAGE DOWN handlers. Disallow negative row
**                           index in setCurrentRow(), except for special
**                           case of END handler.
** 038 ECF 20071030   @35631 Modified refresh(boolean). Added code to draw
**                           after refreshing the data cache. Modified
**                           refreshStartRow() to remove draw() call,
**                           since it is now redundant.
** 039 GES 20071227   @36593 Added hasFocusable() method.
** 040 SIY 20080121   @36873 Changed sending of VALUE-CHANGED. Attempt  
**                           to limit redrawing in some cases.
**                           Ref: #36871
** 041 SIY 20080128   @36920 More correct checking of conditions when 
**                           redraw is required.
**                           Ref: #36919
** 042 SVL 20080208   @36972 Force right-aligning when the column format
**                           is changed from 'number' to 'character'.
** 043 NVS 20080401   @37770 Added isFocusable() method implementation.
** 044 SIY 20080409   @37883 Fixed focus and keyboard and focus handling.
**                           Ref: #37887
** 045 NVS 20080414   @37982 ENTRY event is never consumed by Browse and
**                           may only be consumed by NO-APPLY triggers.
** 046 SIY 20080416   @37995 Postponed refreshing if frame is not visible.
**                           Ref: #37994
** 047 SIY 20080505   @38213 Fixed drawing of not focused BROWSE.
** 048 SIY 20080506   @38217 Force reinitialization of the underlying query
**                           when it is known that previously query
**                           returned other values. Some performance
**                           improvements.
**                           Ref: #38216
** 049 SIY 20080512   @38255 Fixed drawing of disabled widget.
**                           Removed #048, persistence fixes made it 
**                           redundant. More performance improvements.
** 050 SIY 20080521   @38362 Fixed keyboard handling for editing mode.
**                           Ref: #38355
** 051 SIY 20080617   @38796 Fixed initial display of empty browse.
** 052 SIY 20080623   @38925 Fixed drawing of special case when browse is
**                           displayed before query is assigned.
** 053 SIY 20080703   @39036 Fixed drawing of horizontal delimiter.
** 054 CA  20080901   @39644 Fixed startEdit: must show the help for the
**                           the focused widget.
** 055 SIY 20080903   @39695 More fixes in drawing of delimiters and
**                           highlighting.
** 056 CA  20080911   @39784 Bug fix in DataColumnRenderer.addData - the
**                           column and var type must be compared using a
**                           case-insensitive method.
** 057 SVL 20080925   @39941 Call to repaint() was added to
**                           refreshStartRow().
** 058 SIY 20081105   @40313 Added flag which tracks 'focus attractor'
**                           state. See #40284 for more details.
** 059 SIY 20081107   @40368 Rolled back #058, it is not necessary since
**                           more generic approach is discovered.
**                           See #40367 for more details.
** 060 CA  20081114   @40451 Changes due to new clipping implementation.
** 061 CA  20081203   @40757 When an event is received, clear the temp
**                           texts from the message area.
** 062 CA  20081208   @40821 Added ROW-DISPLAY event implementation.
** 063 CA  20081210   @40845 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 064 CA  20081211   @40861 Fixed a regression introduced by H062: after
**                           ROW-DISPLAY trigger invocation, reposition 
**                           the query to the previous 'scrollRow' value.
** 065 CA  20081212   @40884 Yet another regression about ROW-DISPLAY: the
**                           current row must be computed from scrollRow 
**                           and currRow.
** 066 CA  20081212   @40887 On CLOSE QUERY, reset the currRow field.
**                           triggerRowDisplayEvents: row is restored from
**                           the currRow value.
** 067 SIY 20090213   @41309 Added 'force' parameter into refreshStartRow
**                           signature.
** 068 NVS 20090213   @41312 Name changed for clearPutScreenInMessageArea
**                           method in ThinCLient.
** 069 SIY 20090216   @41322 Added tracking of highlighting state.
** 070 CA  20090403   @41728 Fix regression introduced by frame disabling
**                           on destroy - generate ROW-DISPLAY events only
**                           if frame is still alive.
** 071 SIY 20090413   @41784 Fix regression caused by changes in focus 
**                           management. Do not remove highlighting on 
**                           CHARVA focus change since it does not reflect
**                           P2J widget highlighting.
** 072 CA  20090423   @41926 Fixed a regression introduced by ROW-DISPLAY
**                           events - these events must be triggered only
**                           on initial draw or when the rows have scrolled.
**                           The problem was that, on each draw, the query was 
**                           repositioned (which might interfere with an
**                           ongoing CREATE/UPDATE operation).
** 073 SIY 20090902   @43824 Browse is always valid (just like other simple 
**                           components).
** 074 SIY 20090908   @43840 Do not redraw widget if parent frame is hidden.
** 075 SVL 20091011   @44133 Now we take into account the offset of the zero
**                           row of the rowDisplayed array (which equals
**                           dcStartRow).
** 076 SVL 20091029   @44287 Reworked logic into refresh(boolean) responsible
**                           for determination of the top row in a view.
** 077 SVL 20091105   @44315 Into refresh(boolean) do not set the current row
**                           as the start row in a view if the total number of
**                           rows is less than or equals view size.
** 078 SVL 20091119   @44441 Fixed repositioning for editable browse, fixed
**                           adjustment of current row on focus gaining event
**                           if the current row of the backing query row was
**                           changed because of GET statement, fixed
**                           browse:refresh() for the case of server-side
**                           update of rows.
** 079 EVL 20100430   @44839 Fix for incorrect help string with empty browser
**                           data.
** 080 SIY 20100619          Switching to own implementation of character UI.
** 081 SIY 20100804          Changes in imports, some refactorings.
** 082 SIY 20100822          Refactored AttributeContainer support.
** 083 SIY 20100015          locateFrame() is moved to UiUtils, renamed base
**                           class.
** 084 SIY 20101122          Switching from KeyEvent to KeyInput.
** 085 SIY 20101202          Minor change in constant name.
** 086 SIY 20110112          Changes related to new color handling scheme.
** 087 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 088 SIY 20111202          Refactored and moved from CHUI package.
** 089 SIY 20111221          Switched to widget factory.
** 090 CA  20121022          Added support for optional query and column-list clauses.
** 091 GES 20140627          Changed color attribute method names to match naming standards.
** 092 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access 
**                           only to the base APIs provided by the OutputManager.
** 093 HC  20140911          GUI coordinate support - data types refactoring.
** 094 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 095 VIG 20140923          Changed clearPutScreenInMessageArea and message calls to window-aware.
** 096 CA  20141104          Replaced BrowseConfig.width with BaseConfig.widthChars.
** 097 HC  20141021          Implemented window move/position support.
** 098 SVL 20141209          Added support for NO-UNDERLINE. Partially fixed refreshing. Fixed
**                           row selection. Implemented some new functions. Added support for
**                           locked columns.
** 099 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 100 SVL 20150212          Browse reworked.
** 101 SVL 20150305          Fixed SELECT-ROW for editing browse. Added refresh mode for
**                           DELETE-CURRENT-ROW.
** 102 SVL 20150331          Implemented selected row availability mechanics (required for
**                           DELETE-* functions). Implemented ROW-LEAVE triggers.
** 103 SVL 20150413          Implemented SELECT-{NEXT|PREV}-ROW.
** 104 HC  20150323          Removed business logic from config classes.
** 105 SVL 20150420          Implemented REFRESHABLE and ADD-LIKE-COLUMN.
** 106 SVL 20150508          Implemented addition of new rows using insertRow.
** 107 SVL 20150625          Improved INSERT-ROW, DELETE-*-ROW and REFRESH behavior.
** 108 CA  20150722          Refactored so that browse column configurations are transfered via 
**                           pushScreenDefinition calls.  Extracted layout code in doLayout, so 
**                           that it can be performed only once, on the first push.
** 109 SVL 20150722          Minor refactoring. Implemented MODIFIED attribute. Reworked
**                           executePendingRefresh. Implemented SET-REPOSITIONED-ROW.
** 110 IAS 20150801          TAB-STOP attribute support
** 111 SVL 20151013          Moved most of ChUI-specific functionality to BrowseImpl.
** 112 HC  20151013          Fix of repaint issue in nextFocus().
** 113 SVL 20151019          Implemented maxDataGuess.
** 114 SVL 20151027          Added hasRowMarkers. Support for currRowSelected and PageUp/Down in
**                           GUI. Added new functionality to scrollRow().
** 115 SVL 20151126          Changes in processing since column.NO-LABELS attribute is now not
**                           inherited form the parent browse.
** 116 SVL 20151230          Changes related to GUI editing implementation.
** 117 SVL 20160314          Fixed some row deletion cases. Fixed the issue when a new edited row
**                           is out of the view. Added moveColumn.
** 118 SVL 20160722          Refactoring. Fixed initial maxDataGuess on query (re)open.
** 119 HC  20160922          Added getLegacyWidget() method.
**     SVL 20160922          Code refactoring because of switching from BrowseConfig.queryId to
**                           BrowseConfig.queryAssigned.
** 120 CA  20161006          getColumnText() returns empty string if the value is unknown BDT.
** 121 SVL 20161007          Browse is refreshed on CLEAR FRAME instead of "entry" and
**                           "value-changed" events.
** 122 HC  20161025          Comment added.
** 123 EVL 20161020          Adding more chek condition to use character base format in
**                           getDisplayFormat() call.
** 124 SVL 20161029          Changes related to modifications in browse model initialization
**                           process.
** 125 GES 20170131          Protect from NPE by checking isSelectionMultiple() before updating
**                           the selectedRowIndices.
**     EVL 20170203          Fix for wrong changing browse help string to the first column help
**                           string when the browse widget is not visible.
** 126 OM  20170320          ThinClient.setHelp() signature changed. Made hasRowMarkers() public.
** 127 SVL 20170418          Added cellStartsWith().
** 128 SVL 20170814          Switched to server-side ROW-DISPLAY triggering.
** 129 SVL 20170831          Support for toggle-box columns.
** 130 SVL 20171024          Changes because of in-browse validation.
** 131 OM  20171124          Emit LEAVE and ENTRY events when focus enter/leaves browse editors.
**     CA  20171126          Added afterConfigUpdate to repaint if title has changed.
** 132 GES 20171220          Added a TODO comments.
** 133 OM  20171127          The setColumnValue() updates editedRow with SCREEN-VALUE set from
**                           server. Added invokeRowEntryTrigger() method.
**     SVL 20171204          Added refreshCurrentRow.
**     CA  20171204          Fixed setCurrentValue - updateColumn must be called before the cache
**                           gets updated with the new value.
**     CA  20171218          Fixes for BROWSE/BROWSE-COLUMN:ENTRY/LEAVE, BROWSE:ROW-LEAVE/ENTER,
**                           BROWSE:VALUE-CHANGED events.
**     SVL 20171219          Improved management of browse pre-display state.
** 134 SVL 20171225          Fixed comments.
** 135 HC  20180125          Improved synchronization of browse screen buffer between client and
**                           server.
**     EVL 20180130          Adding ability to properly handle SE_ENTRY event starting the cell
**                           edit mode.
** 136 HC  20180223          Prevented unexpected exception during an attempt to start edit mode on
**                           empty browse.
** 137 SVL 20180209          Support for invisible columns. Refactoring related to isRightAligned.
** 138 HC  20180307          Prevented unexpected exception during an attempt to start edit mode
**                           when browse is readonly.
** 139 SBI 20180319          Changed onKeyPressed in order to fire the default-action on ENTER key.
** 140 SVL 20180321          Data for invisible columns should be preserved on the client side too.
** 141 EVL 20180406          Javadoc fix.
** 142 OM  20180606          The BROWSE HELP attribute was lost (replaced by column's null one).
**     HC  20180625          When NO-APPLY is not returned from a browse column ENTRY trigger, that column
**                           must receive focus. Even when the ENTRY trigger itself causes another column to
**                           receive focus and get into edit mode.
** 143 SVL 20180529          Added applyEnhancedConfigClient.
** 144 OM  20180724          Fixed HELP attribute regression in H142 when column was NOT editable.
**                           Avoid repositioning the query when same row is set as current.
**     EVL 20180730          Fix for too early scroll to current row call.  The cached data needs
**                           to be populated first.
** 145 EVL 20180806          Fix for browse row delete with losing data sync.
** 146 SVL 20180813          Reverted H144-2.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.chui.ThinClient.TriggerResult;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

import static com.goldencode.p2j.ui.RowDisplayStrategy.*;

/**
 * Implementation of Progress BROWSE widget.
 * 
 * TODO:  Do we handle options like NO-ASSIGN, NO-VALIDATE, and VALIDATE
 * functions?
 */
public abstract class Browse<O extends OutputManager<?>> 
extends FixedSizeContainer<O>
implements WidgetWithConfig<BrowseConfig>,
           FocusListener,
           KeyListener
{
   /** The size of data cache in screen pages. */
   public static final int CACHE_SIZE = 20;

   /** Index of the first added new row. Following new rows have decreasing row numbers. */
   protected final int NEW_ROWS_NUMERATION_START = -10;
   
   /** A data model instance. */
   protected BrowseModel browseModel = null;

   /** Determines if the current row is selected (single-selection mode only). */
   protected boolean currRowSelected = false;
   
   /**
    * Index of current column. It is the column where the last edit is
    * performed. Does NOT include invisible columns.
    */
   protected int currColumn = 0;
   
   /**
    * An updated data for the current row. These data are passed to model if
    * focus leaves this row. If the array or some element in it is
    * <code>null</code> then no updates are available for the corresponding
    * cell.
    */
   protected BaseDataType[] currRowUpdates = null;
   
   /**
    * Indexes of selected rows in <code>Integer</code> form. Used only in
    * multiple selection mode. This mode is active if and only if this member
    * is not <code>null</code>.
    */
   protected LinkedHashSet<SelectedRow> selectedRowIndices = null;
   
   /**
    * The data model index of the scroll anchor column. Value <code>0</code>
    * means the absence of horizontal scrolling. Otherwise this column will
    * occupy the right part of the view window.
    */
   protected int scrollColumn = 0;
   
   /**
    * The height (in character rows) of the labels area not including the
    * separator line.
    */
   protected int ilLabelsHeight = 0;
   
   /**
    * The minimal positive value of <code>scrollColumn</code>.
    */
   protected int ilFirstScrollCol = 0;
   
   /**
    * Cached rows from the data model. Cache size matches the number of rows displayed in the
    * viewport.
    */
   protected BrowseRow[] dcData = null;

   /**
    * Rendered columns labels area. This cache is reset when horizontal
    * scrolling is performed.
    */
   protected String[] rcLabels = null;
   
   /** Component attribute container. */
   protected BrowseConfig config = null;
   
   /** Draw labels. */
   protected boolean drawLabels = true;

   /** 0-based index of the current row in the browse model. */
   private int currentRow = 0;

   /**
    * Determines if the current row was available (could be fetched) at the moment it has become
    * the current row.
    */
   private boolean currentRowAvail = false;

   /** Waiting for END-ERROR key after edit mode validation error */
   private boolean awaitingEndError = false;

   /** Next unique row number for a new row. */
   private int newRowNumber = NEW_ROWS_NUMERATION_START;
   
   /** Set to true if browse was drawn at least once. */
   protected boolean hadDraw = false;
   
   /** 
    * This variable is used to reproduce special drawing behavior for case
    * when query is initialized after browse is displayed first time. 
    */
   protected boolean pseudoFocus = false;

   /**
    * The number that allows to assess if the total number or rows is more or less than the page
    * size. Can take the following values:
    * <ul>
    *    <li><code>-1</code> if the query is closed or we haven't got any rows yet;</li>
    *    <li>less or equal page size if that represents the number of rows;</li>
    *    <li><code>page size + 1</code> if the number of rows is more than the page size.</li>
    * </ul>
    */
   protected int rowCountAssessment = -1;

   /** Set to <code>true</code> when separators need to be underscored. */
   public boolean separatorQuirk = false;

   /**
    * Width of browse as it was passed to {@link #setSize(double, double)}. Used to recalculate
    * size variables when a new column is added.
    */
   protected double originalWidth;

   /**
    * Height of browse as it was passed to {@link #setSize(double, double)}. Used to recalculate
    * size variables when a new column is added.
    */
   protected double originalHeight;

   /**
    * Contents and cached text of the edited row. Should be saved because if the edited row goes
    * out of the view and then brought back, it doesn't reflect the changes made in database while
    * the row was edited.
    */
   protected BrowseRow editedRow = null;

   /** 0-based index of the current edited row in the browse model. */
   protected int editedRowIndex = -1;

   /** <code>true</code> if we are in the process of saving of the edited row. */
   protected boolean savingRow = false;

   /**
    * Terminal color for current text. It is displayed in inverse colors.
    */
   public Color colorCurrent;

   /**
    * Terminal color for normal text in data area when separators are
    * active. It is displayed with underline.
    */
   protected Color colorHSep;

   /**
    * Reference to the active in-place editor if edit is in progress.
    */
   protected Widget activeEditor;

   /**
    * If editing is active, the index (in the browse model) of the closest exiting row before the
    * currently edited new row. <code>-1</code> if the new row is first row in the result
    * set. <code>null</code> if a new row is not being edited.
    */
   protected Integer editingExistingRowBefore = null;

   /** Terminal attribute for normal text. */
   public Color colorCommon;
   
   /** Cache of all columns. */
   private BrowseColumnConfig[] columnsCache = null;

   /** Cache of visible columns. */
   private BrowseColumnConfig[] visibleColumnsCache = null;

   /**
    * Map of columns' 0-based ordinals among <b>all</b> columns to 0-based ordinals of these
    * columns among <b>visible</b> columns. E.g. if the second column is invisible, the map will
    * be {0 -&gt; 0, 2 -&gt; 1, 3 -&gt; 2}.
    */
   private Map<Integer, Integer> ordinalToVisibleMap = null;

   /**
    * If we are getting new rows, specifies type of action for which browse rows are fetched and
    * defines how rows are iterated / displayed.
    */
   protected RowDisplayStrategy currentRowDisplayStrategy = null;

   /**
    * Cache of current rows supplemented with "intermediate" rows (i.e. the rows that were already
    * fetched in the process of getting new rows).
    */
   protected BrowseRow[] intermediateDcData = null;

   /**
    * After initial enabling of browse it becomes navigable. Initial enabling can be performed
    * using ENABLE statement or SENSITIVE attribute, regardless of the parent frame visibility
    * (so browse can stay invisible). When a browse becomes navigable is does NOT fetch rows
    * immediately, but it allows initial row fetching using HOME or END events (at least).
    */
   protected boolean navigable = false;

   /**
    * If the query was repositioned before the rows data set was fetched for the first time, then
    * this variable keeps the current row.
    */
   protected Integer initialRow = null;

   /**
    * Flag which helps to distinguish if {@link #queryRepositioned(boolean, int)} was caused by
    * query opening or repositioning.
    */
   private boolean queryClosed = true;

   /**
    * <code>true</code> if the rows data set was already fetched for the first time.
    */
   private boolean initialDataSetObtained = false;

   /**
    * Constructs the Browse widget.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Browse(WidgetId id, BrowseConfig cfg)
   {
   }

   /**
    * Notify the browse that some columns have been added to it or visibility of columns has been
    * changed.
    */
   public abstract void notifyColumnsUpdated();

   /**
    * Get the index of the row on the top of the page.
    *
    * @return 0-based index (in the browse model) of the row on the top of the page or
    *         <code>0</code> if there are no rows.
    */
   protected int getScrollRow()
   {
      if (dcData == null || dcData.length == 0)
         return 0;

      return getCachedRowNumber(0);
   }
   
   /**
    * The method is called after the configuration associated with the
    * owner changes. The parameter points to a reference representing
    * the original configuration state. This reference can be used to
    * detect what configuration fields changed and so optimize any
    * processing related to the change.
    * <p>
    * The implemented logic must be idempotent - the method must produce 
    * the same results if executed once or multiple times not relying 
    * on any particular call sequence or related state. This is
    * necessary so that the method can be repeated or retried as often as 
    * necessary without causing unintended effects
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before 
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      BrowseConfig c = (BrowseConfig) config();
      
      if ((c.title == null && beforeUpdate.title != null) ||
          (c.title != null && !c.title.equals(beforeUpdate.title)))
      {
         repaint();
      }
   }

   /**
    * Get the current row.
    *
    * @return  0-based index of the current row in the browse model. Negative numbers are returned
    *          for new rows. <code>0</code> is also returned if there are no rows.
    */
   public int getCurrentRow()
   {
      return currentRow;
   }

   /**
    * Get the current row as {@link com.goldencode.p2j.ui.client.Browse.SelectedRow} object which
    * takes into account its availability (see {@link #currentRowAvail}).
    *
    * @return see above.
    */
   protected SelectedRow getCurrentRowAsSelected()
   {
      return new SelectedRow(currentRow, currentRowAvail);
   }

   /**
    * Get the browse model index of the data row at the specified row index in the specified
    * row cache.
    *
    * @param  rowCache
    *         Cache of rows: {@link #dcData} or {@link #intermediateDcData}.
    * @param  rowIndex
    *         0-based index of the row in the given row cache.
    *
    * @return browse model index of the data row at the specified row index in the specified
    *         row cache or <code>-1</code> if there is no row with the specified index.
    */
   protected int getCachedRowNumber(BrowseRow[] rowCache, int rowIndex)
   {
      if (rowCache == null || rowIndex > rowCache.length - 1)
         return -1;

      return rowCache[rowIndex].rowIndex;
   }

   /**
    * Get the browse model index of the data row positioned at the specified viewport row.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport.
    *
    * @return 0-based index of the target row in the browse model or <code>-1</code> if there are
    *         no rows or there is no data row at the specified viewport row.
    */
   public int getCachedRowNumber(int viewportIndex)
   {
      return getCachedRowNumber(dcData, viewportIndex);
   }

   /**
    * Get the browse model index of the data row in the row cache supplemented with "intermediate"
    * rows ({@link #intermediateDcData}), positioned at the specified viewport row.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport.
    *
    * @return 0-based index of the target row in the browse model or <code>-1</code> if there are
    *         no rows or there is no data row at the specified viewport row.
    */
   public int getCachedRowNumberWithIntermediate(int viewportIndex)
   {
      return intermediateDcData != null ? getCachedRowNumber(intermediateDcData, viewportIndex) :
                                          getCachedRowNumber(viewportIndex);
   }

   /**
    * Get the number of rows in the cache (which matches the number of rows in the viewport).
    *
    * @return the number of rows in the cache.
    */
   public int getCacheSize()
   {
      return dcData == null ? 0 : dcData.length;
   }

   /**
    * Get the number of rows in the cache supplemented with "intermediate" rows
    * ({@link #intermediateDcData}), which matches the number of rows in the viewport.
    *
    * @return the number of rows in the cache supplemented with "intermediate" rows.
    */
   public int getCacheSizeWithIntermediate()
   {
      return intermediateDcData != null ? intermediateDcData.length : getCacheSize();
   }

   /**
    * Get the index of the current row in the viewport.
    *
    * @return 0-based index of the current row in the viewport.
    */
   public int getViewportRow()
   {
      return getViewportRow(getCurrentRow());
   }

   /**
    * Get the index of the specified row in the viewport.
    *
    * @param  rowIndex
    *         0-based index of the target row in the browse model.
    *
    * @return 0-based index of the specified row in the viewport. <code>0</code> if the row out
    *         of the current view.
    */
   public int getViewportRow(int rowIndex)
   {
      if (dcData == null)
         return 0;

      for (int i = 0; i < dcData.length; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
            return i;
      }

      return 0;
   }

   /**
    * Calculate DOWN value for the browse basing on the given row height.
    *
    * @param  rowHeightPixels
    *         Row height in pixels.
    *
    * @return calculated DOWN value.
    */
   public int calcDown(int rowHeightPixels)
   {
      return config.down;
   }

   /**
    * Set "intermediate" rows (i.e. the rows that were already fetched in the process of getting
    * new rows by
    * {@link BrowseWidget#getRows(int, int, boolean, int, ScreenBuffer[], RowDisplayStrategy)}),
    * so these rows can be drawn if a pause occurs in ROW-DISPLAY trigger.
    *
    * @param rows
    *        Rows to set.
    */
   public void setIntermediateRows(BrowseRow[] rows)
   {
      // overridden in GUI version, do nothing for ChUI
   }

   /**
    * Scroll browse by the specified number of rows. If the number of rows is less than the page
    * size, then some rows from the current view remain in the cache (without refresh). The
    * new rows are obtained from the server starting from the first row which row index in the
    * browse model is more than the index of the last row in the viewport (or less than the index
    * of the first row if we scroll in backward direction).
    *
    * @param  rowCount
    *         Number of rows to scroll. Positive count to scroll forward, negative to scroll
    *         backward.
    *
    * @return <code>true</code> if at least one row was obtained from the server.
    */
   protected boolean incrementalScroll(int rowCount)
   {
      return incrementalScroll(rowCount, false, false);
   }

   /**
    * Scroll browse by the specified number of rows. If the number of rows is less than the page
    * size, then some rows from the current view remain in the cache (without refresh). The
    * new rows are obtained from the server starting from the first row which row index in the
    * browse model is more than the index of the last row in the viewport (or less than the index
    * of the first row if we scroll in backward direction).
    *
    * @param  rowCount
    *         Number of rows to scroll. Positive count to scroll forward, negative to scroll
    *         backward.
    * @param  forceTrailingSpace
    *         If number rows returned by server is less than expected, then remaining rows will be
    *         spaces (i.e. <code>(DOWN - rowCount)</code> current rows in view are dropped
    *         unconditionally).
    * @param  incrementalRowDisplay
    *         <code>true</code> if an incremental row display strategy is used.
    *
    * @return <code>true</code> if at least one row was obtained from the server.
    */
   protected boolean incrementalScroll(int rowCount,
                                       boolean forceTrailingSpace,
                                       boolean incrementalRowDisplay)
   {
      if (rowCount == 0)
         return false;

      int down = getDown();

      if (dcData.length != down && rowCount > 0)
      {
         throw new IllegalStateException("There is no point to use incremental scroll for " +
                                         "scrolling forward if the page is not full.");
      }

      boolean forward = rowCount > 0;
      int diff = Math.abs(rowCount);

      int lastRow = forward ? getLastExistingRow() : getFirstExistingRow();
      int startRow = forward ? lastRow + 1 : lastRow - 1;

      BrowseRow[] serverData;
      if (isNewEditingRowOutOfTheView() &&
          Math.abs(rowCount) == 1 &&
          (forward ? editingExistingRowBefore == startRow - 1:
                     editingExistingRowBefore == startRow))
      {
         // iterate with the single editing row
         serverData = insertEditingRow(new BrowseRow[0], forward, diff, lastRow, true);
      }
      else
      {
         RowDisplayStrategy displayStrategy =
               incrementalRowDisplay && Math.abs(rowCount) > 1 ?
                                                 (forward ? INCREMENTAL_SCROLL_FWD :
                                                            INCREMENTAL_SCROLL_BACK) :
                                                 NO_DISPLAY;
         serverData = getRows(startRow,
                              diff,
                              forward,
                              displayStrategy);
      }

      if (serverData.length > 0)
      {
         int overlap = Math.min(down - (forceTrailingSpace ? diff : serverData.length),
                                getCacheSize());
         if (overlap > 0)
         {
            BrowseRow[] newData = new BrowseRow[overlap + serverData.length];

            if (forward)
            {
               System.arraycopy(dcData, getCacheSize() - overlap, newData, 0, overlap);
               System.arraycopy(serverData, 0, newData, overlap, serverData.length);
               dcData = newData;
            }
            else
            {
               System.arraycopy(serverData, 0, newData, 0, serverData.length);
               System.arraycopy(dcData, 0, newData, serverData.length, overlap);
               dcData = newData;
            }
         }
         else
         {
            dcData = serverData;
         }
      }

      if (forward && serverData.length < rowCount)
      {
         // we have reached the end
         updateMaxDataGuess(true);
      }

      notifyBrowseScrolled();

      return serverData.length > 0;
   }

   /**
    * If a new row is being edited, add this row into the given result set if it should be in this
    * set.
    *
    * @param  serverData
    *         Existing rows (order of rows is from top to bottom for any scrolling direction).
    * @param  forward
    *         <code>true</code> for forward scrolling direction, <code>false</code> for backward.
    * @param  maxSize
    *         Maximum allowed size of the resulting set.
    * @param  lastRow
    *         For forward scrolling: last existing row before the rows from the result set.
    *         For backward scrolling: first existing row after the rows from the result set.
    * @param  allowBeforeFirst
    *         If <code>true</code>, the edited row can take the first position in the result set
    *         (for its scrolling direction).
    *
    * @return resulting set with the currently edited new row included, if necessary.
    */
   protected BrowseRow[] insertEditingRow(BrowseRow[] serverData,
                                          boolean forward,
                                          int maxSize,
                                          int lastRow,
                                          boolean allowBeforeFirst)
   {
      int insertPosition = -1;

      if (serverData.length == 0)
      {
         if (forward ? editingExistingRowBefore >= lastRow :
                       editingExistingRowBefore <= lastRow + 1)
         {
            insertPosition = 0;
         }
      }
      else
      {
         if (forward &&
             allowBeforeFirst &&
             editingExistingRowBefore >= lastRow &&
             editingExistingRowBefore < serverData[0].rowIndex)
         {
            // forward, the first element
            insertPosition = 0;
         }
         else if (!forward &&
                  editingExistingRowBefore < serverData[0].rowIndex)
         {
            if (serverData.length == maxSize)
            {
               return serverData;
            }

            // backward, first element
            insertPosition = 0;
         }
         else if (forward &&
                  editingExistingRowBefore >= serverData[serverData.length - 1].rowIndex)
         {
            if (serverData.length == maxSize)
            {
               return serverData;
            }

            // forward, the last element
            insertPosition = serverData.length;
         }
         else if (!forward &&
                  allowBeforeFirst &&
                  editingExistingRowBefore < lastRow &&
                  editingExistingRowBefore >= serverData[serverData.length - 1].rowIndex)
         {
            // backward, the last element
            insertPosition = serverData.length;
         }
         else
         {
            // between rows
            for (int i = 0; i < serverData.length - 1; i++)
            {
               if (serverData[i].rowIndex <= editingExistingRowBefore &&
                   editingExistingRowBefore < serverData[i + 1].rowIndex)
               {
                  insertPosition = i + 1;
               }
            }
         }
      }

      if (insertPosition >= 0)
      {
         BrowseRow[] newData = new BrowseRow[Math.min(maxSize, serverData.length + 1)];

         if (insertPosition > 0)
         {
            System.arraycopy(serverData, forward ? 0 : 1, newData, 0, insertPosition);
         }

         newData[insertPosition] = new BrowseRow(editedRowIndex,
                                                 new BaseDataType[editedRow.data.length],
                                                 ThinClient.getInstance().isChui());
         System.arraycopy(editedRow.data,
                          0,
                          newData[insertPosition].data,
                          0,
                          editedRow.data.length);

         if (insertPosition < newData.length - 1)
         {
            System.arraycopy(serverData,
                             insertPosition,
                             newData,
                             insertPosition + 1,
                             newData.length - insertPosition - 1);
         }

         return newData;
      }

      return serverData;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BrowseConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      
      addKeyListener(this);
      addFocusListener(this);
   }

   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      super.destroy();

      BrowseColumnConfig[] cols = getColumns();
      
      if (cols == null)
      {
         return;
      }
      
      ThinClient tc = ThinClient.getInstance();
      
      for (BrowseColumnConfig col : cols)
      {
         @SuppressWarnings("unchecked")
         BrowseColumn<O> bcol = (BrowseColumn<O>) tc.getWidget(col.id);

         if (bcol != null)
         {
            bcol.destroy();
         }
      }
   }

   /**
    * Get display format which corresponds to format string and data type
    * stored in configuration.
    * 
    * @return  Instance of <code>DisplayFormat</code>.
    */
   public static DisplayFormat getDisplayFormat(BrowseColumnConfig colCfg)
   {
      DisplayFormat colFormat = null;
      
      try
      {
         BaseDataType var = DisplayFormat.instanceOfType(colCfg.dataType);
         
         if (colCfg.format != null && !colCfg.format.isEmpty())
            colFormat = DisplayFormatFactory.makeFormat(var, colCfg.format);
         else
            colFormat = DisplayFormatFactory.makeDefaultFormat(var);
      }
      catch (DisplayFormatParsingException e)
      {
         throw new RuntimeException(e);
      }
      
      return colFormat;
   }

   /**
    * Return label as array of strings.
    * 
    * @return  Parsed label.
    */
   public static String[] getLabelVar(BrowseColumnConfig colCfg)
   {
      String label = colCfg.clabel;
      
      if (label == null)
         label = colCfg.label;
      
      if (label == null || (!colCfg.labels && !colCfg.forceLabel))
         return new String[0]; // no label
   
      int len = label.length();
      boolean singlelined = true;
   
      int offsExcl = 0;
      while (offsExcl < len)
      {
         offsExcl = label.indexOf('!', offsExcl);
         if (offsExcl < 0)
         {
            break; // no single '!'
         }
   
         ++offsExcl;
         if (offsExcl == len)
         {
            break; // no char after single '!'
         }
   
         if (label.charAt(offsExcl) == '!')
         {
            ++offsExcl;
            continue; // skipping double '!'
         }
   
         // found single '!' followed by other char. Now checking two last
         // chars. If both are '!' then label is contains only one line.
         if (label.charAt(len - 1) != '!' || label.charAt(len - 2) != '!')
         {
            singlelined = false;
         }
         break;
      }
   
      if (singlelined)
         return new String[] { label }; // displaying label as-is
   
      ArrayList<String> lines = new ArrayList<String>(4);
      StringBuilder line = new StringBuilder(80);
   
      int offsLine = 0;
      while (true)
      {
         int offsSep = label.indexOf('!', offsLine);
         
         if (offsSep < 0)
         {
            offsSep = len;
         }
         line.append(label.substring(offsLine, offsSep));
   
         // last line is indicated as no '!' or no chars after next '!'
         if (offsSep >= len - 1)
         {
            // finishing last line
            lines.add(line.toString());
            break;
         }
   
         ++offsSep;
         if (label.charAt(offsSep) == '!')
         {
            // double '!' should be treated as one '!' in screen value
            line.append('!');
            ++offsSep;
         }
         else
         {
            // real line separator
            lines.add(line.toString());
            line.setLength(0);
         }
   
         offsLine = offsSep;
      }
   
      return lines.toArray(new String[lines.size()]);
   }

   /**
    * Returns <code>true</code> if the visual representation of the cell content starts with the
    * given character.
    *
    * @param  cellText
    *         Cell text.
    * @param  searchChar
    *         Target character.
    *
    * @return <code>true</code> if visual representation of the cell content starts with the
    *         given character.
    */
   public static boolean cellStartsWith(String cellText, char searchChar)
   {
      return cellText != null &&
             !cellText.isEmpty() &&
             cellText.charAt(0) == searchChar;

   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#setEnabled(boolean)
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      if (enabled && hadDraw)
         separatorQuirk  = true;

      if(enabled != isEnabled())
         pseudoFocus = false;

      if (enabled && isRefreshedOnEnable())
      {
         refresh(false);
      }

      if (enabled)
      {
         navigable = true;
      }

      super.setEnabled(enabled);
   }
   
   /**
    * Tracks the highlighted state  of the component.
    * 
    * @param  highlighted
    *         New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      if (highlighted != highlighted())
         pseudoFocus = false;
      
      super.setHighlighted(highlighted);
   }
   
   /**
    * Second stage of the initialization of the Browse widget which is
    * performed only when query is initially assigned.
    */
   public void initQueryAssigned()
   {
      if (browseModel != null)
         return;

      browseModel = new ClientBrowseModel(config.id.asInt());

      int firstEditable = -1;
      
      BrowseColumnConfig[] cols = getVisibleColumns();
      for (int i = 0; i < cols.length; ++i)
      {
         if (!cols[i].readOnly)
         {
            firstEditable = i;
            break;
         }
      }
      
      if (config.multiple)
         selectedRowIndices = new LinkedHashSet<>();
      rowCountAssessment = -1;
      
      if (isEditPossible())
      {
         currColumn = firstEditable;
      }
      
      if (config.multiple)
         setSelectionMultiple(true);
   }

   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   public BaseDataType getValue()
   {
      // editing is allowed
      if (isEditing())
      {
         return getEditorValue();
      }
         
      // editing is not allowed
      // TODO: change the approach
      // Code below may cause problems in GUI environment, because each 
      // validation will reset view to selected cell. This is not an issue
      // in TUI because there is no way to make current cell not in view port. 
      columnToView(currColumn);
      
      return getValue(visibleToOrdinal(currColumn));
   }

   /**
    * Handle situation when <code>Browse</code> widget receives the focus.
    * <p>
    * When a help string is defined for the browse it will be displayed on
    * the status line.  This method also sets the selected attribute on. 
    * 
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // For editable browse, when event.other is null, it is actually a "focus lost"
      // event, do not update selection in this case.
      boolean updateSelection = !isEditing() && isAggressiveCurrentRowSelection();
      setCurrentRow(getCurrentRow(), updateSelection, true);
      
      Widget other = event.other();
      if (other != null)
      {
         // focus gained
         startEdit(other.getId());
      }
      else
      {
         BrowseColumnConfig colConfig = getColumns()[currColumn];
         ThinClient.setHelp(!colConfig.readOnly && colConfig.help != null ? colConfig.help 
                                                                          : config.help, this);
      }
      
      repaint();
   }
   
   /**
    * Handle situation when <code>Browse</code> widget loses the focus.
    * <p>
    * This method clears the focused flag.
    * 
    * @param    fe_
    *           Reference to the event.
    */
   @Override
   public void onFocusLost(FocusEvent fe_)
   {
      if (isEditing())
      {
         stopEdit(fe_.other() == null ? null : fe_.other().getId());
      }
      
      repaint();
   }
   
   /**
    * Undo editing operation results and reset everything to initial state.
    */
   public void undoValue()
   {
      undoValue(true);
   }

   /**
    * Undo editing operation results and reset everything to initial state.
    *
    * @param    startEdit
    *           <code>true</code> to call {@link #startEdit()} after the value has been undone.
    *        
    * @return   <code>true</code> if the edit was stopped.
    */
   public boolean undoValue(boolean startEdit)
   {
      Widget editor = getActiveEditor();
      
      if (editor != null)
      {
         if (!stopEdit(true, false))
         {
            return false;
         }
      }
      
      currRowUpdates = null;
      editedRow = null;
      editedRowIndex = -1;
      
      if (startEdit && editor != null)
         startEdit();
      
      return true;
   }
   
   /**
    * Browse behaves like simple component in this regard.
    * 
    * @return  <code>true</code>.
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Requesting the focus to the <code>Browse</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;

      setCurrentRow(getCurrentRow(), isAggressiveCurrentRowSelection(), true);

      if (isEditing() && isSelectionMultiple())
      {
         selectedRowIndices.clear();
         selectedRowIndices.add(getCurrentRowAsSelected());
         repaint();
      }

      super.requestFocus();
   }
   
   /**
    * Set READ-ONLY attribute.
    * 
    * @param   readOnly
    *          New value for the attribute.
    */
   public void setReadOnly(boolean readOnly)
   {
      // Config may be overwritten in unexpected places (e.g. in getRows because of maxDataGuess
      // update). So we want to avoid calling startEdit if we have temporarily stopped editing
      // in order to update cache contents.
      if (config.readOnly == readOnly)
         return;

      config.readOnly = readOnly;

      if (browseModel == null)
         return;

      if (readOnly)
         stopEdit(false, true);
      else
         startEdit();
   }

   /**
    * Select all rows.
    */
   public void selectAll()
   {
      if (isSelectionMultiple())
      {
         boolean editing = isEditing();
         if (editing)
         {
            undoValue(false);
         }

         // get the last row
         int bigNumber = Integer.MAX_VALUE - 1000; // decreased to avoid overflow
         BrowseRow[] row = getRows(bigNumber, 1, false, NO_DISPLAY);

         selectedRowIndices.clear();
         if (row.length == 0)
            return;

         int numResults = row[0].rowIndex + 1;
         for (int i = 0; i < numResults; i++)
         {
            SelectedRow selectedRow = new SelectedRow(i);
            selectedRowIndices.add(selectedRow);
         }

         // scroll to the first row in the result set
         boolean anyRows = scrollToRow(0, false, true);
         if (anyRows)
            setCurrentRow(getCachedRowNumber(0));

         if (editing)
         {
            startEdit(false);
         }
      }
   }
   
   /**
    * Remove all selections. 
    */
   public void deselectAll()
   {
      if (isSelectionMultiple())
         selectedRowIndices.clear();
      else
         currRowSelected = false;

      repaint();
   }
   
   /**
    * Select row which is currently in focus. 
    */
   public void selectFocusedRow()
   {
      if (browseModel == null)
      {
         return;
      }
      
      if (isSelectionMultiple())
      {
         SelectedRow currRow = getCurrentRowAsSelected();
         if (!selectedRowIndices.contains(currRow))
         {
            selectedRowIndices.add(currRow);
            repaint();
         }
      }
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   rowIndex
    *          0-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   public boolean selectRow(int rowIndex)
   {
      if (rowIndex < 0 || rowIndex > getCacheSize() - 1)
      {
         return false;
      }

      int row = getCachedRowNumber(rowIndex);
      SelectedRow selectedRow = new SelectedRow(row);
      if (isSelectionMultiple())
      {
         if (!selectedRowIndices.contains(selectedRow))
         {
            boolean avail = browseModel.isRowAvailable(row);
            selectedRowIndices.add(new SelectedRow(selectedRow.getRowIndex(), avail));
            repaint();
         }
      }
      else if (getCurrentRow() != row)
      {
         boolean editing = isEditing();
         if (editing)
         {
            releaseCurrentRow(false, false, true);
         }

         setCurrentRow(row);
         repaint();

         if (editing)
            startEdit();
      }

      return true;
   }

   /**
    * Deselect row which is currently in focus.
    */
   public void deselectFocusedRow()
   {
      if (browseModel == null)
      {
         return;
      }

      if (isSelectionMultiple())
      {
         SelectedRow currRow = getCurrentRowAsSelected();
         if (selectedRowIndices.contains(currRow))
         {
            selectedRowIndices.remove(currRow);
            repaint();
         }
      }
      else
      {
         currRowSelected = false;
         repaint();
      }
   }

   /**
    * Get the number of visible columns.
    *
    * @return number of visible columns.
    */
   public int getVisibleColumnCount()
   {
      return getVisibleColumns().length;
   }

   /**
    * Get the number of columns.
    *
    * @return number of columns.
    */
   public int getColumnCount()
   {
      return getColumns().length;
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p> Deselects the <i>n</i>th selected
    * row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   rowIndex
    *          0-based index of the row (in the list of selected rows) to deselect.
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public boolean deselectSelectedRow(int rowIndex)
   {
      if (rowIndex < 0)
         return false;

      if (isSelectionMultiple())
      {
         if (rowIndex > selectedRowIndices.size() - 1)
            return false;

         // skip elements
         Iterator iter = selectedRowIndices.iterator();
         for (int i = 0; i < rowIndex; i++)
         {
            iter.next();
         }

         // remove target element
         iter.next();
         iter.remove();

      }
      else
      {
         if (rowIndex > 0 || !currRowSelected)
            return false;

         if (hasFocus() && isEnabled() && !isEditing())
         {
            // Progress quirk: when a non-editable browse has focus, the row is NOT deselected,
            // but true is returned.
            return true;
         }

         currRowSelected = false;
      }

      repaint();

      return true;
   }

   /**
    * Determine if the browse column is right-aligned.
    *
    * @param  column
    *         Target browse column.
    *
    * @return <code>true</code> if the target column is right-aligned.
    */
   public static boolean isRightAligned(BrowseColumn column)
   {
      return isRightAligned(column.config());
   }

   /**
    * Determine if the browse column is right-aligned.
    *
    * @param  columnConfig
    *         Config of the target browse column.
    *
    * @return <code>true</code> if the target column is right-aligned.
    */
   public static boolean isRightAligned(BrowseColumnConfig columnConfig)
   {
      boolean rightAlign = getDisplayFormat(columnConfig).isRightPositioning() ||
                           columnConfig.rightAligned;

      // adjust alignment for numbers
      if (getDisplayFormat(columnConfig) instanceof NumberFormat)
         rightAlign = true;

      return rightAlign;
   }

   /**
    * Get cell text to render.
    *
    * @param  columnConfig
    *         Config of the browse column the cell belongs to.
    * @param  var
    *         Cell data.
    *
    * @return Rendered cell text.
    *
    * @throws DisplayFormatCheckException
    *         Data is incompatible with format.
    */
   public static String getCellText(BrowseColumnConfig columnConfig, BaseDataType var)
   throws DisplayFormatCheckException
   {
      if (var == null)
      {
         return null;
      }

      return getDisplayFormat(columnConfig).fromVar(var).toScreenValue();
   }

   /**
    * Get cell text to render, without raising error.
    *
    * @param  columnConfig
    *         Config of the browse column the cell belongs to.
    * @param  var
    *         Cell data.
    *
    * @return Rendered cell text or question marks ("????...") if the value cannot be displayed
    *         using the current format.
    */
   public static String getCellTextNoError(BrowseColumnConfig columnConfig, BaseDataType var)
   {
      String text;
      try
      {
         text = Browse.getCellText(columnConfig, var);
      }
      catch (DisplayFormatCheckException e)
      {
         text = generateQuestionMarks(columnConfig, var);
      }
      return text;
   }

   /**
    * Generate question marks that indicate that the cell value cannot be displayed using the
    * current format.
    *
    * @param  columnConfig
    *         Config of the browse column the cell belongs to.
    * @param  var
    *         Cell data.
    *
    * @return string of question marks to be displayed in the cell.
    */
   protected static String generateQuestionMarks(BrowseColumnConfig columnConfig,
                                                 BaseDataType var)
   {
      int formatLen;
      try
      {
         formatLen = DisplayFormatFactory.makeFormat(var, columnConfig.format).getScreenWidth();
      }
      catch (DisplayFormatParsingException e)
      {
         formatLen = (int) columnConfig.widthChars;
      }

      return DisplayFormat.generateQuestions(formatLen);
   }

   /**
    * Renders the screen value into a cell. This is helper method for the
    * render classes.
    * 
    * @param   screen
    *          Screen value.
    * @param   col
    *          Column to render.
    * @param   colWidth
    *          The full width for formatting.
    * @param   scroll
    *          Scrolling correction. Means the number of characters which
    *          should be erased from the left (if positive) or from the right
    *          (if negative). This scroll differs from the scroll in
    *          <code>DisplayFormat.formatScreenValue()</code> because it
    *          means the scroll of the whole browse column not the scroll of
    *          screen value inside the column.
    * 
    * @return  Rendered screen value.
    */
   protected static String renderCell(String screen, BrowseColumnConfig col, 
                                     int colWidth, int scroll)
   {
      BrowseColumn<?> colw = (BrowseColumn<?>) ThinClient.getInstance().getWidget(col.id);
      boolean rightAlign = isRightAligned(colw);
      
      if (colw.config().rightAligned)
         screen = screen.trim();
      
      String display = DisplayFormat.formatScreenValue(screen, colWidth, 0, 
                                                       rightAlign, rightAlign);
      if (scroll > 0)
      {
         // scroll by truncating from the left
         display = display.substring(scroll);
      }
      else
      {
         display = display.substring(0, display.length() + scroll);
      }

      return display;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   public BrowseConfig config()
   {
      return config;
   }
   
   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public abstract void drawCaret();

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public abstract void draw();

   /**
    * This component will receive an input focus when Tab or Shift-Tab
    * is pressed.
    * 
    * @return <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns the same as <code>getDefaultFocus()</code> because focus placement
    * inside BROWSE can't be switched from initial component until subcomponent
    * creation or deletion.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget<O> currentFocus()
   {
      return defaultFocus();
   }

   /**
    * Returns this component directly if there is no in-place editor.
    * Otherwise returns this editor.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget<O> defaultFocus()
   {
      Widget editor = getActiveEditor();
      if (editor !=null)
      {
         return editor;
      }

      return this;
   }
                                         
   /**                                             
    * Reports if there is at least 1 focusable column that is not the
    * current column.
    *
    * @param    next
    *           <code>true</code> if this must search for a next focusable
    *           widget, <code>false</code> for a previous focusable widget.
    * @param    exclude
    *           This is ignored since the current column is internally known
    *           by the browse.
    *
    * @return   <code>true</code> if there is at least one other column
    *           that can be focused.
    */
   public boolean hasFocusable(boolean next, Widget<O> exclude)
   {
      // quick exit if no edits are possible
      if (!isEditPossible())
         return false;
      
      // when editing is not active, browse acts like a single widget that
      // has no focusable widgets inside
      if (!isEditing())
         return false;
      
      BrowseColumnConfig[] cols = getVisibleColumns();
      
      // can't previous past beginning of first row
      if (!next && isFirstBrowseRow())
      {
         for (int j = currColumn - 1; j >= 0; j--)
         {
            if (!cols[j].readOnly)
               return true;
         }
         
         return false;
      }
      
      // can't next past end of last row
      if (next && getViewportRow() == (getCacheSize() - 1))
      {
         for (int k = currColumn + 1; k < cols.length; k++)
         {
            if (!cols[k].readOnly)
               return true;
         }
         
         return false;
      }
      
      // since we are editable, there must be another field that can be
      // accessed
      return true;
   }
   
   /**
    * Reports if there is at least 1 focusable component that is not the
    * given component.
    *
    * @param    exclude
    *           A component to ignore or <code>null</code> if any focusable
    *           component will do.
    *
    * @return   <code>true</code> if there is at least one other component
    *           that can be focused.
    */
   @Override
   public boolean hasFocusable(Widget<O> exclude)
   {
      return hasFocusable(true, exclude) || parent().hasFocusable(exclude); 
   }
       
   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the next widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#nextFocus()
    */
   @Override
   public void nextFocus()
   {
      if (isEditPossible() && isEditing())
      {
         goDown(true);

         // surround the repaint() with event bracket,
         // so that the generated paint events are processed and
         // the invalidation rectanges added to screen bitmap
         ThinClient.getInstance().eventBracket(new Runnable()
         {
            @Override
            public void run()
            {
               repaint();
            }
         });
      }
      else
         parent().nextFocus();
   }

   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the previous widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#prevFocus()
    */
   @Override
   public void prevFocus()
   {
      if (isEditPossible() && !isEditing())
         switchToPrevCell();
      else
         parent().prevFocus();
   }
   
   /**
    * Returns selection mode state.
    * 
    * @return Is browse in the multiple selection mode.
    */
   public boolean isSelectionMultiple()
   {
      return (selectedRowIndices != null);
   }

   /**
    * Switch multiple selection on and off.
    *
    * @param   multiple
    *          <code>true</code> enable multiple selection,
    *          <code>false</code> - disable multiple selection.
    */
   public void setSelectionMultiple(boolean multiple)
   {
      if (browseModel == null)
         return;
      
      if (multiple && selectedRowIndices == null)
      {
         selectedRowIndices = new LinkedHashSet<>();
         selectedRowIndices.add(getCurrentRowAsSelected());

         return;
      }
      
      if (!multiple && selectedRowIndices != null)
      {
         selectedRowIndices = null;
         return;
      }
      
      // do nothing, other cases should not have any effect
      return;
   }

   /**
    * Returns is given row selected.
    * 
    * @param   rowIndex
    *          0-based row's index in browse model.
    * 
    * @return  Is this row selected.
    */
   public boolean isRowSelected(int rowIndex)
   {
      if (!isSelectionMultiple())
      {
         return (getCurrentRow() == rowIndex) && currRowSelected;
      }

      return selectedRowIndices.contains(new SelectedRow(rowIndex));
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.
    *
    * @return 1-based index of the focused row in the viewport. If there are no rows,
    *         <code>1</code> is returned,
    */
   public int getFocusedRow()
   {
      //TODO focused row shouldn't be bound to the query's current row
      // demonstration: click on a row, hold the mouse button, scroll browse by moving mouse up
      // or down - current row changes while the focused row stays the same
      return getViewportRow() + 1;
   }

   /**
    * Returns the count of selected rows.
    * 
    * @return  The count of selected rows.
    */
   public int getSelectedRowsCount()
   {
      if (!isSelectionMultiple())
      {
         return (getCacheSize() > 0 && currRowSelected) ? 1 : 0;
      }

      return selectedRowIndices.size();
   }

   /**
    * Get the index (in the browse model) of the given selected row.
    *
    * @param   selIndex
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    *
    * @return  0-based index of the row in the browse model.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public int getSelectedRowIndex(int selIndex)
   {
      return getSelectedRowIndex(selIndex, false, false);
   }

   /**
    * Get the index (in the browse model) of the given selected row.
    *
    * @param   selIndex
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    * @param   ignoreNotAvail
    *          If <code>true</code> then exception is not raised if the selected row is not
    *          available.
    * @param   newAsAvail
    *          Count new rows as "available". Affects error handling.
    *
    * @return  0-based index of the row in the browse model.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows and
    *          <code>ignoreNotAvail</code> parameter is not specified.
    */
   public int getSelectedRowIndex(int selIndex, boolean ignoreNotAvail, boolean newAsAvail)
   {
      int numSelectedRows = getSelectedRowsCount();
      if (selIndex == -1)
      {
         // DELETE-CURRENT-ROW mode
         if (numSelectedRows == 0)
         {
            // This is being raised to transfer control to the server where the proper
            // ErrorManager processing will occur.
            throw new ErrorConditionException(382, "");
         }

         selIndex = numSelectedRows - 1;
      }

      if (selIndex >= numSelectedRows || selIndex < 0)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(385, "");
      }

      if (!ignoreNotAvail && !isSelectedRowAvailable(selIndex, newAsAvail))
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(382, "");
      }

      if (!isSelectionMultiple())
      {
         return getCurrentRow();
      }

      Iterator<SelectedRow> itRow = selectedRowIndices.iterator();
      for (int i = selIndex; i > 0; --i)
      {
         itRow.next();
      }

      return itRow.next().getRowIndex();
   }

   /**
    * Retrieve the row indexes of the selected rows in the given browse widget.
    *
    * @return  Array of 0-based row indexes of the selected rows in the browse.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public int[] getSelectedRowIndexes()
   {
      int numSelectedRows = getSelectedRowsCount();
      if (numSelectedRows == 0)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(382, "");
      }

      if (!isSelectedRowAvailable(-1, true))
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(382, "");
      }

      if (!isSelectionMultiple())
      {
         return new int[] { getCurrentRow() };
      }

      int[] res = new int[selectedRowIndices.size()];
      int i = 0;
      for (SelectedRow row : selectedRowIndices)
      {
         res[i++] = row.getRowIndex();
      }

      return res;
   }

   /**
    * Get the maximal row index (in the browse model) among the selected rows,
    *
    * @return maximal row index (in the browse model) among the selected rows or <code>-1</code>
    *         if there are no selected rows.
    */
   public int getMaximalSelectedRowIndex()
   {
      int numSelectedRows = getSelectedRowsCount();
      if (numSelectedRows == 0)
      {
         return -1;
      }

      if (!isSelectionMultiple())
      {
         return getCurrentRow();
      }

      int res = -1;
      for (SelectedRow row : selectedRowIndices)
      {
         int index = row.getRowIndex();
         if (index > res || res == -1)
            res = index;
      }

      return res;
   }

   /**
    * Fires the DefaultAction event.
    */
   public void fireDefaultAction()
   {
      EventManager.postEvent(new KeyInput(this, 
                                          EventType.KEY_PRESSED,
                                          Keyboard.SE_DEFAULT_ACTION));
   }

   /**
    * Fires the ValueChanged event.
    */
   public void fireValueChanged()
   {
      EventManager.postEvent(new KeyInput(this,
                                          EventType.KEY_PRESSED,
                                          Keyboard.SE_VALUE_CHANGED,
                                          true));
   }
   
   /**
    * Return status of the updates of the current row.
    * 
    * @return  <code>true</code> if current row is modified and
    *          <code>false</code> otherwise.
    */
   public boolean isCurrentRowModified()
   {
      if (currRowUpdates != null)
         return true;

      if (isEditing())
      {
         BaseDataType oldValue = dcGetRowData(getCurrentRow())[currColumn];
         if (oldValue == null)
         {
            if (getActiveEditor().getState() != ScreenBuffer.UNINITIALIZED)
               return true;
         }
         else if (!oldValue.equals(getEditorValue()))
            return true;
      }

      return false;
   }

   /**
    * Refresh the current browse row using the provided data.
    *
    * @param row
    *        Container with the new data for the current row. If an {@link BrowseRow#data} element
    *        is <code>null</code>, corresponding cell shouldn't be refreshed.
    */
   public void refreshCurrentRow(BrowseRow row)
   {
      if (getCacheSize() == 0)
      {
         return;
      }

      if (isCurrentRowInView())
      {
         int viewportRow = getViewportRow();
         for (int i = 0; i < row.data.length; i++)
         {
            if (row.data[i] != null)
            {
               dcData[viewportRow].data[i] = row.data[i];
               dcData[viewportRow].text[i] = row.text[i];
            }
         }
      }

      if (editedRow != null)
      {
         for (int i = 0; i < row.data.length; i++)
         {
            if (row.data[i] != null)
            {
               editedRow.data[i] = row.data[i];
               editedRow.text[i] = row.text[i];
            }
         }
      }

      if (currRowUpdates != null)
      {
         for (int i = 0; i < row.data.length; i++)
         {
            if (row.data[i] != null)
            {
               currRowUpdates[i] = null;
            }
         }
      }

      if (isEditing() && row.data[currColumn] != null)
      {
         setEditorValue(row.data[currColumn]);
         if (isEditorFillIn())
         {
            getActiveFillIn().invalidateSelection();
            getActiveFillIn().setZapMode(false);
         }
      }

      repaint();
   }

   /**
    * Determine if the current row is in the view.
    *
    * @return <code>true</code> if the current row is the view.
    */
   public boolean isCurrentRowInView()
   {
      return isRowInView(getCurrentRow());
   }

   /**
    * Returns <code>true</code> if the current row is a new row and is not present in the
    * database.
    *
    * @return  <code>true</code> if the current row is a new row and <code>false</code> otherwise.
    */
   public boolean isNewRow()
   {
      return isNewRow(getCurrentRow());
   }

   /**
    * Returns <code>true</code> if the given row is a new row and is not present in the
    * database.
    *
    * @param  newRowIndex
    *         0-based index of the target row in the browse model (new rows have specific negative
    *         indexes).
    *
    * @return  <code>true</code> if the given row is a new row and <code>false</code> otherwise.
    */

   public boolean isNewRow(int newRowIndex)
   {
      return newRowIndex <= NEW_ROWS_NUMERATION_START;
   }

   /**
    * Inserts a blank row in an updateable browse before or after the last selected row.
    * The blank row is a placeholder for a new record to be added through the browse.
    *
    * @param   after
    *          <code>true</code> if the row is added after the last selected row.
    *          <code>false</code> if the row is added before the last selected row.
    *
    * @return  <code>true</code> if the row was successfully added.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          In specific cases when row cannot be added. The exception contains proper 4GL error
    *          code.
    */
   public boolean insertRow(boolean after)
   throws ErrorConditionException
   {
      // Check if adding a new row will make the last existing database row pushed out of the
      // view, which shouldn't happen.
      boolean singleFirstRow = true;
      boolean singleLastRow  = true;
      boolean noExistingRows         = true;
      int cacheSize = getCacheSize();
      for (int i = 0; i < cacheSize; i++)
      {
         if (!isNewRow(getCachedRowNumber(i)))
         {
            noExistingRows = false;

            if (i == 0)
            {
               singleLastRow = false;
            }
            else if (i == getDown() - 1)
            {
               singleFirstRow = false;
            }
            else
            {
               singleFirstRow = false;
               singleLastRow = false;
               break;
            }
         }
      }

      if (noExistingRows)
      {
         singleFirstRow = false;
         singleLastRow = false;
      }

      int viewportRow = getViewportRow();
      int lastRow = getDown() - 1;
      if ((noExistingRows &&
           getCacheSize() == lastRow &&
           (after || viewportRow == 0))                                                       ||
          (singleFirstRow && (after ? viewportRow == lastRow : viewportRow != 0)) ||
          (singleLastRow &&  (after ? viewportRow != lastRow : viewportRow == 0)))
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         // Error text is "You must have a database row in BROWSE in order to use INSERT-ROW"
         throw new ErrorConditionException(4963, "");
      }

      BaseDataType copiedValue = null;

      if (isEditing())
      {
         if (isCurrentRowModified())
         {
            UnimplementedFeature.todo("4GL does not update the value, the value of the cell is " +
                  "saved in the browse cache and may eventually be saved in the browse model.");
         }

         if (isFirstLastRowCopyQuirk() &&
             ((viewportRow == 0 && !after) || (viewportRow == lastRow && after)))
         {
            // store current column value for the first or last row in the view
            copiedValue = getEditorValue();
         }
      }

      if (isEditPossible() && config.readOnly)
      {
         UnimplementedFeature.todo("In REPLACE mode of an editable browse adding new row " +
               "causes quirk in which the rows above are not shifted.");
      }

      int newViewportRow = addNewRow(-1, null, viewportRow, after);
      if (copiedValue != null)
      {
         dcData[newViewportRow].data[currColumn] = copiedValue;
      }
      
      // events need to be triggered in this order:
      // 1. leave currently edited cell, and if not NO-APPLY then
      // 2. row-leave, and if not NO-APPLY then
      // 3. row-entry
      // 3->4. entry for the edited column, ROW-ENTRY was not NO-APPLY
      // 5. value-changed for browse
      
      // this needs to be picked up by the next event-processing loop, not the current one; so we
      // need to use postOSEvent, which will be picked up by TC.waitForEvent
      ThinClient.getInstance().postOSEvent(new InvocationEvent(() ->
      {
         int nextColumn = getFirstEditableColumn(0);
         WidgetId nextId = nextColumn < 0 ? null : getVisibleColumnCfg(nextColumn).id;

         if (!invokeLeaveCellTrigger(nextId))
         {
            return;
         }
         
         // this performs validation
         if (!validateStopEdit(true, false))
         {
            return;
         }
         
         currRowUpdates = null;
         editedRow = null;
         editedRowIndex = -1;
         
         setCurrentRow(getCachedRowNumber(newViewportRow));
            
         if (invokeRowLeaveTrigger(null))
         {
            if (invokeRowEntryTrigger(false))
            {
               // this is later than 4LG does it, but we need it here because otherwise we lose
               // the current focus (previously editable column...).
               finishStopEdit();
               
               startEdit();
            }
            
            fireValueChanged();
         }
      }));

      repaint();
      
      return true;
   }

   /**
    * Return <code>true</code> if row which has focus is selected.
    * 
    * @return  <code>true</code> if row which has focus is selected and
    *          <code>false</code> otherwise.
    */
   public boolean isFocusedRowSelected()
   {
      return isRowSelected(getCurrentRow());
   }
   
   /**
    * Return number of potential data rows in the <code>Browse</code>.
    * Amount of available data can be less that number of rows which
    * potentially can be displayed by the <code>Browse</code>.
    * 
    * @return  Number of potential rows.
    */
   public int getDown()
   {
      return config.down;
   }

   /**
    * Get currently used <code>BrowseModel</code>.
    * 
    * @return  Reference to <code>BrowseModel</code> used by
    *        <code>Browse</code>.
    */
   public BrowseModel getModel()
   {
      return browseModel;
   }

   /**
    * Return reference to previous column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getPrevColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = getVisibleColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i + 1) < cols.length)
               return cols[i + 1];
            return null;
         }
      }

      return null;
   }

   /**
    * Return reference to next column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getNextColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = getVisibleColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i - 1) > 0)
               return cols[i - 1];
            return null;
         }
      }

      return null;
   }

   /**
    * Get widget IDs of the browse column widgets.
    *
    * @return array of widget IDs of the browse column widgets.
    */
   public int[] getIdsOfColumnWidgets()
   {
      if (browseModel == null)
         return new int[0];

      BrowseColumnConfig[] columns = getColumns();
      int[] res = new int[columns.length];

      for (int i = 0; i < columns.length; i++)
         res[i] = columns[i].id.asInt();

      return res;
   }

   /**
    * Determines if the specific selected row was available (could be fetched) at the moment of
    * its selection.
    *
    * @param  selIndex
    *         0-based position of the row in the list of selected rows. If <code>-1</code> is
    *         passed then all of the selected rows are checked for availability.
    * @param  newAsAvailable
    *         Count new rows as "available".
    *
    * @return <code>true</code> if the specific selected row was available (could be fetched) at
    *         the moment of its selection.
    *         If <code>-1</code> is passed then <code>true</code> is returned if at least one
    *         of the selected rows is available.
    */
   public boolean isSelectedRowAvailable(int selIndex, boolean newAsAvailable)
   {
      if (!isSelectionMultiple())
      {
         if (selIndex > 0)
            return false;

         return isNewRow() ? newAsAvailable : currentRowAvail;
      }
      else
      {
         if (selIndex >= selectedRowIndices.size())
            return false;

         boolean checkAllRows = selIndex < 0;

         Iterator<SelectedRow> iter = selectedRowIndices.iterator();
         int i = 0;
         while (iter.hasNext())
         {
            SelectedRow row = iter.next();
            if (checkAllRows)
            {
               if (isNewRow(row.getRowIndex()) ? newAsAvailable : row.isRowAvailable())
                  return true;
            }
            else
            {
               if (selIndex == i)
                  return isNewRow(row.getRowIndex()) ? newAsAvailable : row.isRowAvailable();
               i++;
            }
         }

         // no available rows
         return false;
      }
   }

   /**
    * Lay out the browse.
    */
   public void doLayout()
   {
      if (!config.performLayout)
      {
         return;
      }

      config.performLayout = false;
      
      BrowseColumnConfig[] cols = getVisibleColumns();

      ThinClient client = ThinClient.getInstance();
      for (BrowseColumnConfig columnConfig : cols)
      {
         BrowseColumn<O> bcol = (BrowseColumn<O>) client.getWidget(columnConfig.id);
         bcol.setBrowse(this);

         // Recalculate column auto-width with the parent browse set.
         bcol.afterConfigUpdate(bcol.config);
      }

      ilLabelsHeight = 0;
      int minHeight = config.down;
      int minWidth  = 4;

      if (config.box)
      {
         minWidth += 2;
         minHeight += 2;
      }
      
      int firstEditable = -1;
      
      drawLabels = config.labels;
      
      for (int i = 0; i < cols.length; ++i)
      {
         if (!config.labels)
         {
            // no need to set the column's LABELS attr: this is set right by the server-side
            if (cols[i].forceLabel)
               drawLabels = true;
         }
         
         if (!cols[i].readOnly)
         {
            if (firstEditable == -1)
               firstEditable = i;
         }
         
         minWidth += cols[i].widthChars;
         
         if (i < (cols.length - 1))
            minWidth += 1;
      }

      updateLabelsHeight();
      
      if (drawLabels)
         minHeight += ilLabelsHeight + (config.noUnderline ? 0 : 1); //Headers
      
      if (config.widthChars > 0)
         minWidth = (int) Coordinate.scale(config.widthChars);
      
      if (config.widthChars == 0)
      {
         config.widthChars = Coordinate.scale(minWidth); // direct access
      }
      if (config.heightChars == 0)
      {
         config.heightChars = Coordinate.scale(minHeight); // direct access
      }
      
      setSize(config.widthChars, config.heightChars);
   }

   /**
    * Returns the widget which corresponds the FOCUS handle if this browse is focused.
    *
    * @return if editing is in process - the editor widget (fill-in by default), otherwise - this
    *         browse widget.
    */
   public Widget getFocusedWidget()
   {
      Widget editor = getActiveEditor();
      return editor != null ? editor : this;
   }

   /**
    * If browse is in editing mode, set editor in 4GL READ-ONLY mode.
    *
    * @param readOnly
    *        <code>true</code> to set editor in 4GL READ-ONLY mode.
    */
   public void setEditorReadOnly(boolean readOnly)
   {
      if (isEnabled() &&
          isEditorFillIn())   // doesn't do anything to TOGGLE-BOX columns
      {
         getActiveFillIn().config.readOnly = readOnly;
      }
   }

   /**
    * Return {@link BrowseColumnConfig} for specified index.
    *
    * @param   columnOrdinal
    *          0-based index of the column among all columns (i.e. column ordinal).
    *
    * @return  <code>BrowseColumn</code> instance or <code>null</code> in
    *          case of invalid index.
    */
   public BrowseColumnConfig getColumnCfg(int columnOrdinal)
   {
      BrowseColumnConfig[] cols = getColumns();

      if (columnOrdinal >= 0 && columnOrdinal < cols.length)
         return cols[columnOrdinal];

      return null;
   }

   /**
    * Return {@link BrowseColumnConfig} for specified index.
    *
    * @param   visibleColumnIndex
    *          0-based index of the column among the visible columns.
    * 
    * @return  <code>BrowseColumn</code> instance or <code>null</code> in
    *          case of invalid index.
    */
   protected BrowseColumnConfig getVisibleColumnCfg(int visibleColumnIndex)
   {
      BrowseColumnConfig[] cols = getVisibleColumns();
      
      if (visibleColumnIndex >= 0 && visibleColumnIndex < cols.length)
         return cols[visibleColumnIndex];
      
      return null;
   }

   /**
    * Returns the active in-place editor if edit is in progress.
    *
    * @return Reference to editor or <code>null</code>.
    */
   @SuppressWarnings("unchecked")
   protected Widget getActiveEditor()
   {
      return activeEditor;
   }

   /**
    * Returns <code>true</code> if editing is in progress.
    *
    * @return <code>true</code> if editing is in progress.
    */
   public boolean isEditing()
   {
      return getActiveEditor() != null;
   }

   /**
    * Get active editor (assumes it is a fill-in). Before using this function, use
    * {@link #isEditorFillIn()} to check if the active editor is a fill-in, otherwise
    * {@link ClassCastException} may be thrown.
    *
    * @return active fill-in editor.
    *
    * @throws ClassCastException
    *         if the active editor is not a fill-in.
    */
   protected FillIn getActiveFillIn()
   {
      return (FillIn) activeEditor;
   }

   /**
    * Get active editor (assumes it is a toggle-box). Before using this function, use
    * {@link #isEditorToggleBox()} to check if the active editor is a toggle-box, otherwise
    * {@link ClassCastException} may be thrown.
    *
    * @return active toggle-box editor.
    *
    * @throws ClassCastException
    *         if the active editor is not a toggle-box.
    */
   protected ToggleBox getActiveToggleBox()
   {
      return (ToggleBox) activeEditor;
   }

   /**
    * Returns <code>true</code> if the current editor is a fill-in.
    *
    * @return see above.
    */
   protected boolean isEditorFillIn()
   {
      return activeEditor instanceof FillIn;
   }

   /**
    * Returns <code>true</code> if the current editor is a toggle-box.
    *
    * @return see above.
    */
   protected boolean isEditorToggleBox()
   {
      return activeEditor instanceof ToggleBox;
   }

   /**
    * Get value of the editor widget. See {@link DataContainer#getValue()}.
    *
    * @return value of the editor widget.
    */
   protected BaseDataType getEditorValue()
   {
      return ((DataContainer) activeEditor).getValue();
   }

   /**
    * Returns <code>true</code> if the widget value was modified. See
    * {@link DataContainer#isModified()}.
    *
    * @return <code>true</code> if the widget value was modified.
    */
   protected boolean isEditorModified()
   {
      return ((DataContainer) activeEditor).isModified();
   }

   /**
    * Set value of the editor widget. See {@link DataContainer#setValue(BaseDataType)}.
    *
    * @param value
    *        value of the editor widget to set.
    */
   protected void setEditorValue(BaseDataType value)
   {
      ((DataContainer) activeEditor).setValue(value);
   }

   /**
    * Returns <code>true</code> if there is a new editing row which is out of the view.
    *
    * @return see above.
    */
   protected boolean isNewEditingRowOutOfTheView()
   {
      return editingExistingRowBefore != null &&
             isEditing() &&
             editedRow != null &&
             !isRowInView(getCurrentRow());
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    * 
    * @param   visibleColumnIndex
    *          0-based index of the column among the visible columns.
    */
   protected abstract void columnToView(int visibleColumnIndex);

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    */
   public boolean startEdit()
   {
      return startEdit(true);
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param    prevFocus
    *           The previous focused widget, which will lose focus.
    * 
    * @return   <code>true</code> if editing was on and ENTRY triggered allowed it to stop.
    */
   public boolean startEdit(WidgetId prevFocus)
   {
      return startEdit(prevFocus, true, false, null);
   }

   /**
    * Finishes edit and destroys an editor.
    *
    * @param   nextFocus
    *          The widget which will be focused next.
    *
    * @return  <code>true</code> if editing was stopped;  <code>false</code>
    *          if we were not in edit mode or validation failed.
    */
   public boolean stopEdit(WidgetId nextFocus)
   {
      return stopEdit(nextFocus, true, true);
   }
   
   /**
    * Finishes edit and destroys an editor.
    *
    * @param   undo
    *          <code>true</code> if the value in the cell is undone rather than committed.
    * @param   validate
    *          <code>true</code> to perform validation.
    *
    * @return  <code>true</code> if editing was stopped;  <code>false</code>
    *          if we were not in edit mode or validation failed.
    */
   protected boolean stopEdit(boolean undo, boolean validate)
   {
      return stopEdit(null, undo, validate);
   }

   /**
    * Finishes edit and destroys an editor.
    *
    * @param   nextFocus
    *          The widget which will be focused next.
    * @param   undo
    *          <code>true</code> if the value in the cell is undone rather than committed.
    * @param   validate
    *          <code>true</code> to perform validation.
    *
    * @return  <code>true</code> if editing was stopped;  <code>false</code>
    *          if we were not in edit mode or validation failed.
    */
   protected boolean stopEdit(WidgetId nextFocus, boolean undo, boolean validate)
   {
      // first the LEAVE trigger is invoked and only after that validation is performed
      
      if (!invokeLeaveCellTrigger(nextFocus))
      {
         return false;
      }

      if (!validateStopEdit(undo, validate))
      {
         return false;
      }

      finishStopEdit();
      
      return true;
   }

   /**
    * Validates current editor.
    *
    * @param   undo
    *          <code>true</code> if the value in the cell is undone rather than committed.
    * @param   validate
    *          <code>true</code> to perform validation.
    *
    * @return  <code>true</code> if validation passed or we were not in edit mode.
    */
   private boolean validateStopEdit(boolean undo, boolean validate)
   {
      Widget editor = getActiveEditor();
      if (editor == null)
      {
         return true;
      }

      if (!undo && isEditing())
      {
         if (isEditorFillIn() && !getActiveFillIn().update())
         {
            return false;
         }
         
         if (validate && !validateEditorValue())
         {
            return false;
         }
         
         if (editor.getState() != ScreenBuffer.UNINITIALIZED)
         {
            BaseDataType var = getEditorValue();
            updateColumn(visibleToOrdinal(currColumn), var);
         }
         
         if (currRowUpdates != null && isEditorModified())
         {
            ConfigHelper.setModified(config, true);
         }
      }
      
      return true;
   }
   
   /**
    * Stops the current editor.
    */
   public void finishStopEdit()
   {
      Widget editor = getActiveEditor();
      if (editor == null)
      {
         return;
      }
      Container parent = editor.parent();

      if (parent == null)
      {
         System.err.println("Editor was somehow deleted from its parent.");
      }
      parent.remove(editor);
      screen().getRegistry().removeWidget(editor.config().id);
      cleanupEditor(editor);
      activeEditor = null;
      editingExistingRowBefore = null;

      Window.setInsertMode(Window.INSERT_MODE_OVR);

      repaint();
   }

   /**
    * Run validation for the current editor.
    *
    * @return <code>true</code> if no validation for the current column or validation was
    *         successful.
    */
   protected boolean validateEditorValue()
   {
      if (!config.autoValidate || config.noValidate)
      {
         return true;
      }

      BrowseColumnConfig column = getVisibleColumnCfg(currColumn);
      if (!column.validatable)
      {
         return true;
      }

      Widget editor = getActiveEditor();
      if (((DataContainer) editor).isModified())
      {
         if (!ThinClient.getInstance().validateWidget(editor, true))
         {
            return false;
         }
      }

      return true;
   }

   /**
    * Initialize editor widget config.
    *
    * @param cfg
    *        Editor widget config.
    */
   protected abstract void initEditorConfig(BaseConfig cfg);

   /**
    * Add editor widget to the browse.
    *
    * @param editor
    *        Editor widget.
    */
   protected abstract void addEditor(Widget editor);

   /**
    * Clean up editor widget.
    *
    * @param editor
    *        Editor widget.
    */
   protected void cleanupEditor(Widget editor)
   {
   }

   /**
    * Add new row to the cache.
    *
    * @param  rowIndex
    *         0-based index of the row in the browse model or <code>-1</code> if an empty row is
    *         generated.
    * @param  rowData
    *         Row cache data. <code>null</code> to generate an empty row.
    * @param  viewportRow
    *         0-based index of the base row in the viewport.
    * @param  after
    *         <code>true</code> if the row will be inserted after the base row. <code>false</code>
    *         if before the base row.
    *
    * @return 0-based index of the new row in the viewport
    */
   protected int addNewRow(int rowIndex, BaseDataType[] rowData, int viewportRow, boolean after)
   {
      BrowseRow[] newData;
      int newRowViewportIndex;

      if (dcData == null)
      {
         newData = new BrowseRow[1];
         newRowViewportIndex = 0;
      }
      else
      {
         int down = getDown();
         if (after)
         {
            if (viewportRow < down - 1)
            {
               newData = new BrowseRow[Math.min(down, dcData.length + 1)];
               System.arraycopy(dcData, 0, newData, 0, Math.min(viewportRow + 1, dcData.length));

               newRowViewportIndex = viewportRow + 1;

               if (dcData.length > viewportRow + 1)
               {
                  System.arraycopy(dcData,
                                   viewportRow + 1,
                                   newData,
                                   viewportRow + 2,
                                   newData.length - viewportRow - 2);
               }
            }
            else
            {
               // we are on the last row in the view
               newData = new BrowseRow[down];
               System.arraycopy(dcData, 1, newData, 0, newData.length - 1);

               newRowViewportIndex = newData.length - 1;
            }
         }
         else
         {
            if (viewportRow > 0)
            {
               newData = new BrowseRow[Math.min(down, dcData.length)];
               System.arraycopy(dcData, 1, newData, 0, Math.min(viewportRow - 1, dcData.length));

               newRowViewportIndex = viewportRow - 1;

               if (dcData.length > viewportRow)
               {
                  System.arraycopy(dcData,
                        viewportRow,
                        newData,
                        viewportRow,
                        newData.length - viewportRow);
               }
            }
            else
            {
               // we are on the first row in the view
               newData = new BrowseRow[Math.min(down, dcData.length + 1)];
               newRowViewportIndex = 0;
               System.arraycopy(dcData, 0, newData, 1, newData.length - 1);
            }
         }
      }

      if (rowData == null)
      {
         int newId = nextNewRowNumber();
         int colCount = getColumns().length;
         newData[newRowViewportIndex] = new BrowseRow(newId,
                                                      new BaseDataType[colCount],
                                                      ThinClient.getInstance().isChui());
      }
      else
      {
         newData[newRowViewportIndex] = new BrowseRow(rowIndex,
                                                      rowData,
                                                      ThinClient.getInstance().isChui());
      }
      newData[newRowViewportIndex].renderRow(getColumns());

      dcData = newData;
      // no ROW-DISPLAY triggering
      notifyBrowseScrolled();
      return newRowViewportIndex;
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param    updateSelection
    *           For multi-selection browse: if <code>true</code>, current row becomes the only
    *           selected row.
    * 
    * @return   <code>true</code> if editing was on and ENTRY triggered allowed it to stop.
    */
   protected boolean startEdit(boolean updateSelection)
   {
      return startEdit(null, updateSelection, false, null);
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param    prevFocus
    *           The previous focused widget, which will lose focus.
    * @param    updateSelection
    *           For multi-selection browse: if <code>true</code>, current row becomes the only
    *           selected row.
    * @param    useValue
    *           Use the specified value for the editor instead of getting it from the browse cache.
    * @param    value
    *           Value to use in the editor. Can be <code>null</code> (valid value). Set
    *           <code>useValue</code> to <code>true</code> if you want to use this parameter.
    * 
    * @return   <code>true</code> if editing was on and ENTRY triggered allowed it to stop.
    */
   @SuppressWarnings("unchecked")
   protected boolean startEdit(WidgetId     prevFocus, 
                               boolean      updateSelection, 
                               boolean      useValue, 
                               BaseDataType value)
   {
      // TODO: this must be generated only after ROW-ENTRY has finished, without NO-APPLY!
      
      if (isEditing() || getCacheSize() <= 0 || !isEditPossible())
      {
         if (!isEditPossible())
         {
            ThinClient.setHelp(config.help, this);
         }
         return false;
      }
      
      BrowseColumnConfig[] cols = getVisibleColumns();
      if (cols[currColumn].readOnly)
      {
         ThinClient.setHelp(config.help, this);
         return false;
      }
      
      Widget<?> oldFocus = currentFocus();
      int oldCol = currColumn;
      
      // allow ENTRY triggers to fire on server side after setting the editor. Note that the
      // trigger is set for browse column not the actual editor
      ThinClient tc = ThinClient.getInstance();
      TriggerResult tr = tc.invokeTriggers(tc.getWidget(getVisibleColumnCfg(currColumn).id),
                                           prevFocus == null ? null : tc.getWidget(prevFocus), 
                                           Keyboard.SE_ENTRY, 
                                           false);

      if (tr.consume)
      {
         return false;
      }

      // when the trigger is not consumed, continue with editing the current column as of before the trigger,
      // note that the trigger could have applied entry (or tab, etc.) to this or another column
      currColumn = oldCol;
      
      // TODO: if the row is not in view, bring it in view.
      
      Widget editor = createEditor(useValue, value);
      
      // the following call is an internal focus change and is safe
      editor.parent().setFocus(editor);
      activeEditor = editor;
      EventManager.postEvent(new FocusEvent(editor, EventType.FOCUS_GAINED));
      
      String columnHelp = cols[currColumn].help;
      ThinClient.setHelp(columnHelp != null ? columnHelp : config.help, this);
      
      if (isSelectionMultiple() && updateSelection)
      {
         selectedRowIndices.clear();
         selectedRowIndices.add(getCurrentRowAsSelected());
      }
      
      return true;
   }

   /**
    * Create editor fill-in for the current column.
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return editor fill-in for the current column.
    */
   protected Widget createEditor(boolean useValue, BaseDataType value)
   {
      return createEditingFillIn(useValue, value);
   }

   /**
    * Create fill-in editor widget for the current column.
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return fill-in editor widget for the current column.
    */
   protected FillIn createEditingFillIn(boolean useValue, BaseDataType value)
   {
      ConfigManager mgr =           ConfigManager.getInstance();
      WidgetConfig controlConfig =  browseModel.getEditorsConfigs()[visibleToOrdinal(currColumn)];
      BrowseColumnConfig[] cols =   getVisibleColumns();
      FillInConfig cfg =            (FillInConfig) controlConfig;

      if (mgr.getActiveConfig(cfg.id) == null)
      {
         // Widget is unregistered when it is destroyed in stopEdit. Register it again with
         // empty backup config so all updated values are pushed to server.
         FillInConfig backupConfig = new FillInConfig();
         backupConfig.cursorOffset = -1;
         mgr.addWidgetConfig(cfg, backupConfig);
      }

      // Change 007 in ConfigManager disables "backup configuration" concept. If it will be
      // turned on back again, that will break client -> server synchronization of MODIFIED
      // attribute for browse columns. The problem is is that backup configuration should match
      // the previous state sent to the server, not the one that widget had on its re-creation
      // time. The solution to this is to:
      // 1. add the map for backup configurations:
      //    private Map<WidgetId, WidgetConfig> editingBackupConfigs = new HashMap<>();
      // 2. add to stopEdit:
      //    WidgetId wid = editor.getId();
      //    editingBackupConfigs.put(wid, ConfigManager.getInstance().getBackupConfig(wid));
      // 3. add to startEdit:
      //    WidgetConfig backupConfig = editingBackupConfigs.get(cfg.id);
      //    ...
      //    mgr.addWidgetConfig(cfg, backupConfig);
      //    mgr.trackAllChanges(cfg);

      if (cols[currColumn].format != null)
         cfg.format = cols[currColumn].format; // direct access

      cfg.dataType = cols[currColumn].dataType; // direct access
      cfg.pfcolor = config.pfcolor;
      cfg.dcolor = config.dcolor;
      cfg.autoReturn = cols[currColumn].autoReturn;
      cfg.autoZap = cols[currColumn].autoZap;
      cfg.help = cols[currColumn].help;
      cfg.appDataPres = null;
      cfg.modified = false;
      cfg.visible = false;
      cfg.readOnly = cols[currColumn].columnReadOnly;

      initEditorConfig(cfg);

      WidgetFactory<?> factory = screen().getFactory();
      WidgetId editorId = cfg.id;
      FillIn fillIn = factory.create(FillIn.class, editorId, cfg);

      screen().getRegistry().addWidget(editorId, fillIn);
      fillIn.initialize(editorId, cfg);

      // reset underline attribute and force Browse colors
      fillIn.setBrowse(this);

      BaseDataType fillInValue = useValue ? value : getValue(visibleToOrdinal(currColumn));
      if (fillInValue != null)
      {
         fillIn.setValue(fillInValue);
         fillIn.setState(ScreenBuffer.CHANGED);
      }
      else
      {
         fillIn.setState(ScreenBuffer.UNINITIALIZED);
      }
      fillIn.setUnderline(false);

      addEditor(fillIn);

      fillIn.setEnabled(true);
      fillIn.setVisible(true);
      fillIn.activate();
      return fillIn;
   }

   /**
    * Display browse-related error message.
    *
    * @param message
    *        Message to display.
    */
   protected void displayErrorMessage(String message)
   {
      window().message(message);
   }

   /**
    * Update specified cell in current row.
    * 
    * @param   columnOrdinal
    *          0-based index of the cell to update.
    * @param   var
    *          New value of the cell.
    * 
    * @return  <code>true</code> if column value was changed.
    */
   protected boolean updateColumn(int columnOrdinal, BaseDataType var)
   {
      if (getCacheSize() == 0)
         return false;

      boolean res = true;

      int currentRow = getCurrentRow();
      BaseDataType oldValue = isRowInView(currentRow) ? dcGetRowData(currentRow)[columnOrdinal] :
                                                        editedRow.data[columnOrdinal];

      if (oldValue != null && oldValue.equals(var))
      { 
         // no real edit
         
         res = false;
         
         if (currRowUpdates != null)
            currRowUpdates[columnOrdinal] = null;
      }
      else
      {
         if (currRowUpdates == null)
            currRowUpdates = new BaseDataType[getColumns().length];

         currRowUpdates[columnOrdinal] = var;
      }
      
      return res;
   }
   
   /**
    * Switches in-line editor to the previous editable cell. Initiates commit
    * of the current row if necessary.
    */
   void switchToPrevCell()
   {
      BrowseColumnConfig[] cols = getVisibleColumns();

      int prevColumn;
      for (prevColumn = currColumn - 1; prevColumn >= 0; 
           --prevColumn)
      {
         if (!cols[prevColumn].readOnly)
            break;
      }

      do
      {
         if (prevColumn >= 0)
         {
            if (!stopEdit(false, true))
            {
               return;
            }

            currColumn = prevColumn;
            startEdit();
            break;
         }

         if (!leaveRow())
            break;

         int oldCurrRow = getCurrentRow();
         goUp(false);

         if (oldCurrRow != getCurrentRow())
         {
            currColumn = cols.length;

            while (cols[--currColumn].readOnly)
            {
               // empty loop body
            }
         }
         
         startEdit();
         // invokeRowEntryTrigger();
         fireValueChanged(); // TODO: investigate whether we should fire it?
      }
      while (false);
   
      repaint();
   }
   
   /**
    * Switches in-line editor to the next editable cell. Initiates commit of
    * the current row if necessary.
    */
   protected void switchToNextCell()
   {
      WidgetId prev = getVisibleColumnCfg(currColumn).id;
      
      BrowseColumnConfig[] cols = getVisibleColumns();

      int nextColumn;
      for (nextColumn = currColumn + 1; nextColumn < cols.length; ++nextColumn)
      {
         if (!cols[nextColumn].readOnly)
            break;
      }

      do
      {
         if (nextColumn < cols.length)
         {
            if (isEditPossible() && !stopEdit(getVisibleColumnCfg(nextColumn).id, false, true))
            {
               return;
            }

            currColumn = nextColumn;
            startEdit(prev);
            break;
         }

         if (!leaveRow())
            break;
   
         int oldCurrRow = getCurrentRow();
         goDown(false);

         if (oldCurrRow != getCurrentRow())
         {
            currColumn = 0;

            while (cols[currColumn].readOnly)
               ++currColumn;
         }
         
         startEdit(prev);
         
      }
      while (false);

      repaint();
   }

   /**
    * Determine if this browse should have the following quirk: is a new row is inserted before
    * the first row or after the last row, the value of the current column is copied from the
    * first/last row to the new row.
    *
    * @return <code>false</code>
    */
   protected boolean isFirstLastRowCopyQuirk()
   {
      return false;
   }

   /**
    * Saves the current row into cache and data model.
    *
    * @param  updateModel
    *         <code>true</code> if data model is updated. Otherwise only cache is updated.
    *
    * @return Is data checking and saving was successful.
    */
   private boolean saveCurrentRow(boolean updateModel)
   {
      if (currRowUpdates == null)
      {
         editedRow = null;
         editedRowIndex = -1;

         notifyRowReleased();
         return true;
      }

      int len = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         if (currRowUpdates[i] != null)
         {
            ++len;
         }
      }

      if (len == 0)
      {
         editedRow = null;
         editedRowIndex = -1;

         notifyRowReleased();
         return true;
      }

      int[] indices = new int[len];
      BaseDataType[] values = new BaseDataType[len];
      
      int p = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         BaseDataType update = currRowUpdates[i];
         if (update != null)
         {
            indices[p] = i;
            values[p] = update;
            ++p;
         }
      }

      try
      {
         int currRow = getCurrentRow();
         savingRow = true;
         dcUpdateRow(currRow, indices, values, updateModel, isRowInView(currRow), editedRow);
      }
      catch (DataValidationException ex)
      {
         // Enable internal, row-level undo.
         awaitingEndError = true;
         String message =
               String.format("Press %s to undo changes to the current row in BROWSE %s. (5133)",
                             Keyboard.kbLabel(Keyboard.END_ERROR),
                             config.name);
         displayErrorMessage(message);
         return false;
      }
      finally
      {
         savingRow = false;
      }

      currRowUpdates = null;
      editedRow = null;
      editedRowIndex = -1;

      return true;
   }

   /**
    * Notify server that the row has been released. You do not need to call this function if you
    * have any data to save.
    */
   private void notifyRowReleased()
   {
      WidgetConfig[] cfgs = browseModel.getEditorsConfigs();
      if (cfgs != null)
      {
         for (WidgetConfig cfg : cfgs)
         {
            if (cfg != null)
            {
               cfg.modified = false;
            }
         }
      }

      try
      {
         browseModel.updateRow(getCurrentRow(), null, null, true);
      }
      catch (DataValidationException e)
      {
         // no data validation is performed
      }
   }

   /**
    * Finishes editing and commits row to model if some updates are available. Restores editor if
    * commit has failed.
    *
    * @param   keepColumn
    *          Flag indicating if we keep the {@link #currColumn} or compute the new editable one,
    *          starting with the first one.
    *
    * @return  <code>true</code> if all actions were successful.
    */
   protected boolean releaseCurrentRow(boolean keepColumn)
   {
      return releaseCurrentRow(true, true, keepColumn);
   }

   /**
    * Finishes editing and commits row to model if some updates are available. Restores editor if
    * commit has failed.
    *
    * @param   updateModel
    *          <code>true</code> if data model is updated. Otherwise only cache is updated and
    *          validation does NOT take place.
    * @param   triggerRowLeave
    *          <code>true</code> to trigger ROW-LEAVE event.
    * @param   keepColumn
    *          Flag indicating if we keep the {@link #currColumn} or compute the new editable one,
    *          starting with the first one.
    *
    * @return  <code>true</code> if all actions were successful.
    */
   private boolean releaseCurrentRow(boolean updateModel,
                                     boolean triggerRowLeave,
                                     boolean keepColumn)
   {
      Widget editor = getActiveEditor();
      
      if (editor == null)
         return true;

      if (updateModel)
      {
         if (!validateEditorValue())
         {
            return false;
         }
      }

      int nextColumn = keepColumn ? currColumn : getFirstEditableColumn(0);
      WidgetId nextId = nextColumn < 0 ? null : getVisibleColumnCfg(nextColumn).id;

      boolean status = stopEdit(nextId, false, false);
      if (!status)
         return false;

      if (triggerRowLeave)
      {
         if (!invokeRowLeaveTrigger(nextId))
         {
            return false;
         }
      }

      return saveCurrentRow(updateModel && !isNewRow());
   }

   /**
    * Returns the row data with the given index. The row must already be in
    * the cache. 
    *
    * @param  rowIndex
    *         Row's index in model.
    *
    * @return The row data from the cache.
    */
   protected BaseDataType[] dcGetRowData(int rowIndex)
   {
      return dcData[dcGetRowIndex(rowIndex)].data;
   }

   /**
    * Returns the cached row with the given index. The row must already be in
    * the cache.
    *
    * @param  rowIndex
    *         Row's index in model.
    *
    * @return The row from the cache.
    */
   protected BrowseRow dcGetRow(int rowIndex)
   {
      return dcData[dcGetRowIndex(rowIndex)];
   }

   /**
    * If the browse has row markers drawn. Note that the returned value may differ from
    * {@link com.goldencode.p2j.ui.BrowseWidget#hasRowMarkers()}.
    *
    * @return <code>true</code> if the browse has row markers drawn.
    */
   public boolean hasRowMarkers()
   {
      return config.rowMarkers != null ? config.rowMarkers :
                                         (isEditPossible() || config.dynamic);
   }

   /**
    * Updates row in the cache and the model. Row must already be in cache.
    *
    * @param  rowIndex
    *         0-based row's index in model.
    * @param  cellIndices
    *         Indexes of cells to update.
    * @param  cellValues
    *         New values of cells to update.
    * @param  updateModel
    *         <code>true</code> if data model is updated. Otherwise only cache is updated.
    * @param  rowInView
    *         <code>true</code> if the edited row is in view.
    * @param  editedRowData
    *         Data of the edited row before editing has started.
    *
    * @throws DataValidationException
    *         New values don't passed database validation rules.
    */
   private void dcUpdateRow(int            rowIndex,
                            int[]          cellIndices,
                            BaseDataType[] cellValues,
                            boolean        updateModel,
                            boolean        rowInView,
                            BrowseRow      editedRowData)
   throws DataValidationException
   {
      updateModel = updateModel && !config.noAssign;
      if (updateModel)
      {
         browseModel.updateRow(rowIndex, cellIndices, cellValues, false);
      }
      else
      {
         notifyRowReleased();
      }

      if (rowInView)
      {
         BrowseRow row = dcData[dcGetRowIndex(rowIndex)];

         if (editedRowData != null)
         {
            System.arraycopy(editedRowData.data, 0, row.data, 0, editedRowData.data.length);
            System.arraycopy(editedRowData.text, 0, row.text, 0, editedRowData.text.length);
         }

         for (int i = 0; i < cellIndices.length; ++i)
         {
            int idx = cellIndices[i];
            row.data[idx] = cellValues[i];
            row.text[idx] = Browse.getCellTextNoError(getColumns()[idx], row.data[idx]);
         }
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Reference to keyboard event.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
      int key = event.actionCode();
      
      Widget editor   = getActiveEditor();
      boolean editing = editor != null;
      
      if (key == Keyboard.KA_TAB)
      {
         event.consume();

         handleTabKey();
         return;
      }
      
      if (key == Keyboard.KA_BACK_TAB && editor == null)
      {
         event.consume();

         if (releaseCurrentRow(false))
         {
            //TODO: editor can only be null at this point (see 'if' 5 lines
            // above. Requires deeper investigation.
            //if (editor != null)
            //   startEdit(); // editor must be present in edit mode
            
            // the following request should be conditional
            ThinClient.previousTabStop(Browse.this);
         }
         return;
      }

      if (key == Keyboard.KA_HOME || key == Keyboard.KA_LEFT_END ||
          key == Keyboard.KA_END || key == Keyboard.KA_RIGHT_END)
      {
         if (initialDataSetObtained ? (getCacheSize() == 0 ||
                                         !isEnabled() ||
                                         (editing && (key == Keyboard.KA_HOME ||
                                                      key == Keyboard.KA_END))   ||
                                         !leaveRow()) :
                                      !navigable)
         {
            return;
         }

         int oldCurrRow = getCurrentRow();

         if (key == Keyboard.KA_HOME || key == Keyboard.KA_LEFT_END)
         {
            boolean anyRows = scrollToRow(0, false, true);
            if (anyRows)
            {
               setCurrentRow(getCachedRowNumber(0), isAggressiveCurrentRowSelection(), true);
            }
         }
         else
         {
            if (getCacheSize() < getDown() && getCacheSize() > 0)
            {
               setCurrentRow(getCachedRowNumber(getCacheSize() - 1),
                             isAggressiveCurrentRowSelection(),
                             true);
            }
            else
            {
               int bigNumber = Integer.MAX_VALUE - 1000; // decreased to avoid overflow
               boolean anyRows = scrollToRow(bigNumber, true, true);
               if (anyRows)
               {
                  incrementalScroll(-(getDown() - 1));
               }
               updateMaxDataGuess(true);
            }
         }

         if (editing)
         {
            startEdit();
         }

         if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
         {
            fireValueChanged();
         }

         event.consume();

         if (isEditing())
         {
            editor = getActiveEditor(); // editor may be recreated
            editor.repaint();
         }
         return;
      }

      if (getCacheSize() == 0)
      { 
         // all other keys requires the presence of current row
         return;
      }
      
      boolean invalid = true;

      do
      {
         if (key == Keyboard.KA_CURSOR_LEFT)
         {
            if (scrollColumn == 0)
            {
               invalid = false;
               break;
            }
               
            --scrollColumn;
            if (scrollColumn < ilFirstScrollCol)
            {
               scrollColumn = 0;
            }

            rcLabels = null;
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_RIGHT)
         {
            int colsCount = getVisibleColumns().length - 1;
            if (scrollColumn >= colsCount)
            {
               invalid = false;
               break;
            }
               
            if (scrollColumn == 0)
            {
               if (ilFirstScrollCol > colsCount)
               {
                  invalid = false;
                  break;
               }
               scrollColumn = ilFirstScrollCol;
            }
            else
            {
               ++scrollColumn;
            }
               
            rcLabels = null;
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_UP)
         {
            goUp(true);
            break;
         }
            
         if (key == Keyboard.KA_CURSOR_DOWN ||
             (key == Keyboard.KA_RETURN && isEditing()))
         {
            goDown(true);
            break;
         }
         
         if (key == Keyboard.KA_RETURN && !isEditing())
         {
           fireDefaultAction();
           break;
         }
         
         if (key == Keyboard.KA_PAGE_UP)
         {
            // In 4GL if the previous page is fully deleted, you have to press Page Up twice in
            // order to move to the page before the deleted one. But eventually gaps of deleted
            // rows can be closed and you'll have to press Page Up only once.
            // In P2J for now you'll have to press Page Up only once for the beginning.
            // TODO can be implemented after gap closing mechanism is implemented

            if (editing && !leaveRow())
               break;

            final int oldScrollRow = getScrollRow();
            int oldCurrRow = getCurrentRow();
            int newScrollRow;
            if (oldScrollRow > 0)
            {
               int pageSize = getPageSize();
               int virtualRow = Math.max(oldScrollRow - pageSize, 0);
               int viewportRow = getViewportRow();

               while (true)
               {
                  boolean anyRows = scrollToRow(virtualRow, false, true);
                  newScrollRow = getScrollRow();

                  if (!anyRows            ||
                      virtualRow == 0     ||
                      newScrollRow < oldScrollRow)
                  {
                     // The result set is empty OR
                     // We reached the first row OR
                     // We found a page with has an least one row.
                     break;
                  }

                  virtualRow = Math.max(virtualRow - pageSize, 0);
               }

               viewportRow = newScrollRow == oldScrollRow ? 0 :
                             Math.min(getCacheSize() - 1, viewportRow);
               setCurrentRow(getCachedRowNumber(viewportRow));
            }
            else
            {
               setCurrentRow(0);
            }

            if (editing)
            {
               startEdit();
            }
            
            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
            {
               // invokeRowEntryTrigger();
               fireValueChanged(); // TODO: investigate whether we should fire it?
            }
            
            break;
         }

         if (key == Keyboard.KA_PAGE_DOWN)
         {
            if (editing && !leaveRow())
               break;

            int oldCurrRow = getCurrentRow();
            int viewportRow = getViewportRow();

            if (getCacheSize() < getDown() ||
                !incrementalScroll(getPageSize(), true, false))
            {
               // no more rows after
               int newRowModelIndex = getCachedRowNumber(getCacheSize() - 1);
               setCurrentRow(newRowModelIndex);
            }
            else
            {
               int row = Math.min(viewportRow, getCacheSize() - 1);
               setCurrentRow(getCachedRowNumber(row));
            }

            if (editing)
            {
               startEdit();
            }
            
            if (!isSelectionMultiple() && oldCurrRow != getCurrentRow())
            {
               // invokeRowEntryTrigger();
               fireValueChanged(); // TODO: investigate whether we should fire it?
            }
            
            break;
         }

         if (key == Keyboard.KA_REPLACE)
         {
            if (editing)
            {
               //TODO check how REPLACE key should be handled correctly (ChUI)
               setReadOnly(true);
               releaseCurrentRow(false);
               invalid = true;
            }
            else
            {
               if (!isEditPossible())
               {
                  invalid = false;
                  break;
               }

               setReadOnly(false);
            }

            break;
         }

         if (key == Keyboard.SE_ENTRY)
         {
            break;
         }

         if (editing && (key == Keyboard.KA_BACK_TAB))
         {
            switchToPrevCell();
            break;
         }

         if (editing && (key == Keyboard.KA_EDITOR_TAB))
         {
            switchToNextCell();
            break;
         }

         if (editing && key == Keyboard.KA_END_ERROR && awaitingEndError)
         {
            awaitingEndError = false;
            undoValue();
            break;
         }

         return; // ignoring other keys
      }
      while (false);

      event.consume(); // consume all handled keys
      
      if (invalid)
      {
         repaint();
      }
      else if (isEditing())
      {
         editor = getActiveEditor(); // editor may be recreated
         editor.repaint();
      }
   }
   
   /**
    * Intercept special events. Using regular keyTyped/keyPressed may not work
    * if component is not in focus.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#processEvent(com.goldencode.p2j.ui.client.event.Event)
    */
   @Override
   public void processEvent(Event evt) 
   {
      if (evt instanceof KeyInput)
      {
         KeyInput ki = (KeyInput) evt;
         int action = ki.actionCode();
         
         if(action == Keyboard.SE_VALUE_CHANGED)
         {
            return;
         }
         
         if (action == Keyboard.SE_ENTRY)
         {
            Widget<?> widPrevFocused = evt.other(); 
            if (widPrevFocused != null)
            {
               startEdit(widPrevFocused.getId());
            }
            else
            {
               startEdit();
            }
            ki.consume();
            return;
         }
         
         if (action == Keyboard.SE_LEAVE)
         {
            if (isEditing())
            {
               WidgetId nextFocus = evt.other() == null ? null : evt.other().getId();
               if (!invokeLeaveCellTrigger(nextFocus) || !validateStopEdit(false, true))
               {
                  ki.consume();
                  return;
               }
               if (!invokeRowLeaveTrigger(nextFocus))
               {
                  ki.consume();
                  return;
               }
               
               finishStopEdit();
            }
         }
         
         onKeyPressed(ki);
         boolean consumed = ki.isConsumed();
         
         if (consumed                     || 
             action == Keyboard.KA_RETURN || 
             action == Keyboard.keyCode(" "))
         {
            // if the key was processed by the browse, clear the message area
            // if space or return was pressed, clear the message area
            window().clearPutScreenInMessageArea(-1);
         }
         
         if (consumed)
            return;
      }
      
      super.processEvent(evt);
   }
   
   /**
    * Handles ' ' key in R/O mode: fires ValueChanged event. Additionally
    * inverts selection status of the current line in multiple selection
    * mode.
    *
    * @see com.goldencode.p2j.ui.client.event.listener.KeyListener#onKeyTyped(KeyInput)
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
      if (getCacheSize() == 0)
      {
         return;
      }
      
      if (event.keyCode() == ' ' && !isEditing())
      {
         switchCurrentRowSelection();

         fireValueChanged();
         repaint();
      }
   }

   /**
    * Notify the browse that the key was pressed when the editor is in focus.
    *
    * @param key
    *        The key which was pressed in the editor.
    */
   public void notifyKeyPressedInEditor(KeyInput key)
   {
      if (key.isRealKey())
      {
         scrollToEditableRow();
      }
   }

   /**
    * For browse with multiple selection: select current row if it is not selected or deselect if
    * it is selected.
    */
   protected void switchCurrentRowSelection()
   {
      if (isSelectionMultiple())
      {
         SelectedRow currRow = getCurrentRowAsSelected();

         if (selectedRowIndices.contains(currRow))
         {
            selectedRowIndices.remove(currRow);
         }
         else
         {
            selectedRowIndices.add(currRow);
         }
      }
   }

   /**
    * Get page size used for Page Up / Page Down.
    *
    * @return increment size (in rows) used for Page Up / Page Down.
    */
   protected int getPageSize()
   {
      return getDown();
   }

   /**
    * Set the current row.
    *
    * @param   currRow
    *          New current row (0-based index in the browse model).
    */
   protected void setCurrentRow(int currRow)
   {
      setCurrentRow(currRow, true);
   }

   /**
    * Set the current row.
    *
    * @param   currRow
    *          New current row (0-based index in the browse model).
    * @param   allowDeletedRow
    *          If <code>true</code>, then <code>currRow</code> becomes the current row even
    *          if it doesn't exist in the result set anymore. If <code>false</code> then the next
    *          available row becomes the current row if the <code>currRow</code> doesn't exist.
    */
   private void setCurrentRow(int currRow, boolean allowDeletedRow)
   {
      setCurrentRow(currRow, true, allowDeletedRow);
   }

   /**
    * Set the current row.
    * 
    * @param   currRow
    *          New current row (0-based index in the browse model).
    * @param   updateSelection
    *          <code>true</code> if selection should be adjusted.
    * @param   allowDeletedRow
    *          If <code>true</code>, then <code>currRow</code> becomes the current row even
    *          if it doesn't exist in the result set anymore. If <code>false</code> then the next
    *          available row becomes the current row if the <code>currRow</code> doesn't exist.
    */
   protected void setCurrentRow(int currRow, boolean updateSelection, boolean allowDeletedRow)
   {
      if (browseModel == null || getCacheSize() == 0)
      {
         this.currentRow = 0;
         currentRowAvail = false;
      }
      else
      {
         if (!isNewRow(currRow))
         {
            int availRow = browseModel.setCurrentRow(currRow);
            this.currentRow = allowDeletedRow ? currRow : availRow;
            currentRowAvail = availRow == currRow;
         }
         else if (currentRow != currRow)
         {
            // a new [virtual] row, without database backing, buffers must be released
            browseModel.setCurrentRow(currRow);
            this.currentRow = currRow;
            currentRowAvail = false;
         }
         
         if (updateSelection)
         {
            currRowSelected = true;
         }
      }
   }

   /**
    * Returns <code>true</code> if the current row of the single-selection browse is selected on
    * FOCUS-GAINED and some other events.
    *
    * @return see above.
    */
   protected boolean isAggressiveCurrentRowSelection()
   {
      return true;
   }

   /**
    * Handle TAB key press.
    */
   protected void handleTabKey()
   {
      if (releaseCurrentRow(false))
      {
         if (isEditing())
         {
            startEdit(false); // editor must be preserved in edit mode
         }

         ThinClient.nextTabStop(Browse.this);
      }
   }
   
   /**
    * Get repositioned row. I.e. the index of the viewport row where records positioned with the
    * REPOSITION statement are displayed.
    *
    * @return 0-based repositioned row.
    */
   public int getRepositionedRow()
   {
      return config.repositionedRow;
   }

   /**
    * Returns reposition mode set be SET-REPOSITIONED-ROW.
    *
    * @return <code>true</code> if reposition mode is "CONDITIONAL", <code>false</code> if
    *         reposition mode is "ALWAYS".
    */
   public boolean isConditionalReposition()
   {
      return config.conditionalReposition;
   }

   /**
    * Get the currently editing column.
    *
    * @return The {@link #currColumn} value or <code>null</code> if not editing..
    */
   public BrowseColumnConfig getCurrentColumn()
   {
      return !isEditing() || currColumn < 0 ? null : getVisibleColumns()[currColumn];
   }

   /**
    * Set the currently editing column.
    */
   public void setCurrentColumn(int colIdx)
   {
      currColumn = colIdx;
   }
   
   /**
    * Get text value of the given column in current row.
    * 
    * @param   columnOrdinal
    *          0-based index of the column among all columns.
    * 
    * @return  The text value of the given column in current row.
    */
   public String getColumnText(int columnOrdinal)
   {
      BaseDataType value = null;
      Integer visIndex = ordinalToVisible(columnOrdinal);
      if (visIndex != null)
      {
         if (visIndex == currColumn && isEditing())
         {
            if (isEditorFillIn())
            {
               return getActiveFillIn().getText();
            }
            else if (isEditorToggleBox())
            {
               value = getEditorValue();
            }
            else
            {
               // TODO combo-box
            }
         }

         columnToView(visIndex);
      }
      
      if (value == null)
      {
         value = getValue(columnOrdinal);
      }

      if (value == null || value.isUnknown())
         return "";

      String fmt = getColumns()[columnOrdinal].format;
      return (fmt == null) ?
                            character.valueOf(value).getValue():
                            character.valueOf(value, fmt).getValue();
   }

   /**
    * Remove given row(s) from the set of selected rows.
    *
    * @param deletedRowIndex
    *        0-based index of the deleted row in the browse model. <code>-1</code> if all selected
    *        rows are deleted.
    */
   public void removeFromSelection(int deletedRowIndex)
   {
      if (!isSelectionMultiple())
         return;

      if (deletedRowIndex == -1)
      {
         selectedRowIndices.clear();
      }
      else
      {
         SelectedRow[] res = selectedRowIndices.toArray(
                              new SelectedRow[selectedRowIndices.size()]);
         selectedRowIndices.clear();

         boolean newRow = isNewRow(deletedRowIndex);
         for (SelectedRow selectedRow : res)
         {
            if (newRow)
            {
               if (selectedRow.getRowIndex() != deletedRowIndex)
               {
                  selectedRowIndices.add(selectedRow);
               }

               UnimplementedFeature.todo("4GL can make selection invalid if there is a new row " +
                                         "in the set of selected rows.");
            }
            else
            {
               if (selectedRow.getRowIndex() < deletedRowIndex)
               {
                  selectedRowIndices.add(selectedRow);
               }
               else if (selectedRow.getRowIndex() > deletedRowIndex)
               {
                  selectedRowIndices.add(new SelectedRow(selectedRow.getRowIndex() - 1,
                        selectedRow.isRowAvailable()));
               }
            }
         }
      }

      repaint();
   }

   /**
    * Get the number of non-new selected rows before the given base row.
    *
    * @param  baseRow
    *         0-based index of the base row in the browse model.
    *
    * @return see above.
    */
   private int getRowDeletionShift(int baseRow)
   {
      if (!isSelectionMultiple())
         return 0;

      int shift = 0;
      for (SelectedRow row : selectedRowIndices)
      {
         int index = row.getRowIndex();
         if (!isNewRow(index) && index < baseRow)
            shift++;
      }

      return shift;
   }

   /**
    * Get value for the specified column.
    * 
    * @param   columnOrdinal
    *          0-based column ordinal.
    *          
    * @return  Column value.
    */
   protected BaseDataType getValue(int columnOrdinal)
   {
      if (currentRowDisplayStrategy != null)
      {
         return null;
      }

      if (currRowUpdates != null && currRowUpdates[columnOrdinal] != null)
         return currRowUpdates[columnOrdinal];
      
      if (getCacheSize() == 0)
         return null;

      return dcGetRowData(getCurrentRow())[columnOrdinal];
   }

   /**
    * Get value of the given column in current row.
    * 
    * @param   columnOrdinal
    *          0-based index of the column among the columns.
    * 
    * @return  The value of the given column in current row.
    */
   public BaseDataType getColumnValue(int columnOrdinal)
   {
      Integer visIndex = ordinalToVisible(columnOrdinal);
      if (visIndex != null && visIndex == currColumn && isEditing())
         return getEditorValue();
         
      BaseDataType value = getValue(columnOrdinal);

      if (value == null)
         return new character();

      return value; 
   }

   /**
    * Set value for the given column in current row.
    *
    * @param   columnOrdinal
    *          0-based index of the column among the visible columns.
    * @param   value
    *          The value for the cell.
    */
   public void setColumnValue(int columnOrdinal, BaseDataType value)
   {
      int currentRow = getCurrentRow();
      if (config.readOnly || getColumns()[columnOrdinal].readOnly)
      {
         // update cache without marking it as changes to commit
         BrowseRow row = dcGetRow(currentRow);
         row.data[columnOrdinal].assign(value);
         row.text[columnOrdinal] =
               Browse.getCellTextNoError(getColumns()[columnOrdinal], value);
         repaint();
         return;
      }
      
      // update currRowUpdates - this needs to be done here otherwise the caches will match the 
      // new value, and old value will be gone
      updateColumn(columnOrdinal, value);

      // update cache
      dcGetRowData(currentRow)[columnOrdinal] = value.duplicate();
      if (editedRow != null && editedRow.rowIndex == currentRow)
      {
         editedRow.data[columnOrdinal] = value.duplicate();
         editedRow.text[columnOrdinal] =
               Browse.getCellTextNoError(getColumns()[columnOrdinal], value);
      }
      
      if (columnOrdinal == currColumn && isEditing())
      {
         setEditorValue(value);
         return;
      }
      
      repaint();
   }

   /**
    * Check if the value in specified column in current row is changed. 
    * 
    * @param   columnOrdinal
    *          0-based index of the column.
    *          
    * @return  <code>true</code> if value is changed.
    */
   public boolean isModified(int columnOrdinal)
   {
      Integer visIndex = ordinalToVisible(columnOrdinal);
      if (visIndex != null)
      {
         if (visIndex == currColumn && isEditing())
         {
            return isEditorModified();
         }

         columnToView(visIndex);
      }

      return currRowUpdates != null && currRowUpdates[columnOrdinal] != null;
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param deleteCurrentRowMode
    *        <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *        <code>false</code> if this function is used for update on REFRESH.
    */
   public void refresh(boolean deleteCurrentRowMode)
   {
      if (config.performLayout)
      {
         return;
      }

      int baseForCurrentNewRow = -1;
      BrowseRow currentNewRow = null;

      boolean editing = isEditing();
      int insState = -1;
      if (editing)
      {
         // preserve Insert state on refresh
         insState = Window.getInsertMode();

         if (isNewRow() && !deleteCurrentRowMode)
         {
            // 4GL preserves a new row on refresh if it is the current row. The first non-new row
            // above it is used as the base row. After refresh the new row will be inserted after
            // the base row. If there is no base row, or the base row is not present in the view
            // after refresh, the current new row will be deleted with other non-current new rows.
            int viewportRow = getViewportRow();
            for (int i = viewportRow - 1; i >= 0; i--)
            {
               int row = getCachedRowNumber(i);
               if (!isNewRow(row))
               {
                  baseForCurrentNewRow = row;
                  break;
               }
            }

            // no ROW-LEAVE event
            releaseCurrentRow(true, false, true);

            if (baseForCurrentNewRow != -1)
               currentNewRow = dcData[viewportRow];
         }
         else
         {
            undoValue(false);
         }
      }

      int currRow;
      int viewportRow;
      int firstExistingRow;

      boolean initialRowSet = initialRow != null;

      if (initialRowSet)
      {
         currRow = initialRow;
         viewportRow = 0;
         firstExistingRow = initialRow;

         initialRow = null;
      }
      else
      {
         currRow = getCurrentRow();
         viewportRow = getViewportRow();

         firstExistingRow = getFirstExistingRow();
         if (firstExistingRow < 0)
         {
            firstExistingRow = 0;
         }
      }

      // preserve currRowSelected state for single-selection browse
      boolean updateSelection = currRowSelected ||
                                isAggressiveCurrentRowSelection() ||
                                deleteCurrentRowMode;

      scrollToRow(firstExistingRow,
                  initialRowSet,
                  false,
                  initialRowSet,
                  false,
                  true,
                  -1,
                  updateSelection);

      if (baseForCurrentNewRow != -1 && isRowInView(baseForCurrentNewRow))
      {
         addNewRow(currentNewRow.rowIndex,
                   currentNewRow.data,
                   getViewportRow(baseForCurrentNewRow),
                   true);
      }

      int cacheSize = getCacheSize();
      if (cacheSize > 0)
      {
         boolean found = deleteCurrentRowMode || isRowInView(currRow);

         boolean fireValueChanged = false;
         if (found)
         {
            // Old current row is in the viewport - preserve viewport row number.
            int newViewportRow = Math.min(viewportRow, getCacheSize() - 1);
            setCurrentRow(getCachedRowNumber(newViewportRow), updateSelection, true);

            if (viewportRow != newViewportRow || deleteCurrentRowMode)
            {
               fireValueChanged = true;
            }
         }
         else
         {
            // Old current row is not in the viewport - use the scroll row as the current row.
            scrollToRow(firstExistingRow, true, false, false, false, true, -1, updateSelection);
            fireValueChanged = true;
         }

         if (editing)
         {
            startEdit();
            
            if (isEditing() && isEditorFillIn())
            {
               getActiveFillIn().setInsertMode(insState == Window.INSERT_MODE_INS);
            }
         }
         
         if (!isSelectionMultiple() && fireValueChanged)
         {
            // invokeRowEntryTrigger();
            fireValueChanged(); // TODO: investigate whether we should fire it?
         }
      }
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      Widget<?> parent = ThinClient.getInstance().getWidget(config.frameId);
      if (!isVisible() && parent != null && parent.isVisible())
      {
         if (config.performLayout)
         {
            doLayout();
         }
         refresh(false);
      }
      
      super.show();
   }
   
   /**
    * Execute pending refresh when REFRESHABLE is turned on. Includes pending query repositions
    * and reopenings.
    *
    * @param   queryReopened
    *          <code>true</code> if there is a pending query reopening.
    * @param   repositionBeforeReopen
    *          <code>true</code> if the last pending reposition was <i>before</i> the last pending
    *          reopening.
    * @param   repositionedRow
    *          0-based index of the last row the query was repositioned to. <code>-1</code> if
    *          there is no repositioned row.
    */
   public void executePendingRefresh(boolean queryReopened,
                                     boolean repositionBeforeReopen,
                                     int repositionedRow)
   {
      if (browseModel != null)
      {
         if (queryReopened || repositionedRow != -1)
         {
            if (isEditing())
            {
               undoValue(false);
            }

            if (queryReopened && isSelectionMultiple())
            {
               selectedRowIndices.clear();
            }

            if (queryReopened && (repositionedRow == -1 || repositionBeforeReopen))
            {
               int viewportRow = getViewportRow();
               scrollToRow(0, false, false, false, false, false);

               viewportRow = Math.min(getCacheSize() - 1, viewportRow);
               setCurrentRow(getCachedRowNumber(viewportRow));
            }
            else
            {
               if (queryReopened)
               {
                  // This initializes rowCountAssessment for proper future scrolling.
                  scrollToRow(0, false, false, false, false, false);
               }

               // conditional reposition is performed as unconditional
               scrollToRow(repositionedRow, true, false, false, true);
            }

            if (isEditPossible() && getCacheSize() > 0)
            {
               startEdit();
            }
         }

         /*
         TODO commented code takes place only in quirky cases. One of these cases is
         1. REFRESHABLE off
         2. reposition to a row
         3. go outside of a view with arrows
         4. REFRESHABLE on

         if (queryReopened)
         {
            if (selectedRowIndices != null)
               selectedRowIndices.clear();

            if (editing)
            {
               undoValue();
               stopEdit();
            }

            if (repositionedRow >= 0 && !repositionBeforeReopen)
            {
               // This initializes rowCountAssessment for proper future scrolling.
               scrollToRow(0, false, false);
               if (editing)
                  stopEdit();

               scrollToRow(repositionedRow, true);
            }
            else
            {
               int viewportRow = getViewportRow();
               scrollToRow(0);
               if (editing)
                  stopEdit();

               if (repositionedRow >= 0 && isRowInView(repositionedRow))
               {
                  setCurrentRow(repositionedRow);
               }
               else
               {
                  viewportRow = Math.min(getCacheSize() - 1, viewportRow);
                  setCurrentRow(getCachedRowNumber(viewportRow));
               }
            }

            if (editing && getCacheSize() > 0)
               startEdit();
         }
         else if (repositionedRow >= 0)
         {
            if (!editing)
            {
               if (!isSelectionMultiple())
               {
                  // read-only single selection browse
                  scrollToRow(getCurrentRow(), false, false);
                  if (isRowInView(repositionedRow))
                     setCurrentRow(repositionedRow);
                  else
                     setCurrentRow(getCachedRowNumber(0));
               }
               else
               {
                  // read-only multi selection browse
                  int lastSelectedRow = getLastAvailableSelectedRow(false);
                  if (lastSelectedRow == -1)
                  {
                     scrollToRow(repositionedRow, true, false);
                  }
                  else
                  {
                     int viewportRow = getViewportRow();
                     scrollToRow(lastSelectedRow, false, false);
                     if (isRowInView(repositionedRow))
                     {
                        setCurrentRow(repositionedRow);
                     }
                     else
                     {
                        viewportRow = Math.min(getCacheSize() - 1, viewportRow);
                        setCurrentRow(getCachedRowNumber(viewportRow));
                     }
                  }
               }
            }
            else
            {
               // editable browse
               undoValue();
               stopEdit();
               scrollToRow(repositionedRow, true, false);
               if (getCacheSize() > 0)
                  startEdit();
            }
         }*/
      }
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    */
   public void refreshScrollRow(boolean leaveOnRowInView, boolean preserveViewportRow)
   {
      refreshScrollRow(leaveOnRowInView, preserveViewportRow, false, false, -1);
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param scrollRow
    *        New scroll row.
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    */
   public void refreshScrollRow(int scrollRow,
                                boolean leaveOnRowInView,
                                boolean preserveViewportRow)
   {
      refreshScrollRow(scrollRow, leaveOnRowInView, preserveViewportRow, false, false);
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    * @param repositioned
    *        If the browse is refreshed because the query was repositioned.
    * @param repositionError
    *        For reposition mode - if an error has happened during reposition.
    * @param targetRepositionRow
    *        For reposition mode - row used as the target for reposition. If it was deleted,
    *        reposition with fetching may end up on a different row.
    */
   public void refreshScrollRow(boolean leaveOnRowInView,
                                boolean preserveViewportRow,
                                boolean repositioned,
                                boolean repositionError,
                                int targetRepositionRow)
   {
      if (browseModel != null)
      {
         refreshScrollRow(browseModel.getCurrentRow(),
                          leaveOnRowInView,
                          preserveViewportRow,
                          repositioned,
                          repositionError,
                          targetRepositionRow);
      }
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param scrollRow
    *        New scroll row.
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    * @param repositioned
    *        If the browse is refreshed because the query was repositioned.
    * @param repositionError
    *        For reposition mode - if an error has happened during reposition.
    */
   public void refreshScrollRow(int scrollRow,
                                boolean leaveOnRowInView,
                                boolean preserveViewportRow,
                                boolean repositioned,
                                boolean repositionError)
   {
      refreshScrollRow(scrollRow,
                       leaveOnRowInView,
                       preserveViewportRow,
                       repositioned,
                       repositionError,
                       -1);
   }
   
   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param scrollRow
    *        New scroll row.
    * @param leaveOnRowInView
    *        If the current row is in the viewport, do not update the scroll row.
    * @param preserveViewportRow
    *        Try to preserve the current viewport row (i.e. the row at the same position in
    *        the viewport becomes the current row after refresh).
    * @param repositioned
    *        If the browse is refreshed because the query was repositioned.
    * @param repositionError
    *        For reposition mode - if an error has happened during reposition.
    * @param targetRepositionRow
    *        For reposition mode - row used as the target for reposition. If it was deleted,
    *        reposition with fetching may end up on a different row.
    */
   public void refreshScrollRow(int scrollRow,
                                boolean leaveOnRowInView,
                                boolean preserveViewportRow,
                                boolean repositioned,
                                boolean repositionError,
                                int targetRepositionRow)
   {
      if (browseModel != null)
      {
         int viewportRow = getViewportRow();

         boolean editing = isEditing();
         if (editing)
         {
            undoValue(false);
         }

         int row = scrollRow;
         if (row < 0)
         {
            // Protection code if the query has gone off-end for some reason.
            row = 0;
         }

         if ((leaveOnRowInView || (repositioned && isConditionalReposition())) &&
              isRowInView(row))
         {
            // target row is in view, no need to scroll in this mode
            if (targetRepositionRow != -1 && targetRepositionRow != row)
            {
               this.currentRow = targetRepositionRow;
               currentRowAvail = false;
            }
            else
            {
               this.currentRow = row;
               currentRowAvail = true;
            }
         }
         else
         {
            scrollToRow(row, true, false, false, repositioned && !repositionError);
         }

         if (preserveViewportRow)
         {
            viewportRow = Math.min(viewportRow, getCacheSize() - 1);
            setCurrentRow(getCachedRowNumber(viewportRow), true);
         }

         if (editing)
            startEdit(!repositioned);

         repaint();
      }
   }

   /**
    * Refresh the browse after some of its rows were deleted.
    *
    * @param   rowIndex
    *          0-based index of the deleted row in the browse model. <code>-1</code> if all of the
    *          selected rows were deleted.
    */
   public void refreshAfterDeletion(int rowIndex)
   {
      boolean editing = isEditing();
      if (editing)
      {
         undoValue(false);
      }

      // refresh procedures in single and multiple rows case are different
      boolean singleRow = rowIndex != -1;

      if (!singleRow)
      {
         int numSelected = getSelectedRowsCount();

         if (numSelected == 1)
         {
            singleRow = true;
            rowIndex = getSelectedRowIndex(0);
         }
         else
         {
            int lastRow    = getMaximalSelectedRowIndex();
            boolean inView = isRowInView(lastRow);

            if (!inView)
            {
               int shift = getRowDeletionShift(lastRow + 1);
               removeFromSelection(-1);
               boolean rowsAfterLast = false;

               int rowCount = browseModel.getRowCount() + shift;
               if (rowCount > lastRow + 1)
               {
                  rowsAfterLast = true;
               }

               // scroll to the row after the last deleted row
               if (rowsAfterLast)
               {
                  refreshScrollRow(lastRow + 1 - shift, false, true);
               }
               else
               {
                  int scrollRow = 0;
                  BrowseRow[] row = getRows(lastRow - shift,
                                            1,
                                            false,
                                            NO_DISPLAY);
                  if (row.length > 0)
                     scrollRow = row[0].rowIndex;
                  refreshScrollRow(scrollRow, false, true);
               }
            }
            else
            {
               // otherwise preserve the scroll row
               int scrollRow = getScrollRow();
               int shift = getRowDeletionShift(scrollRow);
               removeFromSelection(-1);
               scrollRow = Math.max(0, scrollRow - shift);
               refreshScrollRow(scrollRow, false, true);
            }
         }
      }

      if (singleRow)
      {
         int viewportRow = isRowInView(rowIndex) ? getViewportRow(rowIndex) : -1;
         int newRowsBefore = viewportRow >= 0 ? getNewRowsBefore(getViewportRow(rowIndex)) : 0;
         refresh(true);

         removeFromSelection(rowIndex);
         if (viewportRow != -1)
         {
            int newRowIndex = getCachedRowNumber(Math.max(0, viewportRow - newRowsBefore));
            if (newRowIndex != -1)
               setCurrentRow(newRowIndex, false);
         }
         else if (isNewRow(rowIndex))
         {
            UnimplementedFeature.todo("Implement proper refresh after deletion if the " +
                                         "deleted row is a new row outside of the view.");
         }
         else
         {
            refreshScrollRow(rowIndex, true, false);
         }
      }

      if (getCacheSize() < getDown())
      {
         updateMaxDataGuess(true);
      }

      if (editing && getCacheSize() > 0)
      {
         startEdit();
      }

      repaint();
   }

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   protected BrowseRow[] getRows(int startRowIndex,
                                 int count,
                                 boolean forward,
                                 RowDisplayStrategy displayStrategy)
   {
      int triggerId = getRowDisplayTriggerId();
      ScreenBuffer[] sb = getScreenBufferForTrigger(triggerId);

      currentRowDisplayStrategy = displayStrategy;
      try
      {
         return browseModel.getRows(startRowIndex,
                                    count,
                                    forward,
                                    triggerId,
                                    sb,
                                    displayStrategy);
      }
      finally
      {
         currentRowDisplayStrategy = null;
         intermediateDcData = null;
      }
   }

   /**
    * This function is used when search by key is performed in a browse. Retrieves the set of
    * rows which should be in view after search is completed. There are complex rules which
    * determine the top row of the view depending on how the <i>current</i> row is positioned
    * relatively to the current view and how the <i>found</i> row is positioned relatively to the
    * current view and to the current row.
    *
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   protected BrowseRow[] getRows(char searchChar,
                                 int searchColumn,
                                 int viewportTopRow,
                                 int viewportBottomRow)
   {
      int triggerId = getRowDisplayTriggerId();
      ScreenBuffer[] sb = getScreenBufferForTrigger(triggerId);

      currentRowDisplayStrategy = SEARCH_BY_KEY;
      try
      {
         return browseModel.getRows(searchChar,
                                    searchColumn,
                                    viewportTopRow,
                                    viewportBottomRow,
                                    triggerId,
                                    sb);
      }
      finally
      {
         currentRowDisplayStrategy = null;
         intermediateDcData = null;
      }
   }

   /**
    * Get ID of the ROW-DISPLAY trigger for this browse.
    *
    * @return ID of the ROW-DISPLAY trigger for this browse or <code>-1</code> if there is no
    *         such trigger.
    */
   protected int getRowDisplayTriggerId()
   {
      return ThinClient.getInstance().getRowDisplayTriggerId(getId().asInt());
   }

   /**
    * Get screen buffer that can be used for trigger invocation on server side.
    *
    * @param  triggerId
    *         ID of the trigger.
    *
    * @return screen buffer that can be used for trigger invocation on server side or
    *         <code>null</code> if no valid trigger ID is specified.
    */
   protected ScreenBuffer[] getScreenBufferForTrigger(int triggerId)
   {
      return triggerId != -1 ? ThinClient.getInstance().getEditableClientScreenBuffers(this) :
                               null;
   }

   /**
    * Reset the cache and redisplay.
    */
   public void queryClosed()
   {
      if (config.refreshable)
      {
         stopEdit(true, false);
         dcData = null;
         setCurrentRow(0, false, true);
      }

      if (isSelectionMultiple())
         selectedRowIndices.clear();

      rowCountAssessment = -1;
      config.maxDataReached = false;
      config.maxDataGuess = BrowseConfig.DATA_GUESS_DEFAULT_VALUE;
      config.modified = false;
      initialRow = null;
      queryClosed = true;
      notifyBrowseScrolled();
      repaint();
   }

   /**
    * Notify the browse that the underlying query has been repositioned.
    *
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   public void queryRepositioned(boolean repositionError, int targetRepositionRow)
   {
      if (initialDataSetObtained)
      {
         refreshScrollRow(false, false, true, repositionError, targetRepositionRow);
      }
      else
      {
         if (!queryClosed)  // it is REPOSITION rather than QUERY OPEN
         {
            initialRow = targetRepositionRow;
         }
      }

      queryClosed = false;
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public boolean selectNextRow()
   {
      return advanceRowSelection(true);
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public boolean selectPreviousRow()
   {
      return advanceRowSelection(false);
   }

   /**
    * Implements the client side of the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  srcIndex
    *         0-based position of the column to be moved (column ordinal).
    * @param  destIndex
    *         0-based position to which the column is moved (column ordinal).
    */
   public void moveColumn(int srcIndex, int destIndex)
   {
   }

   /**
    * Client-side part of the code which applies the given enhanced configuration to the browse.
    *
    * @param  config
    *         Enhanced browse config.
    */
   public void applyEnhancedConfigClient(EnhancedBrowseConfig config)
   {
   }

   /**
    * Returns <code>true</code> if we are in the process of saving of the edited row.
    *
    * @return <code>true</code> if we are in the process of saving of the edited row.
    */
   public boolean isSavingRow()
   {
      return savingRow;
   }

   /**
    * The method returns the legacy widget representing this widget.
    *
    * @return  A non-wirtual widget reference, or <code>null</code> if none found.
    */
   @Override
   public Widget<O> getLegacyWidget()
   {
      // TODO: this brakes the interface contract which states "If this is already a legacy
      //       widget it is returned." This method must return this, as browse is a legacy
      //       widget itself. I am not removing the method now to fix this as this could break
      //       other logic relying on this implementation.
      // If browse editing is in process then the editor widget (fill-in by default) is
      // returned, otherwise - the browse widget itself is returned.
      return getFocusedWidget();
   }

   /**
    * Clear widget state on CLEAR FRAME call for the parent frame.
    *
    * @return <code>true</code>.
    */
   @Override
   public boolean clearWidget()
   {
      super.clearWidget(); // OK to ignore returned value

      if (!config.performLayout)
      {
         refresh(false);

         // TODO if a record in view was deleted, refresh may cause desynchronization of the
         // current and selected row in single browse, which is not yet supported by the current
         // browse implementation.
      }

      return true;
   }

   /**
    * Scrolls a browse (if necessary) to bring the currently selected row into view. If the browse
    * supports multiple selections, then SCROLL-TO-CURRENT-ROW() brings the most recently selected
    * row into view.<p>
    * The position of the scrolled row is the first row in the browse viewport, unless the current
    * row is already visible. In this case, the current row remains in the original position.
    *
    * @throws ErrorConditionException
    *         if a numbered error occurs.
    */
   public void scrollToCurrentRow()
   {
      if (getCacheSize() == 0)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(2108, "");
      }
      else if (getSelectedRowsCount() == 0)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(382, "");
      }

      int[] selectedRows = getSelectedRowIndexes();
      int selectedRow = selectedRows[selectedRows.length - 1];
      if (!isRowInView(selectedRow))
      {
         scrollToRow(selectedRow, false, true, false, false);
      }
   }

   /**
    * Scrolls a browse (if necessary) to bring the specified selected row into view.
    *
    * @param  selectedRowIndex
    *         Selected row within the browse. A numbered list of selected rows, starting at 1 is
    *         maintained. When the <code>SCROLL-TO-SELECTED-ROW(n)</code> method is encountered,
    *         the list is iterated to find the <i>n</i>th selected row.
    *
    * @throws ErrorConditionException
    *         if a numbered error occurs.
    */
   public void scrollToSelectedRow(int selectedRowIndex)
   {
      if (getCacheSize() == 0)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(2108, "");
      }
      else if (selectedRowIndex < 1 || selectedRowIndex > getSelectedRowsCount())
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         throw new ErrorConditionException(385, "");
      }

      int[] selectedRows = getSelectedRowIndexes();
      int selectedRow = selectedRows[selectedRowIndex - 1];
      if (!isRowInView(selectedRow))
      {
         scrollToRow(selectedRow, false, true, false, false);
      }
   }

   /**
    * Get the number of new rows in the viewport before the given base viewport row.
    *
    * @param  baseViewportRow
    *         0-based index of the base row in the viewport.
    *
    * @return see above.
    */
   protected int getNewRowsBefore(int baseViewportRow)
   {
      int count = 0;
      int maxRow = Math.min(baseViewportRow, getCacheSize());
      for (int i = 0; i < maxRow; i++)
      {
         if (isNewRow(getCachedRowNumber(i)))
            count++;
      }
      return count;
   }

   /**
    * Worker for {@link #selectNextRow()} and {@link #selectPreviousRow()}.
    *
    * @param  forward
    *         If <code>true</code> then it corresponds {@link #selectNextRow()}. If
    *         <code>false</code> - {@link #selectPreviousRow()}.
    *
    * @return <code>true</code> on success.
    */
   protected boolean advanceRowSelection(boolean forward)
   {
      boolean editing = isEditing();

      int selectedRow = getLastAvailableSelectedRow(true);
      if (selectedRow == -1)
      {
         // This is being raised to transfer control to the server where the proper
         // ErrorManager processing will occur.
         // Error: no *available* rows selected.
         throw new ErrorConditionException(382, "");
      }

      BrowseRow[] followingRow = null;
      if (!editing)
      {
         // no more *available* rows
         followingRow = forward ?
               getRows(selectedRow + 1, 1, true, NO_DISPLAY) :
               getRows(selectedRow - 1, 1, false, NO_DISPLAY);
         if (followingRow.length == 0)
            return false;
      }

      boolean inView = isRowInView(selectedRow);

      // special case - read-only multi-selection, last available selected row is in view
      boolean specialCase = inView && isSelectionMultiple() && !editing;

      if (editing || inView)
      {
         int savedViewportRow = getViewportRow();
         boolean refresh      = false;
         if (!editing && isViewportEdge(selectedRow, forward))
         {
            refresh = true;
         }
         else
         {
            boolean res = advanceCurrRowOrSelection(specialCase, forward, editing, selectedRow);
            if (!editing && res)
            {
               // go down until we will find an available row
               while (specialCase ? !isSelectedRowAvailable(0, false) :
                                     getCurrentRow() != browseModel.getCurrentRow())
               {
                  selectedRow = getSelectedRowIndex(0, true, false);
                  if (isViewportEdge(selectedRow, forward))
                  {
                     // we have reached the end of view but didn't find an available row
                     refresh = true;
                     break;
                  }

                  if (!advanceCurrRowOrSelection(specialCase, forward, false, selectedRow))
                     break;
               }
            }
         }

         if (refresh)
         {
            if (!browseModel.isRowAvailable(getCurrentRow()))
            {
               // do not preserve viewport row if current row is not available
               savedViewportRow = 0;
            }

            if (forward)
            {
               // This is triggered in the following case:
               //    ...
               //  > current (available) row
               //    0 or more unavailable rows
               //    --- end of view ---
               scrollToRow(getCachedRowNumber(1));
               selectedRow = getCachedRowNumber(getCacheSize() - 1);
            }
            else
            {
               // This is triggered in the following case:
               //    ...
               //    --- begin of view ---
               //    0 or more unavailable rows
               //  > current (available) row
               int rowIndex = followingRow[0].rowIndex;
               scrollToRow(rowIndex);
               selectedRow = getCachedRowNumber(0);
            }

            if (specialCase)
            {
               selectedRowIndices.clear();
               selectedRowIndices.add(new SelectedRow(selectedRow));
            }
            else
            {
               setCurrentRow(selectedRow);
            }
         }

         if (specialCase)
         {
            // preserve viewport row
            setCurrentRow(getCachedRowNumber(Math.min(getCacheSize() - 1, savedViewportRow)));
         }
      }
      else
      {
         // multi-selection read-only browse, last available selected row is out of the view
         int rowIndex = followingRow[0].rowIndex;
         if (!isRowInView(rowIndex))
         {
            int viewportRow = getViewportRow();
            scrollToRow(rowIndex);
            // preserve viewport row
            setCurrentRow(getCachedRowNumber(Math.min(getCacheSize() - 1, viewportRow)));
         }
         
         if (isSelectionMultiple())
         {
            selectedRowIndices.clear();
            selectedRowIndices.add(new SelectedRow(rowIndex));
         }
      }

      repaint();
      return true;
   }

   /**
    * Determines if the specified row is the first or last row in the view.
    *
    * @param  rowIndex
    *         0-based row index in the browse model.
    * @param  forward
    *         <code>true</code> to check if it is the last row, <code>false</code> to check if it
    *         is the first row.
    *
    * @return <code>true</code> if the specified row is the first (if <code>forward</code> is
    *         <code>true</code>) or last (if <code>forward</code> is <code>false</code>) row in
    *         the view. If the page is not full then the last row on the page is NOT considered to
    *         be the last row in the view.
    */
   protected boolean isViewportEdge(int rowIndex, boolean forward)
   {
      return getViewportRow(rowIndex) == (forward ? getDown() - 1 : 0);
   }

   /**
    * Works in two modes:
    * <ul>
    *    <li>Advance current row - make the row before or after the current row the new
    *    current row.</li>
    *    <li>Advance selection - deselected all rows and make the row before or after the
    *    specified row selected.</li>
    * </ul>
    *
    * @param  advanceSelection
    *         <code>true</code> for selection mode. <code>false</code> for current row mode.
    * @param  forward
    *         <code>true</code> to advance to the next row. <code>false</code> to advance to the
    *         previous row.
    * @param  handleEditing
    *         <code>true</code> to handle start/stop editing. For current row mode only.
    * @param  rowIndex
    *         0-based index of the row which should be used as a base for selection advancing (i.e.
    *         the row before or after it will be selected). For selection mode only.
    *
    * @return <code>true</code> if current/selected row has changed.
    */
   protected boolean advanceCurrRowOrSelection(boolean advanceSelection,
                                               boolean forward,
                                               boolean handleEditing,
                                               int rowIndex)
   {
      if (forward)
      {
         if (advanceSelection)
            return goDownWithSelection(rowIndex);
         else
            return goDown(handleEditing, false);
      }
      else
      {
         if (advanceSelection)
            return goUpWithSelection(rowIndex);
         else
            return goUp(handleEditing, false);
      }
   }

   /**
    * Get last (in the order of selection) "available" (not deleted on the moment of selection)
    * selected row.
    *
    * @param   includeNewRows
    *          Include new rows in the search (consider them "available").
    *
    * @return  0-based index of the selected row in the browse model or <code>-1</code> if there
    *          are no "available" selected rows.
    */
   protected int getLastAvailableSelectedRow(boolean includeNewRows)
   {
      if (!isSelectionMultiple())
      {
         SelectedRow row = getCurrentRowAsSelected();
         if (isNewRow(row.getRowIndex()))
         {
            return includeNewRows ? row.getRowIndex() : -1;
         }
         else
         {
            return row.isRowAvailable() ? row.getRowIndex() : -1;
         }
      }
      else
      {
         int selectedRowsCount = selectedRowIndices.size();
         if (selectedRowsCount == 0)
            return -1;

         SelectedRow[] rows = selectedRowIndices.toArray(new SelectedRow[selectedRowsCount]);

         // find last available selected row
         for (int i = rows.length - 1; i >= 0; i--)
         {
            SelectedRow row = rows[i];
            int index = row.getRowIndex();
            if (isNewRow(index))
            {
               if (includeNewRows)
                  return index;
               // else continue
            }
            else if (row.isRowAvailable())
            {
               return index;
            }
         }

         return -1;
      }
   }

   /**
    * Determines if the specified row is in the viewport.
    *
    * @param  rowIndex
    *         Index of the target row in the browse model.
    *
    * @return <code>true</code> if the specified row is in the viewport.
    */
   protected boolean isRowInView(int rowIndex)
   {
      int cacheSize = getCacheSize();
      for (int i = 0; i < cacheSize; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
         {
            return true;
         }
      }
      return false;
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goUp(boolean handleEditing)
   {
      return goUp(handleEditing, true);
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    * @param  fireValueChanged
    *         If <code>true</code> then VALUE-CHANGED event is fired if current row has changed.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goUp(boolean handleEditing, boolean fireValueChanged)
   {
      boolean editing = handleEditing && isEditing();

      if (editing && !leaveRow())
         return false;

      if (isFirstBrowseRow())
      {
         if (editing)
            startEdit();
         return false;
      }

      int viewportRow = getViewportRow();
      int newRow;
      if (viewportRow > 0)
      {
         newRow = getCachedRowNumber(viewportRow - 1);
      }
      else
      {
         if (!incrementalScroll(-1))
         {
            // no more rows
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(0);
      }

      setCurrentRow(newRow, true, true);

      if (editing)
      {
         startEdit();
         
         if (fireValueChanged)
         {
            invokeRowEntryTrigger(true);
            fireValueChanged();
         }
      }
      else if (!isSelectionMultiple() && fireValueChanged)
      {
         invokeRowEntryTrigger(true);
         fireValueChanged();
      }
      
      return true;
   }

   /**
    * Scroll one row up and adjust current row.
    *
    * @param  handleEditing
    *         If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goDown(boolean handleEditing)
   {
      return goDown(handleEditing, true);
   }

   /**
    * Scroll one row down and adjust current row.
    *
    * @param handleEditing
    *        If <code>true</code> then do {@link #startEdit}/{@link #stopEdit} calls.
    * @param fireValueChanged
    *        If <code>true</code> then VALUE-CHANGED event is fired if current row has changed.
    *
    * @return <code>true</code> if the current row has changed.
    */
   protected boolean goDown(boolean handleEditing, boolean fireValueChanged)
   {
      boolean editing = handleEditing && isEditing();

      if (editing && !leaveRow())
         return false;

      int viewportRow = getViewportRow();
      int newRow;
      int lastRow = getDown() - 1;
      if (viewportRow < lastRow)
      {
         if (viewportRow == getCacheSize() - 1)
         {
            // we are on the last cached row
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(viewportRow + 1);
      }
      else
      {
         if (!incrementalScroll(1))
         {
            // no more rows
            if (editing)
               startEdit();
            return false;
         }

         newRow = getCachedRowNumber(lastRow);
      }

      setCurrentRow(newRow, true, true);

      if (editing)
      {
         startEdit();
         
         if (fireValueChanged)
         {
            invokeRowEntryTrigger(true);
            fireValueChanged();
         }
      }
      else if (!isSelectionMultiple() && fireValueChanged)
      {
         invokeRowEntryTrigger(true);
         fireValueChanged();
      }
      
      return true;
   }

   /**
    * Deselect all rows and select the row before the given one.
    *
    * @param  rowIndex
    *         0-based index of the base row in the browse mode, the previous row before it will be
    *         selected.
    *
    * @return <code>true</code> if there are rows before the given one.
    */
   protected boolean goUpWithSelection(int rowIndex)
   {
      int viewportRow = getViewportRow(rowIndex);
      int newRow;
      boolean rowAvail;
      if (viewportRow > 0)
      {
         newRow = getCachedRowNumber(viewportRow - 1);
         rowAvail = browseModel.isRowAvailable(newRow);
      }
      else
      {
         if (!incrementalScroll(-1))
         {
            // no more rows
            return false;
         }

         newRow = getCachedRowNumber(0);
         rowAvail = true;
      }

      selectedRowIndices.clear();
      selectedRowIndices.add(new SelectedRow(newRow, rowAvail));
      return true;
   }

   /**
    * Deselect all rows and select the row after the given one.
    *
    * @param  rowIndex
    *         0-based index of the base row in the browse mode, the next row after it will be
    *         selected.
    *
    * @return <code>true</code> if there are rows after the given one.
    */
   protected boolean goDownWithSelection(int rowIndex)
   {
      int viewportRow = getViewportRow(rowIndex);
      int newRow;
      boolean rowAvail;
      int lastRow = getDown() - 1;
      if (viewportRow < lastRow)
      {
         if (viewportRow == getCacheSize() - 1)
         {
            // we are on the last cached row
            return false;
         }

         newRow = getCachedRowNumber(viewportRow + 1);
         rowAvail = browseModel.isRowAvailable(newRow);
      }
      else
      {
         if (!incrementalScroll(1))
         {
            // no more rows
            return false;
         }

         newRow = getCachedRowNumber(lastRow);
         rowAvail = true;
      }

      selectedRowIndices.clear();
      selectedRowIndices.add(new SelectedRow(newRow, rowAvail));
      return true;
   }

   /**
    * 4GL has two scrolling modes: "one-page" scrolling mode which is used if the total number
    * of rows is less or equal page size and the "normal" mode. In "one-page" mode scroll row
    * cannot be changed, the first row is the permanent scroll row.
    * 4GL can switch between these modes,
    * TODO switching can be implemented after gap closing mechanism for deleted rows is implemented
    *
    * @return <code>true</code> if "one-page" scrolling mode is active.
    */
   protected boolean isOnePageScrollingMode()
   {
      int down = getDown();
      return (rowCountAssessment <= down) ||
             (getCachedRowNumber(0) == 0 && getCacheSize() < down);
   }

   /**
    * Determines if browse refreshed on browse widget enabling.
    *
    * @return <code>true</code>
    */
   protected boolean isRefreshedOnEnable()
   {
      return true;
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow)
   {
      return scrollToRow(scrollRow, false, true);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow, boolean setAsCurrentRow)
   {
      return scrollToRow(scrollRow, setAsCurrentRow, false);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow, boolean setAsCurrentRow, boolean leaveOnNoRows)
   {
      return scrollToRow(scrollRow, setAsCurrentRow, leaveOnNoRows, false);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    * @param  ignoreOnePageScrollingMode
    *         Perform scroll as if the browse is in multiple-pages scrolling mode.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow,
                                 boolean setAsCurrentRow,
                                 boolean leaveOnNoRows,
                                 boolean ignoreOnePageScrollingMode)
   {
      return scrollToRow(scrollRow,
                         setAsCurrentRow,
                         leaveOnNoRows,
                         ignoreOnePageScrollingMode,
                         false);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    * @param  ignoreOnePageScrollingMode
    *         Perform scroll as if the browse is in multiple-pages scrolling mode.
    * @param  honorRepositionedRow
    *         Honor repositioned row set by SET-REPOSITIONED-ROW.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow,
                                 boolean setAsCurrentRow,
                                 boolean leaveOnNoRows,
                                 boolean ignoreOnePageScrollingMode,
                                 boolean honorRepositionedRow)
   {
      return scrollToRow(scrollRow,
                         setAsCurrentRow,
                         leaveOnNoRows,
                         ignoreOnePageScrollingMode,
                         honorRepositionedRow,
                         true);
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    * @param  ignoreOnePageScrollingMode
    *         Perform scroll as if the browse is in multiple-pages scrolling mode.
    * @param  honorRepositionedRow
    *         Honor repositioned row set by SET-REPOSITIONED-ROW.
    * @param  allowStartEdit
    *         Allow to start editing for the first scroll.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow,
                                 boolean setAsCurrentRow,
                                 boolean leaveOnNoRows,
                                 boolean ignoreOnePageScrollingMode,
                                 boolean honorRepositionedRow,
                                 boolean allowStartEdit)
   {
      return scrollToRow(scrollRow,
                         setAsCurrentRow,
                         leaveOnNoRows,
                         ignoreOnePageScrollingMode,
                         honorRepositionedRow,
                         allowStartEdit,
                         -1,
                         isAggressiveCurrentRowSelection());
   }

   /**
    * Scroll to the specified row (all records are refreshed).
    *
    * @param  scrollRow
    *         The new scroll row (i.e. the first row in the viewport).
    * @param  setAsCurrentRow
    *         Set the new scroll row as the current row.
    * @param  leaveOnNoRows
    *         Return if there are no rows in the result set (otherwise cache will be cleared).
    * @param  ignoreOnePageScrollingMode
    *         Perform scroll as if the browse is in multiple-pages scrolling mode.
    * @param  honorRepositionedRow
    *         Honor repositioned row set by SET-REPOSITIONED-ROW.
    * @param  allowStartEdit
    *         Allow to start editing for the first scroll.
    * @param  oneTimeRepositionedRow
    *         0-based repositioned row used for this scroll action. <code>-1</code> if not
    *         specified.
    * @param  updateSelection
    *         <code>true</code> if selection should be adjusted.
    *
    * @return <code>true</code> if we got any rows from the server.
    */
   protected boolean scrollToRow(int scrollRow,
                                 boolean setAsCurrentRow,
                                 boolean leaveOnNoRows,
                                 boolean ignoreOnePageScrollingMode,
                                 boolean honorRepositionedRow,
                                 boolean allowStartEdit,
                                 int oneTimeRepositionedRow,
                                 boolean updateSelection)
   {
      if (browseModel == null)
      {
         dcData = null;
         setCurrentRow(0, updateSelection, true);
         notifyBrowseScrolled();
         repaint();
         return false;
      }

      boolean firstScroll = rowCountAssessment < 0;
      boolean onePageScrolling = !ignoreOnePageScrollingMode && isOnePageScrollingMode();
      int down = getDown();

      BrowseRow[] serverData;
      int repositionedRow = oneTimeRepositionedRow >= 0 ? oneTimeRepositionedRow :
                                                          getRepositionedRow();
      boolean repositionedRowMode = honorRepositionedRow &&
                                    repositionedRow > 0 &&
                                    !onePageScrolling &&
                                    !firstScroll;
      if (repositionedRowMode)
      {
         BrowseRow[] upperPart = getRows(scrollRow,
                                         repositionedRow + 1,
                                         false,
                                         NO_DISPLAY);
         if (upperPart == null)
         {
            return false;
         }

         int lowerPartSize = down - upperPart.length;
         BrowseRow[] lowerPart;
         if (lowerPartSize == 0)
            lowerPart = new BrowseRow[0];
         else
            lowerPart = getRows(scrollRow + 1,
                                lowerPartSize,
                                true,
                                NO_DISPLAY);

         int totalSize = upperPart.length + lowerPart.length;
         serverData = new BrowseRow[totalSize];
         System.arraycopy(upperPart, 0, serverData, 0, upperPart.length);
         System.arraycopy(lowerPart, 0, serverData, upperPart.length, lowerPart.length);

         if (totalSize < down || serverData[totalSize - 1].rowIndex < scrollRow)
         {
            updateMaxDataGuess(true);
         }

         if (isEditing() && editingExistingRowBefore != null)
         {
            serverData = insertEditingRow(serverData,
                                          true,
                                          down,
                                          serverData.length > 0 ?
                                             serverData[0].rowIndex - 1 :
                                             scrollRow - 1,
                                          false);
         }
      }
      else
      {
         serverData = getRows(onePageScrolling ? 0 : scrollRow,
                              firstScroll ? down + 1 : down,
                              true,
                              firstScroll ? FIRST_SCROLL : NO_DISPLAY);
         if (serverData == null)
         {
            if (firstScroll)
            {
               initialDataSetObtained = true;
            }

            return false;
         }

         if (!firstScroll && isEditing() && editingExistingRowBefore != null)
         {
            serverData = insertEditingRow(serverData,
                                          true,
                                          down,
                                          serverData.length > 0 ?
                                             serverData[0].rowIndex - 1 :
                                             scrollRow - 1,
                                          true);
         }
      }

      if (serverData.length > 0)
      {
         if (firstScroll)
         {
            int len = Math.min(serverData.length, down);
            dcData = new BrowseRow[len];
            System.arraycopy(serverData, 0, dcData, 0, dcData.length);
            rowCountAssessment = serverData.length;
            initialDataSetObtained = true;
         }
         else
         {
            dcData = serverData;
         }

         if (setAsCurrentRow)
         {
            if (repositionedRowMode)
            {
               scrollRow = getCachedRowNumber(Math.min(Math.min(scrollRow, repositionedRow),
                                                       getCacheSize() - 1));
            }

            setCurrentRow(scrollRow, updateSelection, true);
         }
      }
      else
      {
         if (firstScroll)
         {
            rowCountAssessment = 0;
            initialDataSetObtained = true;
            dcData = null;
            setCurrentRow(0, updateSelection, true);
         }
         else
         {
            // we are beyond the end of the result set
            serverData = getRows(scrollRow, 1, false, NO_DISPLAY);

            if (serverData.length == 0)
            {
               if (leaveOnNoRows)
                  return false;

               // the result set is empty
               dcData = null;
               setCurrentRow(0, updateSelection, true);
            }
            else
            {
               // we have scrolled to the last row
               dcData = serverData;
               if (setAsCurrentRow)
               {
                  setCurrentRow(getCachedRowNumber(0), updateSelection, true);
               }
            }
         }
      }

      if (firstScroll && (rowCountAssessment <= down))
      {
         updateMaxDataGuess(true);
      }

      notifyBrowseScrolled();
      repaint();

      if (allowStartEdit &&
          firstScroll &&
          isEditPossible() &&
          !isEditing() &&
          getCacheSize() > 0)
      {
         startEdit();
      }

      return true;
   }

   /**
    * Update height of the labels area.
    */
   protected void updateLabelsHeight()
   {
      BrowseColumnConfig[] cols = getVisibleColumns();
      for (BrowseColumnConfig col : cols)
      {
         ilLabelsHeight = Math.max(ilLabelsHeight, getLabelVar(col).length);
      }
   }

   /**
    * Get next unique row number for a new row.
    *
    * @return see above.
    */
   protected int nextNewRowNumber()
   {
      return newRowNumber--;
   }

   /**
    * Determines if the current row is the first browse row, i.e. we cannot scroll higher this
    * row.
    *
    * @return  <code>true</code> if we cannot scroll higher current row.
    */
   protected boolean isFirstBrowseRow()
   {
      if (getViewportRow() == 0)
      {
         int topExistingRow = getFirstExistingRow();
         if (topExistingRow == 0)
         {
            return true;
         }

         if (topExistingRow == -1)
         {
            // no rows or only new rows
            return editingExistingRowBefore == null;
         }
      }

      return false;
   }

   /**
    * Get first existing (not new) row in the view.
    *
    * @return 0-based index (in the browse model) of the first existing row in the view or
    *          <code>-1</code> if there are no existing rows.
    */
   protected int getFirstExistingRow()
   {
      return getEdgeExistingRow(true);
   }

   /**
    * Get last existing (not new) row in the view.
    *
    * @return 0-based index (in the browse model) of the last existing row in the view or
    *          <code>-1</code> if there are no existing rows.
    */
   protected int getLastExistingRow()
   {
      return getEdgeExistingRow(false);
   }

   /**
    * Get first or last existing (not new) row in the view.
    *
    * @param   first
    *          <code>true</code> to look for the first row. <code>false</code> to look for the last
    *          row.
    *
    * @return  0-based index (in the browse model) of the first existing row in the view or
    *          <code>-1</code> if there are no existing rows.
    */
   protected int getEdgeExistingRow(boolean first)
   {
      if (dcData == null || dcData.length == 0)
         return -1;

      for (int i = first ? 0 :
            dcData.length - 1;
           first ? i < dcData.length :
                 i >= 0;
           i = first ? i + 1 :
                 i - 1)
      {
         int rowNumber = getCachedRowNumber(i);
         if (!isNewRow(rowNumber))
            return rowNumber;
      }

      return -1;
   }

   /**
    * Notify browse that the parent frame became visible. That means that the browse should be
    * refreshed and displayed.
    */
   public void parentFrameViewed()
   {
      if (config.performLayout)
      {
         doLayout();
      }

      refresh(false);
      _setVisible(true);
      _setHidden(false);
      repaint();
   }

   /**
    * Determine if in-browse editing is possible.
    *
    * @return <code>true</code> if there is at least one editable column.
    */
   public boolean isEditPossible()
   {
      return config.editPossible;
   }

   /**
    * Return information about invisible columns.
    *
    * @return  Meta information about invisible columns.
    */
   public BrowseColumnConfig[] getInvisibleColumns()
   {
      ArrayList<BrowseColumnConfig> res = new ArrayList<>();
      for (BrowseColumnConfig column : getColumns())
      {
         if (!column.visible)
         {
            res.add(column);
         }
      }

      return res.toArray(new BrowseColumnConfig[res.size()]);
   }

   /**
    * Calculate row index for internal cache.
    * 
    * @param   rowIndex
    *          Index in the model.
    *          
    * @return  Calculated index in the local cache.
    */
   protected int dcGetRowIndex(int rowIndex)
   {
      int size = getCacheSize();
      for (int i = 0; i < size; i++)
      {
         if (getCachedRowNumber(i) == rowIndex)
            return i;
      }

      StringBuilder sb = new StringBuilder();
      for (int i = 0; i < size; i++)
      {
         sb.append(getCachedRowNumber(i));
         if (i != size - 1)
            sb.append(", ");
      }

      throw new IllegalArgumentException
            ("Invalid cache read request: " + rowIndex +
             ". Cache contains only [" + sb.toString() + "]");
   }
   
   /**
    * Invoke the LEAVE trigger for the currently editable column.
    * 
    * @param    nextFocus
    *           The next focusable widget.
    *           
    * @return   <code>true</code> if the LEAVE trigger allows the leave.
    */
   public boolean invokeLeaveCellTrigger(WidgetId nextFocus)
   {
      if (!isEditing())
      {
         return true;
      }

      ThinClient tc = ThinClient.getInstance();
      Widget src = tc.getWidget(getVisibleColumnCfg(currColumn).id);
      Widget next = nextFocus == null ? null : tc.getWidget(nextFocus);
      
      // allow LEAVE triggers to fire on server side before removing the editor. Note that the
      // trigger is set for browse column not the actual editor
      TriggerResult tr = tc.invokeTriggers(src, next, Keyboard.SE_LEAVE, false);
      return !tr.executed || !tr.consume;
   }
   
   /**
    * Invoke ROW-ENTRY trigger.
    *
    * @param   keepColumn
    *          Flag indicating if we keep the {@link #currColumn} or compute the new editable one,
    *          starting with the first one.
    *
    * @return {@code true} if there is no ROW-ENTRY trigger or it has returned normally.
    *         {@code false} if NO-APPLY has been returned by the trigger.
    */
   public boolean invokeRowEntryTrigger(boolean keepColumn)
   {
      if (!isEditPossible())
      {
         return true;
      }

      int code = Keyboard.SE_ROW_ENTRY;
      Widget focus = UiUtils.getCurrentFocus();
      ThinClient.TriggerResult res = ThinClient.getInstance().invokeTriggers(this, focus, 
                                                                             code, false);
      if (!res.executed || !res.consume)
      {
         if (!keepColumn)
         {
            int nextColumn = getFirstEditableColumn(0);
         
            currColumn = nextColumn >= 0 ? nextColumn : 0;
         }
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Invoke ROW-LEAVE trigger.
    *
    * @param    nextId
    *           The ID of the next focusable widget.
    *
    * @return   <code>true</code> if there is no ROW-LEAVE trigger or it has returned normally.
    *           <code>false</code> if NO-APPLY has been returned by the trigger.
    */
   protected boolean invokeRowLeaveTrigger(WidgetId nextId)
   {
      if (!isEditPossible())
      {
         return true;
      }
      ThinClient tc = ThinClient.getInstance();
      
      Widget<?> next = null;
      if (nextId == null)
      {
         int nextColumn = getFirstEditableColumn(0);
         nextId = nextColumn < 0 ? null : getVisibleColumnCfg(nextColumn).id;
      }

      next = nextId == null ? null : tc.getWidget(nextId);
      
      int code = Keyboard.SE_ROW_LEAVE;
      ThinClient.TriggerResult res = tc.invokeTriggers(this, next, code, false);
      return !res.executed || !res.consume;
   }
   
   /**
    * Release current row (finish editing and commit). If row cannot be released (e.g. validation
    * fails), scroll to editable row and restart editing, if necessary.
    *
    * @return <code>true</code> if row was successfully released.
    */
   protected boolean leaveRow()
   {
      if (!releaseCurrentRow(true))
      {
         scrollToEditableRow();
         startEdit();
         return false;
      }

      return true;
   }

   /**
    * Scroll to the editable row if it is out of the view.
    */
   protected void scrollToEditableRow()
   {
   }
   
   /**
    * Get the first editable column in the browse, starting from the specified one.
    * 
    * @param    start
    *           The start index.
    *
    * @return   0-based index of the first editable column among the <code>visible</code> columns.
    */
   private int getFirstEditableColumn(int start)
   {
      if (!isEditPossible())
      {
         return -1;
      }
      
      BrowseColumnConfig[] cols = getVisibleColumns();
      int nextColumn;
      for (nextColumn = start; nextColumn < cols.length; ++nextColumn)
      {
         if (!cols[nextColumn].readOnly)
            break;
      }
      
      return nextColumn < cols.length ? nextColumn : -1;
   }

   /**
    * Return information about visible columns.
    * 
    * @return  Meta information about visible columns.
    */
   protected BrowseColumnConfig[] getVisibleColumns()
   {
      if (visibleColumnsCache != null)
      {
         return visibleColumnsCache;
      }

      populateColumnsCache();
      return visibleColumnsCache;
   }

   /**
    * Return information about all columns regardless of their visibility.
    *
    * @return  Meta information about all columns.
    */
   protected BrowseColumnConfig[] getColumns()
   {
      if (columnsCache != null)
      {
         return columnsCache;
      }

      populateColumnsCache();
      return columnsCache;
   }

   /**
    * Reset cached information about columns.
    */
   protected void resetColumnsCache()
   {
      columnsCache = null;
      visibleColumnsCache = null;
      ordinalToVisibleMap = null;
   }

   /**
    * Update MAX-DATA-GUESS value which holds estimated number of rows.
    *
    * @param lastReached
    *        <code>true</code> if the last row in the browse model has been reached.
    */
   protected void updateMaxDataGuess(boolean lastReached)
   {
      int lastRow = getLastExistingRow() + 1;
      if (lastReached || config.maxDataGuess < lastRow)
      {
         config.maxDataGuess = lastRow;
      }

      if (lastReached)
      {
         config.maxDataReached = true;
      }
   }

   /**
    * Notify that the browse has been scrolled (i.e. the top row has changed).
    */
   protected void notifyBrowseScrolled()
   {
      updateMaxDataGuess(false);
   }

   /**
    * Get 0-based index of the column among the visible columns.
    *
    * @param columnOrdinal
    *        Column ordinal, i.e. 0-based index of the column among <i>all</i> columns
    *
    * @return 0-based index of the column among the visible columns or <code>null</code> if the
    *         column is not visible.
    */
   protected Integer ordinalToVisible(int columnOrdinal)
   {
      return ordinalToVisibleMap.get(columnOrdinal);
   }

   /**
    * Get 0-based index of the column among <i>all</i> columns (i.e. column ordinal).
    *
    * @param  visibleColumnIndex
    *         0-based index of the column among the visible columns.
    *
    * @return 0-based index of the column among <i>all</i> columns (i.e. column ordinal).
    */
   protected int visibleToOrdinal(int visibleColumnIndex)
   {
      return getVisibleColumns()[visibleColumnIndex].ordinal;
   }

   /**
    * Populate cache of columns using updated {@link BrowseConfig#columnIds}.
    */
   protected void populateColumnsCache()
   {
      int colNo = config.columnIds.length;
      List<BrowseColumnConfig> columns = new ArrayList<>();
      List<BrowseColumnConfig> visibleColumns = new ArrayList<>();
      ordinalToVisibleMap = new HashMap<>();
      ConfigManager cfgMgr = ConfigManager.getInstance();

      for (int i = 0; i < colNo; i++)
      {
         int colId = config.columnIds[i];
         BrowseColumnConfig colCfg = cfgMgr.getActiveConfig(new WidgetId(colId));

         if (colCfg != null)
         {
            if (colCfg.visible)
            {
               visibleColumns.add(colCfg);
               int visIndex = visibleColumns.size() - 1;
               ordinalToVisibleMap.put(colCfg.ordinal, visIndex);
            }

            columns.add(colCfg);
         }

         // Configuration may be null if it is a dynamic column, in this case columnsCache
         // is updated in notifyColumnsUpdated.
      }
      columnsCache = columns.toArray(new BrowseColumnConfig[columns.size()]);
      visibleColumnsCache = visibleColumns.toArray(new BrowseColumnConfig[visibleColumns.size()]);
   }

   /**
    * Container which holds row index and row availability (i.e. if it could be fetched) at the
    * moment of selection.
    */
   protected static class SelectedRow
   {
      /** Row index. */
      private final int rowIndex;

      /** Row availability. */
      private final boolean rowAvailable;

      /**
       * Creates a new selected row which is available.
       *
       * @param rowIndex
       *        Row index.
       */
      public SelectedRow(int rowIndex)
      {
         this.rowIndex = rowIndex;
         this.rowAvailable = true;
      }

      /**
       * Creates a new selected row.
       *
       * @param rowIndex
       *        Row index.
       * @param rowAvailable
       *        Row availability.
       */
      public SelectedRow(int rowIndex, boolean rowAvailable)
      {
         this.rowIndex = rowIndex;
         this.rowAvailable = rowAvailable;
      }

      /**
       * Get row index.
       *
       * @return row index.
       */
      public int getRowIndex()
      {
         return rowIndex;
      }

      /**
       * Get row availability.
       *
       * @return <code>true</code> if the row is available (i.e. if it could be fetched) at the
       *         moment of selection.
       */
      public boolean isRowAvailable()
      {
         return rowAvailable;
      }

      /**
       * Return a hash code for the instance.
       *
       * @return  Hash code for the instance.
       */
      @Override
      public int hashCode()
      {
         return rowIndex;
      }

      /**
       * Compares two objects for equality. Two selected rows are considered equal if they have
       * the same row index (regardless of their availability).
       *
       * @param   obj
       *          Object to compare with.
       *
       * @return  <code>true</code> if instances are equal and <code>false</code> otherwise.
       */
      @Override
      public boolean equals(Object obj)
      {
         if (this == obj)
            return true;

         if (obj == null || getClass() != obj.getClass())
            return false;

         SelectedRow other = (SelectedRow) obj;
         return other.rowIndex == this.rowIndex;
      }
   }
}
