/*
** Module   : Editor.java
** Abstract : client side common code implementation of the Editor component
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20060428   @26083 First stage of the implementation of the 
**                           EDITOR widget. Initial version is capable 
**                           to properly calculate size and display 
**                           itself taking into account assigned text, 
**                           size (assigned via SIZE phrase and via 
**                           INNER-CHARS/INNER-LINES), box flag and 
**                           scroll-bars flags. No real editing 
**                           operations are implemented.
** 002 SIY 20060515   @26151 Added getColumnLabelText().
** 003 NVS 20060703   @27801 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 004 SIY 20060810   @28557 Fixed displaying of enabled but not assigned
**                           widget.
** 005 SIY 20060816   @28644 Second stage of implementation. All editing 
**                           functions are implemented.
** 006 SIY 20060824   @28849 Removed getP2JType() method.
** 007 GES 20061110   @31089 Fixed drawing during editing when the widget
**                           starts in the uninitialized state. There
**                           needed to be a sync request at the end of
**                           the key processing (when we re-draw) too
**                           because wait-for loops won't force a sync
**                           otherwise.
** 008 NVS 20061204   @31556 Switched to using a private widgetFocused
**                           flag instead of widgetFocused
** 009 SIY 20061204   @31545 Fixed handling of backspace.
** 010 NVS 20070426   @33251 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 011 SIY 20070807   @34811 Added isColumnFixed() method.
** 012 SVL 20071221   @36530 Label is shown/hidden together with the
**                           widget.
** 013 EVL 20071224   @36559 Synchronization of the insert mode indicator
**                           with the state of the bottom-right insert mode
**                           area painting when the editor is gaining the
**                           input focus.
** 014 SVL 20080125   @36904 An empty value is set inside the constructor
**                           to prevent NPE upon initialisation.
** 015 EVL 20080303   @37231 Adding code to repaint insert mode indicator
**                           and to turn the insert mode off when the focus
**                           is leaving the editor.
** 016 NVS 20080320   @37616 Fixed drawing of editors in focus when
**                           they are disabled.
** 017 NVS 20080401   @37774 Added isFocusable() method implementation.
** 018 GES 20080421   @38060 I18N charset support for key events.
** 019 SIY 20080711   @39100 Fixed activation behavior to match 4GL one.
** 020 NVS 20080812   @39368 Added fullHeight() method implementation. It
**                           calculates the number of lines of text the
**                           editor is holding.
** 021 CA  20090423   @41899 Added drawCaretUnconditionally(), which draws the
**                           editor caret even if the editor is not in focus.
**                           This is needed to fix special read-only editor
**                           behavior, related to caret.
** 022 SVL 20090424   @41923 Fixed insert indicator state for read-only
**                           mode.
** 023 GES 20090424   @41962 Import change.
** 024 SIY 20090428   @42026 Fixed word wrapping edge case, when word ends 
**                           exactly on the widget boundary.
** 025 SIY 20100619          Switching to own implementation of character UI.
** 026 SIY 20100804          Changes in imports, some refactorings.
** 027 SIY 20100822          Refactored AttributeContainer support.
** 028 SIY 20100915          Renamed base class.
** 029 SIY 20101122          Switching from KeyEvent to KeyInput.
** 030 SIY 20110112          Changes related to new color handling scheme.
** 031 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 032 SIY 20111202          Refactored and moved from CHUI package.
** 033 GES 20140627          Changed color attribute method names to match naming standards.
** 034 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 035 HC  20140911          GUI coordinate support - data types refactoring.
** 036 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 037 HC  20141021          Implemented window move/position support.
** 038 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 039 GES 20150104          Rework to move some state into the configuration. Added support for
**                           the DELETE-LINE() method. Some other attributes can now be supported
**                           on the server side since the state was moved to the config class.
** 040 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 041 EVK 20150220          Added runtime support for save-file and read-file methods.
** 042 IAS 20150225          Added support for new attributes/methods, fixed the widget behavior
** 043 HC  20150323          Removed business logic from config classes.
** 044 CA  20150509          Refactored to remove ChUI specific code and improved in impreparation 
**                           for GUI EDITOR support.
** 045 HC  20150626          Javadoc fixes.
** 046 IAS 20150627          Minor fix for SET-SELECTION
** 047 CA  20150616          Refactored to: remove the redundant "content" field; move more ChUI
**                           code to the EditorImpl class; changes to allow GUI widget.
** 048 IAS 20150801          TAB-STOP attribute support.
** 049 CA  20150923          Repaint on loss of focus.
** 050 EVL 20151104          Added abstract method to clean different EOL symbols in incoming
**                           character data.  The implementation is UI dependent.
** 051 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 052 CA  20160323          Fixed setCursorLine when the offset is below the last visible line
**                           (i.e. greater than config.scrollY + config.innerLines). Allow
**                           sub-classes to optimize the content parsing, in parseContentImpl.
** 053 OM  20170320          Changed signature of ThinClient.setHelp().
** 054 OM  20170426          Added line-level getters.
** 055 OM  20170504          Moved FIND_* constant to EditorWidget. Added support for MOVE-TO-EOF,
**                           SEARCH and INSERT-FILE. Fixed cursor positioning.
** 056 SBI 20180521          Changed to use getFileSystemDaemon instance method of ThinClient.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.nio.charset.*;
import java.nio.file.*;
import java.util.*;
import java.util.regex.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implementation of Progress EDITOR component.
 * <p>
 * Besides what was known from Progress manuals, the following facts describe
 * some peculiarities in the redirected output found experimentally:
 * <ul>
 *   <li>If the number of lines the editor holds is less than or equal to the
 *       height of the widget, the output will consist of the text and some  
 *       empty lines so that the height of the widget equals the number of
 *       output lines.
 *   <li>If the number of lines the editor holds is greater than the height of  
 *       the widget, all lines will be output, not just the visible portion.
 *   <li>No matter how many lines of text the editor holds, the visible lines
 *       is what is counted when placing the frame on the current page.
 *   <li>Unlike other widgets, editors can split the output into multiple
 *       pages. The page overflow condition is checked after every line of
 *       the editor's output. If the current page spills, the form feed 
 *       character is inserted as well as frame headers, if any. Page headers
 *       appear on the page as soon as the editor's output terminates, but no
 *       form feed character is printed with the page headers in this case.
 *   <li>If the editor is long enough, multiple page breaks and frame headers
 *       will be created.
 *   <li>Iterations of the down frames containing editors may have editors of
 *       various sizes.
 *   <li>Every iteration of the down frames containing editors can cause page
 *       break as described.
 * </ul>
 */
public abstract class Editor<O extends OutputManager<?>>
extends LabeledDataContainer<O,EditorConfig>
implements TextSelection
{
   /** line feed character */ 
   protected static final char LF = '\n';

   /** Carrage return character. */ 
   protected static final char CR = '\r';

   /** single space string */
   protected static final String SPACE = " ";

   /** space &amp; lf */
   protected static final String SPACE_LF = " \n";
   
   /** trailing spaces */
   private static final Pattern RTRIM = Pattern.compile("\\s+$"); 

   /** Editor content parsed into separate string. */
   protected List<Line> parsedText = null;

   /** Widget in focus flag. */
   protected boolean widgetFocused = false;

   /**
    * Max data size in widget. Depends on width-chars or inner-chars
    */
   // TODO this data was received experimental during testing on a ChUI-based Linux system and 
   // TODO it is not known if these values are universally applicable to all runtime conditions 
   // TODO and/or platforms.
   protected static final Map<Integer, Integer> maxSizes = new HashMap<>();
   
   /** Insert mode. */
   protected boolean insert = true;

   /** Flag indicating if the processed event forced a re-draw. */
   protected boolean forceRedraw = false;
   
   static
   {
      maxSizes.put(1, 68280);
      maxSizes.put(2, 103766);
      maxSizes.put(3, 139232);
      maxSizes.put(4, 174716);
      maxSizes.put(5, 174765);
      maxSizes.put(6, 245657);
      maxSizes.put(7, 245703);
      maxSizes.put(8, 245748);
      maxSizes.put(9, 245787);
      maxSizes.put(10, 245836);
      maxSizes.put(11, 245878);
      maxSizes.put(12, 458636);
      maxSizes.put(13, 458701);
      maxSizes.put(14, 458701);
      maxSizes.put(15, 458701);
      maxSizes.put(16, 458727);
      maxSizes.put(17, 458727);
      maxSizes.put(18, 458727);
      maxSizes.put(19, 458779);
      maxSizes.put(20, 458883);
      maxSizes.put(21, 458987);
      maxSizes.put(22, 459091);
      maxSizes.put(23, 459195);
      maxSizes.put(24, 459299);
      maxSizes.put(25, 884594);
      maxSizes.put(26, 884607);
      maxSizes.put(27, 884607);
      maxSizes.put(28, 884607);
      maxSizes.put(29, 884620);
      maxSizes.put(30, 884620);
      maxSizes.put(31, 884620);
      maxSizes.put(32, 884659);
      maxSizes.put(33, 884823);
      maxSizes.put(34, 884823);
      maxSizes.put(35, 884823);
      maxSizes.put(36, 884823);
      maxSizes.put(37, 884823);
      maxSizes.put(38, 1310396);
      maxSizes.put(39, 1310526);
      maxSizes.put(40, 1310526);
      maxSizes.put(41, 1310526);
      maxSizes.put(42, 1310539);
      maxSizes.put(43, 1310539);
      maxSizes.put(44, 1310539);
      maxSizes.put(45, 1310617);
      maxSizes.put(46, 1310807);
      maxSizes.put(47, 1310807);
      maxSizes.put(48, 1310807);
      maxSizes.put(49, 1310807);
      maxSizes.put(50, 1310807);
      maxSizes.put(51, 1736419);
      maxSizes.put(52, 1736419);
      maxSizes.put(53, 1736419);
      maxSizes.put(54, 1736419);
      maxSizes.put(55, 1736432);
      maxSizes.put(56, 1736432);
      maxSizes.put(57, 1736432);
      maxSizes.put(58, 1736432);
      maxSizes.put(59, 1736432);
      maxSizes.put(60, 1736791);
      maxSizes.put(61, 1736791);
      maxSizes.put(62, 1736791);
      maxSizes.put(63, 1736791);
      maxSizes.put(64, 2162208);
      maxSizes.put(65, 2162312);
      maxSizes.put(66, 2162312);
      maxSizes.put(67, 2162312);
      maxSizes.put(68, 2162325);
      maxSizes.put(69, 2162325);
      maxSizes.put(70, 2162325);
      maxSizes.put(71, 2162416);
      maxSizes.put(72, 2162775);
      maxSizes.put(73, 2162775);
      maxSizes.put(74, 2162775);
      maxSizes.put(75, 2162775);
      maxSizes.put(76, 2162775);
      maxSizes.put(77, 2588192);
      maxSizes.put(78, 2588205);
      maxSizes.put(79, 2588205);
      maxSizes.put(80, 2588205);
      maxSizes.put(81, 2588218);
      maxSizes.put(82, 2588218);
      maxSizes.put(83, 2588218);
      maxSizes.put(84, 2588231);
      maxSizes.put(85, 2588759);
      maxSizes.put(86, 2588759);
      maxSizes.put(87, 2588759);
      maxSizes.put(88, 2588759);
      maxSizes.put(89, 2588759);
      maxSizes.put(90, 3013994);
      maxSizes.put(91, 3014098);
      maxSizes.put(92, 3014098);
      maxSizes.put(93, 3014098);
      maxSizes.put(94, 3014111);
      maxSizes.put(95, 3014111);
      maxSizes.put(96, 3014111);
      maxSizes.put(97, 3014202);
      maxSizes.put(98, 3014743);
      maxSizes.put(99, 3014743);
      maxSizes.put(100, 3014743);
      maxSizes.put(101, 3014743);
      maxSizes.put(102, 3014743);
      maxSizes.put(103, 3439991);
      maxSizes.put(104, 3439991);
      maxSizes.put(105, 3439991);
      maxSizes.put(106, 3439991);
      maxSizes.put(107, 3440004);
      maxSizes.put(108, 3440004);
      maxSizes.put(109, 3440004);
      maxSizes.put(110, 3440004);
      maxSizes.put(111, 3440004);
      maxSizes.put(112, 3440030);
      maxSizes.put(113, 3440030);
      maxSizes.put(114, 3440727);
      maxSizes.put(115, 3440727);
      maxSizes.put(116, 3865793);
      maxSizes.put(117, 3865884);
      maxSizes.put(118, 3865884);
      maxSizes.put(119, 3865884);
      maxSizes.put(120, 3865897);
      maxSizes.put(121, 3865897);
      maxSizes.put(122, 3865897);
      maxSizes.put(123, 3866001);
      maxSizes.put(124, 3866711);
      maxSizes.put(125, 3866711);
      maxSizes.put(126, 3866711);
      maxSizes.put(127, 3866711);
      maxSizes.put(128, 3866711);
      maxSizes.put(129, 4291764);
      maxSizes.put(130, 4291777);
      maxSizes.put(131, 4291777);
      maxSizes.put(132, 4291777);
      maxSizes.put(133, 4291790);
      maxSizes.put(134, 4291790);
      maxSizes.put(135, 4291790);
      maxSizes.put(136, 4291803);
      maxSizes.put(137, 4292695);
      maxSizes.put(138, 4292695);
      maxSizes.put(139, 4292695);
      maxSizes.put(140, 4292695);
      maxSizes.put(141, 4292695);
      maxSizes.put(142, 4717540);
      maxSizes.put(143, 4717670);
      maxSizes.put(144, 4717670);
      maxSizes.put(145, 4717670);
      maxSizes.put(146, 4717683);
      maxSizes.put(147, 4717683);
      maxSizes.put(148, 4717683);
      maxSizes.put(149, 4717761);
      maxSizes.put(150, 4717761);
      maxSizes.put(151, 4717761);
      maxSizes.put(152, 4717761);
      maxSizes.put(153, 4718047);
      maxSizes.put(154, 4718047);
      maxSizes.put(155, 5143563);
      maxSizes.put(156, 5143563);
      maxSizes.put(157, 5143563);
      maxSizes.put(158, 5143563);
      maxSizes.put(159, 5143576);
      maxSizes.put(160, 5143576);
      maxSizes.put(161, 5143576);
      maxSizes.put(162, 5143576);
      maxSizes.put(163, 5143576);
      maxSizes.put(164, 5144663);
      maxSizes.put(165, 5144663);
      maxSizes.put(166, 5144663);
      maxSizes.put(167, 5144663);
      maxSizes.put(168, 5569339);
      maxSizes.put(169, 5569456);
      maxSizes.put(170, 5569456);
      maxSizes.put(171, 5569456);
      maxSizes.put(172, 5569469);
      maxSizes.put(173, 5569469);
      maxSizes.put(174, 5569469);
      maxSizes.put(175, 5569547);
      maxSizes.put(176, 5569560);
      maxSizes.put(177, 5569586);
      maxSizes.put(178, 5569846);
      maxSizes.put(179, 5570015);
      maxSizes.put(180, 5570015);
      maxSizes.put(181, 5995336);
      maxSizes.put(182, 5995349);
      maxSizes.put(183, 5995349);
      maxSizes.put(184, 5995349);
      maxSizes.put(185, 5995362);
      maxSizes.put(186, 5995362);
      maxSizes.put(187, 5995362);
      maxSizes.put(188, 5995375);
      maxSizes.put(189, 5996631);
      maxSizes.put(190, 5996631);
      maxSizes.put(191, 5996631);
      maxSizes.put(192, 5996631);
      maxSizes.put(193, 5996631);
      maxSizes.put(194, 6421151);
      maxSizes.put(195, 6421242);
      maxSizes.put(196, 6421242);
      maxSizes.put(197, 6421242);
      maxSizes.put(198, 6421255);
      maxSizes.put(199, 6421255);
      maxSizes.put(200, 6421255);
      maxSizes.put(201, 6421359);
      maxSizes.put(202, 6422615);
      maxSizes.put(203, 6422615);
      maxSizes.put(204, 6422615);
      maxSizes.put(205, 6422615);
      maxSizes.put(206, 6422615);
      maxSizes.put(207, 6847135);
      maxSizes.put(208, 6847135);
      maxSizes.put(209, 6847135);
      maxSizes.put(210, 6847135);
      maxSizes.put(211, 6847148);
      maxSizes.put(212, 6847148);
      maxSizes.put(213, 6847148);
      maxSizes.put(214, 6847148);
      maxSizes.put(215, 6847148);
      maxSizes.put(216, 6847148);
      maxSizes.put(217, 6847148);
      maxSizes.put(218, 6847148);
      maxSizes.put(219, 6847148);
      maxSizes.put(220, 7272937);
      maxSizes.put(221, 7273028);
      maxSizes.put(222, 7273028);
      maxSizes.put(223, 7273028);
      maxSizes.put(224, 7273041);
      maxSizes.put(224, 7273041);
      maxSizes.put(225, 7273041);
      maxSizes.put(226, 7273041);
      maxSizes.put(227, 7273145);
      maxSizes.put(228, 7273392);
      maxSizes.put(229, 7273431);
      maxSizes.put(230, 7274263);
      maxSizes.put(231, 7274276);
      maxSizes.put(232, 7276265);
      maxSizes.put(233, 7698908);
      maxSizes.put(234, 7698921);
      maxSizes.put(235, 7698921);
      maxSizes.put(236, 7698921);
      maxSizes.put(237, 7698934);
      maxSizes.put(238, 7698934);
      maxSizes.put(239, 7698934);
      maxSizes.put(240, 7698947);
      maxSizes.put(241, 7700567);
      maxSizes.put(242, 7700567);
      maxSizes.put(243, 7700567);
      maxSizes.put(244, 7700567);
      maxSizes.put(245, 7700567);
      maxSizes.put(245, 7700567);
      maxSizes.put(246, 8124658);
      maxSizes.put(247, 8124814);
      maxSizes.put(248, 8124814);
      maxSizes.put(249, 8124814);
      maxSizes.put(250, 8124827);
      maxSizes.put(251, 8124827);
      maxSizes.put(252, 8124827);
      maxSizes.put(253, 8124892);
      maxSizes.put(254, 8124892);
      maxSizes.put(255, 8124892);
      maxSizes.put(256, 8124892);
      maxSizes.put(257, 8124892);
      maxSizes.put(258, 8124892);
      maxSizes.put(259, 8550707);
      maxSizes.put(260, 8550707);
      maxSizes.put(261, 8550707);
      maxSizes.put(262, 8550707);
      maxSizes.put(263, 8550720);
      maxSizes.put(264, 8550720);
      maxSizes.put(265, 8550720);
      maxSizes.put(266, 8550720);
      maxSizes.put(267, 8550720);
      maxSizes.put(268, 8552535);
      maxSizes.put(269, 8552535);
      maxSizes.put(270, 8552535);
      maxSizes.put(271, 8552535);
      maxSizes.put(272, 8976509);
      maxSizes.put(273, 8976600);
      maxSizes.put(274, 8976600);
      maxSizes.put(275, 8976600);
      maxSizes.put(275, 8976600);
      maxSizes.put(276, 8976613);
      maxSizes.put(278, 8976613);
      maxSizes.put(279, 8976717);
      maxSizes.put(280, 8978519);
      maxSizes.put(281, 8978519);
      maxSizes.put(282, 8978519);
      maxSizes.put(283, 8978519);
      maxSizes.put(284, 8978519);
      maxSizes.put(285, 9402480);
      maxSizes.put(286, 9402493);
      maxSizes.put(287, 9402493);
      maxSizes.put(288, 9402493);
      maxSizes.put(289, 9402506);
      maxSizes.put(290, 9402506);
      maxSizes.put(291, 9402506);
      maxSizes.put(292, 9402519);
      maxSizes.put(293, 9402662);
      maxSizes.put(294, 9402662);
      maxSizes.put(295, 9402662);
      maxSizes.put(296, 9402675);
      maxSizes.put(297, 9403247);
      maxSizes.put(298, 9828321);
      maxSizes.put(299, 9828386);
      maxSizes.put(300, 9828386);
      maxSizes.put(301, 9828386);
      maxSizes.put(302, 9828399);
      maxSizes.put(303, 9828399);
      maxSizes.put(304, 9828399);
      maxSizes.put(305, 9828594);
      maxSizes.put(306, 9828594);
      maxSizes.put(307, 9828594);
      maxSizes.put(307, 9828594);
      maxSizes.put(308, 9828594);
      maxSizes.put(309, 9828594);
      maxSizes.put(310, 9828594);
      maxSizes.put(311, 10254279);
      maxSizes.put(312, 10254279);
      maxSizes.put(313, 10254279);
      maxSizes.put(314, 10254279);
      maxSizes.put(315, 10254292);
      maxSizes.put(316, 10254292);
      maxSizes.put(317, 10254292);
      maxSizes.put(318, 10254292);
   }
   
   /** Maximum capacity of the editor. */
   private static int MAX_SIZE = 32 * 1024;
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Editor(WidgetId id, EditorConfig cfg)
   {
   }
   
   /**
    * Sets position of terminal's cursor unconditionally (even if this editor is not in focus).
    */
   public abstract void drawCaretUnconditionally();

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   public abstract void draw(Color color);

   /**
    * Check if the given text needs to output on a new line.
    * 
    * @param    subtext
    *           The text.
    * @param    placeChar
    *           flag indicating if this call is from {@link #placeChar}.
    *
    * @return   <code>true</code> if the text will output on a new line.
    */
   protected abstract boolean goToNewLine(String subtext, boolean placeChar);

   /** 
    * Determine if a blank line needs to be inserted when RETURN key is pressed on a empty line.
    * 
    * @return   See above.
    */
   protected abstract boolean insertBlankLineOnLF();
   
   /**
    * Merge line with subsequent one(s) if possible
    * 
    * @param   row
    *          line's row
    */
   protected abstract void mergeLines(int row);
   
   /**
    * Get the next split position of this line, so that it fits the editor.
    * 
    * @param    line
    *           The line to be split.
    * @param    splitAfter
    *            minimal split position.
    *           
    * @return   An index where the split ends.
    */
   protected abstract int getNextSplitPosition(String line, int splitAfter);

   /**
    * Get absolute horizontal position in text.
    * 
    * @return  The horizontal position in the text.
    */
   protected abstract int absoluteX();

   /**
    * Processing for unusual EOL characters in different UI modes.
    * 
    * @param   src
    *          the string to be processed.
    *    
    * @return  the string with only '\n' characters as line seperator(EOL).
    */
   protected abstract String processEOLChars(String src);
   
   /**
    * Loads the file from the client side filesystem.
    * 
    * @param    name
    *           The file name for the file to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean editorReadFile(String name)
   {
      //todo client-side stream classes should probably be used for read/write of the files 
      //todo (instead of NIO) to ensure that the I18N behavior of the 4GL is honored.
      FileSystemDaemon fsd  = ThinClient.getInstance().getFileSystemDaemon();
      String fileNameToLoad = fsd.searchPath(name);
      if (fileNameToLoad == null)
      {
         return false;
      }
      boolean result = true;
      BufferedReader reader = null;
      try
      {
         Path path = FileSystems.getDefault().getPath(fileNameToLoad);

         reader = Files.newBufferedReader(path, Charset.defaultCharset());
         //capacity is int, maybe a problem for big files
         Integer maxSize = maxSizes.get((int)config.widthChars);
         if (maxSize == null)
         {
            maxSize = maxSizes.get(318); //known max size
         }
         StringBuilder sb = new StringBuilder(maxSize);
         int counter = 0;
         int c;
         while ((c = reader.read()) != -1)
         {
            if (counter++ > maxSize) 
            {
               String msg = "Editor is full. Some lines could not be inserted and have been lost";
               ErrorManager.recordOrShowError(5903, msg, false, false, false);
               result = false;
               break;
            }
            sb.append((char) c);
         }
         setValue(new character(sb.toString()));
         ConfigHelper.setModified(config, true);
      }
      catch (InvalidPathException | IOException e)
      {
         return false;
      }
      finally
      {
         if (reader != null)
         {
            try
            {
               reader.close();
            }
            catch (IOException e)
            {
               //no-op
            }
         }
      }

      return result;
   }

   /**
    * Save the content in the widgets in the client side filesystem.
    *
    * @param    name
    *           The file name for the file to be written.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean editorSaveFile(String name)
   {
      //todo client-side stream classes should probably be used for read/write of the files 
      //todo (instead of NIO) to ensure that the I18N behavior of the 4GL is honored.
      try
      {
         Path path = FileSystems.getDefault().getPath(name);
         String content = getValue().toStringMessage();

         Files.write(path, content.getBytes(), StandardOpenOption.CREATE); //create new, overwrite
         ConfigHelper.setModified(config, false);
      }
      catch (InvalidPathException | IOException e)
      {
         return false;
      }

      return true;
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, EditorConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      
      setValue(new character(resolveContent()));
      addFocusListener(getFocusListener());
      setEnabled(false);
   }
   
   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      return new character(resolveContent());
   }
   
   /**
    * Requesting the focus to the widget.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;
      
      super.requestFocus();
   }

   /**
    * Sets position of terminal's cursor.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      drawCaretUnconditionally();
   }

   /**
    * Draws text. <code>DisplayFormat.Presentation</code> instance is
    * rendered into screen value. Screen value is left justified or padded to
    * the right bound with spaces. Non first lines are filled with spaces. If
    * widget is focused then after rendering terminal's cursor is set to
    * appropriate position in the widget.
    */
   @Override
   public void draw()
   {
      Color dc = UiUtils.getDColor(this);
      Color pf = config.pfcolor;

      Color color = (widgetFocused && isEnabled()) ? pf : dc;

      if (focusTraversable() /*&& useUnderline*/)
         color = color.withUnderline();

      draw(color);
   }

   /**
    * Refresh the editor.
   */
   protected void refresh()
   {
      draw();
      requestSync();
   }
   
   /**
    * Get screen representation of the data.
    * 
    * @return  Text representation of the variable value according to all
    *          current settings, flags, format, etc.
    */
   @Override
   public String getText()
   {
      return resolveContent();
   }

   /**
    * Clear input variable.
    */
   public void clear()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      ConfigHelper.setModified(config, true);
      cursorHome();
      parseContent("", -1, false);
   }

   /**
    * This component will receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns this component directly because it is not a container (it has no child
    * components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget<O> currentFocus()
   {
      return this;
   }

   /**
    * Returns this component directly because it is not a container (it has no child
    * components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget<O> defaultFocus()
   {
      return this;
   }

   /**
    * Sets the new variable value in the editor.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (value == null)
         return;
      
      cursorHome();
      parseContent(value.toStringMessage(), -1, false);
   }

   /**
    * Process KeyEvents for this widget.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      // Call all KeyListener objects that may have been registered
      // for this component.
      super.processKeyEvent(ke);
      forceRedraw = false;

      // check if any of the KeyListeners consumed the KeyEvent
      if (ke.isConsumed() || !isEnabled())
      {
         return;
      }

      int key = ke.actionCode();
      
      // handle all printable characters and '\n'
      if (Keyboard.isPrintableKey(key) || 
          key == Keyboard.KA_RETURN    ||
          key == Keyboard.KA_NEW_LINE)
      {
         if (config.readOnly)
         {
            return;
         }
         
         ConfigHelper.setModified(config, true);

         char ch = Utils.toChar(key);

         if (key == Keyboard.KA_RETURN ||
             key == Keyboard.KA_NEW_LINE)
         {
            ch = LF;
            
            if (key == Keyboard.KA_NEW_LINE)
               cursorEndOfLine();
         }
         
         placeChar(ch);
         
         forceRedraw = true;
      }
      else
      {
         if (!hasFocus())
            return;
      }
      
      // key actions
      
      if (key == Keyboard.KA_TAB)
      {
         // the following call should be conditional
         ThinClient.nextTabStop(this);
         return;
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         // the following call should be conditional
         ThinClient.previousTabStop(this);
         return;
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         cursorDown();
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         cursorUp();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         cursorLeft();
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         cursorRight();
      }
      else if (key == Keyboard.KA_BACKSPACE)
      {
         backspace();
      }
      else if (key == Keyboard.KA_DELETE_CHARACTER)
      {
         deleteCharacterKey();
      }
      else if (key == Keyboard.KA_PAGE_DOWN)
      {
         cursorPageDown();
      }
      else if (key == Keyboard.KA_PAGE_UP)
      {
         cursorPageUp();
      }
      else if (key == Keyboard.KA_BELL)
      {
         bell();
      }
      else if (!forceRedraw)
      {
         // ignore all unknown actions
         return; 
      }

      refresh();
   }
   
   /**
    * Delete the line of text in which the cursor currently resides.  Leave the cursor at the
    * leftmost position of the same line number (which will be the line previously below this
    * one if the cursor is not already on the last line) or at the leftmost position of the
    * last line (if there are no lines below this one).
    *
    * @return   <code>true</code> if the line was deleted.
    */
   public boolean deleteLine()
   {
      boolean result = deleteLineWorker();
      
      refresh();
      
      return result;
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public boolean setSelection(int start, int end)
   {
      int contentLength = contentLength();
      int p1 = Math.min(start, end);
      int p2 = Math.max(start, end);
      int l1 = parsedText == null ? 0 : contentLength + 1;
      int l2 = contentLength() + 1;
      if (p1 < 1 || p2 > l2)
      {
         config.selectionActive = false;
         config.selectionStart = -1;
         config.selectionEnd = -1;
         config.selectionText = null;
         return false;
      }
      
      setCursorOffset(Math.min(config.selectionEnd, l1)); // may invalidate selection!
      config.selectionActive = true;
      config.selectionStart = p1;
      config.selectionEnd = p2;
      config.selectionText = resolveContent().substring(p1 - 1, Math.min(p2 - 1, contentLength));
      
      refresh();
      return true;
   }

   /**
    * Sets the cursor line
    *
    * @param    line
    *           the new cursor line value
    *           
    * @return   <code>true</code> if the cursor line is valid and set 
    */
   public boolean setCursorLine(int line)
   {
      if (line > parsedText.size())
      {
         return false;
      }
      
      // convert it to 0-based
      line -= 1;
      if (config.scrollY <= line && line < config.scrollY + config.innerLines)
      {
         config.cursorY = (line - config.scrollY) % config.innerLines;
      }
      else
      {
         config.scrollY = Math.max(0, Math.min(parsedText.size() - config.innerLines,
                                               line - line % config.innerLines));
         config.cursorY = line - config.scrollY;
      }
      config.cursorOffset = getCursorOffset();
      refresh();
      return true;
   }
      
   /**
    * Gets the cursor liner 
    * 
    * @return   the cursor line 
    */
   public int getCursorLine()
   {
      return absoluteY() + 1;
   }

   /**
    * Sets the cursor position in the current line
    *
    * @param    pos
    *           the new cursor position value
    *           
    * @return   <code>true</code> if the cursor position is valid and set 
    */
   public boolean setCursorChar(int pos)
   {
      return setCursorChar(pos, true);
   }

   /**
    * Sets the cursor position in the current line
    *
    * @param    pos
    *           The new cursor position value.
    * @param    redraw
    *           Request redraw/sync.
    *           
    * @return   <code>true</code> if the cursor position is valid and set. 
    */
   public boolean setCursorChar(int pos, boolean redraw)
   {
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
      if (redraw)
      {
         refresh();
      }
      return true;
   }
      
   /**
    * Gets the cursor in the current line 
    * 
    * @return   the cursor position 
    */
   public int getCursorChar()
   {
      Line line = getLine();
      return Math.min(absoluteX() + 1, line == null ? 1: line.length());
   }

   /**
    * Sets the cursor offset.
    *
    * @param   offset
    *          the new offset value. Use -1 to move the caret at the end (MOVE-TO-EOF).
    *
    * @return  {@code true} if the cursor offset is valid and set.
    */
   public boolean setCursorOffset(int offset)
   {
      if (offset == -1)
      {
         // move at the end of the editor. Always (?) returns true.
         Line lastLine = parsedText.get(parsedText.size() - 1);
         return setCursorLineAndChar(parsedText.size(), lastLine.length());
      }
      
      offset--;
      for (int i = 0; i <  config.numLines; i++)
      {
         Line line = parsedText.get(i);
         int lineLen = line.getOffset() + line.length() + (line.isNewLine() ? 1 : 0);
         if (offset <= lineLen)
         {
            return setCursorLineAndChar(i + 1, offset - line.getOffset() + 1);
         }
      }
      if (config.numLines > 0)
      {
         Line line = parsedText.get(config.numLines - 1); 
         return setCursorLineAndChar(config.numLines, line.length());
      }
      return false;
   }
   
   /**
    * Sets the cursor line offset
    *
    * @param    line
    *           the new cursor line value
    * @param    offset
    *           the new offset value
    * @return   <code>true</code> if the cursor offset is valid and set 
    */
   protected boolean setCursorLineAndChar(int line, int offset)
   {
      // the line needs to be set first
      if (line > parsedText.size())
      {
         return false;
      }
      setCursorLine(line);
      
      // set the offset within the line
      return setCursorChar(offset, false);
   }
      
   /**
    * Gets the cursor offset 
    * 
    * @return   the cursor offset 
    */
   public int getCursorOffset()
   {
      Line line = getLine();
      return line == null ? 1 : line.getOffset()  + Math.min(absoluteX(), line.length()) + 1;
   }

   /**
    * Gets the content length, compatible with the LENGTH attribute. 
    * 
    * @return   the content length 
    */
   public int getLength()
   {
      return contentLength() + Math.max(config.numLines - 1 , 0);
   }

   /**
    * Gets the actual content length (not the LENGTH attribute). 
    * 
    * @return   the content length 
    */
   public int contentLength()
   {
      if (parsedText == null || parsedText.isEmpty())
      {
         return 0;
      }
      
      Line lastLine = parsedText.get(parsedText.size() - 1);
      
      return lastLine.getOffset() + lastLine.length();
   }
   
   /**
    * Obtain the number of current lines displayed (parsed) by this editor. If no text was parsed
    * 0 is returned.
    * 
    * @return  the number of current lines displayed (parsed) by this editor.
    */
   public int getLineCount()
   {
      return parsedText == null ? 0 : parsedText.size();
   }
   
   /**
    * Obtain the text displayed on a specified line. If text has not been parsed yet or the
    * requested line is invalid, {@code null} is returned.
    * 
    * @param   k
    *          The line whose text is requested. 
    *          
    * @return  The content of the {@code k} line (0-base).
    */
   public String getLineText(int k)
   {
      if (parsedText == null || k < 0 || k >= parsedText.size())
      {
         return null;
      }
      
      return parsedText.get(k).getText();
   }

   /**
    * Inserts string at the current cursor position.
    * // OM: todo: implementation may be too slow from large texts
    *
    * @param   txt
    *          The string to be inserted.
    *
    * @return  {@code true} if the string is inserted. 
    */
   public boolean insertString(String txt)
   {
      if (txt == null || txt.isEmpty())
      {
         return true;
      }
      
      // check size of the buffer, on overflow, the text to be added is clipped
      int count = getLength();
      if (count + txt.length() > MAX_SIZE)
      {
         txt = txt.substring(0, MAX_SIZE - count);
      }
      
      String finalText = txt;
      // next lines need access to GUI driver/window to compute the width of the texts
      ThinClient.getInstance().eventDrawingBracket(this, () ->
      {
         for (int i = 0; i < finalText.length(); i++)
         {
            placeChar(finalText.charAt(i));
         }
         
         refresh();
      });
      
      return true;
   }
   
   /**
    * Inserts content of a file at the current cursor position.
    *
    * @param   fileNameToLoad
    *          The file to be inserted.
    *
    * @return  {@code true} if the file was successfully inserted. 
    */
   public boolean insertFile(String fileNameToLoad)
   {
      BufferedReader reader = null;
      try
      {
         Path path = FileSystems.getDefault().getPath(fileNameToLoad);
         reader = Files.newBufferedReader(path, Charset.defaultCharset());
         
         StringBuilder sb = new StringBuilder(MAX_SIZE);
         int counter = getLength();
         int c;
         while ((c = reader.read()) != -1)
         {
            if (counter++ > MAX_SIZE)
            {
               // abort with error message
               String msg = "The file (" + fileNameToLoad + 
                            ") for INSERT-FILE on the EDITOR widget is too large for the widget";
               ErrorManager.recordOrShowError(225, msg, true, false, false);
               return false;
            }
            sb.append((char) c);
         }
         
         insertString(sb.toString()); // at current offset
         ConfigHelper.setModified(config, true);
      }
      catch (InvalidPathException | IOException e)
      {
         return false;
      }
      finally
      {
         if (reader != null)
         {
            try
            {
               reader.close();
            }
            catch (IOException e)
            {
               //no-op
            }
         }
      }
      
      return true;
   }
   
   /**
    * Replace selection of the editor content with a given string.
    *
    * @param    txt
    *           the replacement string
    *
    * @return   <code>true</code> if the selection is replaced 
    */
   public boolean replaceSelectionText(String txt)
   {
      if (!config.selectionActive)
      {
         return false;
      }
      
      int selStart = config.selectionStart;
      int selEnd = config.selectionEnd;
      invalidateSelection();
      if (!replaceInterval(selStart, selEnd, txt))
      {
         return false;
      }

      config.selectionActive = false;

      refresh();
      return true;
   }
   
   /**
    * Replace subinterval of the editor content
    *  
    * @param start
    *        start of the interval
    * @param end
    *        end of the interval
    * @param txt
    *        replacement text
    *
    * @return   <code>true</code> if replacement done. 
    */
   private boolean replaceInterval(int start, int end, String txt)
   {
      int contentLength = contentLength();
      int p1 = Math.min(start, contentLength + 1);
      int p2 = Math.min(end, contentLength + 1);
      if (p1 < 0 || p2 < 0 )
      {
         return false;
      }

      int r1 = findRow(p1);
      int r2 = findRow(p2);
      if (r1 < 0 || r2 < 0)
      {
         return false;
      }

      setCursorOffset(p1);

      if (r1 == r2)
      {
         Line l = parsedText.get(r1);
         StringBuilder sb = new StringBuilder(l.getText());
         sb.delete(p1 - 1 - l.getOffset(),  p2 - 1 - l.getOffset());
         parsedText.set(r1, new Line(sb.toString(), l.getOffset(), l.isNewLine()));
      }
      else
      {
         Line l1 = parsedText.get(r1);
         Line l2 = parsedText.get(r2);
         for (int i = 0; i < r2 - r1; i++)
         {
            parsedText.remove(r1 + 1);
         }
         String s1 = l1.getText().substring(0, p1 - 1 - l1.getOffset());
         String s2 = l2.getText().substring(p2 - 1 - l2.getOffset());
         splitLine(s1+s2, r1, s1.length());
      }
      updateOffset();
      for (int i = 0; i < txt.length(); i++)
      {
         placeChar(txt.charAt(i));
      }
      
      // [mergeLines] needs access to GUI driver/window to compute the width of the texts
      ThinClient.getInstance().eventDrawingBracket(this, () -> mergeLines(absoluteY()));
      
      updateOffset();
      return true;
   }

   /**
    * Replace given substring of the editor content with a given string
    *
    * @param    from
    *           the string to be replaced
    * @param    to
    *           the replacement string
    * @param    flag
    *           the replacement mode flag as described in 4GL EDITOR:REPLACE documentation 
    * 
    * @return   <code>true</code> if the replacement was performed 
    */
   public boolean replaceString(String from, String to, int flag)
   {
      int contentLength = contentLength();
      if (contentLength == 0 || from == null || from.length() == 0)
      {
         return false;
      }
      int cs = (flag & EditorWidget.FIND_CASE_SENSITIVE) != 0 ? 0 : Pattern.CASE_INSENSITIVE;
      if ((flag & EditorWidget.FIND_NEXT_OCCURRENCE) != 0 && 
          (flag & EditorWidget.FIND_PREV_OCCURRENCE) != 0)
      {
         return false;
      }
      boolean next = (flag & EditorWidget.FIND_PREV_OCCURRENCE) == 0;
      boolean global = (flag & EditorWidget.FIND_GLOBAL) != 0;
      boolean wrap = (flag & EditorWidget.FIND_WRAP_AROUND) != 0;
      if (global && wrap)
      {
         return false;
      }
      
      int pos = getCursorOffset() - 1;
      int newPos = -1;
      
      StringBuffer buff = new StringBuffer(resolveContent());
      
      // Wildcards are not supported in ChUI
      Matcher m = Pattern.compile(from, Pattern.LITERAL | cs).matcher(buff);
      
      int foundAtRow = -1;
      
      int row;
      if (global)
      {
         int firstFoundAtRow = -1;
         StringBuffer result = new StringBuffer();
         while (m.find())
         {
            if ((row = inSameLime(m.start(), m.end())) >= 0)
            {
               foundAtRow = row;
               if (firstFoundAtRow < 0)
               {
                  firstFoundAtRow = row;
               }
               m.appendReplacement(result, to);
               newPos = result.length();
            }
         }
         m.appendTail(result);
         buff = result;
         if (foundAtRow >= 0)
         {
            parseContent(buff.toString(), firstFoundAtRow, true);
         }
      } 
      else
      {
         int foundAtPos = -1;
         if (next)
         {
            m = m.region(pos, contentLength);
            if (m.find())
            {
               if ((row = inSameLime(m.start(), m.end())) >= 0)
               {
                  foundAtRow = row;
                  foundAtPos = m.start();
               }
            }
            if (foundAtPos < 0 && wrap)
            {
               m = m.reset().region(0, pos);
               if (m.find())
               {
                  if ((row = inSameLime(m.start(), m.end())) >= 0)
                  {
                     foundAtRow = row;
                     foundAtPos = m.start();
                  }
               }
            }
         } 
         else
         {
            m = m.region(0, pos);
            while (m.find())
            {
               if ((row = inSameLime(m.start(), m.end())) >= 0)
               {
                  foundAtRow = row;
                  foundAtPos = m.start();
               }
            }
            if (foundAtPos < 0 && wrap)
            {
               m = m.reset().region(pos, contentLength);
               while (m.find())
               {
                  if ((row = inSameLime(m.start(), m.end())) >= 0)
                  {
                     foundAtRow = row;
                     foundAtPos = m.start();
                  }
               }
            }
         }
         if (foundAtPos >= 0)
         {
            replaceInterval(foundAtPos+1, foundAtPos + from.length(), to);
            newPos = foundAtPos + to.length();
         }
      }

      if (foundAtRow >= 0)
      {
         for (int i = 0; i < to.length(); i++)
         {
            if (to.charAt(i) == LF)
            {
               newPos--;
            }
         }
         
         setCursorOffset(newPos + 1);
         setCursorLine(foundAtRow);
         refresh();
         return true;
      }
      return false;
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>EditorWidget.FIND_NEXT_OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>EditorWidget.FIND_PREV_OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>EditorWidget.FIND_CASE_SENSITIVE</td><td>4</td></tr>
    *              <tr><td>EditorWidget.FIND_WRAP_AROUND</td><td>16</td></tr>
    *              <tr><td>EditorWidget.FIND_SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  This base class always returns {@code false}. Objects that extends it will
    *          implement their search and return the result accordingly.
    */
   public Boolean search(String pattern, int mode)
   {
      return false;
   }
   
   /**
    * Find row number for a position
    * 
    * @param    pos
    *           position
    *
    * @return   row number or -1 if position is out of range
    */
   private int findRow(int pos)
   {
      --pos; // lines counts chars 0-based
      for (int i = 0; i < parsedText.size(); i++)
      {
         Line line = parsedText.get(i);
         // check whether the EOLN (CR/LF) is at the end of same line
         int lastOffset = line.start + line.length() + (line.isNewLine() ? 1 : 0);
         if (line.start <= pos && pos < lastOffset)
         {
            return i;
         }
      }
      return -1;
   }

   /**
    * Check if there are now LFs between positions
    * 
    * @param start
    *        first position
    * @param end
    *        second position
    *        
    * @return
    *    number of row for the first position or -1 if provided positions
    *    are invalid or are not in the same line.  
    *    
    */
   private int inSameLime(int start, int end)
   {
      int row = findRow(start);
      if (row < 0)
      {
         return row;
      }
      for (int i = row; i < parsedText.size(); i++)
      {
         Line line = parsedText.get(i);
         if (line.start <= end && end <= line.start + line.length())
         {
            return row;
         }
         if (line.isNewLine())
         {
            return -1;
         }
      }
      return -1;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * Scroll one page up.
    */
   private void cursorPageUp()
   {
      if (absoluteY() >= config.innerLines - 1)
      {
         config.scrollY -= config.innerLines - 1;
         
         if (config.scrollY < 0)
            config.scrollY = 0;
      }
      else
      {
         config.scrollY = 0;
      }
   }

   /**
    * Scroll one page down.
    */
   private void cursorPageDown()
   {
      if ((config.scrollY + config.innerLines - 1) < lastRow())
      {
         config.scrollY += config.innerLines - 1;
         
         if (config.scrollY + config.innerLines - 1 > lastRow())
            config.scrollY = lastRow() - config.innerLines + 1;
      }
   }
   
   /**
    * Delete the line of text in which the cursor currently resides.  Leave the cursor at the
    * leftmost position of the same line number (which will be the line previously below this
    * one if the cursor is not already on the last line) or at the leftmost position of the
    * last line (if there are no lines below this one).
    *
    * @return   <code>true</code> if anything is deleted.
    */
   private boolean deleteLineWorker()
   {
      cursorBeginOfLine();
      
      // remember if this was the last line (must be here because the deletion and re-parse
      // of the text will change the number of lines if the deleted line was there because
      // of word wrapping instead of because of an explicit newline char
      boolean wasLast = (absoluteY() >= lastRow());
      
      // these first two statements delete all the content of the current line, except for the
      // newline character
      boolean resLine = deleteToEndOfLine();
      
      // at this point if we are on the last line, the last line is empty and there is no
      // following newline char to delete but instead we must remove the line, which requires
      // that we delete the newline character from the previous line
      if (wasLast)
      {
         cursorEndOfPrevLine();
      }
      
      boolean resChar = false;
      
      // if there is a newline, delete it
      if (getLine().isNewLine())
      {
         resChar = deleteCharacter();
      }
      
      // in the last line case we are at the end of the previous line which is now the new
      // last line, we must move to the beginning of the line
      if (wasLast)
      {
         cursorBeginOfLine();
      }
      
      // if neither of the above operations deleted anything, return false
      return (resLine || resChar);
   }  

   /**
    * Maximal value of the absolute Y coordinate.
    * 
    * @return  The last row index.
    */
   protected int lastRow()
   {
      return parsedText.size() - 1;
   }

   /**
    * Returns the number of lines the editor is holding.
    * 
    * @return  number of lines of text.
    */
   public int fullHeight()
   {
      int len = parsedText.size();

      // do not count the last line if it's empty
      if (len > 0)
      {
         Line line = parsedText.get(len - 1);
         if (line.getText().length() == 0 && !line.isNewLine())
         {
            len--;
         }
      }

      return len;
   }

   /**
    * Move cursor to beginning of the line.
    */
   protected void cursorBeginOfLine()
   {
      config.cursorX = 0;
      config.scrollX = 0;
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Delete to end of line.
    *
    * @return   <code>true</code> if text was deleted.
    */
   protected boolean deleteToEndOfLine()
   {
      if (config.readOnly || absoluteX() >= getLine().length())
      {
         bell();
         return false;
      }

      ConfigHelper.setModified(config, true);
      
      Line current = getLine();
      
      int len = current.length(); 
      
      if ((len - absoluteX()) > 0)
      {
         parsedText.set(absoluteY(), 
                        new Line(current.getText().substring(0, absoluteX()), 
                                 current.start, 
                                 current.isNewLine()));
         mergeLines(absoluteY());
         updateOffset();
         return true;
      }
      
      return false;
   }

   /**
    * Handle BELL event.
    */
   protected void bell()
   {
      screen().beep();
   }

   /**
    * Handle BACKSPACE key.
    */
   protected void backspace()
   {
      boolean checkMove = true;
      
      if (absoluteX() == 0 && config.cursorY > 0)
      {
         cursorEndOfPrevLine();
         checkMove = false;
         
         if (config.cursorX > 0 && absoluteX() == getLine().length())
         {
            config.cursorX--;
         }
      }

      int col = absoluteX();
      
      // move cursor first
      
      cursorLeft();
      
      if (checkMove && col == absoluteX())
         return;

      String line = getLine().getText();
      if (!checkMove && line.length() > 0 && line.charAt(absoluteX()) != ' ')
      {
         // we're about to merge lines
         cursorRight();
      }
      
      deleteCharacter();
   }

   /**
    * Move the cursor to the end of the previous line.
    */
   protected void cursorEndOfPrevLine()
   {
      cursorUp();
      cursorEndOfLine();
   }

   /**
    * Move cursor to the start position. 
    */
   protected void cursorHome()
   {
      config.cursorX = 0;
      config.cursorY = 0;
      config.scrollX = 0;
      config.scrollY = 0;
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
  }

   /**
    * Move cursor to the end of line.
    */
   protected void cursorEndOfLine()
   {
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }
   
   /**
    * Get reference to the line under absolute cursor position.
    * 
    * @return  The instance of the <code>Line</code> at current position.
    */
   protected Line getLine()
   {
      return parsedText == null ? null
                                : parsedText.get(Math.min(absoluteY(), parsedText.size() - 1));
   }

   /**
    * Split absolute horizontal cursor position into two components - position
    * of the first visible character and relative cursor position in the
    * visible region.
    * 
    * @param   pos
    *          The new absolute cursor position.
    */
   protected void cursorPos(int pos)
   {
      config.scrollX = pos - config.cursorX;
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Move cursor one row down, if possible.
    */
   protected void cursorDown()
   {
      int low = Math.min(parsedText.size(), config.innerLines);
      
      // cursor does not reach boundary
      if (config.cursorY < (low - 1))
      {
         config.cursorY++;
      }
      else
      {
         // Cursor reached the low boundary
         if (parsedText.size() > config.innerLines && 
             parsedText.size() > (config.scrollY + config.cursorY + 1))
         {
            // scroll down
            config.scrollY++;
         }
      }
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Move cursor one row up, if possible.
    */
   protected void cursorUp()
   {
      // move cursor up
      if (config.cursorY > 0)
      {
         config.cursorY--;
      }
      else
      {
         // scroll up and keep cursor at the first row
         if (config.scrollY > 0)
            config.scrollY--;
      }
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Move cursor one char right.
    */
   protected void cursorRight()
   {
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Move cursor one char left, if possible.
    */
   protected void cursorLeft()
   {
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Place new line character at current cursor position. 
    */
   protected void placeLF()
   {
      Line current = getLine();

      ConfigHelper.setModified(config, true);

      int absX          = absoluteX();
      int absY          = absoluteY();
      int currentOffset = current.getOffset();
      int offset        = currentOffset + absX;
      
      if (absX >= current.length())
      {
         if (!current.isNewLine())
         {
            current.setNewLine(true);
            
            // if this is past the last line in the parsed text, add a new line
            if (absY >= parsedText.size() - 1)
            {
               parsedText.add(absY + 1, new Line("", offset + 1, false));
            }
         }
         else 
         {
            // this is already a new line, just add another one
            parsedText.add(absY + 1, new Line("", offset + 1, true));
         }
      }
      else 
      {
         // split the line in two
         parsedText.set(absY, new Line(current.getText().substring(0, absX), currentOffset, true));
         if (insertBlankLineOnLF() && absX == 0 && absY > 0)
         {
            Line prev = parsedText.get(absY - 1);
            prev.setNewLine(true);
         }
         parsedText.add(absY + 1 , 
                        new Line(current.getText().substring(absX),
                                 currentOffset + absX + 1, 
                                 current.isNewLine()));

         // depending on how the editor is configured (scrollable/not scrollable, GUI/ChUI), the
         // lines beyond these may need to be merged
         mergeLines(absY + 1);
      }
      
      setCursorChar(0);
      cursorDown();
   }
   
   /**
    * Determine if a character can be placed in the editor.
    * 
    * @return    <code>true</code> if the editor is not read-only.
    */
   protected boolean canPlaceChar()
   {
      return !config.readOnly;
   }
   
   /**
    * Place character at current cursor position. Depending on the current
    * insertion mode character can be either inserted into text or can replace
    * existing character.
    * 
    * @param   ch
    *          Character to place.
    */
   protected void placeChar(char ch)
   {
      if (!canPlaceChar())
      {
         bell();
         return;
      }

      if (ch == LF)
      {
         placeLF();
         config.numLines = parsedText.size();
         config.length = getLength();
         return;
      }

      ConfigHelper.setModified(config, true);
      
      Line current = getLine();
      StringBuilder line = new StringBuilder(current.getText());
      
      if (absoluteX() >= current.length())
      {
         int len = absoluteX() - current.length();
         
         if (len > 0)
         {
            String spaces = StringHelper.repeatChar(' ', len);
            line.insert(current.length(), spaces);
         }
      }
      
      if (insert)
      {
         int absX = absoluteX();
         if (absX >= line.length())
         {
            line.append(ch);
         }
         else
         {
            line.insert(absX, ch);
         }
      }
      else
      {
         int absX = absoluteX();
         if (absX >= line.length())
         {
            line.append(ch);
         }
         else
         {
            line.setCharAt(absX, ch);
         }
      }
      
      int cursorOffset = getCursorOffset();
      splitLine(line.toString(), absoluteY(), 0);
      
      if (absoluteX() >= getLine().length() - 1 && goToNewLine(line.toString(), true))
      {
         cursorDown();
         config.cursorX = cursorOffset - getLine().getOffset();
      }
      else
         cursorRight();
      
      updateOffset();
   }

   /**
    * Replace line, splitting it if necessary
    * 
    * @param line
    *        line's content
    * @param row
    *        line's row
    * @param splitAfter
    *        minimal split position 
    */
   protected void splitLine(String line, int row, int splitAfter)
   {
      for (; row < parsedText.size(); row++)
      {
         Line current = parsedText.get(row);
         if (!goToNewLine(line, true))
         {
            parsedText.set(row, new Line(line, current.getOffset(), current.isNewLine()));
            break;
         }
         
         // TODO: optimize

         int pos = getNextSplitPosition(line, splitAfter);
         parsedText.set(row, new Line(line.substring(0, pos), current.getOffset(), false));
         String tail = line.substring(pos);
         if (absoluteY() >= parsedText.size() - 1)
         {
            parsedText.add(new Line(tail, current.getOffset() + pos, false));
            break;
         }
         else if (current.isNewLine())
         {
            parsedText.add(row + 1, new Line(tail, current.getOffset() + pos, true));
            break;
         }
         else 
         {
            Line next = null;
            if (row == parsedText.size() - 1)
            {
               // add a new, empty, line
               next = new Line("", 0, false);
               parsedText.add(next);
            }
            else
            {
               next = parsedText.get(row + 1);
            }
            
            next.setOffset(current.getOffset() + pos);
            line = tail + next.getText();
            splitAfter = tail.length();
         }
      }
   }

   /**
    * Get absolute vertical position in text.
    * 
    * @return  The vertical position in the text.
    */
   protected int absoluteY()
   {
      return config.cursorY + config.scrollY;
   }

   /**
    * Delete character at current cursor position - called by {@link #processKeyEvent}.
    *
    */
   protected void deleteCharacterKey()
   {
      deleteCharacter();
   }
   
   /**
    * Delete character at current cursor position.
    *
    * @return   <code>true</code> if something was deleted.
    */
   protected final boolean deleteCharacter()
   {
      if (config.readOnly)
      {
         bell();
         return false;
      }
      
      ConfigHelper.setModified(config, true);
      
      Line oldLine = getLine();
      
      if (absoluteX() >= oldLine.length())
      {
         // at the end of the text, just ignore the request
         if (absoluteY() == lastRow())
            return false;
         
         // move cursor at the end of the text
         cursorEndOfLine();
      }
      
      boolean result = false;
      
      int idx = oldLine.getOffset() + absoluteX();
      
      if (oldLine.isNewLine() && absoluteX() >= oldLine.length())
      {
         // delete from the next line
         idx = idx + 1;
      }
      
      if (idx < contentLength())
      {
         if ((oldLine.isNewLine() || oldLine.length() == 0) && absoluteX() >= oldLine.length())
         {
            if (oldLine.length() == 0)
            {
               parsedText.remove(absoluteY());
            }
            else
            {
               oldLine.setNewLine(false);
            }
         }
         else if (absoluteX() < oldLine.getText().length())
         {
            StringBuilder line = new StringBuilder(oldLine.getText());
            line.deleteCharAt(absoluteX());
            parsedText.set(absoluteY(), 
                           new Line(line.toString(), oldLine.getOffset(), oldLine.isNewLine()));
         }

         mergeLines(absoluteY());
         updateOffset();
         
         int newdelta = getLine().length() - absoluteX();
         
         if (newdelta < 0)
         {
            cursorDown();
            cursorBeginOfLine();
            
            for (int i = 0; i > newdelta; i--)
               cursorRight();   
         }
         
         result = true;
      }
      
      return result;
   }

   /**
    * Parse content of the editor and prepare array of displayable values.
    * 
    * @param   source
    *          Source string to parse.
    * @param   row
    *          Row index to start parsing from.
    * @param   placeChar
    *          Set to {@code true} if called from {@link #placeChar(char)} method in insert mode.
    */
   protected void parseContent(String source, int row, boolean placeChar)
   {
      ArrayList<Line> lines = new ArrayList<>();
      
      for (int i = 0; i < row; i++)
         lines.add(parsedText.get(i));

      // prepare source to pasing, this is UI dependent 
      source = processEOLChars(source);
      
      int start  = 0;
      int offset = 0;
      
      if (row >= 0)
         start = offset = parsedText.get(row).getOffset();
      
      lines.addAll(parseContentImpl(source, start, offset, placeChar));

      if (placeChar && absoluteX() == (config.innerChars - 1))
      {
         start = 0;
         if (!lines.isEmpty())
         {
            Line last = lines.get(lines.size() - 1);
            start = last.getOffset() + last.getText().length() + (last.isNewLine() ? 1 : 0);
         }
         lines.add(new Line("", start, false));
      }

      parsedText = lines;
      
      updateOffset();
   }
   
   /**
    * Worker to parse content of the editor and prepare array of displayable values.
    * <p>
    * Sub-classes can override this, to i.e. use cached metrics.
    * 
    * @param   source
    *          Source string to parse.
    * @param   start
    *          Row index to start parsing from.
    * @param   offset
    *          Text offset to start parsing from.
    * @param   placeChar
    *          Set to {@code true} if called from {@link #placeChar(char)} method in insert mode.
    */
   protected List<Line> parseContentImpl(String source, int start, int offset, boolean placeChar)
   {
      List<Line> lines = new ArrayList<>();
      
      char[] text = source.toCharArray();
      
      for (int i = offset; i < text.length; i++)
      {
         if (text[i] == LF)
         {
            lines.add(new Line(text, start, i - start, true));
            start = i + 1;
            continue;
         }

         if (i > start && text[i - 1] == ' ' && text[i] != ' ')
         {
            // check for word wrapping
            int j = i;

            boolean fullLine = false;
            boolean newWord = false;
            while (j < text.length)
            {
               if (text[j] == ' '  || text[j] == LF)
               {
                  newWord = true;
                  j--;
                  break;
               }

               if (goToNewLine(source.substring(start, j), false))
               {
                  fullLine = true;
                  break;
               }
               
               j++;
            }
            
            if (j == text.length)
            {
               newWord = true;
               j--;
            }
            
            // if this is a full line, determine if this is the end of the word; if so, split just
            // after it, but without moving the space to the next line
            if (fullLine)
            {
               int delta = (placeChar ? 1 : 0);
               
               if (j < text.length - delta && (text[j + delta] == ' ' || text[j + delta] == LF))
               {
                  // if the word has not finished yet, split from the beginning of the word
                  lines.add(new Line(text, start, j - start, false));

                  // advance just after the word, consuming the whitespace
                  start = j + delta;
                  i = start;
               }
               else
               {
                  // if the word has not finished yet, split from the beginning of the word
                  lines.add(new Line(text, start, i - start, false));

                  start = i;
               }
               
               continue;
            }
            
            // if we have found the end of this word, determine if if fits the current line; if
            // it does not fit, the word will go to the next line
            int wordEnd = Math.min(j + 1, source.length());
            if (newWord && goToNewLine(source.substring(start, wordEnd), false))
            {
               lines.add(new Line(text, start, i - start, false));
               start = i;
               continue;
            }
         }

         // last check

         if (goToNewLine(source.substring(start, i + 1), false))
         {
            lines.add(new Line(text, start, i - start, false));
            
            if (text[i] == ' ')
               i++;
            
            start = i;
         }
      }

      if (lines.size() != 0 && lines.get(lines.size() - 1).isNewLine() && text.length - start <= 0)
      {
         lines.add(new Line("", start, false));
         placeChar = false;
      }

      if (text.length - start > 0 || lines.size() == 0)
      {
         lines.add(new Line(text, start, text.length - start, false));
         start = text.length;
      }
      
      return lines;
   }
   
   /**
    * Get a string representation of this editor's content.
    * 
    * @return   See above.
    */
   protected final String resolveContent()
   {
      if (parsedText == null)
      {
         return "";
      }

      StringBuilder buff = new StringBuilder();
      for (Line line : parsedText)
      {
         buff.append(line.getText());
         if (line.isNewLine())
         {
            buff.append(LF);
         }
      }
      
      return buff.toString();
   }
   
   /**
    * Get the focus listener instance.
    * 
    * @return   See above.
    */
   protected FocusListener getFocusListener()
   {
      return new LocalFocusListener();
   }
   
   /**
    * Remove trailing whitespaces
    * 
    * @param   value
    *          the string to be truncated
    *    
    * @return  the truncated argument
    */
   protected String rtrim(String value)
   {
      return RTRIM.matcher(value).replaceAll("");
   }

   /**
    * Update widget content;
    */
   protected void updateOffset()
   {
      config.length = getLength();
      config.numLines = parsedText.size();
      int offset = 0;
      for (Line line: parsedText)
      {
         line.setOffset(offset);
         offset += line.length();
         if (line.isNewLine())
         {
            offset++;
         }
      }
   }
   
   /**
    * Some actions on focus hanging.
    */
   protected class LocalFocusListener
   implements FocusListener
   {
      /**
       * Setups <code>FillIn</code> as it is just activated by Tab key.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusGained(FocusEvent e)
      {
         if (e == null)
         {
            return;
         }
         
         widgetFocused = true;
         
         ThinClient.setHelp(config.help, Editor.this);
         
         repaint();
      }
      
      /**
       * Reset SELECTED attribute when focus is lost.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusLost(FocusEvent e)
      {
         if (e == null)
            return;
         
         widgetFocused = false;
         
         repaint();
      }
   }

   /**
     * Internal text representation element.
     */
   protected static class Line
   {
      /** Initial character index in the source text. */
      private int start;
      
      /** Line text. */
      private final String text;
      
      /**
       * <code>true</code> if instance was created for '\n' in the source text.
       */
      private boolean newLine;
      
      /**
       * The intermediate widths (char by char) for this line; index 0 is for char 1, index 1 is 
       * for chars 1 and 2, etc
       */
      private int[] widths = null;
      
      /**
       * Construct an instance from given character array, offset and length.
       * 
       * @param   text
       *          Source character array.
       * @param   start
       *          Initial character offset.
       * @param   length
       *          Line length.
       * @param   newLine
       *          <code>true</code> means that instance will is created for
       *          the text line with '\n' at the end. The character itself
       *          should not be included into <code>length</code> parameter.
       */
      public Line(char[] text, int start, int length, boolean newLine)
      {
         this.start   = start;
         this.newLine = newLine;
         
         if (text != null && length > 0)
            this.text = new String(text, start, length);
         else
            this.text = "";
      }

      /**
       * Construct an instance for given string and offset.
       * 
       * @param   text
       *          Source string.
       * @param   start
       *          Initial offset.
       * @param   newLine
       *          <code>true</code> means that instance will is created for
       *          the text line with '\n' at the end.
       */
      public Line(String text, int start, boolean newLine)
      {
         this.start   = start;
         this.newLine = newLine;
         
         if (text != null)
            this.text = text;
         else
            this.text = "";
      }
      
      /**
       * Get the widths for these metrics.
       * 
       * @return   See above.
       */
      public int[] getWidths()
      {
         return widths;
      }
      
      /**
       * Set the widths for these metrics.
       *
       * @param    widths
       *           The intermediate char-by-char widths for the entire line. Index 0 is for
       *           char 1, index 1 is for chars 1 and 2, and so on.
       */
      public void setWidths(int[] widths)
      {
         this.widths = widths;
      }
      
      /**
       * Return {@code true} if instance represents the text with the '\n' at the end.
       * 
       * @return  The state of the flag.
       */
      public boolean isNewLine()
      {
         return newLine;
      }
      
      /**
       * Set 'newLine' attribute
       *
       * @param value
       *        new value of the 'newLine' attribute
       */
      public void setNewLine(boolean value)
      {
         newLine = value;
      }
      
      /**
       * Get text length.
       * 
       * @return  The text length.
       */
      public int length()
      {
         return text.length();
      }
      
      /**
       * Return text.
       * 
       * @return  The line text.
       */
      public String getText()
      {
         return text;
      }
      
      /**
       * Return the offset of this instance in the source text.
       * 
       * @return  The index of the first line character in the source text.
       */
      public int getOffset()
      {
         return start;
      }
      
      /**
       * Set the offset of this instance in the source text.
       * 
       * @param value
       *        new offset value
       */
      public void setOffset(int value)
      {
         start = value;
      }
   }
}
