/*
** Module   : LabeledPanel.java
** Abstract : Panel with border and (optionally) - side or top label.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20108020 Refactored common code from RadioSet and SelectionList.
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 EVL 20140528 Adding implementation stubs for two new methods.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 007 CA  20150114 Fixed support for changing the ROW/COLUMN attrs from server-side, for a 
**                  side-label.
** 008 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
** 009 CA  20150202 Side-labels needs to track the HIDDEN state, too.
** 010 HC  20150214 Improved positioning of side labels.
** 011 CA  20150223 Allow the label to be removed.
** 012 HC  20150323 Removed business logic from config classes.
** 013 CA  20150722 Added postprocessConfig - required additional processing, as layout is 
**                  performed only on the first frame definition push.
** 014 HC  20150831 Renamed from LabeledBorderedPanel to LabeledPanel and removed border
**                  functionality from the class.
** 015 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Common implementation of {@link LabeledWidget} interface for 
 * {@link RadioSet} and {@link SelectionList}.
 * 
 * @author 	SIY
 * @version 1.0
 */
public abstract class LabeledPanel<O extends OutputManager<?>, C extends ControlSetConfig>
extends AbstractContainer<O>
implements LabeledWidget<O>,
           WidgetWithConfig<C>,
           MnemonicTrait<O>
{
   /** Component attribute container. */
   protected C config;
   
   /** Component label. */
   private Label<O> label = null;

   /** Mnemonic information */
   private MnemonicInfo mnemonic = null;

   /** Column mnemonic information */
   private MnemonicInfo columnMnemonic = null;

   /**
    * Default constructor.
    */
   public LabeledPanel()
   {
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public C config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(C beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      updateMnemonic(beforeUpdate.clabel, config.clabel, true);
      updateMnemonic(beforeUpdate.label, config.label, false);
   }

   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public double getColon()
   {
      return ConfigHelper.getColon(config());
   }
   
   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return getMnemonicLabel(true).orElse(config().clabel);
   }
   
   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label<O> getLabelInstance()
   {
      return label;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return getMnemonicLabel(false).orElse(config.label);
   }

   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public double getTo()
   {
      return ConfigHelper.getTo(config());
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().forceLabel;
   }
   
   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return !config().labels;
   }
   
   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      if (label != this.label && this.label != null)
      {
         this.label.destroy();
      }
      
      if (label == null)
      {
         this.label = null;
         return;
      }

      int frameId = ThinClient.frameIdFromWidgetId(this.config().id.asInt());
      
      // link the label to the frame of this widget
      label.linkTo(frameId, config().sideLabelId);

      this.label = label;
      this.label.setVisible(isVisible());
      this.label.setHidden(hidden());
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config().labels = !labels;
   }

   /**
    * Shows or hides the radio-set and its label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (label != null)
         label.setVisible(visible);
   }
   
   /**
    * Tracks the hidden state of the widget and attached label.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      super.setHidden(hidden);
      
      if (label != null)
         label.setHidden(hidden);
   }

   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
   
   /**
    * Additional processing required when widget configuration is updated AFTER the frame layout
    * has been performed.
    */
   @Override
   public void postprocessConfig()
   {
      super.postprocessConfig();
      
      if (label != null)
      {
         label.postprocessConfig(this);
      }
   }
   
   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      if (label != null)
      {
         label.destroy();
         label = null;
      }
      
      super.destroy();
   }

   /**
    * MnemonicInfo instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic is advised.
    *
    * @return   See above.
    */
   public MnemonicInfo getMnemonic()
   {
      return mnemonic;
   }

   /**
    * MnemonicInfo instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic is advised.
    */
   public void setMnemonic(MnemonicInfo mnemonic)
   {
      this.mnemonic = mnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic in the implemented method is advised.
    * <p>
    * The method will return a valid value when the implementing widget supports column label.
    *
    * @return   See above.
    */

   public MnemonicInfo getColumnMnemonic()
   {
      return columnMnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic in the implemented method is advised.
    */
   public void setColumnMnemonic(MnemonicInfo mnemonic)
   {
      this.columnMnemonic = mnemonic;
   }
}