/*
** Module   : Menu.java
** Abstract : Menu abstract client implementation
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-------------------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150226 Added method getFirstEnabledWidget.
**                  Added PopupKeyListener and move focus and event processing methods.
** 004 VIG 20150618 Moved setVisible method from ChUI implementation.
** 005 VIG 20150905 Fixed event processing. 
** 006 VIG 20151027 Changed isMenubarElement to return false instead of throwing exception
**                  in cases when specified component is not MenuItem or SubMenu.
** 007 VIG 20151226 Fixed first item focusing of CHUI in showPopupMenu().
** 008 VIG 20160313 Added popupShown() method, added menu location set to showPopupMenu(). 
** 009 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 010 VIG 20160811 Fixed mnemonic and accelerators processing.
** 011 SBI 20161101 getWidgetsList() renamed to widgetsAsList(). 
** 012 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 013 HC  20180528 Fixed key events in top-level window being handled by menu when it shouldn't.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
/**
 * <code>Menu</code> client abstract implementation class. 
 */
public abstract class Menu<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<MenuConfig>
{
   /** Executed when a menu item is selected */
   private Runnable menuClosed;

   /** Menu attribute container. */
   private MenuConfig config = null;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Menu(WidgetId id, MenuConfig cfg)
   {
      config = cfg;
   }
   
   /**
    * Check if this menu-item is right inside MENUBAR (first level)
    * 
    * @param    item
    *           {@link MenuItem} or {@link SubMenu}
    *           
    * @return  <code>true</code> if yes.
    */
   public static boolean isMenubarElement(Widget<?> item)
   {
      if (!(item instanceof MenuItem) && !(item instanceof SubMenu))
      {
         return false;
      }
      
      Container<?> parent = item.parent();
      return parent instanceof Menu && !((Menu<?>)parent).config().popupOnly;
   }

   /**
    * Checks whether the menu of the passed in menu element has focus, i.e. any of the child elements
    * are focused and accept input.
    *
    * @param   element
    *          A menu element.
    *
    * @return  {@code true} if the menu has focus.
    */
   public static boolean hasMenuFocus(MenuStructure element)
   {
      Menu m = element.getParentMenu();
      return m != null && m.isCurrent();
   }
   
   /**
    * Shows this pop-up menu.
    * 
    * @param   src
    *          Source widget.
    * @param   location
    *          The popup location.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   public final void showPopup(Widget<O> src, NativePoint location, Runnable dismissHandler)
   {
      screen().getRegistry().addWidgetsRecursive((Widget) this);
      doShowPopup(src, location, dismissHandler);
      KeyInput e = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_MENU_DROP);
      EventManager.postEvent(e);
   }

   /**
    * Get menu widget from menu-item or sub-menu.
    * 
    * @param    mel
    *           {@link MenuItem} or {@link SubMenu}
    *           
    * @return   The menu, containing specified item.
    */
   public static Menu<?> getMenu(MenuElement<?> mel)
   {
      Menu<?> menu = mel.getParentMenu();
      
      if (menu == null)
      {
         throw new IllegalStateException("Specified widget is not contained in a menu");
      }
      
      return menu;
   }
   
   /**
    * Looks up the very first enabled widget.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   public abstract Widget<O> getFirstEnabledWidget();
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, MenuConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public MenuConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(MenuConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Set displayed flag of this menu.
    * 
    * @param   displayed
    *          <code>true</code> means that menu is dropped down, 
    *          <code>false</code> otherwise.
    */
   public void setDisplayed(boolean displayed)
   {
      if (config.popupOnly)
      {
         config.displayed = displayed;
      }
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      if (config.popupOnly)
      {
         return config.displayed;
      }

      Widget<?> owner = screen().getRegistry().getComponent(config.ownerId);
      
      return owner == null ? false : owner.isDisplayed();
   }

   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      setDisplayed(visible);
      super.setVisible(visible);
      
      for (Widget<O> w : widgets())
      {
         w.setVisible(visible);
      }
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      if (config.popupOnly && config.visible)
      {
         Widget<?> owner = screen().getRegistry().getComponent(config.ownerId);
         return (owner == null || 
                (owner instanceof Container && ((Container<?>)owner).focus() != this)) ? 
                false : owner.hasFocus();
         
      }
      
      return super.hasFocus();
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }
      
      //int key = event.keyCode();
      int actionCode = event.actionCode();
      
      // menu-item choosing events processing
      if ((actionCode == Keyboard.SE_VALUE_CHANGED ||
           actionCode == Keyboard.SE_CHOOSE) && 
          config().popupOnly)
      {
         hide();
         return;
      }
      
      // accelerator
      if (event.isRealKey())
      {
         MenuItem<O> m = findMenuItem(event.keyCode(), this);
         if (m != null && processAccelerator(event, m))
         {
            return;
         }
         // mnemonic 
         else if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
         {
            if (SubMenu.processMnemonicEvent(this, event))
            {
               event.consume();
               return;
            }
         }
      }
      else
      {
         // Can lead to infinite loop, because AbstractContainer.processKeyEvent()
         // delegates processing to its current focused widget.
         // super.processKeyEvent(event);
      }
   }

   /**
    * Processes menu key accelerator event.
    *
    * @param   event
    *          The key event.
    */
   public void processKeyAccelerator(KeyInput event)
   {
      // TODO: ATM accelerators (Alt + key) for menus are not implemented
   }

   /**
    * Close all opened bodies and remove focus from the menu.
    */
   public void leaveMenu()
   {
      // TODO: ChUI
   }

   /**
    * Process accelerator event. Can be overridden. 
    * 
    * @param   event
    *          KeyInput object, related to key event.
    * @param   m
    *          Target accelerator source menu-item.
    *          
    * @return  <code>true</code> means accelerator was processed.
    *          <code>false</code> means accelerator was ignored.
    */
   protected boolean processAccelerator(KeyInput event, MenuItem<?> m)
   {
      EventManager.postEvent(new ItemEvent(m));
      event.consume();
      return true;
   }

   /**
    * Destroy this widget.
    */
   @Override
   public void destroy()
   {
      screen().getRegistry().removeWidgetsRecursive((Widget) this);
      super.destroy();
   }

   /**
    * Sets a handler that will be executed when a menu item is selected. Only one (or no) handler
    * can be set at a time.
    *
    * The handler is executed after the menu UI is made invisible but before the related progress events
    * are posted (CHOOSE, VALUE-CHANGED).
    *
    * @param   menuClosed
    *          May be {@code null}.
    */
   public void setMenuClosedHandler(Runnable menuClosed)
   {
      this.menuClosed = menuClosed;
   }

   /**
    * Triggers the menu-closed event handler.
    */
   public void triggerMenuClosed()
   {
      if (menuClosed != null)
      {
         menuClosed.run();
      }
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      cycleFocus(direction);
   }
   
   /**
    * The abstract portion of the logic making pop-up menu shown.
    * 
    * @param   src
    *          Source widget.
    * @param   location
    *          The popup location.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    */
   protected void doShowPopup(Widget<O> src, NativePoint location, Runnable dismissHandler)
   {
      if (parent() == null)
      {
         src.topLevelWindow().getContentPane().add(this);
      }
      
      if (!isVisible())
      {
         setPhysicalLocation(location.x, location.y);
         setVisible(true);
         doLayout();
         firstFocus();
      }

      repaint();
   }
   
   /**
    * Find MENU-ITEM by specified accelerator key code in specified MENU (SUB-MENU) recursively.
    * If there are several menu-items have the same accelerator - the first found by DFS 
    * menu-item with specified accelerator key code is returned.
    * 
    * @param   accelCode
    *          Key code of accelerator.
    *          
    * @param   menu
    *          Current menu or sub-menu
    *          
    * @return  MENU-ITEM with accelerator code equals accelCode or <code>null</code> 
    */
   private MenuItem<O> findMenuItem(int accelCode, AbstractContainer<O> menu)
   {
      for(Widget<O> w : menu.widgetsAsList())
      {
         if (w instanceof MenuItem)
         {
            String accel = ((MenuItem<O>) w).config().accelerator;
            if (accel != null && Keyboard.keyCode(accel) == accelCode)
            {
               return (MenuItem<O>) w;
            }
         }
         else
         {
            MenuItem<O> m = findMenuItem(accelCode, (AbstractContainer<O>) w);
            if (m != null)
            {
               return m;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Key listener, intended to show pop-up menu. Used for different components. 
    * Keys:
    *   ESC-U in CHUI and SHIFT-F10 in GUI
    */
   public static class PopupKeyListener 
   implements KeyListener
   {
      /**
       * This method is invoked when key is pressed.
       * 
       * @param   evt
       *          Event which triggered this call.
       */
      @Override
      public void onKeyPressed(KeyInput evt)
      {
         // Should we process TriggerResult here?
         
         Widget<GuiOutputManager> src = evt.source();
         int action = (evt == null ? KeyInput.CHAR_UNDEFINED  : evt.actionCode());

         // an active pop-up menu key. TODO check for GUI
         if (action == Keyboard.KA_DEFAULT_POP_UP && src != null)
         {
            int menuId = src.config().popupMenuId;
            if (menuId != -1)
            {
               Menu<GuiOutputManager> menu = 
                        (Menu<GuiOutputManager>) src.screen().getRegistry().getComponent(menuId);
               if (menu != null)
               {
                  menu.showPopup(src, null, null);
                  evt.consume();
               }
            }
         }
      }

      /**
       * This method is invoked when key is typed.
       * 
       * @param   event
       *          Event which triggered this call.
       */
      @Override
      public void onKeyTyped(KeyInput event)
      {
         // Nothing
      }
   }
}