/*
** Module   : MnemonicTrait.java
** Abstract : Common interface for all widgets with mnemonic support.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 HC  20161016 Created initial version
** 002 CA  20161101 Refactored to eliminate GUI/ChUI dependency and create the MnemonicInfo via 
**                  the driver/widget factory. Added support for mnemonic processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

import java.util.*;

/**
 * The interface adds mnemonic support to widgets.
 * <p>
 * The idea is to have the common logic implemented in the interface as default methods and
 * free up the widgets from having to duplicate code. The widgets only have to implement two
 * methods {@link #getMnemonic()} and {@link #setMnemonic(MnemonicInfo)} to handle state.
 */
public interface MnemonicTrait<O extends OutputManager<?>>
{
   /**
    * Check whether this instance supports mnemonics.
    *
    * @return   <code>true</code> if this widget supports mnemonic.
    */
   default boolean isMnemonicSupported()
   {
      return true;
   }

   /**
    * Creates new {@link MnemonicInfo} instance.
    *
    * @return   MnemonicInfo or its descendant.
    */
   default MnemonicInfo createMnemonic()
   {
      return OutputManager.getDriver().getFactory().createMnemonic();
   }

   /**
    * Updates the mnemonic label. The update will be performed only when <code>oldLabel</code>
    * differs from <code>newLabel</code>. To force the label to be updated pass
    * <code>null</code> in <code>oldLabel</code>.
    *
    * @param    oldLabel
    *           The widget label being changed.
    * @param    newLabel
    *           The new label widget.
    * @param    columnMnemonic
    *           Indicates whether the regular or column mnemonic should be updated.
    */
   default void updateMnemonic(String oldLabel, String newLabel, boolean columnMnemonic)
   {
      if (!isMnemonicSupported())
      {
         return;
      }

      MnemonicInfo mnemonic = columnMnemonic ? getColumnMnemonic() : getMnemonic();

      if (mnemonic == null && newLabel == null ||
          mnemonic != null && oldLabel != null && newLabel != null && oldLabel.equals(newLabel))
      {
         // the label has not changed, do not update the mnemonic info
         return;
      }

      if (newLabel == null)
      {
         if (columnMnemonic)
         {
            setColumnMnemonic(null);
         }
         else
         {
            setMnemonic(null);
         }
      }
      else
      {
         mnemonic = createMnemonic();
         mnemonic.prepare(newLabel);

         if (columnMnemonic)
         {
            setColumnMnemonic(mnemonic);
         }
         else
         {
            setMnemonic(mnemonic);
         }
      }
   }

   /**
    * Returns the preprocessed mnemonic label, i.e. with unescaped mnemonic characters.
    *
    * @param    columnMnemonic
    *           Indicates whether the label for regular or column mnemonic should be returned.
    *
    * @return   See above.
    */
   default Optional<String> getMnemonicLabel(boolean columnMnemonic)
   {
      MnemonicInfo mnemonic = columnMnemonic ? getColumnMnemonic() : getMnemonic();
      if (mnemonic == null)
      {
         return Optional.empty();
      }

      return Optional.ofNullable(mnemonic.getPrepLabel());
   }

   /**
    * Returns the mnemonic character from the processed widget label.
    *
    * @param    columnMnemonic
    *           Indicates whether the character for regular or column mnemonic should be returned.
    *
    * @return   See above.
    */
   default Optional<Character> getMnemonicCharacter(boolean columnMnemonic)
   {
      MnemonicInfo mnemonic = columnMnemonic ? getColumnMnemonic() : getMnemonic();
      if (mnemonic == null)
      {
         return Optional.empty();
      }

      return Optional.ofNullable(mnemonic.getMnemonic());
   }
   
   /**
    * Returns the associated key code for this mnemonic character.
    * 
    * @param    columnMnemonic
    *           Indicates whether the character for regular or column mnemonic should be returned.
    * @param    alt
    *           Indicates if the key combination includes ALT+character or the character alone.
    * 
    * @return   See above.
    */
   default Optional<Integer> getMnemonicKeyCode(boolean columnMnemonic, boolean alt)
   {
      Character ch = getMnemonicCharacter(columnMnemonic).orElse(null);
      // all mnemonics are case-insensitive...
      if (ch != null) 
      {
         ch = Character.toUpperCase(ch);
      }
      
      String lbl = ch == null ? null : (alt ? "ALT-" : "") + ch.toString();
      
      return ch == null ? Optional.empty() : Optional.ofNullable(Keyboard.keyCode(lbl));
   }

   /**
    * {@link MnemonicInfo} instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic in the implemented method is advised.
    *
    * @return   See above.
    */
   MnemonicInfo getMnemonic();

   /**
    * {@link MnemonicInfo} instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic in the implemented method is advised.
    */
   void setMnemonic(MnemonicInfo mnemonic);

   /**
    * Column {@link MnemonicInfo} instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic in the implemented method is advised.
    * <p>
    * The method will return a valid value when the implementing widget supports column label.
    *
    * @return   See above.
    */

   default MnemonicInfo getColumnMnemonic()
   {
      UnimplementedFeature.unsupported("The widget doesn't allow a column mnemonic.");
      return null;
   }

   /**
    * Column {@link MnemonicInfo} instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic in the implemented method is advised.
    */
   default void setColumnMnemonic(MnemonicInfo mnemonic)
   {
      UnimplementedFeature.unsupported("The widget doesn't allow a column mnemonic.");
   }
}