/*
** Module   : NativePoint.java
** Abstract : Screen point primitive.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140816 Created initial version
** 002 MAG 20140926 Added minus operation.
** 003 CA  20150615 Added equals and hashCode.
** 004 HC  20150921 Added method translate(NativeInsets).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

/**
 * Point primitive.
 * The coordinate values are expressed in units native to UI type.
 */
public class NativePoint
{
   /** Point X coordinate. */
   public int x;

   /** Point Y coordinate. */
   public int y;

   /**
    * Constructor.
    * 
    * @param   x
    *          Point X.
    * @param   y
    *          Point Y.
    */
   public NativePoint(int x, int y)
   {
      this.x = x;
      this.y = y;
   }

   /**
    * Copy constructor.
    * 
    * @param   point
    *          Source point.
    */
   public NativePoint(NativePoint point)
   {
      x = point.x;
      y = point.y;
   }
   
   /**
    * Shift point to specified offset.
    * 
    * @param   x
    *          X offset.
    * @param   y
    *          Y offset.
    * 
    * @return  new shifted point. 
    */
   public NativePoint offset(int x, int y)
   {
      return new NativePoint(this.x + x, this.y + y);
   }
   
   /**
    * Translate point object by specified offset without creating new instance.
    * 
    * @param   dx
    *          Delta X. 
    * @param   dy 
    *          Delta Y.
    * 
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public NativePoint translate(int dx, int dy)
   {
      x += dx;
      y += dy;
      return this;
   }

   /**
    * Translate point object by specified offset without creating new instance.
    *  
    * @param   point
    *          Source point.
    *          
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public NativePoint translate(NativePoint point)
   {
      return translate(point.x, point.y);
   }
   
   /**
    * Translate point object by specified offset without creating new instance.
    *  
    * @param   insets
    *          Source insets.
    *          
    * @return  <code>this</code> for fluent syntax chained calls.
    */
   public NativePoint translate(NativeInsets insets)
   {
      return translate(insets.left, insets.top);
   }
   
   /**
    * Shift point to specified negative offset.
    * 
    * @param   point
    *          Source point.
    * 
    * @return  new shifted point. 
    */
   public NativePoint minus(NativePoint point)
   {
      return offset(-point.x, -point.y);
   }

   /**
    * For debug purposes
    */
   public String toString()
   {
      return x + ":" + y;
   }
   
   /**
    * Return the unique hash code for this object.
    *
    * @return  Unique hash code for this object.
    */
   @Override
   public int hashCode()
   {
      return x * 37 + y;
   }
   
   /**
    * Determine if this instance is equal with the specified object.
    * <p>
    * The two objects are equal if the <code>obj</code> is a {@link NativePoint} and its 
    * {@link #x} and {@link #y} coordinates match with this instance.
    * 
    * @param    obj
    *           The object to check.
    * 
    * @return   See above.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (!(obj instanceof NativePoint))
      {
         return false;
      }
      
      NativePoint np = (NativePoint) obj;
      
      return np.x == this.x && np.y == this.y;
   }
}