/*
** Module   : OutputPrimitives.java
** Abstract : Common interface for output primitives implementation.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20110112 Changes related to new color handling scheme.
** 004 SIY 20110927 Updated javadoc.
** 005 GES 20111005 Updated import to match refactoring.
** 006 CA  20140806 Extracted the ChUI related primitives to the BasePrimitives class.
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 010 HC  20160628 Prevented unhandled exceptions when window/frame closed/destroyed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.chui.driver.DriverPrimitives;

/**
 * Provides an API for an object which can render formatted output as a
 * terminal.
 * <p>
 * This interface holds vast majority of methods necessary for P2J to perform 
 * user interaction. Particular implementation may use {@link DriverPrimitives}
 * class as a base and share most of the code required to implement this 
 * interface.
 */
public interface OutputPrimitives
{
   /**
    * Clear specified area at the screen.
    * <p>
    * The driver-level implementation does not need to implement this method 
    * because it is never reaches the driver.
    * 
    * @param   left
    *          Left coordinate of the area.
    * @param   top
    *          Top coordinate of the area.
    * @param   right
    *          Right coordinate of the area.
    * @param   bottom
    *          Bottom coordinate of the area.
    * @param   color
    *          Color.
    */
   public void clearArea(int left, int top, int right, int bottom, Color color);
   
   /**
    * Refresh specified area.
    * <p>
    * This method is used to refresh current state of the implementations which
    * use buffering. For driver-level implementation this method should do 
    * nothing. 
    * 
    * @param   rect
    *          Area to refresh. 
    */
   public void refresh(NativeRectangle rect);

   /**
    * Get width of the screen represented by this instance.
    * <p>
    * Driver-level implementation of this method should return physical 
    * dimension of the screen used for user interaction. The dimension is 
    * expressed in characters.
    * 
    * @return  Screen width.
    */
   public int screenWidth();

   /**
    * Get height of the screen represented by this instance.
    * <p>
    * Driver-level implementation of this method should return physical 
    * dimension of the screen used for user interaction. The dimension is 
    * expressed in characters.
    * 
    * @return  Screen height.
    */
   public int screenHeight();
   
   /**
    * Synchronize screen.
    * <p>
    * Driver level implementation of this method should force flushing internal
    * low-level buffers to screen. This method requires implementation only if
    * operating system or hardware uses internal buffering for the terminal 
    * output. Otherwise this method can do nothing. 
    */
   public void sync();
   
   /**
    * Get the screen bitmap for this output primitives implementation. 
    * <p>
    * This method is used by implementations which use internal buffering.
    * The driver-level implementation does not need to implement this method 
    * because it is never reaches the driver.
    * 
    * @return   The screen bitmap.
    */
   public ScreenBitmap getScreenBitmap();
   
   /**
    * Replace the current bitmap instance with this one.
    * <p>
    * This method is used by implementations which use internal buffering.
    * The driver-level implementation does not need to implement this method 
    * because it is never reaches the driver.
    * 
    * @param    bitmap
    *           The new bitmap instance.
    */
   public void setScreenBitmap(ScreenBitmap bitmap);
   
   /**
    * Get the ID of the window currently selected. If no window is currently
    * selected -1 is returned.
    * 
    * @return   See above.
    */
   public int getWindowId();
}