/*
** Module   : RadioSet.java
** Abstract : RadioSet implementation for CHUI Radio-set
**
** Copyright (c) 2005-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050301   @20143 Creation based on CHARVA JPanel class.
**                           Necessary modification to get compiled
** 002 EVL 20050303   @20149 Adding initial data arbitrary association
**                           for radio buttons
** 003 EVL 20050304   @20170 Handling of auto-resize feature
** 004 EVL 20050524   @21284 Adding the Progress like methods
** 005 EVL 20050604   @21405 Changing the inner variable names to sync
**                           with the rest of the project.
** 006 EVL 20050604   @21406 Adding the Progress specific methods to work
**                           with Progress radio-set attributes
** 007 EVL 20050607   @21435 Syntax for comments checking
** 008 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also 
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change
** 009 EVL 20050723   @21806 Properly handling of the change the
**                           RADIO_BUTTONS attribute.
** 010 EVL 20050822   @22199 Using the P2JFrame/Container interfaces to
**                           detect the class.
** 011 SIY 20050831   @22510 Refactoring (see @@22394-22409). 
** 012 SIY 20050907   @22646 Fixed use of @inheritDoc tag in javadoc.
** 013 SIY 20050909   @22668 Fixed double initialization of items.
** 014 EVL 20050912   @22676 Adding Validateable interface to this widget.
** 015 EVL 20050913   @22698 Adding focus listener interface.
** 016 EVL 20050914   @22714 Adding one more focus lost event to identify
**                           actual focus lost for radio-set.
** 017 EVL 20050916   @22736 Returned value class is changed to be the
**                           BaseDataType.
** 018 EVL 20050920   @22762 Adding the implementation of the new method
**                           isModified() for the Validateable interface.
** 019 EVL 20050921   @22766 Adding method to be able to undo the widget's
**                           value.
** 020 SIY 20050920   @22846 Added support for SENSITIVE and SELECTED 
**                           attributes.
** 021 GES 20060114   @24207 Implemented DataContainer interface and the
**                           core methods.
** 022 SIY 20060131   @24208 Use ThinClient special method instead of 
**                           direct calls to Window.message().
** 023 SIY 20060209   @24462 Major refactoring and cleanups.
** 024 SIY 20060210   @24519 Added getSelection() method, added 
**                           ControlSetComponent interface.
** 025 SIY 20060222   @24640 Added necessary methods to support changes in
**                           DataContainer interface.
** 026 NVS 20060227   @24869 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 027 NVS 20060314   @25069 Added reset() method that resets the set of
**                           buttons to the first button selected.
** 028 SIY 20060322   @25212 Added setNoLabels() to match new 
**                           LabeledWidget interface.
** 029 SIY 20060410   @25465 Fixed size calculation. 
** 030 SIY 20060417   @25593 More minor fixes and cleanups.
** 031 SIY 20060515   @26155 Encapsulated access to help field.
** 032 SIY 20060515   @26156 Added getColumnLabelText().
** 033 SIY 20060518   @26277 Changed initial widget state to disabled.
** 034 NVS 20060630   @27729 Reviewed requestFocus() calls. Those in
**                           processKeyEvents() are safe, and those in
**                           refreshItems() are *likely* to be safe.
**                           However, nextFocus() and previousFocus()
**                           calls made from processKeyEvent() are
**                           definitely UNSAFE and need separate
**                           consideration.
** 035 NVS 20060703   @27803 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 036 SIY 20060824   @28855 Removed getP2JType() method.
** 037 NVS 20061204   @31559 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 038 NVS 20061214   @31692 Added getIndex() method and enabled drawing
**                           of uninitialized radio buttons.
** 039 NVS 20061221   @31733 Currently focused button within the group
**                           should be preserved when tabbing through the
**                           widgets of the frame.
** 040 NVS 20070426   @33254 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 041 SIY 20070807   @34805 Added isColumnFixed() method.
** 042 SVL 20071221   @36531 Label is shown/hidden together with the
**                           widget. 
** 043 SIY 20080320   @37600 Fixed handling of the SCREEN-VALUE and data
**                           value.
** 044 NVS 20080401   @37779 Added isFocusable() method implementation.
** 045 SIY 20080411   @37921 Significant rewrite in order to accommodate
**                           changes in focus handling.
**                           Ref: #37923
** 046 SIY 20100619          Switching to own implementation of character UI.
** 047 SIY 20100804          Changes in imports, some refactorings.
** 048 SIY 20100822          Refactored AttributeContainer support.
** 049 SIY 20101122          Switching from KeyEvent to KeyInput.
** 050 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 051 SIY 20111202          Refactored and moved from CHUI package.
** 052 SIY 20111221          Switched to widget factory.
** 053 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 054 HC  20140911          GUI coordinate support - data types refactoring.
** 055 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 056 HC  20141021          Implemented window move/position support.
** 057 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 058 IAS 20150214          Implemented ADD-LAST method.
** 059 IAS 20150312          Implemented ENABLE/DISABLE methods.
** 060 HC  20150323          Removed business logic from config classes.
** 061 IAS 20150801          TAB-STOP attribute support
** 062 HC  20150831          Scrolling support overhaul.
** 063 EVL 20150925          Change the buttons, current and selectButton() access to proteced to
**                           be used in RadioSetGuiImpl.
** 064 CA  20151125          Fixed computing default-focus when the first widget in frame is a 
**                           radio-set.
** 065 HC  20160628          A workaround fix for a radio button not being displayed when it
**                           doesn't fit into the radio-set visible area.
** 066 HC  20160922          RADIO-SET should be initially disabled.
** 067 CA  20160930          Fixed radio-button position in horizontal and expanded mode.
** 068 CA  20161101          Fixed RADIO-SET enable - set the ENABLE flag via the widget's 
**                           setEnabled() API, not just the config.enabled flag.
** 069 OM  20170320          ThinClient.setHelp() signature change.
** 070 CA  20180226          Use reflectAncestorChange() to notify the 'buttons', as this doesn't 
**                           use the AbstractContainer.widgets to keep the children.
** 071 OM  20180321          SensitiveScrollContainer.ensureVisibility() has new parameter.
** 072 CA  20180502          Fixed disabled/enabled state (better fix for H068).
** 073 EVL 20180813          Fixed incorrect change for MODIFIED attribute for ENTRY event
**                           processing.  Limit to GUI mode only.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;


import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * RadioSet is a implementation of radio-set widget, a mutually exclusive set
 * of {@link RadioButton} objects.
 * <p>
 * This is the ChUI specific implementation of the widget.
 */
@SuppressWarnings("unchecked")
public class RadioSet<O extends OutputManager<?>>
extends LabeledPanel<O,RadioSetConfig>
implements FocusListener,
           DataContainer,
           ControlSetComponent
{
   /** current button */
   protected int current = 0;
   
   /** Button collection. */
   protected ArrayList<RadioButton<O>> buttons = new ArrayList<>();
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RadioSet(WidgetId id, WidgetConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RadioSetConfig cfg)
   {
      config = (RadioSetConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      addFocusListener(this);

      disable();
   }
   
   /**
    * This overriding is used to move the focus from this widget to the
    * next/previous one, rather than moving between radio buttons inside the
    * current radio-set.
    * 
    * @param   keyEvent
    *          The key event to be processed.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (keyEvent.isConsumed())
         return;
      
      int action = keyEvent.actionCode();
      int key = keyEvent.keyCode();

      int save_current = current;

      if (action == Keyboard.SE_ENTRY && ThinClient.getInstance().isChui())
      {
         // force setting of SCREEN-VALUE on entry
         ConfigHelper.setModified(config, true);
      }
      if (action == Keyboard.KA_CURSOR_UP || 
               action == Keyboard.KA_CURSOR_LEFT)
      {
         for (int i = current - 1; i >= 0; i--)
         {
            if (buttons.get(i).isEnabled())
            {
               current = i;
               break;
            }
         }
      }
      else if (action == Keyboard.KA_CURSOR_DOWN ||
               action == Keyboard.KA_CURSOR_RIGHT)
      {
         for (int i = current + 1; i < buttons.size(); i++)
         {
            if (buttons.get(i).isEnabled())
            {
               current = i;
               break;
            }
         }
      }
      else if (action == Keyboard.KA_TAB) 
      {
         ThinClient.nextTabStop(this);
      }
      else if (action == Keyboard.KA_BACK_TAB)
      {
         ThinClient.previousTabStop(this);
      }
      else if (key == ' ')
      {
         RadioButton<O> currentButton = buttons.get(current);
         
         if (!currentButton.isSelected())
         {
            selectButton(currentButton);
            ConfigHelper.setModified(config, true);
            EventManager.postEvent(new KeyInput(this,
                  EventType.KEY_PRESSED,
                  Keyboard.SE_VALUE_CHANGED,
                  true));
        }
      }
      else
         super.processKeyEvent(keyEvent);
      
      if (current != save_current)
      {
         buttons.get(save_current).setFocused(false);
         buttons.get(current).setFocused(true);

         repaint();
         requestSync();

         Container<O> parent = parent();
         if (parent instanceof SensitiveScrollContainer)
         {
            ((SensitiveScrollContainer<O>)parent).ensureVisibility(this, false);
         }
      }
   }

   /**
    * Must return the value of the radio-set as a BaseDataType.
    * 
    * @return  The BaseDataType value of the selected button inside the
    *          button group.
    */
   @Override
   public BaseDataType getValue()
   {
      RadioButton<O> button = getSelected();
      
      return (button != null) ? button.getValue() : null;
   }
   
   /**
    * Get text representation of the currently selected value.
    * 
    * @return  The screen representation of the value.
    */
   @Override
   public String getText()
   {
      RadioButton<O> button = getSelected();
      
      return (button != null) ? button.getText() : null;
   }
   
   /**
    * Sets the value of the radio-set (the selected button) based on a
    * BaseDataType value.
    * 
    * @param    value
    *           The BaseDataType value of the button to be selected inside
    *           the button group.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      for (RadioButton<O> button : buttons)
      {
         if (button.getValue().equals(value))
            selectButton(button);
      }
   }

   /**
    * Resets the value of the radio-set (the selected button) to the first
    * button of the set.
    */
   public void reset()
   {
      if (buttons.isEmpty())
      {
         return;
      }
      RadioButton<O> b = buttons.get(0);
      
      if (b.isSelected())
      {
         ConfigHelper.setModified(config, true);
         setInitial();
      }
      else
      {
         selectButton(b);
      }
   }

   /**
    * Indicates whether this component can be traversed using TAB or BACK-TAB
    * events.
    * 
    * @return  <code>true</code> if widget can be traversed.
    */
   @Override
   public boolean focusTraversable() 
   {
       return isEnabled() && isVisible();
   }
   
   /**
    * Get default focusable widget.
    * <p>
    * If at least one button is focusable, then return this radio-set (not individual button).
    *
    * @return    default focusable widget.
    */
   @Override
   public Widget<O> defaultFocus()
   {
      Widget<O> defFocus = null;

      for (Widget<O> c : buttons)
      {
         if (c.focusTraversable())
         {
            defFocus = c;
            break;
         }
      }
      
      // the loop above should always find a focusable button; 4GL does not allow direct access
      // to the radio-set's buttons, so when disabling/enabling the radio-set, it affects all 
      // buttons (their state is in sync with the parent's state).  If the radio-set can't find
      // a focusable button, then there is a problem in P2J, and the exception bellow will be
      // raised

      if (defFocus == null)
         throw new IllegalStateException("no focus-traversable widgets found");

      return this;
   }
   
   /**
    * Return reference to this, since REDIO-SET is always retains focus on
    * itself.
    * 
    * @return  this
    */
   @Override
   public Widget<O> currentFocus()
   {
      return this;
   }
   
   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.modified;
   }

   /**
    * Set up the width of the button dynamically.
    * 
    * @param   newWidth
    *          The new width of the radio-button.
    */
   public void setWidthChars(double newWidth)
   {
      newWidth = Coordinate.scale(newWidth);
      
      if (newWidth <= 0 || config.horizontal)
         return;

      for (RadioButton<O> button : buttons)
         button.setWidthChars(newWidth);

      repaint();
   }

   /**
    * Getting the width of the button. The final width includes the button's
    * label size.
    * 
    * @return  The integer number of the width of the radio-buttons.
    */
   public double getCharWidth()
   {
      return config.widthChars;
   }

   /**
    * Draw RadioSet widget.
    */
   @Override
   public void draw()
   {
      for (RadioButton<O> button : buttons)
         button.draw();
   }
   
   /**
    * This is called when the radio-set takes the input focus. When the help string is defined for
    * the radio-set it should be displayed at the window status line.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent focusEvent)
   {
      if (focusEvent == null) // just make argument referenced
      {
         return;
      }
      
      buttons.get(current).setFocused(true);
      
      ThinClient.setHelp(config.help, this);
      
      repaint();
      requestSync();
   }

   /**
    * This is called when the radio-set looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      if (focusEvent == null)
         return;
      
      buttons.get(current).setFocused(false);

      repaint();
      requestSync();
   }

   /**
    * Refresh widget look using new set of buttons.
    */
   @Override
   public void refreshItems()
   {
      int selectedId = -1;
      boolean selectedDisabled = false;
      for (RadioButton<O> button: buttons) 
      {
         if (button.isSelected())
         {
            selectedId = button.getItemId();
            selectedDisabled = !button.isEnabled();
            break;
         }
      }
      buttons.clear();
      
      WidgetFactory<O> factory = screen().getFactory();
      
      ControlSetItem[] items = config.items;
      for (int i = 0; i < items.length; i++)
      {
         // TODO: Do not share this.config with radio buttons.
         //       Also remove RadioButton.location()/physicalLocation() when this is resolved.
         RadioButton<O> button = factory.createRadioButton(config, 
                                                           items[i].getLabel().getValue(), 
                                                           items[i].getValue());
         button.setParent(this);

         buttons.add(button);
      }
      
      refreshItemsImpl(selectedId, selectedDisabled);
    }

   /**
    * Request indexes of selected items.
    * 
    * @return  array of 0-based indexes of selected items.
    */
   @Override
   public int[] getSelection()
   {
      int index = buttons.indexOf(getSelected());
      
      if (index < 0)
         index = 0;

      return new int[] {index};
   }
   
   /**
    * Request index of some button.
    * 
    * @param   rb 
    *          Radio button.
    *          
    * @return  0-based index of the button or -1 if not in the set.
    */
   public int getIndex(RadioButton<O> rb)
   {
      return buttons.indexOf(rb);
   }

   /**
    * Enable/disable widget.
    *
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      super.setEnabled(enabled);

      // TODO: ConfigManager doesn't implement deep comparison when looking for changes
      //       to synchronize the changes up the server, as a quick solution just copy the
      //       array, this will change the items reference and so will be recognized as dirty
      config.items = Arrays.copyOf(config.items, config.items.length);
      
      for (ControlSetItem csi : config.items)
      {
         csi.setEnabled(enabled);
      }

      for (RadioButton<O> btn : buttons)
      {
         btn.setEnabled(enabled);
      }
   }

   /**
    * Track index of currently selected button. 
    */
   private void setInitial()
   {
      config.initial = getSelection()[0];
   }
   
   /**
    * Get selected button.
    * 
    * @return  Selected button.
    */
   public RadioButton<O> getSelected()
   {
      for (RadioButton<O> button : buttons)
      {
         if (button.isSelected())
            return button;
      }
      
      return null;
   }
   
   /**
    * Gets the sub-rectangle which should be visible.
    * 
    * @return   The sub-rectangle which should be visible 
    */
   public Rectangle getVisiblePart()
   {
      if (buttons == null || buttons.isEmpty())
      {
         return new Rectangle(0, 0, config.heightChars, config.widthChars);
      }
      RadioButton<O> button = buttons.get(current);
      return new Rectangle(button.location().y, button.location().x, 
            button.location().y + 1, button.location().x + button.width());
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           Original configuration state.
    */
   @Override
   public void afterConfigUpdate(RadioSetConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);
      
      RadioSetConfig c = config();
      
      setWidthChars(c.widthChars);
      refreshItems();
      
      if (c.horizontal)
      {
         repaint();
      }
   }
   
   /**
    * Called when any of the widget's ancestors gets re-assigned.  It will clear all cached values
    * ({@link #ancestorCache} and {@link #window}) for itself and the entire widget tree starting
    * from this container.
    */
   @Override
   public void reflectAncestorChange()
   {
      super.reflectAncestorChange();
      
      for (RadioButton<O> button : buttons)
      {
         button.reflectAncestorChange();
      }
   }

   /**
    * Disable widget.
    */
   @Override
   protected void disable()
   {
      super.disable();
      
      config.items = Arrays.copyOf(config.items, config.items.length);
      
      for (ControlSetItem csi : config.items)
      {
         csi.setEnabled(false);
      }

      for (RadioButton<O> btn : buttons)
      {
         btn.setEnabled(false);
      }
   }
   
   /**
    * Enable widget.
    */
   @Override
   protected void enable()
   {
      super.enable();
      
      config.items = Arrays.copyOf(config.items, config.items.length);
      
      for (ControlSetItem csi : config.items)
      {
         csi.setEnabled(true);
      }
      
      for (RadioButton<O> btn : buttons)
      {
         btn.setEnabled(true);
      }
   }
   
   /**
    * Refresh widget look using new set of buttons, after the concrete implementation has done any
    * other needed calculations.
    * 
    * @param    selectedId
    *           The selected button ID; -1 if none selected.
    * @param    selectedDisabled
    *           Flag indicating if the selected button is disabled.
    */
   protected void refreshItemsImpl(int selectedId, boolean selectedDisabled)
   {
      ControlSetItem[] items = config.items;
   
      Point pNext = new Point(0, 0);
   
      RadioButton<O> selected = null;
      
      double bwidth = -1.0;
      if (config.horizontal && config.expand)
      {
         bwidth = getButtonsExpandWidth(items);
      }
   
      double listWidthChars = getListWidthChars();
      double maxSize = config.horizontal ? config.fixedWidth ? config.widthChars 
                                                             : listWidthChars
                                         : config.heightChars > 0.0 ? config.heightChars
                                                                    : config.listHeightChars;
      double size = 0.0;
      double hExtraDist = getButtonsExtraDistance();
      double maxButtonWidth = 0;
      for (int i = 0; i < items.length; i++)
      {
         RadioButton<O> button = buttons.get(i);
         
         if (bwidth >= 0.0)
            button.setWidthChars(bwidth);
         
         maxButtonWidth = Math.max(maxButtonWidth, button.width());
         size += config.horizontal ? button.width() : button.height();
         
         // TODO: In GUI the button sizes are not calculate correctly and so it may
         //       happen that the button does not fit by a pixel or two. This is a 
         //       temporary fix to let the buttons be displayed but the correct
         //       solution will be to get the size calculation right.
         if (screen().isChui())
         {
            if (size > maxSize)
            {
               break;
            }
         }
   
         button.setEnabled(items[i].isEnabled());
         button.setItemId(items[i].getItemId());
         
         button.setLocation(pNext);
   
         if (config.horizontal)
         {
            size += hExtraDist;
            pNext.x += button.width() + hExtraDist;
         }
         else
            pNext.y += button.height();
         
         if (items[i].getItemId() == selectedId)
         {
            selected = button;
            config.initial = i;
         }
      }
      
      boolean hasEnabled = selected != null && selected.isEnabled();
      if (!hasEnabled) 
      {
         RadioButton<O> candidate = null;
         int candidateIdx = -1;
         for (int i = 0; i < items.length; i++)
         {
            if (buttons.get(i).isEnabled())
            {
               hasEnabled = true;
               candidate = buttons.get(i);
               candidateIdx = i;
               break;
            }
         }
   
         if (candidate != null)
         {
            current = candidateIdx;
            if (selected == null || !selectedDisabled)
            {
               selected = candidate;
               config.initial = candidateIdx;
            }
         }
      }
   
      if (selected != null)
      {
         selected.setSelected(true);
      }
   
      if (config.autoResize)
      {
         if (!config.horizontal)
         {
            listWidthChars = maxButtonWidth;
         }
         else
         {
            listWidthChars = size;
         }
         
         setSize(new Dimension(listWidthChars, getListHeightChars()));
      }
   
      // update widget width/height if necessary
      if (!config.fixedWidth || config.widthChars == 0.0)
      {
         config.widthChars = width();
      }
      if (!config.fixedHeight || config.heightChars == 0.0)
      {
         config.heightChars = height();
      }
   
      // and repaint
      super.setEnabled(hasEnabled);
      repaint();
   }

   /**
    * Select specified button and remove current selection.
    * 
    * @param   button
    *          Button to select.
    */
   protected void selectButton(RadioButton<O> button)
   {
      if (button == null)
         return;
      
      RadioButton<O> selected = getSelected();
      
      if (selected != null)
         selected.setSelected(false);
      
      button.setSelected(true);
      
      repaint();
      requestSync();
   }
   
   /**
    * Gets button width for expand mode.
    * 
    * @param    items
    *           Radio-button set to check.
    * 
    * @return   The UI dependent button width for expand mode. 
    */
   protected double getButtonsExpandWidth(ControlSetItem[] items)
   {
      if (items.length == 0)
      {
         return 0;
      }
      
      double wRet = -1.0;
      
      for (ControlSetItem item: items)
      {
         character label = item.getLabel();
         if (label != null)
         {
            wRet = Math.max(wRet, label.toStringMessage().length());
         }
      }
      
      return wRet + 3.0; // "( )".length()
   }
   
   /**
    * Gets extra distance between radio-buttons.
    * 
    * @return   The UI dependent distance between buttons. 
    */
   protected double getButtonsExtraDistance()
   {
      return 1.0;
   }
   
   /**
    * Gets the list width in character units for radio-set.
    * 
    * @return   The UI dependent width for item list. 
    */
   protected double getListWidthChars()
   {
      return config.listWidthChars;
   }
   
   /**
    * Gets the list height in character units for radio-set.
    * 
    * @return   The UI dependent height for item list. 
    */
   protected double getListHeightChars()
   {
      return config.listHeightChars;
   }
}
