/*
** Module   : Rectangle.java
** Abstract : Rectangular area descriptor.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 CA  20140729 Added size, difference and hashCode APIs.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 HC  20140924 Hynek placeholder text
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20150429 Fixed computing the rectangle from a location and dimension (in GUI, the base 
**                  units must be relative to pixels).
** 008 HC  20150803 Added width() and height() methods.
** 009 HC  20150913 Added toString().
** 010 HC  20170629 Added difference().
** 011 SBI 20171208 Changed intersection(.), intersects(.) and union(.) methods to consider the
**                  case if this object or its parameter are represented by empty rectangles.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.Coordinate;

import java.util.*;

/**
 * Rectangular area descriptor.
 * All coordinate values are in character units. 
 */
public class Rectangle
{
   /** Rectangle bottom row. */
   private final double bottom;

   /** Rectangle left column. */
   private final double left;

   /** Rectangle right column. */
   private final double right;

   /** Rectangle top row. */
   private final double top;

   /**
    * Constructor.
    * 
    * @param   top
    *          Top row.
    * @param   left
    *          Left column.
    * @param   bottom
    *          Bottom row.
    * @param   right
    *          Right column.
    */
   public Rectangle(double top, double left, double bottom, double right)
   {
      this.top = Coordinate.scale(top);
      this.left = Coordinate.scale(left);
      this.bottom = Coordinate.scale(bottom);
      this.right = Coordinate.scale(right);
   }

   /**
    * Constructor.
    * 
    * @param   topleft
    *          Top-left corner.
    * @param   size
    *          Rectangle dimension.
    * @param   baseUnits
    *          The base units for this rectangle.
    */
   public Rectangle(Point topleft, Dimension size, Point baseUnits)
   {
      top = topleft.y;
      left = topleft.x;
      
      bottom = Coordinate.scale(top + size.height - baseUnits.y);
      right = Coordinate.scale(left + size.width - baseUnits.x);
   }

   /**
    * Copy constructor.
    * 
    * @param   rect
    *          Source rectangle.
    */
   public Rectangle(Rectangle rect)
   {
      top = rect.top();
      left = rect.left();
      bottom = rect.bottom();
      right = rect.right();
   }

   /**
    * Get bottom coordinate of this rectangle.
    * 
    * @return  bottom.
    */
   public double bottom()
   {
      return bottom;
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   x
    *          X coordinate.
    * @param   y
    *          Y coordinate.
    * 
    * @return <code>true</code> if coordinates are inside rectangle.
    */
   public boolean contains(double x, double y)
   {
      return ((x >= left) && (x <= right) && (y >= top) && (y <= bottom));
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   p
    *          Point to check.
    *          
    * @return  <code>true</code> if specified point is inside rectangle.
    */
   public boolean contains(Point p)
   {
      return contains(p.x, p.y);
   }

   /**
    * Returns <code>true</code> is specified rectangle is completely inside
    * this rectangle.
    * 
    * @param   r
    *          Rectangle to check.
    *          
    * @return  <code>true</code> if specified rectangle is inside this
    *          rectangle.
    */
   public boolean contains(Rectangle r)
   {
      return ((r.left >= left) && (r.right <= right) && (r.top >= top) && (r.bottom <= bottom));
   }

   /**
    * Check if rectangle is empty.
    * 
    * @return  <code>true</code> if rectangle is empty.
    */
   public boolean empty()
   {
      return top > bottom || right < left;
   }

   /**
    * Check if this rectangle is equal to specified one.
    * 
    * @param   rect
    *          Rectangle to check.
    * 
    * @return  <code>true</code> if rectangles are equal.
    */
   public boolean equals(Rectangle rect)
   {
      if (rect == null)
         return false;
      if (top != rect.top)
         return false;
      if (bottom != rect.bottom)
         return false;
      if (left != rect.left)
         return false;
      if (right != rect.right)
         return false;
      return true;
   }
   
   /**
    * Compute the hashcode of this rectangle.
    * 
    * @return   A hashcode uniquely describing this rectangle.
    */
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      long temp;
      temp = Double.doubleToLongBits(bottom);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      temp = Double.doubleToLongBits(left);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      temp = Double.doubleToLongBits(right);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      temp = Double.doubleToLongBits(top);
      result = prime * result + (int) (temp ^ (temp >>> 32));
      return result;
   }

   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (getClass() != obj.getClass())
         return false;
      Rectangle other = (Rectangle) obj;
      if (Double.doubleToLongBits(bottom) != Double.doubleToLongBits(other.bottom))
         return false;
      if (Double.doubleToLongBits(left) != Double.doubleToLongBits(other.left))
         return false;
      if (Double.doubleToLongBits(right) != Double.doubleToLongBits(other.right))
         return false;
      if (Double.doubleToLongBits(top) != Double.doubleToLongBits(other.top))
         return false;
      return true;
   }

   /**
    * Calculate rectangle which is intersection between this and specified
    * rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    *          
    * @return  rectangle which is an intersection between this and specified
    *          rectangle or <code>null</code> if there is no intersection.
    */
   public Rectangle intersection(Rectangle rect)
   {
      if (empty() || rect.empty())
      {
         return null;
      }
      
      if (top > rect.bottom)
         return null;
      if (bottom < rect.top)
         return null;
      if (left > rect.right)
         return null;
      if (right < rect.left)
         return null;

      return new Rectangle((top > rect.top) ? top
                                           : rect.top,
                           (left > rect.left) ? left
                                             : rect.left,
                           (bottom < rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right < rect.right) ? right
                                               : rect.right);
   }

   /**
    * Check if this and specified rectangle have intersection.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  <code>true</code> if rectangles have intersection.
    */
   public boolean intersects(Rectangle rect)
   {
      if (empty() || rect.empty())
      {
         return false;
      }
      
      if (rect.left > right)
         return false;
      if (rect.right < left)
         return false;
      if (rect.top > bottom)
         return false;
      if (rect.bottom < top)
         return false;
      return true;
   }

   /**
    * Create new rectangles which is the difference between this and the specified rectangle.
    *
    * @param   rect
    *          Removed rectangle.
    * @param   baseUnits
    *          The base measurement units.
    *
    * @return  All the rectangles which are part of this rectangle but not part of the specified
    *          rectangle.
    */
   public List<Rectangle> difference(Rectangle rect, Point baseUnits)
   {
      List<Rectangle> res = new ArrayList<>(4);

      Rectangle intersection = this.intersection(rect);
      if (intersection == null)
      {
         // nothing to remove, return this
         res.add(this);

         return res;
      }

      // at this point, is just a matter of removing the intersection
      rect = intersection;

      if (rect.top - baseUnits.y >= top)
      {
         // rectangle found above the intersection
         res.add(new Rectangle(top, left, rect.top - baseUnits.y, right));
      }

      if (rect.bottom + baseUnits.y <= bottom)
      {
         // rectangle found bellow the intersection
         res.add(new Rectangle(rect.bottom + baseUnits.y, left, bottom, right));
      }

      if (rect.left - baseUnits.x >= left)
      {
         // rectangle found on the left-side of the intersection
         res.add(new Rectangle(rect.top, left, rect.bottom, rect.left - baseUnits.x));
      }

      if (rect.right + baseUnits.x <= right)
      {
         // rectangle found on the right-side of the intersection
         res.add(new Rectangle(rect.top, rect.right + baseUnits.x, rect.bottom, right));
      }

      return res;
   }

   /**
    * Get left coordinate of this rectangle.
    * 
    * @return  left.
    */
   public double left()
   {
      return left;
   }

   /**
    * Get right coordinate of this rectangle.
    * 
    * @return  right.
    */
   public double right()
   {
      return right;
   }

   /**
    * Get top coordinate of this rectangle.
    * 
    * @return top.
    */
   public double top()
   {
      return top;
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   dx
    *          Delta X.
    * @param   dy
    *          Delta Y.
    * 
    * @return  translated rectangle.
    */
   public Rectangle translate(double dx, double dy)
   {
      return new Rectangle(top + dy, left + dx, bottom + dy, right + dx);
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   p
    *          Offset represented as a point.
    *           
    * @return  translated rectangle.
    */
   public Rectangle translate(Point p)
   {
      return translate(p.x, p.y);
   }

   /**
    * Create new rectangle which is a union of this and specified rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  union rectangle.
    */
   public Rectangle union(Rectangle rect)
   {
      return new Rectangle((top < rect.top) ? top
                                           : rect.top,
                           (left < rect.left) ? left
                                             : rect.left,
                           (bottom > rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right > rect.right) ? right
                                               : rect.right);
   }
   
   /**
    * Returns the rectangle's width.
    * <p> 
    * Because the rectangle's coordinates are inclusive and
    * base units differ in ChUI and GUI, they must be provided
    * in the only parameter.
    * 
    * @param    baseUnits
    *           Reference to {@link Point} instance representing the
    *           base units.
    * 
    * @return   See above.
    */
   public double width(Point baseUnits)
   {
      return right - left + baseUnits.x;
   }
   
   /**
    * Returns the rectangle's height.
    * <p> 
    * Because the rectangle's coordinates are inclusive and
    * base units differ in ChUI and GUI, they must be provided
    * in the only parameter.
    *
    * @param    baseUnits
    *           Reference to {@link Point} instance representing the
    *           base units.
    * 
    * @return   See above.
    */
   public double height(Point baseUnits)
   {
      return bottom - top + baseUnits.y;
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return  See above.
    */
   @Override
   public String toString()
   {
      return "Rectangle[top=" + top + ", left=" + left + ", bottom=" + bottom + ", right="
               + right + "]";
   }
}