/*
** Module   : TempContainer.java
** Abstract : Temporary container used for layout processing.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100915 Created initial version
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 SIY 20111221 Implemented screen() method.
** 005 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 HC  20150323 Removed business logic from config classes.
** 008 CA  20150429 Added NPE protection for screen().
** 009 CA  20150509 Allow scroll container for any widgets, not just frames.
** 010 HC  20150831 Scrolling support overhaul.
** 011 CA  20151123 Added APIs to place a widget on a specified location.
** 012 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.client.widget.*;

/**
 * Temporary container used for second and subsequent layout processing.
 * <p>
 * ENABLE processing may change order of the widgets in the container. Since
 * order also affects layout processing, original widget order is saved and
 * then all layout processing is done using this special container as a 
 * temporary holder for frame widgets. 
 */
public class TempContainer<O extends OutputManager<?>>
extends SensitiveScrollContainer<O>
{
   /** Source scroll container used to retrieve some useful info. */
   private AbstractContainer<O> base = null;
   
   /**
    * Construct an instance of container using given list of widgets and
    * dimension.
    * 
    * @param    widgets
    *           Array of <code>Widget</code> instances.
    * @param    base
    *           Base container for which this one will work as substitute.
    * @param    frame
    *           instance of <code>FrameHelper</code> for this frame
    * @param    minDim
    *           The minimum dimension for the temp container.
    */
   public TempContainer(Widget<O>[] widgets, 
                        AbstractContainer<O> base, 
                        Frame<O> frame, 
                        Dimension minDim)
   {
      this.base = base;
      this.setParent(base.parent());
      setSize(minDim != null ? minDim : base.dimension());
      
      // note that direct call is used instead of Component.add
      // to prevent changing of the reference to parent
      for (int i = 0; i < widgets.length; i++)
      {
         addDirectly(widgets[i]);
      }
      
      Widget<O>[] c = base.widgets();
      
      for (int i = 0; i < c.length; i++)
      {
         if (c[i] instanceof Label && ((Label<O>) c[i]).isDelimiter())
         {
            addDirectly(c[i]);
         }
      }
   }

   /**
    * Forwarder for add(Component) method.
    * 
    * @param    component
    * 
    * @return   <code>this</code> for fluent syntax support.
    */
   @Override
   public Container<O> add(Widget<O> component)
   {
      return base.add(component);
   }
   

   /**
    * Add the widget on the specified position.
    * 
    * @param    pos
    *           The widget's position.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   @Override
   public Container<O> add(int pos, Widget<O> widget)
   {
      return base.add(pos, widget);
   }

   /**
    * Add the widget before the specified widget.
    * 
    * @param    before
    *           The widget before the new widget will be placed.
    * @param    widget
    *           The child widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    * 
    * @throws  IllegalArgumentException
    *          when the argument is null or instance of {@link TopLevelWindow}
    */
   @Override
   public Container<O> add(Widget<O> before, Widget<O> widget)
   {
      return base.add(before, widget);
   }
   
   /**
    * Forwarder for remove(Component) method.
    * 
    * @param    component
    */
   @Override
   public void remove(Widget<O> component) 
   {
      base.remove(component);
   }
   
   /**
    * Extended implementation of replace method.
    * 
    * @param    oldc
    *           Existing component.
    * @param    newc
    *           Component which will replace old component.
    */
   @Override
   public void replace(Widget<O> oldc, Widget<O> newc)
   {
      // regular action
      base.replace(oldc, newc);
      
      // extension - changing non-body list
      @SuppressWarnings("unchecked")
      Frame<O> frame = parent(Frame.class);
      frame.fixNonBody(oldc, newc);
   }
   
   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public O screen()
   {
      return base == null ? null : base.screen();
   }
}