/*
** Module   : TopLevelWindow.java
** Abstract : A base class for all top-level windows.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20050509 Initial version.
** 002 HC  20150626 Javadoc fixes.
** 003 HC  20150702 Improved parent-child relationship of top-level windows.
** 004 CA  20150829 Fixed repaint() coordinates - must use (0,0), not screen origin.
** 005 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
**                  Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 006 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 007 CA  20151013 Added support for window resize for the GUI web client.
** 008 HC  20151024 Prevented processing of mouse events for non-displayed widgets.
**     CA  20151106 A fix for HIDDEN attribute: determine if any of the ancestors (or self) is 
**                  marked HIDDEN.
** 009 CA  20151127 Fix for ACTIVE-WINDOW: when a window is being activated, the ACTIVE-WINDOW
**                  must reference it regardless of what happens with the old window.
** 010 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 011 HC  20160113 Redrawing window title bar inside of an event-drawing bracket when handling 
**                  WindowActivated event.
**     EVL 20160114 Window title should not be repaining working with overaly window.
** 012 EVL 20160122 Moving activation to native window should cause the setting current active
**                  window internal variable(not ACTIVE-WINDOW handler) to null.  Window title
**                  should be repainted after changing active window, not before. 
** 013 HC  20160203 Improved runtime support of KEEP-FRAME-Z-ORDER and TOP-ONLY attributes.
** 014 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 015 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation().
** 016 SBI 20160310 Changed to reuse the common code
** 017 HC  20160406 Overhaul of window-activation logic. Added close() method.
** 018 VIG 20160307 Moved system menus here from WindowGuiImpl, moved init* and create* methods.
**     SBI 20160310 Changed to reuse the common code
**     VIG 20160401 Fixed currentFocus method.
** 019 HC  20160413 Fixed top-level windows to always receive focus event when activated, also
**                  deduplicated processing of focus events.
** 020 HC  20160421 Focused frame is repainted when the focus of a top-level window changes.
** 021 HC  20160511 Fixed unbalanced selected windows in GUI driver. Fixed COMBO-BOX focusing 
**                  issues and related changes.
** 022 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 023 HC  20160922 Changed FramePlacementManager to operate on the level of TopLevelWindow.
** 024 VIG 20160708 Added window exist checks to processEvent for WINDOW_ACTIVATED event.
**                  added repainting of focused item.
** 025 CA  20161101 Fixed a bug in findFrames() - in GUI, window.getContentPane().widgets() can't 
**                  be used, as the frames are attached to the workspace instance; 
**                  use window.getFrames() instead.
** 026 OM  20170320 Repaint window on focus gained/lost.
** 027 OM  20170407 Repaint window on focus gained/lost clipped to borders/titlebar.
** 028 HC  20170823 Added runtime support of WINDOW:SHOW-IN-TASKBAR.
** 029 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
** 030 CA  20180208 Made reflectLocationChange() public.
** 031 HC  20180420 Implemented popup menu resolution in the tree of widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.FocusEvent;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.google.common.collect.*;

/**
 * The class represents a top-level window, that is a window that sits
 * at the top of widget hierarchy (it has no parent), has its own independent 
 * screen, a possible window caption, etc.
 * <p>
 * A top level window may exist on a "desktop" spanning multiple physical
 * screen devices. These screen devices form a virtual device and common 
 * coordinate system with origin at top-left corner of the primary screen.
 * Depending on the location of the primary screen in the virtual device, 
 * negative coordinates are possible.
 * <p>
 * To determine bounds of the physical screen a window is positioned on
 * call {@link GuiDriver#getDisplayBounds(int)}. These bounds are needed
 * for example when positioning the window relative to the top-left corner
 * of the physical screen. 
 */
public abstract class TopLevelWindow<O extends OutputManager<?>>
extends OuterFrame<O>
{
   /** Data type passing as initial window icon. */
   public static enum IconDataType
   {
      ICON_CUSTOM_JAR_RESOURCE, 
      ICON_CUSTOM_CLIENT_FILENAME, 
      ICON_DEFAULT
   }

   /** Default window icon. */
   private static byte[] iconDef = null;
   
   /** The owner of this window. */
   private Optional<TopLevelWindow<O>> owner = Optional.empty();
   
   /** LEAVE event result after window deactivated. */
   private boolean lastLeaveResult;

   /** Frame placement manager. */
   private FramePlacementManager placementManager = new FramePlacementManager(this);

   /**
    * Default constructor.
    */
   public TopLevelWindow()
   {
      super(new Insets(0, 0, 0, 0));
   }
   
   /**
    * Constructs a window widget with title. 
    * 
    * @param   title
    *          Window title.
    */
   public TopLevelWindow(String title)
   {
      super(title);
   }
   
   /**
    * Returns the owner.
    * 
    * @return   See above.
    */
   public TopLevelWindow<O> getOwner()
   {
      return owner.orElse(null);
   }
   
   /**
    * Returns the owner as {@linkplain Optional}.
    * 
    * @return   See above.
    */
   public Optional<TopLevelWindow<O>> getOwnerOpt()
   {
      return owner;
   }
   
   /**
    * Get the title bar widget associated witn this window.
    * 
    * @return   See above.
    */
   public WindowTitleBar getTitleBar()
   {
      return null;
   }

   /**
    * Sets the default window icon for all new windows on the client side.
    * 
    * @param    type
    *           Data type for second parameter.
    * @param    data
    *           Data to create new icon on client side.  Can be the byte array, filename string
    *           or nothing for predefined icon from p2j.jar client side file. 
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean setWindowDefaultIcon(IconDataType type, Object data)
   {
      switch (type)
      {
         case ICON_CUSTOM_JAR_RESOURCE:
         {
            // getting icon data from server side application jar
            iconDef = (byte[])data;
   
            break;
         }
         case ICON_CUSTOM_CLIENT_FILENAME:
         {
            // find the icon on local file system
            InputStream is = null;
            try
            {
               is = new FileInputStream((String) data);
            }
            catch (FileNotFoundException ex)
            {
               is = null;
            }
   
            // the file is found, try to read
            if (is != null)
            {
               iconDef = UiUtils.getData(is);
               break;
            }
            else
            {
               // we are here - means fallback to the default one, make recursive call
               return setWindowDefaultIcon(IconDataType.ICON_DEFAULT, null);
            }
         }
         case ICON_DEFAULT:
         {
            // loading icon embedded into p2j.jar
            InputStream is = Window.class.getResourceAsStream(WindowConfig.DEFAULT_ICON);
         
            if (is != null)
            {
               iconDef = UiUtils.getData(is);
            }
         
            break;
         }
      }
      
      return iconDef != null;
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      // for the window's coordinate system, the window's location is always (0,0)
      return new Point(0, 0);
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  absolute widget location.
    */
   @Override
   public NativePoint screenPhysicalLocation()
   {
      // for the window's coordinate system, the window's location is always (0,0)
      return new NativePoint(0, 0);
   }
   
   /**
    * Get widget bounding rectangle.
    * <p>
    * For windows, the boundary starts from (0,0) regardless where is placed on screen.
    * 
    * @return   bounding rectangle.
    */
   @Override
   public Rectangle bounds()
   {
      Rectangle bounds = super.bounds();
      
      bounds = bounds.translate(-bounds.left(), -bounds.top());
      
      return bounds;
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * <p>
    * For windows, the boundary starts from (0,0) regardless where is placed on screen.
    * 
    * @return   See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      NativeRectangle bounds = super.physicalBounds();
      
      bounds = bounds.translate(-bounds.left(), -bounds.top());
      
      return bounds;
   }
   
   /**
    * Repaints window's decorations (titlebar with its components and window's border).
    */
   public void repaintDecorations()
   {
      CoordinatesConversion c = screen().coordinates();
      WindowTitleBar tbar = getTitleBar();
      NativeInsets insets = ThemeManager.getCurrentTheme().getWindowBorderInsets(isModal());
      double bN = c.rowFromNative(insets.top) + ((tbar == null) ? 0 : tbar.height());
      double bS = c.rowFromNative(insets.bottom);
      double bE = c.columnFromNative(insets.right);
      double bW = c.columnFromNative(insets.left);
      double ww = width();
      double wh = height();
      // gd.startBatch()
      repaint(new Rectangle(0,       0,       bN,      ww)); // N
      repaint(new Rectangle(wh - bS, 0,       wh,      ww)); // S
      repaint(new Rectangle(bN,      0,       wh - bS, bW)); // W
      repaint(new Rectangle(bN,      ww - bE, wh - bS, ww)); // E
      // gd.endBatch()
   }
   
   /**
    * Get window content pane.
    * 
    * @return  content pane.
    */
   public abstract Container<O> getContentPane();
   
   // TODO: review the usage of this method. it is abused by combobox to 
   //       add its drop down to the window's content pane. 
   /**
    * Adds the frame to the window at the specified position.
    * 
    * @param    widget
    *           The handle of the frame to be added to the window.
    * @param    row
    *           The row number for the upper-left frame corner.
    * @param    column
    *           The column number for the upper-left frame corner.
    */
   public void addFrame(Widget<O> widget, double row, double column)
   {
      getContentPane().add(widget);
   }
   
   /**
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   public abstract Dimension getScreenDimension();
   
   /**
    * Returns the frames contained in this top-level window. If no frames
    * then the result is an empty list.
    * 
    * @return  See above.
    */
   public List<Frame<O>> getFrames()
   {
      Container<O> content = getContentPane();
      ArrayList<Frame<O>> result = new ArrayList<>(content.widgets().length);
      
      for (Widget<O> w : content.widgets())
      {
         if (w instanceof Frame)
         {
            result.add((Frame<O>) w);
         }
      }
      
      return result;
   }

   /**
    * Gets the frame from the list of currently displayed frames that
    * immediately follows the specified frame.
    * 
    * @param   frame
    *          frame that designates the starting point 
    * @return  the next frame in the list or <code>null</code>
    */
   public Frame<?> getNextVisibleFrame(Frame<?> frame)
   {
      List<Frame<O>> items = getFrames();
      
      int nref = 0;
      
      // scan the list of top level items for the reference point
      for (nref = 0; nref < items.size(); nref++)
      {
         if (items.get(nref).equals(frame))
         {
            break;
         }
      }
   
      if (nref == items.size())
      {
         // reference point not found
         return null;
      }
   
      // scan the list of top level items and find the closest visible frame
      for (int i = nref + 1; i < items.size(); i++)
      {
         if (items.get(i).isVisible())
         {
            return (Frame<?>) items.get(i);
         }
      }
   
      // no visible frames
      return null;
   }

   /**
    * Gets the first frame from the list of currently displayed frames.
    * 
    * @return  the first frame in the list or <code>null</code>
    */
   public Frame<?> getFirstVisibleFrame()
   {
      List<Frame<O>> frames = getFrames();
      
      Frame<?> firstVisible = null;
      Frame<?> lastDialog   = null;
      
      // scan the list of top level items and find the first visible frame
      // if there is a dialog box, find and return the one on the top
      for (Frame<O> frame : frames)
      {
         if (frame.isVisible())
         {
            if (frame.isDialog())
            {
               lastDialog = frame;
            }
            else
            {
               if (firstVisible == null)
               {
                  firstVisible = frame;
               }
            }
         }
      }
   
      // return the last dialog if found or else the first visible frame
      if (lastDialog != null)
      {
         return lastDialog;
      }
   
      return firstVisible;
   }

   /**
    * Returns the frame that is on top of the Z-order or <code>null</code>
    * if no such frame present. If requested method also can filter out all
    * invisible, disabled and out of scope frames.
    * 
    * @param   ignoreState
    *          if <code>true</code> then ignore frame state during search.
    * 
    * @return  The topmost active frame or <code>null</code>.
    */
   public Frame<?> getTopFrame(boolean ignoreState)
   {
      List<Frame<O>> frames = Lists.reverse(getFrames());
   
      // the components list is ordered by Z-order such that the first entry
      // is at the bottom and the last entry is at the top, so search "down"
      // by walking backwards through the array
      for (Frame<O> frame : frames)
      {
         if (ignoreState)
            return frame;

         // avoid all cases where a frame is not in scope or is otherwise
         // not a valid frame to be used
         if (!frame.isVisible() || !frame.isEnabled() || !frame.isInScope())
         {
            continue;
         }

         // we are done
         return frame;
      }
      
      return null;
   }

   /**
    * Get the first enabled, visible frame, in z-order, after the specified
    * frame. The found frame must have at least one enabled, visible widget.
    * <p>
    * If no frame is specified, it will get the bottommost enabled and visible
    * frame.
    * 
    * @param   frame
    *          Reference frame. When <code>null</code>, it will return the
    *          bottommost enabled and visible frame.
    * 
    * @return  Next enabled frame, above this frame, in z-order, or null if no
    *          such frame exists.
    */
   public Frame<?> getNextEnabledFrame(Frame<?> frame)
   {
      List<Frame<O>> frames = Lists.reverse(getFrames());
      Frame<?> found = null;
   
      // the components list is ordered by Z-order such that the first entry
      // is at the bottom and the last entry is at the top, so search "down"
      // by walking backwards through the array
      for (Frame<O> frm : frames)
      {
         if (frame == frm)
            break;

         if (frm.isEnabled() && frm.isVisible() && frm.isAlive())
         {
            int wid = Frame.getFirstFocusableId(frm);
            if (wid != -1)
            {
               found = frm;
            }
         }
      }
      
      return found;
   }

   /**
    * Get the first enabled, visible frame, in z-order, before the specified
    * frame. The found frame must have at least one enabled, visible widget.
    * <p>
    * If no frame is specified, it will get the bottommost enabled and visible
    * frame.
    * 
    * @param   frame
    *          Reference frame. When <code>null</code>, it will return the
    *          topmost enabled and visible frame.
    * 
    * @return  Previous enabled frame, below this frame, in z-order, or null if
    *          no such frame exists.
    */
   public Frame<?> getPrevEnabledFrame(Frame<?> frame)
   {
      List<Frame<O>> frames = Lists.reverse(getFrames());
      Frame<?> found = null;
      
      // the components list is ordered by Z-order such that the first entry
      // is at the bottom and the last entry is at the top, so search "up"
      // by walking foreword through the array
      for (Frame<O> frm : frames)
      {
         if (frame == frm)
            break;

         if (frm.isEnabled() && frm.isVisible() && frm.isAlive())
         {
            int wid = Frame.getFirstFocusableId(frm);
            if (wid != -1)
            {
               found = frm;
            }
         }
      }
      
      return found;
   }

   /**
    * Get top visible frame.
    * 
    * @return   reference to top visible frame or <code>null</code> if there is
    *           no such frame.
    */
   public Frame<?> getTopVisibleFrame()
   {
      List<Frame<O>> frames = Lists.reverse(getFrames());
      
      // the components list is ordered by Z-order such that the first entry
      // is at the bottom and the last entry is at the top, so search "down"
      // by walking backwards through the array
      for (Frame<O> frame : frames)
      {
         if (!frame.isVisible())
            continue;

         // we are done
         return frame;
      }
      
      return null;
   }
   
   /**
    * Sets the new application icon for given window on the client side.
    * 
    * @param    newIcon
    *           A new icon image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setWindowIcon(byte[] newIcon)
   {
      // this is generic and CHUI version returning false
      return false;
   }

   /**
    * Loads the new application icon for given window from the client side filesystem.
    * 
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    * @param    name
    *           The fully qualified file name for the icon to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadWindowIcon(boolean small, String name)
   {
      // this is generic and CHUI version returning false
      return false;
   }

   /**
    * Sets the default window icon for all new windows on the client side.
    * 
    * @return   The byte array to construct the default window icon.
    */
   public static byte[] getWindowDefaultIcon()
   {
      return iconDef;
   }
   
   /**
    * Check if the window is already maximized - it can't be resized.
    * 
    * @return   This implementation always returns <code>false</code>.
    */
   public boolean isMaximized()
   {
      return false;
   }

   /**
    * Check if the window is minimized.
    *
    * @return   If minimized returns <code>true</code>, <code>false</code> otherwise.
    */
   public boolean isMinimized()
   {
      return false;
   }

   /**
    * Resolve the font associated with the window's title.
    * 
    * @return   See above.
    */
   @Override
   public int resolveTitleFont()
   {
      return FontManager.WINDOW_TITLE_FONT;
   }
   
   /**
    * Get ancestor Window for this widget. This method must be called
    * only for widgets that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @Override
   public Window<O> window()
   {
      if (!owner.isPresent())
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
                                         + " is not attached to a Window instance.");
      }
      
      return owner.get().window();
   }
   
   /**
    * Get ancestor window for this widget.This method must be called
    * only for widgets that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @Override
   public TitledWindow<O> ancestor()
   {
      if (!owner.isPresent())
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
                                         + " is not attached to a Window instance.");
      }

      return owner.get().ancestor();
   }

   /**
    * For windows, repaint must use the window's top-left corner (0,0) as origin, not the 
    * screen's.
    */
   @Override
   public void repaint()
   {
      if (!isDisplayed())
         return;

      Rectangle bounds = new Rectangle(0, 0, height(), width());
      
      repaint(bounds);
   }
   
   /**
    * Replacement method for {@link AbstractContainer#defaultFocus()} 
    * implementation which takes into account that frames are listed in
    * reverse order by the means of Z-order (i&#46;e&#46; topmost window is
    * last in the list).
    */
   @Override
   public Widget<O> defaultFocus()
   {
      for (Frame<O> f : getFrames())
      {
         if (f.focusTraversable())
         {
            return f.defaultFocus();
         }
      }
      
      return null;
   }
   
   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }
   
   /**
    * Check if the current window or one of its ancestors is hidden.
    * 
    * @return   <code>true</code> if this window or one of its ancestors is hidden.
    */
   public boolean parentOrSelfHidden()
   {
      TopLevelWindow parent = this;
      while (parent != null)
      {
         if (parent.config() != null && parent.config().hidden)
         {
            return true;
         }
         
         BaseConfig cfg = UiUtils.getConfig(parent, BaseConfig.class);
         parent = (cfg == null ? null : WindowManager.findWindow(cfg.parentId));
      }
      
      return false;
   }
   
   /**
    * The method hides the window if visible and then destroys it by the means of
    * calling {@link #destroy()} method. Both operations are enqueued on the OS
    * event queue, this allows all pending events to be processed before the
    * window is closed. 
    */
   public void close()
   {
      ThinClient.getInstance().invokeLaterOS(() ->
      {
         if (isVisible())
         {
            hide();
         }
         
         destroy();
      });
   }
   
   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processFocusEvent(FocusEvent event)
   {
      if (event.id() == EventType.FOCUS_GAINED)
      {
         WindowManager.setFocusWindow(this);
      }
      
      Widget<?> focus = UiUtils.getCurrentFocus();
      
      if (focus != null)
      {
         Frame<?> frame = UiUtils.locateRootFrame(focus);
         if (frame != null)
         {
            frame.repaint();
         }
      }
      
      super.processFocusEvent(event);
   }

   /**
    * Collect all frames which are intersected with specified rectangle into
    * provided list.
    *
    * @param   list
    *          List where frames will be stored.
    * @param   rect
    *          Bounding rectangle.
    * @param   full
    *          If <code>true</code> collect all frames. Otherwise frames
    *          with OVERLAY option are excluded.
    * @param   redirected
    *          If in redirected mode, return only redirected frames. Else,
    *          return only the interactive frames.
    */
   public void findFrames(List<Frame<?>> list, Rectangle rect, boolean full, boolean redirected)
   {
      findFrames(list, rect, full, redirected, false);
   }

   /**
    * Collect all frames which are intersected with specified rectangle into
    * provided list.
    *
    * @param   list
    *          List where frames will be stored.
    * @param   rect
    *          Bounding rectangle.
    * @param   full
    *          If <code>true</code> collect all frames. Otherwise frames
    *          with OVERLAY option are excluded.
    * @param   redirected
    *          If in redirected mode, return only redirected frames. Else,
    *          return only the interactive frames.
    * @param   special
    *          Enable special handling of empty NO-BOX frames.
    */
   public void findFrames(List<Frame<?>> list,
                          Rectangle      rect,
                          boolean        full,
                          boolean        redirected,
                          boolean        special)
   {
      List<Frame<O>> c = getFrames();

      for (int i = 0; i < c.size(); i++)
      {
         if (c.get(i) instanceof Frame)
         {
            Frame<?> frm = c.get(i);
            FrameConfig cfg = frm.config();

            if (!frm.isVisible())
               continue;

            if (cfg.overlay && !full)
               continue;

            Rectangle r = frm.bounds();

            // special handling of empty frames with NO-BOX option
            if (special  &&
                    !cfg.box &&
                    (r.right() - r.left()) <= 0 &&
                    (r.bottom() - r.top()) >= 0)
            {
               r = new Rectangle(r.top(), r.left(),
                       r.bottom() + 1, r.right() + 1);
            }

            if (rect.intersects(r) && frm.isRedirected() == redirected)
               list.add(frm);
         }
      }
   }

   /**
    * Get the {@link FramePlacementManager} associated with this window.
    *
    * @return   The {@link #placementManager}.
    */
   public FramePlacementManager getPlacementManager()
   {
      return this.placementManager;
   }

   /**
    * Check the window's {@link #lastLeaveResult} state.
    * 
    * @return   See above.
    */
   public boolean isLastLeaveResult()
   {
      return lastLeaveResult;
   }

   /**
    * Set the window's {@link #lastLeaveResult} state.
    * 
    * @param    lastLeaveResult
    *           The flag's state.
    */
   public void setLastLeaveResult(boolean lastLeaveResult)
   {
      this.lastLeaveResult = lastLeaveResult;
   }

   /**
    * Called when the widget's location changes. This allows the subclasses to react
    * on this, for example by notifying the driver.
    */
   @Override
   public void reflectLocationChange()
   {
      if (!screen().isChui())
      {
         // fetch the location in case there have been any adjustments
         // by the super implementations
         NativePoint p = physicalLocation();

         GuiDriver<?,?> gd = (GuiDriver<?,?>) screen().getInstanceDriver();
         gd.selectWindow(getId().asInt());
         gd.setWindowLocation(p.x, p.y);
         gd.releaseWindow();
      }
   }

   /**
    * Sets the owner window.
    * 
    * @param    owner
    *           Reference to the owner window.
    */
   protected void setOwner(TopLevelWindow<O> owner)
   {
      this.owner = Optional.ofNullable(owner);
   }
   
   /**
    * Build widget clipped rectangle.
    * <p>
    * For top-level windows, their boundary always starts from <code>(0,0)</code>, and not their
    * physical location on screen.
    * 
    * @return  see above.
    */
   @Override
   protected NativeRectangle clipRectangle()
   {
      NativeDimension dim = physicalDimension();
      return new NativeRectangle(0, 0, dim.height, dim.width);
   }
}
