/*
** Module   : FileDialogImpl.java
** Abstract : abstracts common code of dialogs that allows the user to select a file form local
**            disk in either GUI or ChUI modes 
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170815 Initial commit. Just a stub with mandatory methods. Work in progress.
** 002 OM  20171027 Added filters validation.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;

import java.io.*;

/**
 * Abstracts common code of dialogs that allows the user to select a file form local disk in
 * either GUI or ChUI modes. Note that these both dialogs are rendered using GUI driver!
 */
public class FileDialogImpl
extends AbstractFileChooserDialog
implements ActionListener
{
   /** The preferred title, if any set by ABL programmer. */
   protected String preferredTitle = null;
   
   /** Flags "Save As" dialogs. In this case the selected file may not exists. */
   protected boolean saveAs = false;
   
   /**
    * The dialog will asl whether the user allows the file to be overwritten. Only in case of
    * save-as dialogs.
    */
   protected boolean askOverwrite = false;
   
   /**
    * The dialog creates a temporary file at the chosen location to validate the write access.
    * Only in case of save-as dialogs.
    */
   protected boolean createTestFile = false;
   
   /** The dialog does not allow to select files that do not exist. */
   protected boolean mustExist = false;
   
   /** The set of filters, with their labels. The array must have an even size. */
   protected String[] filters = null;
   
   /** The index of initial filter to be preselected when the dialog opens. */
   protected int initialFilter = 0;
   
   /** The default extension for the selected files. */
   private String defaultExtension = null;
   
   /**
    * The constructor build an object by forwarding the required parameters to super class. 
    *
    * @param   id
    *          The {@link WidgetId} of the dialog. 
    * @param   cfg
    *          The {@link FrameConfig} of the dialog.
    */
   public FileDialogImpl(WidgetId id, FrameConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Configures the filters in the current dialog.
    * 
    * @param   filters
    *          The set of filters, with their labels. The array must have an even size.
    * @param   initialFilter
    *          The index of initial filter to be preselected when the dialog opens.
    * @param   defaultExtension
    *          The default extension for the selected files.
    */
   public void setFilters(String[] filters, int initialFilter, String defaultExtension)
   {
      // quick validation:
      if (filters != null && (filters.length & 0x01) == 1)
      {
         throw new RuntimeException("The filters array must contain both label and pattern.");
      }
      
      this.filters = filters;
      this.initialFilter = initialFilter;
      this.defaultExtension = defaultExtension;
   }
   
   /**
    * Configures additional constraints for the file to be selected.
    * 
    * @param   saveAs
    *          {@code true} if this is gonna be a save-as dialog.
    * @param   askOverwrite
    *          If {@code true} the dialog ask user permission to return an existing file (that can
    *          be later overwritten programmatically).
    * @param   createTestFile
    *          If {@code true} the dialog checks the write access right by creating a temporary
    *          file. If this fail, the user is not allowed to select a file in that location. 
    * @param   mustExist
    *          If {@code true} the user is only allowed to select a file that already exist.
    */
   public void setValidators(boolean saveAs,
                             boolean askOverwrite,
                             boolean createTestFile,
                             boolean mustExist)
   {
      if (preferredTitle == null)
      {
         // use the default titles if none specified in constructor
         config().title = saveAs ? "Save As" : "Open";
      }
      this.saveAs = saveAs;
      this.askOverwrite = askOverwrite;
      this.createTestFile = createTestFile;
      this.mustExist = mustExist;
   }
   
   /**
    * Sets the value to be returned by the dialog. Must be absolute path. If {@code path} do not
    * have an extension, the {@code defaultExtension} is set, if defined.
    * 
    * @param   path
    *          The full path of the selected file. 
    */
   @Override
   public void setSelectedPath(String path)
   {
      // System.out.println("Prepared to return [" + path + "].");
      if (path != null && defaultExtension != null && !defaultExtension.isEmpty())
      {
         File f = new File(path);
         String name = f.getName();
         int k = name.indexOf('.');
         if (k == -1)
         {
            // no extension detected? add the default extension
            path += (defaultExtension.startsWith(".") ? "" : ".") + defaultExtension;  
         }
      }
      this.selectedPath = path;
   }
   
   /**
    * This method is invoked when action event is generated.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
   }
}
