/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150226 Added processKeyEvent method.
**                  Implemented draw, location methods.
** 004 VIG 20150618 Moved setVisible to base class.
** 005 VIG 20151020 Force set MENUBAR width to window's width.
** 006 HC  20180425 Fixed an NPE during menu initialization.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.driver.ChuiOutputManager;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Menu</code> CHUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class MenuChuiImpl
extends Menu<ChuiOutputManager>
{
   /** Border around the panel. */
   private Border<ChuiOutputManager> border = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuChuiImpl(WidgetId id, MenuConfig cfg)
   {
      super(id, cfg);
      
      if (cfg.popupOnly)
      {
         border = new LineBorder();
      }
   }
   
   /**
    * Drawing worker method.
    */
   public void drawPanel()
   {
      screen().clear(WidgetId._DEFAULT_WINDOW_ID, 
               location(), 
               dimension(false), 
               screen().getInteractiveOutput().widgetColor(this));

      if (border != null)
         border.drawBorder(this, dimension(false));
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      drawPanel();
      super.draw();
   }

   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   @Override
   public Point childrenLocation(Widget<ChuiOutputManager> child)
   {
      Point bodyLoc = location();
      double x = config().popupOnly ? bodyLoc.x + 1 : bodyLoc.x; 
      return new Point(x, bodyLoc.y); // border
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      return dimension(true);
   }
   
   /**
    * Lay out container. Set children locations.
    */
   @Override
   public void doLayout()
   {
      setLocation(calcLocation());
      super.doLayout();
      
      Widget<ChuiOutputManager>[] w = widgets();
      double x=0;
      
      for(int i=0; i<w.length; i++)
      {
         if (!config().popupOnly) // menu-bar
         {
            w[i].setLocation(new Point(x, 0));
            x += 1 + w[i].width(); // shift horizontally and add space
         }
         else
         {
            w[i].setLocation(new Point(1, i + 1)); // borders
         }
         
         if(w[i] instanceof SubMenu)
         {
            ((SubMenu<?>)w[i]).doLayout();
         }
      }
   }
   
   /**
    * Looks up the very first enabled widget.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   @Override
   public Widget<ChuiOutputManager> getFirstEnabledWidget()
   {
      return this;
   }
   
   /**
    * Calculates MENU location relative to its parent if it is POPUP-ONLY or 
    * [0, 0] if it is MENUBAR.
    * 
    * @return   point of location relative to its parent.
    */
   private Point calcLocation()
   {
      if (config().popupOnly)
      {
         Widget<?> owner = screen().getRegistry().getComponent(config().ownerId);
         if (owner != null)
         {
            AbstractWidget<?> awOwner = (AbstractWidget<?>) owner;
            Point ownloc = awOwner.location();
            
            // pop-up menu is shown right under owner, shifted by 1 horizontally 
            return new Point(ownloc.x + 1, ownloc.y + awOwner.height() + 1);
         }
      }
      
      return new Point(0, 0); // menu-bar is always on top
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    * 
    * @param   includeNested
    *          <code>true</code> if we should process nested sub-menus
    */
   private Dimension dimension(boolean includeNested)
   {
      double height = SubMenuChuiImpl.getMaxItemHeight(this, includeNested);
      double width = SubMenuChuiImpl.getMaxItemWidth(this, includeNested);
      
      if (!config().popupOnly)
      {
         /*
         double menubarWidth = 0; 
         for(Widget<?> w : widgets())
         {
            AbstractWidget<?> aw = (AbstractWidget<?>) w;
            menubarWidth += aw.width() + 1; // space between items in a menu-bar
         }
         width = Math.max(menubarWidth, width);
         */

         // window width, get the reference from window manager, the menu may not have been initialized
         width = WindowManager.getDefaultWindow().width();
      }
      
      return new Dimension(width + 4, height); // left and right borders with insets
   }
}