/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150225 Added event processing and drawing methods.
**                  Implemented draw, location methods.
** 004 VIG 20150618 Moved some methods to base class.
** 005 VIG 20150905 Changed title drawing using new mnemonic logic, 
**                  Added accelerators drawing support.
** 006 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 007 VIG 20160222 Label is renamed to MnemonicInfo.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Sub-menu</code> CHUI client implementation class. 
 */
public class SubMenuChuiImpl
extends SubMenu<ChuiOutputManager>
{
   /** Border around the panel. */
   private Border<ChuiOutputManager> border = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuChuiImpl(WidgetId id, SubMenuConfig cfg)
   {
      super(id, cfg);
      
      border = new LineBorder()
      {
         /**
          * Draw line border.
          * 
          * @param   widget
          *          Base widget. 
          * @param   size
          *          Border dimensions.
          */
         @Override
         public void drawBorder(Widget<ChuiOutputManager> widget, Dimension size)
         {
            widget.screen().box(bodyLocation(), 
                     size, 
                     widget.screen().borderColor(widget));
         }
      };
   }
   
   /**
    * Calculates max menu/sub-menu item accelerator width.
    * 
    * @return   max accelerator width.
    */
   public static int getLongestItemAccelWidth(Container<?> menu)
   {
      int width = 0;
      for(Widget<?> w : menu.widgets())
      {
         // ignore SKIP or RULE or SUB-MENU
         if (!(w instanceof MenuItem) || !((MenuItem<?>)w).isNormal())
         {
            continue;
         }
         
         String accel = ((MenuItem<?>) w).config().accelerator;
         int itemWidth = 0;
         if (accel != null)
         {
            itemWidth = accel.length();
         }
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      
      return width;
   }
   
   /**
    * Calculates max menu/sub-menu item title (config's) width.
    * 
    * @param    menu
    *           Parent container: MENU or SUB-MENU.
    * @param    exclude
    *           Current item, to exclude from processing.
    * 
    * @return   max title width in pixels.
    */
   public static int getLongestItemTitleWidth(Container<?> menu, Widget<?> exclude)
   {
      if (exclude != null && !exclude.parent().equals(menu))
      {
         throw new IllegalArgumentException("Exclude item must be a child of menu.");
      }
      
      int width = 0;
      for (Widget<?> w : menu.widgets())
      {
         // prevent possible infinite loop during width calculation.
         if (w.equals(exclude))
         {
            continue;
         }
         
         // ignore SKIP or RULE 
         if (w instanceof MenuItem && !((MenuItem<?>) w).isNormal())
         {
            continue;
         }

         String title = w instanceof MenuItem ? 
                        ((MenuItem<?>) w).getPreparedTitle() : 
                        ((SubMenu<?>) w).getPreparedTitle();
                        
         int itemWidth = title.length(); 
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      
      return width;
   }
   
   /**
    * Calculates a max height of specified Menu or SubMenu.
    * Analyzes nested dropped-down sub-menus.
    * 
    * @param   menu
    *          Menu or SubMenu.
    *          
    * @param   includeNested
    *          <code>true</code> process nested sub-menus.
    * 
    * @return  a height of menu.
    */
   public static double getMaxItemHeight(AbstractContainer<?> menu, boolean includeNested)
   {
      Widget<?>[] widgets = menu.widgets();
      double height = widgets.length + 2; // borders
      
      // MENUBAR
      if (menu instanceof Menu && !((Menu<?>) menu).config().popupOnly)
         height = 1;
      
      if (includeNested)
      {
         for (int i = 0; i < widgets.length; i++)
         {
            if (widgets[i] instanceof SubMenu)
            {
               SubMenu<?> sm = (SubMenu<?>) widgets[i];
               if (sm.config().displayed)
               {
                  height = Math.max(height, sm.height() + i + 1 + 1); // this.y + title height + top border
               }
            }
         }
      }
      return height;
   }
   
   /**
    * Calculates max menu/sub-menu item width with borders.
    * 
    * @param    menu
    *           MENU or SUB-MENU
    * @param    includeNested
    *           <code>true</code> means, that we should process nested sub-menus, 
    *           <code>false</code> otherwise.          
    *           
    * @return   max width.
    */
   public static double getMaxItemWidth(Container<?> menu, boolean includeNested)
   {
      double width = 0; 
      for(Widget<?> w : menu.widgets())
      {
         AbstractWidget<?> aw = (AbstractWidget<?>) w;
         
         // ignore SKIP or RULE
         if (aw instanceof MenuItem && !((MenuItem<?>) aw).isNormal())
         {
            continue;
         }
         
         double itemWidth;
         if (aw instanceof SubMenu && !includeNested)
         {
            itemWidth = ((SubMenu<?>) aw).titleWidth() + 2 + 2; // left and right padding
         }
         else
         {
            itemWidth = aw.width();
         }
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      
      return width;
   }
   
   /**
    * Drawing worker method.
    */
   public void drawPanel()
   {
      Dimension dimension = dimension(false, false);
      
      screen().clear(WidgetId._DEFAULT_WINDOW_ID, 
                     bodyLocation(), 
                     dimension, 
                     screen().getInteractiveOutput().widgetColor(this));

      if (border != null)
      {
         border.drawBorder(this, dimension);
      }
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      super.doLayout();

      Widget<ChuiOutputManager>[] w = widgets();

      for (int i = 0; i < w.length; i++)
      {
         w[i].setLocation(new Point(1, i));

         if (w[i] instanceof SubMenu)
         {
            ((SubMenu<?>) w[i]).doLayout();
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      Point childrenLoc = parent().childrenLocation(this);
      Point loc = location();
      drawTitle(new Point(loc.x + childrenLoc.x, loc.y + childrenLoc.y));

      if (config().displayed)
      {
         drawPanel();
         super.draw();
      }
   }
   
   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      return dimension(true, true);
   }

   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   @Override
   public Point childrenLocation(Widget<ChuiOutputManager> child)
   {
      Point bodyLoc = bodyLocation();
      return new Point(bodyLoc.x + 1, bodyLoc.y + 1); // border
   }

   /**
    * Get widget width with borders.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      double w = titleWidth();
      
      if (config().displayed)
      {
         w = Math.max(w, SubMenuChuiImpl.getMaxItemWidth(this, true)); // borders
      }
      
      return w;
   }

   /**
    * Gets sub-menu title width (with '-&gt;' if sub-menu is in other sub-menu or in popup menu)
    * 
    * @return   width of sub-menu title
    */
   @Override
   public double titleWidth()
   {
      double w = getPreparedTitle().length();
      
      if (!Menu.isMenubarElement(this))
      {
         w += 3; // " ->" will be added

         // menu-items accelerators
         int accelWidth = SubMenuChuiImpl.getLongestItemAccelWidth(parent()); 
         w += accelWidth == 0 ? 0 : accelWidth + 2;
      }
      
      return w;
   }
   
   /**
    * Gets sub-menu title height
    * 
    * @return   height of sub-menu title
    */
   public double titleHeight()
   {
      return 1;
   }
   
   /**
    * Get sub-menu height.
    * 
    * @return  sub-menu height.
    */
   @Override
   public double height()
   {
      double h = 1; // title height
      
      if (config().displayed)
      {
         h = SubMenuChuiImpl.getMaxItemHeight(this, true);
      }
      
      return h;
   }

   /**
    * Raise the MENU-DROP 4GL events.
    */
   @Override
   protected void menuDropEvents()
   {
      KeyInput e = new KeyInput(this, EventType.ACTION_PERFORMED, Keyboard.SE_MENU_DROP);
      EventManager.postEvent(e);
      
      // if we are in ChUI, this needs to be posted again
      KeyInput e1 = new KeyInput(this, EventType.ACTION_PERFORMED, Keyboard.SE_MENU_DROP);
      EventManager.postEvent(e1);
   }
   
   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   @Override
   protected List<Widget<ChuiOutputManager>> prepareWidgets(List<Widget<ChuiOutputManager>> widgets)
   {
      List<Widget<ChuiOutputManager>> result = new ArrayList<>();
      Widget<ChuiOutputManager> displayedSubmenu = null;
      
      for (Widget<ChuiOutputManager> w : widgets)
      {
         if (w instanceof SubMenu && ((SubMenu<?>)w).config().displayed)
         {
            displayedSubmenu = w;
         }
         else
         {
            result.add(w);
         }
      }
      
      if (displayedSubmenu != null)
      {
         result.add(displayedSubmenu);
      }
      
      return Collections.unmodifiableList(result);
   }
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   @Override
   protected Point bodyLocation()
   {
      Point loc = parent().childrenLocation(this).translate(location());
      return new Point(loc.x, loc.y + 1); // under the title
   }

   /**
    * Calculates body dimension
    * 
    * @return  Dimension
    */
   @Override
   protected Dimension bodyDimension()
   {
      return dimension();
   }
   
   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component.
    *           
    * @return   corrected parent location.
    */
   @Override
   protected Point prepParentLocation(Widget<?> parent)
   {
      Point loc = parent.location();
      if (parent instanceof SubMenu)
      {
         loc.x += 1; // border
         loc.y += 2; // top border and title
      }
      return loc;
   }
   
   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   p
    *          Title point.  
    */
   private void drawTitle(Point p)
   {
      Color dColor = hasFocus() ? Color.MESSAGES : Color.NORMAL; // highlighted() ?
      Color mnemonicColor = Color.INPUT;
      
      MnemonicInfo label = getPreparedLabel();
      String title = getPreparedTitle();
      Point tp = new Point(p);

      if (label.getMnemonic() == null)
      {
         title += '_';
         int longTitleWidth = SubMenuChuiImpl.getLongestItemTitleWidth(parent(), null);

         // Special case, when this item label ends with odd sequence of '&' and it is a longest 
         // label between neighbors in this sub-menu body or menu. Focused item is drawn in the  
         // following way: last 2 characters including added '_' are drawn using mnemonic color.
         boolean specialCase = (title.length() - 1 >= longTitleWidth) && hasFocus();;
         
         // draw
         if (specialCase)
         {
            screen().at(tp).append(title.substring(0, title.length() - 1), dColor);
            tp.x += title.length() - 1;
            screen().at(tp).append(title.substring(title.length() - 1), mnemonicColor);
            tp.x += 1;
         }
         else
         {
            screen().at(tp).append(title, dColor);  
            tp.x += title.length();
         }
      }
      else
      {
         Integer ampIdx = label.getMnemIdx();

         String leftPart = title.substring(0, ampIdx);
         screen().at(tp).append(leftPart, dColor);

         tp.x += ampIdx;

         screen().at(tp).append(title.substring(ampIdx, ampIdx + 1), mnemonicColor);

         tp.x += 1;
         String rightPart = title.substring(ampIdx + 1);
         screen().at(tp).append(rightPart, dColor);
         tp.x += rightPart.length();
      }
      
      if (!Menu.isMenubarElement(this))
      {
         if (label.getMnemonic() != null)
         {
            screen().at(tp).append(" ", dColor);
            tp.x += 1;
         }
         screen().at(tp).append("->", dColor);
      }
   }

   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @param   includeNested
    *          <code>true</code> calculates dimension processing nested sub-menus.
    * @param   includeTitle
    *          <code>true</code> add title height (1).
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   private Dimension dimension(boolean includeNested, boolean includeTitle)
   {
      if (!config().displayed)
      {
         return new Dimension(width(), 1); 
      }
      else
      {
         double width = getMaxItemWidth(this, includeNested);

         int height = widgets().length + 2; // height borders
         
         if (includeTitle)
            height++;
         
         return new Dimension(width, height); // borders and title
      }
   }
}