/*
** Module   : AbstractChuiDriver.java
** Abstract : simple base class for ChUI screen driver implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150402 Created initial version.
** 002 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.chui.*;

/**
 * Simple base class for ChUI-specific screen driver implementation features. GUI-specific
 * features are not to be implemented here.  This class is intended to be specific to a given
 * UI-modality (ChUI in this case). However, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public abstract class AbstractChuiDriver
extends AbstractDriver<BasePrimitives, ChuiOutputManager>
{
   /**
    * Construct a new instance.
    *
    * @param    direct
    *           The output primitives.
    * @param    term
    *           Terminal type string.
    */
   public AbstractChuiDriver(BasePrimitives direct, String term)
   {
      super(new ChuiWidgetFactory(), direct, true, term);
   }

   /**
    * Read configuration values from the bootstrap configuration and convert
    * these into an options instance.
    * <p>
    * The following parameters are supported:
    * <p>
    * <pre>
    * Category:Group:Key       Type
    * -----------------------  ---------------------------
    * client:chui:rows         integer
    * client:chui:columns      integer
    * client:chui:background   text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:foreground   text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:selection    text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:fontname     text containing a monospaced (fixed width) font family name
    * client:chui:fontsize     integer
    * </pre>
    *
    * @param    cfg
    *           The bootstrap configuration.
    *
    * @return   A new options instance with any defaults overridden as specified
    *           in the bootstrap configuration.
    */
   public static TerminalOptions convertToOptions(BootstrapConfig cfg)
   {
      TerminalOptions options = new TerminalOptions();
      
      String next;
      
      if ((next = cfg.getString("client", "chui", "rows", null)) != null)
      {
         options.rows = parseInt(next, options.rows);
      }
      
      if ((next = cfg.getString("client", "chui", "columns", null)) != null)
      {
         options.columns = parseInt(next, options.columns);
      }
      
      if ((next = cfg.getString("client", "chui", "background", null)) != null)
      {
         options.background = ColorRgb.fromString(next, options.background);
      }
      
      if ((next = cfg.getString("client", "chui", "foreground", null)) != null)
      {
         options.foreground = ColorRgb.fromString(next, options.foreground);
      }
      
      if ((next = cfg.getString("client", "chui", "selection", null)) != null)
      {
         options.selection = ColorRgb.fromString(next, options.selection);
      }
      
      if ((next = cfg.getString("client", "chui", "fontname", null)) != null)
      {
         options.fontname = next;
      }
      
      if ((next = cfg.getString("client", "chui", "fontsize", null)) != null)
      {
         options.fontsize = parseInt(next, options.fontsize);
      }
      
      if ((next = cfg.getString("client", "chui", "windows", null)) != null)
      {
         options.underWindows = "true".equalsIgnoreCase(next);
      }
      
      return options;
   }

   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    *
    * @return   The instance.
    */
   @Override
   public ChuiOutputManager createOutputManager(BootstrapConfig cfg)
   {
      return new ChuiOutputManager(this);
   }
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * For GUI, this does nothing.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    *
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean setCursorStatus(boolean on)
   {
      return false;
   }

   /**
    * Enable or disable capture of mouse events.
    * <p>
    * This implementation is a no-op.
    *  
    * @param    state
    *           <code>true</code> to capture mouse events, <code>false</code> to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      // no-op
      return false;
   }

   /**
    * Handle the specified event, in the context of the specified widget.
    * <p>
    * This implementation is a no-op.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   @Override
   public void handleMouseEvent(int sourceID, MouseEvent evt)
   {
      // no-op
   }
   
   /**
    * Attempt to convert the given string into an integer using
    * <code>Integer.parseInt</code>.
    *
    * @param    spec
    *           String encoded number specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed number or the default if there is any parsing
    *           problem.
    */
   private static int parseInt(String spec, int def)
   {
      int val = def;
      
      try
      {
         val = Integer.parseInt(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return val;
   }
}