/*
** Module   : VT100EmulatorChildProcess.java
** Abstract : interactive child process for clients that have an emulated vt100 terminal
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150413 First version.
** 002 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.terminal.*;

/**
 * Provides an implementation for an interactive child process for clients that do NOT have a
 * native terminal built-in to the JVM process.
 */
public abstract class VT100EmulatorChildProcess
extends AbstractChildProcess
{
   /** Child process wrapped by a VT100 terminal emulator. */
   protected VT100Emulator emulator = null;
   
   /** Emulator worker thread. */
   protected Thread worker = null;
   
   /**
    * Construct an instance.
    */
   public VT100EmulatorChildProcess()
   {
      super(ThinClient.getInstance(), false, true);
   }
   
   /**
    * Create a shell emulator and execute the given command.
    * 
    * @param    cmdline
    *           The command.
    *
    * @return   A VT100 emulator executing the given command.
    */
   public abstract VT100Emulator createShellEmulator(String[] cmdline);
   
   /**
    * Launch the specified command line as an interactive child process.
    *
    * @param    cmdline
    *           The command line to execute.  The first element must be the program name.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    *
    * @return   The operating system process ID (pid) of the child process.
    */
   @Override
   public int launch(String[] cmdline, boolean silent)
   {
      emulator = createShellEmulator(cmdline);
            
      // get child process pid.
      int pid = emulator.getPid();
      
      if (pid >= 0)
      {
         // start pseudo terminal, this is the thread that will read output from the child
         // process and write that output to the simulated terminal (displayed to the user)
         worker = new Thread(emulator);
         worker.start();
      }
      
      return pid;
   }
   
   /**
    * Block execution until the interactive child process has exited. This will ensure that
    * any operating system resources are released before returning.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    */
   @Override
   public void waitForExit(int pid, boolean silent)
   throws InterruptedException
   {
      if (emulator == null || worker == null)
      {
         // TODO: log this?
         return;
      }
      
      // wait for interactive process to die
      emulator.waitFor();
      
      // stop the emulator and wait for the worker thread to stop
      emulator.stopRunning();
      worker.join();
   }
   
   /**
    * Asynchronously cleanup any operating system resources when the interactive child process
    * has exited.  This processing should be done an a secondary thread and this method will
    * return immediately after starting that thread.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    */
   @Override
   public void asynchCleanup(int pid)
   {
      Thread t1 = new Thread(new Cleaner());
      t1.start();
   }
   
   /**
    * Wait for the user to press a key.
    *
    * @return   The key code of the key pressed to clear the pause or -1 on any failure.
    */
   @Override
   public int pauseForUser()
   {
      WindowManager.resolveWindow().status("Press space bar to continue.");
      return super.pauseForUser();
   }
   
   /**
    * Provides a simple mechanism to wait for the exit of a process on an asynchronous basis
    * (on a new thread dedicated to this wait).
    */
   private class Cleaner
   implements Runnable
   {
      /**
       * Blocks until the child process exits.
       */
      public void run()
      {
         emulator.waitFor();
      }
   }
}