/*
** Module   : ChuiWebPageHandler.java
** Abstract : ChUI-specific page handler
**
** Copyright (c) 2013-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20131105 Created initial version.
** 002 MAG 20140114 Add new replacements.
** 003 MAG 20140217 Add referrer replacement.
** 004 MAG 20140303 Add authorization token.
** 005 MAG 20140711 Enable/Disable clipboard.
** 006 MAG 20140723 Get authorization token as a query parameter and set a cookie for this domain
**                  if token match the expected value.
** 007 GES 20150312 Moved most code to a common class for shared usage.
** 008 GES 20150717 Moved clipboard processing back here because it is not conditional for GUI.
** 009 SBI 20160601 Added graphicsCached parameter to the index.html template.
** 010 SBI 20180405 Changed to use template key and value provider.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;


import java.util.*;
import java.util.function.*;

import org.eclipse.jetty.server.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.client.driver.web.*;

/**
 * This code primarily exists to parse the HTML text and substitute place holders, but all the
 * rest of the handler functionality is in the parent class where it can be shared. 
 */
public class ChuiWebPageHandler
extends WebPageHandler
{
   /**
    * Constructor.
    *
    * @param    config
    *           Configuration parameters.
    * @param    root
    *           Prefix for all URLs relating to this page.
    * @param    skel
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public ChuiWebPageHandler(BootstrapConfig config, String root, String skel)
   {
      super(config, root, skel);
   }
   
   /**
    * Generates a map of template keys with their value suppliers in order to fill gaps in
    * the html skeleton template page.
    * 
    * @param    base
    *           The request
    * 
    * @return   The target map of template keys with their value suppliers
    */
   @Override
   protected Map<String, Supplier<String>> getTemplateKeys(Request base)
   {
      ChuiWebPageKeysProvider provider = new ChuiWebPageKeysProvider(base);
      
      return provider.getTemplateKeys();
   }

   /**
    * Builds the target map of keys and their value providers for the associated template
    * given by the skeleton page.
    */
   class ChuiWebPageKeysProvider
   extends WebPageKeysProvider
   {
      /**
       * Creates this key and value provider.
       * 
       * @param    base
       *           The request
       */
      public ChuiWebPageKeysProvider(Request base)
      {
         super(base);
         
         add("client_title", () -> { return "ChUI";});
         
         add("isGui", () -> { return "false";});
         
         add("height_name", () -> { return "rows";});
         
         add("height_val", () ->
         {
            int rows = config.getInt("client", "chui", "rows", 24);
            
            return  String.valueOf(rows);
         });
         
         add("width_name", () -> { return  "cols";});
         
         add("width_val", () ->
         {
            int cols = config.getInt("client", "chui", "columns", 80);
            
            return String.valueOf(cols);
         });
         
         add("font.name", () ->
         {
            String fontName = config.getString("client", "chui", "fontname", "monospace");
            
            return String.valueOf(fontName);
         });
         
         add("font.size", () ->
         {
            int fontSize = config.getInt("client", "chui", "fontsize", 12);
            
            return String.valueOf(fontSize);
         });

         add("font.color", () ->
         {
            String fontColor = config.getString("client", "chui", "foreground", "#FFA500");
            
            return htmlColor(fontColor);
         });

         add("font.background", () ->
         {
            String fontBackground = config.getString("client", "chui", "background", "#000000");
            
            return htmlColor(fontBackground);
         });
         
         add("clipboard.enabled", () ->
         {
            boolean enabled = config.getBoolean("client", "web", "clipboard", true);
            
            return enabled ? "true" : "false";
         });
         
         // the taskbar is always disabled in ChUI, we don't honor the directory setting
         add("taskbar.enabled", () ->
         {
            return "false";
         });
         
         add("graphicsCached", () ->
         {
            return "false";
         });
      }
   }
}