/*
** Module   : p2j.clipboard_helpers.js
** Abstract : ChUI-specific text clipboard processing
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150324 First version, with refactored content moved from tty.clipboard.js.
** 002 SBI 20150731 Applied "use strict" directive, fixed undeclared variables
**                  and corrected the selection rectangle coordinates.
** 003 GES 20150710 Reworked approach and moved most of the previously "common" implementation
**                  to this module, since it was really ChUI-specific. The external interface
**                  is extended to match the new common approach.  All selection processing 
**                  and the popup menu is here since GUI doesn't do that.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Core clipboard implementation that provides ChUI-specific callbacks for selection processing
 * and rendering.  The user drags the mouse to select text and the hooks are called to capture
 * selection positions and to render the selection highlight. After the mouse drag is complete
 * any CTRL-C will copy the text to the clipboard. ESC key will clear the selection. For paste,
 * the CTRL-V key combination is monitored and text read from the clipboard is sent to the server
 * as input. 
 */
p2j.clipboard_helpers = (function() 
{
   /** Clipboard helper object which will be returned from this initialization function. */
   var me = {};
   
   /** Drag flag. */
   var drag = false;
      
   /** Active selection flag. */
   var selection = false;
   
   /** Draw rectangle selection using stroke. */
   var stroke = true;
   
   /** Selection color. */
   var color = 'blue';

   /** Font width metric. */
   var charWidth;
   
   /** Font height metric. */
   var charHeight;
      
   /** Canvas object. */
   var canvas;
   
   /** 2D graphics context. */
   var ctx;
   
   /** Selected cells. */
   var cells = {};
   
   /** Selection rectangle. */
   var rect = {};
   
   /**
    * Init module.
    * 
    * @param {type} cfg configurartion object
    */
   me.init = function(cfg)
   {
      charWidth = p2j.screen.charWidth;
      charHeight = p2j.screen.charHeight;
      
      var settings = 
      {
         id : cfg.clipboard.id,
         contextmenu : "menu-id",
         width : p2j.screen.canvasWidth,
         height : p2j.screen.canvasHeight,
         style : { zIndex : 1, position : "absolute"}
      };

      canvas = p2j.createCanvas(null, settings);
      
      stroke = cfg.clipboard.stroke;
      
      ctx = canvas.getContext('2d');
      ctx.fillStyle = "white"; 
      ctx.globalAlpha = 0;
      ctx.clearRect(0, 0, canvas.width, canvas.height);
      
      canvas.addEventListener('mousedown', mouseDown, false);
      canvas.addEventListener('mouseup', mouseUp, false);
      canvas.addEventListener('mousemove', mouseMove, false);

      require(["dijit/Menu", "dijit/MenuItem", "dijit/CheckedMenuItem", "dijit/MenuSeparator", "dijit/PopupMenuItem"], 
      function(Menu, MenuItem, CheckedMenuItem, MenuSeparator, PopupMenuItem)
      {   
         var pMenu = new dijit.Menu(
         {
            targetNodeIds:[cfg.clipboard.id]
         });

         // select 
         var pMenuSelect = new dijit.CheckedMenuItem(
         {
            iconClass:"dijitEditorIcon dijitEditorIconSelectAll",
            label: "Enable Clipboard",
            disabled: false,
            checked: p2j.clipboard.enabled,
            onClick: function()
            {
               p2j.clipboard.enabled = this.checked;
               p2j.clipboard.clearSelection();
            }
         });
         
         pMenu.addChild(pMenuSelect);
         pMenu.startup();      
      });
   };
   
   /**
    * Mouse down event handler.
    * 
    * @param {type} e Event object
    *
    * @returns {undefined}
    */
   function mouseDown(e) 
   {
      if (e.button === 0)
      {
         var x = (e.pageX - canvas.offsetLeft) | 0;
         var y = (e.pageY - canvas.offsetTop) | 0;
         
         cells.startX = (x / charWidth) | 0;
         cells.startY = (y / charHeight) | 0;         
         rect.startX = cells.startX * charWidth;
         rect.startY = cells.startY * charHeight;
         
         drag = true;
      }
   }

   /**
    * Mouse up event handler. Draw selection rectangle.
    * 
    * @param {type} e Event object
    *
    * @returns {undefined}
    */
   function mouseUp(e) 
   {
      if (drag && e.button === 0)
      {
         drag = false;
      }
   }

   /**
    * Mouse move event handler.
    * 
    * @param {type} e Event object
    *
    * @returns {undefined}
    */
   function mouseMove(e) 
   {
      if (p2j.clipboard.enabled && drag && e.button === 0) 
      {
         var x = (e.pageX - canvas.offsetLeft | 0);
         var y = (e.pageY - canvas.offsetTop | 0);
         cells.startX1 = (x / charWidth) | 0;
         cells.startY1 = (y / charHeight) | 0;
   
         rect.w = (cells.startX1 - cells.startX + 1) * charWidth;
         rect.h = (cells.startY1 - cells.startY + 1) * charHeight;
   
         ctx.clearRect(0, 0, canvas.width, canvas.height);         
   
         if (stroke)
         {
            ctx.globalAlpha = 1.0;
            ctx.beginPath();
            ctx.strokeStyle = color;      
            ctx.strokeRect(rect.startX, rect.startY, rect.w, rect.h);
            ctx.stroke();            
         }
         else
         {
            ctx.globalAlpha = 0.2;
            ctx.fillStyle = color;
            ctx.fillRect(rect.startX, rect.startY, rect.w, rect.h);
         }
         
         selection = true;
      }
   }
   
   /**
    * Return the currently selected text.
    * 
    * @return   The text that is selected.
    */
   me.getSelectedText = function()
   {
      // rows
      var begRow = Math.min(cells.startY, cells.startY1);
      var endRow = Math.max(cells.startY, cells.startY1);
      
      // cols
      var begCol = Math.min(cells.startX, cells.startX1);
      var endCol = Math.max(cells.startX, cells.startX1);
      
      return p2j.screen.copy(begRow, endRow, begCol, endCol);
   };
   
   /**
    * Check if this client needs an active selection area for copy to work.
    * 
    * @returns   {boolean} Always returns true.
    */
   me.needsSelection = function()
   {
      return true;
   };
   
   /**
    * Clear selection on screen.
    * 
    * @returns   {undefined}
    */
   me.clearSelection = function()
   {
      selection = false;
      ctx.clearRect(0, 0, canvas.width, canvas.height);
   };
   
   /**
    * Check if there is an active selection area.
    * 
    * @returns   {boolean} true if have a selection false if not.
    */
   me.haveSelection = function()
   {
      return selection;
   };
   
   return me;   
}());