/*
** Module   : WidgetFactoryAdapter.java
** Abstract : provide empty implementations for WidgetFactory interface
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140521 Created initial version
** 002 CA  20140711 Added MESSAGE-AREA widget support.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added support for IMAGE widget creation.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Added support for scroll bar widget creation.
** 006 MAG 20141009 Deleted createMessageLine.
** 007 EVL 20141118 Adding scroll container creation method.
** 008 CA  20150323 Pass the frame ID when creating a label (to properly resolve the font in GUI).
** 009 HC  20150323 Removed business logic from config classes.
** 010 EVL 20150428 Adding drop-down container creation method to be overridden.
** 011 CA  20150509 Allow scroll container for any widgets, not just frames.
** 012 HC  20150831 Scrolling support overhaul.
** 013 IAS 20160217 Added new flavor for the createScrollPane method.
** 014 OM  20170815 Dropped unnecessary imports.
** 015 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provide default implementations for {@link WidgetFactory} interface.  Any API not overridden
 * by sub-classes will throw a {@link RuntimeException} by default.
 */
public abstract class WidgetFactoryAdapter<O extends OutputManager<?>>
implements WidgetFactory<O>
{
   /** Mapping between abstract and UI-specific widget classes. */
   @SuppressWarnings("rawtypes")
   protected Map<Class<? extends Widget>, Class<? extends Widget>> classMap = new HashMap<>();
   
   /**
    * A cache of widget constructors having as single parameter a {@link WidgetConfig} instance.
    */
   private Map<Class<? extends Widget<?>>, Constructor<?>> widgetCtorsCache = new HashMap<>();
      
   /**
    * Instantiate UI-specific version of the specified widget class.
    * 
    * @param   <T>
    *          Widget abstract class type.
    * @param   clazz
    *          Widget abstract class for which UI-specific implementation should
    *          be instantiated.
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Widget configuration container.
    *          
    * @return  new UI-specific instance of the requested class. 
    */
   @Override
   public <T extends Widget<O>> T create(Class<T> clazz, WidgetId id, WidgetConfig cfg)
   {
      @SuppressWarnings("unchecked")
      Class<T> implClazz = (Class<T>) classMap.get(clazz);
      
      if (implClazz == null) {
         throw new RuntimeException("Attempt to instantiate class" + clazz + 
                                    " which has no proper mapping");
      }
      
      try
      {
         @SuppressWarnings("unchecked")
         Constructor<T> constructor = (Constructor<T>) widgetCtorsCache.get(implClazz);
         if (constructor == null)
         {
            // find a compatible constructor - two arguments, first assignable to 
            // WidgetId and second assignable to WidgetConfig
            Constructor<?>[] ctors = implClazz.getConstructors();
            
            for (Constructor<?> ctor : ctors)
            {
               Class<?>[] params = ctor.getParameterTypes();
               if (params.length != 2)
               {
                  continue;
               }
               
               if (!WidgetId.class.isAssignableFrom(params[0]))
               {
                  continue;
               }
               
               if (!WidgetConfig.class.isAssignableFrom(params[1]))
               {
                  continue;
               }
               
               if (constructor != null)
               {
                  throw new RuntimeException("Multiple compatible constructors found.");
               }
               
               @SuppressWarnings("unchecked")
               Constructor<T> foundCtor = (Constructor<T>) ctor;
               constructor = foundCtor;
            }
            
            if (constructor == null)
            {
               throw new RuntimeException("No compatible constructor found.");
            }
            
            widgetCtorsCache.put(implClazz, constructor);
         }
         
         // this is an expensive operation in a heavy-duty method, but at least we are not 
         // resolving the c'tor each time a new widget is created.
         return constructor.newInstance(id, cfg);
      }
      catch (Exception e)
      {
         throw new RuntimeException("Attempt to instantiate class" + implClazz + 
                                    " for class " + clazz + " failed", e);
      }
   }

   /**
    * Create new toggle-box for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          ToggleBox label text.
    *          
    * @return  new {@link ToggleBox} instance.
    */
   @Override
   public ToggleBox<O> createToggleBox(WidgetId id, Container<O> container, String text)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create new rectangle for given container.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    *          
    * @return  new {@link Rect} instance.
    */
   @Override
   public Rect<O> createRectangle(WidgetId id, Container<O> container)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link Image} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Image}.
    */
   @Override
   public Image<O> createImage(WidgetId id)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create UI-specific {@link ScrollBar} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   orientation
    *          Scroll bar orientation horizontal or vertical.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollBar}.
    */
   public ScrollBar<O> createScrollBar(WidgetId id, 
                                       ScrollBar.Orientation orientation,
                                       Viewport<O> viewport)
   {
      throw new RuntimeException("Not implemented.");
   }   
}
