/*
** Module   : SwingKeyboardReader.java
** Abstract : process swing key events and convert them to 4GL keys
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20111012 Made some static data final and added a comment.
** 003 CA  20141029 Refactored into Linux/Windows specific implementations.
** 004 GES 20150506 Moved to a driver package.
** 005 SBI 20151008 Improvements to support better GUI key processing compatibility.
** 006 VIG 20151012 Added a comment to keyReleased method about toggle-box exception.
** 007 GES 20151020 Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.event.*;
import java.util.*;
import java.util.concurrent.*;

// selectively import because this is more swing than p2j
import com.goldencode.p2j.ui.client.event.Key;

/**
 * This class processes Swing key events and converts them into 4GL keys. The
 * 7-bit ASCII range is generally handled using {@link #keyTyped} and all other
 * key codes are handled by {@link #keyPressed}.  {@link #keyReleased} is not
 * used since Progress generates all key input on the key press instead of on
 * the key releasse.
 * <p>
 * All input that is not dropped is translated as needed and then is stored
 * in a FIFO queue which represents the key input buffer for the application.
 * This key input is read using {@link #read}.
 * <p>
 * Known Issues and/or TODOs
 * <p>
 * <ul>
 *   <li> refactor this code into CHUI-specific and generic Swing code (this
 *        generic code would be usable by a GUI implementation)
 *   <li> confirm that the numeric keypad (i.e. numpad) works the same on a
 *        keyboard with NUMLOCK off
 * </ul> 
 */
public abstract class SwingKeyboardReader
extends KeyAdapter
{
   // these maps are read-only after initialization so it is OK for them to
   // be static
   
   /** Translation map of Java virtual keys to Progress key codes. */
   protected final Map<Integer, Integer> xlate = new HashMap<>();
   
   /** Incoming keystroke buffer. */
   protected final LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>();
   
   /** The stack of keystrokes to parse, contains key down events. */
   protected final Deque<KeyEvent>  keyDownEvents = new ArrayDeque<KeyEvent>(256);
   
   /**
    * Default c'tor.  Populates the key {@link #xlate translations}. 
    */
   public SwingKeyboardReader()
   {
      // The following do not need any translation since the keyTyped() method
      // naturally generates these values:
      //    VK_BACKSPACE
      //    VK_CTRL_*
      //    VK_DELETE
      //    VK_ESCAPE (this is never returned to Progress code anyway)
      //    VK_TAB (the containing frame's focus traversal keys must be
      //            forced to the empty set in order for this to work)
      
      xlate.put(KeyEvent.VK_DOWN, Key.VK_CURSOR_DOWN);
      xlate.put(KeyEvent.VK_END, Key.VK_END);
      xlate.put(KeyEvent.VK_ENTER, Key.VK_ENTER); // Java (10), 4GL (13)
      xlate.put(KeyEvent.VK_F1, Key.VK_F1);
      xlate.put(KeyEvent.VK_F2, Key.VK_F2);
      xlate.put(KeyEvent.VK_F3, Key.VK_F3);
      xlate.put(KeyEvent.VK_F4, Key.VK_F4);
      xlate.put(KeyEvent.VK_F5, Key.VK_F5);
      xlate.put(KeyEvent.VK_F6, Key.VK_F6);
      xlate.put(KeyEvent.VK_F7, Key.VK_F7);
      xlate.put(KeyEvent.VK_F8, Key.VK_F8);
      xlate.put(KeyEvent.VK_F9, Key.VK_F9);
      xlate.put(KeyEvent.VK_F10, Key.VK_F10);
      xlate.put(KeyEvent.VK_F11, Key.VK_F11);
      xlate.put(KeyEvent.VK_F12, Key.VK_F12);
      xlate.put(KeyEvent.VK_F13, Key.VK_F13);
      xlate.put(KeyEvent.VK_F14, Key.VK_F14);
      xlate.put(KeyEvent.VK_F15, Key.VK_F15);
      xlate.put(KeyEvent.VK_F16, Key.VK_F16);
      xlate.put(KeyEvent.VK_F17, Key.VK_F17);
      xlate.put(KeyEvent.VK_F18, Key.VK_F18);
      xlate.put(KeyEvent.VK_F19, Key.VK_F19);
      xlate.put(KeyEvent.VK_F20, Key.VK_F20);
      xlate.put(KeyEvent.VK_HOME, Key.VK_HOME);
      xlate.put(KeyEvent.VK_INSERT, Key.VK_INSERT);
      xlate.put(KeyEvent.VK_LEFT, Key.VK_CURSOR_LEFT);
      xlate.put(KeyEvent.VK_PAGE_DOWN, Key.VK_PAGE_DOWN);
      xlate.put(KeyEvent.VK_PAGE_UP, Key.VK_PAGE_UP);
      xlate.put(KeyEvent.VK_RIGHT, Key.VK_CURSOR_RIGHT);
      xlate.put(KeyEvent.VK_TAB, Key.VK_TAB);  // not used for VK_TAB, instead  
                                               // used with SHIFT = VK_BACKTAB
      xlate.put(KeyEvent.VK_UP, Key.VK_CURSOR_UP);
   }
   
   /**
    * Tests if the given key code is a modifier.
    * 
    * @param    key
    *           The key code of the pressed key
    * 
    * @return   {Boolean}
    *           True iff a key is from the following set: SHIFT, ALT, CTRL.
    */
   public final boolean isModifier(int key)
   {
      return key == KeyEvent.VK_SHIFT || key == KeyEvent.VK_CONTROL || key == KeyEvent.VK_ALT;
   }
   
   /**
    * Merges the java keyboard event with the last key event.
    * 
    * @param    last
    *           The last key event in the keystrokes queue.
    * @param    evt
    *           The java keyboard event
    */
   public final void mergeWithLastEvent(KeyEvent last, KeyEvent evt)
   {
      last.setKeyCode(evt.getKeyCode());
      if (evt.isShiftDown() || evt.getKeyCode() == KeyEvent.VK_SHIFT)
      {
         last.setModifiers(last.getModifiers() | InputEvent.SHIFT_MASK);
      }
      if (evt.isAltDown() || evt.getKeyCode() == KeyEvent.VK_ALT)
      {
         last.setModifiers(last.getModifiers() | InputEvent.ALT_MASK);
      }
      if (evt.isControlDown() || evt.getKeyCode() == KeyEvent.VK_CONTROL)
      {
         last.setModifiers(last.getModifiers() | InputEvent.CTRL_MASK);
      }
   }

   /**
    * Tests if the pressed key is @, ^, _ , \, ]
    * 
    * @param    evt
    *           The java generated keyboard event.
    * 
    * @return   {Boolean}
    *           True iff the pressed key is one of @, ^, _ , \, ]
    */
   public final boolean isSpecialCharacter(KeyEvent evt)
   {
      int keyCode = evt.getKeyCode();
      if (evt.isShiftDown())
      {
         return (keyCode == '@') || (keyCode == '^') || (keyCode == '_') ||
         (keyCode == '|') || (keyCode == '}');
      }
      return (keyCode == '2') || (keyCode == '6') || (keyCode == '-') ||
               (keyCode == '\\') || (keyCode == ']');
   }
   
   /**
    * Read a keystroke from the key input buffer. This method will block until
    * a key is available.
    *
    * @return   A 4GL-compatible key code matching the input combination which
    *           the user pressed on the keyboard.
    */
   public int read()
   {
      int key = -1;
      
      try
      {
         // use of the queue is thread-safe
         key = queue.take();
      }
      
      catch (InterruptedException ie)
      {
         // ignore (TODO: is this right?)
      }
      
      return key; 
   }
   
   /**
    * Add each character of the string to end of the key input buffer in FIFO
    * order (left to right). Each character is converted directly to an
    * integer during the addition. This is fine for the 7-bit ASCII range, but
    * it may be incorrect for other UNICODE characters.
    *
    * @param    data
    *           The "keys" to insert.
    */
   public void insert(String data)
   {
      // use of the queue is thread-safe
      for (int i = 0; i < data.length(); i++)
      {
         queue.add(new Integer((int) data.charAt(i)));
      }
   }

   /**
    * This method is called when a pressed key is released. This is not used
    * since Progress processes key events on the key press, not on the key
    * release.  In addition, no state needs to be cleaned up using this method
    * so it is a NO-OPERATION.
    * 
    * TODO: There is an exception from this rule found for Windows toggle-boxes, 
    *       so we should implement this method in {@link WinKeyboardReader}.
    *       Toggle-box (or other, possible) component must be able to process this 
    *       event of releasing key along with pressing.
    *
    * @param    evt
    *           The key event.
    */
   public synchronized void keyReleased(KeyEvent evt)
   {
   }

   /**
    * Finds the first key down event that has the given key code.
    * 
    * @param    keyCode
    *           The key event code
    * 
    * @return   KeyEvent
    *           The down key event with the given key code 
    */
   protected KeyEvent findKeyEventByKeyCode(int keyCode)
   {
      Iterator<KeyEvent> eventsIterator = keyDownEvents.iterator();
      while (eventsIterator.hasNext())
      {
         KeyEvent evt = eventsIterator.next(); // modifier key down events are merged
         if (evt.getKeyCode() == keyCode || (evt.isAltDown() && keyCode == KeyEvent.VK_ALT)
                  || (evt.isShiftDown() && keyCode == KeyEvent.VK_SHIFT)
                  || (evt.isControlDown() && keyCode == KeyEvent.VK_CONTROL))
         {
            eventsIterator.remove();
            return evt;
         }
      }
      return null;
   }
}