/*
** Module   : ClientProtocolHooks.java
** Abstract : API that describes callbacks for WebSockets usage
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 GES 20150318 First version.
** 002 CA  20150821 Added raiseMouseEvent, setWindowLocation, windowActivated.
** 003 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
** 004 CA  20151013 Added support for window resize for the GUI web client.
** 005 SBI 20151208 Added new parameters, scroll amount and scroll unit for wheel events and
**                  a mouse modifiers mask for all mouse events.
**     GES 20160201 Javadoc fix.
** 006 HC  20160406 Overhaul of window-activation logic.
** 007 EVL 20170825 Adding method to initiate file list dropping.
**     SBI 20170903 Used file channels for uploading large files.
**     EVL 20170905 Changing signature of the hook function to use list of strings.
** 008 HC  20170823 Made web js client to notify the java client when desktop (browser window)
**                  resized. This makes the Web GUI driver to report correct desktop dimensions.
** 009 HC  20171203 Added notification of Web embedded client when program execution enters modal or wait-for
**                  event loop.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.nio.channels.*;
import java.util.*;

/**
 * API that describes callbacks for WebSockets driver processing.
 */
public interface ClientProtocolHooks
{
   /**
    * Called each time a connect occurs, this allows caller-controlled initialization of the
    * remote client.  There can only be 1 connection open at a time, but it is possible for
    * the connection to be opened, closed and then re-opened.
    */
   public void initRemoteClient();
   
   /**
    * Input key events when in VT100 mode.
    * 
    * @param    keyCode
    *           The key code from the input event.
    * @param    charCode
    *           The character code from the input event.
    */
   public void injectVT100Key(int keyCode, char charCode);
   
   /**
    * Raise a mouse event.
    * 
    * @param    windowId
    *           The window ID.
    * @param    mouseOp
    *           The mouse event code.
    * @param    tstamp
    *           The event timestamp.
    * @param    mouseX
    *           The absolute mouse X for this event.
    * @param    mouseY
    *           The absolute mouse Y for this event.
    * @param    button
    *           The button which was pressed.
    * @param    clickCount
    *           The click count.
    * @param    modifiersMask
    *           The bit mask that represents the modifiers keys are pressed. The 0-bit corresponds
    *           to the shift key, the 1-bit to the ctrl key, the 2-bit to the meta key and
    *           the 3-bit to the alt key.
    * @param    wheelParameters
    *           The array that represents the wheel rotation amount and its unit.
    * @param    widgetId
    *           The explicit widget ID to which this event needs to be posted.
    */
   public void raiseMouseEvent(int   windowId,
                               int   mouseOp, 
                               long  tstamp, 
                               int   mouseX, 
                               int   mouseY,
                               int   button,
                               int   clickCount, 
                               int   modifiersMask,
                               int[] wheelParameters,
                               int   widgetId);

   /**
    * Set the location for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    x
    *           The X coordinate of the top-left corner.
    * @param    y
    *           The Y coordinate of the top-left corner.
    */
   public void setWindowLocation(int windowId, int x, int y);

   /**
    * Set the dimension for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    width
    *           The window width.
    * @param    height
    *           The window height.
    */
   public void setWindowDimension(int windowId, int width, int height);

   /**
    * Raise an event indicating that the target window has been activated or deactivated.
    * 
    * @param    deactivatingWindowId
    *           Id of the window being deactivated, may be negative.
    * @param    activatingWindowId
    *           Id of the window being activated, always valid.
    */
   public void windowActivated(int deactivatingWindowId, int activatingWindowId);

   /**
    * Raise an event as the window was (de)iconified.
    * 
    * @param    windowId
    *           The window ID.
    * @param    state
    *           <code>true</code> if the window was iconified.
    */
   public void windowIconified(int windowId, boolean state);

   /**
    * Initiates the process of the file list dropping for web clients.
    * 
    * @param    windowId
    *           The window ID.
    * @param    widgetId
    *           The widget ID.
    * @param    fileNameList
    *           The ordered file name list of objects to be dropped.
    */
   public default void startDropTarget(int windowId, int widgetId, List<String> fileNameList)
   {
      // do nothing by default for ChUI emulator version
   }
   
   /**
    * Stores the given byte array data as file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create.
    * @param    data
    *           The data to store in the file.
    * @param    size
    *           The size of the to create, used to check the input array.
    */
   public default void saveDropTargetFile(String name, byte[] data, int size)
   {
      // do nothing by default for ChUI emulator version
   }
   
   /**
    * Reads the data from the given file channel starting from its current position and stores
    * the retrieved data into the file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create
    * @param    channel
    *           The file channel
    * @param    length
    *           The length of the content
    */
   public default void saveDropTargetFile(String name, FileChannel channel, int length)
   {
      // do nothing by default for ChUI emulator version
   }

   /**
    * Called when the desktop on the client has changed its dimensions.
    *
    * @param    width
    *           New desktop width.
    * @param    height
    *           New desktop height.
    */
   public void desktopResized(int width, int height);

   /**
    * The method is executed by the JS client when it requires to be notified when 4GL program execution
    * enters modal or wait-for event loop, in other words when user input is expected.
    *
    * @param    newState
    *           {@code true} turns ON the notification while {@code false} turns it OFF.
    */
   public default void setNeedsNotificationOnWait(boolean newState)
   {
      // no-op by default
   }
}
