/*

** Module   : WebTaskWorker.java
** Abstract : a worker to process JS requests
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 CA  20150904 Created initial version.
** 002 CA  20151014 Protect the task runner loop against abends during task execution.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.util.*;

/**
 * This class is responsible of processing the requests received from javascript in a separate
 * thread.  The requests are processed in the order they are received.
 * <p>
 * This helps in releasing the websocket (and thus the javascript's single-thread processing) so
 * JS events can be captured/
 */
class WebTaskWorker
extends Thread
{
   /** Flag to control thread abort. */
   private volatile boolean running = true;

   /** List of tasks sent by the JS side. */
   private final LinkedList<Runnable> webTasks = new LinkedList<>();

   /**
    * Create a new worker.
    */
   public WebTaskWorker()
   {
      this.setName("webtaskworker");
      this.start();
   }

   /**
    * Listening loop which waits for tasks to be posted to {@link #webTasks}.
    */
   @Override
   public void run()
   {
      LinkedList<Runnable> copy = null;
      
      while (running)
      {
         synchronized (webTasks)
         {
            if (webTasks.isEmpty())
            {
               try
               {
                  webTasks.wait();
               }
               catch (InterruptedException e)
               {
                  e.printStackTrace();
               }
            }

            copy = new LinkedList<>(webTasks);
            webTasks.clear();
         }
         
         while (!copy.isEmpty())
         {
            Runnable task = copy.removeFirst();
            try
            {
               task.run();
            }
            catch (Throwable t)
            {
               t.printStackTrace();
            }
         }
      }
   }
   
   /**
    * Terminate the thread.
    */
   public void kill()
   {
      running = false;
      this.interrupt();
      
      // wait for thread to exit
      try
      {
         this.join(1000);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
   }

   /**
    * Post the given task to the queue.
    * 
    * @param    task
    *           The task.
    */
   public void post(Runnable task)
   {
      synchronized (webTasks)
      {
         webTasks.addLast(task);
         webTasks.notify();
      }
   }
}