/*
** Module   : p2j.keyboard.js
** Abstract : keystroke management
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140228 Add javadoc.
** 003 MAG 20140402 Implements interactive process.
** 004 MAG 20140507 Fix DEL key code for VT100 mode.
** 005 MAG 20140701 Use CTRL-C and CTRL-V to copy/paste.
** 006 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients.
** 007 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 008 GES 20150709 Modifications to support GUI clipboard processing. Changes were made to make
**                  the common clipboard implementation more generic. Removed reference to
**                  window.Clipboard since that is undefined on any browser (it must have been
**                  copied from some samples on the web which had a typo because I can find no
**                  evidence it is a real thing).
** 009 SBI 20150922 Rewrite to implement separate implementations for ChUI and GUI. GUI key
**                  processing is intended to be compatible with WinKeyboardReader, though
**                  it may not yet be 100% correct.
** 010 SBI 20151006 GUI rewrite to support better key processing compatibility.
** 011 SBI 20151014 Auto-repeats are fixed. Added GUI copy/cut/paste alternative keys support
**                  that used by clipboard processing.
** 012 SBI 20160119 Changed to support Chrome and IE.
**     SBI 20160125 Fixed CTRL + CTRL + key, the permitted default actions are copy, paste and cut. 
** 013 SBI 20160523 Fixed copy/cut/paste work flow.
**     SBI 20160603 Changed to process Safari COMMAND as it would be a CTRL key pressed, fixed
**                  INSERT key and automatic repeats for Safari.
** 014 EVL 20180321 Fix for CTRL-BREAK detect on Windows, accept 3 as key code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Returns the last array element.
 * 
 * @return   {Object}
 *           The last element or undefined if this array is empty.
 */
Array.prototype.peek = function()
{
   if (this.length > 0)
   {
      return this[this.length - 1];
   }
   return undefined;
};

p2j.keyboard = (function()
{
   /** Keyboard object which will be returned from this initialization function. */
   var me = {};    
   
   /** Interactive child process mode if true, normal 4GL compatibility mode if false. */
   me.vt100 = false;

   /**
    *  The key event handler.
    */
   var keyboardReader;
   
   /**
    * Initialize module.
    * 
    * @param    {Object} cfg
    *           The client configuration
    */
   me.init = function(cfg)
   {      
      if (keyboardReader)
      {
         document.removeEventListener('keydown', keyboardReader.onkeydown, true);
         document.removeEventListener('keypress', keyboardReader.onkeypress, true);
         document.removeEventListener('keyup', keyboardReader.onkeyup, true);
      }
      // init mapping   
      p2j.keymap.init(cfg);
      if(cfg.isGui)
      {
         keyboardReader = new GUIKeyboardReader();
      }
      else
      {
         keyboardReader = new CHUIKeyboardReader();
      }
      // TODO: confirm that this must exist after the on* function definitions
      document.addEventListener('keydown', keyboardReader.onkeydown, true);
      document.addEventListener('keypress', keyboardReader.onkeypress, true);
      document.addEventListener('keyup', keyboardReader.onkeyup, true);
   };
   
   /**
    * Process key strokes on VT100 mode.
    * 
    * @param {number} key key code
    * @param {event}  evt event object
    */
   var doVT100 = function(key, evt)
   {
      // process COMMAND key for Safari as CTRL key
      var metaToCtrl = p2j.isSafari && evt.metaKey;
      
      if (evt.ctrlKey || metaToCtrl)
      {
         // CTRL keys
         var charCode = p2j.keymap.mapKey(evt);
         if (charCode >= 97 && charCode <= 122)
         {
            charCode -= 32;
         }
         var ctrlCode = charCode & 0xffbf;
         
         sendKeyPair(0, ctrlCode, evt);
      }
      else if (evt.altKey)
      {
         // ALT keys are not used on VT100
         p2j.consumeEvent(evt);
      }
      else
      {
         switch (key)
         {  
            case keys.TAB:
            case keys.ENTER:
            case keys.ESCAPE:
               sendKeyPair(0, key, evt);
               break;
            case keys.DELETE:
               sendKeyPair(0x7f, 0, evt);
               break;
            case keys.BACKSPACE:
            case keys.PAGE_UP:
            case keys.PAGE_DOWN:
            case keys.LEFT_ARROW:
            case keys.RIGHT_ARROW:
            case keys.UP_ARROW:
            case keys.DOWN_ARROW:
            case keys.INSERT:
            case keys.HOME:
            case keys.END:
            case keys.F1:
            case keys.F2:
            case keys.F3:
            case keys.F4:
            case keys.F5:
            case keys.F6:
            case keys.F7:
            case keys.F8:
            case keys.F9:
            case keys.F10:
            case keys.F11:
            case keys.F12:
               sendKeyPair(key, 0, evt);
               break;
         }
      }
   };
                                                                              
   /**
    * Send key.
    * 
    * @param {number} key key code
    * @param {event}  evt event object
    */
   var sendKey = function(key, evt)
   {
      // if have clipboard
      if (p2j.clipboard.enabled)
      {
         // in GUI, alternate cut (shift-del), copy (ctrl-ins) and paste (shift-ins) keys
         // will generate the clipboard events; we need processing here for those events
         // when we generate them properly (shift-ins == 1022, shift-del == 639 and
         // ctrl-ins == 2558)
         if (p2j.isGui)
         {
            if (((key === 3 || key === 2558) && (!p2j.clipboard.needsSelection() || p2j.clipboard.haveSelection())) ||
                  ((key === 24 || key === 639) && !p2j.clipboard.needsSelection()))
            {
               p2j.clipboard.notifyAboutUserInitiatedAction();
               p2j.clipboard.prepareCopy();
               sendKeyCode(key);
               // let the browser generate the corresponding clipboard event by not suppressing the
               // default processing 
               return;
            }
         }
         else
         {
            // CTRL-C (copy) or CTRL-X (cut), CUT to system clipboard is not supported for ChUI
            if ((key === 3 && (!p2j.clipboard.needsSelection() || p2j.clipboard.haveSelection())) ||
                (key === 24 && !p2j.clipboard.needsSelection()))
            {
               // prepare clipboard operations on CTRL key down (copies selected text into a
               // hidden input field, focuses that field and selects the text)
               p2j.clipboard.prepareCopy();
               
               // let the browser generate the corresponding clipboard event by not suppressing the
               // default processing 
               return;
            }
         }
            
         // CTRL-V (paste)
         if (key === 22 || (p2j.isGui && (key === 1022)))
         {
            // use a not-visible input field and select all contents of that field
            p2j.clipboard.preparePaste();
            
            // default processing will overwrite the contents with the clipboard contents and
            // a paste listener in the p2j.clipboard module will read that text and send it to
            // the server; in GUI, the server will generate the CTRL-V key event in response to
            // the paste message, while in ChUI the key is eaten unless the clipboard is empty
            // (this last case is handled in the clipboard code, not here)
            return;
         }
      }
      sendKeyCode(key);
      
      p2j.consumeEvent(evt);
   };

   /**
    * Sends a compatible Progress 4GL key code message. 
    * 
    *  @param    {number} keyCode
    *            The key code
    */
   function sendKeyCode(keyCode)
   {
      p2j.socket.sendKeyCode(keyCode);
   }
   
   /**
    * Send key code and character code.
    * 
    * @param {number} keyCode key code
    * @param {number} charCode key character code
    * @param {event}  evt event object
    */
   function sendKeyPair(keyCode, charCode, evt)
   {
      p2j.socket.sendKeyPair(keyCode, charCode);
      
      p2j.consumeEvent(evt);
   };
   
   /**
    * The browser keyboard handler provided for CHUI interface.
    */
   function CHUIKeyboardReader()
   {
      /** Track escape key status. */
      var escape = false;
      
      /** Track alt char status. */
      var altChar = false;

      this.onkeypress = onkeypress;
      
      /** The last key down event */
      var lastEvent;
      
      /** The last key press code */
      var lastKeyPressCode;
      /**
       * Key press event.
       * 
       * @param {event} evt  event object.
       */
      function onkeypress(evt)
      {
         var key = evt.charCode;
         
         lastKeyPressCode = key;
         
         // process COMMAND key for Safari as CTRL key 
         var metaToCtrl = p2j.isSafari && evt.metaKey;

         // prevent Firefox keypress on CTRL-C and CTRL-V
         // Chrome doesn't generate keypress event for CTRL-C and CTRL-V
         if (p2j.clipboard.enabled && (p2j.isFirefox && evt.ctrlKey || metaToCtrl))
         {
            if (key === 99 || key === 118 || (p2j.isSafari && (key === 3 || key === 22)))
            {
               return;
            }
         }
        
         if (key !== 0)
         {
            if (me.vt100)
            {
               sendKeyPair(0, evt.charCode, evt);
            }
            else
            {
               sendKey(key, evt);
            }
         }
      };
      
      this.onkeyup = onkeyup;
      /**
       * Key up event handler.
       * 
       * @param {event} evt  event object.
       */
      function onkeyup(evt)
      {
         if (evt.keyCode === keys.ALT && evt.location === evt.DOM_KEY_LOCATION_RIGHT)
         {
            altChar = false;
         }
         // Safari generates only keyup event if INSERT key has been pressed
         if (p2j.isSafari && evt.keyCode === keys.INSERT)
         {
            onkeydown(evt);
         }
         // stop event propagation
         p2j.consumeEvent(evt);
      };

      this.onkeydown = onkeydown;
      /**
       * Key down event handler.
       * 
       * @param {event} evt  event object.
       */
      function onkeydown(evt)
      {
         var key;
         var peer;
         if (evt.keyCode === keys.IME_PROCESSKEY)
         {
            if (lastEvent !== undefined && lastEvent.keyCode !== keys.IME_PROCESSKEY)
            {
               key  = lastEvent.keyCode;
               peer = lastEvent;
               
               var keyPressEvt = p2j.createKeyEvent("keypress", false, true, null,
                     lastEvent.ctrlKey,
                     lastEvent.altKey,
                     lastEvent.shiftKey,
                     lastEvent.metaKey,
                     key, lastKeyPressCode ? lastKeyPressCode : key);
               
               // check if the keyboard event is created successfully
               if (keyPressEvt)
               {
                  onkeypress(keyPressEvt);
               }
               
               p2j.consumeEvent(evt);
               return;
            }
            else
            {
               key  = evt.keyCode;
               peer = evt;
            }
         }
         else
         {
            lastEvent = evt;
            key  = evt.keyCode;
            peer = evt;
         }
         
         if (altChar || key === keys.CTRL || key === keys.SHIFT || key === keys.LEFT_WINDOW ||
             key === keys.RIGHT_WINDOW)
         {
            
            return;
         }

         if (key === keys.ALT)
         {
            if (evt.location === evt.DOM_KEY_LOCATION_RIGHT)
            {
               altChar = evt.ctrlKey;
            }
            
            return;
         }
         
         // in interactive child process mode handle as a key pair 
         if (me.vt100)
         {
            doVT100(key, peer);
            return;
         }
         
         if (key === keys.ESCAPE)
         {
            evt.preventDefault();
            
            // ESC is used to clear selection (sometimes)
            if (p2j.clipboard.needsSelection() && p2j.clipboard.haveSelection())
            {
               p2j.clipboard.clearSelection();
            }
            else
            {
               escape = !escape;
            }
            
            return;
         };
         
         // process COMMAND key for Safari as CTRL key 
         var metaToCtrl = p2j.isSafari && peer.metaKey;
         
         var p4code = -1;
         var charCode = p2j.keymap.mapKey(peer);
         
         if (charCode >= 97 && charCode <= 122)
         {
            charCode -= 32;
         }
         
         // handle ALT or ESC
         if (peer.altKey || escape)
         {
            p4code = p2j.keymap.xlateAlt[key];
            
            if (p4code === undefined)
            {
               if (charCode !== -1 && ((!escape && peer.altKey) || (escape && !peer.altKey)))
               {
                  p4code = p2j.keymap.processCtrl(peer, charCode);
                  
                  if(p4code === -1)
                  {
                     p4code = charCode;
                  }
                  
                  p4code |= 1024;
               }
               else
               {
                  p4code = -1;
               }
            }
            else
            {
               if (!escape && peer.altKey && key === keys.DELETE)
               {
                  // special case for ALT+DELETE key combo
                  p4code = 512;
               }     
            }    
                       
            escape = false;
         }
         else
         {
            // normal map
            var p4code = p2j.keymap.xlate[key];
            
            if (p4code === undefined)
            {
               if (charCode > 0)
               {
                  // check CTRL
                  p4code = p2j.keymap.processCtrl(peer, charCode);
               }
               else
               {
                  p4code = -1;
                  if (peer.ctrlKey || metaToCtrl)
                  {
                     switch(key)
                     {
                        case keys.PAUSE:
                           p4code = 541;
                           break;
                     }
                  }
               }
            }
            else
            {
               if (peer.shiftKey)
               {
                  switch(p4code)
                  {
                     case Progress4GLKeys.VK_TAB:
                        p4code = 509;
                        break;
                     case Progress4GLKeys.VK_BACKSPACE:
                        p4code = 520;
                        break;
                     case Progress4GLKeys.VK_ENTER:
                        p4code = 525;
                        break;
                     case Progress4GLKeys.VK_F1:
                        p4code = 813;
                        break;
                     case Progress4GLKeys.VK_F2:
                        p4code = 814;
                        break;
                     case Progress4GLKeys.VK_F3:
                        p4code = 815;
                        break;
                     case Progress4GLKeys.VK_F4:
                        p4code = 816;
                        break;
                     case Progress4GLKeys.VK_F5:
                        p4code = 817;
                        break;
                     case Progress4GLKeys.VK_F6:
                        p4code = 818;
                        break;
                     case Progress4GLKeys.VK_F7:
                        p4code = 819;
                        break;
                     case Progress4GLKeys.VK_F8:
                        p4code = 820;
                        break;
                     case Progress4GLKeys.VK_F9:
                        p4code = 821;
                        break;
                     case Progress4GLKeys.VK_F10:
                        p4code = 822;
                        break;
                     case Progress4GLKeys.VK_F11:
                        p4code = 823;
                        break;
                     case Progress4GLKeys.VK_F12:
                        p4code = 824;
                        break;
                     case Progress4GLKeys.VK_CURSOR_UP:
                        p4code = 1013;
                        break;
                     case Progress4GLKeys.VK_CURSOR_DOWN:
                        p4code = 1014;
                        break;
                     case Progress4GLKeys.VK_CURSOR_RIGHT:
                        p4code = 1015;
                        break;
                     case Progress4GLKeys.VK_CURSOR_LEFT:
                        p4code = 1016;
                        break;
                     case Progress4GLKeys.VK_HOME:
                        p4code = 1017;
                        break;
                     case Progress4GLKeys.VK_END:
                        p4code = 1018;
                        break;
                     case Progress4GLKeys.VK_PAGE_UP:
                        p4code = 1019;
                        break;
                     case Progress4GLKeys.VK_PAGE_DOWN:
                        p4code = 1020;
                        break;
                     case Progress4GLKeys.VK_INSERT:
                        p4code = 1022;
                        break;
                  }
               }
            }
         }
         
         if (p4code !== -1)
         {
            sendKey(p4code, peer);
         }      
      };
   };
   
   function GUIKeyboardReader()
   {
      /**
       * The keystrokes queue.
       */
      var queue = [];
      
      /**
       * Tests if the keys strokes can invoke the permitted default actions: cut (shift + del,
       * ctrl + x), copy (ctrl + ins, ctrl + c) and paste (shift + ins, ctrl + v). For Safari,
       * pressing COMMAND + C, COMMAND + X and COMMAND + V can invoke the system clipboard copy,
       * cut and paste actions.
       * 
       * @param    {Event} evt
       *           The merged key stroke event that represents two pressed keys: a modifier and 
       *           a standard key.
       * 
       * @return   True if the keys strokes are from this set: shift + del, ctrl + x, ctrl+ins,
       *           ctrl + c, shift + ins and ctrl + v, otherwise false.
       */
      function permittedKeyStrokes(evt)
      {
         if (p2j.isSafari)
         {
            if (evt.metaKey &&
                (evt.keyCode === keys.X || evt.keyCode === keys.C || evt.keyCode === keys.V))
            {
               return true;
            }
         }
         else
         {
            if (evt.shiftKey && (evt.keyCode === keys.DELETE || evt.keyCode === keys.INSERT))
            {
               return true;
            }
            else if ((evt.ctrlKey && (evt.keyCode === keys.X || evt.keyCode === keys.C ||
                        evt.keyCode === keys.V || evt.keyCode === keys.INSERT)))
            {
               return true;
            }
         }
         
         return false;
      }
      
      /**
       * Tests if the given key code is a modifier.
       * 
       * @param    {Number} key
       *           The browser code of the pressed key
       * 
       * @return   {Boolean}
       *           True iff a key is from the following set: SHIFT, ALT, CTRL, COMMAND.
       */
      function isModifier(key)
      {
         return key === keys.SHIFT ||
                key === keys.CTRL  ||
                key === keys.ALT   ||
                key === keys.LEFT_WINDOW || key === keys.RIGHT_WINDOW;
      }
      
      /**
       * Merges the browser keyboard event with the last key event.
       * 
       * @param    {Object} last
       *           The last key event in the keystrokes queue.
       * @param    {KeyboardEvent} evt
       *           The browser keyboard event
       */
      function mergeWithLastEvent(last, evt)
      {
         last.keyCode  = evt.keyCode;
         last.charCode = evt.charCode;
         if (evt.shiftKey || evt.keyCode === keys.SHIFT)
         {
            last.shiftKey = true;
         }
         if (evt.altKey || evt.keyCode === keys.ALT)
         {
            last.altKey = true;
         }
         if (evt.ctrlKey || evt.keyCode === keys.CTRL)
         {
            last.ctrlKey = true;
         }
         if (evt.keyCode === keys.CTRL)
         {
            last.ctrlKey = true;
            last.ctrlCount++;
         }
         if (evt.metaKey || evt.keyCode === keys.LEFT_WINDOW || evt.keyCode === keys.RIGHT_WINDOW)
         {
            last.metaKey = true;
         }
      }

      this.onkeydown = onkeydown;

      /**
       * Handles the keydown events and pushes them into the the keystrokes queue.
       * 
       * @param    {KeyboardEvent} evt
       *           The browser generated keyboard event.
       */
      function onkeydown(evt)
      {
         if (p2j.isDialogDisplayed())
         {
            return false;
         }
         var last = queue.peek();
         if (last === undefined)
         {
            last = {
                  keyCode   : evt.keyCode,
                  shiftKey  : evt.shiftKey,
                  altKey    : evt.altKey,
                  ctrlKey   : evt.ctrlKey,
                  metaKey   : evt.metaKey,
                  charCode  : evt.charCode,
                  ctrlCount : (evt.ctrlKey || evt.keyCode === keys.CTRL) ? 1 : 0
               };
         }
         else
         {
            if (isModifier(last.keyCode))
            {
               mergeWithLastEvent(last, evt);
               var permitted = permittedKeyStrokes(last);
               if (!permitted)
               {
                  p2j.consumeEvent(evt);
               }
               return permitted;
            }
            else if (last.keyCode === evt.keyCode || evt.keyCode === keys.IME_PROCESSKEY)
            {  
               console.debug(last.ctrlCount);
               var repeatCode = last.keyCode;
               // auto-repeats that have no typed keys.
               last = {
                         keyCode   : repeatCode,
                         shiftKey  : evt.shiftKey,
                         altKey    : evt.altKey,
                         ctrlKey   : evt.ctrlKey,
                         metaKey   : evt.metaKey,
                         charCode  : evt.charCode,
                         ctrlCount : last.ctrlCount
                      };
               
               if (repeatCode !== keys.IME_PROCESSKEY)
               {
                  queue.push(last);
                  if (evt.keyCode === keys.IME_PROCESSKEY)
                  {
                     var keyUpEvt = p2j.createKeyEvent("keyup", false, true, null, evt.ctrlKey,
                           evt.altKey, evt.shiftKey, evt.metaKey, repeatCode, 0);
                     
                     // check if the keyboard event is created successfully
                     if (keyUpEvt)
                     {
                        onkeyup(keyUpEvt);
                     }
                     
                     p2j.consumeEvent(evt);
                     return;
                  }
                  else
                  {
                     onkeyup(evt);
                  }
               }
               
               var permitted = permittedKeyStrokes(last);
               if (!permitted)
               {
                  p2j.consumeEvent(evt);
               }
               return permitted;
            }
            else
            {
               last = {
                         keyCode   : evt.keyCode,
                         shiftKey  : evt.shiftKey,
                         altKey    : evt.altKey,
                         ctrlKey   : evt.ctrlKey,
                         metaKey   : evt.metaKey,
                         charCode  : evt.charCode,
                         ctrlCount : (evt.ctrlKey || evt.keyCode === keys.CTRL) ? 1 : 0
                      };
            }
         }
         queue.push(last);

         var permitted = permittedKeyStrokes(last);
         if (!permitted)
         {
            p2j.consumeEvent(evt);
         }
         return permitted;
      };

      /**
       * Tests if the pressed key is @, ^, _ , \, ]
       * 
       * @param    {KeyboardEvent} evt
       *           The browser generated keyboard event.
       * 
       * @return   {Boolean}
       *           True iff the pressed key is one of @, ^, _ , \, ]
       */
      function isSpecialCharacter(evt)
      {
         var code = evt.charCode == 0 ? p2j.keymap.mapKey(evt) : evt.charCode;
         if (evt.shiftKey)
         {
            return (code === 64) || (code === 94) || (code === 95) ||
            (code === 124) || (code === 125);
         }
         return (code === 50) || (code === 54) || (code === 45) ||
                  (code === 92) || (code === 93);
      }
      
      /**
       * Tests if the pressed key is a letter.
       * 
       * @param    {KeyboardEvent} evt
       *           The browser generated keyboard event.
       * 
       * @return   {Boolean}
       *           True iff the pressed key is one of [A-Za-z]
       */
      function isAlphabeticCharacter(evt)
      {
         var charCodeIsNotSet = evt.charCode == 0;
         var code = charCodeIsNotSet ? evt.keyCode : evt.charCode; 
         if (evt.shiftKey || charCodeIsNotSet)
         {
            return (code >= 65) && (code <= 90);
         }
         else
         {
            return (code >= 97) && (code <= 122);
         }
      }

      this.onkeypress = onkeypress;

      /**
       * This method is called when a pressed key (or a multi-key combination) is converted to a 
       * higher level representation. As a general rule, all key codes between 0 and 127 inclusive 
       * are generated through this callback. The reason for this is that Java translates the keys or
       * multi-key combinations into the higher level representation which almost perfectly matches 
       * the 4GL key codes 0 through 127.  In other words, the keyboard inputs of the user are 
       * very compatible between the 4GL and Java for those portions of the key codes (0 through 127).
       * <p>
       * Once a valid key code is generated, that key code is added to the key input queue in FIFO 
       * order.
       * <p>
       * The non-visual values 0 through 31 are handled almost "for free". The user generates these 
       * in the following manner:
       * <p>
       * <pre>
       * Key Code  User Input               Notes
       * --------  -----------------------  --------------------------------------
       * 0         CTRL+2                   Special code in this method supports this.
       * 1         CTRL+A
       * 2         CTRL+B
       * 3         CTRL+C
       * 4         CTRL+D
       * 5         CTRL+E
       * 6         CTRL+F
       * 7         CTRL+G
       * 8         CTRL+H or VK_BACKSPACE
       * 9         CTRL+I or VK_TAB         The containing frame's focus traversal
       *                                    keys must be forced to the empty set in
       *                                    order for this to work.
       * 10        CTRL+J or VK_ENTER       NOT DELIVERED, handled in keyPressed()
       * 11        CTRL+K
       * 12        CTRL+L
       * 13        CTRL+M
       * 14        CTRL+N
       * 15        CTRL+O
       * 16        CTRL+P
       * 17        CTRL+Q
       * 18        CTRL+R
       * 19        CTRL+S
       * 20        CTRL+T
       * 21        CTRL+U
       * 22        CTRL+V
       * 23        CTRL+W
       * 24        CTRL+X
       * 25        CTRL+Y
       * 26        CTRL+Z
       * 27        VK_ESCAPE                DELIVERED; Windows mode allows this.
       * 28        CTRL+\
       * 29        CTRL+]
       * 30        CTRL+6                   Special code in this method supports this.
       * 31        CTRL+-                   Special code in this method supports this.
       * </pre>
       * <p>
       * ALT key is the modifier which sets the key code's 10th bit.  If this is determined to be
       * pressed together with the key, than bit 10 will be set and this key will be pushed to the
       * {@link SwingKeyboardReader#queue}, but only if the key code's label is not the empty string.
       * 
       * @param    {KeyboardEvent} evt
       *           The browser generated keyboard event.
       */
      function onkeypress(evt)
      {
         if (p2j.isDialogDisplayed())
         {
            return false;
         }
         
         var lastKey = queue.pop();
         
         if (lastKey === undefined)
         {
            return false;
         }
         
         handleKeyPress(evt, lastKey);
      };
      
      /**
       * Handles the typed keypress events.
       * 
       * @param    {KeyboardEvent} evt
       *           The current keyboard event.
       * @param    {KeyboardEvent} lastKey
       *           The latest keyboard event merged with keys modifiers.
       * @param    {Number} printedCode
       *           The ASCII code of the latest keyboard event.
       * @param    {Number} mappedCode
       *           The Progress4GL key code for the latest keyboard event.
       * @param    {Boolean} ctrl
       *           The flag indicating the ctrl modifier is active or not.
       * @param    {Boolean} alt
       *           The flag indicating the alt modifier is active or not.
       * @param    {Boolean} shift
       *           The flag indicating the shift modifier is active or not.
       * @param    {Boolean} meta
       *           The flag indicating the meta modifier is active or not.
       */
      function handleTypedKey(evt, lastKey, printedCode, mappedCode, ctrl, alt, shift, meta)
      {
         // In general (printedCode === evt.charCode) is true but if the keys are permuted,
         // then we can rely on the lastKey that is merged with keys modifiers
         var charCode = evt.charCode > 0 ? evt.charCode : printedCode;
         var code = -1;
         
         // process COMMAND key for Safari as CTRL key
         var metaToCtrl = p2j.isSafari && meta;
         
         if (ctrl || metaToCtrl)
         {
            if (isAlphabeticCharacter(lastKey))
            {
               if (!shift)
               {
                  charCode -= 32;
               }
            }
            code = p2j.keymap.processCtrl(lastKey, charCode);
            // not alphabetic character
            if ((lastKey.keyCode < 65 || lastKey.keyCode > 90) && !isSpecialCharacter(lastKey))
            {
               code = lastKey.keyCode;
               code = code | 1 << 11;
            }
            // CTRL + CTRL + key
            if (lastKey.ctrlCount && lastKey.ctrlCount > 1 || (lastKey.ctrlCount && metaToCtrl))
            {
               code = code | 1 << 11;
            }
            if (alt)
            {
               code = processAlt(code);
            }
            if (shift)
            {
               // not alphabetic character
               if ((lastKey.keyCode < 65 || lastKey.keyCode > 90) && !isSpecialCharacter(lastKey))
               {
                  code = code | 1 << 9;
                  
               }
               else
               {
                  code = processShift(code);
               }
            }
         }
         else
         {
            code = mappedCode === undefined ? charCode : mappedCode;
            if (alt)
            {
               // Remapping lowercase Latin to uppercase. This differs
               // with Progress but seems to be more convenient to user because
               // Ctrl+<letter> combinations are case insensitive (and even
               // keyboard layout insensitive). Note the usage of direct
               // calculation not Character class methods. This is
               // because we are dealing with integer key codes not characters
               // so this is dirty hack anyway.
               if (code >= 97 && code <= 122)
               {
                  code -= 32;
               }
               code = processAlt(code);
            }
            if (shift)
            {
               code = processShift(code);
            }
         }

         sendKey(code, evt);
      }
      
      /**
       * Handles the keypress events that have no typed characters.
       * 
       * @param    {KeyboardEvent} evt
       *           The current keyboard event.
       * @param    {KeyboardEvent} lastKey
       *           The latest keyboard event merged with keys modifiers.
       * @param    {Number} mappedCode
       *           The Progress4GL key code for the latest keyboard event.
       * @param    {Boolean} ctrl
       *           The flag indicating the ctrl modifier is active or not.
       * @param    {Boolean} alt
       *           The flag indicating the alt modifier is active or not.
       * @param    {Boolean} shift
       *           The flag indicating the shift modifier is active or not.
       * @param    {Boolean} meta
       *           The flag indicating the meta modifier is active or not.
       */
      function handleNonTypedKey(evt, lastKey, mappedCode, ctrl, alt, shift, meta)
      {
         // In general (lastKey.keyCode === evt.keyCode) is true but if the keys are permuted,
         // then we can rely on the lastKey that is merged with keys modifiers
         var code = mappedCode === undefined ? lastKey.keyCode : mappedCode;
         
         // process COMMAND key for Safari as CTRL key
         var metaToCtrl = p2j.isSafari && meta;
         
         if (ctrl || metaToCtrl)
         {
            // in Windows we have 3 as CTRL-BREAK key code
            if (lastKey.keyCode === keys.PAUSE || lastKey.keyCode === keys.CTRL_BREAK)
            {
               code = Progress4GLKeys.VK_CTRL_BREAK;
            }
            else
            {
               code = code | 1 << 11;
            }
            // CTRL + CTRL + key
            if (lastKey.ctrlCount && lastKey.ctrlCount > 1 || (lastKey.ctrlCount && metaToCtrl))
            {
               code = code | 1 << 11;
            }
         }
         if (shift)
         {
            if (code === Progress4GLKeys.VK_TAB)
            {
               code = Progress4GLKeys.VK_BACK_TAB;
            }
            else
            {
               code = processShift(code);
            }
         }
         if (alt)
         {
            code = processAlt(code);
         }

         return sendKey(code, evt);
      }
      
      /**
       * Handles the keypress events.
       * 
       * @param    {KeyboardEvent} evt
       *           The current keyboard event.
       * @param    {KeyboardEvent} lastKey
       *           The latest keyboard event merged with keys modifiers.
       */
      function handleKeyPress(evt, lastKey)
      {
          var ctrl  = lastKey.ctrlKey  || evt.ctrlKey;
          var alt   = lastKey.altKey   || evt.altKey;
          var shift = lastKey.shiftKey || evt.shiftKey;
          var meta  = lastKey.metaKey  || evt.metaKey;
          // in general (lastKey.keyCode === evt.keyCode) is true but if the keys are permuted,
          // then we can rely on the lastKey that is merged with keys modifiers
          var mappedCode  = p2j.keymap.xlate[lastKey.keyCode];
          
          var printedCode = p2j.keymap.mapKey(lastKey);
          if (printedCode === -1)
          {
             //  a non printed key is pressed 
             handleNonTypedKey(evt, lastKey, mappedCode, ctrl, alt, shift, meta);
          }
          else
          {
             // a printed key is pressed
             handleTypedKey(evt, lastKey, printedCode, mappedCode, ctrl, alt, shift, meta);
          }
      }
      
      this.onkeyup = onkeyup;
      
      /**
       * Handles the keyup events.
       * 
       * @param    {KeyboardEvent} evt
       *           The browser generated keyboard event.
       */
      function onkeyup(evt)
      {
         if (p2j.isDialogDisplayed())
         {
            return false;
         }
         // Safari generates only keyup event if INSERT key has been pressed
         if (p2j.isSafari && evt.keyCode === keys.INSERT)
         {
            onkeydown(evt);
         }
         
         var lastKey = findKeyEventByKeyCode(evt.keyCode);
         
         if (lastKey === undefined)
         {
            return false;
         }
         
         handleKeyPress(evt, lastKey);
      };

      /**
       * Finds the first key down event that has the given key code.
       * 
       * @param    {Number} keyCode
       *           The key event code
       * 
       * @return   {KeyboardEvent}
       *           The down key event with the given key code 
       */
      function findKeyEventByKeyCode(keyCode)
      {
         var pos = queue.length - 1;
         for (;pos >= 0; pos--)
         {
            var evt = queue[pos];
            // modifier key down events are merged
            if ((evt.keyCode === keyCode)                ||
                (evt.altKey   && keyCode === keys.ALT)   ||
                (evt.shiftKey && keyCode === keys.SHIFT) ||
                (evt.ctrlKey  && keyCode === keys.CTRL)  ||
                (evt.metaKey  && keyCode === keys.LEFT_WINDOW) ||
                (evt.metaKey  && keyCode === keys.RIGHT_WINDOW))
            {
               queue.splice(pos, 1);
               return evt;
            }
         }
         return undefined;
      }

      /**
       * Process the given key by setting the 10th bit.
       *
       * @param    {Number} code
       *           The key code to be processed.
       *           
       * @return   {Number}
       *           The processed key code or -1 if the ALT-key combination has no label.
       */
      function processAlt(code)
      {
         // modify the next character (in GUI 4GL, ALT is a modifier key)
         code = code | 1 << 10;
         
         if (!p2j.keymap.keyboardMapping.hasLabelForCode(code))
         {
            code = -1;
         }
         
         return code;
      };
      
      /**
       * Process the given key by setting the 9th bit.
       *
       * @param    {Number} code
       *           The key code to be processed.
       *           
       * @return   {Number}
       *           The processed key code or -1 if the SHIFT-key combination has no label.
       */
      function processShift(code)
      {
         // ignore tab - this is treated separately
         if (code == keys.TAB)
         {
            return -1;
         }

         var orig = code;
         // modify the next character (in GUI 4GL, SHIFT is a modifier key)
         code = code | 1 << 9;
         // return original code if there is no label or if the label doesn't end with
         // the original key
         if (!p2j.keymap.keyboardMapping.hasLabelForCode(code) ||
               p2j.keymap.keyboardMapping.shiftModifierExceptions(code) ||
               p2j.keymap.keyboardMapping.shiftModifierExceptions(orig))
         {
            return orig;
         }

         return code;
      };

   };
   
   return me;
}());