/*
** Module   : WindowEvent.java
** Abstract : Container for an event generated by client-side, having as source a window event.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141020 Created initial version.
** 002 CA  20150415 Allow window events to be chained.  This is required because acting on a 
**                  window can cause events to be raised for the entire sub-tree to which this
**                  acts like a root.  And although OS events are generated for the entire 
**                  sub-tree, 4GL events are dependent on the trigger's consume state (it ends
**                  on the first consumed trigger in the chain).
** 004 HC  20160104 Implemented toString().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Container for an event generated by client-side, having as source a OS-level event generated 
 * on a window.
 */
public class WindowEvent
extends Event
{
   /** The 4GL action associated with this event. */
   private int action;

   /**
    * The id of the source window which was the root cause of this event.  This is different  
    * from the {@link Event#source()}, which represents the window on which the event is applied.
    */
   private int origin = -1;

   /** The next event to be raised, if the trigger does not consume this one. */
   private WindowEvent nextEvent = null;
   
   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    id
    *           The event ID.
    */
   public WindowEvent(Widget<?> widget, int id)
   {
      this(widget, id, id);
   }

   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    id
    *           The event ID.
    * @param    action
    *           The associated action.
    */
   public WindowEvent(Widget<?> widget, int id, int action)
   {
      super(widget, id);
      
      this.action = action;
   }
   
   /**
    * Get the {@link #action}.
    * 
    * @return    See above.
    */
   public int getAction()
   {
      return action;
   }

   /**
    * Get the origin of this event (the id of the source window which was the root cause of this 
    * event).
    * 
    * @return   The {@link #origin}.
    */
   public int getOrigin()
   {
      return origin;
   }
   
   /**
    * Set the origin of this event (the id of the source window which was the root cause of this 
    * event).
    * 
    * @param    windowId
    *           The window ID. 
    */
   public void setOrigin(int windowId)
   {
      this.origin = windowId;
   }
   
   /**
    * Set the next event to be raised, if the trigger does not consume this one.
    * 
    * @param    next
    *           The next window event to be raised.
    */
   public void setNextEvent(WindowEvent next)
   {
      this.nextEvent = next;
   }
   
   /**
    * Get the next event to be raised, if the trigger does not consume this one.
    * 
    * @return   The {@link #nextEvent}.
    */
   public WindowEvent getNextEvent()
   {
      return this.nextEvent;
   }

   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      String actionStr = "unknown";
      switch (action)
      {
         case Keyboard.SE_WINDOW_CLOSE:
            actionStr = "CLOSE";
            break;
         case Keyboard.SE_WINDOW_MAXIMIZED:
            actionStr = "MAXIMIZED";
            break;
         case Keyboard.SE_WINDOW_MINIMIZED:
            actionStr = "MINIMIZED";
            break;
         case Keyboard.SE_WINDOW_RESTORED:
            actionStr = "RESTORED";
            break;
         case Keyboard.SE_WINDOW_RESIZED:
            actionStr = "RESIZED";
            break;
         default:
            actionStr = "unknown";
      }
      
      return "WindowEvent [source="    + Utils.nonNull(source(), "<null>") + ", " +
                          "action="    + actionStr + ", " + 
                          "origin="    + Utils.nonNull(origin, "<null>") + ", " + 
                          "nextEvent=" + Utils.nonNull(nextEvent, "<null>") + ", " +
                          "]";
   }
}