/*
** Module   : AlertBoxGuiImpl.java
** Abstract : ALERT-BOX GUI client widget implementation. 
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150612 Initial version.
** 002 HC  20150702 Improved positioning of GUI alert-box on systems with multiple displays.
** 003 HC  20150831 Scrolling support overhaul.
** 004 CA  20150826 Added movable support.
** 005 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 006 HC  20151002 Draw fixes.
** 007 HC  20151021 Fixed an abend in GUI ALERT-BOX caused by wrong button initialization. Fixed 
**                  highlighting of default button in ALERT-BOX. Improved layout. Proper handling 
**                  of close window caption button.
** 008 HC  20151023 Fixed GUI ALERT-BOX keyboard input processing.
**         20151110 Changed ALERT-BOX owner to be always the DEFAULT-WINDOW.
** 009 IAS 20160315 Support for the DEBUG-ALERT attribute
** 010 IAS 20160328 Client log support
** 011 CA  20160216 Cache the intermediate widget states.
**     SBI 20160329 Fixed NPE in the Web client embedded mode.
** 012 HC  20160418 Fixed ALERT-BOX title bar close button to close the box with correct result.
** 013 HC  20160428 Improved ALERT-BOX layout to closer match the native 4GL. ALERT-BOX icons are
**                  loaded directly from SVG resources.
** 014 VIG 20160315 Removed underscore setting for buttons.
** 015 SBI 20160811 The alert box owner should be CURRENT-WINDOW.
** 016 OM  20170320 Added support for Themes.
** 017 SBI 20161101 Renamed getWidgetList() to widgetsAsList().
** 018 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 019 HC  20171219 Fixed alert-box window owner and the dismiss behavior not to close its parent
**                  window.
** 020 HC  20180116 Prevented an unexpected exception.
** 021 OM  20180321 Fixed message geometry.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.KeyInput;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.logical;

/**
 * This class is the specific implementation of the AlertBox for GUI.
 */
public class AlertBoxGuiImpl
extends ModalWindow
implements AlertBox<GuiOutputManager>
{
   /** Max message width in pixels. */
   private static int MAX_MESSAGE_PIXEL_WIDTH = 500;

   /** Min message width in pixels. */
   private static int MIN_MESSAGE_PIXEL_WIDTH = 300;
   
   /** Min message height in pixels. */
   private static int MIN_MESSAGE_PIXEL_HEIGHT = 22;
   
   /** Width of alert box buttons in characters. */
   private static double BUTTONS_CHAR_WIDTH = 15;
   
   /** Height of alert box buttons in characters. */
   private static double BUTTONS_CHAR_HEIGHT = 1.04;
   
   /** The alert type. */
   private int alertType;
   
   /** Alert box message text. */
   private String messageText;
   
   /** Button list. */
   private List<Button<GuiOutputManager>> buttons;
   
   /** The alert box icon. */
   private ImageWrapper<?> icon;
   
   /** The message font. */
   private FontDetails font;
   
   /** The externalized implementation. */
   private AlertBoxCommons<GuiOutputManager> commons;
   
   /** Cached driver. */
   private GuiDriver gd;
   
   /** Message panel. */
   private MessagePanel messagePanel;
   
   /** Buttons panel. */
   private ButtonsPanel buttonsPanel = new ButtonsPanel();

   /**
    * Default constructor.
    * 
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    content
    *           Array of elements which form alert box content.
    * @param    buttonSet
    *           Button set type to be constructed for this AlertBox.
    * @param    var
    *           The variable to be edited.  The initial value sets the 
    *           default button and the resulting value is changed by the
    *           user's press of a button.
    * @param    title
    *           AlertBox title to be rendered at the top of the one line
    *           border.
    * @param    stack
    *           Progress stack trace
    */
   public AlertBoxGuiImpl(int alertType, 
                          Object[] content, 
                          int buttonSet, 
                          logical var, 
                          String title,
                          String[] stack)
   {
      super(title, WindowManager.getAlertBoxOwnerWindow());
      
      this.alertType = alertType;
      commons = new AlertBoxCommons<>(this,
                                      buttonsPanel,
                                      screen().getFactory(), 
                                      content, 
                                      buttonSet, 
                                      var, 
                                      null,
                                      stack);

      init();
      if (stack != null && alertType == LogicalTerminal.APPL_ALERT)
      {
         ThinClient.getInstance().clientLog(messageText, stack);
      }
   }
   
   /**
    * Layout.
    */
   @Override
   public void doLayout()
   {
      messagePanel.doLayout();
      
      NativeDimension messageDiv = messagePanel.physicalDimension();
      
      buttonsPanel.setPhysicalLocation(0, messageDiv.height);
      buttonsPanel.setWidth(messagePanel.width());
      buttonsPanel.doLayout();
      
      Container<GuiOutputManager> content = getContentPane();
      
      Dimension messageDim = messagePanel.dimension();
      Dimension buttonsDim = buttonsPanel.dimension();
      
      double effectiveWidth = Math.max(messageDim.width, buttonsDim.width);
      
      messageDim.width = effectiveWidth;
      buttonsDim.width = effectiveWidth; 
      
      messagePanel.setSize(messageDim);
      buttonsPanel.setSize(buttonsDim);
      
      content.setSize(new Dimension(effectiveWidth, messageDim.height + buttonsDim.height));
      
      super.doLayout();
      
      // place the window at the display center
      NativeRectangle bounds = ThinClient.getInstance().getDisplayBounds(getId().asInt());
      NativeDimension dim = physicalDimension();
      setPhysicalLocation(bounds.left() + (bounds.width()  - dim.width) / 2, 
                          bounds.top()  + (bounds.height() - dim.height) / 2);
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      init();

      super.show();
   }
   
   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      super.hide();
      
      destroy();
   }

   /**
    * Close this window. This should be only called when window close caption
    * button is cliced.
    */
   @Override
   public void close()
   {
      logical result = null;
      switch (commons.buttonSet)
      {
         case LogicalTerminal.BTN_OK:
            result = new logical(true);
            break;
         case LogicalTerminal.BTN_OK_CANCEL:
            result = new logical(false);
            break;
         case LogicalTerminal.BTN_RETRY_CANCEL:
            result = new logical(false);
            break;
         case LogicalTerminal.BTN_YES_NO_CANCEL:
            result = new logical();
            break;
         default:
            // when YES_NO the close window caption button is disabled
            // so this method should not be called
            throw new IllegalStateException("Unexpected button set!");
      }

      setResult(result);
      dismiss();
   }
   
   /**
    * Move the input focus to the first widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the last one within the container and
    * we have to simulate the Progress behavior in which the next focused
    * widget will be the first widget which can receive a focus in a
    * container.
    */
   @Override
   public void nextFocus()
   {
      commons.nextFocus();
   }
   
   /**
    * Move the input focus to the last widget which can receive a focus in
    * the peer container. This is required to handle the situation when the
    * current widget with the focus is the first one within the container and
    * we have to simulate the Progress behavior in which the previous focused
    * widget will be the last widget which can receive a focus in a container.
    */
   @Override
   public void prevFocus()
   {
      commons.prevFocus();
   }
   
   /**
    * Find the button that matches a passed key event.
    * 
    * @param    keyEvent
    *           Event to check.
    * 
    * @return   <code>true</code> if matching button is found and event is
    *           processed.
    */
   @Override
   public boolean findButton(KeyInput keyEvent)
   {
      return commons.findButton(keyEvent);
   }

   /**
    * Let the modal UI element now that it should finish its modal input processing.
    */
   @Override
   public void dismiss()
   {
      commons.dismiss();
   }

   /**
    * Returns <code>true</code> when the modal UI element has dismissed (finished its modal input).
    *
    * @return  see above.
    */
   @Override
   public boolean isDismissed()
   {
      return commons.isDismissed();
   }

   /**
    * Helper method to get the edited result when finished.
    * 
    * @return   The result of the editing operation which was set depending
    *           on the pressed button.
    */
   @Override
   public logical getResult()
   {
      return commons.getResult();
   }

   /**
    * Sets the alert-box result value.
    *
    * @param   value
    *          The result logical value.
    */
   @Override
   public void setResult(logical value)
   {
      commons.setResult(value);
   }
   
   /**
    * Returns the tab item list for this container. If the tab item list is not given, then
    * an empty list is returned.
    * 
    * @return   The tab item list.
    */
   @Override
   public List<Widget<GuiOutputManager>> getTabItemList()
   {
      return getContentPane().widgetsAsList();
   }
   
   /**
    * Create window for the stack trace
    * 
    * @param title
    *        window title
    * 
    * @return window for the stack trace
    */
   @Override
   public TitledWindow createStackTraceWindow(String title)
   {
      final ModalWindow win = new DialogBoxWindow(title);
      ThinClient.getInstance().eventBracket(() -> 
         {
            win.realize();
            win.setEnabled(true);
         }
      );
      return win; 
   }

   /**
    * Create effectively invisible widget (the gap) for the stack trace dialog
    * to avoid explicit coordinates' usage 
    * TODO: at this moment the layout manager doesn't support this approach.
    *   
    * @param registry 
    *        widget registry
    * 
    * @return the gap widget
    */
   @Override
   public Widget<GuiOutputManager> createGap(WidgetRegistry<GuiOutputManager> registry)
   {
      RectangleConfig rcfg = new RectangleConfig(WidgetId.nextID());
      rcfg.clientHeightPixels = 15;
      rcfg.widthChars = 6;
      rcfg.filled = false;
      rcfg.edgePixels = 0;
      registry.reconstructWidget(rcfg);
      return registry.getComponent(rcfg.id); 
   }
 
   /** 
    * Initializes this instance. 
    */
   private void init()
   {
      if (gd == null)
      {
         gd = (GuiDriver<?,?>) OutputManager.getDriver();
         
         realize();

         Container<GuiOutputManager> content = getContentPane();
         
         messagePanel = new MessagePanel();
         content.add(messagePanel);
         
         content.add(buttonsPanel);
         
         buttons = commons.createButtonSet();
         
         for (Button<GuiOutputManager> button : buttons)
         {
            button.config().font = FontManager.ALERT_BOX_FONT;
         }
         
         messageText = commons.buildMessageText();
         
         font = FontManager.resolveFont(window(), FontManager.ALERT_BOX_FONT, gd);

         String iconName = null;
         switch (alertType)
         {
            case LogicalTerminal.ALERT_ERROR: iconName = "error"; break;
            case LogicalTerminal.ALERT_INFO: iconName = "info"; break;
            case LogicalTerminal.ALERT_QUESTION: iconName = "question"; break;
            case LogicalTerminal.ALERT_WARNING: iconName = "warning"; break;
         }
         icon = ThemeManager.getCurrentTheme().getImage(iconName, 32, 32, true);
         
         GuiColorResolver gc = new GuiColorResolver(gd, 
                                                    new BaseConfig(), 
                                                    SysColor.COLOR_3DFACE, 
                                                    SysColor.COLOR_WINDOWTEXT);
         
         windowPane.setColorResolver(gc);
         contentPane.setColorResolver(gc);
         
         // the close window caption button is only disabled for yes/no
         boolean enabled = commons.buttonSet != LogicalTerminal.BTN_YES_NO;
         
         // the titlebar can be null when the web client is in embedded mode
         WindowTitleBar tbar = getTitleBar();
         if (tbar != null)
         {
            tbar.setCaptionButtonEnabled(CaptionButtonType.CLOSE, enabled);
            if (enabled)
            {
               tbar.setClickHandler(CaptionButtonType.CLOSE, () ->
               {
                  logical res = null;
                  switch (commons.buttonSet)
                  {
                     case LogicalTerminal.BTN_YES_NO_CANCEL:
                        res = new logical(); 
                        break;
                     case LogicalTerminal.BTN_OK:
                        res = new logical(true);
                        break;
                     case LogicalTerminal.BTN_OK_CANCEL:
                     case LogicalTerminal.BTN_RETRY_CANCEL:
                        res = new logical(false);
                        break;
                  }

                  commons.setResult(res);
                  commons.dismiss();
               });
            }
         }
      }
   }
   
   /**
    * The class implements the message-panel part of the alert box widget.
    */
   private class MessagePanel
   extends AbstractContainer<GuiOutputManager>
   {
      /** The expected width of the alert box message. */
      private int paragraphWidth;
      
      /**
       * Layout.
       */
      @Override
      public void doLayout()
      {
         Theme theme = ThemeManager.getCurrentTheme();
         
         NativeDimension iconDim = null;
         
         if (icon != null)
         {
            iconDim = new NativeDimension(icon.getWidth(), icon.getHeight());
         }
         
         CoordinatesConversion cc = screen().coordinates();
         
         // determine max line width
         int maxLineWidth = 0;
         String[] messageLines = messageText.split(System.getProperty("line.separator"));
         
         for (String messageLine : messageLines)
         {
            maxLineWidth = Math.max(getTextWidthNative(messageLine, FontManager.ALERT_BOX_FONT), 
                                    maxLineWidth);
         }
         
         // do not exceed the max message box width
         paragraphWidth = Math.min(maxLineWidth, MAX_MESSAGE_PIXEL_WIDTH);
         paragraphWidth = Math.max(paragraphWidth, MIN_MESSAGE_PIXEL_WIDTH);
         
         TopLevelWindow<?> defWindow = getOwner();
         gd.selectWindow(defWindow.getId().asInt());
         
         // determine message height
         int paragraphHeight = gd.getParagraphHeight(messageText, font, paragraphWidth);
         gd.releaseWindow();
         
         paragraphHeight = Math.max(paragraphHeight, MIN_MESSAGE_PIXEL_HEIGHT);
         
         gd.selectWindow(AlertBoxGuiImpl.this.getId().asInt());
         
         int messageAreaWidth = (int) (paragraphWidth + 
                                       2.5 * theme.getAlertBoxPadding() + 
                                       (iconDim == null ? 0 : iconDim.width));
         
         int messageAreaHeight = paragraphHeight + 2 * theme.getAlertBoxPadding();
         
         NativeDimension size = new NativeDimension(messageAreaWidth, messageAreaHeight); 
         
         setSize(cc.dimensionFromNative(size));
         
         gd.releaseWindow();
      }
      
      /**
       * Draw widget.
       */
      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeRectangle clipRect = clipRectangle();
         NativeDimension dim = physicalDimension();
         
         gd.draw(getId(), p, clipRect, () ->
               ThemeManager.getCurrentTheme().drawAlertBox(gd, 
                     0, 0, dim.width, dim.height, paragraphWidth, icon, messageText, font));
      }
   }
   
   /**
    * The class implements the buttons-panel part of the alert box widget.
    */
   private class ButtonsPanel
   extends AbstractContainer<GuiOutputManager>
   {
      /** Panel padding in pixels. */
      private static final int PADDING = 10;
      
      /** Button spacing in characters. */
      private static final double SPACING = 1.36;  
      
      /** Extra left border padding. */
      private static final int LEFT_BORDER_PADDING = 33; 
      
      /** Extra right border padding. */
      private static final int RIGHT_BORDER_PADDING = 8;
      
      /**
       * Layout.
       */
      @Override
      public void doLayout()
      {
         CoordinatesConversion cc = screen().coordinates();

         int buttonNativeWidth  = cc.widthToNative(BUTTONS_CHAR_WIDTH);
         int buttonNativeHeight = cc.heightToNative(BUTTONS_CHAR_HEIGHT);
         int buttonSpacing      = cc.widthToNative(SPACING);

         // it is expected that panel width has been set
         NativeDimension dim = physicalDimension();
         
         // windows message box keeps MIN_X left padding
         int minWidth = buttons.size() * buttonNativeWidth + 
                        (buttons.size() - 1) * buttonSpacing + 
                        PADDING + LEFT_BORDER_PADDING;
         
         dim.width  = Math.max(dim.width, minWidth);
         dim.height = buttonNativeHeight + 2*PADDING;
         
         setSize(cc.dimensionFromNative(dim));
         
         int buttonX = dim.width - RIGHT_BORDER_PADDING - buttonNativeWidth;
         int buttonY = dim.height / 2 - buttonNativeHeight / 2; 
         
         // iterate buttons in reverse order
         ListIterator<Button<GuiOutputManager>> it = buttons.listIterator(buttons.size());
         while (it.hasPrevious())
         {
            Button<GuiOutputManager> button = it.previous();
            button.config().autoResize = false;
            button.config().widthChars = BUTTONS_CHAR_WIDTH;
            button.config().heightChars = BUTTONS_CHAR_HEIGHT;
            button.setPhysicalLocation(buttonX, buttonY);
            
            buttonX -= buttonNativeWidth + buttonSpacing; 
         }
         
         // get default button
         Button<GuiOutputManager> defaultButton = buttons.get(commons.getDefaultButton());
         
         // service focus change call; is safe
         defaultButton.parent().setFocus(defaultButton);
         defaultButton.setHighlighted(true);
      }
      
      /**
       * Draw widget.
       */
      @Override
      public void draw()
      {
         NativePoint p = physicalLocation();
         NativeRectangle clipRect = clipRectangle();
         
         gd.draw(getId(), p, clipRect, ButtonsPanel.super::draw);
      }

      /**
       * Select next focusable widget.
       */
      @Override
      public void nextFocus()
      {
         commons.nextFocus();
      }

      /**
       * Select previous focusable widget.
       */
      @Override
      public void prevFocus()
      {
         commons.prevFocus();
      }
   }
}
