/*
** Module   : BrowseColumnGuiImpl.java
** Abstract : GUI implementation of browse column.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 SVL 20151013 Initial version.
** 002 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 003 SVL 20151019 Layout procedure refactoring, misc fixes.
** 004 SVL 20151110 Support for multi-line label. Added setWidth.
** 005 SVL 20151126 Fixed label drawing order for multi-line labels.
** 006 SVL 20151126 Use browse.isLabelsVisible.
** 007 SVL 20151206 Update column width if its WIDTH attribute has been updated.
** 008 SVL 20151230 Coloring fix, support for editable browse.
** 009 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 010 IAS 20160229 LOAD-MOUSE-POINTER support.
** 011 SVL 20160407 Support for column dragging.
** 012 SVL 20160412 Updated clipping rectangles. Implemented column resizing.
** 013 SVL 20160719 Updated translate() signature.
** 014 SVL 20160722 Reflect signature change in browse.rowClicked.
** 015 HC  20160716 Implemented runtime support for BROWSE-COLUMN:RESIZABLE attribute.
** 016 EVL 20160921 Fix for column header vertical text position mismatch.
** 017 SBI 20160921 Fixed casual NPE.
** 018 VIG 20160927 Added label update to afterConfigUpdate.
** 019 SVL 20170126 Added support for pressed columns, START-SEARCH/END-SEARCH events and
**                  SORT-ASCENDING/SORT-NUMBER attributes.
** 020 OM  20170320 Added support for Themes.
** 021 SVL 20170418 Minor fixes related to drawing of pressed labels and labels with sort numbers.
** 022 EVL 20170725 Adding handler for direct manipulation events.  The events are routed to main
**                  browse widget owning the current column.  Browse column can have custom mouse
**                  pointer like other hoverable widgets.
** 023 SVL 20170814 Changes related to mouse wheel, COLUMN-READ-ONLY and WIDTH-PIXELS.
** 024 SVL 20170831 Support for toggle-box columns.
** 025 SVL 20171116 Added coordination with MouseDirectManipulation.
** 026 CA  20171218 Fixes for BROWSE/BROWSE-COLUMN:ENTRY/LEAVE, BROWSE:ROW-LEAVE/ENTER,
**                  BROWSE:VALUE-CHANGED events.
**     HC  20171221 Fixed font resolution for browse column cells so that the column width is calculated
**                  correctly.
** 027 SVL 20180115 Fixed text offset and clipping inside cells and labels.
**     EVL 20180124 Adding method to gel pressed label state.
**     SVL 20180126 Edited row is committed on a column header press.
** 028 SVL 20180209 Added support for mouse right-click. Refactoring because of support for
**                  invisible columns.
** 029 SVL 20180321 Renamed columnIndex to visibleColumnIndex;
** 030 EVL 20180328 Fixed wrong column width for date formatted fields.
** 031 HC  20180420 Implemented popup menu resolution in the tree of widgets.
** 032 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 033 SBI 20180606 Changed to override getActualBounds() and isVisible() in order to update mouse
**                  and hover widget regions correctly.
** 034 SVL 20180529 Support for enhanced config.
** 035 SBI 20180727 Changed getActualBounds() to check if this column is in its container.
** 036 OM  20180809 Fixed drawing of browse with emphasis to Windows8  theme.
**         20180811 Fixed behaviour on mouse click events.
** 037 HC  20180813 Implemented initial version of dynamic layout.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.InvocationEvent;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.event.*;
import java.util.*;
import java.util.logging.*;

import static com.goldencode.p2j.ui.BrowseColumnWidgetType.*;
import static com.goldencode.p2j.ui.client.gui.BrowseGuiImpl.*;

/**
 * GUI implementation of browse column.
 */
public class BrowseColumnGuiImpl
extends BrowseColumn
implements Hoverable
{
   /** Offset of the label text if the column is in pressed state. */
   public final static int LABEL_PRESSED_OFFSET = 1;

   /** Thickness of the shadow inside the column header. */
   public final static int LABEL_SHADOW_THICKNESS = 1;

   /** Maximum offset between sort arrow and label text. */
   public final static int SORT_ARROW_MAX_LEFT_OFFSET = 10;

   /** Left clipping for sort area (sort arrow and number). */
   public final static int SORT_AREA_LEFT_CLIP = 1;

   /** Right clipping for sort area (sort arrow and number). */
   public final static int SORT_AREA_RIGHT_CLIP = 2;

   /** Gap between sort arrow and sort number. */
   public static final int GAP_BETWEEN_SORT_ARROW_AND_NUMBER = 1;

   /** Difference between data width and actual column width (in pixels). */
   private final static int PADDING = 4;

   /** Logger. */
   private final static Logger LOG = LogHelper.getLogger(BrowseColumnGuiImpl.class.getName());

   /** GUI driver. */
   private GuiDriver gd;

   /** Parent browse. */
   private BrowseGuiImpl browse;

   /** Label font. */
   private FontDetails labelFont;

   /** Font for sort number. */
   private FontDetails sortNumberFontDetails;

   /** Column font. */
   private FontDetails cellsFont;

   /** Coordinates conversion. */
   private CoordinatesConversion cc;

   /** Column width, in chars. */
   private double width;

   /** 0-based index of the column among the visible columns. */
   private int visibleColumnIndex = -1;
   
   /** If it is the markers (leftmost) column. */
   private final boolean markersColumn;

   /** The lines of which the label consist. */
   private String[] labelLines = new String[0];

   /** <code>true</code> if the column is being dragged. */
   private boolean isDragged = false;

   /** <code>true</code> if the column is being resized. */
   private boolean isResized = false;

   /**
    * <code>true</code> if user pressed down mouse button on a column label if
    * ALLOW-COLUMN-SEARCHING is <code>true</code>.
    */
   private boolean startedPress = false;

   /**
    * If column is pressed (column label is pressed if ALLOW-COLUMN-SEARCHING is
    * <code>true</code>).
    */
   private boolean isPressed = false;

   /**
    * Indicates the resize area over which the mouse pointer is hovering now.
    * See {@link ResizeArea} for more information. <code>null</code> if the
    * mouse pointer is not hovering over a resize area.
    */
   private ResizeArea currentResizeArea = null;
   
   /** If column label is pressed this becoming true when owning window has 3-d falg on. */
   private boolean labelPressed = false;

   /**
    * <code>true</code> if the column has implicit width. <code>false</code> if the column has
    * explicitly specified width.
    */
   private Boolean implicitWidth = null;

   /**
    * Constructor.
    *
    * @param markersColumn
    *        <code>true</code> if it is a markers (leftmost) column.
    */
   public BrowseColumnGuiImpl(boolean markersColumn)
   {
      super(WidgetId.nextID());
      this.markersColumn = markersColumn;
   }

   /**
    * Constructor.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Configuration data.
    */
   public BrowseColumnGuiImpl(WidgetId id, BrowseColumnConfig cfg)
   {
      this(id, cfg, false);
   }

   /**
    * Constructor.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Configuration data.
    * @param markersColumn
    *        <code>true</code> if it is a markers (leftmost) column.
    */
   public BrowseColumnGuiImpl(WidgetId id, BrowseColumnConfig cfg, boolean markersColumn)
   {
      super(id, cfg);
      this.markersColumn = markersColumn;
   }

   /**
    * Set column width and update WIDTH attribute accordingly.
    *
    * @param width
    *        Column width in characters.
    */
   public void setWidth(double width)
   {
      this.width = width;
      config.widthChars = width - cc.widthFromNative(getColumnPadding());
      config.widthPixels = cc.widthToNative(config.widthChars);
   }

   /**
    * Initialize the markers (leftmost) column.
    *
    * @param browse
    *        Parent browse.
    */
   public void initMarkersColumn(BrowseGuiImpl browse)
   {
      init(browse, 0, -1);
      OutputManager.instance().getRegistry().addWidget(this);
   }
   
   /**
    * Test whether this is a marker column. 
    * 
    * @return  {@code true} if this a marker column.
    */
   public boolean isMarkersColumn()
   {
      return markersColumn;
   }

   /**
    * Check if the column is visible via its parent view port.
    * 
    * @return   True if this column is via its parent view port, otherwise false.
    */
   public boolean isVisible()
   {
      if (config.visible)
      {
         NativeRectangle actualBounds = getActualBounds();
         
         if (actualBounds != null && !actualBounds.empty())
         {
            return true;
         }
      }
      
      return false;
   }

   /**
    * Get the column actual bounds.
    * 
    * @return   The column actual bounds. <code>null</code> if the column is outside of the browse
    *           viewport.
    */
   @Override
   public NativeRectangle getActualBounds()
   {
      BrowseGuiImpl browseGuiImpl = getBrowse();
      
      if (parent() == null || browseGuiImpl == null)
      {
         return null;
      }
      
      NativeRectangle viewPortRect = browseGuiImpl.getViewportBoundedRectangle();
      
      NativeRectangle actualBounds = viewPortRect.intersection(super.getActualBounds());
      
      return actualBounds;
   }

   /**
    * Initialize the widget.
    *
    * @param browse
    *        Parent browse.
    * @param leftOffset
    *        Column offset in pixels (i.e. width of preceding columns).
    * @param visibleColumnIndex
    *        0-based index of the column among the visible columns.
    */
   public void init(BrowseGuiImpl browse, int leftOffset, int visibleColumnIndex)
   {
      this.browse = browse;
      this.visibleColumnIndex = visibleColumnIndex;

      gd = (GuiDriver) OutputManager.getDriver();
      cc = screen().coordinates();

      setPhysicalLocation(leftOffset, 0);

      if (implicitWidth == null)
      {
         implicitWidth = config().widthChars <= 0;
      }
      setWidth(cc.widthFromNative(calcColumnWidth()));
   }
   
   /**
    * Obtain the {@link BrowseGuiImpl} this column is part of.
    * 
    * @return  the parent {@code BrowseGuiImpl} widget.
    */
   public BrowseGuiImpl getBrowse()
   {
      return browse;
   }

   /**
    * Calculate column width.
    *
    * @return column width.
    */
   public int calcColumnWidth()
   {
      return calcColumnWidth(false);
   }
   
   /**
    * Calculate column width.
    *
    * @param  ignoreSaved
    *         <code>true</code> to ignore the saved width for this column.
    *
    * @return column width.
    */
   public int calcColumnWidth(boolean ignoreSaved)
   {
      int width = calcDataWidth(ignoreSaved);
      width += getColumnPadding();
      return width;
   }

   /**
    * Get column padding, i.e. difference between data width and actual column width.
    *
    * @return column padding in native units.
    */
   public int getColumnPadding()
   {
      if (markersColumn)
      {
         return 0;
      }

      return PADDING + (isFirstColumn() ? BORDER_THICKNESS : 0);
   }

   /**
    * The method is called after the configuration associated with
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    *
    * @param    beforeUpdate
    *           Config reference capturing the config state before
    *           the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseColumnConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      if (browse != null)
      {
         resolveFonts();

         if (beforeUpdate.widthChars != config().widthChars && !browse.isPerformLayout())
         {
            setWidth(cc.widthFromNative(calcColumnWidth()));
            browse.notifyColumnsUpdated();
         }

         if (!Utils.equals(beforeUpdate.clabel, config.clabel))
         {
            updateLabel();
         }

         if (beforeUpdate.columnReadOnly != config().columnReadOnly && browse.isEditing())
         {
            browse.setEditorReadOnly(config().columnReadOnly);
         }

         if (config().readOnly      &&
             !beforeUpdate.readOnly && 
             browse.isEditing()     && 
             beforeUpdate.ordinal == browse.getCurrentColumn().ordinal)
         {
            // if read-only state for this column has changed, and the browse is in editable mode,
            // then stop it
            EventManager.postEvent(new InvocationEvent(() -> browse.finishStopEdit()));
         }

         if (beforeUpdate.hyperlinked != config.hyperlinked && browse.isVisible())
         {
            Map texts = browse.prepareHyperlinks();
            if (config.hyperlinked)
            {
               if (gd.needsExplicitTextMetricsCaching())
               {
                  gd.cacheTextMetrics(texts, null);
               }
               browse.layoutHyperlinks();
            }
         }
         
         browse.repaint();
      }
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      Container p = parent();
      return new Dimension(width, (p != null) ? p.height() : 0);
   }

   /**
    * Draw the widget.
    */
   @Override
   public void draw()
   {
      final NativePoint p = this.physicalLocation();
      final NativeDimension d = this.physicalDimension();
      final NativeDimension clipDimension = new NativeDimension(d.width, d.height);
      if (browse.config().separators && isLastLockedColumn() && !isLastColumn())
      {
         clipDimension.width += BORDER_THICKNESS; // for separator
      }
      final NativeRectangle clip = clipRectangle(p, clipDimension);

      final int labelsHeight = browse.getLabelsHeight();

      gd.draw(p, clip, () ->
      {
         Theme theme = ThemeManager.getCurrentTheme();
         
         int labelHeight = 0; // including border
         
         if (browse.isLabelsVisible())
         {
            labelHeight += labelsHeight + BORDER_THICKNESS * 2;
            
            theme.drawBrowseColumnLabel(this, gd, 0, 0, d.width, labelHeight);
         }
         
         if (markersColumn)
         {
            theme.drawBrowseMarkersColumn(this, gd, 0, labelHeight, d.width, d.height - labelHeight);
         }
         else 
         {
            theme.drawBrowseNormalColumn(this, gd, 0, labelHeight, d.width, d.height - labelHeight);
         }
      });
   }
   
   /**
    * Obtain the current {@code FontDetails} of this label.
    *
    * @return  the current {@code FontDetails} of the label.
    */
   public FontDetails getLabelFont()
   {
      return labelFont;
   }
   
   /**
    * Obtain the current {@code FontDetails} for this column.
    *
    * @return the current {@code FontDetails}.
    */
   public FontDetails getCellsFont()
   {
      return cellsFont;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      super.mouseClicked(e);

      if (e.getButton() == MouseEvent.BUTTON1)
      {
         NativePoint point = translate(e);
         browse.rowClicked(browse.getRowByCoordinates(point.y), visibleColumnIndex, e);
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      NativePoint point = translate(e);
      
      // let the parent BrowseGuiImpl know who generates the next focus events
      browse.setMouseActionColumn(this);
      
      if (currentResizeArea != null)
      {
         isResized = true;
      }
      else if (browse.isColumnMovable() && !markersColumn && browse.isLabelsVisible())
      {
         if (isHeaderArea(point))
         {
            isDragged = true;
         }
      }
      else if (!browse.config().columnMovable &&
               !markersColumn &&
               browse.isLabelsVisible() &&
               browse.config().allowColumnSearching)
      {
         if (isHeaderArea(point) && browse.beforeColumnHeaderPressed())
         {
            browse.resetPressedColumns(true);
            startedPress = true;
            isPressed = true;
            browse.repaint();
         }
      }

      if (isHeaderArea(point))
      {
         // if the top level window style is 3-d we need to keep track on label pressing
         TopLevelWindow tlWnd = topLevelWindow();
         if (tlWnd instanceof WindowGuiImpl && ((WindowGuiImpl)tlWnd).config().threeD)
         {
            labelPressed = true;
            repaint();
         }
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      NativePoint point = translate(e);
      
      if (isDragged)
      {
         isDragged = false;
         browse.stopColumnDrag(e);
      }

      if (isResized)
      {
         isResized = false;
         browse.stopColumnResize();
         gd.setCursor(MousePtrWrapper.DEFAULT);
         currentResizeArea = null;
      }

      boolean fireStartSearch = false;

      if (startedPress)
      {
         config.pressed = isHeaderArea(point);
         if (config.pressed)
         {
            fireStartSearch = true;
         }
         startedPress = false;
         isPressed = false;
      }

      // update resize cursor
      mouseMoved(e);

      if (fireStartSearch)
      {
         ThinClient.getInstance().invokeTriggers(browse, Keyboard.SE_START_SEARCH, false);
      }

      if (isHeaderArea(point) && labelPressed)
      {
         labelPressed = false;
      }
      
      // let my browse that focus events are not generated by mouse events on this column
      browse.setMouseActionColumn(null);
      browse.repaint();
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (isDragged)
      {
         if (!browse.isColumnDrag())
         {
            int leftDragColumnMouseOffset = translate(e).x;
            browse.startColumnDrag(this, leftDragColumnMouseOffset);
         }

         browse.mouseDragged(e);
      }
      else if (isResized)
      {
         if (!browse.isColumnResize() && currentResizeArea != null)
         {
            int index = 0;
            switch (currentResizeArea)
            {
               case LEFT:
                  index = visibleColumnIndex - 1;
                  break;
               case RIGHT:
                  index = visibleColumnIndex;
                  break;
               case LAST_LOCKED_RIGHT:
                  index = browse.config().numLockedColumns - 1;
                  break;
            }

            browse.startColumnResize(index);
         }

         browse.mouseDragged(e);
      }
      else if (startedPress)
      {
         NativePoint p = translate(e);
         isPressed = isHeaderArea(p);
         repaint();
      }
   }

   /**
    * Notification of a mouse moved event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // switch to resize cursor if required
      if ((browse.isColumnResizable() || config().resizable) && !markersColumn)
      {
         NativePoint p = translate(e);

         if (browse.getRowResizeArea(p) != null)
         {
            // row resize is handled by MouseDirectManipulation
            currentResizeArea = null;
            return;
         }

         if (getVisibleColumnIndex() >= browse.config().numLockedColumns &&
             browse.isLastLockedColumnResizeArea(e))
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.LAST_LOCKED_RIGHT;
         }
         else if (hasLeftResizeArea() && p.x >= 0 && p.x < COLUMN_RESIZE_AREA_WIDTH)
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.LEFT;
         }
         else if (hasRightResizeArea() &&
                  p.x >= physicalDimension().width - COLUMN_RESIZE_AREA_WIDTH &&
                  p.x < physicalDimension().width)
         {
            gd.setCursor(MousePtrWrapper.E_RESIZE);
            currentResizeArea = ResizeArea.RIGHT;
         }
         else
         {
            gd.setCursor(MousePtrWrapper.DEFAULT);
            currentResizeArea = null;
         }
      }
   }

   /**
    * Process direct manipulation event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processDirectManipulationEvent(DirectManipulationEvent event)
   {
      int id = event.id();
      
      // process events depending on ID code
      if (id == Keyboard.SE_SELECTION || id == Keyboard.SE_DESELECTION)
      {
         boolean newSelected = (id == Keyboard.SE_SELECTION); 
         BaseConfig brc = (BaseConfig)browse.config();
         // check if we need to sync the widget selection flag 
         if (brc.selected != newSelected)
         {
            // yes, need to reroute the event to main browse widget
            EventManager.postEvent(new DirectManipulationEvent(browse,
                                                               event.getParentContainer(),
                                                               event.isRepaintRequired(),
                                                               id));
         }
      }
   }
   
   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      browse.mouseWheelMoved(e);
   }

   /**
    * Get ancestor Window for this widget.
    *
    * @return  ancestor Window or <code>null</code> if the widget is not yet attached to a Window.
    */
   @Override
   public Window window()
   {
      return browse.window();
   }

   /**
    * Update the label (it should be parsed to the consisting lines).
    */
   void updateLabel()
   {
      labelLines = getLabelVar(config());
   }

   /**
    * Get the array of labels for this column. You may need to call {@link #updateLabel()} first
    * to have then initialized.
    *
    * @return the lines of which the label consist.
    */
   public String[] getLabelLines()
   {
      return labelLines;
   }

   /**
    * Initialize editor widget config.
    *
    * @param cfg
    *        Editor widget config.
    */
   void initEditorConfig(BaseConfig cfg)
   {
      if (TOGGLE_BOX.equals(config.widgetType))
      {
         ToggleBoxConfig toggleConfig = (ToggleBoxConfig) cfg;
         toggleConfig.x = browse.getCurrentEditorX();
         toggleConfig.y = browse.getCurrentEditorY();
         toggleConfig.fixedWidth = true;
         toggleConfig.fixedHeight = true;

         NativeDimension d = ThemeManager.getCurrentTheme().getRowToggleSize(this, true);
         toggleConfig.widthPixels = d.width;
         toggleConfig.heightPixels = d.height;
      }
      else
      {
         FillInConfig fillInConfig = (FillInConfig) cfg;
         fillInConfig.selectionStart = ControlConfig.INVALID_SELECTION;
         fillInConfig.selectionEnd = ControlConfig.INVALID_SELECTION;

         fillInConfig.x = browse.getCurrentEditorX();
         fillInConfig.y = browse.getCurrentEditorY();

         fillInConfig.fixedWidth = true;
         fillInConfig.fixedHeight = true;
         fillInConfig.widthPixels = physicalDimension().width;
         fillInConfig.heightPixels = browse.getRowHeight();

         fillInConfig.ehFont = getCellsFont();

         // Background color
         int bgColor = config().columnBgColor;
         if (bgColor < 0)
         {
            bgColor = browse.config().columnBgColor;
         }

         if (bgColor < 0)
         {
            bgColor = browse.config().bgcolor;
         }

         fillInConfig.bgcolor = bgColor;


         // Foreground color
         int fgColor = config().columnFgColor;
         if (fgColor < 0)
         {
            fgColor = browse.config().columnFgColor;
         }

         if (fgColor < 0)
         {
            fgColor = browse.config().fgcolor;
         }

         fillInConfig.fgcolor = fgColor;
      }
      //TODO combo-box
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Index of the browse column in the set of visible columns.
    *
    * @return  0-based index of the browse column in the set of visible columns (also except the
    *          markers column).
    */
   public int getVisibleColumnIndex()
   {
      return visibleColumnIndex;
   }

   /**
    * Sub-classes will override this to return a list of mouse actions on which the widget
    * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
    * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_MOVED,
         MouseEvent.MOUSE_WHEEL
      };
   }

   /**
    * Returns <code>true</code> if the given point is inside the column header.
    *
    * @param  point
    *         Point to check.
    *
    * @return See above.
    */
   private boolean isHeaderArea(NativePoint point)
   {
      int labelsHeight = browse.getLabelsAreaHeight();
      return point.x >= 0 &&
            point.x <= physicalDimension().width &&
            point.y >= 0 &&
            point.y < labelsHeight;
   }

   /**
    * Get width of the label text, in pixels.
    *
    * @return see above.
    */
   public int getColumnLabelTextWidth()
   {
      int res = 0;
      for (String labelLine : labelLines)
      {
         res = Math.max(0, getTextWidth(labelLine, labelFont));
      }
      return res;
   }

   /**
    * Get width of the sort area (sort arrow and number), in pixels.
    *
    * @return see above.
    */
   public int getSortAreaWidth()
   {
      if (config.sortAscending == null)
      {
         return 0;
      }

      int res = SORT_ARROW_WIDTH;

      if (config.sortNumber != null && config.sortNumber > 0)
      {
         res += GAP_BETWEEN_SORT_ARROW_AND_NUMBER;
         res += SORT_NUMBER_WIDTH;
      }

      return res;
   }

   /**
    * Returns <code>true</code> if the column is visually in pressed state.
    *
    * @return see above.
    */
   public boolean isDrawnAsPressed()
   {
      if (markersColumn)
      {
         return false;
      }

      return startedPress ? isPressed : isPressed || config.pressed;
   }

   /**
    * Returns <code>true</code> if the column label must visually be in a pressed state.
    *
    * @return see above.
    */
   public boolean isLabelPressed()
   {
      return labelPressed;
   }

   /**
    * Get clipping point for a text inside a label.
    *
    * @return see above.
    */
   public NativePoint getLabelTextClipPoint()
   {
      int x = ThemeManager.getCurrentTheme().getBrowseLabelOffsetLeft(Browse.isRightAligned(this))
                + (isFirstColumn() ? BrowseGuiImpl.BORDER_THICKNESS : 0);
      
      // need to add one pixel more to take into account frame around header
      int y = CELL_OFFSET_TOP + BORDER_THICKNESS + 1;
      
      if (isDrawnAsPressed())
      {
         // there is no reason to shift x, not even for ClassicTheme
         y += LABEL_PRESSED_OFFSET;
      }
      
      return new NativePoint(x, y);
   }

   /**
    * Get clipping rectangle for a text inside a label.
    *
    * @param  clipPoint
    *         Clipping point.
    *
    * @return see above.
    */
   public NativeRectangle getLabelTextClipRectangle(NativePoint clipPoint)
   {
      int clipHeight = browse.getLabelsHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      int labelAreaWidth = getLabelAreaWidth();
      int sortAreaWidth = getSortAreaWidth();

      int clipWidth;
      if (sortAreaWidth == 0)
      {
         clipWidth = labelAreaWidth -
                       clipPoint.x -
                       ThemeManager.getCurrentTheme().getBrowseLabelOffsetRight() -
                       (isLastColumn() ? 0 : BORDER_THICKNESS);
         // when drawing as pressed the text is shifted, but width is the same
      }
      else
      {
         clipWidth = labelAreaWidth - clipPoint.x -
                        SORT_AREA_LEFT_CLIP - sortAreaWidth  - SORT_AREA_RIGHT_CLIP -
                        (isLastColumn() ? 0 : BORDER_THICKNESS);

         int labelTextWidth = getColumnLabelTextWidth();
         if (clipWidth > labelTextWidth)
         {
            clipWidth = Math.max(labelTextWidth,
                                 clipWidth - (SORT_ARROW_MAX_LEFT_OFFSET - SORT_AREA_LEFT_CLIP));
         }
      }

      return clipRectangle(clipPoint, new NativeDimension(clipWidth, clipHeight));
   }

   /**
    * Get width of the column header, starting from the left column header border (inclusive) and
    * ending with the right column header border (exclusive).
    *
    * @return see above.
    */
   public int getLabelAreaWidth()
   {
      NativeDimension d = physicalDimension();
      return d.width - (isLastColumn() ? BORDER_THICKNESS : 0);
   }

   /**
    * Get top offset for positioning text inside a label.
    *
    * @return see above.
    */
   public int getLabelTextTopOffset()
   {
      return labelFont.legacyHeight + CELL_OFFSET_BOTTOM -
                FontManager.getBrowseCellBottomTextOffset(labelFont);
   }

   /**
    * Get top offset for positioning text inside a cell.
    *
    * @return see above.
    */
   public int getCellTextTopOffset()
   {
      return cellsFont.legacyHeight
            + CELL_OFFSET_BOTTOM - FontManager.getBrowseCellBottomTextOffset(cellsFont);
   }

   /**
    * Get clipping point for a text inside a cell relatively to the column top.
    *
    * @param viewportIndex
    *        0-based viewport index of the row the cell belongs to.
    *
    * @return see above.
    */
   public NativePoint getCellTextClipPoint(int viewportIndex)
   {
      return new NativePoint(
            ThemeManager.getCurrentTheme().getBrowseDataOffsetLeft(Browse.isRightAligned(this)) +
                                (isFirstColumn() ? BrowseGuiImpl.BORDER_THICKNESS : 0),
            browse.getTopRowOffset(viewportIndex) + CELL_OFFSET_TOP + 1);
   }

   /**
    * Get clipping rectangle for a text inside a cell.
    *
    * @param  point
    *         Clipping point.
    *
    * @return see above.
    */
   public NativeRectangle getCellTextClipRectangle(NativePoint point)
   {
      final NativeDimension d = physicalDimension();
      int bottom = browse.getRowHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      int right = d.width - DATA_OFFSET_RIGHT - point.x + 1;

      return clipRectangle(point, new NativeDimension(right - 1, bottom - 1));
   }

   /**
    * If the column has an implicit width, recalculate it according to the new column font.
    */
   void updateImplicitColumnWidth()
   {
      if (implicitWidth != null && implicitWidth)
      {
         setWidth(cc.widthFromNative(calcColumnWidth(true)));
      }
   }

  /**
    * Calculate width of the data part of the column.
    *
    * @param  ignoreSaved
    *         <code>true</code> to ignore the saved width for this column.
    *
    * @return width of the data part of the column.
    */
   int calcDataWidth(boolean ignoreSaved)
   {
      resolveFonts();

      Window window = window();

      if (markersColumn)
      {
         int fontWidth = FontManager.getFontWidth(window, browse.getBrowseFont(), gd);
         return (int) (fontWidth * 1.6) + 2 * BORDER_THICKNESS; // TODO: is this formula correct?
      }

      if (!ignoreSaved && config().widthChars > 0)
      {
         return cc.widthToNative(config().widthChars);
      }

      int dataWidth = 0;
      String dataType = config().dataType;
      String format = config().format;
      if (format == null)
      {
         format = DisplayFormat.instanceOfType(dataType).defaultFormatString();
      }

      boolean characterFormat = "character".equals(dataType);
      boolean logicalFormat = "logical".equals(dataType);

      if (characterFormat || logicalFormat)
      {
         try
         {
            DisplayFormat formatInstance;
            if (characterFormat)
            {
               formatInstance = new StringFormat(format);
            }
            else
            {
               formatInstance = new BoolFormat(format);
            }

            dataWidth += FontManager.getFontWidth(window, cellsFont, gd) *
                           formatInstance.getScreenWidth();
         }
         catch (DisplayFormatParsingException e)
         {
            LOG.severe("Cannot parse format for browse column: " + format);
            return 0;
         }
      }
      else
      {
         int len = format.length();
         for (int i = 0; i < len; i++)
         {
            char c = format.charAt(i);
            switch (c)
            {
               case '9':
               case '>':
               case 'z':
               case 'Z':
               case '*':
                  // width of an arbitrary digit (they all have the same widths)
                  dataWidth += getTextWidth("0", cellsFont);
                  break;
               default:
                  // other symbols like ', . - + /' widths are added as is
                  dataWidth += getTextWidth(String.valueOf(c), cellsFont);
            }
         }
      }

      updateLabel();
      int labelWidth = 0;
      for (String labelLine : labelLines)
      {
         labelWidth = Math.max(labelWidth, getTextWidth(labelLine, labelFont));
      }
      
      return Math.max(dataWidth, labelWidth);
   }

   /**
    * Resolve browse column fonts.
    */
   void resolveFonts()
   {
      if (browse == null)
      {
         return;
      }

      if (browse.config().enhancedBrowse)
      {
         int windowId = window().getId().asInt();
         config.ehFont = FontManager.registerFontDetails(windowId, config.ehFont);
         config.ehLabelFont = FontManager.registerFontDetails(windowId, config.ehLabelFont);
      }

      Window window = window();

      // base font used for calculation of browse elements sizes
      FontDetails browseFont = browse.getBrowseFont();

      if (markersColumn)
      {
         return;
      }

      // label font
      if (config().ehLabelFont != null)
      {
         labelFont = config().ehLabelFont;
      }
      else
      {
         int font = config().labelFont;
         if (font != FontManager.UNDEFINED_FONT)
         {
            labelFont = FontManager.resolveFont(window, font, gd);
         }
         else
         {
            labelFont = browse.getBrowseLabelFont();
         }
      }

      // cells font
      if (config().ehFont != null)
      {
         cellsFont = config().ehFont;
      }
      else
      {
         int font = config().columnFont;
         if (font != FontManager.UNDEFINED_FONT)
         {
            cellsFont = FontManager.resolveFont(window, font, gd);
         }
         else
         {
            cellsFont = browseFont;
         }
      }

      sortNumberFontDetails = FontManager.resolveFont(window(), FontManager.SORT_NUMBER_FONT, gd);
   }

   /**
    * Returns <code>true</code> if it is the first column (not including markers column).
    *
    * @return see above.
    */
   public boolean isFirstColumn()
   {
      return visibleColumnIndex == 0;
   }

   /**
    * Returns <code>true</code> if it is the last column (not including markers column).
    *
    * @return see above.
    */
   public boolean isLastColumn()
   {
      return !markersColumn && (visibleColumnIndex == browse.getVisibleColumnCount() - 1);
   }

   /**
    * Determines if this column has left resize area (see {@link ResizeArea} for more
    * information).
    *
    * @return <code>true</code> if this column has left resize area.
    */
   private boolean hasLeftResizeArea()
   {
      if (isFirstColumn())
      {
         return false;
      }
      else if (visibleColumnIndex < browse.config().numLockedColumns)
      {
         // non-first locked column
         return true;
      }
      else
      {
         // scrolling set
         return physicalLocation().x + parent().physicalLocation().x >
                  browse.getLockedColumnsWidth();
      }
   }

   /**
    * Determines if this column has right resize area (see {@link ResizeArea} for more
    * information).
    *
    * @return <code>true</code> if this column has right resize area.
    */
   private boolean hasRightResizeArea()
   {
      return physicalLocation().x + parent().physicalLocation().x + physicalDimension().width <
               browse.getViewportDimension().width;
   }

   /**
    * Returns <code>true</code> if it is the last column if the set of locked columns (not
    * including markers column).
    *
    * @return see above.
    */
   public boolean isLastLockedColumn()
   {
      return !markersColumn && (visibleColumnIndex == browse.config().numLockedColumns - 1);
   }

   /**
    * Determines if the mouse event happened over the toggle box image area in the current row in
    * this column.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if the mouse event happened over the toggle box image area in the
    *         current row in this column.
    */
   boolean isInToggleBoxArea(MouseEvent e)
   {
      NativePoint point = translate(e);
      NativeDimension toggleSize = ThemeManager.getCurrentTheme().getRowToggleSize(this, false);
      NativePoint toggleOffset = ThemeManager.getCurrentTheme().getRowToggleOffset(this, false);
      int toggleY = browse.getTopRowOffset(browse.getViewportRow()) + toggleOffset.y;

      return point.x >= toggleOffset.x && point.x < toggleOffset.x + toggleSize.width &&
             point.y >= toggleY && point.y < toggleY + toggleSize.height;
   }

   /**
    * The method shows the popup menu for this widget. The menu is resolved with the help of
    * {@link #resolvePopupMenuId()}.
    *
    * @param   e
    *          The related mouse event that caused a popup to be shown.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    *
    * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
    */
   @Override
   protected Menu showPopupMenu(MouseEvent e, Runnable dismissHandler)
   {
      return browse.showPopupMenu(this, e, dismissHandler);
   }

   /**
    * Enumeration for different resize areas. Resize area is an area over which the mouse pointer
    * changes to resize pointer and user can resize column or row by dragging this area. Vertical
    * resize areas have width of several pixels and height equal to viewport height. Horizontal
    * resize areas have width equal to viewport width and height of several pixels.
    */
   public enum ResizeArea
   {
      /** Area on the left edge of a column - resizes previous column. */
      LEFT,

      /** Area on the right edge of a column - resizes current column. */
      RIGHT,

      /** Area after the right edge of the last locked column - resizes last locked column. */
      LAST_LOCKED_RIGHT,

      /** Area on the upper edge of a row - resizes the row above. */
      UPPER,

      /** Area on the lower edge of a row - resizes the current row. */
      LOWER
   }
}
