/* 
** Module   : FillInGuiImpl.java
** Abstract : text field editor/display implementation, GUI-specific version
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Fixed calculations of widget dimensions in GUI.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed fill-in height: is computed using the same formula as pixels-per-row, 
**                  but using the explicit font, if set.
** 005 CA  20150507 RETURN doesn't clear selection.
**                  Fixed text drawing - centered vertically.
** 006 CA  20150510 Fixed selection/border color and added a NPE protection.
** 007 GES 20150425 Eliminated AWT dependencies.
** 008 CA  20150617 Use Keyboard.isPrintableKey to determine if a key is printable or not.
** 009 EVL 20150601 Adding protected method to draw the border.  Other subclasses can have special
**                  draw approach for the border.  Separating drawSelection() and making it
**                  protected.  Adding color/logic to draw the fill-in in disabled state. 
** 010 CA  20150723 Fixed widget content clearing - it was adding 1 pixel to its width/height.
** 011 CA  20150728 Fixed refresh() and drawBorder().
** 012 EVL 20150723 Moving disabled color setup to the subclass level.
** 013 CA  20150730 The caret can be drawn outside of window's draw call - so ensure the widget's
**                  window is selected before drawing it.
** 014 CA  20150805 Fixed FILL-IN cursor problems when editable width is smaller than content
**                  length.
** 015 GES 20150710 Added getSelectedText() implementation.
** 016 CA  20150907 Fixed popup-menu support for fill-in.
** 017 CA  20150826 Added mouse actions. Modified editable portion dimensions calcuations as used
**                  by text drawing and caret drawing (to ignore the full fixed width that can be
**                  explicit set).
** 018 IAS 20150924 AUTO-RESIZE, TOOLTIP and MENU-MOUSE attributes support
** 019 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 020 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 021 SBI 20151014 Added to notify the client the current selection is changed.
**                  Improvements in content placement: horizontally is OK, but vertically there 
**                  are some differences (maybe is related to font ascent/descent?).
**                  Right-aligned formats are "pushed" to the right (by simulating a space prefix)
**                  - this should be removed when the NumberFormat is fixed to handle left-side
**                  spaces (see the TODOs in editableNativeDimension and editableNativeLocation).
**                  Fixed auto-resize for FILL-IN widgets.
**     EVL 20151021 Fix for selection replacement. This covers resolution for tiny input issue
**                  when selected text replacing incorrectly.
**     HC  20151112 Implemented THREE-D runtime support.
** 022 HC  20151124 Utilized Keep3DFillinBorder environment variable when drawing FILL-IN border.
** 023 EVL 20151204 Fixed the usage of the disabled color.  This is important to properly draw
**                  EntryFieldGuiImpl subclass.
** 024 SVL 20151230 Repaint on focus gaining.
** 025 IAS 20151230 LOAD-MOUSE-POINTER support.
** 026 IAS 20160331 Fixed potential NPE
** 027 CA  20160216 Cache the intermediate widget states.
** 028 SVL 20160314 Fixed scrolling to the editable row which is outside of the view.
** 029 VIG 20160128 Added isRMBEntrySupported() method, added system popup support.
**                  Changed showPopupMenu() call: added parameter. Fixed onFocusLost listener.
** 030 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 031 SBI 20160614 Changed to notify the client the current selection is invalid.
** 032 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 033 EVL 20160816 Fix for missing initial selection drawing.  The ENTRY artificial key does not
**                  invalidate current selection.
** 034 EVL 20161103 Fix for text position calculation.  The content of the fill-in must be
**                  horizontally scrollable when the text exceeds the fill-in width.
** 035 HC  20170103 BLANK must show empty spaces instead of question marks in GUI. Prevented an
**                  unhandled StringIndexOutOfBoundsException.
**     GES 20170104 Added support for PASSWORD-FIELD to display text as asterisks.
** 036 CA  20161101 Added support for mnemonic processing.
**     SVL 20170206 If foreground color is set, use it as the foreground color in disabled state
**                  too.
** 037 OM  20170320 Added support for Themes.
** 038 OM  20170406 Content of password cannot be copied. Method signature change.
** 039 OM  20170511 Added support for EDIT-CLEAR.
** 040 EVL 20170824 Adding file list drag-and-drop support.
** 041 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
** 042 SVL 20171026 Fixed mouse wheel for in-browse editor.
** 043 OM  20171027 Avoided NPE in initialize().
** 044 CA  20171111 Added WIDGET:BGCOLOR-RGB and WIDGET:FGCOLOR-RGB FWD extension attributes
** 045 CA  20171127 Avoid gd.setCurrentSelection(null) if the selection was invalid already.  
**                  This call is expensive on the web driver.
** 046 SBI 20180108 Fixed the case when the cursor is at the "before text" position.
**         20180207 Fixed the mouse text selection by reusing virtual right and left keys.
** 047 SBI 20180209 Added double click word selection.
** 048 OM  20171122 moveCursorToLast() is protected.
** 049 SBI 20180403 Fixed StringIndexOutOfBoundsException exception.
** 050 HC  20180419 Fixed StringIndexOutOfBoundsException exception. Fixed an NPE when showing
**                  popup menu. Implemented popup menu resolution in the tree of widgets.
** 051 HC  20180503 Fixed FILL-IN stops responding to mouse clicks after its popup menu
**                  displayed. Improved MENU focus management.
** 052 CA  20180502 Fixed COLOR option, COLOR statement and other color-related issues.
**     CA  20180530 Fixed DISABLE-AUTO-ZAP - cursor must not be set to 0 on LEAVE and content 
**                  must not be selected on ENTRY.
**     HC  20180530 Fixed the widget stops responding to mouse clicks after its popup menu
**                  displayed. This is second attempt to resolve this issue.
** 053 CA  20180615 Do not draw the caret if we are still in invalidation mode (the root drawing
**                  bracket is not finished yet).
** 054 SVL 20180529 Added support for enhanced font.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.gui.theme.Theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * Implementation of Progress FILL-IN component for GUI.
 * 
 * TODO: selection via mouse drag
 */
public class FillInGuiImpl
extends FillIn<GuiOutputManager,FillInConfig>
implements AutoResizable<FillInConfig>,
           StreamableWidget,
           LightweightTextWidget
{
   /** Screen driver */
   private GuiDriver gd;
   
   /** Font resolver */
   private GuiFontResolver gf;

   /**
    * Flag indicating initialization is not needed on a focus-gained event, as it was treated
    * explicitly by some other logic (like mouse clicked).
    */
   private boolean bypassFocusInit = false;

   /** Disabled state text color. */
   protected ColorRgb disabledFore;

   /**
    * <code>true</code> if the foreground color of a disabled field is fixed (grey).
    * <code>false</code> if the foreground color of a disabled field is defined by the FGCOLOR
    * attribute.
    */
   protected boolean disabledForeFixedColor = false;
   
   /** The absolute x-coordinate of the last mouse pressed event */
   private int lastPressedX;
   
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FillInGuiImpl(WidgetId id, FillInConfig cfg)
   {
      super(id, cfg);
      
      this.nativeInsets = new NativeInsets(2, 2, 2, 2);
   }
   
   /**
    * Convert this widget instance to its streamable equivalent.
    * 
    * @return   This widget's streamable equivalent.
    */
   @SuppressWarnings("unchecked")
   @Override
   public Widget<?> convertToStreamable()
   {
      return OutputManager.instance().getFactory().create(FillIn.class, getId(), config());
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Compute the widget's width in native units.
    * 
    * @param    widget
    *           The FILL-IN or TEXT widget.
    * @param    gd
    *           The GUI driver in use.
    * @param    gf
    *           The font resolver.
    * @param    isText
    *           <code>true</code> if this is for a TEXT widget.
    *           
    * @return   The widget's width in native units.
    */
   protected static int nativeWidth(FillIn<GuiOutputManager,?> widget, 
                                    GuiDriver                  gd,
                                    GuiFontResolver            gf,
                                    boolean                    isText)
   {
      Window<?> window = WindowManager.resolveWindow(widget);
      FillInConfig config = widget.config();
   
      if (gf == null || config == null || config.appDataPres == null)
         return 0;
   
      boolean explicit = (config.font != FontManager.UNDEFINED_FONT); 
      boolean fixedFontFormat = config.appDataPres.isFixedFont();
      FontDetails font = gf.font().font;
      int fmtWidth = widget.isAtField() ? config.atFormatLength : config.appFormat.getScreenWidth();
      int avgFontWidth = FontManager.getFontWidth(window, font, gd);
      boolean fixedFont = false;
      if (!fixedFontFormat && !explicit)
      {
         fixedFont = font.fixedSize;
      }
      int maxFontWidth = FontManager.getMaxFontWidth(window, font, gd);

      // In GUI, the width has a deviation from the normal "multiply format width with the font's 
      // average character width" formula.  After gathering fill-in metrics for various fonts and 
      // format lengths, it was determined that a special formula is used when all these are true:
      // - the format length is between 1 and 12 characters
      // - the font is not set explicitly via the FONT phrase 
      // - the widget normally expects a fixed font (like the case for a fill-in with a numeric 
      //   value)
      // A java program used to prove this formula can be found in the testcases project, 
      // the uast/fillin/CheckMetricsFillin.java file.
      
      int pixelWidth = !fixedFontFormat && !explicit && fmtWidth >= 1 && fmtWidth <= 12
                          // handle font sizes 1 to 12 
                          ? Math.min(fmtWidth, 3) * 
                           (fixedFont && !explicit ? maxFontWidth 
                                                    : (avgFontWidth * 2 + 
                                                       (int) Math.floor(avgFontWidth / 2))) +  
                           Math.max(fmtWidth - 3, 0) * avgFontWidth
   
                           // default case
                           : fmtWidth * avgFontWidth;
      
      return pixelWidth + (isText ? 0 : 8 + (FontManager.isThreeD() ? 2 : 0));
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, FillInConfig cfg)
   {
      super.initialize(id, cfg);
      
      gd = (GuiDriver) OutputManager.getDriver();
      Window<?> window = WindowManager.resolveWindow(this);
      boolean fixedFont = config.appDataPres != null && config.appDataPres.isFixedFont();
      gf = new GuiFontResolver(window, gd, config, fixedFont, config.ehFont);
      // disabled text color, use fg color for regular fill-ins
      disabledFore  = gd.getSysColor(SysColor.COLOR_FILLIN_FG);
      updateSize();
      
      addFocusListener(new FocusListener()
      {
         @Override
         public void onFocusLost(FocusEvent event)
         {
            @SuppressWarnings("unchecked")
            Widget<GuiOutputManager> currentFocus = 
               (Widget<GuiOutputManager>) UiUtils.getCurrentFocus();
            if (currentFocus != null)
            {
               // owned editor pop-up menu
               EditorPopupGuiImpl editorPopup = currentFocus.parent(EditorPopupGuiImpl.class);
               if (editorPopup == null || 
                   editorPopup.config().ownerId != FillInGuiImpl.this.getId().asInt())
               {
                  invalidateSelection();
                  
                  resetEditScroller();
                  
                  repaint();
               }
            }
         }
         
         @Override
         public void onFocusGained(FocusEvent event)
         {
            if (bypassFocusInit || config.disableAutoZap)
            {
               bypassFocusInit = false;
               return;
            }

            int lastOffset = getLastOffset();
            
            if (getText(true).length() > 0)
            {
               // select entire text
               config.selectionStart = 0;
               config.selectionEnd = lastOffset + 1;
               moveCursorToLast();
            }
            else
            {
               invalidateSelection();
               setCursorOffset(0);
            }

            repaint();
         }
      });
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (config.fixedWidth)
      {
         return config.widthPixels;
      }
      else if (config.fontChanged)
      {
         int fmtLength = getCurrentFormat().getScreenWidth();
         config.fontChanged = false;
         
         String txt = isUninitialized() ? "" : getText(false);
         txt = StringHelper.leftAlignText(txt, fmtLength);
         
         int txtWidth = getTextWidthNative(txt, gf.font().font);
         
         txtWidth += 8 + (FontManager.isThreeD() ? 2 : 0);
         
         // if the font has changed, mark the width as fixed (until the font changes again)
         
         config.fixedWidth = true;
         config.fontChanged = false;
         return txtWidth;
      }
      
      return nativeWidth(this, gd, gf, false);
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      if (config == null)
      {
         return 0;
      }

      if (config.fixedHeight)
      {
         return config.heightPixels;
      }

      if (gf == null || config.appDataPres == null)
         return 0;

      Window<?> window = WindowManager.resolveWindow(this);

      int legacyHeight;
      boolean threeD = FontManager.isThreeD();
      ScreenDriver sd = screen().getInstanceDriver();
      if (config().ehFont == null)
      {
         int font = gf.resolveFontNum();
         if (font == FontManager.DEFAULT_FIXED_FONT)
         {
            // fixed-font is used only for width
            font = FontManager.DEFAULT_FONT;
         }
         legacyHeight = FontManager.getFontHeight(window, font, sd);
      }
      else
      {
         legacyHeight = FontManager.getFontHeight(window, gf.resolveFontNum(), sd);
      }

      legacyHeight = legacyHeight + (int) (legacyHeight / 2) + (threeD ? 2 : 0);

      return legacyHeight;
   }

   /**
    * Returns the last valid cursor position from the left bound of the FillIn.
    * 
    * @return  The last valid cursor offset.
    */
   @Override
   public int getLastOffset()
   {
      return Math.max(getText(true).length() - 1, 0);
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return processMnemonicImpl(key, this);
   }
   
   /**
    * Creates new {@link MnemonicInfoGuiImpl} instance.
    *
    * @return   {@link MnemonicInfoGuiImpl} instance.
    */
   @Override
   public MnemonicInfo createMnemonic()
   {
      return new MnemonicInfoGuiImpl(false);
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      lastPressedX = e.getX();
      
      if (!hasFocus())
      {
         // focus management is done at the TC.processProgressEvent
         bypassFocusInit = true;
         invalidateSelection();
      }

      // popup menu is only shown on click
      if (e.getButton() != getMenuMouse())
      {
         invalidateSelection();
         mouseClicked(e);
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      super.mouseClicked(e);

      // when popup-menu is opened due to this event, do not proceed further
      if (e.isConsumed())
      {
         repaint();
         return;
      }
      
      if (!isEnabled())
      {
         return;
      }
      
      String content = getText();
      
      int oldOffset = getCursorOffset();
      setCursorOffset(content, e);
      
      // if this is a double-click, select the current word and its right-side spaces
      if (e.getClickCount() == 2 && !content.isEmpty())
      {
         int cursorOffset = getCursorOffset();
         
         if (cursorOffset > 0)
         {
            cursorOffset -= 1;
         }
         else
         {
            cursorOffset = 0;
         }
         
         int[] limits = StringHelper.getWordWithRightSpacesAt(content, cursorOffset);
         setSelection(limits[0], limits[1]);
         // notify about the current selection
         gd.setCurrentSelection(getSelectedText());
      }
      
      if (getCursorOffset() != oldOffset)
      {
         repaint();
      }
   }

   /**
    * The method shows the popup menu for this widget. The menu is resolved with the help of
    * {@link #resolvePopupMenuId()}.
    *
    * @param   e
    *          The related mouse event that caused a popup to be shown.
    * @param   dismissHandler
    *          When not null, the runnable instance is executed when the popup menu is dismissed, regardless
    *          whether an item is chosen or not.
    *
    * @return  Returns the popup menu shown or {@code null} if no popup is resolved for this widget.
    */
   protected Menu<GuiOutputManager> showPopupMenu(MouseEvent e, Runnable dismissHandler)
   {
      Menu<GuiOutputManager> menu = super.showPopupMenu(e, dismissHandler);
      if (menu != null || !isEnabled())
      {
         return menu;
      }

      String selection = getSelectedText();
      boolean selectionEmpty = selection == null || selection.isEmpty();
      menu = EditorPopupGuiImpl.create(this, selectionEmpty);
      menu.showPopup(this, new NativePoint(e.getX(), e.getY()), dismissHandler);

      return menu;
   }

   /**
    * Handle the mouse dragged events in order to support fillin functionality.
    * <p>
    * If the fillin is enabled and has focus, it will re-adjust the selected text depending on
    * where the mouse is moving. 
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (!isEnabled() || !hasFocus())
      {
         return;
      }

      boolean fromLeftToRight = e.getX() >= lastPressedX;

      int virtualKey;
      
      if (fromLeftToRight)
      {
         virtualKey = Key.VK_SHIFT_CURSOR_RIGHT;
      }
      else
      {
         virtualKey = Key.VK_SHIFT_CURSOR_LEFT;
      }
      
      KeyInput keyInput = new KeyInput(this, EventType.KEY_PRESSED, virtualKey);
      processKeyEvent(keyInput);
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (browse != null)
      {
         browse.mouseWheelMoved(e);
      }
   }

   /**
    * Process GUI specific keys:
    * <ul>
    *    <li><code>VK_SHIFT_CURSOR_RIGHT</code> - expands text selection to the right.</li>
    *    <li><code>VK_SHIFT_CURSOR_LEFT</code> - expands text selection to the left.</li>
    *    <li><code>VK_SHIFT_HOME</code> - expands text selection to the first char.</li>
    *    <li><code>VK_SHIFT_END</code> - expands text selection to the last char.</li>
    *    <li><code>VK_CTRL_A</code> - selects the entire content.</li>
    *    <li><code>KA_COPY</code> - copies the selected text to the OS clipboard.</li>
    *    <li><code>KA_CUT</code> - copies and deletes the selected text.</li>
    *    <li><code>KA_PASTE</code> - inserts the OS cliboard content into the widget.</li>
    * </ul>
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      int key = ke.actionCode();
      boolean invalidateSelection = true;
      int lastOffset = getLastOffset();
      int offset = getCursorOffset();

      if (browse != null)
      {
         browse.notifyKeyPressedInEditor(ke);
      }

      if (key == Key.VK_SHIFT_CURSOR_RIGHT)
      {
         // expand selection
         if (config.selectionStart == FillInConfig.INVALID_SELECTION)
         {
            config.selectionStart = offset;
            config.selectionEnd = Math.min(offset + 1, lastOffset + 1);
         }
         else if (offset == config.selectionStart)
         {
            config.selectionStart = Math.min(config.selectionStart + 1, lastOffset);
            
            if (config.selectionStart > config.selectionEnd)
            {
               int aux = config.selectionStart;
               config.selectionStart = config.selectionEnd;
               config.selectionEnd = aux;
            }
         }
         else if (config.selectionEnd != offset + 1)
         {
            config.selectionEnd = Math.min(offset + 1, lastOffset + 1);
         }
         else
         {
            config.selectionEnd = Math.min(config.selectionEnd + 1, lastOffset + 1);
         }
         // notify about the current selection
         gd.setCurrentSelection(getSelectedText());
         
         ke.setKeyCode(Keyboard.KA_CURSOR_RIGHT);
         ke.setActionCode(Keyboard.KA_CURSOR_RIGHT);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_CURSOR_LEFT)
      {
         // expand selection
         if (config.selectionEnd == FillInConfig.INVALID_SELECTION)
         {
            config.selectionEnd = Math.min(offset, lastOffset + 1);
            config.selectionStart = Math.max(offset - 1, 0);
         }
         else if (offset == config.selectionEnd)
         {
            config.selectionEnd = Math.max(config.selectionEnd - 1, 0);
            
            if (config.selectionStart > config.selectionEnd)
            {
               int aux = config.selectionStart;
               config.selectionStart = config.selectionEnd;
               config.selectionEnd = aux;
            }
         }
         else if (config.selectionStart != offset)
         {
            config.selectionStart = Math.max(offset - 1, 0);
         }
         else
         {
            config.selectionStart = Math.max(config.selectionStart - 1, 0);
         }
         // notify about the current selection
         gd.setCurrentSelection(getSelectedText());
         
         ke.setKeyCode(Keyboard.KA_CURSOR_LEFT);
         ke.setActionCode(Keyboard.KA_CURSOR_LEFT);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_HOME)
      {
         // expand selection
         if (config.selectionEnd == FillInConfig.INVALID_SELECTION)
         {
            config.selectionEnd =  Math.min(offset + 1, lastOffset + 1);
         }
         else if (offset == config.selectionEnd)
         {
            config.selectionEnd = config.selectionStart;
         }
         config.selectionStart = 0;
         // notify about the current selection
         gd.setCurrentSelection(getSelectedText());
         
         ke.setKeyCode(Keyboard.KA_HOME);
         ke.setActionCode(Keyboard.KA_HOME);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_END)
      {
         // expand selection
         if (config.selectionStart == FillInConfig.INVALID_SELECTION)
         {
            config.selectionStart = offset;
         }
         else if (offset == config.selectionStart)
         {
            config.selectionStart = config.selectionEnd;
         }

         config.selectionEnd = lastOffset + 1;
         // notify about the current selection
         gd.setCurrentSelection(getSelectedText());
         
         ke.setKeyCode(Keyboard.KA_END);
         ke.setActionCode(Keyboard.KA_END);
         invalidateSelection = false;
      }
      else if (key == Key.VK_CTRL_A)
      {
         if (getText(true).length() == 0)
         {
            invalidateSelection();
         }
         else
         {
            config.selectionStart = 0;
            config.selectionEnd = lastOffset + 1;
            // notify about the current selection
            gd.setCurrentSelection(getSelectedText());

            repaint();
         }
         
         return;
      }
      else if (key == Keyboard.KA_COPY || key == Keyboard.KA_CUT)
      {
         // do not copy the password
         if (!config.passwordField)
         {
            if (isSelectionValid())
            {
               // copy first
               gd.copyToClipboard(getSelectedText());
               
               // clear as needed
               if (key == Keyboard.KA_CUT)
               {
                  clear(config.selectionStart, config.selectionEnd);
               }
            }
         }
         
         ke.consume();
      }
      else if (key == Keyboard.KA_PASTE)
      {
         if (isSelectionValid())
         {
            // clear everything first
            clear(config.selectionStart, config.selectionEnd);
         }
         
         // insert the text
         String txt = gd.clipboardContents();
         
         for (char ch : txt.toCharArray())
         {
            if (!config.appDataPres.input(ch))
            {
               // widget no longer accepts this...
               break;
            }
         }
         
         repaint();
         
         ke.consume();
      }
      
      key = ke.actionCode();
      
      if (invalidateSelection && isSelectionValid())
      {
         // TODO: date format does not allow clearing...
         // TODO: date format does not allow BACKSPACE/DELETE
         if (Keyboard.isPrintableKey(key)        ||
             key == Keyboard.KA_BACKSPACE        ||
             key == Keyboard.KA_DELETE_CHARACTER)
         {
            // clear the selection
            clear(config.selectionStart, config.selectionEnd);
            
            if (key == Keyboard.KA_BACKSPACE || key == Keyboard.KA_DELETE_CHARACTER)
            {
               ke.consume();
            }

            // turn on auto-zap to properly replace selection region with char entered 
            setZapMode(true);
         }

         // no more selection
         if (key != Keyboard.KA_CLEAR && key != Keyboard.SE_ENTRY)
         {
            // selection needs to survive, it will be used by the FillIn.processKeyEvent
            invalidateSelection();
         }
      }
      
      DisplayFormat dispFormat = getCurrentFormat();
      
      if (!(dispFormat instanceof NumberFormat)            &&
          getCursorOffset() >= dispFormat.getScreenWidth() &&
          Keyboard.isPrintableKey(key))
      {
         // can not add keys past the last char...
         return;
      }
      
      if (key == Keyboard.KA_END)
      {
         moveCursorToLast();
         adjustEditScroller(Keyboard.KA_CURSOR_RIGHT);
         refresh();
         
         return;
      }
      
      if (key == Keyboard.KA_HOME)
      {
         moveCursorToFirst();
         refresh();
         return;
      }
      
      if (key == Keyboard.SE_ENTRY)
      {
         // nothing special done on ENTRY for GUI
         return;
      }
      
      
      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         if (offset == lastOffset)
         {
            // this needs to be treated explicitly... otherwise the DisplayFormat classes will
            // not allow the cursor to get past the last position via cursor-right key
            setCursorOffset(offset + 1);
            
            // the cursor will remain on the current position, but will be drawn right-side
            adjustEditScroller(Keyboard.KA_CURSOR_RIGHT);
            if (editScroller == rightAnchor)
            {
               editScroller.setAnchorIndex(offset + 1);
            }
            refresh();
            
            return;
         }
      }
      if (key == Keyboard.KA_CURSOR_LEFT && isRightAlignedInt())
      {
         if (offset >= lastOffset + 1)
         {
            // this needs to be treated explicitly... otherwise the DisplayFormat classes will
            // not allow the cursor to get past the last position via cursor-right key
            setCursorOffset(lastOffset);
            
            refresh();
            
            return;
         }
      }
      
      super.processKeyEvent(ke);

      if (parent() != null) // check that the widget wasn't deleted (e.g. it was in-browse widget)
      {
         adjustEditScroller(key);
      }
   }
   
   /**
    * Draws the caret.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }
      
      // drawing worker
      drawCaretInt();
   }

   /**
    * Actual drawing routine.
    *
    * @param   aColor
    *          Color for drawing.
    */
   @Override
   public void draw(Color aColor)
   {
      Theme theme = ThemeManager.getCurrentTheme();
      final GuiColorResolver gce = new GuiColorResolver(gd, config,
                                                        SysColor.COLOR_HIGHLIGHT, 
                                                        SysColor.COLOR_HIGHLIGHTTEXT);
      
      int bgColor = isThreeD() && (!isEnabled() || config.readOnly) ? 
                    SysColor.COLOR_3DFACE : 
                    SysColor.COLOR_FILLIN_BG;
      
      final GuiColorResolver gcd = new GuiColorResolver(gd, config, 
                                                        bgColor,
                                                        SysColor.COLOR_FILLIN_FG);
      gcd.boxColor = gd.getSysColor(SysColor.COLOR_FILLIN_FG);
      gce.selectionBgColor = gd.getSysColor(SysColor.COLOR_HOTLIGHT);
      gce.selectionFgColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      
      // If uninitialized and not for input and not date fill-in (in this case
      // separators should be displayed), display blanks instead.
      final boolean uninit = isUninitialized();

      gce.colors();
      gcd.colors();

      if (!disabledForeFixedColor)
      {
         disabledFore = gcd.fgColor;
      }
      
      NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      final String txt = getDrawText(theme.getPasswordCharacter());
      final NativeRectangle clip = clipRectangle(p, dim);
      final FontCache fc = gf.font();
      final FontDetails<?> font = fc.font;
      Integer fontNum = config.ehFont != null ? null : gf.resolveFontNum();
      int fontHeight = FontManager.getFontHeight(window(), gf.font().font, fontNum, gd);
      
      gd.draw(getId(), p, clip, () ->
      {
         // clear the area
         gd.setColor(gcd.bgColor);
         gd.fillRect(0, 0, dim.width, dim.height);
         
         ThinClient tc = ThinClient.getInstance();
         
         boolean enabled = isEnabled();
         boolean threeDBorder = isThreeD() && (tc.isKeep3DFillinBorder() || enabled);
         
         if (enabled || threeDBorder) 
         {
            drawBorder(gcd, 0, 0, dim, threeDBorder);
         }
         
         int xOffset = threeDBorder ? 1 : 0;
         
         if (!uninit)
         {
            NativeDimension edim = editableNativeDimension();
            // here we need to get text coord taking into account possible horizontal scrolling
             
            NativePoint txtLoc = textNativeLocation();
            gd.setGuiFont(font);
            
            if (isEnabled())
            {
               NativeRectangle txtClip = clipRectangle(txtLoc, edim);
               
               gd.draw(txtLoc, txtClip, () ->
               {
                  TextLineSelection textLineSelection = getTextSelection(0);
                  if (textLineSelection != null)
                  {
                     TextLookAndFeel textLookAndFeel = getTextLookAndFeel(gcd, gce);
                     gd.drawTextSelection(textLineSelection,
                                          textLookAndFeel,
                                          xOffset, fontHeight / 2, fontHeight, 0);
                  }
                  else
                  {
                     // draw full text (as if nothing is selected), first
                     gd.setColor(gcd.fgColor);
                     gd.drawStringCentered(txt, xOffset, fontHeight / 2);
                  }
               });
            }
            else
            {
               // set text color for disabled state
               gd.setColor(disabledFore);
               
               edim.width -= 1;
               
               NativeRectangle txtClip = clipRectangle(txtLoc, edim);
               // needs to be centered vertically ?
               gd.draw(txtLoc, txtClip, () ->
                     gd.drawStringCentered(txt, 0, fontHeight / 2));
            }
         }
         
         drawCaret();
         
         ThemeManager.debug(() ->
         {
            gd.setColor(java.awt.Color.BLUE);
            gd.drawRect(0, 0, dim.width - 1, dim.height - 1);
         });
      });
   }
   
   /**
    * Draws the border box around the fill-in.
    * 
    * @param    threeDBorder
    *           Indicates whether 3d or flat border should be drawn.
    */
   protected void drawBorder(GuiColorResolver gcd,
                             int x,
                             int y,
                             NativeDimension dim,
                             boolean threeDBorder)
   {
      ThemeManager.getCurrentTheme().drawFillInBorder(gd, gcd,
               x, y, dim.width, dim.height, threeDBorder, hasFocus(), false);
   }
   
   /**
    * Returns text look and feel.
    * 
    * @param    gcd
    *           The gui color resolver used for normal text.
    * @param    gce
    *           The gui color resolver used for selected text.
    * 
    * @return   The Fillin text look and feel.
    */
   protected TextLookAndFeel getTextLookAndFeel(GuiColorResolver gcd, GuiColorResolver gce)
   {
      return ThemeManager.getCurrentTheme().getTextLookAndFeel(gcd, gce);
   }
   
   /**
    * Compute the initial cursor position, depending on the widget's implementation.
    * <p>
    * In GUI, the initial cursor position is always on the right-side text edge.
    *  
    * @return   See above.
    */
   @Override
   public int getInitialCursorPos()
   {
      return getText(true).length();
   }
   
   /**
    * Obtain any selected text in the widget for copying to the clipboard.
    *
    * @return   The selected text or <code>null</code> if there is no selected text or if the
    *           widget type does not support clipboard operations.
    */
   @Override
   public String getSelectedText()
   {
      String selected = null;
      
      if (isSelectionValid())
      {
         String str = getText(true);
         if (str == null)
         {
            return null;
         }

         int len = str.length();
         if (config.selectionStart < 0 || config.selectionStart > len ||
             config.selectionEnd < 0 || config.selectionEnd > len)
         {
            return null;
         }

         selected = str.substring(config.selectionStart, config.selectionEnd);
      }
      
      return selected;
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvt.MOUSE_HOVERABLE
      };
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(FillInConfig beforeUpdate)
   {
      if (beforeUpdate == null)
      {
         return;
      }

      if (!StringHelper.equal(config.format, beforeUpdate.format ) || config.fontChanged)
      {
         updateSize();
      }
   }

   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   public boolean isRMBEntrySupported()
   {
      return false;
   }
   
   /**
    * Invalidate the text selection state. Delegate to the base method and notify the client that
    * the current selection is invalid.
    */
   @Override
   public void invalidateSelection()
   {
      boolean selValid = isSelectionValid();
      
      super.invalidateSelection();
      
      if (selValid)
      {
         gd.setCurrentSelection(null);
      }
   }
   
   /**
    * Removes selected text. If no text is selected acts as pressing DEL key by removing the
    * character to the right of the cursor. 
    */
   @Override
   public void removeSelectedText()
   {
      processKeyEvent(new KeyInput(this, EventType.KEY_TYPED, Key.VK_DELETE));
   }
   
   /**
    * Creates a new selection, eventually dropping the already existing one.
    *
    * @param   start
    *          The offset of the start of the new selection.
    * @param   end
    *          The offset of the end of the new selection.
    *
    * @return  {@code true} is operation is successful.
    */
   @Override
   public boolean setSelection(int start, int end)
   {
      int contentLength = getText(true).length();
      if (start > contentLength || end > contentLength)
      {
         return false; // invalid offset
      }
      
      int p1 = Math.min(start, end);
      int p2 = Math.max(start, end);
      
      setCursorOffset(p2); // may invalidate selection!
      config.selectionActive = true;
      config.selectionStart = p1;
      config.selectionEnd = p2;
      
      refresh();
      return true;
   }
   
   /**
    * Checks if the given widget is able to be the target of the drop operation.
    * 
    * @return   <code>TRUE</code> if this widget can be droppable at this time,
    *           <code>FALSE</code> otherwise.
    */
   @Override
   public boolean isDroppable()
   {
      return config.dropTarget; 
   }

   /**
    * Tests if the text selection is displayed.
    * 
    * @return   True if the current text selection is displayed, otherwise false.
    */
   protected boolean isTextSelectionDisplayed()
   {
      return hasFocus() && isEnabled() && isSelectionValid();
   }
   
   /**
    * Draws the caret.  Internal version.
    */
   protected void drawCaretInt()
   {
      if (ThinClient.getInstance().getOutputManager().isInvalidate())
      {
         // if this is called from outside of the actual driver drawing mechanism, ignore it.
         return;
      }
      
      NativePoint eorig = this.screenPhysicalLocation().translate(editableNativeLocation());
      eorig.x--;
      
      NativeDimension edim = editableNativeDimension();
      edim.width = edim.width + nativeInsets.right;
      
      final NativeRectangle eclip = clipRectangle(eorig, edim);

      // ensure we are on the correct window, as this can be called outside of window's draw
      TopLevelWindow<?> window = topLevelWindow();
      gd.selectWindow(window.getId().asInt());
      
      gd.drawFromOrigin(eorig, eclip, () -> gd.drawCursor(this));
      
      gd.releaseWindow();
   }
   
   /**
    * Adjust the edit scroller depending on the key being processed and the cursor location.
    * 
    * @param    key
    *           The currently processed key.
    */
   @Override
   protected void adjustEditScroller(int key)
   {
      int offset = getCursorOffset();
      int last = getLastOffset();
      
      if (key == Keyboard.KA_DELETE_CHARACTER)
      {
         String txt = getText(true);
         txt = txt.substring(0, offset);
         
         int tlen = getTextWidthNative(txt, gf.font().font);
         
         if (tlen > editableNativeDimension().width)
         {
            editScroller = rightAnchor;
            editScroller.setAnchorIndex(Math.min(offset + 1, last));
         }
         else
         {
            editScroller = leftAnchor;
            editScroller.setAnchorIndex(0);
         }
      }
      else
      {
         super.adjustEditScroller(key);
      }
   }

   /**
    * Get valid cursor position.
    * 
    * @return   valid cursor position.
    */
   protected int getValidCursorPos()
   {
      int pos = Math.min(getCursorOffset() + editScroller.getDelta(), getLastOffset() + 1);
      return pos;
   }
   
   /**
    * Get the cursor position, in native units, as it would be positioned in the fully-visible
    * text.
    * 
    * @return   See above.
    */
   @Override
   protected int getNativeCursorPos()
   {
      if (gd == null)
      {
         return 0;
      }

      int pos = getCursorOffset();
      
      String fullTxt = getDrawText(ThemeManager.getCurrentTheme().getPasswordCharacter());
      
      NativePoint np;
      
      if (pos >= 0)
      {
         String txt = pos >= fullTxt.length() ? fullTxt : fullTxt.substring(0, pos);
         np = new NativePoint(getTextWidthNative(txt, gf.font().font), 0);
      }
      else
      {
         np = new NativePoint(0, 0);
      }
      
      np = editScroller.adjustLocation(fullTxt, np);
      
      return np.x;
   }
   
   /**
    * Returns the insert mode state.
    * 
    * @return  In GUI, this is always <code>false</code>.
    */
   protected boolean getInsertMode()
   {
      return false;
   }
   
   /**
    * Reset the edit scroller to the initial state.
    */
   @Override
   protected void resetEditScroller()
   {
      boolean focused = hasFocus();
      if (editScroller == null || !focused)
      {
         // everything starts from the left-side
         editScroller = leftAnchor;
         editScroller.setAnchorIndex(0);
         
         if (!focused)
         {
            return;
         }
      }
      
      if (gd == null || gf == null)
      {
         // not yet initialized, nothing to do
         return;
      }
      
      int cursorPos = getInitialCursorPos();
      int txtWidth = getTextWidthNative(getText(true), gf.font().font);
      boolean rightEndCursor = txtWidth >= editableNativeDimension().width;

      editScroller = (rightEndCursor ? rightAnchor : leftAnchor);

      editScroller.setAnchorIndex(rightEndCursor ? cursorPos : 0);
   }
   
   /**
    * Refresh widget by performing an immediate drawing.
    * 
    * @param    delayed
    *           Flag indicating if the drawing should be immediately or delayed.
    */
   @Override
   protected void refresh(boolean delayed)
   {
      repaint();
   }
   
   /**
    * Create the left-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createLeftAnchor()
   {
      return new LeftAnchorEditScroller();
   }

   /**
    * Create the right-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createRightAnchor()
   {
      return new RightAnchorEditScroller();
   }

   /**
    * Determine the interior portion of the widget.  This may be different than the widget's
    * width, as determined by the WIDTH-CHARS or WIDTH-PIXELS attribute.
    * <p>
    * For GUI widgets, this is the physical dimension without the insets.
    * 
    * @return   The editable dimension of this widget, in native units.
    */
   @Override
   protected NativeDimension editableNativeDimension()
   {
      // ensure the explicit width (if set) is checked
      int width = nativeWidth();
      
      NativeDimension dim = physicalDimension();
      dim.width = width - nativeInsets.right - nativeInsets.left;
      dim.height-= nativeInsets.top + nativeInsets.bottom;
   
      String txt = getText(true);
      if (isRightAligned() && txt.length() > 0)
      {
         // TODO: this needs to be removed after the padding is moved at the 
         // DisplayFormat
         int fmtWidth = getCurrentFormat().getScreenWidth();
         // for numeric values, right-aligned is done via spaces (and after that
         // the text is assumed left-aligned)
         String paddedText = StringHelper.rightAlignText(txt, fmtWidth);
         String prefix = paddedText.substring(0, paddedText.length() - txt.length());
         int padLength = getTextWidthNative(prefix, gf.font().font);
         
         dim.width -= padLength;
      }

      
      return dim;
   }

   /**
    * Implements screen value blanking when BLANK modifier is used.
    *
    * @param   curPres
    *          Effective screen value.
    * @param   fmtWidth
    *          Effective format width.
    *
    * @return  The blanked display value.
    */
   @Override
   protected String blankScreenValue(DisplayFormat.Presentation curPres, int fmtWidth)
   {
      String txt = curPres.toScreenValue(isEnabled());
      char[] data = new char[txt.length()];
      Arrays.fill(data, ' ');
      return String.valueOf(data);
   }

   /**
    * Compute the location of the fill-in's editable content, relative to the fill-in's top-left
    * corner.
    * 
    * @return   See above.
    */
   private NativePoint editableNativeLocation()
   {
      NativePoint np = new NativePoint(nativeInsets.left, nativeInsets.top);
      String txt = getText(true);
      
      if (isRightAligned() && txt.length() > 0)
      {
         // TODO: this needs to be removed after the padding is moved at the 
         // DisplayFormat
         int fmtWidth = getCurrentFormat().getScreenWidth();
         // for numeric values, right-aligned is done via spaces (and after that
         // the text is assumed left-aligned)
         String paddedText = StringHelper.rightAlignText(txt, fmtWidth);
         String prefix = paddedText.substring(0, paddedText.length() - txt.length());
         int padLength = getTextWidthNative(prefix, gf.font().font);
         
         np = np.translate(padLength, 0);
      }
      

      return np;
   }

   /**
    * Compute the fill-in's content location, adjusted so that only the required portion will be
    * visible.
    * <p>
    * The returned location is relative to the fill-in's top-left corner, and the X coordinate
    * may be negative.  When used during drawing, the clipping rectangle ensures that nothing is
    * drawn outside of the required area.
    * 
    * @return   See above.
    */
   private NativePoint textNativeLocation()
   {
      String txt = getText(true);

      NativePoint eloc = editableNativeLocation();
      
      NativePoint txtLoc = editScroller.adjustLocation(txt, eloc);

      return txtLoc;
   }

   /**
    * Move the cursor to the rightmost index.
    */
   protected void moveCursorToLast()
   {
      setCursorOffset(getLastOffset() + 1);
   }

   /**
    * Move the cursor to the leftmost index.
    */
   private void moveCursorToFirst()
   {
      setCursorOffset(0);
   }

   /**
    * Check if this fill-in acts as "uninitialized".
    * 
    * @return   <code>true</code> if the fill-in is in the {@link ScreenBuffer#UNINITIALIZED} 
    *           state.
    */
   private boolean isUninitialized()
   {
      // TODO: probably it should be replaced by variable initialization
      //       in FrameHelper.clearComponents().
      return (getState() == ScreenBuffer.UNINITIALIZED)    && 
              !(config.appFormat != null               && 
                config.appFormat instanceof DateFormat &&
                isEnabled());
   }

   /**
    * Sets the current cursor position by the mouse position within the given text.
    * 
    * @param    stxt
    *           The given text
    * @param    e
    *           The mouse event holding the current mouse coordinates
    */
   private void setCursorOffset(String stxt, MouseEvent e)
   {
      // determine which letter is hovered
      int x = textNativeLocation().translate(screenPhysicalLocation()).x;

      int idx = 1;
      FontDetails font = gf.font().font;
      String txt = "";
      while (idx < stxt.length())
      {
         txt = stxt.substring(0, idx);

         int txtWidth = getTextWidthNative(txt, font);
         
         if (e.getX() <= x + txtWidth)
         {
            break;
         }

         idx = idx + 1;
      }
      
      // hovered letter is at index "idx".  need to found which side to draw the caret
      if (idx == stxt.length())
      {
         // mouse is on the right-side of the right-most letter
         moveCursorToLast();
      }
      else if (txt.length() > 0)
      {
         // find where is the hovering - left-side or right-side
         int leftSideWidth = getTextWidthNative(txt.substring(0, txt.length() - 1), font);
         int fullWidth = getTextWidthNative(txt, font);
         int letterWidth = fullWidth - leftSideWidth;
         int midLetter = x + leftSideWidth + Math.round(letterWidth / 2);
         setCursorOffset(e.getX() >= midLetter ? txt.length() : txt.length() - 1);
      }
   }

   /**
    * Content scroller implementation when the anchor is positioned on the left-side editable edge.
    */
   private class LeftAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         int delta = getTextWidthNative(txt.substring(0, Math.min(anchorIndex, txt.length())), 
                                        gf.font().font);
   
         return new NativePoint(np.x - delta, np.y);
      }
   
      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         return -anchorIndex;
      }
      
      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return anchorIndex;
      }
   }

   /**
    * Content scroller implementation when the anchor is positioned on the right-side editable edge.
    */
   private class RightAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         int nwidth = editableNativeDimension().width;
         int twidth = getTextWidthNative(txt.substring(0, Math.min(anchorIndex, txt.length())), 
                                         gf.font().font);
         int delta = nwidth - twidth - 1;
         
         return new NativePoint(np.x + delta, np.y);
      }
      
      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         return (int) width() - anchorIndex;
      }
      
      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return Math.max(anchorIndex - (int) width(), 0);
      }
   }

   /**
    * Draws a special cursor relative to the current origin.
    */
   @Override
   public void drawCursor()
   {
      FontCache fc = getFontCache();
      ThemeManager.getCurrentTheme().drawFillInCaret(this, gd, 1, 0, 1, fc.fontHeight - 1);
   }

   /**
    * Gets the screen text content.
    * 
    * @param    line
    *           The given line number
    * 
    * @return   The screen text content
    */
   @Override
   public String getScreenText(int line)
   {
      return getScreenText();
   }
   
   /**
    * Gets the screen text content.
    * 
    * @return   The screen text content
    */
   @Override
   public String getScreenText()
   {
      StringBuilder text = new StringBuilder(
               getDrawText(ThemeManager.getCurrentTheme().getPasswordCharacter()));
      
      int offset = getCursorOffset();
      
      int rightTrimmedLength = text.length();
      
      if (offset >= rightTrimmedLength)
      {
         text.setLength(offset + 1);
         for(int i = rightTrimmedLength; i <= offset; i++)
         {
            text.setCharAt(i, ' ');
         }
      }
      
      return text.toString();
   }

   /**
    * Gets the current cursor column within its text content.
    * 
    * @return   The current cursor column
    */
   @Override
   public int getCursorColumnOffset()
   {
      return getCursorOffset();
   }

   /**
    * Gets the current cursor row within its text content.
    * 
    * @return   The current cursor row
    */
   @Override
   public int getCursorRowOffset()
   {
      return 0;
   }

   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll column
    */
   @Override
   public int getScrollColumnOffset()
   {
      return 0;
   }

   /**
    * Gets the current scroll offset within its text content.
    * 
    * @return   The current scroll row
    */
   @Override
   public int getScrollRowOffset()
   {
      return 0;
   }

   /**
    * Returns the current font cache.
    * 
    * @return   The current font cache
    */
   @Override
   public FontCache getFontCache()
   {
      return gf.font();
   }

   /**
    * Gets the current text selection object for the given line. If the given line doesn't have
    * selected text, then this method returns null.
    * 
    * @param    line
    *           The given line number
    * 
    * @return   The current text selection object if there is selected text in the given line,
    *           otherwise null.
    */
   @Override
   public TextLineSelection getTextSelection(int line)
   {
      String txt = getScreenText();

      int txtLen = txt.length();

      boolean select = isTextSelectionDisplayed() &&
                       (config.selectionStart <= config.selectionEnd) &&
                       (config.selectionStart < txtLen);
      
      if (!select)
      {
         return null;
      }
      
      int start = config.selectionStart;
      
      int end = config.selectionEnd;
      
      if (end > txtLen)
      {
         end = txtLen;
      }
      
      String[] texts = new String[3];
      
      // prefix not-selected text
      texts[0] = (start == 0 ? "" : txt.substring(0, start));
      // selected text
      texts[1] = txt.substring(start, end);
      // suffix not-selected text
      texts[2] = (end == txtLen ? "" : txt.substring(end));

      int[] hintWidths = new int[3];
      
      for(int i = 0; i < 3; i++)
      {
         if (texts[i].isEmpty())
         {
            hintWidths[i] = 0;
         }
         else
         {
            hintWidths[i] = getTextWidthNative(texts[i], gf.font().font);
         }
      }
      
      /**
       * Defines the current text selection object.
       */
      return new TextLineSelection()
      {
         /**
          * Returns the 3-tuple of string widths used as hint widths for this line split into outer
          * and selected text.
          * 
          * @return   The 3-tuple of string hint widths in pixels
          */
         @Override
         public int[] getTextWidthsHint()
         {
            return hintWidths;
         }
         
         /**
          * Returns the column and row, where the current text selection gets the its origin.
          * 
          * @return   The beginning of the current text selection
          */
         @Override
         public NativePoint getSelectionStart()
         {
            return new NativePoint(config.selectionStart, 0);
         }

         /**
          * Returns the most closest point, column and row, that breaks the continuous selection text
          * part.
          * 
          * @return   The break point, column and row, of the current text selection
          */
         @Override
         public NativePoint getSelectionEnd()
         {
            return new NativePoint(config.selectionEnd, 0);
         }
         
         /**
          * Returns the 3-tuple of text parts into which this line has been split by text selection.
          * 
          * @return   The 3-tuple of text parts
          */
         @Override
         public String[] getLineParts()
         {
            return texts;
         }
      };
   }
}
