/*
** Module   : MnemonicInfoGuiImpl.java
** Abstract : GUI client object accumulating prepared label and mnemonic info.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20160222 Created initial version.
** 002 VIG 20160624 init() method: changed visibility, return value, renamed to prepare().
** 003 CA  20170929 Added MNEMONIC and PREPROCESSED-LABEL FWD extension attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.client.MnemonicInfo;

/**
 * This class accumulates label information: 
 * <ul>
 *   <li>label after processing of special character '&amp;' denoting mnemonic.</li> 
 *   <li>mnemonic character, used to call menu item, that it is assigned on.</li>
 *   <li>mnemonic index in preprocessed label for mnemonic drawing.</li>
 * </ul>
 */
public class MnemonicInfoGuiImpl 
extends MnemonicInfo
{
   /**
    * Default constructor.
    */
   public MnemonicInfoGuiImpl()
   {
      super();
   }
   
   /**
    * Constructor.
    * 
    * @param   defaultFirst
    *          <code>true</code> if first character is a default mnemonic.
    */
   public MnemonicInfoGuiImpl(boolean defaultFirst)
   {
      super(defaultFirst);
   }
   
   /**
    * Initialized fields with new menu item label. For GUI '&amp;' can't be a mnemonic.
    * 
    * @param   title
    *          New menu item label.
    *          
    * @return  prepared label String.
    */
   @Override
   public String prepare(String title)
   {
      String prepLabel = super.prepare(title);
      
      // Choose the first character in prepared label, that differs from '&'.
      Integer mnemIdx = getMnemIdx();
      Character mnem = getMnemonic(); 
      if (mnem != null && mnem == '&' && mnemIdx != null && mnemIdx == 0)
      {
         int i = 0;
         for (; i < prepLabel.length(); i++)
         {
            if (prepLabel.charAt(i) != '&')
               break;
         }
         
         if (i < prepLabel.length())
         {
            setMnemIdx(i);
            setMnemonic(prepLabel.charAt(i));
         }
      }
      
      if (cfg != null)
      {
         cfg.preprocessedLabel = prepLabel;
         cfg.mnemonic = mnemIdx;
      }

      return prepLabel;
   }
   
   /**
    * Process one '&amp;' characters sequence.
    * 
    * @param   start
    *          the first '&amp;' index from the right.
    * @param   end
    *          the last '&amp;' index from the right.
    * @param   mnemAssnd
    *          <tt>true</tt> means that a sequence of '&amp;' of odd length was met, wherein
    *          mnemonic wasn't necessarily being assigned.
    *          
    * @param   result
    *          Resulting string buffer.
    */
   @Override
   protected boolean processSequence(int start, int end, StringBuilder result, boolean mnemAssnd)
   {
      int seqLen = start - end + 1;
      
      // If the odd sequence of '&' was met, but mnemonic wasn't assigned 
      // save mnemonic index only for drawing.
      if (seqLen % 2 != 0 && mnemAssnd && getMnemIdx() == null && result.length() > start + 1)
      {
         setMnemIdx(start + 1);
      }
      
      return super.processSequence(start, end, result, mnemAssnd);
   }
}