/*
** Module   : FontHelper.java
** Abstract : a helper class which works with AWT fonts.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20171204 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.awt.*;
import java.awt.font.*;
import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.swing.*;

/**
 * Helper class which exposes APIs to work with AWT fonts.
 */
public class FontHelper
{
   /** A special map containing the attributes of a underlined font. */
   public static Map<TextAttribute, Integer> UNDERLINE_FONT = new HashMap<>();
   
   /** A canvas used to compute the font metrics.*/
   private static final Canvas FONT_CANVAS = new Canvas();
   
   /** A cache of AWT fonts. */
   private static final Map<String, Font> FONT_CACHE = new HashMap<>();
   
   static 
   {
      UNDERLINE_FONT.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
      UNDERLINE_FONT = Collections.unmodifiableMap(UNDERLINE_FONT);
   }

   /**
    * Create a AWT font with the specified details.
    * 
    * @param    details
    *           The font details.
    *           
    * @return   The AWT font.
    */
   public static Font createFont(FontDetails<?> details)
   {
      int style = 0;
      
      style += (details.style.isBold() ? Font.BOLD : 0);
      style += (details.style.isItalic() ? Font.ITALIC : 0);
      
      // TODO: bitmap fonts
      
      Font f = null;
      String name = details.fontName;
      int size = details.pointSize;
      if (details.fontDefinition != null)
      {
         try
         {
            f = Font.createFont(Font.TRUETYPE_FONT, 
                                new ByteArrayInputStream(details.fontDefinition));
            f = f.deriveFont((float) details.pointSize);
            f = f.deriveFont(style);
            
            name = f.getName();
         }
         catch (FontFormatException | IOException e)
         {
            throw new RuntimeException("Could not define custom font " + name + ", size " + size,
                                       e);
         }
      }
      else
      {
         f = new Font(name, style, size);
      }
      
      if (details.style.isUnderline())
      {
         f = f.deriveFont(UNDERLINE_FONT);
      }

      return f;
   }

   /**
    * Derive the given font to match the screen resolution.
    * 
    * @param    font
    *           The AWT font to derive.
    * @param    fd
    *           The font details.
    * @param    screenRes
    *           The screen resolution.
    *           
    * @return   The derived AWT font.
    */
   public static Font deriveFont(Font font, FontDetails<?> fd, int screenRes)
   {
      float factor = (float) (fd.pointSize * screenRes / 72.0);
      font = font.deriveFont(factor);
      
      return font;
   }
   
   /**
    * Get the {@link FontMetricsHelper} for the cached font, with the given key.
    * <p>
    * These are build using the detached {@link #FONT_CANVAS}.
    * 
    * @param    key
    *           The font key.
    *           
    * @return   The font metrics.
    */
   public static FontMetricsHelper getFontMetrics(String key)
   {
      Font font = FONT_CACHE.get(key);
      
      FontMetrics metrics = FONT_CANVAS.getFontMetrics(font);

      return new SwingFontMetrics(metrics);
   }
   
   /**
    * Get a font from cache.
    * 
    * @param    key
    *           The font key.
    *           
    * @return   The cached font.
    */
   public static Font get(String key)
   {
      return FONT_CACHE.get(key);
   }
   
   /**
    * Add a font to cache.
    * 
    * @param    key
    *           The font key.
    * @param    font
    *           The AWT font.
    */
   public static void add(String key, Font font)
   {
      FONT_CACHE.put(key, font);
   }
}
