/*
** Module   : GuiOutputManager.java
** Abstract : Provides a generic interface to low level GUI primitives.
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20140805 Created initial version.  ThinClient is imported explicitly.
** 002 VIG 20140826 Changed default window ID constant to dynamic ID definition.
** 003 MAG 20140912 Implement GUI status bar.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 MAG 20140925 Check if window has renderer when process default window.
** 007 CA  20141021 Removed the repaint(Widget) API: all drawing needs to be done via the common
**                  event processing.   Added translatedOrigin, to inform the caller about the 
**                  real location of the current origin point.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Added API to check if a certain key can raise a STOP condition.
** 011 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 012 EVL 20150107 Fix for dynamic windows renderer missing exception.
** 013 HC  20150323 WidgetId.getId() renamed to asInt().
** 014 GES 20150425 Match refactoring changes elsewhere.
** 015 HC  20150514 Support for modal windows.
** 016 HC  20150702 Implemented window modality independently on the actual UI driver.
** 017 GES 20150730 Removed translatedOrigin() since it isn't actually used.
** 018 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 019 HC  20151013 enableOSEvents() moved to WindowManager.
** 020 IAS 20160217 Special handling of the overlay windows.
** 021 HC  20160926 Removed disabling of window input events during server processing.
** 022 CA  20170104 Added flag to track window's modal state.
** 023 HC  20170823 Added runtime support of WINDOW:SHOW-IN-TASKBAR. Implemented runtime support
**                  for WINDOW:SHOW-IN-TASKBAR.
** 024 CA  20171018 Allow PUT, EXPORT and MESSAGE statements to work with streams and/or
**                  redirected terminal, in GUI.
** 025 CA  20171020 Added support for FRAMEs sent to named or unnamed streams, in GUI.
** 026 OM  20171029 Small fix in javadoc.
** 027 EVL 20171025 Javadoc fix.
** 028 HC  20171123 Fixed a stack overlow condition in GUI redirected mode.
** 029 CA  20180201 When cleaning the ChUI window, call detach() and not remove() for the frame -
**                  remove() will also destroy it, which is not OK.
** 030 HC  20180212 Fixed GUI output redirection to also work in DIALOG-BOX widgets.
** 031 CA  20180313 Added FWD extension SESSION:DISABLE-REDRAW. 
** 032 HC  20180629 Added a check to independentEventDrawingBracket for window existence.
** 033 CA  20180727 Removed setStatusText APIs.
** 034 CA  20180813 When switching output to redirected mode, re-build only ChUI frames for which
**                  their GUI counterpart has been changed - to improve performance.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.util.List;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provides a generic interface to low level graphic user interface primitives.
 */
public class GuiOutputManager
extends OutputManager<GuiPrimitives>
{
   /** Widget registry. */
   private WidgetRegistry<GuiOutputManager> registry = new WidgetRegistry<>();

   /** The initial configuration of this manager. */
   private final BootstrapConfig cfg;
   
   /** The {@link StreamableManager} instance used in redirected mode. */
   private final OutputManager<?> streamableMgr;
   
   /**
    * Create a new output manager instance, with the specified driver.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be <code>null</code>.
    * @param    cfg
    *           The initial configuration.
    */
   public GuiOutputManager(GuiDriver driver, BootstrapConfig cfg)
   {
      super(driver);
      
      this.cfg = cfg;
      
      streamableMgr = StreamableManager.newInstance();
   }
   
   /**
    * Get the context-specific instance, if it exists. In a single-client
    * JVM environment, this is equivalent to a singleton.
    * 
    * @return   The instance or <code>null</code> if none exists.
    */
   public static synchronized GuiOutputManager instance()
   {
      return (GuiOutputManager) defaultTerminal.get();
   }

   /**
    * Get the ChUI manager for this manager instance, used for redirected terminal support.
    * 
    * @return   The ChUI {@link #streamableMgr} instance.
    */
   @Override
   public StreamableManager getStreamableManager()
   {
      return (StreamableManager) streamableMgr;
   }
   
   /**
    * Reports if the terminal is currently redirected to a stream.
    * 
    * @return <code>true</code> if the terminal is redirected to a stream,
    *         <code>false</code> if the terminal is interactive.
    */
   @Override
   public boolean isRedirected()
   {
      return streamableMgr.isRedirected();
   }
   
   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    */
   @Override
   public void setInvalidate(Widget<?> widget, boolean mode, boolean flush)
   {
      if (isRedirected())
      {
         streamableMgr.setInvalidate(widget, mode, flush);
      }
      else
      {
         super.setInvalidate(widget, mode, flush);
      }
   }
   
   /**
    * Redirect terminal output to the drawing worker represented by
    * <code>newOps</code> (or back to the interactive terminal if
    * <code>newOps</code> is <code>null</code>) and return the instance
    * representing the current output destination. This may be another
    * <code>null</code> if the interactive terminal is in use.
    * 
    * @param    newOps
    *           The drawing worker or <code>null</code> to remove redirections
    *           and activate the interactive terminal.
    * 
    * @return   The previous drawing worker instance.
    */
   public OutputPrimitives switchOutput(OutputPrimitives newOps)
   {
      try
      {
         // temporarly switch the output manager to chuiMgr
         defaultTerminal.set(streamableMgr);
         
         OutputPrimitives old = streamableMgr.switchOutput(newOps);
         
         if (newOps != null)
         {
            List<TitledWindow<?>> windows = WindowManager.windowList();
            
            Window<?> streamableWindow = getStreamableManager().getStreamableWindow();
            Container<?> windowContainer = streamableWindow.getContentPane();
            WidgetRegistry<?> registry = streamableMgr.getRegistry();
            
            streamableWindow.getFrames()
                            .stream()
                            .forEach((f) -> 
                            {
                               Frame<?> guiFrame = 
                                (Frame<?>) GuiOutputManager.this.registry.getComponent(f.getId());
                               
                               // changed means the frame (or any child widget) state has 
                               // changed, or the frame has been destroyed.  we need to check only
                               // its GUI couterpart - otherwise keep all ChUI frame state as is
                               if (guiFrame == null || guiFrame.isChanged())
                               {
                                  // we need to detach here - calling 'remove' will destroy the
                                  // widgets (which is not OK, as the business logic must decide
                                  // when to destroy them).
                                  windowContainer.detach((Frame) f);

                                  // do not call registry.removeFrame, as this will destroy it.
                                  registry.removeWidgetsRecursive((Frame) f);
                               }
                            });
            
            windows.stream()
                   .filter((w) -> (w instanceof TopLevelWindow))
                   .forEach(w -> 
                   {
                      List<Frame<?>> frames = ((TopLevelWindow) w).getFrames();
                      
                      try
                      {
                         ConfigManager.getInstance().setTracking(false);

                         for (Frame<?> frame : frames)
                         {
                            if (!frame.isChanged())
                            {
                               continue;
                            }

                            Frame<?> streamableFrame = 
                               (Frame<?>) StreamableWidget.convertToStreamable(frame);

                            streamableWindow.getContentPane().add((Frame) streamableFrame);
                            streamableFrame.postprocess(true, true);
                            
                            // we track changes only in the GUI frame
                            frame.markChanged(false);
                         }
                      }
                      finally
                      {
                         ConfigManager.getInstance().setTracking(true);
                      }
                   });
         }
         
         return old;
      }
      finally
      {
         defaultTerminal.set(this);
      }
   }
   
   /**
    * Determine the screen bitmap associated with the given window.
    * <p>
    * After this call, the {@link #output} will have selected the specified window.
    * <p>
    * WARNING! If the {@link ScreenBitmap} reference is saved and a code is called which in turn
    * ends up changing the {@link ScreenBitmap} reference (via the {@link #setBitmap} API), then
    * ensure the proper reference is used!
    * 
    * @param    windowId
    *           The window ID.
    *           
    * @return   The {@link ScreenBitmap} associated with this window.
    */
   @Override
   protected ScreenBitmap screenBitmap(int windowId)
   {
      return isRedirected() ? getStreamableManager().screenBitmap() 
                            : super.screenBitmap(windowId);
   }

   /**
    * Registers new window with the UI driver.
    *
    * @param    window
    *           A window.
    */
   @Override
   public void registerWindow(TopLevelWindow<?> window)
   {
      int windowId = window.getId().asInt();

      WidgetConfig c = window.config();
  
      boolean showInTaskbar = false;
      if (c instanceof WindowConfig)
      {
         showInTaskbar = ((WindowConfig) c).showInTaskbar;
      }

      // this part of code need to be called for dynamically created windows not yet active to
      // register simulator
      registerWindow(windowId, showInTaskbar);
   }

   /**
    * Registers new window with the UI driver.
    *
    * @param    windowId
    *           A window ID to be realized.
    * @param    showInTaskbar
    *           If set to {@code true} the registered window will be shown in the system taskbar.
    */
   @Override
   public void registerWindow(int windowId, boolean showInTaskbar)
   {
      GuiDriver driver = (GuiDriver) this.driver;
      
      if (!driver.hasWindowEmulator(windowId))
      {
         driver.registerWindow(windowId, showInTaskbar);
      }
   }
   
   /**
    * Realize new child window.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           A window ID to be realized.
    * @param    windowTitle
    *           Title of the realized window.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    */
   public void realizeChildWindow(int ownerId, 
                                  int windowId, 
                                  String windowTitle, 
                                  boolean resizable, 
                                  boolean isShareActivationWithOwner)
   {
      realizeChildWindow(ownerId, windowId, windowTitle, 
                         resizable, isShareActivationWithOwner, false);
   }
   
   /**
    * Realize new child window.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           A window ID to be realized.
    * @param    windowTitle
    *           Title of the realized window.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    */
   public void realizeChildWindow(int ownerId, 
                                  int windowId, 
                                  String windowTitle, 
                                  boolean resizable, 
                                  boolean isShareActivationWithOwner,
                                  boolean modal)
   {
      GuiDriver driver = (GuiDriver) this.driver;
      
      if (!driver.hasWindowEmulator(windowId))
      {
         driver.registerChildWindow(ownerId, windowId, windowTitle, resizable, 
                                    isShareActivationWithOwner, modal);
      }
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(Rectangle bounds)
   {
      NativeRectangle nbounds = bounds == null ? null : coordinates().rectangleToNative(bounds);
      ((GuiDriver) driver).setDrawingArea(nbounds);
   }
   
   /**
    * Sets the global SESSION:DISABLE-REDRAW attribute.
    * When this mode is active, all output is disabled.
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   @Override
   public void setDisableRedraw(boolean value)
   {
      super.setDisableRedraw(value);

      ((GuiDriver) driver).setDisableRedraw(value);
   }
   
   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   @Override
   public void clear(Widget<?> widget)
   {
      if (widget == null || isRedirected())
         return;

      // TODO: which color?
      int windowId = widget.topLevelWindow().getId().asInt();
      
      clear(windowId, widget.physicalLocation(), widget.physicalDimension(), null);
   }

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public GuiWidgetFactory getFactory()
   {
      return (GuiWidgetFactory) driver.getFactory();
   }

   /**
    * Get the widget registry for this output manager.
    * 
    * @return   See above.
    */
   @Override
   public WidgetRegistry<GuiOutputManager> getRegistry()
   {
      return isRedirected() ? (WidgetRegistry) streamableMgr.getRegistry() : registry;
   }

   /**
    * Get the manager instance, depending if the mode is redirected or not.
    * 
    * @return   The manager instance.
    */
   public OutputManager managerInstance()
   {
      return isRedirected() ? streamableMgr : this;
   }

   /**
    * On a device which allows multiple windows to be created, this allows selecting an explicit
    * window before using the primitives or before drawing a widget.
    * 
    * @param    windowId
    *           Window ID.
    */
   @Override
   public void selectWindow(int windowId)
   {
      if (isRedirected())
      {
         Window<?> window = getStreamableManager().getStreamableWindow();
         streamableMgr.selectWindow(window.getId().asInt());
      }
      else
      {
         ((GuiDriver) driver).selectWindow(windowId);
      }
   }

   /**
    * On a device which allows multiple windows to be created, this allows releasing the window to
    * allow drawing from other threads.
    */
   @Override
   public void releaseWindow()
   {
      if (isRedirected())
      {
         streamableMgr.releaseWindow();
      }
      else
      {
         ((GuiDriver) driver).releaseWindow();
      }
   }

   /**
    * Check if window has an emulator for the given window id.
    *
    * @param    windowId
    *           Window ID.
    * @return   <code>true</code> if window has an associated emulator, <code>false</code>
    *           otherwise.
    */
   @Override
   public boolean hasWindowEmulator(int windowId)
   {
      return ((GuiDriver) driver).hasWindowEmulator(windowId);
   }

   /**
    * Check if the given key can raise the STOP condition.
    * 
    * @param    key
    *           The key code to be checked.
    * 
    * @return   <code>true</code> if the key is the {@link Key#VK_CTRL_BREAK} key.
    */
   @Override
   protected boolean raiseStopCondition(int key)
   {
      return key == Key.VK_CTRL_BREAK;
   }
   
   /**
    * Check if the window being drawn needs to be overridden.
    * <p>
    * In GUI, the <code>window</code> parameter is returned.
    *  
    * @param    window
    *           The currently active window.
    *           
    * @return   See above.
    */
   protected TitledWindow<?> overrideWindow(TopLevelWindow<?> window)
   {
      return window;
   }
}
