/*
** Module   : MouseDirectManipulation.java
** Abstract : Mouse direct manipulation event handler.
**
** Copyright (c) 2015-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description--------------------------------------
** 001 EVL 20170620 Initial version.
** 002 EVL 20170816 Adding frame move/resize handling.
** 003 SVL 20171116 Fixes for row resizing.
** 004 CA  20180530 Fixed problems with BUTTON:MOVABLE and mouse manipulation.
** 005 SVL 20180620 Fixed NPE because of right-clicking.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui.driver;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.web.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Mouse direct manipulation event handler.
 */
public class MouseDirectManipulation 
extends MouseWidgetAction
{
   /** Definition for possible resize directions. Starting with undefined */
   private final static int RESIZE_DIRECTION_UNDEF = -1;

   /** North-West resize direction. */
   public final static int RESIZE_DIRECTION_NW = 0;

   /** West resize direction. */
   public final static int RESIZE_DIRECTION_N  = 1;

   /** North-East resize direction. */
   public final static int RESIZE_DIRECTION_NE = 2;

   /** East resize direction. */
   public final static int RESIZE_DIRECTION_E  = 3;

   /** South-East resize direction. */
   public final static int RESIZE_DIRECTION_SE = 4;

   /** South resize direction. */
   public final static int RESIZE_DIRECTION_S  = 5;

   /** South-West resize direction. */
   public final static int RESIZE_DIRECTION_SW = 6;

   /** Resize direction. */
   public final static int RESIZE_DIRECTION_W  = 7;
   
   /** Resize handle size. */
   public final static int RESIZE_HANDLE_SIZE = 5;
   
   /** Point where drag operation is starting. */
   private NativePoint startPoint = null;

   /** Flag to separate frames and dialogs from other widgets. */
   private boolean isFrame = false;

   /** Parent container for widget the handler is attached to. */
   private Container<?> parentContainer = null;
   
   /** Actual widget whose component is attached to handler. */
   private Widget<?> widgetEff = null;
   
   /** Flag indicating box selection mode was activated.*/
   private boolean boxSelecting = false;

   /** Flag indicating widgets moving mode was activated.*/
   private boolean widgetsMoving = false;

   /** Flag indicating widget resizing mode was activated.*/
   private boolean widgetResizing = false;
   
   /** Flag indicating browse widget row resizing mode was activated.*/
   private boolean rowResizing = false;

   /** Flag to suppress the click events handling after mouse release with some conditions. */
   private boolean clickExpected = true;
   
   /** Widget reference to resize. */
   private Widget<?> widgetToResize = null;
   
   /** Resized row number for browse widget. */
   private int resizedRowNumber = -1;
   
   /** Resize direction. */
   private int resizeDirection = RESIZE_DIRECTION_UNDEF;

   /** Cache for X mouse coordinate. */
   private int xOld;
   
   /** Cache for Y mouse coordinate. */
   private int yOld;
   
   /** Cache for rectangle width. */
   private int wOld;
   
   /** Cache for rectangle height. */
   private int hOld;
   
   /** Color for selection box. */
   private ColorRgb boxColor = new ColorRgb(0x00000000); 

   /** The mouse pointer currently in use. */
   private MousePtrWrapper ptrCurr = MousePtrWrapper.DEFAULT;
   
   /** Flag used to control first drag event in sequential drag operations. */
   private boolean dragInProgress = false;
   
   /** Event code recently consumed as result of the 4GL NO-APPLY trigger execution. */
   private int eventNoApply = -1;
   
   /** Parent window workspace for widget the handler is attached too. */
   private Container<?> windowWorkSpace = null;
   
   /** Flag to indicate we are looking at the frame from window point of view. */
   private boolean isWindow = false;

   /** The widget's custom mouse pointer (if any). */
   private MousePtrWrapper mousePtr = null;
   
   /**
    * Constructor.
    * 
    * @param    widget
    *           The widget which contains the objects for direct manipulation.
    */
   public MouseDirectManipulation(Widget<?> widget)
   {
      super(widget, 
            new int[] { MouseEvent.MOUSE_CLICKED, MouseEvent.MOUSE_PRESSED,
                        MouseEvent.MOUSE_RELEASED, MouseEvent.MOUSE_MOVED,
                        MouseEvent.MOUSE_DRAGGED }, 
            Integer.MIN_VALUE);

      // check if the widget is frame, dialog or other widget
      if (widget instanceof ScrollContainer || widget instanceof ScrollPaneGuiImpl)
      {
         widgetEff = widget.parent();
         isFrame = true;
      }
      else
      {
         widgetEff = widget;
      }

      parentContainer = UiUtils.locateFrame(widget);

      // on the frame level we need to know the window based parent container
      // get the window workspace associated with given widget
      WindowGuiImpl parentWnd = (WindowGuiImpl)WindowManager.resolveWindow(
                                                    (AbstractWidget<GuiOutputManager>)widget);
      if (parentWnd != null)
      {
         windowWorkSpace = parentWnd.getWorkspace();
         if (widget instanceof FrameGuiImpl)
         {
            isWindow = true;
         }
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // possible issue here, drag was started somewhere else or other error
      if (startPoint == null)
      {
         return;
      }

      // the current mouse coordinates related to widget area
      NativePoint n = ((AbstractWidget)widget).translate(e);
      
      // the first priority is widget resizing
      if (widgetResizing && eventNoApply != Keyboard.SE_START_RESIZE)
      {
         // clean up previous drawings
         if (dragInProgress)
         {
            if (isFrameResizing())
            {
               drawSelectionBox(windowWorkSpace, xOld, yOld, wOld, hOld, true);
            }
            else
            {
               drawSelectionBox(((Frame)parentContainer).getContentPane(),
                                xOld, yOld, wOld, hOld, true);
            }
         }
         else
         {
            dragInProgress = true;
         }
         // get the resizable widget bounds
         NativeRectangle wr = new NativeRectangle(widgetToResize.physicalLocation(),
                                                  widgetToResize.physicalDimension());
         // change location and size depending on resize direction
         int xShift = n.x - startPoint.x;
         int yShift = n.y - startPoint.y;
         // calculate the new suggested widget bounds depending on direction to resize
         int x = wr.left();
         int y = wr.top();
         int w = wr.width();
         int h = wr.height();
         switch (resizeDirection)
         {
            case RESIZE_DIRECTION_NW:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + xShift, false) - x;
                  yShift = alignToGrid(y + yShift, true) - y;
               }
               // then get the new coordinates
               x += xShift;
               y += yShift;
               w -= xShift;
               h -= yShift;
               break;
            case RESIZE_DIRECTION_N:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  yShift = alignToGrid(y + yShift, true) - y;
               }
               // then get the new coordinates
               y += yShift;
               h -= yShift;
               break;
            case RESIZE_DIRECTION_NE:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + w + xShift, false) - x - w;
                  yShift = alignToGrid(y + yShift, true) - y;
               }
               // then get the new coordinates
               y += yShift;
               w += xShift;
               h -= yShift;
               break;
            case RESIZE_DIRECTION_E:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + w + xShift, false) - x - w;
               }
               // then get the new coordinates
               w += xShift;
               break;
            case RESIZE_DIRECTION_SE:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + w + xShift, false) - x - w;
                  yShift = alignToGrid(y + h + yShift, true) - y - h;
               }
               // then get the new coordinates
               w += xShift;
               h += yShift;
               break;
            case RESIZE_DIRECTION_S:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  yShift = alignToGrid(y + h + yShift, true) - y - h;
               }
               // then get the new coordinates
               h += yShift;
               break;
            case RESIZE_DIRECTION_SW:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + xShift, false) - x;
                  yShift = alignToGrid(y + h + yShift, true) - y - h;
               }
               // then get the new coordinates
               x += xShift;
               w -= xShift;
               h += yShift;
               break;
            case RESIZE_DIRECTION_W:
               // if the grid snap is on we need to recalculate the shift first
               if (isGridSnapOn())
               {
                  xShift = alignToGrid(x + xShift, false) - x;
               }
               // then get the new coordinates
               x += xShift;
               w -= xShift;
               break;
         }
         // draw the new suggested size 
         if (isFrameResizing())
         {
            drawSelectionBox(windowWorkSpace, x, y, w, h, false);
         }
         else
         {
            drawSelectionBox(((Frame)parentContainer).getContentPane(), x, y, w, h, false);
         }
         // save parameters for resizing complete
         xOld = x;
         yOld = y;
         wOld = w;
         hOld = h;
      }
      // then row resizing
      else if (rowResizing && eventNoApply != Keyboard.SE_START_ROW_RESIZE)
      {
         // get the main browse widget
         BrowseGuiImpl br = ((BrowseColumnGuiImpl) widgetEff).getBrowse();
         // get the resizable row bounds
         NativeDimension nd = br.getViewportDimension();
         // clean up previous drawings
         if (dragInProgress)
         {
            //TODO left and right borders of selection box are inaccurate, also check NO-BOX and
            // NO-SCROLLBAR-VERTICAL cases
            drawSelectionBox(br.getScrollPane(),
                             BrowseGuiImpl.BORDER_THICKNESS, yOld, nd.width, hOld, true);
         }
         else
         {
            dragInProgress = true;
         }
         // calc new y and height
         int y = br.getTopYforRow(resizedRowNumber);
         int h = br.getRowHeight() + BrowseGuiImpl.BORDER_THICKNESS;
         int brY = isGridSnapOn()
                      ? br.physicalLocation().y + br.getScrollPane().physicalLocation().y : 0;

         // change location and size depending on resize direction
         int yShift = n.y - startPoint.y;
         // calculate the new suggested row bounds depending on direction to resize
         if (isGridSnapOn()) // if the grid snap is on we need to recalculate the shift first
         {
            yShift = alignToGrid(y + brY + h + yShift, true) - y - brY - h;
         }
         // then get the new coordinates
         h += yShift;
         h = limitRowHeight(h, br);

         // draw the new suggested size
         drawSelectionBox(br.getScrollPane(), BrowseGuiImpl.BORDER_THICKNESS, y,
                          nd.width, h, false);
         // save parameters for resizing complete
         yOld = y;
         hOld = h;
      }
      // then widgets moving
      else if (widgetsMoving && eventNoApply != Keyboard.SE_START_MOVE)
      {
         // new move distance
         int xShift = n.x - startPoint.x;
         int yShift = n.y - startPoint.y;
         // clean up previous drawings
         if (dragInProgress)
         {
            drawMovableWidgetsBounds(xOld, yOld, true);
         }
         else
         {
            dragInProgress = true;
         }
         // draw new bounds
         drawMovableWidgetsBounds(xShift, yShift, false);
         xOld = xShift;
         yOld = yShift;
      }
      // drag in a frame enters box selection mode
      else if (boxSelecting && eventNoApply != Keyboard.SE_START_BOX_SELECTION)
      {
         // calculate rectangle parameters
         int x = n.x < startPoint.x ? n.x : startPoint.x;
         int y = n.y < startPoint.y ? n.y : startPoint.y;
         int w = Math.abs(n.x - startPoint.x);
         int h = Math.abs(n.y - startPoint.y);
         // selection rectangle drawing clearing canvas from previous draw
         if (dragInProgress)
         {
            // erase old selection box
            drawSelectionBox(widget, xOld, yOld, wOld, hOld, true);
         }
         else
         {
            dragInProgress = true;
            EventManager.postEvent(new DirectManipulationEvent(this, widgetEff,
                                                               Keyboard.SE_START_BOX_SELECTION));
         }
         drawSelectionBox(widget, x, y, w, h, false);
         xOld = x;
         yOld = y;
         wOld = w;
         hOld = h;
      }
   }

   /**
    * Notification of a mouse moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // if we are in any of the dragged mode suppress pointer change
      if (widgetResizing || rowResizing || widgetsMoving || boxSelecting)
      {
         return;
      }
      
      MousePtrWrapper ptrNew = null;
      // the pointer can be changed either by widget resize handles or by browse row resize
      if (isFrame || isWindow)
      {
         ptrNew = getNewPointer(e);
      }
      else if (isInsideRowResizable())
      {
         ptrNew = getRowResizePointer(e);
      }
      // set up new pointer if got one
      if (ptrNew != null && ptrNew != ptrCurr)
      {
         widgetEff.setMousePointer(ptrNew);
         ptrCurr = ptrNew;
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * In GUI mode, if the widget has a popup-menu attached, it will display it.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }

      // sometimes the incoming click is not correct, the web driver after drag operation
      // finished is an example
      // see handling for mouseReleased to understand how it works
      if (!clickExpected)
      {
         return;
      }

      // reset starting point for further use
      startPoint = null;
      
      // when widget selectable inverts the widget selected flag
      if (widgetEff != null && widgetEff.config() != null && widgetEff.config().selectable)
      {
         // mouse EXTENDED selection means selection inversion
         if (isExtended(e))
         {
            // invert selection for EXTENDED button press
            if ((widgetEff instanceof BrowseColumnGuiImpl &&
                 ((BrowseColumnGuiImpl)widgetEff).getBrowse().config().selected) ||
                ((BaseConfig)widgetEff.config()).selected)
            {
               // deselect selected widget
               EventManager.postEvent(new DirectManipulationEvent(widgetEff, parentContainer,
                                                                  true, Keyboard.SE_DESELECTION));
            }
            else
            {
               // select deselected widget
               EventManager.postEvent(new DirectManipulationEvent(widgetEff, parentContainer,
                                                                  true, Keyboard.SE_SELECTION));
            }
         }
         // simple selection means select single widget deselecting all others
         else
         {
            // post the event to target widget
            selectSingleWidget(true);
         }
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }
      
      mousePtr = widget.getCustomMousePointer().orElse(null);

      // reset NO-APPLY cache
      eventNoApply = -1;
      startPoint = ((AbstractWidget)widget).translate(e);
      // prepare mouse coord to resize operation check 
      NativePoint resizePoint = ((Frame)parentContainer).getContentPane().translate(e);
      widgetResizing = isInsideFrameResizeHandle(resizePoint) ||
                       isInsideResizeHandle(resizePoint);
      // widget resizing has higher priority comparing to widgets moving
      if (!widgetResizing)
      {
         rowResizing = isInsideRowResizable() && isInsideRowResizeArea(e);
         
         if (rowResizing)
         {
            // start browse widget row resize mode
            EventManager.postEvent(new DirectManipulationEvent(
                                                     ((BrowseColumnGuiImpl)widgetEff).getBrowse(),
                                                                   Keyboard.SE_START_ROW_RESIZE));
         }
         else
         {
            widgetsMoving = isInsideMovable(e) && !isFrame;
            // the case when we start moving nonselected widget
            if (widgetsMoving)
            {
               if (!((BaseConfig)widgetEff.config()).selected && !isExtended(e))
               {
                  selectSingleWidget();
               }
               // start widgets move mode
               EventManager.postEvent(new DirectManipulationEvent(this, widgetEff,
                                                                  Keyboard.SE_START_MOVE));
            }
            else if (isFrame && !isExtended(e))
            {
               // deselect everything selected
               deselectEverything();
               // empty selection events after everything is deselected
               EventManager.postEvent(new DirectManipulationEvent(widgetEff,
                                                                  Keyboard.SE_EMPTY_SELECTION));
               if (((PaneConfig)widgetEff.config()).boxSelectable)
               {
                  // we will enter box selecting mode if drag next
                  boxSelecting = true;
               }
            }
         }
      }
      else
      {
         // start widget resizing mode
         EventManager.postEvent(new DirectManipulationEvent(this, widgetToResize,
                                                            Keyboard.SE_START_RESIZE));
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }

      NativePoint endPoint = ((AbstractWidget)widget).translate(e);
      // for every END-* events we need to check if the trigger did not consume the
      // START-* event counterpart
      if (boxSelecting && dragInProgress && eventNoApply != Keyboard.SE_START_BOX_SELECTION)
      {
         // build rectangle from start and end points depending on drag direction
         int top = endPoint.y < startPoint.y ? endPoint.y : startPoint.y;
         int left = endPoint.x < startPoint.x ? endPoint.x : startPoint.x;
         int bottom = endPoint.y < startPoint.y ? startPoint.y : endPoint.y;
         int right = endPoint.x < startPoint.x ? startPoint.x : endPoint.x;

         // getting the selection box rectangle
         NativeRectangle selRect = new NativeRectangle(top, left, bottom, right);
         // getting extended flag form mouse event
         boolean isExtended = isExtended(e);
         
         // check if there are widgets inside selection box
         Widget<?>[] widLst = ((Frame)parentContainer).getContentPane().widgets();
         for (Widget<?> w : widLst)
         {
            // skip widgets must be excluded
            if (!w.isDirectManipulable())
            {
               continue;
            }
            
            BaseConfig bc = w.config();
            if (bc.selectable)
            {
               NativeRectangle widBounds = new NativeRectangle(w.physicalLocation(),
                                                               w.physicalDimension());
               if (selRect.contains(widBounds))
               {
                  if (isExtended)
                  {
                     if (!bc.selected)
                     {
                        EventManager.postEvent(new DirectManipulationEvent(w, parentContainer,
                                                                          Keyboard.SE_SELECTION));
                     }
                     else
                     {
                        EventManager.postEvent(new DirectManipulationEvent(w, parentContainer,
                                                                        Keyboard.SE_DESELECTION));
                     }
                  }
                  else
                  {
                     if (!bc.selected)
                     {
                        EventManager.postEvent(new DirectManipulationEvent(w, parentContainer,
                                                                          Keyboard.SE_SELECTION));
                     }
                  }
               }
               else if (!isExtended && bc.selected)
               {
                  EventManager.postEvent(new DirectManipulationEvent(w, parentContainer,
                                                                     Keyboard.SE_DESELECTION));
               }
            }
         }
         
         // post message the box selection mode is completed
         // clean up the drawing box and repaint selected widget status
         EventManager.postEvent(new DirectManipulationEvent(widgetEff,
                                                            Keyboard.SE_END_BOX_SELECTION));

         // erase old selection box
         drawSelectionBox(widget, xOld, yOld, wOld, hOld, true);
      }
      else if (widgetResizing && eventNoApply != Keyboard.SE_START_RESIZE)
      {
         // size change values
         int xShift = endPoint.x - startPoint.x;
         int yShift = endPoint.y - startPoint.y;
         // consider possible grid snap on
         if (isGridSnapOn())
         {
            NativePoint cLoc = widgetToResize.physicalLocation();
            NativeDimension cDim = widgetToResize.physicalDimension();
            // adjust X and Y shifts depending on resize directions
            if (resizeDirection == RESIZE_DIRECTION_NW ||
                resizeDirection == RESIZE_DIRECTION_SW ||
                resizeDirection == RESIZE_DIRECTION_W)
            {
               xShift = alignToGrid(cLoc.x + xShift, false) - cLoc.x;
            }
            else if (resizeDirection == RESIZE_DIRECTION_NE ||
                     resizeDirection == RESIZE_DIRECTION_E  ||
                     resizeDirection == RESIZE_DIRECTION_SE)
            {
               xShift = alignToGrid(cLoc.x + cDim.width + xShift, false) - cLoc.x - cDim.width;
            }
            if (resizeDirection == RESIZE_DIRECTION_NW ||
                resizeDirection == RESIZE_DIRECTION_N  ||
                resizeDirection == RESIZE_DIRECTION_NE)
            {
               yShift = alignToGrid(cLoc.y + yShift, true) - cLoc.y;
            }
            else if (resizeDirection == RESIZE_DIRECTION_SE ||
                     resizeDirection == RESIZE_DIRECTION_S  ||
                     resizeDirection == RESIZE_DIRECTION_SW)
            {
               yShift = alignToGrid(cLoc.y + cDim.height + yShift, true) - cLoc.y - cDim.height;
            }
         }
         // reconstruct the resize event, enclosing frame will be repainted
         DirectManipulationEvent dmEvent = new DirectManipulationEvent(widgetToResize,
                                                                       isFrameResizing()
                                                                          ? windowWorkSpace
                                                                          : parentContainer,
                                                                       true,
                                                                       Keyboard.SE_END_RESIZE);
         // set up move shift
         dmEvent.setDistance(xShift, yShift);
         // set up resize direction
         dmEvent.setResizeDirection(resizeDirection);
         // post event
         EventManager.postEvent(dmEvent);

         // clean up previous drawings
         if (isFrameResizing())
         {
            drawSelectionBox(windowWorkSpace, xOld, yOld, wOld, hOld, true);
         }
         else
         {
            drawSelectionBox(((Frame)parentContainer).getContentPane(),
                             xOld, yOld, wOld, hOld, true);
         }
      }
      // then row resizing 
      else if (rowResizing && eventNoApply != Keyboard.SE_START_ROW_RESIZE)
      {
         // get the main browse widget
         BrowseGuiImpl br = ((BrowseColumnGuiImpl)widgetEff).getBrowse();
         // size change value
         int yShift = endPoint.y - startPoint.y;
         // consider possible grid snap on
         if (isGridSnapOn())
         {
            // calc new y and height(this is frame oriented Y value)
            int y = br.getTopYforRow(resizedRowNumber) +
                    br.physicalLocation().y +
                    br.getScrollPane().physicalLocation().y;
            int h = br.getRowHeight() + BrowseGuiImpl.BORDER_THICKNESS;
            // adjust shift value considering grid snap
            yShift = alignToGrid(y + h + yShift, true) - y - h;
         }

         int newHeight = limitRowHeight(br.getRowHeight() + yShift, br);
         yShift = newHeight - br.getRowHeight();

         // reconstruct the row resize event, enclosing frame will be repainted
         DirectManipulationEvent dmEvent =
            new DirectManipulationEvent(br, parentContainer, true, Keyboard.SE_END_ROW_RESIZE);
         // integrate height change
         dmEvent.setDistance(0, yShift);
         // set up resize direction
         dmEvent.setResizeDirection(resizeDirection);
         // post event
         EventManager.postEvent(dmEvent);

         // get the resizable row bounds
         NativeDimension nd = br.getViewportDimension();
         // clean up previous drawings
         drawSelectionBox(br.getScrollPane(),
                          BrowseGuiImpl.BORDER_THICKNESS, yOld, nd.width, hOld, true);
      }
      else if (widgetsMoving && eventNoApply != Keyboard.SE_START_MOVE)
      {
         // position change values
         int xMove = endPoint.x - startPoint.x;
         int yMove = endPoint.y - startPoint.y;
         // obtain widgets to be moved
         Widget<?>[] widMove = getWidgetsToMove();
         // post them END_MOVE event with distance to move
         for (int i = 0; i < widMove.length; i++)
         {
            DirectManipulationEvent dmEvent = null;
            if (i == widMove.length - 1)
            {
               // clean up the movable widgets enclosing frame
               if (isWindow)
               {
                  dmEvent = new DirectManipulationEvent(widMove[i], windowWorkSpace,
                                                        true, Keyboard.SE_END_MOVE);
               }
               else
               {
                  dmEvent = new DirectManipulationEvent(widMove[i], parentContainer,
                                                        true, Keyboard.SE_END_MOVE);
               }
            }
            else
            {
               // regular processing
               dmEvent = new DirectManipulationEvent(widMove[i], parentContainer,
                                                     Keyboard.SE_END_MOVE);
            }
            // set up move shift taking into account possible grid snap flag
            if (isGridSnapOn())
            {
               NativePoint currLoc = widMove[i].physicalLocation();
               dmEvent.setDistance(alignToGrid(currLoc.x + xMove, false) - currLoc.x,
                                   alignToGrid(currLoc.y + yMove, true) - currLoc.y);
            }
            else
            {
               dmEvent.setDistance(xMove, yMove);
            }
            // and post event
            EventManager.postEvent(dmEvent);
         }

         // clean up previous drawings
         drawMovableWidgetsBounds(xOld, yOld, true);
      }
      // for some reasons web client driver unlike swing one generates click even for
      // pres-drag-release sequence so we disable click handling here for drag operations
      clickExpected = !dragInProgress;
      // clean up all mode flags
      startPoint = null;
      boxSelecting = false;
      widgetsMoving = false;
      widgetResizing = false;
      rowResizing = false;
      widgetToResize = null;
      if (dragInProgress)
      {
         // finish any selection box paintings
         gd.resetSelection();
         dragInProgress = false;
      }
      resizeDirection = RESIZE_DIRECTION_UNDEF;
      ptrCurr = MousePtrWrapper.DEFAULT;
      // restore mouse pointer
      widgetEff.setMousePointer(mousePtr == null ? ptrCurr : mousePtr);
   }

   /**
    * Saves the message code consumed by the trigger.  Used by some message processing for event
    * sequences with mutual dependency.
    * 
    * @param    code
    *           The event code that was consumed.
    */
   public void setNoApplyEvent(int code)
   {
      eventNoApply = code;
   }

   /**
    * Return row height trimmed according to its minimum and maximum allowed values.
    *
    * @param  h
    *         Row height.
    * @param  browse
    *         Browse which contains the resized row.
    *
    * @return row height trimmed according to its minimum and maximum allowed values.
    */
   private int limitRowHeight(int h, BrowseGuiImpl browse)
   {
      h = Math.max(h, BrowseGuiImpl.MIN_RESIZE_ROW_HEIGHT);
      h = Math.min(h, browse.getViewportDimension().height -
            browse.getTopYforRow(resizedRowNumber) - BrowseGuiImpl.ROW_RESIZE_BOTTOM_OFFSET);
      return h;
   }

   /**
    * Immediate drawing the selection box area with dashed line style.  The coordinates provided
    * are related to the widget the mouse handler is attached to.
    * 
    * @param    widBase
    *           The widget which context will be used for drawing.
    * @param    x
    *           The X starting coordinate.
    * @param    y
    *           The Y starting coordinate.
    * @param    w
    *           The width of the area.
    * @param    h
    *           The height of the area.
    * @param    clear
    *           <code>TRUE</code> means clean up area instead of drawing selections.
    */
   private void drawSelectionBox(Widget<?> widBase, int x, int y, int w, int h, boolean clear)
   {
      gd.draw(widBase.getId(), widBase.screenPhysicalLocation(), null, () ->
      {
         try
         {
            gd.startBatch();
            try
            {
               gd.setDrawingArea(new NativeRectangle(new NativePoint(x, y),
                                 new NativeDimension(w + 1, h + 1)));
               gd.setColor(boxColor);
               gd.drawSelection(x, y, w, h, clear);
            }
            finally
            {
               gd.setDrawingArea(null);
               gd.endBatch();
            }
         }
         catch (Exception exc)
         {
            System.out.printf("Unexpected exception in direct manipulation handler."+
                              " Widget: %s\n", widBase.getClass().getName());

            exc.printStackTrace();
         }
      });
   }
   
   /**
    * Immediate drawing the rectangular dotted line bounds for movable selected widgets.  The
    * distance provided are the shifts from original location in X and Y directions.
    * 
    * @param    xDistance
    *           The X direction distance to move from original position.
    * @param    yDistance
    *           The Y direction distance to move from original position.
    * @param    clear
    *           <code>TRUE</code> means clean up area instead of drawing selections.
    */
   private void drawMovableWidgetsBounds(int xDistance, int yDistance, boolean clear)
   {
      // get rectangles for movable widgets
      NativeRectangle [] nrMovable = getMovableWidgetsBounds();
      // draw rectangles for movable widgets
      if (nrMovable != null && nrMovable.length > 0)
      {
         for (NativeRectangle nr : nrMovable)
         {
            int xNewPos = nr.left() + xDistance;
            int yNewPos = nr.top() + yDistance;
            // if GRID-SNAP is active align new coords to grid lines
            if (isGridSnapOn())
            {
               xNewPos = alignToGrid(xNewPos, false);
               yNewPos = alignToGrid(yNewPos, true);
            }
            
            if (isWindow)
            {
               drawSelectionBox(windowWorkSpace,
                                xNewPos, yNewPos, nr.width(), nr.height(), clear);
            }
            else
            {
               drawSelectionBox(((Frame)parentContainer).getContentPane(),
                                xNewPos, yNewPos, nr.width(), nr.height(), clear);
            }
         }
      }
   }
   
   /**
    * Checks if the mouse event is extended or not.  Usually the <code>Ctrl</code> key pressing
    * turns extended key mode on.
    * 
    * @param    e
    *           The mouse event to check.
    * 
    * @return   <code>TRUE</code> if mouse event is extended, <code>FALSE</code> otherwise.
    */
   private boolean isExtended(MouseEvent e)
   {
      int extSelectionMask = InputEvent.CTRL_DOWN_MASK;
      
      return (e.getModifiersEx() & extSelectionMask) == extSelectionMask;
   }
   
   /**
    * Checks if the mouse event is inside movable widget.
    * 
    * @param    e
    *           The mouse event to check.
    * 
    * @return   <code>TRUE</code> if mouse event is inside movable widget,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isInsideMovable(MouseEvent e)
   {
      WidgetConfig wc = widgetEff.config();
      return wc instanceof BaseConfig ? ((BaseConfig)wc).movable : false;
   }
   
   /**
    * Checks if the given point is over the widget's resize handle.
    * <p>
    * The method remember the widget and resize direction if the point is over resize handle.
    * 
    * @param    np
    *           The native point to check.
    * 
    * @return   <code>TRUE</code> if the point is inside resize handle,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isInsideResizeHandle(NativePoint np)
   {
      // first exclude simple cases
      if (!isFrame)
      {
         return false;
      }
      
      // here we are within container, check if pointing to some selected widget resize handle
      Widget<?>[] widLst = ((Frame)widgetEff).getContentPane().widgets();
      for (Widget<?> w : widLst)
      {
         // skip widgets must be excluded
         if (!w.isDirectManipulable())
         {
            continue;
         }
            
         BaseConfig bwc = w.config();
         if (bwc.selected && bwc.resizable)
         {
            // generate resize handles coordinates for current widget
            NativeRectangle[] resizeHandles = getResizeHandles(w);
            // check if we are inside one of them
            for(int i = 0; i < resizeHandles.length; i++)
            {
               if (resizeHandles[i].contains(np))
               {
                  // found inclusion, return immediately
                  // remembering widget to resize and direction
                  widgetToResize = w;
                  resizeDirection = i;
                  return true;
               }
            }
         }
      }
      
      return false;
   }
   
   /**
    * Checks if the given point is over the frame's resize handle.
    * <p>
    * The method remember the widget and resize direction if the point is over resize handle.
    * 
    * @param    np
    *           The native point to check.
    * 
    * @return   <code>TRUE</code> if the point is inside resize handle,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isInsideFrameResizeHandle(NativePoint np)
   {
      // need a little fix to take into account frame title height
      NativePoint npCheck = new NativePoint(np);
      npCheck.y += ((FrameGuiImpl)parentContainer).getTitleHeight();
      
      // check the frame itself first
      BaseConfig fc = parentContainer.config();
      if (fc.selected && fc.resizable)
      {
         // generate resize handles coordinates for current widget
         NativeRectangle[] frResizeHandles = getFrameResizeHandles(parentContainer);
         // check if we are inside one of them
         for(int i = 0; i < frResizeHandles.length; i++)
         {
            if (frResizeHandles[i].contains(npCheck))
            {
               // found inclusion, return immediately
               // remembering widget to resize and direction
               widgetToResize = parentContainer;
               resizeDirection = i;
               return true;
            }
         }
      }
      
      return false;
   }
   
   /**
    * Checks if the given point is over the browse widget's row resize area.
    * <p>
    * The method remember the widget and resize direction if the point is over resize handle.
    * 
    * @param    event
    *           Mouse event.
    * 
    * @return   <code>TRUE</code> if the point is inside row resize area,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isInsideRowResizeArea(MouseEvent event)
   {
      // first exclude simple cases
      if (isFrame)
      {
         return false;
      }
      
      // here we are within browse column artificial widget implementation
      BrowseColumnGuiImpl col = (BrowseColumnGuiImpl) widgetEff;
      BrowseGuiImpl br = col.getBrowse();
      NativePoint point = col.translate(event);
      BrowseColumnGuiImpl.ResizeArea resizeArea = br.getRowResizeArea(point);
      if (resizeArea != null)
      {
         int rowNum = br.getRowByCoordinates(point.y);
         resizeDirection = RESIZE_DIRECTION_S;
         resizedRowNumber = BrowseColumnGuiImpl.ResizeArea.UPPER.equals(resizeArea) ? rowNum - 1:
                                                                                      rowNum;
         return true;
      }
      else
      {
         resizeDirection = RESIZE_DIRECTION_UNDEF;
         return false;
      }
   }

   /**
    * Checks if inside the row resizable widget.  The browse column is an example. 
    * 
    * @return   <code>TRUE</code> if the point is inside row resizable widget,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isInsideRowResizable()
   {
      return widgetEff instanceof BrowseColumnGuiImpl &&
             ((BrowseColumnGuiImpl) widgetEff).getBrowse().isRowResizable();
   }
   
   /**
    * Selects single widget the current handler is attached to.  All other selected widget inside
    * current container will be de-selected, including container itself.
    */
   private void selectSingleWidget()
   {
      selectSingleWidget(false);
   }
   
   /**
    * Selects single widget the current handler is attached to.  All other selected widget inside
    * current container will be de-selected, including container itself.
    * 
    * @param    repaintParent
    *           If <code>TRUE</code> the parent repaint should be performed after message
    *           processing.
    */
   private void selectSingleWidget(boolean repaintParent)
   {
      // deselect everything selected
      deselectEverything();
      // select single widget
      EventManager.postEvent(new DirectManipulationEvent(widgetEff, parentContainer,
                                                         repaintParent, Keyboard.SE_SELECTION));
   }
   
   /**
    * Deselects all selected widget inside the current container, including container itself.
    */
   private void deselectEverything()
   {
      if (parentContainer != null)
      {
         // first the frame itself
         BaseConfig fc = parentContainer.config();
         if (fc.selectable && fc.selected)
         {
            EventManager.postEvent(new DirectManipulationEvent(parentContainer, parentContainer,
                                                               Keyboard.SE_DESELECTION));
         }
         // then all widgets inside
         Widget<?>[] widLst = ((Frame)parentContainer).getContentPane().widgets();
         for (Widget<?> wid : widLst)
         {
            // skip widgets must be excluded
            if (!wid.isDirectManipulable())
            { 
               continue;
            }
            
            BaseConfig bc = wid.config();
            if (bc.selectable && bc.selected)
            {
               EventManager.postEvent(new DirectManipulationEvent(wid, parentContainer,
                                                                  Keyboard.SE_DESELECTION));
            }
         }
      }
   }
   
   /**
    * Collects the widget resize handles bounds for selected resizable widget.
    * 
    * @param    widResize
    *           The resizable widget to compute handles coordinates.
    * 
    * @return   Array of the bounds as {@link NativeRectangle} elements.
    */
   private NativeRectangle [] getResizeHandles(Widget<?> widResize)
   {
      // get widget position and size
      NativePoint     pos = widResize.physicalLocation();
      NativeDimension dim = widResize.physicalDimension();

      // reconstruct the widget resize handles from top-left clockwise 
      return buildResizeHandles(pos, dim);
   }
   
   /**
    * Collects the widget resize handles bounds for selected resizable widget.
    * 
    * @param    frResize
    *           The resizable frame to compute handles coordinates.
    * 
    * @return   Array of the bounds as {@link NativeRectangle} elements.
    */
   private NativeRectangle [] getFrameResizeHandles(Widget<?> frResize)
   {
      // get the frame resize area
      NativeRectangle resizeArea = ((FrameGuiImpl)frResize).getResizeRectangle();
      
      // get widget position and size
      NativePoint     pos = new NativePoint(resizeArea.left(), resizeArea.top());
      NativeDimension dim = new NativeDimension(resizeArea.width(), resizeArea.height());

      // reconstruct the widget resize handles from top-left clockwise 
      return buildResizeHandles(pos, dim);
   }
   
   /**
    * Builds the widget resize handles bounds for area with given position and size.
    * 
    * @param    pos
    *           The area top-left location to compute handles array.
    * @param    dim
    *           The area dimension to compute handles array.
    * 
    * @return   Array of the bounds as {@link NativeRectangle} elements.
    */
   private NativeRectangle [] buildResizeHandles(NativePoint pos, NativeDimension dim)
   {
      // it is known the total numbers of the resize handles are 8 
      NativeRectangle [] res = new NativeRectangle[8];
      
      // reconstruct the widget resize handles from top-left clockwise 
      int x = pos.x - 2;
      int y = pos.y - 2;
      res[0] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      x = pos.x - 2 + dim.width/2;
      res[1] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      x = pos.x + dim.width - 2;
      res[2] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      y = pos.y - 2 + dim.height/2;
      res[3] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      y = pos.y + dim.height - 2;
      res[4] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      x = pos.x - 2 + dim.width/2;
      res[5] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      x = pos.x - 2;
      res[6] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      y = pos.y - 2 + dim.height/2;
      res[7] = new NativeRectangle(y, x, y + RESIZE_HANDLE_SIZE, x + RESIZE_HANDLE_SIZE);
      
      return res;
   }
   
   /**
    * Collects the widget bounds for selected movable widgets belonging to current container.
    * 
    * @return   Array of the bounds as {@link NativeRectangle} elements.
    */
   private NativeRectangle [] getMovableWidgetsBounds()
   {
      ArrayList<NativeRectangle> res = new ArrayList<NativeRectangle>();
      if (parentContainer != null)
      {
         // check first the frame itself
         BaseConfig fc = parentContainer.config();
         if (fc.selected && fc.movable)
         {
            // add another widget's bounding rectangle
            NativeRectangle widBounds = new NativeRectangle(parentContainer.physicalLocation(),
                                                            parentContainer.physicalDimension());
            res.add(widBounds);
         }
         
         // enumerate all other widgets
         Widget<?>[] widLst = ((Frame)parentContainer).getContentPane().widgets();
         for (Widget<?> wid : widLst)
         {
            // skip widgets must be excluded
            if (!wid.isDirectManipulable())
            { 
               continue;
            }
            
            BaseConfig bc = wid.config();
            if (bc.selected && bc.movable)
            {
               // add another widget's bounding rectangle
               NativeRectangle widBounds = new NativeRectangle(wid.physicalLocation(),
                                                               wid.physicalDimension());
               res.add(widBounds);
            }
         }
      }
      
      return res.toArray(new NativeRectangle[res.size()]);
   }
   
   /**
    * Collects the widgets to be moved, selected movable widgets belonging to current container.
    * 
    * @return   Array of the {@link Widget} elements to be moved.
    */
   private Widget<?> [] getWidgetsToMove()
   {
      ArrayList<Widget<?>> res = new ArrayList<Widget<?>>();
      if (parentContainer != null)
      {
         // check first the frame itself
         BaseConfig fc = parentContainer.config();
         if (fc.selected && fc.movable)
         {
            // add frame widget to move
            res.add(parentContainer);
         }
         
         // enumerate all other widgets
         Widget<?>[] widLst = ((Frame)parentContainer).getContentPane().widgets();
         for (Widget<?> wid : widLst)
         {
            // skip widgets must be excluded
            if (!wid.isDirectManipulable())
            { 
               continue;
            }
            
            BaseConfig bc = wid.config();
            if (bc.selected && bc.movable)
            {
               // add another widget to move
               res.add(wid);
            }
         }
      }
      
      return res.toArray(new Widget<?>[res.size()]);
   }
   
   /**
    * Checks if new mouse pointer is need to be used.  An example is pointer is over the resize
    * handles.
    * 
    * @param    e
    *           The mouse event to check.
    * 
    * @return   New mouse pointer or default one from the set predefined in
    *           {@link MousePtrWrapper}.
    */
   private MousePtrWrapper getNewPointer(MouseEvent e)
   {
      MousePtrWrapper resPtr = MousePtrWrapper.DEFAULT;
      
      // look for mouse to be inside one of the resize handle, frame or widget
      NativePoint resizePoint = ((Frame)parentContainer).getContentPane().translate(e);
      if (isInsideFrameResizeHandle(resizePoint) || isInsideResizeHandle(resizePoint))
      {
         switch (resizeDirection)
         {
            case RESIZE_DIRECTION_NW:
               resPtr = MousePtrWrapper.NW_RESIZE;
               break;
            case RESIZE_DIRECTION_N:
               resPtr = MousePtrWrapper.N_RESIZE;
               break;
            case RESIZE_DIRECTION_NE:
               resPtr = MousePtrWrapper.NE_RESIZE;
               break;
            case RESIZE_DIRECTION_E:
               resPtr = MousePtrWrapper.E_RESIZE;
               break;
            case RESIZE_DIRECTION_SE:
               resPtr = MousePtrWrapper.SE_RESIZE;
               break;
            case RESIZE_DIRECTION_S:
               resPtr = MousePtrWrapper.S_RESIZE;
               break;
            case RESIZE_DIRECTION_SW:
               resPtr = MousePtrWrapper.SW_RESIZE;
               break;
            case RESIZE_DIRECTION_W:
               resPtr = MousePtrWrapper.W_RESIZE;
               break;
         }
      }
      
      return resPtr;
   }

   /**
    * Checks if new mouse pointer for browse row resize is need to be used.
    * 
    * @param    e
    *           The mouse event to check.
    * 
    * @return   New mouse pointer or default one from the set predefined in
    *           {@link MousePtrWrapper}.
    */
   private MousePtrWrapper getRowResizePointer(MouseEvent e)
   {
      return isInsideRowResizeArea(e) ? MousePtrWrapper.S_RESIZE : MousePtrWrapper.DEFAULT;
   }

   /**
    * Makes coordinate alignment according to grid line network position.
    * 
    * @param    coord
    *           The coordinate to process.
    * @param    isVertical
    *           Flag indicating if the given coordinate is in vertical direction or horizontal.
    * 
    * @return   New coordinate value closest to the grid line.
    */
   private int alignToGrid(int coord, boolean isVertical)
   {
      int res = coord;
      PaneConfig pc = parentContainer.config();
      int gridStep = isVertical ? pc.gridUnitHeightPixels : pc.gridUnitWidthPixels;

      // now calculate the closest grid line value
      // the braces order is important here
      res = gridStep * (coord / gridStep);
      int mod = coord % gridStep;
      // if modulo is greater than half of the grid step we need to go to the next grid line
      if (2 * mod >= gridStep)
      {
         res += gridStep;
      }
      
      return res;
   }
   
   /**
    * Checks if the grid snap mode is active.  If yes the position and size of the widgets must
    * be adjusted to the closest grid line.
    * 
    * @return   <code>TRUE</code> if the grid snap is active, <code>FALSE</code> otherwise.
    */
   private boolean isGridSnapOn()
   {
      return parentContainer == null ? false
                                     : !isFrameResizing() &&
                                       ((PaneConfig)parentContainer.config()).gridSnap;
   }
   
   /**
    * Checks if the current mode is widget resizing and widget under resize is a frame.
    * 
    * @return   <code>TRUE</code> if the currently frame widget is resizing,
    *           <code>FALSE</code> otherwise.
    */
   private boolean isFrameResizing()
   {
      return widgetToResize != null &&
             widgetToResize.getId().asInt() == parentContainer.getId().asInt();
   }
}