/*
** Module   : SwingGuiClient.java
** Abstract : GUI-specific OS-level window component for the AWT/Swing client
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 MAG 20140925 Hide constructor.
** 003 CA  20141021 Do not allow GUI windows to be closed via OS (i.e. ALT-F4).
** 004 CA  20141029 Each client needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 005 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 006 CA  20150415 Moved the default title to the WindowGuiImpl class.
** 007 GES 20150504 Changed imports/member names to match package moves and class renames.
**                  Renamed this class to better describe the specific nature.
** 008 CA  20150921 Mouse and focus fixes.
** 009 HC  20151022 A fix for JWindow not able to receive focus when its owner window is not
**                  visible. This resolves ALERT-BOX and DIALOG-BOX not receiving key input when
**                  no main window visible.
**     SBI 20151104 Setup the driver's level keyboard focus manager.
** 010 HC  20160406 An attempt to fix swing windows sometimes not appearing on the taskbar.
** 011 SBI 20160729 Changed to make a window's background transparent, otherwise it is possible
**                  to view the empty background under the changed window.
** 012 OM  20170214 Replaced base class from JFrame to JDialog which suits better our purpose.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.lang.reflect.*;
import javax.swing.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.driver.swing.*;

/** 
 * Main GUI client class which instantiates the primary frame window.
 */
@SuppressWarnings("serial")
public class SwingGuiClient 
extends AbstractClientFrame
{
   /** Window emulator instance. */
   final private SwingEmulatedWindow sew;

   /**
    * Construct the new frame by instantiating the {@link SwingEmulatedWindow} as the
    * content pane and initializing it using the given <code>options</code>.
    * 
    * @param   focusManager
    *          The driver's level keyboard focus manager.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   private SwingGuiClient(KeyboardFocusManager focusManager, SwingKeyboardReader keyReader)
   {
      super(WindowGuiImpl.DEFAULT_TITLE);
      
      KeyboardFocusManager.setCurrentKeyboardFocusManager(focusManager);
      // common initialization
      sew = new SwingEmulatedWindow(this, keyReader);
      init(sew.getContentPane(), false);
   }

   /**
    * Initialize the application frame by creating a new instance of this class
    * and of the content pane which is an instance of {@link SwingEmulatedWindow}.
    * 
    * @param   focusManager
    *          The driver's level keyboard focus manager.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    * @param   parent
    *          The parent components. If {@code null} this is a top-level component equivalent to
    *          a {@code JFrame}, with button in taskbar. If not-null, the modeness of the newly
    *          created window is relative to this parent.
    * @param   popup
    *           
    * @return   The simulator instance.
    */
   public static EmulatedWindowState load(final KeyboardFocusManager focusManager,
                                          final SwingKeyboardReader keyReader,
                                          final Window parent,
                                          final boolean popup) 
   throws InterruptedException,
          InvocationTargetException
   {
      if (parent == null)
      {
         // lame hack because of Java lack of parameter passing by reference (only by value)
         SwingGuiClient[] client = new SwingGuiClient[1];

         // instantiate the application's main frame
         SwingUtilities.invokeAndWait(() ->
         {
            client[0] = new SwingGuiClient(focusManager, keyReader);
         });
         return client[0].sew;
      }
      else
      {
         // lame hack because of Java lack of parameter passing by reference (only by value)
         SwingGuiDialog[] client = new SwingGuiDialog[1];
   
         // instantiate the application's main frame
         SwingUtilities.invokeAndWait(() ->
         {
            client[0] = new SwingGuiDialog(focusManager, keyReader, parent, popup);
         });
         return client[0].sew;
      }
   }

   /**
    * Initialize.  GUI windows are not allowed to be closed via OS (i.e. ALT-F4).
    * 
    * @param   pane
    *          Content pane.
    * @param   resizable
    *          Resizable attribute.
    */
   @Override
   protected void init(JComponent pane, boolean resizable)
   {
      super.init(pane, resizable);
      
      setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
      
      // by default, the simulator is hidden
      // Do not call sew.setWindowVisible(false), this is an attempt to fix the problem
      // with the window sometime not appearing on the taskbar
      pack();
      
      setVisible(false);
   }
   
   static class SwingGuiDialog
   extends AbstractClientDialog
   {
      /** Window emulator instance. */
      final private SwingEmulatedWindow sew;
      
      /**
       * Construct the new frame by instantiating the {@link SwingEmulatedWindow} as the
       * content pane and initializing it using the given {@code options}.
       *
       * @param   focusManager
       *          The driver's level keyboard focus manager.
       * @param   keyReader
       *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
       * @param   parent
       *          The parent components. Cannot be {@code null} .
       * @param   popup
       */
      SwingGuiDialog(KeyboardFocusManager focusManager,
                     SwingKeyboardReader keyReader,
                     Window parent,
                     boolean popup)
      {
         super(WindowGuiImpl.DEFAULT_TITLE, parent, !popup);
         
         if (popup)
         {
            setType(javax.swing.JFrame.Type.POPUP);
         }
         
         KeyboardFocusManager.setCurrentKeyboardFocusManager(focusManager);
         // common initialization
         sew = new SwingEmulatedWindow(this, keyReader);
         init(sew.getContentPane(), false);
      }
      
      /**
       * Initialize.  GUI windows are not allowed to be closed via OS (i.e. ALT-F4).
       *
       * @param   pane
       *          Content pane.
       * @param   resizable
       *          Resizable attribute.
       */
      @Override
      protected void init(JComponent pane, boolean resizable)
      {
         super.init(pane, resizable);
         
         setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
         
         // by default, the simulator is hidden
         pack();
         setVisible(false);
      }
   }
}
