/*
** Module   : DocumentOutputHandler.java
** Abstract : Document output handler.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20171024 Initial version.
** 002 SBI 20180410 Fixed EOF exception, changed document content type and disposition, renamed
**                  PDFPrintOutputHandler to DocumentOutputHandler.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client.gui.driver.web;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.*;

import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.regex.*;


/**
 * Jetty handler for serving document outputs.
 */
public class DocumentOutputHandler
extends AbstractHandler
{
   /**
    * The regular expression that is matched against the request path. The handler will handle only those
    * targets matching this expression.
    */
   private static final Pattern TARGET_REGEXP = Pattern.compile("/document/(.*?)");

   /** The storage of print outputs. */
   private DocumentOutputStorage storage;

   /**
    * Ctor.
    *
    * @param   storage
    *          The storage of print outputs.
    */
   public DocumentOutputHandler(DocumentOutputStorage storage)
   {
      this.storage = storage;
   }

   /**
    * The main logic of the handler.
    *
    * @param   s
    *          The target path.
    * @param   request
    *          The request.
    * @param   httpServletRequest
    *          The original request.
    * @param   httpServletResponse
    *          The response.
    *
    * @throws  IOException
    *          when an IO error occurs
    * @throws  ServletException
    *          when a servlet error occurs
    */
   @Override
   public void handle(String s,
                      Request request,
                      HttpServletRequest httpServletRequest,
                      HttpServletResponse httpServletResponse)
   throws IOException,
          ServletException
   {
      if (request.isHandled())
      {
         return;
      }

      Matcher m = TARGET_REGEXP.matcher(s);
      if (!m.matches())
      {
         return;
      }

      String id = m.group(1);
      DocumentOutput output = storage.dequeueDocumentOutput(id);

      if (output == null)
      {
         return;
      }

      httpServletResponse.setContentType(output.mimeType);
      httpServletResponse.setHeader("Content-disposition", "inline; filename=" + output.fileName);

      try
      {
         ByteArrayOutputStream bout = new ByteArrayOutputStream();
         
         output.outputConsumer.accept(bout);
         
         ServletOutputStream out = httpServletResponse.getOutputStream();
         out.write(bout.toByteArray());
         out.flush();
         
         if (output.finalizer != null)
         {
            output.finalizer.run();
         }
      }
      catch (Exception ex)
      {
         // in case of an error assume the client didn't successfully download the print output and
         // so enqueue for another attempt
         storage.enqueueDocumentOutput(id, output);
         throw ex;
      }
      finally
      {
         request.setHandled(true);
      }
   }
}
