/*
** Module   : GuiWebDriver.java
** Abstract : WebSocket-based GUI driver
**
** Copyright (c) 2013-2018 Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150325 Created initial version patterned off of the ChuiWebDriver.
** 002 CA  20150520 Changed signature for createFont.
** 003 HC  20150612 Compatibility changes related to support of modal windows.
** 004 HC  20150702 Implemented window modality independently on the actual UI driver.
** 005 GES 20150701 Added GUI implementation for previously stubbed-out methods,
**                  including image drawing, font and metrics-related APIs.
** 006 GES 20150818 Reworked z-order implementation and finished remaining stubbed methods.
**                  Added waitInitialization, which waits for a notification from the web page
**                  that it has finished initialized. Added mouse support. deriveFont() is
**                  now driver specific. It scales the font depending on the DPI.
** 007 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt. Z-order
**                  fixes: the full z-order widget list must always be sent if there are changes
**                  in the widget tree (adjusting the new/dead widget list can not provide
**                  correct z-order information). Fixed multi-window drawing issues caused by
**                  concurrent drawing access to the GuiDriver: all access to the driver is done
**                  exclusively, each thread taking and releasing ownership for the duration of
**                  the drawing operation or other GUI API invocation which requires access to
**                  to the underlying physical window.
** 008 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
** 009 SBI 20151108 Changed moveToTop, moveToBottom in order to check if the target window is
**                  visible. Added a mouse event modifier for left, middle and right mouse
**                  buttons.
** 010 SBI 20151124 Changed/added: a widget region is considered to be new if there doesn't exist
**                  in the last widgets cache or its bounded rectangle is changed.
**     SBI 20151201 Changed to provide processMouseWidgets with calculated widgets bounds, added
**                  the special EditorGuiImpl case to consider its dimension as its viewport size
**                  in order to fix a mouse cursor type over its scroll bars.
** 011 HC  20160104 Added the capability to identify active window in gui drivers.
** 012 EVL 20160106 Adding overlay window support.
** 013 EVL 20160122 Change the logic for setting active window.  Overlay window can not be active
**                  from the higher layer activation point of view.  Removing isWindowActive()
**                  method.  Logic is now in WindowManager. 
** 014 IAS 20160130 Renamed registerEditableWidget to registerHoverableWidget
**                  LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 015 EVL 20160301 Small javadoc fix.
** 016 SBI 20151208 Added new parameters, scroll amount and scroll unit for wheel events and
**                  a mouse modifiers mask for all mouse events.
**     SBI 20160321 Fixed an inconsistent state of the web page handlers to process requests by
**                  the dedicated web application handler, implemented new method
**                  isWindowDecorationSupported().
**     CA  20160323 Added parseEditorContent() and getTextWidths().
** 017 HC  20160407 Overhaul of window-activation logic. Removed setOverlayWindowLocation(), the
**                  method was unnecessary and was causing the COMBO-BOX drop-down location
**                  incorrect.
** 018 HC  20160413 Fixed top-level windows to always receive focus event when activated, also
**                  deduplicated processing of focus events.
** 019 SBI 20160418 Added openURL(String path) to open the target page in the browser. Fixed
**                  maximal window's width and height values if they are not set by the server.
**     CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 020 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
**     SBI 20160506 Fixed RuntimeException if window id is -1.
** 021 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 022 EVL 20161028 Fixed the case when visible widget is inside invisible frame.  In this case
**                  the mouse events should not be enabled and/or registered.
** 023 CA  20170104 Added flag to track window's modal state.
** 024 HC  20170202 Fixed potential race condition between GUI and WindowManager.
**     HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
**     CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
**     CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 025 CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
** 026 EVL 20170811 Adding resetSelection() method to clean up web client after selection boxes
**                  drawing completed.
** 027 EVL 20170825 Adding method to initiate file list dropping, saving uploaded file to temp
**                  file system directory.  Adding method to init file size limits.
**     SBI 20170903 Used file channels for uploading large files.
**     EVL 20170905 Moving web client specific code from general drag drop helper here.
** 028 HC  20170819 Implemented runtime support for WINDOW:ALWAYS-ON-TOP and SHOW-IN-TASKBAR.
**                  Added extension widget method GET-MOUSE-POSITION. Made web js client to
**                  notify the java client when desktop (browser window) resized. This makes the
**                  Web GUI driver to report correct desktop dimensions.
** 029 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 030 HC  20171123 Fixed a stack overlow condition in GUI redirected mode.
** 031 CA  20171111 Allow the iframe size to be reported as desktop size in embedded mode.
** 032 HC  20171203 Added notification of Web client when program execution enters modal or wait-for loop.
** 033 CA  20171204 Allow the Web driver to measure text on the client-side, via a AWT font; this
**                  avoids expensive JS calls to measure the text.
** 034 CA  20180201 Performance improvements for registerMouseWidgets().
**     CA  20180208 registerMouseWidgets() is called only once, after control switches from server
**                  side to client-side.
** 035 CA  20180210 Renamed deliverPDFPrintOutput to deliverPrintOutput and added report format 
**                  parameter.
** 036 SBI 20180403 Changed message MSG_OPEN_PRINT_OUTPUT to handle document mime type, added
**                  MSG_OPEN_MIME_RESOURCE.
** 037 HC  20180422 Fixed menu input events sent to wrong widgets in web GUI.
** 038 SBI 20180517 Changed openMimeResource to use a temporary copy of the target resource.
** 039 SBI 20180606 Changed updateCachedWidgetBounds to use Widget.getActualBounds() generally.
** 040 HC  20180525 Fixed mouse pointer loading.
** 041 HC  20180712 Fixed image data not sent to JS causing garbage drawn on the screen.
**     SBI 20180716 Checked for a valid image id to prevent NPE while the window is destroyed.
** 042 SVL 20180718 Added cacheTextMetrics and needsExplicitTextMetricsCaching.
** 043 CA  20180727 Status and message texts can be sent to the remote embedded app.
** 044 SBI 20180731 Implemented getTextWidths in order to use the average string widths.
** 045 HC  20180813 Implemented initial version of dynamic layout.
** 046 SBI 20180815 Changed according to GuiDriver.isFontInstalled.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

/// TODO: remove this, it is there to allow compilation
import java.awt.Component;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.io.*;
import java.net.*;
import java.nio.channels.*;
import java.nio.file.*;
import java.util.function.*;
// don't import the entire logging package as it has naming conflicts with jetty
import java.util.logging.Logger;
import java.util.stream.*;
import java.util.*;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.util.resource.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.Utils.InstrumentedMap;
// don't import the entire package due to LogHelper is ambiguous
import com.goldencode.util.StringHelper;
// don't import the entire logging package as it has naming conflicts with jetty
import com.google.common.io.ByteStreams;

/**
 * GUI driver using WebSockets to communicate with a very thin javascript client hosted in
 * a browser. This class creates and uses an embedded web server to accomplish this.
 */
public class GuiWebDriver
extends AbstractGuiDriver<Integer>
implements EmbeddedWebServer,
           ClientProtocolHooks,
           EmbeddedClient,
           DocumentOutputStorage
{
   /** Taskbar height in pixels when it exists. */
   public static final int TASKBAR_HEIGHT = 32;
   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(GuiWebDriver.class.getName());

   /** Main page root target. */
   private static final String MAIN_TARGET = "/";

   /** Welcome page. */
   private static final String WELCOME_PAGE = "/index.html";
   
   /** Define the file schema to get local resources */
   private static final String FILE_SCHEMA = "file:";
   
   /** A bogus AWT component used to raise mouse events, on notification from the JS code. */
   private static final Component BOGUS_AWT_COMPONENT = new Component()
   {
   };
   
   /** Actions for hoverable widgets */
   private static final Set<Integer> HOVER_ACTIONS =
         Collections.unmodifiableSet(
                  new HashSet<Integer>() {{
                     add(MouseEvent.MOUSE_ENTERED);
                     add(MouseEvent.MOUSE_EXITED);
                     add(MouseEvt.MOUSE_HOVERABLE);
                  }}
               );

   /**
    * Max age of print outputs in milliseconds. When a PDF print output is delivered to the driver it is
    * stored in {@link #documentOutputs} form where it is later served to JS client. Print outputs are removed
    * from the map either upon successful delivery to the JS client or when the age extends over this value.
    */
   private static final long MAX_PRINT_OUTPUT_AGE = 60000;

   /** Temporary files directory. */
   private static String tmpDir = null;
   
   /** Delgated implementation for the embedded web server. */
   private EmbeddedWebServerImpl websrv;
   
   /** Web-Socket which provides a GUI-specific protocol for communication with the browser. */
   private GuiWebSocket websock;
   
   /** Lock object for shared syncrhonization of the web socket. */
   private Object lock = new Object();
   
   /** Desktop bounds. */ 
   private NativeDimension desktop;
   
   /** Work-Area bounds. */ 
   private NativeDimension workarea;
   
   /** Virtual screen for drawing in a memory */
   private VirtualScreen virtualScreen;

   /** Maps the image seal to the unique image identity. */
   private final Map<List<Integer>, Integer > imageSeals;

   /**
    * Image usages is a mapping of an image loaded by a client to its owner windows.
    * The loaded image is represented by its image seal 6-tuple integer array.
    **/
   private final Map<List<Integer>, Set<Integer> > imageUsages;
   
   /** Owned images is a mapping of a window to its owned images seals. */
   private final Map<Integer, Set<List<Integer>> > ownedImages;
   
   /** A cache of font metrics objects, by their remote font ID. */
   private Map<Integer, FontMetricsHelper> fontMetricsCache = new HashMap<>();
   
   /** A counter for font IDs to be mapped on the js-side. */
   private int nextFontId = 0;
   
   /** 
    * A map with the last widgets sent to the JS side in [ windowid : [widgetId : actions] ]
    * form.
    */
   private Map<Integer, Map<Integer, Set<Integer>>> lastMouseWidgets = new HashMap<>();
   
   /** 
    * A map with the last widgets listening for "any mouse action" sent to the JS side in 
    * [ windowid : [widgetId(s) ... ] ] form.
    */
   private Map<Integer, Set<Integer>> lastAnyWidgets = new HashMap<>();
   
   /** Collects the mouse widgets belonging to a window. */
   private Map<Integer, Map<Integer, NativeRectangle>> widgetsRegions = new HashMap<>();

   /** 
    * A queue of pending repaint requests which needs to be processed by the 
    * {@link RepaintThread} thread.
    */
   private LinkedList<GuiWebEmulatedWindow> repaints = new LinkedList<>();
   
   /** The active window. */
   private int activeWindowId = -1;
   
   /** Indicates if the web client embedded mode is on or off. */
   private final boolean embedded;

   /** Flag indicating if the virtual desktop taskbar is shown. */
   private boolean taskbar = true;

   /** Map of document outputs pending to be sent to the Javascript part of the client. */
   private Map<UUID, DocumentOutput> documentOutputs = new HashMap<>();

   /** The notify-user-input state. */
   private volatile boolean needsNotificationOnWait = false;

   /**
    * Flag indicating if {@link #registerMouseWidgets()} must be executed.  This is to minimize
    * the amount of calls, and is performed only when client-side receives control.
    */
   private boolean registerMouseWidgets = true;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public GuiWebDriver(BootstrapConfig config)
   {
      super(new GuiPrimitivesImpl()
      {
         @Override
         public void sync()
         {
            ews.repaint();
         };
      }, config);
      
      this.embedded = config.getBoolean("client", "web", "embedded", false);
      this.imageUsages = new HashMap<List<Integer>, Set<Integer>>();
      this.ownedImages = new HashMap<Integer, Set<List<Integer>>>();
      this.imageSeals  = new HashMap<List<Integer>, Integer>();
      
      new RepaintThread().start();
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      long timeout = -1;
      long wdtimeout = -1;
      int  height = -1;
      int  width  = -1;
      
      this.taskbar = true;
      boolean graphicsCached = true;
      
      if (config != null)
      {
         
         timeout   = config.getInt("client", "web", "socketTimeout", -1);
         wdtimeout = config.getInt("client", "web", "watchdogTimeout", -1);
         height    = config.getInt("client", "gui", "desktopHeight", height);
         width     = config.getInt("client", "gui", "desktopWidth", width);
         taskbar   = config.getBoolean("client", "gui", "taskbar", taskbar);
         graphicsCached = config.getBoolean("client", "gui", "graphicsCached", true);
      }
      
      // provide some meaningful defaults, the actual size will be fetched from the client
      // later on
      if (height == -1 || width == -1)
      {
         height = 768;
         width = 1024;
      }

      updateDesktopDimensions(width, height);

      // calculate our resource base
      Resource cpres = EmbeddedWebServerImpl.calcResourceBase(GuiWebDriver.class, "res");
     
      // calculate our welcome page
      String welcomeUrl = Utils.calculateResourcePath(MAIN_TARGET,
                                                      WebPageHandler.class.getPackage(),
                                                      WELCOME_PAGE);
      
      // create our handlers
      Handler[] hdlrs = new Handler[3];
      
      hdlrs[0] = new GuiWebPageHandler(config, WELCOME_PAGE, welcomeUrl);
      String openResourceTemplate = Utils.calculateResourcePath(
                                                      MAIN_TARGET,
                                                      OpenResourceTemplate.class.getPackage(),
                                                      OpenResourceTemplate.OPEN_RESOURCE_TEMPLATE);
      hdlrs[1] = new OpenResourceHandler(config, openResourceTemplate); 
      hdlrs[2] = new DocumentOutputHandler(this);
      
      // create our web socket
      websock = new GuiWebSocket(lock, this, timeout, wdtimeout, graphicsCached);
      
      // create the embedded server instance
      websrv = new EmbeddedWebServerImpl("GuiWebDriver",
                                         new GenericWebSocketCreator<WebClientProtocol>(websock),
                                         cpres,
                                         hdlrs);
   }

   /**
    * Notify the driver that the client is ready (useful for embedded mode, to notify the embedded
    * app the client can handle messages).
    */
   @Override
   public void clientReady()
   {
      websock.clientReady();

      syncDesktopDimensions();
   }

   /**
    * Called each time a connect occurs, this allows caller-controlled initialization of the
    * remote client.  There can only be 1 connection open at a time, but it is possible for
    * the connection to be opened, closed and then re-opened.
    * <p>
    * Whenever the main page is loaded or re-loaded, the WebSocket will connect and we must
    * put the virtual terminal into a known state.
    */
   @Override
   public void initRemoteClient()
   {
   }

   /**
    * Wait for the page to notify that it has fully initialized its JS code.
    */
   @Override
   public void waitInitialization()
   {
      websrv.waitInitialization(websock);
   }
   
   /**
    * Initializes several web client specific internal variables used with drag-n-drop.
    * 
    * @param    srv
    *           Server to make config related calls.
    */
   public void initDragAndDrop(ServerExports srv)
   {
      // this is used in web client only and only once
      if (tmpDir != null)
      {
         return;
      }

      // setting up temporary files upload directory
      try
      {
         Path tmpFile = Files.createTempFile("tmp-file", ".tmp");
         String absolutePath = tmpFile.toAbsolutePath().toString();
         tmpDir = absolutePath.substring(0, absolutePath.lastIndexOf(File.separator) + 1);
         Files.delete(tmpFile);
      }
      catch (IOException e)
      {
         // for the case the regular way above does not work
         tmpDir = System.getProperty("user.home");
      }
      // create separate directory for fwd files
      tmpDir = tmpDir.concat("fwd");
      File tmpDirObj = new File(tmpDir);
      if (!tmpDirObj.exists())
      {
         tmpDirObj.mkdir();
      }
      // adding file separator to the end
      tmpDir = tmpDir.concat(File.separator);
      
      // getting the file size limit values considering overrides in server's directory
      int [] uploadLimits = srv.getUploadFileSizeLimits();
      // set up gui driver
      if (uploadLimits[0] > 0 && uploadLimits[1] > uploadLimits[0])
      {
         setUploadFileSizeLimits(uploadLimits[0], uploadLimits[1]);
      }
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    request
    *           The request data incoming from 4GL.
    * 
    * @return   The response from the customer-specific application, which will be sent back to 
    *           the converted 4GL code.
    */
   @Override
   public UnformattedPayload remoteCall(UnformattedPayload request)
   {
      return websock.remoteCall(request);
   }
   
   /**
    * Send a PUBLISH notification to a remote application. 
    * 
    * @param    subId
    *           The subscriber's ID, known by the remote application.
    * @param    event
    *           The published event name.
    * @param    param
    *           The parameters for this event.
    */
   @Override
   public void publish(long subId, String event, String[] param)
   {
      websock.publish(subId, event, param);
   }
   
   /**
    * Set status text message of specified window.
    * 
    * @param   windowId
    *          Window ID.
    * @param   message
    *          Text message.
    */
   @Override
   public void setStatusText(int windowId, String message)
   {
      if (embedded)
      {
         websock.setWindowStatusText(windowId, message);
      }
   }
   
   /**
    * Add message texts of specified window.
    * 
    * @param   windowId
    *          Window ID.
    * @param   message
    *          Text message.
    */
   @Override
   public void setMessageText(int windowId, String[] message)
   {
      if (embedded)
      {
         websock.addWindowMessageText(windowId, message);
      }
   }
   
   /**
    * Clear all message texts of specified window.
    * 
    * @param   windowId
    *          Window ID.
    */
   @Override
   public void clearMessages(int windowId)
   {
      if (embedded)
      {
         websock.clearWindowMessages(windowId);
      }
   }

   /**
    * Input key events when in VT100 mode.
    * 
    * @param    keyCode
    *           The key code from the input event.
    * @param    charCode
    *           The character code from the input event.
    */
   @Override
   public void injectVT100Key(int keyCode, char charCode)
   {
   }
   
   /**
    * Since web interface is inherently a single-display setup the method 
    * effectively returns a rectangle with top-level corner equal to 0,0.
    * 
    * @return   The physical screen bounds.
    */
   public NativeRectangle getDisplayBounds()
   {
      NativeDimension desktop;
      synchronized (lock)
      {
         desktop = this.desktop;
      }
      return new NativeRectangle(new NativePoint(0, 0), desktop);
   }
   
   /**
    * Returns its virtual screen that represents the physical screen in a memory
    * using a sRGB pixel model.
    * 
    * @return The virtual screen to draw images in a memory.
    */
   public VirtualScreen getVirtualScreen()
   {
      synchronized (lock)
      {
         return virtualScreen;
      }
   }
   
   /**
    * Register a usage of the given image by the given window.
    * 
    * @param    imageHash
    *           The image hash id.
    * @param    windowId
    *           The window id.
    * 
    * @return   True iff it is the first usage of the given image.
    */
   public boolean addImageUsage(List<Integer> imageHash, int windowId)
   {
      boolean newLoadedImage = false;
      
      Set<Integer> usages = imageUsages.get(imageHash);

      newLoadedImage = usages == null || !usages.contains(windowId);

      if (usages == null)
      {
         usages = new HashSet<Integer>();
         imageUsages.put(imageHash, usages);
      }
      usages.add(windowId);
      
      Set<List<Integer>> images = ownedImages.get(windowId);
      if (images == null)
      {
         images = new HashSet<List<Integer>>();
         ownedImages.put(windowId, images);
      }
      images.add(imageHash);
      
      return newLoadedImage;
   }
   
   /**
    * Returns all unused images. Removes the given window from all image usages.
    * 
    * @param    windowId
    *           The window id.
    * 
    * @return The set of all unused images.
    */
   public Set<Integer> removeUnusedImages(int windowId)
   {
      Set<Integer> imagesToRemove = new HashSet<Integer>();
      Set<List<Integer>> images = ownedImages.remove(windowId);
      if (images != null && !images.isEmpty())
      {
         for (List<Integer> imageSeal : images)
         {
            Set<Integer> usages = imageUsages.get(imageSeal);
            if (usages.remove(windowId))
            {
               if (usages.isEmpty())
               {
                  Integer imageUniqueId = imageSeals.remove(imageSeal);
                  
                  if (imageUniqueId != null)
                  {
                     imagesToRemove.add(imageUniqueId);
                  }
                  
                  imageUsages.remove(imageSeal);
               }
            }
         }
      }
      
      return imagesToRemove;
   }
   
   /**
    * Maps the image identity array to the unique image id.
    * 
    * @param    imageSeal
    *           The image identity array.
    * @param    uniqueId
    *           The image unique id.
    */
   public void mapSealToUniqueId(List<Integer> imageSeal, Integer uniqueId)
   {
      imageSeals.put(imageSeal, uniqueId);
   }
   
   /**
    * Gets the unique image id by its identity array.
    * 
    * @param    imageSeal
    *           The image identity array.
    * 
    * @return   The image unique id.
    */
   public Integer getUniqueIdForSeal(List<Integer> imageSeal)
   {
      return imageSeals.get(imageSeal);
   }
   
   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area and it is always at the bottom of the
    * virtual screen (when present).
    * 
    * @return   The area in pixels.
    */
   public NativeRectangle getDisplayWorkArea()
   {
      synchronized (lock)
      {
         return new NativeRectangle(new NativePoint(0, 0), this.workarea);
      }
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    */
   @Override
   public void beep()
   {
      if (websock != null)
      {
         websock.beep();
      }
   }

   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return   The pressed key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()                      
   {
      return websock.readKey();
   }

   /**
    * Create a new window emulator.
    * <p>
    * The method ignores the alwaysOnTop flag.
    * 
    * @param    windowId
    *           The window ID.
    * @param    showInTaskbar
    *           If set to {@code true} the window will be shown in the virtual desktop's taskbar.
    *
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newEmulator(int windowId,
                                          boolean showInTaskbar)
   {
      GuiWebEmulatedWindow ews = new GuiWebEmulatedWindow(lock, websock, this);
         
      // the real window size is unknown at this time; it will be set later, when the layout
      // is performed
      ews.setScreenBitmap(new ScreenBitmap(0, 0));
      
      // remote setup
      websock.createWindow(windowId, showInTaskbar);
      
      return ews;
   }
   
   /**
    * Create a new child window emulator.
    * 
    * @param    windowId
    *           The window ID.
    * @param    parentWindow
    *           The parent window.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *           
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newChildWindowEmulator(
                               int                 windowId,
                               EmulatedWindowState parentWindow, 
                               String              windowTitle,
                               boolean             resizable,
                               boolean             isShareActivationWithOwner,
                               boolean             modal)
   {
      // TODO: we may need to remember the window type (child vs top-level) in the GWEW
      GuiWebEmulatedWindow ews = new GuiWebEmulatedWindow(lock, websock, this);
         
      // the real window size is unknown at this time; it will be set later, when the layout
      // is performed
      ews.setScreenBitmap(new ScreenBitmap(0, 0));
      
      // remote setup
      websock.createChildWindow(windowId, parentWindow.getWindowId(), windowTitle, resizable, 
                                isShareActivationWithOwner, modal);
      
      return ews;
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    details
    *           The font specifications.
    * 
    * @return   A font instance.
    */
   @Override
   public FontDetails<Integer> createFont(FontDetails<Integer> details)
   {
      FontDetails<Integer> res = (details.fontAlias == null ? details : details.fontAlias);

      FontHelper.add(details.key, FontHelper.createFont(res));
      
      // get the ID for this font
      res.font = ++nextFontId;
      
      int jsFont = websock.createFont(res.font,
                                      res.fontName,
                                      res.pointSize,
                                      res.style.toInt(),
                                      res.boldFontDefinition,
                                      res.fontDefinition);
      
      res = details.copy();
      res.font = jsFont;
      res.pointSize = details.pointSize;
      res.fontName = details.fontName;
      
      return res;
   }
   
   /**
    * Scale the given font to match the screen resolution.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    */
   @Override
   public void deriveFont(FontDetails<Integer> fd)
   {
      FontHelper.add(fd.key, 
                     FontHelper.deriveFont(FontHelper.get(fd.key), fd, getScreenResolution()));
      
      // this is done in place, as is expected for deriveFont to be called always after createFont
      // is called
      websock.deriveFont(fd.font);
   }

   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   @Override
   public int getScreenResolution()
   {
      // web browsers use a "CSS pixel" which is defined as 96 DPI; this may differ from the
      // actual number of pixels on the device; in other words the CSS pixel is device
      // independent and is the basis for everything we do; the only output that may not
      // be handled as well in such an environment is bitmap rendering, which may not scale
      // well to the device-specific resolution
      return 96;
   }
   
   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToClipboard(String txt)
   {
      websock.writeClipboard(txt);
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   The clipboard contents.
    */
   @Override
   public String clipboardContents()
   {
      return websock.readClipboard();
   }
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void restackWindows(int[] windowIds)
   {
      websock.restackWindows(windowIds);
   }
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window 
    *           and its state changed to active. 
    */
   @Override
   public void moveToTop(int windowId, boolean focus)
   {
      EmulatedWindowState<?, ?> window = direct.getWindowEmulator(windowId);
      if (!window.isWindowVisible())
      {
         return;
      }
      websock.moveToTop(windowId, focus);
      
      if (focus)
      {
         // activate the window and notify the upper layers;
         // do not wait for the JS layer to notify us back as
         // the focus request always succeeds
         windowActivated(activeWindowId, windowId);
      }
   }
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    */
   @Override
   public void moveToBottom(int windowId)
   {
      EmulatedWindowState window = direct.getWindowEmulator(windowId);
      if (!window.isWindowVisible())
      {
         return;
      }
      websock.moveToBottom(windowId);
   }
   
   /**
    * Perform driver shutdown.
    */
   @Override
   public void shutdown()
   {
      // close the emulated windows first
      super.shutdown();
      
      // notify the remote client about application exit    
      websock.quit();
      
      // since the application will exit shutdown the server
      shutdownServer();
   }
   
   /**
    * Startup the embedded server using the server's key store to build the web server SSL
    * context.
    * <p>
    * There are two page handlers.  One is for the root ("/") target and the other is for
    * the index ("/index.html") target.  Both load the same page index.html.
    * <p>
    * When the user is first redirected the URL is <code>https://server:port/</code> and the
    * target is "/". The index.html page is loaded for this target.
    * <p>
    * After redirection and the main page is loaded, the URL in the browser address bar will
    * become <code>https://server:port/index.html</code> and the corresponding target is now
    * "/index.html".
    * <p>
    * When the user reloads the main page within the browser environment the "/index.html"
    * target is sent to the web server and the same main page index.html is loaded. 
    * 
    * @param    keyStore
    *           The exported server KeyStore. 
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      return websrv.startupServer(keyStore, config, host, port);
   }
   
   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      websrv.shutdownServer();
   }
   
   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      return websrv.getServerURI();
   }
   
   /**
    * Enable or disable capture of mouse events.
    * <p>
    * It also notifies the JS side to enable/disable mouse events.
    *  
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      boolean prev = super.captureMouseEvents(state);
      
      websock.captureMouseEvents(state);
      
      return prev;
   }
   
   /**
    * Raise a mouse event.
    * 
    * @param    windowId
    *           The window ID.
    * @param    mouseOp
    *           The mouse event code.
    * @param    tstamp
    *           The event timestamp.
    * @param    mouseX
    *           The absolute mouse X for this event.
    * @param    mouseY
    *           The absolute mouse Y for this event.
    * @param    button
    *           The button which was pressed.
    * @param    clickCount
    *           The click count.
    * @param    modifiersMask
    *           The bit mask that represents the modifiers keys are pressed. The 0-bit corresponds
    *           to the shift key, the 1-bit to the ctrl key, the 2-bit to the meta key and
    *           the 3-bit to the alt key.
    * @param    wheelParameters
    *           The array that represents the wheel rotation amount and its unit.
    * @param    widgetId
    *           The explicit widget ID to which this event needs to be posted.
    */
   @Override
   public void raiseMouseEvent(int   windowId,
                               int   mouseOp, 
                               long  tstamp, 
                               int   mouseX, 
                               int   mouseY, 
                               int   button,
                               int   clickCount, 
                               int   modifiersMask,
                               int[] wheelParameters,
                               int   widgetId)
   {
      if (!direct.hasWindowEmulator(windowId))
      {
         return;
      }

      MouseOps mop = MouseOps.fromCode(mouseOp);
      
      int modifiers = getMouseModifiers(button, modifiersMask);
      EmulatedWindowState window = direct.getWindowEmulator(windowId);
      NativePoint wndLoc = window.getWindowLocation();

      MouseEvent event;
      if (mop.isWheel())
      {
         // the scroll unit
         int unit = MouseWheelEvent.WHEEL_UNIT_SCROLL;
         // the scroll direction, 1 means down or right and -1 means up or left 
         int direction = 1;
         // the scroll amount in the scroll unit
         int scrollAmount = 0;
         if (wheelParameters != null && wheelParameters.length == 2)
         {
            direction    = (wheelParameters[0] > 0) ? 1 : -1;
            scrollAmount = Math.abs(wheelParameters[0]);
            if (wheelParameters[1] == 2)
            {
               unit = MouseWheelEvent.WHEEL_BLOCK_SCROLL;
            }
         }
         event = new MouseWheelEvent(BOGUS_AWT_COMPONENT, 
                                     mop.awtEventCode, 
                                     tstamp, 
                                     modifiers, 
                                     mouseX, 
                                     mouseY, 
                                     mouseX + wndLoc.x,
                                     mouseY + wndLoc.y,
                                     clickCount, 
                                     false, 
                                     unit, 
                                     scrollAmount, 
                                     direction);
      }
      else
      {
         event = new MouseEvent(BOGUS_AWT_COMPONENT, 
                                mop.awtEventCode, 
                                tstamp, 
                                modifiers, 
                                mouseX, 
                                mouseY, 
                                mouseX + wndLoc.x,
                                mouseY + wndLoc.y,
                                clickCount, 
                                false,
                                button);
      }
      

      ((WebMouseHandler) window.getMouseHandler()).raiseMouseEvent(event, windowId, widgetId);
   }

   /**
    * Initiates the process of the file list dropping for web clients.
    * 
    * @param    windowId
    *           The window ID.
    * @param    widgetId
    *           The widget ID.
    * @param    fileNameList
    *           The oredered file name list to be dropped.
    */
   @Override
   public void startDropTarget(int windowId, int widgetId, List<String> fileNameList)
   {
      List<File> fileList = new ArrayList<File>();
      // change the file name
      for (String fn : fileNameList)
      {
         fileList.add(new File(tmpDir.concat(fn)));
      }
      // call web window emulator
      EmulatedWindowState window = direct.getWindowEmulator(windowId);
      if (window != null)
      {
         ((GuiWebEmulatedWindow)window).startDropTarget(widgetId, fileList);
      }
   }
   
   /**
    * Stores the given byte array date as file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create.
    * @param    data
    *           The data to store in the file.
    * @param    size
    *           The size of the to create, used to check the input array.
    */
   @Override
   public void saveDropTargetFile(String name, byte [] data, int size)
   {
      // route to drag drop helper to save the file
      DragDropHelper.saveFile(tmpDir.concat(name), data, size);
   }
   
   /**
    * Reads the data from the given file channel starting from its current position and stores
    * the retrieved data into the file in Java temporary directory with the given name.
    * 
    * @param    name
    *           The file name to create
    * @param    channel
    *           The file channel
    * @param    length
    *           The length of the content
    */
   @Override
   public void saveDropTargetFile(String name, FileChannel channel, int length)
   {
      DragDropHelper.readDataAndSaveAsFile(tmpDir.concat(name), channel, length);
   }
   
   /**
    * Gets the mouse modifiers mask for this event.
    * 
    * @param    button
    *           The mouse button that is currently pressed
    * @param    modifiersMask
    *           The bit mask that represents the modifiers keys are pressed. The 0-bit corresponds
    *           to the shift key, the 1-bit to the ctrl key, the 2-bit to the meta key and
    *           the 3-bit to the alt key.
    * 
    * @return   The key's modifiers bit mask for this mouse event.
    */
   private int getMouseModifiers(int button, int modifiersMask)
   {
      int modifiers = 0;
      // button codes from JS (same as from AWT) 
      // 0 : No button
      // 1 : Left mouse button
      // 2 : Wheel button or middle button (if present)
      // 3 : Right mouse button
      switch (button)
      {
         case 1:
            modifiers = InputEvent.BUTTON1_MASK;
            break;
         case 2:
            modifiers = InputEvent.BUTTON2_MASK;
            break;
         case 3:
            modifiers = InputEvent.BUTTON3_MASK;
            break;
         default:
      }
      if ((modifiersMask & MouseEvent.SHIFT_MASK) != 0)
      {
         modifiers |= MouseEvent.SHIFT_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.CTRL_MASK) != 0)
      {
         modifiers |= MouseEvent.CTRL_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.META_MASK) != 0)
      {
         modifiers |= MouseEvent.META_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.ALT_MASK) != 0)
      {
         modifiers |= MouseEvent.ALT_DOWN_MASK; 
      }
      
      return modifiers;
   }
   
   /**
    * Set the location for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    x
    *           The X coordinate of the top-left corner.
    * @param    y
    *           The Y coordinate of the top-left corner.
    */
   @Override
   public void setWindowLocation(int windowId, int x, int y)
   {
      TopLevelWindow<?> w = WindowManager.findWindow(windowId);
      ThinClient.getInstance().postOSEvent(new WindowMoved(w, x, y));
   }

   /**
    * Set the dimension for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    width
    *           The window width.
    * @param    height
    *           The window height.
    */
   @Override
   public void setWindowDimension(int windowId, int width, int height)
   {
      synchronized (this)
      {
         // invalidate all widget locations
         lastMouseWidgets.clear();
         lastAnyWidgets.clear();
      }

      TopLevelWindow<?> w = WindowManager.findWindow(windowId);
      ThinClient.getInstance().postOSEvent(new WindowResized(w, width, height));
   }
   
   /**
    * Raise an event indicating that the target window has been activated or deactivated.
    * 
    * @param    deactivatingWindowId
    *           Id of the window being deactivated, may be negative.
    * @param    activatingWindowId
    *           Id of the window being activated, always valid.
    */
   @Override
   public void windowActivated(int deactivatingWindowId, int activatingWindowId)
   {
      if (activatingWindowId == activeWindowId)
      {
         return;
      }
      
      activeWindowId = activatingWindowId;
      if (deactivatingWindowId != -1)
      {
         // this always executes on the main thread no need to synchronize through
         // ThinClient.invokeLaterOS()
         WindowManager.windowDeactivated(deactivatingWindowId, false);
      }
      if (activatingWindowId != -1)
      {
         // this always executes on the main thread no need to synchronize through
         // ThinClient.invokeLaterOS()
         WindowManager.windowActivated(activatingWindowId, false);
      }
   }

   /**
    * Raise an event as the window was (de)iconified.
    * 
    * @param    windowId
    *           The window ID.
    * @param    state
    *           <code>true</code> if the window was iconified.
    */
   @Override
   public void windowIconified(int windowId, boolean state)
   {
      ThinClient.getInstance().invokeLaterOS(() ->
      {
         WindowGuiImpl wnd = WindowManager.findWindow(windowId);
         if (wnd != null)
         {
            if (state)
            {
               wnd.iconify(true);
            }
            else
            {
               ThinClient.getInstance().eventBracket(() -> wnd.restore(true));
            }
         }
      });
   }

   /**
    * Called when the desktop on the client has changed its dimensions.
    *
    * @param    width
    *           New desktop width.
    * @param    height
    *           New desktop height.
    */
   @Override
   public void desktopResized(int width, int height)
   {
      updateDesktopDimensions(width, height);
   }

   /**
    * Notify the driver which side has control (client-side (when <code>true</code>) or server,
    * otherwise).
    * <p>
    * When control is received, {@link #registerMouseWidgets()} is enabled.
    *  
    * @param    control
    *           Flag indicating (when <code>true</code>) that the client-side has control.
    */
   @Override
   public void setControl(boolean control)
   {
      super.setControl(control);
      
      this.registerMouseWidgets = control;
   }
   
   /**
    * Register all the mouse-aware widgets (in all Windows) with the JS side.
    * <p>
    * To register a widget, it must be visible, displayed and enabled, and part of a visible
    * window.
    * <p>
    * This method is executed only when {@link #registerMouseWidgets} flag is set 
    * (via {@link #setControl}).  Once executed, it disables this flag.
    */
   @Override
   public synchronized void registerMouseWidgets()
   {
      if (!registerMouseWidgets)
      {
         return;
      }
      
      this.registerMouseWidgets = false;
      
      // track widgets registered for complex mouse actions
      InstrumentedMap<Integer, Map<Integer, Set<Integer>>> allWidgets = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      InstrumentedMap<Integer, Map<Integer, Set<Integer>>> newWidgets = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      // track widgets bounds
      InstrumentedMap<Integer, Map<Integer, NativeRectangle>> newWidgetsBounds = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      // track widgets registered for "any mouse action"
      InstrumentedMap<Integer, Set<Integer>> newAnyWidgets = new InstrumentedMap<>(TreeSet::new);
      InstrumentedMap<Integer, Set<Integer>> allAnyWidgets = new InstrumentedMap<>(TreeSet::new);
      // track existed widgets: mouse widgets and any widgets
      Map<Integer,List<Integer>> existingMouseWidgets = new HashMap<>();
      Map<Integer,List<Integer>> existingAnyWidgets = new HashMap<>();
      NativeRectangle[] actualBounds = new NativeRectangle[1];
      Integer[] collectorWindowId = new Integer[1];
      
      Consumer<Widget> actionsCollector = widget ->
      {
         int widgetId = widget.getId().asInt();
         Set<Integer> webActions = getWebActions(widget);
         boolean hasAnyAction = MouseHandler.hasAnyAction(widgetId);
         Integer windowId = collectorWindowId[0];

         if (!webActions.isEmpty())
         {
            allWidgets.modifyValue(windowId, actions -> actions.put(widgetId, webActions));
            
            // determine if the widget was previously registered with the web side
            // or it changed its bounds.
            boolean isExisted = isInLastMouseWidgets(widgetId, windowId);
            if (isExisted)
            {
               existingMouseWidgets.get(windowId).add(widgetId);
            }
            boolean isSizeChanged = updateCachedWidgetBounds(windowId, widget, actualBounds);
            if (!isExisted || isSizeChanged)
            {
               // new widget, track it
               newWidgets.modifyValue(windowId, actions -> actions.put(widgetId, webActions));

               // track its bounds
               newWidgetsBounds.modifyValue(windowId,
                                            bounds -> bounds.put(widgetId, actualBounds[0]));
            }
         }

         if (hasAnyAction)
         {
            allAnyWidgets.modifyValue(windowId, actions -> actions.add(widgetId));
            
            if (!isInLastAnyWidgets(widgetId, windowId))
            {
               newAnyWidgets.modifyValue(windowId,  actions -> actions.add(widgetId));
            }
            else
            {
               // added the given any widget to the existed any widget map
               existingAnyWidgets.get(windowId).add(widgetId);
            }
         }
      };
      
      // go through all windows which are active/shown/etc and collect widgets,
      // title bars, message area widgets
      // collect all enabled, visible widgets.  the title bar widgets
      // are collected regardless if they are enabled or not
      // the visible widget inside invisible frame is considering as invisible
      // and is not including into result
     
      Predicate<Widget> isVisibleNonVirtual = widget -> 
                                                   widget.isVisible()                       &&
                                                   UiUtils.getWidgetId(widget) != null      &&
                                                   isInsideVisibleFrameOrStandalone(widget) &&
                                                   widget.isDisplayed()                     &&
                                                   !widget.isTotallyObscured();
      Predicate<Widget> enabledWidget = Widget::isEnabled;
      
      WindowManager.windowList().stream().filter(window -> 
                                                      window.isVisible()   &&
                                                      window.isDisplayed() &&
                                                      window instanceof TopLevelWindow
                                          ).
            forEach(
               window -> 
                  {
                     Integer windowId = window.getId().asInt();
                     collectorWindowId[0] = windowId;
                     
                     // the window check is not needed here, as we are processing this window's
                     // widgets only
                     Predicate<Widget> isValidChild = isVisibleNonVirtual;
                           
                     existingMouseWidgets.put(windowId, new ArrayList<>());
                     existingAnyWidgets.put(windowId, new ArrayList<>());

                     // window border is managed by the JS side
                     if (GuiWindow.class.isInstance(window))
                     {
                        GuiWindow gw = (GuiWindow) window;
                        collectWidgets(gw.getTitleBar(), isValidChild, actionsCollector);
                        
                        collectWidgets((MessageAreaGuiImpl) gw.getMessageArea(), 
                                       isValidChild,
                                       actionsCollector);
                     }
                     
                     collectWidgets(window, enabledWidget.and(isValidChild), actionsCollector);
                  }
      );

      // remove existing mouse widgets from the last mouse widgets map
      //    to determine dead mouse widgets.
      existingMouseWidgets.keySet().stream().forEach(
               windowId -> lastMouseWidgets.getOrDefault(windowId, Collections.emptyMap()).
                  keySet().
                  removeAll(existingMouseWidgets.getOrDefault(windowId, Collections.EMPTY_LIST))
            );
      // remove existing any widgets from the last any widgets map
      //    to determine dead any widgets.
      existingAnyWidgets.keySet().stream().forEach(
            windowId -> lastMouseWidgets.getOrDefault(windowId, Collections.emptyMap()).
               keySet().
               removeAll(existingMouseWidgets.getOrDefault(windowId, Collections.EMPTY_LIST))
         );
      // remove all empty window maps from dead widgets
      Utils.removeAllIf(lastMouseWidgets.keySet(), wid -> lastMouseWidgets.get(wid).isEmpty());
      
      // register the new widgets and delete the widgets which are no longer in the set
      websock.processMouseWidgets(allWidgets, newWidgets, newWidgetsBounds, lastMouseWidgets,
                                  newAnyWidgets, lastAnyWidgets);
      
      // save the entire set
      lastMouseWidgets.clear();
      lastMouseWidgets.putAll(allWidgets);
      
      lastAnyWidgets.clear();
      lastAnyWidgets.putAll(allAnyWidgets);
   }

   /**
    * Split the text for a editor widget, so that it fits the specified maximum width.  The 
    * <code>delta</code> parameter represents the start position of this text, in the editor's 
    * content; when the full content is parsed, this will be zero.
    *  
    * @param    widgetId
    *           The editor's ID.
    * @param    windowId
    *           The editor's window ID.
    * @param    maxWidth
    *           The width for the content to fit (the editor's editable area).
    * @param    text
    *           The text to split into lines.
    * @param    delta
    *           The start position of this text in the editor's full content.
    * @param    font
    *           The editor's font.
    *           
    * @return   Two arrays: one with the split indexes and one with the split indexes which were
    *           determined to be caused by a new line.
    */
   @Override
   public int[][] parseEditorContent(int                  widgetId,
                                     int                  windowId,
                                     int                  maxWidth,
                                     String               text,
                                     int                  delta,
                                     FontDetails<Integer> font)
   {
      return websock.parseEditorContent(widgetId,
                                        windowId,
                                        maxWidth,
                                        text,
                                        delta,
                                        font.font);
   }
   
   /**
    * Reset selection boxes painting mode and return to normal drawing mode.
    */
   @Override
   public void resetSelection()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESET_SELECTION);

      ews.offer(ps);
   }
   
   /**
    * Setting up file size limits for cases when drag and drop needs file uploads from web client
    * to Java side.
    * 
    * @param    single
    *           Single file size limit.
    * @param    total
    *           Total file size limit for single operation. 
    */
   @Override
   public void setUploadFileSizeLimits(int single, int total)
   {
      websock.setUploadFileSizeLimits(single, total);
   }
   
   /**
    * Updates the cached widgets bounds and returns true iff the cached widgets bounds has been
    * updated.
    * 
    * @param    windowId
    *           The target window id.
    * @param    widget
    *           The target widget with its cached bounds.
    * @param    widgetBounds
    *           On index 0, the computed bounds will be saved.
    *            
    * @return   true value indicates that the widget's bounds has been changed and must be sent
    *           to the JS client.
    */
   private boolean updateCachedWidgetBounds(int               windowId,
                                            Widget            widget, 
                                            NativeRectangle[] widgetBounds)
   {
      boolean updated = false;
      Integer widgetId = widget.getId().asInt();
      NativeRectangle actualBounds = widget.getActualBounds();
      Map<Integer, NativeRectangle> cachedRegions = widgetsRegions.get(windowId);

      if (cachedRegions == null)
      {
         widgetsRegions.put(windowId, cachedRegions = new LinkedHashMap<>());
         updated = true;
      }
      else
      {
         NativeRectangle bounds = cachedRegions.get(widgetId);
         if (bounds == null)
         {
            updated = true;
         }
         else
         {
            updated = !bounds.equals(actualBounds);
         }
      }

      if (updated)
      {
         cachedRegions.put(widgetId, actualBounds);
      }

      widgetBounds[0] = actualBounds;
      return updated;
   }

   /**
    * Returns the set of mouse actions handled by this widget.
    * 
    * @param    widget
    *           The collected mouse widget.
    * @return   The web actions that are handled by the given widget.
    */
   private Set<Integer> getWebActions(Widget widget)
   {
      Set<Integer> mouseActions = new TreeSet<>();
      mouseActions.addAll(widget.getMouseActions());
      mouseActions.addAll(MouseHandler.getWidgetActions(widget.getId().asInt()));
      if (widget.supportsCustomMousePointer())
      {
         mouseActions.addAll(HOVER_ACTIONS);
      }
      
      return Arrays.stream(MouseOps.values()).
                     filter(op -> mouseActions.contains(op.awtEventCode)).
                     map(op -> op.code).collect(Collectors.toSet());
   }
   
   /**
    * Tests if the given mouse widget is in the last mouse widgets map.
    * 
    * @param    widgetId
    *           The widget id.
    * @param    windowId
    *           The window id.
    * 
    * @return   Returns true iff the given widget is in the target map.
    */
   private boolean isInLastMouseWidgets(Integer widgetId, Integer windowId)
   {
      // determine if the widget was previously registered with the web side
      return lastMouseWidgets.getOrDefault(windowId,
                                           Collections.emptyMap()).containsKey(widgetId);
   }
   
   /**
    * Tests if the given any mouse widget is in the last any mouse widgets map.
    * 
    * @param    widgetId
    *           The widget id.
    * @param    windowId
    *           The window id.
    * 
    * @return   Returns true iff the given widget is in the target map.
    */
   private boolean isInLastAnyWidgets(Integer widgetId, Integer windowId)
   {
      return lastAnyWidgets.getOrDefault(windowId, Collections.emptySet()).contains(widgetId);
   }
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    *  
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerMoveableWidget(int wid)
   {
      // no-op
   }

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    * 
    * @param    action
    *           The widget mouse action.
    */
   @Override
   public void registerHoverableWidget(MouseHoverAction action)
   {
      Widget<?> widget = action.widget();
      Widget<?> parent = widget.enclosingFrame().orElse(null);

      if (parent == null && (widget instanceof Frame))
      {
         parent = ((Frame)widget).window();
      }

      websock.registerHoverableWidget(
         widget.topLevelWindow().getId().asInt(), 
         widget.getId().asInt(), 
         parent != null ? parent.getId().asInt() : -1, 
         widget.getActualBounds(),
         widget.getLoadedMousePointer().orElse(null),
         widget.defaultMousePtr().orElse(MousePtrWrapper.DEFAULT),
         widget.enforceDefaultMousePtr()
      );
   }

   /**
    * Propagate wait state
    * 
    * @param isWaitState
    *        wait state
    */
   public void propagateWaitState(boolean isWaitState)
   {
      websock.propagateWaitState(isWaitState);
   }

   /**
    * Register the specified widget as a resizeable widget, via mouse.  Only WINDOW widgets are
    * supported.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerResizeableWidget(int wid)
   {
      // no-op
   }

   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      super.deregisterWidget(wid);
      websock.deregisterWidget(ews.getWindowId(), wid);
   }

   /**
    * Set the window details for resize; this includes the minimum and maximun dimensions, and 
    * also the window's resizeable state.
    * 
    * @param    resize
    *           Flag indicating if the window can be resized or not.
    * @param    minWidth
    *           The window's minimum width.
    * @param    minHeight
    *           The window's minimum height.
    * @param    maxWidth
    *           The window's maximum width.
    * @param    maxHeight
    *           The window's maximum height.
    */
   @Override
   public void setResizeableWindow(boolean resize,
                                   int     minWidth, 
                                   int     minHeight, 
                                   int     maxWidth, 
                                   int     maxHeight)
   {
      // only windows are resized
      Window<?> window = WindowManager.findWindow(ews.getWindowId());
      
      if (window == null)
      {
         return;
      }
      
      WindowConfig wcfg = window.config();
      // get actual size
      NativeDimension size = window.physicalDimension();
      
      // if the actual size is greater than the configuration size, then let the maximal size be
      // equal to its actual size.
      int maxWidthPixels  = Math.max(ConfigHelper.getMaxWidthPixels(wcfg), size.width);
      
      int maxHeightPixels = Math.max(ConfigHelper.getMaxHeightPixels(wcfg), size.height);
      
      int minWidthPixels  = ConfigHelper.getMinWidthPixels(wcfg);
      
      int minHeightPixels = ConfigHelper.getMinHeightPixels(wcfg);

      websock.setResizeableWindow(ews.getWindowId(), 
                                  wcfg.resize,
                                  minWidthPixels,
                                  minHeightPixels,
                                  maxWidthPixels,
                                  maxHeightPixels);
   }

   /**
    * The method stores document outputs pending to be sent to the Javascript part of FWD client.
    * <p>
    * Before the supplied print output is inserted into the storage, the method removes all outputs
    * older than {@link #MAX_PRINT_OUTPUT_AGE}. The actual age is calculated from the current time
    * stamp and the time stamp of the document output when enqueued.
    *
    * @param   id
    *          Id of the document output.
    * @param   docOutput
    *          The object that encapsulates the document output being enqueued.
    */
   @Override
   public void enqueueDocumentOutput(String id, DocumentOutput docOutput)
   {
      // garbage collect old document outputs first
      long currTime = new Date().getTime();
      Set<UUID> losers = new HashSet<>();
      for (Map.Entry<UUID, DocumentOutput> out : documentOutputs.entrySet())
      {
         long delta = currTime - out.getValue().timestamp;
         if (delta > MAX_PRINT_OUTPUT_AGE)
         {
            losers.add(out.getKey());
         }
      }

      for (UUID loser : losers)
      {
         documentOutputs.remove(loser);
      }

      // enqueue the provided document output
      UUID uuid = UUID.fromString(id);
      docOutput.timestamp = new Date().getTime();
      documentOutputs.put(uuid, docOutput);
   }

   /**
    * Returns the document outputs previously enqueued with
    * {@link #enqueueDocumentOutput(String, DocumentOutput)}.
    * The returned entry is removed from the storage before it is returned to the caller.
    *
    * @param    id
    *           Id of the document output.
    *
    * @return   See above. The result may be <code>null</code> when the id is invalid or doesn't
    *           represent a previously enqueued document output.
    */
   @Override
   public DocumentOutput dequeueDocumentOutput(String id)
   {

      UUID uuid = UUID.fromString(id);
      return documentOutputs.remove(uuid);
   }

   /**
    * Post a new repaint for the specified window to the {@link #repaints} queue.  This will
    * be processed by the {@link RepaintThread}.
    * 
    * @param    ews
    *           The window requiring repaint.
    */
   protected void postRepaint(GuiWebEmulatedWindow ews)
   {
      synchronized (repaints)
      {
         repaints.addLast(ews);
         repaints.notifyAll();
      }
   }
   
   /**
    * Add the widget (and all its possible children) in the specified set, maintaining z-order in
    * a left-to-right (topmost to bottommost) order.
    * 
    * @param    w
    *           The parent widget.
    * @param    filter
    *           The condition predicate that must be evaluated for collected widgets.
    * @param    actionsCollector
    *           Collects a widget's action if the target is selected by the filter condition.
    */
   private void collectWidgets(Widget            w,
                               Predicate<Widget> filter, 
                               Consumer<Widget>  actionsCollector)
   {
      if (w == null)
      {
         return;
      }
   
      if (w instanceof MouseWidgetsEnumerator)
      {
         List<Widget> widgets = ((MouseWidgetsEnumerator) w).getMouseWidgets();

         if (widgets != null)
         {
            for (Widget widget : widgets)
            {
               if (filter.test(widget))
               {
                  actionsCollector.accept(widget);
               }
            }
         }
      }
      else
      {
         if (w instanceof AbstractContainer)
         {
            Widget[] children = ((AbstractContainer) w).widgets();

            // the javascript expects them in z-order, but JS keeps them left-to-right not
            // right-to-left
            for (int i = children.length - 1; i >= 0; i--)
            {
               collectWidgets(children[i], filter, actionsCollector);
            }
         }

         if (filter.test(w))
         {
            actionsCollector.accept(w);
         }
      }
   }

   /**
    * Checks if the given widget is belonging to visible frame or not inside any frame, like
    * window decorations.
    * 
    * @param   wid
    *          Widget to check.
    * 
    * @return  <code>TRUE</code> if the widget is inside visible frame or standalone,
    *          <code>FALSE</code> otherwise.
    */
   private boolean isInsideVisibleFrameOrStandalone(Widget wid)
   {
      // get the enclosing frame
      Frame fr = UiUtils.locateFrame(wid);
      
      // and return if it exists and is visible or does not exist
      return fr == null || fr.isVisible();
   }

   /**
    * Dedicated thread responsible of processing repaint requests.  For the Web driver, this 
    * requires calling {@link GuiWebEmulatedWindow#drawOnScreenBitmap()} for each window in the
    * {@link GuiWebDriver#repaints} queue.
    */
   private class RepaintThread
   extends Thread
   {
      /**
       * Initialize the repaint thread and start it.
       * <p>
       * The thread's name will be set to <code>WebRepaint</code> and is set as daemon.  It will
       * be alive for the entire life of the client instance.
       */
      public RepaintThread()
      {
         setName("WebRepaint");
         setDaemon(true);
      }
      
      /**
       * Listening loop for repaint requests.
       */
      @Override
      public void run()
      {
         while (true)
         {
            GuiWebEmulatedWindow ews = null;

            synchronized (repaints)
            {
               if (repaints.isEmpty())
               {
                  try
                  {
                     repaints.wait();
                  }
                  catch (InterruptedException e)
                  {
                     break;
                  }
               }
               
               if (!repaints.isEmpty())
               {
                  ews = repaints.removeFirst();
               }
               else
               {
                  continue;
               }
            }
            
            ews.drawOnScreenBitmap();
         }
      }
   }

   /**
    * Sets the current editors selection.
    * 
    * @param txt
    *        The current editors selection.
    */
   @Override
   public void setCurrentSelection(String txt)
   {
      websock.setSelection(txt);
   }

   /**
    * Load mouse pointer from the InputStream
    * 
    * @param data
    *        cursor data
    * @param ptrName
    *        pointer name
    *        
    * @return
    *        mouse pointer or <code>null</code>
    */
   @Override
   public MousePtrWrapper createMousePointer(InputStream data, String ptrName)
   {
      if (data == null)
      {
         return null;
      }

      byte[] bdata = UiUtils.getData(data);
      if (bdata == null || bdata.length == 0)
      {
         return null;
      }

      return MousePtrWrapper.createStaticWebInstance(bdata, Utils.mimeType(ptrName));
   }

   /**
    * Lock the mouse pointer (suppress widget-specific changes) or unlock it
    * 
    * @param widgetId
    *          the id of the active DropDown which caused the pointer lock or -1
    * @param ptr
    *          the locked mouse pointer or <code>null</code> for unlock
    *           
    */
   @Override
   public void lockMousePointer(int widgetId, MousePtrWrapper ptr)
   {
      websock.lockMousePointer(widgetId, ptr);
   }

   /**
    * Push information about the loaded mouse pointer for widget 
    * 
    * @param windowId
    *           to level window id
    * @param wid
    *           widgetId
    * @param ptr
    *           mouse pointer
    */
   public void propagate(int windowId, int wid, MousePtrWrapper ptr)
   {
      websock.setMousePtr(windowId, wid, ptr); 
   }

   /**
    * Returns the boolean value if the target font is installed or not.
    * 
    * @param    fontName
    *           The target font name.
    * @param    boldFont
    *           Indicates that the bold font is queried.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   @Override
   public boolean isFontInstalled(String fontName, boolean boldFont)
   {
      return websock.isFontInstalled(fontName, boldFont);
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(int[] windowIds, boolean capture)
   {
      websock.enableOsEvents(windowIds, capture);
   }

   /**
    * Returns true if a window can be decorated in this graphics environment, otherwise false.
    * 
    * @return   true if a window can have its title, caption buttons and border,
    *           otherwise false.
    */
   @Override
   public boolean isWindowDecorationSupported()
   {
      // In the web client embedded mode all windows must be drawn without decorations.
      return !embedded;
   }
   
   /**
    * Returns true if this client runs in embedded mode.
    * 
    * @return   See above.
    */
   @Override
   public boolean isEmbeddedClient()
   {
      return embedded;
   }

   /**
    * Returns current mouse cursor display position relative to the display origin.
    *
    * @return  mouse cursor position, never <code>null</code>
    */
   @Override
   public NativePoint getMousePosition()
   {
      return websock.getMousePosition();
   }
   /**
    * Get the widths of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's widths.
    */
   @Override
   public int[] getTextWidths(String text, FontDetails<Integer> fd)
   {
      if (isEmptyText(text))
      {
         return null;
      }
      
      TextMetrics metrics = resolveTextMetrics(text, fd);
      if (metrics.getWidths() == null)
      {
         metrics.setWidths(StringHelper.getAverageWidths(text.length(), metrics.getWidth()));
      }
      
      return metrics.getWidths();
   }

   /**
    * The method delivers document output to the client. The supplied consumer is invoked by
    * the driver providing a suitable output stream. The consumer implementation is responsible
    * for sending the data to the provided stream. The consumer implementation must not close
    * the provided stream. When the finalizer is supplied, it will be called by the driver when
    * the document output is no longer needed. The finalizer should free up any additional
    * resources the document output may need.
    *
    * <p>
    * This implementation stores the provided consumer with
    * {@link #enqueueDocumentOutput(String, DocumentOutput)} and notifies the JavaScript part
    * to download the document output. The document output is identified with a random UUID
    * generated by this method.
    *
    * @param   consumer
    *          Output consumer, must not be null. See above.
    * @param   finalizer
    *          Finalizer callback, may be null. See above.
    * @param   format
    *          Report output format.
    */
   @Override
   public void deliverDocumentOutput(Consumer<OutputStream> consumer, 
                                  Runnable               finalizer, 
                                  MediaType format)
   {
      UUID uuid = UUID.randomUUID();
      DocumentOutput printOut = new DocumentOutput(consumer,
                                                   getPrintOutputFile(format),
                                                   format.getMimeType(),
                                                   finalizer);
      String id = uuid.toString();
      enqueueDocumentOutput(id, printOut);
      websock.openPrintOutput(id, format.getMimeType());
   }

   /**
    * Opens the specified resource in a browser's window.
    * 
    * @param    mimeType
    *           The resource mime type.
    * @param    urlString
    *           The string presentation of the target resource url.
    * @param    embedded
    *           The document is embedded in the template document.
    * 
    * @throws   MalformedURLException, IOException
    *           If the string presentation of the target resource url is malformed. 
    */
   @Override
   public void openMimeResource(String mimeType, String urlString, boolean embedded)
   throws MalformedURLException
   {
      // test that urlString is well-formed. 
      URL url = new URL(urlString);
      
      URI uri;
      
      try
      {
         uri = url.toURI();
      }
      catch (URISyntaxException ex)
      {
         throw new MalformedURLException(ex.getMessage());
      }
      
      String id = uri.toASCIIString();
      
      boolean local = false;
      
      if (urlString.startsWith(FILE_SCHEMA))
      {
         if (url.getHost() == null || url.getHost().isEmpty())
         {
            local = true;
            final File temporaryFile;
            try
            {
               temporaryFile = File.createTempFile("openMimeResource",
                                                   null,
                                                   new File(tmpDir));
               try(InputStream input = url.openStream();
                   FileOutputStream output = new FileOutputStream(temporaryFile))
               {
                  ByteStreams.copy(input, output);
               }
            }
            catch (IOException ex)
            {
               throw new RuntimeException(ex);
            }
            
            Consumer<OutputStream> consumer = output ->
            {
               try(FileInputStream input = new FileInputStream(temporaryFile))
               {
                  ByteStreams.copy(input, output);
               }
               catch (IOException e)
               {
                  throw new RuntimeException(e);
               }
            };
            
            Runnable finalizer = () -> { temporaryFile.delete(); };
            
            UUID uuid = UUID.randomUUID();
            DocumentOutput docOut = new DocumentOutput(consumer,
                                                       url.getFile(),
                                                       mimeType,
                                                       finalizer);
            id = uuid.toString();
            enqueueDocumentOutput(id, docOut);
         }
      }
      
      websock.openMimeResource(id, mimeType, embedded, local);
   }

   /**
    * The method is called when the 4GL program execution enters a modal or wait-for event loop, in other
    * words when user input is expected.
    */
   @Override
   public void notifyWait()
   {
      if (needsNotificationOnWait)
      {
         websock.notifyWait();
      }
   }

   /**
    * The method is executed by the JS client when it requires to be notified when 4GL program execution
    * enters modal or wait-for event loop, in other words when user input is expected.
    *
    * @param    newState
    *           {@code true} turns ON the notification while {@code false} turns it OFF.
    */
   @Override
   public void setNeedsNotificationOnWait(boolean newState)
   {
      needsNotificationOnWait = newState;
   }

   /**
    * Get and cache metrics for the given texts.
    *
    * @param  texts
    *         Map of texts for which metrics should be calculated, keyed by fonts which are used
    *         for drawing specific texts.
    * @param  windowId
    *         ID of the parent window. Can be <code>null</code> if there is no need to select a
    *         window.
    */
   @Override
   public void cacheTextMetrics(Map<FontDetails<Integer>, Set<String>> texts, Integer windowId)
   {
      if (texts == null || texts.size() == 0)
      {
         return;
      }

      Map<FontDetails<Integer>, List<String>> textsToCache = null;

      for (FontDetails<Integer> font : texts.keySet())
      {
         for (String text : texts.get(font))
         {
            if (!isEmptyText(text) && !isActualTextMetricsCached(text, font))
            {
               if (textsToCache == null)
               {
                  textsToCache = new HashMap<>();
               }

               List<String> fontTextsToCache =
                     textsToCache.computeIfAbsent(font, k -> new ArrayList<>());
               fontTextsToCache.add(text);
            }
         }
      }

      if (textsToCache != null)
      {
         // convert map to arrays
         int size = textsToCache.size();
         String[][] textsToCacheArr = new String[size][];
         FontDetails[] fontArr      = new FontDetails[size];
         int[] fontNumArr           = new int[size];

         int counter = 0;
         for (FontDetails<Integer> font : textsToCache.keySet())
         {
            List<String> fontTextsToCache = textsToCache.get(font);
            fontArr[counter] = font;
            fontNumArr[counter] = font.font;
            textsToCacheArr[counter] = fontTextsToCache.toArray(new String[0]);
            counter++;
         }

         int[][] widths = websock.getTextWidths(fontNumArr, textsToCacheArr);

         try
         {
            if (windowId != null)
            {
               selectWindow(windowId);
            }

            for (int i = 0; i < size; i++)
            {
               String[] fontTextArr = textsToCacheArr[i];
               FontDetails font = fontArr[i];
               for (int j = 0; j < fontTextArr.length; j++)
               {
                  String text = fontTextArr[j];
                  // use swing metrics for height
                  int height = ews.getTextHeight(text, font);
                  int width = widths[i][j];
                  setTextMetrics(text, font, new TextMetrics(width, height, true));
                  FontManager.setTextMetrics(font.key, text, width, height);
               }
            }
         }
         finally
         {
            if (windowId != null)
            {
               releaseWindow();
            }
         }
      }
   }

   /**
    * Returns <code>true</code> if getting actual text metrics is an expensive operation for this
    * driver and you want to manually manage metrics cache using
    * {@link GuiDriver#cacheTextMetrics(Map, Integer)}.
    *
    * @return see above.
    */
   @Override
   public boolean needsExplicitTextMetricsCaching()
   {
      return true;
   }

   /*
    * Fetches the desktop dimension from the client and calls
    * {@link #updateDesktopDimensions(int, int)} with the result.
    */
   private void syncDesktopDimensions()
   {
      NativeDimension dim = websock.getDesktopDimension();
      if (dim == null)
      {
         // if it couldn't be computed for any reason...
         dim = new NativeDimension(1024, 768);
      }

      updateDesktopDimensions(dim.width, dim.height);
   }

   /**
    * Updates the class state that relies on the client desktop dimensions.
    *
    * @param   width
    *          New desktop width.
    * @param   height
    *          New desktop height.
    */
   private void updateDesktopDimensions(int width, int height)
   {
      synchronized (lock)
      {
         desktop = new NativeDimension(width, height);
         virtualScreen = new VirtualScreenImpl(width, height);
         workarea = new NativeDimension(width, taskbar ? height - TASKBAR_HEIGHT : height);
      }
   }
}
