/*
** Module   : p2j.emain.js
** Abstract : Defines common APIs for working with the embedded application's index.html.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20170727 Created initial version.
** 002 CA  20180801 Allow ADM window switching, perserving the embedded state (toolbar, pages, etc).
**                  Added standard code to handle MDC drawers (plus pinned/unpinned support).
** 003 CA  20180814 The /launch target must be relative to the current base path.
**                  Allow a delay before showing the loading overlay.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};
   
   /** Custom implementation of p2j.emain.js APIs. */
   var customEmain;

   /** Debugging flag. */
   var DEBUG = false;
   
   /** Should application be reloaded after user disconnection from FWD client driver? */
   var refreshOnDisconnect = true;
   
   /**
    * The JS object for the already loaded windows.  If the index for the window is not populated,
    * then the window's JS will be loaded via DOJO (if a custom one exists).
    */
   var activeWindows = [];
   
   /** User name. */
   var fwd_u = "";
   
   /** User password. */
   var fwd_p = "";
   
   /** The current window's title. */
   var currentWindowTitle = null;
   
   /*- The HTML code to define the iframe (taken on first login attempt, from the DOM). */
   var emptyIframe = undefined;
   
   /** The loading overlay timer. */
   var loading = undefined;
   
   /** The number of milliseconds to wait before showing the loading screen. */
   var loadingInterval = 500;
   
   /**
   * The state of the configured drawers:
   * {
   *    "object"     : // the MDCTemporaryDrawer instance
   *    "pinParent"  : // the ID of the element where the drawer content is moved when pinned.
   *    "sideParent" : // the ID of the element where the drawer content is moved when using MDC
   *    "drawerEl"   : // the ID of the drawer content
   *    "pinned"     : true/false, // state of the drawer - pinned or not
   *    "opened"     : true/false // state of the drawer - opened or not
   * } 
   */
   var drawers = {};
   
   /**
    * Set the custom emain implementation.
    *
    * @param    {object} cEmain
    *           An object overriding p2j.emain.js APIs.
    */
   me.initialize = function(cEmain)
   {
      customEmain = cEmain;

      if (customEmain && customEmain.getLoadingInterval)
      {
         loadingInterval = customEmain.getLoadingInterval();
      }

      // these need to be set before login...
      p2j.embedded.setWaitForInputCallback(waitForInput);
      p2j.embedded.setSocketDisconnectedCallback(notifyDisconnect);
      p2j.embedded.registerIncomingRequest("notifyDisconnect", notifyDisconnect);
   };
   
   /**
    * Initialize and register this drawer.
    * 
    * @param    {string} drawerName
    *           The ID of the DOM element used to construct a MDCTemporaryDrawer.
    * @param    {string} drawerEl
    *           The ID of the element with the drawer content.
    * @param    {string} pinParent
    *           The ID of the element where to move the drawer content when pinned.
    * @param    {string} sideParent
    *           The ID of the element where to move the drawer content when using MDC.
    */
   me.initializeDrawer = function(drawerName, drawerEl, pinParent, sideParent)
   {
      var dr = {};
      dr["object"] = new mdc.drawer.MDCTemporaryDrawer(document.querySelector("#" + drawerName));
      dr["pinParent"] = pinParent;
      dr["sideParent"] = sideParent;
      dr["drawerEl"] = drawerEl;
      dr["pinned"] = false;
      dr["opened"] = false;
      
      drawers[drawerName] = dr;
   };
   
   /**
    * Check if the specified drawer is opened.
    *
    * @param    {string} drawerName
    *           The ID of the DOM element used to construct a MDCTemporaryDrawer.
    */
   me.isDrawerOpened = function(drawerName)
   {
      var drawer = drawers[drawerName];
      var opened = drawer["opened"];
      var pinned = drawer["pinned"];
      
      return pinned ? opened : drawer["object"].open;
   };
   
   /**
    * Toggle the specified drawer between pinned and unpinned.
    *
    * @param   {string} drawerName
    *          The drawer ID.
    */
   me.pinDrawer = function(drawerName)
   {
      var drawer = drawers[drawerName];
      var opened = drawer["opened"];
      var pinned = drawer["pinned"];
      
      var pinParent = $("#" + drawer["pinParent"]);
      var sideParent = $("#" + drawer["sideParent"]);
      var drawerEl = $("#" + drawer["drawerEl"]);
      var newParent = pinned ? sideParent : pinParent;
      
      if (opened)
      {
         p2j.emain.toggleDrawer(drawerName);
      }
      
      if (!pinned)
      {
         pinParent.width(drawerEl.width());
      }
      
      drawerEl.detach().appendTo(newParent);
      drawer["pinned"] = !pinned;

      if (opened)
      {
         p2j.emain.toggleDrawer(drawerName);
      }
   }
   
   
   /**
    * Toggle the specified drawer between opened and closed, keeping the pinned/unpinned state.
    *
    * @param   {string} drawerName
    *          The drawer ID.
    * @param   {boolean} onlySide
    *          Flag to consider only the unpinned state.
    */
   me.toggleDrawer = function(drawerName, onlySide)
   {
      var drawer = drawers[drawerName];
      var opened = drawer["opened"];
      var pinned = drawer["pinned"];
      
      if (pinned && (onlySide == undefined || !onlySide))
      {
         var pinParent = $("#" + drawer["pinParent"]);
         if (opened)
         {
            pinParent.hide();
         }
         else
         {
            pinParent.show();
         }
         
         opened = !opened;
      }
      else if (!pinned)
      {
         opened = drawer["object"].open;
         drawer["object"].open = !opened;
         
         opened = !opened;
      }

      drawer["opened"] = opened;
   };

   /**
    * Register the specified window as active.
    *
    * @param   {string} windowId
    *          The window ID.
    * @param   {object} wobject
    *          The window object handler.
    */
   me.registerActiveWindow = function(windowId, wobject)
   {
      activeWindows[windowId] = wobject;
   };

   /**
    * Show the message in the specified JSON in an alert box.
    *
    * @param   {string} apiName
    *          The API to which the message belongs.
    * @param   {object} msg
    *          The JSON object with the message.
    */
   me.showMessage = function(apiName, msg)
   {
      var smsg = JSON.stringify(msg);
      dbg(apiName + " " + smsg);
   };
   
   /**
    * Enable/Disable the debug of incoming messages.
    *
    * @param   {boolean} state
    *          The debug state.
    */
   me.setDebug = function(state)
   {
      DEBUG = state;
   };
   
   /**
    * Execute the specified API, with the specified arguments.  This will execute the 
    * p2j.embedded.invokeProcedure API, with the program name set to "inMsg_" + apiName.  Also, 
    * a "messageId" is added to the arguments.  No callback is used for this call.
    *
    * @param    remoteProgramHandle
    *           The handle in witch to invoke the API.
    * @param    {string} apiName
    *           The name of the remote API to execute.
    * @param    {object} args
    *           The argument map.  The values must be strings.
    */
   me.invoke = function(remoteProgramHandle, apiName, args)
   {
      var msgId = p2j.embedded.getNextMessageId();
      args["messageId"] = msgId.toString();

      p2j.embedded.invokeProcedure(null, "inMsg_" + apiName, args, remoteProgramHandle);
   };
   
   /**
    * Callback executed when the P2J (FWD) page has loaded.
    *
    * @param    {string} origin
    *           The origin of the P2J page.
    */
   me.pageLoaded = function(origin)
   {
      p2j.embedded.registerIncomingRequest("p2jInitialized", p2jInitialized);
      p2j.embedded.registerIncomingRequest("showError", showError);
      if (customEmain && customEmain.getDefaultBgColor)
      {
         var bgcolor = customEmain.getDefaultBgColor();
         if (bgcolor)
         {
            p2j.embedded.setDefaultBgColorRgb(bgcolor);
         }
      }

      // do not call the remote program here, wait for the "p2jInitialized" notification
      p2j.embedded.silentErrorEnable();

      // perform login
      var msgId = p2j.embedded.getNextMessageId();
      var args = 
      {
         "messageId": msgId.toString(),
         "username" : fwd_u,
         "password" : fwd_p
      };
      p2j.embedded.invokeProcedure(null, "doLogin", args, null);
   };
   
   /**
    * Log into the client driver in order to load the embedded, converted ABL content.
    *
    * @param   {string} u
    *          User name.
    * @param   {string} p
    *          Password.
    */
   me.login = function(u, p)
   {
      if (emptyIframe == undefined)
      {
         emptyIframe = $("#p2jIframeParent").html();
      }
   
      setIframeToNone();
      refreshOnDisconnect = true;

      var ifr = document.getElementById("embeddedP2J");

      fwd_u = u;
      fwd_p = p;
 
      $(".for-login").hide();
      showLoading();
   
      // mostly hide background image with a nearly opaque overlay
      $("#bgFader").show().css("opacity", 0.87);

      if (ifr.src.startsWith("https://"))
      {
         var msgId = p2j.embedded.getNextMessageId();

         var args = 
         {
            "messageId": msgId.toString(),
            "username" : fwd_u,
            "password" : fwd_p
         };

         p2j.embedded.invokeProcedure(null, "doLogin", args, null);
      }
      else
      {
         var xmlHttp = new XMLHttpRequest();
         xmlHttp.onreadystatechange = function()
         { 
            if (xmlHttp.readyState == 4 && xmlHttp.status == 200)
            {
               var fwd_url = xmlHttp.responseText;
               
               if (fwd_url.startsWith("null"))
               {
                  showError({ "error": "Could not connect to the FWD server!"} );
               }
               else
               {
                  ifr.src = fwd_url;

                  // the main view is transparent at this point; however, we need to realize it here to
                  // work around a firefox bug (https://bugzilla.mozilla.org/show_bug.cgi?id=941146),
                  // which we would otherwise hit due to FWD's use of an unrealized canvas to calculate
                  // some font metrics
                  $("#mainView").show();

                  p2j.embedded.init("embeddedP2J", "embeddedP2JOverlay", "embedded-overlay");
                  p2j.embedded.setPageLoadedCallback(p2j.emain.pageLoaded);
               }
            }
         }
         
         xmlHttp.open("GET", p2j.embedded.getBasePath() + "/launch", true); 
         xmlHttp.send(null);
      }
   };
   
   /**
    * Notification from the business logic that the embedded app has initialized.
    *
    * @param   {object} msg
    *          An empty object.
    */
   function p2jInitialized(msg)
   {
      var fcb = function()
      {
         // make all drawers pinned
         for (var drawerName in drawers)
         {
            p2j.emain.pinDrawer(drawerName);
            p2j.emain.toggleDrawer(drawerName);
         }
      };
   
      if (customEmain && customEmain.p2jInitialized)
      {
         customEmain.p2jInitialized(msg, fcb);
      }
      else
      {
         fcb();
      }
   };

   /**
    * Notification the window title must be set.
    *
    * @param    {string} wtitle
    *           Optional, the title to use, if set.
    */
   function setTitle(wtitle)
   {
      if (wtitle)
      {
         currentWindowTitle = wtitle;
      }
      
      if (currentWindowTitle)
      {
         $("#iframe-menubar").show();
      }
      else
      {
         $("#iframe-menubar").hide();
      }
      
      $("#windowTitle").text(currentWindowTitle ? currentWindowTitle : ""); 
   };
   me.setTitle = setTitle;
   
   /**
    * Notification from the business logic the ADM page has changed.
    *
    * @param   {object} msg
    *          The JSON object, with 'page' and 'windowId' values.
    */
   function admPageChanged(msg)
   {
      var pageId = msg["page"];
      var thisPage = document.getElementById("fwd-page-" + pageId);

      if (customEmain && customEmain.admPageChanged)
      {
         customEmain.admPageChanged(pageId, thisPage, false);
      }
   };
   me.admPageChanged = admPageChanged;

   /**
    * Set the text for the "status-area" DOM component.
    * 
    * @param   {string} txt
    *          The status text.
    */
   me.showStatus = function(txt)
   {
      if (txt == "" || txt == undefined)
      {
         $("#status-area").html("&nbsp;");
      }
      else
      {
         $("#status-area").text(txt);
      }
   };

   /**
    * Set the text for the "message-area" DOM component.
    * 
    * @param   {string[]} msgs
    *          The messages.
    */
   me.showMessages = function(msgs)
   {
      var txt = "";
      if (msgs != undefined)
      {
         for (var msg in msgs)
         {
            txt = txt + msg + " ";
         }
      }
      if (txt == "")
      {
         $("#message-area").html("&nbsp;");
      }
      else
      {
         $("#message-area").text(txt);
      }
   };

   /**
    * Callback when the window needs to be removed (via the window list).
    *
    * @param   {number} windowId
    *          The window ID.
    */
   me.removeWindow = function(windowId)
   {
      showLoading();
   
      var fcb = function()
      {
         showLoading();
         
         // call apply "window-close" for this window
         p2j.embedded.applyEvent("window-close", windowId);
      };
      
      p2j.embedded.switchWindow(windowId, fcb);
   };

   /**
    * Callback when the window needs to be switched (via the window list).
    *
    * @param   {number} windowId
    *          The window ID, as returned by HWND.
    * @param   {function} switchCallback
    *          Callback executed after the window has been switched.
    */
   me.switchWindow = function(windowId, switchCallback)
   {
      if (p2j.embedded.getCurrentWindowId() == windowId)
      {
         hideLoading();
         return;
      }
   
      if (p2j.chips)
      {
         // deselect everything, the further 'VisibleWindows' message will process the top window
         // (including decorations - toolbar, menu, pages, etc)
         p2j.chips.selectChip(null);
      }
      
      p2j.embedded.switchWindow(windowId, switchCallback);
   };
   
   /**
    * Callback when the window needs to be destroyed.
    *
    * @param   {number} windowId
    *          The window ID.
    */
   me.windowDestroyed = function(windowId)
   {
      var wjs = p2j.emain.getWindowScript();
      if (wjs)
      {
         delete activeWindows[p2j.embedded.getCurrentWindow()];
         
         if (wjs.windowDestroyed)
         {
            wjs.windowDestroyed();
         }
      }

      currentWindowTitle = undefined;
      setTitle();
   };
   
   /**
    * Callback after the window has been created.
    */
   function windowCreated()
   {
      p2j.emain.hideLoading();

      var wjs = p2j.emain.getWindowScript();
      if (wjs && wjs.windowCreated)
      {
         wjs.windowCreated();
      }
      
      if (customEmain && customEmain.windowCreated)
      {
         customEmain.windowCreated();
      }
   };
   me.windowCreated = windowCreated;
   
   /**
    * Express a message, either as an alert (in debug mode), or as an entry in the console log.
    */
   function dbg(msg)
   {
      if (DEBUG)
      {
         alert(msg);
      }
      else
      {
         console.error(msg);
      }
   };
   
   /**
    * Reload the current page after a slight delay.
    */
   function refreshPage()
   {
      if (refreshOnDisconnect)
      {
         setTimeout(function ()
         {
            location.reload(true);
         }, 750);
      }
   };
   me.refreshPage = refreshPage;
  
   /**
    * Set the URL for the iframe containing embedded, converted ABL content to 'about:blank'.
    */
   function setIframeToNone()
   {
      p2j.embedded.handleMessages = false;
      var ifr = document.getElementById("embeddedP2J");
      ifr.src = "about:blank";

      $("#p2jIframeParent").html(emptyIframe);
   }
   
   /**
    * Get the javascript object loaded for the specified window, from the <code>activeWindows</code>
    * map.
    * 
    * @param    {string} windowId
    *           The window ID.  If undefined, use the current window.
    *
    * @return   The object or undefined if none was loaded.
    */
   me.getWindowScript = function(windowId)
   {
      if (windowId == undefined)
      {
         windowId = p2j.embedded.getCurrentWindow();
      }
      
      return activeWindows[windowId] ? activeWindows[windowId] : undefined;
   };

   /**
    * Notification that the 4GL business logic is in a state where is waiting for input.  This 
    * will be sent only if there is a modal/event processing loop started while a 
    * p2j.embedded.INVOKE call is still in progress.
    */
   function waitForInput()
   {
      hideLoading();

      document.getElementById("p2jIframeParent").style.visibility = "visible";
      document.getElementById("p2jIframeParent").style.display = "";

      $(".for-runtime").show();
      $(".for-runtime").css("visibility", "visible");
      $(".for-login").hide();
      
      if (customEmain && customEmain.waitForInput)
      {
         customEmain.waitForInput();
      }
   };
   me.waitForInput = waitForInput;

   /**
    * Show the 'Loading...' overlay and hide the iframe.
    * <p>
    * If the loading interval is 0, then the overlay is shown immediately.  Otherwise, 
    * wait the number of milliseconds and, if hideLoading was not called, show it.
    */
   function showLoading()
   {
      if (loading != undefined)
      {
         return;
      }
      
      if (loadingInterval == 0)
      {
         showLoadingInt();
         return;
      }
   
      loading = window.setTimeout(showLoadingInt, loadingInterval);
   };
   me.showLoading = showLoading;

   /**
    * Show the 'Loading...' overlay and hide the iframe.
    */
   function showLoadingInt()
   {
      $("#bgFader").css("display", "table");
      $("#waitMsg").css("opacity", 1).text("Loading...");

      $("#mainView").css("opacity", 0);
      $("#mainView").css("visibility", "hidden");
      
      loading = undefined;

      if (customEmain && customEmain.showLoading)
      {
         customEmain.showLoading();
      }
   };
      
   /**
    * Hide the 'Loading...' overlay and show the iframe.
    */
   function hideLoading()
   {
      if (loading != undefined)
      {
         window.clearTimeout(loading);
         loading = undefined;
      }

      $("#bgFader").hide();
      $("#waitMsg").css("opacity", 0).text("Loading...");
      
      $("#mainView").css("display", "inline-flex");
      $("#mainView").css("opacity", 1);
      $("#mainView").css("visibility", "visible");

      setTitle();
      
      if (customEmain && customEmain.hideLoading)
      {
         customEmain.hideLoading();
      }
   };
   me.hideLoading = hideLoading;
   
   /**
    * Notification that the remote side has disconnected.
    */
   function notifyDisconnect()
   {
      var mainView = $("#mainView");
      mainView.css("opacity", 0);
      setTimeout(function ()
      {
         if (refreshOnDisconnect)
         {
            $(".for-runtime").hide();
         }
         mainView.hide();
         $(".for-login").show();

         $("#bgFader").hide();
      }, 1000);

      if (refreshOnDisconnect)
      {
         refreshPage();
      }
      else
      {
         setIframeToNone();
      }
   };
   me.notifyDisconnect = notifyDisconnect;

   /**
    * Show the current embedded window.
    *
    * @param   {string} windowId
    *          The window's ID (the ADM program name).
    * @param   {string} windowTitle.
    *          The window's title.
    * @param   {number} pageNum
    *          If set, the 1-based page number to select.
    */
   function showCurrentWindow(windowId, windowTitle, pageNum)
   {
      p2j.embedded.setCurrentWindow(windowId);
      currentWindowTitle = windowTitle;
      
      document.getElementById("p2jIframeParent").style.visibility = "visible";
      document.getElementById("p2jIframeParent").style.display = "";

      if (activeWindows[windowId] || 
          !(customEmain && customEmain.loadWindowScript && 
            customEmain.loadWindowScript(windowId, windowTitle)))
      {
         p2j.embedded.showWindow(windowId, windowCreated, windowTitle, pageNum);
      }
   };
   me.showCurrentWindow = showCurrentWindow;

   /**
     * Show the embedded page targeted by this event. 
     *
     * @param    {Event} evt
     *           The event object.
     */
   me.showEmbeddedPage = function(evt)
   {
      var id = this.id;
      var pageId;
      
      if (id.startsWith("fwd-page-"))
      {
         pageId = id.substring("fwd-page-".length);
      }
      else
      {
         return;
      }
      
      evt.preventDefault();
      
      if (customEmain && customEmain.admPageChanged)
      {
         customEmain.admPageChanged(pageId, this, true);
      }
   };

   /**
    * Toggle visibility of the given component and bring all its parents to the same z-order.
    *
    * @param    {string} compId
    *           The DOM component ID.
    * @param    {boolean} forceHide
    *           Flag indicating if the component is force to be hidden.
    */
   me.toggleVisible = function(compId, forceHide)
   {
      var comp = $("#" + compId);
      var isVis = comp.is(':visible');
      if (isVis || forceHide)
      {
         comp.hide();
         comp.parents().css("z-index", 1);
      }
      else
      {
         comp.parents().css("z-index", comp.css("z-index"));
         comp.show();
      }
   };

   /**
    * Show an error from a JSON message in the alert area of the login container.
    *
    * @param   {object} jsonMsg
    *          Object containing error message.
    */
   function showError(jsonMsg)
   {
      $("#p2jAlertArea").text(jsonMsg.error).show();

      refreshOnDisconnect = false;
   };

   return me;
});