/*
** Module   : p2j.embedded.js
** Abstract : The embedded application controller that is responsible to extend embedded dialogs
**            modality to the external web application.
**
** Copyright (c) 2016-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 SBI 20160330 Added the embedded application module to be loaded by the DOJO loader.
** 002 CA  20160425 Added the remote APIs which can be used by customer-specific applications, 
**                  to call into the converted code.
** 003 CA  20160429 Added a missing 'return' when treating the SOCKET_DISCONNECTED request in 
**                  receiveRemoteMessage
** 004 CA  20170105 Fixed overlay panel state with nested modal windows.
** 005 CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 006 CA  20170825 Enhanced to allow an automatic way to drive the embedded windows via the 
**                  fwd-embedded-driver.p program.
** 007 CA  20170928 Changes to maximize common code managed by FWD.
** 008 CA  20171004 Expose legacy menubar to the embedded application.
** 009 CA  20171023 ADM version is kept on a per-window basis.
** 010 CA  20171111 Added possibility to set default BGCOLOR for the ABL windows.
** 011 HC  20171203 Added hiding of Loading... overlay when 4GL app blocks for user input during 
**                  window/page switch.
** 012 CA  20180313 Added module and toolbar support.  Other misc fixes.
** 013 CA  20180529 Fixed some bugs in H012 related to Hotel GUI (which isn't using modules).
** 014 CA  20180727 Embedded mode enhancements: added APPLY extension, allow notifications for 
**                  window status and message texts; allow visible window notification; other misc
**                  fixes.
** 015 CA  20180801 Allow ADM window switching, perserving the embedded state (toolbar, pages, etc).
** 016 CA  20180813 When top window has been switched, send a notification.  This allows the 
**                  embedded app to track the topmost window, even if is not an ADM one - so that
**                  if is not an ADM window (and non-modal), it can be i.e. closed.
**                  The /adm_windows.json file must be relative to the current base path.
**     CA  20180816 Allow the toolbar images to be prefixed by a custom path.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};
      
   me.REQUEST         = "request";
   me.RESPONSE        = "response";
   me.REMOTE_CALL     = "remote-call";
   me.INVOKE          = "invoke";
   me.DELETE_HANDLE   = "delete-handle";
   me.SUBSCRIBE       = "subscribe";
   me.UNSUBSCRIBE     = "unsubscribe";
   me.PUBLISH         = "publish";
   me.CREATE_RESOURCE = "create-resource";
   me.APPLY_EVENT     = "apply-event";
   me.SET_STATUS_TEXT = "set-status";
   me.CLEAR_MESSAGES  = "clear-messages";
   me.ADD_MESSAGES    = "add-messages";

   me.EXTERNAL_PROGRAM = 0;
   me.PROCEDURE        = 1;
   me.FUNCTION         = 2;

   me.PAGE_LOADED         = -1;
   me.PAGE_CLOSED         = -2;
   me.SOCKET_DISCONNECTED = -3;
   me.SOCKET_RECONNECTED  = -4;
   me.WAIT_USER_INPUT     = -5;

   /** Flag indicating if the remote side can start receiving messages. */
   me.handleMessages = false;

   /** Called when the 4gl app blocks for user input during window/page switch. */
   me.waitForInputCallback = null;
   
   /** Map of parsers for each API. */
   var incomingRequestParser = {};
   
   /** Map of handlers for each API. */
   var incomingRequestHandlers = {};
   
   /**
    * If set, it represents a default RGB color for the window background, in the "#rrggbb" format
    */
   var defaultBgColorRgb = undefined;

   /**
    * The details for each window, used to build pages,toolbar, etc. The format is:
    * {
    *    "messageText" : string,
    *    "statusText"  : string,
    *    "pages"       : string[],
    *    "toolbar"     : JSON,
    *    "menubar"     : JSON,
    *    "currentPage" : number
    * }
    */
   var windowDetails = {}; 

   /** Callback executed when the status needs to be shown. */
   var showStatusCallback;

   /** Callback executed when the message needs to be shown. */
   var showMessagesCallback;

   /** The chips handler used to create the window chips. */
   var chipsHandler;

   /**
    * The object which is responsible for creating, destroying and processing the modules. 
    * If not set, the implicit FWD module handler will be used.
    */
   var moduleHandler;

   /** The ID of the DOM element where the modules will be attached. */
   var moduleContainer;
   
   /** The callback to be executed when a module is selected. */
   var moduleEvent;

   /**
    * The object which is responsible for creating, destroying and processing the menubar. 
    * If not set, the implicit FWD menu handler will be used.
    */
   var menuHandler;

   /** The ID of the DOM element where the menubar will be attached. */
   var menuContainer;

   /**
    * The object which is responsible for creating, destroying and processing the toolbar. 
    * If not set, the implicit FWD toolbar handler will be used.
    */
   var toolbarHandler;

   /** The ID of the DOM element where the toolbar will be attached. */
   var toolbarContainer;

   /** The data for embedded windows.*/
   var admWindows = {};
   
   /** Map of program name to module name. */
   var modules = {};
   
   /** The current ADM window - the program name. */
   var currentADMWindow;
   
   /** The current ADM window ID. */
   var currentFWDWindowId;
   
   /** Flag identifying if we are switching ADM modules via the user's action. */
   var admWindowSwitch = false;
   
   /** The handle ID for the FWD embedded driver. */
   var fwdEmbeddedDriverHandle;
   
   /** A callback to initialize/populate the embedded window list in the UI. */
   var populateWindowsFunc;
   
   /** A callback to initialize/populate the pages for a embedded window. */
   var populatePagesFunc;
   
   /** A callback executed when the window has been destroyed by the business logic. */
   var windowDestroyedFunc;
   
   /** A callback executed when the top window has changed. */
   var topWindowChangedCallback;
   
   /**
    * A registry of callbacks, to be executed when a response is received for a message.  If a
    * callback is not set for a message ID, then the global callback is executed.
    */
   var callbacks = {};

   /** Flag indicating if the API calls have the NO-ERROR mode enabled. */
   var noError = false;

   /** The generator for the message IDs, on the remote side (always positive). */ 
   var nextMessageId = 0;
   
   /**
    * A register with all subscriptions.  This acts like a map with the following structure:
    * - each key is a resource ID
    * - the value is an associative array with "event" key representing the event name and 
    *   the "method" key representing the function callback to be executed for the event.
    */
   var subscriptions = {};
   
   /**
    * A register with the pending subscriptions.  Acts the same as the <code>subscriptions</code>
    * object, except the keys in this object represent message IDs.
    * <p>
    * Once a response is received for the SUBSCRIBE request, the subscription will be registered
    * in the <code>subscriptions</code> register.
    */
   var pendingSubscriptions = {};
   
   /**
    * A register with the pending unsubscriptions.  Acts like a map with the following structure:
    * - each key is the message ID
    * - each value is an associative array with "subscriber-id" the ID of the subscriber which
    *   initiated the UNSUBSCRIBE and the "event" key the event name.
    * <p>
    * Once a response is received for the UNSUBSCRIBE request, the subscription will be removed
    * from the <code>subscriptions</code> register.
    */
   var pendingUnsubscriptions = {};

   /**
    * It is responsible to create the overlay panel and to put it over all page elements except
    * the embedded frame that must be on the top. The method must be invoked on the page load
    * event to be sure that all page elements have been loaded.
    * 
    * @param    {String} frameId
    *           The embedded frame unique identifier.
    * @param    {String} overlayId
    *           The overlay panel unique identifier.
    * @param    cls
    *           The overlay panel css name.
    */
   me.init = function(frameId, overlayId, cls)
   {
      var frame  = document.getElementById(frameId);
      this.fwdIframe = frame;

      var origin = getUrlOrigin(frame.src);
      // calculates the maximal zIndex over the page elements
      var zIndex  = getMaxZIndex();
      // supposed that IFRAME is wrapped by DIV that is used to place
      // the embedded content within the page
      var frameParent = frame.parentElement;
      // create the overlay panel that must cover the displayed page
      // let the overlay panel be a sibling of the frame parent - frameParent
      var overlayPanel = createOverlayPanel(overlayId, zIndex, frameParent.parentElement, cls);
      // let the frame be on the top of the overlay panel
      while (frameParent != overlayPanel.parentElement)
      {
         frameParent.style.zIndex = zIndex + 1;
         frameParent = frameParent.parentElement;
         zIndex = zIndex + 1;
      }
      frame.style.zIndex = zIndex;
      
      var overlayPanelStates = [];
      overlayPanelStates.push("none");
      
      window.addEventListener(
                              "message",
                              function (event)
                              {
                                 if (event.origin != origin)
                                 {
                                    console.error("event.origin " + event.origin +
                                          " but expected " + origin);
                                    return;
                                 }
                                 
                                 var handled = false;
                                 
                                 // expected to get { target : "OverlayPanel", visible: [true|false]}.
                                 var msg = event.data;
                                 if (msg.target === "OverlayPanel" && msg.visible === true)
                                 {
                                    handled = true;
                                    
                                    overlayPanelStates.push(overlayPanel.style.display);
                                    overlayPanel.style.display = "block";
                                    frame.focus();
                                 }
                                 else if (msg.target === "OverlayPanel" && msg.visible === false)
                                 {
                                    handled = true;
                                    
                                    var state = overlayPanelStates.pop();
                                    overlayPanel.style.display = state;
                                 }
                                 else if (msg.target === "Frame" && msg.bgcolor)
                                 {
                                    handled = true;
                                    
                                    if (defaultBgColorRgb != undefined)
                                    {
                                       frame.style.background = defaultBgColorRgb;
                                    }
                                    else
                                    {
                                       frame.style.background = msg.bgcolor;
                                    }
                                 }
                                 else if (msg.target === "WindowDestroyed")
                                 {
                                    handled = true;
                                    
                                    p2j.embedded.destroyWindow(parseInt(msg.windowId), true);
                                 }
                                 else if (msg.target === "VisibleWindows")
                                 {
                                    handled = true;
                                    
                                    p2j.embedded.captureVisibleWindows(msg.windows, msg.topWindow);
                                 }
                                 
                                 if (handled)
                                 {
                                    event.stopPropagation();
                                 }
                              },
                              false);

      this.targetWindow = window.frames[frameId];
      this.targetOrigin = "*";

      if (window.addEventListener)
      {
         window.addEventListener("message", receiveRemoteMessage, false);
      }
      else
      {
         window.attachEvent("onmessage", receiveRemoteMessage);
      }
   };
   
   /**
    * Mark the status area (for module windows) as being handled by the embedded application.
    *
    * @param   {boolean} l
    *          Flag indicating if status area is handled by embedded app (when true).
    * @param   {function} callback
    *          Callback executed when the remote call is finished.
    */
   me.embeddedStatusArea = function(l, callback)
   {
      var args = 
      {
         "l" : l
      };

      p2j.embedded.invokeProcedure(callback,
                                   "inMsg_embeddedStatusArea", 
                                   args,
                                   fwdEmbeddedDriverHandle);
   };
   
   /**
    * Mark the message area (for module windows) as being handled by the embedded application.
    *
    * @param   {boolean} l
    *          Flag indicating if message area is handled by embedded app (when true).
    * @param   {function} callback
    *          Callback executed when the remote call is finished.
    */
   me.embeddedMessageArea = function(l, callback)
   {
      var args = 
      {
         "l" : l
      };

      p2j.embedded.invokeProcedure(callback,
                                   "inMsg_embeddedMessageArea", 
                                   args,
                                   fwdEmbeddedDriverHandle);
   };
   
   /**
    * Notification executed when the remote side has computed a new visible window stack.
    * 
    * @param   {object} windowList
    *          The associative array with window ID as key and window title as value.
    * @param   {number} topWindow
    *          The id of the top window, as returned by HWND.
    */
   me.captureVisibleWindows = function(windowList, topWindow)
   {
      if (topWindow != undefined && currentFWDWindowId != topWindow && topWindowChangedCallback)
      {
         topWindowChangedCallback(topWindow);
      }
      
      if (!admWindowSwitch                &&
          fwdEmbeddedDriverHandle         &&
          currentFWDWindowId != topWindow &&
          topWindow != undefined)
      {
         if (chipsHandler)
         {
            chipsHandler.interact(parseInt(topWindow));
         }
         else
         {
            p2j.embedded.switchWindow(parseInt(topWindow));
         }
      }

      if (chipsHandler)
      {
         chipsHandler.createChips(windowList, topWindow);
      }
   };
   
   /**
    * Init the details for the chips support.
    *
    * @param    {string} cContainer
    *           The ID of the DOM element where the chips will be attached.
    * @param    {function} cRemove
    *           The callback to be executed when a chip is being removed.
    * @param    {function} cInteract
    *           The callback to be executed when a chip is being selected.
    * @param    {object} cHandler
    *           The object which will handle the chips.
    */
   me.initChips = function(cContainer, cRemove, cInteract, cHandler)
   {
      chipsHandler = cHandler;
      cHandler.init(cContainer, cRemove, cInteract);
   };
   
   /**
    * Init the details for the module support.
    *
    * @param    {string} mcontainer
    *           The ID of the DOM element where the modules will be attached.
    * @param    {function} mevent
    *           The callback to be executed when a module is selected.
    * @param    {object} mhandler
    *           Optional, the object which will handle the module.  If not set, the implicit 
    *           FWD module support will be used.
    */
   me.initModule = function(mcontainer, mevent, mhandler)
   {
      moduleContainer = mcontainer;
      moduleEvent = function(moduleName, windowTitle)
      {
         var args = 
         {
            "module" : moduleName
         };

         var mcb = function(msg)
         {
            var payload = msg.payload;
            
            var pname = payload["arguments"]["programname"];
            
            if (pname == undefined)
            {
               // can't run this...
               return;
            }
            
            pname = pname.replace(/\\/g, '/');
            pname = pname.replace(/^.\//g, '');

            if (pname.startsWith("./"))
            {
               pname = pname.substring(2);
            }

            modules[pname] = moduleName;

            // don't check the admWindows map, because that contains only the 'paged' ADM windows
            // single-frame ADM windows can still be run and will not be included here.
            mevent(pname, windowTitle);
         };

         p2j.embedded.invokeProcedure(mcb, "inMsg_fwdGetProgramModule", args,
                                      fwdEmbeddedDriverHandle);
         
      };
      if (mhandler)
      {
         moduleHandler = mhandler;

         moduleHandler.init(moduleContainer, moduleEvent);
      }
   };
   
   /**
    * Init the details for the menubar support.
    *
    * @param    {string} mcontainer
    *           The ID of the DOM element where the menubar will be attached.
    * @param    {function} mevent
    *           Callback executing when selecting a menu item.
    * @param    {object} mhandler
    *           Optional, the object which will handle the menu.  If not set, the implicit 
    *           FWD menu support will be used.
    */
   me.initMenuBar = function(mcontainer, mevent, mhandler)
   {
      menuContainer = mcontainer;
      
      var fcb = function(hmenu, callback)
      {
         if (mevent)
         {
            mevent(hmenu, callback);
         }
         
         menuEvent(hmenu, callback);
      };
      
      if (mhandler)
      {
         menuHandler = mhandler;

         menuHandler.init(menuContainer, fcb);
      }
   };
   
   /**
    * Init the details for the toolbar support.
    *
    * @param    {string} tbcontainer
    *           The ID of the DOM element where the toolbar will be attached.
    * @param    {object} tbhandler
    *           Optional, the object which will handle the toolbar.  If not set, the implicit 
    *           FWD toolbar support will be used.
    * @param    {string} imageFolder
    *           The folder to add as a prefix for all images.
    */
   me.initToolBar = function(tbcontainer, tbhandler, imageFolder)
   {
      toolbarContainer = tbcontainer;
      
      if (tbhandler)
      {
         toolbarHandler = tbhandler;

         toolbarHandler.init(toolbarContainer, toolbarEvent, imageFolder);
      }
   };

   /**
    * Notify that a menu event has been raised.
    *
    * @param    {string} hmenu
    *           The resource ID of source menu element for this event.
    * @param    {function} callback
    *           Optional, the function to execute to handle the result of the FWD call.
    */
   function menuEvent(hmenu, callback)
   {
      var args = 
      {
         "hch" : hmenu
      };

      p2j.embedded.invokeProcedure(callback,
                                   "inMsg_fwdMenuItemSelected", 
                                   args,
                                   fwdEmbeddedDriverHandle);
   };

   /**
    * Notify that a toolbar event has been raised.
    *
    * @param    {string} hbutton
    *           The resource ID of source button element for this event.
    * @param    {function} callback
    *           Optional, the function to execute to handle the result of the FWD call.
    */
   function toolbarEvent(hbutton, callback)
   {
      var args = 
      {
         "hch" : hbutton
      };

      p2j.embedded.invokeProcedure(callback,
                                   "inMsg_fwdToolbarItemSelected", 
                                   args,
                                   fwdEmbeddedDriverHandle);
   };

   /**
    * Check if the NO-ERROR mode is enabled for messages.
    */
   me.isSilentError = function()
   {
      return noError;
   };
   
   /**
    * Enable NO-ERROR (silent error) mode for all future requests.  This will add a 
    * <code>msg.noError = true</code> flag into the message.
    */
   me.silentErrorEnable = function()
   {
      noError = true;
   };
   
   /**
    * Disable NO-ERROR (silent error) mode for all future requests.  This will add a 
    * <code>msg.noError = true</code> flag into the message.
    */
   me.silentErrorDisable = function()
   {
      noError = false;
   };
   
   /**
     * Send a message to the target window.
     */
   me.send = function(msg)
   {
      this.targetWindow.postMessage(msg, this.targetOrigin);
   };

   /**
    * Get the the callback handler for the specified message ID.
    *
    * @param   {number} msgId
    *          The message ID.
    *
    * @return  The message-specific callback or the global callback (if defined).
    */
   me.getHandler = function(msgId)
   {
      var msgCallback = msgId ? callbacks[msgId] : undefined;
      if (msgCallback != undefined)
      {
         delete callbacks[msgId];
      }
      else if (p2j.embedded.messageCallback)
      {
         msgCallback = p2j.embedded.messageCallback;
      }
      else
      {
         msgCallback = defaultMessageCallback;
      }
      
      return msgCallback;
   };

   /**
    * Generic handler for the messages incoming from the converted code.  These can be either
    * responses for a INVOKE request or REMOTE_CALL requests.
    * The payload for the message (msg.payload) will always be a JSON string.
    *
    * @param   {object} msg
    *          The received message.
    */
   function defaultMessageCallback(msg)
   {
      if (msg.type == p2j.embedded.REMOTE_CALL)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            // do the work
            handleIncomingRequest(msg.payload);
            
            // send just an OK response
            p2j.embedded.sendRemoteCallResponse(msg.id, "OK");
         }
         else
         {
            // a RESPONSE mode for REMOTE_CALL should never happen 
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a REMOTE_CALL message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
      }
      else if (msg.type == p2j.embedded.INVOKE)
      {
         if (msg.mode == p2j.embedded.RESPONSE)
         {
            if (msg.payload.error)
            {
               console.error(msg.payload["error-num"] + " " + msg.payload["error-msg"]);
            }
            
            // otherwise, don't do anything
         }
         else
         {
            // a REQUEST mode for INVOKE should never happen
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a INVOKE message with " + msgMode + 
                        " mode - remote side can handle only RESPONSE modes!");
         }
      }
   };

   /** 
    * Generic handler for incoming REMOTE_CALL requests.  It delegates the call to a dedicated
    * handler, based on the "api" value.
    * 
    * @param   {string} msg
    *          The request message, always a JSON string.
    */
   function handleIncomingRequest(msg)
   {
      var msgJson = JSON.parse(msg);
      
      var api = msgJson["api"];
      var incomingParser = getIncomingRequestParser(api);
      var incomingHandler = getIncomingRequestHandler(api);

      if (incomingParser)
      {
         // parse the message
         msgJson = incomingParser(msg);
         
         // invoke the appropriate handle
         incomingHandler(msgJson);
      }
      else
      {
         console.error("Received message " + msg + " with unknown handler for " + api);
      }
   };
   
   /**
    * Resolve the request parser for the specified API name.
    * 
    * @param    {string} api
    *           The API name.
    *
    * @return   The request parser (as a function), or undefined.
    */
   function getIncomingRequestParser(api)
   {
      var incomingParser;
      
      if (incomingRequestParser[api])
      {
         incomingParser = incomingRequestParser[api];
      }
      
      return incomingParser;
   };
   
   /**
    * Resolve the request handler for the specified API name.
    * 
    * @param    {string} api
    *           The API name.
    *
    * @return   The request handler (as a function), or undefined.
    */
   function getIncomingRequestHandler(api)
   {
      var incomingHandler;

      if (incomingRequestHandlers[api])
      {
         incomingHandler = incomingRequestHandlers[api];
      }
      
      return incomingHandler;
   };
   
   /**
    * Register a response and parse callbacks for the given API.
    * 
    * @param  {string} api
    *         The API name.
    * @param  {function} respCallback
    *         Callback executed when this API is being received.  It has as single argument a 
    *         JSON object.
    * @param  {function} parseCallback
    *         Custom callback used to parse this message.
    *
    */
   me.registerIncomingRequest = function(api, respCallback, parseCallback)
   {
      if (parseCallback == undefined)
      {
         parseCallback = function(msg)
         {
            var jsonMsg = {};
            if (msg)
            {
               jsonMsg = JSON.parse(msg);
            }
      
            return jsonMsg;
         };
      }

      incomingRequestParser[api] = parseCallback;
      incomingRequestHandlers[api] = respCallback;
   };
   
   /**
    * Deregister the given API.
    * 
    * @param  {string} api
    *         The API name.
    */
   me.deregisterIncomingRequest = function(api)
   {
      delete incomingRequestHandlers[api];
   };

   /**
    * Set a global callback to be executed when a message is received.
     *
     * @param    {function} callback
     *           The global callback.
    */
   me.setMessageCallback = function(callback)
   {
      this.messageCallback = callback;
   };
   
   /**
    * Set the default background color for the iframe's body - it will send a message to the 
    * remote side.
    *
    * @param    {string} colorRgb
    *           Color in "#rrggbb" CSS-style format.
    */
   me.setDefaultBgColorRgb = function(colorRgb)
   {
      defaultBgColorRgb = colorRgb;
      var msg = 
      {
         "target" : "Frame", 
         "bgcolor" : defaultBgColorRgb
      };
      
      p2j.embedded.send(msg);
   };
   
   /**
    * Set a callback to be executed when a PAGE_LOADED message is received.
    *
    * @param    {function} callback
    *           The page-loaded callback.
    */
   me.setPageLoadedCallback = function(callback)
   {
      this.pageLoadedCallback = callback;
   };
   
   /**
    * Set a callback to be executed when a PAGE_CLOSED message is received.
    *
    * @param    {function} callback
    *           The page-closed callback.
    */
   me.setPageClosedCallback = function(callback)
   {
      this.pageClosedCallback = callback;
   };
   
   /**
    * Set a callback to be executed when a SOCKET_DISCONNECTED message is received.
    *
    * @param    {function} callback
    *           The socket-disconnected callback.
    */
   me.setSocketDisconnectedCallback = function(callback)
   {
      this.socketDisconnectedCallback = callback;
   };
      
   /**
    * Set a callback to be executed when a SOCKET_RECONNECTED message is received.
    *
    * @param    {function} callback
    *           The socket-reconnected callback.
    */
   me.setSocketReconnectedCallback = function(callback)
   {
      this.socketReconnectedCallback = callback;
   };
   
   /**
    * This API creates a DELETE_HANDLE message request and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} handleId
    *          A string with a hex representation of the handleId. 
    */
   me.deleteHandle = function(callback, handleId)
   {
      var msgId = getNextMessageId();
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      var msg = {};
      msg["type"]    = this.DELETE_HANDLE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = handleId;
      msg["noError"] = noError;
      
      this.send(msg);
   };

   /**
    * Execute a SUBSCRIBE request.  If the <code>subscriber</code> is undefined, then it will be 
    * built and its id can be saved via the <code>callback</code> function.
    *
    * @param   {function} callback
    *          A callback with the response for the SUBSCRIBE request. 
    * @param   {boolean} global
    *          Flag indicating if this SUBSCRIBE is for notifications from all contexts.
    * @param   {string} subscriber
    *          The hex representation of this subscriber resource ID, or undefined if it is not 
    *          yet built.
    * @param   {string} event
    *          The event for which to register.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.  If undefined or
    *          empty string, acts as ANYWHERE legacy option.
    * @param   {function} method
    *          The callback which will be executed when the event is raised via PUBLISH.
    */
   me.subscribe = function(callback, global, subscriber, event, publisher, method)
   {
      var msgId = getNextMessageId();
      
      if (subscriber == undefined)
      {
         subscriber = "";
      }
      
      if (event == undefined)
      {
         event = "";
      }
      
      if (publisher == undefined)
      {
         publisher = "";
      }
      
      // 'callback' is used to save the subscriber ID (if is unknown here) 
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      pendingSubscriptions[msgId] = 
      {
         "event"  : event,
         "method" : method
      };
      
      var payload = 
      {
         "global"     : global,
         "subscriber" : subscriber,
         "event"      : event,
         "publisher"  : publisher
      };

      var msg = {};
      msg["type"]    = this.SUBSCRIBE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["noError"] = noError;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a UNSUBSCRIBE request.
    *
    * @param   {function} callback
    *          A callback with the response for the UNSUBSCRIBE request. 
    * @param   {boolean} global
    *          Flag indicating if this UNSUBSCRIBE is for global registrations or for local ones.
    * @param   {string} subscriber
    *          The hex representation of this subscriber resource ID.
    * @param   {string} event
    *          The event the subscription was registered.  undefined or empty string will 
    *          act as null on the remote side.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.  If undefined or
    *          empty string, acts as ANYWHERE legacy option.
    */
   me.unsubscribe = function(callback, global, subscriber, event, publisher)
   {
      var msgId = getNextMessageId();
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
            
      pendingUnsubscriptions[msgId] = 
      {
         "event"         : event,
         "subscriber-id" : subscriber
      };
      
      var payload = 
      {
         "global"     : global,
         "subscriber" : subscriber,
         "event"      : event,
         "publisher"  : publisher
      };
      
      var msg = {};
      msg["type"]    = this.UNSUBSCRIBE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["noError"] = noError;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a PUBLISH request.
    *
    * @param   {string} event
    *          The event name to PUBLISH.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.
    * @param   {array} args
    *          An array with the arguments for this event.
    */
   me.publish = function(event, publisher, args)
   {
      // no callbacks for this api, as no response is sent back
      var msgId = getNextMessageId();
      
      if (args == undefined)
      {
         args = [];
      }

      var payload = 
      {
         "event"      : event,
         "publisher"  : publisher,
         "args"       : args
      };
      
      var msg = {};
      msg["type"]    = this.PUBLISH;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a APPLY event TO widget-id. request.
    *
    * @param   {string} event
    *          The event name to APPLY.
    * @param   {integer} widgetId
    *          The widget ID (use HWND attribute to compute this on server-side).
    */
   me.applyEvent = function(event, widgetId)
   {
      // no callbacks for this api, as no response is sent back
      var msgId = getNextMessageId();
      
      var payload = 
      {
         "event"    : event,
         "widgetId" : widgetId,
      };
      
      var msg = {};
      msg["type"]    = this.APPLY_EVENT;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a CREATE-RESOURCE request.
    *
    * @param   {function} callback
    *          A callback with the response for the CREATE-RESOURCE request. 
    */
   me.createResource = function(callback)
   {
      // no callbacks for this api, as no response is sent back
      var msgId = getNextMessageId();
      
      // 'callback' is used to save the resource ID 
      if (callback)
      {
         callbacks[msgId] = callback;
      }

      var msg = {};
      msg["type"]    = this.CREATE_RESOURCE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;

      this.send(msg);
   };

   /**
    * This API creates a INVOKE message request for an external program, and sends it to the
    * target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An external program name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    */
   me.invokeExternalProgram = function(callback, name, args)
   {
      invoke(callback, name, this.EXTERNAL_PROGRAM, args, null, true);
   };

   /**
    * This API creates a INVOKE message request for an external program in persistent mode, and 
    * sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An external program name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    */
   me.invokeExternalProgramPersistent = function(callback, name, args)
   {
      invoke(callback, name, this.EXTERNAL_PROGRAM, args, null, true);
   };

   /**
    * This API creates a INVOKE message request for a function, and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          A function name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle. 
    */
   me.invokeFunction = function(callback, name, args, handle)
   {
      invoke(callback, name, this.FUNCTION, args, handle, true);
   };

   /**
    * This API creates a INVOKE message request for an internal procedure, and sends it to the 
    * target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An internal procedure name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle.
    */
   me.invokeProcedure = function(callback, name, args, handle)
   {
      invoke(callback, name, this.PROCEDURE, args, handle, false, true);
   };

   /**
     * Send a response for a REMOTE_CALL request, to the target window.
     *
     * @param   {number} msgId
     *          The request message ID.
     * @param   {string} payload
     *          The payload to send for this response.
     */
   me.sendRemoteCallResponse = function(msgId, payload)
   {
      // this is a notification that a request 4GL-to-remote has finished, and the response needs
      // to be sent back, with this payload.
      
      var msg = {};
      msg["type"]    = this.REMOTE_CALL;
      msg["mode"]    = this.RESPONSE;
      msg["id"]      = msgId;
      msg["payload"] = payload;
      
      this.send(msg);
   };
   
   /**
    * Initialize the embedded windows by retrieveing the data from adm_windows.json.
    * 
    * The configuration looks like:
    *    {
    *       "toolbar" :
    *       {
    *          "container"   : "fwdToolbar",       // ID of the toolbar element
    *          "customImage" : customImageFunc,    // function(imagePath, tooltip, label)
    *          "imageFolder" : "embedded/custom/", // the root folder to add as a prefix
    *          "handler"     : p2j.toolbar         // instance of p2j.toolbar
    *       },
    *       "menubar" :
    *       {
    *          "container" : "fwdMenubar", // ID of the menubar element
    *          "event"     : openMenuItem, // function(handleId, callback)
    *          "handler"   : p2j.menubar   // instance of p2j.menubar
    *       },
    *       "module" :
    *       {
    *          "container" : "fwdModule",        // ID of the module element
    *          "event"     : showEmbeddedWindow, // function(moduleId, windowTitle)
    *          "handler"   : p2j.modules         // instance of p2j.modules
    *       },
    *       "chipset" :
    *       {
    *          "container" : "window-chip-set", // ID of chipset element
    *          "remove"    : removeWindow,      // function(chipId)
    *          "interact"  : showWindow,        // function(chipId)
    *          "handler"   : p2j.chips          // instance of p2j.chips
    *       },
    *       "initCallback"    : hideLoading,     // function(), executed after initialization
    *       "populateWindows" : undefined,       // function(JSON) - populate the window list, if
    *                                            // not using a p2j.module
    *       "topWindowChanged": undefined,       // function(windowId), executed when the top
    *                                            // window has switched.  windowId is a HWND value. 
    *       "populatePages"   : populatePages,   // function(windowId, pagesTitles[], currentPage)
    *       "windowDestroyed" : windowDestroyed, // function(windowId) 
    *       "defaultBgColor"  : defaultBGColor,  // function(colorRgb)
    *
    *       "embedded-status-area"    : true,
    *       "show-status-callback"    : showStatus, // function(txt)
    *
    *       "embedded-message-area"   : true,
    *       "show-messages-callback"  : showMessages // function(txt[])
    *    };
    *
    * @param    {object} cfg
    *           JSON-style configuration.
    */
   me.initializeEmbeddedWindows = function(cfg)
   {
      p2j.embedded.registerIncomingRequest("admPageChanged", admPageChanged);

      if (cfg["chipset"])
      {
         p2j.embedded.initChips(cfg.chipset.container, 
                                cfg.chipset.remove,
                                cfg.chipset.interact,
                                cfg.chipset.handler);
      }
      
      if (cfg["menubar"])
      {
         p2j.embedded.initMenuBar(cfg.menubar.container, cfg.menubar.event, cfg.menubar.handler);
      }
      
      if (cfg["toolbar"])
      {
         p2j.embedded.initToolBar(cfg.toolbar.container, cfg.toolbar.handler, cfg.toolbar.imageFolder);
         if (cfg.toolbar.customImage)
         {
            cfg.toolbar.handler.setCustomImage(cfg.toolbar.customImage);
         }
      }
      
      if (cfg["module"])
      {
         p2j.embedded.initModule(cfg.module.container, cfg.module.event, cfg.module.handler);
      }
      
      if (cfg["defaultBgColor"])
      {
         defaultBgColorRgb = cfg.defaultBgColor;
      }
      
      if (cfg["topWindowChanged"])
      {
         topWindowChangedCallback = cfg.topWindowChanged;
      }
      
      var initCallback = function()
      {
         if (cfg["embedded-status-area"])
         {
            p2j.embedded.embeddedStatusArea(true);
            showStatusCallback = cfg["show-status-callback"];
         }
         if (cfg["embedded-message-area"])
         {
            p2j.embedded.embeddedMessageArea(true);
            showMessagesCallback = cfg["show-messages-callback"];
         }
         
         if (cfg["initCallback"])
         {
            cfg.initCallback();
         }
      };
      
      initilizeEmbeddedWindowsInt(initCallback,
                                  cfg["populateWindows"] ? cfg.populateWindows : function() {},
                                  cfg["populatePages"],
                                  cfg["windowDestroyed"]);
   };
   
   /**
    * Initialize the embedded windows by retrieveing the data from adm_windows.json.
    *
    * @param    {function} initCallback
    *           Callback executed when all initialization has been done.
    * @param    {function} populateWindows
    *           Callback to populate the ADM windows in the UI.
    * @param    {function} populatePages
    *           Callback to populate the pages for a ADM window in the UI.
    * @param    {function} windowDestroyed
    *           Callback executed when a window gets destroyed.
    */
    function initilizeEmbeddedWindowsInt(initCallback, 
                                         populateWindows,
                                         populatePages,
                                         windowDestroyed)
   {
      windowDestroyedFunc = windowDestroyed;
      
      var moduleWindows = function(windowDetails)
      {
         var fcb = function(msg)
         {
            var payload = msg.payload;
            
            var callCallback = false;
            
            if (payload["error"] || payload["quit"] || payload["stop"])
            {
               populateWindows(windowDetails);
               callCallback = true;
            }
            else
            {
               // destroy it
               if (moduleHandler)
               {
                  // on first call it's already empty
                  moduleHandler.destroy();
               }

               var jsonModules;
               try
               {
                  jsonModules = JSON.parse(payload.return);
               }
               catch(e)
               {
                  // ignore
               }

               if (jsonModules)
               {
                  if (moduleHandler)
                  {
                     // create it
                     moduleHandler.create(jsonModules);
                     callCallback = true;
                  }
                  else
                  {
                     // load the modules
                     require(["p2j.module", "dojo/dom", "dojo/dom-construct", "dojo/on", "dojo/domReady!"],
                             function(fwdModule, dom, domConstruct, on)
                             {
                                moduleHandler = fwdModule;
                                moduleHandler.init(moduleContainer, moduleEvent);
                                
                                // create it
                                moduleHandler.create(jsonModules);

                                if (initCallback)
                                {
                                   initCallback();
                                }
                             });
                  }
               }
               else
               {
                  populateWindows(windowDetails);
                  callCallback = true;
               }
            }
            
            if (callCallback && initCallback)
            {
               initCallback();
            }
         };
         
         p2j.embedded.invokeFunction(fcb, "inMsg_fwdGetModules", {}, fwdEmbeddedDriverHandle);
      }
   
      // TODO: prefix the json file with the basepath, /embedded/custom/, but make it configurable
      $.ajax({
         url      : getBasePath() + "/adm_windows.json",
         type     : "GET",
         async    : true,
         dataType : "json",
         
         success  : 
         function(data)
         {
            loadEmbeddedWindows(data, moduleWindows, populatePages);
         },
         
         error    : 
         function (xhr, textStatus, errorThrown)
         {
            console.error("ERROR (ajax get): " + xhr.status + " - " + errorThrown);
         },
      });
   };

   /**
    * Get the base path for this embedded web app.
    *
    * @return    See above.
    */
   function getBasePath()
   {
      return window.location.pathname == "/" ? "" : window.location.pathname;
   };
   me.getBasePath = getBasePath;

   /**
    * Sets the wait-for-input callback.
    *
    * @param    {function} waitForInputCallback
    *           Called when the UI enters a "wait-for-event" loop during window switch. This indicates that
    *           user input is required and the UI logic should take appropriate actions (for example make the
    *           4GL iframe accessible to the user to allow him take any required actions.).
    */
   me.setWaitForInputCallback = function(waitForInputCallback)
   {
      me.waitForInputCallback = waitForInputCallback;
   };

   /**
    * Logout from the application.
    *
    * @param   {function} callback
    *          Callback executed when the API returns.
    */
   me.logout = function(callback)
   {
      p2j.embedded.invokeProcedure(callback, "inMsg_logout", {}, fwdEmbeddedDriverHandle);
   };
   
   /**
    * Get the menubar definition
    *
    * @param    {function} callback
    *           Callback to be executed after the menubar has been retrieved, 
    *           with the menubar definition as first argument, a JSON-encoded string.
    */
   me.getMenuBar = function(callback)
   {
      var fcb = function(msg)
      {
         var mbJson = msg.payload.return;
      
         if (callback)
         {
            callback(mbJson);
         }
      };
      
      p2j.embedded.invokeFunction(fcb, "inMsg_fwdGetMenuBar", {}, fwdEmbeddedDriverHandle);
   };
   
   /**
    * Get the toolbar definition
    *
    * @param    {function} callback
    *           Callback to be executed after the toolbar has been retrieved, 
    *           with the toolbar definition as first argument, a JSON-encoded string.
    */
   me.getToolBar = function(callback)
   {
      var fcb = function(msg)
      {
         var tbJson = msg.payload.return;
      
         if (callback)
         {
            callback(tbJson);
         }
      };
      
      p2j.embedded.invokeFunction(fcb, "inMsg_fwdGetToolBar", {}, fwdEmbeddedDriverHandle);
   };
   
   /*
    * Check if the specified window ID is an embedded window.
    *
    * @param    {string} windowId
    *           The window ID (program name).
    *
    * @return   <code>true</code> if there is a key in the admWindows for the window ID.
    */
   me.isEmbeddedWindow = function(windowId)
   {
      return admWindows[windowId] ? true : false;
   };
   
   /*
    * Check if the specified window ID is an embedded window already loaded.
    *
    * @param    {string} windowId
    *           The window ID (program name).
    *
    * @return   <code>true</code> if there is a HWND ID in the windowDetails for the given program.
    */
   me.isLoadedWindow = function(windowId)
   {
      return windowDetails[windowId] ? windowDetails[windowId] : undefined;
   };

   /*
    * If the current ADM window is the specified window, then destroy it.
    *
    * @param    {string} windowId
    *           The window ID (FWD internal identifier as reported by HWND attribute).
    * @param    {boolean} force
    *           If this is a registered window, then force its destroy.
    */
   me.destroyWindow = function(windowId, force)
   {
      if ((force && windowDetails[windowId] != undefined) || windowId === currentFWDWindowId)
      {
         var fdw = function(msg)
         {
            if (windowId === currentFWDWindowId)
            {
               if (menuHandler)
               {
                  menuHandler.destroy();
               }
         
               if (toolbarHandler)
               {
                  toolbarHandler.destroy();
               }
               
               currentFWDWindowId = undefined;
               currentADMWindow = undefined;

               if (showStatusCallback)
               {
                  showStatusCallback("");
               }
               if (showMessagesCallback)
               {
                  showMessagesCallback("");
               }
            
               if (windowDestroyedFunc)
               {
                  windowDestroyedFunc(windowId);
               }
            }
            
            var pname = windowDetails[windowId]["pname"];
            delete windowDetails[windowId];
            delete windowDetails[pname];
            
            p2j.emain.hideLoading();
         };
         
         var args =
         {
            "windowId" : windowId
         };

         p2j.emain.showLoading();

         p2j.embedded.invokeProcedure(fdw, "inMsg_fwdDestroyWindow", args,
                                      fwdEmbeddedDriverHandle);
      }
   };

   /*
    * Switch to the specified window ID (as returned by HWND).  If this window is not yet
    * initialized, it will be initialized.
    *
    * @param    {string} windowId
    *           The window ID (FWD internal identifier as reported by HWND attribute).
    * @param    {function} switchCallback
    *           Callback to execute once switching is done.
    */
   me.switchWindow = function(windowId, switchCallback)
   {
      if (admWindowSwitch || (windowId != undefined && currentFWDWindowId == windowId))
      {
         if (switchCallback)
         {
            switchCallback();
         }
         
         p2j.emain.hideLoading();
         
         return;
      }
      
      admWindowSwitch = true;
      
      var fcb = function(msg)
      {
         var payload = msg.payload;
         
         var pname = payload["arguments"]["windowprogram"];
         var wtitle = payload["arguments"]["windowtitle"];
         
         if (pname == undefined)
         {
            // couldn't switch...
            p2j.emain.hideLoading();
            return;
         }

         currentFWDWindowId = parseInt(windowId);
            
         if (windowDetails[windowId])
         {
            var wd = windowDetails[windowId];
            
            if (wd["currentPage"] != undefined && wd["pages"] == undefined)
            {
               delete windowDetails[windowId];
               delete windowDetails[pname];

               p2j.emain.showCurrentWindow(pname, wtitle, wd["currentPage"]);
            }
            else
            {
               currentADMWindow = wd["pname"];
               
               if (showStatusCallback)
               {
                  showStatusCallback(wd["statusText"]);
               }
               
               if (showMessagesCallback)
               {
                  showMessagesCallback(wd["messageText"]);
               }
               
               if (toolbarHandler)
               {
                  toolbarHandler.destroy();
                  if (wd["toolbar"])
                  {
                     toolbarHandler.create(wd["toolbar"]);
                  }
               }
   
               if (menuHandler)
               {
                  menuHandler.destroy();
                  if (wd["menubar"])
                  {
                     menuHandler.create(wd["menubar"]);
                  }
               }
               
               if (wd["pages"])
               {
                  populatePagesFunc(currentADMWindow, wd["pages"], wd["currentPage"] - 1);
               }
               
               if (p2j.emain && p2j.emain.setTitle)
               {
                  p2j.emain.setTitle(wtitle);
               }
   
               p2j.emain.hideLoading();
            }
         }
         else
         {
            p2j.emain.showCurrentWindow(pname, wtitle);
         }

         if (switchCallback)
         {
            switchCallback();
         }
         
         admWindowSwitch = false;
      };
      
      p2j.emain.showLoading();

      var wd = windowDetails[windowId];
      var args = 
      {
         "pageNum" : wd && wd["currentPage"] ? wd["currentPage"] : 1,
         "windowId" : windowId
      };
      p2j.embedded.invokeProcedure(fcb, "inMsg_fwdSwitchWindow", args, fwdEmbeddedDriverHandle);
   };

   /**
    * Notification from the business logic the ADM page has changed.
    *
    * @param   {object} msg
    *          The JSON object, with 'page' and 'windowId' values.
    */
   function admPageChanged(msg)
   {
      var pageId = msg["page"];
      var windowId = msg["windowId"];
      
      setWindowDetail(windowId, "currentPage", pageId);
      
      if (windowId == currentFWDWindowId)
      {
         p2j.emain.admPageChanged(msg);
      }
   };
   me.admPageChanged = admPageChanged;

   /**
    * Show the specified window.
    *
    * @param   {string} windowId
    *          The window key (program name) in <code>admWindows</code>.
    * @param   {function} callbackCreated
    *          Callback executed when the API returns, for the created window.
    * @param   {string} windowTitle
    *          Optional, the title to use, if set.
    * @param   {number} pageNum
    *          If set, the 1-based page number to select.
    */
   me.showWindow = function(windowId, callbackCreated, windowTitle, pageNum)
   {
      if (windowDetails[windowId] != undefined)
      {
         p2j.emain.switchWindow(windowDetails[windowId]);
         return;
      }
   
      currentADMWindow = windowId;
      admWindowSwitch = true;
      
      if (menuHandler)
      {
         menuHandler.destroy();
      }

      if (toolbarHandler)
      {
         toolbarHandler.destroy();
      }
   
      p2j.embedded.showWindowPage(pageNum ? pageNum : 1,
         function ()
         {
            setWindowDetail(currentFWDWindowId, "title", windowTitle);
            setWindowDetail(currentFWDWindowId, "pname", currentADMWindow);
            p2j.emain.setTitle(windowTitle);
            
            admWindowSwitch = false;
            
            if (menuContainer)
            {
               p2j.embedded.getMenuBar(function(menubar)
               {
                  var jsonMenubar;
                  try
                  {
                     jsonMenubar = JSON.parse(menubar);
                  }
                  catch(e)
                  {
                     // ignore
                  }
   
                  if (jsonMenubar)
                  {
                     setWindowDetail(currentFWDWindowId, "menubar", jsonMenubar);

                     if (menuHandler)
                     {
                        // create it
                        menuHandler.create(jsonMenubar);
                     }
                     else
                     {
                        require(["p2j.menu", "dojo/dom", "dojo/dom-construct", "dojo/on", "dojo/domReady!"],
                                function(fwdMenu, dom, domConstruct, on)
                                {
                                   menuHandler = fwdMenu;
                                   menuHandler.init(menuContainer, menuEvent);
                                   
                                   // create it
                                   menuHandler.create(jsonMenubar);
   
                                   if (callbackCreated)
                                   {
                                      callbackCreated();
                                   }
                                });
                        
                        return;
                     }
                  }
   
                  if (callbackCreated)
                  {
                     callbackCreated();
                  }
               });
            }
            else
            {
               if (callbackCreated)
               {
                  callbackCreated();
               }
            }
         });
   };

   /**
    * Show the specified page in the current ADM window.
    *
    * @param   {integer} pageNum
    *          The page number.
    * @param   {function} callback
    *          Callback executed when the API returns.
    */
   me.showWindowPage = function(pageNum, callback)
   {
      var bgcolor = defaultBgColorRgb;
      if (bgcolor != undefined)
      {
         bgcolor = parseInt(bgcolor.substring(1), 16);
      }
      else
      {
         bgcolor = -1;
      }
      var args = 
      {
         "pfile"      : currentADMWindow,
         "pnum"       : pageNum,
         "module"     : modules[currentADMWindow],
         "bgColorRgb" : bgcolor
      };
      
      var fcb = function(msg)
      {
         if (msg.payload.arguments["windowid"] == undefined || 
             isNaN(parseInt(msg.payload.arguments.windowid)))
         {
            currentADMWindow = undefined;
            currentFWDWindowId = undefined;
            admWindowSwitch = false;
            
            p2j.emain.hideLoading();
            p2j.emain.windowDestroyed();
            
            return;
         }

         currentFWDWindowId = parseInt(msg.payload.arguments.windowid);
         windowDetails[currentADMWindow] = currentFWDWindowId;
         
         var pageNames = msg.payload.arguments.pages;
         var pages = JSON.parse("[" + pageNames + "]");

         setWindowDetail(currentFWDWindowId, "currentPage", pageNum);
         setWindowDetail(currentFWDWindowId, "pages", pages);

         populatePagesFunc(currentADMWindow, pages, pageNum - 1);
         
         p2j.embedded.initWindowPage(callback);
      };

      p2j.embedded.invokeProcedure(fcb, "inMsg_fwdShowWindow", args, fwdEmbeddedDriverHandle);
   };
   
   me.initWindowPage = function(callback)
   {
      if (toolbarContainer)
      {
         p2j.embedded.getToolBar(function(toolbar)
         {
            if (toolbarHandler)
            {
               toolbarHandler.destroy();
            }

            var jsonToolbar;
            try
            {
               jsonToolbar = JSON.parse(toolbar);
            }
            catch(e)
            {
               // ignore
            }

            if (jsonToolbar)
            {
               setWindowDetail(currentFWDWindowId, "toolbar", jsonToolbar);
            
               if (toolbarHandler)
               {
                  // create it
                  toolbarHandler.create(jsonToolbar);
               }
               else
               {
                  require(["p2j.toolbar", "dojo/dom", "dojo/dom-construct", "dojo/on", "dojo/domReady!"],
                          function(fwdToolBar, dom, domConstruct, on)
                          {
                             toolbarHandler = fwdToolBar;
                             toolbarHandler.init(toolbarContainer, toolbarEvent);
                             
                             // create it
                             toolbarHandler.create(jsonToolbar);

                             if (callback)
                             {
                                callback();
                             }
                          });
                  
                  return;
               }
            }
            
            if (callback)
            {
               callback();
            }
         });
      }
      else
      {
         if (callback)
         {
            callback();
         }
      }
   };
   
   /**
    * Get the current ADM window.
    *
    * @return   The <code>currentADMWindow</code>.
    */
   me.getCurrentWindow = function()
   {
      return currentADMWindow;
   };
   
   /**
    * Get the current HWND window ID..
    *
    * @return   The <code>currentFWDWindowId</code>.
    */
   me.getCurrentWindowId = function()
   {
      return currentFWDWindowId;
   };
   
   /**
    * If the current window is an ADM tabbed window in "admWindows", return its "window-title".
    *
    * @return    The ADM window's title.
    */
   me.getCurrentWindowTitle = function()
   {
      return admWindows[currentADMWindow] ? admWindows[currentADMWindow]["window-title"] : "";
   };
   
   /**
     * Set the current ADM window.
     *
     * @param    {string} windowId
     *           The current window ID.
     */
   me.setCurrentWindow = function(windowId)
   {
      currentADMWindow = windowId;
   };
   
   /**
    * Get the current page in the current ADM window.
    *
    * @return   The current ADM window's page.
    */
   me.getCurrentPage = function()
   {
      return windowDetails[currentFWDWindowId]["currentPage"];
   };
   
   /**
    * Initialize the embedded driver and load the embedded windows.  
    * API executed once the remote side sends the ADM window data.
    *
    * @param    {array} data
    *           The ADM windows data.
    * @param    {function} populateWindows
    *           Callback to populate the ADM windows in the UI.
    * @param    {function} populatePages
    *           Callback to populate the pages for a ADM window in the UI.
    */
   function loadEmbeddedWindows(data, populateWindows, populatePages)
   {
      populateWindowsFunc = populateWindows;
      populatePagesFunc = populatePages;

      admWindows = {};
      for (var i = 0; i < data.length; i++)
      {
         var program = data[i]["program-name"];
         
         admWindows[program] = data[i];
      }

      currentADMWindow = null;
      
      p2j.embedded.invokeExternalProgramPersistent(
         function(msg)
         {
            fwdEmbeddedDriverHandle = msg.payload.handle;
            
            populateWindowsFunc(data);
         },
         "fwd-embedded-driver.p",
         {});
   };

   /**
    * Generate a new message ID for a request into the P2J client.
    *
    * @return   a message ID: always a positive value.
    */
   function getNextMessageId()
   {
      // the messages incoming from P2J have a negative value; the messages from other windows
      // must have a positive value
      var msgId = nextMessageId + 1;
      
      if (msgId < 0)
      {
         msgId = 1;
      }
      
      nextMessageId = msgId;
      
      return msgId;
   };
   me.getNextMessageId = getNextMessageId;

   /**
    * Creates the overlay panel that must cover the displayed page.
    * 
    * @param    {String} id
    *           The element unique identifier.
    * @param    {Number} zIndex
    *           The maximal z-index calculated for the current page.
    * @param    {HtmlElement} container
    *           The container element.
    * @param    cls
    *           The overlay panel css name.
    * 
    * @return   The overlay panel.
    */
   function createOverlayPanel(id, zIndex, container, cls)
   {
      var overlayPanel = document.getElementById(id);

      if (!overlayPanel)
      {
         overlayPanel    = document.createElement("div");
         overlayPanel.id = id;
         overlayPanel.className = cls;
         container.appendChild(overlayPanel);
      }

      overlayPanel.style.position = "fixed";
      overlayPanel.style.top      = "0px";
      overlayPanel.style.left     = "0px";
      overlayPanel.style.width    = "100%";
      overlayPanel.style.height   = "100%";

      overlayPanel.style.background = "black";
      overlayPanel.style.opacity    = 0.5;
      overlayPanel.style.display    = "none";
      overlayPanel.style.zIndex     = zIndex;

      return overlayPanel;
   };

   /**
    * Calculates the maximal z-index on the current page.
    * 
    * @return   The maximal z-index on the current page.
    */
   function getMaxZIndex()
   {
      var zIndex;
      var maxZIndex = 1;
      var all  = document.getElementsByTagName('*');
      for (var i = 0, n = all.length; i < n; i++)
      {
          zIndex = window.getComputedStyle(all[i], null).getPropertyValue("z-index");
          zIndex = parseInt(zIndex, 10);
          if (zIndex != null && isFinite(zIndex))
          {
             maxZIndex = Math.max(maxZIndex, zIndex);
          }
      }
      
      return maxZIndex;
   };
   
   /**
    * Calculates the url origin. It is supposed that url has "https" protocol part. It retrieves
    * the url domain part and forms the origin. If the url doesn't satisfy this assumption, then
    * it returns "*" as an origin indicating any origin is accepted.
    * 
    * @param    {String} url
    *           The url path.
    * 
    * @return   The url origin.
    */
   function getUrlOrigin(url)
   {
      var protocol = "https";
      if (url != null && url.indexOf(protocol) >= 0)
      {
         var components = url.substr(8).split("/");
         return protocol + "://" + components[0];
      }
      else
      {
         return "*";
      }
   };
   
   /**
    * This API creates a INVOKE message request and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          The legacy name of the target external program, function or procedure. 
    * @param   {number} type
    *          The target's type: EXTERNAL_PROGRAM - (0), PROCEDURE - (1) or FUNCTION - (2) 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle.
    *          Only used by functions or internal procedure calls.
    * @param   {boolean} persistent
    *          Only used when invoking external programs; it indicates if it is ran persistent.
    * @param   {boolean} needsNotificationOnWait
    *          The flag indicates that the server should send a notification when 4GL program execution
    *          enters modal or wait-for event loop during the procedure invocation.
    */
   function invoke(callback, name, type, args, handle, persistent, needsNotificationOnWait)
   {
      var msgId = getNextMessageId();
      var vargs = {};
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      for (var arg in args)
      {
         vargs[arg] = "" + args[arg];
      }
      
      var payload = {};
      payload["type"]       = type;
      payload["name"]       = name;
      payload["persistent"] = persistent;
      payload["handle"]     = (handle ? handle : "");
      payload["arguments"]  = vargs;
      payload["noError"]    = noError;

      if (needsNotificationOnWait)
      {
         payload["needsNotificationOnWait"] = true;
      }

      var msg = {};
      msg["type"]    = p2j.embedded.INVOKE;
      msg["mode"]    = p2j.embedded.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = payload;

      p2j.embedded.send(msg);
      
      return msgId;
   };

   /**
    * Handler for message events.
    *
    * @param   {object} event
    *          The event object.
    */
   function receiveRemoteMessage(event)
   {
      /**
        * Worker to handle the specified message.
        * <p>
        * If a custom callback is set for this message ID in the <code>p2j.embedded.callbacks</code>
        * map, then that will be executed.  Otherwise, the global 
        * <code>p2j.embedded.messageCallback</code> callback will be executed.  If not set, a message
        * is logged in the JS console.
        *
        * @param   {object} msg
        *          The message to handle.
        */
      function handleMessage(msg)
      {
         var msgId = msg.id;
         var msgCallback = p2j.embedded.getHandler(msgId);

         if (msgCallback)
         {
            msgCallback(msg);
         }
         else
         {
            var smsg = JSON.stringify(msg);
            console.log("No callback is set for the message: " + smsg);
         }
      };
      
      if (event.cancelBubble || (event.isPropagationStopped && event.isPropagationStopped()))
      {
         // already handled
         return;
      }
      
      // WARNING! here "this" will be the Window object (not "me"!)

      // For Chrome, the origin property is in the event.originalEvent object.
      var origin = event.origin || event.originalEvent.origin;

      var msg = event.data;

      // if the message is a request for PAGE_LOADED, then check that the source is the same as
      // the targetWindow and the secret
      if (msg.type == p2j.embedded.PAGE_LOADED)
      {
         if (p2j.embedded.handleMessages)
         {
            var smsg = JSON.stringify(msg);
            console.log("p2j.embedded received a PAGE_LOADED request after  it was initialized: " +
                        smsg);
            return;
         }

         if (msg.mode == p2j.embedded.REQUEST)
         {
            if (event.source == p2j.embedded.targetWindow)
            {
               p2j.embedded.handleMessages = true;

               var secret = msg.payload;
               
               // send back a notification with this page's uri
               p2j.embedded.targetOrigin = origin;
               
               var msg = {};
               msg["type"] = p2j.embedded.PAGE_LOADED;
               msg["mode"] = p2j.embedded.RESPONSE;
               msg["payload"] = secret;
               
               p2j.embedded.send(msg);
               
               if (p2j.embedded.pageLoadedCallback)
               {
                  p2j.embedded.pageLoadedCallback(origin);
               }
            }
            else
            {
               console.log("Received a PAGE_LOADED message from incorrect window: " + origin);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a PAGE_LOADED message with " + msgMode + 
                        " mode - remote side can handle only RESPONSE modes!");
         }
         
         return;
      }
      else if (msg.type == p2j.embedded.PAGE_CLOSED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.targetOrigin = "*";
            p2j.embedded.handleMessages = false;

            // this is a request for a PAGE_CLOSED notification (no response is required)
            if (p2j.embedded.pageClosedCallback)
            {
               p2j.embedded.pageClosedCallback(origin);
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a PAGE_CLOSED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
         
         return;
      }
      else if (msg.type == p2j.embedded.SOCKET_RECONNECTED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.handleMessages = true;

            // this is a request for a SOCKET_RECONNECTED notification (no response is required)
            if (p2j.embedded.socketReconnectedCallback)
            {
               p2j.embedded.socketReconnectedCallback();
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a SOCKET_RECONNECTED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }

         return;
      }
      else if (msg.type == p2j.embedded.SOCKET_DISCONNECTED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.handleMessages = false;

            // this is a request for a SOCKET_DISCONNECTED notification (no response is required)
            if (p2j.embedded.socketDisconnectedCallback)
            {
               p2j.embedded.socketDisconnectedCallback();
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a SOCKET_DISCONNECTED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
         
         return;
      }
      else if (msg.type == p2j.embedded.WAIT_USER_INPUT)
      {
         if (me.waitForInputCallback)
         {
            me.waitForInputCallback();
         }
         return;
      }

      if (!p2j.embedded.handleMessages)
      {
         var smsg = JSON.stringify(msg);
         console.log("p2j.embedded received a message before it was initialized: " + smsg);
         return;
      }

      if (origin !== p2j.embedded.targetOrigin)
      {
         var smsg = JSON.stringify(msg);

         console.log("Received message " + smsg + " from unexpected window origin: " + origin +
                     "(expected origin is " + p2j.embedded.targetOrigin + ")");
         return;
      }
      
      var types = 
      [
         p2j.embedded.REMOTE_CALL,     // 0
         p2j.embedded.INVOKE,          // 1
         p2j.embedded.DELETE_HANDLE,   // 2
         p2j.embedded.SUBSCRIBE,       // 3
         p2j.embedded.UNSUBSCRIBE,     // 4
         p2j.embedded.PUBLISH,         // 5
         p2j.embedded.CREATE_RESOURCE, // 6
         // p2j.embedded.APPLY_EVENT,     // 7 - not received, just sent
         p2j.embedded.SET_STATUS_TEXT, // 8
         p2j.embedded.ADD_MESSAGES,    // 9
         p2j.embedded.CLEAR_MESSAGES   // 10
      ];
      var modes =
      [
         p2j.embedded.REQUEST,  // 0
         p2j.embedded.RESPONSE, // 1
         p2j.embedded.RESPONSE, // 2
         p2j.embedded.RESPONSE, // 3
         p2j.embedded.RESPONSE, // 4
         p2j.embedded.REQUEST,  // 5
         p2j.embedded.RESPONSE,  // 6
         // p2j.embedded.REQUEST,  // 7
         p2j.embedded.REQUEST,  // 8
         p2j.embedded.REQUEST,  // 9
         p2j.embedded.REQUEST   // 10
      ];
      var customCallbacks = 
      [
         undefined,           // 0
         undefined,           // 1
         undefined,           // 2
         subscribeCallback,   // 3
         unsubscribeCallback, // 4
         publishCallback,     // 5
         undefined,           // 6
         // undefined,           // 7
         statusTextCallback,   // 8
         addMessagesCallback,  // 9
         clearMessagesCallback // 10
      ];
      
      var found = false;
      
      // handle other message types
      for (var i = 0; i < types.length; i++)
      {
         if (msg.type == types[i])
         {
            if (msg.mode == modes[i])
            {
               if (customCallbacks[i] != undefined)
               {
                  customCallbacks[i](msg);
               }

               handleMessage(msg);
            }
            else
            {
               var msgMode = (msg.mode ? msg.mode : "undefined");
               console.log("Received a " + types[i] + " message with " + msgMode + 
                           " mode - remote side can handle only " + modes[i] + " modes!");
            }
            
            found = true;
            break;
         }
      }
      
      if (!found)
      {
         // unknown payload, discard
         var data = JSON.stringify(msg);
         console.log("remote side has received an unknown message, and will be discarded: " + data);
      }
   };
   
   /**
    * Notification that the status for a window has been set.
    *
    * @param    {object} msg
    *           The object with 'text' and 'window' keys.
    */
   function statusTextCallback(msg)
   {
      if (showStatusCallback && msg.window == currentFWDWindowId)
      {
         setWindowDetail(msg.window, "statusText", msg.text);
         showStatusCallback(msg.text);
      }
   };
   
   /**
    * Notification that the message for a window has been set.
    *
    * @param    {object} msg
    *           The object with 'text' and 'window' keys.
    */
   function addMessagesCallback(msg)
   {
      if (showMessagesCallback && msg.window == currentFWDWindowId)
      {
         setWindowDetail(msg.window, "messageText", msg.text);
         showMessagesCallback(msg.text);
      }
   };
   
   /**
    * Notification that the message for a window has been cleared.
    *
    * @param    {object} msg
    *           The object with 'window' key.
    */
   function clearMessagesCallback(msg)
   {
      if (showMessagesCallback && msg.window == currentFWDWindowId)
      {
         setWindowDetail(msg.window, "messageText", undefined);
         showMessagesCallback("");
      }
   };
   
   /**
    * Save the specified window detail.
    *
    * @param    {string} windowId
    *           The window ID (program name).
    * @param    {string} detailKey
    *           The detail key (i.e. "pages", "toolbar", etc).
    * @param    {string} detailVal
    *           The value.
    */
   function setWindowDetail(windowId, detailKey, detailVal)
   {
      if (windowDetails[windowId] == undefined)
      {
         windowDetails[windowId] = {};
      }
      windowDetails[windowId][detailKey] = detailVal;
   };
   me.setWindowDetail = setWindowDetail;
   
   /**
    * Callback executed when a UNSUBSCRIBE response is received.
    * <p>
    * If "msg.payload.error == true", then it will not remove the event.  Otherwise, it will
    * be removed from the "subscriptions" register.
    * 
    * @param    {object} msg
    *           The response message.
    */
   function unsubscribeCallback(msg)
   {
      var payload = msg["payload"];
   
      var msgId = msg["id"];
      
      var ps = pendingUnsubscriptions[msgId];
      
      delete pendingUnsubscriptions[msgId];

      if (payload["error"])
      {
         // unsubscribe failed, do not remove
         return;
      }

      var event = ps["event"];
      var subId = ps["subscriber-id"];

      var subs = subscriptions[subId];
      delete subs[event];
   };
   
   /**
    * Callback executed when a SUBSCRIBE response is received.
    * <p>
    * If "msg.payload.error == true", then it will not register the event.  Otherwise, it will
    * be added to the "subscriptions" register.
    * 
    * @param    {object} msg
    *           The response message.
    */
   function subscribeCallback(msg)
   {
      var payload = msg["payload"];

      var msgId = msg["id"];
      
      var ps = pendingSubscriptions[msgId];
      
      delete pendingSubscriptions[msgId];

      if (payload["error"])
      {
         // subscription failed, do not register
         return;
      }

      var method = ps["method"];
      var event = ps["event"];

      var subId = payload["subscriber-id"];
      if (subscriptions[subId] == undefined)
      {
         subscriptions[subId] = {};
      }
      
      subscriptions[subId][event] = method;
   };
   
   /**
    * Callback executed when the PUBLISH request is received.
    * If there is a subscription for the "msg.payload.subscriber-id" and "msg.payload.event"
    * in the "subscriptions" register, then that callback will be executed.
    * 
    * @param    {object} msg
    *           The request message.
    */
    function publishCallback(msg)
    {
      var payload = msg["payload"];
      var event   = payload["event"];
      var subId   = payload["subscriber-id"];
      var params  = payload["parameters"];
      
      if (subscriptions[subId])
      {
         var subscriberCallback = subscriptions[subId][event];
         if (subscriberCallback)
         {
            subscriberCallback(params);
         }
      }
   };

   return me;
});