/*
** Module   : p2j.menu.js
** Abstract : Defines menubar navigation widget.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 CA  20180419 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};

   /** The ID of the DOM element where to attach the module. */
   var menuContainer;
   
   /** The callback executed when a event is performed. */
   var menuEventCallback;
   
   /** The menu object. */
   var pmenu;

   /** The animation speed (millis). */
   var speed = 200;

   /** Left-side padding for children. */
   var padding = 15;
   
   /**
    * Initialize the menu.
    *
    * @param    {string} vmenuContainer
    *           The ID of the DOM element where to attach the menu.
    * @param    {function} vmenuEventCallback
    *           The callback executed when a menu-related event is performed (menu-drop, choose or 
    *           value-changed).
    */
   me.init = function(vmenuContainer, vmenuEventCallback)
   {
      menuContainer = $("#" + vmenuContainer);
      menuEventCallback = vmenuEventCallback;
   };

   /*
    * Destroy the menu.
    */
   me.destroy = function()
   {
      menuContainer.empty();
   };

   /*
    * Create a menubar and attach it, with the specified details.
    *
    * @param   {array} jsonMenubar
    *          The array with menu specifications.
    */
   me.create = function(jsonMenubar)
   {
      me.destroy();
   
      buildMenuTree(menuContainer, jsonMenubar["menu-items"], 1);

      if (menuEventCallback)
      {
         menuContainer.find("a").click(function(evt)
         {
            var a = $(this);
            if (a.attr("handle"))
            {
               menuEventCallback(a.attr("handle"));
            }
         });
         menuContainer.find("input").click(function(evt)
         {
            var a = $(this).parent();
            if (a.attr("handle"))
            {
               menuEventCallback(a.attr("handle"));
            }
         });
      }
   };

   /**
    * Build the collapsable menubar tree.
    * 
    * @param    {Object} mparent
    *           The parent element where children are attached.
    * @param    {array} menuItems
    *           The array tree with menu specifications.
    * @param    {integer} lvl
    *           The current nesting level.
    */
   function buildMenuTree(mparent, menuItems, lvl)
   {
      if (menuItems == undefined)
      {
         return mparent;
      }

      var pList = $("<ul/>");
      mparent.append(pList);

      for (var i = 0; i < menuItems.length; i++)
      {
         var menuItem = menuItems[i];

         var msubtype  = menuItem["subtype"];
         var mhandle   = menuItem["handle"];
         var mtype     = menuItem["type"];
         var disabled  = menuItem["sensitive"] == false;
         var accel     = menuItem["accelerator"];
         var menuLabel = (menuItem["label"] ? menuItem["label"] : " ");;
         var mnemonic;
         if (menuItem["mnemonic"])
         {
            var p1 = "";
            var p2 = "";
            var p3 = "";
         
            var mnemIdx = parseInt(menuItem["mnemonic"]);
            if (mnemIdx > 0)
            {
               p1 = menuLabel.substring(0, mnemIdx);
            }
            p2 = menuLabel.substring(mnemIdx, mnemIdx + 1);
            if (mnemIdx < menuLabel.length - 1)
            {
               p3 = menuLabel.substring(mnemIdx + 1);
            }
            mnemonic = p2;
            menuLabel = p1 + "<u>" + p2 + "</u>" + p3;
         }

         var chLi = $("<li/>");
         var chA;
         // TODO: disabled
         if (mtype == "SUB-MENU")
         {
            chA = $("<a href='#'><i class='material-icons'></i>" + menuLabel + "</a>");
         }
         else
         {
            if (msubtype == "RULE")
            {
               chA = $("<div/>");
            }
            else if (msubtype == "SKIP")
            {
               chA = $("<div/>");
            }
            else
            {
               if (menuItems[i]["toggle-box"] == true)
               {
                  if (menuItems[i]["checked"] == true)
                  {
                     chA = $("<a href='#'><input type='checkbox' checked>" + menuLabel + "</input></a>");
                  }
                  else
                  {
                     chA = $("<a href='#'><input type='checkbox'>" + menuLabel + "</input></a>");
                  }
               }
               else
               {
                  chA = $("<a href='#'>" + menuLabel + "</a>");
               }
            }
         }

         if (mnemonic)
         {
            chA.attr("accesskey", mnemonic);
         }

         chA.css("padding-left", lvl * padding + (menuItem["menu-items"] ? 0 : padding));
         chA.attr("handle", mhandle);

         chLi.append(chA);
         pList.append(chLi);   

         if (mtype == "SUB-MENU")
         {
            chA.attr("data-option", "off");
            chA.find("i").text("keyboard_arrow_right");

            chA.unbind('click').on("click", function() 
            {
               var a = $(this);
               a.parent().parent().find("a[data-option='on']").parent("li").children("ul").slideUp(speed / 1.2, function() 
               {
                  $(this).parent("li").children("a").attr("data-option", "off");
                  $(this).parent("li").children("a").find("i").text("keyboard_arrow_right");
               })
               if (a.attr("data-option") == "off") 
               {
                  a.parent("li").children("ul").slideDown(speed, function()
                  {
                     $(this).parent("li").children("a").attr("data-option", "on");
                     $(this).parent("li").children("a").find("i").text("keyboard_arrow_down");
                  });
               }
               if (a.attr("data-option") == "on") 
               {
                  a.attr("data-option", "off");
                  a.find("i").text("keyboard_arrow_right");
                  a.parent("li").children("ul").slideUp(speed, function()
                  {
                     $(this).parent("li").children("a").attr("data-option", "off");
                     $(this).parent("li").children("a").find("i").text("keyboard_arrow_right");
                  });
               }
            });

            buildMenuTree(chLi, menuItem["menu-items"], lvl + 1);
         }
         else
         {
            if (msubtype == "RULE")
            {
               chLi.append($("<hr/>"));
            }
            else if (msubtype == "SKIP")
            {
               chLi.append($("&nbsp;"));
            }
         }
         
      }
   };

   return me;
});
