/*
** Module   : p2j.screen.js
** Abstract : GUI-specific screen management
**
** Copyright (c) 2014-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150325 Created GUI version from ChUI version. Right now it is not functional.
**                  Renamed objects to match the more generic web client approach (not just
**                  ChUI). Moved some logic to common code locations.
** 002 SBI 20150731 Applied "use strict" directive, fixed undeclared variables.
** 003 GES 20150720 Total rewrite to implement a first pass at GUI support,
**                  added a drawing image operation.
** 004 GES 20150818 Added more GUI support including fonts, cursors, z-order, mouse events.
**                  Fixed many problems with the initial implementation. Implemented SET_ICON.
**                  Added the draggable task bar implementation.
** 005 CA  20150925 Added explicit track of widget z-order.
**                  Finished posibility to raise a mouse event for an explicit widget (needed by
**                  mouse enter/exit events for P2J widgets).
** 006 CA  20151013 Added support for window resize, z-order operatons for the GUI web client.
**                  Fixed incorrect arguments provided for layoutParagraphWorker and findNextPos.
** 007 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151026 Fixed mouse processing for modal windows.
**         20151108 Added widget z-order logic for canProcessWidget.
**         20151112 Added mouse event dispatcher, fixed incorrect parameters number for
**                  sendMouseEvent invocations.
** 008 SBI 20151124 Fixed cleanup() to remove the corresponding window canvas from DOM.
**     SBI 20151201 Fixed two times clicks for menus.
** 009 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 010 IAS 20151214 New cursor types added.
**                  LOAD-MOUSE-POINTER ans SET-WAIT-STATE support
** 011 SBI 20151208 Added new parameters, scroll amount and scroll unit for wheel events and
**                  a mouse modifiers mask for all mouse events.
**     SBI 20151209 Changed to create task icons only for main windows (the child windows includes
**                  all modal windows), to cleanup all mouse handlers for destroyed windows.
**     SBI 20151217 Changed to add the offscreen canvas drawings.
**     CA  20160129 Added a drawing cache.  Removed the "ctx" parameter from CanvasRenderer APIs.
**                  Added parseEditorContent().
**     SBI 20160328 Changed to hide the virtual desktop and the task bar iff the web client
**                  embedded mode is on, added SET_DESKTOP_BGCOLOR paint primitive to set
**                  the desktop background color and to notify the enclosed window about it,
**                  fixed TypedArray.slice is unsupported by IE. 
** 011 HC  20160406 Overhaul of window-activation logic.
** 012 SBI 20160414 Fixed  doWindowActivationChange(...) and activateTopVisibleWindow(...).
**     SBI 20160422 The resize drawing operations must not be cached.
** 013 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
**     SBI 20160505 Fixed canProcessWidget to return false if the event has been prevented by
**                  some mouse handler.
**     SBI 20160506 Fixed to send the current window changes its minimized state.
**     SBI 20160613 Added OS actions: copy, cut, paste, maximize, minimize and close.
** 014 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
**     SBI 20160713 Fixed p2j.screen.lockMousePointer() to set the active window.
** 015 HC  20160922 Prevented mouse double click to be sent twice down to the Java client.
** 016 CA  20170104 Verify the windows in embedded mode: only the first visible non-modal, 
**                  non-shared-state window and visible modal dialogs remain visible, in z-order.
**                  Added an internal overlay panel for embedded mode.
** 017 HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 018 CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
**     SBI 20170330 Fixed unexpected dismissal of overlay in embedded mode when combo-box
**                  drop-down list is dismissed.
** 019 EVL 20170707 Adding XOR canvas invert mode, draw selection.
** 020 EVL 20170824 Adding os originated file list drop handling.
** 021 HC  20170913 Implemented runtime support for WINDOW:ALWAYS-ON-TOP and SHOW-IN-TASKBAR.
**                  Added extension widget method GET-MOUSE-POSITION. Made web js client to
**                  notify the java client when desktop (browser window) resized. This makes the
**                  Web GUI driver to report correct desktop dimensions.
** 022 SBI 20171218 Fixed activations of the task icon in the virtual desktop.
** 023 SBI 20180130 Added native implementation for text selection and cursor positioning within
**                  the given text.
** 024 SBI 20180425 Fixed the click count value for IE mouse click events.
** 025 SBI 20180531 Changed p2j.screen.error function in order to use beep by business logic only.
** 026 CA  20180527 Fixed a typo related to window height in Window c'tor function.
**     CA  20180531 Keep the window's children above it, when moving the window to top.
** 027 HC  20180712 Fixed image data not sent to JS causing garbage drawn on the screen.
** 028 CA  20180727 Embedded mode enhancements: send the window list to the remote side and 
**                  other fixes.
** 029 CA  20180803 Center horizontally all non-modal embedded windows; modal windows are centered
**                  on the parent.  Also, on browser resize, the embedded windows are recentered.
** 030 CA  20180813 The overlay panel in embedded mode must track the window visible state, not 
**                  the window life (create/destroy).
** 031 OM  20180815 Added support for drawing ellipse primitives.
** 032 CA  20180820 Fixed overlay z-order when multiple modal dialogs are shown.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.screen = (function()
{
   /** Screen object which will be returned from this initialization function. */
   var me = {};
   
   /** List of active windows. */
   var winlist = [];

   /** Windows in z-order (lowest at index 0 to highest at length - 1). */
   var zlist = [];

   /** Loaded images map */
   var loadedImages = new Map();
   
   /** The current font ID, in the js font-table. */
   var currentFont;

   /** locked mouse pointer */
   var lockedMousePointer = null;
   
   /** the id of the widget which caused the mouse pointer lock*/
   var activeDropDown = -1;

   /** wait state flag */
   var isWaitState = false;
   
   /** most recent mouse position */
   var lastMousePos = null;

   /** Indicates that the graphics cache is in use. */
   var graphicsCached;
   
   /** PaintPrimitives to ordinal mappings. Change this to match Java PaintPrimitives changes! */
   var ops = 
   {
      DRAW_STRING             : 0,
      DRAW_STRING_SCALED      : 1,
      DRAW_PARAGRAPH          : 2,
      DRAW_RECT               : 3,
      DRAW_ROUND_RECT         : 4,
      DRAW_IMAGE              : 5,
      FILL_RECT               : 6,
      FILL_ROUND_RECT         : 7,
      FILL_POLYGON            : 8,
      DRAW_LINE               : 9,
      DRAW_3D_RECT            : 10,
      FILL_3D_RECT            : 11,
      SET_COLOR               : 12,
      SET_LINE_STROKE         : 13,
      SET_FONT                : 14,
      SET_FONT_STYLE          : 15,
      TRANSLATE_PUSH          : 16,
      TRANSLATE_POP           : 17,
      CLEAR                   : 18,
      CLEAR_AREA              : 19,
      NO_CLIP                 : 20,
      CLIP                    : 21,
      FLUSH                   : 22,
      SET_LOCATION            : 23,
      RESIZE_WINDOW           : 24,
      SET_WINDOW_BOUNDS       : 25,
      DRAW_HIGHLIGHT          : 26,
      CLEAR_HIGHLIGHT         : 27,
      START_BATCH             : 28,
      END_BATCH               : 29,
      SET_TITLE               : 30,
      SET_ICON                : 31,
      SET_XOR_COMPOSITE       : 32,
      RESET_COMPOSITE         : 33,
      START_RECORDING         : 34,
      STOP_RECORDING          : 35,
      DRAW_CACHED             : 36,
      SET_DESKTOP_BGCOLOR     : 37,
      SET_XOR_MODE            : 38,
      DRAW_SELECTION          : 39,
      RESET_SELECTION         : 40,
      START_DRAW_CURSOR       : 41,
      END_DRAW_CURSOR         : 42,
      TEXT_SHIFT              : 43,
      UNDO_TEXT_SHIFT         : 44,
      FILL_RECT_AROUND_TEXT   : 45,
      DRAW_ELLIPSE            : 46,
      FILL_ELLIPSE            : 47
   };
   
   /* map cursor id to name */
   var mouseCursors = 
   {
          0: "default",
          1: "crosshair",
          2: "text",
          3: "wait",
          4: "sw-resize",
          5: "se-resize",
          6: "nw-resize",
          7: "ne-resize",
          8: "n-resize",
          9: "s-resize",
         10: "w-resize",
         11: "e-resize",
         12: "pointer",
         13: "move"
   };
   
   me.mouseCursors = mouseCursors;

   /** Flag indicating if the screen should send mouse events to the client-side. */
   me.processMouse = false;
   
   
   // reverse map the types to their property names
   for (var oi in ops)
   {
      if (ops.hasOwnProperty(oi))
      {
         ops[ops[oi]] = oi;
      }
   }
   
   /**
     * Mouse event names to their encoded values.  The IDs are the same as in the 
     * gui.driver.web.MouseOps enum.
     */
   var mouseOps =
   {
      "click"        :  1, // left-click
      "contextmenu"  :  2, // right-click
      "dblclick"     :  3, // double click
      "mousedown"    :  4, // mouse pressed
      "mouseenter"   :  5, // mouse entered
      "mouseleave"   :  6, // mouse exited
      "mousemove"    :  7, // mouse move
      "mouseover"    :  8, // mouse over
      "mouseout"     :  9, // mouse exited (and all its children)
      "mouseup"      : 10, // mouse released
      "wheel"        : 11, // mouse wheel
      "drag"         : 12  // mouse drag
   };
   me.mouseOps = mouseOps;

   /**
    * Complex mouse operations handled at the driver level.
    */
   var osOps =
   {
      "hoverable"         : 13,
      "movable"           : 14,
      "popupable_left"    : 18,
      "popupable_right"   : 19,
      "copy"              : 20,
      "cut"               : 21,
      "paste"             : 22,
      "minimize"          : 23,
      "maximize"          : 24,
      "close"             : 25
   };
   me.osOps = osOps;

   /**
    * Creates a line strokes object that plays a manager role, used to apply stroke styles
    * to lines and to the native canvas context.
    */
   var strokesManager = new LineStrokes();
   
   /**
    * Holds the virtual desktop with task bar and bottom and right drop zones.
    */
   var desktop;

   /**
    * Holds the task bar.
    */
   var taskBar;

   /**
    * Helps to deactivate/activate the external input html elements when the P2j client opens
    * the modal dialogs in the case if the web client is used as an embedded IFRAME content.
    */
   var embeddedAppController;

   /**
    * Drawing cache for intermediate widget states.  This is not kept in a per-window map to 
    * allow states to be re-used between windows.
    */
   me.globalCache = {};
   
   var mouseEventDispatcher = new MouseEventDispatcher();
   
   /** The value of the single file size limit available for uploading. Default value is 20MB. */
   var singleFileLimit = 20 * 1024 * 1024;
   
   /**
    * The value of the total files size limit available for uploading in single drop. Defailt
    * value is 100MB.
    */
   var totalFilesLimit = 100 * 1024 * 1024;
   
   /**
    * Dispatch mouse events to the registered widgets. 
    */
   function MouseEventDispatcher()
   {
      var down = false;
      var up = false;
      var targetWidget;
      var targetWindow;
      var dragTimeout = null;
      var downTimeStamp;

      document.addEventListener("mousedown", mouseDownHandler, false);
      document.addEventListener("mouseup", mouseUpHandler, false);
      document.addEventListener("mousemove", mouseMoveHandler, false);
      document.addEventListener("click", mouseClickHandler, false);
      // adding file list drop handler
      document.addEventListener("drop", mouseDropHandler, false);
      document.addEventListener("dragover", mouseDragOverHandler, false);
      window.addEventListener("resize", desktopResized, false);

      /** The id of the scheduled resize function */
      var resizeTimeout;

      /**
       * Notify the server about new desktop size when the browser window changes its size.
       */
      function desktopResized()
      {
         if (resizeTimeout)
         {
            window.clearTimeout(resizeTimeout);
         }

         resizeTimeout = window.setTimeout(
            function()
            {
               resizeTimeout = null;
               var dim = me.getDesktopDimension();
               p2j.socket.sendDesktopResized(dim.width, dim.height);
               verifyEmbeddedWindows();
            }, 500);
      };

      /**
       * Handles a document widen "mousedown" event.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseDownHandler(evt)
      {
         // clean up pending drag events (here is almost impossible here to have this)
         cancelDrag();
         
         if (p2j.isDialogDisplayed())
         {
            p2j.consumeEvent(evt);
         }
         down = true;
         up = false;
         downTimeStamp = Date.now();
         var activeWindowId = p2j.screen.getActiveWindowId();
         if (activeWindowId)
         {
            targetWindow = p2j.screen.getWindow(activeWindowId);
         }
         else
         {
            targetWindow = undefined;
         }

         if (targetWindow)
         {
            targetWidget = targetWindow.findMouseSource(evt, targetWindow.mouseAwareWidgets["mousedown"]);
         }
         dispatchEvent("mousedown", evt);
      }
      
      /**
       * Handles a document widen "mouseup" event.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseUpHandler(evt)
      {
         // clean up pending drag events
         cancelDrag();
         
         if (p2j.isDialogDisplayed())
         {
            p2j.consumeEvent(evt);
         }
         up = true;
         down = false;
         dispatchEvent("mouseup", evt);
      }

      /**
       * Handles a document widen "click" event.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseClickHandler(evt)
      {
         // clean up pending drag events
         cancelDrag();
         
         if (p2j.isDialogDisplayed())
         {
            p2j.consumeEvent(evt);
         }
         dispatchEvent("click", evt);
      }

      /**
       * Handles a document widen "mousemove" event and dispatches artificial "drag" events.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseMoveHandler(evt)
      {
         if (p2j.isDialogDisplayed())
         {
            p2j.consumeEvent(evt);
         }
         if (!down)
         {
            // clean up pending drag events
            cancelDrag();
            return;
         }

         if (!dragTimeout)
         {
            //Sends drag notifications with the frequency that doesn't exceed one event per 100 ms.
            dragTimeout = setTimeout(
                  function()
                  {
                     dragTimeout = null;
                     dispatchEvent("drag", evt);
                  }, 50);
         }
      }
      
      /**
       * Handles a document widen "dragover" event.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseDragOverHandler(evt)
      {
         evt.preventDefault();
         
         // limit the actions allowed 
         evt.dataTransfer.dropEffect = "move";
         
         return false;
      }
      
      /**
       * Handles a document widen "drop" event.
       * 
       * @param    {Event} evt
       *           The raised mouse event.
       */
      function mouseDropHandler(evt)
      {
         evt.stopPropagation();
         evt.preventDefault();

         // check for empty data transfer
         if (evt.dataTransfer == null)
         {
            return false;
         }
         
         // getting the file list
         var fileNames = [];
         var files = evt.dataTransfer.files;
         // the file to upload can differ from initial file set due to size limits
         var filesToUpload = [];
         // accumulate the total file size per single drop operation
         var totalSize = 0;
         for (var i = 0, f; f = files[i]; i++)
         {
            // preparing the file name array to send and check file sizes to match the limits
            if (f.size <= singleFileLimit && totalSize + f.size <= totalFilesLimit)
            {
               // yet another file to upload
               filesToUpload[i] = f;
               // extract the name
               fileNames[i] = f.name;
               // total size available is decreasing
               totalSize = totalSize + f.size;
            }
         }
         
         targetWindow = p2j.screen.topVisibleWindow();
         
         // detect the drop target candidate
         if (targetWindow !== undefined)
         {
            targetWidget = targetWindow.findMouseSource(evt,
                                                   targetWindow.mouseAwareWidgets["mouseenter"]);
            // after all files uploaded finalize drop target by noting it can use files
            p2j.socket.sendDropFileNotify(targetWindow.id, targetWidget, fileNames);
         }
         
         // this ensures the drop file notify message arriving before file uploading starts
         for (var i = 0, f; f = filesToUpload[i]; i++)
         {
            // sending file content with names associated
            p2j.socket.uploadFileToJava(f);
         }
         
         return false;
      }
      
      /**
       * Interrupts any pendind drag event.
       */
      function cancelDrag()
      {
         // clean up pending drag events
         if (!dragTimeout)
         {
            clearTimeout(dragTimeout);
            dragTimeout = null;
         }
      }

      /**
       * Dispatches a mouse event to the registered mouse widgets.
       * 
       * @param    {String} mouseOp
       *           The mouse events name
       * @param    {MouseEvent} evt
       *           The raised mouse event
       */
      function dispatchEvent(mouseOp, evt)
      {
         if (targetWindow === undefined)
         {
            return;
         }
         
         var mousePos = targetWindow.getMousePos(evt);
         // check if targetWidget belongs to "any mouse" widgets (listen all events)
         // than sending mouseOp message is send to targetWidget
         if (!targetWindow.anyMouseAwareWidgets.hasOwnProperty(targetWidget))
         {
            dispatchEventToAnyMouseWidgets(mouseOp, evt, mousePos);
         }
         
         if (targetWidget === undefined)
         {
            return;
         }
         
         if (!targetWindow.processOsEvents || !p2j.screen.processMouse)
         {
            return;
         }

         if (p2j.screen.hasOsEvent(evt, targetWindow, mouseOp, targetWidget))
         {
            // high-level event, do not process here
            return;
         }

         targetWindow.sendMouseEvent(evt, mousePos, targetWindow, p2j.screen.mouseOps[mouseOp], targetWidget);
      }

      /**
       * Dispatches event to any mouse widgets.
       * 
       * @param    {String} mouseOp
       *           The mouse events name
       * @param    {MouseEvent} evt
       *           The raised mouse event
       * @param    {Object} mousePos
       *           The mouse event pointer coordinates {x : x, y: y} related to the target window.
       */
      function dispatchEventToAnyMouseWidgets(mouseOp, evt, mousePos)
      {
         if (targetWindow === undefined)
         {
            return;
         }

         // if there are "any mouse" widgets, process them all
         for (var wid in targetWindow.anyMouseAwareWidgets)
         {
            if (targetWindow.anyMouseAwareWidgets.hasOwnProperty(wid))
            {
               // check if a given widget is a registered widget for mouseOp
               // to prevent sending the same event twice
               if (!targetWindow.mouseAwareWidgets[mouseOp].hasOwnProperty(wid))
               {
                  targetWindow.sendMouseEvent(evt, mousePos, targetWindow, p2j.screen.mouseOps[mouseOp], wid);
               }
            }
         }
      }
   }
   
   /**
    * Constructor for a Window class that represents a Progress 4GL window instance.
    *
    * @param   wid
    *          The window id by which this window can be uniquely identified.
    * @param   settings
    *          Optional settings for the window.
    */
   function Window(wid, settings)
   {
      /** Window id by which this window can be uniquely identified. */
      this.id = wid;
      this.visible = false;
      this.clientMove = false;
      
      var options = settings || { };
      
      options.id     = options.id    || ("pwindow_" + wid);
      options.width  = options.width || 400;
      options.height = options.height || 600;
      options.style  = options.style || { };
      
      // do NOT set the tabIndex, that causes the canvas to get a selection highlight when a
      // mouse clicks into the canvas (it is a small dotted rectangle just outside of the
      // canvas' border), to avoid this we don't set tabIndex
      // options.tabIndex = options.tabIndex || 1;
      
      // this must always be set so that the canvas element is removed from the normal layout
      // processing of the document
      options.style.position = "absolute";
      
      /** Backing canvas on which all output for this window will be rendered. */
      this.canvas = p2j.createCanvas(null, options);
      this.canvas.style.visibility = "hidden";

      this.resizeable = false;

      this.drawCache = {};

      /** The off screen canvas to be used to make batch drawing. */
      this.offscreenCanvas = document.createElement('canvas');
      this.offscreenContext = this.offscreenCanvas.getContext('2d', {alpha : true});

      setImageSmoothing(this.offscreenContext, false);

      /** Canvas 2D graphics context. */
      this.ctx = this.canvas.getContext('2d', {alpha : true});

      setImageSmoothing(this.ctx, false);

      /** List of owned windows. */
      this.children = [];
      
      /** Defines the window icon id. */
      this.iconId = -1;
      
      /** Defines the window icon's width. */
      this.iconWidth  = 0;
      
      /** Defines the window icon's height. */
      this.iconHeight = 0;
      
      /** Creates canvas renderer. */
      this.canvasRenderer = new CanvasRenderer(this.offscreenCanvas, this.offscreenContext,
                                               this.canvas, strokesManager, p2j.fonts, p2j.logger);
      
      /** 
        * Flag indicating if the window should send OS events to the client-side. The default value
        * for new created window must be true due to modal windows application logic.
        */
      this.processOsEvents = true;

      /**
        * List of widgets which are aware of mouse events. They are in z-order, with the top-most
        * widgets having a having a lower index than the bottom most widgets.  This provides a 
        * left-to-right iteration of this list.
        */
      this.mouseAwareWidgets = {};    // cannot use array here, some indexes are negative numbers
      this.anyMouseAwareWidgets = {}; // cannot use array here, some indexes are negative numbers
      this.zOrder = [];

      /** List of widgets which are aware of OS events. */
      this.osAwareWidgets = {}; // cannot use array here, some indexes are negative numbers
      
      /** Tracks cumulative number of coordinate translations in effect. */
      this.trans = 0;
      
      /**
       * The horizontal cursor position within the text content relative to the beginning of
       * text content given in pixels.
       */
      this.cursorOffsetX = 0;
      
      /**
       * The vertical cursor position within the text content relative to the beginning of
       * text content given in pixels.
       */
      this.cursorOffsetY = 0;
      
      /** The current text shift along X relative to the origin */
      this.textShiftX = 0;
      
      /** Tracks cumulative number of clip rectangle unions in effect. */
      this.clips = 0;
      
      // WARNING: following code must keep the listener registration in this order
      this.mouseHandler = new MouseHandler(this);

      /** currently hovered widget */
      this.currentHovered = null;
      
      /** currently cursor */
      this.currentMousePtr = "default";
      
      /** on cleanup callback */
      this.cleanupCallback;

      // register default mouse listeners last
      for(var mouseOp in mouseOps)
      {
         if (mouseOps.hasOwnProperty(mouseOp))
         {
            // cannot use an array here because some indexes are negative numbers
            this.mouseAwareWidgets[mouseOp] = {};
            if (mouseOp !== "drag" && mouseOp !== "click" && mouseOp !== "dblclick" && 
                mouseOp !== "mouseup" && mouseOp !== "mousedown")
            {
               this.canvas.addEventListener(mouseOp, this.raiseMouseEvent(this, mouseOp));
            }
         }
      }

      for(var osOp in osOps)
      {
         if (osOps.hasOwnProperty(osOp))
         {
           // cannot use an array here because some indexes are negative numbers
            if (!this.osAwareWidgets[osOp])
            {
               this.osAwareWidgets[osOp] = {};
            }
         }
      }
      // END WARNING
   };
   
   /**
    * Set the currently hovered widget
    * 
    * @param {Object} widget
    *        The currently hovered widget
    */
   Window.prototype.setHovered = function(widget)
   {
      this.currentHovered = widget;
   }

   /**
    * Recalculate hoverable widget for a mouse position
    * 
    * @param {Object} pos
    *        mouse position
    */
   Window.prototype.recalcHovered = function(pos)
   {
//      var rect = this.canvas.getBoundingClientRect();
//      var pos = 
//      {
//         "x" : pos.clientX - rect.left,
//         "y" : pos.clientY - rect.top
//      };
      var wid = this.findMouseSource(pos, this.mouseAwareWidgets["mouseenter"]);
      var widget = this.getHoverableWidget(wid);
      if (widget)
      {
         this.currentHovered = widget;
      }
   }
   /**
    * Determine the mouse location relative to the window's canvas.
    *
    * @param    {Event} evt
    *           The raised mouse event.
    *
    * @return   The mouse position, in an associative array.
    */
   Window.prototype.getMousePos = function(evt)
   {
      return this.mouseHandler.getMousePos(evt);
   };
   
   /**
    * Determine the topmost widget which is at the coordinates specified by the given mouse event.
    *
    * @param    {Event} evt
    *           The raised mouse event.
    * @param    {Object} widgetList
    *           The widgets where the search needs to be performed.
    *
    * @return   The topmost widget which is targeted by the event.
    */
   Window.prototype.findMouseSource = function(evt, widgetList)
   {
      var mousePos = this.getMousePos(evt);

      // go through the widgets via the z-order
      for (var idx = 0; idx < this.zOrder.length; idx++)
      {
         var wid = this.zOrder[idx];

         if (this.isInside(mousePos, wid, widgetList))
         {
            return wid;
         }
      }
      return undefined;
   };

   /**
    * Check if the widgets is in the list and mouse position is within its bounds
    * 
    * @param  {Object} mousePos
    *          The mouse position
    * @param  {Number} wid
    *          widget id
    * @param    {Object} widgetList
    *           The widgets where the search needs to be performed.
    *           
    * @returns <code>true</code> if the widgets is in the list and mouse position 
    *          is within its bounds         
    */
   Window.prototype.isInside = function(mousePos, wid, widgetList)
   {
      if (!widgetList || !(wid in widgetList))
      {
         return false;
      }

      var ma = widgetList[wid];

      return ma                            &&
             mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height;
   };

   
   /**
    * Determine if the specified widget can process this event.
    * This function returns true only if there are no other widgets in the same coordinates where
    * the mouse event was captured AND the mouse event is within the boundaries of the 
    * specified widget ID .
    * <p>
    * A more complete implementation should keep track of the widget's z-order.
    *
    * @param    {Number} widgetId
    *           The widget ID.
    * @param    {Event} evt
    *           The raised mouse event.
    * @param    {String} mouseOp
    *           The string representation of the event type.
    *
    * @return   <code>true</code> if no other widget is registered for event listening, on the
    *           same coordinates.
    */
   Window.prototype.canProcessWidget = function(widgetId, evt, mouseOp)
   {
      if (this.canvas.disabled || evt.defaultPrevented || p2j.isDialogDisplayed())
      {
         return false;
      }
   
      // determine if there are other widgets on top of this one which can process the event
      var mousePos = this.getMousePos(evt);
      var widgetZorder = this.zOrder.indexOf(widgetId);
      // go through all the registered widgets for this action
      var mouseOpListeners = this.mouseAwareWidgets[mouseOp];
      for (var wid in mouseOpListeners)
      {
         if (!mouseOpListeners.hasOwnProperty(wid))
         {
            continue;
         }
         var ma = mouseOpListeners[wid];
         if (wid == widgetId)
         {
            if (!(mousePos.x >= ma.x            &&
                  mousePos.x <= ma.x + ma.width &&
                  mousePos.y >= ma.y            &&
                  mousePos.y <= ma.y + ma.height))
            {
               // outside the widget, ignore event
               return false;
            }

            continue;
         }

         var widZorder = this.zOrder.indexOf(wid);
         if (mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height && (widZorder < widgetZorder))
         {
            return false;
         }
      }
      
      return true;
   };
   
   /**
    * Return an object which will act as a callback for the specified mouse operation.
    * 
    * @param    {Object} wThis
    *           The window object to which the callback is registered.
    * @param    {String} mouseOp
    *           The mouse operation to which the callback is attached.
    *
    * @return   An function acting as a callback for the specified mouse op.
    */
   Window.prototype.raiseMouseEvent = function(wThis, mouseOp)
   {
      var opCode = mouseOps[mouseOp] & 0xff;
      
      return function(evt)
      {
         var mousePos = wThis.getMousePos(evt);

         // if there are "any mouse" widgets, process them all
         for (var wid in wThis.anyMouseAwareWidgets)
         {
            if (wThis.anyMouseAwareWidgets.hasOwnProperty(wid))
            {
               wThis.sendMouseEvent(evt, mousePos, wThis, opCode, wid);
            }
         }
         
         if (!me.canProcessOsEvent(evt, wThis))
         {
            return;
         }

         var wid = wThis.findMouseSource(evt, wThis.mouseAwareWidgets[mouseOp]);
         
         if (wid == undefined)
         {
            return;
         }

         if (me.hasOsEvent(evt, wThis, mouseOp, wid))
         {
            // high-level event, do not process here
            return;
         }

         wThis.sendMouseEvent(evt, mousePos, wThis, opCode, wid);

         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };
   
   /**
    * Save the z-order widget list.
    *
    * @param    {Array} zOrder.
    *           The array of widget IDs, in z-order, topmost widget having a lower index.
    */
   Window.prototype.setWidgetZOrder = function(zOrder)
   {
//      console.debug(JSON.stringify(zOrder));
      this.zOrder = zOrder;
   }

   /**
    * Get the registered hoverable widget by id.
    *
    * @param    {Number} wid
    *           The widget ID.
    * 
    * @returns  the MouseHoverable widget
    */
   Window.prototype.getHoverableWidget = function(wid)
   {
      var mactions = this.osAwareWidgets["hoverable"];
      var widget = mactions[wid];
      if (widget)
      {
         return widget;
      }
      var parentWindow = getWindow(this.owner);
      return parentWindow && parentWindow.id != this.id ? 
            parentWindow.getHoverableWidget(wid) : undefined;
   }

   /**
    * Register the hoverable widget.
    *
    * @param    {Number} wid
    *           The widget ID.
    * 
    * @param    {Number} pid
    *           The parent widget ID.
    * 
    * @param    {Object} ptr
    *           The mouse pointer (style or data url).
    */
   Window.prototype.registerHoverableWidget = function(wid, pid, x, y, width, height, 
                                       ptr, defaultPtr, enforceDefault)
   {
//      console.log("%s:registerHoverableWidget(%s, %s, %s, %s, %s, %s, %s, %s, %s)", 
//            this.id, wid, pid, x, y, width, height, ptr, defaultPtr, enforceDefault);
       var mactions = this.osAwareWidgets["hoverable"];
      if (!mactions)
      {
         mactions = {};
         this.osAwareWidgets["hoverable"] = mactions;
      }
      mactions[wid] = new MouseHoverable(this, wid, pid, x, y, width, height, ptr, 
                                          defaultPtr, enforceDefault);
   }

   /**
    * Set the custom mouse pointer for the hoverable widget.
    *
    * @param    {Number} wid
    *           The widget ID.
    * 
    * @param    {Object} ptr
    *           The mouse pointer (style or data url).
    */
   Window.prototype.setMousePtr = function(wid, ptr)
   {
      var widget = this.getHoverableWidget(wid);
      if (widget != null)
      {
         widget.setPtr(ptr);
      }
   }

   /**
    * Register the specified coordinates for mouse support.
    *
    * @param    {Number} wid
    *           The widget ID.
    * @param    {Number} x
    *           The X coordinate.
    * @param    {Number} y
    *           The Y coordinate.
    * @param    {Number} width
    *           The widget width.
    * @param    {Number} height
    *           The widget height.
    * @param    {Number} actions
    *           Bit encoded actions (each bit, if set, represents an action with the specified
    *           ID from <code>mouseOps</code>).
    */
   Window.prototype.registerMouseWidget = function(wid, x, y, width, height, actions)
   {
//      var widActions = "";
//      for (var opName in mouseOps)
//      {
//         if (mouseOps.hasOwnProperty(opName))
//         {
//            if ((actions & (1 << mouseOps[opName])) == (1 << mouseOps[opName]))
//            {
//               widActions = widActions + " " + opName;
//            }
//         }
//      }
//      for (var opName in osOps)
//      {
//         if (osOps.hasOwnProperty(opName))
//         {
//            if ((actions & (1 << osOps[opName])) == (1 << osOps[opName]))
//            {
//               widActions = widActions + " " + opName;
//            }
//         }
//      }
//      console.debug("registerMouseWidget( " + wid + ", x=" + x + ", y=" + y + ", width=" + width
//            + ", height=" + height + ", actions=" + widActions + ")");
      // register the widget aware of mouse (existing entry will be overwritten)
      for (var opName in mouseOps)
      {
         if (mouseOps.hasOwnProperty(opName))
         {
            var opCode = mouseOps[opName];
            
            if ((actions & (1 << opCode)) == (1 << opCode))
            {
               var mactions = this.mouseAwareWidgets[opName];
               
               mactions[wid] = { "x": x, "y": y, "width": width, "height": height };
            }
         }
      }

      for (var opName in osOps)
      {
         if (osOps.hasOwnProperty(opName))
         {
            var opCode = osOps[opName];
            
            if ((actions & (1 << opCode)) == (1 << opCode))
            {
               var mactions = this.osAwareWidgets[opName];
               var widget = createOSMouseWidget(opName, this, wid, x, y, width, height);
               if (widget)
               {
                  mactions[wid] = widget;
               }
            }
         }
      }
   };

   /**
    * The factory method to create an OS mouse widget.
    * 
    * @param    {String} opName
    *           The os operation associated with this widget.
    * @param    {Window} win
    *           The widget owner.
    * @param    {Number} wid
    *           The widget ID.
    * @param    {Number} x
    *           The X coordinate.
    * @param    {Number} y
    *           The Y coordinate.
    * @param    {Number} width
    *           The widget width.
    * @param    {Number} height
    *           The widget height.
    */
   function createOSMouseWidget(opName, win, wid, x, y, width, height)
   {
      var osMouseWidget;
      if (opName === "movable")
      {
         osMouseWidget = new MouseMovable(win, wid, x, y, width, height);
      }
      else if (opName == "popupable_left")
      {
         osMouseWidget = new MousePopupable(win, wid, x, y, width, height, 1);
      }
      else if (opName == "popupable_right")
      {
         osMouseWidget = new MousePopupable(win, wid, x, y, width, height, 3);
      }
      else  if (opName == "copy" || opName == "cut" || opName == "paste" || opName === "minimize")
      {
         osMouseWidget = new SystemAction(win, wid, x, y, width, height, opName);
      }
      
      return osMouseWidget;
   }
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    {Number} wid
    *           The widget ID.
    * @parem    {boolean} retainHoverable
    *           <code>true</code> if the Hoverable widgets should be retained
    */
   Window.prototype.deregisterMouseWidget = function(wid, retainHoverable)
   {
      for (var opName in mouseOps)
      {
         if (mouseOps.hasOwnProperty(opName))
         {
            var mactions = this.mouseAwareWidgets[opName];
            if (wid in mactions)
            {
               delete mactions[wid];
            }
         }
      }

      for (var opName in osOps)
      {
         if (osOps.hasOwnProperty(opName))
         {
            var mactions = this.osAwareWidgets[opName];
            if (wid in mactions)
            {
               if (!retainHoverable && opName != "hoverable")
               {
                  var opHandler = mactions[wid];
                  opHandler.deregisterListeners();
                  delete mactions[wid];
               }
            }
         }
      }
   };
   
   /**
    * Deregister all mouse widgets for this window.
    */
   Window.prototype.deregisterAllMouseWidget = function()
   {
      for (var opName in mouseOps)
      {
         if (mouseOps.hasOwnProperty(opName))
         {
            var mactions = this.mouseAwareWidgets[opName];
            for (var wid in mactions)
            {
               if (mactions.hasOwnProperty(wid))
               {
                  delete mactions[wid];
               }
            }
         }
      }
      
      for (var opName in osOps)
      {
         if (osOps.hasOwnProperty(opName))
         {
            var mactions = this.osAwareWidgets[opName];
            for (var wid in mactions)
            {
               if (mactions.hasOwnProperty(wid))
               {
                  var opHandler = mactions[wid];
                  opHandler.deregisterListeners();
                  delete mactions[wid];
               }
            }
         }
      }
      
      for (var wid in this.anyMouseAwareWidgets)
      {
         if (this.anyMouseAwareWidgets.hasOwnProperty(wid))
         {
            delete this.anyMouseAwareWidgets[wid];
         }
      }
   };

   /**
    * Register the specified widget for any mouse events.
    *
    * @param    {Number} wid
    *           The widget ID.
    */
   Window.prototype.registerAnyMouseWidget = function(wid)
   {
      this.anyMouseAwareWidgets[wid] = true;
   };

   /**
    * Deregister the specified widget for any mouse events.
    * 
    * @param    {Number} wid
    *           The widget ID.
    */
   Window.prototype.deregisterAnyMouseWidget = function(wid)
   {
      if (wid in this.anyMouseAwareWidgets)
      {
         delete this.anyMouseAwareWidgets[wid];
      }
   };
         
   /**
    * Split the text for a editor widget, so that it fits the specified maximum width.  The 
    * <code>delta</code> parameter represents the start position of this text, in the editor's 
    * content; when the full content is parsed, this will be zero.
    *  
    * @param    {Number} widgetId
    *           The editor's ID.
    * @param    {Number} offset
    *           The start position of this text in the editor's full content.
    * @param    {Number} fontNum
    *           The editor's font.
    * @param    {Number} maxWidth
    *           The width for the content to fit (the editor's editable area).
    * @param    {String} text
    *           The text to split into lines.
    *           
    * @return   A list of <code>Line</code> objects.
    */
   Window.prototype.parseEditorContent = function(widgetId, offset, fontNum, maxWidth, text)
   {
      function Line(text, start, length, newLine)
      {
         this.start   = start;
         this.newLine = newLine;
         /*
         if (text != undefined && length > 0)
            this.text = text.substring(start, start + length);
         else
            this.text = "";
         */
         this.isNewLine = function()
         {
            return newLine;
         }
      };
      
      function goToNewLine(txt, width, fontNum)
      {
         var mw = p2j.fonts.getFontMaxWidth(fontNum);
         if (txt.length * mw < width - 2 * mw)
         {
            return false;
         }
      
         var txtWidth = p2j.fonts.getTextWidth(fontNum, txt);
         // console.log(txt + " " + txtWidth);
         return txtWidth > width;
      };
         
      var lines = [];
      var start = 0;
      var LF = '\n';
      // lines.push(text);
      for (var i = 0; i < text.length; i++)
      {
         if (text[i] == LF)
         {
            lines.push(new Line(text, start + offset, i - start + offset, true));
            start = i + 1;
            continue;
         }
         
         if (i > start && text[i - 1] == ' ' && text[i] != ' ')
         {
            // check for word wrapping
            var j = i;
         
            var fullLine = false;
            var newWord = false;
            while (j < text.length)
            {
               if (text[j] == ' '  || text[j] == LF)
               {
                  newWord = true;
                  j--;
                  break;
               }
      
               if (goToNewLine(text.substring(start, j), maxWidth, fontNum))
               {
                  fullLine = true;
                  break;
               }
          
               j++;
            }
       
            if (j == text.length)
            {
               newWord = true;
               j--;
            }
         
            // if this is a full line, determine if this is the end of the word; if so, split just
            // after it, but without moving the space to the next line
            if (fullLine)
            {
               var delta = 0; // (placeChar ? 1 : 0);
          
               if (j < text.length - delta && (text[j + delta] == ' ' || text[j + delta] == LF))
               {
                  // if the word has not finished yet, split from the beginning of the word
                  lines.push(new Line(text, start + offset, j - start + offset, false));
            
                  // advance just after the word, consuming the whitespace
                  start = j + delta;
                  i = start;
               }
               else
               {
                  // if the word has not finished yet, split from the beginning of the word
                  lines.push(new Line(text, start + offset, i - start + offset, false));
         
                  start = i;
               }
          
               continue;
            }
       
            // if we have found the end of this word, determine if if fits the current line; if
            // it does not fit, the word will go to the next line
            var wordEnd = Math.min(j + 1, text.length);
            if (newWord && goToNewLine(text.substring(start, wordEnd), maxWidth, fontNum))
            {
               lines.push(new Line(text, start + offset, i - start + offset, false));
               start = i;
               continue;
            }
         }
         
         // last check
         
         if (goToNewLine(text.substring(start, i + 1), maxWidth, fontNum))
         {
            lines.push(new Line(text, start + offset, i - start + offset, false));
            
            if (text[i] == ' ')
               i++;
       
            start = i;
         }
      } 
      
      if (lines.length != 0 && lines[lines.length - 1].isNewLine() && text.length - start <= 0)
      {
         lines.push(new Line("", start + offset, false));
      }

      if (text.length - start > 0 || lines.length == 0)
      {
         lines.push(new Line(text, start + offset, text.length - start, false));
         start = text.length;
      }
      
      return lines;
   };

   /**
    * Cleanup any resources, including any linkages to the window from an owning window.
    */
   Window.prototype.cleanup = function()
   {
      this.deregisterAllMouseWidget();
      // remove the child linkage from the owner if it exists
      if (isExistingWindowId(this.owner))
      {
         var parent = getWindow(this.owner);
         
         var changed = true;
         while (changed)
         {
            changed = false;
            for (var i = 0; i < parent.children.length; i++)
            {
               var childId = parent.children[i];
               if (childId === this.id)
               {
                  parent.children.splice(i, 1);
                  changed = true;
                  break;
               }
            }
         }
      };
      
      // remove from the DOM
      this.canvas.parentNode.removeChild(this.canvas);
      
      if (this.cleanupCallback)
      {
         this.cleanupCallback();
      }
      // TODO: compress the zIndex values and manage z-order state
   };
      
   /**
    * Enable/disable the window.
    *                              
    * @param    {boolean} enabled  
    *           Flag that determines the new sensitivity state.
    */
   Window.prototype.setEnabled = function(enabled)
   {
      //p2j.logger.log(this.id + " .setEnabled(" + enabled + ")");
      //console.debug(this.id + " .setEnabled(" + enabled + ")");
      this.canvas.disabled = !enabled;
   };
   
   /**
    * Show or hide the window.
    *                              
    * @param    {boolean} visible  
    *           Flag that determines the new visibility state.
    */
   Window.prototype.setVisible = function(visible)
   {
      var oldVisible = this.visible;
      
      //p2j.logger.log(this.id + " .setVisible(" + visible + ")");
      //console.debug(this.id + " .setVisible(" + visible + ")");
      // this code depends on each canvas being style.position = 'absolute' so that it is
      // outside of the browser's layout processing
      this.canvas.style.visibility = visible ? "visible" : "hidden";
      this.visible = visible;
      
      // do not bring the window to top, leave this up to the app logic

      if (visible != oldVisible            &&
          !this.isShareActivationWithOwner &&
          embeddedAppController            &&
          this.modal)
      {
         if (visible)
         {
            embeddedAppController.displayOverlayPanel(this.id);
         }
         else
         {
            embeddedAppController.hideOverlayPanel(this.id);
         }
      }
   };
   
   /**
    * Is visible.
    */
   Window.prototype.isVisible = function()
   {
      return this.canvas.style.visibility === "visible";
   };
   
   /**
    * Resize the window to the given width and height.
    *
    * @param    {Number} width
    *           The new window width in pixels.
    * @param    {Number} height
    *           The new window height in pixels.
    */
   Window.prototype.resize = function(width, height)
   {
      this.canvasRenderer.resize(width, height);
      
      verifyEmbeddedWindows();
   };

   /**
    * Sets the mouse cursor style.
    * 
    * @param    {String} style     
    *           Denotes the predefined cursor style.
    * @param    {boolean} ignoreLockedPtr     
    *           Ignore locked mouse pointer.
    */
   Window.prototype.setCursorStyle = function(style, ignoreLockedPtr)
   {
      if (!ignoreLockedPtr)
      {
         var es = p2j.screen.lockedMousePtr();
         if (es != null)
         {
            style = es; 
         }
      }
      if (style != this.currentMousePtr) // avoid overhead for large file-based custom cursors
      {
//         console.log("%s.setCursorStyle(%s, %s)", this.id, style, ignoreLockedPtr);
         this.canvas.style.cursor = style;
         this.currentMousePtr = style;
      }
   };

   /**
    * Gets the current mouse cursor style.
    * 
    * @return   See above.
    */
   Window.prototype.getCursorStyle = function()
   {
      return this.canvas.style.cursor;
   };
   
   /**
    * Enables/disables the OS events for this window.
    * 
    * @param    {Boolean} capture
    *           When <code>true</code>, it will process OS events like window close/iconify/etc.
    */
   Window.prototype.enableOsEvents = function(capture)
   {
      this.processOsEvents = capture;
   };

   /**
    * Draw the given list of operations in the canvas.
    *
    * @param    {byte[]} message
    *           The encoded list of drawing operations.
    * @param    {String} md5
    *           The MD5 hash for this current drawing message.
    */
   Window.prototype.draw = function(message, md5)
   {
      var numOps = p2j.socket.readInt32BinaryMessage(message, 21); // 1 + 16 + 4
                                   
      p2j.logger.log("START DRAWING CYCLE FOR WINDOW " + this.id);

      var idx = 25; // 1 + 16 + 4 + 4

      drawLoop:
      for (var i = 0; i < numOps; i++)
      {
         var sz = p2j.socket.readInt32BinaryMessage(message, idx);
                                   
         idx += 4;
         
         var type = message[idx];
         
         var x;
         var y;                    
         var height;
         var width;
         var diameter;
         var raised;
         var filled;
         var text;
         var textLength;           
         var offset;
         var centered;
         var fontId;               
         var fname;
         var path;
         var clear;
         var extra = "";
                                   
         // the following are implemented on the server-side and require no additional
         // implementation here:
         // ops.CLEAR_AREA         
         // ops.CLEAR:
         // ops.FLUSH
         // ops.START_BATCH
         // ops.END_BATCH
               
         switch (type)
         {
            case ops.DRAW_STRING:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               centered = (p2j.socket.readInt32BinaryMessage(message, idx + offset) == 1);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               this.canvasRenderer.drawText(text, x, y, centered);

               extra = " text = " + text + "; x = " + x + "; y = " + y +
                       "; centered = " + centered;
               break;
            case ops.DRAW_STRING_SCALED:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               centered = (p2j.socket.readInt32BinaryMessage(message, idx + offset) == 1);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               var lwidth = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               var lheight = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;
               this.canvasRenderer.drawScaledText(currentFont, text, x, y, centered, lwidth, lheight);
               extra = " text = " + text + "; x = " + x + "; y = " + y +
                       "; centered = " + centered +
                       "; lwidth = " + lwidth + "; lheight = " + lheight;
               break;
            case ops.DRAW_PARAGRAPH:
               offset = 1;

               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;

               fontId = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               x = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               y = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;
               
               width = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset = offset + 4;

               // TODO: this might need to be 'bottom', too
               this.offscreenContext.textBaseline = 'top';
               me.layoutParagraphWorker(this.offscreenContext, text, fontId, x, y, width);

               break;
            case ops.DRAW_LINE:
               var x1 =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var y1 =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               var x2 =  p2j.socket.readInt32BinaryMessage(message, idx + 9);
               var y2 =  p2j.socket.readInt32BinaryMessage(message, idx + 13);
               extra = " x1 = " + x1 + "; y1 = " + y1 + "; x2 = " + x2 + "; y2 = " + y2;
               this.canvasRenderer.strokeLineSegment(x1, y1, x2, y2, this.canvasRenderer.rawColor);
               break;
            case ops.DRAW_RECT:
            case ops.FILL_RECT:
            case ops.DRAW_ELLIPSE:
            case ops.FILL_ELLIPSE:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               filled = (type === ops.FILL_RECT || type === ops.FILL_ELLIPSE);
               extra  = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height;
               if (type === ops.FILL_RECT || type === ops.DRAW_RECT)
               {
                  this.canvasRenderer.strokeRect(x, y, width, height, this.canvasRenderer.rawColor, filled);
               }
               else
               {
                  this.canvasRenderer.strokeEllipse(x, y, width, height, this.canvasRenderer.rawColor, filled);
               }
               break;
            case ops.FILL_RECT_AROUND_TEXT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               textLength = p2j.socket.readInt32BinaryMessage(message, idx + 17);
               text  = p2j.socket.readStringBinaryMessageByLength(message, idx + 21, textLength);
               width = p2j.fonts.getTextWidth(currentFont, text);
               filled = true;
               extra  = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height;
               this.canvasRenderer.strokeRect(x, y, width, height, this.canvasRenderer.rawColor, filled);
               break;
            case ops.DRAW_ROUND_RECT:
               x        = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y        = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width    = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height   = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               diameter = p2j.socket.readInt32BinaryMessage(message, idx + 17);
               extra = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height
                  + "; diameter = " + diameter;
               this.canvasRenderer.drawRoundRect(x, y,
                     width, height, diameter,
                     this.canvasRenderer.rawColor, false);
               break;
            case ops.DRAW_IMAGE:
               x = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y = p2j.socket.readInt32BinaryMessage(message, idx + 5);

               width = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);

               extra = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height;
               var encoding = message[idx + 17];
               var key = p2j.socket.readInt32BinaryMessage(message, idx + 18);
               var pixelsInBytes = width * height * 4;
               var imgData;
               var imgDataOffset;

               // RAW encoding
               if (encoding == 0)
               {
                  imgData = message;
                  imgDataOffset = idx + 22;
                  loadedImages.set(key, message.subarray(imgDataOffset, imgDataOffset + pixelsInBytes));
               }
               // HASH encoding
               else if (loadedImages.has(key))
               {
                  imgData = loadedImages.get(key);
                  imgDataOffset = 0;
               }
               else
               {
                  console.log("DRAW_IMAGE: No image data received, but image not cached!")
               }

               if (imgData)
               {
                  this.canvasRenderer.drawImage(x, y, width, height, imgData, imgDataOffset);
               }
               break;
            case ops.FILL_ROUND_RECT:
               x        = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y        = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width    = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height   = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               diameter = p2j.socket.readInt32BinaryMessage(message, idx + 17);
               extra = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height
                        + "; diameter = " + diameter;
               this.canvasRenderer.drawRoundRect(x, y, width, height,
                     diameter, this.canvasRenderer.rawColor, true);
               break;
            case ops.FILL_POLYGON:
               offset = idx + 1;
               var numPoints = p2j.socket.readInt32BinaryMessage(message, offset);
               var xPoints = [];
               var yPoints = [];
               
               for (var j = 0; j < numPoints; j++)
               {
                  offset += 4;
                  xPoints[j] = p2j.socket.readInt32BinaryMessage(message, offset);
                  offset += 4;
                  yPoints[j] = p2j.socket.readInt32BinaryMessage(message, offset);
               }
               
               this.canvasRenderer.strokePolygon(xPoints, yPoints,
                     numPoints, this.canvasRenderer.rawColor, true);
               break;
            case ops.DRAW_3D_RECT:
            case ops.FILL_3D_RECT:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               raised = (message[idx + 17] === 1);
               filled = (type == ops.FILL_3D_RECT);
               extra = " x = " + x +"; y = " + y + "; width = " + width + "; height = " +
                        height + "; raised = " + raised;
               this.canvasRenderer.draw3DRect(x, y, width, height,
                     this.canvasRenderer.rawColor, filled, raised);
               break;
            case ops.DRAW_SELECTION:
               x      = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y      = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               clear  = (message[idx + 17] === 1);
               extra  = " x = " + x + "; y = " + y + "; width = " + width + "; height = " +
                        height + "; clear = " + clear;
               this.canvasRenderer.drawSelection(x, y, width, height,
                                                 this.canvasRenderer.rawColor, clear);
               break;
            case ops.RESET_SELECTION:
               this.canvasRenderer.resetSelection();
               break;
            case ops.SET_COLOR:
            case ops.SET_DESKTOP_BGCOLOR:
               var r  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var g  = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               var b  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               if (type == ops.SET_COLOR)
               {
                  this.canvasRenderer.setColor([ r, g, b ]);
                  extra = " r = " + r + "; g = " + g + "; b = " + b;
               }
               else
               {
                  var bgcolor = this.canvasRenderer.createColorString([ r, g, b ]);
                  p2j.setDesktopBgColor(bgcolor);

                  if (embeddedAppController)
                  {
                     embeddedAppController.setFrameBackground(bgcolor);
                  }
               }
               break;
            case ops.SET_LINE_STROKE:
               var strokeStyleId = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var strokeWidth = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.canvasRenderer.setLineStroke(strokeStyleId, strokeWidth);
               extra = " strokeStyleId = " + strokeStyleId + "; strokeWidth = " + strokeWidth;
               break;              
            case ops.SET_FONT:
               fontId = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               setFont(this.offscreenContext, fontId);
               break;
            case ops.SET_FONT_STYLE:
               var style = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               p2j.fonts.setFontStyle(currentFont, style);
               fname = p2j.fonts.getFontName(currentFont);
               this.offscreenContext.font = fname;
               break;
            case ops.START_DRAW_CURSOR:
               offset = 1;
               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               text = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               offset = offset + textLength * 2;
               var columnOffset = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               var rowOffset = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               var columnScrollOffset = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               var rowScrollOffset = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               var rowHeight = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               
               this.cursorOffsetY = (rowOffset - rowScrollOffset) * rowHeight;
               
               text = text.substring(columnScrollOffset, columnOffset);
               
               this.cursorOffsetX = p2j.fonts.getTextWidth(currentFont, text);
               this.trans++;
               this.canvasRenderer.translate(this.cursorOffsetX, this.cursorOffsetY);
               break;
            case ops.END_DRAW_CURSOR:
               this.canvasRenderer.untranslate(-this.cursorOffsetX, -this.cursorOffsetY);
               this.cursorOffsetX = 0;
               this.cursorOffsetY = 0;
               this.trans--;
               break;
            case ops.TEXT_SHIFT:
               offset = 1;
               textLength = p2j.socket.readInt32BinaryMessage(message, idx + offset);
               offset += 4;
               text = p2j.socket.readStringBinaryMessageByLength(message, idx + offset, textLength);
               
               var textShift = p2j.fonts.getTextWidth(currentFont, text);
               
               if (this.textShiftX > 0)
               {
                  this.trans--;
                  this.canvasRenderer.untranslate(-this.textShiftX, 0);
               }
               
               
               
               if (textShift > 0)
               {
                  this.textShiftX += textShift;
                  this.trans++;
                  this.canvasRenderer.translate(this.textShiftX, 0);
               }
               else
               {
                  this.textShiftX = 0;
               }
               
               break;
            case ops.UNDO_TEXT_SHIFT:
               if (this.textShiftX > 0)
               {
                  this.trans--;
                  this.canvasRenderer.untranslate(-this.textShiftX, 0);
                  this.textShiftX = 0;
               }
               break;
            case ops.TRANSLATE_PUSH:                                                   
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.trans++;
               extra = " " + this.trans.toString() + "; x = " + x.toString()
                     + "; y = " + y.toString();
               this.canvasRenderer.translate(x, y);
               break;
            case ops.TRANSLATE_POP:                                                   
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.trans--;
               extra = " " + this.trans.toString() + "; x = " + x.toString() +"; y = " + y.toString();
               this.canvasRenderer.untranslate(x, y);
               break;
            case ops.CLIP:
               var clip = [];
               var clipCount = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               for (var j = 0; j < clipCount; j++)
               {
                  var off = idx + 5 + 16*j;
                  x = p2j.socket.readInt32BinaryMessage(message, off);
                  y = p2j.socket.readInt32BinaryMessage(message, off + 4);
                  width = p2j.socket.readInt32BinaryMessage(message, off + 8);
                  height = p2j.socket.readInt32BinaryMessage(message, off + 12);
                  clip.push
                  ({
                     x : x,
                     y : y,
                     width : width,
                     height : height
                  });
               }
               this.clips++;
               extra = " " + this.clips.toString() + JSON.stringify(clip);
               this.canvasRenderer.clip(clip);
               break;
            case ops.NO_CLIP:
               this.clips--;
               extra = " " + this.clips.toString();
               this.canvasRenderer.unclip();
               break;
            case ops.SET_WINDOW_BOUNDS:
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);
               
               this.originalX = x;
               this.originalY = y;
               
               var parentWindow = getWindow(this.owner);
               if (!this.clientMove &&
                   !(parentWindow == undefined           ||
                     parentWindow.originalX == undefined ||
                     parentWindow.originalY == undefined))
               {
                  var cbounds = parentWindow.canvas.getBoundingClientRect();
                  var deltaX = parentWindow.originalX - cbounds.left;
                  var deltaY = parentWindow.originalY - cbounds.top;
                  
                  x = x - deltaX;
                  y = y - deltaY;
               }
               
               this.canvas.style.left = x.toString() + "px";
               this.canvas.style.top  = y.toString() + "px";
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 13);
               this.resize(width, height);
               this.canvas.width  = width;
               this.canvas.height = height;

               extra = " x = " + x + "; y = " + y + "; width = " + width + "; height = " + height;
               
               this.embeddedCentered = false;
               verifyEmbeddedWindows();
               break;
            case ops.SET_LOCATION:
               x =  p2j.socket.readInt32BinaryMessage(message, idx + 1);
               y =  p2j.socket.readInt32BinaryMessage(message, idx + 5);

               this.originalX = x;
               this.originalY = y;
               
               var parentWindow = getWindow(this.owner);
               if (!this.clientMove &&
                   !(parentWindow == undefined           ||
                     parentWindow.originalX == undefined ||
                     parentWindow.originalY == undefined))
               {
                  var cbounds = parentWindow.canvas.getBoundingClientRect();
                  var deltaX = parentWindow.originalX - cbounds.left;
                  var deltaY = parentWindow.originalY - cbounds.top;
                  
                  x = x - deltaX;
                  y = y - deltaY;
               }

               this.canvas.style.left = x.toString() + "px";
               this.canvas.style.top  = y.toString() + "px";
               extra = " x = " + x + "; y = " + y;
               
               this.embeddedCentered = false;
               verifyEmbeddedWindows();
               break;
            case ops.RESIZE_WINDOW:
               width  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               height = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               this.resize(width, height);
               this.canvas.width  = width;
               this.canvas.height = height;
               extra = " width = " + width + "; height = " + height;
               
               this.embeddedCentered = false;
               verifyEmbeddedWindows();
               break;
            case ops.DRAW_HIGHLIGHT:
               break;
            case ops.CLEAR_HIGHLIGHT:
               break;
            case ops.SET_TITLE:
               textLength = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               text = p2j.socket.readStringBinaryMessageByLength(message, idx + 5, textLength);
               extra = " title = " + text;
               this.title = text;
               if (taskBar)
               {
                  taskBar.setTitleForTask(this.id, this.title);
               }
               break;
            case ops.SET_ICON:
               var iconWidth  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var iconHeight = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               extra = " x = 0; y = 0; width = " + iconWidth + "; height = " + iconHeight;
               var encoding = message[idx + 9];
               var key = p2j.socket.readInt32BinaryMessage(message, idx + 10);
               var pixelsInBytes = iconWidth * iconHeight * 4;
               var imgData;
               var imgDataOffset;
               if(!loadedImages.has(key))
               {
                  imgData = message;
                  imgDataOffset = idx + 14;
                  loadedImages.set(key, message.subarray(imgDataOffset, imgDataOffset + pixelsInBytes));
                  this.iconId     = key;
                  this.iconWidth  = iconWidth;
                  this.iconHeight = iconHeight;
               }
               if (taskBar)
               {
                  taskBar.setIconForTask(this.id, this.iconId, loadedImages.get(this.iconId),
                        this.iconWidth, this.iconHeight);
               }
               break;
            case ops.SET_XOR_COMPOSITE:
               this.canvasRenderer.setXORComposite();
               break;
            case ops.SET_XOR_MODE:
               var r  = p2j.socket.readInt32BinaryMessage(message, idx + 1);
               var g  = p2j.socket.readInt32BinaryMessage(message, idx + 5);
               var b  = p2j.socket.readInt32BinaryMessage(message, idx + 9);
               this.canvasRenderer.setXORMode([ r, g, b ]);
               break;
            case ops.RESET_COMPOSITE:
               this.canvasRenderer.resetComposite();
               break;
            case ops.START_RECORDING:
               var offset = idx + 1;

               var x = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var y = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var width = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var height = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var md5a = new message.constructor(16);
               for (var h = 0; h < 16; h++)
               {
                  md5a[h] = message[offset];
                  offset = offset + 1;
               }
               var rmd5 = p2j.socket.createHexString(md5a);
               
               this.canvasRenderer.startRecording(x, y, width, height, rmd5);
               break;
            case ops.STOP_RECORDING:
               this.canvasRenderer.stopRecording();
               break;
            case ops.DRAW_CACHED:
               var offset = idx + 1;

               var x = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var y = p2j.socket.readInt32BinaryMessage(message, offset);
               offset += 4;
               var md5a = new message.constructor(16);
               for (var h = 0; h < 16; h++)
               {
                  md5a[h] = message[offset];
                  offset = offset + 1;
               }
               var rmd5 = p2j.socket.createHexString(md5a);
               
               this.canvasRenderer.drawCached(x, y, rmd5);
               break;
            default:
               if (typeof type !== "undefined")
               {
                  me.error("Unknown operation " + type.toString() + "!");
               }
               else
               {
                  me.error("Your session is corrupted buddy!");
                  /*
                  var binMsg  = '';
                  for (var j = 0; j < message.length; j++)
                  {
                     binMsg += String.fromCharCode(message[j]);
                  }
                  var b64msg = window.btoa(binMsg);
                  p2j.logger.log(b64msg);
                  */
               }
               console.trace();
               break drawLoop;
         }
         
         p2j.logger.log("PaintPrimitives." + ops[type] + " " + extra);
         idx += sz;
      }

      // to blast the offscreen image on the window canvas
      if (this.offscreenCanvas.width > 0 && this.offscreenCanvas.height > 0)
      {
         this.ctx.drawImage(this.offscreenCanvas, 0, 0);
         // if this is the end of a drawing bracket (clip and translate are zero), save the canvas
         if (this.clips == 0 && this.trans == 0 && graphicsCached)
         {
            // save the offscreen canvas
            var img = new Image();
            img.src = this.offscreenCanvas.toDataURL("image/png");
            this.cacheDraw(md5, img);
            
            // clear the offscreen canvas for next drawing
            this.canvasRenderer.clear();
         }
         
      }
      
      p2j.logger.log("END DRAWING CYCLE FOR WINDOW " + this.id);
   }; 

   /**
    * Cache the specified image, for this window, to be reused.
    *
    * @param   {String} md5
    *          The hash of the drawing operations which produced this image.
    * @param   {Image} img
    *          The image to be cached.
    */
   Window.prototype.cacheDraw = function(md5, img)
   {
      this.drawCache[md5] = img;
   }

   /**
    * Removed the cached image for the specified hash.
    *
    * @param   {String} md5
    *          The hash of the drawing operations which produced this image.
    */
   Window.prototype.removeCachedDraw = function(md5)
   {
      delete this.drawCache[md5];
   }
   
   /**
    * Check if there is a cached image for this hash.
    *
    * @param   {String} md5
    *          The hash of the drawing operations which produced this image.
    *
    * @return  {boolean}  <code>true</code> if there is an image for the given hash.
    */
   Window.prototype.hasCachedDraw = function(md5)
   {
      return !(this.drawCache[md5] == undefined);
   }

   /**
    * Get the cached image for the specified hash.
    *
    * @param   {String} md5
    *          The hash of the drawing operations which produced this image.
    *
    * @return  {Image}  The cached image
    */
   Window.prototype.getCachedDraw = function(md5)
   {
      return this.drawCache[md5];
   }

   Window.prototype.sendExplicitMouseEvent = function(wid, evt, mousePos, wThis, opCode)
   {
      this.sendMouseEvent(evt, mousePos, wThis, opCode, wid);
   }
   
   /**
    * Send a mouse event to be processed by the Java side.
    *
    * @param   {Event} evt
    *          The mouse event.
    * @param   {Object} mousePos
    *          The mouse position.
    * @param   {Window} wThis
    *          The window reference.
    * @param   {Number} opCode
    *          The ID of this mouse event.
    */
   Window.prototype.sendMouseEvent = function(evt, mousePos, wThis, opCode, wid)
   {
      var mouseEvent = {
                           windowId   : wThis.id,
                           opCode     : opCode,
                           wid        : wid,
                           x          : mousePos.x,
                           y          : mousePos.y,
                           button     : evt.which,
                           clicks     : evt.detail
                        };
      
      // fix mouse click count
      if (p2j.isIe && (opCode == mouseOps["click"]) && (evt.detail == 0))
      {
         mouseEvent.clicks = 1;
      }
      
      mouseEvent.bitmask = getModifiersBitMask(evt);
      
      var wheelEvent = (evt.deltaMode !== undefined);
      
      mouseEvent.wheelEvent = wheelEvent;
      
      if (wheelEvent)
      {
         mouseEvent.rotation = getScrollAmountDelta(evt);
         mouseEvent.unit     = getScrollAmountUnit(evt);
      }
      //console.debug("sendMouseEvent " + JSON.stringify(mouseEvent));
      // send the mouse event
      p2j.socket.sendMouseEvent(mouseEvent);
      
      // events are generated for the first widget which should respond for this mouse
      // action; it's the P2J runtimes responsibility to determine which actual widget
      // should interpret the event (i.e. z-order, etc)
   }
   
   /**
    * Returns the bit mask that represents the current modifier keys are pressed during this
    * mouse event occured.
    * 
    * @param    {Event} evt
    *           The mouse event.
    * 
    * @return   The keys modifier mask that represents which of the modifier keys are down.
    *           If 0, 1, 2, or 3-bit is set, then respectivelly the shift, the ctrl, the meta or
    *           the alt key is pressed.
    */
   function getModifiersBitMask(evt)
   {
      var modifiers = 0;
      if (evt.shiftKey !== undefined && evt.shiftKey)
      {
         modifiers |= 1;
      }
      if (evt.ctrlKey !== undefined && evt.ctrlKey)
      {
         modifiers |= 2;
      }
      if (evt.metaKey !== undefined && evt.metaKey)
      {
         modifiers |= 4;
      }
      if (evt.altKey !== undefined && evt.altKey)
      {
         modifiers |= 8;
      }
      
      return modifiers;
   }
   
   /**
    * Returns the scroll amount unit given by deltaMode. The valid values are 0x00, 0x01 and 0x02.
    * 0x00 deltaMode indicates that the delta values are specified in pixels.
    * 0x01 deltaMode indicates that the delta values are specified in lines.
    * 0x02 deltaMode indicates that the delta values are specified in pages.
    * 
    * @param    {Event} evt
    *           The mouse event.
    * 
    * @return   The scroll amount unit.
    */
   function getScrollAmountUnit(evt)
   {
      var unit = 0;
      if (evt.deltaMode !== undefined)
      {
         unit = evt.deltaMode;
      }
      
      return unit;
   }
   
   /**
    * Returns the scroll amount along the direction axis.
    * 
    * @param    {Event} evt
    *           The mouse event.
    * 
    * @return   The scroll amount.
    */
   function getScrollAmountDelta(evt)
   {
      var direction = 0;
      if (evt.deltaX !== undefined && evt.deltaX != 0)
      {
         direction = evt.deltaX;
      }
      else if (evt.deltaY !== undefined && evt.deltaY != 0)
      {
         direction = evt.deltaY;
      }
      
      return direction;
   }
   
   /**
    * Move the window to top, in z-order.
    */
   Window.prototype.moveToTop = function()
   {
      var entries = [];
      // remove all its children
      for (var i = 0; i < this.children.length; i++)
      {
         var ch = this.children[i];
         var entry = removeZOrderEntry(ch);
         if (entry)
         {
            entries.push(entry);
         }
      }

      moveZOrderEntryToTop(this.id);
      
      // add them back AFTER this, in z-order
      for (var i = 0; i < entries.length; i++)
      {
         var entry = entries[i];
         addZOrderEntry(entry.id, entry.win);
      }
   };
   
   /**
    * Close this window.
    */
   Window.prototype.close = function()
   {
      // no-op, as the OS can't decide to close the window; only the business logic can
      console.log("Closing window " + this.id);
   };
   
   /**
    * Iconify this window.
    */
   Window.prototype.iconify = function()
   {
      //p2j.logger.log("Window.prototype.iconify(" + this.id + ")");

      if (this.showInTaskBar)
      {
         this.setVisible(false);
      }

      if (taskBar && this.showInTaskBar)
      {
         taskBar.iconify(this.id);
      }
      activateTopVisibleWindow(this.id);
   };
   
   /**
    * De-iconify this window.
    */
   Window.prototype.deiconify = function()
   {
      //p2j.logger.log("Window.prototype.deiconify(" + this.id + ")");

      if (this.showInTaskBar)
      {
         this.setVisible(true);
      }

      if (taskBar && this.showInTaskBar)
      {
         taskBar.deiconify(this.id);
      }
   };
   
   /**
    * Maximize/Restore this window.
    */
   Window.prototype.maximize = function()
   {
      // no-op as this is done by the java side
   };
   
   /**
    * Checks if the given value is a valid integer. Modal dialogs windows can have negative ids.
    *
    * @param    wid
    *           The value to test.
    *
    * @return   <code>false</code> if the given value is <code>null</code>,
    *           <code>undefined</code>.  
    */
   function isValidWindowId(wid)
   {
      return (typeof wid === 'number');
   };
   
   /**
    * Checks if the given value already exists as a window id.
    *
    * @param    wid
    *           The value to test.
    *
    * @return   <code>false</code> if the given value does not exist in the window list.
    */
   function isExistingWindowId(wid)
   {
      return winlist[wid] instanceof Window;
   };

   /**
    * Obtain the specified window.
    *
    * @param    wid
    *           The window id to obtain.
    *
    * @return   The window instance or <code>null</code> if a window of the given id does not
    *           exist in the window list.
    */
   function getWindow(wid)
   {
      return winlist[wid] instanceof Window ? winlist[wid] : null;
   };

   /**
    * Move the first upper visible window to the top of other visible windows
    * after the selected window is minimized.
    * 
    * @param    {Number} id
    *           The id of the window that has been minimized.
    */
   function activateTopVisibleWindow(id)
   {
      var topWindow = findTopVisibleWindow(id);
      if (topWindow)
      {
         topWindow.moveToTop();
         p2j.screen.setActiveWindowId(topWindow.id);
         p2j.socket.sendWindowActivated(id, topWindow.id);
      }
      else
      {
         p2j.socket.sendWindowActivated(id, -1);
      }
   }

   /**
    * Find a visible window having the maximal z-index from all windows that are different
    * from the selected window.
    * 
    * @param    {Number} id
    *           The id of the selected window.
    * 
    * @return   {Window}
    *           The first different visible window having the maximal z-index.
    */
   function findTopVisibleWindow(id)
   {
      for(var j = zlist.length - 1; j >=0; j--)
      {
         var win = zlist[j].win;
         if ((win.id !== id) && win.isVisible())
         {
            return win;
         }
      }
      return null;
   }

   /**
    * Find the index of the given window in the z-order list.
    *
    * @param    {Number} wid
    *           The window ID to lookup.
    */
   function findZOrderEntry(wid)
   {
      var found = -1;
      
      // find the window in the z-order list
      for (var zidx = 0; zidx < zlist.length; zidx++)
      {
         // we keep the array non-sparse so there should never be a missing element
         if (zlist[zidx].id != wid)
         {
            // the id doesn't match, check the next one
            continue;
         }
         else
         {
            // found it!
            found = zidx;
            
            break;
         }
      }
      
      if (found == -1)
      {
         me.error("Missing window " + wid.toString() + " in z-order list.");
      }
      
      return found;
   };

   /**
    * Adds the given window to the end of the z-order list, setting the z-order value as needed.
    *
    * @param    {Number} wid
    *           The window ID to add.
    * @param    {Window} win
    *           The window instance.
    */
   function addZOrderEntry(wid, win)
   {
      // add to the end of the list
      zlist.push({ "id" : wid, "win" : win });

      // the z-order is based on the array index
      var zidx = zlist.length * 10;
      // assign the z-order
      win.canvas.style.zIndex = zidx;
      
      verifyEmbeddedWindows();
   };
   
   /**
    * Removes the given window from the z-order list and resets the z-order values of all
    * windows that are above the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} wid
    *           The window ID to remove.
    *
    * @return   {id : Number , win : Window} The removed entry.
    */
   function removeZOrderEntry(wid)
   {
      var found = findZOrderEntry(wid);
      
      if (found == -1)
      {
         // error was already raised
         return;
      }
      
      return removeZOrderEntryWorker(found);
   };
      
   /**
    * Removes the given element from the z-order list and resets the z-order values of all
    * windows that are above the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} idx
    *           The index of the element to remove.
    *
    * @return   {id : Number , win : Window} The removed entry.
    */
   function removeZOrderEntryWorker(idx)
   {
      // delete the entry and maintain the non-sparse nature of the list
      var removed = zlist.splice(idx, 1);
      
      // all subsequent entries are kept in the same order, but we compress the z-index
      // values so there are no gaps
      for (var midx = idx; midx < zlist.length; midx++)
      {
         // the z-order is based on the array index
         zlist[midx].win.canvas.style.zIndex = (midx + 1) * 10;
      }
      
      verifyEmbeddedWindows();

      // splice returns an array, so return the first element
      return removed[0];
   };
   
   /**
    * Moves the given window to the top of the z-order list and resets the z-order values of all
    * windows that are now moved below the given window.  This compresses the z-index values and
    * eliminates any gaps.
    *
    * @param    {Number} wid
    *           The window ID to move.
    */
   function moveZOrderEntryToTop(wid)
   {
      var last  = zlist.length - 1;
      var found = findZOrderEntry(wid);
      
      if (found != last)
      {
         var entry = removeZOrderEntryWorker(found);
         
         addZOrderEntry(entry.id, entry.win);
      }
   };
   
   /**
    * Set the font in the current context.
    * <p>
    * The context font is always set, as someone else might have change it...
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw.
    * @param    {Number}  fontId
    *           The font identifier.
    */
   function setFont(ctx, fontId)
   {
      var fname = p2j.fonts.getFontName(fontId);
   
      ctx.font = fname;
   
      currentFont = fontId;
   };

   /**
    * Find the next split position, so that the sub-text fits the given width.
    *
    * @param    {Number} font
    *           The font used to measure the text.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} start
    *           The index where to start splitting.
    * @param    {Number} width
    *           The width in which the sub-text must fit.
    *
    * @return   {Number} The end split position, exclusive.
    */
   function findNextPos(font, txt, start, width)
   {
      var end = start;
      do
      {
         var nextWord = regexIndexOf(txt, /\s/g, end);

         if (nextWord == -1)
         {
            // no more words, use the entire text from "start"
            nextWord = txt.length;
         }

         var candidate = txt.substring(start, nextWord);
         var cwidth = p2j.fonts.getTextWidth(font, candidate);

         if (cwidth > width)
         {
            // two cases
            if (end == start)
            {
               // if it's the start of the line, find where to split the word
               end = splitWord(font, txt, start, width);
            }
            else
            {
               // not the start of the line, break on the previous word
            }

            break;
         }

         // the line is not yet filled, continue
         end = nextWord + 1;
      }
      while (end < txt.length);

      return end;
   };

   /**
    * Split the word starting on the given index so that it fits the specified width.
    *
    * @param    {Number} font
    *           The font used to measure the text.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} start
    *           The index where to start splitting.
    * @param    {Number} width
    *           The width in which the sub-text must fit.
    *
    * @return   {Number} The end split position, exclusive.
    */
   function splitWord(font, txt, start, width)
   {
      var end = start + 1;
      while (end < txt.length)
      {
         var candidate = txt.substring(start, end);
         var cwidth = p2j.fonts.getTextWidth(font, candidate);
         
         if (cwidth > width)
         {
            break;
         }

         end = end + 1;
      }

      return end;
   };

   /**
    * Find the next occurrence of the character given by the specified regex, starting from the
    * specified index.
    *
    * @param    {String} txt
    *           The text.
    * @param    {regex} regex
    *           The regular expression used for search.
    * @param    {Number} start
    *           The index where to start searching.
    *
    * @return   {Number} The next occurrent of the given regex, or -1 if not found.
    */
   function regexIndexOf(txt, regex, start)
   {
      var indexOf = txt.substring(start).search(regex);
      return (indexOf >= 0) ? (indexOf + start) : indexOf;
   };
   
   /**
    * Helps to support the dialogs modality of the P2j embedded web application.
    */
   function EmbeddedAppController()
   {
      /** The top most enclosed window */
      var parent    = window.top;
      
      /** Indicates that the web client is embedded in IFRAME. */
      var isInFrame = (parent !== window);
      
      /** The target domain that represents the enclosed window */
      var origin;
      try
      {
         origin = parent.location.origin;
      }
      catch(e)
      {
         console.error(e);
         origin = "*";
      }
      
      function createOverlayPanel(id, zIndex, container)
      {
         var overlayPanel = document.getElementById(id);
   
         if (!overlayPanel)
         {
            overlayPanel    = document.createElement("div");
            overlayPanel.id = id;
            container.appendChild(overlayPanel);
         }
   
         overlayPanel.style.position = "fixed";
         overlayPanel.style.top      = "0px";
         overlayPanel.style.left     = "0px";
         overlayPanel.style.width    = "100%";
         overlayPanel.style.height   = "100%";
   
         overlayPanel.style.background = "black";
         overlayPanel.style.opacity    = 0.5;
         overlayPanel.style.display    = "none";
         overlayPanel.style.zIndex     = zIndex;
   
         return overlayPanel;
      };
      
      // create an overlay panel to block all canvas elements except the top-most one
      var internalOverlayPanel = createOverlayPanel("p2jInternalOverlayPanel", 1, document.body);
      var overlayPanelStates = [];
      overlayPanelStates.push("none");
      
      /**
       * Sends the "display" notification to the top window, crossing the source domain,
       * in order to display the top window overlay panel.
       */
      function displayOverlayPanel(id)
      {
         if (isInFrame)
         {
            parent.postMessage({"target" : "OverlayPanel", "visible" : true, "id" : id}, origin);

            overlayPanelStates.push(internalOverlayPanel.style.display);
            internalOverlayPanel.style.display = "block";

            setInternalZIndex();
         }
      };
      this.displayOverlayPanel = displayOverlayPanel;
      
      /**
       * Sends the "hide" notification to the top window, crossing the source domain,
       * in order to hide the top window overlay panel.
       */
      function hideOverlayPanel(id)
      {
         if (isInFrame)
         {
            parent.postMessage({"target" : "OverlayPanel", "visible" : false, "id" : id}, origin);

            var state = overlayPanelStates.pop();
            internalOverlayPanel.style.display = state;
            
            setInternalZIndex();
         }
      };
      this.hideOverlayPanel = hideOverlayPanel;
      
      /**
       * Set the z-index for the internal overlay panel; execute only if the panel is still
       * visible.
       */
      function setInternalZIndex()
      {
         if (!isInFrame)
         {
            return;
         }
      
         var state = internalOverlayPanel.style.display;
         if (state != "block")
         {
            return;
         }
      
         var idx = zlist.length - 1;
         var c = undefined;
         while (idx >= 0)
         {
            var win = zlist[idx].win;
            c = win.canvas;
            
            if (win.visible)
            {
               break;
            }
            
            idx = idx - 1;
         }
         
         internalOverlayPanel.style.zIndex = (c == undefined ? 0 : c.style.zIndex - 1);
      };
      this.setInternalZIndex = setInternalZIndex;

      /**
       * Send a message the given window has been destroyed.
       *
       * @param   {integer} windowID
       *          The window ID.
       */
      function windowDestroyed(windowId)
      {
         parent.postMessage({target : "WindowDestroyed", "windowId" : windowId}, origin);
      };
      this.windowDestroyed = windowDestroyed;
      
      /**
       * Sends the "bgcolor" notification to the top window, crossing the source domain,
       * in order to set the emedded frame background color according to the window workspace.
       * 
       * @param    {String} bgcolor
       *           The background color as an html string.
       */
      function setFrameBackground(bgcolor)
      {
         if (isInFrame)
         {
            parent.postMessage({"target" : "Frame", "bgcolor" : bgcolor}, origin);
         }
      };
      this.setFrameBackground = setFrameBackground;
      
      /**
       * Verify the windows in embedded mode: only the first visible non-modal, non-shared-state window
       * and visible modal dialogs remain visible, in z-order.
       *
       * @param   {Array} zlist
       *          The window list by their z-order.
       */
      function verifyEmbeddedWindows(zlist)
      {
         var hidden = false;
         var screenDim = me.getDesktopDimension();
         
         var topHeight = 0; // screenDim.height
         var topWidth = screenDim.width;
         var topX = 0;
         var topY = 0;
         
         var visibleWindows = { };
         var topWindow = undefined;
         
         for (var i = zlist.length - 1; i >= 0; i--)
         {
            var win = zlist[i].win;
   
            var c = win.canvas;
            
            if (hidden || !win.isVisible())
            {
               c.style.display = "none";
            }
            else
            {
               c.style.display = "";
               
               if (!(win.modal || win.isShareActivationWithOwner))
               {
                  topWidth = c.width;
                  topHeight = c.height;
   
                  // center only horizontally
                  topX = (screenDim.width - c.width) / 2;
                  if (topX < 0) topX = 0;
                  topY = 0; // (screenDim.height - c.height) / 2;
                  
                  c.style.left = topX + "px";
                  c.style.top = topY + "px";
                  
                  topWindow = win.id;
               }
            }
            
            // TODO: if first is a modal dialog, then remove the decorations;
            
            // only first non-modal window is left visible; overlay windows or modal windows are left
            // visible
            if (!(win.modal || win.isShareActivationWithOwner))
            {
               hidden = true;
            }
         }
   
         for (var i = zlist.length - 1; i >= 0; i--)
         {
            var win = zlist[i].win;
   
            var c = win.canvas;
            
            if (!win.clientMove && win.modal && !win.embeddedCentered)
            {
               var left = topX + (topWidth - c.width) / 2;
               if (left < 0) left = 0;
               c.style.left = left + "px";
               
               var top = topY + (topHeight - c.height) / 2;
               if (top < 0) top = 0;
               c.style.top = top + "px";
               
               win.embeddedCentered = true;
            }
         }
         
         for (var wid in winlist)
         {
            var win = winlist[wid];
   
            if (win.visible && !(win.modal || win.isShareActivationWithOwner))
            {
               visibleWindows[wid] = win.title;
            }
         }
         
         var msg =
         {
            "target"    : "VisibleWindows", 
            "windows"   : visibleWindows,
            "topWindow" : topWindow
         };
         parent.postMessage(msg, origin);
      };
      this.verifyEmbeddedWindows = verifyEmbeddedWindows;

     /**
      * Get the desktop's dimension.
      *
      * @return   {object}
      *           An associative array with width/height properties.
      */
      function getDesktopDimension()
      {
         var w = window.innerWidth;
         var h = window.innerHeight;
         
         var dim =
         {
            "width" : w,
            "height" : h
         };
         
         return dim;
      };
      this.getDesktopDimension = getDesktopDimension;
   }

   /**
    * Get the desktop's dimension.
    *
    * @return   {object}
    *           An associative array with width/height properties.
    */
   me.getDesktopDimension = function()
   {
      if (embeddedAppController)
      {
         return embeddedAppController.getDesktopDimension();
      }
      else
      {
         return { "width" : window.innerWidth, "height" : window.innerHeight };
      }
   };

   /**
    * Determine if the specified event can be processed.  If the OS events or the mouse events are
    * disabled, return false.  Also, return false if the event is consumed.
    *
    * @param    {MouseEvent} evt
    *           The event details.
    * @param    {Window} win
    *           The window object.
    *
    * @return   <code>true</code> if the event can be processed.
    */
   me.canProcessOsEvent = function(evt, win)
   {
      return (win.processOsEvents   &&
              me.processMouse       &&
              !evt.defaultPrevented &&
              !p2j.isDialogDisplayed());
   }
   
   /**
    * Check if there is a high-level OS event registered for this widget and mouse operation.
    *
    * @param    {Event} evt
    *           The mouse event.
    * @param    {Window} wThis
    *           The window object.
    * @param    {String} mouseOp
    *           The mouse operation which raised the event.
    * @param    {Number} id
    *           The widget ID.
    *
    * @return   <code>true</code> if there is an OS event for this widget and mouse action.
    */
   me.hasOsEvent = function(evt, wThis, mouseOp, id)
   {
      return me.getOsEvent(evt, wThis, mouseOp, id) != undefined;
   }
   
   /**
    * Check if there is a high-level OS event registered for this widget and mouse operation.
    *
    * @param    {Event} evt
    *           The mouse event.
    * @param    {Window} wThis
    *           The window object.
    * @param    {String} mouseOp
    *           The mouse operation which raised the event.
    * @param    {Number} id
    *           The widget ID.
    *
    * @return   An object representing the found OS event for this widget and mouse operation.
    */
   me.getOsEvent = function(evt, wThis, mouseOp, id)
   {
      var mousePos = wThis.getMousePos(evt);

      for (var osOp in wThis.osAwareWidgets)
      {
         var ma = wThis.osAwareWidgets[osOp][id];
         
         if (ma == undefined)
         {
            continue;
         }
         
         var osEvents = ma.getEvents();
         
         if (osEvents[mouseOp] != undefined)
         {
            return wThis.osAwareWidgets[osOp][id];
         }
      }
      
      return undefined;
   }
   
   /**
    * Create a new top-level window (and the canvas that backs it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    showInTaskbar
    *           Indicates whether the window should appear on the task bar.
    *
    * @return   The created window instance or <code>null</code> on any error.
    */
   me.createWindow = function(wid, showInTaskbar)
   {
      return createWindowInternal(wid, true, showInTaskbar, false);
   }
   
   /**
    * Create a new top-level window (and the canvas that backs it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    canBeMinimized
    *           The boolean flag indicates that this window must be represented by a task icon on
    *           the tools bar.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *
    * @return   The created window instance or <code>null</code> on any error.
    */
   function createWindowInternal(wid, canBeMinimized, showInTaskBar, modal)
   {
      if (!isValidWindowId(wid))
      {
         me.error("Window ids must be non-positive integers.");
         return null;
      }
      
      if (isExistingWindowId(wid))
      {
         me.error("Cannot create window with id " + wid.toString() + " because it already exists.");
         return null;                       
      }
      
      var newwin = new Window(wid);
      newwin.modal = modal;
      newwin.showInTaskBar = showInTaskBar;
      
      // store in our window list
      winlist[wid] = newwin;
      
      // add to the z-order list
      addZOrderEntry(wid, newwin);
      
      if (canBeMinimized && showInTaskBar)
      {
         // creates new task icon widget for new window
         if (taskBar)
         {
            taskBar.addTaskIcon(newwin.id);
         }
      }

      return newwin;
   };
   
   /**
    * Set up file limit valuses used to upload files with drag-n-drop operations.
    *
    * @param    single
    *           New value for single file size limit.
    * @param    total
    *           New value for total file size limit for single operation.
    */
   me.setFileSizeLimits = function(single, total)
   {
      singleFileLimit = single;
      totalFilesLimit = total;
   };
   
   /**
    * Create a new child window (and the canvas that backs it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    owner
    *           The id of the owning window.
    * @param    title
    *           The title text for the new window.
    * @param    resizable
    *           <code>true</code> if the window can be resized.
    * @param    isShareActivationWithOwner
    *           This flag indicates if this window is active with its owner or not.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *
    * @return   The created window instance or <code>null</code> on any error.
    */
   me.createChildWindow = function(wid, owner, title, resizable, isShareActivationWithOwner, modal)
   {
      if (!isValidWindowId(owner))
      {
         me.error("Owner window ids must be non-positive integers.");
         return null;
      }
      
      if (!isExistingWindowId(owner))
      {
         me.error("Owner window with id " + wid.toString() + " must already exist.");
         return null;
      }
      
      var newwin = createWindowInternal(wid, false, false, modal);
      
      if (newwin === null)
         return null;
      
      newwin.owner = owner;
      newwin.title = title || "";
      newwin.isResizable = resizable;
      newwin.isShareActivationWithOwner = isShareActivationWithOwner;
      
      var parent = getWindow(owner);
      
      // add to the owned window list
      parent.children.push(wid);
      
      verifyEmbeddedWindows();
   };
   
   /**
    * Destroy the given window (and all resources that back it).
    *
    * @param    wid
    *           The window id by which this window can be uniquely identified.
    * @param    images
    *           The cached images to remove.
    */
   me.destroyWindow = function(wid, images)
   {
      if (!isValidWindowId(wid))
      {
         me.error("Cannot destroy window for id that is not a non-positive integer.");
         return;
      }
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot destroy window with id " + wid.toString() + " because it doesn't exist.");
         return;
      }
      
      var win = getWindow(wid);
      
      // remove its task icon
      if (taskBar)
      {
         taskBar.removeTaskIcon(wid);
      }
      
      for (var image in images)
      {
         loadedImages.delete(image);
      }
      
      // remove from our window list
      delete winlist[wid];
      
      // remove the window from the z-order
      removeZOrderEntry(wid);
      
      win.cleanup();

      if (win != null                     &&
          win.visible                     &&
          !win.isShareActivationWithOwner &&
          embeddedAppController           &&
          win.modal)
      {
         embeddedAppController.hideOverlayPanel(wid);
      }

      if (embeddedAppController)
      {
         // notify the embedded app that this window has been destroyed...
         embeddedAppController.windowDestroyed(wid);
      }
   };
      
   /**
    * Change the sensitivity of the given window.
    *
    * @param    {int} wid
    *           The window id by which this window can be uniquely identified.
    * @param    {boolean} enabled
    *           Flag that determines the new sensitivity state.
    */
   me.setWindowEnabled = function(wid, enabled)
   {
      //p2j.logger.log("setWindowEnabled: wid = " + wid + " enabled = " + enabled);
      if (!isValidWindowId(wid))
      {
         me.error("Cannot change sensitivity for window for id that is invalid.");
         return;
      }
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot change sensitivity for non-existent window with id " + wid.toString());
         return;
      }
      
      var win = getWindow(wid);
      
      win.setEnabled(enabled);
   };
   
   /**
    * Change the visibility of the given window.
    *
    * @param    {int} wid
    *           The window id by which this window can be uniquely identified.
    * @param    {boolean} visible
    *           Flag that determines the new visibility state.
    */
   me.setWindowVisible = function(wid, visible)
   {
      //p2j.logger.log("setWindowVisible: wid = " + wid + " visible = " + visible);
      if (!isValidWindowId(wid))
      {
         me.error("Cannot change visibility for window for id that is invalid.");
         return;
      }
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot change visibility for non-existent window with id " + wid.toString());
         return;
      }
      
      var win = getWindow(wid);
      
      win.setVisible(visible);
      // view/hide task icon usecases 
      if (taskBar)
      {
         taskBar.setTaskIconVisible(wid, visible);
      }

      verifyEmbeddedWindows();
   };
   
   /**
    * Get the ID of the topmost window
    * 
    * @return   The window ID.
    */
   me.topWindowId = function()
   {
      var last  = zlist.length - 1;
      return zlist[last].id;
   };
   
   /**
    * Get the topmost visible window
    * 
    * @return   The window.
    */
   me.topVisibleWindow = function()
   {
      var last  = zlist.length - 1;
      while (last >= 0)
      {
         if (zlist[last].win.isVisible())
         {
            return zlist[last].win;
         }
         
         last = last - 1;
      }
      
      return undefined;
   };
   
   /**
    * Moves the given window to the top of the z-order.
    *
    * @param    Number wid
    *           The window ID to move.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window.
    */
   me.moveToTop = function(wid, focus)
   {
      var win = getWindow(wid);
      if (win)
      {
         win.moveToTop();
         if (focus)
         {
            p2j.screen.setActiveWindowId(wid);
         }
      }
   }
   
   /**
    * Moves the given window to the bottom of the z-order.
    *
    * @param    Number wid
    *           The window ID to move.
    */
   me.moveToBottom = function(wid)
   {
      var i;
      
      var wlist = [];
      
      wlist.push(wid);
      
      for (i = 0; i < zlist.length; i++)
      {
         if (zlist[i].id !== wid)
         {
            wlist.push(zlist[i].id);
         }
      }
      
      me.restackZOrderEntries(wlist);
   }

   /**
    * Set the id of the active window.
    *
    *  @param    win
    *            the active window
    */
   me.setActiveWindowId = function(winId)
   {
      var prevId = me.activeWindowId;
      me.activeWindowId = winId;
      var win = getWindow(winId);
      if (taskBar && !win.isShareActivationWithOwner && win.showInTaskBar)
      {
         if (prevId)
         {
            taskBar.setTaskIconActive(prevId, false);
         }

         if (winId)
         {
            taskBar.setTaskIconActive(winId, true);
         }
      }
   }

   /**
    * Returns id of the current active window.
    *
    * @returns    window id
    */
   me.getActiveWindowId = function()
   {
      return me.activeWindowId;
   }

   /**
    * Returns the current active window.
    *
    * @returns    window
    */
   me.getActiveWindow = function()
   {
      if (me.activeWindowId)
      {
         return getWindow(me.activeWindowId);
      }
      else
      {
         return undefined;
      }
   }

   /**
    * Set the id of the active mouse hover window.
    * 
    *  @param win
    *         the active window 
    */
   me.setActiveHoverWindow = function(win)
   {
      me.activeHoverWindow = win;
   }

   /**
    * Resets the z-order values of all windows to match the given order.
    *
    * @param    {Number[]} wids
    *           The list of window IDs in the new z-order.
    */
   me.restackZOrderEntries = function(wids)
   {
      console.log("restack request");
      for (var i = 0; i < wids.length; i++)
      {
         console.log(wids[i]);
      }
      
      var newZlist = zlist; 
      
      // move the windows in wids to the top of the temp array
      for (var j = 0; j < wids.length; j++)
      {
         // wids cannot be a sparse array
         if (!(typeof wids[j] === 'number'))
         {
            me.error("Element " + j.toString() + " in restacking list is invalid!");
            continue;
         }
         
         for (var i = 0; i < newZlist.length; i++)
         {
            if (newZlist[i] == null)
            {
               continue;
            }
            
            var id = newZlist[i].id;

            if (id == wids[j])
            {
               newZlist.push(newZlist[i]);
               newZlist[i] = null;
               break;
            }
         }
      }
      
      // drop the old list
      zlist = [];
      
      // compact newZlist and re-add each element in order
      for (var i = 0; i < newZlist.length; i++)
      {
         if (newZlist[i] == null)
         {
            continue;
         }
         
         addZOrderEntry(newZlist[i].id, newZlist[i].win);
      }
      
      console.log("restack");
      for (var i = 0; i < zlist.length; i++)
      {
         console.log(zlist[i]);
      }
      
      if (embeddedAppController)
      {
         embeddedAppController.setInternalZIndex();
      }
   };
   
   /**
    * Obtain the specified window.
    *
    * @param    wid
    *           The window id to obtain.
    *
    * @return   The window instance or <code>null</code> if a window of the given id does not
    *           exist in the window list.
    */
   me.getWindow = function(wid)
   {
      return getWindow(wid);
   };
   
   /*
    * Retrieve the current font used to draw text.
    *
    * @return   See above.
    */
   me.getCurrentFont = function()
   {
      return currentFont;
   };
   
   /**
    * Checks if the given value already exists as a window id.
    *
    * @param    wid
    *           The value to test.
    *
    * @return   <code>false</code> if the given value does not exist in the window list.
    */
   me.isExistingWindowId = function(wid)
   {
      return isExistingWindowId(wid);
   };

   /**
    * Disable or enable mouse processing.
    *
    * @param   {boolean} capture
    *          When <code>true</code>, it will process mouse events.
    */
   me.captureMouseEvents = function(capture)
   {
      me.processMouse = capture;
   }

   /**
    * Disable or enable OS event processing.
    *
    * @param    {Number} wids
    *           The target window IDs for which OS event processing should be enabled or disabled.
    * @param    {Boolean} capture
    *           When <code>true</code>, it will process OS events like window close/iconify/etc.
    */
   me.enableOsEvents = function(wids, capture)
   {
      for (var j = 0; j < wids.length; j++)
      {
         var win = getWindow(wids[j]);
         if (win)
         {
            win.enableOsEvents(capture);
         }
      }
   }
   
   /**
    * Display an error message.
    * 
    * @param    {string} text
    *           The error text to display.
    */
   me.error = function(text)
   {
      console.error(text);
   };
   
   /**
    * (Re)set wait state
    * 
    * @param isWaitState
    *        the new wait state value
    */
   me.setWaitState = function(isWaitState)
   {
      me.isWaitState = isWaitState;
   }

   /**
    * Lock the mouse pointer
    * 
    * @param {Number} wid
    *        the id of the widget which caused the lock 
    * @param {Object} ptr
    *        the locked mouse pointer 
    */
   me.lockMousePointer = function(wid, ptr)
   {
      if (me.isWaitState && !ptr)
      {
         return;
      }
      me.lockedMousePointer = ptr;
      me.activeDropDown = wid;
      me.setActiveHoverWindow(me.getActiveWindow());
      me.refreshHovered();
   }
   
   /**
    * Set current mouse position (absolute)
    * @param {Object} pos
    *        the current mouse position
    * 
    */
   me.setMousePos = function(pos)
   {
      me.lastMousePos = pos;
   }

   /**
    * Returns current mouse position (absolute)
    *
    * @return mouse position
    *
    */
   me.getMousePos = function()
   {
      return me.lastMousePos;
   }

   /**
    * Refresh mouse cursor for the currently hovered widget
    */
   me.refreshHovered = function()
   {
      if (me.activeHoverWindow && me.lastMousePos)
      {
         me.activeHoverWindow.recalcHovered(me.lastMousePos);
         if (me.activeHoverWindow.currentHovered)
         {
            me.activeHoverWindow.currentHovered.setCursorPtr();
         }
      }
   }
   /**
    * Check if the mouse pointer is locked
    * 
    * @returns <code>true>/code> if the mouse pointer is locked
    */
   me.isMousePointerLocked = function()
   {
      return lockedMousePointer != null;
   }

   /**
    * Get the locked mouse pointer
    * 
    * @returns the locked mouse pointer
    */
   me.lockedMousePtr = function()
   {
      return me.lockedMousePointer;
   }

   
   /**
    * Get the id of the widget which caused the mouse pointer lock
    *
    *  @returns The id of the widget which caused the mouse pointer lock
    */
   me.getActiveDropDown = function()
   {
      return me.activeDropDown;
   }

   /**
    * Notifies the server that the target window should be active and restores its state
    * if the target window has been minimized before.
    * 
    * @param    {Number} deactivatingWindowId
    *           Id of the window being deactivated. If -1, there is no deactivating
    *           window, i.e. previous active was a non-P2J window.
    * 
    * @param    {Number} activatingWindowId
    *           Id of the window being activated. If -1, there is no activating
    *           window, i.e. new active is a non-P2J window.
    */
   function doWindowActivationChange(deactivatingWindowId, activatingWindowId)
   {
      // do we have a valid window being deactivated?
      if (deactivatingWindowId != -1)
      {
         var win = getWindow(deactivatingWindowId);
         if (win && !win.isVisible())
         {
            p2j.socket.sendWindowIconState(deactivatingWindowId, false);
         }
      }
      
      // do we have a valid window being activated?
      if (activatingWindowId != -1)
      {
         var win = getWindow(activatingWindowId);
         // de-iconify the window
         if (win)
         {
            // Notify GuiWebDriver that this window changes its minimized state
            p2j.socket.sendWindowIconState(activatingWindowId, false);
         }
      }
      
      // send the window activation to the java side
      if (activatingWindowId != -1 || deactivatingWindowId != -1)
      {
         p2j.screen.setActiveWindowId(activatingWindowId);
         p2j.socket.sendWindowActivated(deactivatingWindowId, activatingWindowId);
      }
   }

   /**
    * Initialize module.
    *  
    * @param {object} cfg configuration.
    */
   me.init = function(cfg) 
   {
      // disable mouse event
      me.captureMouseEvents(false);
      if (cfg.embedded)
      {
         me.embedded = true;
         embeddedAppController = new EmbeddedAppController();
      }
      else
      {
         desktop = new VirtualDesktop(
               doWindowActivationChange,
               [174, 174, 174],
               [144, 144, 144],
               p2j.fonts,
               p2j.logger);
         taskBar = desktop.getTaskBar();
         taskBar.setFont(cfg);
         
         // register listener for window resize; this is added to the document, to allow event 
         // notification during mouse drag even if mouse is outside of the window
         var resizeHandler = new MouseResizeable();
      }
      graphicsCached = cfg.graphicsCached;
      document.addEventListener("mousemove", function (evt) { me.processScreenMouseMove(evt); });
   };

   /**
    * Set cursor position.
    */
   me.setCursorPosition = function()
   {
      me.error("setCursorPosition() should not be used in GUI!");
   };

   /**
    * Set cursor visibility
    */
   me.setCursorStatus = function()
   {
      me.error("setCursorStatus() should not be used in GUI!");
   };

   /**
    * Set the mouse cursor to the given predifined cursor style. The cursor style id is mapped
    * to the predefined cursors styles: auto | default | none | context-menu | help | pointer
    *  | progress | wait | cell | crosshair | text | vertical-text | alias | copy | move | no-drop
    *  | not-allowed | e-resize | n-resize | ne-resize | nw-resize | s-resize | se-resize
    *  | sw-resize | w-resize | ew-resize | ns-resize | nesw-resize | nwse-resize | col-resize
    *  | row-resize | all-scroll | zoom-in | zoom-out | grab | grabbing 
    * 
    * @param    {Number} cursorStyleId
    *           Denotes the cursor style id.
    * @param    {Number} wid
    *           Denotes the active window id.
    */
   me.setCursorStyle = function(cursorStyleId, wid)
   {
      var style;
      if (((typeof(cursorStyleId) == "string") || (cursorStyleId instanceof String)) && 
         cursorStyleId.startsWith("url("))
      {
         style = cursorStyleId;
      }
      else
      {
         style = me.mouseCursors[cursorStyleId];
         if (!style)
         {
            style = "default";
         }
      }

      var win = getWindow(wid);
      
      if (win)
      {
         win.setCursorStyle(style);
      }
      else
      {
         console.log("Unknown window: %s", wid);
      }
   };

   /**
    * Clear the screen.
    */
   me.clear = function()
   {
      me.error("clear() should not be used in GUI!");
   };

   /**
    * Enable or disable image smoothing for the given graphics context.
    * 
    * @param    {CanvasRenderingContext2D} graphicsContext
    *           The given graphics context
    * @param    {boolean} enabled
    *           The enable flag that must be true to enable image smoothing and false to disable.
    */
   function setImageSmoothing(graphicsContext, enabled)
   {
      if (graphicsContext.imageSmoothingEnabled)
      {
         graphicsContext.imageSmoothingEnabled = enabled;
      }
      else
      {
         if (graphicsContext.mozImageSmoothingEnabled)
         {
            graphicsContext.mozImageSmoothingEnabled = enabled;
         }
         
         if (graphicsContext.webkitImageSmoothingEnabled)
         {
            graphicsContext.webkitImageSmoothingEnabled = enabled;
         }
         
         if (graphicsContext.msImageSmoothingEnabled)
         {
            graphicsContext.msImageSmoothingEnabled = enabled;
         }
      }
   }

   /**
    * Draw the given list of operations in the specified window. This is a completely different
    * implementation than for ChUI, but we are re-using the message type.
    *
    * @param    {byte[]} message
    *           The encoded list of drawing operations.
    */
   me.drawRectangles = function(message)
   {
      // get the MD5 code
      var md5a = new Uint8Array(message.buffer.slice(1, 17));
      var md5 = p2j.socket.createHexString(md5a);

      var wid = p2j.socket.readInt32BinaryMessage(message, 17);
      
      if (!isValidWindowId(wid))
      {
         me.error("Cannot draw window for id that is invalid.");
         return;
      }                                    
      
      if (!isExistingWindowId(wid))
      {
         me.error("Cannot draw non-existent window with id " + wid.toString());
         return;
      }
      
      var win = getWindow(wid);
      
      var oldMouse = me.processMouse;
      me.processMouse = false;

      // if hash sum are filled with zeros, then don't cache drawing operations
      var magic = true;
      if (graphicsCached)
      {
         for (var i = 0; i < 16; i++)
         {
            if (md5a[i] != 0)
            {
               magic = false;
               break;
            }
         }
      }
      
      if (win.hasCachedDraw(md5) && !magic)
      {
         var cachedDraw = win.getCachedDraw(md5);
         
         // blast the image from cache unto the canvas
         win.ctx.save();
         win.ctx.globalAlpha = 1;
         win.ctx.drawImage(cachedDraw, 0, 0);
         win.ctx.restore();
      }
      else
      {
         win.draw(message, md5);
      }
      
      me.processMouse = oldMouse;
   };
   
   /**
    * The method performs a layout operation on the supplied text and returns the resulting 
    * paragraph height while maintaining the supplied maximum width.
    *
    * @param    {CanvasRenderingContext2D} ctx
    *           The canvas 2D graphics context on which to draw, when <code>ctx</code> is set,
    *           the text should be drawn.
    * @param    {String} txt
    *           The paragraph text.
    * @param    {Number} font
    *           The font ID with which the layout/drawing will be performed.
    * @param    {Number} x
    *           The drawing X position, ignored if drawing is not required.
    * @param    {Number} y
    *           The drawing Y position, ignored if drawing is not required.
    * @param    {Number} width
    *           The maximum width used to layout the paragraph.
    * 
    * @return   {Number} The paragraph height.
    */
   me.layoutParagraphWorker = function(ctx, txt, font, x, y, width)
   {
      var newLine = true;
      var i = 0;
      if (ctx)
      {
         // unless we are drawing, there is no need in changing the context font
         ctx.save();
      
         setFont(ctx, font);
      }
      
      var height = 0;
      var fheight = p2j.fonts.getFontHeight(font);
      while (i < txt.length)
      {
         var nextPos = findNextPos(font, txt, i, width);

         var line = txt.substring(i, nextPos);

         if (ctx)
         {
            ctx.fillText(line, x, y);
         }

         y = y + fheight;
         i = nextPos;
         
         height = height + fheight;
      }
      
      if (ctx)
      {
         ctx.restore();
      }
      
      return height;
   };

   /**
    * Verify the windows in embedded mode: only the first visible non-modal, non-shared-state window
    * and visible modal dialogs remain visible, in z-order.
    *
    * @param   {Array} zlist
    *          The window list by their z-order.
    */
   function verifyEmbeddedWindows()
   {
      if (!embeddedAppController)
      {
         return;
      }
      
      embeddedAppController.verifyEmbeddedWindows(zlist);
   }

   /**
    * Handles mouse move on the screen and preserves the last event.
    *
    * @param   evt
    *          The mouse move event.
    */
   me.processScreenMouseMove = function(evt)
   {
      me.setMousePos(evt);
   };

   return me;
}());
