/*
** Module   : p2j.strokes.js
** Abstract : GUI-specific screen management
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI 20150819 The first version defines LineStrokes class as a strokes manager.
** 002 SBI 20151012 Added XOR processing and dotted large line stroke support.
** 003 SBI 20160215 Reduced putImageData invocations for horizontal and vertical lines.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Defines the known stroke styles and methods to draw styled outer lines.
 */
function LineStrokes()
{
   /** Joins path segments by extending their outside edges until they meet. */
   var JOIN_MITER = 0;

   /** Joins path segments by rounding off the corner at a radius of half the line width. */
   var JOIN_ROUND = 1;

   /**
    * Joins path segments by connecting the outer corners of their
    * wide outlines with a straight segment.
    */
   var JOIN_BEVEL = 2;

   /** Ends unclosed subpaths and dash segments with no added decoration. */
   var CAP_BUTT = 0;

   /**
    * Ends unclosed subpaths and dash segments with a round
    * decoration that has a radius equal to half of the width
    * of the pen.
    */
   var CAP_ROUND = 1;

   /**
    * Ends unclosed subpaths and dash segments with a square
    * projection that extends beyond the end of the segment
    * to a distance equal to half of the line width.
    */
   var CAP_SQUARE = 2;
   
   /** This is the list of all possible line stoke types. */
   var LineStrokeEnum = 
   {
      DEFAULT           : 0,
      SPECIFIED_WIDTH   : 1,
      DOTTED            : 2,
      DOTTED_SMALL      : 3,
      DOTTED_SMALL_THIN : 4,
      DOTTED_LARGE      : 5
   };

   var LineStroke =
   {
      /** Dotted line specific parameter. */
      DOT_STEP            : 2.0,
      
      /** Dotted line specific parameter. */
      DOT_STEP_SMALL      : 1.0,
      
      /** Dotted line width. */
      DOT_LINE_WIDTH      : 1.0,
      
      /** Dotted thin line width. */
      DOT_LINE_THIN_WIDTH : 0.5,
      
      /** The dash step for the dotted large style. */
      DOT_STEP_LARGE      : 3
   };

   /**
    * Represents a dot style with 2 pixels to draw and then 2 pixels to pass.
    */
   var dotStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH, CAP_BUTT, JOIN_BEVEL, 0.0,
         [LineStroke.DOT_STEP, LineStroke.DOT_STEP], LineStroke.DOT_STEP);

   /**
    * Represents a dot style with 1 pixel to draw and then 1 pixel to pass.
    */
   var dotSmallStroke  = new BasicStroke(LineStroke.DOT_LINE_WIDTH, CAP_BUTT, JOIN_BEVEL, 0.0,
         [LineStroke.DOT_STEP_SMALL, LineStroke.DOT_STEP_SMALL], LineStroke.DOT_STEP_SMALL);

   /**
    * Represents a dot style with 0.5 pixel to draw and then 0.5 pixel to pass.
    */
   var dotSmallThinStroke = new BasicStroke(LineStroke.DOT_LINE_THIN_WIDTH, CAP_BUTT,
         JOIN_BEVEL, 0.0, [LineStroke.DOT_STEP_SMALL, LineStroke.DOT_STEP_SMALL],
         LineStroke.DOT_STEP_SMALL);

   /**
    * Represents a dot style with 3 pixel to draw and then 3 pixel to pass.
    */
   var dotLargeStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH, CAP_BUTT, JOIN_MITER,
         1.0, [LineStroke.DOT_STEP_LARGE, LineStroke.DOT_STEP_LARGE], 0);
   /**
    * Represents a default style to draw a 1 pixel width solid line.
    */
   var defaultStroke = new BasicStroke(1.0, CAP_SQUARE, JOIN_MITER, 10.0, null, 0.0);
   
   /** Holds all java script styles of how the ends of lines are drawn. */
   var caps = [ "butt", "round", "square"];
   
   /**
    * Holds all java script styles of how two connecting segments with non-zero lengths
    * in a shape are joined together.
    */
   var joins = [ "miter", "round", "bevel"];
   
   /**
    * Holds properties for line drawing styles.
    * 
    * @param    width
    *           The width of lines to draw.
    * @param    cap
    *           The decoration of the line ends.
    * @param    join
    *           The decoration applied where path segments meet.
    * @param    miterlimit
    *           The limit to trim the miter join.
    * @param    dash
    *           The array representing the dashing pattern.
    * @param    dash_phase
    *           The offset to start the dashing pattern.
    */
   function BasicStroke(width, cap, join, miterLimit, dash, dash_phase)
   {
      if (width < 1)
      {
         width = 1;
      }
   
      this.getWidth = function() { return width; };
      this.getCap   = function() { return caps[cap]; };
      
      this.getJoin        = function() { return joins[join]; };
      this.getMiterLimit  = function() { return miterLimit; };
      
      this.getDash       = function() { return dash; };
      this.getDashPhase  = function() { return dash_phase; };
   };

   /**
    * Returns the stroke style object.
    * 
    * @param    {Number} strokeStyle
    *           The stroke style id according to the LineStroke enumeration.
    * @param    {Number} width
    *           The line width.
    * 
    * @return   The basic line stroke object.
    */
   function getLineStroke(strokeStyle, width)
   {
      switch (strokeStyle)
      {
         case LineStrokeEnum.SPECIFIED_WIDTH:
            return new BasicStroke(width, CAP_SQUARE, JOIN_MITER, 10.0, null, 0.0);
         case LineStrokeEnum.DOTTED:
            return dotStroke;
         case LineStrokeEnum.DOTTED_SMALL:
            return dotSmallStroke;
         case LineStrokeEnum.DOTTED_SMALL_THIN:
            return dotSmallThinStroke;
         case LineStrokeEnum.DOTTED_LARGE:
            return dotLargeStroke;
         default:
            return defaultStroke;
      }
   };
   
   /**
    * Returns the stroke path renderer.
    * 
    * @param    {CanvasRenderer} canvasRenderer
    *           The canvas renderer.
    * @param    {Number} strokeStyle
    *           The stroke style id according to the LineStroke enumeration.
    * @param    {Number} width
    *           The line width.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    * 
    * @return   The stroke path renderer.
    */
   this.getStrokePathRenderer = function(canvasRenderer, strokeStyle, width, strokeColor)
   {
      var basicStroke = getLineStroke(strokeStyle, width);
      
      switch (strokeStyle)
      {
         case LineStrokeEnum.SPECIFIED_WIDTH:
            return new WidenPathRenderer(canvasRenderer, basicStroke, strokeColor);
         case LineStrokeEnum.DOTTED:
         case LineStrokeEnum.DOTTED_SMALL:
         case LineStrokeEnum.DOTTED_SMALL_THIN:
         case LineStrokeEnum.DOTTED_LARGE:
            return new DotsPathRenderer(canvasRenderer, basicStroke, strokeColor);
         default:
            return new DefaultPathRenderer(canvasRenderer, strokeColor);
      }
   };
   
   /**
    * Apply the given stroke style to the JS native canvas renderer.
    * 
    * @param    {CanvasRenderingContext2D} context
    *           The canvas renderer context.
    * @param    {Number} strokeStyle
    *           The stroke style id.
    * @param    {Number} width
    *           The line width.
    */
   this.applyStrokeToCanvas = function(context, strokeStyle, width)
   {
      if (context instanceof CanvasRenderingContext2D)
      {
         var basicStroke = getLineStroke(strokeStyle, width);
         context.lineWidth = basicStroke.getWidth();
         
         context.lineCap = basicStroke.getCap();
         
         context.lineJoin = basicStroke.getJoin();
         
         context.miterLimit = basicStroke.getMiterLimit();
         if (basicStroke.getDash())
         {
            context.setLineDash(basicStroke.getDash());
         }
         else
         {
            context.setLineDash([]);
         }
         context.lineDashOffset = basicStroke.getDashPhase();
      }
   }
   
   /**
    * Returns the default stroke that is a solid line style with 1 pixel line width.
    * 
    * @return   {Number} The default stroke style.
    */
   this.getDefaultStrokeStyle = function()
   {
      return LineStrokeEnum.DEFAULT;
   }

   /**
    * Defines StrokeRenderer class
    */
   function StrokeRenderer()
   {
   }

   /**
    * Defines the constructor function
    */
   StrokeRenderer.prototype.constructor = StrokeRenderer;

   /**
    * Prepare the object. This method is used to initialize and to reset the stroke renderer. 
    */
   StrokeRenderer.prototype.beginStroke = function()
   {
   };

   /**
    * Apply the stroke to lines given by the absolute coordinates. This method is used to stroke
    * only vertical and horizontal lines.
    * 
    * @param    {Number} x1
    *           X coordinate of the starting pixel to be drawn.
    * @param    {Number} y1
    *           Y coordinate of the starting pixel to be drawn.
    * @param    {Number} x2
    *           X coordinate of the ending pixel to be drawn.
    * @param    {Number} y2
    *           Y coordinate of the ending pixel to be drawn.
    */
   StrokeRenderer.prototype.strokeLine  = function(x1, y1, x2, y2)
   {
   };
   
   /**
    * Apply the stroke to the next point given by its coordinates (x, y). This method is used to
    * stroke sloped and curved lines. 
    * 
    * @param    {Number} x
    *           X coordinate of the point to be drawn.
    * @param    {Number} y
    *           Y coordinate of the point to be drawn.
    */
   StrokeRenderer.prototype.strokePoint = function(x, y)
   {
   };
   
   /**
    * Used to draw dash patterns on lines of pixels.
    * 
    * @param    {CanvasRenderer} canvasRenderer
    *           The canvas renderer.
    * @param    {Object} basicStroke
    *           The stroke style object.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    */
   function DotsPathRenderer(canvasRenderer, basicStroke, strokeColor)
   {
      var pixel = canvasRenderer.createImageData(1, 1);
      var pixelData = pixel.data;
      pixelData[3] = 0xFF;
      var state = {};

      /**
       * Normalizes a dash offset distance according to the given dash pattern.
       * A normalized dash offset distance is a distance that is cut the corresponding segment
       * from the dash pattern that is extended periodically in order to cover the target dash
       * offset.
       * 
       * @param    {Array} dash
       *           The dash array that specifies the number of units to be drawn followed
       *           by the number of units to be skipped.
       * @param    {Number} dashOffset
       *           The offset distance from the first pattern point to the point which becomes
       *           a start point of the dash pattern.
       * 
       * @return   The container {startPhase : phase, startDashOn : dashOn, startIdx : idx},
       *           where the startPhase field holds a normalized dash offset, startDashOn defines
       *           to draw or to skip, the value of startIdx is the current dash pattern index.
       */
      function normalize(dash, dashOffset)
      {
         var phase = dashOffset;
         var idx = 0;
         var dashOn = true;
         var d;
         
         while (phase >= (d = dash[idx]))
         {
             phase -= d;
             idx = (idx + 1) % dash.length;
             dashOn = !dashOn;
         }

         return {startPhase : phase, startDashOn : dashOn, startIdx : idx};
      }
      
      /**
       * Prepares the given stroke to be applied to the line of pixels called a path.
       */
      this.beginStroke = function()
      {
         var dash = basicStroke.getDash();
         if (!dash || dash.length == 0)
         {
            return;
         }
         var dashOffset = basicStroke.getDashPhase();
         var len = dash.length;
         
         var patternObj = normalize(dash, dashOffset);
         
         var dashOn = patternObj.startDashOn;
         var phase  = patternObj.startPhase;
         var idx    = patternObj.startIdx;
         var rest = dash[idx] - phase;
         
         state.dash   = dash;
         state.len    = len;
         state.dashOn = dashOn;
         state.idx    = idx;
         state.rest   = rest;
      }
      
      /**
       * Apply the stroke to vertical or horizontal lines given by the absolute coordinates that
       * satisfy x1 <= x2 and y1 <= y2.
       * 
       * @param    {Number} x1
       *           X coordinate of the starting pixel to be drawn.
       * @param    {Number} y1
       *           Y coordinate of the starting pixel to be drawn.
       * @param    {Number} x2
       *           X coordinate of the ending pixel to be drawn.
       * @param    {Number} y2
       *           Y coordinate of the ending pixel to be drawn.
       */
      this.strokeLine = function(x1, y1, x2, y2)
      {
         var rest   = state.rest;
         var idx    = state.idx;
         var len    = state.len;
         var dashOn = state.dashOn;
         var dash   = state.dash;
         // precondition: dx should be greater or equal than zero.
         var dx = x2 - x1;
         var x = x1;
         var y = y1;
         // precondition: y2 should be greater or equal than y1.
         while(x <= x2 && y <= y2)
         {
            if (rest < 1)
            {
               idx    = (idx + 1) % len;
               rest   = state.dash[idx];
               dashOn = !dashOn;
            }
            rest -= 1;
            
            if (dashOn)
            {
               // fill a pixel and put it on the canvas.
               pixelData[0] = strokeColor[0];
               pixelData[1] = strokeColor[1];
               pixelData[2] = strokeColor[2];
               canvasRenderer.putImageData(pixel, x, y);
            }
            // vertical line: increase Y-coordinate
            if (dx === 0)
            {
               y = y + 1;
            }
            else // horizontal line: increase X-coordinate
            {
               x = x + 1;
            }
         }
         // save the dash pattern state in order to continue this dash pattern for next drawings 
         state.dashOn = dashOn;
         state.rest   = rest;
         state.idx    = idx;
      }
      
      /**
       * Apply the stroke to the next point given by its coordinates (x, y).
       * 
       * @param    {Number} x
       *           X coordinate of the point to be drawn.
       * @param    {Number} y
       *           Y coordinate of the point to be drawn.
       */
      this.strokePoint = function(x, y)
      {
         var rest   = state.rest;
         var idx    = state.idx;
         var len    = state.len;
         var dashOn = state.dashOn;
         if (rest < 1)
         {
            idx    = (idx + 1) % len;
            rest   = state.dash[idx];
            dashOn = !dashOn;
         }
         rest -= 1;
         if (dashOn)
         {
            // fill a pixel and put it on the canvas.
            pixelData[0] = strokeColor[0];
            pixelData[1] = strokeColor[1];
            pixelData[2] = strokeColor[2];
            canvasRenderer.putImageData(pixel, x, y);
         }
         // save the dash pattern state in order to continue this dash pattern for next drawings 
         state.dashOn = dashOn;
         state.rest   = rest;
         state.idx    = idx;
      }
      
   }
   
   /**
    * DotsPathRenderer implements StrokeRenderer methods
    */
   DotsPathRenderer.prototype = Object.create(StrokeRenderer.prototype);

   /**
    * Defines the constructor function
    */
   DotsPathRenderer.prototype.constructor = DotsPathRenderer;
   
   /**
    * Used to widen lines of pixels.
    * 
    * @param    {CanvasRenderer} canvasRenderer
    *           The canvas renderer.
    * @param    {Object} basicStroke
    *           The stroke style object.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    */
   function WidenPathRenderer(canvasRenderer, basicStroke, strokeColor)
   {
      var width = basicStroke.getWidth();
      var image = canvasRenderer.createImageData(width, width);
      var imageData = image.data;
      
      for (var i = 0; i < 4 * width * width; i += 4)
      {
         imageData[i] = strokeColor[0];
         imageData[i + 1] = strokeColor[1];
         imageData[i + 2] = strokeColor[2];
         imageData[i + 3] = 0xFF;
      }
      
      /**
       * Prepares the given stroke to be applied to the line of pixels called a path.
       */
      this.beginStroke = function()
      {
      }
      
      /**
       * Apply the stroke to vertical or horizontal lines given by the absolute coordinates that
       * satisfy x1 <= x2 and y1 <= y2.
       * 
       * @param    {Number} x1
       *           X coordinate of the starting pixel to be drawn.
       * @param    {Number} y1
       *           Y coordinate of the starting pixel to be drawn.
       * @param    {Number} x2
       *           X coordinate of the ending pixel to be drawn.
       * @param    {Number} y2
       *           Y coordinate of the ending pixel to be drawn.
       */
      this.strokeLine = function(x1, y1, x2, y2)
      {
         // calculate the minimal rectangle width and height that covers and fills
         //a line that should be widen to have the given width
         var lineWidth  = x2 - x1 + width;
         var lineHeight = y2 - y1 + width;
         var x0 = x1 - (width >> 1);
         var y0 = y1 - (width >> 1);
         // create an image to draw the widen line
         var screenImage = canvasRenderer.ctx.createImageData(lineWidth, lineHeight);
         var data = screenImage.data;
         var length = 4 * lineWidth * lineHeight;
         
         for (var i = 0; i < length; i += 4)
         {
            data[i]     = strokeColor[0];
            data[i + 1] = strokeColor[1];
            data[i + 2] = strokeColor[2];
            data[i + 3] = 255;
         }
         canvasRenderer.putImageData(screenImage, x0, y0);
      }
      
      /**
       * Apply the stroke to the next point given by its coordinates (x, y).
       * 
       * @param    {Number} x
       *           X coordinate of the point to be drawn.
       * @param    {Number} y
       *           Y coordinate of the point to be drawn.
       */
      this.strokePoint = function(x, y)
      {
         canvasRenderer.putImageData(image, x - (width >> 1), y - (width >> 1));
      }

   }
   
   /**
    * WidenPathRenderer implements StrokeRenderer methods
    */
   WidenPathRenderer.prototype = Object.create(StrokeRenderer.prototype);

   /**
    * Defines the constructor function
    */
   WidenPathRenderer.prototype.constructor = WidenPathRenderer;

   /**
    * The default lines renderer.
    * 
    * @param    {CanvasRenderer} canvasRenderer
    *           The canvas renderer.
    * @param    {Number[]} strokeColor
    *           The array of 3 integer values between 0 and 255 inclusive, representing
    *           a stroke color.
    */
   function DefaultPathRenderer(canvasRenderer, strokeColor)
   {
      var width = defaultStroke.getWidth();
      var image = canvasRenderer.createImageData(width, width);
      var imageData = image.data;
      
      for (var i = 0; i < 4 * width * width; i += 4)
      {
         imageData[i]     = strokeColor[0];
         imageData[i + 1] = strokeColor[1];
         imageData[i + 2] = strokeColor[2];
         imageData[i + 3] = 255;
      }
      
      /**
       * Prepares the given stroke to be applied to the line of pixels called a path.
       */
      this.beginStroke = function()
      {
      }
      
      /**
       * Apply the stroke to vertical or horizontal lines given by the absolute coordinates that
       * satisfy x1 <= x2 and y1 <= y2.
       * 
       * @param    {Number} x1
       *           X coordinate of the starting pixel to be drawn.
       * @param    {Number} y1
       *           Y coordinate of the starting pixel to be drawn.
       * @param    {Number} x2
       *           X coordinate of the ending pixel to be drawn.
       * @param    {Number} y2
       *           Y coordinate of the ending pixel to be drawn.
       */
      this.strokeLine = function(x1, y1, x2, y2)
      {
         // The default line width is 1 pixel.
         // Calculate the line with and height to draw a line as an image
         var lineWidth  = x2 - x1 + 1;
         var lineHeight = y2 - y1 + 1;
         // Create a line image
         var screenImage = canvasRenderer.ctx.createImageData(lineWidth, lineHeight);
         var data = screenImage.data;
         var length = 4 * lineWidth * lineHeight;
         // Fill a line image with the given stroke color
         for (var i = 0; i < length; i += 4)
         {
            data[i]     = strokeColor[0];
            data[i + 1] = strokeColor[1];
            data[i + 2] = strokeColor[2];
            data[i + 3] = 255;
         }
         // Put the line image on the canvas
         canvasRenderer.putImageData(screenImage, x1, y1);
      }
      
      /**
       * Apply the stroke to the next point given by its coordinates (x, y).
       * 
       * @param    {Number} x
       *           X coordinate of the point to be drawn.
       * @param    {Number} y
       *           Y coordinate of the point to be drawn.
       */
      this.strokePoint = function(x, y)
      {
         canvasRenderer.putImageData(image, x, y);
      }

   }

   /**
    * DefaultPathRenderer implements StrokeRenderer methods
    */
   DefaultPathRenderer.prototype = Object.create(StrokeRenderer.prototype);

   /**
    * Defines the constructor function
    */
   DefaultPathRenderer.prototype.constructor = DefaultPathRenderer;

}