/*
** Module   : DirDialogGuiImpl.java
** Abstract : Allows user to select a directory from current file system in a GUI dialog.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171028 Initial commit.
** 002 OM  20180321 Fixed keyboard navigation, virtual folders and initial directory. Added Theme
**                  and key navigation support.
** 003 EVL 20180723 Fixed NPE with not defined initial directory.
**     OM  20180727 Added handle for system close button.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import java.awt.image.*;
import java.io.*;
import java.nio.file.*;
import java.util.*;
import java.util.logging.*;

import static com.goldencode.p2j.ui.client.gui.filechooser.FileChooserUtils.*;

/**
 * This GUI dialog allows user to select a directory from current file system. 
 */
public class DirDialogGuiImpl
extends AbstractFileChooserDialog
implements ActionListener
{
   /** Logger for all related classes. */
   private final Logger LOG = LogHelper.getLogger(DirDialogGuiImpl.class.getName());
   
   /** The icon that marks a folder to be expanded, usually a down arrow or [-]. */
   private ImageWrapper<BufferedImage> expImg = null;
   
   /** The icon that marks a folder to be collapsed, usually a right arrow or [+]. */
   private ImageWrapper<BufferedImage> colImg = null;
   
   /** The native width of the dialog. */
   private final int NATIVE_WIDTH = 318;
   
   /** The native height of the dialog. */
   private final int NATIVE_HEIGHT = 288;
   
   /** The height of the folders in tree. */
   private final int FOLDER_HEIGHT = 21;
   
   /** Cached driver. */
   protected GuiDriver gd = null;
   
   /** Cached coordinates. */
   private CoordinatesConversion cc;
   
   /**
    * The container that holds the file-system tree.
    */
   private FolderTree dirTree;
   
   /**
    * The scroll pane that allows the {@code dirTree} to be scrolled up and down. 
    */
   private ScrollPaneGuiImpl treeScroll;
   
   /** A label that displays the title. */
   private SimpleLabel titleLabel;
   
   /** A label for directory name editor. */
   private SimpleLabel nameLabel;
   
   /**
    * The directory name editor should allow the user to manually enter the name of the folder.
    * Note: Not implemented yet. It only work a a view for the file selected by mouse. 
    */
   private FillInGuiImpl dirName;
   
   /** The {@code Make New Folder} button. */
   private ButtonGuiImpl newFolder;
   
   /** The {@code OK} button. Validates the current selection and closes the dialog. */
   private ButtonGuiImpl accept;
   
   /** The {@code Cancel} button. Rejects the dialog. */
   private ButtonGuiImpl reject;
   
   /** The title of the dialog. */
   private String title = "";
   
   /** The folder that is currently selected. */
   private Folder crtFolder = null;
   
   /** Cache {@code This PC} folder. */
   private Folder thisPC = null;
   
   /**
    * The constructor build an object by forwarding the required parameters to super class.
    *
    * @param   fid
    *          The {@link WidgetId} of the dialog. 
    * @param   fCfg
    *          The {@link FrameConfig} of the dialog.
    */
   public DirDialogGuiImpl(WidgetId fid, FrameConfig fCfg)
   {
      super(fid, fCfg);
      setLayout(new DirChooserLayoutManager());
   }
   
   /**
    * Sets the title of this dir chooser dialog.
    * 
    * @param   title
    *          The new title.
    */
   public void setTitle(String title)
   {
      this.title = title;
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    */
   protected void init()
   {
      if (gd != null)
      {
         return; // already initialized
      }
      
      // install system close handler
      super.init();
      
      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      cc = screen().coordinates();
      
      expImg = FileChooserUtils.loadImage("expanded");
      colImg = FileChooserUtils.loadImage("collapsed");
      
      titleLabel = createSimpleLabel(title, 288, 18);
      addWidget(titleLabel);
      
      nameLabel = createSimpleLabel("&Folder:", 53, 23);
      addWidget(nameLabel);
      
      WidgetId fid = WidgetId.nextID();
      FillInConfig fcfg = new FillInConfig(fid);
      fcfg.format = "x(255)";
      fcfg.state = ScreenBuffer.CHANGED;
      fcfg.autoResize = false;
      fcfg.realized = true;
      try
      {
         fcfg.appFormat = new StringFormat("x(255)");
      }
      catch (DisplayFormatParsingException e)
      {
         // just in case, no expected exceptions for this format
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "Unexpected StringFormat", e);
         }
      }
      
      dirName = new FillInGuiImpl(fid, fcfg);
      dirName.initialize(fid, fcfg);
      dirName.setDataType("character");
      dirName.setEnabled(true);
      dirName.setForceLeft(true);
      addWidget(dirName);
      
      dirTree = new FolderTree();
      dirTree.setScrollable(true);
      treeScroll = new ScrollPaneGuiImpl(dirTree);
      treeScroll.setShowBars(ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS);
      treeScroll.setScrollBars(ScrollBar.Bars.VERTICAL);
      addWidget(treeScroll);
      
      newFolder = createButton("&Make New Folder", "MAKE-NEW", 105, 23, false);
      newFolder.setEnabled(false); // TODO: implement "New Folder" features 
      addWidget(newFolder);
      
      accept = createButton("OK", "OK", 75, 23, true);
      config().defaultButtonId = accept.getId().asInt();
      addWidget(accept);
      
      reject = createButton("Cancel", "CANCEL", 75, 23, false);
      config().cancelButtonId = reject.getId().asInt();
      addWidget(reject);
      
      screen().getRegistry().addFrame(this);
      setSize(cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH, NATIVE_HEIGHT)));
      
      String userHome = System.getProperty("user.home") + File.separator;
      String userName = System.getProperty("user.name");
      
      Folder root = createFolder(null, "Desktop", userHome + "Desktop", 0, "desktop");
      root.content = new Folder[6];
      root.content[0] = createFolder(root, userName, userHome, 1, "small-folder");
      root.content[1] = createFolder(root, getDisplayName(VF_THISPC), VF_THISPC, 1, "thispc");
      root.content[2] = createFolder(
            root, getDisplayName(VF_LIBRARIES), VF_LIBRARIES, 1, "libraries");
      root.content[3] = createFolder(root, getDisplayName(VF_NETWORK), VF_NETWORK, 1, "network");
      root.content[4] = createFolder(
            root, getDisplayName(VF_CONTROLPANEL), VF_CONTROLPANEL, 1, "control-panel");
      root.content[5] = createFolder(
            root, getDisplayName(VF_RECYCLEBIN), VF_RECYCLEBIN, 1, "recycle-bin");
      
      // cache [This PC] folder for further lookups 
      thisPC = root.content[1];
      dirTree.setRoot(root);
      
      // set up home directory as initial if none yet specified
      if (initialDir == null || initialDir.isEmpty())
      {
         setInitialDir(userHome);
      }
      
      // set the startup location
      Path iniPath = Paths.get(initialDir);
      if (Files.exists(iniPath) && Files.isDirectory(iniPath))
      {
         crtFolder = selectPath(iniPath);
      }
      else
      {
         crtFolder = root.content[0]; //user's folder
      }
      dirName.setValue(new character(crtFolder.textLabel()));
      
      dirTree.doLayout();
      treeScroll.repaint();
      
      // everything done, go to initial dir (select current file):
      // navigate(initialDir, true, false);
      crtFolder.requestFocus();
   }
   
   /**
    * This method is invoked when action event is generated.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
      String cmd = event.command();
      if (cmd == null)
      {
         return; // unknown command
      }
      
      switch (cmd)
      {
         case "OK":
            setSelectedPath(crtFolder.path);
            accepted = true;
            dismiss();
            return;
         
         case "CANCEL":
            setSelectedPath(null);
            accepted = false;
            dismiss();
            return;
         
         case "MAKE-NEW":
            UnimplementedFeature.missing("DirectoryChooser.newFolder");
            return;
         
         default:
            if (cmd.startsWith("OPEN "))
            {
               boolean doLayout = false;
               String location = cmd.substring(5);
               if (event.source() instanceof Folder)
               {
                  Folder folder = (Folder) event.source();
                  Folder oldSel = crtFolder;
                  crtFolder = folder;
                  dirName.setValue(new character(crtFolder.textLabel()));
                  if (oldSel != null && oldSel != crtFolder)
                  {
                     oldSel.repaint();
                  }
                  
                  if (folder.content == null)
                  {
                     if (isVirtualFolder(location))
                     {
                        populateVirtualFolder(folder, location);
                        if (folder.content.length != 0)
                        {
                           doLayout = true;
                        }
                     }
                     else
                     {
                        Path newPath = Paths.get(folder.path);
                        try
                        {
                           Object[] dirs = Files.list(newPath)
                                                .filter(FileChooserUtils::isPathVisible)
                                                .filter(Files::isDirectory)
                                                .toArray();
                           folder.content = new Folder[dirs.length];
                           for (int k = 0; k < folder.content.length; k++)
                           {
                              Path path = (Path) dirs[k];
                              folder.content[k] = createFolder(folder,
                                                               getFileName(path),
                                                               path.toAbsolutePath().toString(),
                                                               folder.indent + 1,
                                                               "small-folder");
                           }
                           doLayout = true;
                        }
                        catch (IOException e)
                        {
                           LOG.log(Level.WARNING, "Failed to explore [" + folder.path + "]");
                           folder.content = new Folder[0];
                        }
                     }
                  }
                  
                  if (!folder.expanded)
                  {
                     doLayout = true;
                  }
                  folder.expanded = true;
               }
               if (doLayout)
               {
                  dirTree.doLayout();
                  treeScroll.repaint();
               }
            }
      }
   }
   
   /**
    * Adding new widget to the container. This method does not require the coordinates to be
    * defined. The location will be selected automatically. Makes sure this frame is also set on
    * target {@code widget}.
    *
    * @param   widget
    *          The new widget to be added to the container.
    */
   @Override
   public void addWidget(Widget<GuiOutputManager> widget)
   {
      super.addWidget(widget);
      widget.config().frameId = getId().asInt();
   }
   
   /**
    * Creates a customized {@code SimpleLabel} to be used on this dialog. 
    *
    * @param   text
    *          The text to be displayed.
    * @param   w
    *          The physical width of the label.
    * @param   h
    *          The physical height of the label.
    *
    * @return  A customized {@code SimpleLabel}.
    */
   private SimpleLabel createSimpleLabel(String text, int w, int h)
   {
      SimpleLabel label = new SimpleLabel(text);
      label.setSize(cc.dimensionFromNative(new NativeDimension(w, h)));
      LabelConfig lcfg = (LabelConfig) label.config();
      lcfg.align = ControlEntity.ALIGN_LEFT;
      lcfg.frameId = getId().asInt();
      return label;
   }
   
   /**
    * Creates and initializes a button widget.
    *
    * @return   A reference to {@link Button} instance.
    */
   private ButtonGuiImpl createButton(String text, String action, int w, int h, boolean def)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = w;
      bcfg.heightPixels = h;
      bcfg.isDefault = def;
      bcfg.flatButton = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.frameId = getId().asInt();
      
      ButtonGuiImpl b = new ButtonGuiImpl(bid, bcfg);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(action);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      return b;
   }
   
   /**
    * Creates a customized {@code Folder} to be used on this dialog. The new object is
    * automatically added to {@code dirTree} container. 
    *
    * @param   parent
    *          The parent folder. {@code null} if node is the root.
    * @param   text
    *          The text to be displayed on the button.
    * @param   path
    *          The absolute file path.
    * @param   xOffset
    *          The indentation of the image and text on the button.
    * @param   img
    *          The image to be displayed on the button.
    *
    * @return  A customized {@code SimpleImageButton}.
    */
   private Folder createFolder(Folder parent, String text, String path, int xOffset, String img)
   {
      WidgetId bid = WidgetId.nextID();
      
      ButtonConfig bcfg = new ButtonConfig(bid);
      bcfg.widthPixels = -1;
      bcfg.heightPixels = FOLDER_HEIGHT;
      bcfg.isDefault = false;
      bcfg.autoResize = false;
      bcfg.fixedHeight = true;
      bcfg.fixedWidth = true;
      bcfg.frameId = getId().asInt();
      
      Folder b = new Folder(bid, parent, text, path, img, xOffset);
      b.initialize(bid, bcfg);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand("OPEN " + path);
      b.setEnabled(true);
      b.setVisible(true);
      b.addListener(this);
      
      dirTree.add(b);
      
      return b;
   }
   
   /**
    * Populates a virtual folder.
    *
    * @param   folder
    *          The folder to be populated.
    * @param   virtualLocation
    *          The virtual location path (the $ prefix dropped).
    */
   private void populateVirtualFolder(Folder folder, String virtualLocation)
   {
      if (folder == null || folder.content!= null)
      {
         return;
      }
      
      String userHome = System.getProperty("user.home") + File.separator;
      switch (virtualLocation)
      {
         case VF_RECYCLEBIN:
         case VF_CONTROLPANEL:
         case VF_NETWORK:
            folder.content = new Folder[0];
            return;
         case VF_LIBRARIES:
            folder.content = new Folder[4];
            folder.content[0] = createFolder(
                  folder, "Documents", userHome + "Documents", folder.indent + 1, "documents");
            folder.content[1] = createFolder(
                  folder, "Music", userHome + "Music", folder.indent + 1, "music");
            folder.content[2] = createFolder(
                  folder, "Pictures", userHome + "Pictures", folder.indent + 1, "pictures");
            folder.content[3] = createFolder(
                  folder, "Videos", userHome + "Videos", folder.indent + 1, "videos");
            return;
         case VF_THISPC:
            File[] roots = File.listRoots();
            folder.content = new Folder[roots.length + 6];
            folder.content[0] = createFolder(
                  folder, "Desktop", userHome + "Desktop", folder.indent + 1, "desktop");
            folder.content[1] = createFolder(
                  folder, "Documents", userHome + "Documents", folder.indent + 1, "documents");
            folder.content[2] = createFolder(
                  folder, "Downloads", userHome + "Downloads", folder.indent + 1, "downloads");
            folder.content[3] = createFolder(
                  folder, "Music", userHome + "Music", folder.indent + 1, "music");
            folder.content[4] = createFolder(
                  folder, "Pictures", userHome + "Pictures", folder.indent + 1, "pictures");
            folder.content[5] = createFolder(
                  folder, "Videos", userHome + "Videos", folder.indent + 1, "videos");
            for (int k = 0; k < roots.length; k++)
            {
               folder.content[6 + k] = createFolder(
                     folder,
                     "Local Disk (" + roots[k] + ")",
                     roots[k].getAbsolutePath(),
                     folder.indent + 1,
                     "drive");
            }
            return;
         default:
            LOG.log(Level.WARNING, "Unknown Virtual Folder: [" + virtualLocation + "]");
      }
   }
   
   /**
    * Recursively creates a path (in $THIS-PC/$drive/..) of expanded folders and returns the last
    * folder component.
    * 
    * @param   path
    *          The path to be created
    *
    * @return  The 'leaf'. 
    */
   private Folder selectPath(Path path)
   {
      if (path.getFileName() == null)
      {
         onAction(new ActionEvent(thisPC, thisPC.getCommand()));
         return thisPC.lookup("Local Disk (" + path.toString() + ")");
         // we have a root folder
      }
      Folder parent = selectPath(path.getParent());
      onAction(new ActionEvent(parent, parent.getCommand()));
      String folder = path.getFileName().toString();
      return parent.lookup(folder);
   }
   
   /**
    * The {@code LayoutManager} that handles the placement and size of the widgets on this dialog.
    */
   private class DirChooserLayoutManager
   implements LayoutManager<GuiOutputManager>
   {
      /**
       * Calculate minimum size required for widgets stored in passed container.
       *
       * @param   container
       *          Container to process.
       *
       * @return  calculated space.
       */
      @Override
      public Dimension minimumSize(Container<GuiOutputManager> container)
      {
         return cc.dimensionFromNative(new NativeDimension(NATIVE_WIDTH - 5, NATIVE_HEIGHT - 5));
      }
      
      /**
       * Place components in specified container.
       *
       * @param   container
       *          Container to process.
       */
      @Override
      public void doLayout(Container<GuiOutputManager> container)
      {
         NativeDimension dim = container.physicalDimension();
         
         titleLabel.setPhysicalLocation(15, 11);
         titleLabel.setSize(cc.dimensionFromNative(new NativeDimension(dim.width - 2 * 15, 17)));
         treeScroll.setPhysicalLocation(15, 57);
         treeScroll.setSize(cc.dimensionFromNative(
               new NativeDimension(dim.width - 2 * 15, dim.height - 141)));
         
         nameLabel.setPhysicalLocation(15, dim.height - 71);
         nameLabel.setSize(cc.dimensionFromNative(new NativeDimension(53, 23)));
         dirName.setPhysicalLocation(67, dim.height - 71);
         dirName.setSize(cc.dimensionFromNative(new NativeDimension(dim.width - 83, 23)));
         
         newFolder.setPhysicalLocation(15, dim.height - 35);
         accept.setPhysicalLocation(dim.width - 172, dim.height - 35);
         reject.setPhysicalLocation(dim.width -  91, dim.height - 35);
      }
   }
   
   /**
    * A dedicated container for holding the shortcuts in a vertical flow.
    */
   private class FolderTree
   extends SensitiveScrollContainer
   {
      /** X inset for the tree. */
      static final int X_OFFSET = 7;
      
      /** The root of the file tree. This is the virtual Desktop node. */
      private Folder root = null;
      
      /** Indexes the shown Folder. Used for key navigation. */
      private Folder[] shown = null;
      
      /**
       * Sets the root folder.
       * 
       * @param   root
       *          The new root folder.
       */
      public void setRoot(Folder root)
      {
         this.root = root;
      }
      
      /**
       * Draw widget (white background).
       */
      @Override
      public void draw()
      {
         if (!isVisible())
         {
            return;
         }
         
         NativePoint p = physicalLocation();
         NativeDimension dim = physicalDimension();
         NativeRectangle clip = clipRectangle(p, dim);
         gd.draw(getId(), p, clip, () ->
         {
            gd.setColor(new ColorRgb(0xFFFFFF)); // white
            gd.fillRect(0, 0, dim.width, dim.height);
            
            // make sure the draw origin is properly translated
            super.draw();
         });
      }
      
      /**
       * Lay out container (in a vertical flow).
       */
      @Override
      public void doLayout()
      {
         // invalidate all:
         Widget[] allWidgets = widgets();
         for (Widget w : allWidgets)
         {
            w.setVisible(false);
         }
         shown = new Folder[allWidgets.length]; // all elements are [null] 
         
         int[] yOffset = {0};
         doLayout(root, yOffset);
      }
      
      /**
       * Recursive worker that do the layout of the widgets in a tree fashion.
       * 
       * @param   node
       *          The current node to be processed.
       * @param   yOffset
       *          The vertical position of the node. To avoid using a class field, the procedure
       *          is made re-entrant by using a single element array (to pass value by reference). 
       */
      private void doLayout(Folder node, int[] yOffset)
      {
         ButtonConfig bcfg = node.config();
         if (bcfg.widthPixels == -1)
         {
            bcfg.widthPixels = node.computePreferredWidth();
         }
         
         int xOffset = X_OFFSET + ((node.indent == 0) ? 0 : Folder.X_INDENT * (node.indent - 1));
         node.setVisible(true);
         node.setPhysicalLocation(xOffset, yOffset[0] * FOLDER_HEIGHT);
         assert (shown.length > yOffset[0]);
         shown[yOffset[0]] = node;
         ++yOffset[0];
         
         if (node.isExpanded() && node.content != null && node.content.length != 0)
         {
            for (Folder folder : node.content)
            {
               doLayout(folder, yOffset);
            }
         }
      }
      
      /**
       * Find the folder above {@code dir}, in tree structure.
       *
       * @param   dir
       *          A reference folder.
       *
       * @return  The folder above {@code dir}, or {@code null} if there is none.
       */
      Folder getPreviousFolder(Folder dir)
      {
         if (dir.isRoot())
         {
            // the root is the first, cannot go UP
            return null; 
         }
         
         NativePoint p = dir.physicalLocation();
         int idx = p.y / FOLDER_HEIGHT;
         assert (shown[idx] == dir && idx > 0);
         
         return shown[idx - 1];
      }
      
      /**
       * Find the next folder under {@code dir}, in tree structure.
       *  
       * @param   dir
       *          A reference folder.
       *
       * @return  The folder under {@code dir}, or {@code null} if there is none.
       */
      Folder getNextFolder(Folder dir)
      {
         NativePoint p = dir.physicalLocation();
         int idx = p.y / FOLDER_HEIGHT;
         assert (idx < shown.length && shown[idx] == dir);
         
         if (idx + 1 == shown.length || shown[idx + 1] == null)
         {
            // this is the last folder, cannot go down
            return null;
         }
         
         return shown[idx + 1];
      }
   }
   
   /**
    * A thumbnail for a directory.
    */
   private class Folder 
   extends ButtonGuiImpl
   {
      /** Indentation step (in pixels). */
      static final int X_INDENT = 10;
      
      private final Folder parent;
      
      /**
       * The list of sub-folders or {@code null} if it was never expanded.
       */
      private Folder[] content = null;
      
      /**
       * The absolute path of this folder.
       */
      private final String path;
      
      /**
       * The image to be displayed in the lef-side of the button.
       */
      private ImageWrapper<BufferedImage> img = null;
      
      /**
       * Indentation of this folder.
       */
      private final int indent;
      
      /** {@code true} only if this folder is expanded, showing its children. */
      private boolean expanded = false;
      
      /**
       * The constructor creates a button with given properties.
       *
       * @param   id
       *          The widget's ID.
       * @param   parent
       *          The parent folder. {@code null} if node is the root.
       * @param   text
       *          The string to be used as the button text.
       * @param   path
       *          The absolute path of the folder.
       * @param   icon
       *          The image to be displayed.
       * @param   indent
       *          The indentation of image and text.
       */
      public Folder(WidgetId id, Folder parent, String text, String path, String icon, int indent)
      {
         super(id, null/* ignored */);
         setTextLabel(text);
         this.img = FileChooserUtils.loadImage(icon);
         this.path = path;
         this.indent = indent;
         this.parent = parent;
      }
      
      /**
       * Draw the button.
       */
      @Override
      public void draw()
      {
         if (!isVisible())
         {
            return;
         }
         
         // get location and size
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         
         // set up clipping rectangle
         NativeRectangle clip = clipRectangle(p, d);
         
         // clippable drawing start 
         gd.draw(this.getId(), p, clip, () -> {
            int x = isRoot() ? 0 : 10;
            int w = isRoot() ? d.width : d.width - X_INDENT;
            
            Theme theme = ThemeManager.getCurrentTheme();
            theme.drawFileChooserThumbnailBackground(
                  gd, x, 0, w, d.height, this == crtFolder, isMouseOver(), hasFocus());
            
            if (!isRoot() && (content == null || content.length != 0))
            {
               FileChooserUtils.drawImage(gd, expanded ? expImg : colImg, 0, 6);
            }
            
            FileChooserUtils.drawImage(gd, img, x, 0);
            
            gd.setColor(theme.getFileChooserForegroundColor(
                  gd, hasFocus(), this == crtFolder, isMouseOver(), true, false, false));
            
            FontDetails<?> fd = gf.font().font;
            gd.setGuiFont(fd);
            gd.setFontStyle(fd.style);
            int tX = x + (img == null ? 0 : img.getWidth()) + 3;
            gd.drawStringCentered(textLabel(), tX, d.height / 2);
         });
      }
      
      /**
       * Checks whether this is the root of the tree. 
       * 
       * @return  {@code true} if this is the root of the tree.
       */
      boolean isRoot()
      {
         return parent == null; // also indent == 0
      }
      
      /**
       * Checks whether this node is expanded. The root node is always expanded. 
       *
       * @return  {@code true} if this node is expanded.
       */
      boolean isExpanded()
      {
         return isRoot() || expanded;
      }
      
      /**
       * Computes the preferred width for this thumbnail.
       *
       * @return the preferred width for this thumbnail.
       */
      int computePreferredWidth()
      {
         int tW = getTextWidthNative(textLabel(), gf.resolveFontNum());
         int ret = 2 + (img != null ? img.getWidth() + 2 : 0) + tW + 4;
         if (!isRoot()) 
         {
            ret += X_INDENT;
         }
         return ret;
      }
      
      /**
       * Searches for a child among all direct children.
       * 
       * @param   child
       *          The {@code Folder} to search.
       *
       * @return  The requested {@code Folder} or {@code null} if not found.
       */
      Folder lookup(String child)
      {
         if (content == null)
         {
            return null;
         }
         
         for (Folder folder : content)
         {
            if (Objects.equals(folder.textLabel(), child))
            {
               return folder;
            }
         }
         
         // not found?
         return null;
      }
      
      /**
       * Process basic events - key navigation event.
       *
       * @param event Key event to process.
       */
      @Override
      protected void handleBasicEvents(KeyInput event)
      {
         if (this == crtFolder)
         {
            Folder newSelFolder = null;
            switch (event.keyCode())
            {
               case Key.VK_ENTER:
                  onAction(new ActionEvent(accept, accept.getCommand()));
                  break;
               case Key.VK_CURSOR_RIGHT:
                  // expand this node
                  onAction(new ActionEvent(this, getCommand()));
                  break;
               case Key.VK_CURSOR_LEFT:
                  if (this.expanded && (content == null || content.length != 0))
                  {
                     // collapse this node if expanded
                     this.expanded = false;
                     dirTree.doLayout();
                     treeScroll.repaint();
                  }
                  else if (!isRoot())
                  {
                     // if already collapsed move to parent
                     newSelFolder = parent;
                  }
                  break;
               case Key.VK_CURSOR_UP:
                  // move focus to above Folder
                  newSelFolder = dirTree.getPreviousFolder(this);
                  break;
               case Key.VK_CURSOR_DOWN:
                  // move focus to Folder below
                  newSelFolder = dirTree.getNextFolder(this);
                  break;
            }
            
            if (newSelFolder != null)
            {
               crtFolder = newSelFolder;
               dirName.setValue(new character(crtFolder.textLabel()));
               crtFolder.requestFocus();
               dirTree.ensureVisibility(crtFolder, true);
            }
         }
         
         super.handleBasicEvents(event);
      }
   }
}
