/*
** Module   : SimpleImageButton.java
** Abstract : a button that renders using only images. 
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.  
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import javax.imageio.*;
import java.awt.image.*;
import java.io.*;

/**
 * This button rewrites the rendering routine and instead of using multiple primitives to draw
 * it, a pre-rendered image is used for each of its four states: disabled/pressed/hovered and 
 * normal (enabled). This is useful when the images to be displayed on the buttons are too complex
 * for a standard drawing process. 
 */
public class SimpleImageButton
extends ButtonGuiImpl
{
   /**
    * A flag to mark whether the images were loaded (at least attempted). It prevents multiple
    * attempts to load non-existing resource.
    */
   private boolean imgLoaded = false;
   
   /**
    * The pre-rendered image of the button in normal, enabled state.
    */
   private ImageWrapper<BufferedImage> imgEnabled = null;
   
   /**
    * The pre-rendered image of the button in disabled state.
    */
   private ImageWrapper<BufferedImage> imgDisabled = null;
   
   /**
    * The pre-rendered image of the button when mouse hovers it.
    */
   private ImageWrapper<BufferedImage> imgHighlighted = null;
   
   /**
    * The pre-rendered image of the button when pressed down.
    */
   private ImageWrapper<BufferedImage> imgPressed = null;
   
   /**
    * This constructor creates a button with a given text label.
    *
    * @param id   The widget's ID.
    * @param text The string to be used as the button text.
    */
   public SimpleImageButton(WidgetId id, String text)
   {
      super(null/* ignored */, null/* ignored */);
      initButtonText(text);
   }
   
   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public void draw()
   {
      if (!imgLoaded)
      {
         String command = getCommand().toLowerCase();
         imgEnabled = FileChooserUtils.loadImage(command + "-" + "en");
         imgDisabled = FileChooserUtils.loadImage(command + "-" + "dis");
         imgHighlighted = FileChooserUtils.loadImage(command + "-" + "hi");
         imgPressed = FileChooserUtils.loadImage(command + "-" + "down");
         imgLoaded = true;
      }
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);
      
      // clippable drawing start 
      gd.draw(this.getId(), p, clip, () -> {
         if (!isEnabled())
         {
            drawImage(d, imgDisabled);
         }
         else if (isPressed())
         {
            drawImage(d, imgPressed);
         }
         else if (isMouseOver())
         {
            drawImage(d, imgHighlighted);
         }
         else
         {
            drawImage(d, imgEnabled);
         }
      });
   }
   
   /**
    * Draws the image on a button. The image is centered both horizontally and vertically.
    *
    * @param d   The button dimension (in pixels).
    * @param img The image to be drawn.
    */
   private void drawImage(NativeDimension d, ImageWrapper<BufferedImage> img)
   {
      if (img == null)
      {
         return;
      }
      gd.drawImage(img, (d.width - img.getWidth()) / 2, (d.height - img.getHeight()) / 2,
                   img.getWidth(), img.getHeight(), 0, 0, false, false, true, false);
   }
}
