/*
** Module   : SmallIconsThumbView.java
** Abstract : This view shows a grid of small-sized thumbnails.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20171026 Initial commit.
** 002 OM  20180321 Fixed frameId of thumbs. Added Theme and Edit support.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.filechooser;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import java.io.*;
import java.nio.file.*;
import java.nio.file.attribute.*;
import java.util.*;
import java.util.List;

/**
 * The thumbnails are displayed in a grid, with no details. They are displayed on columns  
 * of fixed width.  
 */
public class SmallIconsThumbView
extends EmptyThumbView
{
   /** The constant width of the thumbnails. */
   private final int ICON_WIDTH = 230;

   /** The constant height of the thumbnails. */
   private final int ICON_HEIGHT = 20;
   
   /** The horizontal gap between two thumbnails. */
   private final int H_GAP = 1;
   
   /** The vertical gap between two thumbnails. */
   private final int V_GAP = 1;
   
   /**
    * Query the preferred scroll types. Each type of view uses a different way of scrolling.
    *
    * @return the preferred scroll type for this thumbnail view.  
    */
   @Override
   public ScrollPaneGuiImpl.ShowBars getScrollType()
   {
      return ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS;
   }
   
   /**
    * Query the preferred scroll bars. Each type of view uses a different way of scrolling.
    *
    * @return the preferred scroll type for this thumbnail view.  
    */
   @Override
   public ScrollBar.Bars getScrollBars()
   {
      return ScrollBar.Bars.VERTICAL;
   }
   
   /**
    * Factory method for creating file thumbnails depending on the current view.
    *
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          The widget's configuration.
    * @param   path
    *          The path this thumbnail represents.
    * @param   isDir
    *          Flags the thumbnails that represents directories.
    * @param   click2Command
    *          The action to be performed when the thumbnail is double clicked.
    */
   @Override
   FileThumb createThumb(WidgetId id,
                         ButtonConfig cfg,
                         Path path,
                         boolean isDir,
                         String click2Command)
   {
      cfg.widthPixels = ICON_WIDTH;
      cfg.heightPixels = ICON_HEIGHT;
      return new SmallIcon(id, cfg, path, isDir, click2Command);
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      NativeDimension dim = physicalDimension();
      int xOffset = 17;
      int yOffset = 6;
      
      List<Widget> col = new ArrayList<>(13);
      Widget[] widgets = widgets();
      for (Widget widget : widgets)
      {
         if (widget instanceof SmallIcon)
         {
            SmallIcon mi = (SmallIcon) widget;
            mi.setPhysicalLocation(xOffset, yOffset);
            xOffset += ICON_WIDTH + H_GAP;
            if (xOffset + ICON_WIDTH > dim.width)
            {
               xOffset = 17;
               yOffset += ICON_HEIGHT + V_GAP;
            }
         }
      }
   }
   
   /**
    * Obtain the name of this view.
    *
    * @return  the name of this view.
    */
   @Override
   public String getViewName()
   {
      return "Small icons";
   }
   
   /**
    * This thumbnail that shows a file in Small Icons view. This view is similar to 
    * {@link ListThumbView}, except that the thumbnails are distributed horizontally, on rows.
    * When the number of rows exceeds the available space, the user is required to scroll the view
    * vertically. 
    */
   private class SmallIcon
   extends FileThumb
   {
      /**
       * Constructor used to create component using given configuration.
       *
       * @param   id
       *          The widget's ID.
       * @param   cfg
       *          The widget's configuration.
       * @param   path
       *          The path this thumbnail represents.
       * @param   isDir
       *          Flags the thumbnails that represents directories.
       * @param   click2Command
       *          The action to be performed when the thumbnail is double clicked.
       */
      SmallIcon(WidgetId id, ButtonConfig cfg, Path path, boolean isDir, String click2Command)
      {
         super(id, cfg, path, isDir ? "small-folder" : "small-file", isDir, click2Command);
      }
      
      /**
       * Draws the text and icon on this thumbnail.
       *
       * @param   width
       *          The width of the widget
       * @param   height
       *          The height of the widget.
       * @param   theme
       *          The current theme.
       */
      @Override
      protected void drawForeground(int width, int height, Theme theme)
      {
         int tX = 2;
         if (img != null)
         {
            FileChooserUtils.drawImage(gd, img, 2, (height - img.getHeight()) / 2);
            tX += img.getWidth() + 2;
         }
         
         FontDetails<?> fd = gf.font().font;
         gd.setGuiFont(fd);
         gd.setFontStyle(fd.style);
         String txt = textLabel();
         
         gd.setColor(theme.getFileChooserForegroundColor(
               gd, hasFocus(), isSelected(), isMouseOver(), true, false, false));
         gd.drawStringCentered(txt, tX, height / 2);
      }
      
      /**
       * Query the preferred width of this thumbnail. It is used by some thumbnail containers in
       * order to do the internal layout. If width is not important, the thumbnail can return -1.
       *
       * @return  the preferred width of this thumbnail.
       */
      @Override
      public int getPreferredWidth()
      {
         return ICON_WIDTH;
      }
      
      /**
       * Obtain the tooltip to be displayed when hovering this thumbnail.
       *
       * @return  the tooltip to be displayed when hovering this thumbnail. 
       */
      @Override
      String getTooltip()
      {
         Path path = getPath();
         String fileName = FileChooserUtils.getFileName(path);
         if (FileChooserUtils.isVirtualFolder(fileName))
         {
            return null;
         }
         
         StringBuilder sb = new StringBuilder(fileName);
         if (isDir())
         {
            sb.append("\nDate created: ").append(FileChooserUtils.getDateAsString(created));
         }
         sb.append("\nSize: ").append(fileSize).append(" bytes");
         if (!isDir())
         {
            sb.append("\nDate modified: ").append(FileChooserUtils.getDateAsString(lastModified));
         }
         
         return sb.toString();
      }
      
      /**
       * Obtain the dimension of the editor that when renaming this thumbnail.  
       *
       * @return  the dimension of the editor that when renaming this thumbnail.
       */
      @Override
      protected NativeDimension getEditorDimension()
      {
         NativeDimension d = physicalDimension();
         int tX = 2 + (img != null ? img.getWidth() + 2 : 0);
         return new NativeDimension(d.width - tX, d.height);
      }
      
      /**
       * Obtain the physical location of the editor that when renaming this thumbnail. The
       * location is relative to this widget.
       *
       * @return  the physical location of the editor that when renaming this thumbnail.
       */
      @Override
      protected NativePoint getEditorLocation()
      {
         int tX = 2 + (img != null ? img.getWidth() + 2 : 0);
         return new NativePoint(tX, 0);
      }
   }
}
