/*
** Module   : BorderLayout.java
** Abstract : Simple border layout manager. 
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 HC  20141021 Implemented window move/position support.
** 006 HC  20150831 Code cleanup.
** 007 HC  20150917 Do not attempt to layout when multiple inner widgets.
** 008 HC  20151030 Fixed insets translation in ChUI mode.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Simple implementation of border layout.
 */
public class BorderLayout<O extends OutputManager<?>>
implements LayoutManager<O>
{
   /**
    * Calculate minimum space required for contained widgets.
    * 
    * @param   container
    *          Container to process.
    *          
    * @return  calculated dimension.
    */
   @Override
   public Dimension minimumSize(Container<O> container)
   {
      double width = 0;
      double height = 0;

      Widget<O> innerPane = container.get(0);
      
      if (innerPane != null)
      {
         Dimension inner = innerPane.minimumSize();
         height = inner.height;
         width = inner.width;
      }
      
      Insets insets = container.insets();
      
      height = Coordinate.scale(height + insets.heightDelta());
      width = Coordinate.scale(width + insets.widthDelta());

      return new Dimension(width, height);
   }

   /**
    * Process container and place components at required locations.
    * 
    * @param   container
    *          Container to process.
    */
   @Override
   public void doLayout(Container<O> container)
   {
      // do not layout if multiple inner widgets,
      // the result would be unexpected
      Widget<O>[] widgets = container.widgets();
      if (widgets == null || widgets.length > 1)
      {
         return;
      }
      
      Dimension size = container.dimension();
      Insets insets = container.insets();

      Widget<O> innerPane = container.get(0);

      if (innerPane == null)
         return;

      fillSpace(container, innerPane);

      double availWidth = size.width - insets.widthDelta();
      double availHeight = size.height - insets.heightDelta();
      
      // BorderedPanelGuiImpl already takes care of translating the insets for its children,
      // so for the children of BorderedPanelGuiImpl the origin is relative to the parent's
      // origin plus the insets
      Point loc = container instanceof BorderedPanelGuiImpl ? new Point(0, 0) : 
                                                              new Point(insets.left, insets.top);
      
      // non-container widgets are centered when they are smaller than avail space
      if (availWidth > innerPane.dimension().width)
      {
         loc.x = loc.x + (availWidth - innerPane.dimension().width) / 2;
      }
      
      if (availHeight > innerPane.dimension().height)
      {
         loc.y = loc.y + (availHeight - innerPane.dimension().height) / 2;
      }
      
      innerPane.setLocation(loc);
   }

   /**
    * Helper method for {@link #doLayout(Container)}.
    * 
    * @param   container 
    *          Main widget container.
    * @param   innerPane
    *          Inner pane of the container.
    */
   private void fillSpace(Container<O> container, Widget<O> innerPane)
   {
      Dimension size = container.dimension();
      
      double availableHeight = size.height - container.insets().heightDelta();
      double availableWidth = size.width - container.insets().widthDelta();

      if (innerPane instanceof Container)
      {
         ((Container<O>) innerPane).setSize(availableWidth, availableHeight);
         ((Container<O>) innerPane).doLayout();
      }
   }
}