/*
** Module   : DynamicLayoutConfig.java
** Abstract : Dynamic layout configuration.
**
** Copyright (c) 2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 HC  20180813 Initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Dynamic layout configuration options read from directory.
 */
public class DynamicLayoutConfig
{
   /** The default frame padding */
   public static final int    FRAME_PADDING_PIXELS = 2;

   /** The default window padding */
   public static final int    WINDOW_PADDING_PIXELS = 0;

   /** The default minimum column width ratio for browse widget */
   public static final double MIN_COLUMN_WIDTH_RATIO = 0.5;

   /** The root configuration node name */
   private static final String ROOT_CFG_NODE = "dynamic-layout";

   /** Context class instance */
   private static ContextLocal<DynamicLayoutConfig> instance = new ContextLocal<DynamicLayoutConfig>()
   {
      /**
       * Initializes the instance, the first time it is requested within a new context.
       *
       * @return   The newly instantiated config class.
       */
      protected synchronized DynamicLayoutConfig initialValue()
      {
         return new DynamicLayoutConfig();
      }
   };

   /* The dynamic layout enabled flag read from directory */
   private boolean dynamicLayoutEnabled;

   /* Frame padding on the X axis */
   private double framePaddingX;

   /* Frame padding on the Y axis */
   private double framePaddingY;

   /* Window padding on the X axis */
   private double windowPaddingX;

   /* Window padding on the Y axis */
   private double windowPaddingY;

   /* Minimum column value will be calculated as the column's data width times this value */
   private double minColumnWidthRatio;

  /**
   * When the flag enabled dynamic layout will determine the optimal size of the widgets in a window
   * and will increase the window size accordingly.
   */
   private boolean optimalSizeEnabled;

   /**
    * Returns the class context instance.
    *
    * @return  see above
    */
   public static DynamicLayoutConfig getInstance()
   {
      return instance.get();
   }

   /**
    * Constructor.
    */
   public DynamicLayoutConfig()
   {
      Directory dir = DirectoryManager.getInstance();

      dynamicLayoutEnabled = dir.getBoolean(Directory.ID_RELATIVE,
                                            ROOT_CFG_NODE + "/enabled",
                                            false);

      if (dynamicLayoutEnabled)
      {
         int framePaddingPixels = dir.getInt(Directory.ID_RELATIVE,
                                             ROOT_CFG_NODE + "/frame-padding-pixels",
                                             FRAME_PADDING_PIXELS);

         int windowPaddingPixels = dir.getInt(Directory.ID_RELATIVE,
                                              ROOT_CFG_NODE + "/window-padding-pixels",
                                              WINDOW_PADDING_PIXELS);

         OutputManager om = ThinClient.getInstance().getOutputManager();
         CoordinatesConversion cc = om.coordinates();

         framePaddingX = cc.widthFromNative(framePaddingPixels);
         framePaddingY = cc.heightFromNative(framePaddingPixels);
         windowPaddingX = cc.widthFromNative(windowPaddingPixels);
         windowPaddingY = cc.heightFromNative(windowPaddingPixels);

         minColumnWidthRatio = dir.getDouble(Directory.ID_RELATIVE,
                                             ROOT_CFG_NODE + "/min-column-width-ratio",
                                             MIN_COLUMN_WIDTH_RATIO);

         optimalSizeEnabled = dir.getBoolean(Directory.ID_RELATIVE,
                                             ROOT_CFG_NODE + "/optimal-size",
                                             false);
      }
   }

   /**
    * Returns the dynamic layout enabled flag read from directory.
    *
    * @return  see above
    */
   public boolean isDynamicLayoutEnabled()
   {
      return dynamicLayoutEnabled;
   }

   /**
    * Returns the frame padding on the X axis read from directory.
    *
    * @return  see above
    */
   public double getFramePaddingX()
   {
      return framePaddingX;
   }

   /**
    * Returns the frame padding on the Y axis read from directory.
    *
    * @return  see above
    */
   public double getFramePaddingY()
   {
      return framePaddingY;
   }

   /**
    * Returns the window padding on the X axis read from directory.
    *
    * @return  see above
    */
   public double getWindowPaddingX()
   {
      return windowPaddingX;
   }

   /**
    * Returns the window padding on the Y axis read from directory.
    *
    * @return  see above
    */
   public double getWindowPaddingY()
   {
      return windowPaddingY;
   }

   /**
    * Returns the minimum column value will be calculated as the column's data width times this value read
    * from directory.
    *
    * @return see above
    */
   public double getMinColumnWidthRatio()
   {
      return minColumnWidthRatio;
   }

   /**
    * Returns the optimal size enabled flag read from directory.
    *
    * @return  see above
    */
   public boolean isOptimalSizeEnabled()
   {
      return optimalSizeEnabled;
   }
}
