/*
** Module   : StackLayout.java
** Abstract : Stack layout manager. 
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150831 Created initial version
** 002 EVL 20151030 Fix for max size calculation in horizontal direction. 
** 003 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.layout;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * An implementation of stack layout. Widgets are stacked on top of each 
 * other in a vertical or horizontal fashion. Widgets are resized to
 * fill up the whole space given by the target container. When widgets
 * are stacked horizontally, the vertical space is filled up. When widgets
 * are stacked vertically, the horizontal space is filled up.
 */
public class StackLayout<O extends OutputManager<?>>
implements LayoutManager<O> 
{
   /**
    * Stacking direction.
    */
   public enum Direction
   {
      HORIZONTAL,
      VERTICAL
   }
   
   /** Stacking direction. */
   private Direction direction;
   
   /**
    * Creates new instance.
    * 
    * @param   direction
    *          The desired stacking direction.
    */
   public StackLayout(Direction direction)
   {
      super();
      this.direction = direction;
   }

   /**
    * Calculate minimum space required for contained widgets.
    * 
    * @param   container
    *          Container to process.
    *          
    * @return  calculated dimension.
    */
   @Override
   public Dimension minimumSize(Container<O> container)
   {
      Dimension result = new Dimension(0, 0);
      
      if (direction == Direction.VERTICAL)
      {
         Insets insets = container.insets(); 
         result.width  = container.width() + insets.left + insets.right;
         result.height = insets.top + insets.bottom; 
      }
      else
      {
         Insets insets = container.insets();
         result.height = container.height() + insets.top + insets.bottom;
         result.width  = insets.left + insets.right; 
      }
      
      // sum all valid minimum sizes in the desired direction of all the container widgets
      for (Widget<O> w : container.widgets())
      {
         Dimension size = w.minimumSize();
         
         if (direction == Direction.VERTICAL)
         {
            if (size.height != BaseConfig.INV_COORD)
            {
               result.height += size.height;
            }
         }
         else
         {
            if (size.width != BaseConfig.INV_COORD)
            {
               result.width += size.width;
            }
         }
      }
      
      return result;
   }

   /**
    * Process container and place components at required locations.
    * 
    * @param   container
    *          Container to process.
    */
   @Override
   public void doLayout(Container<O> container)
   {
      Insets insets = container.insets(); 

      // determine max size in the desired direction  
      double maxSize = 0;
      
      if (direction == Direction.VERTICAL)
      {
         maxSize = container.width() - insets.left - insets.right;
      }
      else
      {
         maxSize = container.height() - insets.top - insets.bottom;      
      }
      
      Point origin = new Point(insets.left, insets.top);
      
      // position the widgets in the container order
      for (Widget<O> w : container.widgets())
      {
         if (!w.isVisible())
         {
            continue;
         }
         
         w.setLocation(origin);
         
         if (direction == Direction.VERTICAL)
         {
            // only containers can be assigned size externally
            if (w instanceof Container)
            {
               ((Container<O>) w).setSize(maxSize, ((Container<O>) w).height());
            }
            
            origin.y += w.height();
         }
         else
         {
            // only containers can be assigned size externally
            if (w instanceof Container)
            {
               ((Container<O>) w).setSize(((Container<O>) w).width(), maxSize);
            }
            
            origin.x += w.width();
         }
      }
   }
}