/*
** Module   : Viewport.java
** Abstract : View port for widget. It allows to see only a part of the 
**            widget visible through "window" - view port.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20101122 setExtent() not is public.
** 004 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 HC  20150405 Refactoring of config coordinate fields.
** 009 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 010 HC  20150831 Scrolling support overhaul.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import com.goldencode.p2j.ui.client.*;

/**
 * Viewport container.
 */
public class Viewport<O extends OutputManager<?>>
extends AbstractContainer<O>
{
   /** Widget which is visible through this viewport. */
   private ScrollableWidget<O> scrollWidget;
   
   /**
    * Constructor.
    * 
    * @param   view
    *          Widget which will be visible through viewport.          
    */
   public Viewport(ScrollableWidget<O> view)
   {
      this.scrollWidget = view;
      add(view);
   }
   
   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      Point origin = screenLocation();
      Rectangle rect = new Rectangle(origin, dimension(), screen().coordinates().baseUnits());

      ScreenBitmap bitmap = getAndSetBitmap(rect);
      scrollWidget.draw();
      setBitmap(bitmap);
   }

   /**
    * Get underlying widget.
    * 
    * @return  underlying widget.
    */
   public ScrollableWidget<O> getScrollWidget()
   {
      return scrollWidget;
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      super.setSize(width, height);
      
      if (!(scrollWidget instanceof AbstractContainer))
      {
         return;
      }
      
      Dimension dim = dimension();
      Dimension scrollDim = scrollWidget.dimension();
      
      boolean dirty = false;
      
      // the scroll widget must be at least as large as its viewport
      if (dim.width > scrollDim.width)
      {
         scrollDim.width = dim.width;
         dirty = true;
      }
      
      if (dim.height > scrollDim.height)
      {
         scrollDim.height = dim.height;
         dirty = true;
      }
      
      if (dirty)
      {
         ((AbstractContainer<?>) scrollWidget).setSize(scrollDim);
      }
   }

   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return scrollWidget.minimumSize();
   }
}