/*
** Module   : WidgetBrowser.java
** Abstract : Implements Widget Browser UI.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141221 New features added: property table now contains all widget and its associated
**                  config fields, added filter textbox, if property value is a weak reference it
**                  is dereferenced, property value string formatter recognizes more well-known 
**                  types.
** 003 HC  20150216 Widget properties can be copied to clipboard.
** 004 HC  20150323 Removed business logic from config classes.
** 005 CA  20151006 Include the 4GL window title in the browser.
** 006 HC  20160210 Minor layout improvements.
** 007 HC  20160511 Do not let Widget Browser to steal focus from P2J windows. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.event.*;
import javax.swing.tree.TreePath;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.widget.Container;

/**
 * Implements Widget Browser UI.
 */
public class WidgetBrowser
{
   /**
    * The main window.
    */
   private JFrame windowFrame;
   
   /**
    * The tree component for displaying widget hierarchy.
    */
   private ContainerTree tree;

   /**
    * The constructor.
    * 
    * @param   container
    *          The root container, usually a window reference.
    *          
    * @param   actionListener
    *          An optional events listener, see {@link Event} for available events.
    */
   public WidgetBrowser(final Container<?> container, final ActionListener actionListener) 
   {
      // build main window title
      String windowTitle = "<" + container.getId() + ">" + 
                           (container.name() != null ? container.name() : "");
      if (container.config() instanceof BaseConfig)
      {
         String title = ((BaseConfig) container.config()).title;
         if (title != null)
         {
            windowTitle += title;
         }
      }
      windowTitle = String.format("Widget Browser %s", windowTitle);
      
      // main window
      windowFrame = new JFrame(windowTitle);
      windowFrame.setSize(1280, 800);
      windowFrame.setLocation(100, 100);
      windowFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
      windowFrame.addWindowListener(new WindowAdapter()
      {
         @Override
         public void windowClosed(WindowEvent e)
         {
            if (actionListener != null)
            {
               actionListener.actionPerformed(new Event(this, Event.BROWSER_CLOSED, container));
            }
         }
      });

      // add content panel
      JPanel contentPanel = new JPanel();
      contentPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
      contentPanel.setLayout(new BorderLayout(0, 0));
      windowFrame.setContentPane(contentPanel);

      // property table
      PropertyTable table = new PropertyTable();
      final ContainerTree tree = new ContainerTree(container, table);
      tree.setActionListener(actionListener);

      // table event handler
      table.setActionListener(new ActionListener() 
      {
         @Override
         public void actionPerformed(ActionEvent paramActionEvent) 
         {
            if (paramActionEvent.getID() != Event.PROPERTY_VALUE_NAVIGATED)
            {
               return;
            }
            
            Object propValue = ((Event)paramActionEvent).propertyValue;
            if (propValue instanceof Widget<?>)
            {
               TreePath result = tree.findNode((Widget<?>) propValue);
               if (result != null) {
                  tree.setSelectionPath(result);
               }
            }
         }
      });
      
      this.tree = tree;

      // add top panel for buttons and controls
      JPanel masterPanelTop = new JPanel();
      masterPanelTop.setLayout(new BorderLayout());
      contentPanel.add(masterPanelTop, BorderLayout.NORTH);
      
      JPanel panelTop = new JPanel();
      panelTop.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
      masterPanelTop.add(panelTop, BorderLayout.WEST);

      // refresh button
      JButton refreshButton = new JButton("Refresh");
      panelTop.add(refreshButton, BorderLayout.WEST);
      refreshButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.refresh();
         }
      });

      // collapse all button
      JButton collapseAllButton = new JButton("Collapse All");
      panelTop.add(collapseAllButton, BorderLayout.WEST);
      collapseAllButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.collapseAll();
         }
      });

      // expand all button
      JButton expandAllButton = new JButton("Expand All");
      panelTop.add(expandAllButton, BorderLayout.WEST);
      expandAllButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.expandAll();
         }
      });

      // copy to clipboard button
      JButton clipboardButton = new JButton("To Clipboard");
      clipboardButton.setToolTipText("Refreshes and copies the properties of the selected "
                                   + "widget recursively to system clipboard.");
      panelTop.add(clipboardButton, BorderLayout.WEST);
      clipboardButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.refresh();
            tree.copyPropertiesToClipboard();
         }
      });
      
      // full class names check-box
      final JCheckBox showFullClassName = new JCheckBox("Full Class Names");
      showFullClassName.setSelected(tree.isShowFullClassName());
      showFullClassName.addItemListener(new ItemListener() 
      {
         public void itemStateChanged(ItemEvent e) 
         {
            tree.setShowFullClassName(showFullClassName.isSelected());
            tree.refresh();
         }
      });
      panelTop.add(showFullClassName, BorderLayout.WEST);
      
      // field name filter
      final JTextField filter = new JTextField();
      filter.setToolTipText("Case insensitive white space delimited list of inclusions. " +
                            "Example: 'Hello ello Hell ell'.");
      filter.getDocument().addDocumentListener(new DocumentListener()
      {
         @Override
         public void insertUpdate(DocumentEvent e)
         {
            refreshProperties();
         }

         @Override
         public void removeUpdate(DocumentEvent e)
         {
            refreshProperties();
         }

         @Override
         public void changedUpdate(DocumentEvent e)
         {
            refreshProperties();
         }
         
         private void refreshProperties()
         {
            // parse the inclusion string
            String inclString = filter.getText().trim();
            tree.setPropertyInclusions(Arrays.asList(inclString.split("\\s")));
            tree.refreshPropertyTable();
         }
      });
      masterPanelTop.add(filter, BorderLayout.CENTER);
      
      // add splitter
      JSplitPane splitPane = new JSplitPane();
      splitPane.setDividerLocation(550);
      contentPanel.add(splitPane, BorderLayout.CENTER);
      splitPane.setLeftComponent(new JScrollPane(tree));
      splitPane.setRightComponent(new JScrollPane(table));
   }
   
   /**
    * Shows the main window.
    */
   public void show()
   {
      // do not steal focus from P2J windows
      windowFrame.setAutoRequestFocus(false);
      windowFrame.setVisible(true);
   }
   
   /**
    * Returns <code>true</code> is the supplied widget is selected to be highlighted.
    * 
    * @param   widget
    *          A {@link Widget} reference.
    * 
    * @return  <code>true</code> if the supplied widget is selected to be highlighted,
    *          <code>false</code> otherwise.
    */
   public boolean isHighlighted(Widget<?> widget)
   {
      return widget != null && tree.isSelected(widget);
   }
}