/*
** Module   : AbstractSimpleParameter.java
** Abstract : Common logic for assigns [input-]output parameter values (database fields and normal
**            variables) upon exit from procedure/function
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 OM  20150626 First version based on the old FieldAssigner code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;

/**
 * Delegate worker for the variables or data-type wrapper implementations, which assigns an OUTPUT
 * or INPUT-OUTPUT parameter's value back to the variable / database field with which it is
 * associated, upon exit from a procedure or function.
 * <p>
 * Upon exit from the procedure or function with which it is associated, it attempts to assign the
 * current value of its contained variable back into the associated item (variable / DMO)
 * optionally using the conversion constructor provided. This may trigger normal validation
 * processing, which may produce error conditions.
 */
public abstract class AbstractSimpleParameter
extends AbstractParameter
{
   /** 
    * Data wrapper which is used as a procedure or function parameter.
    * <p>
    * The variable is used as a key for fast-finding the original item that needs to be updated
    * at the return form the called procedure. In this regard it must be unique in the scope!
    * Indeed, it's at least a strange construct to have the same variable declared twice in
    * OUTPUT mode when calling a procedure.
    */
   protected BaseDataType outputValue;
   
   /**
    * The optional conversion constructor that will cast the new variable to old variable type in
    * order to be assigned to field.
    */
   protected Constructor<?> convCtor = null;
   
   /**
    * Constructor. Register this as output parameter. 
    *
    * @param   outputValue
    *          Parameter value to and from which we will assign.
    */
   protected AbstractSimpleParameter(BaseDataType outputValue)
   {
      this.outputValue = outputValue;
      getCurrentScope().addSimpleParameter(this);
   }
   
   /**
    * Get the variable container used as parameter inside the procedure body.
    * 
    * @return  The data wrapper which is used as a procedure or function parameter.
    */
   protected BaseDataType getLocalParam()
   {
      return this.outputValue;
   }
   
   /**
    * Assign the current value of the parameter variable back to the associated item (variable or
    * DMO property. This is expected to occur upon exit from the function or procedure with which
    * this object is associated. Must be implemented in subclasses.
    */
   abstract protected void assign();
   
   /**
    * Update the wrapper variable for a {@code OutputParameterAssigner}. Sometimes is required to
    * change the wrapper variable for an item (Eg. when the {@link ControlFlowOps} detects a
    * conversion type is necessary). 
    *
    * @param   key
    *          The old key variable. May be the intermediary wrapper for a field or the real
    *          variable passed as *OUTPUT parameter.
    * @param   newVar
    *          The new wrapper variable that will be used inside the calling procedure. At the end
    *          of the call its value will be converted using {@code convCtor} and assigned to
    *          wrapped field or original variable.
    * @param   convCtor
    *          The conversion constructor that will cast the new variable to old variable type in
    *          order to be assigned to field. 
    */
   public static void update(BaseDataType key, BaseDataType newVar, Constructor<?> convCtor)
   {
      if (key == newVar) // comparing 'pointers', not their value
      {
         return; // same var, NO-OP
      }
      
      WorkArea local = context.get();
      Scope scope = local.getCurrentScope();
      scope.updateSimpleParameter(key, newVar, convCtor);
   }
}