/*
** Module   : BaseDataType.java
** Abstract : Progress 4GL compatible operator logic
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050518   @21222 Created initial version with full support for all Progress logical
**                           operators and common functions.
** 002 GES 20050606   @21423 Minor additions for function support.
** 003 ECF 20050629   @21601 Implement Serializable. Will need this to push wrapper objects
**                           over the wire.
** 004 SVG 20050720   @21762 Cloning support added.
** 005 GES 20050815   @22117 Added Undoable interface support.
** 006 GES 20050825   @22252 Added getTypeName().
** 007 ECF 20051005   @22966 Introduced an abstract hashCode() method. This forces subclasses
**                           to provide an implementation which is consistent with this class'
**                           implementation of the equals() method.
** 008 GES 20060208   @24385 Added defaultFormatString().
** 009 GES 20060301   @24840 Added generateUnknown() static factory method.
** 010 NVS 20060407   @25407 Added isUnknownValue() and notUnknownValue() methods.
** 011 GES 20060417   @25551 Added default value generation.
** 012 GES 20060724   @28156 Added a forced assign() abstract method.
** 013 GES 20060807   @28469 Moved to externalizable interface to optimize network performance.
** 014 GES 20061109   @31049 Added formatLength().
** 015 ECF 20070720   @34626 Added safety check to equals(). Check type of object before casting
**                           it to BaseDataType to prevent ClassCastException.
** 016 SIY 20100812          Removed unnecessary cast.
** 017 CA  20130119          Added castLogical and castBoolean, which cast the current value
**                           using DYNAMIC-FUNCTION's rules.
** 018 GES 20130128          Clarified some javadoc.
** 019 GES 20130312          Removed castLogical() and castBoolean(). This same functionality is
**                           implemented in the logical(BDT) c'tor.
** 020 EVK 20131130          Added elementsOfType static method.
** 021 MAG 20140602          Added incompatibleTypesOnConversion() method.
** 022 OM  20150723          Updated javadoc.
** 023 CA  20151214          Added APIs to implicitly initialize an extent var and to check if two
**                           objects are of the same type.
** 024 GES 20160617          Added assign(BDT) for converting polymorphic return values.
** 025 CA  20160627          Reworked undoable support - the undoables register themselves with
**                           all blocks up the stack, until either 1. the tx block which created
**                           it is reached or 2. the last tx block where it was saved is reached.
** 026 GES 20171207          Removed explicit bypass on pending error (new silent error mode).
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.reflect.*;

/**
 * An abstract class that implements a Progress 4GL compatible data type
 * which has common logical operator processing.  Meant to be the parent
 * class for specific Progress compatible data type wrappers.
 * <p>
 * Each subclass should implement <code>Comparable</code> but is only
 * required to implement <code>compareTo()</code> as the <code>equals()</code>
 * method is implemented in this class.
 * <p>
 * <strong>Note:</strong> As result, classes from this hierarchy cannot be used as keys for normal
 * java associative collections because of collisions caused by objects (Ex: an unknown 
 * {@link integer} and an unknown {@link logical} are considered equals in P4GL so only one
 * unknown can be the key in such collection). If such collection is required, use a special one
 * like {@link java.util.IdentityHashMap}.
 * <p>
 * All processing relies upon the implementation of the following methods
 * in each concrete subclass:
 * <p>
 * <pre>
 * boolean      isUnknown()
 * void         setUnknown(boolean)
 * int          compareTo(Object) - this comes from the Comparable interface
 * BaseDataType instantiateUnknown()
 * BaseDataType duplicate()
 * void         assign(Undoable) - this comes from the Undoable interface
 * void         assign(BaseDataType)
 * String       toString()
 * String       toStringMessage()
 * String       toStringExport()
 * String       defaultFormatString()
 * </pre>
 *
 * @author    GES
 */
public abstract class BaseDataType
extends LazyUndoable
implements Comparable,
           Externalizable
{
   /**
    * Default constructor (only used for de-serialization).
    */
   public BaseDataType()
   {
   }
   
   /**
    * Creates a new instance of the given type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public static BaseDataType generateUnknown(Class cls)   
   {
      BaseDataType bdt = null;
      
      try
      {
         // use the default constructor
         bdt = (BaseDataType) cls.newInstance();
      }
      
      catch (Exception expt)
      {
         // what else to do here (the calling code will probably fail with
         // a class cast exception)?
         return new unknown();
      }
      
      return bdt.instantiateUnknown();
   }
   
   /**
    * Creates a new instance of the given type that represents the 
    * default initialized value for that type.
    *
    * @return   An instance that represents the default value.
    */
   public static BaseDataType generateDefault(Class cls)   
   {
      BaseDataType bdt = null;
      
      try
      {
         // use the default constructor
         bdt = (BaseDataType) cls.newInstance();
      }
      
      catch (Exception expt)
      {
         throw new RuntimeException(expt);
      }
      
      return bdt.instantiateDefault();
   }
   
   /**
    * Calculate the length of a formatted value (the text form of the value)
    * using this format string.
    *
    * @param    fmt
    *           The format string.
    *
    * @return   The length in characters of the any resulting formatted
    *           value.
    */
   public static int calcFormatLength(String fmt, Class cls)
   {
      BaseDataType bdt = null;
      
      try
      {
         // use the default constructor
         bdt = (BaseDataType) cls.newInstance();
      }
      
      catch (Exception expt)
      {
         throw new RuntimeException(expt);
      }
      
      return bdt.formatLength(fmt);
   }

   /**
    * Check list of BaseType of arguments and return {@code true} if all BaseTypes are specific
    * type.
    *
    * @param    cls
    *           Specific BaseDataType class witch will be checked to the arguments.
    * @param    args
    *           list of BaseType instances.
    *
    * @return   The {@code true} if all BaseTypes are character.
    */
   public static boolean elementsOfType(Class<? extends BaseDataType> cls, BaseDataType... args)
   {
      if (args.length == 0)
      {
         return true;
      }

      for (BaseDataType arg : args)
      {
         if (!cls.isInstance(arg))
         {
            return false;
         }
      }
      return true;
   }
   
   /**
    * Check if the types of the specified objects match.
    * <p>
    * In case of extent vars, the size of each var must be the same, and both must be extent.
    * 
    * @param    source
    *           The reference variable.
    * @param    target
    *           The candidate variable, which must match.
    *           
    * @return   <code>true</code> if the type of these two variables matches.
    */
   protected static boolean sameType(Object source, Object target)
   {
      // extent vars
      if (target.getClass().isArray())
      {
         if (!source.getClass().isArray())
         {
            // shared is not an array
            return false;
         }
         
         Class<?> tcls = target.getClass().getComponentType();
         Class<?> scls = source.getClass().getComponentType();
         
         // check component type
         if (!tcls.equals(scls))
         {
            return false;
         }
         else
         {
            // check extent
            int slen = Array.getLength(source);
            int tlen = Array.getLength(target);
            if (slen != tlen)
            {
               return false;
            }
         }
         
         return true;
      }
      else
      {
         // check type directly (this will also fail if source is array and target
         // is not array)
         return source.getClass().equals(target.getClass());
      }
   }

   /**
    * Initialize the given array with the default value, based on the following table
    * (the rules are from variable_definition.rules):
    * 
    * <pre>
    *  integer                new integer(0)
    *  int64                  new int64(0)
    *  decimal                new decimal(0)
    *  character              new character(&quot;&quot;)
    *  logical                new logical(false)
    *  longchar               new longchar(&quot;&quot;)
    *  date                   new date()
    *  datetime               new datetime()
    *  datetime-tz            new datetimetz()
    *  raw                    new raw()
    *  memptr                 new memptr()
    *  recid                  new recid()
    * </pre>
    * 
    * @param    var
    *           The extent variable which needs to be initialized.
    */
   protected static void initializeDefaultExtent(BaseDataType[] var)
   {
      if (var == null || var.length == 0)
      {
         return;
      }
   
      Class<? extends BaseDataType> vcls = 
         (Class<? extends BaseDataType>) var.getClass().getComponentType();
      
      try
      {
         BaseDataType bdinit = vcls.newInstance().instantiateDefaultExtent();
         ArrayAssigner.assignMulti(var, bdinit);
      }
      catch (SecurityException        |
             InstantiationException   |
             IllegalAccessException   |
             IllegalArgumentException e)
      {
         throw new RuntimeException(e);
      }
   }

   /**
    * Calculate the length of a formatted value (the text form of the value)
    * using this format string.
    *
    * @param    fmt
    *           The format string.
    *
    * @return   The length in characters of the any resulting formatted
    *           value.
    */
   public int formatLength(String fmt)
   {
      if (fmt == null)
      {
         fmt = defaultFormatString();
      }
      
      return fmt.length();
   }
   
   /**
    * Determines if this instance and the given instance are equivalent. Only an object of type
    * {@code BaseDataType} is compared with this instance; objects of other types cause this test
    * to return {@code false}.
    * <p>
    * <strong>Note:</strong> Unknown values form different classes from this hierarchy are 
    * considered equals in P4GL, even if they have different types (Ex: an unknown {@link integer}
    * and an unknown {@link logical}.
    * <strong>Note:</strong> Classes from this hierarchy cannot be used as keys for normal
    * java associative collections because of collisions caused by objects (Ex: only one
    * unknown can be the key in such collection). If such collection is required, use a special
    * one like {@link java.util.IdentityHashMap}.
    *
    * @param   obj
    *          The instance to compare against.
    *
    * @return  {@code true} if the objects compare equivalently; {@code false} if they do not,
    *          or if the parameter is not a {@code BaseDataType} instance.
    */
   public boolean equals(Object obj)   
   {
      if (!(obj instanceof BaseDataType))
      {
         return false;
      }
      
      return CompareOps.equals(this, (BaseDataType) obj);
   }
   
   /**
    * The default implementation of this method provided by
    * <code>Object</code> is overridden here with an abstract method to force
    * subclasses to provide a concrete implementation which is consistent with
    * this class' implementation of {@link #equals}.
    *
    * @return  The hash code for this object instance.
    */
   public abstract int hashCode();
   
   /**
    * Returns the largest instance.
    *
    * @param    b
    *           The instance to compare against.
    *
    * @return   The largest instance.
    */
   public BaseDataType maximum(BaseDataType b)   
   {
      if (isUnknown() || b.isUnknown())
         return instantiateUnknown();
      
      return (CompareOps.greaterThan(this, b)) ? this : b;
   }
      
   /**
    * Returns the smallest instance.
    *
    * @param    b
    *           The instance to compare against.
    *
    * @return   The smallest instance.
    */
   public BaseDataType minimum(BaseDataType b)   
   {
      if (isUnknown() || b.isUnknown())
         return instantiateUnknown();
      
      return (CompareOps.lessThan(this, b)) ? this : b;
   }
   
   /**
    * This is a form of a copy constructor that makes a deep copy of the
    * current instance and returns this copy as a new instance of the same
    * class.
    *
    * @return   An instance of the same class whose data is identical to
    *           that of this instance.
    */
   public Undoable deepCopy()   
   {
      return (Undoable) duplicate();
   }
   
   /**
    * Returns the legacy type name of the sub-class as the type was known in
    * the 4GL.  Generally, this is equivalent to an uppercased version of the
    * base name of the class file (all package information is removed). Child
    * classes can override this default behavior.
    * 
    * @return   The legacy type name.
    */
   public String getTypeName()
   {
      String cls = this.getClass().getName();
      
      int idx = cls.lastIndexOf('.') + 1;
      
      return cls.substring(idx).toUpperCase();
   }

   /**
    * Reports if this instance represents the <code>unknown value</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public logical isUnknownValue()
   {
      return new logical(isUnknown());
   }
   
   /**
    * Reports if this instance represents some known value.
    *
    * @return   <code>true</code> if this instance is not set to the
    *           <code>unknown value</code>.
    */
   public logical notUnknownValue()
   {
      return new logical(!isUnknown());
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()   
   {
      BaseDataType bdt = null;
      
      try
      {
         // use the default constructor
         bdt = this.getClass().newInstance();
      }
      
      catch (Exception expt)
      {
         throw new RuntimeException(expt);
      }
      
      return bdt;
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the state of the passed
    * instance.
    * <p>
    * If the value is not of the same type, some form of automatic type conversion may occur, or
    * an error will be raised.
    * <p>
    * This variant is meant to handle the cases of built-in functions and methods in the 4GL
    * which have polymorphic return types (e.g. DYNAMIC-FUNCTION()). This should NOT be used for
    * non-polymorphic assignments.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public abstract void assign(BaseDataType value);
   
   /**
    * Reports if this instance represents the <code>unknown value</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public abstract boolean isUnknown();
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag or
    * state to <code>true</code>.
    * <p>
    * <b>Warning: the data stored in this instance may be invalid after
    * calling this method.</b>
    */
   public abstract void setUnknown();
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.
    */
   public abstract BaseDataType duplicate();

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public abstract BaseDataType instantiateUnknown();   
   
   /**
    * Creates a string representation of the instance data using the given
    * format string.  If the instance represents the
    *  <code>unknown value</code>, a '?' will be returned.
    *
    * @param    fmt
    *           The format string to use.
    *
    * @return   The formatted string.
    */
   public abstract String toString(String fmt);   
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public abstract String toStringMessage();   
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return   The 'export' formatted string.
    */
   public abstract String toStringExport();      
   
   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public abstract String defaultFormatString();
   
   /**
    * Get the default initialization for an extent variable of this type.
    * 
    * @return   See above.
    */
   protected BaseDataType instantiateDefaultExtent()
   {
      return instantiateDefault();
   }
   
   /**
    * Display error message in case of incompatible types for a POLY conversion.
    * (e.g. DYNAMIC-FUNCTION()).
    */
   protected void incompatibleTypesOnConversion()
   {
      String err = "Incompatible datatypes found during runtime conversion";
      ErrorManager.recordOrThrowError(5729, err);      
   }
}