/*
** Module   : BlockRunner.java
** Abstract : Helper to execute a 4GL code block with managed errors.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20171206 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.function.*;
import com.goldencode.p2j.net.*;

/**
 * Helper to execute a 4GL code block with managed errors. Some of the data is intentionally
 * public so that it is easily accessed after block execution.
 */
public class BlockRunner
{
   /** Return value from the code block execution. */
   public BaseDataType result = null;
   
   /** <code>true</code> if an error condition is raised in the code execution. */
   public boolean error = false;
   
   /** <code>true</code> if a stop condition is raised in the code execution. */
   public boolean stop = false;
   
   /** <code>true</code> if a quit condition is raised in the code execution. */
   public boolean quit = false;
   
   /** Error number of the error condition that was raised in the code execution. */
   public int errorNumber = 0;
   
   /** Error message of the error condition that was raised in the code execution. */
   public String errorMessage = null;
   
   /** Main code block to execute. */
   private Function<BlockRunner, BaseDataType> block = null;
   
   /** Post-processing block (optional). */
   private Runnable post = null;
   
   /** Flag to determine if the code is run in silent error mode or not. */
   private boolean silent = false;
   
   /**
    * Create an instance to execute the given code block(s) WITHOUT silent mode.
    *
    * @param    block
    *           The main code block to execute. Must NOT be <code>null</code>.
    */
   public BlockRunner(Function<BlockRunner, BaseDataType> block)
   {
      this.block  = block;
   }
   
   /**
    * Create an instance to execute the given code block(s).
    *
    * @param    block
    *           The main code block to execute. Must NOT be <code>null</code>.
    * @param    post
    *           Optional block to execute as post-processing. It will happen in a finally block
    *           if this is non-null.
    * @param    silent
    *           <code>true</code> if this should be executed in silent error mode.
    */
   public BlockRunner(Function<BlockRunner, BaseDataType> block, Runnable post, boolean silent)
   {
      this.block  = block;
      this.post   = post;
      this.silent = silent;
   }
   
   /**
    * Run the code block.
    */
   public void execute()
   {
      if (silent)
      {
         ErrorManager.silent(() -> this.process());
      }
      else
      {
         this.process();
      }
   }  
   
   /**
    * Execute the code block with integrated error management.
    */
   private void process()
   {
      try
      {
         result = block.apply(this);
      }
      catch (Throwable t)
      {
         // check for quit, stop, error
         Throwable err = t;
         
         while (err != null)
         {
            if (err instanceof SilentUnwindException)
            {
               throw (SilentUnwindException) err;
            }
            else if (err instanceof ErrorConditionException)
            {
               error        = true;
               errorNumber  = ((ErrorConditionException) err).getProgressErrorCode();
               errorMessage = err.getMessage();
               break;
            }
            else if (err instanceof QuitConditionException)
            {
               quit = true;
               break;
            }
            else if (err instanceof StopConditionException)
            {
               stop = true;
               break;
            }
            
            err = err.getCause();
         }
         
         if (!(error || quit || stop))
         {
            // assume an ERROR condition for everything else 
            error        = true;
            errorNumber  = -1;
            errorMessage = t.getMessage();
         }
      }
      finally
      {
         if (post != null)
         {
            post.run();
         }
      }
      
      if (silent && error && ErrorManager.isErrorFlag())
      {
         // check the error messages - only first message will be reported
         error = true;
         
         if (ErrorManager.numErrors().intValue() > 0)
         {
            errorNumber  = ErrorManager.getErrorNumber(1).intValue();
            errorMessage = ErrorManager.getErrorText(1).toStringMessage();
         }
      }
   }
}
