/* 
** Module   : ImageExts.java
** Abstract : Enumerates default image files' extensions 
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 IAS 20160112 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.function.*;
import java.util.stream.*;

/** Default image files' extensions */
public enum ImageExt
{
   BMP("bmp", "image/bmp"),
   
   ICO("ico", "image/ico"),
   
   CUR("cur", "application/cur"),
   
   NONE("", "application/octet-stream");

   /** list of extensions with dot */
   public final static List<String> EXT_BASES = Collections.unmodifiableList(
            Arrays.stream(values()).map(e -> e.base).collect(Collectors.toList())
         );
   
   /** list of extension variants with dot */
   public final static List<String> EXT_VARIANTS = Collections.unmodifiableList( 
            Arrays.stream(values()).map(e -> e.variants).
                  flatMap(List::stream).collect(Collectors.toList())
         );
   
   /** mime type */
   public final String mimeType;
   
   /** extension with dot */
   public final String base;
   
   /** extension variants with dot */
   public final List<String> variants;
   

   /** map of ImageExt by  extension*/
   private final static Map<String, ImageExt> EXTS = Collections.unmodifiableMap(
            Arrays.stream(values()).
               collect(Collectors.toMap((ImageExt e) -> e.ext, Function.identity()))
         );

   /** extension */
   private final String ext;
   
   /**
    * Constructor
    * 
    * @param ext
    *        extension
    */
   private ImageExt(String ext, String mimeType)
   {
      this.ext = ext;
      this.mimeType = mimeType;
      this.base = "." + ext.toLowerCase();
      this.variants = generateVariants(ext.toLowerCase());
   }

   /**
    * Retrieve ImageExt by extension string
    *  
    * @param ext
    *        string extension string
    *        
    * @return ImageExt instance or <code>null</code>
    */
   public static ImageExt fromExt(String ext)
   {
      return ext == null ? NONE : EXTS.getOrDefault(ext.toLowerCase(), NONE);
   }

   /**
    * Generate extension variants
    * 
    * @param ext
    *        extension
    *        
    * @return list of extension variants with dot
    *       
    */
   private static List<String> generateVariants(String ext)
   {
      char[] extChars = ext.toCharArray();
      int le = extChars.length;
      int nv = 1 << le;
      List<String> list = new ArrayList<>(nv);
      char[] extVariant = new char[le];
      for (int i = 0; i < nv; i++)
      {
         for (int nc = 0; nc < le; nc++)
         {
            char c = extChars[nc];
            extVariant[nc] = (i & (1 << nc)) == 0 ? c : Character.toUpperCase(c); 
         }
         list.add("." + String.valueOf(extVariant));
      }

      return Collections.unmodifiableList(list);
   }
}