/*
** Module   : InvocationResponsePayload.java
** Abstract : A payload with data resulted from an invoke request into the legacy converted code, 
**            which will be passed back to the customer-specific application which initiated the
**            request.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20160418 Created initial version
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * This class holds the data for a response, after invoking a 4GL external program, procedure or
 * function, and associated with a request incoming from a customer-specific application.
 * <p>
 * The response contains details for the handle ID (in case of an external program ran persistent),
 * values for OUTPUT/INPUTPUT arguments, function (or procedure)'s return value, state of any
 * ERROR, QUIT or STOP conditions.
 */
public class InvocationResponsePayload
implements Externalizable
{
   /** 
    * When running an external program in persistent mode, this is the handle ID associated with
    * the resource of that external program.
    */
   public long handle;

   /** 
    * This map holds the string-representation of all OUTPUT or INPUT-OUTPUT arguments, as resulted
    * after the request was performed.
    */
   public Map<String, String> outputArguments;

   /**
    * In case of functions, this will hold the value returned by the function.  Otherwise, this
    * value is the same value as the RETURN-VALUE 4GL function would return, after the call of
    * an external program or internal procedure.
    */
   public String returnValue;
   
   /** Flag indicating if the request ended with an ERROR condition. */
   public boolean error;
   
   /** If {@link #error} flag is <code>true</code>, this represents the error number. */
   public int errorNumber;
   
   /** If {@link #error} flag is <code>true</code>, this represents the error message. */
   public String errorMessage;
   
   /** Flag indicating if the request ended with a QUIT condition. */
   public boolean quit;
   
   /** Flag indicating if the request ended with a STOP condition. */
   public boolean stop;

   /**
    * Replacement for the default object reading method. The response data is read from the input 
    * source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException, 
          ClassNotFoundException
   {
      handle          = in.readLong();
      returnValue     = (String) in.readObject();
      error           = in.readBoolean();
      errorNumber     = in.readInt();
      errorMessage    = (String) in.readObject();
      quit            = in.readBoolean();
      stop            = in.readBoolean();
      outputArguments = new HashMap<>();
      
      byte argNum = in.readByte();
      for (byte i = 0; i < argNum; i++)
      {
         String argName = (String) in.readObject();
         String argVal = (String) in.readObject();
         
         outputArguments.put(argName, argVal);
      }
   }
   
   /**
    * Replacement for the default object writing method. The response data is written to the output 
    * destination.
    * 
    * @param    out
    *           The output destination to which the fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out) 
   throws IOException
   {
      out.writeLong(handle);
      out.writeObject(returnValue);
      out.writeBoolean(error);
      out.writeInt(errorNumber);
      out.writeObject(errorMessage);
      out.writeBoolean(quit);
      out.writeBoolean(stop);
      
      out.writeByte(outputArguments.size());
      for (String key : outputArguments.keySet())
      {
         out.writeObject(key);
         out.writeObject(outputArguments.get(key));
      }
   }
}