/*
** Module   : OnPhrase.java
** Abstract : represents the condition/action pair that defines how blocks
**            react to conditions
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20080522   @38524 Created initial version which represents the condition/action pair
**                           that defines how blocks react to conditions.
** 002 GES 20080702   @39063 Support cases where 2 different labels are targets (for UNDO and
**                           the action).
** 003 GES 20160518          Code formatting cleanup.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import static com.goldencode.p2j.util.BlockManager.*;

/**
 * Represents the condition/action pair that defines how a given block is
 * configured to react to conditions. Optionally, this class can contain
 * a label that references the block name that is the target of the undo
 * and action.  Optionally, a return value can be stored.  This class is
 * simply a storage mechanism to associate all this data as a single tuple.
 */
public class OnPhrase
{
   /** Condition being raised for a block. */
   public Condition condition = null;
   
   /** Action to be taken when the given condition is raised. */
   public Action action = null;
   
   /** Optional block name of the target for the UNDO. */
   public String undoLabel = null;
   
   /** Optional block name of the target for the ACTION. */
   public String actionLabel = null;
   
   /** Optional return value if the action is one of the RETURN variants. */
   public BaseDataType value = null;
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           LEAVE, NEXT, RETRY, RETURN_NORMAL, RETURN_CONSUME or
    *           RETURN_ERROR.
    */
   public OnPhrase(Condition condition, Action action)
   {
      this.condition = condition;
      this.action    = action;
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           LEAVE, NEXT, RETRY, RETURN_NORMAL, RETURN_CONSUME or
    *           RETURN_ERROR.
    * @param    label
    *           The name of the block that is the target of BOTH the UNDO and
    *           the <code>action</code>.
    */
   public OnPhrase(Condition condition, Action action, String label)
   {
      this(condition, action);
      this.undoLabel   = label;
      this.actionLabel = label;
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           LEAVE, NEXT, RETRY, RETURN_NORMAL, RETURN_CONSUME or
    *           RETURN_ERROR.
    * @param    undoLabel
    *           The name of the block that is the target of the UNDO.
    * @param    actionLabel
    *           The name of the block that is the target of the 
    *           <code>action</code>.
    */
   public OnPhrase(Condition condition,
                   Action    action,
                   String    undoLabel,
                   String    actionLabel)
   {
      this(condition, action);
      this.undoLabel   = undoLabel;
      this.actionLabel = actionLabel;
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, BaseDataType value)
   {
      this(condition, action);
      this.value = value;
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, int value)
   {
      this(condition, action);
      this.value = new integer(value);
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, double value)
   {
      this(condition, action);
      this.value = new decimal(value);
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, boolean value)
   {
      this(condition, action);
      this.value = new logical(value);
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    label
    *           The name of the block that is the target of the UNDO.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, String label, BaseDataType value)
   {
      this(condition, action, label);
      this.value = value;
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    label
    *           The name of the block that is the target of the UNDO.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, String label, int value)
   {
      this(condition, action, label);
      this.value = new integer(value);
   }
   
   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    label
    *           The name of the block that is the target of the UNDO.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, String label, double value)
   {
      this(condition, action, label);
      this.value = new decimal(value);
   }

   /**
    * Construct an association between the given condition and action, which
    * operates on the nearest enclosing block.
    *
    * @param    condition
    *           ERROR, ENDKEY, STOP or QUIT.
    * @param    action
    *           RETURN_NORMAL, RETURN_CONSUME or RETURN_ERROR.
    * @param    label
    *           The name of the block that is the target of the UNDO.
    * @param    value
    *           The data to be returned to the caller of the top-level block
    *           from which the control flow is going to return.
    */
   public OnPhrase(Condition condition, Action action, String label, boolean value)
   {
      this(condition, action, label);
      this.value = new logical(value);
   }
}