/*
** Module   : ParsedDateFormat.java
** Abstract : parsed form of the date format string
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20150115 Shared support for date format parsing and parse results access.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * This centralizes 4GL date format parsing and provides public access to the parse results as
 * simple member fields.
 */
public class ParsedDateFormat
{
   /** <code>true</code> if this format has a separator (or if it has 2 separators). */
   public boolean hasSep = true;
   
   /** Separator character between the first two date components. */
   public char sep1 = '\0';
   
   /** Separator character between the last two date components. */
   public char sep2 = '\0';
   
   /** Split format string MDY components (but not necessarily in MDY order). */
   public String[] pieces = null;
   
   /** Size of each date component. */
   public int[] size = null;
   
   /** Format string with separators converted to the / character. */
   public String normalized = null;
   
   /** Original format string. */
   public String original = null;
   
   /** <code>true</code> if parsing was successful. */
   public boolean valid = false;
   
   /**
    * Parse the given format and use the parse results to create a new instance.
    *
    * @param    fmt      
    *           The 4GL date format string to parse.           
    */
   public ParsedDateFormat(String fmt)
   {
      this.original = fmt;
      
      normalized = fmt.replaceAll("[.-]", "/");
      pieces     = normalized.split("/"); 
      
      // special case of "999999"
      if (normalized.equals("999999"))
      {
         pieces = new String[] { "99", "99", "99" };
         hasSep = false;
      }
      
      // special case of "99999999"
      if (normalized.equals("99999999"))
      {
         pieces = new String[] { "99", "99", "9999" };
         hasSep = false;
      }
      
      // the format string must have 3 parts
      if (pieces.length != 3)
      {
         String spec = "Date format %s is incomplete";
         ErrorManager.recordOrThrowError(154, String.format(spec, fmt));
         return;
      }
      
      int idx = 0;
      
      // more validity testing: each component must be non-empty
      for (int i = 0; i < 3; i++)
      {
         // empty string
         if (pieces[i].length() == 0)
         {
            date.genBadCharInFormatError(idx, fmt);
            return;
         }
         
         // keep track of the index of the next possible empty string
         idx += (pieces[i].length()) + 1;
      }
      
      int x   = 0;
      int len = normalized.length();
      
      // not done with validity test yet: the normalized format string must
      // only have '/' or '9' chars
      while (x < len)
      {
         char c = normalized.charAt(x);
         
         if (c != '/' && c != '9')
         {
            // invalid format string
            date.genBadCharInFormatError(x, fmt);
            return;
         }
         
         x++;
      }
      
      size = new int[3];
      
      // setup field sizes and separators
      for (int j = 0; j < 3; j++)
      {
         size[j] = pieces[j].length();
      }
      
      if (hasSep)
      {
         int idx2  = size[0] + 1 + size[1];
         
         sep1 = fmt.charAt(size[0]);
         sep2 = fmt.charAt(idx2);
      }
      
      valid = true;
   }
}