/*
** Module   : ProcedureManager.java
** Abstract : Implements access to procedure system handles and procedure
**            related attributes and methods. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130130 Added stubbed methods to implement the PUBLISH, SUBSCRIBE and UNSUBSCRIBE 
**                  language statements.
** 003 CA  20130131 Fixed problem determining THIS-PROCEDURE when the trigger invoked is defined 
**                  as ON ... PERSISTENT RUN.
** 004 CA  20130202 Update from character to TextOps.
** 005 CA  20130529 Some refactoring related to procedure deletion and added appserver support.
** 006 CA  20130918 Added runtime support for widget-pools: the pools need scope notifications,
**                  and this must be done by the ProcedureManager, to ensure proper access to the 
**                  THIS-PROCEDURE handle.
** 007 CA  20130925 Added support for named events.
** 008 CA  20131013 Implemented implicit deletion of static resources and delayed processing of
**                  finalizable/scopeable instances. Added support for the INSTANTIATING-PROCEDURE 
**                  attribute.
** 009 CA  20131208 Emit SEARCH-TARGET and SEARCH-SELF literals as integer.  Fixed runtime for 
**                  RUN SUPER and SUPER() calls, in search-target and search-self modes.
** 010 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
**                  INSTANTIATING-PROCEDURE is now managed by ProcedureManager.
** 011 EVK 20131223 Made method isProcedure public.
** 012 CA  20140109 WorkArea.scopeFinished has no target-procedure to pop, when terminating an 
**                  appserver request.
** 013 VIG 20140109 Implemented CURRENT-WINDOW attribute.
** 014 CA  20140218 Implemented PROGRAM-NAME. Fixed some SOURCE- and TARGET-PROCEDURE cases (the
**                  RUN PERSISTENT or RUN ext-prog cases).  Also, the external program name used
**                  at the RUN statement must be used in error messages; the absolute name must be
**                  used each time SourceNameMapper APIs are invoked.  Now we allow java-style
**                  calls for both internal and external procedures.
** 015 CA  20140221 Fixed SUPER-PROCEDURES - emit correct handle ID.
** 016 CA  20140223 Fixed deletion of persistent procedures via DELETE OBJECT <system-proc-handle>.
** 017 CA  20140301 Procedure's scope is notified first for start and last for finish.
** 018 HC  20140424 Fixed resolveClosestMethod failing to resolve extent functions.
** 019 CA  20140502 Fixed super-procedure chaining in SUPER and RUN SUPER calls.
** 020 SVL 20140414 Modified INSTANTIATING-PROCEDURE assignment logic in order to support agent
**                  contexts.
** 021 CA  20140513 Added a weight for the context-local var, to ensure predetermined order during
**                  context reset.
** 022 HC  20140629 Implemented access to the internal callee information structure.
** 023 CA  20140729 Made getRelativeName public.
** 024 VIG 20140810 Added invalid param processing for setCurrentWindow method.
** 025 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 026 EVL 20141024 The dynamic flag need to be set early enough for dynamic widgets to avoid NPE
**                  in processResource() method.
** 027 OM  20150112 Made _thisProcedure() public.
** 028 CA  20150306 When an external, non-persistent procedure, gets terminated, execute the same
**                  work as when deleting a persistent procedure.
** 029 CA  20150501 DELETE OBJECT THIS-PROCEDURE forces the delete of the current procedure,
**                  regardless of where is called (in an external proc block, inner proc, function
**                  or trigger).
**                  Track the resources created by each procedure, as dynamic resources don't get  
**                  deleted when the creator proc gets deleted, but their INSTANTIATING-PROCEDURE 
**                  attr needs to be unknown.
** 030 ECF 20150810 Removed redundant Throwable.fillInStackTrace call from resolveClosestMethod
**                  as a performance improvement.
** 031 ECF 20150907 Made resolveClosestMethod public. Re-organized methods by access modifier.
** 032 CA  20151027 When running an internal entry (procedure, function) of a persistent procedure,
**                  all its shared data (variables, tables, etc) needs to be pushed and popped, 
**                  so that the internal entry (and any other external program ran from it) has
**                  access to it.
**                  When a procedure gets deleted (persistent or not), then all created triggers
**                  must be destroyed.
**                  resolveInstantiatingProcedure must not use the handle's c'tor, as this will
**                  not perform the assignment if a NO-ERROR clause is in effect.
** 033 ECF 20160105 Fixed concurrent modification errors in publish method.
**     CA  20151219 Track the name of the external program being instantiated - required for 
**                  ERROR conditions raised during instantiation (i.e. by shared vars).
**     OM  20161026 Added setDefaultWindow() method.
** 034 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 035 OM  20161026 Added cleanupPending() method.
** 036 IAS 20160314 DEBUG-ALERT support.
** 037 CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 038 GES 20160630 Some ControlFlowOps cases call removeProcedureImpl() with a null handle, we
**                  treat that as a no-op.
** 039 HC  20160718 PUBLISH/SUBSCRIBE runtime support modified so that events are fired in the
**                  order of subscriptions. Procedure cleanup will also remove the procedure's
**                  subscriptions.
** 040 CA  20160728 Fixed undoable support for data in external programs ran persistent.
** 041 HC  20160817 Fixed PUBLISH to properly invoke all procedures even in case there is a
**                  failure due to procedure not found in the procedure file.
**                  Fixed PUBLISH to resolve the target internal procedure in the subscribed
**                  procedure as well as in its SUPER procedures or SESSION procedures.
**                  Fixed UNSUBSCRIBE to only unsubscribe the procedure passed in the statement.
** 042 CA  20160812 Performance improvements for H040: use IdentityHashMap when using
**                  ext prog instances as keys.
** 043 CA  20160824 Fixed a regression with delete-on-exit support: the external program must be 
**                  deleted only when the last of its internal entry on the stack is finished. 
** 044 CA  20160902 Fixed H043 - only this-procedure stack is checked for delete-on-exit postpone.
**                  Fixed a possible leak when deleting ext prog at scope finish.  Fixed order of
**                  created static resources in an ext prog.
** 045 CA  20160907 For the duration of a PUBLISH call, the SOURCE-PROCEDURE must be set as the
**                  publisher's procedure handle.
** 046 OM  20160901 Replaced ArrayDeque with custom Deque implementation for persistentProcedures.
**                  Small optimizations.
**     CA  20160923 Reworked H045 to explicitly set the PUBLISH's SOURCE-PROCEURE without 
**                  'hacking' the WorkArea.thisProcedures stack, as this is used in conjunction
**                  with the WorkArea.internalEntryStack, when determining PROGRAM-NAME or 
**                  4GL-style stacktraces.
** 047 HC  20161110 Fixed an abend in publish() when subscriptions are removed as a result of the
**                  published event processing.
**                  Fixed several cases of referencing deleted system handles TARGET-PROCEDURE,
**                  SOURCE-PROCEDURE and THIS-PROCEDURE.
** 048 CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 049 GES 20171206 Switch to ErrorManager.silent().
** 050 CA  20171202 Added searchInStack.
** 051 ECF 20171228 Refactored to remove redundant context-local lookups where possible.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.scheduler.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.UniqueIdGenerator.*;
import com.goldencode.util.*;

import static com.goldencode.p2j.util.ErrorManager.*;

/**
 * Provides implementation for the following procedure handles:
 * <table summary="">
 * <tr>
 *   <td>THIS-PROCEDURE</td>
 *   <td>{@link #thisProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>SOURCE-PROCEDURE</td>
 *   <td>{@link #sourceProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>TARGET-PROCEDURE</td>
 *   <td>{@link #targetProcedure()}</td>
 * </tr>
 * </table>
 * The following procedure-related attributes and methods are implemented:
 * <table summary="">
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>SESSION:SUPER-PROCEDURES</td>
 *    <td>{@link #superProcedures()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:ADD-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #addSuperProcedure(handle)}</li>
 *          <li>{@link #addSuperProcedure(handle, SearchMode)}</li>
 *          <li>{@link #addSuperProcedure(handle, int)}</li>
 *       </ul>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:REMOVE-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #removeSuperProcedure(handle)}</li>
 *       </ul>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:FIRST-PROCEDURE</td>
 *    <td>{@link #firstProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:LAST-PROCEDURE</td>
 *    <td>{@link #lastProcedure()}</td>
 * </tr>
 * </table>
 * <p>
 * The {@link #registerFunctionHandle(String, Resolvable)} is emitted at the
 * beginning of an external procedure, for any function defined 
 * <code>IN handle</code>.
 */
public class ProcedureManager
{
   /** Constant defining search mode: SEARCH-SELF */
   public static final int SEARCH_SELF = 1;
   
   /** Constant for socket search mode: SEARCH_TARGET */
   public static final int SEARCH_TARGET = 2;
   
   /** Marks the external program's root procedure - this is a P2J reserver name. */
   public static final String EXTERNAL_PROGRAM = "__P2J__EXTERNAL__PROGRAM__";
   
   /** Marks the trigger block - this is what 4GL reports for triggers. */
   public static final String TRIGGER_PROCEDURE = "USER-INTERFACE-TRIGGER";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ProcedureManager.class.getName());
   
   /** Worker for managing global subscriptions. */
   private static GlobalEventSubscriptions globalSubscriptions;
   
   /**
    * Defines search modes for the ADD-SUPER-PROCEDURE statement. Needed to
    * be able to identify the correct target in RUN SUPER and SUPER() cases.
    */
   public static enum SearchMode
   {
      SEARCH_SELF
      {
         @Override
         public String toString()
         {
            return "SEARCH-SELF";
         }
      },
      
      SEARCH_TARGET
      {
         @Override
         public String toString()
         {
            return "SEARCH-TARGET-PROCEDURE";
         }
      }
   }

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();
 
   /**
    * API executed on client startup, to initialize this class.
    */
   public static void initialize()
   {
      // WARNING: do not initialize static fields in the default class c'tor, as the 
      // ProcedureManager currently gets resolved on client-side, too, because of a client-side
      // dependency on TM.setNestedMode
      
      // initialize the static fields.
      globalSubscriptions = new GlobalEventSubscriptions();
   }

   /**
    * Register the handle expression for the given function.
    * 
    * @param    pname
    *           The function's legacy 4GL name.
    * @param    handleExpr
    *           The {@link Resolvable} instance used to determine its IN 
    *           handle target.
    */
   public static void registerFunctionHandle(String pname, Resolvable handleExpr)
   {
      // always work with thisProcedure
      handle h = thisProcedure();
      Object referent = h.get();
      ProcedureData pp = locateProcedure(referent);
      pp.inHandleFunctions.put(pname, handleExpr);
   }
   
   /**
    * Get the THIS-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle thisProcedure()
   {
      return thisProcedure(work.get());
   }
   
   /**
    * Get the TARGET-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle targetProcedure()
   {
      return targetProcedure(work.get());
   }
   
   /**
    * Get the SOURCE-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle sourceProcedure()
   {
      WorkArea wa = work.get();
      ProcedureData pd = wa.sourceProceduresData.peek();
      
      // Do not use locateProcedure() here, the system handle may have been deleted
      // with DELETE PROCEDURE. Note that when a procedure is deleted it can be
      // referenced and queried through the system procedure handles.
      return pd == null ? new handle() : new handle(pd.phandle);
   }
   
   /**
    * Get the SESSION:LAST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle lastProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }
      
      Object referent = wa.persistentProcedures.peekLast();
      return new handle(locateProcedure(wa, referent).phandle);
   }
   
   /**
    * Get the SESSION:FIRST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle firstProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }
      
      Object referent = wa.persistentProcedures.peekFirst();
      
      return new handle(locateProcedure(wa, referent).phandle);
   }
   
   /**
    * Get the SESSION:SUPER-PROCEDURES attribute.
    * 
    * @return   See above.
    */
   public static character superProcedures()
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSessionSuperProcedures().values());
   }
   
   /**
    * Add the given handle as a super-procedure for the SESSION handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h)
   {
      return addSuperProcedure(h, (SearchMode) null);
   }
   
   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, int searchMode)
   {
      return addSuperProcedure(h, getSearchMode(searchMode));
   }
   
   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, integer searchMode)
   {
      if (searchMode.isUnknown())
      {
         return addSuperProcedure(h, (SearchMode) null);
      }
      else
      {
         return addSuperProcedure(h, searchMode.intValue());
      }
   }
   
   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, null);
   }
   
   /**
    * Remove the given handle from the SESSION's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical removeSuperProcedure(handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      removeProcedureImpl(wa, h, wa.getSessionSuperProcedures().keySet());
      
      return new logical(true);
   }
   
   /**
    * Process the given resource by either setting its INSTANTIATING-PROCEDURE attribute 
    * immediately (in case of non-static resources) or add the static resource to the 
    * {@link WorkArea#pendingResources} set, to be processed when the next scope is entered.
    * 
    * @param    res
    *           The resource to be processed.
    * @param    dynamic
    *           Flag indicating if this is a dynamic or static resource.
    */
   public static void processResource(WrappedResource res, boolean dynamic)
   {
      // be a no-op in case of headless mode
      if (ErrorManager.isHeadless())
      {
         return;
      }
      
      WorkArea wa = work.get();
      
      boolean pending = !(res instanceof StaticTempTable);
      
      if (res instanceof DynamicResource && !dynamic && pending)
      {
         // all static resources are created as instance fields; thus, the actual processing is 
         // delayed, as the THIS-PROCEDURE is not available at the time of the call (i.e. we are
         // in an instance field initialization code, at the business logic class)
         wa.pendingResources.add(res);
      }
      else if (res instanceof HandleResource)
      {
         wa.pendingResources.remove(res);
         
         // all resources which don't have dynamic counterparts are processed immediately
         Object referent = _thisProcedure(wa);
         ProcedureData pdata = locateProcedure(wa, referent);
         
         // make sure the resource is registered as static
         if (!dynamic && !pending)
         {
            pdata.staticResources.peek().add(res);
         }
         
         // need to init dynamic flag here to avoid NPE
         if (dynamic && res instanceof GenericWidget)
         {
            ((GenericWidget)res)._setDynamic(dynamic);
         }
         
         ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
      }
   }
   
   /**
    * The {@code ProcedureManager} should know the DEFAULT-WINDOW in order to prevent releasing /
    * deleting it when the top-level scope is popped out.
    * <p>
    * The DEFAULT-WINDOW was eventually added in {@link #processResource(WrappedResource, 
    * boolean)} but, at that time, we were not able to distinct it. The {@link LogicalTerminal}
    * must call this as soon as the internal variable has been set.
    *
    * @param   defaultWindow
    *          The default window for this session.
    */
   public static void deregisterPendingResource(WindowWidget defaultWindow)
   {
      // a no-op in case of headless mode
      if (ErrorManager.isHeadless())
      {
         return;
      }
      
      WorkArea wa = work.get();
      wa.pendingResources.remove(defaultWindow);
   }
   
   /**
    * Cleanup of pending resources. This is called when the external procedure constructor failed 
    * to build a new object because of an exception thrown in a member initialization (shared 
    * frame/table).
    */
   public static void cleanupPending()
   {
      // a no-op in case of headless mode
      if (ErrorManager.isHeadless())
      {
         return;
      }
      
      WorkArea wa = work.get();
      wa.pendingResources.clear();
   }
   
   /**
    * Get the current external program (an application logic instance) being processed for 
    * scopeable or finalizable notifications.
    * 
    * @return   See above
    */
   public static Object getProcessedProcedure()
   {
      WorkArea wa = work.get();
      
      return wa.processedProcedure;
   }
   
   /**
    * Save the information about the program being called, to be used on the top-level scope 
    * start notification: at that time, the info about the callee (this, source- and target-
    * procedures and other info) will be resolved.
    * 
    * @param    superCall
    *           Flag indicating if the destination is in a super-procedure.
    * @param    target
    *           The target procedure, which was targeted by the caller.
    * @param    destination
    *           The procedure where the internal entry was resolved.
    * @param    iename
    *           The internal entry name. In case of triggers or external programs, use the 
    *           {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
    * @param    pname
    *           The external program name; will be non-null only when we are invoking an external
    *           program.
    * @param    function
    *           Flag indicating we are calling a user-defined function.
    * @param    persistentProc
    *           Flag indicating the external program was ran persistent.
    */
   public static void pushCalleeInfo(boolean superCall,
                                     Object  target,
                                     Object  destination,
                                     String  iename,
                                     String  pname,
                                     boolean function,
                                     boolean persistentProc)
   {
      CalleeInfoImpl info = new CalleeInfoImpl(superCall, target, destination, iename, pname,
                                               function, persistentProc);
      
      work.get().calleeInfo.push(info);
   }
   
   /**
    * Pop the last callee info from the stack.
    */
   public static void popCalleeInfo()
   {
      work.get().calleeInfo.pop();
   }
   
   /**
    * Return the last callee info from the stack.
    * 
    * @return  {@link CalleeInfo} instance.
    */
   public static CalleeInfo peekCalleeInfo()
   {
      return work.get().calleeInfo.peek();
   }
   
   /**
    * Get the THIS-PROCEDURE referent.
    * 
    * @return   See {@link #thisProcedure()}.
    */
   public static Object _thisProcedure()
   {
      return _thisProcedure(work.get());
   }
   
   /**
    * Get the SOURCE-PROCEDURE referent.
    * 
    * @return   See {@link #sourceProcedure()}.
    */
   public static Object _sourceProcedure()
   {
      WorkArea wa = work.get();
      return wa.sourceProcedures.peek();
   }
   
   /**
    * Check if the given referent is for a persistent procedure. This implements the PERSISTENT
    * attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of a converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   public static boolean _isPersistent(Object referent)
   {
      return _isPersistent(work.get(), referent);
   }
   
   /**
    * Register an external program which was ran persistent by a remote side.
    *  
    * @param   referent
    *          The external program instance.
    */
   public static void registerRemotePersistentProcedure(Object referent)
   {
      work.get().remotePersistentProcedures.add(referent);
   }
   
   /**
    * Deregister an external program which was ran persistent by a remote side.
    *  
    * @param   referent
    *          The external program instance.
    */
   public static void deregisterRemotePersistentProcedure(Object referent)
   {
      work.get().remotePersistentProcedures.remove(referent);
   }
   
   /**
    * Check if an external program was ran persistent by a remote side.
    *  
    * @param   referent
    *          The external program instance.
    *          
    * @return  <code>true</code> if the referent is part of the 
    *          {@link WorkArea#remotePersistentProcedures} map.
    */
   public static boolean isRemotePersistentProcedure(Object referent)
   {
      return work.get().remotePersistentProcedures.contains(referent);
   }
   
   /**
    * Check if the handle's referent is a procedure or not. This means that the value must be set 
    * and must be an instance of {@link ExternalProgramWrapper}. The enclosed referent may or may
    * not be registered with the {@link ProcedureManager}.
    *
    * @param    h
    *           handle object to check.
    *  
    * @return   See above.
    */
   public static boolean isProcedure(handle h)
   {
      return h != null && !h.isUnknown() && (h.getResource() instanceof ExternalProgramWrapper);
   }
   
   /**
    * Gets the value of CURRENT-WINDOW attribute of specified procedure
    * @param   referent
    *          Procedure wrapper 
    * @return  handle.
    */
   public static handle getCurrentWindow(Object referent)
   {
      return new handle(locateProcedure(referent).currentWindow);
   }
   
   /**
    * Sets the value of CURRENT-WINDOW attribute of specified procedure
    * @param   referent
    *          Procedure wrapper
    * @param   winHandle 
    *          New value for the attribute.          
    */
   public static void setCurrentWindow(Object referent, handle winHandle)
   {
      if (winHandle == null || !winHandle._isValid() || winHandle.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083, 
            "Unable to assign UNKNOWN value to attribute CURRENT-WINDOW on PROCEDURE widget",
            false, false, false);
         
         return;
      }
      
      if (!(winHandle.getResource() instanceof WindowWidget))
      {
         String ms = "Unable to set attribute CURRENT-WINDOW of procedure widget";
         ErrorManager.recordOrShowError(4078, ms, false, false, false);
         
         return;
      }
      
      locateProcedure(referent).currentWindow.assign(winHandle);
   }
   
   /**
    * Get the relative name used to create the given external program instance.
    * 
    * @param    instance
    *           The external program instance.
    * 
    * @return   See above.
    */
   public static String getRelativeName(Object instance)
   {
      return work.get().relativeNames.get(instance);
   }
   
   /**
    * Resolve the closest business logic method in the stack trace.
    * 
    * @param    referent
    *           The external program to which the method must belong.
    * @param    function
    *           Flag indicating if a function or internal procedure will be searched.
    * @param    externalProc
    *           Flag indicating if we are looking for an external procedure.
    *           
    * @return   The legacy method name.
    */
   public static String resolveClosestMethod(Object referent,
                                             boolean function,
                                             boolean externalProc)
   {
      // get a helper to process the stack
      Throwable stack = new Throwable();
      
      // obtain the stack trace
      StackTraceElement[] trace = stack.getStackTrace();
      
      String rClsName   = referent.getClass().getName();
      String methodName = null;
      String legacyName = null;
      
      String bmClass = BlockManager.class.getName();
      String procApi = (externalProc ? "internalProcedure" : "externalProcedure");
      String funcApi = "function";
      
      // walk the stack trace until we find a method defined by the referent's class
      boolean foundBlockManager = false;
      for (int i = 1; i < trace.length; i++)
      {
         String cname = trace[i - 1].getClassName();
         String mname = trace[i - 1].getMethodName();
         
         if (cname.equals(bmClass) && 
             ((function && mname.toLowerCase().endsWith(funcApi)) || 
              (!function && mname.equals(procApi))))
         {
            foundBlockManager = true;
         }
         else if (foundBlockManager)
         {
            methodName = mname;
            break;
         }
      }
      
      if (methodName != null)
      {
         if ("execute".equals(methodName))
         {
            // this is a special 4GL string.
            legacyName = EXTERNAL_PROGRAM;
         }
         else
         {
            // resolve the legacy name of this method
            legacyName = SourceNameMapper.convertJavaMethod(rClsName, methodName, function);
            
            if (legacyName == null)
            {
               // assume the java name as legacy name
               legacyName = methodName;
            }
         }
      }
      
      if (legacyName == null)
      {
         final String msg = "Could not resolve closest legacy method for class %s";
         throw new RuntimeException(String.format(msg, rClsName) , stack);
      }
      
      return legacyName;
   }
   
   /**
    * Get the name of the currently instantiated external program.
    * 
    * @return   The 4GL external program name being instantiated, or <code>null</code> if this is 
    *           not the case.
    */
   public static String getInstantiatingExternalProgram()
   {
      WorkArea wa = work.get();
      return wa.initiatingExtProg;
   }
   
   /**
    * Get the {@link Scopeable} instance for this context.
    * 
    * @return   See above.
    */
   static Scopeable getScopeable()
   {
      return work.get();
   }
   
   /**
    * Check if we are executing finalizable or scopeable notifications.
    * 
    * @return   See above.
    */
   static boolean inNotifications()
   {
      return work.get().processedProcedure != null;
   }
   
   /**
    * Set the external procedure beeing processed for scopeable or finalizable notifications.
    * 
    * @param    referent
    *           The external procedure instance.
    */
   static void setProcessedProcedure(Object referent)
   {
      WorkArea wa = work.get();
      
      wa.processedProcedure = referent;
   }
   
   /**
    * Notify all subscribers (to the given event) that the event has occurred. This method is
    * executed in a {@link ErrorManager#silent} bracket, as the <code>PUBLISH</code>
    * statement is executed with a default NO-ERROR clause.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    * <p>
    * The subscriptions are fired in the order they were subscribed. Although 4GL
    * documentation specifically states, that "If a named event has multiple subscribers, 
    * the order in which the AVM notifies subscribers is undefined.", the actual
    * 4GL implementation (OpenEdge 10.2B) does fire the events in the order the subscriptions
    * get subscribed.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           Must not be <code>null</code>.
    * @param    modes
    *           The mode for each parameter, encoded in a string. When <code>null</code>, 
    *           the parameter modes will not be checked against the procedure definition.
    * @param    parameters
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   static void publish(character event, handle publisher, String modes, Object[] parameters)
   {
      final Object[] param = (parameters == null ? new Object[0] : parameters);
      
      silent(() ->
      {
         if (!publisher.isExternalResource())
         {
            if (publisher.isUnknown())
            {
               final String msg = "Could not evaluate publisher handle expression while trying " +
                                  "to execute PUBLISH statement";
               ErrorManager.recordOrThrowError(6472, msg, false);
               return;
            }
            if (!isProcedure(publisher) || !publisher._isValid())
            {
               final String msg = 
                  "Invalid or inappropriate publisher handle value given to PUBLISH statement";
               ErrorManager.recordOrThrowError(9072, msg, false);
               return;
            }
         }
         
         // Progress does not validate the event
         
         Object rpub = publisher.get();
         boolean externalPublisher = publisher.isExternalResource();
         
         WorkArea wa = work.get();
         
         List<String> errors = new LinkedList<>();
         
         // determine the subscriptions
         // business logic can unsubscribe during this loop, so we have to use a copy of
         // the subscription list to avoid concurrent modification
         List<Subscription> subCopy = new ArrayList<>(wa.subscriptions);
         subCopy.addAll(globalSubscriptions.getAllSubscriptions());
         
         Session currentSession = SessionManager.get().getSession();
         for (Subscription<?> sub : subCopy)
         {
            // if the events match and the subscription is ANYWHERE or the publishers match,
            // then invoke the procedure
            if (sub.event.equals(event) &&
               (sub.publisher == null || sub.publisher.equals(rpub)))
            {
               if (currentSession == sub.session)
               {
                  sub.publish((externalPublisher ? null : rpub), errors, modes, param);
               }
               else
               {
                  Object[] params = new Object[param.length];
                  // create a copy of the params - as the OUTPUT/INPUT-OUTPUT params must not
                  // affect this session, and following are not supported:
                  // - extent parameters
                  // - field parameters
                  // - memptr
                  // - handle
                  // - buffers
                  for (int i = 0; i < param.length; i++)
                  {
                     if (param[i] instanceof BaseDataType)
                     {
                        params[i] = ((BaseDataType) param[i]).duplicate();
                     }
                     else if (param[i] instanceof AbstractParameter ||
                              param[i] instanceof memptr            ||
                              param[i] instanceof handle            ||
                              param[i] instanceof WrappedResource   ||
                              param[i] instanceof BufferReference)
                     {
                        throw new UnsupportedOperationException("Parameter of type " + 
                                                                param[i].getClass().toString() + 
                                                                " is not supported!");
                     }
                     else if (param[i] != null && param[i].getClass().isArray())
                     {
                        throw new UnsupportedOperationException(
                           "Array parameters are not supported!");
                     }
                     else
                     {
                        params[i] = param[i];
                     }
                  }
                  
                  Runnable task = () -> sub.publishRemote(modes, params);
                  globalSubscriptions.post(new PublishTask(sub, task));
               }
            }
         }
         
         if (!errors.isEmpty())
         {
            int[] nums = new int[errors.size()];
            Arrays.setAll(nums, operand -> 6480);
            ErrorManager.recordOrThrowError(nums, errors.toArray(new String[0]), false);
         }
      });
   }
   
   /**
    * Create a new subscription to the given event for a specified program. This program will 
    * receive a call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribe(handle subscriber, character event, handle publisher, character method)
   {
      subscribe(false, subscriber, event, publisher, method);
   }
   
   /**
    * Create a new subscription to the given event for a specified program. This program will 
    * receive a call-back when the event is published.
    *
    * @param    global
    *           Flag indicating if the subscription is for global (cross-session) events.
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribe(boolean   global,
                         handle    subscriber, 
                         character event, 
                         handle    publisher, 
                         character method)
   {
      // validate subscriber
      if (!subscriber.isExternalResource())
      {
         if (subscriber.isUnknown())
         {
            final String msg = "Could not evaluate subscriber handle  expression while trying to " +
                               "execute SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6463, msg, false);
            return;
         }
         if (!isProcedure(subscriber) || !subscriber._isValid())
         {
            final String msg = "Invalid or inappropriate subscriber handle value given to " +
                               "SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6466, msg, false);
            return;
         }
      }
      
      // validate publisher; if null, is OK, as it means ANYWHERE option
      if (publisher != null && !publisher.isExternalResource())
      {
         if (publisher.isUnknown())
         {
            final String msg = "Could not evaluate publisher handle  expression while trying " +
                               "to execute SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6464, msg, false);
            return;
         }
         if (!isProcedure(publisher) || !publisher._isValid())
         {
            final String msg = "Invalid or inappropriate publisher handle value given to " +
                               "SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6467, msg, false);
            return;
         }
      }
      // Progress does not validate event and method name
      
      Object rsub = subscriber.get();
      Object rpub = (publisher == null ? null : publisher.get());
      WorkArea wa = null;
      
      List<Subscription> subs;
      if (global)
      {
         subs = globalSubscriptions.getSubscriptions();
      }
      else
      {
         if (wa == null)
         {
            wa = work.get();
         }
         subs = wa.subscriptions;
      }
      
      if (subs == null)
      {
         return;
      }
      
      // 4GL doesn't raise an error if a subscription already exists, but doesn't replace it 
      // either
      boolean found = false;
      for (Subscription<?> sub : subs)
      {
         if (sub.readOnly)
         {
            // read-only subscriptions are not checked.
            continue;
         }
         
         // get an exact match
         if (sub.subscriber.equals(rsub) &&
             event.equals(sub.event)     &&
             ((rpub == null && sub.publisher == null) ||
              (rpub != null && sub.publisher != null && rpub.equals(sub.publisher))))
         {
            found = true;
            break;
         }
      }
      
      // create only if there isn't an existing one;
      // the subscription is added into a "global" list of subscriptions, the list
      // also holds the order when the subscriptions got subscribed so that they
      // can be fired in this order
      if (!found)
      {
         Subscription<?> sub = (rsub instanceof ExternalResource)
            ? new ExternalSubscription((ExternalResource) rsub, event, rpub)
            : new LegacySubscription(rsub, event, rpub, method);
         
         if (global)
         {
            globalSubscriptions.subscribe(sub);
         }
         else
         {
            if (wa == null)
            {
               wa = work.get();
            }
            wa.subscriptions.add(sub);
         }
      }
   }
   
   /**
    * Create a new subscription to the given event for any program. This program will receive a 
    * call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribeAnywhere(handle subscriber, character event, character method)
   {
      subscribe(subscriber, event, null, method);
   }
   
   /**
    * Create a new subscription to the given event for a specified external resource. 
    * This resource will receive a notification when the event is published.
    *
    * @param    global
    *           Flag indicating if the subscription is for global (cross-session) events.
    * @param    subscriber
    *           The resource to notify when the specified event is published. If <code>null</code>,
    *           a {@link ExternalResource} will be built and its ID return.
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The handle representing the program or resource that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * 
    * @return   The subscriber's ID.
    */
   static handle subscribeExternal(boolean global,
                                   handle  subscriber,
                                   String  event, 
                                   handle  publisher)
   {
      if (subscriber == null)
      {
         subscriber = handle.buildExternalResource();
      }
      
      subscribe(global, subscriber, new character(event), publisher, null);
      
      return subscriber;
   }
   
   /**
    * Delete a subscription to the given event for the specified external resource. This method is 
    * executed in a {@link ErrorManager#silent} bracket, as the <code>UNSUBSCRIBE</code>
    * statement is executed with a default NO-ERROR clause.
    *
    * @param    global
    *           Flag indicating if the search is done in the global (cross-session) event register.
    * @param    subscriber
    *           The resource whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure or resource handle representing the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribeExternal(boolean global,
                                   handle  subscriber,
                                   String  event,
                                   handle  publisher)
   {
      unsubscribe(global, subscriber, event == null ? null : new character(event), publisher);
   }
   
   /**
    * Notify all subscribers (to the given event) that the event has occurred.  This method is
    * executed in a {@link ErrorManager#silent} bracket, as the <code>PUBLISH</code>
    * statement is executed with a default NO-ERROR clause.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    * <p>
    * The subscriptions are fired in the order they were subscribed. Although 4GL
    * documentation specifically states, that "If a named event has multiple subscribers, 
    * the order in which the AVM notifies subscribers is undefined.", the actual
    * 4GL implementation (OpenEdge 10.2B) does fire the events in the order the subscriptions
    * get subscribed.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure or resource handle representing the program that is the source of  
    *           the event. Must not be <code>null</code>.
    * @param    args
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   static void publishExternal(String event, handle publisher, String[] args)
   {
      // convert the arguments from String to an Object element array
      Object[] param = new Object[args.length];
      System.arraycopy(args, 0, param, 0, args.length);
      publish(new character(event), publisher, null, param);
   }
   
   /**
    * Remove all subscriptions registered by the given resource.  This may be a program handle
    * or a {@link ExternalResource}.
    * 
    * @param    resource
    *           All subscriptions for this resource will be removed.
    */
   static void cleanupSubscriptions(Object resource)
   {
      WorkArea wa = work.get();
      Iterator<Subscription> iter = wa.subscriptions.iterator();
      while (iter.hasNext())
      {
         Subscription sub = iter.next();
         
         if (sub.subscriber.equals(resource))
         {
            iter.remove();
         }
      }
      
      globalSubscriptions.removeAll(resource);
   }
   
   /**
    * Delete a subscription to the given event for the specified program. This method is executed
    * executed in a {@link ErrorManager#silent} bracket, as the <code>UNSUBSCRIBE</code>
    * statement is executed with a default NO-ERROR clause.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribe(handle subscriber, character event, handle publisher)
   {
      unsubscribe(false, subscriber, event, publisher);
   }
   
   /**
    * Delete a subscription to the given event for the specified program. This method is executed
    * executed in a {@link ErrorManager#silent} bracket, as the <code>UNSUBSCRIBE</code>
    * statement is executed with a default NO-ERROR clause.
    *
    * @param    global
    *           Flag indicating if the search is done in the global (cross-session) event register.
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribe(boolean global, handle subscriber, character event, handle publisher)
   {
      silent(() ->
      {
         if (!subscriber.isExternalResource())
         {
            // validate subscriber
            if (subscriber.isUnknown())
            {
               final String msg = "Could not evaluate subscriber handle  expression while trying " +
                                  "to execute UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6475, msg, false);
               return;
            }
            if (!isProcedure(subscriber) || !subscriber._isValid())
            {
               final String msg = "Invalid or inappropriate subscriber handle value given to " +
                                  "UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6477, msg, false);
               return;
            }
         }
         
         // validate publisher; if null, is OK, as it means ALL option
         if (publisher != null && !publisher.isExternalResource())
         {
            if (publisher.isUnknown())
            {
               final String msg = "Could not evaluate publisher handle  expression while " +
                                  "trying to execute UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6476, msg, false);
               return;
            }
            if (!isProcedure(publisher) || !publisher._isValid())
            {
               final String msg = "Invalid or inappropriate publisher handle value given to " +
                                  "UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6478, msg, false);
               return;
            }
         }
         
         // Progress does not validate the event
         
         Object rsub = subscriber.get();
         Object rpub = (publisher == null ? null : publisher.get());
         
         boolean found = false;
         
         List<Subscription> subs;
         if (global)
         {
            subs = globalSubscriptions.getSubscriptions();
         }
         else
         {
            subs = work.get().subscriptions;
         }
         
         Iterator<Subscription> itr = subs.iterator();
         while (itr.hasNext())
         {
            Subscription sub = itr.next();
            
            // a special case: if a publisher is specified and ALL clause is used, than all 
            // ANYWERE subscriptions are marked read-only, for the event.
            if (rpub != null && event == null && sub.publisher == null)
            {
               sub.readOnly = true;
            }
            
            if (sub.readOnly)
            {
               // read-only subscriptions can't be removed
               continue;
            }
            
            // if removes all subscriptions considering:
            // - if the passed event is null, then ALL clause is in effect, else event names must
            // match
            // - the publisher is not matched as stated in the documentation. the correct approach
            // is that if the publisher is missing, then ANY publisher is matched
            if (sub.subscriber.equals(rsub)                &&
                (event == null || event.equals(sub.event)) &&
                ((rpub == null || rpub.equals(sub.publisher))))
            {
               if (global)
               {
                  globalSubscriptions.removeAll(sub);
               }
               else
               {
                  // remove this subscription
                  itr.remove();
               }
               
               found = true;
            }
         }
         
         // if a specific event is used and no subscriptions found, then error
         if (!found && event != null)
         {
            final String msg = "UNSUBSCRIBE failed.  No subscriptions for %s";
            String sevt = (event.isUnknown() ? "" : event.toStringMessage());
            ErrorManager.recordOrThrowError(6479, String.format(msg, sevt), false);
            return;
         }
      });
   }
   
   /**
    * Delete a subscription to ALL events for the specified program. This method is 
    * executed in a {@link ErrorManager#silent} bracket, as the <code>UNSUBSCRIBE</code>
    * statement is executed with a default NO-ERROR clause.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source.  When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribeAll(handle subscriber, handle publisher)
   {
      unsubscribe(subscriber, null, publisher);
   }
   
   /**
    * Ensure all the context-local data has initialized (to be able to listen
    * for scope notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope
      // notifications
      work.get();
   }
   
   /**
    * Add the <code>h</code> handle as a persistent procedure. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)}.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    * @param   name
    *          The name of this procedure, used at the RUN statement.
    * @param   persistent
    *          Flag indicating if this is a persistent procedure.
    */
   static void addProcedure(handle h, String name, boolean persistent)
   {
      addProcedure(work.get(), h, name, persistent);
   }
   
   /**
    * Remove the given handle from the list of persistent procedures. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)} in
    * cases when errors were encountered while invoking the procedure.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void removePersistentProcedure(handle h)
   {
      WorkArea wa = work.get();
      removeProcedureImpl(wa, h, wa.persistentProcedures);
   }
   
   /**
    * Implementation for the PROGRAM-NAME functions. Resolves the legacy top level block (external
    * program, function, internal procedure or trigger), at the given depth. The index starts with 
    * 0 for the currently executing top-level block.
    *  
    * @param    idx
    *           The stack entry index.
    * 
    * @return   The name of the stack entry or {@code null} if outside of stack.
    */
   static String getStackEntry(int idx)
   {
      WorkArea wa = work.get();
      
      if (idx < 0 || idx >= wa.thisProcedures.size())
      {
         return null;
      }
      
      Iterator<Object> procItr = wa.thisProcedures.iterator();
      Iterator<String> ieItr = wa.internalEntryStack.iterator();
      
      Object referent = null;
      String iename = null;
      
      for (int i = 0; i <= idx; i++)
      {
         referent = procItr.next();
         iename = ieItr.next();
      }
      
      String name = wa.relativeNames.get(referent);
      
      return EXTERNAL_PROGRAM.equals(iename) ? name : (iename + " " + name);
   }
   
   /**
    * Get the Progress stack trace
    * 
    * @return the Progress stack trace
    */
   
   public static List<String> getStackTrace()
   {
      WorkArea wa = work.get();
      List<String> stack = new ArrayList<>();
      
      Iterator<Object> procItr = wa.thisProcedures.iterator();
      Iterator<String> ieItr = wa.internalEntryStack.iterator();
      
      while (procItr.hasNext())
      {
         Object referent = procItr.next();
         String iename = ieItr.next();
         String name = wa.relativeNames.get(referent);
         stack.add(EXTERNAL_PROGRAM.equals(iename) ? name : (iename + " " + name));
      }
      return stack;
   }
   
   /**
    * Get the absolute name of the external program associated with the given java instance.
    * <p>
    * This is used internally by P2J, as we need absolute names when resolving a program's
    * signature, events, etc.
    * <p>
    * For the name associated with the procedure's NAME and FILE-NAME attributes, and returned by
    * the PROGRAM-NAME function or used in 4GL error messages, see {@link #getRelativeName}, which
    * returns the real name used when the external procedure was instantiated first.
    * 
    * @param    referent
    *           The referent for which the name is needed.
    *           
    * @return   The absolute name.
    */
   static String getAbsoluteName(Object referent)
   {
      String jname = referent.getClass().getName();
      String pname = SourceNameMapper.convertJavaProg(jname);
      return pname;
   }
   
   /**
    * Check if the given referent has been (or is still) registered with the
    * {@link ProcedureManager} in the {@link WorkArea#pMap procedure} map.
    * 
    * @param     referent
    *            The referent to be checked if is still loaded.
    * 
    * @return    <code>true</code> if the referent was found.
    */
   static boolean hasReferent(Object referent)
   {
      return work.get().pMap.containsKey(referent);
   }
   
   /**
    * Delete the given procedure handle (i.e. execute the DELETE OBJECT or DELETE PROCEDURE 
    * handle).
    * 
    * @param    referent
    *           The persistent procedure instance to be deleted.
    * 
    * @return   <code>true</code> if this procedure handle can be deleted.
    */
   static boolean delete(Object referent)
   {
      // handle is always valid
      WorkArea wa = work.get();
      
      ProcedureData pdata = locateProcedure(wa, referent);
      
      // do not postpone the delete when the referent is a TARGET-PROCEDURE or
      // a SOURCE-PROCEDURE
      if (_isPersistent(wa, referent) && wa.thisProcedures.contains(referent))
      {
         // if this is a persistent system handle, the actual delete is postponed until we
         // are finished.
         wa.deleteOnExit.put(referent, true);
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Postponing deletion of procedure " + getAbsoluteName(referent));
         }
         
         return false;
      }
      
      wa.persistentProcedures.remove(referent);
      wa.sessionSuperProcs.remove(referent);
      
      for (ProcedureData pp : wa.pMap.values())
      {
         pp.superProcedures.remove(referent);
      }
      
      // force the cleanup here (do not use cleanup()), we don't want to cause any inconsistency
      // here, when this method proceeds with clean up do the cleanup completely
      wa.forceCleanup(referent);
      
      try
      {
         wa.processedProcedure = referent;
         
         // cleanup after the procedure
         for (Finalizable target : pdata.finalizables)
         {
            target.deleted();
         }
         
         for (Scopeable target : pdata.scopeables)
         {
            target.scopeDeleted();
         }
         
         // as the procedure no longer exists, delete the static resources
         deleteResources(pdata.staticResources.pop());
      }
      finally
      {
         wa.processedProcedure = null;
      }
      
      handle unknown = new handle();
      
      // remove it from the instantiating procedure
      for (HandleResource res : pdata.createdResources.keySet())
      {
         res.setInstantiatingProcedure(unknown);
      }
      
      // delete the widget pools for this external program
      WidgetPool.deleteWidgetPools(referent);
      
      // get rid of references...
      pdata.createdResources.clear();
      pdata.scopeables.clear();
      pdata.finalizables.clear();
      
      // remove any triggers created in this-procedure
      LogicalTerminal.cleanupTriggers(referent);
      
      handle.removeResource(pdata.phandle.getResource());
      
      // clear all undoable variables created by this procedure (they might have been tracked
      // in parent tx blocks).
      TransactionManager.untrackUndoables(referent);
      
      return true;
   }
   
   /**
    * Using the name of a function defined IN handle and a procedure handle 
    * to which this function definition belongs, resolve the target handle,
    * to be able to invoke it.
    * 
    * @param    phandle
    *           The procedure handle where the function was defined.
    * @param    fname
    *           The function's legacy name.
    *           
    * @return   The procedure handle where to search for the function's body.
    */
   static handle getHandleForFunction(handle phandle, String fname)
   {
      Object referent = phandle.get();
      ProcedureData pp = locateProcedure(referent);
      
      handle h = (handle) pp.inHandleFunctions.get(fname).resolve();
      return h == null ? new handle() : h;
   }
   
   /**
    * For procedures or functions defined IN SUPER, this computes the target procedure handle 
    * where the procedure's or function's body resides.
    * <p>
    * In case when no super procedure is found, it will throw proper error messages and return
    * <code>null</code>. 
    * 
    * @param    iename
    *           Internal-entry's name (this may be a procedure or function).
    * @param    function
    *           <code>true</code> if this is a function.
    *           
    * @return   A procedure handle where to search for the internal-entry's body or
    *           <code>null</code> of none found.
    */
   static handle getSuperHandle(String iename, boolean function)
   {
      WorkArea wa = work.get();
      
      handle hthis           = thisProcedure(wa);
      handle htarget         = targetProcedure(wa);
      Object refThis         = hthis.get();
      Object refTarget       = htarget.get();
      ProcedureData ppThis   = locateProcedure(wa, refThis);
      ProcedureData ppTarget = locateProcedure(wa, refTarget);
      
      boolean searchSelf = ppThis.isSearchSelf();
      handle psuper = null;
      
      @SuppressWarnings("unchecked")
      Collection<Object>[] search = new Collection[2];
      
      if (searchSelf)
      {
         search[0] = ppThis.superProcedures.keySet();
         search[1] = wa.sessionSuperProcs.keySet();
      }
      else
      {
         Collection<Object> targetProcs = ppTarget.superProcedures.keySet();
         search[0] = new ArrayDeque<>(targetProcs);
         search[1] = new ArrayDeque<>(wa.sessionSuperProcs.keySet());
         
         boolean inTarget = search[0].contains(refThis);
         boolean inSess   = search[1].contains(refThis);
         
         if (inTarget || inSess)
         {
            // remove referents until this-procedure is found in the list
            prepareSearches(search, refThis);
         }
         
         if (!(inTarget || inSess) || (inSess && search[1].isEmpty() && !targetProcs.isEmpty()))
         {
            if (!function)
            {
               // in some cases when invoking internal procedures, it can end up in an infinite 
               // recursion
               
               search[0] = new ArrayDeque<>(targetProcs);
               search[1] = new ArrayDeque<>(wa.sessionSuperProcs.keySet());
               
               // step 1. find the first already-invoked proc part of target's super-procs
               Object ref = null;
               for (CalleeInfoImpl info : wa.calleeInfo)
               {
                  if (targetProcs.contains(info.destination))
                  {
                     ref = info.destination;
                     break;
                  }
                  
                  if (!info.superCall && refTarget != info.destination)
                  {
                     break;
                  }
               }
               
               if (ref != null)
               {
                  // step 2. remove all until ref is found, from all searches
                  prepareSearches(search, ref);
               }
               else
               {
                  search[0] = null;
                  search[1] = null;
               }
            }
            else
            {
               // not found, clear the lists
               search[0] = null;
               search[1] = null;
            }
         }
      }
      
      for (int i = 0; i < search.length && psuper == null; i++)
      {
         if (search[i] == null || search[i].isEmpty())
         {
            continue;
         }
         
         // search is done backwards (last set super-procedure is searched first)
         Deque<Object> queue = new ArrayDeque<>(search[i]);
         while (!queue.isEmpty())
         {
            Object referent = queue.removeLast();
            ProcedureData pp = locateProcedure(wa, referent);
            
            if (pp == null || !pp.phandle._isValid())
            {
               // deleted or invalid procedures are ignored
               continue;
            }
            
            String spname = getAbsoluteName(pp.phandle.get());
            String jname = SourceNameMapper.getMethodName(spname, iename, function);
            if (jname != null)
            {
               // found it
               psuper = pp.phandle;
               break;
            }
         }
      }
      
      if (psuper == null)
      {
         String err = null;
         int errId = -1;
         if (function)
         {
            final String errMsg = 
               "SUPER version of user-defined function '%s' invoked but could not be found";
            errId = 6454;
            err = String.format(errMsg, iename);
            
            ErrorManager.recordOrShowError(errId, err, false, false, false);
         }
         else
         {
            String pname = getRelativeName(searchSelf ? refThis : refTarget);
            
            errId = 6439;
            if (searchSelf)
            {
               final String errMsg = 
                  "Procedure %s %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, iename, pname, iename);
            }
            else
            {
               final String errMsg = 
                 "Procedure %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, pname, iename);
            }
            
            ErrorManager.recordOrThrowError(errId, err, false);
         }
      }
      
      return psuper;
   }
   
   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   static logical isPersistent(Object referent)
   {
      return new logical(_isPersistent(work.get(), referent));
   }
   
   /**
    * Check if this is a "delete-on-exit" procedure.
    * 
    * @param    referent
    *           An instance of a converted external procedure class.
    *           
    * @return   <code>true</code> if the referent is a "delete-on-exit" procedure.
    */
   static boolean _isDelayedDelete(Object referent)
   {
      Boolean delete = work.get().deleteOnExit.get(referent);
      
      return delete != null && delete;
   }
   
   /**
    * Check if the given referent is the top-level procedure (the startup procedure).
    * 
    * @param    referent
    *           The instance to be verified.
    *           
    * @return   <code>true</code> if the referent is the startup procedure.
    */
   static boolean isTopLevel(Object referent)
   {
      Deque<Object> procs = work.get().thisProcedures;
      return procs.getLast() == referent;
   }
   
   /**
    * Check if the given handle is a proxy procedure.
    * 
    * @param    h
    *           The handle to check.
    *           
    * @return   <code>true</code> if the handle is a proxy procedure.
    */
   static boolean isProxy(handle h)
   {
      return h != null && h._isValid() && (h.getResource() instanceof ProxyProcedureWrapper);
   }
   
   /**
    * Get the <code>SUPER-PROCEDURES</code> attribute of the specified 
    * external procedure instance.
    *
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character superProcedures(Object referent)
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSuperProcedures(referent).values());
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h)
   {
      return addSuperProcedure(referent, h, (SearchMode) null);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the explicit {@link SearchMode mode}.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h,SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, referent);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the search-mode set as a string (one of the 
    * <code>SEARCH-TARGET</code> or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h, int mode)
   {
      return addSuperProcedure(referent, h, getSearchMode(mode));
   }
   
   /**
    * Remove the given handle from the specified referent's list of 
    * super-procedures.
    *  
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical removeSuperProcedure(Object referent, handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      removeProcedureImpl(wa, h, wa.getSuperProcedures(referent).keySet());
      
      return new logical(true);
   }
   
   /**
    * Get the {@code NEXT-SIBLING} attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getNextSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object next = wa.persistentProcedures.getNext(referent);
      
      if (next == null)
      {
         return new handle();
      }
      
      return locateProcedure(wa, next).phandle;
   }
   
   /**
    * Get the {@code PREV-SIBLING} attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getPrevSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = wa.persistentProcedures.getPrevious(referent);
      
      if (prev == null)
      {
         return new handle();
      }
      
      return locateProcedure(wa, prev).phandle;
   }
   
   static void setPrivateData(Object referent, character data)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.privateData = new character(data);
   }
   
   /**
    * Get the <code>PRIVATE-DATA</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static character getPrivateData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      
      return pp.privateData == null ? new character() : new character(pp.privateData);
   }
   
   /**
    * Get the value of the ADM-DATA attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character getADMData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new character(pp.admData);
   }
   
   /**
    * Set the value of the ADM-DATA attribute for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @param    value
    *           The new value.
    */
   static void setADMData(Object referent, String value)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.admData.assign(value);
   }
   
   /**
    * Gets the the unique ID number associated to the referent. Not the 
    * same as the handle value itself.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   The integer number of the children.
    */
   static integer getUniqueID(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new integer(pp.uniqueID);
   }
   
   /**
    * Get a handle list of super-procedures for the given referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures(Object referent)
   {
      WorkArea wa = work.get();
      
      List<handle> superProcs = new LinkedList<>();
      // need to emit them in reverse order (they are kept in LIFO)
      for (ProcedureData pp : wa.getSuperProcedures(referent).values())
      {
         superProcs.add(0, pp.phandle);
      }
      
      return superProcs;
   }
   
   /**
    * Get a handle list of super-procedures for the session handle.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures()
   {
      return getSuperProcedures(null);
   }
   
   /**
    * Check if the given referent is for a system handle (i.e. for one of the
    * THIS-PROCEDURE, SOURCE-PROCEDURE or TARGET-PROCEDURE handles).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static boolean systemHandle(Object referent)
   {
      WorkArea wa = work.get();
      return wa.thisProcedures.contains(referent) ||
             wa.targetProcedures.contains(referent) ||
             wa.sourceProcedures.contains(referent);
   }
   
   /**
    * Register the given {@link Scopeable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Scopeable} instance which needs delayed notification.
    */
   static void addScopeable(Object referent, Scopeable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.scopeables.add(target);
   }
   
   /**
    * Register the given {@link Finalizable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Finalizable} instance which needs delayed notification.
    */
   static void addFinalizable(Object referent, Finalizable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.finalizables.add(target);
   }
   
   /**
    * Gets the INSTANTIATING-PROCEDURE associated to the referent. 
    * 
    * @param    referent
    *           The external program.
    * 
    * @return   See above.
    */
   static handle getInstantiatingProcedure(Object referent)
   {
      ProcedureData pdata = locateProcedure(referent);
      return new handle(pdata.instProc);
   }
   
   /**
    * Resolve the instantiating procedure from the given handle: if the handle contains a valid
    * procedure, return a copy of it; else, return an unknown handle.
    * 
    * @param    hproc
    *           The handle instance representing an instantiating procedure.
    * 
    * @return   See above.
    */
   static handle resolveInstantiatingProcedure(handle hproc)
   {
      handle res = new handle();
      
      if (hproc._isValid())
      {
         res.assign(hproc);
      }
      
      return res;
   }
   
   /**
    * Find the root enclosing instance of the specified inner class instance.
    * It gets computed using the <code>this$</code> fields inserted by the 
    * compiler; it returns the instance refered by the <code>this$0</code> 
    * fields, when reached.
    * 
    * @param    instance
    *           The inner class instance.
    *           
    * @return   The root instance.
    */
   static Object findRootEnclosingInstance(Object instance)
   {
      Field[] fields = instance.getClass().getDeclaredFields();
      for (Field f : fields)
      {
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            try
            {
               f.setAccessible(true);
               Object outer = f.get(instance);
               
               return fname.equals("this$0") ? outer : findRootEnclosingInstance(outer);
            }
            catch (Exception e)
            {
               e.printStackTrace();
               return null;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Register the given resource as being created by the specified procedure.
    * 
    * @param    referent
    *           The procedure which creates the resource.
    * @param    resource
    *           The resource being created.
    */
   static void registerResource(Object referent, HandleResource resource)
   {
      ProcedureData pdata = locateProcedure(referent);
      if (pdata != null)
      {
         pdata.createdResources.put(resource, null);
      }
      else if (LOG.isLoggable(Level.FINE))
      {
         Long id = handle.resourceId(resource);
         LOG.log(Level.FINE, "Could not register resource " + resource.type() + 
                             " with id " + id + " - the data for procedure " + 
                              referent.getClass().getName() + " could not be found!");
      }
   }
   
   /**
    * Deregister the given resource from the specified creator procedure.
    * 
    * @param    referent
    *           The procedure which creates the resource.
    * @param    resource
    *           The resource being created.
    */
   static void deregisterResource(Object referent, HandleResource resource)
   {
      ProcedureData pdata = locateProcedure(referent);
      
      if (pdata != null)
      {
         pdata.createdResources.remove(resource);
      }
      else if (LOG.isLoggable(Level.FINE))
      {
         Long id = handle.resourceId(resource);
         LOG.log(Level.FINE, "Could not deregister resource " + resource.type() + 
                             " with id " + id + " - the data for procedure " + 
                              referent.getClass().getName() + " could not be found!");
      }
   }
   
   /**
    * Set the name of the 4GL program currently being instantiated, into the 
    * {@link WorkArea#initiatingExtProg} field.
    * 
    * @param    name
    *           The 4GL external program name.
    */
   static void setInstantingExternalProgram(String name)
   {
      WorkArea wa = work.get();
      wa.initiatingExtProg = name;
   }
   
   /**
    * Walk the {@link WorkArea#thisProcedures THIS-PROCEDURE} stack and apply the search function;
    * return the first value given by the search function, once it returns a non-null value.
    * 
    * @param    search
    *           The search function.
    *           
    * @return   The first non-null value returned by the search function.
    */
   static Object searchInStack(Function<Object, Object> search)
   {
      Map<Object, Object> searched = new IdentityHashMap<>();
      
      WorkArea wa = work.get();
      Iterator<Object> iter = wa.thisProcedures.iterator();
      while (iter.hasNext())
      {
         Object referent = iter.next();
         if (searched.containsKey(referent))
         {
            continue;
         }
         
         Object ret = search.apply(referent);
         
         if (ret != null)
         {
            return ret;
         }
         
         searched.put(referent, referent);
      }
      
      return null;
   }
   
   /**
    * Get the THIS-PROCEDURE system handle.
    * 
    * @param    wa
    *           Context-local work area.
    * 
    * @return   See above.
    */
   private static handle thisProcedure(WorkArea wa)
   {
      Object referent = wa.thisProcedures.peek();
      
      return new handle(locateProcedure(wa, referent).phandle);
   }
   
   /**
    * Get the THIS-PROCEDURE referent.
    * 
    * @param    wa
    *           Context-local work area.
    * 
    * @return   See {@link #thisProcedure()}.
    */
   private static Object _thisProcedure(WorkArea wa)
   {
      return wa.thisProcedures.peek();
   }
   
   /**
    * Get the TARGET-PROCEDURE system handle.
    * 
    * @param    wa
    *           Context-local work area.
    * 
    * @return   See above.
    */
   private static handle targetProcedure(WorkArea wa)
   {
      // Do not use locateProcedure() here, the system handle may have been deleted
      // with DELETE PROCEDURE. Note that when a procedure is deleted it can be
      // referenced and queried through the system procedure handles.
      return new handle(wa.targetProceduresData.peek().phandle);
   }
   
   /**
    * Check if the given referent is for a persistent procedure. This implements the PERSISTENT
    * attribute for a procedure handle.
    * 
    * @param    wa
    *           Context-local work area.
    * @param    referent
    *           An instance of a converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   private static boolean _isPersistent(WorkArea wa, Object referent)
   {
      return wa.persistentProcedures.contains(referent);
   }
   
   /**
    * Push an internal entry to the stack.  The passed internal entry name is resolved and is
    * pushed as it was defined, not as it is being called.
    * <p>
    * Use {@link #EXTERNAL_PROGRAM} for the name of the root external procedure block and 
    * {@link #TRIGGER_PROCEDURE} for the root trigger block.
    * 
    * @param    wa
    *           Context-local work area.
    * @param    instance
    *           The external program instance.
    * @param    iename
    *           The legacy internal-entry name.
    * @param    function
    *           Flag indicating if the internal entry is a user-defined function.
    */
   private static void pushInternalEntry(WorkArea wa,
                                         Object instance,
                                         String iename,
                                         boolean function)
   {
      if (!(iename.equals(EXTERNAL_PROGRAM) || iename.equals(TRIGGER_PROCEDURE)))
      {
         // the internal entry names are reported as they are defined, not as they are called.
         // thus, we need to determine the definition and get it from there
         
         String pname = getAbsoluteName(instance);
         InternalEntry ie = SourceNameMapper.getInternalEntry(pname, iename, function);
         
         iename = (ie == null ? iename : ie.pname);
      }
      
      wa.internalEntryStack.push(iename);
   }
   
   /**
    * Add the <code>h</code> handle as a persistent procedure.
    * 
    * @param   wa
    *          Context-local work area.
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    * @param   name
    *          The name of this procedure, used at the RUN statement.
    * @param   persistent
    *          Flag indicating if this is a persistent procedure.
    */
   private static void addProcedure(WorkArea wa, handle h, String name, boolean persistent)
   {
      Object referent = h.get();
      ProcedureData pp = new ProcedureData(h, null);
      
      wa.pMap.put(referent, pp);
      
      if (persistent)
      {
         wa.persistentProcedures.add(referent);
      }
      wa.relativeNames.put(referent, name);
   }
   
   /**
    * Implementation for the SUPER-PROCEDURES attribute.
    *  
    * @param    superProcs
    *           A list of {@link ProcedureData} objects, in the order in 
    *           which they were added.
    *           
    * @return   The string representation of each super-procedure, separated
    *           by a comma. The last added super-procedure must be first in
    *           this list, the second-to-last must be the second in the list,
    *           etc.
    */
   private static character superProceduresImpl(Collection<ProcedureData> superProcs)
   {
      character sp = new character("");
      boolean first = true;
      ArrayDeque<ProcedureData> queue = new ArrayDeque<>();
      queue.addAll(superProcs);
      // need to emit them in a LIFO order
      while (!queue.isEmpty())
      {
         ProcedureData pp = queue.removeLast();
         long hashCode = handle.resourceId(pp.phandle.getResource());
         sp.assign(TextOps.concat(sp, (!first ? "," : ""), hashCode));
         first = false;
      }
      
      return sp;
   }
   
   /**
    * The implementation of the ADD-SUPER-PROCEDURE method. When the
    * <code>referent</code> is null, this call is for the SESSION handle.
    * 
    * @param    h
    *           A handle to be added as super-procedure.
    * @param    mode
    *           The {@link SearchMode} for the passed handle.
    * @param    referent
    *           When <code>null</code> add to the SESSION:SUPER-PROCEDURES
    *           list. Else, add to the procedure associated with this 
    *           referent.
    * 
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   private static logical addSuperProcedureImpl(handle h, SearchMode mode, Object referent)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      Map<Object, ProcedureData> superProcs = wa.getSuperProcedures(referent);
      ProcedureData pData = locateProcedure(wa, h.get());
      
      // the super-procedure to be added to the chain will always exist (either as the current
      // this-procedure or as a persistent procedure). the remove call is needed just to eliminate
      // it from the referent's super procedure chain, if was previously added 
      removeProcedureImpl(wa, h, superProcs.keySet());
      
      if (!(mode == null || pData.searchMode == null || pData.searchMode.equals(mode)))
      {
         // if the mode is changed, show warning
         int errNum = 8914;
         String msg = "**Changing proc-search for procedure %s from  '%s' to '%s'";
         
         String progName = getRelativeName(referent);
         String oldMode  = pData.searchMode.toString();
         String newMode  = mode.toString();
         
         msg = String.format(msg, progName, oldMode, newMode);
         
         // this is just a warning, the mode will be changed
         ErrorManager.recordOrShowError(errNum, msg, false, false, false);
      }
      
      pData.searchMode = mode;
      superProcs.put(h.get(), pData);
      
      return new logical(true);
   }
   
   /**
    * Removes the given procedure handle from the given procedure instances list.
    * 
    * @param    wa
    *           Context-local work area.
    * @param    h
    *           The procedure handle to be removed.
    * @param    ppList
    *           The collection from which the procedure must be removed.
    */
   private static void removeProcedureImpl(WorkArea wa, handle h, Collection<Object> ppList)
   {
      // if the handle is null, there is nothing to do
      if (h == null)
      {
         return;
      }
      
      Object hreferent = h.get();
      
      Iterator<Object> iter = ppList.iterator();
      while (iter.hasNext())
      {
         Object referent = iter.next();
         if (hreferent == referent)
         {
            iter.remove();
         }
      }
      
      wa.cleanup(hreferent);
   }
   
   /**
    * Get the {@link SearchMode} constant associated with the given mode, used
    * by ADD-SUPER-PROCEDURE calls. If the mode is not one of the SEARCH-SELF 
    * or SEARCH-TARGET values, or an {@link IllegalArgumentException} will be 
    * thrown.
    * <p>
    * The given <code>mode</code> is not compatible with the mode returned by
    * {@link SearchMode#toString()}.
    * 
    * @param    mode
    *           The string representation of the search-mode.
    *           
    * @return   For SEARCH-TARGET value, {@link SearchMode#SEARCH_TARGET} 
    *           will be returned. For SEARCH-SELF value, 
    *           {@link SearchMode#SEARCH_SELF} will be returned.
    *
    * @throws   IllegalArgumentException
    *           If the mode is not of a known value.
    */
   private static SearchMode getSearchMode(int mode)
   {
      if (mode == SEARCH_SELF)
         return SearchMode.SEARCH_SELF;
      
      if (mode == SEARCH_TARGET)
         return SearchMode.SEARCH_TARGET;
      
      // if not a known mode, default to "search mode not specified"
      return null;
   }
   
   /**
    * Locate the {@link ProcedureData} instance associated with the given external program
    * instance.
    * <p>
    * This will find any persistent procedure or any procedure active on the stack.
    * 
    * @param    referent
    *           The external program instance.
    * 
    * @return   See above.
    */
   private static ProcedureData locateProcedure(Object referent)
   {
      return locateProcedure(work.get(), referent);
   }
   
   /**
    * Locate the {@link ProcedureData} instance associated with the given external program
    * instance.
    * <p>
    * This will find any persistent procedure or any procedure active on the stack.
    * 
    * @param    wa
    *           Context-local work area.
    * @param    referent
    *           The external program instance.
    * 
    * @return   See above.
    */
   private static ProcedureData locateProcedure(WorkArea wa, Object referent)
   {
      return wa.pMap.get(referent);
   }
   
   /**
    * Delete the given resources, as long as they are still valid.
    * 
    * @param    resources
    *           The resources to be deleted.
    */
   private static void deleteResources(Set<WrappedResource> resources)
   {
      resources.stream().filter(WrappedResource::valid).forEach(target -> {
         ((Deletable) target).delete();
      });
   }
   
   /**
    * Remove elements from the end of each queue until the specified referent is found.
    * 
    * @param   queues
    *          The queues which need to be prepared.
    * @param   ref
    *          The referent on which removing will stop.  This will be also removed, if found.
    */
   private static void prepareSearches(Collection<?>[] queues, Object ref)
   {
      l1: for (Collection<?> queue_i : queues)
      {
         Deque<Object> queue = new ArrayDeque<>(queue_i);
         while (!queue.isEmpty())
         {
            Object referent = queue.removeLast();
            
            queue_i.remove(referent);
            
            if (referent == ref)
            {
               break l1;
            }
         }
      }
   }
   
   /**
    * Container for a PUBLISH task to be executed in the actual subscription's context.
    */
   private static class PublishTask
   {
      /** The subscription on which {@link Subscription#session context } the task is executed. */
      private final Subscription subscription;

      /** The PUBLISH task. */
      private final Runnable task;
      
      /**
       * Create a new task.
       * 
       * @param   subscription
       *          The subscription.
       * @param   task
       *          The PUBLISH task.
       */
      public PublishTask(Subscription subscription, Runnable task)
      {
         this.subscription = subscription;
         this.task = task;
      }
   }

   /**
    * Helper class to manage the global subscriptions.
    */
   private static class GlobalEventSubscriptions
   {
      /** The global list of subscriptions. */
      private final Map<Session, List<Subscription>> subscriptions = new HashMap<>();

      /** Map of session listener for each subscription. */
      private final Map<Subscription, SessionListener> listeners = new HashMap<>();

      /** Tasks for PUBLISH are posted here. */
      private Queue<PublishTask> queue = new LinkedList<>();
      
      /** Flag indicating the listening loop was terminated. */
      private boolean terminated = false;
      
      /**
       * Create a dedicated thread, in the server's context, which will {@link #listen()} for 
       * PUBLISH tasks.
       */
      public GlobalEventSubscriptions()
      {
         String jobName = "global-publisher";
         Runnable launchJob = () ->
         {
            new AssociatedThread(new Runnable()
            {
               @Override
               public void run()
               {
                  listen();
               }
            })
            {
               {
                  this.setName("Global Publisher Worker");
                  this.setDaemon(true);
               };
            }.start();
         };

         Job job = new Job(jobName, launchJob, true, JobMode.NOW);
         if (!Scheduler.addJob(job))
         {
            throw new IllegalStateException(
               "A job with for " + jobName + " is already configured!");
         }
      }
      
      /**
       * Listen for incoming {@link PublishTask tasks}, on the {@link #queue}.
       * <p>
       * Once a task is received, an {@link AssociatedThread} will be created and will execute the
       * task in the {@link Subscription#session subscription's session}, using 
       * {@link SessionManager#executeInContext}.
       */
      private void listen()
      {
         SessionManager sm = SessionManager.get();

         List<PublishTask> tasks = new ArrayList<>(); 

         while (true)
         {
            tasks.clear();

            synchronized (queue)
            {
               if (queue.isEmpty())
               {
                  try
                  {
                     queue.wait();
                  }
                  catch (InterruptedException e)
                  {
                     LOG.log(Level.INFO, "global publish event listening loop interrupted");
                     
                     // terminate the loop
                     terminated = true;
                     break;
                  }
               }
               
               tasks.addAll(queue);
               queue.clear();
            }
            
            for (PublishTask task : tasks)
            {
               Runnable ptask = () ->
               {
                  Thread t = Thread.currentThread();
                  
                  try
                  {
                     sm.registerAsyncThread(t);
                     
                     task.task.run();
                  }
                  finally
                  {
                     sm.deregisterAsyncThread(t);
                  }
               };
               
               try
               {
                  sm.executeInContext(task.subscription.session, 
                                      () -> new AssociatedThread(ptask).start());
               }
               catch (RestrictedUseException e)
               {
                  String msg = "Not allowed to use SessionManager.executeInContext";
                  throw new IllegalStateException(msg, e);
               }
            }
         }
      }
      
      /**
       * Post the given task to the {@link #queue}.
       * 
       * @param    task
       *           The task for a PUBLISH event.
       */
      public void post(PublishTask task)
      {
         synchronized (queue)
         {
            if (terminated)
            {
               LOG.log(Level.INFO,
                      "global publish event listening loop is terminated - could not post event");
               return;
            }
            
            queue.offer(task);
            queue.notify();
         }
      }
      
      /**
       * Get all the global subscriptions, for all contexts.
       * 
       * @return   See above.
       */
      public synchronized List<Subscription> getAllSubscriptions()
      {
         if (subscriptions.isEmpty())
         {
            return Collections.EMPTY_LIST;
         }

         List<Subscription> subs = new LinkedList<>();
         subscriptions.values().forEach(l -> subs.addAll(l));
         return subs;
      }
      
      /**
       * Remove all the global subscriptions for the given resource.  This will remove them only
       * from the current context's global subscriptions.
       * 
       * @param    resource
       *           The resource which will be matched against the {@link Subscription#subscriber}.
       */
      public synchronized void removeAll(Object resource)
      {
         if (subscriptions.isEmpty())
         {
            return;
         }
         
         Session session = SessionManager.get().getSession();
         List<Subscription> subs = subscriptions.get(session);
         
         if (subs == null)
         {
            return;
         }
         
         Iterator<Subscription> iter = subs.iterator();
         while (iter.hasNext())
         {
            Subscription sub = iter.next();
            
            if (sub.subscriber.equals(resource))
            {
               session.removeSessionListener(listeners.remove(sub));
               iter.remove();
            }
         }
         
         if (subs.isEmpty())
         {
            subscriptions.remove(session);
         }
      }
      
      /**
       * Remove the specified subscription from the current context's global subscriptions.
       * 
       * @param    subscription
       *           The subscription which will removed.
       */
      public synchronized void removeAll(Subscription subscription)
      {
         if (subscriptions.isEmpty())
         {
            return;
         }
         
         Session session = SessionManager.get().getSession();
         List<Subscription> subs = subscriptions.get(session);
            
         Iterator<Subscription> iter = subs.iterator();
         while (iter.hasNext())
         {
            Subscription sub = iter.next();
            
            if (sub == subscription)
            {
               session.removeSessionListener(listeners.remove(sub));
               iter.remove();
            }
         }
         
         if (subs.isEmpty())
         {
            subscriptions.remove(session);
         }
      }
      
      /**
       * Get all the global subscriptions, for this context.
       * 
       * @return   See above.
       */
      public synchronized List<Subscription> getSubscriptions()
      {
         if (subscriptions.isEmpty())
         {
            return Collections.EMPTY_LIST;
         }

         Session session = SessionManager.get().getSession();
         List<Subscription> subs = subscriptions.get(session);
         
         return subs == null ? Collections.EMPTY_LIST 
                             : new ArrayList<>(subscriptions.get(session));
      }
      
      /**
       * Register the specified {@link Subscription} as global, for this context.
       * 
       * @param    subscription
       *           The subscription to register.
       */
      public synchronized void subscribe(Subscription subscription)
      {
         Session session = SessionManager.get().getSession();

         List<Subscription> subs = subscriptions.get(session);
         if (subs == null)
         {
            subs = new ArrayList<>();
            subscriptions.put(session, subs);
         }

         SessionListener listener = new SessionListener()
         {
            @Override
            public void terminate(Session session)
            {
               synchronized (GlobalEventSubscriptions.this)
               {
                  subscriptions.remove(session);
                  listeners.remove(subscription);
               }
            }
            
            @Override
            public void initialize(Session session)
            {
               // no-op
            }
         };
         
         session.addSessionListener(listener);
         listeners.put(subscription, listener);

         subs.add(subscription);
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /** The list of all registered persistent procedures, via RUN PERSISTENT calls. */
      // private Deque<Object> persistentProcedures = new ArrayDeque<>();
      private FastSiblingDeque<Object> persistentProcedures = new FastSiblingDeque<>();
      
      /** A map containing the list of super-procedures, for the session, by each instance. */
      private Map<Object, ProcedureData> sessionSuperProcs = new LinkedHashMap<>();
      
      /** 
       * This map holds the attributes for all registered external procedures, by instance.
       * It includes all persistent procedures and all procedures active on the stack.
       */
      private Map<Object, ProcedureData> pMap = new IdentityHashMap<>();
      
      /** Stack of target-procedures, by instance. */
      private Deque<Object> targetProcedures = new ArrayDeque<>();
      
      /**
       * Stack of target-procedures' data. Must be kept in sync with targetProcedures!
       * This stack holds procedure data for deleted TARGET-PROCEDURE handles.
       */
      private Deque<ProcedureData> targetProceduresData = new ArrayDeque<>();
      
      /** Stack of source-procedures, by instance. */
      private Deque<Object> sourceProcedures = new ArrayDeque<>();
      
      /**
       * Stack of source-procedures' data. Must be kept in sync with sourceProcedures!
       * This stack holds procedure data for deleted SOURCE-PROCEDURE handles.
       */
      private Deque<ProcedureData> sourceProceduresData = new ArrayDeque<>();
      
      /** Stack of this-procedures, by instance. */
      private Deque<Object> thisProcedures = new ArrayDeque<>();
      
      /** Explicit referent used as SOURCE-PROCEDURE for a PUBLISH statement. */
      private Object publishSourceProcedure = null;
      
      /**
       * Stack of internal entry names.  Evolves together with the {@link #thisProcedures} stack. 
       */
      private Deque<String> internalEntryStack = new ArrayDeque<>();
      
      /** Resources pending to registered with THIS-PROCEDURE when the next block is entered. */
      private Set<WrappedResource> pendingResources = new LinkedHashSet<>();
      
      /** Map of relative names (i.e. used at the RUN statement) for each external program. */
      private Map<Object, String> relativeNames = new IdentityHashMap<>();
      
      /**
       * Indicates the current external program when a scope or finalizable notification occurs.
       */
      private Object processedProcedure = null;
      
      /** A stack containing the info about the next called program or internal entry. */
      private Deque<CalleeInfoImpl> calleeInfo = new ArrayDeque<>();
      
      /** The 4GL program name being currently initiating. */
      private String initiatingExtProg = null;
      
      /** Map with flags indicating if this procedure should be deleted on exit. */
      private Map<Object, Boolean> deleteOnExit = new IdentityHashMap<>();
      
      /** A registry of persistent procedures which were instantiated by a remote side. */
      private Set<Object> remotePersistentProcedures = new HashSet<>();
      
      /** The global list of subscriptions, for this context. */
      private List<Subscription> subscriptions = new LinkedList<>();
      
      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeStart()
      {
         BlockType bt = TransactionManager.getBlockType();
         
         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // push always before pushing new THIS-PROCEDURE
            CalleeInfoImpl info = calleeInfo.peek();
            info.push();
            
            ProcedureData pdata = pMap.get(info.destination);
            
            // setting INSTANTIATING-PROCEDURE for external procedure block can be done only 
            // after ProcedureData is created
            if (bt.equals(BlockType.EXTERNAL_PROC))
            {
               int nestingLevel = TransactionManager.getNestingLevel();
               if (AppServerManager.isRemote() ? nestingLevel > 3 : nestingLevel > 2)
               {
                  // if this is not the entry point, set the INSTANTIATING-PROCEDURE as the
                  // current THIS-PROCEDURE
                  pdata.instProc.assign(thisProcedure());
               }
            }
            
            thisProcedures.push(info.destination);
            
            WidgetPool.scopeStart();
            
            pdata.staticResources.push(new LinkedHashSet<>());
         }
         
         if (!TransactionManager.isGlobalBlock())
         {
            ProcedureData pdata = locateProcedure(this, _thisProcedure(this));
            
            // process the newly created resources
            for (WrappedResource res : pendingResources)
            {
               if (res instanceof HandleResource)
               {
                  ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
               }
            }
            
            // collect the static resources in the correct scope
            pdata.staticResources.peek().addAll(pendingResources);
            pendingResources.clear();
         }
      }
      
      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeFinished()
      {
         BlockType bt = TransactionManager.getBlockType();
         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // pop this-procedure from the stack and cleanup
            Object referent = thisProcedures.pop();
            ProcedureData pdata = locateProcedure(referent);
            
            CalleeInfoImpl info = calleeInfo.peek();
            info.pop();
            
            boolean extProc = bt.equals(BlockType.EXTERNAL_PROC);
            boolean persistent = _isPersistent(this, referent);
            
            // check if we have a postponed delete
            Boolean delete = deleteOnExit.get(referent);
            
            if (delete != null && delete)
            {
               // check if this is the last usage of the referent
               
               if (thisProcedures.contains(referent))
               {
                  // still used, wait for the last proc on the stack to finish.
                  delete = null;
               }
               else
               {
                  // this is no longer a persistent procedure being used, can be deleted
                  
                  deleteOnExit.remove(referent);
                  
                  persistentProcedures.remove(referent);
                  persistent = false;
               }
            }
            
            WidgetPool.scopeFinished(referent, extProc, persistent);
            
            boolean deleted = false;
            
            // notify the widget-pool management a procedure has finished; this needs to be done
            // here to ensure that the correct THIS-PROCEDURE handle is available
            if (!(extProc && persistent))
            {
               // bypass this if there was a postponed delete, as the resources were deleted by 
               // the delete(referent) call; otherwise delete this external, non-persistent,
               // procedure
               if (extProc && !persistent)
               {
                  deleted = delete(referent);
                  
                  // already deleted, nothing else to do
                  delete = null;
               }
               else
               {
                  // delete the static resources created in an internal proc/function/trigger 
                  deleteResources(pdata.staticResources.pop());
               }
            }
            else
            {
               cleanup(referent);
            }
            
            if (delete != null && delete)
            {
               String name = getAbsoluteName(referent);
               
               // force the delete of this external procedure, as some internal proc/func/trig
               // has called DELETE OBJECT THIS-PROCEDURE
               deleted = delete(referent);
               
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.log(Level.FINE, "Processing postponed delete for procedure " + name + 
                                      " has " + (deleted ? "passed" : "failed"));
               }
            }
            
            if (deleted)
            {
               handle.removeResource(new ExternalProgramWrapper(referent));
            }
         }
      }
      
      /**
       * Provides a notification that an external scope is about to be deleted.
       * <p>
       * This implementation is a no-op.
       */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }
      
      /**
       * Get the map of session's super-procedures.
       */
      public Map<Object, ProcedureData> getSessionSuperProcedures()
      {
         return getSuperProcedures(null);
      }

      /**
       * Get the map of the super-procedures for the given referent.
       * 
       * @param    referent
       *           The external program instance for which the super-procedures
       *           are needed.
       */
      public Map<Object, ProcedureData> getSuperProcedures(Object referent)
      {
         return referent == null ? sessionSuperProcs : pMap.get(referent).superProcedures;
      }
      
      /**
       * Attempt cleanup of the given external program instance. If is no
       * longer used as a persistent procedure, super procedure, or as a 
       * system handle, the referent will be removed from the {@link #pMap}
       * map.
       * 
       * @param    referent
       *           The external program instance for which needs to be cleaned
       *           up.
       */
      public void cleanup(Object referent)
      {
         // if no longer used, remove from global map
         if (!thisProcedures.contains(referent)                   &&
             !targetProcedures.contains(referent)                 &&
             !sourceProcedures.contains(referent)                 &&
             !persistentProcedures.contains(referent)             &&
             !sessionSuperProcs.containsKey(referent)             &&
             !superProcForExtProc(referent, persistentProcedures) &&
             !superProcForExtProc(referent, sessionSuperProcs.keySet()))
         {
            forceCleanup(referent);
         }
      }

      /**
       * Cleanup of the given external program instance. The referent will be
       * removed from the {@link #pMap} map.
       *
       * @param    referent
       *           The external program instance for which needs to be cleaned
       *           up.
       */
      public void forceCleanup(Object referent)
      {
         // remove from global map
         ProcedureData pd = pMap.remove(referent);
         relativeNames.remove(referent);

         ProcedureManager.cleanupSubscriptions(referent);
      }

      /**
       * Check if the given referent is a super-procedure of one of the
       * procedures in the given list.
       * 
       * @param    referent
       *           The external program instance.
       * @param    progs
       *           The list of external programs for which their 
       *           super-procedures list will be searched for the referent.
       *            
       * @return   <code>true</code> if it was found as a super-procedure for
       *           one of the given procedures.
       */
      private boolean superProcForExtProc(Object referent, Collection<Object> progs)
      {
         for (Object prog : progs)
         {
            ProcedureData pp = pMap.get(prog);
            for (Object sreferent : pp.superProcedures.keySet())
            {
               if (sreferent == referent)
               {
                  return true;
               }
            }
         }
         
         return false;
      }
   }

   /**
    * Contains data about a subscription.
    */
   private static abstract class Subscription<T>
   {
      /** The client session which registered this subscription. */
      private final Session session = SessionManager.get().getSession();

      /** The subscriber. */
      protected final T subscriber;
      
      /** The subscribed event. */
      protected final character event = new character();
      
      /** The publisher to match. <code>null</code> represents any source. */
      protected final Object publisher;
      
      /** Flag indicating that this subscription can't be removed. */
      protected boolean readOnly = false;

      /**
       * Build a new subscription with the specified data.
       * 
       * @param    subscriber
       *           The subscriber.
       * @param    event
       *           The event to subscribe.
       * @param    publisher
       *           The event's publisher. May be <code>null</code>.
       */
      public Subscription(T subscriber, character event, Object publisher)
      {
         this.subscriber = subscriber;
         this.publisher = publisher;
         this.event.assign(event);
      }
      
      /**
       * Execute a PUBLISH request, by posting the event to the actual {@link #session context}.
       * 
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      public abstract void publishRemote(String modes, Object[] params);
      
      /**
       * Execute a PUBLISH request, in this context.
       * 
       * @param    publisher
       *           The publisher which posted the PUBLISH event.
       * @param    errors
       *           A list to collect any errors during PUBLISH.
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      public abstract void publish(Object       publisher, 
                                   List<String> errors, 
                                   String       modes, 
                                   Object[]     params);
   }
   
   /**
    * Contains data about a subscription from an {@link ExternalResource}.
    */
   private static class ExternalSubscription
   extends Subscription<ExternalResource>
   {
      /**
       * Build a new subscription with the specified data.
       * 
       * @param    subscriber
       *           The subscriber.
       * @param    event
       *           The event to subscribe.
       * @param    publisher
       *           The event's publisher. May be <code>null</code>.
       */
      public ExternalSubscription(ExternalResource subscriber, character event, Object publisher)
      {
         super(subscriber, event, publisher);
      }

      /**
       * Execute a PUBLISH request, by posting the event to the actual {@link #session context}.
       * 
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      @Override
      public void publishRemote(String modes, Object[] params)
      {
         publishWorker(params, true);
      }
      
      /**
       * Execute a PUBLISH request, in this context.
       * 
       * @param    publisher
       *           The publisher which posted the PUBLISH event.
       * @param    errors
       *           A list to collect any errors during PUBLISH.
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      @Override
      public void publish(Object publisher, List<String> errors, String modes, Object[] params)
      {
         publishWorker(params, false);
      }
      
      /**
       * Worker to perform PUBLISH to a remote side.
       * 
       * @param    params
       *           The PUBLISH parameters.
       * @param    delayed
       *           Flag indicating if the PUBLISH is executed immediately or to be posted to the
       *           client's event queue, when <code>false</code>.
       */
      private void publishWorker(Object[] params, boolean delayed)
      {
         // stringify all params...
         String[] sparams = new String[params.length];
         for (int i = 0; i < params.length; i++)
         {
            Object p = params[i];
            
            if (p instanceof BaseDataType)
            {
               sparams[i] = ((BaseDataType) p).toStringMessage();
            }
            else
            {
               sparams[i] = (p == null ? null : p.toString());
            }
         }

         LogicalTerminal.publish(this.subscriber, this.event.toStringMessage(), sparams, delayed);
      }
   }

   /**
    * Contains data about a legacy subscription made by a certain procedure.
    */
   private static class LegacySubscription
   extends Subscription<Object>
   {
      /** The method to be invoked. May be <code>null</code>. */
      private final character method;
      
      /**
       * Build a new subscription with the specified data.
       * 
       * @param    subscriber
       *           The subscriber procedure.
       * @param    event
       *           The event to subscribe.
       * @param    publisher
       *           The event's publisher. May be <code>null</code>.
       * @param    method
       *           The method to be invoked. When <code>null</code>, the event name will be used
       *           as the method name.
       */
      public LegacySubscription(Object subscriber, character event, Object publisher, character method)
      {
         super(subscriber, event, publisher);
         this.method = new character(method == null ? event : method);
      }

      /**
       * Execute a PUBLISH request, by posting the event to the actual {@link #session context}.
       * 
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      @Override
      public void publishRemote(String modes, Object[] params)
      {
         throw new UnsupportedOperationException(
            "PUBLISH to a remote side running in legacy ABL mode is not yet supported!");
         
         // TODO: publisher can't be used here, as it will have no correspondent on the remote
         // side.  what surrogate to use?
         
         // calling publish() at the ServerEvent seems safe, as will be executed in the 
         // subscriber's context
         
         // TODO: remove undoable flag from BDT vars?
         
         /*
         ServerEvent evt = new ServerEvent(-1,
                                           "PUBLISH", 
                                           () -> publish(null, new ArrayList<>(), modes, params));
         
         LogicalTerminal.postServerEvent(evt);
         */
      }

      /**
       * Execute a PUBLISH request, in this context.
       * 
       * @param    publisher
       *           The publisher which posted the PUBLISH event.
       * @param    errors
       *           A list to collect any errors during PUBLISH.
       * @param    modes
       *           The parameter modes.
       * @param    params
       *           The PUBLISH parameters.
       */
      @Override
      public void publish(Object publisher, List<String> errors, String modes, Object[] params)
      {
         WorkArea wa = work.get();
         
         ProcedureData pd = wa.pMap.get(this.subscriber);
         
         // business logic can unsubscribe during this loop
         if (pd == null)
         {
            return;
         }
         
         String relName = getRelativeName(pd.phandle.get());
         
         // invoke the associated procedure. there are some rules here:
         // - changes of INPUT-OUTPUT parameters are inherited from one call to another
         // - OUTPUT parameters are always overridden
         String iename = (this.method.isUnknown() ? event.toStringMessage() :
                                                   this.method.toStringMessage());
         
         // check if current procedure (or its SUPER) has a definition for target
         handle phandle = ControlFlowOps.resolveProgramForProcedure(pd.phandle, iename);
         
         if (phandle == null)
         {
            final String msg =
               "PUBLISH statement failed. Procedure %s has no entry point for %s";
            
            // postpone error reporting until all the invocations are done,
            // we want to prevent setting of deferred error as this would
            // produce an invalid handle in the call to handle.ctor below
            errors.add(String.format(msg, relName, iename));
            return;
         }
         
         // for the duration of this call, the SOURCE-PROCEDURE must be assumed as the
         // publisher.
         // the received publisher argument may be null, if the PUBLISH originates from a remote
         // side.  in this case, THIS-PROCEDURE will remain unchanged.
         wa.publishSourceProcedure = publisher;
         
         try
         {
            // even if silent error mode is enabled by default, when the procedure is
            // executed it will be disabled for the duration of the execution and re-enabled
            // when the procedure ends. TransactionManager.push/popScope takes care of this.
            ControlFlowOps.invokeInWithMode(iename, pd.phandle, modes, params);
         }
         finally
         {
            // ensure it remains null, in case for any reason the internal proc is not 
            // executed
            wa.publishSourceProcedure = null;
         }
      }
   }
   
   /**
    * A container with various data about the next called program/internal-entry, needed by
    * {@link WorkArea#scopeStart()}.
    */
   private static class CalleeInfoImpl 
   implements CalleeInfo
   {
      /** Flag indicating if the destination is in a super-procedure. */
      private final boolean superCall;
      
      /** The target procedure, which was targeted by the caller. */
      private final Object target;
      
      /** The procedure where the internal entry was resolved. */
      private final Object destination;
      
      /** 
       * The internal entry name. In case of triggers or external programs, use the 
       * {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
       */
      private final String iename;
      
      /** 
       * The external program name; will be non-null only when we are invoking an external 
       * program.
       */
      private final String pname;
      
      /** Flag indicating we are calling a user-defined function. */
      private final boolean function;
      
      /** Flag indicating the external program was ran persistent. */
      private final boolean persistentProc;
      
      /**
       * Create a new instance.
       * 
       * @param    superCall
       *           Flag indicating if the destination is in a super-procedure.
       * @param    target
       *           The target procedure, which was targeted by the caller.
       * @param    destination
       *           The procedure where the internal entry was resolved.
       * @param    iename
       *           The internal entry name. In case of triggers or external programs, use the 
       *           {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
       * @param    pname
       *           The external program name; will be non-null only when we are invoking an external
       *           program.
       * @param    function
       *           Flag indicating we are calling a user-defined function.
       * @param    persistentProc
       *           Flag indicating the external program was ran persistent.
       */
      public CalleeInfoImpl(boolean superCall,
                            Object  target,
                            Object  destination,
                            String  iename,
                            String  pname,
                            boolean function,
                            boolean persistentProc)
      {
         this.superCall      = superCall;
         this.target         = target;
         this.destination    = destination;
         this.iename         = iename;
         this.pname          = pname;
         this.function       = function;
         this.persistentProc = persistentProc;
      }
      
      /**
       * Called by {@link WorkArea#scopeStart()} - depending on the callee's configuration, it
       * populates the procedure system handles and maintains persistent procedures.
       */
      public void push()
      {
         WorkArea wa = work.get();
         
         if (pname != null && !wa.pMap.containsKey(target))
         {
            // this is an external procedure invocation...
            handle htarget = new handle(new ExternalProgramWrapper(target));
            ProcedureManager.addProcedure(wa, htarget, pname, persistentProc);
         }
         
         if (!superCall)
         {
            if (!wa.thisProcedures.isEmpty())
            {
               if (wa.publishSourceProcedure != null)
               {
                  wa.sourceProcedures.push(wa.publishSourceProcedure);
                  wa.sourceProceduresData.push(wa.pMap.get(wa.publishSourceProcedure));
                  wa.publishSourceProcedure = null;
               }
               else
               {
                  Object thisProcedure = wa.thisProcedures.peek();
                  wa.sourceProcedures.push(thisProcedure);
                  wa.sourceProceduresData.push(wa.pMap.get(thisProcedure));
               }
            }
            wa.targetProcedures.push(target);
            wa.targetProceduresData.push(wa.pMap.get(target));
         }
         
         pushInternalEntry(wa, destination, iename, function);
      }
      
      /**
       * Called by {@link WorkArea#scopeFinished()} - depending on the callee's configuration, it
       * backs out the data pushed by {@link #push()}, so the state will be the same as before
       * the callee as invoked.
       */
      public void pop()
      {
         WorkArea wa = work.get();
         
         wa.internalEntryStack.pop();
         
         if (!superCall)
         {
            if (!wa.thisProcedures.isEmpty())
            {
               wa.sourceProcedures.pop();
               wa.sourceProceduresData.pop();
            }
            
            wa.targetProcedures.pop();
            wa.targetProceduresData.pop();
         }
      }
      
      /** 
       * The internal entry name. In case of triggers or external programs, returns the 
       * {@link ProcedureManager#TRIGGER_PROCEDURE} and {@link ProcedureManager#EXTERNAL_PROGRAM} 
       * constants respectively.
       * 
       * @return   The internal entry name.
       */
      @Override
      public String getInternalEntryName()
      {
         return iename;
      }
      
      /**
       * The procedure where the internal entry was resolved.
       * 
       * @return   A valid instance.
       */
      @Override
      public Object getDestination()
      {
         return destination;
      }
      
      /** 
       * The external program name; will be non-null only when we are invoking an external 
       * program.
       * 
       * @return   A {@link String} representing the program name.
       */
      @Override 
      public String getExternalProgramName()
      {
         return pname;
      }
      
      /** 
       * Flag indicating a user-defined function is called. 
       *
       * @return   A boolean flag.
       */
      @Override
      public boolean isFunction()
      {
         return function;
      }
   }
   
   /**
    * Container for various procedure data.
    */
   private static class ProcedureData
   {
      /** The list of super-procedures for this procedure, by external program instance. s*/
      private final Map<Object, ProcedureData> superProcedures = new LinkedHashMap<>();
      
      /** 
       * The map of name of functions defined IN handle in this external program, to their 
       * {@link Resolvable} instance. 
       */
      private final Map<String, Resolvable> inHandleFunctions = new HashMap<>();
      
      /**
       * A handle instance for this procedure, so that other handles can be instantiated from it.
       */
      private handle phandle = null;
      
      /** The current search-mode for this external program. */
      private SearchMode searchMode = null;
      
      /** The private data of this procedure handle. */
      private character privateData = null;
      
      /** Implementation of ADM-DATA attribute. */
      private final character admData = new character();

      /** Implementation of UNIQUE-ID attribute. */
      private final integer uniqueID = new integer();
      
      /** Handle referencing the instantiating procedure. */
      private final handle instProc = new handle();
      
      /**
       * Set of {@link Scopeable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Scopeable> scopeables = new HashSet<>();

      /**
       * Set of {@link Finalizable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Finalizable> finalizables = new HashSet<>();

      /**
       * A stack holding the resources created in each top-level block.
       */
      private Deque<Set<WrappedResource>> staticResources = new ArrayDeque<>();
      
      /**
       * A value of CURRENT-WINDOW attribute of THIS-PROCEDURE handle
       */
      private handle currentWindow = new handle();
      
      /** A set of all created (and still alive) resources, in this procedure (dynamic or not). */
      private Map<HandleResource, Object> createdResources = new WeakHashMap<>();

      /**
       * Create a new {@link ProcedureData} instance using the given 
       * procedure handle and search-mode.
       * 
       * @param    phandle
       *           The procedure handle.
       * @param    searchMode
       *           The search-mode. May be null.
       */
      ProcedureData(handle phandle, SearchMode searchMode)
      {
         this.phandle = new handle(phandle);
         this.searchMode = searchMode;
         this.uniqueID.assign(new integer(UniqueIdGenerator.getUniqueId(IdKind.PROCEDURE)));
      }
      
      /**
       * Check if the search-mode for this external program is SEARCH-SELF.
       * 
       * @return   See above.
       */
      boolean isSearchSelf()
      {
         return searchMode == null || searchMode == SearchMode.SEARCH_SELF;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LEVEL_5}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_5;
      }
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context. Registers the {@link WorkArea} instance with the
       * {@link TransactionManager}, for scope notifications.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   }
}
