/*
** Module   : SessionUtils.java
** Abstract : Defines APIs used to access the SESSION attributes and methods.
**
** Copyright (c) 2013-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CS  20130214 Added support for SESSION: EXPORT, SERVER-OPERATING-MODE, REMOTE attributes.
** 003 CA  20130221 getType was renamed to getResourceType.
** 004 CA  20130222 Added stubs for more SESSION-related attributes and methods: REMOTE,
**                  PARAMETER. The implementations for EXPORT, SERVER-OPERATING-MODE and REMOVE 
**                  attrs call their ServerImpl counterparts.
** 005 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 006 CA  20130529 Added implementations for FIRST-SERVER, LAST-SERVER, REMOTE, EXPORT and
**                  SERVER-OPERATING-MODE.
** 007 OM  20130510 Added support for date-related SESSION attributes so they can be queried from 
**                  client side.
** 008 CA  20130822 Allow access to the DATE-FORMAT attribute (set/getDateOrder) outside of a P2J
**                  server.
** 009 CA  20130927 Static proxies need the unknown() API. Resource type is determined from 
**                  LegacyResource constants. Implemented the following attributes: FIRST-BUFFER,
**                  FIRST-QUERY, FIRST-/LAST-SOCKET, FIRST-/LAST-SERVER-SOCKET.
** 010 CA  20131002 Implemented the PARAMETER attribute.
** 011 CA  20140701 Added PIXELS-PER-ROW and PIXELS-PER-COLUMN.
** 012 CA  20140702 Implemented PIXELS-PER-ROW and PIXELS-PER-COLUMN.
** 013 OM  20141110 Refactored get/setDateOrder() to get/setDateFormat().
** 014 EVL 20141216 Adding SESSION:TOOLTIPS attribute handling, server side part.
** 015 GES 20141228 Completed PARAMETER attribute implementation by storing it in a context-local
**                  and allowing the client command line/bootstrap to override any default value
**                  found in the directory.
** 016 GES 20150227 Added FIRST-CHILD and LAST-CHILD attribute support.
** 017 HC  20150508 Improvements to GUI window scrolling support.
** 018 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
** 019 OM  20150915 Added stub for SESSION:DEBUG-ALERT, DATA-ENTRY-RETURN and APPL-ALERT-BOXES
**                  attributes and method. Fixed set3D method.  Added LOAD-ICON.
** 020 HC  20160222 Fixed error reporting when assigning session read-only attributes.
** 021 IAS 20160302 Support for the APPL-ALERT-BOXES, DATA-ENTRY-RETURN, DEBUG-ALERT and
**                  IMMEDIATE-DISPLAY SESSION attributes.
** 022 IAS 20160328 Client log support.
** 023 OM  20160401 Fixed method name collision for static proxy.
** 024 HC  20170811 Added conversion support for SESSION attributes PRINTER-NAME, PRINTER-PORT,
**                  PRINTER-CONTROL-HANDLE, PRINTER-HDC and the method GET-PRINTERS.
**     HC  20170817 Added conversion support and runtime stubs for SESSION:STARTUP-PARAMETERS and
**                  SESSION:SUPPRESS-WARNINGS.
** 025 HC  20171003 Implemented SESSION:SUPPRESS-WARNINGS runtime support and the ability to
**                  handle (display or output depending on the runtime state) legacy warnings.
** 026 HC  20171024 Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 027 EVL 20180220 Now the current printer name is gettimg from client OS base.
** 028 CA  20180502 Fixed issues with time-related attributes used by SessionExports. 
** 029 CA  20180503 Added MULTITASKING-INTERVAL, SYSTEM-ALERT-BOXES and SERVER-CONNECTION-ID. 
**                  Renamed/moved some SESSION attribute getters/setters to fix issues with using
**                  the SESSION via a handle.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.proxy.*;

/**
 * This class provides implementation for SESSION attributes and methods.
 * Mainly, the APIs implemented here are defined in the {@link CommonSession}
 * interface (or one of its super interfaces) and are not statically 
 * implemented by the classes used by the {@link #asHandle()}.
 */
public class SessionUtils
{
   /** Context local proxy that allows SESSION attributes to be pushed from server to client. */
   private static final ContextLocal<WorkArea> work = new ContextLocal<WorkArea>()
   {
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       * <p>
       * This will obtain a unique instance of <code>SessionExports</code> depending on which
       * side (server or client) it is called the returned value can be a local or a network
       * proxy instance.
       *
       * @return   The newly instantiated work area.
       */
      @Override
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         if (!ErrorManager.isHeadless() && !SessionManager.get().isLeaf())
         {
            wa.exports = (SessionExports) RemoteObject.obtainInstance(SessionExports.class, true);
         }
         
         ClientParameters cp = StandardServer.getClientParameters();
         if (cp != null)
         {
            wa.isDebugAlert = cp.debugAlert;
            wa.isClientLog  = cp.clientlog != null;
            wa.param = cp.param;
         }
         if (wa.param == null)
         {
             Directory dir = DirectoryManager.getInstance();
             wa.param = dir.getString(Directory.ID_RELATIVE, "startup_parameter", "");
            
         }
         
         return wa;
      }
   };
   
   /**
    * Get a the instance for the SESSION system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link CommonSession} interface and its methods implemented by these
    * classes: {@link SessionUtils}, {@link EnvironmentOps}, {@link date},
    * {@link LogicalTerminal}, {@link ProcedureManager}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = work.get();

      Class[] classes = 
      {
         SessionUtils.class,
         EnvironmentOps.class,
         date.class,
         LogicalTerminal.class,
         ProcedureManager.class,
         NumberType.class
      };
      CommonSession proxy = StaticProxy.obtain(CommonSession.class, classes);

      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }

      return new handle(proxy);
   }
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = work.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      // the following attribute names are reported in their shortest forms
      if (attribute.toUpperCase().startsWith("DISPLAY-TY"))
      {
         attribute = "DISPLAY-T";
      }
      else if (attribute.toUpperCase().startsWith("HEIGHT-C"))
      {
         attribute = "HEIGHT";
      }
      else if (attribute.toUpperCase().startsWith("WIDTH-C"))
      {
         attribute = "WIDTH";
      }
      
      // the error message for session read-only attributes is different 
      // from the one emitted by handle.readOnlyError() - settable is spelled with 
      // two 't' and there is no word 'widget' at the end of the message.
      final String msg = "**%s is not a settable attribute for %s";
      String err = String.format(msg,
                                 attribute.toUpperCase(),
                                 asHandle().unwrapSession().getResourceType().toStringMessage());
      ErrorManager.recordOrShowError(4052, err, false, false, false);
   }
   
   /**
    * Indicates if the session is for the context of an appserver.
    * 
    * @return   <code>true</code> if the session is for the context of an appserver.
    */
   public  static logical isRemote()
   {
      return new logical(_isRemote());
   }

   /**
    * Indicates if the session is for the context of an appserver.
    * 
    * @return   <code>true</code> if the session is for the context of an appserver.
    */
   public static boolean _isRemote()
   {
      return AppServerManager.isRemote();
   }
   
   /**
    * Method implementation for getting the SERVER-OPERATING-MODE attribute of the session handle.
    * <p>
    * Possible values:
    * <ul>
    *    <li>State-reset
    *    <li>State-aware
    *    <li>Stateless
    *    <li>State-free
    * </ul>
    * 
    * @return   The corresponding SERVER-OPERATING-MODE attribute of the session handle.
    */
   public static character getServerOperatingMode()
   {
      return AppServerManager.getServerOperatingMode();
   }
   
   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and  name-patterns separated by comma.
    *         
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(String procList)
   {
      return AppServerManager.export(procList);
   }
   
   /**
    * Method that specifies the remote procedures the client can execute from the AppServer 
    * session.
    * 
    * @param    procList
    *           The list of procedure  and  name-patterns separated by comma.
    *         
    * @return   <code>true</code> if the method succeeded.
    */
   public static logical export(character procList)
   {
      return AppServerManager.export(procList);
   }

   /**
    * Get the value of the PARAMETER attribute, which is set to the value of the 4GL 
    * <code>-param</code> command line option specified for the current session.
    * <p>
    * The following is the precedence order (highest to lowest) for obtaining the
    * value of this parameter:
    * <p>
    * <ul>
    *    <li> The client command line.
    *    <li> The client bootstrap configuration.
    *    <li> The directory using relative lookup and the "startup_parameter" node.
    * </ul>
    *
    * @return   The <code>SESSION:PARAMETER</code> value for this context.
    */
   public static character getParameter()
   {
      return new character(work.get().param);
   }

   /**
    * Returns the handle for the first dynamic buffer in the first table containing a dynamic 
    * buffer.
    * 
    * @return   See above.
    */
   public static handle firstBuffer()
   {
      // TODO: this should be OK, but more testing is needed to check if 4GL doesn't sort them in
      // a different order (i.e. using the table name)
      return HandleChain.firstResource(LegacyResource.BUFFER);
   }

   /**
    * Returns the handle for the first window in the session.
    * 
    * @return   The first window in the session's list of windows.
    */
   public static handle firstChild()
   {
      return HandleChain.firstResource(LegacyResource.WINDOW);
   }

   /**
    * Returns the handle for the last window in the session.
    * 
    * @return   The last window in the session's list of windows.
    */
   public static handle lastChild()
   {
      return HandleChain.lastResource(LegacyResource.WINDOW);
   }

   /**
    * Returns the handle for the first created dynamic query.
    * <p>
    * TODO: this should call a worker method from the persist package.
    * 
    * @return   See above.
    */
   public static handle firstQuery()
   {
      return HandleChain.firstResource(LegacyResource.QUERY);
   }
   
   /**
    * Get the FIRST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstServer()
   {
      return HandleChain.firstResource(LegacyResource.SERVER);
   }
   
   /**
    * Get the FIRST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstServerSocket()
   {
      return HandleChain.firstResource(LegacyResource.SERVER_SOCKET);
   }
   
   /**
    * Get the FIRST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle firstSocket()
   {
      return HandleChain.firstResource(LegacyResource.SOCKET);
   }
   
   /**
    * Get the LAST-SERVER attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastServer()
   {
      return HandleChain.lastResource(LegacyResource.SERVER);
   }
   
   /**
    * Get the LAST-SERVER-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastServerSocket()
   {
      return HandleChain.lastResource(LegacyResource.SERVER_SOCKET);
   }
   
   /**
    * Get the LAST-SOCKET attribute of this session.
    * 
    * @return   See above.
    */
   public static handle lastSocket()
   {
      return HandleChain.lastResource(LegacyResource.SOCKET);
   }
   
   /** 
    * Sets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using "M", "D" and "Y" once 
    * each.  The order from index 0 to index 2 represents the left to right ordering of the 
    * components. So the leftmost date component will be defined by the character at index 
    * position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   public static void setDateFormat(String dateFormat)
   {
      setDateFormat(new character(dateFormat));
   }
   
   /** 
    * Sets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using "M", "D" and "Y" once 
    * each.  The order from index 0 to index 2 represents the left to right ordering of the 
    * components. So the leftmost date component will be defined by the character at index 
    * position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   public static void setDateFormat(character dateFormat)
   {
      date.setDateOrder(dateFormat);
      
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setDateFormat(dateFormat);
      }
   }
   
   /** 
    * Gets the order of date components (SESSION:DATE-FORMAT) for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components is a 3 character string using
    * "M", "D" and "Y" once each.  The order from index 0 to index 2
    * represents the left to right ordering of the components. So the leftmost
    * date component will be defined by the character at index position 0.
    * <p>
    * The value accessed is specific to the current user's context.
    *
    * @return   Defines the order of date components for conversion to/from strings.
    */
   public static character getDateFormat()
   {
      return date.getDateOrder();
   }
   
   /**
    * Get the state of SESSION:DEBUG-ALERT attribute. 
    * 
    * @return  the state of SESSION:DEBUG-ALERT attribute.
    */
   public static logical isDebugAlert()
   {
      return new logical(work.get().isDebugAlert);
   }
   
   /**
    * Set the state of SESSION:DEBUG-ALERT attribute.
    * 
    * @param   on
    *          The new state for SESSION:DEBUG-ALERT attribute.
    */
   public static void setDebugAlert(logical on)
   {
      setDebugAlert(on == null || on.isUnknown() ? false : on.booleanValue());
   }
   
   /**
    * Set the state of SESSION:DEBUG-ALERT attribute.
    * 
    * @param   on
    *          The new state for SESSION:DEBUG-ALERT attribute.
    */
   public static void setDebugAlert(boolean on)
   {
      work.get().isDebugAlert = on;
   }
   
   /**
    * Get the state of SESSION:DATA-ENTRY-RETURN attribute. 
    * 
    * @return  the state of SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static logical isDataEntryReturn()
   {
      return LogicalTerminal._isDataEntryReturn();
   }
   
   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static void setDataEntryReturn(logical on)
   {
      if (on == null || on.isUnknown())
      {
         return;
      }
      setDataEntryReturn(on.booleanValue());
   }
   
   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static void setDataEntryReturn(boolean on)
   {
      LogicalTerminal.setDataEntryReturn(on);
   }
   
   /**
    * Set the state of SESSION:LOAD-ICON attribute.
    *
    * @param   iconName
    *          The file name of the icon to be loaded
    */
   public static logical loadIcon(String iconName)
   {
      UnimplementedFeature.missing("LOAD-ICON attribute is not implemented");
      return new logical();
   }
   
   /**
    * Sets the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @param    offset
    *           The new timezone offset which must be between -840 and 840
    *           inclusive.  The value may also be <code>unknown</code> which
    *           is the equivalent of resetting this attribute to the default.
    */
   public static void setSessionTimeZone(NumberType offset)
   {
      date.setTimeZone(offset);

      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionTimeZone(offset);
      }
   }
   
   /**
    * Accesses the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @return   The timezone offset. If this has never been set, then
    *           <code>unknown</code> will be returned.
    */
   public static integer getSessionTimeZone()
   {
      return date.getTimeZone();
   }

   /**
    * Obtains the currently configured TIME-SOURCE attribute for the current
    * user's context.  This attribute is can be read and written but it is
    * not honored for date/time processing in P2J.  The intention of the
    * attribute is to define the system from which time values will be obtained.
    * In P2J, time values are always obtained from the application server. 
    * <p>
    * The default value is "local" which in Progress means that the client
    * machine is the source  (this is the same as the empty string "").  It is
    * common to use a database name in the 4GL, which makes the application
    * use the database server as the time source.
    *
    * @return   The time source attribute.
    */
   public static character getSessionTimeSource()
   {
      return date.getTimeSource();
   }
   
   /**
    * Sets the <code>TIME-SOURCE</code> attribute for the user's context.
    * This attribute is can be read and written but it is not honored for 
    * date/time processing in P2J.  The intention of the attribute is to define
    * the system from which time values will be obtained. In P2J, time values
    * are always obtained from the application server.
    *
    * @param    timeSource
    *           The new value to set. If <code>null</code>, then the empty
    *           string will be used, which in the 4GL has the same affect as
    *           "local".
    */
   public static void setSessionTimeSource(character timeSource)
   {
      date.setTimeSource(timeSource);

      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionTimeSource(timeSource);
      }
   }
   
   /**
    * Obtains the currently configured DISPLAY-TIMEZONE attribute for the current
    * user's context. This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @return   The display-timezone attribute.
    */
   public static integer getSessionDisplayTimeZone()
   {
      return date.getDisplayTimeZone();
   }
   
   /**
    * Sets the <code>DISPLAY-TIMEZONE</code> attribute for the user's context.
    * This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @param    displayTimezone
    *           The new value to set.
    */
   public static void setSessionDisplayTimeZone(NumberType displayTimezone)
   {
      date.setDisplayTimeZone(displayTimezone);
      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setSessionDisplayTimeZone(displayTimezone);
      }
   }
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute getter. Returns the 
    * currently configured start year for the two-digit Y2K windowing range.
    * This is the equivalent to the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @return  The start year for the two-digit windowing range.
    */
   public static integer getYearOffset()
   {
      return date.getYearOffset();
   }
   
   /**
    * Implementation of SESSION:YEAR-OFFSET attribute setter. Sets the 
    * configured start year for the two-digit Y2K windowing range. This is
    * the equivalent of the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date#getWindowingYear}).
    * 
    * @param   year
    *          The new start year for the two-digit windowing range.
    */
   public static void setYearOffset(NumberType year)
   {
      date.setYearOffset(year);

      WorkArea wa = work.get();
      
      if (wa.exports != null)
      {
         wa.exports.setYearOffset(year);
      }
   }
                                                                                                                          
   /**
    * Get the PIXELS-PER-ROW attribute of this session.
    * 
    * @return   See above.
    */
   public static integer getPixelsPerRow()
   {
      return new integer(LogicalTerminal.getClient().getPixelsPerRow());
   }
   
   /**
    * Get the PIXELS-PER-COLUMN attribute of this session.
    * 
    * @return   See above.
    */
   public static integer getPixelsPerColumn()
   {
      return new integer(LogicalTerminal.getClient().getPixelsPerColumn());
   }

   /**
    * Sets the SESSION:TOOLTIPS attribute for this session.
    *
    * @param    tooltips
    *           The new value for session related TOOLTIPS attribute.
    */
   public static void setSessionTooltips(logical tooltips)
   {
      WorkArea wa = work.get();

      // change the local copy
      if (tooltips != null && !tooltips.isUnknown())
      {
         wa.tooltips = tooltips.booleanValue();
      }
      else
      {
         wa.tooltips = false;
      }

      // and push new value to the client
      LogicalTerminal.getClient().setSessionTooltips(wa.tooltips);
   }
   
   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   <code>true</code> if tooltips are available for this session
    *           <code>false</code> otherwise.
    */
   public static logical getSessionTooltips()
   {
      return new logical(work.get().tooltips);  
   }
   
   /**
    * Get THREE-D attribute value for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static logical is3D()
   {
      return LogicalTerminal.is3D();
   }
   
   /**
    * Set THREE-D attribute value for SESSION handle.
    * 
    * @param   value
    *          The value to set.
    */
   public static void set3D(logical value)
   {
      LogicalTerminal.set3D(value);
   }
   
   /**
    * Check if client log is active
    * 
    * @return <code>true</code> if client log is active
    */
   public static boolean isClientlog()
   {
      return work.get().isClientLog;
   }

   /**
    * Get the state of SESSION:PRINTER-NAME attribute.
    *
    * @return  the state of SESSION:PRINTER-NAME attribute.
    */
   public static character getPrinterName()
   {
      return EnvironmentOps.getCurrentPrinterName();
   }

   /**
    * Set the state of SESSION:PRINTER-NAME attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-NAME attribute.
    */
   public static void setPrinterName(character value)
   {
      // currently only the built-in PDF printer is supported, so this method is a no-op
   }

   /**
    * Get the state of SESSION:PRINTER-PORT attribute.
    *
    * @return  the state of SESSION:PRINTER-PORT attribute.
    */
   public static character getPrinterPort()
   {
      // currently only the built-in PDF printer is supported, so this method always returns an unknown
      return new character();
   }

   /**
    * Set the state of SESSION:PRINTER-PORT attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-PORT attribute.
    */
   public static void setPrinterPort(character value)
   {
      readOnlyError("PRINTER-PORT");
   }

   /**
    * Get the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    *
    * @return  the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    */
   public static integer getPrinterControlHandle()
   {
      return new integer(PrintingService.getControlHandle());
   }

   /**
    * Set the state of SESSION:PRINTER-CONTROL-HANLDE attribute.
    *
    * @param   value
    *          The new state for SESSION:PRINTER-CONTROL-HANLDE attribute.
    */
   public static void setPrinterControlHandle(integer value)
   {
      PrintingService.resetControlHandle();
   }

   /**
    * Get the state of SESSION:PRINTER-HDC attribute.
    *
    * @return  the state of SESSION:PRINTER-HDC attribute.
    */
   public static integer getPrinterHDC()
   {
      return new integer(PrintingService.getHDC());
   }

   /**
    * The SESSION:GET-PRINTERS method.
    *
    * @return  the result of SESSION:GET-PRINTERS method.
    */
   public static character getPrinters()
   {
      // currently only the built-in PDF printer is supported
      return new character(PrintingService.PDF_PRINTER_DESCR);
   }

   /**
    * Get the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @return  the state of SESSION:SUPPRESS-WARNINGS attribute.
    */
   public static logical isSuppressWarnings()
   {
      return new logical(ErrorManager.isSuppressWarnings());
   }

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   public static void setSuppressWarnings(logical value)
   {
      if (value.isUnknown())
      {
         // UNKNOWN is silently ignored
         return;
      }

      setSuppressWarnings(value.booleanValue());
   }

   /**
    * Set the state of SESSION:SUPPRESS-WARNINGS attribute.
    *
    * @param   value
    *          The new state for SESSION:SUPPRESS-WARNINGS attribute.
    */
   @LegacyAttribute(name = "SUPPRESS-WARNINGS", setter = true)
   public static void setSuppressWarnings(boolean value)
   {
      ErrorManager.setSuppressWarnings(value);
   }

   /**
    * Get the state of SESSION:STARTUP-PARAMETERS attribute.
    *
    * @return  the state of SESSION:STARTUP-PARAMETERS attribute.
    */
   public static character getStartupParameters()
   {
      return new character(LegacyCLI.getInstance().getStartupParameters());
   }

   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static void setSystemAlertBoxes(boolean on)
   {
      ErrorManager.setSystemAlertBoxes(on);
   }
   
   /**
    * Set the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    * 
    * @param   on
    *          The new state for SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static void setSystemAlertBoxes(logical on)
   {
      setSystemAlertBoxes(on == null || on.isUnknown() ? false : on.booleanValue());
   }

   /**
    * Get the state of SESSION:SYSTEM-ALERT-BOXES attribute. 
    * 
    * @return  the state of SESSION:SYSTEM-ALERT-BOXES attribute.
    */
   public static logical isSystemAlertBoxes()
   {
      return new logical(ErrorManager.isSystemAlertBoxes());
   }
   
   /**
    * Set the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @param   interval 
    *          The new value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   public static void setMultitaskingInterval(NumberType interval)
   {
      work.get().multitaskingInterval = interval.isUnknown() ? 0 : interval.longValue();
   }
   
   /**
    * Get the value of SESSION:MULTITASKING-INTERVAL attribute. 
    * 
    * @return  the value of SESSION:MULTITASKING-INTERVAL attribute.
    */
   public static integer getMultitaskingInterval()
   {
      return new integer(work.get().multitaskingInterval);
   }
   
   /**
    * Get the value of SESSION:SERVER-CONNECTION-ID attribute. 
    * 
    * @return  the value of SESSION:SERVER-CONNECTION-ID attribute.
    */
   public static character getServerConnectionId()
   {
      String id = AppServerManager.getServerConnectionId();
      return new character(id == null ? "" : id);
   }

   /**
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** The resource's ID. */
      private Long id = null;

      /** Remote client interface. */
      private SessionExports exports = null;
      
      /** Command line <code>-param</code> value (SESSION:PARAM attribute). */
      private String param = null;
      
      /** The state of SESSION:DEBUG-ALERT attribute */
      private boolean isDebugAlert = false;

      /** The flag indicating whether the widgets have the tooltip feature. */
      private boolean tooltips = Utils.getDirectoryNodeBoolean(null, "session-tooltips", true);
      
      /** The flag indicating whether the clieet log is active */ 
      private boolean isClientLog = false;
      
      /** This is just a holder for SESSION:MULTITASKING-INTERVAL - no behavior in FWD. */
      private long multitaskingInterval = 0;
   }
}
