/*
** Module : StreamBuilder.java
** Public : interface for remote construction of streams
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060211  @24552 Created initial version providing remote construction of streams.
** 002 CS  20130103         Implement OS-DIR support as an INPUT stream.
** 003 VIG 20130918         Added printer stream support (OUTPUT TO PRINTER).
** 004 GES 20140612         Added explicit terminal stream support.
** 005 HC  20171024         Runtime support for SYSTEM-DIALOG PRINT-SETUP and OUTPUT TO PRINTER.
** 006 SBI 20180213         Added openClipboardStream().
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * An interface providing remote construction of file and process streams.
 */
public interface StreamBuilder
{
   /**
    * Constructs a remote stream instance representing a file using a filename, this file or
    * device will be opened for reading or writing based on the given <code>write</code> flag.
    * The current read/write position will be at offset 0 unless in <code>write</code> mode and 
    * the <code>append</code> flag is on (in which case the write position will be at the end of
    * the file).
    * <p>
    * This stream will NOT have paging active.
    *
    * @param    filename
    *           The name of the file or device to open.  Must not be <code>null</code>.
    * @param    write
    *           If <code>true</code>, open in write mode, otherwise open the file for reading.
    * @param    append
    *           In <code>write</code> mode, if <code>true</code>, the current file pointer is
    *           set to the end of the file. Otherwise if in <code>write</code> mode, the file
    *           is truncated to 0 length. This value is ignored in <code>read</code> mode.
    *           
    * @return   The stream id which is used to route operations to the newly created instance.
    *
    * @throws   ErrorConditionException
    *           If the file does not exist OR if the current user does not have the proper rights
    *           to open the file or device.
    */
   public int openFileStream(String filename, boolean write, boolean append)
   throws ErrorConditionException;
      
   /**
    * Constructs a remote stream instance representing a process' standard I/O.
    *
    * @return   The stream id which is used to route operations to the newly created instance.
    */
   public int openProcessStream();

   /**
    * Constructs the output stream into the client's system clipboard.
    *
    * @return   The stream id which is used to route operations to the newly created instance.
    */
   public int openClipboardStream();

   /**
    * Constructs a remote stream instance representing the OsDirStream result of the INPUT 
    * THROUGH OS-DIR (dir-name) statement.
    * <p>
    * This will take the directory and export the list of files, directory and device names
    * and properties.
    * 
    * @param    dir
    *           The <code>String</code> path to the directory on which OS-DIR will be applied.
    * @param    noAttrList
    *           <code>true</code> if property indexes will be added to the results.
    *        
    * @return   The stream id which is used to route operations to the newly created instance.
    */
   public int openDirStream(String dir, boolean noAttrList);

   /**
    * Constructs a remote stream instance representing the terminal.
    *
    * @return   The stream id which is used to route operations to the newly created instance.
    */
   public int openTerminalStream();
   
   /**
    * Constructs a remote printer stream instance.  This can be assigned as unnamed output
    * (OUTPUT TO PRINTER statement) or as a named stream (OUTPUT STREAM someStream TO PRINTER).
    * 
    * @param    name
    *           The name of the target printer.
    * @param    options
    *           Print options.
    * 
    * @return   The stream id which is used to route operations to the newly created instance.
    */
   public int openPrinterStream(String name, PrintOptions options);
}