/*
** Module   : PojoWebSocketHandler.java
** Abstract : create and register WebSockets POJO objects.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20131105 First version based on jetty 9.1
** 002 SBI 20160220 Added websocket configuration parameters.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.web;

import java.io.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.*;
import org.eclipse.jetty.websocket.server.*;
import org.eclipse.jetty.websocket.servlet.*;

/**
 * This class is used to register and handle WebSockets POJO objects. 
 * <p>
 * Methods to create WebSockets:
 * <ul>
 *    <li>POJO WebSocket using annotations.</li>
 *    <li>Implements WebSocketListener interface.</li>
 *    <li>Extends WebSocketAdapter class.</li>
 * </ul>
 */
public class PojoWebSocketHandler 
extends WebSocketHandler
{
   /** WebSocket class */
   private Class<?> webSocketPojo;
   
   /** WebSocket target should be unique on the server context */ 
   private String webSocketTarget;
   
   /** WebSocketCreator used to create WebSockets by hand */
   private WebSocketCreator creator;
   
   /** Websocket configuration */
   private WebSocketConfig config;
   
   /**
    * Constructor.
    *
    * @param   webSocketTarget
    *          The WebSocket target string.
    *          This value should be unique on the WebSocket collection
    *          of handlers for the same server context.
    * @param   creator
    *          The WebSocketCreator instance used tom create 
    *          WebSockets by hand.
    * @param   config
    *          The websocket configuration parameters.
    */
   public PojoWebSocketHandler(String webSocketTarget, WebSocketCreator creator, WebSocketConfig config)
   {
      this.webSocketTarget = webSocketTarget;
      this.creator = creator;
      this.config  = config;
   }
   
   /**
    * Constructor.
    * 
    * @param   webSocketTarget
    *          The WebSocket target string.
    *          This value should be unique on the WebSocket collection
    *          of handlers for the same server context.
    * @param   webSocketPojo
    *          The WebSocket POJO class.
    * @param   config
    *          The websocket configuration parameters.
    */
   public PojoWebSocketHandler(String webSocketTarget, Class<?> webSocketPojo, WebSocketConfig config)
   {
      this.webSocketTarget = webSocketTarget;
      this.webSocketPojo = webSocketPojo;
      this.config = config;
   }

   /**
    * Register a WebSocket class POJO with the 
    * default {@link WebSocketCreator} or using a custom
    * creator.
    *
    * @param  wsServletFactory
    *         Basic WebSocketServletFactory for working with 
    *         Jetty-based WebSocketServlet
    */
   @Override
   public void configure(WebSocketServletFactory wsServletFactory)
   {
      if (config != null)
      {
         wsServletFactory.getPolicy().setIdleTimeout(config.getMaxIdleTime());
         wsServletFactory.getPolicy().setMaxBinaryMessageSize(config.getMaxBinaryMessage());
         wsServletFactory.getPolicy().setMaxTextMessageBufferSize(config.getMaxTextMessage());
         // used only by streams 
         wsServletFactory.getPolicy().setMaxBinaryMessageBufferSize(config.getMaxBinaryMessage());
      }
      if (creator == null)
      {
         wsServletFactory.register(webSocketPojo);
      }
      else
      {
         wsServletFactory.setCreator(creator);
      }
   }

   /**
    * Handle WebSocket's upgrade requests. If the request is not handled and the target is
    * equal to this target handler, then set handled to <code>true</code>
    *
    * @param    target 
    *           The target of the request - either a URI or a name.
    * @param    baseRequest 
    *           The original unwrapped request object.
    * @param    request 
    *           The request either as the {@link Request} object or a wrapper of that request. 
    * @param    response
    *           The response as the {@link Response} object or a wrapper of that request. 
    *
    * @throws IOException
    * @throws ServletException
    */
   @Override
   public void handle(String              target, 
                      Request             baseRequest,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      if (!baseRequest.isHandled() && webSocketTarget.equals(target))
      {
         super.handle(target, baseRequest, request, response);
         baseRequest.setHandled(true);
      }
   }
}