/*
** Module   : XEntity
** Abstract : DOM and SAX common interface.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130204 Created initial version.
** 002 EVL 20130214 Adding the base interface XCommon to place the XML related common features.  
** 003 CA  20130221 Added UNIQUE-ID and NAMESPACE-URI support. 
** 004 EVL 20130222 Adding SUBTYPE attribute and INSERT-BEFORE method support.
** 005 EVK 20130527 Adding support CHILD-NUM, NAMESPACE-PREFIX attribute and REPLACE-CHILD
**                  method.
** 006 EVL 20131002 Adding legacy attributes and methods annotation.
** 007 VIG 20141213 setSubtype methods are moved to SubTypeAttribute interface.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.xml;

import com.goldencode.p2j.util.*;

/**
 * Interface describing the common features for DOM and SAX XML support objects.
 */
public interface XEntity
extends XCommon,
        UniqueID,
        NamespaceURI,
        SubTypeAttribute
{
   /**
    * Gets the the number of children objects for the current one.
    * 
    * @return   The integer number of the children.
    */
   @LegacyAttribute(name = "NUM-CHILDREN")
   public integer getNumChildren();

   /**
    * Gets the number of child for the current one. Only for XNodRef
    *
    * @return   The integer number of the child, 1-based index of the child node.
    */
   @LegacyAttribute(name = "CHILD-NUM")
   public integer getChildNum();

   /**
    * Inserts a node as a chaild of this document before another node.
    *
    * @param    handle1
    *           The handle representing the child to insert.
    * @param    handle2
    *           The handle representing the XML node to be inserted before.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @LegacyMethod(name = "INSERT-BEFORE")
   public logical insertBefore(handle handle1, handle handle2);

   /**
    * Appending the new child to the current object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be appended.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @LegacyMethod(name = "APPEND-CHILD")
   public logical appendChild(handle hChild);

  /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put
    *           the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @LegacyMethod(name = "GET-CHILD")
   public logical getChild(handle hChild, integer index);
   
   /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put
    *           the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @LegacyMethod(name = "GET-CHILD")
   public logical getChild(handle hChild, int index);
   
   /**
    * Removing the specified child from the node.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be removed.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @LegacyMethod(name = "REMOVE-CHILD")
   public logical removeChild(handle hChild);

   /**
    * Removing the specified child from the node.
    *
    * @param    newChildHandle
    *           The handle class instance that represents the node to insert in the tree.
    * @param    oldChildHandle
    *           The handle class instance that represents the node to remove from the tree.
    *
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @LegacyMethod(name = "REPLACE-CHILD")
   public logical replaceChild(handle newChildHandle, handle oldChildHandle);

   /**
    * Getting the Document the current reference belongs to.
    *
    * @return   The handle of the Document containing tne current reference..
    */
   @LegacyAttribute(name = "OWNER-DOCUMENT")
   public handle getOwnerDocument();

   /**
    * Returns the value of the the NAMESPACE-PREFIX attribute. Only for ELEMENT or ATTRIBUTE,
    * for other attributes return unknown value.
    *
    * @return    See above.
    */
   @LegacyAttribute(name = "NAMESPACE-PREFIX")
   public character getNamespacePrefix();

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @LegacyAttribute(name = "NAMESPACE-PREFIX", setter = true)
   public void setNamespacePrefix(String prefix);

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @LegacyAttribute(name = "NAMESPACE-PREFIX", setter = true)
   public void setNamespacePrefix(character prefix);
}