/*
** Module   : ClassInfo.java
** Abstract : Type information specific to a class.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20101119 Created initial version.
** 002 ECF 20150309 Upgraded to ASM 5.0.3.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

import org.objectweb.asm.*;
import com.goldencode.asm.*;

/**
 * Type information for a class (as opposed to a primitive data type).  This
 * includes access to the class itself, and convenience methods to return the
 * JVM-internal name of the class, as well as its descriptor.  Wrap and unwrap
 * method information is not provided, as this is only meaningful for
 * primitive data types.  Width, as well as load and return instructions are
 * hard-coded.
 * 
 * @author ECF
 */
class ClassInfo
implements TypeInfo,
           Opcodes
{
   /** Class which represents data type */
   private final Class<?> type;
   
   /** Internal class name used with CHECKCAST instruction */
   private final String castName;
   
   /** Internal data type descriptor */
   private final String descriptor;
   
   /**
    * Constructor.
    * 
    * @param   type
    *          Class which represents data type.
    */
   protected ClassInfo(Class<?> type)
   {
      this.type = type;
      this.castName = AsmUtils.commonToInternalTypeName(type);
      if (type.isArray())
      {
         this.descriptor = castName;
      }
      else
      {
         StringBuilder buf = new StringBuilder("L");
         buf.append(castName);
         buf.append(";");
         this.descriptor = buf.toString();
      }
   }
   
   /**
    * Get the class object associated with this data type.
    * 
    * @return  Class which represents data type.
    */
   @Override
   public Class<?> getType()
   {
      return type;
   }
   
   /**
    * Get the internal name of the class which should be used with a CHECKCAST
    * instruction when casting an <code>Object</code> returned from a call to
    * <code>InvocationHandler.invoke</code> to an instance of this data type.
    * 
    * @return  Internal class name for CHECKCAST instruction, or
    *          <code>null</code> for the <code>void</code> return type.
    */
   @Override
   public String getCastName()
   {
      return castName;
   }
   
   /**
    * Get the descriptor which is used internally within the JVM to represent
    * this data type.
    * 
    * @return  Type descriptor.
    */
   @Override
   public String getDescriptor()
   {
      return descriptor;
   }
   
   /**
    * Get information for the method, if any, used to wrap instances of this
    * data type in a J2SE data wrapper object.  Note that this only applies to
    * primitive data types.
    * 
    * @return  <code>null</code>.
    */
   @Override
   public MethodInfo getWrapInfo()
   {
      return null;
   }
   
   /**
    * Get information for the method, if any, used to unwrap instances of this
    * data type from a J2SE data wrapper object.  Note that this only applies
    * to primitive data types.
    * 
    * @return  <code>null</code>.
    */
   @Override
   public MethodInfo getUnwrapInfo()
   {
      return null;
   }
   
   /**
    * Get the width, in stack or local variable slots for instances of this
    * data type, when compiling bytecode instructions which manipulate such
    * data.
    * 
    * @return  Width of this data type in slots (always 1).
    */
   @Override
   public int getWidth()
   {
      return 1;
   }
   
   /**
    * Get the ASM constant representing the instruction used to load an
    * instance of this data type from a local variable and push it onto the
    * operand stack (e.g., ALOAD, ILOAD, etc.).
    * 
    * @return  Load instruction for this data type (always ALOAD).
    */
   @Override
   public int getLoadInsn()
   {
      return ALOAD;
   }
   
   /**
    * Get the ASM constant representing the instruction used to pop an
    * instance of this data type from the operand stack and return it from a
    * method (e.g., ARETURN, IRETURN, etc.).
    * 
    * @return  Return instruction for this data type (always ARETURN).
    */
   @Override
   public int getReturnInsn()
   {
      return ARETURN;
   }
}