/*
** Module   : PrimitiveInfo.java
** Abstract : Type information specific to a primitive data type.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20101119 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

/**
 * Type information specific to a primitive data type.  This differs from the
 * type information for a class in that primitives each have a corresponding
 * object wrapper type (e.g., <code>boolean</code> has
 * <code>java/lang/Boolean</code>).  Information about the wrapper type is
 * recorded in this class to allow auto-boxing code to be assembled, which
 * enables, for instance:
 * <ul>
 * <li>a primitive to be wrapped before being passed as an argument to
 *     <code>InvocationHandler.invoke</code>;
 * <li>the <code>Object</code> return value from
 *     <code>InvocationHandler.invoke</code> to be cast into the appropriate
 *     object wrapper type, and then unwrapped to the corresponding primitive
 *     value.
 * </ul>
 * 
 * @author ECF
 */
final class PrimitiveInfo
extends ClassInfo
{
   /** Internal class name (of wrapper) used with CHECKCAST instruction */
   private final String castName;
   
   /** Internal data type descriptor */
   private final String descriptor;
   
   /** Information for wrapper method */
   private final MethodInfo wrapInfo;
   
   /** Information for unwrapper method */
   private final MethodInfo unwrapInfo;
   
   /** Width of data type in stack or local variable slots */
   private final int width;
   
   /** Load instruction */
   private final int loadInsn;
   
   /** Return instruction */
   private final int returnInsn;
   
   /**
    * Constructor.
    * 
    * @param   type
    *          Class representing primitive data type.
    * @param   wrapper
    *          Class representing J2SE object wrapper data type.
    * @param   descriptor
    *          Internal data type descriptor.
    * @param   width
    *          Width of data type in stack or local variable slots.
    * @param   loadInsn
    *          Load instruction.
    * @param   returnInsn
    *          Return instruction.
    */
   PrimitiveInfo(Class<?> type,
                 Class<?> wrapper,
                 String descriptor,
                 int width,
                 int loadInsn,
                 int returnInsn)
   {
      super(type);
      
      this.descriptor = descriptor;
      this.width = width;
      this.loadInsn = loadInsn;
      this.returnInsn = returnInsn;
      
      if (wrapper != null)
      {
         TypeInfo clsInfo = new ClassInfo(wrapper);
         castName = clsInfo.getCastName();
         String typeName = type.getName();
         wrapInfo = makeWrapInfo(clsInfo, descriptor);
         unwrapInfo = makeUnwrapInfo(typeName, clsInfo, descriptor);
      }
      else
      {
         castName = null;
         wrapInfo = null;
         unwrapInfo = null;
      }
   }
   
   /**
    * Generate the information for the data wrapper method, given the wrapper
    * type information and the primitive type descriptor.
    * 
    * @param   wrapperInfo
    *          Type information for object wrapper.
    * @param   typeDescriptor
    *          Internal type descriptor of primitive data type.
    * 
    * @return  Method information for the data wrapper method.
    */
   private static MethodInfo makeWrapInfo(TypeInfo wrapperInfo,
                                          String typeDescriptor)
   {
      StringBuilder buf = new StringBuilder("(");
      buf.append(typeDescriptor);
      buf.append(")");
      buf.append(wrapperInfo.getDescriptor());
      String methodDescriptor = buf.toString();
      
      MethodInfo info = new MethodInfo(wrapperInfo.getCastName(),
                                       "valueOf",
                                       methodDescriptor);
      
      return info;
   }
   
   /**
    * Generate the information for the data unwrapper method, given the
    * primitive type name, wrapper type information, and primitive type
    * descriptor.
    * 
    * @param   typeName
    *          External name of primitive data type (e.g., "boolean").
    * @param   wrapperInfo
    *          Type information for object wrapper.
    * @param   typeDescriptor
    *          Internal type descriptor of primitive data type.
    * 
    * @return  Method information for the data unwrapper method.
    */
   private static MethodInfo makeUnwrapInfo(String typeName,
                                            TypeInfo wrapperInfo,
                                            String typeDescriptor)
   {
      StringBuilder buf = new StringBuilder(typeName);
      buf.append("Value");
      String methodName = buf.toString();
      
      buf.setLength(0);
      buf.append("()");
      buf.append(typeDescriptor);
      String methodDescriptor = buf.toString();
      
      MethodInfo info = new MethodInfo(wrapperInfo.getCastName(),
                                       methodName,
                                       methodDescriptor);
      
      return info;
   }
   
   /**
    * Get the descriptor which is used internally within the JVM to represent
    * this data type.
    * 
    * @return  Type descriptor.
    */
   @Override
   public String getDescriptor()
   {
      return descriptor;
   }
   
   /**
    * Get the internal name of the class which should be used with a CHECKCAST
    * instruction when casting an <code>Object</code> returned from a call to
    * <code>InvocationHandler.invoke</code> to an instance of this data type.
    * <p>
    * For primitive data types, this is the internal name of the wrapper data
    * type (e.g., "java/lang/Boolean").
    * 
    * @return  Internal class name for CHECKCAST instruction, or
    *          <code>null</code> for the <code>void</code> return type.
    */
   @Override
   public String getCastName()
   {
      return castName;
   }
   
   /**
    * Get information for the method, if any, used to wrap instances of this
    * data type in a J2SE data wrapper object.  Note that this only applies to
    * primitive data types.
    * 
    * @return  Method information for the wrapper method invoked to wrap an
    *          instance of this data type in an object, or <code>null</code>
    *          for the <code>void</code> data type.
    */
   @Override
   public MethodInfo getWrapInfo()
   {
      return wrapInfo;
   }
   
   /**
    * Get information for the method, if any, used to unwrap instances of this
    * data type from a J2SE data wrapper object.  Note that this only applies
    * to primitive data types.
    * 
    * @return  Method information for the unwrapper method invoked to unwrap
    *          an instance of this data type from an object, or
    *          <code>null</code> for the <code>void</code> data type.
    */
   @Override
   public MethodInfo getUnwrapInfo()
   {
      return unwrapInfo;
   }
   
   /**
    * Get the width, in stack or local variable slots for instances of this
    * data type, when compiling bytecode instructions which manipulate such
    * data.  This will be 2 for <code>long</code> and <code>double</code>
    * types, and 1 for all others.
    * 
    * @return  Width of this data type in slots.
    */
   @Override
   public int getWidth()
   {
      return width;
   }
   
   /**
    * Get the ASM constant representing the instruction used to load an
    * instance of this data type from a local variable and push it onto the
    * operand stack (e.g., DLOAD, ILOAD, etc.).
    * 
    * @return  Load instruction for this data type.
    */
   @Override
   public int getLoadInsn()
   {
      return loadInsn;
   }
   
   /**
    * Get the ASM constant representing the instruction used to pop an
    * instance of this data type from the operand stack and return it from a
    * method (e.g., DRETURN, IRETURN, etc.).
    * <p>
    * Note:  the RETURN instruction (used for return type <code>void</code>)
    * does not pop an operand off the stack;  it simply returns.
    * 
    * @return  Return instruction for this data type.
    */
   @Override
   public int getReturnInsn()
   {
      return returnInsn;
   }
}