/*
** Module   : ProxyFactory.java
** Abstract : Utility class to create and manage dynamic proxy classes.
**
** Copyright (c) 2010-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20101115 Created initial version.
** 002 ECF 20101214 Added variants of getProxy methods which do not accept
**                  InvocationHandler instances. These assume the proxy's
**                  superclass implement InvocationHandler.
** 003 ECF 20110101 Fixed CacheKey implementation to deal correctly with array
**                  hash codes. Fixed class-loading logic to use correct
**                  loader when cached class name is found.
** 004 ECF 20130120 Added isProxyClass method to test whether a class was
**                  generated by this factory; fix to enable nested interfaces
**                  to be used to define the proxy.
** 005 ECF 20130124 Changed naming convention, so that generated proxy class
**                  has the same package as its superclass.
** 006 ECF 20150309 Moved inner class ProxyClassLoader to top-level class AsmClassLoader in
**                  com.goldencode.asm package.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 008 ECF 20160219 Added support for an optional plugin to override the assembly of arbitrary
**                  methods.
** 009 ECF 20160827 Cache proxy classes directly, rather than their names, to avoid expensive
**                  calls to Class.forName(). If we ever allow these proxy classes to be unloaded
**                  and garbage collected, the weak references used for the cache's values will
**                  require a ReferenceQueue implementation, or we will need some external API to
**                  clean the cache explicitly.
** 010 ECF 20180103 Use AsmClassLoader.getInstance to prevent creation of a second instance of the
**                  class loader.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

import java.io.*;
import java.lang.ref.*;
import java.lang.reflect.*;
import java.util.*;
import com.goldencode.asm.*;

/**
 * A factory for dynamic proxy objects, which defines new classes using the
 * ASM bytecode engineering library (http://asm.ow2.org).  This class manages
 * a cache of proxy classes and implements the loading of these dynamically
 * generated classes using a custom class loader.  However, the work of
 * actually assembling the proxy classes is delegated to the {@link
 * ProxyAssembler}.  Please see the description of that class for the
 * rationale of this mechanism as an alternative to Java's standard, dynamic
 * proxy implementation.
 * <p>
 * The access points to this factory's functionality are the public, static
 * variants of the <code>getProxy</code> method, which allow the definition
 * of a new proxy class, based on a given superclass, list of interfaces to
 * implement, and an invocation handler to which method invocations are
 * dispatched.  Several additional options are available to tailor the proxy
 * class to specific needs.  Please see:
 * <ul>
 * <li>{@link #getProxy(Class, Class[], InvocationHandler)}
 * <li>{@link #getProxy(Class, Class[], boolean, Method[], ProxyAssemblerPlugin,
 *     InvocationHandler)}
 * <li>{@link #getProxy(Class, Class[])}
 * <li>{@link #getProxy(Class, Class[], boolean, Method[])}
 * </ul>
 * 
 * <strong>Please note that generics and annotations are not supported at this
 * time.</strong>
 * 
 * @author ECF
 */
public final class ProxyFactory
{
   /** Singleton instance of this factory class */
   private static final ProxyFactory singleton = new ProxyFactory();
   
   /** Cache of proxy classes which already have been assembled */
   private final Map<CacheKey, WeakReference<Class<?>>> cache = new HashMap<>();
   
   /** Custom class loader to load proxy classes */
   private final AsmClassLoader loader = AsmClassLoader.getInstance();
   
   /** Suffix to ensure unique proxy class names */
   private int nextSuffix = 0;
   
   /**
    * Private constructor to ensure singleton instance.
    */
   private ProxyFactory()
   {
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * the same names, compatible signatures and return types, but different
    * lists of exception types, the first method discovered which has a return
    * type assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  Therefore, callers should take
    * care to specify methods with matching exception lists, if methods with 
    * the same name, signature, and compatible return types are specified
    * across interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   handler
    *          Handler to which to dispatch method invocations.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<?> parent,
                                Class<?>[] interfaces,
                                InvocationHandler handler)
   {
      return (T) getProxy(parent, interfaces, true, null, null, handler);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * @param   handler
    *          Handler to which to dispatch method invocations.  This value
    *          must not be <code>null</code>.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if <code>handler</code> is <code>null</code>;
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<?> parent,
                                Class<?>[] interfaces,
                                boolean declaredOnly,
                                Method[] doNotProxy,
                                ProxyAssemblerPlugin plugin,
                                InvocationHandler handler)
   {
      if (handler == null)
      {
         throw new IllegalArgumentException(
            "Invocation handler must not be null");
      }
      
      return (T) singleton.getProxyImpl(parent,
                                        interfaces,
                                        declaredOnly,
                                        doNotProxy,
                                        plugin,
                                        handler);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  This method variant does not
    * accept an <code>InvocationHandler</code> instance.  It should be used
    * when the parent class is expected to fulfill the role of invocation
    * handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class, which must implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<? extends InvocationHandler> parent, Class<?>[] interfaces)
   {
      return (T) getProxy(parent, interfaces, true, null);
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  This method variant does not
    * accept an <code>InvocationHandler</code> instance.  It should be used
    * when the parent class is expected to fulfill the role of invocation
    * handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class, which must implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   @SuppressWarnings("unchecked")
   public static <T> T getProxy(Class<? extends InvocationHandler> parent,
                                Class<?>[] interfaces,
                                boolean declaredOnly,
                                Method[] doNotProxy)
   {
      return (T) singleton.getProxyImpl(parent,
                                        interfaces,
                                        declaredOnly,
                                        doNotProxy,
                                        null,
                                        null);
   }
   
   /**
    * Indicate whether the given class represents a proxy class generated by
    * this factory.
    * 
    * @param   clazz
    *          The class to be tested.
    * 
    * @return  <code>true</code> if the given class represents a proxy class
    *          generated by this factory, else <code>false</code>.
    */
   public static boolean isProxyClass(Class<?> clazz)
   {
      return singleton.loader.equals(clazz.getClassLoader());
   }
   
   /**
    * Get an instance of a proxy class which subclasses the given parent class
    * and implements the specified interfaces.  Invocations of the methods in
    * those interfaces will be dispatched to the specified invocation handler.
    * <p>
    * If a proxy class does not yet exist which meets the given criteria, one
    * is assembled and loaded on the fly.
    * <p>
    * <b>Implementation Note:</b> if multiple interfaces specify methods with
    * compatible signatures and return types, but different lists of exception
    * types, the first method discovered which has a return type which is
    * assignable from all other, matching methods is used for the dynamic
    * proxy, regardless of the exception list.  The order in which methods are
    * checked is indeterminate.  Therefore, callers should be careful to
    * specify methods with matching exception lists, if methods with the same
    * name, signature, and compatible return types are specified across
    * interfaces.
    * 
    * @param   <T>
    *          Type of the proxy class which will be returned.
    * @param   parent
    *          Superclass of the proxy class.  If <code>handler</code> is
    *          <code>null</code>, this class is expected to implement the
    *          <code>InvocationHandler</code> interface.
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * @param   plugin
    *          An optional plugin which can override the assembly of arbitrary methods.
    * @param   handler
    *          Handler to which to dispatch method invocations.  If
    *          <code>null</code>, the dynamic proxy instance itself will be
    *          used as the invocation handler.
    * 
    * @return  Instance of a dynamically generated proxy class which is a
    *          subclass of <code>parent</code> and which implements the
    *          methods defined by <code>interfaces</code>.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types;
    *          if <code>parent</code> is an interface;
    *          if any of the specified <code>interfaces</code> are not
    *          actually interfaces.
    */
   @SuppressWarnings("unchecked")
   private <T> T getProxyImpl(Class<?> parent,
                              Class<?>[] interfaces,
                              boolean declaredOnly,
                              Method[] doNotProxy,
                              ProxyAssemblerPlugin plugin,
                              InvocationHandler handler)
   {
      if (parent.isInterface())
      {
         throw new IllegalArgumentException("Proxy parent must be a concrete class");
      }
      
      for (Class<?> next : interfaces)
      {
         if (!next.isInterface())
         {
            throw new IllegalArgumentException(
               next.getName() +
               " is not an interface; it cannot be used to define a proxy");
         }
      }
      
      Class<?> clazz = null;
      T proxy = null;
      boolean proxyIsHandler = (handler == null);
      
      if (doNotProxy == null)
      {
         doNotProxy = new Method[0];
      }
      CacheKey key = new CacheKey(parent, interfaces, declaredOnly, doNotProxy);
      
      synchronized (this)
      {
         WeakReference<Class<?>> ref = cache.get(key);
         if (ref != null)
         {
            clazz = ref.get();
            if (clazz == null)
            {
               // class was unloaded, so remove it from the cache
               cache.remove(key);
            }
         }
         
         try
         {
            if (clazz == null)
            {
               String parentName = parent.getName();
               String proxyName = makeProxyClassName(parent, parentName);
               Method[] methods = gatherMethods(interfaces, declaredOnly, doNotProxy);
               
               // assemble code for the proxy class
               ProxyAssembler asm = new ProxyAssembler(proxyName,
                                                       parentName,
                                                       interfaces,
                                                       methods,
                                                       plugin,
                                                       proxyIsHandler);
               byte[] code = asm.assembleClass();
               
               /*
               // debug code
               FileOutputStream fos;
               try
               {
                  int pos = proxyName.lastIndexOf(".");
                  String fname = (pos >= 0
                                  ? proxyName.substring(pos + 1)
                                  : proxyName);
                  System.out.println("Proxy class file written to: " + fname);
                  fos = new FileOutputStream(fname);
                  fos.write(code);
                  fos.close();
               }
               catch (IOException exc)
               {
                  exc.printStackTrace();
               }
               */
               
               // load the proxy class with a custom class loader
               clazz = loader.loadClass(proxyName, code);
               
               // call the proxy class' __initializeMethods method, to assign
               // the array of Method objects instances will pass to their
               // invocation handlers
               Method initMethsMethod =
                  clazz.getDeclaredMethod(ProxyAssembler.INIT_METHODS_METHOD, Method[].class);
               initMethsMethod.invoke(clazz, (Object) asm.getProxiedMethods());
               
               // remember that we've assembled a proxy for the current set of criteria;
               // store it with a weak reference, so we don't pin it in memory if it otherwise
               // should be unloaded and garbage collected
               // TODO: currently, proxies created here are never unloaded; when this feature
               // is added, set up a ReferenceQueue to remove unloaded proxy classes from this
               // cache to prevent a memory leak
               cache.put(key, new WeakReference<>(clazz));
            }
            
            if (proxyIsHandler)
            {
               // use default constructor
               proxy = (T) clazz.newInstance();
            }
            else
            {
               // use constructor which stores separate handler
               Constructor<?> ctor = clazz.getDeclaredConstructor(InvocationHandler.class);
               proxy = (T) ctor.newInstance(handler);
            }
         }
         catch (ClassNotFoundException    |
                NoSuchMethodException     |
                InvocationTargetException |
                IllegalAccessException    |
                InstantiationException exc)
         {
            InternalError error = new InternalError(exc.getMessage());
            error.initCause(exc);
            
            throw error;
         }
      }
      
      return proxy;
   }
   
   /**
    * Given the proxy's superclass and its name, generate a unique name of the
    * proxy class, which can be used to load the class in the custom, proxy
    * class loader.  This name differs from the internal name of the proxy
    * class, in that it uses the dot (<code>.</code>) character as its
    * separator, whereas the internal name uses the forward slash
    * (<code>/</code>).
    * <p>
    * The package of the superclass becomes the package of the proxy class.
    * The simple name of the proxy class is <code>$__Proxy<i>N</i></code>,
    * where <i>N</i> is an integer unique across all proxy classes generated
    * by this factory.
    * 
    * @param   parent
    *          Parent class of proxy.
    * @param   parentName
    *          Fully qualified name of parent class.
    * 
    * @return  Unique name for the proxy class, qualified by the
    *          <code>com.goldencode.proxy</code> package name.
    */
   private String makeProxyClassName(Class<?> parent, String parentName)
   {
      String pkgName = parent.getPackage().getName();
      StringBuilder buf = new StringBuilder();
      if (pkgName != null)
      {
         buf.append(pkgName);
         buf.append(".");
      }
      buf.append("$__Proxy");
      buf.append(nextSuffix++);
      
      return buf.toString();
   }
   
   /**
    * Gather the final list of methods to be implemented explicitly by the
    * dynamic proxy class.  The list is determined by iterating through the
    * methods declared by the given interfaces (and optionally, by their
    * ancestor interfaces), omitting those explicitly identified as not to be
    * proxied.  In the event the same method name and parameter signatures are
    * found across interfaces, any potential conflicts are resolved by using
    * only that method (if any) whose return type is assignable from all
    * other, potentially conflicting methods.  In the event any such conflict
    * cannot be resolved, an exception is thrown.
    * 
    * @param   interfaces
    *          Array of interfaces which must be implemented by the proxy
    *          class.
    * @param   declaredOnly
    *          If <code>true</code>, only methods explicitly declared by the
    *          given interfaces are used;  if <code>false</code>, methods
    *          declared by the given interfaces and by any of their ancestor
    *          interfaces are used.
    * @param   doNotProxy
    *          An optional array of <code>Method</code> objects which are not
    *          to be implemented by the dynamically generated proxy.  This
    *          option should be used only to represent methods which already
    *          are implemented by the proxy class' superclass.
    * 
    * @return  The final array of methods to be implemented explicitly by the
    *          dynamic proxy class.
    * 
    * @throws  IllegalArgumentException
    *          if multiple interfaces specify methods which share a name and
    *          parameter signature, but which have incompatible return types.
    */
   private Method[] gatherMethods(Class<?>[] interfaces,
                                  boolean declaredOnly,
                                  Method[] doNotProxy)
   {
      // create a set of methods which are to be omitted from proxy class
      Set<Method> omit = new HashSet<Method>();
      for (Method method : doNotProxy)
      {
         omit.add(method);
      }
      
      // map signatures (w/o return type) to sets of unique methods
      Map<String, Set<Method>> sig2Methods = new LinkedHashMap<String, Set<Method>>();
      
      // list of methods which survive screening process
      List<Method> survivors = new ArrayList<Method>();
      
      // collect list of methods declared by array of interfaces
      for (Class<?> iface : interfaces)
      {
         Method[] toBeProxied = (declaredOnly
                                 ? iface.getDeclaredMethods()
                                 : iface.getMethods());
         for (Method method : toBeProxied)
         {
            if (omit.contains(method))
            {
               continue;
            }
            
            StringBuilder buf = new StringBuilder(method.getName());
            buf.append(ProxyUtils.makeMethodSignatureDescriptor(method));
            String sig = buf.toString();
            Set<Method> methods = sig2Methods.get(sig);
            if (methods == null)
            {
               methods = new LinkedHashSet<Method>();
               sig2Methods.put(sig, methods);
            }
            methods.add(method);
         }
      }
      
      // walk through all groups of methods;  collect survivors
      for (Map.Entry<String, Set<Method>> entry : sig2Methods.entrySet())
      {
         Set<Method> methods = entry.getValue();
         Method survivor = identifyBestMethod(methods);
         survivors.add(survivor);
      }
      
      Method[] methods = survivors.toArray(new Method[survivors.size()]);
      
      return methods;
   }
   
   /**
    * Given a set of methods, identify the best method to implement in the
    * dynamic proxy.  If the set contains only a single method, it is used.
    * In the event the set contains multiple methods, the method whose return
    * type is assignable from all others in the set is chosen.  If no such
    * method exists, an exception is thrown.
    * 
    * @param   methods
    *          A set of methods which share a name and parameter signature,
    *          but not necessarily the same return type.
    * 
    * @return  The best method to implement in the dynamic proxy, according to
    *          the above criteria.
    * 
    * @throws  IllegalArgumentException
    *          if multiple methods exist in the set, but none has a return
    *          type that is assignable from all the others.
    */
   private Method identifyBestMethod(Set<Method> methods)
   {
      if (methods.size() == 1)
      {
         return methods.iterator().next();
      }
      
      outerLoop:
      for (Iterator<Method> outer = methods.iterator(); outer.hasNext(); )
      {
         Method mOuter = outer.next();
         Class<?> rtOuter = mOuter.getReturnType();
         
         for (Iterator<Method> inner = methods.iterator(); inner.hasNext(); )
         {
            Method mInner = inner.next();
            Class<?> rtInner = mInner.getReturnType();
            if (!rtOuter.isAssignableFrom(rtInner))
            {
               continue outerLoop;
            }
         }
         
         return mOuter;
      }
      
      throw new IllegalArgumentException(
         "Methods have same signature, but incompatible return types: " +
         methods);
   }
   
   /**
    * A cache key which takes into account a proxy class' parent, implemented
    * interfaces, whether only declared methods are implemented, and list of
    * methods to be omitted from the proxy.
    * <p>
    * If all the above criteria match between two keys, the keys are
    * equivalent;  else they are not.
    * <p>
    * Instances of this class are used in a cache, to ensure redundant proxy
    * classes are not assembled and loaded.
    * 
    * TODO: Since these cache keys rely in part on the proxies' parent classes
    * and interfaces, their presence in the cache will prevent these parent
    * classes and interfaces from being unloaded and garbage collected.
    * This constitutes a potential memory leak, which should be resolved.
    */
   private static class CacheKey
   {
      /** Parent class of proxy class */
      private final Class<?> parent;
      
      /** Interfaces implemented by proxy class */
      private final Class<?>[] interfaces;
      
      /** Flag indicating only declared methods of interfaces are proxied */
      private final boolean declaredOnly;
      
      /** Methods which are not to be proxied */
      private final Method[] doNotProxy;
      
      /**
       * Constructor.
       * 
       * @param   parent
       *          Parent class of proxy class.
       * @param   interfaces
       *          Interfaces implemented by proxy class.
       * @param   declaredOnly
       *          Flag indicating only declared methods of interfaces are
       *          proxied.
       * @param   doNotProxy
       *          Methods which are not to be proxied.  May be an empty array,
       *          but must not be <code>null</code>.
       */
      CacheKey(Class<?> parent,
               Class<?>[] interfaces,
               boolean declaredOnly,
               Method[] doNotProxy)
      {
         this.parent = parent;
         this.interfaces = interfaces;
         this.declaredOnly = declaredOnly;
         this.doNotProxy = doNotProxy;
      }
      
      /**
       * Return a well-distributed hash code which is consistent with this
       * class' implementation of {@link #equals(Object)}.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + parent.hashCode();
         result = 37 * result + Arrays.hashCode(interfaces);
         result = 37 * result + (declaredOnly
                                 ? Boolean.TRUE.hashCode()
                                 : Boolean.FALSE.hashCode());
         result = 37 * result + Arrays.hashCode(doNotProxy);
         
         return result;
      }
      
      /**
       * Determine whether this object is equivalent to the given object.
       * <p>
       * Two cache keys are considered equivalent if their various fields are
       * equivalent.
       * 
       * @param   o
       *          Another object to test for equivalence with this one.
       * 
       * @return  <code>true</code> if <code>o</code> is another instance of
       *          this class with equivalent fields, else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof CacheKey))
         {
            return false;
         }
         
         if (this == o)
         {
            return true;
         }
         
         CacheKey that = (CacheKey) o;
         
         return (this.parent.equals(that.parent) &&
                 Arrays.equals(this.interfaces, that.interfaces) &&
                 this.declaredOnly == that.declaredOnly &&
                 Arrays.equals(this.doNotProxy, that.doNotProxy));
      }
   }
}