/*
** Module   : ProxyUtils.java
** Abstract : Static utility helper methods.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20101119 Created initial version.
** 002 ECF 20150309 Refactored some methods to com.goldencode.asm.AsmUtils; upgraded to ASM 5.0.3.
** 003 ECF 20160219 Made methods public. Minor performance improvement.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.proxy;

import java.lang.reflect.*;
import java.util.*;
import org.objectweb.asm.*;
import com.goldencode.asm.*;

/**
 * A collection of helper methods for use in assembling and managing dynamic
 * proxy classes.
 * <p>
 * These methods are static;  the class is not meant to be instantiated.
 * 
 * @author ECF
 */
public final class ProxyUtils
implements Opcodes
{
   /** Information for primitive types, mapped by type name */
   private static final Map<Class<?>, TypeInfo> primitives = new HashMap<>();
   
   static
   {
      // populate the map of type name to TypeInfo objects for all primitive types
      TypeInfo[] info =
      {
         new PrimitiveInfo(boolean.class, Boolean.class  , "Z", 1, ILOAD, IRETURN),
         new PrimitiveInfo(byte.class   , Byte.class     , "B", 1, ILOAD, IRETURN),
         new PrimitiveInfo(char.class   , Character.class, "C", 1, ILOAD, IRETURN),
         new PrimitiveInfo(double.class , Double.class   , "D", 2, DLOAD, DRETURN),
         new PrimitiveInfo(float.class  , Float.class    , "F", 1, FLOAD, FRETURN),
         new PrimitiveInfo(int.class    , Integer.class  , "I", 1, ILOAD, IRETURN),
         new PrimitiveInfo(long.class   , Long.class     , "J", 2, LLOAD, LRETURN),
         new PrimitiveInfo(short.class  , Short.class    , "S", 1, ILOAD, IRETURN),
         new PrimitiveInfo(void.class   , null           , "V", 0, NOP  , RETURN ),
      };
      
      int len = info.length;
      for (int i = 0; i < len; i++)
      {
         primitives.put(info[i].getType(), info[i]);
      }
   }
   
   /**
    * Private constructor;  this class is never instantiated.
    */
   private ProxyUtils()
   {
   }
   
   /**
    * Get a {@link TypeInfo} object specific to the given class.  If the class
    * represents a primitive type, a cached object is returned;  otherwise, a
    * new object is instantiated.
    * 
    * @param   type
    *          Class for which type information is required.
    * 
    * @return  Type information for the specified class.
    */
   public static TypeInfo getTypeInfo(Class<?> type)
   {
      if (type.isPrimitive())
      {
         return primitives.get(type);
      }
      
      return (new ClassInfo(type));
   }
   
   /**
    * Compose the parameter signature portion of a method descriptor.  This
    * consists of an opening parenthesis, type descriptor list, and closing
    * parenthesis.  The return type descriptor is not included.
    * 
    * @param   method
    *          Method for which a parameter signature descriptor is required.
    * 
    * @return  Parameter signature portion of the method descriptor.
    */
   public static String makeMethodSignatureDescriptor(Method method)
   {
      StringBuilder buf = new StringBuilder("(");
      
      for (Class<?> parmType : method.getParameterTypes())
      {
         TypeInfo info = getTypeInfo(parmType);
         String typeDesc = info.getDescriptor();
         buf.append(typeDesc);
      }
      
      buf.append(")");
      
      String desc = buf.toString();
      
      return desc;
   }
   
   /**
    * Compose a method descriptor for the given method.  This consists of an
    * opening parenthesis, type descriptor list, and closing parenthesis,
    * followed by the return type descriptor.
    * 
    * @param   method
    *          Method for which a descriptor is required.
    * 
    * @return  Method descriptor.
    */
   public static String makeMethodDescriptor(Method method)
   {
      String desc = makeMethodSignatureDescriptor(method);
      
      StringBuilder buf = new StringBuilder(desc);
      
      Class<?> returnType = method.getReturnType();
      TypeInfo info = getTypeInfo(returnType);
      String typeDesc = info.getDescriptor();
      buf.append(typeDesc);
      
      desc = buf.toString();
      
      return desc;
   }
   
   /**
    * Get the list of caught exception types (i.e., not including those
    * assignable from <code>RuntimeException</code>) defined for the given
    * method.
    * 
    * @param   method
    *          Method whose exception types should be gathered.
    * 
    * @return  Array of fully qualified exception type names, for caught
    *          exceptions only.
    */
   public static String[] getCaughtExceptionTypeNames(Method method)
   {
      Class<?>[] excTypes = method.getExceptionTypes();
      int size = excTypes.length;
      
      if (size == 0)
      {
         return new String[0];
      }
      
      List<String> names = new ArrayList<>(size);
      
      for (int i = 0; i < size; i++)
      {
         Class<?> exc = excTypes[i];
         
         // omit uncaught (runtime) exception types
         if (RuntimeException.class.isAssignableFrom(exc))
         {
            continue;
         }
         
         names.add(AsmUtils.commonToInternalTypeName(excTypes[i]));
      }
      
      return names.toArray(new String[names.size()]);
   }
}