/*
** Module   : WalkListener.java
** Abstract : listener for tree walk events which are translated into callbacks
**            to user-defined tree walkers
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090601   @42541 Listener for tree walk events which are translated
**                           into callbacks to user-defined tree walkers.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl;

import com.goldencode.ast.*;

/**
 * Listener for tree walk events which are translated into callbacks to
 * user-defined tree walkers. A new instance of this class needs to be
 * created for each tree because some of the contained state is specific
 * to a single tree.
 */
public class WalkListener
implements AstWalkListener
{
   /** State shared by all callbacks (across multiple different trees). */
   private final BaseState base;
   
   /** Provides basic services to assist tree walkers (specific to the tree). */
   private final Resolver resolver;
   
   /** Walker to notify of events. */
   private final TreeWalker walker;
   
   /**
    * Construct an instance with the given state which is common for a single
    * tree walk.
    *
    * @param    base
    *           State shared by all callbacks (across multiple different trees).
    * @param    resolver
    *           Provides basic services to assist tree walkers (specific to the
    *           tree).
    * @param    walker
    *           Walker to notify of events.
    */
   public WalkListener(BaseState base, Resolver resolver, TreeWalker walker)
   {
      this.base     = base;
      this.resolver = resolver;
      this.walker   = walker;
   }
   
   /**
    * Called whenever a transition from a parent node to its first child
    * takes place during an AST walk. This will be called <em>before</em> the
    * next AST node in the tree traversal is visited.
    *
    * @param   ast
    *          Source AST node which represents the parent in the parent to
    *          child transition of the current AST walk.
    */
   public void descent(Aast ast)
   {
      WalkState state = prepareState(ast);
      walker.descent(state);
   }
   
   /**
    * Called whenever a transition from a child node to its parent takes
    * place during an AST walk. This will be called <em>before</em> the
    * next AST node in the tree traversal is visited.
    *
    * @param   ast
    *          Source AST node which represents the parent in the child to
    *          parent transition of the current AST walk.
    */
   public void ascent(Aast ast)
   {
      WalkState state = prepareState(ast);
      walker.ascent(state);
   }
   
   /**
    * Called whenever a transition from a child node to its next right
    * sibling takes place during an AST walk. This will be called
    * <em>before</em> the next AST node in the tree traversal is visited.
    * The parent node of the children is in scope at the time of
    * notification. This effectively mimics recursion by notifying the
    * listener when one child node has been processed, before the next is
    * visited. This allows the listener to perform parent-level processing
    * between visits to child nodes.
    * <p>
    * This method is <em>not</em> invoked after the last child is visited,
    * as this is the purpose of the {@link #ascent} method.
    *
    * @param   ast
    *          Source AST node which represents the parent of the children
    *          between which the lateral transition is taking place in the
    *          current AST walk.
    * @param   index
    *          The 0-based index of the child which is about to be processed.
    */
   public void nextChild(Aast ast, int index)
   {
      WalkState state = prepareState(ast);
      walker.nextChild(state, index);
   }
   
   /**
    * Setup the resolver and walk state for an event callback for a specific
    * node of the tree.
    *
    * @param    here
    *           This is the target node for the read-only "source" tree.
    *
    * @return   The walk state. 
    */
   private WalkState prepareState(Aast here)
   {
      Aast copy = resolver.getAst(here.getId());
      return new WalkState(base, resolver, here, copy);
   }   
}