/*
** Module   : LogHelper.java
** Abstract : static utility methods to assist logging 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20090212 Static utility methods to assist logging.
** 002 HC  20150702 Implemented support for call-stack logging.
** 003 HC  20150913 Enhancements call-stack logging.
** 004 HC  20160729 Introduced trace method family.
** 005 HC  20161016 Fixed abend in formatTrace().
** 006 HC  20180419 Improvements to the trace methods. Every trace now also
**                  outputs the method name, file and location.
** 007 HC  20180509 Trace method disambiguation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.text.*;

/**
 * Static utility methods to assist logging.
 */
public class LogHelper
{
   /** Platform-specific line separator. */
   private static final String SEP = System.getProperty("line.separator");

   /** System.nanoTime() at program startup. */
   private static final long startNanoTime = System.nanoTime();

   /**
    * Instances of this class cannot be created.
    */
   private LogHelper()
   {
   }
   
   /**
    * Dump all details of the throwable into a string. Includes exception type,
    * message, complete stack trace, and chained throwables, if any.
    *
    * @param    t
    *           The throwable to dump.
    * 
    * @return   String containing throwable details.
    */
   public static String dumpThrowable(Throwable t)
   {
      StringBuilder sb = new StringBuilder();
      dumpThrowable(t, sb);
      
      return sb.toString();
   }

   /**
    * Dump all details of the throwable into the given buffer. Includes
    * exception type, message, complete stack trace, and chained throwables,
    * if any.
    *
    * @param    t
    *           The throwable to dump.
    * @param    buf
    *           Buffer into which details are written.
    *
    * @return   The given buffer that was modified. This allows chaining.
    */
   public static StringBuilder dumpThrowable(Throwable t, StringBuilder buf)
   {
      Throwable cause = t;
      
      do
      {
         if (cause != t)
         {
            buf.append("Caused by: ");
         }
         
         buf.append(cause.getClass().getName());
         
         String msg = cause.getMessage();
         
         if (msg != null && msg.trim().length() > 0)
         {
            buf.append(": ");
            buf.append(msg);
         }
         buf.append(SEP);
         
         StackTraceElement[] trace = cause.getStackTrace();
         
         for (int i = 0; i < trace.length; i++)
         {
            buf.append("        at ");
            buf.append(trace[i]);
            buf.append(SEP);
         }
         
         cause = cause.getCause();
      }
      while (cause != null);
      
      return buf;
   }
   
   /**
    * Takes the result of {@link #getStackTrace()} and sends it to the
    * standard output.
    */
   public static void dumpStackTrace()
   {
      System.err.println(getStackTraceImpl(-1, 3));
   }
   
   /**
    * Takes the result of {@link #getStackTrace()} and sends it to the
    * standard output.
    * 
    * @param    entries
    *           Number of stack entries to dump.
    */
   public static void dumpStackTrace(int entries)
   {
      System.err.println(getStackTraceImpl(entries, 3));
   }
   
   /**
    * Takes the result of {@link #getStackTrace()} and sends it to the
    * standard error output stream together with the value of the only 
    * argument.
    * 
    * @param    message
    *           A valid string or <code>null</code>.
    */
   public static void dumpStackTrace(String message)
   {
      if (message == null)
         message = "<null>";
      
      System.err.println(message);
      System.err.println(getStackTraceImpl(-1, 3));
   }
   
   /**
    * Takes the result of {@link #getStackTrace()} and sends it to the
    * standard error output stream together with the value of the only 
    * argument.
    * 
    * @param    message
    *           A valid string or <code>null</code>.
    * @param    entries
    *           Number of stack entries to dump.
    */
   public static void dumpStackTrace(String message, int entries)
   {
      if (message == null)
         message = "<null>";
      
      System.err.println(message);
      System.err.println(getStackTraceImpl(entries, 3));
   }
   
   /**
    * Returns a string representing the current call stack. The call stack
    * entries are separated with new-lines.
    * Note that the call stack will exclude this method.
    * 
    * @param    entries
    *           Number of stack entries to dump.
    *           
    * @return   See above.
    */
   public static String getStackTrace(int entries)
   {
      return getStackTraceImpl(entries, 3);
   }
   
   /**
    * Returns a string representing the current call stack. The call stack
    * entries are separated with new-lines.
    * Note that the call stack will exclude this method.
    * 
    * @return   See above.
    */
   public static String getStackTrace()
   {
      return getStackTraceImpl(-1, 3);
   }
   
   /**
    * Returns a string representing the current call stack. The call stack
    * entries are separated with new-lines.
    * 
    * @param    entries
    *           Number of stack entries to return, use negative number to get all.
    * @param    startAtIndex
    *           The index which will become the top of the resulting stack.
    *
    * @return   See above.
    */
   private static String getStackTraceImpl(int entries, int startAtIndex)
   {
      StringBuilder result = new StringBuilder();
      StackTraceElement[] stack = Thread.currentThread().getStackTrace();

      if (entries < 0)
      {
         entries = stack.length;
      }

      entries =  Math.min(stack.length - startAtIndex, entries);
      int lastIndex = entries + startAtIndex;

      for (int i = startAtIndex; i < lastIndex; i++)
      {
         result.append(stack[i]);
         result.append(SEP);
      }
      
      return result.toString();
   }

   /**
    * The method outputs an empty trace message together with nano time and method location to the
    * system error stream.
    */
   public static void trace()
   {
      System.err.println(formatTrace(null, null));
   }

   /**
    * The method formats <code>msg</code> and <code>args</code> using
    * {@link String#format(String, Object...)} and outputs the result together with nano
    * time to the system error stream.
    *
    * @param    msg
    *           The message to output.
    * @param    args
    *           The arguments to format the message with.
    */
   public static void trace(Object msg, Object... args)
   {
      System.err.println(formatTrace(msg, args));
   }

   /**
    * The method formats <code>msg</code> and <code>args</code> using
    * {@link String#format(String, Object...)} and outputs the result together with nano
    * time to the system error stream. The method also outputs the current call stack with the
    * <code>callstackEntries</code> number of entries.
    *
    * @param    msg
    *           The message to output.
    * @param    args
    *           The arguments to format the message with.
    */
   public static void traces(int callstackEntries, Object msg, Object... args)
   {
      System.err.println(formatTrace(msg, args));

      if (callstackEntries > 0)
      {
         dumpStackTrace(callstackEntries);
      }
   }

   /**
    * Formats the input message and arguments for trace output.
    *
    * @param    msg
    *           The message to output.
    * @param    args
    *           The arguments to format the message with.
    *
    * @return   The formatted message.
    */
   private static String formatTrace(Object msg, Object... args)
   {
      DecimalFormat fmt = new DecimalFormat("#,###");
      String method = getStackTraceImpl(1, 4);

      if (msg != null)
      {
         if (args != null && args.length > 0)
         {
            msg = String.format(msg instanceof String ? (String) msg : msg.toString(), args);
         }
      }
      else
      {
         msg = "<null>";
      }

      return fmt.format(System.nanoTime() - startNanoTime) + "ns [" + method + "] " + msg;
   }
}