/*
** Module   : comauto_win.c
** Abstract : Windows specific implementation for COM automation functions.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20171024 Created initial version.
** 002 EVL 20171031 Fix for OLE DATE to handle properly.  Adding RAW data type support, extent
**                  based arrays handling implementation.  Different access modes for method call
**                  parameters support.  Adding Indexed properties support, garbage collector for
**                  internal objects created during life time.  Adding array support for property
**                  getter.
** 003 EVL 20180220 Fix for wrong method type interpretation in certain cases.  The methods with
**                  no parameters can be cosidered as property get while requesting indexed
**                  property can be considered as calling method.  Adding ability to log errors
**                  into dedicated file to be visible for headless environment as well.
** 004 EVL 20180302 Changed registry compact code to minimize pointers copy.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include <windows.h>
#include <objbase.h>
#include <math.h>
#include "comauto.h"

// internal structures to keep active COM object's info
#define MAX_HANDLES_INFO_KEEP 4096

// error log file stream
#define COM_OLE_ERRORS_LOG_FILE "com_ole_errors.log"
FILE * comerr;

// this info will be stored loclly for every COM object in use
typedef struct
{
   jlong id;
   jlong * parent_ids;
   int parentsSize;
   jchar name[40]; // Unicode name 39 chars in max
} ComHandleInfo;
typedef ComHandleInfo * PComHandleInfo;

// local process heap handle
HANDLE hProcHeap = NULL;
// process heap usage mode, no serialization for now
DWORD heapUseMode = HEAP_NO_SERIALIZE;

// local cache to keep COM objects info
// this is an array of pointers for memory to allocate
PComHandleInfo phInfo[MAX_HANDLES_INFO_KEEP];
// current COM handle registry size
int hRegSize = 0;

// define local cache for BST strings allocated (1024 max)
#define MAX_BSTR_TO_KEEP 1024
int bstrStorageSize = 0;
BSTR bstrStorage[MAX_BSTR_TO_KEEP];

// define local cache for SAFEARRAY objectss allocated (1024 max)
#define MAX_SAFEARRAY_TO_KEEP 1024
int saStorageSize = 0;
SAFEARRAY * saStorage[MAX_SAFEARRAY_TO_KEEP];

// define local cache for variables passed by pointer
#define MAX_LOCAL_VAR_IN_BYTES 4096
int nextVarStorageStart = 0;
jbyte varByPointerStorage[MAX_LOCAL_VAR_IN_BYTES];

// DataType enum values mapping (must be in sync with 'p2j/comauto/ComParameter' enum definitioin)
#define AS_DT_NOT_DEFINED  -1
#define AS_DT_SHORT         0
#define AS_DT_FLOAT         1
#define AS_DT_CURRENCY      2
#define AS_DT_UNSIGNED_BYTE 3
#define AS_DT_ERROR_CODE    4
#define AS_DT_IUNKNOWN      5

// ------ base data types constants ------
// The data types constant below were derived from string based data type received from Java.
// It is the uppercases single byte per char strings with more than 3 characters, so minimum
// amount of bytes consumed by string is 4 and this allows us to interpret the first 4 bytes as
// unique integer value. And this in turn allows to replace slow ctrcmp() functions with fast
// and simple integer values comparisons.
// Modify the values only if the data type string on Java side changed or become lowercased.
// The example of computing integer constant: "INT64" has 'I'(0x49), 'N'(0x4E), 'T'(0x54),
// '6'(0x36) chars inside. Place them in integer with reverse order.
//
#define DT_INTEGER   0x45544e49  // instead of strcmp(chFwdDataType, "INTEGER") == 0
#define DT_DECIMAL   0x49434544  // instead of strcmp(chFwdDataType, "DECIMAL") == 0
#define DT_LOGICAL   0x49474F4C  // instead of strcmp(chFwdDataType, "LOGICAL") == 0
#define DT_CHARACTER 0x52414843  // instead of strcmp(chFwdDataType, "CHARACTER") == 0
#define DT_LONGCHAR  0x474E4F4C  // instead of strcmp(chFwdDataType, "LONGCHAR") == 0
#define DT_DATETTZ   0x45544144  // instead of (strcmp(chFwdDataType, "DATE") == 0 ||
                                 //             strcmp(chFwdDataType, "DATETIME") == 0 ||
                                 //             strcmp(chFwdDataType, "DATETIME-TZ") == 0)
#define DT_INT64     0x36544E49  // instead of strcmp(chFwdDataType, "INT64") == 0
#define DT_COMHANDLE 0x484D4F43  // instead of strcmp(chFwdDataType, "COMHANDLE") == 0
#define DT_RAW       0x00574152  // instead of strcmp(chFwdDataType, "RAW") == 0
// ------ base data types constants ------

// functions used in this module only
/**
 * Allocate new BSTR string and keep it into local storage.
 * 
 * @param    olechar
 *           The java string to use as source.
 *           
 * @return   New BSTR string or NULL in case of error. When BSTR variable no more required call
 *           FreeLocalBSTR() to free BSTR consumed memory storage. It has limitation to keep
 *           data if not to free on time.
 */
BSTR LocalAllocBSTR(jchar * olechar)
{
   // string to return
   BSTR bstrNew = NULL;
   
   // check if we have enough room for new string
   if (bstrStorageSize < MAX_BSTR_TO_KEEP && olechar != NULL)
   {
      // null character pointer will mean empty string
      bstrNew = SysAllocString((OLECHAR*)olechar);
      if (bstrNew != NULL)
      {
         bstrStorage[bstrStorageSize] = bstrNew;
         bstrStorageSize++;
      }
   }

   return bstrNew; 
}

/**
 * Clean up local BSTR string storage in order reversed to allocation.
 */
void FreeLocalBSTR()
{
   // nothing to clear
   if (bstrStorageSize == 0)
   {
      return;
   }
   
   int i;
   for (i = bstrStorageSize - 1; i >= 0; i--)
   {
      SysFreeString(bstrStorage[i]);
      bstrStorage[i] = NULL;
   }
   
   // and clean up the size counter
   bstrStorageSize = 0;
}

/**
 * Makes a copy of the given variable into local storage for variables passing by pointer.  After
 * all local variables passed by pointer aro no more required reset the buffer by FreeLocalVar().
 * 
 * @param    varIn
 *           The pointer to variable to copy.
 *           
 * @param    varSize
 *           The size in bytes of the memory required to store the variable.
 *           
 * @return   New pointer to memory location the variable can be accessed or NULL if buffer is
 *           full.
 */
void * LocalVarAlloc(void * varIn, jsize varSize)
{
   void * ptrRes = NULL;
   if (nextVarStorageStart + varSize < MAX_LOCAL_VAR_IN_BYTES)
   {
      // this is the pointer to next free space in by-pointer vars storage
      ptrRes = varByPointerStorage + nextVarStorageStart;
      // make a copy of the input variable
      memcpy(ptrRes, varIn, varSize);
      // prepare next storage starting point
      nextVarStorageStart += varSize;
   }
      
   return ptrRes;
}

/**
 * Resets local buffer to store the variables passing by pointer.
 */
void FreeLocalVar()
{
   // simple returning to the begin
   nextVarStorageStart = 0;
}

/**
 * Gets SAFEARRAY object to hold the byte array coming from Java side.  The memory allocated for
 * SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to free
 * the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joByteArrIn
 *           The byte array Java object to get SAFEARRAY.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial byte array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromRaw(JNIEnv *env, jobject joByteArrIn)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // getting the java array size and data itself
      jsize jszRawSize = (*env)->GetArrayLength(env, joByteArrIn);
      jbyte * jbRawData = (*env)->GetByteArrayElements(env, joByteArrIn, 0);
      if (jbRawData != NULL)
      {
         // allocate 1-dimensional array
         psaRes = SafeArrayCreateVector(VT_UI1, 0, jszRawSize);
         // copy data from java array to SAFEARRAY
         if (psaRes != NULL)
         {
            jbyte * jbSAData = NULL;
            // get SAFEARRAY data access
            HRESULT hr = SafeArrayAccessData(psaRes, (void**)&jbSAData);
            if (hr == S_OK && jbSAData != NULL)
            {
               // copy whole memory region
               memcpy(jbSAData, jbRawData, jszRawSize);
               // release SAFEARRAY data
               SafeArrayUnaccessData(psaRes);
            }
            // keep the SAFEARRAY memory until explicit de-allocation
            saStorage[saStorageSize] = psaRes;
            saStorageSize++;
         }
      }
   
      // free java array size
      (*env)->ReleaseByteArrayElements(env, joByteArrIn, jbRawData, 0);
   }
   
   return psaRes;
}

/**
 * Clean up local SAFEARRAY object storage in order reversed to allocation.
 */
void FreeSafeArrays()
{
   // nothing to clear
   if (saStorageSize == 0)
   {
      return;
   }
   
   int i;
   for (i = saStorageSize - 1; i >= 0; i--)
   {
      SafeArrayDestroy(saStorage[i]);
      saStorage[i] = NULL;
   }
   
   // and clean up the size counter
   saStorageSize = 0;
}

// constand defining days from OLE DATE base to Java Date base
#define DAYS_SHIFT 25568.0
#define MILLIS_PER_DAY 86400000.0

/**
 * Converts LONG64 millisecond Java Date value to OLE DATE value.  The OLE DATE is a floating
 * point double value with days including fractions since 31 December 1899.  Java Date is based
 * on 1 January 1970 in milliseconds.
 * 
 * @param    jlMillisec
 *           The milliseconds of the Date since Java base.
 *           
 * @return   OLE DATA value.
 */
DATE OleDateFromJavaLong(jlong jlMillisec)
{
   // starting with bases distance
   double dblRes = DAYS_SHIFT;

   // OLE DATE is a double in format 'num_days.part_of_the_day' 
   dblRes += (double)(jlMillisec / (jlong)MILLIS_PER_DAY);
   dblRes += ((double)(jlMillisec % (jlong)MILLIS_PER_DAY)) / MILLIS_PER_DAY;
           
   return (DATE)dblRes;
}

/**
 * Converts OLE DATE value to LONG64 millisecond Java Date value.  The OLE DATE is a floating
 * point double value with days including fractions since 31 December 1899.  Java Date is based
 * on 1 January 1970 in milliseconds.
 * 
 * @param    dataIn
 *           The double OLE DATA value in days and fractions.
 *           
 * @return   Java compatible LONG64 value of milliseconds.
 */
jlong OleDateToJavaLong(DATE dataIn)
{
   // OLE DATE is a double in format 'num_days.part_of_the_day' 
   return (jlong)roundl(((double)dataIn - DAYS_SHIFT) * MILLIS_PER_DAY);
}

/**
 * Reconstructs the parent objects list with given immediate parent.
 * 
 * @param    startObjId
 *           ID of the immediate parent object.
 *
 * @param    ptrgArray
 *           The address ot the pointer to COM parent list to generate in target object.
 *           
 * @return   The number of the parent objects found or 0 if error or target is root object.
 */
int getCOMParents(jlong startObjId, jlong **ptrgArray)
{
   // loop through full registry to search the object
   // startObjId == 0 means target object is root(no parents)
   if (startObjId != 0)
   {
      int i;
      for (i = 0; i < hRegSize; i++)
      {
         // found item in registry
         if (phInfo[i]->id == startObjId)
         {
            // parent has a parents
            if (phInfo[i]->parentsSize > 0)
            {
               *ptrgArray = HeapAlloc(hProcHeap, heapUseMode,
                                      (phInfo[i]->parentsSize + 1) * sizeof(**ptrgArray));
               // copy all parents of the found object into target array
               if (*ptrgArray != NULL)
               {
                  memcpy(*ptrgArray + 1, phInfo[i]->parent_ids,
                         phInfo[i]->parentsSize * sizeof(**ptrgArray));
               }
            }
            // the parent is root
            else
            {
               // only single parent
               *ptrgArray = HeapAlloc(hProcHeap, heapUseMode, sizeof(**ptrgArray));
            }
            // the first element will be the immediate parent  
            if (*ptrgArray != NULL)
            {
               *ptrgArray[0] = startObjId;
            }
            // return the amount of the copied objects
            return (*ptrgArray != NULL) ? phInfo[i]->parentsSize + 1 : 0;
         }
      }
   }
   
   // not found in registry or regstry is empty or root object case
   return 0;
}

/**
 * Adding the specified COM object into local registry.
 * 
 * @param    objId
 *           Pointer to a COM object to add.
 *
 * @param    objName
 *           The COM property object name.
 *           
 * @param    parentId
 *           The parent ID of the object that is adding to registry.
 */
void addCOMToRegistry(jlong objId, jchar * objName, jlong parentId)
{
   // simple check for registry not to exceed the size limit
   if (hRegSize < MAX_HANDLES_INFO_KEEP)
   {
      // index of new entry to add
      int newIdx = hRegSize;
      // allocate memory and save COM object info
      phInfo[newIdx] = (PComHandleInfo)GlobalAlloc(GMEM_FIXED, sizeof(ComHandleInfo));
      if (phInfo[newIdx] != NULL)
      {
         phInfo[newIdx]->id = objId;
         wcscpy(phInfo[newIdx]->name, objName);
         // reconstruct parents array
         phInfo[newIdx]->parent_ids = NULL;
         phInfo[newIdx]->parentsSize = getCOMParents(parentId, &phInfo[newIdx]->parent_ids);
         // yet another item in registry
         hRegSize++;
      }
   }
}

/**
 * Removing the specified COM object from local registry and compacting the registry.
 * 
 * @param    objId
 *           Pointer to a COM object to add.
 */
void removeCOMFromRegistry(jlong objId)
{
   int i;
   // loop through registry to search the object
   for (i = 0; i < hRegSize; i++)
   {
      if (phInfo[i]->id == objId)
      {
         // free memory allocated for parent list
         if (phInfo[i]->parent_ids != NULL)
         {
            HeapFree(hProcHeap, heapUseMode, phInfo[i]->parent_ids);
            phInfo[i]->parent_ids = NULL;
         }
         // free memory allocated to given object info
         GlobalFree(phInfo[i]);
         phInfo[i] = NULL;
         break;
      }
   }
   // compact remaining registry entries
   // the idea if size is 2 and more copy last entry into deleted position 
   if (hRegSize > 1 && i < hRegSize - 1)
   {
      phInfo[i] = phInfo[hRegSize - 1];
   }
   // zero for last entry
   if (hRegSize > 0)
   {
      hRegSize--;
      phInfo[hRegSize] = NULL;
   }
}

/**
 * Checks if the specified COM object is in local registry.
 * 
 * @param    objId
 *           The COM object ID.
 *           
 * @return   JNI_TRUE is object in registry, JNI_FALSE otherwise.
 */
jboolean isCOMInRegistry(jlong objId)
{
   int i;
   // loop through registry to search the object
   for (i = 0; i < hRegSize; i++)
   {
      // found item in registry
      if (phInfo[i]->id == objId)
      {
         return JNI_TRUE;
      }
   }
   
   // not found in registry or regstry is empty
   return JNI_FALSE;
}

/**
 * Gets the string name of the specified COM object from local registry.
 * 
 * @param    objId
 *           The COM object ID.
 *           
 * @return   The character string name of the given object.
 */
jchar * getCOMName(jlong objId)
{
   int i;
   // loop through registry to search the object
   for (i = 0; i < hRegSize; i++)
   {
      // found item in registry
      if (phInfo[i]->id == objId)
      {
         return phInfo[i]->name;
      }
   }
   
   // not found in registry or regstry is empty
   return NULL;
}

/**
 * Getting next child for given parent from the top of the list or 0 if no more children.
 * 
 * @param    parentId
 *           Parent object ID for children to search.
 *           
 * @return   The child object ID or 0 if none.
 */
jlong getNextCOMChild(jlong parentId)
{
   int i, j;
   // loop through full registry to search the object
   for (i = 0; i < hRegSize; i++)
   {
      // found item in registry
      for (j = 0; j < phInfo[i]->parentsSize; j++)
      {
         if (phInfo[i]->parent_ids[j] == parentId)
         {
            return phInfo[i]->id;
         }
      }
   }
   
   // not found in registry or regstry is empty
   return 0;
}

/**
 * Releasing COM object according to usage counter value.
 * 
 * @param    objId
 *           Object ID to release.
 */
void releaseCOMObject(jlong objId)
{
   // at least one mandatory release we have
   // convert object ID to interface pointer
   IDispatch * objPtr = (IDispatch *)objId;
   // release object completely
   ULONG ulRemaining;
   do
   {
      // we need to remove all instances of the object
      ulRemaining = objPtr->lpVtbl->Release(objPtr);
   } while (ulRemaining > 0);
}

/**
 * Releasing all child objects for given parent.
 * 
 * @param    parentId
 *           Parent object ID for children to release.
 */
void releaseCOMChildren(jlong parentId)
{
   jlong objId;
   while ((objId = getNextCOMChild(parentId)) != 0)
   {
      // release object
      releaseCOMObject(objId);
      // refresh our registry
      removeCOMFromRegistry(objId);
   }
}

/**
 * Gets the ID of the method or property setter or getter for the given Unicode name for COM
 * object specified.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    objPtr
 *           The COM object IDispatch pointer.
 *           
 * @param    name
 *           The Unicode name of the method or property to search.
 *           
 * @param    dispIdRes
 *           The pointer to DISPID variable to return the ID.
 *           
 * @return   JNI_TRUE if the ID was found, JNI_FALSE otherwise.
 */
jboolean getMethOrPropId(JNIEnv *env, IDispatch * objPtr, jstring name, DISPID * dispIdRes)
{
   // result to return
   jboolean idRes  = JNI_FALSE;
   
   // getting the Unicode name for method or property setter/getter
   jchar* nameMethOrProp = (jchar*) (*env)->GetStringChars(env, name, NULL);
   if (nameMethOrProp != NULL)
   {
      // convert long ID to object pointer
      // attempt to identify the ID of property getter/setter or method
      HRESULT hr = objPtr->lpVtbl->GetIDsOfNames(objPtr, &IID_NULL, (LPOLESTR*)&nameMethOrProp,
                                                 1, LOCALE_NEUTRAL, dispIdRes);
      // the ID was properly set
      if (hr == S_OK)
      {
         idRes = JNI_TRUE;
      }
   }
   
   // release java string
   (*env)->ReleaseStringChars(env, name, nameMethOrProp);
   
   return idRes;
}

/**
 * Constructs the appropriate jobject of BaseDataType from VARIANT provided.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    jlParentID
 *           The parent ID of the object whose property or method called.
 *           
 * @param    vInput
 *           The pointer to VARIANT source variable.
 *           
 * @return   The BaseDataType base object representing the VARIANT data.
 */
jobject getJObjectFromVariant(JNIEnv *env, jlong jlParentID, VARIANT * vInput)
{
   jobject retVal = NULL;
   jclass cls = NULL;
   jmethodID cid = NULL;
   
   // BY-VARIANT-POINTER data is not supported at this time
   
   // start checking the incoming data
   if ((vInput->vt & 0xFF) == VT_I4)
   {
      // INTEGER value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/integer");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
         if (cid != NULL)
         {
            // and make new instance for single object or array
            if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
            {
               // we have an extent based array of integers
               SAFEARRAY * psaIntSArr;
               // extracting array data from VARIANT
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  psaIntSArr = (SAFEARRAY *)*vInput->pparray;
               }
               else
               {
                  psaIntSArr = (SAFEARRAY*)vInput->parray;
               }
               // reconstruct java integer objects array from SAFEARRAY
               if (psaIntSArr != NULL)
               {
                  // array size to return to Java
                  jsize jszArrSize = psaIntSArr->rgsabound[0].cElements;
                  // placeholder for objects array to return
                  retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                  if (retVal != NULL)
                  {
                     jint * jintIntArr = NULL;
                     // get the SAFEARRAY data access
                     if (SafeArrayAccessData(psaIntSArr, (void**)&jintIntArr) == S_OK &&
                         jintIntArr != NULL)
                     {
                        int i;
                        for (i = 0; i < jszArrSize; i++)
                        {
                           // create yet another 'integer' class object from int value
                           // and put it into resulting object array
                           (*env)->SetObjectArrayElement(env, retVal, i,
                                                  (*env)->NewObject(env, cls, cid, *jintIntArr));
                           // move to the next integer value address in SAFEARRAY data region
                           jintIntArr++;
                        }
                        // release SAFEARRAY data access
                        SafeArrayUnaccessData(psaIntSArr);
                     }
                  }
               }
            }
            else
            {
               // we have single instance
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jint)*(vInput->plVal));
               }
               else
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jint)vInput->lVal);
               }
            }
         }
      }
   }
   else if ((vInput->vt & VT_I8) == VT_I8)
   {
      // INT64 value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/int64");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(J)V");
         if (cid != NULL)
         {
            // and make new instance for single object or array
            if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
            {
               // we have an extent based array of integers
               SAFEARRAY * psaInt64SArr;
               // extracting array data from VARIANT
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  psaInt64SArr = (SAFEARRAY *)*vInput->pparray;
               }
               else
               {
                  psaInt64SArr = (SAFEARRAY*)vInput->parray;
               }
               // reconstruct java integer objects array from SAFEARRAY
               if (psaInt64SArr != NULL)
               {
                  // array size to return to Java
                  jsize jszArrSize = psaInt64SArr->rgsabound[0].cElements;
                  // placeholder for objects array to return
                  retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                  if (retVal != NULL)
                  {
                     jlong * jlInt64Arr = NULL;
                     // get the SAFEARRAY data access
                     if (SafeArrayAccessData(psaInt64SArr, (void**)&jlInt64Arr) == S_OK &&
                         jlInt64Arr != NULL)
                     {
                        int i;
                        for (i = 0; i < jszArrSize; i++)
                        {
                           // create yet another 'int64' class object from int value
                           // and put it into resulting object array
                           (*env)->SetObjectArrayElement(env, retVal, i,
                                                  (*env)->NewObject(env, cls, cid, *jlInt64Arr));
                           // move to the next integer value address in SAFEARRAY data region
                           jlInt64Arr++;
                        }
                        // release SAFEARRAY data access
                        SafeArrayUnaccessData(psaInt64SArr);
                     }
                  }
               }
            }
            else
            {
               // and make new instance
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jlong)*(vInput->pllVal));
               }
               else
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jlong)vInput->llVal);
               }
            }
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_R8)
   {
      // DECIMAL value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/decimal");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(D)V");
         if (cid != NULL)
         {
            // and make new instance for single object or array
            if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
            {
               // we have an extent based array of integers
               SAFEARRAY * psaDoubleSArr;
               // extracting array data from VARIANT
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  psaDoubleSArr = (SAFEARRAY *)*vInput->pparray;
               }
               else
               {
                  psaDoubleSArr = (SAFEARRAY*)vInput->parray;
               }
               // reconstruct java integer objects array from SAFEARRAY
               if (psaDoubleSArr != NULL)
               {
                  // array size to return to Java
                  jsize jszArrSize = psaDoubleSArr->rgsabound[0].cElements;
                  // placeholder for objects array to return
                  retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                  if (retVal != NULL)
                  {
                     jdouble * jdblDoubleArr = NULL;
                     // get the SAFEARRAY data access
                     if (SafeArrayAccessData(psaDoubleSArr, (void**)&jdblDoubleArr) == S_OK &&
                         jdblDoubleArr != NULL)
                     {
                        int i;
                        for (i = 0; i < jszArrSize; i++)
                        {
                           // create yet another 'decimal' class object from int value
                           // and put it into resulting object array
                           (*env)->SetObjectArrayElement(env, retVal, i,
                                                (*env)->NewObject(env, cls, cid, *jdblDoubleArr));
                           // move to the next integer value address in SAFEARRAY data region
                           jdblDoubleArr++;
                        }
                        // release SAFEARRAY data access
                        SafeArrayUnaccessData(psaDoubleSArr);
                     }
                  }
               }
            }
            else
            {
               // and make new instance
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  retVal = (*env)->NewObject(env, cls, cid, (double)*(vInput->pdblVal));
               }
               else
               {
                  retVal = (*env)->NewObject(env, cls, cid, (double)vInput->dblVal);
               }
            }
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_BOOL)
   {
      // LOGICAL value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/logical");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(Z)V");
         if (cid != NULL)
         {
            // and make new instance for single object or array
            if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
            {
               // we have an extent based array of integers
               SAFEARRAY * psaLogicalSArr;
               // extracting array data from VARIANT
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  psaLogicalSArr = (SAFEARRAY *)*vInput->pparray;
               }
               else
               {
                  psaLogicalSArr = (SAFEARRAY*)vInput->parray;
               }
               // reconstruct java integer objects array from SAFEARRAY
               if (psaLogicalSArr != NULL)
               {
                  // array size to return to Java
                  jsize jszArrSize = psaLogicalSArr->rgsabound[0].cElements;
                  // placeholder for objects array to return
                  retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                  if (retVal != NULL)
                  {
                     jboolean * jbLogicalArr = NULL;
                     // get the SAFEARRAY data access
                     if (SafeArrayAccessData(psaLogicalSArr, (void**)&jbLogicalArr) == S_OK &&
                         jbLogicalArr != NULL)
                     {
                        int i;
                        for (i = 0; i < jszArrSize; i++)
                        {
                           // create yet another 'logical' class object from int value
                           // and put it into resulting object array
                           (*env)->SetObjectArrayElement(env, retVal, i,
                                                (*env)->NewObject(env, cls, cid, *jbLogicalArr));
                           // move to the next integer value address in SAFEARRAY data region
                           jbLogicalArr++;
                        }
                        // release SAFEARRAY data access
                        SafeArrayUnaccessData(psaLogicalSArr);
                     }
                  }
               }
            }
            else
            {
               // and make new instance
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jboolean)*(vInput->pboolVal));
               }
               else
               {
                  retVal = (*env)->NewObject(env, cls, cid, (jboolean)vInput->boolVal);
               }
            }
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_BSTR)
   {
      // CHARACTER or LONGCHAR value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/character");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/lang/String;)V");
         if (cid != NULL)
         {
            // and make new instance for single object or array
            if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
            {
               // we have an extent based array of integers
               SAFEARRAY * psaCharSArr;
               // extracting array data from VARIANT
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  psaCharSArr = (SAFEARRAY *)*vInput->pparray;
               }
               else
               {
                  psaCharSArr = (SAFEARRAY*)vInput->parray;
               }
               // reconstruct java integer objects array from SAFEARRAY
               if (psaCharSArr != NULL)
               {
                  // array size to return to Java
                  jsize jszArrSize = psaCharSArr->rgsabound[0].cElements;
                  // placeholder for objects array to return
                  retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                  if (retVal != NULL)
                  {
                     BSTR * pbstrCharacter = NULL;
                     // get the SAFEARRAY data access
                     if (SafeArrayAccessData(psaCharSArr, (void**)&pbstrCharacter) == S_OK &&
                         pbstrCharacter != NULL)
                     {
                        int i;
                        for (i = 0; i < jszArrSize; i++)
                        {
                           // create new string from BSTR
                           jstring joStringVal = (*env)->NewString(env, *pbstrCharacter,
                                                                   wcslen(*pbstrCharacter));
                           // create yet another 'character' class object from BSTR value
                           // and put it into resulting object array with
                           // character(String) entries
                           (*env)->SetObjectArrayElement(env, retVal, i,
                                                (*env)->NewObject(env, cls, cid, joStringVal));
                           // move to the next integer value address in SAFEARRAY data region
                           pbstrCharacter++;
                        }
                        // release SAFEARRAY data access
                        SafeArrayUnaccessData(psaCharSArr);
                     }
                  }
               }
            }
            else
            {
               // Get the pointer to Unicode string
               jchar * jchStr = NULL;
               if ((vInput->vt & VT_BYREF) == VT_BYREF)
               {
                  jchStr = (jchar*)*vInput->pbstrVal;
               }
               else
               {
                  jchStr = (jchar*)vInput->bstrVal;
               }
               // and make new instance of Unicode string
               retVal = (*env)->NewObject(env, cls, cid,
                                          (*env)->NewString(env, jchStr, wcslen(jchStr)));
            }
            
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_UNKNOWN || (vInput->vt & 0xFF) == VT_DISPATCH)
   {
      // COM-HANDLE object inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/comhandle");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/lang/Object;)V");
         if (cid != NULL)
         {
            // create NativeComObject
            jclass clsNativeCom = (*env)->FindClass(env,
                                                    "com/goldencode/p2j/comauto/NativeComObject");
            // and find the constructor
            if (clsNativeCom != NULL)
            {
               // getting the NativeComObject(long) constructor
               jmethodID cidNativeCom = (*env)->GetMethodID(env, clsNativeCom, "<init>", "(J)V");
               if (cidNativeCom != NULL)
               {
                  // and make new instance for single object or array
                  if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
                  {
                     // we have an extent based array of integers
                     SAFEARRAY * psaComSArr;
                     // extracting array data from VARIANT
                     if ((vInput->vt & VT_BYREF) == VT_BYREF)
                     {
                        psaComSArr = (SAFEARRAY *)*vInput->pparray;
                     }
                     else
                     {
                        psaComSArr = (SAFEARRAY*)vInput->parray;
                     }
                     // reconstruct java integer objects array from SAFEARRAY
                     if (psaComSArr != NULL)
                     {
                        // array size to return to Java
                        jsize jszArrSize = psaComSArr->rgsabound[0].cElements;
                        // placeholder for objects array to return
                        retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                        if (retVal != NULL)
                        {
                           jlong * jlComHandle = NULL;
                           // get the SAFEARRAY data access
                           if (SafeArrayAccessData(psaComSArr, (void**)&jlComHandle) == S_OK &&
                               jlComHandle != NULL)
                           {
                              int i;
                              for (i = 0; i < jszArrSize; i++)
                              {
                                 // Java NativeComObject class
                                 jobject jobNativeCom = (*env)->NewObject(env, clsNativeCom,
                                                                          cidNativeCom,
                                                                          *jlComHandle);
                                 // create yet another 'comhandle' class object from int value
                                 // and put it into resulting object array for
                                 // comhandle(NativeComObject) objects
                                 (*env)->SetObjectArrayElement(env, retVal, i,
                                                 (*env)->NewObject(env, cls, cid, jobNativeCom));
                                 // move to the next integer value address in SAFEARRAY data region
                                 jlComHandle++;
                              }
                              // release SAFEARRAY data access
                              SafeArrayUnaccessData(psaComSArr);
                           }
                        }
                     }
                  }
                  else
                  {
                     // create comhandle object based on NativeComObject
                     jlong comHandleID;
                     // we must differentiate returned VT_UNKNOWN from VT_DISPATCH here
                     // due to different target data location 
                     if ((vInput->vt & 0xFF) == VT_UNKNOWN)
                     {
                        // got by pointer or by value
                        if ((vInput->vt & VT_BYREF) == VT_BYREF)
                        {
                           comHandleID = (jlong)*(vInput->ppunkVal);
                        }
                        else
                        {
                           comHandleID = (jlong)vInput->punkVal;
                        }
                     }
                     else
                     {
                        // got by pointer or by value
                        if ((vInput->vt & VT_BYREF) == VT_BYREF)
                        {
                           comHandleID = (jlong)*(vInput->ppdispVal);
                        }
                        else
                        {
                           comHandleID = (jlong)vInput->pdispVal;
                        }
                     }
                     // requested com handle must be added to local registry
                     if (!isCOMInRegistry(comHandleID))
                     {
                        // reserved room for 64 bit hexadecimal integer(16 chars) + 2 chars for
                        // '0x' prefix and 1 char for terminating null character, one char is 2
                        // bytes in size
                        jchar name[19];
                        // adding handle ID to local registry
                        swprintf(name, 19, L"%#018I64x", comHandleID);
                        addCOMToRegistry(comHandleID, name, jlParentID);
                     }
                     // this is new comhandle(NativeComObject(comId)) object to return
                     retVal = (*env)->NewObject(env, cls, cid,
                                 (*env)->NewObject(env, clsNativeCom, cidNativeCom, comHandleID));
                  }
               }
            }
         }
      }
   }
   else if ((vInput->vt & VT_ARRAY) == VT_ARRAY && (vInput->vt & 0xFF) == VT_UI1)
   {
      // RAW data indide(array of bytes)
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/raw");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "([B)V");
         if (cid != NULL)
         {
            SAFEARRAY * psaByteSArr;
            // extracting array data from VARIANT
            if ((vInput->vt & VT_BYREF) == VT_BYREF)
            {
               psaByteSArr = (SAFEARRAY *)*vInput->pparray;
            }
            else
            {
               psaByteSArr = (SAFEARRAY*)vInput->parray;
            }
            // reconstruct java byte array from SAFEARRAY
            if (psaByteSArr != NULL)
            {
               jsize jszRawSize = psaByteSArr->rgsabound[0].cElements;
               jobject joByteArr = (*env)->NewByteArray(env, jszRawSize);
               if (joByteArr != NULL)
               {
                  jbyte * jbBytesArr = NULL;
                  // get the SAFEARRAY data access
                  if (SafeArrayAccessData(psaByteSArr, (void**)&jbBytesArr) == S_OK &&
                      jbBytesArr != NULL)
                  {
                     // byte array to return constructed, populate it with SAFEARRAY data
                     (*env)->SetByteArrayRegion(env, joByteArr, psaByteSArr->rgsabound[0].lLbound,
                                                jszRawSize, jbBytesArr);
                     // and make new instance fpr 'raw' class
                     retVal = (*env)->NewObject(env, cls, cid, joByteArr);
                     // release SAFEARRAY data access
                     SafeArrayUnaccessData(psaByteSArr);
                  }
               }
            }
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_DATE)
   {
      // DATE data passed by reference
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/datetime");
      if (cls != NULL)
      {
         // get the object constructor based on BaseDataType
         cid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/util/Date;)V");
         if (cid != NULL)
         {
            jclass clsDate = (*env)->FindClass(env, "java/util/Date");
            if (clsDate != NULL)
            {
               jmethodID cidDate = (*env)->GetMethodID(env, clsDate, "<init>", "(J)V");
               if (cidDate != NULL)
               {
                  // and make new instance for single object or array
                  if ((vInput->vt & VT_ARRAY) == VT_ARRAY)
                  {
                     // we have an extent based array of integers
                     SAFEARRAY * psaDateSArr;
                     // extracting array data from VARIANT
                     if ((vInput->vt & VT_BYREF) == VT_BYREF)
                     {
                        psaDateSArr = (SAFEARRAY *)*vInput->pparray;
                     }
                     else
                     {
                        psaDateSArr = (SAFEARRAY*)vInput->parray;
                     }
                     // reconstruct java integer objects array from SAFEARRAY
                     if (psaDateSArr != NULL)
                     {
                        // array size to return to Java
                        jsize jszArrSize = psaDateSArr->rgsabound[0].cElements;
                        // placeholder for objects array to return
                        retVal = (*env)->NewObjectArray(env, jszArrSize, cls, NULL);
                        if (retVal != NULL)
                        {
                           DATE * dtDateArr = NULL;
                           // get the SAFEARRAY data access
                           if (SafeArrayAccessData(psaDateSArr, (void**)&dtDateArr) == S_OK &&
                               dtDateArr != NULL)
                           {
                              int i;
                              for (i = 0; i < jszArrSize; i++)
                              {
                                 // long value of the DATE
                                 jlong jlDateVal = OleDateToJavaLong(*dtDateArr);
                                 // Java Date class
                                 jobject joDate = (*env)->NewObject(env, clsDate, cidDate,
                                                                    jlDateVal);
                                 // create yet another 'integer' class object from int value
                                 // and put it into resulting object array for
                                 // datetime(Date) objects
                                 (*env)->SetObjectArrayElement(env, retVal, i,
                                                       (*env)->NewObject(env, cls, cid, joDate));
                                 // move to the next integer value address in SAFEARRAY data region
                                 dtDateArr++;
                              }
                              // release SAFEARRAY data access
                              SafeArrayUnaccessData(psaDateSArr);
                           }
                        }
                     }
                  }
                  else
                  {
                     // now get the value from variant
                     DATE dateVal;
                     if ((vInput->vt & VT_BYREF) == VT_BYREF)
                     {
                        dateVal = *(vInput->pdate);
                     }
                     else
                     {
                        dateVal = vInput->date;
                     }
                     // reconstructing LONG value from DATE
                     jlong jlDateVal = OleDateToJavaLong(dateVal);
                     retVal = (*env)->NewObject(env, cls, cid,
                                            (*env)->NewObject(env, clsDate, cidDate, jlDateVal));
                  }
               }
            }
         }
      }
   }
   else if ((vInput->vt & 0xFF) == VT_ERROR)
   {
      // assume INTEGER ERROR CODE single onject value inside
      cls = (*env)->FindClass(env, "com/goldencode/p2j/util/integer");
      if (cls != NULL)
      {
         // get the object constructor
         cid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
         if (cid != NULL)
         {
            // we have single instance
            if ((vInput->vt & VT_BYREF) == VT_BYREF)
            {
               retVal = (*env)->NewObject(env, cls, cid, (jint)*(vInput->pscode));
            }
            else
            {
               retVal = (*env)->NewObject(env, cls, cid, (jint)vInput->scode);
            }
         }
      }
   }

   return retVal;
}

/**
 * Constructs the appropriate array of jobjects of BaseDataType from VARIANT provided.  In case
 * the variant contains single object the returnig array will have size 1 elements.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    jlParentID
 *           The parent ID of the object whose property or method called.
 *           
 * @param    vInput
 *           The pointer to VARIANT source variable.
 *           
 * @return   The BaseDataType base object array representing the VARIANT data.
 */
jobjectArray getJObjectArrayFromVariant(JNIEnv *env, jlong jlParentID, VARIANT * vInput)
{
   jobjectArray joArrayRet = NULL;
   // class to construct objects array
   jclass clsRet = (*env)->FindClass(env, "com/goldencode/p2j/util/BaseDataType");

   // do we have the class to construct from?
   if (clsRet != NULL)
   {
      // convert variant to java object, single or array
      jobject joRetVal = getJObjectFromVariant(env, jlParentID, vInput);
      // consider raw data type must be excluded as ready to use array
      if ((vInput->vt & VT_ARRAY) == VT_ARRAY && (vInput->vt & 0xFF) != VT_UI1)
      {
         // we already have an array inside variant
         joArrayRet = joRetVal;
      }
      else
      {
         // single object, make 1-sized array
         joArrayRet = (*env)->NewObjectArray(env, 1, clsRet, joRetVal);
      }
   }
   
   return joArrayRet;
}

/**
 * Gets SAFEARRAY object to hold the integer array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The integer array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial integer array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromIntJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_I4, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         jint * jintSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&jintSAData);
         if (hr == S_OK && jintSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joIntObj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsInt = (*env)->GetObjectClass(env, joIntObj);
               jmethodID midGetIntegerVal = (*env)->GetMethodID(env, clsInt, "intValue", "()I");
               // getting next data entry
               if (midGetIntegerVal != NULL)
               {
                  jint jintVal = (*env)->CallIntMethod(env, joIntObj, midGetIntegerVal);
                  // copy data value to the next location inside SAFEARRAY
                  *jintSAData = jintVal;
                  // location in SAFEARRAY memory to store the next data item
                  jintSAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the character or longchar array coming from Java side.  The
 * memory allocated for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer
 * needs.  Attempting to free the SAFEARRAY object before using finished can produce application
 * to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The integer array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial character or longchar
 *           array from Java, or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromCharJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_BSTR, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         BSTR * pbstrSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&pbstrSAData);
         if (hr == S_OK && pbstrSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jstring joStrVal = (*env)->GetObjectArrayElement(env, joArrIn, i);
               // getting java Unicode string
               jchar * jchVal = (jchar*) (*env)->GetStringChars(env, joStrVal, NULL);
               // create OLE compatible string
               // we will free it only after finish using it(after Invoke for given VARIANT)
               BSTR bstr = LocalAllocBSTR(jchVal);
               // getting next data entry
               if (bstr != NULL)
               {
                  // copy data value to the next location inside SAFEARRAY
                  *pbstrSAData = bstr;
                  // location in SAFEARRAY memory to store the next data item
                  pbstrSAData++;
               }
               // release Java string
               (*env)->ReleaseStringChars(env, joStrVal, jchVal);
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the int64 array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The integer array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @param    vt
 *           The VARTYPE constant that must be used to create 8-bytes long compatible SAFEARRAY.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial int64 array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromInt64JArray(JNIEnv *env, jobject joArrIn, jsize jszSize, VARTYPE vt)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(vt, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         jlong * jlSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&jlSAData);
         if (hr == S_OK && jlSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joInt64Obj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsInt64 = (*env)->GetObjectClass(env, joInt64Obj);
               jmethodID midGetInt64Val = (*env)->GetMethodID(env, clsInt64, "longValue", "()J");
               // getting next data entry
               if (midGetInt64Val != NULL)
               {
                  jlong jlVal = (*env)->CallLongMethod(env, joInt64Obj, midGetInt64Val);
                  // copy data value to the next location inside SAFEARRAY
                  *jlSAData = jlVal;
                  // location in SAFEARRAY memory to store the next data item
                  jlSAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the Date array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The Date array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial Date array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromDateJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_DATE, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         DATE * dtSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&dtSAData);
         if (hr == S_OK && dtSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joDateObj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsDate = (*env)->GetObjectClass(env, joDateObj);
               jmethodID midGetDateVal = (*env)->GetMethodID(env, clsDate, "getTime", "()J");
               // getting next data entry
               if (midGetDateVal != NULL)
               {
                  // Date to store value as jlong
                  jlong jlDateVal = (*env)->CallLongMethod(env, joDateObj, midGetDateVal);
                  // COM compatible DATE value
                  DATE dateValue = OleDateFromJavaLong(jlDateVal);
                  // copy data value to the next location inside SAFEARRAY
                  *dtSAData = dateValue;
                  // location in SAFEARRAY memory to store the next data item
                  dtSAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the decimal array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The decimal array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial decimal array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromDecimalJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_R8, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         jdouble * jdblSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&jdblSAData);
         if (hr == S_OK && jdblSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joDecObj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsDec = (*env)->GetObjectClass(env, joDecObj);
               jmethodID midGetDecVal = (*env)->GetMethodID(env, clsDec, "doubleValue", "()D");
               // getting next data entry
               if (midGetDecVal != NULL)
               {
                  jdouble jdblVal = (*env)->CallDoubleMethod(env, joDecObj, midGetDecVal);
                  // copy data value to the next location inside SAFEARRAY
                  *jdblSAData = jdblVal;
                  // location in SAFEARRAY memory to store the next data item
                  jdblSAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the currency array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The currency array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial currency array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromCurrencyJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_CY, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         CY * cySAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&cySAData);
         if (hr == S_OK && cySAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joDecObj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsDec = (*env)->GetObjectClass(env, joDecObj);
               jmethodID midGetDecVal = (*env)->GetMethodID(env, clsDec, "doubleValue", "()D");
               // getting next data entry
               if (midGetDecVal != NULL)
               {
                  jdouble jdblVal = (*env)->CallDoubleMethod(env, joDecObj, midGetDecVal);
                  // copy data value to the next location inside SAFEARRAY
                  // transforming DECIMAL to CURRENCY 
                  VarCyFromR8(jdblVal, cySAData);
                  // location in SAFEARRAY memory to store the next data item
                  cySAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Gets SAFEARRAY object to hold the logical array coming from Java side.  The memory allocated
 * for SAFEARRAY must explicitly be freed by FreeSafeArrays() when no longer needs. Attempting to
 * free the SAFEARRAY object before using finished can produce application to crash.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    joArrIn
 *           The logical array Java object to get SAFEARRAY.
 *           
 * @param    jszSize
 *           The target array size on output.  The value is independent on the Java array size
 *           integrated into Java object array and has preference to compute final array size.
 *           
 * @return   Pointer to the SAFEARRAY object containing data from initial logical array from Java,
 *           or NULL in case of error for SAFEARRAY creation.
 */
SAFEARRAY * GetSafeArrayFromLogicalJArray(JNIEnv *env, jobject joArrIn, jsize jszSize)
{
   SAFEARRAY * psaRes = NULL;
   
   if (saStorageSize < MAX_SAFEARRAY_TO_KEEP)
   {
      // check the array size
      jsize jszArrSize = (*env)->GetArrayLength(env, joArrIn);
      if (jszArrSize < jszSize)
      {
         // TODO: what actions if size mismatch
         return psaRes;
      }
      
      // allocate 1-dimensional array
      psaRes = SafeArrayCreateVector(VT_BOOL, 0, jszSize);
      if (psaRes != NULL)
      {
         // populate the array with elements from java array
         jboolean * jbSAData = NULL;
         // get SAFEARRAY data access
         HRESULT hr = SafeArrayAccessData(psaRes, (void**)&jbSAData);
         if (hr == S_OK && jbSAData != NULL)
         {
            int i;
            for (i = 0; i < jszSize; i++)
            {
               // getting tools to obtain integer value
               jobject joBoolObj = (*env)->GetObjectArrayElement(env, joArrIn, i);
               jclass clsBool = (*env)->GetObjectClass(env, joBoolObj);
               jmethodID midGetBoolVal = (*env)->GetMethodID(env, clsBool, "booleanValue", "()Z");
               // getting next data entry
               if (midGetBoolVal != NULL)
               {
                  jboolean jbData = (*env)->CallBooleanMethod(env, joBoolObj, midGetBoolVal);
                  // copy data value to the next location inside SAFEARRAY
                  *jbSAData = jbData;
                  // location in SAFEARRAY memory to store the next data item
                  jbSAData++;
               }
            }
            // release SAFEARRAY data
            SafeArrayUnaccessData(psaRes);
         }
         // keep the SAFEARRAY memory until explicit de-allocation
         saStorage[saStorageSize] = psaRes;
         saStorageSize++;
      }
   }
   
   return psaRes;
}

/**
 * Constructs the appropriate VARIANT variable from BaseDataType based jobject provided.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    vOutput
 *           The pointer to VARIANT result variable.
 *           
 * @param    valObj
 *           The BaseDataType based object used as source.
 *           
 * @return   JNI_TRUE if the input oject has OUTPUT or INPUT-OUTPUT mode embedded,
 *           JNI_FALSE otherwise (the object is used only as INPUT variable).
 */
jboolean setVariantFromJObject(JNIEnv *env, VARIANT * vOutput, jobject valObj)
{
   jclass cls = NULL;
   jboolean rc = JNI_FALSE;
   
   // sometimes input object can be NULL for example if parameter is optional and missing
   if (valObj == NULL)
   {
      // in this case we notify COM there is no parameter for this data slot
      vOutput->vt = VT_ERROR;
      vOutput->scode = DISP_E_PARAMNOTFOUND;
   }
   // expecting to have ComParameter class here
   else if ((cls = (*env)->GetObjectClass(env, valObj)) != NULL)
   {
      // method to get FWD data type
      jmethodID midGetFwdDT = (*env)->GetMethodID(env, cls, "getFwdDataType",
                                                  "()Ljava/lang/String;");
      // method to get parameter mode
      jmethodID midGetMode = (*env)->GetMethodID(env, cls, "getMode", "()C");
      // method to get BY-POINTER mode
      jmethodID midGetByPtr = (*env)->GetMethodID(env, cls, "isByPointer", "()Z");
      // method to get BY-VARIANT-POINTER mode
      jmethodID midGetByVariantPtr = (*env)->GetMethodID(env, cls, "isByVariantPointer", "()Z");
      // method to get AS-DATATYPE option
      jmethodID midGetAsDataType = (*env)->GetMethodID(env, cls, "getDataType",
                                          "()Lcom/goldencode/p2j/comauto/ComParameter$DataType;");
      // method to get if the variable is extent based array
      jmethodID midIsExtent = (*env)->GetMethodID(env, cls, "isExtent", "()Z");
      // method to get the extent array size
      jmethodID midExtentLength = (*env)->GetMethodID(env, cls, "getLength", "()I");
      
      if (midGetFwdDT != NULL && midGetMode != NULL && midGetByPtr != NULL &&
          midGetByVariantPtr != NULL && midGetAsDataType != NULL &&
          midIsExtent != NULL && midExtentLength != NULL)
      {
         // get as-datatype option
         int asDataType = AS_DT_NOT_DEFINED;
         jobject joAsDataType = (*env)->CallObjectMethod(env, valObj, midGetAsDataType);
         if (joAsDataType != NULL)
         {
            // find the method to get enum value
            jclass clsDataType = (*env)->GetObjectClass(env, joAsDataType);
            if (clsDataType != NULL)
            {
               // this is method to get integer order value
               jmethodID midGetDataTypeValue = (*env)->GetMethodID(env, clsDataType,
                                                                   "ordinal", "()I");
               if (midGetDataTypeValue != NULL)
               {
                  // now we refresh the as-datatype value for given parameter
                  asDataType = (*env)->CallIntMethod(env, joAsDataType, midGetDataTypeValue);
               }
            }
         }
         // get FWD data mode
         jchar jchParamMode = (*env)->CallCharMethod(env, valObj, midGetMode);
         if (jchParamMode == 'O' || jchParamMode == 'U')
         {
            // this means the option to be returned back to Java
            rc = JNI_TRUE;
         }
         // get data passing modes (by pointer or by variant pointer)
         // access modes 'O' or 'U' requires default by-pointer passing to change the value
         jboolean isByPointer = rc || (*env)->CallBooleanMethod(env, valObj, midGetByPtr);
         jboolean isByVariantPtr = (*env)->CallBooleanMethod(env, valObj, midGetByVariantPtr);
         // get extent option values
         jboolean isExtent = (*env)->CallBooleanMethod(env, valObj, midIsExtent);
         jint extentSize = (*env)->CallIntMethod(env, valObj, midExtentLength);
         // get FWD data type
         jstring jstFwdDataType = (*env)->CallObjectMethod(env, valObj, midGetFwdDT);
         char * chFwdDataType = (char*) (*env)->GetStringUTFChars(env, jstFwdDataType, NULL);

         // check FWD data type
         if (chFwdDataType != NULL)
         {
            // int representation of the string first 4 chars to avoid strcmp() calls
            int * iDataType = (int *)chFwdDataType;
            // method to get the real data
            jmethodID midGetData = NULL;

            // BY-VARIANT-POINTER option specified
            if (isByVariantPtr)
            {
               // get the actial data from Java side
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/util/Object;");
               if (midGetData != NULL)
               {
                  jobject joLongVal = (*env)->CallObjectMethod(env, valObj, midGetData);
                  jclass clsLongVal = (*env)->GetObjectClass(env, joLongVal);
                  jmethodID midGetLongValue = (*env)->GetMethodID(env, clsLongVal, "longValue",
                                                                  "()J");
                  jlong jlongVal = (*env)->CallLongMethod(env, joLongVal, midGetLongValue);
                  // now we can set up the output VARIANT
                  vOutput->vt = VT_BYREF | VT_VARIANT;
                  vOutput->pvarVal = (PVOID)jlongVal;
               }
            }
            // simple INTEGER data type
            else if (*iDataType == DT_INTEGER)
            {
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // TODO: add checking for not-null
                  jobject joIntVal = (*env)->CallObjectMethod(env, valObj, midGetData);
                  // check if we need to process extent arrays instead of simple data type
                  if (isExtent && extentSize > 0)
                  {
                     // construct SAFEARRAY fro Java array of integer objects 
                     SAFEARRAY * saJArray = GetSafeArrayFromIntJArray(env, joIntVal, extentSize);
                     if (saJArray != NULL)
                     {
                        if (isByPointer)
                        {
                           vOutput->vt = VT_BYREF | VT_ARRAY | VT_I4;
                           vOutput->pparray = &saJArray;
                        }
                        else
                        {
                           vOutput->vt = VT_ARRAY | VT_I4;
                           vOutput->parray = saJArray;
                        }
                     }
                  }
                  else
                  {
                     jclass clsInt = (*env)->GetObjectClass(env, joIntVal);
                     jmethodID midGetIntegerVal = (*env)->GetMethodID(env, clsInt, "intValue",
                                                                      "()I");
                     jint jintVal = (*env)->CallIntMethod(env, joIntVal, midGetIntegerVal);
                     if (isByPointer)
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_I4 | VT_BYREF;
                        vOutput->plVal = (LONG*)LocalVarAlloc(&jintVal, sizeof(jintVal));
                     }
                     else
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_I4;
                        vOutput->lVal = jintVal;
                     }
                  }
               }
            }
            // simple DECIMAL data expected
            else if (*iDataType == DT_DECIMAL)
            {
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  jobject joDecVal = (*env)->CallObjectMethod(env, valObj, midGetData);

                  // check if we need to process extent arrays instead of simple data type
                  if (isExtent && extentSize > 0)
                  {
                     // construct SAFEARRAY fro Java array of integer objects
                     SAFEARRAY * saJArray = NULL;
                     if (asDataType == AS_DT_CURRENCY)
                     {
                        saJArray = GetSafeArrayFromCurrencyJArray(env, joDecVal, extentSize);
                        if (saJArray != NULL)
                        {
                           if (isByPointer)
                           {
                              vOutput->vt = VT_BYREF | VT_ARRAY | VT_CY;
                              vOutput->pparray = &saJArray;
                           }
                           else
                           {
                              vOutput->vt = VT_ARRAY | VT_CY;
                              vOutput->parray = saJArray;
                           }
                        }
                     }
                     else
                     {
                        saJArray = GetSafeArrayFromDecimalJArray(env, joDecVal, extentSize);
                        if (saJArray != NULL)
                        {
                           if (isByPointer)
                           {
                              vOutput->vt = VT_BYREF | VT_ARRAY | VT_R8;
                              vOutput->pparray = &saJArray;
                           }
                           else
                           {
                              vOutput->vt = VT_ARRAY | VT_R8;
                              vOutput->parray = saJArray;
                           }
                        }
                     }
                  }
                  else
                  {
                     jclass clsDec = (*env)->GetObjectClass(env, joDecVal);
                     jmethodID midGetDecimalVal = (*env)->GetMethodID(env, clsDec, "doubleValue",
                                                                      "()D");
                     jdouble jdblVal = (*env)->CallDoubleMethod(env, joDecVal, midGetDecimalVal);
                     if (isByPointer)
                     {
                        // now we can set up the output VARIANT
                        if (asDataType == AS_DT_CURRENCY)
                        {
                           CY cyValue;
                           // transforming DECIMAL to CYRRENCY 
                           VarCyFromR8(jdblVal, &(cyValue));
                           vOutput->vt = VT_CY | VT_BYREF;
                           vOutput->pcyVal = (CY*)LocalVarAlloc(&cyValue, sizeof(cyValue));
                        }
                        else
                        {
                           vOutput->vt = VT_R8 | VT_BYREF;
                           vOutput->pdblVal = (DOUBLE*)LocalVarAlloc(&jdblVal, sizeof(jdblVal));
                        }
                     }
                     else
                     {
                        // now we can set up the output VARIANT
                        if (asDataType == AS_DT_CURRENCY)
                        {
                           vOutput->vt = VT_CY;
                           // transforming DECIMAL to CYRRENCY 
                           VarCyFromR8(jdblVal, &(vOutput->cyVal));
                        }
                        else
                        {
                           vOutput->vt = VT_R8;
                           vOutput->dblVal = jdblVal;
                        }
                     }
                  }
               }
            }
            // simple LOGICAL data expected
            else if (*iDataType == DT_LOGICAL)
            {
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // TODO: add checking for not-null
                  jobject joLogVal = (*env)->CallObjectMethod(env, valObj, midGetData);
                  // check if we need to process extent arrays instead of simple data type
                  if (isExtent && extentSize > 0)
                  {
                     // construct SAFEARRAY fro Java array of logical objects 
                     SAFEARRAY * saJArray = GetSafeArrayFromLogicalJArray(env, joLogVal,
                                                                          extentSize);
                     if (saJArray != NULL)
                     {
                        if (isByPointer)
                        {
                           vOutput->vt = VT_BYREF | VT_ARRAY | VT_BOOL;
                           vOutput->pparray = &saJArray;
                        }
                        else
                        {
                           vOutput->vt = VT_ARRAY | VT_BOOL;
                           vOutput->parray = saJArray;
                        }
                     }
                  }
                  else
                  {
                     jclass clsLog = (*env)->GetObjectClass(env, joLogVal);
                     jmethodID midGetLogicalVal = (*env)->GetMethodID(env, clsLog, "booleanValue",
                                                                      "()Z");
                     jboolean jboolVal = (*env)->CallBooleanMethod(env, joLogVal,
                                                                   midGetLogicalVal);
                     if (isByPointer)
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_BOOL | VT_BYREF;
                        vOutput->pboolVal = (VARIANT_BOOL*)LocalVarAlloc(&jboolVal,
                                                                         sizeof(jboolVal));
                     }
                     else
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_BOOL;
                        vOutput->boolVal = jboolVal;
                     }
                  }
               }
            }
            // simple CHARACTER or LONGCHAR data expected
            else if (*iDataType == DT_CHARACTER || *iDataType == DT_LONGCHAR)
            {
               // get the actial data from Java side
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // we use Unicode strings here
                  jstring jstrVal = (*env)->CallObjectMethod(env, valObj, midGetData);
                  // need to protect from null string to avoid JVM crash
                  if (jstrVal != NULL)
                  {
                     // check if we need to process extent arrays instead of simple data type
                     if (isExtent && extentSize > 0)
                     {
                        // construct SAFEARRAY fro Java array of logical objects 
                        SAFEARRAY * saJArray = GetSafeArrayFromCharJArray(env, jstrVal,
                                                                          extentSize);
                        if (saJArray != NULL)
                        {
                           if (isByPointer)
                           {
                              vOutput->vt = VT_BYREF | VT_ARRAY | VT_BSTR;
                              vOutput->pparray = &saJArray;
                           }
                           else
                           {
                              vOutput->vt = VT_ARRAY | VT_BSTR;
                              vOutput->parray = saJArray;
                           }
                        }
                     }
                     else
                     {
                        // reconstruct the BSTR variable from jobject
                        jchar * jchVal = (jchar*) (*env)->GetStringChars(env, jstrVal, NULL);
                        // create OLE compatible string
                        // we will free it only after finish using it(after Invoke for given VARIANT)
                        BSTR bstr = LocalAllocBSTR(jchVal);
                        if (bstr != NULL)
                        {
                           // now we can set up the output VARIANT
                           if (isByPointer)
                           {
                              vOutput->vt = VT_BSTR | VT_BYREF;
                              vOutput->pbstrVal = &bstr;
                           }
                           else
                           {
                              vOutput->vt = VT_BSTR;
                              vOutput->bstrVal = bstr;
                           }
                        }
                        // release string taken from Java
                        if (jchVal != NULL)
                        {
                           (*env)->ReleaseStringChars(env, jstrVal, jchVal);
                        }
                     }
                  }
               }
            }
            // DATE, DATETIME, DATETIME-TZ data expected
            else if (*iDataType == DT_DATETTZ)
            {
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // getting the respective 'date*' class
                  jobject jobDate = (*env)->CallObjectMethod(env, valObj, midGetData);
                  if (jobDate != NULL)
                  {
                     if (isExtent && extentSize > 0)
                     {
                        // construct SAFEARRAY fro Java array of integer objects 
                        SAFEARRAY * saJArray = GetSafeArrayFromDateJArray(env, jobDate,
                                                                          extentSize);
                        if (saJArray != NULL)
                        {
                           if (isByPointer)
                           {
                              vOutput->vt = VT_BYREF | VT_ARRAY | VT_DATE;
                              vOutput->pparray = &saJArray;
                           }
                           else
                           {
                              vOutput->vt = VT_ARRAY | VT_DATE;
                              vOutput->parray = saJArray;
                           }
                        }
                     }
                     else
                     {
                        jclass clsDate = (*env)->GetObjectClass(env, jobDate);
                        if (clsDate != NULL)
                        {
                           jmethodID midGetDateLongVal = (*env)->GetMethodID(env, clsDate,
                                                                             "getTime", "()J");
                           if (midGetDateLongVal != NULL)
                           {
                              // getting Date value as long
                              jlong jlDate = (*env)->CallLongMethod(env, jobDate,
                                                                    midGetDateLongVal);
                              DATE dateValue = OleDateFromJavaLong(jlDate);
                              // convert LONG64 value to DATE
                              if (isByPointer)
                              {
                                 vOutput->vt = VT_DATE | VT_BYREF;
                                 vOutput->pdate = (DATE*)LocalVarAlloc(&dateValue,
                                                                       sizeof(dateValue));
                              }
                              else
                              {
                                 vOutput->vt = VT_DATE;
                                 vOutput->date = dateValue;
                              }
                           }
                        }
                     }
                  }
               }
            }   
            // INT64 data expected
            else if (*iDataType == DT_INT64)
            {
               // get the actial data from Java side
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // TODO: add checking for not-null
                  jobject joInt64Val = (*env)->CallObjectMethod(env, valObj, midGetData);
                  if (isExtent && extentSize > 0)
                  {
                     // construct SAFEARRAY fro Java array of integer objects 
                     SAFEARRAY * saJArray = GetSafeArrayFromInt64JArray(env, joInt64Val,
                                                                        extentSize, VT_I8);
                     if (saJArray != NULL)
                     {
                        if (isByPointer)
                        {
                           vOutput->vt = VT_BYREF | VT_ARRAY | VT_I8;
                           vOutput->pparray = &saJArray;
                        }
                        else
                        {
                           vOutput->vt = VT_ARRAY | VT_I8;
                           vOutput->parray = saJArray;
                        }
                     }
                  }
                  else
                  {
                     jclass clsInt64 = (*env)->GetObjectClass(env, joInt64Val);
                     jmethodID midGetInt64Value = (*env)->GetMethodID(env, clsInt64, "longValue",
                                                                      "()J");
                     jlong jlongVal = (*env)->CallLongMethod(env, joInt64Val, midGetInt64Value);
                     if (isByPointer)
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_I8 | VT_BYREF;
                        vOutput->pllVal = (LONGLONG*)LocalVarAlloc(&jlongVal, sizeof(jlongVal));
                     }
                     else
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_I8;
                        vOutput->llVal = jlongVal;
                     }
                  }
               }
            }
            // COM-HANDLE data expected
            else if (*iDataType == DT_COMHANDLE)
            {
               // get the actial data from Java side
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // TODO: add checking for not-null
                  jobject joHandleVal = (*env)->CallObjectMethod(env, valObj, midGetData);
                  if (isExtent && extentSize > 0)
                  {
                     // construct SAFEARRAY fro Java array of integer objects 
                     SAFEARRAY * saJArray = GetSafeArrayFromInt64JArray(env, joHandleVal,
                                                                        extentSize, VT_DISPATCH);
                     if (saJArray != NULL)
                     {
                        if (isByPointer)
                        {
                           vOutput->vt = VT_BYREF | VT_ARRAY | VT_DISPATCH;
                           vOutput->pparray = &saJArray;
                        }
                        else
                        {
                           vOutput->vt = VT_ARRAY | VT_DISPATCH;
                           vOutput->parray = saJArray;
                        }
                     }
                  }
                  else
                  {
                     jclass clsHandle = (*env)->GetObjectClass(env, joHandleVal);
                     jmethodID midGetHandleValue = (*env)->GetMethodID(env, clsHandle, "longValue",
                                                                      "()J");
                     jlong jlongVal = (*env)->CallLongMethod(env, joHandleVal, midGetHandleValue);
                     if (isByPointer)
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_DISPATCH | VT_BYREF;
                        vOutput->ppdispVal = (IDispatch**)LocalVarAlloc(&jlongVal,
                                                                        sizeof(jlongVal));
                     }
                     else
                     {
                        // now we can set up the output VARIANT
                        vOutput->vt = VT_DISPATCH;
                        vOutput->pdispVal = (IDispatch*)jlongVal;
                     }
                  }
               }
            }
            // RAW data expected
            else if (*iDataType == DT_RAW)
            {
               // get the actial data from Java side
               midGetData = (*env)->GetMethodID(env, cls, "getValue", "()Ljava/lang/Object;");
               if (midGetData != NULL)
               {
                  // getting byte array object
                  jobject joByteArray = (*env)->CallObjectMethod(env, valObj, midGetData);
                  if (joByteArray != NULL)
                  {
                     // getting the byte array as SAFEARRAY vector
                     SAFEARRAY * psaRawData = GetSafeArrayFromRaw(env, joByteArray);
                     // now we can set up the output VARIANT
                     if (psaRawData != NULL)
                     {
                        if (isByPointer)
                        {
                           vOutput->vt = VT_BYREF | VT_ARRAY | VT_UI1;
                           vOutput->pparray = &psaRawData;
                        }
                        else
                        {
                           vOutput->vt = VT_ARRAY | VT_UI1;
                           vOutput->parray = psaRawData;
                        }
                     }
                  }
               }
            }
         }

         // release FWD data type string
         (*env)->ReleaseStringUTFChars(env, jstFwdDataType, chFwdDataType);
      }
   }
   
   return rc;
}

/**
 * Checks if the given method ID belongs to the property or method or indexed property.
 *
 * @param    objPtr
 *           The pointer to object interface.
 *           
 * @param    mID
 *           The ID of the method to check for property.
 *           
 * @param    isPropertyOrig
 *           The boolean flag indicating call from method execution or property set/get.
 *           - JNI_TRUE means calling from method with property get hint
 *           - JNI_FALSE means calling from method as pure method.
 *           
 * @return   The WORD type flag to make Invoke call for method, can ave three possible values:
 *           - DISPATCH_METHOD for pure method.
 *           - DISPATCH_PROPERTYGET for pure property getter
 *           - DISPATCH_METHOD | DISPATCH_PROPERTYGET for call that can be used for both method
 *                                                    and property set.
 */
WORD getMethodType(IDispatch * objPtr, DISPID mID, jboolean isPropertyOrig)
{
   // assume pure method by default
   WORD wRes = DISPATCH_METHOD;
   
   ITypeInfo * pti = NULL;
   HRESULT hr = objPtr->lpVtbl->GetTypeInfo(objPtr, 0, LOCALE_NEUTRAL, &pti);
   if (hr == S_OK && pti != NULL)
   {
      FUNCDESC * pMethDesc = NULL;
      hr = pti->lpVtbl->GetFuncDesc(pti, mID, &pMethDesc);
      if (hr == S_OK && pMethDesc != NULL)
      {
         // consider the function is an exact method only if invkind is INVOKE_FUNC
         // all other cases can be used as property get or method call simultaneously
         if (pMethDesc->invkind == INVOKE_PROPERTYGET)
         {
            wRes |= DISPATCH_PROPERTYGET;
         }
         else if (pMethDesc->invkind == INVOKE_PROPERTYPUT)
         {
            wRes |= DISPATCH_PROPERTYPUT;
         }
         else if (pMethDesc->invkind != INVOKE_FUNC)
         {
            // this is the case when the call is not an exact method but we hav no
            // modifiers change at this time
            fwprintf(comerr, L"getMethodType(): Unsupported pMethDesc->invkind = %d\n",
                     pMethDesc->invkind);
            fflush(comerr);
         }
         pti->lpVtbl->ReleaseFuncDesc(pti, pMethDesc);
      }
      else
      {
         // when no method description found we consider property get or method case depending
         // on the hint flag provided
         if (isPropertyOrig)
         {
            wRes = DISPATCH_PROPERTYGET;
         }
      }
   }
   else
   {
      // when no method found we consider pure property get case
      wRes = DISPATCH_PROPERTYGET;
   }
   
   return wRes;
}

/**
 * Releases COM subsystem and destroy the COM object internal registry.
 */
void freeComBridge()
{
   // free allocated memory
   int i;
   // loop through registry to search the object
   for (i = hRegSize - 1; i >= 0; i--)
   {
      // free memory allocated to given object info
      if (phInfo[i] != NULL)
      {
         // free memory allocated for parent list
         if (phInfo[i]->parent_ids != NULL)
         {
            HeapFree(hProcHeap, heapUseMode, phInfo[i]->parent_ids);
            phInfo[i]->parent_ids = NULL;
         }
         GlobalFree(phInfo[i]);
         phInfo[i] = NULL;
      }
   }
   // and free up COM itself
   CoUninitialize();
   // close log file
   fclose(comerr);
}

// JNI implementations
/**
 * This function will initialize COM subsystem in Windows and returns TRUE in cases of success.
 *
 * @return   TRUE in success init FALSE otherwise.
 */
jboolean initComBridge()
{
   // clean old log file and open new one for appent mode once per session 
   comerr = fopen(COM_OLE_ERRORS_LOG_FILE, "w");
   fclose(comerr);
   comerr = fopen(COM_OLE_ERRORS_LOG_FILE, "a");
   // init subsystem itself
   return ((hProcHeap = GetProcessHeap()) != NULL && CoInitialize(NULL) == S_OK);
}

/**
 * Gets the automation COM object. It can be a new instance or a reference to an already existing
 * top-level automation. 
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    name
 *           The name of the automation.
 *
 * @param    topLevel
 *           Only return top-level. If the requested COM name is not a top-level, return null.
 *
 * @param    noCreate
 *           Must exist. Do not allow creation of new instances. Only used with topLevel
 *           requests.
 *          
 * @return  The ID of the COM object, or 0 if such automation does not exist
 *          or it does not satisfy the parameter constraints. 
 */
jlong createCom(JNIEnv *env, jstring name, jboolean topLevel, jboolean noCreate)
{
   jlong idRes = 0;
   IDispatch *iDispPtr;
   // getting name of the com object to request
   jchar* comName = (jchar*) (*env)->GetStringChars(env, name, NULL);
   if (comName != NULL)
   {
      CLSID classID;
      HRESULT hr = CLSIDFromProgID((LPOLESTR)comName, &classID);
      if (hr == S_OK)
      {
         // trying to connect to existed COM object
         if (topLevel)
         {
            hr = GetActiveObject(&classID, NULL, (IUnknown **)&iDispPtr);
            // create object if not active if not to suppress creation
            if (hr != S_OK && !noCreate)
            {
               hr = CoCreateInstance(&classID, NULL, CLSCTX_SERVER, &IID_IDispatch,
                                     (void **)&iDispPtr);
            }
         }
         else
         {
            hr = CoCreateInstance(&classID, NULL, CLSCTX_SERVER,
                                  &IID_IDispatch, (void **)&iDispPtr);
         }
         
         // here we must have object created
         if (hr == S_OK)
         {
            idRes = (jlong)iDispPtr;
            // refresh our registry
            if (!isCOMInRegistry(idRes))
            {
               // add new COM id to registry only if it is not already in
               // this is root object, no parent
               addCOMToRegistry(idRes, comName, 0);
            }
         }
      }
   }

   (*env)->ReleaseStringChars(env, name, comName);
   
   return idRes;
}

/**
 * Release and destroy the COM object with the specified ID.
 * 
 * @param    comPtrID
 *           The COM object ID.
 */
void releaseCom(jlong comPtrID)
{
   // release object
   releaseCOMObject(comPtrID);
   // refresh our registry
   removeCOMFromRegistry(comPtrID);
   // release all child objects
   releaseCOMChildren(comPtrID);
}

/**
 * Get the specified COM object property.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    comPtrID
 *           The COM object ID.
 *
 * @param    name
 *           The COM property name to retrieve.
 *           
 * @return   A BaseDataType instance with the COM property's details.
 */
jobject getComProperty(JNIEnv *env, jlong comPtrID, jstring name)
{
   jobjectArray propValue = NULL; 
   // we need to get ID of the property getter
   // convert long ID to object pointer
   IDispatch * objPtr = (IDispatch *)comPtrID;
   DISPID prID;
   // attempt to identify the ID of property getter
   if (getMethOrPropId(env, objPtr, name, &prID))
   {
      // have got ID of the getter, calling it
      DISPPARAMS   args;
      VARIANT varRes;
      // prepare arguments to pass via object method call
      // zero input parameters
      ZeroMemory(&args, sizeof(DISPPARAMS));
      // and storage for one variavle to return
      VariantInit(&varRes);
      // call the property getter
      HRESULT hr = objPtr->lpVtbl->Invoke(objPtr, prID, &IID_NULL, LOCALE_NEUTRAL,
                                          DISPATCH_PROPERTYGET, &args, &varRes, 0, 0);
      // got property extract data to return
      if (hr == S_OK)
      {
//         propValue = getJObjectArrayFromVariant(env, comPtrID, &varRes);
         propValue = getJObjectFromVariant(env, comPtrID, &varRes);
      }
   }

   return propValue;
}

/**
 * Set the specified COM object property.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    comPtrID
 *           The COM object ID.
 *
 * @param    name
 *           The COM property name to set.
 *
 * @param    val
 *           The details about property's new value.
 *           
 * @param    ndxs
 *           Indices array values to find out indexed property.
 *           
 * @return   JNI_TRUE if the property was set successfully, JNI_FALSE otherwise.
 */
jboolean setComProperty(JNIEnv *env, jlong comPtrID, jstring name, jobject val, jobjectArray ndxs)
{
   jboolean rc = JNI_FALSE;
   // convert long ID to object pointer
   IDispatch * objPtr = (IDispatch *)comPtrID;
   DISPID prID;
   // attempt to identify the ID of property getter
   if (getMethOrPropId(env, objPtr, name, &prID))
   {
      // get the input variables size(new value always exists, indices are optional)
      int numInputParm = 1;
      // check indexes array
      if (ndxs != NULL)
      {
         numInputParm += (*env)->GetArrayLength(env, ndxs);
      }
      // have got ID of the setter, calling it
      DISPPARAMS   args;
      DISPID didPropPut = DISPID_PROPERTYPUT;
      VARIANT var[numInputParm];
      VARIANT varRes;
      // prepare arguments to pass via object method call
      ZeroMemory(&args, sizeof(DISPPARAMS));
      int i;
      for (i = 0; i < numInputParm; i++)
      {
         VariantInit(&var[i]);
      }
      VariantInit(&varRes);
      // transferring data from Java side into VARIANT for value to be set
      // the value to set the first
      setVariantFromJObject(env, &var[0], val);
      // start processing the indexes in reverse order
      for (i = 0; i < numInputParm - 1; i++)
      {
         // first index will occupy last - 1 slot reserving last slot for real value
         setVariantFromJObject(env, &var[numInputParm - i - 1],
                               (*env)->GetObjectArrayElement(env, ndxs, i));
      }
      // perpare property setter parameters
      args.rgvarg = var;
      // total number of parametes with indices
      args.cArgs = numInputParm;
      args.rgdispidNamedArgs = &didPropPut;
      args.cNamedArgs = 1;
      // call the property getter
      HRESULT hr = objPtr->lpVtbl->Invoke(objPtr, prID, &IID_NULL, LOCALE_NEUTRAL,
                                          DISPATCH_PROPERTYPUT, &args, &varRes, 0, 0);
      
      // return JNI_TRUE if property set was OK
      if (hr == S_OK)
      {
         rc = JNI_TRUE;
      }

      // clean up BSTR string cache
      FreeLocalBSTR();
      // and SAFEARRAY objects too
      FreeSafeArrays();
      // and by-pointer var storage
      FreeLocalVar();
   }
   
   return rc;
}

/**
 * Execute a COM object method call.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    comPtrID
 *           The COM object ID.
 *
 * @param    name
 *           The COM method name to call.
 *
 * @param    valArray
 *           An array with this call's arguments supplied to the COM object method call.
 *           
 * @return   An 2d array where on index 0 are details of the returned value(single or array), and
 *           further indices correspond with argument's 1-based indices, where there will be a
 *           value(single or array) for each OUTPUT or INPUT-OUTPUT argument.
 */
jobjectArray callComMethod(JNIEnv *env, jlong comPtrID, jstring name, jobjectArray valArray)
{
   jobjectArray joMethRes = NULL; 
   // we need to get ID of the property getter
   IDispatch * objPtr = (IDispatch *)comPtrID;
   DISPID methID;
   // attempt to identify the ID of property getter
   if (getMethOrPropId(env, objPtr, name, &methID))
   {
      // get the input variables size(return value always exists)
      int numInputParm = 0;
      // at least one output to return we always have
      int numOutputRes = 1;
      // add number of items from incoming Java array
      if (valArray != NULL)
      {
         numInputParm += (*env)->GetArrayLength(env, valArray);
      }
      // prepare VARIANT based array of input data
      VARIANT varIn[numInputParm];
      // flags indicating which parameters must be returned back to Java
      jboolean bToReturn[numInputParm];
      int i;
      for (i = 0; i < numInputParm; i++)
      {
         VariantInit(&varIn[i]);
      }
      // have got ID of the setter, calling it
      DISPPARAMS   args;
      // prepare arguments to pass via object method call
      ZeroMemory(&args, sizeof(DISPPARAMS));
      // reserved variant for resilt
      VARIANT vResult;
      // init return value storage
      VariantInit(&vResult);
      vResult.vt = VT_EMPTY;
      // copy methods parameters from Java object array
      for (i = 0; i < numInputParm; i++)
      {
         // the order is reversed, last object should be first entry
         bToReturn[i] = setVariantFromJObject(env, &varIn[numInputParm - i - 1],
                                              (*env)->GetObjectArrayElement(env, valArray, i));
            
         // the parameter must be returned as having OUTPUT or INPUT-OUTPUT mode 
         if (bToReturn[i])
         {
            numOutputRes++;
         }
      }
      // perpare method parameters
      if (numInputParm > 0)
      {
         args.rgvarg = varIn;
      }
      else
      {
         args.rgvarg = NULL;
      }
      args.cArgs = numInputParm;
      // check if we have indexed property getter instead of method call
      WORD invFlags = getMethodType(objPtr, methID, numInputParm == 0);
      // call the object method
      HRESULT hr = objPtr->lpVtbl->Invoke(objPtr, methID, &IID_NULL, LOCALE_NEUTRAL,
                                          invFlags, &args, &vResult, 0, 0);
      // check if we need to re-run for a indexed property instead
      if (hr != S_OK)
      {
         fwprintf(comerr, L"Method execution failed, hr = %x, objID:methID = %x:%x\n",
                  hr, objPtr, methID);
         fwprintf(comerr, L"Another try with a method as property...\n");
         fflush(comerr);
         // another try to consider the method as indexed property
         invFlags = getMethodType(objPtr, methID, JNI_TRUE);
         hr = objPtr->lpVtbl->Invoke(objPtr, methID, &IID_NULL, LOCALE_NEUTRAL,
                                          invFlags, &args, &vResult, 0, 0);
      }
      // got property extract data to return
      if (hr == S_OK)
      {
         // the return date will be based on BaseDataType class
         jclass clsArrRet = (*env)->FindClass(env, "[Lcom/goldencode/p2j/util/BaseDataType;");
         // now prepare data to return, array of the arrays
         joMethRes = (*env)->NewObjectArray(env, numOutputRes, clsArrRet, NULL);
         // fill the array with return objects
         if (joMethRes != NULL && clsArrRet != NULL)
         {
            // store the method result as 0 indexed object, inner array
            (*env)->SetObjectArrayElement(env, joMethRes, 0,
                                          getJObjectArrayFromVariant(env, comPtrID, &vResult));
            // the index of the returned oject in returned array
            int j = 0;
            for (i = 0; i < numInputParm; i++)
            {
               if (bToReturn[i])
               {
                  // yet another object inner array to return
                  j++;
                  // reconstruct the object(single or array) from VARIANT (in reverse order)
                  (*env)->SetObjectArrayElement(env, joMethRes, j,
                         getJObjectArrayFromVariant(env, comPtrID, &varIn[numInputParm - i - 1]));
               }
            }
         }
      }
      else
      {
         fwprintf(comerr, L"COM OLE method execution failed, hr = %x, objID:methID = %x:%x\n",
                  hr, objPtr, methID);
         fflush(comerr);
      }

      // clean up BSTR string cache
      FreeLocalBSTR();
      // and SAFEARRAY objects too
      FreeSafeArrays();
      // and by-pointer var storage
      FreeLocalVar();
   }

   return joMethRes;
}

/**
 * Get the COM object name associated with the given COM object ID.
 * 
 * @param    env
 *           Pointer to a structure that provides access to JNI environment.
 *
 * @param    comPtrID
 *           The COM object ID.
 *           
 * @return   See above.
 */
jstring getComName(JNIEnv *env, jlong comPtrID)
{
   // the object name can be no more than 39 chars long
   jchar * nameCom;

   // get the char string name of the given COM object ID
   nameCom = getCOMName(comPtrID);

   // reconstruct java string to return
   return nameCom != NULL ? (*env)->NewString(env, nameCom, wcslen(nameCom)) : NULL;
}

/**
 * Reports if this object is valid for use.  
 *
 * @param    comPtrID
 *           The COM object ID.
 *           
 * @return   JNI_TRUE if we are valid (can be used),
 *           JNI_FALSE otherwise(vas released or not created).
 */
jboolean isComValid(jlong comPtrID)
{
   // for now just chech if COM was created and not yet released
   return isCOMInRegistry(comPtrID);
}
