/*
** Module   : process_linux.c
** Abstract : implementation for process creation/waiting logic in libp2j.so native library which
**            is called via JNI
**            This is used by process.c
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Linux specifics for process launching
**                  operations.
** 002 EVL 20140117 Moving the init steps to the special location.
** 003 ECF 20150126 Added flag for background mode to pseudoTerminalLaunch. Do not check whether
**                  STDIN_FILENO is a TTY when in background mode.
** 004 EVL 20160425 Fix for read line blocking.  Moved the child process command line preparation
**                  code before the fork() call.  Removed static jfd chached field because it can
**                  be changed and not valid to be cached.
** 005 EVL 20160511 Adding more checks for child process to be terminated.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include "process_launch.h" 
#include <sys/wait.h>
#ifdef SOLARIS
#include <curses.h>
#endif
#include <term.h>

/**
 * initializeFileDescriptor - initialize the passed file descriptor object 
 *                            with the pty fd
 *
 * @param   env         pointer to an object providing access to JNI environment
 *
 * @param   descriptor  object representing the FileDescriptor Java instance 
 *
 * @param   fd          the pty file descriptor
 *
 */ 
void initializeFileDescriptor(JNIEnv* env, jobject descriptor, int fd)
{
   jclass fdcls = (*env)->GetObjectClass(env, descriptor);

   jfieldID jfd = (*env)->GetFieldID(env, fdcls, "fd", "I");

   (*env)->SetIntField(env, descriptor, jfd, fd);
}

/**
 * pseudoTerminalLaunch - spawns a child process redirecting its standard input, output and
 *                        error to particular file descriptors (FileDescriptor Java instances). 
 *                        Linux version.
 *
 * @param   env         pointer to a structure that provides access to JNI
 *                      environment
 *
 * @param   cls         reference to the Java class who called this function
 *                      (from static context)
 *
 * @param   cmdline     the command line used for launching the child process  
 *
 * @param   in          FileDescriptor Java object for stdin
 *
 * @param   out         FileDescriptor Java object for stdout
 *
 * @param   err         FileDescriptor Java object for stderr
 *
 * @param   silent      boolean flag controlling interactive mode
 *
 * @param   background  boolean flag indicating client is running in the
 *                      background (disconnected from its controlling terminal)
 *
 * @return  > 1  PID of spawned child process if successful
 *          < 0  on error
 */ 

JNIEXPORT jint JNICALL pseudoTerminalLaunch(JNIEnv* env, jclass cls, jobjectArray cmdline,
                                            jobject in, jobject out, jobject err, jboolean silent,
                                            jboolean background)
   
{
   jint pid = 0;

   // make sure we have valid argument references
   if (cmdline == NULL)
   {
      return -1;
   }
   
   // if not in background mode, make sure we have a tty as stdin
   if (!background && !isatty(STDIN_FILENO))
   {
      return -2;
   }
   
   // access redirection flags and create pipes
   int pipe_i[2];                // pipe for stdin
   if (in != NULL)
   {
      if (pipe(pipe_i) < 0)
      {
         return -3;
      }
   }
   
   int pipe_o[2];                // pipe for stdout
   if (out != NULL)
   {
      if (pipe(pipe_o) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         return -4;
      }
   }
   
   int pipe_e[2];                // pipe for stderr
   if (err != NULL)
   {
      if (pipe(pipe_e) < 0)
      {
         if (in != NULL)
         {
            close(pipe_i[0]);
            close(pipe_i[1]);
         }
         if (out != NULL)
         {
            close(pipe_o[0]);
            close(pipe_o[1]);
         }
         return -5;
      }
   }

   // due to a bug in the JVM 1.4.2 and earlier, we must override any
   // signal handler for SIGCHLD that may exist because otherwise the JVM
   // may reap our process instead of letting us do it (see Sun J2SE bugs 
   // #4945203, #4785154 and #4763362)
   signal(SIGCHLD, SIG_IGN);

   // must build the parameter list
   int len = (int) (*env)->GetArrayLength(env, cmdline);
   int i;

   char** cmds = malloc(sizeof(char*) * (len + 1)); 
   char** current = cmds;

   // copy the strings into our own array
   for (i = 0; i < len; i++, current++)
   {
      jobject next = (*env)->GetObjectArrayElement(env, cmdline, i); 
      char*   tmp  = (char*) (*env)->GetStringUTFChars(env, (jstring) next, NULL);

      if (tmp == NULL)
      {
         // out of memory!
         return -6;
      }

      *current = strdup(tmp);

      (*env)->ReleaseStringUTFChars(env, (jstring) next, tmp);
   }

   // terminate the argv[] with a null pointer (current will be pointing
   // at cmds[len] after the loop ends)
   *current = NULL;

   // create a child process
   pid = fork();
   
   // WARNING: in the child process there can be NO JNI USAGE AFTER THIS POINT!
   
   if (pid < 0)
   {
      // error, nothing more to do
      return -7;
   }
   else if (pid == 0)
   {
      // child process
      
      if (!silent)
      {
         // we have to leave the screen clear for the program being launched
         system("tput clear");

         // in interactive mode, we redirect stderr to stdout
         dup2(STDOUT_FILENO, STDERR_FILENO);
      }

      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need write end for child's stdin in child
         close(pipe_i[1]);
         
         // redirect this pipe to stdin
         if (pipe_i[0] != STDIN_FILENO)
         {
            // duplicate read end of stdin
            dup2(pipe_i[0], STDIN_FILENO);
            
            close(pipe_i[0]);    // not required after duplication
         }
      }
      
      if (out != NULL)
      {
         // don't need read end for child's stdout in child
         close(pipe_o[0]);
         
         // redirect this pipe to stdout
         if (pipe_o[1] != STDOUT_FILENO)
         {
            // duplicate write end of stdout
            dup2(pipe_o[1], STDOUT_FILENO);
            
            close(pipe_o[1]);    // not required after duplication
         }
      }
      
      if (err != NULL)
      {
         // don't need read end for child's stderr in child
         close(pipe_e[0]);
         
         // redirect this pipe to stderr
         if (pipe_e[1] != STDERR_FILENO)
         {
            // duplicate write end of stderr
            dup2(pipe_e[1], STDERR_FILENO);
            
            close(pipe_e[1]);    // not required after duplication
         }
      }

      // exec (never returns unless there is an error)
      int rc = execvp(*cmds, cmds);
      
      // if we are here, there was a problem!
      fprintf(stderr, "ERROR: execvp rc = %i; errno = %i\n", rc, errno);
      exit(-1);
   }
   else
   {
      // parent process
      // initialize the file descriptors for the redirected handles
      if (in != NULL)
      {
         // don't need read end for child's stdin in parent
         close(pipe_i[0]);
         
         // pass the write end back
         initializeFileDescriptor(env, in, pipe_i[1]); 
      }
      
      if (out != NULL)
      {
         // don't need write end for child's stdout in parent
         close(pipe_o[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, out, pipe_o[0]); 
      }
      
      if (err != NULL)
      {
         // don't need write end for child's stderr in parent
         close(pipe_e[1]);
         
         // pass the read end back
         initializeFileDescriptor(env, err, pipe_e[0]); 
      }
   }
   
   return pid;
}

/**
 * pseudoTerminalWait - The Linux version of waits for a child process to terminate. 
 *
 * @param   env     pointer to a structure that provides access to JNI environment
 *
 * @param   cls     reference to the Java class who called this function from
 *                  a static context
 *
 * @param   pid     PID of the process to be waited
 *
 * @param   silent  boolean flag to control interactive mode
 *
 */ 
JNIEXPORT void JNICALL pseudoTerminalWait(JNIEnv* env, jclass cls, int pid, jboolean silent)
{
   pid_t child = (pid_t) pid;
   int   status;
   int rcWaitpid;

   do
   {
      rcWaitpid = waitpid(child, &status, WUNTRACED | WCONTINUED);
   }
   while (!WIFEXITED(status) && !WIFSIGNALED(status) && rcWaitpid != -1);

   if (!silent)
   {
      // in interactive mode we're calling functions in NCURSES directly, 
      // as opposed to executing command line utilities for the same purpose
      // (NCURSES library is a requirement anyway so we want to have a more
      // efficient and more controlled use of that interface as well, while 
      // also avoiding the hard requirement that other utility programs be
      // installed in addition to P2J)

      // put "Press space bar to continue." message on the last line

      // move cursor to the beginning of the last terminal line 
      char *s;
      s = tparm(tigetstr("cup"), tigetnum("lines"), 0, 0, 0, 0, 0, 0, 0, 0);
      putp(s);

      // display prompt
      printf("Press space bar to continue.");

      // move cursor to the end of the last terminal line (as in Progress)
      s = tparm(tigetstr("cup"), tigetnum("lines"), tigetnum("cols"),
                0, 0, 0, 0, 0, 0, 0);
      putp(s);
   }
}
